(PackageInfo named: 'System') postscript: '"Re-create all themes."
SqueakTheme create apply.
SqueakTheme createClassic; createDuller.
SolarizedTheme createDark; createLight.
MonokaiTheme createDark.
CommunityTheme createDark.'!
SystemOrganization addCategory: #'System-Applications'!
SystemOrganization addCategory: #'System-Change Notification'!
SystemOrganization addCategory: #'System-Changes'!
SystemOrganization addCategory: #'System-Digital Signatures'!
SystemOrganization addCategory: #'System-Download'!
SystemOrganization addCategory: #'System-Exceptions'!
SystemOrganization addCategory: #'System-FilePackage'!
SystemOrganization addCategory: #'System-FileRegistry'!
SystemOrganization addCategory: #'System-Finalization'!
SystemOrganization addCategory: #'System-Localization'!
SystemOrganization addCategory: #'System-Object Events'!
SystemOrganization addCategory: #'System-Object Storage'!
SystemOrganization addCategory: #'System-Preferences'!
SystemOrganization addCategory: #'System-Recovery'!
SystemOrganization addCategory: #'System-Serial Port'!
SystemOrganization addCategory: #'System-Support'!
SystemOrganization addCategory: #'System-Tools'!

Categorizer subclass: #SystemOrganizer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SystemOrganizer commentStamp: '<historical>' prior: 0!
My instances provide an organization for the classes in the system, just as a ClassOrganizer organizes the messages within a class. The only difference is the methods for fileIn/Out.!

!SystemOrganizer class methodsFor: 'class initialization' stamp: 'ar 2/26/2010 23:01'!
cleanUp: agressive
	"Remove empty categories when cleaning aggressively"

	agressive ifTrue:[SystemOrganization removeEmptyCategories].
! !

!SystemOrganizer class methodsFor: 'class initialization' stamp: 'spd 5/19/2010 15:54'!
default
	"I have a global instance, SystemOrganization.  In case someone stumbles upon my class, and does not know about the global, it can be accessed from here, too."

  ^ SystemOrganization.
! !

!SystemOrganizer methodsFor: 'accessing' stamp: 'cwp 2/12/2013 00:40'!
addCategory: catString before: nextCategory
	self == SystemOrganization 
		ifTrue: 
			[SystemChangeNotifier uniqueInstance
				doSilently: [super addCategory: catString before: nextCategory];
				classCategoryAdded: catString]
		ifFalse:
			[super addCategory: catString before: nextCategory]! !

!SystemOrganizer methodsFor: 'remove' stamp: 'di 3/3/2001 16:07'!
categoriesMatching: matchString
	"Return all matching categories"
	^ self categories select: [:c | matchString match: c]! !

!SystemOrganizer methodsFor: 'accessing' stamp: 'dtl 1/17/2013 19:49'!
classify: element under: newCategory
	| oldCategory class |
	oldCategory := self categoryOfElement: element.
	super classify: element under: newCategory.
	class := Smalltalk at: element ifAbsent: [^ self].
	self == SystemOrganization ifTrue: [
		SystemChangeNotifier uniqueInstance
			class: class
			recategorizedFrom: oldCategory
			to: newCategory]! !

!SystemOrganizer methodsFor: 'query' stamp: 'dtl 8/26/2004 11:18'!
commentInventory: categoryName

	"SystemOrganization commentInventory: 'Morphic*'"

	| classes commentedClasses |
	classes := OrderedCollection new.
	self categories withIndexCollect: [:cat :idx |
		(categoryName match: cat)
			ifTrue: [classes addAll: (self listAtCategoryNumber: idx)]
			ifFalse: [nil]].
	commentedClasses := classes select: [:catCls | (Smalltalk at: catCls) hasComment].
	^ 'There are ' , classes size asString , ' classes in ' , categoryName ,
		' of which ' , commentedClasses size asString , ' have comments and ',
		(classes size - commentedClasses size) asString , ' do not yet have comments.'
! !

!SystemOrganizer methodsFor: 'fileIn/Out' stamp: 'yo 7/5/2004 20:22'!
fileOut  "SystemOrganization fileOut"

	| internalStream |
	internalStream := WriteStream on: (String new: 30000).
	internalStream nextPutAll: 'SystemOrganization changeFromCategorySpecs: #('; cr;
		print: SystemOrganization;  "ends with a cr"
		nextPutAll: ')!!'; cr.

	FileStream writeSourceCodeFrom: internalStream baseName: (FileDirectory default nextNameFor: 'SystemOrganization' extension: 'st') isSt: true useHtml: false
! !

!SystemOrganizer methodsFor: 'fileIn/Out' stamp: 'di 6/28/97 19:01'!
fileOutCategory: category 
	"Store on the file named category (a string) concatenated with '.st' all the 
	classes associated with the category."
	^ self fileOutCategory: category asHtml: false! !

!SystemOrganizer methodsFor: 'fileIn/Out' stamp: 'yo 7/5/2004 20:22'!
fileOutCategory: category asHtml: useHtml
	"FileOut all the classes in the named system category."
	| internalStream |
	internalStream := WriteStream on: (String new: 1000).
	self fileOutCategory: category on: internalStream initializing: true.

	FileStream writeSourceCodeFrom: internalStream baseName: category isSt: true useHtml: useHtml.
! !

!SystemOrganizer methodsFor: 'fileIn/Out' stamp: 'ar 12/22/1999 17:28'!
fileOutCategory: category on: aFileStream 
	"Store on the file associated with aFileStream, all the classes associated 
	with the category and any requested shared pools."
	^self fileOutCategory: category on: aFileStream initializing: true! !

!SystemOrganizer methodsFor: 'fileIn/Out' stamp: 'ar 3/5/2010 20:42'!
fileOutCategory: category on: aFileStream initializing: aBool
	"Store on the file associated with aFileStream, all the traits and classes associated 
	with the category and any requested shared pools in the right order."

	| first poolSet tempClass classes traits |
	traits := self orderedTraitsIn: category.
	classes := self superclassOrder: category.
	poolSet := Set new.
	classes do:  [:class | class sharedPools do: [:eachPool | poolSet add: eachPool]].
	poolSet size > 0 ifTrue: [
		tempClass := Class new.
		tempClass shouldFileOutPools ifTrue: [
			poolSet := poolSet select: [:aPool |
				tempClass shouldFileOutPool: (Smalltalk globals keyAtIdentityValue: aPool)].
			poolSet do: [:aPool | tempClass fileOutPool: aPool onFileStream: aFileStream]]].
	first := true.
	traits, classes do: [:each | 
		first
			ifTrue: [first := false]
			ifFalse: [aFileStream cr; nextPut: Character newPage; cr].
		each
			fileOutOn: aFileStream
			moveSource: false
			toFile: 0
			initializing: false].
	aBool ifTrue: [classes do: [:cls | cls fileOutInitializerOn: aFileStream]].! !

!SystemOrganizer methodsFor: 'private' stamp: 'rw 7/31/2003 17:23'!
ifClassOrganizerDo: aBlock
	"Do nothing, since this is not a class organizer"! !

!SystemOrganizer methodsFor: 'fileIn/Out' stamp: 'tk 9/28/2000 15:50'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a path to me in the other system instead."

self == SystemOrganization ifTrue: [
	dp := DiskProxy global: #SystemOrganization selector: #yourself args: #().
	refStrm replace: self with: dp.
	^ dp].
^ self
! !

!SystemOrganizer methodsFor: 'fileIn/Out' stamp: 'nice 8/19/2010 16:35'!
orderedTraitsIn: category 
	"Answer an OrderedCollection containing references to the traits in the 
	category whose name is the argument, category (a string). The traits 
	are ordered so they can be filed in."

	| behaviors traits |
	behaviors := (self listAtCategoryNamed: category asSymbol) 
			collect: [:title | Smalltalk at: title].
	traits := behaviors reject: [:each | each isBehavior].
	^traits asArray sort: [:t1 :t2 |
		(t2 traitComposition allTraits includes: t1)
			or: [(t1 traitComposition allTraits includes: t2) not]]! !

!SystemOrganizer methodsFor: 'remove' stamp: 'di 3/3/2001 16:08'!
removeCategoriesMatching: matchString
	"Remove all matching categories with their classes"
	(self categoriesMatching: matchString) do:
		[:c | self removeSystemCategory: c]! !

!SystemOrganizer methodsFor: 'accessing' stamp: 'dtl 1/17/2013 19:50'!
removeCategory: cat 
	| r |
	r := super removeCategory: cat.
	self == SystemOrganization
		ifTrue: [SystemChangeNotifier uniqueInstance classCategoryRemoved: cat].
	^ r! !

!SystemOrganizer methodsFor: 'remove' stamp: 'jm 5/20/1998 19:38'!
removeMissingClasses
	"Remove any class names that are no longer in the Smalltalk dictionary. Used for cleaning up after garbage collecting user-generated classes."
	"SystemOrganization removeMissingClasses"

	elementArray copy do: [:el |
		(Smalltalk includesKey: el) ifFalse: [self removeElement: el]].
! !

!SystemOrganizer methodsFor: 'remove' stamp: 'al 4/25/2004 11:38'!
removeSystemCategory: category
	"remove all the classes and traits associated with the category"

	 (self orderedTraitsIn: category) , (self superclassOrder: category)
		reverseDo: [:each | each removeFromSystem].

	self removeCategory: category.
! !

!SystemOrganizer methodsFor: 'accessing' stamp: 'dtl 1/17/2013 19:51'!
renameCategory: oldCatString toBe: newCatString
	| r |
	r := super renameCategory: oldCatString toBe: newCatString.
	self == SystemOrganization ifTrue: [
		SystemChangeNotifier uniqueInstance 
			classCategoryRenamedFrom: oldCatString to: newCatString].
	^ r! !

!SystemOrganizer methodsFor: 'fileIn/Out' stamp: 'dvf 9/27/2005 19:06'!
superclassOrder: category 
	"Answer an OrderedCollection containing references to the classes in the 
	category whose name is the argument, category (a string). The classes 
	are ordered with superclasses first so they can be filed in."

	| behaviors classes |
	behaviors := (self listAtCategoryNamed: category asSymbol) 
			collect: [:title | Smalltalk at: title].
	classes := behaviors select: [:each | each isBehavior].
	^ChangeSet superclassOrder: classes! !

!SystemOrganizer methodsFor: 'query' stamp: 'dtl 8/26/2004 11:23'!
uncommentedClassesIn: categoryName

	"SystemOrganization uncommentedClassesIn: 'Morphic*'"

	| classes |
	classes := OrderedCollection new.
	self categories withIndexCollect: [:cat :idx |
		(categoryName match: cat)
			ifTrue: [classes addAll: (self listAtCategoryNumber: idx)]
			ifFalse: [nil]].
	^ (classes collect: [:clsName | Smalltalk at: clsName]
		thenSelect: [:cls | cls hasComment not]) asArray
! !

!StandardFileStream class methodsFor: '*System-Files-error handling' stamp: 'mtf 1/20/2011 14:14'!
fileDoesNotExistUserHandling: fullFileName ifDebug: debugBlock

	| selection newName |
	selection := UIManager default chooseFrom: {
		'create a new file' translated.
		'choose another name' translated.
		'debug' translated.
		'cancel' translated
	} title: (FileDirectory localNameFor: fullFileName) , '
does not exist.'.
	selection = 1 ifTrue:
		[^ self new open: fullFileName forWrite: true].
	selection = 2 ifTrue:
		[ newName := UIManager default request: 'Enter a new file name'
						initialAnswer:  fullFileName.
		^ self oldFileNamed:
			(self fullName: newName)].
	selection = 3 ifTrue: [^ debugBlock value].
	self halt! !

!StandardFileStream class methodsFor: '*System-Files-error handling' stamp: 'mtf 1/20/2011 14:12'!
fileExistsUserHandling: fullFileName ifDebug: debugBlock
	| dir localName choice newName newFullFileName |
	dir := FileDirectory forFileName: fullFileName.
	localName := FileDirectory localNameFor: fullFileName.
	choice := (UIManager default 
		chooseFrom: #('overwrite that file' 'append (risky!!!!)' 'choose another name' 'debug' 'cancel')
		title: localName, ' already exists.').

	choice = 1 ifTrue: [
		dir deleteFileNamed: localName
			ifAbsent: [self error: 'Could not delete the old version of that file'].
		^ self new open: fullFileName forWrite: true].

	choice = 2 ifTrue: [
		^ (self new open: fullFileName forWrite: true) setToEnd].

	choice = 3 ifTrue: [
		newName := UIManager default request: 'Enter a new file name' initialAnswer: fullFileName.
		newFullFileName := self fullName: newName.
		^ self newFileNamed: newFullFileName].

	choice = 4 ifTrue: [^ debugBlock value].

	self error: 'Please close this to abort file opening'! !

!StandardFileStream class methodsFor: '*System-Files-error handling' stamp: 'mtf 1/20/2011 14:13'!
readOnlyFileDoesNotExistUserHandling: fullFileName ifDebug: debugBlock

	| dir files choices selection newName fileName |
	dir := FileDirectory forFileName: fullFileName.
	files := dir fileNames.
	fileName := FileDirectory localNameFor: fullFileName.
	choices := fileName correctAgainst: files.
	choices add: 'Choose another name'.
	choices add: 'Debug'.
	choices add: 'Cancel'.
	selection := UIManager default chooseFrom: choices lines: (Array with: 5)
		title: (FileDirectory localNameFor: fullFileName), '
does not exist.'.
	selection = choices size ifTrue:["cancel" ^ nil "should we raise another exception here?"].
	selection < (choices size - 1) ifTrue: [
		newName := (dir pathName , FileDirectory slash , (choices at: selection))].
	selection = (choices size - 2) ifTrue: [
		newName := UIManager default 
							request: 'Enter a new file name' 
							initialAnswer: fileName].
	selection = (choices size - 1) ifTrue: [^ debugBlock value].
	newName = '' ifFalse: [^ self readOnlyFileNamed: (self fullName: newName)].
	^ self error: 'Could not open a file'! !

TextURL subclass: #TextSqkProjectLink
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!TextSqkProjectLink class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:16'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$p! !

!TextSqkProjectLink methodsFor: 'mouse events' stamp: 'dtl 11/27/2009 14:18'!
actOnClickFor: textMorph

	Project current enterIfThereOrFind: url.
	^ true! !

!TextSqkProjectLink methodsFor: 'as yet unclassified' stamp: 'RAA 10/19/2000 16:24'!
analyze: aString

	^url := aString! !

!TextSqkProjectLink methodsFor: 'as yet unclassified' stamp: 'RAA 10/19/2000 16:29'!
writeScanOn: strm

	strm nextPut: $p; nextPutAll: url; nextPut: $;! !

!TTFileDescription methodsFor: '*System-Tools-profiling' stamp: 'ar 7/30/2009 18:15'!
profileAll
	"Profile reading all the glyphs"
	MessageTally spyOn:[
		1 to: numGlyphs do:[:glyphIndex| self readGlyphAt: glyphIndex-1].
	].! !

Notification subclass: #ProgressTargetRequestNotification
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Exceptions'!

!ProgressTargetRequestNotification commentStamp: '<historical>' prior: 0!
I am used to allow the ComplexProgressIndicator one last chance at finding an appropriate place to display. If I am unhandled, then the cursor location and a default rectangle are used.!

!ProgressTargetRequestNotification methodsFor: 'handling' stamp: 'RAA 7/7/2000 12:25'!
defaultAction

	self resume: nil! !

Notification subclass: #ProjectEntryNotification
	instanceVariableNames: 'projectToEnter'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Exceptions'!

!ProjectEntryNotification commentStamp: '<historical>' prior: 0!
I provide a way to override the style of Project entry (which is buried deep in several different methods). My default is a normal full-screen enter.!

!ProjectEntryNotification class methodsFor: 'as yet unclassified' stamp: 'ajh 1/22/2003 23:52'!
signal: aProject

	| ex |
	ex := self new.
	ex projectToEnter: aProject.
	^ex signal: 'Entering ',aProject printString! !

!ProjectEntryNotification methodsFor: 'handling' stamp: 'RAA 6/6/2000 18:55'!
defaultAction

	self resume: projectToEnter enter! !

!ProjectEntryNotification methodsFor: 'accessing' stamp: 'RAA 6/6/2000 19:02'!
projectToEnter

	^projectToEnter! !

!ProjectEntryNotification methodsFor: 'accessing' stamp: 'RAA 6/6/2000 18:53'!
projectToEnter: aProject

	projectToEnter := aProject! !

Notification subclass: #ProjectPasswordNotification
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Exceptions'!

!ProjectPasswordNotification methodsFor: 'handling' stamp: 'RAA 4/23/2001 16:47'!
defaultAction

	self resume: ''! !

Notification subclass: #ProjectViewOpenNotification
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Exceptions'!

!ProjectViewOpenNotification commentStamp: '<historical>' prior: 0!
ProjectViewOpenNotification is signalled to determine if a ProjectViewMorph is needed for a newly created project. The default answer is yes.!

!ProjectViewOpenNotification methodsFor: 'handling' stamp: 'RAA 7/4/2000 16:24'!
defaultAction

	self resume: true! !

!Behavior methodsFor: '*System-Support' stamp: 'ul 8/8/2013 13:18'!
allCallsOn
	"Answer a sorted collection of all the methods that refer to me."
	
	^self systemNavigation allCallsOnClass: self! !

!Behavior methodsFor: '*System-Support' stamp: 'ul 3/14/2011 20:49'!
allCallsOn: aSymbol
	"Answer a sorted collection of all the methods that call on aSymbol."

	^self systemNavigation allCallsOn: aSymbol from: self.
	! !

!Behavior methodsFor: '*System-Support' stamp: 'ul 3/14/2011 20:55'!
allLocalCallsOn: aSymbol
	"Answer a sorted collection of all the methods that call on aSymbol, anywhere in my class hierarchy."

	^self systemNavigation allCallsOn: aSymbol from: self theNonMetaClass! !

!Behavior methodsFor: '*System-Support' stamp: 'ul 3/14/2011 20:56'!
allUnsentMessages
	"Answer an array of all the messages defined by the receiver that are not sent anywhere in the system."

	^self systemNavigation allUnSentMessagesIn: self selectors! !

!Behavior methodsFor: '*System-Changes' stamp: 'al 11/28/2005 21:48'!
changeRecordsAt: selector
	"Return a list of ChangeRecords for all versions of the method at selector. Source code can be retrieved by sending string to any one.  Return nil if the method is absent."

	"(Pen changeRecordsAt: #go:) collect: [:cRec | cRec string]"
	^ChangeSet 
		scanVersionsOf: (self compiledMethodAt: selector ifAbsent: [^ nil])
		class: self meta: self isMeta
		category: (self whichCategoryIncludesSelector: selector)
		selector: selector.! !

Exception subclass: #Abort
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Exceptions'!

!Abort methodsFor: 'handling' stamp: 'ajh 3/24/2003 00:55'!
defaultAction
	"No one has handled this error, but now give them a chance to decide how to debug it.  If none handle this either then open debugger (see UnhandedError-defaultAction)"

	UnhandledError signalForException: self! !

!StringHolder methodsFor: '*System-preferences' stamp: 'mt 7/9/2016 08:52'!
defaultWindowColor
	^ (Color r: 0.9 g: 0.9 b: 0.719)! !

!WeakArray class methodsFor: '*System-Finalization' stamp: 'eem 8/25/2015 10:29'!
finalizationProcess

	[FinalizationSemaphore wait.
	 FinalizationLock
		critical:
			[FinalizationDependents do:
				[ :weakDependent |
				weakDependent ifNotNil:
					[weakDependent finalizeValues]]]
		ifError:
			[:msg :rcvr | rcvr error: msg]] repeat! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'ul 3/22/2010 08:16'!
backChunk
	"Answer the contents of the receiver back to the previous terminator character.  Doubled terminators indicate an embedded terminator character."

	| output character |
	output := WriteStream on: (String new: 1000).
	self back. "oldBack compatibility"
	[ (character := self back) == nil ] whileFalse: [
		character == $!! ifTrue: [
			self back == $!! ifFalse: [ 
				self skip: 2. "oldBack compatibility"
				^output contents reversed ] ].
		output nextPut: character].
	self skip: 1. "oldBack compatibility"
	^output contents reversed! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'yo 8/7/2003 13:04'!
basicNextChunk
	"Answer the contents of the receiver, up to the next terminator character. Doubled terminators indicate an embedded terminator character."
	| terminator out ch |
	terminator := $!!.
	out := WriteStream on: (String new: 1000).
	self skipSeparators.
	[(ch := self next) == nil] whileFalse: [
		(ch == terminator) ifTrue: [
			self peek == terminator ifTrue: [
				self next.  "skip doubled terminator"
			] ifFalse: [
				^ out contents  "terminator is not doubled; we're done!!"
			].
		].
		out nextPut: ch.
	].
	^ out contents! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'sd 5/23/2003 14:40'!
checkForPreamble: chunk
	((chunk beginsWith: '"Change Set:') and: [ChangeSet current preambleString == nil])
		ifTrue: [ChangeSet current preambleString: chunk].
	((chunk beginsWith: '"Postscript:') and: [ChangeSet current postscriptString == nil])
		ifTrue: [ChangeSet current postscriptString: chunk].
							
! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: ''!
command: aString
	"Overridden by HtmlFileStream to append commands directly without translation.  4/5/96 tk"
	"We ignore any HTML commands.  Do nothing"! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'di 2/3/98 14:44'!
copyMethodChunkFrom: aStream
	"Copy the next chunk from aStream (must be different from the receiver)."
	| chunk |
	chunk := aStream nextChunkText.
	chunk runs values size = 1 "Optimize for unembellished text"
		ifTrue: [self nextChunkPut: chunk asString]
		ifFalse: [self nextChunkPutWithStyle: chunk]! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'yo 10/15/2003 19:02'!
copyMethodChunkFrom: aStream at: pos
	"Copy the next chunk from aStream (must be different from the receiver)."
	| chunk |
	aStream position: pos.
	chunk := aStream nextChunkText.
	chunk runs values size = 1 "Optimize for unembellished text"
		ifTrue: [self nextChunkPut: chunk asString]
		ifFalse: [self nextChunkPutWithStyle: chunk]! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'SergeStinckwich 7/31/2009 15:25'!
copyPreamble: preamble from: aStream at: pos 
	"Look for a changeStamp for this method by peeking backward.
	Write a method preamble, with that stamp if found."
	| terminator last50 stamp i |
	terminator := $!!.

	"Look back to find stamp in old preamble, such as...
	Polygon methodsFor: 'private' stamp: 'di 6/25/97 21:42' prior: 34957598!! "
	aStream position: pos.
	aStream backChunk.	"to beginning of method"
	last50 := aStream backChunk.	"to get preamble"
	aStream position: pos.
	stamp := String new.
	(i := last50 
		findLastOccurrenceOfString: 'stamp:'
		startingAt: 1) > 0 ifTrue: 
		[ stamp := (last50 
			copyFrom: i + 8
			to: last50 size) copyUpTo: $' ].

	"Write the new preamble, with old stamp if any."
	self
		cr;
		nextPut: terminator.
	self nextChunkPut: (String streamContents: 
			[ :strm | 
			strm nextPutAll: preamble.
			stamp size > 0 ifTrue: 
				[ strm
					nextPutAll: ' stamp: ';
					print: stamp ] ]).
	self cr! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'nice 2/26/2013 00:17'!
decodeString: string andRuns: runsRaw

	| strm runLength runValues newString index |
	strm := runsRaw readStream.
	(strm peekFor: $( ) ifFalse: [^ nil].
	runLength := OrderedCollection new.
	[strm skipSeparators.
	 strm peekFor: $)] whileFalse: 
		[runLength add: (Number readFrom: strm)].

	runValues := OrderedCollection new.
	[strm atEnd not] whileTrue: 
		[runValues add: (Number readFrom: strm).
		strm next.].

	newString := WideString new: string size.
	index := 1.
	runLength with: runValues do: [:length :leadingChar |
		index to: index + length - 1 do: [:pos |
			newString at: pos put: (Character leadingChar: leadingChar code: (string at: pos) charCode).
		].
		index := index + length.
	].

	^ newString.
! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'tk 12/13/97 13:36'!
decodeStyle: runsObjData version: styleVersion
	"Decode the runs array from the ReferenceStream it is stored in."
	"Verify that the class mentioned have the same inst vars as we have now"

	| structureInfo |
	styleVersion = RemoteString currentTextAttVersion ifTrue: [
		"Matches our classes, no need for checking"
		^ (ReferenceStream on: runsObjData) next].
	structureInfo := RemoteString structureAt: styleVersion.	"or nil"
		"See SmartRefStream instVarInfo: for dfn"
	^ SmartRefStream read: runsObjData withClasses: structureInfo! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'mir 7/26/2000 13:28'!
fileIn
	"This is special for reading expressions from text that has been formatted 
	with exclamation delimitors. The expressions are read and passed to the 
	Compiler. Answer the result of compilation."

	^ self fileInAnnouncing: 'Reading ' , self name! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'cmm 3/26/2015 15:32'!
fileInAnnouncing: announcement 
	"This is special for reading expressions from text that has been formatted 
	with exclamation delimitors. The expressions are read and passed to the 
	Compiler. Answer the result of compilation.  Put up a progress report with
     the given announcement as the title."

	| val |
	announcement 
		displayProgressFrom: 0
		to: self size
		during: 
			[:bar | 
			[self atEnd] whileFalse: 
					[bar value: self position.
					self skipSeparators.
					
					[ | chunk |
					val := (self peekFor: $!!) 
								ifTrue: [(Compiler evaluate: self nextChunk logged: true) scanFrom: self]
								ifFalse: 
									[chunk := self nextChunk.
									self checkForPreamble: chunk.
									Compiler evaluate: chunk logged: true]] 
							on: InMidstOfFileinNotification
							do: [:ex | ex resume: true].
					self skipStyleChunk].
			self close].
	"Note:  The main purpose of this banner is to flush the changes file."
	Smalltalk logChange: '----End fileIn of ' , self name , '----'.
	self flag: #ThisMethodShouldNotBeThere.	"sd"
	^val! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'fbs 7/25/2013 07:15'!
fileInFor: client announcing: announcement
	"This is special for reading expressions from text that has been formatted 
	with exclamation delimitors. The expressions are read and passed to the 
	Compiler. Answer the result of compilation.  Put up a progress report with
     the given announcement as the title.
	Does NOT handle preambles or postscripts specially."
	| val |
	announcement
		displayProgressFrom: 0
		to: self size
		during:
		[:bar |
		[self atEnd]
			whileFalse: 
				[bar value: self position.
				self skipSeparators.
				[ | chunk |
				val := (self peekFor: $!!) ifTrue: [
						(Compiler evaluate: self nextChunk for: client logged: false) scanFrom: self
					] ifFalse: [
						chunk := self nextChunk.
						self checkForPreamble: chunk.
						Compiler evaluate: chunk for: client logged: true ].
				] on: InMidstOfFileinNotification
				  do: [ :ex | ex resume: true].
				self atEnd ifFalse: [ self skipStyleChunk ]].
		self close].
	"Note:  The main purpose of this banner is to flush the changes file."
	Smalltalk logChange: '----End fileIn of ' , self name , '----'.
	^ val! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'fbs 7/25/2013 07:15'!
fileInSilentlyAnnouncing: announcement 
	"This is special for reading expressions from text that has been formatted 
	with exclamation delimitors. The expressions are read and passed to the 
	Compiler. Answer the result of compilation.  Put up a progress report with
     the given announcement as the title."

	| val chunk |
	[self atEnd] whileFalse: 
			[self skipSeparators.
			
			[val := (self peekFor: $!!) 
						ifTrue: [(Compiler evaluate: self nextChunk logged: false) scanFrom: self]
						ifFalse: 
							[chunk := self nextChunk.
							self checkForPreamble: chunk.
							Compiler evaluate: chunk logged: true]] 
					on: InMidstOfFileinNotification
					do: [:ex | ex resume: true].
			self skipStyleChunk].
	self close.
	"Note:  The main purpose of this banner is to flush the changes file."
	Smalltalk  logChange: '----End fileIn of ' , self name , '----'.
	self flag: #ThisMethodShouldNotBeThere.	"sd"
	SystemNavigation new allBehaviorsDo: 
			[:cl | 
			cl
				removeSelectorSimply: #DoIt;
				removeSelectorSimply: #DoItIn:].
	^val! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: ''!
header
	"If the stream requires a standard header, override this message.  See HtmlFileStream"! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'yo 8/13/2003 11:59'!
nextChunk
	"Answer the contents of the receiver, up to the next terminator character. Doubled terminators indicate an embedded terminator character."
	| terminator out ch |
	terminator := $!!.
	out := WriteStream on: (String new: 1000).
	self skipSeparators.
	[(ch := self next) == nil] whileFalse: [
		(ch == terminator) ifTrue: [
			self peek == terminator ifTrue: [
				self next.  "skip doubled terminator"
			] ifFalse: [
				^ self parseLangTagFor: out contents  "terminator is not doubled; we're done!!"
			].
		].
		out nextPut: ch.
	].
	^ self parseLangTagFor: out contents.
! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'nice 2/25/2013 23:25'!
nextChunkText
	"Deliver the next chunk as a Text.  Decode the following ]style[ chunk if present.  Position at start of next real chunk."
	| string runs peek pos |
	"Read the plain text"
	string := self nextChunk.
	
	"Test for ]style[ tag"
	pos := self position.
	peek := self skipSeparatorsAndPeekNext.
	peek = $] ifFalse: [self position: pos. ^ string asText].  "no tag"
	(self upTo: $[) = ']style' ifFalse: [self position: pos. ^ string asText].  "different tag"

	"Read and decode the style chunk"
	runs := RunArray scanFrom: self basicNextChunk readStream.

	^ Text basicNew setString: string setRunsChecking: runs.
! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'sumim 11/20/2003 18:11'!
parseLangTagFor: aString

	| string peek runsRaw pos |
	string := aString.
	"Test for ]lang[ tag"
	pos := self position.
	peek := self skipSeparatorsAndPeekNext.
	peek = $] ifFalse: [self position: pos. ^ string].  "no tag"
	(self upTo: $[) = ']lang' ifTrue: [
		runsRaw := self basicNextChunk.
		string := self decodeString: aString andRuns: runsRaw
	] ifFalse: [
		self position: pos
	].
	^ string.
! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'di 1/13/98 16:08'!
skipSeparatorsAndPeekNext
	"A special function to make nextChunk fast"
	| peek |
	[self atEnd]
		whileFalse:
		[(peek := self next) isSeparator
			ifFalse: [self position: self position-1. ^ peek]]! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: 'tk 12/29/97 12:37'!
skipStyleChunk
	"Get to the start of the next chunk that is not a style for the previous chunk"

	| pos |
	pos := self position.
	self skipSeparators.
	self peek == $] 
		ifTrue: [(self upTo: $[) = ']text' 	"old -- no longer needed"
				"now positioned past the open bracket"
			ifFalse: [self nextChunk]]	"absorb ]style[ and its whole chunk"
				
		ifFalse: [self position: pos]	"leave untouched"
! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: ''!
trailer
	"If the stream requires a standard trailer, override this message.  See HtmlFileStream"! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: ''!
unCommand
	"If this read stream is at a <, then skip up to just after the next >.  For removing html commands."
	| char |
	[self peek = $<] whileTrue: ["begin a block"
		[self atEnd == false and: [self next ~= $>]] whileTrue.
		"absorb characters"
		].
 ! !

!PositionableStream methodsFor: '*System-Changes-fileIn/Out' stamp: ''!
verbatim: aString
	"Do not attempt to translate the characters.  Use to override nextPutAll:"
	^ self nextPutAll: aString! !

OrderedCollection subclass: #ObjectFinalizerCollection
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Finalization'!

!ObjectFinalizerCollection commentStamp: 'ul 2/26/2010 14:23' prior: 0!
An ObjectFinalizerCollection is simple collection which is intended to hold ObjectFinalizers and be used by WeakRegistry.!

!ObjectFinalizerCollection methodsFor: 'finalization' stamp: 'ul 10/17/2010 03:34'!
finalize
	"Finalize all elements in this collection. The elements are expected to be ObjectFinalizers, but can be any object which understands #finalize."

	self do: [ :each | each finalize ]! !

!ObjectFinalizerCollection methodsFor: 'testing' stamp: 'Igor.Stasenko 5/25/2010 04:58'!
hasMultipleExecutors
	^ true! !

!Warning methodsFor: '*System-exceptionDescription' stamp: 'mt 8/25/2015 14:31'!
defaultAction
	"Inform the user of a Warning, giving them the choice of ignoring the warning (proceeding), debugging, or terminating the computation."
	UnhandledWarning signalForException: self! !

!GZipReadStream class methodsFor: '*System-FileRegistry' stamp: 'sd 2/1/2002 22:15'!
serviceViewDecompress

	^ SimpleServiceEntry 
				provider: self 
				label: 'view decompressed'
				selector: #viewContents:
				description: 'view decompressed' 
! !

!GZipReadStream class methodsFor: '*System-FileRegistry' stamp: 'fbs 6/27/2013 17:36'!
viewContents: fullFileName
	"Open the decompressed contents of the .gz file with the given name.  This method is only required for the registering-file-list of Squeak 3.3a and beyond, but does no harm in an earlier system"

	(FileStream readOnlyFileNamed: fullFileName) ifNotNil:
		[:aStream | | contents |
		contents := Cursor wait showWhile: [aStream viewGZipContents].
		UIManager default
			edit: contents
			label: 'Decompressed contents of: ', aStream localName].! !

!UnhandledWarning methodsFor: '*System-priv handling' stamp: 'mt 8/25/2015 14:40'!
defaultAction

	^ ToolSet
		debugContext: self exception signalerContext
		label: 'Warning'
		contents: self exception messageText , '\\Select Proceed to continue, or close this window to cancel the operation.' withCRs! !

MessageSend subclass: #ObjectFinalizer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Finalization'!

!ObjectFinalizer commentStamp: 'jcg 4/27/2010 01:01' prior: 0!
Represents an action that is to be taken when a target object (not referenced by the ObjectFinalizer) is garbage-collected.!

!ObjectFinalizer methodsFor: 'finalization' stamp: 'jcg 1/15/2010 01:37'!
finalize
	"Finalize the resource associated with the receiver. This message should only be sent during the finalization process. There is NO garantuee that the resource associated with the receiver hasn't been free'd before so take care that you don't run into trouble - this all may happen with interrupt priority."
	[self value] on: Error do:[:ex| ex return].
! !

!ObjectFinalizer methodsFor: 'initialize' stamp: 'jcg 1/15/2010 01:36'!
receiver: anObject selector: aSymbol argument: aParameter
	"Not really necessary now that we're a subclass of MessageSend, but we keep it around because existing code might expect it."
	self 
		receiver: anObject;
		selector: aSymbol;
		arguments: (Array with: aParameter).! !

Halt subclass: #BreakPoint
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Tools'!

!BreakPoint commentStamp: 'md 11/18/2003 09:32' prior: 0!
This exception is raised on executing a breakpoint.

"BreakPoint signal" is called from "Object>>break".!

ProtoObject subclass: #ImageSegmentRootStub
	instanceVariableNames: 'shadowSuper shadowMethodDict shadowFormat imageSegment'
	classVariableNames: 'FaultLogs LoggingFaults'
	poolDictionaries: ''
	category: 'System-Object Storage'!

!ImageSegmentRootStub commentStamp: '<historical>' prior: 0!
An ImageSegmentRootStub is a stub that replaces one of the root of an ImageSegment that has been extracted from the Squeak ObjectMemory.  It has two very simple roles:

1.  If any message is sent to one of these objects, it will be caught by doesNotUnderstand:, and bring about a reinstallation of the missing segment.  This exception is caused by the fact that no other messages are defined in this class, and neither does it inherit any from above, since its superclass is nil.  When the reinstallation has been accomplished, the message will be resent as though nothing was amiss.

2.  If one of these objects is a class, and a message is sent to one of its instances, it will cause a similar fault which will be caught by cannotInterpret:.  This exception is caused by a somewhat more subtle condition:  the primitive operations of the virtual machine do not have time to check whether classes are resident or not -- they assume that all classes are resident.  However every non-cached message lookup does test for a nil in the methodDictionary slot.  If a rootStub replaces a class (or any behavior), it masquerades as the class, but it will have a nil in the slot where the method Dictionary is expected.  This will cause the VM to send cannotInterpret:, eventually leading to the same process for reinstalling the missing segment and resending the message as above.

Just to be on the safe side, a rootStub that replaces a Behavior also carries a copy of both the superclass and format fields from the original class.  This insures that, even if some operations of the VM require these values, things will continue to operate properly when the segment is absent.!

!ImageSegmentRootStub class methodsFor: 'settings' stamp: 'di 2/7/2000 23:43'!
doLogFaults  "ImageSegmentRootStub doLogFaults"

	FaultLogs := Dictionary new.
	LoggingFaults := true.! !

!ImageSegmentRootStub class methodsFor: 'settings' stamp: 'di 2/7/2000 23:17'!
dontLogFaults  "ImageSegmentRootStub dontLogFaults"

	FaultLogs := Dictionary new.
	LoggingFaults := false.! !

!ImageSegmentRootStub class methodsFor: 'accessing' stamp: 'di 2/11/2000 12:33'!
faultLogs  "ImageSegmentRootStub faultLogs"  "<-- inspect it"

	^ FaultLogs! !

!ImageSegmentRootStub class methodsFor: 'settings' stamp: 'di 2/11/2000 12:31'!
startLoggingFaults  "ImageSegmentRootStub startLoggingFaults"

	FaultLogs := Dictionary new.
	LoggingFaults := true.! !

!ImageSegmentRootStub class methodsFor: 'settings' stamp: 'di 2/11/2000 12:33'!
stopLoggingFaults  "ImageSegmentRootStub stopLoggingFaults"

	FaultLogs := Dictionary new.
	LoggingFaults := false.! !

!ImageSegmentRootStub methodsFor: 'fetch from disk' stamp: 'di 3/4/2001 22:45'!
doesNotUnderstand: aMessage 
	 | segmentName |
"Any normal message sent to this object is really intended for another object that is in a non-resident imageSegment.  Reinstall the segment and resend the message."

	segmentName := imageSegment segmentName.
	imageSegment install.
	LoggingFaults ifTrue:		"Save the stack printout to show who caused the fault"
		[FaultLogs at: Time millisecondClockValue printString
			put: (String streamContents:
				[:strm | 
				strm nextPutAll: segmentName; cr.
				strm print: self class; space; print: aMessage selector; cr.
				(thisContext sender stackOfSize: 30)
					do: [:item | strm print: item; cr]])].

	"NOTE:  The following should really be (aMessage sentTo: self)
		in order to recover properly from a fault in a super-send,
		however, the lookupClass might be bogus in this case, and it's
		almost unthinkable that the first fault would be a super send."
	^ self perform: aMessage selector withArguments: aMessage arguments! !

!ImageSegmentRootStub methodsFor: 'basics' stamp: 'tk 8/13/1999 15:59'!
isInMemory
	"We are a place holder for an object that is out."
	^ false! !

!ImageSegmentRootStub methodsFor: 'fetch from disk' stamp: 'di 3/27/1999 12:19'!
xxSuperclass: superclass format: format segment: segment

	"Set up fields like a class but with null methodDict"
	shadowSuper := superclass.
	shadowMethodDict := nil.
	shadowFormat := format.
	imageSegment := segment.
! !

!ImageSegmentRootStub methodsFor: 'fetch from disk' stamp: 'tk 4/9/1999 10:32'!
xxxClass
	"Primitive. Answer the object which is the receiver's class. Essential. See 
	Object documentation whatIsAPrimitive."

	<primitive: 111>
	self primitiveFailed! !

!ImageSegmentRootStub methodsFor: 'fetch from disk' stamp: 'tk 10/24/1999 10:57'!
xxxSegment
	^ imageSegment! !

ProtoObject subclass: #WrappedBreakpoint
	instanceVariableNames: 'method'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Tools'!

!WrappedBreakpoint commentStamp: 'topa 3/26/2015 21:51' prior: 0!
I am a wrapper around an actual method that should be debugged.
Contrary to my siblings BreakPoint and BreakpointManager I do not need
to modify the original method nor its source but rather implant me in my
method's position.

I am based on OBBreakpoint from the OmniBrowser framework.

Instance Variables
	method:		<CompiledMethod>

method
	- actual method I wrap
!

!WrappedBreakpoint class methodsFor: 'instance creation' stamp: 'topa 3/26/2015 21:54'!
on: aCompiledMethod
	^ self basicNew initializeOn: aCompiledMethod! !

!WrappedBreakpoint methodsFor: 'private' stamp: 'lr 9/20/2010 13:15'!
doesNotUnderstand: aMessage
	^ method 
		perform: aMessage selector
		withArguments: aMessage arguments! !

!WrappedBreakpoint methodsFor: 'private' stamp: 'lr 4/3/2011 11:08'!
flushCache
	method selector flushCache! !

!WrappedBreakpoint methodsFor: 'public' stamp: 'topa 3/26/2015 22:17'!
hasBreakpoint
	^ true! !

!WrappedBreakpoint methodsFor: 'initialization' stamp: 'lr 9/20/2010 13:12'!
initializeOn: aCompiledMethod
	method := aCompiledMethod! !

!WrappedBreakpoint methodsFor: 'public' stamp: 'lr 9/20/2010 13:14'!
install
	method methodClass methodDictionary
		at: method selector
		put: self! !

!WrappedBreakpoint methodsFor: 'literals' stamp: 'topa 3/26/2015 21:52'!
literalsDo: aBlock
	"This method is necessary to show the breakpoint-flag in the browser."
	
	aBlock value: #break.
	method literalsDo: aBlock! !

!WrappedBreakpoint methodsFor: 'private' stamp: 'topa 3/26/2015 21:56'!
printOn: aStream

	aStream nextPutAll: 'Break in '.
	method printOn: aStream.! !

!WrappedBreakpoint methodsFor: 'private' stamp: 'topa 3/26/2015 22:06'!
printStringLimitedTo: limit
	"Answer a String whose characters are a description of the receiver.
	If you want to print without a character limit, use fullPrintString."
	| limitedString |
	limitedString := String streamContents: [:s | self printOn: s] limitedTo: limit.
	limitedString size < limit ifTrue: [^ limitedString].
	^ limitedString , '...etc...'! !

!WrappedBreakpoint methodsFor: 'evaluation' stamp: 'mt 5/9/2016 12:50'!
run: aSelector with: anArray in: aReceiver
	| process |
	process := Process 
		forContext: (MethodContext
			sender: thisContext sender
			receiver: aReceiver
			method: method
			arguments: anArray)
		priority: Processor activeProcess priority.
	ToolSet
		debug: process context: process suspendedContext
		label:  'Breakpoint in ' , method methodClass name , '>>#' , method selector
		contents: nil fullView: true.
	Project current spawnNewProcessIfThisIsUI: Processor activeProcess.
	thisContext swapSender: nil.
	Processor activeProcess terminate.! !

!WrappedBreakpoint methodsFor: 'public' stamp: 'lr 9/20/2010 13:14'!
uninstall
	method methodClass methodDictionary
		at: method selector
		put: method! !

!UnhandledError methodsFor: '*System-priv handling' stamp: 'mt 8/25/2015 14:43'!
defaultAction
	"The current computation is terminated. The cause of the error should be logged or reported to the user. If the program is operating in an interactive debugging environment the computation should be suspended and the debugger activated."
	^ToolSet debugError: self exception! !

!OutOfMemory methodsFor: '*System-exceptionDescription' stamp: 'ar 2/10/2007 10:43'!
defaultAction
	Smalltalk signalLowSpace.! !

!WriteStream methodsFor: '*System-Object Storage-fileIn/Out' stamp: 'RAA 4/6/2001 18:32'!
fileOutChangeSet: aChangeSetOrNil andObject: theObject
	"Write a file that has both the source code for the named class and an object as bits.  Any instance-specific object will get its class written automatically."

	"An experimental version to fileout a changeSet first so that a project can contain its own classes"


	self setFileTypeToObject.
		"Type and Creator not to be text, so can attach correctly to an email msg"
	self header; timeStamp.

	aChangeSetOrNil ifNotNil: [
		aChangeSetOrNil fileOutPreambleOn: self.
		aChangeSetOrNil fileOutOn: self.
		aChangeSetOrNil fileOutPostscriptOn: self.
	].
	self trailer.	"Does nothing for normal files.  HTML streams will have trouble with object data"

	"Append the object's raw data"
	(SmartRefStream on: self)
		nextPut: theObject;  "and all subobjects"
		close.		"also closes me"
! !

!WriteStream methodsFor: '*System-Object Storage-fileIn/Out' stamp: 'sd 5/23/2003 14:41'!
fileOutChanges
	"Append to the receiver a description of all class changes."
	Cursor write showWhile:
		[self header; timeStamp.
		ChangeSet current fileOutOn: self.
		self trailer; close]! !

!WriteStream methodsFor: '*System-Object Storage-fileIn/Out' stamp: 'yo 8/16/2004 13:45'!
fileOutClass: extraClass andObject: theObject
	"Write a file that has both the source code for the named class and an object as bits.  Any instance-specific object will get its class written automatically."

	| class srefStream |
	self setFileTypeToObject.
		"Type and Creator not to be text, so can attach correctly to an email msg"
	self text.
	self header; timeStamp.

	extraClass ifNotNil: [
		class := extraClass.	"A specific class the user wants written"
		class sharedPools size > 0 ifTrue:
			[class shouldFileOutPools
				ifTrue: [class fileOutSharedPoolsOn: self]].
		class fileOutOn: self moveSource: false toFile: 0].
	self trailer.	"Does nothing for normal files.  HTML streams will have trouble with object data"
	self binary.

	"Append the object's raw data"
	srefStream := SmartRefStream on: self.
	srefStream nextPut: theObject.  "and all subobjects"
	srefStream close.		"also closes me"
! !

!WriteStream methodsFor: '*System-Object Storage-fileIn/Out' stamp: 'tk 3/13/98 22:25'!
fileOutClass: extraClass andObject: theObject blocking: anIdentDict
	"Write a file that has both the source code for the named class and an object as bits.  Any instance-specific object will get its class written automatically.  Accept a list of objects to map to nil or some other object (blockers).  In addition to teh choices in each class's objectToStoreOnDataStream"

	| class srefStream |
	self setFileTypeToObject.
		"Type and Creator not to be text, so can attach correctly to an email msg"
	self header; timeStamp.

	extraClass ifNotNil: [
		class := extraClass.	"A specific class the user wants written"
		class sharedPools size > 0 ifTrue:
			[class shouldFileOutPools
				ifTrue: [class fileOutSharedPoolsOn: self]].
		class fileOutOn: self moveSource: false toFile: 0].
	self trailer.	"Does nothing for normal files.  HTML streams will have trouble with object data"

	"Append the object's raw data"
	srefStream := SmartRefStream on: self.
	srefStream blockers: anIdentDict.
	srefStream nextPut: theObject.  "and all subobjects"
	srefStream close.		"also closes me"
! !

!WriteStream methodsFor: '*System-Object Storage-fileIn/Out' stamp: 'nice 7/27/2014 22:51'!
padToEndWith: aChar
	"We don't have pages, so we are at the end, and don't need to pad."! !

!WriteStream methodsFor: '*System-Object Storage-fileIn/Out' stamp: 'nice 7/27/2014 15:32'!
setFileTypeToObject

	"ignore"! !

!SyntaxErrorNotification methodsFor: '*System-exceptionDescription' stamp: 'ar 9/27/2005 19:13'!
defaultAction
	^ToolSet debugSyntaxError: self! !

Stream subclass: #DataStream
	instanceVariableNames: 'byteStream topCall basePos'
	classVariableNames: 'TypeMap'
	poolDictionaries: ''
	category: 'System-Object Storage'!

!DataStream commentStamp: 'nice 3/25/2010 23:01' prior: 0!
This is the save-to-disk facility. A DataStream can store one or more objects in a persistent form.

To handle objects with sharing and cycles, you must use a
ReferenceStream instead of a DataStream.  (Or SmartRefStream.)  ReferenceStream is typically
faster and produces smaller files because it doesn't repeatedly write the same Symbols.

Here is the way to use DataStream and ReferenceStream:
	rr := ReferenceStream fileNamed: 'test.obj'.
	rr nextPut: <your object>.
	rr close.

To get it back:
	rr := ReferenceStream fileNamed: 'test.obj'.
	<your object> := rr next.
	rr close.

Each object to be stored has two opportunities to control what gets stored.  On the high level, objectToStoreOnDataStream allows you to substitute another object on the way out.  The low level hook is storeDataOn:. The read-in counterparts to these messages are comeFullyUpOnReload and (class) readDataFrom:size:. See these methods, and the class DiskProxy, for more information about externalizing and internalizing.

NOTE: A DataStream should be treated as a write-stream for writing.  It is a read-stream for reading.  It is not a ReadWriteStream.
!

!DataStream class methodsFor: 'class initialization' stamp: 'ar 2/26/2010 23:02'!
cleanUp
	"Re-initialize DataStream to avoid hanging onto obsolete classes"

	self initialize! !

!DataStream class methodsFor: 'as yet unclassified' stamp: 'topa 6/25/2015 08:52'!
example
    "An example and test of DataStream/ReferenceStream.
     11/19/92 jhm: Use self testWith:."
    "DataStream example"
    "ReferenceStream example"
    | input sharedPoint |

    "Construct the test data."
    input := Array new: 10.
    input at: 1 put: nil.
    input at: 2 put: true.
    input at: 3 put: (Form extent: 63 @ 50 depth: 8).
		(input at: 3) fillWithColor: Color lightBlue.
    input at: 4 put: #(3 3.0 'three').
    input at: 5 put: false.
    input at: 6 put: 1024 @ -2048.
    input at: 7 put: #x.
    input at: 8 put: (Array with: (sharedPoint := 0 @ -30000)).
    input at: 9 put: sharedPoint.
    input at: 10 put: (Character value: 16r200d).

    "Write it out, read it back, and return it for inspection."
    ^ self testWith: input! !

!DataStream class methodsFor: 'as yet unclassified' stamp: ''!
exampleWithPictures
	"DataStream exampleWithPictures"
	| file result |
	file := FileStream fileNamed: 'Test-Picture'.
	file binary.
	(DataStream on: file) nextPut: (Form fromUser).
	file close.

	file := FileStream fileNamed: 'Test-Picture'.
	file binary.
	result := (DataStream on: file) next.
	file close.
	result display.
	^ result! !

!DataStream class methodsFor: 'as yet unclassified' stamp: ''!
fileNamed: aString
	"Here is the way to use DataStream and ReferenceStream:
rr := ReferenceStream fileNamed: 'test.obj'.
rr nextPut: <your object>.
rr close.
"

	| strm |
	strm := self on: (FileStream fileNamed: aString).		"will be binary"
	strm byteStream setFileTypeToObject.
		"Type and Creator not to be text, so can attach correctly to an email msg"
	^ strm! !

!DataStream class methodsFor: 'as yet unclassified' stamp: 'fbs 7/12/2013 08:38'!
initialize
	"TypeMap maps Smalltalk classes to type ID numbers which identify the data stream primitive formats.  nextPut: writes these IDs to the data stream.  NOTE: Changing these type ID numbers will invalidate all extant data stream files.  Adding new ones is OK.  
	Classes named here have special formats in the file.  If such a class has a subclass, it will use type 9 and write correctly.  It will just be slow.  (Later write the class name in the special format, then subclasses can use the type also.)
	 See nextPut:, next, typeIDFor:, & ReferenceStream>>isAReferenceType:"
	"DataStream initialize"

	| refTypes t |
	refTypes := OrderedCollection new.
	t := TypeMap := WeakIdentityKeyDictionary new. "It has to be weak, because some classes may go away, leaving obsolete versions in this dictionary which may make it corrupt."

	t at: UndefinedObject put: 1.   refTypes add: 0.
	t at: True put: 2.   refTypes add: 0.
	t at: False put: 3.   refTypes add: 0.
	t at: SmallInteger put: 4.	 refTypes add: 0.
	t at: ByteString put: 5.   refTypes add: 1.
	t at: ByteSymbol put: 6.   refTypes add: 1.
	t at: ByteArray put: 7.   refTypes add: 1.
	t at: Array put: 8.   refTypes add: 1.
	"(type ID 9 is for arbitrary instances of any class, cf. typeIDFor:)"
		refTypes add: 1.
	"(type ID 10 is for references, cf. ReferenceStream>>tryToPutReference:)"
		refTypes add: 0.
	t at: Bitmap put: 11.   refTypes add: 1.
	t at: Metaclass put: 12.   refTypes add: 0.
	"Type ID 13 is used for HyperSqueak User classes that must be reconstructed."
		refTypes add: 1.
	t at: Float put: 14.  refTypes add: 1.
	t at: Rectangle put: 15.  refTypes add: 1.	"Allow compact Rects."
	"type ID 16 is an instance with short header.  See beginInstance:size:"
		refTypes add: 1.
self flag: #ByteArray.
	t at: ByteString put: 17.   refTypes add: 1.	"new String format, 1 or 4 bytes of length"
	t at: WordArray put: 18.  refTypes add: 1.	"bitmap-like"
	t at: WordArrayForSegment put: 19.  refTypes add: 1.		"bitmap-like"
	t at: SoundBuffer put: 20.  refTypes add: 1.	"And all other word arrays, both 
		16-bit and 32-bit.  See methods in ArrayedCollection.  Overridden in SoundBuffer."
	t at: CompiledMethod put: 21.  refTypes add: 1.	"special creation method"
	"t at:  put: 22.  refTypes add: 0."
	ReferenceStream refTypes: refTypes.		"save it"

	"For all classes that are like WordArrays, store them the way ColorArray is stored.  As bits, and able to change endianness."
	Smalltalk globals do: [:cls |
		(cls isInMemory and: [
			cls isBehavior and: [
			cls isObsolete not and: [
			cls isPointers not and: [
			cls isVariable and: [
			cls isWords and: [
			(t includesKey: cls) not ] ] ] ] ] ]) 
				ifTrue: [ t at: cls put: 20 ] ]! !

!DataStream class methodsFor: 'as yet unclassified' stamp: 'di 2/15/98 14:03'!
new
	^ self basicNew! !

!DataStream class methodsFor: 'as yet unclassified' stamp: ''!
newFileNamed: aString
	"Here is the way to use DataStream and ReferenceStream:
rr := ReferenceStream fileNamed: 'test.obj'.
rr nextPut: <your object>.
rr close.
"

	| strm |
	strm :=  self on: (FileStream newFileNamed: aString).		"will be binary"
	strm byteStream setFileTypeToObject.
		"Type and Creator not to be text, so can attach correctly to an email msg"
	^ strm! !

!DataStream class methodsFor: 'as yet unclassified' stamp: ''!
oldFileNamed: aString
	"Here is the way to use DataStream and ReferenceStream:
rr := ReferenceStream oldFileNamed: 'test.obj'.
^ rr nextAndClose.
"

	| strm ff |
	ff := FileStream oldFileOrNoneNamed: aString.
	ff ifNil: [^ nil].
	strm := self on: (ff binary).
	^ strm! !

!DataStream class methodsFor: 'as yet unclassified' stamp: 'di 6/24/97 00:18'!
on: aStream
	"Open a new DataStream onto a low-level I/O stream."

	^ self basicNew setStream: aStream
		"aStream binary is in setStream:"
! !

!DataStream class methodsFor: 'as yet unclassified' stamp: 'RAA 7/28/2000 08:38'!
streamedRepresentationOf: anObject

	| file |
	file := (RWBinaryOrTextStream on: (ByteArray new: 5000)).
	file binary.
	(self on: file) nextPut: anObject.
	^file contents! !

!DataStream class methodsFor: 'as yet unclassified' stamp: 'eem 2/3/2015 10:23'!
testWith: anObject
	"As a test of DataStream/ReferenceStream, write out anObject and read it back.
	11/19/92 jhm: Set the file type. More informative file name."
	"DataStream testWith: 'hi'"
	"ReferenceStream testWith: 'hi'"
	| file result |

	file := FileStream fileNamed: (self name, ' test').
	file binary.
	(self on: file) nextPut: anObject.
	file close.

	file := FileStream fileNamed: (self name, ' test').
	file binary.
	result := (self on: file) next.
	file close.
	(anObject class whichClassIncludesSelector: #=) == Object ifFalse:
		[self assert: result = anObject].
	^result! !

!DataStream class methodsFor: 'as yet unclassified' stamp: 'RAA 7/28/2000 08:33'!
unStream: aString

	^(self on: ((RWBinaryOrTextStream with: aString) reset; binary)) next! !

!DataStream methodsFor: 'other' stamp: ''!
atEnd
    "Answer true if the stream is at the end."

    ^ byteStream atEnd! !

!DataStream methodsFor: 'write and read' stamp: '6/9/97 08:14 tk'!
beginInstance: aClass size: anInteger
	"This is for use by storeDataOn: methods.
	 Cf. Object>>storeDataOn:."

		"Addition of 1 seems to make extra work, since readInstance
		has to compensate.  Here for historical reasons dating back
		to Kent Beck's original implementation in late 1988.

		In ReferenceStream, class is just 5 bytes for shared symbol.

		SmartRefStream puts out the names and number of class's instances variables for checking."

	byteStream nextNumber: 4 put: anInteger + 1.

	self nextPut: aClass name! !

!DataStream methodsFor: 'write and read' stamp: 'yo 12/3/2004 17:12'!
beginReference: anObject
    "We're starting to read anObject. Remember it and its reference
     position (if we care; ReferenceStream cares). Answer the
     reference position."

    ^ 0! !

!DataStream methodsFor: 'other' stamp: ''!
byteStream
	^ byteStream! !

!DataStream methodsFor: 'other' stamp: ''!
close
	"Close the stream."

	| bytes |
	byteStream closed 
		ifFalse: [
			bytes := byteStream position.
			byteStream close]
		ifTrue: [bytes := 'unknown'].
	^ bytes! !

!DataStream methodsFor: 'other' stamp: 'nk 3/12/2004 21:56'!
contents
	^byteStream contents! !

!DataStream methodsFor: 'other' stamp: 'yo 12/3/2004 17:14'!
errorWriteReference: anInteger
    "PRIVATE -- Raise an error because this case of nextPut:'s perform:
     shouldn't be called. -- 11/15/92 jhm"

    self error: 'This should never be called'! !

!DataStream methodsFor: 'other' stamp: ''!
flush
    "Guarantee that any writes to me are actually recorded on disk. -- 11/17/92 jhm"

    ^ byteStream flush! !

!DataStream methodsFor: 'write and read' stamp: ''!
getCurrentReference
    "PRIVATE -- Return the currentReference posn.
     Overridden by ReferenceStream."

    ^ 0! !

!DataStream methodsFor: 'write and read' stamp: 'nice 3/11/2012 16:30'!
insideASegment
	^ false! !

!DataStream methodsFor: 'write and read' stamp: 'tk 4/8/1999 13:11'!
maybeBeginReference: internalObject
	"Do nothing.  See ReferenceStream|maybeBeginReference:"

	^ internalObject! !

!DataStream methodsFor: 'write and read' stamp: 'ul 11/15/2010 11:38'!
next
	"Answer the next object in the stream."
	| type selector anObject isARefType pos internalObject |

	type := byteStream next.
	type ifNil: [pos := byteStream position.	"absolute!!!!"
		byteStream close.	"clean up"
		byteStream position = 0 
			ifTrue: [self error: 'The file did not exist in this directory'] 
			ifFalse: [self error: 'Unexpected end of object file'].
		pos.	"so can see it in debugger"
		^ nil].
	type = 0 ifTrue: [pos := byteStream position.	"absolute!!!!"
		byteStream close.	"clean up"
		self error: 'Expected start of object, but found 0'.
		^ nil].
	isARefType := self noteCurrentReference: type.
	selector := #(readNil readTrue readFalse readInteger	"<-4"
			readStringOld readSymbol readByteArray		"<-7"
			readArray readInstance readReference readBitmap	"<-11"
			readClass readUser readFloat readRectangle readShortInst 	"<-16"
			readString readWordArray readWordArrayForSegment 	"<-19"
			readWordLike readMethod "<-21") at: type ifAbsent: [
				pos := byteStream position.	"absolute!!!!"
				byteStream close. 
				self error: 'file is more recent than this system'. ^ nil].
	anObject := self perform: selector. "A method that recursively
		calls next (readArray, readInstance, objectAt:) must save &
		restore the current reference position."
	isARefType ifTrue: [self beginReference: anObject].

		"After reading the externalObject, internalize it.
		 #readReference is a special case. Either:
		   (1) We actually have to read the object, recursively calling
			   next, which internalizes the object.
		   (2) We just read a reference to an object already read and
			   thus already interalized.
		 Either way, we must not re-internalize the object here."
	selector == #readReference ifTrue: [^ anObject].
	internalObject := anObject comeFullyUpOnReload: self.
	internalObject == String ifTrue:[
		"This is a hack to figure out if we're loading a String class 
		that really should be a ByteString. Note that these days this
		will no longer be necessary since we use #withClassVersion:
		for constructing the global thus using a different classVersion
		will perfectly do the trick."
		((anObject isKindOf: DiskProxy) 
			and:[anObject globalObjectName == #String
			and:[anObject constructorSelector == #yourself]]) ifTrue:[
				internalObject := ByteString]].
	^ self maybeBeginReference: internalObject! !

!DataStream methodsFor: 'other' stamp: 'tk 3/5/2002 09:51'!
nextAndClose
	"Speedy way to grab one object.  Only use when we are inside an object binary file.  Do not use for the start of a SmartRefStream mixed code-and-object file."

	| obj |
	obj := self next.
	self close.
	^ obj! !

!DataStream methodsFor: 'write and read' stamp: ''!
nextPutAll: aCollection
    "Write each of the objects in aCollection to the
     receiver stream. Answer aCollection."

    ^ aCollection do: [:each | self nextPut: each]! !

!DataStream methodsFor: 'write and read' stamp: 'nice 10/31/2015 23:24'!
nextPut: anObject
	"Write anObject to the receiver stream. Answer anObject."
	| typeID selector objectToStore |

	typeID := self typeIDFor: anObject.
	(self tryToPutReference: anObject typeID: typeID)
		ifTrue: [^ anObject].

	objectToStore := (self objectIfBlocked: anObject) objectForDataStream: self.
	objectToStore == anObject
		ifFalse:
			[typeID := self typeIDFor: objectToStore.
			(self tryToPutReference: objectToStore typeID: typeID)
				ifTrue: [^ anObject]].

	byteStream nextPut: typeID.
	selector := #(writeNil: writeTrue: writeFalse: writeInteger: 
		writeStringOld: writeSymbol: writeByteArray:
		writeArray: writeInstance: errorWriteReference: writeBitmap:
		writeClass: writeUser: writeFloat: writeRectangle: == "<-16 short inst" 
		writeString: writeBitmap: writeBitmap: writeWordLike: 
		writeInstance: "CompiledMethod") at: typeID.
	self perform: selector with: objectToStore.

	^ anObject


"NOTE: If anObject is a reference type (one that we write cross-references to) but its externalized form (result of objectForDataStream:) isn't (e.g. CompiledMethod and ViewState), then we should remember its externalized form
 but not add to 'references'. Putting that object again should just put its
 external form again. That's more compact and avoids seeks when reading.
 But we just do the simple thing here, allowing backward-references for
 non-reference types like nil. So objectAt: has to compensate. Objects that
 externalize nicely won't contain the likes of ViewStates, so this shouldn't
 hurt much.
	 writeReference: -> errorWriteReference:."! !

!DataStream methodsFor: 'other' stamp: ''!
next: anInteger
    "Answer an Array of the next anInteger objects in the stream."
    | array |

    array := Array new: anInteger.
    1 to: anInteger do: [:i |
        array at: i put: self next].
    ^ array! !

!DataStream methodsFor: 'write and read' stamp: 'yo 12/3/2004 17:25'!
noteCurrentReference: typeID
    "PRIVATE -- If we support references for type typeID, remember
     the current byteStream position so we can add the next object to
     the 'objects' dictionary, and return true. Else return false.
     This method is here to be overridden by ReferenceStream"

    ^ false! !

!DataStream methodsFor: 'write and read' stamp: 'tk 6/9/1997 08:18 '!
objectAt: anInteger
	"PRIVATE -- Read & return the object at a given stream position.  anInteger is a relative file position. "
	| savedPosn anObject refPosn |

	savedPosn := byteStream position.	"absolute"
	refPosn := self getCurrentReference.	"relative position"

	byteStream position: anInteger + basePos.	"was relative"
	anObject := self next.

	self setCurrentReference: refPosn.	"relative position"
	byteStream position: savedPosn.		"absolute"
	^ anObject! !

!DataStream methodsFor: 'write and read' stamp: 'tk 3/13/98 22:16'!
objectIfBlocked: anObject
	"We don't do any blocking"

	^ anObject! !

!DataStream methodsFor: 'write and read' stamp: '6/9/97 08:46 tk'!
outputReference: referencePosn
	"PRIVATE -- Output a reference to the object at integer stream position referencePosn (relative to basePos). To output a weak reference to an object not yet written, supply (self vacantRef) for referencePosn."

	byteStream nextPut: 10. "reference typeID"
	byteStream nextNumber: 4 put: referencePosn	"relative position"! !

!DataStream methodsFor: 'other' stamp: 'ar 2/24/2001 22:45'!
project
	^nil! !

!DataStream methodsFor: 'write and read' stamp: '6/9/97 08:32 tk'!
readArray
	"PRIVATE -- Read the contents of an Array.
	 We must do beginReference: here after instantiating the Array
	 but before reading its contents, in case the contents reference
	 the Array. beginReference: will be sent again when we return to
	 next, but that's ok as long as we save and restore the current
	 reference position over recursive calls to next."
	| count array refPosn |

	count := byteStream nextNumber: 4.

	refPosn := self beginReference: (array := Array new: count).		"relative pos"
	1 to: count do: [:i |
		array at: i put: self next].
	self setCurrentReference: refPosn.		"relative pos"
	^ array! !

!DataStream methodsFor: 'write and read' stamp: ''!
readBitmap
	"PRIVATE -- Read the contents of a Bitmap."

	^ Bitmap newFromStream: byteStream
	"Note that the reader knows that the size is in long words, but the data is in bytes."! !

!DataStream methodsFor: 'write and read' stamp: ''!
readBoolean
	"PRIVATE -- Read the contents of a Boolean.
	 This is here only for compatibility with old data files."

	^ byteStream next ~= 0! !

!DataStream methodsFor: 'write and read' stamp: 'jm 8/19/1998 17:00'!
readByteArray
	"PRIVATE -- Read the contents of a ByteArray."

	| count |
	count := byteStream nextNumber: 4.
	^ byteStream next: count  "assume stream is in binary mode"
! !

!DataStream methodsFor: 'write and read' stamp: 'topa 6/25/2015 09:17'!
readCharacterOfSize: instSize
	| refPosn val |
	self assert: instSize = 1.
	refPosn := self getCurrentReference.
	self setCurrentReference: refPosn.
	val := self next.
	^ Character value: val.
! !

!DataStream methodsFor: 'write and read' stamp: 'tk 3/24/98 10:29'!
readClass
	"Should never be executed because a DiskProxy, not a clas comes in."

	^ self error: 'Classes should be filed in'! !

!DataStream methodsFor: 'write and read' stamp: ''!
readFalse
    "PRIVATE -- Read the contents of a False."

    ^ false! !

!DataStream methodsFor: 'write and read' stamp: ''!
readFloat
	"PRIVATE -- Read the contents of a Float.
	 This is the fast way to read a Float.
	 We support 8-byte Floats here.  Non-IEEE"

	| new |
	new := Float new: 2.		"To get an instance"
	new at: 1 put: (byteStream nextNumber: 4).
	new at: 2 put: (byteStream nextNumber: 4).
	^ new! !

!DataStream methodsFor: 'write and read' stamp: 'yo 12/3/2004 17:12'!
readFloatString
	"PRIVATE -- Read the contents of a Float string.
	 This is the slow way to read a Float--via its string rep'n.
	 It's here for compatibility with old data files."

	^ Float readFrom: (byteStream next: (byteStream nextNumber: 4))! !

!DataStream methodsFor: 'write and read' stamp: 'topa 6/25/2015 09:17'!
readInstance
	"PRIVATE -- Read the contents of an arbitrary instance.
	 ASSUMES: readDataFrom:size: sends me beginReference: after it
	   instantiates the new object but before reading nested objects.
	 NOTE: We must restore the current reference position after
	   recursive calls to next.
	Let the instance, not the class read the data.  "
	| instSize aSymbol refPosn anObject newClass |

	instSize := (byteStream nextNumber: 4) - 1.
	refPosn := self getCurrentReference.
	aSymbol := self next.
	aSymbol == #Character
		ifTrue: [^ self readCharacterOfSize:  instSize].
	newClass := Smalltalk at: aSymbol asSymbol.
	anObject := newClass isVariable 	"Create object here"
			ifFalse: [newClass basicNew]
			ifTrue: [newClass basicNew: instSize - (newClass instSize)].
	self setCurrentReference: refPosn.  "before readDataFrom:size:"
	anObject := anObject readDataFrom: self size: instSize.
	self setCurrentReference: refPosn.  "before returning to next"
	^ anObject! !

!DataStream methodsFor: 'write and read' stamp: ''!
readInteger
    "PRIVATE -- Read the contents of a SmallInteger."

    ^ byteStream nextInt32	"signed!!!!!!"! !

!DataStream methodsFor: 'write and read' stamp: 'eem 7/5/2016 17:45'!
readMethod
	"PRIVATE -- Read the contents of an arbitrary instance.
	 ASSUMES: readDataFrom:size: sends me beginReference: after it
	   instantiates the new object but before reading nested objects.
	 NOTE: We must restore the current reference position after
	   recursive calls to next.
	Let the instance, not the class read the data.  "
	| instSize refPosn newClass className xxHeader nLits byteCodeSizePlusTrailer newMethod lits |

	instSize := (byteStream nextNumber: 4) - 1.
	refPosn := self getCurrentReference.
	className := self next.
	newClass := Smalltalk at: className asSymbol.

	xxHeader := self next.
		"nArgs := (xxHeader >> 24) bitAnd: 16rF."
		"nTemps := (xxHeader >> 18) bitAnd: 16r3F."
		"largeBit := (xxHeader >> 17) bitAnd: 1."
	nLits := xxHeader bitAnd: 16r7FFF.
	byteCodeSizePlusTrailer := instSize - (newClass instSize "0") - (nLits + 1 * Smalltalk wordSize).

	newMethod := newClass 
						newMethod: byteCodeSizePlusTrailer
						header: xxHeader.

	self setCurrentReference: refPosn.  "before readDataFrom:size:"
	self beginReference: newMethod.
	lits := newMethod numLiterals + 1.	"counting header"
	2 to: lits do:
		[:ii | newMethod objectAt: ii put: self next].
	lits*Smalltalk wordSize+1 to: newMethod basicSize do:
		[:ii | newMethod basicAt: ii put: byteStream next].
			"Get raw bytes directly from the file"
	self setCurrentReference: refPosn.  "before returning to next"
	^newMethod! !

!DataStream methodsFor: 'write and read' stamp: ''!
readNil
    "PRIVATE -- Read the contents of an UndefinedObject."

    ^ nil! !

!DataStream methodsFor: 'write and read' stamp: ' 6/9/97'!
readRectangle
    "Read a compact Rectangle.  Rectangles with values outside +/- 2047 were stored as normal objects (type=9).  They will not come here.  17:22 tk"

	"Encoding is four 12-bit signed numbers.  48 bits in next 6 bytes.  17:24 tk"
	| acc left top right bottom |
	acc := byteStream nextNumber: 3.
	left := acc bitShift: -12.
	(left bitAnd: 16r800) ~= 0 ifTrue: [left := left - 16r1000].	"sign"
	top := acc bitAnd: 16rFFF.
	(top bitAnd: 16r800) ~= 0 ifTrue: [top := top - 16r1000].	"sign"

	acc := byteStream nextNumber: 3.
	right := acc bitShift: -12.
	(right bitAnd: 16r800) ~= 0 ifTrue: [right := right - 16r1000].	"sign"
	bottom := acc bitAnd: 16rFFF.
	(bottom bitAnd: 16r800) ~= 0 ifTrue: [bottom := bottom - 16r1000].	"sign"
	
    ^ Rectangle left: left right: right top: top bottom: bottom
! !

!DataStream methodsFor: 'write and read' stamp: 'tk 1/5/2000 11:47'!
readReference
	"Read the contents of an object reference. (Cf. outputReference:)  File is not now positioned at this object."
	| referencePosition |

	^ (referencePosition := (byteStream nextNumber: 4)) = self vacantRef	"relative"
		ifTrue:  [nil]
		ifFalse: [self objectAt: referencePosition]		"relative pos"! !

!DataStream methodsFor: 'write and read' stamp: 'tk 1/8/97'!
readShortInst
	"Read the contents of an arbitrary instance that has a short header.
	 ASSUMES: readDataFrom:size: sends me beginReference: after it
	   instantiates the new object but before reading nested objects.
	 NOTE: We must restore the current reference position after
	   recursive calls to next.
	Let the instance, not the class read the data.  "
	| instSize aSymbol refPosn anObject newClass |

	instSize := (byteStream next) - 1.	"one byte of size"
	refPosn := self getCurrentReference.
	aSymbol := self readShortRef.	"class symbol in two bytes of file pos"
	newClass := Smalltalk at: aSymbol asSymbol.
	anObject := newClass isVariable 	"Create object here"
			ifFalse: [newClass basicNew]
			ifTrue: [newClass basicNew: instSize - (newClass instSize)].
	self setCurrentReference: refPosn.  "before readDataFrom:size:"
	anObject := anObject readDataFrom: self size: instSize.
	self setCurrentReference: refPosn.  "before returning to next"
	^ anObject! !

!DataStream methodsFor: 'write and read' stamp: 'tk 7/12/1998 13:32'!
readShortRef
	"Read an object reference from two bytes only.  Original object must be in first 65536 bytes of the file.  Relative to start of data.  vacantRef not a possibility."

	^ self objectAt: (byteStream nextNumber: 2)! !

!DataStream methodsFor: 'write and read' stamp: 'nice 7/18/2014 00:42'!
readString
	^byteStream nextString
! !

!DataStream methodsFor: 'write and read' stamp: 'tk 6/8/1998 21:27'!
readStringOld

   ^ byteStream nextStringOld! !

!DataStream methodsFor: 'write and read' stamp: ''!
readSymbol
    "PRIVATE -- Read the contents of a Symbol."

    ^ self readString asSymbol! !

!DataStream methodsFor: 'write and read' stamp: ''!
readTrue
    "PRIVATE -- Read the contents of a True."

    ^ true! !

!DataStream methodsFor: 'write and read' stamp: 'tk 3/4/1999 22:58'!
readUser
	"Reconstruct both the private class and the instance.  Still used??"

	^ self readInstance.		"Will create new unique class"
! !

!DataStream methodsFor: 'write and read' stamp: 'tk 1/24/2000 23:20'!
readWordArray
	"PRIVATE -- Read the contents of a WordArray."

	^ WordArray newFromStream: byteStream
	"Size is number of long words."! !

!DataStream methodsFor: 'write and read' stamp: 'tk 1/24/2000 23:23'!
readWordArrayForSegment
	"Read the contents of a WordArray ignoring endianness."

	^ WordArrayForSegment newFromStream: byteStream
	"Size is number of long words."! !

!DataStream methodsFor: 'write and read' stamp: 'tk 2/3/2000 21:11'!
readWordLike
	| refPosn aSymbol newClass anObject |
	"Can be used by any class that is bits and not bytes (WordArray, Bitmap, SoundBuffer, etc)."

	refPosn := self getCurrentReference.
	aSymbol := self next.
	newClass := Smalltalk at: aSymbol asSymbol.
	anObject := newClass newFromStream: byteStream.
	"Size is number of long words."
	self setCurrentReference: refPosn.  "before returning to next"
	^ anObject
! !

!DataStream methodsFor: 'write and read' stamp: 'tk 9/24/2000 15:39'!
replace: original with: proxy
	"We may wish to remember that in some field, the original object is being replaced by the proxy.  For the hybred scheme that collects with a DummyStream and writes an ImageSegment, it needs to hold onto the originals so they will appear in outPointers, and be replaced."

	"do nothing"! !

!DataStream methodsFor: 'other' stamp: ''!
reset
    "Reset the stream."

    byteStream reset! !

!DataStream methodsFor: 'other' stamp: 'tk 5/29/97'!
rootObject
	"Return the object at the root of the tree we are filing out.  "

	^ topCall! !

!DataStream methodsFor: 'other' stamp: 'tk 5/29/97'!
rootObject: anObject
	"Return the object at the root of the tree we are filing out.  "

	topCall := anObject! !

!DataStream methodsFor: 'write and read' stamp: ''!
setCurrentReference: refPosn
    "PRIVATE -- Set currentReference to refPosn.
     Noop here. Cf. ReferenceStream."! !

!DataStream methodsFor: 'other' stamp: '6/9/97 08:03 di'!
setStream: aStream
	"PRIVATE -- Initialization method."

	aStream binary.
	basePos := aStream position.	"Remember where we start.  Earlier part of file contains a class or method file-in.  Allow that to be edited.  We don't deal in absolute file locations."
	byteStream := aStream.! !

!DataStream methodsFor: 'other' stamp: 'tk 8/18/1998 08:59'!
setStream: aStream reading: isReading
	"PRIVATE -- Initialization method."

	aStream binary.
	basePos := aStream position.	"Remember where we start.  Earlier part of file contains a class or method file-in.  Allow that to be edited.  We don't deal in absolute file locations."
	byteStream := aStream.! !

!DataStream methodsFor: 'other' stamp: ''!
size
    "Answer the stream's size."

    ^ byteStream size! !

!DataStream methodsFor: 'write and read' stamp: 'yo 12/3/2004 16:59'!
tryToPutReference: anObject typeID: typeID
    "PRIVATE -- If we support references for type typeID, and if
       anObject already appears in my output stream, then put a
       reference to the place where anObject already appears. If we
       support references for typeID but didn't already put anObject,
       then associate the current stream position with anObject in
       case one wants to nextPut: it again.
     Return true after putting a reference; false if the object still
       needs to be put.
     For DataStream this is trivial. ReferenceStream overrides this."

    ^ false! !

!DataStream methodsFor: 'write and read' stamp: 'topa 6/25/2015 10:22'!
typeIDFor: anObject
	"Return the typeID for anObject's class.  This is where the tangle of objects is clipped to stop everything from going out.  
	Classes can control their instance variables by defining objectToStoreOnDataStream.
	Any object in blockers is not written out.  See ReferenceStream.objectIfBlocked: and DataStream nextPut:.
	Morphs do not write their owners.  See Morph.storeDataOn:   Each morph tells itself to 'prepareToBeSaved' before writing out."


	anObject isFloat ifTrue: [^14]. "shortcut floats of any representation."
	^ TypeMap at: anObject class ifAbsent: [9 "instance of any normal class"]	
"See DataStream initialize.  nil=1. true=2. false=3. a SmallInteger=4. (a String was 5). a Symbol=6.  a ByteArray=7. an Array=8. other = 9.  a Bitmap=11. a Metaclass=12. a Float=14.  a Rectangle=15. any instance that can have a short header=16.  a String=17 (new format). a WordArray=18."! !

!DataStream methodsFor: 'other' stamp: 'tk 7/12/1998 13:16'!
vacantRef
	"Answer the magic 32-bit constant we use ***ON DISK*** as a stream 'reference
	 position' to identify a reference that's not yet filled in. This must be a
	 value that won't be used as an ordinary reference. Cf. outputReference: and
	 readReference. -- 
	 NOTE: We could use a different type ID for vacant-refs rather than writing
		object-references with a magic value. (The type ID and value are
		overwritten by ordinary object-references when weak refs are fullfilled.)"

	^ SmallInteger maxVal! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeArray: anArray
	"PRIVATE -- Write the contents of an Array."

	byteStream nextNumber: 4 put: anArray size.
	self nextPutAll: anArray.! !

!DataStream methodsFor: 'write and read' stamp: 'tk 6/8/1998 21:07'!
writeBitmap: aBitmap
	"PRIVATE -- Write the contents of a Bitmap."

	aBitmap writeOn: byteStream
	"Note that this calls (byteStream nextPutAll: aBitmap) which knows enough to put 4-byte quantities on the stream!!  Reader must know that size is in long words."! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeBoolean: aBoolean
    "PRIVATE -- Write the contents of a Boolean.
     This method is now obsolete."

    byteStream nextPut: (aBoolean ifTrue: [1] ifFalse: [0])! !

!DataStream methodsFor: 'write and read' stamp: 'tk 6/8/1998 21:06'!
writeByteArray: aByteArray
	"PRIVATE -- Write the contents of a ByteArray."

	byteStream nextNumber: 4 put: aByteArray size.
	"May have to convert types here..."
	byteStream nextPutAll: aByteArray.! !

!DataStream methodsFor: 'write and read' stamp: 'tk 3/24/98 10:27'!
writeClass: aClass
	"Write out a DiskProxy for the class.  It will look up the class's name in Smalltalk in the new sustem.  Never write classes or methodDictionaries as objects.  For novel classes, front part of file is a fileIn of the new class."

	"This method never executed because objectToStoreOnDataStream returns a DiskProxy.  See DataStream.nextPut:"
    ^ self error: 'Write a DiskProxy instead'! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeFalse: aFalse
    "PRIVATE -- Write the contents of a False."! !

!DataStream methodsFor: 'write and read' stamp: 'yo 12/3/2004 17:07'!
writeFloatString: aFloat
    "PRIVATE -- Write the contents of a Float string.
     This is the slow way to write a Float--via its string rep'n."

    self writeByteArray: (aFloat printString)! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeFloat: aFloat
	"PRIVATE -- Write the contents of a Float.
	  We support 8-byte Floats here."

	byteStream nextNumber: 4 put: (aFloat at: 1).
	byteStream nextNumber: 4 put: (aFloat at: 2).
! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeInstance: anObject
    "PRIVATE -- Write the contents of an arbitrary instance."

    ^ anObject storeDataOn: self! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeInteger: anInteger
	"PRIVATE -- Write the contents of a SmallInteger."

	byteStream nextInt32Put: anInteger	"signed!!!!!!!!!!"! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeNil: anUndefinedObject
    "PRIVATE -- Write the contents of an UndefinedObject."! !

!DataStream methodsFor: 'write and read' stamp: 'jm 7/31/97 16:16'!
writeRectangle: anObject
    "Write the contents of a Rectangle.  See if it can be a compact Rectangle (type=15).  Rectangles with values outside +/- 2047 were stored as normal objects (type=9).  17:22 tk"

	| ok right bottom top left acc |
	ok := true.
	(right := anObject right) > 2047 ifTrue: [ok := false].
	right < -2048 ifTrue: [ok := false].
	(bottom := anObject bottom) > 2047 ifTrue: [ok := false].
	bottom < -2048 ifTrue: [ok := false].
	(top := anObject top) > 2047 ifTrue: [ok := false].
	top < -2048 ifTrue: [ok := false].
	(left := anObject left) > 2047 ifTrue: [ok := false].
	left < -2048 ifTrue: [ok := false].
	ok := ok & left isInteger & right isInteger & top isInteger & bottom isInteger.

	ok ifFalse: [
		byteStream skip: -1; nextPut: 9; skip: 0. "rewrite type to be normal instance"
	    ^ anObject storeDataOn: self].

	acc := ((left bitAnd: 16rFFF) bitShift: 12) + (top bitAnd: 16rFFF).
	byteStream nextNumber: 3 put: acc.
	acc := ((right bitAnd: 16rFFF) bitShift: 12) + (bottom bitAnd: 16rFFF).
	byteStream nextNumber: 3 put: acc.! !

!DataStream methodsFor: 'write and read' stamp: 'tk 6/8/1998 21:23'!
writeStringOld: aString
	"PRIVATE -- Write the contents of a String."

	| length |
	aString size < 16384 
		ifTrue: [
			(length := aString size) < 192
				ifTrue: [byteStream nextPut: length]
				ifFalse: 
					[byteStream nextPut: (length // 256 + 192).
					byteStream nextPut: (length \\ 256)].
			aString do: [:char | byteStream nextPut: char asciiValue]]
		ifFalse: [self writeByteArray: aString].	"takes more space"! !

!DataStream methodsFor: 'write and read' stamp: 'tk 6/8/1998 20:57'!
writeString: aString
	"PRIVATE -- Write the contents of a String."

	byteStream nextStringPut: aString.! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeSymbol: aSymbol
    "PRIVATE -- Write the contents of a Symbol."

    self writeString: aSymbol! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeTrue: aTrue
    "PRIVATE -- Write the contents of a True."! !

!DataStream methodsFor: 'write and read' stamp: ''!
writeUser: anObject
    "Write the contents of an arbitrary User instance (and its devoted class)."
    " 7/29/96 tk"

	"If anObject is an instance of a unique user class, will lie and say it has a generic class"
    ^ anObject storeDataOn: self! !

!DataStream methodsFor: 'write and read' stamp: 'tk 2/5/2000 21:53'!
writeWordLike: aWordArray
	"Note that we put the class name before the size."

	self nextPut: aWordArray class name.
	aWordArray writeOn: byteStream
	"Note that this calls (byteStream nextPutAll: aBitmap) which knows enough to put 4-byte quantities on the stream!!  Reader must know that size is in long words or double-bytes."! !

DataStream subclass: #ReferenceStream
	instanceVariableNames: 'references objects currentReference fwdRefEnds blockers skipping insideASegment'
	classVariableNames: 'RefTypes'
	poolDictionaries: ''
	category: 'System-Object Storage'!

!ReferenceStream commentStamp: 'nice 3/25/2010 23:03' prior: 0!
This is a way of serializing a tree of objects into disk file. A ReferenceStream can store
one or more objects in a persistent form, including sharing and cycles.

Here is the way to use DataStream and ReferenceStream:
	rr := ReferenceStream fileNamed: 'test.obj'.
	rr nextPut: <your object>.
	rr close.

To get it back:
	rr := ReferenceStream fileNamed: 'test.obj'.
	<your object> := rr next.
	rr close.

ReferenceStreams can now write "weak" references. nextPutWeak:
writes a "weak" reference to an object, which refers to that object
*if* it also gets written to the stream by a normal nextPut:.

A ReferenceStream should be treated as a read-stream *or* as a write-stream, *not* as a read/write-stream. The reference-remembering mechanism would probably do bad things if you tried to read and write from the same ReferenceStream.

[TBD] Should we override "close" to do (self forgetReferences)?

Instance variables
 references -- an IdentityDictionary mapping objects already written
	to their byteStream positions. If asked to write any object a
	second time, we just write a reference to its stream position.
	This handles shared objects and reference cycles between objects.
	To implement "weak references" (for Aliases), the references
	dictionary also maps objects not (yet?) written to a Collection
	of byteStream positions with hopeful weak-references to it. If
	asked to definitely write one of these objects, we'll fixup those
	weak references.
 objects -- an IdentityDictionary mapping relative byte stream positions to
	objects already read in. If asked to follow a reference, we
	return the object already read.
	This handles shared objects and reference cycles between objects.
 currentReference -- the current reference position. Positon relative to the 
	start of object data in this file.  (Allows user to cut and paste smalltalk 
	code from the front of the file without effecting the reference values.)  
	This variable is used to help install each new object in "objects" as soon
	as it's created, **before** we start reading its contents, in
	case any of its content objects reference it.
 fwdRefEnds -- A weak reference can be a forward reference, which
	requires advance-reading the referrent. When we later come to the
	object, we must get its value from "objects" and not re-read it so
	refs to it don't become refs to copies. fwdRefEnds remembers the
	ending byte stream position of advance-read objects.
 skipping -- true if <what?>
 insideASegment -- true if we are being used to collect objects that will be 
	included in an ImageSegment.  If so, UniClasses must be noted and traced.

If the object is referenced before it is done being created, it might get created twice.  Just store the object the moment it is created in the 'objects' dictionary.  If at the end, comeFullyUpOnReload returns a different object, some refs will have the temporary object (this is an unlikely case).  At the moment, no implementor of comeFullyUpOnReload returns a different object except DiskProxy, and that is OK.
!

!ReferenceStream class methodsFor: 'examples' stamp: ''!
example2
"Here is the way to use DataStream and ReferenceStream:
	rr := ReferenceStream fileNamed: ''test.obj''.
	rr nextPut: <your object>.
	rr close.

To get it back:
	rr := ReferenceStream fileNamed: ''test.obj''.
	<your object> := rr next.
	rr close.
"
"An example and test of DataStream/ReferenceStream.
	 11/19/92 jhm: Use self testWith:."
	"ReferenceStream example2"
	| input sharedPoint |

	"Construct the test data."
	input := Array new: 9.
	input at: 1 put: nil.
	input at: 2 put: true.
	input at: 3 put: false.
	input at: 4 put: #(-4 -4.0 'four' four).
	input at: 5 put: (Form extent: 63 @ 50 depth: 8).
		(input at: 5) fillWithColor: Color lightOrange.
	input at: 6 put: 1024 @ -2048.
	input at: 7 put: input. "a cycle"
	input at: 8 put: (Array with: (sharedPoint := 0 @ -30000)).
	input at: 9 put: sharedPoint.

	"Write it out, read it back, and return it for inspection."
	^ self testWith: input
! !

!ReferenceStream class methodsFor: 'instance creation' stamp: 'tk 4/19/2001 16:50'!
on: aStream
	"Open a new ReferenceStream on a place to put the raw data."

	aStream class == ReadWriteStream ifTrue: [
		self inform: 'Please consider using a RWBinaryOrTextStream 
instead of a ReadWriteStream'].

	^ super on: aStream
! !

!ReferenceStream class methodsFor: 'accessing' stamp: ''!
refTypes: oc
	RefTypes := oc! !

!ReferenceStream class methodsFor: 'accessing' stamp: 'tk 5/26/1998 14:51'!
versionCode
    "Answer a number representing the 'version' of the ReferenceStream facility; this is stashed at the beginning of ReferenceStreams, as a secondary versioning mechanism (the primary one is the fileTypeCode).   At present, it serves for information only, and is not checked for compatibility at reload time, but could in future be used to branch to variant code. "

	" 1 = original version 1992"
	" 2 = HyperSqueak.  PathFromHome used for Objs outside the tree.  SqueakSupport SysLibrary for shared globals like Display and StrikeFonts.  File has version number, class structure, then an IncomingObjects manager.  8/16/96 tk.  
	Extended to SmartRefStream.  class structure also keeps superclasse chain.  Does analysis on structure to see when translation methods are needed.  Embedable in file-ins.  (factored out HyperSqueak support)  Feb-May 97 tk"
	" 3 = Reference objects are byte offsets relative to the start of the object portion of the file.  Rectangles with values -2048 to 2047 are encoded compactly."
	" 4 = If UniClasses (subclasses of Player) have class instance variables, append their values in the form (#Class43 (val1 val2 vla3)).  An array of those.  Can still read version 3."
	^ 4! !

!ReferenceStream methodsFor: 'writing' stamp: 'nice 7/28/2014 22:09'!
beginInstance: aClass size: anInteger
	"This is for use by storeDataOn: methods.  Cf. Object>>storeDataOn:."
	"Addition of 1 seems to make extra work, since readInstance has to compensate.  Here for historical reasons dating back to Kent Beck's original implementation in late 1988.
	In ReferenceStream, class is just 5 bytes for shared symbol.
	SmartRefStream puts out the names and number of class's instances variables for checking.
6/10/97 16:09 tk: See if we can put on a short header. Type = 16. "

	| short ref |
	short := true.	"All tests for object header that can be written in 4 bytes"
	anInteger <= 254 ifFalse: [short := false].	"one byte size"
	ref := references at: aClass name ifAbsent: [short := false. nil].
	ref isInteger ifFalse: [short := false].
	short ifTrue: [short := (ref < 65536) & (ref > 0) "& (ref ~= self vacantRef)"].  "vacantRef is big"
	short ifTrue: [short := byteStream peekLast = 9].	"ugly workaround"
	short 
		ifTrue: ["passed all the tests!!"
			byteStream skip: -1; nextPut: 16; "type = short header"
				nextPut: anInteger + 1;	"size is short"
				nextNumber: 2 put: ref]
		ifFalse: [
			"default to normal longer object header"
			byteStream nextNumber: 4 put: anInteger + 1.
			self nextPut: aClass name].
	insideASegment ifTrue: [
		aClass isSystemDefined ifFalse: [self nextPut: aClass]].
			"just record it to put it into roots"! !

!ReferenceStream methodsFor: 'reading' stamp: 'RAA 1/18/2001 11:52'!
beginReference: anObject
	"Remember anObject as the object we read at the position recorded by
	 noteCurrentReference:. This must be done after instantiating anObject but
	 before reading any of its contents that might (directly or indirectly) refer to
	 it. (It's ok to do this redundantly, which is convenient for #next.)
	 Answer the reference position."

	objects at: currentReference ifAbsent: [
		objects at: currentReference put: anObject.
		^ currentReference].
	(skipping includes: currentReference) ifFalse: [
		"If reading just to skip it, don't record this copy."
		objects at: currentReference put: anObject
	].
	^ currentReference		"position relative to start of data portion of file"! !

!ReferenceStream methodsFor: 'writing' stamp: 'tk 3/15/98 19:54'!
blockers

	^ blockers! !

!ReferenceStream methodsFor: 'writing' stamp: 'tk 3/13/98 20:00'!
blockers: anIdentDict
	"maps objects -> nil if they should not be written.  object -> anotherObject if they need substitution."

	anIdentDict class == IdentityDictionary ifFalse: [self error: 'must be IdentityDictionary'].
	blockers := anIdentDict! !

!ReferenceStream methodsFor: 'reading' stamp: '6/9/97 08:26 tk'!
getCurrentReference
    "PRIVATE -- Return the currentReference posn.  Always a relative position.  So user can cut and paste the Smalltalk source code at the beginning of the file."

    ^ currentReference	"relative position"! !

!ReferenceStream methodsFor: 'writing' stamp: 'tk 9/27/2000 11:37'!
insideASegment
	^ insideASegment! !

!ReferenceStream methodsFor: 'writing' stamp: 'tk 9/27/2000 11:36'!
insideASegment: aBoolean
	insideASegment := aBoolean! !

!ReferenceStream methodsFor: 'writing' stamp: 'ul 11/15/2010 12:00'!
isAReferenceType: typeID
	"Return true iff typeID is one of the classes that can be written as a reference to an instance elsewhere in the stream."

	"too bad we can't put Booleans in an Array literal"
	^ (RefTypes at: typeID) = 1
		"NOTE: If you get a bounds error here, the file probably has bad bits in it.  The most common cause is a file unpacking program that puts linefeeds after carriage returns."! !

!ReferenceStream methodsFor: 'reading' stamp: 'tk 4/8/1999 13:11'!
maybeBeginReference: internalObject
	"See if need to record a reference.  In case in the file twice"

	(self isAReferenceType: (self typeIDFor: internalObject))
			ifTrue: [self beginReference: internalObject].
			"save the final object and give it out next time."
	^ internalObject! !

!ReferenceStream methodsFor: 'reading' stamp: 'RAA 1/18/2001 16:46'!
next
	"Answer the next object in the stream.  If this object was already read, don't re-read it.  File is positioned just before the object."
	| curPosn skipToPosn haveIt theObject wasSkipping |

	haveIt := true.
	curPosn := byteStream position - basePos.
	theObject := objects at: curPosn ifAbsent: [haveIt := false].
		"probe in objects is done twice when coming from objectAt:.  This is OK."
	skipToPosn := fwdRefEnds at: curPosn ifAbsent: [nil].
	haveIt ifFalse: [ ^ super next].

	skipToPosn ifNotNil: [
		"Skip over the object and return the already-read-in value."
		byteStream position: skipToPosn + basePos		"make absolute"
	] ifNil: [
		"File is not positioned correctly.  Read object and throw it away."
		wasSkipping := skipping includes: curPosn.
		skipping add: curPosn.
		"fake :=" super next.
		wasSkipping ifFalse: [skipping remove: curPosn ifAbsent: []].
	].
	^ theObject
		! !

!ReferenceStream methodsFor: 'writing' stamp: 'nice 1/2/2013 23:22'!
nextPutWeak: anObject
    "Write a weak reference to anObject to the receiver stream. Answer anObject.
     If anObject is not a reference type of object, then just put it normally.
     A 'weak' reference means: If anObject gets written this stream via nextPut:,
     then its weak references will become normal references. Otherwise they'll
     read back as nil. -- "
    | typeID referencePosn |

    "Is it a reference type of object? If not, just write it normally."
    typeID := self typeIDFor: anObject.
    (self isAReferenceType: typeID) ifFalse: [^ self nextPut: anObject].

    "Have we heard of and maybe even written anObject before?"
    referencePosn := references at: anObject ifAbsentPut: [OrderedCollection new].

    "If referencePosn is an Integer, it's the stream position of anObject.
     Else it's a collection of hopeful weak-references to anObject."
    referencePosn isInteger ifFalse:
        [referencePosn add: byteStream position - basePos.		"relative"
        referencePosn := self vacantRef].
    self outputReference: referencePosn.		"relative"

    ^ anObject! !

!ReferenceStream methodsFor: 'reading' stamp: '6/9/97 09:00 tk'!
noteCurrentReference: typeID
	"PRIVATE -- If we support references for type typeID, remember
	 the current byteStream position so beginReference: can add the
	 next object to the 'objects' dictionary of reference positions,
	 then return true. Else return false."
	| answer |

	(answer := self isAReferenceType: typeID)
		ifTrue: [self setCurrentReference: (byteStream position - 1) - basePos "relative"
				"subtract 1 because we already read the object's type ID byte"].
	^ answer! !

!ReferenceStream methodsFor: 'reading' stamp: 'nice 12/26/2009 20:43'!
objectAt: anInteger
    "PRIVATE -- Read & return the object at a given stream position.
     If we already read it, just get it from the objects dictionary.
     (Reading it again wouldn't work with cycles or sharing.)
     If not, go read it and put it in the objects dictionary.
     NOTE: This resolves a cross-reference in the ReferenceStream:
       1. A backward reference to an object already read (the normal case).
       2. A forward reference which is a sated weak reference (we record where
          the object ends so when we get to it normally we can fetch it from
          'objects' and skip over it).
       3. A backward reference to a 'non-reference type' per the long NOTE in
          nextPut: (we compensate here--seek back to re-read it and add the object
          to 'objects' to avoid seeking back to read it any more times).
       4. While reading a foward weak reference (case 2), we may recursively hit an
          ordinary backward reference to an object that we haven't yet read because
          we temporarily skipped ahead. Such a reference is forward in time so we
          treat it much like case 2.
     11/16-24/92 jhm: Handle forward refs. Cf. class comment and above NOTE.
	08:57 tk   anInteger is a relative position"
    
    ^ objects at: anInteger "relative position.  case 1: It's in 'objects'"
        ifAbsent:   "do like super objectAt:, but remember the fwd-ref-end position"
            [| savedPosn refPosn anObject |
		savedPosn := byteStream position.		"absolute"
            refPosn := self getCurrentReference.	"relative position"

            byteStream position: anInteger + basePos.	"was relative"
            anObject := self next.

            (self isAReferenceType: (self typeIDFor: anObject))
                ifTrue:  [fwdRefEnds at: anInteger put: byteStream position - basePos] "cases 2, 4"
                ifFalse: [objects at: anInteger put: anObject]. "case 3"

            self setCurrentReference: refPosn.		"relative position"
            byteStream position: savedPosn.		"absolute"
            anObject]! !

!ReferenceStream methodsFor: 'writing' stamp: 'tk 9/24/2000 09:18'!
objectIfBlocked: anObject
	"See if this object is blocked -- not written out and another object substituted."

	^ blockers at: anObject ifAbsent: [anObject]! !

!ReferenceStream methodsFor: 'writing' stamp: 'nk 5/4/2004 14:40'!
project
	"Return the project we are writing or nil"

	(topCall respondsTo: #isCurrentProject) ifTrue: [^ topCall].
	(topCall respondsTo: #do:) ifTrue: [1 to: 5 do: [:ii | 
		((topCall at: ii ifAbsent: [ ^nil ]) respondsTo: #isCurrentProject) ifTrue: [^ topCall at: ii]]].
	^ nil! !

!ReferenceStream methodsFor: 'writing' stamp: 'tk 9/23/2000 08:40'!
projectChangeSet
	| pr |
	"The changeSet of the project we are writing"
	(pr := self project) ifNil: [^ nil].
	^ pr projectChangeSet! !

!ReferenceStream methodsFor: 'writing' stamp: ''!
references
	^ references! !

!ReferenceStream methodsFor: 'writing' stamp: 'tk 9/24/2000 16:44'!
replace: original with: proxy
	"We may wish to remember that in some field, the original object is being replaced by the proxy.  For the hybred scheme that collects with a DummyStream and writes an ImageSegment, it needs to hold onto the originals so they will appear in outPointers, and be replaced."

	blockers at: original put: proxy! !

!ReferenceStream methodsFor: 'writing' stamp: 'tk 8/18/1998 08:38'!
reset
	"PRIVATE -- Reset my internal state.
	   11/15-17/92 jhm: Added transients and fwdRefEnds.
	   7/11/93 sw: Give substantial initial sizes to avoid huge time spent growing.
	   9/3/93 sw: monster version for Sasha"

	super reset.
	references := IdentityDictionary new: 4096 * 5.
"	objects := IdentityDictionary new: 4096 * 5.
	fwdRefEnds := IdentityDictionary new.
"
	blockers ifNil: [blockers := IdentityDictionary new].
 ! !

!ReferenceStream methodsFor: 'writing' stamp: '6/9/97 08:24 tk'!
setCurrentReference: refPosn
    "PRIVATE -- Set currentReference to refPosn.  Always a relative position."

    currentReference := refPosn		"relative position"! !

!ReferenceStream methodsFor: 'writing' stamp: 'RAA 1/18/2001 11:51'!
setStream: aStream
	"PRIVATE -- Initialization method."

	super setStream: aStream.
	references := IdentityDictionary new: 4096 * 5.
	objects := IdentityDictionary new: 4096 * 5.
	fwdRefEnds := IdentityDictionary new.
	skipping := IdentitySet new.
	insideASegment := false.
	blockers ifNil: [blockers := IdentityDictionary new].	"keep blockers we just passed in"
! !

!ReferenceStream methodsFor: 'writing' stamp: 'RAA 1/18/2001 11:51'!
setStream: aStream reading: isReading
	"PRIVATE -- Initialization method."

	super setStream: aStream reading: isReading.
	"isReading ifFalse: [  when we are sure"
	references := IdentityDictionary new: 4096 * 5.
	isReading ifTrue: [
		objects := IdentityDictionary new: 4096 * 5.
		skipping := IdentitySet new.
		insideASegment := false.
		fwdRefEnds := IdentityDictionary new].
	blockers ifNil: [blockers := IdentityDictionary new].	"keep blockers we just passed in"
! !

!ReferenceStream methodsFor: 'statistics' stamp: 'nice 1/2/2013 23:23'!
statisticsOfRefs
	"Analyze the information in references, the objects being written out"

	| parents ownerBags tallies n nm owners |
	parents := IdentityDictionary new: references size * 2.
	n := 0.
	'Finding Owners...'
	displayProgressFrom: 0 to: references size
	during: [:bar |
	references keysDo:
		[:parent | | kids |
		bar value: (n := n+1).
		kids := parent class isFixed
			ifTrue: [(1 to: parent class instSize) collect: [:i | parent
instVarAt: i]]
			ifFalse: [parent class isBits ifTrue: [Array new]
					 ifFalse: [(1 to: parent basicSize) collect: [:i | parent basicAt:
i]]].
		(kids select: [:x | references includesKey: x])
			do: [:child | parents at: child put: parent]]].
	ownerBags := Dictionary new.
	tallies := Bag new.
	n := 0.
	'Tallying Owners...'
	displayProgressFrom: 0 to: references size
	during: [:bar |
	references keysDo:  "For each class of obj, tally a bag of owner
classes"
		[:obj | | objParent | bar value: (n := n+1).
		nm := obj class name.
		tallies add: nm.
		owners := ownerBags at: nm ifAbsentPut: [Bag new].
		(objParent := parents at: obj ifAbsent: [nil]) == nil
			ifFalse: [owners add: objParent class name]]].
	^ String streamContents:
		[:strm |  tallies sortedCounts do:
			[:assn | n := assn key.  nm := assn value.
			owners := ownerBags at: nm.
			strm cr; nextPutAll: nm; space; print: n.
			owners size > 0 ifTrue:
				[strm cr; tab; print: owners sortedCounts]]]! !

!ReferenceStream methodsFor: 'writing' stamp: 'tk 3/28/2000 22:19'!
tryToPutReference: anObject typeID: typeID
	"PRIVATE -- If we support references for type typeID, and if
	   anObject already appears in my output stream, then put a
	   reference to the place where anObject already appears. If we
	   support references for typeID but didn't already put anObject,
	   then associate the current stream position with anObject in
	   case one wants to nextPut: it again.
	 Return true after putting a reference; false if the object still
	   needs to be put.
	 : Added support for weak refs. Split out outputReference:.
	08:42 tk  references stores relative file positions."
	| referencePosn nextPosn |

	"Is it a reference type of object?"
	(self isAReferenceType: typeID) ifFalse: [^ false].

	"Have we heard of and maybe even written anObject before?"
	referencePosn := references at: anObject ifAbsent:
			["Nope. Remember it and let the sender write it."
			references at: anObject put: (byteStream position - basePos).	"relative"
			^ false].

	"If referencePosn is an Integer, it's the stream position of anObject."
	referencePosn isInteger ifTrue:
		[self outputReference: referencePosn.	"relative"
		^ true].

	referencePosn == #none ifTrue: ["for DiskProxy"
			references at: anObject put: (byteStream position - basePos).	"relative"
			^ false].


	"Else referencePosn is a collection of positions of weak-references to anObject.
	 Make them full references since we're about to really write anObject."
	references at: anObject put: (nextPosn := byteStream position) - basePos.	"store relative"
	referencePosn do: [:weakRefPosn |
			byteStream position: weakRefPosn + basePos.		"make absolute"
			self outputReference: nextPosn - basePos].	"make relative"
	byteStream position: nextPosn.		"absolute"
	^ false! !

ReferenceStream subclass: #SmartRefStream
	instanceVariableNames: 'structures steady reshaped renamed renamedConv superclasses progressBar objCount classInstVars'
	classVariableNames: 'ScannedObject'
	poolDictionaries: ''
	category: 'System-Object Storage'!

!SmartRefStream commentStamp: 'nice 3/25/2010 23:03' prior: 0!
Ordinary ReferenceStreams assume that the names and order of instance variables is exactly the same when an object file is written and read.  
	SmartRefStream allows object files to be read even after instance variables have changed or the entire class has been renamed.

When an object file is written, no one knows how the classes will change in the future.  Therefore, all conversion must be done when the file is read.  The key is to store enough information in the file about the names of the instance variables of all outgoing classes.  

SmartRefStream works best with only one tree of objects per file.  You can nextPut: more than once, but each object tree gets its own class structure description, which is big.  

Conversion of old objects is done by a method in each class called (convertToCurrentVersion: varDict refStream: smartRefStrm).  At fileOut time, ChangeSet>>checkForConversionMethods creates a prototype of this method (if Preference #conversionMethodsAtFileOut is true).  The programmer must edit this method to (1) test if the incoming object needs conversion, (2) put non-nil values into any new inst vars that need them, and (3) save the data of any inst vars that are being deleted. 

Determining which old version is represented by the incoming object can be done in several ways: noticing that a current inst var is nil when it should have data, noticing that there is an older inst var name in the variable dictionary (varDict), checking kinds of objects in one or more inst vars, or retrieving the classVersion of the incoming object from the ref stream.  

If a class is renamed, a method goes into SmartRefStream telling the new name.  The conversion method of the new class must be prepared to accept instances of the old class also.  If no inst var names have changed, the conversion method does nothing.

An example:  
	Suppose we change the representation of class Rectangle from ('origin' 'corner') to ('origin' 'extent').  Suppose lots of Rectangle instances are already out on files (in .pr project files, especially).  
	The programmer changes the class definition, modifies all the methods, and filesOut.  A series of dialogs appear, asking if instances Rectangle might be in an object file, if 'extent' needs to be non-nil (yes), and if the info in 'corner' needs to be preserved (yes).  This method appears:

Rectangle >> convertToCurrentVersion: varDict refStream: smartRefStrm
	"These variables are automatically stored into the new instance: #('origin').
	Test for this particular conversion.  Get values using expressions like (varDict at: 'foo')."

	"New variables: #('extent').  If a non-nil value is needed, please assign it."
	"These are going away #('corner').  Possibly store their info in some other variable?"
	"Move your code above the ^ super...  Delete extra comments."
	^ super convertToCurrentVersion: varDict refStream: smartRefStrm

The programmer modifies it to be:

Rectangle >> convertToCurrentVersion: varDict refStream: smartRefStrm

(varDict includesKey: 'extent') ifFalse: ["old version!!"
	"Create the new extent, and preserve the info from the old corner"
	extent := (varDict at: 'corner') - origin.
	].
^ super convertToCurrentVersion: varDict refStream: smartRefStrm

	This conversion method stays in the system and is ready to convert the old format of Rectangle whenever one is encountered in an object file.  Note that the subclasses of Rectangle, (B3DViewport, CharacterBlock, and Quadrangle) do not need conversion methods.  Their instances will be converted by the code in Rectangle.  

	Files written by SmartRefStream are in standard fileout format.  You can mix raw objects with code to be filed in.  The file starts out in the normal fileOut format.  Definitions of new classes on the front.

structures 	Dictionary of (#Rectangle -> #(<classVersionInteger> 'origin' 'corner')).  Inst 
				var names are strings.
steady 		Set of Classes who have the same structure now as on the incoming file.
				Includes classes with same inst vars except for new ones added on the end.
reshaped 	Dictionary of Classes who have a different structure now from the incoming file.  
				Includes those with same inst vars but new version number.
				(old class name -> method selector to fill in data for version to version)
renamed	Dictionary of Classes who have a different name.  Make an instance of the new
			class, and send it the conversion call.
				(old class name symbol -> new class name).  
renamedConv	Dictionary of conversion selector for Classes who have a different name.
				(old class name symbol -> conversion selector).  
topCall		Tells if next or nextPut: are working on the top object in the tree.  
			nil if outside, the top object if deep inside.

See DataStream.typeIDFor: for where the tangle of objects is clipped, so the whole system will not be written on the file.

No object that is written on the file is ever a class.  All class definitions are filed in.  A class may be stored inside an ImageSegment that itself is stored in a SmartRefStream.

UniClasses are classes for the instance specific behavior of just one instance.  Subclasses of Player are an example.  When a UniClass is read in, and a class of the same name already exists, the incoming one is renamed.  ObjectScanner converts the filed-in code.

Values in instance variables of UniClasses are stored in the array that tells the class structure.  It is the fourth of the four top level objects.  #(version (class-structure) the-object ((#Player25 scripts slotInfo costumeDictionary) (#Player26 scripts slotInfo costumeDictionary))).

There is a separate subclass for doing veryDeepCopy (in memory).  Currently, any object for which objectToStoreOnDataStream return an object other than self, does this:  The new object (a DiskProxy) is traced.  When it comes time to go through the fields of the old object, they are not found as keys in references (DiskProxies are there instead).  So the old field value is left in the new object.  That is OK for StrikeFont, Class, MetaClass, DisplayScreen.  But the DiskProxies are evaluated, which takes a lot of time.

Some metaclasses are put into the structures table.  This is for when a block has a receiver that is a class.  See checkFatalReshape:.

ImageSegments:
	A ReferenceStream is used to enumerate objects to put inside an ImageSegment.  If an instance of a UniClass is seen, the class is put in also.
	A SmartRefStream is used to store the ImageSegment.  Roots are nil, and the segment is a wordArray.  We are encoding the outPointers.  Structures contains all classes from both places.  Must filter out UniClasses for some things, and do include them for putting source code at end of file.  Do not write any class inst vars in file.

--Ted Kaehler and Bob Arning.
!

!SmartRefStream class methodsFor: 'initialize-release' stamp: 'nice 12/27/2009 03:11'!
cleanUpCategories
	| list valid removed |
	"Look for all conversion methods that can't be used any longer.  Delete them."
	" SmartRefStream cleanUpCategories "

	"Two part selectors that begin with convert and end with a digit."
	"convertasossfe0: varDict asossfeu0: smartRefStrm"
	list := Symbol selectorsContaining: 'convert'.
	list := list select: [:symb | (symb beginsWith: 'convert') & (symb allButLast last isDigit)
				ifTrue: [(symb numArgs = 2)]
				ifFalse: [false]].
	valid := 0.  removed := 0.
	list do: [:symb |
		(self systemNavigation allClassesImplementing: symb) do: [:newClass | | newList newVers |
			newList := (Array with: newClass classVersion), (newClass allInstVarNames).
			newVers := self new versionSymbol: newList.
			(symb endsWith: (':',newVers,':')) 
				ifFalse: [
					"method is useless because can't convert to current shape"
					newClass removeSelector: symb.	"get rid of it"
					removed := removed + 1]
				ifTrue: [valid := valid + 1]]].
	Transcript cr; show: 'Removed: '; print: removed; 
		show: '		Kept: '; print: valid; show: ' '.! !

!SmartRefStream class methodsFor: 'i/o' stamp: 'RAA 7/9/2000 05:48'!
objectFromStreamedRepresentation: someBytes

	| file |

	file := RWBinaryOrTextStream with: someBytes.
	file reset.
	^file fileInObjectAndCode! !

!SmartRefStream class methodsFor: 'i/o' stamp: 'tk 12/9/97 21:31'!
read: aByteStream withClasses: structureArray
	"Read an object off the stream, but first check structureArray against the current system."

	| me |
	me := self on: aByteStream.
	me noHeader.
	me structures: (structureArray at: 2).
	me superclasses: (structureArray at: 4).
	(me verifyStructure = 'conversion method needed') ifTrue: [^ nil].
	^ super next
! !

!SmartRefStream class methodsFor: 'i/o' stamp: 'tk 5/20/97'!
scanFrom: aByteStream
	"During a code fileIn, we need to read in an object, and stash it in ScannedObject.  "

	| me |
	me := self on: aByteStream.
	ScannedObject := me next.
	aByteStream ascii.
	aByteStream next == $!! ifFalse: [
		aByteStream close.
		self error: 'Object did not end correctly']. 
	"caller will close the byteStream"
	"HandMorph.readMorphFile will retrieve the ScannedObject"! !

!SmartRefStream class methodsFor: 'i/o' stamp: 'cwp 6/20/2012 17:42'!
scanFrom: aByteStream environment: anEnvironment
	^ self scanFrom: aByteStream! !

!SmartRefStream class methodsFor: 'accessing' stamp: 'tk 5/20/97'!
scannedObject
	"The most recently read in object.  Watch out for read-in that is interrupted and resumed.  May want to make this a dictionary?  "

	^ ScannedObject! !

!SmartRefStream class methodsFor: 'accessing' stamp: 'tk 5/20/97'!
scannedObject: objOrNil
	"Used to free up the last object stashed here.  "

	ScannedObject := objOrNil! !

!SmartRefStream class methodsFor: 'utilities' stamp: 'tk 5/4/1998 17:34'!
statsOfSubObjects: anObject
	"Open a window with statistics on what objects would be written out with anObject.  Does not actually write on the disk.  Stats in the form:
	ScriptEditorMorph 51
		SortedCollection (21->LayoutMorph 15->SimpleButtonMorph 9->Array 4->CompoundTileMorph 2->StringMorph )"

	| dummy printOut |
	dummy := ReferenceStream on: (DummyStream on: nil).
		"Write to a fake Stream, not a file"
	"Collect all objects"
	dummy rootObject: anObject.	"inform him about the root"
	dummy nextPut: anObject.
	"(dummy references) is the raw data"
	printOut := dummy statisticsOfRefs.
	(StringHolder new contents: printOut) 
		openLabel: 'ReferenceStream statistics'.! !

!SmartRefStream class methodsFor: 'accessing' stamp: 'tk 3/11/98 09:45'!
subObjects: anObject ofClass: aClass
	"Return a collection of all instances of aClass that would be written out with anObject.  Does not actually write on the disk.  Inspect the result and ask for 'references to this object'."

	| dummy coll |
	dummy := ReferenceStream on: (DummyStream on: nil).
		"Write to a fake Stream, not a file"
	"Collect all objects"
	dummy rootObject: anObject.	"inform him about the root"
	dummy nextPut: anObject.
	coll := OrderedCollection new.
	dummy references keysDo: [:each |
		each class == aClass ifTrue: [coll add: each]].
	^ coll! !

!SmartRefStream class methodsFor: 'utilities' stamp: 'tk 5/4/1998 17:34'!
tallyOfSubObjects: anObject
	"Open a window with statistics on what objects would be written out with anObject.  Does not actually write on the disk.  Stats are simply the number of instances of each class:
	1450->Point   835->Rectangle   549->Array   300->String"

	| dummy bag |
	dummy := ReferenceStream on: (DummyStream on: nil).
		"Write to a fake Stream, not a file"
	"Collect all objects"
	dummy rootObject: anObject.	"inform him about the root"
	dummy nextPut: anObject.
	bag := Bag new.
	dummy references keysDo: [:key | bag add: key class name].
	"(bag sortedCounts) is the SortedCollection"
	(StringHolder new contents: bag sortedCounts printString) 
		openLabel: 'ReferenceStream statistics'.! !

!SmartRefStream methodsFor: 'strings-conversion' stamp: 'ar 4/10/2005 15:44'!
abstractStringx0

	^ String! !

!SmartRefStream methodsFor: 'read write' stamp: 'nice 7/18/2014 01:18'!
appendClassDefns
	"Make this a fileOut format file.  For each UniClass mentioned, prepend its source code to the file.  Class name conflicts during reading will be resolved then.  Assume instVarInfo: has already been done."

| asciiStream |
asciiStream := (String new: 1024) writeStream.
byteStream position = 0 ifTrue: [
	byteStream setFileTypeToObject.
		"Type and Creator not to be text, so can attach correctly to an email msg"
	asciiStream header; timeStamp].

asciiStream cr; nextPutAll: '!!ObjectScanner new initialize!!'; cr; cr.
self uniClasesDo: [:class | class
		class sharedPools size > 0 ifTrue:  "This never happens"
			[class shouldFileOutPools
				ifTrue: [class fileOutSharedPoolsOn: self]].
		class fileOutOn: asciiStream moveSource: false toFile: 0].	
		"UniClasses are filed out normally, no special format."

	asciiStream trailer.	"Does nothing for normal files.  
		HTML streams will have trouble with object data"

	"Append the object's raw data"
	asciiStream cr; cr; nextPutAll: '!!self smartRefStream!!'.
	byteStream nextPutAll: asciiStream contents asByteArray.		"get ready for objects"
! !

!SmartRefStream methodsFor: 'import image segment' stamp: 'tk 11/26/2004 05:53'!
applyConversionMethodsTo: objectIn className: className varMap: varMap
	"Modify the object's instance vars to have the proper values
for its new shape.  Mostly, fill in defaut values of new inst vars.
Can substitute an object of a different class.  (Beware: if
substituted, varMap will not be correct when the new object is asked
to convert.)"
	| anObject prevObject |

	self flag: #bobconv.

	anObject := objectIn.
	[
		prevObject := anObject.
		anObject := anObject convertToCurrentVersion: varMap
refStream: self.
		prevObject == anObject
	] whileFalse.
	^anObject! !

!SmartRefStream methodsFor: 'class changed shape' stamp: 'tk 1/7/97'!
catalogValues: instVarList size: varsOnDisk
	"Create a dictionary of (name -> value) for the inst vars of this reshaped object.  Indexed vars as (1 -> val) etc.  "

	| dict sz |
	dict := Dictionary new.
	2 to: instVarList size do: [:ind |
		dict at: (instVarList at: ind) put: self next].
	sz := varsOnDisk - (instVarList size - 1).
	1 to: sz do: [:ii | 
		dict at: ii put: self next].
	"Total number read MUST be equal to varsOnDisk!!"
	sz > 0 ifTrue: [dict at: #SizeOfVariablePart put: sz].
	^ dict! !

!SmartRefStream methodsFor: 'read write' stamp: 'nice 12/27/2009 03:10'!
checkCrLf
	| save isCrLf cc loneLf |
	"Watch for a file that has had all of its Cr's converted to CrLf's.  Some unpacking programs like Stuffit 5.0 do this by default!!"

	save := byteStream position.
	isCrLf := false.  loneLf := false.
	cc := 0.
	350 timesRepeat: [ | prev |
		prev := cc.
		(cc := byteStream next) = 16r0A "Lf" ifTrue: [
			prev = 16r0D "Cr" ifTrue: [isCrLf := true] ifFalse: [loneLf := true]].
		].
	isCrLf & (loneLf not) ifTrue: [
		self inform: 'Carriage Returns in this file were converted to CrLfs 
by an evil unpacking utility.  Please set the preferences in 
StuffIt Expander to "do not convert file formats"'].
	byteStream position: save.
! !

!SmartRefStream methodsFor: 'import image segment' stamp: 'nice 12/27/2009 03:10'!
checkFatalReshape: setOfClasses
	| suspects |
	"Inform the user if any of these classes were reshaped.  A block has a method from the old system whose receiver is of this class.  The method's inst var references might be wrong.  OK if inst vars were only added."

	self flag: #bobconv.	

	setOfClasses isEmpty ifTrue: [^ self].
	suspects := OrderedCollection new.
	setOfClasses do: [:aClass | | newInstVars oldInstVars bad className |
		className := renamed keyAtValue: aClass name ifAbsent: [aClass name].
		oldInstVars := (structures at: className ifAbsent: [#(0)]) allButFirst.		"should be there"
		newInstVars := aClass allInstVarNames.
		oldInstVars size > newInstVars size ifTrue: [bad := true].
		oldInstVars size = newInstVars size ifTrue: [
			bad := oldInstVars ~= newInstVars].
		oldInstVars size < newInstVars size ifTrue: [
			bad := oldInstVars ~= (newInstVars copyFrom: 1 to: oldInstVars size)].
		bad ifTrue: [suspects add: aClass]].

	suspects isEmpty ifFalse: [
		self inform: ('Imported foreign methods will run on instances of:\',
			suspects asArray printString, 
			'\whose shape has changed.  Errors may occur.') withCRs].! !

!SmartRefStream methodsFor: 'class changed shape' stamp: 'nice 12/27/2009 03:10'!
conversionMethodsFor: classList
	| list |
	"Each of these needs a conversion method.  Hard part is the comment in it.  Return a MessageSet."

	list := OrderedCollection new.
	classList do: [:cls | | oldStruct newStruct |
		oldStruct := structures at: cls name ifAbsent: [#()].
		newStruct := (Array with: cls classVersion), (cls allInstVarNames).
		self writeConversionMethodIn: cls fromInstVars: oldStruct to: newStruct 
				renamedFrom: nil.
		list add: cls name, ' convertToCurrentVersion:refStream:'.
		].

	^list.! !

!SmartRefStream methodsFor: 'import image segment' stamp: 'RAA 12/20/2000 11:06'!
convert1: misShapenInst to: goodClass allVarMaps: allVarMaps
	"Go through the normal instance conversion process and return a modern object."

	| className oldInstVars anObject varMap |

	self flag: #bobconv.	

	goodClass isVariable ifTrue: [
		goodClass error: 'shape change for variable class not implemented yet'
	].
	(misShapenInst class name beginsWith: 'Fake37') ifFalse: [self error: 'why mapping?'].
	className := (misShapenInst class name allButFirst: 6) asSymbol.
	oldInstVars := structures at: className.
	anObject := goodClass basicNew.

	varMap := Dictionary new.	"later, indexed vars as (1 -> val) etc."
	2 to: oldInstVars size do: [:ind |
		varMap at: (oldInstVars at: ind) put: (misShapenInst instVarAt: ind-1)].
	varMap at: #ClassName put: className.	"original"
	varMap at: #NewClassName put: goodClass name.	"new"
	self storeInstVarsIn: anObject from: varMap. 	"ones with the same names"
	allVarMaps at: misShapenInst put: varMap.
	^ anObject
! !

!SmartRefStream methodsFor: 'import image segment' stamp: 'RAA 12/20/2000 17:15'!
convert2: partiallyCorrectInst allVarMaps: allVarMaps
	"Go through the normal instance conversion process and return a modern object."

	| className varMap |

	self flag: #bobconv.	

	varMap := allVarMaps at: partiallyCorrectInst.
	className := varMap at: #ClassName.	"original"
	^self applyConversionMethodsTo: partiallyCorrectInst className: className varMap: varMap.

! !

!SmartRefStream methodsFor: 'read write' stamp: 'bf 4/10/2013 16:24'!
initKnownRenames
	renamed
		at: #FlasherMorph put: #Flasher;
		at: #AlansTextPlusMorph put: #TextPlusMorph;
		at: #Project put: #MorphicProject;
		at: #Presenter put: #EtoysPresenter;
		yourself! !

!SmartRefStream methodsFor: 'read write' stamp: 'ar 4/10/2005 19:27'!
initShapeDicts
	"Initialize me. "

	self flag: #bobconv.	

	"These must stay constant.  When structures read in, then things can change."
	steady := {Array. Dictionary. Association. ByteString. SmallInteger} asSet.

	renamed ifNil: [
		renamed := Dictionary new.  "(old class name symbol -> new class name)"
		renamedConv := Dictionary new "(oldClassNameSymbol -> conversionSelectorInNewClass)"
	].
	self initKnownRenames! !

!SmartRefStream methodsFor: 'read write' stamp: 'nice 1/2/2013 23:24'!
instVarInfo: anObject
	"Return the object to write on the outgoing file that contains the structure of each class we are about to write out.  Must be an Array whose first element is 'class structure'.  Its second element is a Dictionary of pairs of the form #Rectangle -> #(<classVersion> 'origin' 'corner').  "

	"Make a pass through the objects, not writing, but recording the classes.  Construct a database of their inst vars and any version info (classVersion)."

	| dummy refs cls newSupers |
	structures := Dictionary new.
	superclasses := Dictionary new.
	dummy := ReferenceStream on: (DummyStream on: nil).
		"Write to a fake Stream, not a file"
	"Collect all objects"
	dummy rootObject: anObject.	"inform him about the root"
	dummy nextPut: anObject.
	refs := dummy references.
	objCount := refs size.		"for progress bar"
		"Note that Dictionary must not change its implementation!!  If it does, how do we read this reading information?"
	(refs includesKey: #AnImageSegment) 
		ifFalse: [
			self uniClassInstVarsRefs: dummy.	"catalog the extra objects in UniClass inst vars"
			refs keysDo: [:each | 
				cls := each class.
				"cls isObsolete ifTrue: [self error: 'Trying to write ', cls name]."
				(cls class ~~ Metaclass) & (cls isObsolete not) ifTrue: [
					structures at: cls name put: false]]]
		ifTrue: [self recordImageSegment: refs].
	"Save work by only computing inst vars once for each class"
	newSupers := Set new.
	structures at: #Point put: false.	"writeRectangle: does not put out class pointer"
	structures at: #Rectangle put: false.
	structures at: #LargePositiveInteger put: false.	"used in slow case of WordArray"
	structures keysDo: [:nm | 
		cls := (nm endsWith: ' class') 
			ifFalse: [Smalltalk at: nm]
			ifTrue: [(Smalltalk at: nm substrings first asSymbol) class].
		cls allSuperclasses do: [:aSuper |
			structures at: aSuper name ifAbsent: [newSupers add: aSuper name]]].
			"Don't modify structures during iteration"
	newSupers do: [:nm | structures at: nm put: 3].	"Get all superclasses into list"
	structures keysDo: [:nm | "Nothing added to classes during loop"
		cls := (nm endsWith: ' class') 
			ifFalse: [Smalltalk at: nm]
			ifTrue: [(Smalltalk at: nm substrings first asSymbol) class].
		structures at: nm put: 
			((Array with: cls classVersion), (cls allInstVarNames)).
		superclasses at: nm ifAbsentPut: [cls superclass name]].
	(refs includesKey: #AnImageSegment) 
		ifTrue: [classInstVars := #()]
		ifFalse: [self saveClassInstVars].	"of UniClassses"
	^ (Array with: 'class structure' with: structures with: 'superclasses' with: superclasses)! !

!SmartRefStream methodsFor: 'read write' stamp: 'ar 4/10/2005 18:52'!
mapClass: incoming
	"See if the old class named nm exists.  If so, return it.  If not, map it to a new class, and save the mapping in renamed.  "

	| cls oldVer sel nm |

	self flag: #bobconv.	


	nm := renamed at: incoming ifAbsent: [incoming].	"allow pre-mapping around collisions"
	(nm endsWith: ' class') 
		ifFalse: [cls := Smalltalk at: nm ifAbsent: [nil].
			cls ifNotNil: [^ cls]]  	"Known class.  It will know how to translate the instance."
		ifTrue: [cls := Smalltalk at: nm substrings first asSymbol ifAbsent: [nil].
			cls ifNotNil: [^ cls class]]. 	"Known class.  It will know how to translate the instance."
	oldVer := self versionSymbol: (structures at: nm).
	sel := nm asString.
	sel at: 1 put: (sel at: 1) asLowercase.
	sel := sel, oldVer.	"i.e. #rectangleoc4"
	Symbol hasInterned: sel ifTrue: [:symb | 
		(self class canUnderstand: sel asSymbol) ifTrue: [
			reshaped ifNil: [reshaped := Dictionary new].
			cls := self perform: sel asSymbol]].	"This class will take responsibility"
	cls ifNil: [cls := self writeClassRenameMethod: sel was: nm
					fromInstVars: (structures at: nm).
			   cls isString ifTrue: [cls := nil]].
	cls ifNotNil: [renamed at: nm put: cls name].
	^ cls
! !

!SmartRefStream methodsFor: 'import image segment' stamp: 'bf 4/9/2013 15:33'!
mapClass: newClass origName: originalName
	"See if instances changed shape.  If so, make a fake class for the old shape and return it.  Remember the original class name."

	| newName oldInstVars fakeClass |
	newClass isMeta ifTrue: [^ newClass].
	newName := newClass name.
	(steady includes: newClass) & (newName == originalName) ifTrue: [^ newClass].
		"instances in the segment have the right shape"
	oldInstVars := structures at: originalName ifAbsent: [
			self error: 'class is not in structures list'].	"Missing in object file"

	"Allow mapping from old to new string names"
	(newName == #ByteString and:[originalName == #String]) ifTrue:[^newClass].
	(newName == #WideString and:[originalName == #MultiString]) ifTrue:[^newClass].
	(newName == #WideSymbol and:[originalName == #MultiSymbol]) ifTrue:[^newClass].

	"Variable classes are not handled fully. This one is fine. --bf"
	newName == #MethodContext ifTrue: [^newClass].

	fakeClass := Object subclass: ('Fake37', originalName) asSymbol
		instanceVariableNames: oldInstVars allButFirst
		classVariableNames: ''
		poolDictionaries: ''
		category: 'Obsolete'.
	ChangeSet current removeClassChanges: fakeClass name.	"reduce clutter"
	^ fakeClass
! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 5/26/1998 15:09'!
moreObjects
	"Return true if there appears to be another object following this one on the file."

	| byte |
	byteStream atEnd ifTrue: [^ false].	"off end of file"
	(byte := byteStream peek) ifNil: [^ false].	"off end of file"
	byte = 33 "$!! asciiValue" ifTrue: [^ false].
	byte = 0 ifTrue: [^ false].
	^ byte <= RefTypes size		"between 1 and 16"! !

!SmartRefStream methodsFor: 'strings-conversion' stamp: 'ar 4/12/2005 17:38'!
multiStringx0

	^ WideString! !

!SmartRefStream methodsFor: 'strings-conversion' stamp: 'ar 4/12/2005 17:38'!
multiSymbolx0

	^ WideSymbol! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 6/19/2000 17:06'!
next
	"Really write three objects: (version, class structure, object). But only when called from the outside.  "

	| version ss object |
	^ topCall == nil 
		ifTrue: 
			[topCall := #marked.
			version := super next.
			version class == SmallInteger ifFalse: [^ version].	
				"version number, else just a regular object, not in our format, "
			self checkCrLf.
			ss := super next.
			ss class == Array ifFalse: [^ ss].  "just a regular object"
			(ss at: 1) = 'class structure' ifFalse: [^ ss].
			structures := ss at: 2.
			superclasses := (ss size > 3 and: [(ss at: 3) = 'superclasses']) 
				ifTrue: [ss at: 4]		"class name -> superclass name"
				ifFalse: [Dictionary new].
			(self verifyStructure = 'conversion method needed') ifTrue: [^ nil].
			object := super next.	"all the action here"
			self restoreClassInstVars.		"for UniClasses. version 4"

			topCall := nil.	"reset it"
			object]
		ifFalse:
			[super next]
! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 3/5/2002 09:52'!
nextAndClose
	"Speedy way to grab one object.  Only use when we are inside an object binary file.  If used for the start of a SmartRefStream mixed code-and-object file, tell the user and then do the right thing."

	| obj |
	byteStream peek = ReferenceStream versionCode "4" ifFalse: [
		"OK it is a fileIn afterall..."
		self inform: 'Should be using fileInObjectAndCode'.
		byteStream ascii.
		byteStream fileIn.
		obj := SmartRefStream scannedObject.
		SmartRefStream scannedObject: nil.
		^ obj].

	obj := self next.
	self close.
	^ obj! !

!SmartRefStream methodsFor: 'read write' stamp: 'ul 6/17/2011 12:32'!
nextPutObjOnly: anObject
	"Really write three objects: (version, class structure, object).  But only when called from the outside.  Not in fileOut format.  No class definitions will be written for instance-specific classes.  Error if find one.  (Use nextPut: instead)"

	| info |
	topCall == nil 
		ifTrue:
			[topCall := anObject.
			super nextPut: ReferenceStream versionCode.
			'Please wait while objects are counted'
				displayProgressFrom: 0 to: 10
				during: [:bar |
					info := self instVarInfo: anObject].
			self uniClasesDo: [:cls | cls error: 'Class defn not written out.  Proceed?'].
			'Writing an object file'
				displayProgressFrom: 0 to: objCount*4	"estimate"
				during: [:bar |
					objCount := 0.
					progressBar := bar.
					super nextPut: info.
					super nextPut: anObject.	"<- the real writing"
					"Class inst vars not written here!!"].
			"references is an IDict of every object that got written
			(in case you want to take statistics)"
			"Transcript cr; show: structures keys printString."		"debug"
			topCall := progressBar := nil]	"reset it"
		ifFalse:
			[super nextPut: anObject.
			progressBar ifNotNil: [progressBar value: (objCount := objCount + 1)]].! !

!SmartRefStream methodsFor: 'read write' stamp: 'ul 6/17/2011 12:32'!
nextPut: anObject
	"Really write three objects: (version, class structure, object).  But only when called from the outside.  If any instance-specific classes are present, prepend their source code.  byteStream will be in fileOut format.
	You can see an analysis of which objects are written out by doing: 
	(SmartRefStream statsOfSubObjects: anObject)
	(SmartRefStream tallyOfSubObjects: anObject)
	(SmartRefStream subObjects: anObject ofClass: aClass)"

| info |
topCall == nil 
	ifTrue:
		[topCall := anObject.
		'Please wait while objects are counted' 
			displayProgressFrom: 0 to: 10
			during: [:bar | info := self instVarInfo: anObject].
		self appendClassDefns.	"For instance-specific classes"
		'Writing an object file'
			displayProgressFrom: 0 to: objCount*4	"estimate"
			during: [:bar |
				objCount := 0.
				progressBar := bar.
				self setStream: byteStream reading: false.
					"set basePos, but keep any class renames"
				super nextPut: ReferenceStream versionCode.
				super nextPut: info.
				super nextPut: anObject.		"<- the real writing"
				classInstVars size > 0 ifTrue: [super nextPut: classInstVars]].
					"Note: the terminator, $!!, is not doubled inside object data"
		"references is an IDict of every object that got written"
		byteStream ascii.
		byteStream nextPutAll: '!!'; cr; cr.
		byteStream padToEndWith: $ .	"really want to truncate file, but can't"
		topCall := progressBar := nil]	"reset it"
	ifFalse:
		[super nextPut: anObject.
		progressBar ifNotNil: [progressBar value: (objCount := objCount + 1)]].
		"return the argument - added by kwl"
	^ anObject
! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 6/23/1998 11:13'!
noHeader
	"Signal that we've already dealt with the version and structure array, and are now reading objects."

	topCall := #marked.
! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 11/3/2000 17:59'!
readInstance
	"Read the contents of an arbitrary instance.
	 ASSUMES: readDataFrom:size: sends me beginReference: after it
	   instantiates the new object but before reading nested objects.
	 NOTE: We must restore the current reference position after
	   recursive calls to next.
Three cases for files from older versions of the system:
1) Class has not changed shape, read it straight.
2) Class has changed instance variables (or needs fixup).  Call a particular method to do it.
3) There is a new class instead.  Find it, call a particular method to read.
	All classes used to construct the structures dictionary *itself* need to be in 'steady' and they must not change!!  See setStream:"
	| instSize className refPosn |

	instSize := (byteStream nextNumber: 4) - 1.
	refPosn := self getCurrentReference.
	className := self next asSymbol.
	^ self readInstanceSize: instSize clsname: className refPosn: refPosn
! !

!SmartRefStream methodsFor: 'read write' stamp: 'topa 6/25/2015 10:27'!
readInstanceSize: instSize clsname: className refPosn: refPosn
	"The common code to read the contents of an arbitrary instance.
	 ASSUMES: readDataFrom:size: sends me beginReference: after it
	   instantiates the new object but before reading nested objects.
	 NOTE: We must restore the current reference position after
	   recursive calls to next.
Three cases for files from older versions of the system:
1) Class has not changed shape, read it straight.
2) Class has changed instance variables (or needs fixup).  Call a particular method to do it.
3) There is a new class instead.  Find it, call a particular method to read.
	All classes used to construct the structures dictionary *itself* need to be in 'steady' and they must not change!!  See setStream:"
	| anObject newName newClass dict oldInstVars isMultiSymbol |

	self flag: #bobconv.	

	self setCurrentReference: refPosn.  "remember pos before readDataFrom:size:"
	newName := renamed at: className ifAbsent: [className].
	newName == #Character
		ifTrue: [^ self readCharacterOfSize:  instSize].
	isMultiSymbol := newName = #MultiSymbol or: [newName = #WideSymbol].
	"isMultiSymbol ifTrue: [self halt]."
	newClass := Smalltalk at: newName asSymbol.
	(steady includes: newClass) & (newName == className) ifTrue: [
	 	anObject := newClass isVariable "Create it here"
			ifFalse: [newClass basicNew]
			ifTrue: [newClass basicNew: instSize - (newClass instSize)].

		anObject := anObject readDataFrom: self size: instSize.
		self setCurrentReference: refPosn.  "before returning to next"
		isMultiSymbol ifTrue: [^ Symbol intern: anObject asString].
		^ anObject].
	oldInstVars := structures at: className ifAbsent: [
			self error: 'class is not in structures list'].	"Missing in object file"
	anObject := newClass createFrom: self size: instSize version: oldInstVars.
		"only create the instance"
	self beginReference: anObject.
	dict := self catalogValues: oldInstVars size: instSize.
		"indexed vars as (1 -> val) etc."
	dict at: #ClassName put: className.	"so conversion method can know it"

	"Give each superclass a chance to make its changes"
	self storeInstVarsIn: anObject from: dict.	"ones with the same names"

	anObject := self applyConversionMethodsTo: anObject className: className varMap: dict.

	self setCurrentReference: refPosn.  "before returning to next"
	isMultiSymbol ifTrue: [^ Symbol intern: anObject asString].
	^ anObject! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 11/3/2000 18:04'!
readShortInst
	"Instance has just one byte of size.  Class symbol is encoded in two bytes of file position.  See readInstance."
	| instSize className refPosn |

	instSize := (byteStream next) - 1.	"one byte of size"
	refPosn := self getCurrentReference.
	className := self readShortRef.	"class symbol in two bytes of file pos"
	^ self readInstanceSize: instSize clsname: className refPosn: refPosn
! !

!SmartRefStream methodsFor: 'read write' stamp: 'ar 7/25/2005 21:30'!
readWordLike
	| refPosn newClass anObject className |
	"Can be used by any class that is bits and not bytes (WordArray, Bitmap, SoundBuffer, etc)."

	refPosn := self getCurrentReference.
	className := self next asSymbol.
	className := renamed at: className ifAbsent: [className].
	newClass := Smalltalk at: className.
	anObject := newClass newFromStream: byteStream.
	"Size is number of long words."
	self setCurrentReference: refPosn.  "before returning to next"
	^ anObject
! !

!SmartRefStream methodsFor: 'read write' stamp: 'nice 12/27/2009 03:10'!
recordImageSegment: refs
	"Besides the objects being written out, record the structure of instances inside the image segment we are writing out."

	| list |
	"Do not record Player class inst vars.  They are in the segement."
	refs keysDo: [:each | | cls | 
		cls := each class.
		cls isObsolete ifTrue: [self error: 'Trying to write ', cls name].
		cls class == Metaclass 
			ifFalse: [structures at: cls name put: false.
				(each isKindOf: ImageSegment) ifTrue: [
					each outPointers do: [:out |
						(out isKindOf: Class) ifTrue: [
							structures at: out theNonMetaClass name put: false].
						out class == DiskProxy ifTrue: [
							out simpleGlobalOrNil ifNotNil: [
								(out simpleGlobalOrNil isKindOf: Class) ifTrue: [
									structures at: out simpleGlobalOrNil name put: false]]]].
					"each arrayOfRoots do: [:rr | (rr isKindOf: Class) ifTrue: [
							structures at: rr theNonMetaClass name put: false]]."
					 	"all classes in roots are local to seg"]]].
	list := refs at: #BlockReceiverClasses ifAbsent: [^ self].
	list do: [:meta | structures at: meta name put: false].
		"Just the metaclasses whose instances are block receivers.  Otherwise metaclasses are not allowed."! !

!SmartRefStream methodsFor: 'read write' stamp: 'RAA 12/20/2000 11:08'!
renamed

	self flag: #bobconv.	


	^ renamed! !

!SmartRefStream methodsFor: 'read write' stamp: 'RAA 12/20/2000 11:10'!
renamedConv
	self flag: #bobconv.	


	^ renamedConv! !

!SmartRefStream methodsFor: 'import image segment' stamp: 'nice 12/27/2009 03:10'!
reshapedClassesIn: outPointers
	"Look for classes in the outPointer array that have changed shape.  Make a fake class for the old shape.  Return a dictionary mapping Fake classes to Real classes.  Substitute fake classes for real ones in outPointers."

	| mapFakeClassesToReal |

	self flag: #bobconv.	


	mapFakeClassesToReal := IdentityDictionary new.
	outPointers withIndexDo: [:outp :ind | | originalName fakeCls | 
		outp isBehavior ifTrue: [
			originalName := renamedConv at: ind ifAbsent: [outp name].
				"in DiskProxy>>comeFullyUpOnReload: we saved the name at the index"
			fakeCls := self mapClass: outp origName: originalName.
			fakeCls == outp ifFalse: [
				mapFakeClassesToReal at: fakeCls put: outp.
				outPointers at: ind put: fakeCls]]].
	^ mapFakeClassesToReal! !

!SmartRefStream methodsFor: 'read write' stamp: 'ul 11/16/2010 04:41'!
restoreClassInstVars
	"Install the values of the class instance variables of UniClasses
(i.e. scripts slotInfo).  classInstVars is ((#Player25 scripts slotInfo)
...).  Thank you Mark Wai for the bug fix."

	| normal trans classPlayer |

	self flag: #bobconv.	

	classPlayer := Smalltalk at: #Player ifAbsent:[^self].

	self moreObjects ifFalse: [^ self]. 	"are no UniClasses with class inst vars"
	classInstVars := super next.	"Array of arrays"
	normal := Object class instSize.	"might give trouble if Player class superclass changes size"
	(structures at: #Player ifAbsent: [#()]) = #(0 'dependents' 'costume') ifTrue:
		[trans := 1].	"now (0 costume costumes).  Do the conversion of Player class
			inst vars in Update 509."
	classInstVars do: [:list | | aName newCls rList newName start |
		aName := (list at: 1) asSymbol.
		rList := list.
		newName := renamed at: aName ifAbsent: [aName].
		newCls := Smalltalk at: newName
				ifAbsent: [self error: 'UniClass definition missing'].
		("old conversion" trans = 1 and: [newCls inheritsFrom: classPlayer]) ifTrue: [
			"remove costumeDictionary from Player class inst vars"
			rList := rList asOrderedCollection.
			rList removeAt: 4].	"costumeDictionary's value"
		start := list second = 'Update to read classPool' ifTrue: [4] ifFalse: [2].
		newCls class instSize = (normal + (rList size) - start + 1) ifFalse:
			[self error: 'UniClass superclass class has changed size'].
			"Need to install a conversion method mechanism"
		start = 4 ifTrue: [newCls instVarAt: normal - 1 "classPool" put: (list at: 3)].
		start to: rList size do: [:ii |
			newCls instVarAt: normal + ii - start + 1 put: (rList at: ii)]].
! !

!SmartRefStream methodsFor: 'read write' stamp: 'nice 12/27/2009 03:10'!
saveClassInstVars
	"Install the values of the instance variables of UniClasses.
classInstVars is an array of arrays (#Player3 (Player3 class's inst var
scripts) (Player3 class's inst var slotInfo) ...) "

	| normal clsPoolIndex |
	classInstVars := OrderedCollection new: 100.
	normal := Object class instSize.
	clsPoolIndex := Object class allInstVarNames indexOf: 'classPool'.
	self uniClasesDo: [:aUniClass | | list mySize |
		list := OrderedCollection new.
		mySize := aUniClass class instSize.
		mySize = normal ifFalse:
			[list add: aUniClass name.	"a symbol"
			list add: 'Update to read classPool'.	"new
convention for saving the classPool"
			list add: (aUniClass instVarAt: clsPoolIndex)
"classPool".
						"write actual value of nil
instead of Dictionary()"
			normal + 1 to: mySize do: [:ii |
				list addLast: (aUniClass instVarAt: ii)].
			classInstVars add: list asArray]].
	classInstVars := classInstVars asArray.
	! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 8/18/1998 09:02'!
scanFrom: aByteStream
	"During a code fileIn, we need to read in an object, and stash it in ScannedObject.  "

	self setStream: aByteStream reading: true.
	ScannedObject := self next.
	byteStream ascii.
	byteStream next == $!! ifFalse: [
		byteStream close.
		self error: 'Object did not end correctly']. 
	"caller will close the byteStream"
	"HandMorph.readMorphFile will retrieve the ScannedObject"! !

!SmartRefStream methodsFor: 'read write' stamp: 'cwp 6/20/2012 17:41'!
scanFrom: aByteStream environment: anEnvironment
	^ self scanFrom: aByteStream! !

!SmartRefStream methodsFor: 'read write' stamp: 'RAA 12/20/2000 16:57'!
setStream: aStream
	"Initialize me. "

	self flag: #bobconv.	

	super setStream: aStream.
	self initShapeDicts.

! !

!SmartRefStream methodsFor: 'read write' stamp: 'RAA 12/20/2000 16:57'!
setStream: aStream reading: isReading
	"Initialize me. "

	self flag: #bobconv.	

	super setStream: aStream reading: isReading.
	isReading ifFalse: [^ false].
	self initShapeDicts.

! !

!SmartRefStream methodsFor: 'class changed shape' stamp: 'tk 5/26/97'!
storeInstVarsIn: anObject from: dict
	"For instance variables with the same names, store them in the new instance.  Values in variable-length part also.  This is NOT the normal inst var transfer!!  See Object.readDataFrom:size:.  This is for when inst var names have changed and some additional conversion is needed.  Here we handle the unchanged vars.  "

	(anObject class allInstVarNames) doWithIndex: [:varName :index |
		(dict includesKey: varName) ifTrue: [
			anObject instVarAt: index put: (dict at: varName)]].
	"variable part"
	(dict includesKey: #SizeOfVariablePart) ifFalse: [^ anObject].
	1 to: (dict at: #SizeOfVariablePart) do: [:index | 
		anObject basicAt: index put: (dict at: index)].
	^ anObject! !

!SmartRefStream methodsFor: 'read write' stamp: ''!
structures
	^ structures! !

!SmartRefStream methodsFor: 'accessing' stamp: 'tk 5/19/1999 15:47'!
structures: anObject
	structures := anObject! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 9/28/97 11:17'!
superclasses
	^superclasses! !

!SmartRefStream methodsFor: 'accessing' stamp: 'tk 5/19/1999 15:47'!
superclasses: anObject
	superclasses := anObject! !

!SmartRefStream methodsFor: 'read write' stamp: 'nice 12/27/2009 03:10'!
uniClasesDo: aBlock
	"Examine structures and execute the block with each instance-specific class"

	
	structures keysDo: [:clsName | | cls | 
		(clsName endsWith: ' class') ifFalse: [
			(cls := Smalltalk at: clsName) isSystemDefined ifFalse: [
					aBlock value: cls]]]! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 1/18/2001 15:54'!
uniClassInstVarsRefs: dummy
	"If some of the objects seen so far are instances UniClasses, check the UniClasses for extra class inst vars, and send them to the steam also.  The new objects get added to (dummy references), where they will be noticed by the caller.  They will wind up in the structures array and will be written on the disk by class.
	Return all classes seen." 
| uniClasses normal more aUniClass mySize allClasses |

"Note: Any classes used in the structure of classInstVars must be written out also!!"
uniClasses := Set new.
allClasses := IdentitySet new.
normal := Object class instSize.
more := true.
[more] whileTrue: [
	more := false.
	dummy references keysDo: [:each | "any faster way to do this?"
		(aUniClass := each class) isSystemDefined ifFalse: [
			(uniClasses includes: aUniClass name) ifFalse: [
				mySize := aUniClass class instSize.
				normal+1 to: mySize do: [:ii | 
					more := true.
					dummy nextPut: (aUniClass instVarAt: ii)].
				uniClasses add: aUniClass name]].
		each class class isMeta ifFalse: ["it is a class" allClasses add: each]]].
"References dictionary is modified as the loop proceeds, but we will catch any we missed on the next cycle."

^ allClasses! !

!SmartRefStream methodsFor: 'read write' stamp: 'nice 12/27/2009 03:10'!
verifyStructure
	"Compare the incoming inst var name lists with the existing classes.  Prepare tables that will help to restructure those who need it (renamed, reshaped, steady).    If all superclasses are recorded in the file, only compare inst vars of this class, not of superclasses.  They will get their turn.  "


	| converting |

	self flag: #bobconv.	

	converting := OrderedCollection new.
	structures keysDo: [:nm "an old className (symbol)" | | newClass newList oldList |
		"For missing classes, there needs to be a method in SmartRefStream like 
			#rectangleoc2 that returns the new class."
		newClass := self mapClass: nm.	   "does (renamed at: nm put: newClass name)"
		newClass isString ifTrue: [^ newClass].  "error, fileIn needed"
		newList := (Array with: newClass classVersion), (newClass allInstVarNames).
		oldList := structures at: nm.
		newList = oldList 
			ifTrue: [steady add: newClass]  "read it in as written"
			ifFalse: [converting add: newClass name]
	].
	false & converting isEmpty not ifTrue: ["debug" 
			self inform: 'These classes are being converted from existing methods:\' withCRs,
				converting asArray printString].
! !

!SmartRefStream methodsFor: 'read write' stamp: 'tk 1/7/97'!
versionSymbol: instVarList
	"Create the symbolic code (like a version number) for this class in some older version.  First initials of all the inst vars, followed by the class version number.  Returns a string, caller makes it into a compound selector.  "

	| str |
	str := instVarList size = 1 ifFalse: [''] ifTrue: ['x'].		"at least one letter"
	2 to: instVarList size do: [:ind |
		str := str, (instVarList at: ind) first asString].
	str := str, instVarList first printString.	"the number"
	^ str

" | list | list := (Array with: Paragraph classVersion), (Paragraph alistInstVarNames).
(SmartRefStream  on: (DummyStream on: nil)) versionSymbol: list
"! !

!SmartRefStream methodsFor: 'conversion' stamp: 'di 8/16/2000 16:37'!
worldMorphbosfcebbfgccpmcpbttloiairfidcuwhavcdsll0
	^ 'PutNewClassHere'  " <-- Replace this by a class name (no string quotes)"! !

!SmartRefStream methodsFor: 'class changed shape' stamp: 'CdG 10/17/2005 21:01'!
writeClassRenameMethod: sel was: oldName fromInstVars: oldList
	"The class coming is unknown.  Ask the user for the existing class it maps to.  If got one, write a method, and restart the obj fileIn.  If none, write a dummy method and get the user to complete it later.  "

| tell choice  newName answ code |

	self flag: #bobconv.	


tell := 'Reading an instance of ', oldName, '.
Which modern class should it translate to?'.
answ := (UIManager default 
		chooseFrom: #('Let me type the name now' 'Let me think about it'
'Let me find a conversion file on the disk') 
		title: tell). 

answ = 1 ifTrue: [
	tell := 'Name of the modern class {1} should translate to:' translated format: {oldName}.
	choice := UIManager default request: tell.		"class name"
	(choice size = 0) 
		ifTrue: [answ := 'conversion method needed']
		ifFalse: [newName := choice.
			answ := Smalltalk at: newName asSymbol 
				ifAbsent: ['conversion method needed'].
			answ isString ifFalse: [renamed at: oldName asSymbol put: answ name]]].
(answ = 3) | (answ = 0) ifTrue: [self close.
		^ 'conversion method needed'].
answ = 2 ifTrue: [answ := 'conversion method needed'].
answ = 'conversion method needed' ifTrue: [
		self close.  
		newName := 'PutNewClassHere'].

code := WriteStream on: (String new: 500).
code nextPutAll: sel; cr.
code cr; tab; nextPutAll: '^ ', newName.	"Return new class"

self class compile: code contents classified: 'conversion'.

newName = 'PutNewClassHere' ifTrue: [
	self inform: 'Please complete the following method and 
then read-in the object file again.'.
	SystemNavigation default browseAllImplementorsOf: sel asSymbol]. 

	"The class version number only needs to change under one specific circumstance.  That is when the first letters of the instance variables have stayed the same, but their meaning has changed.  A conversion method is needed, but this system does not know it.  
	If this is true for class Foo, define classVersion in Foo class.  
	Beware of previous object fileouts already written after the change in meaning, but before bumping the version number.  They have the old (wrong) version number, say 2.  If this is true, your method must be able to test the data and successfully read files that say version 2 but are really 3."

	^ answ! !

!SmartRefStream methodsFor: 'class changed shape' stamp: 'tk 6/8/2001 09:57'!
writeClassRename: newName was: oldName
	"Write a method that tells which modern class to map instances to."
	| oldVer sel code |

	oldVer := self versionSymbol: (structures at: oldName).
	sel := oldName asString.
	sel at: 1 put: (sel at: 1) asLowercase.
	sel := sel, oldVer.	"i.e. #rectangleoc4"

	code := WriteStream on: (String new: 500).
	code nextPutAll: sel; cr.
	code cr; tab; nextPutAll: '^ ', newName.	"Return new class"

	self class compile: code contents classified: 'conversion'.

! !

!SmartRefStream methodsFor: 'class changed shape' stamp: 'fbs 7/25/2013 07:19'!
writeConversionMethodIn: newClass fromInstVars: oldList to: newList renamedFrom: oldName
	"The method convertToCurrentVersion:refStream: was not found in newClass.  Write a default conversion method for the author to modify.  If method exists, append new info into the end."

	| code newOthers oldOthers copied newCode |

	newOthers := newList asOrderedCollection "copy".
	oldOthers := oldList asOrderedCollection "copy".
	copied := OrderedCollection new.
	newList do: [:instVar |
		(oldList includes: instVar) ifTrue: [
			instVar isInteger ifFalse: [copied add: instVar].
			newOthers remove: instVar.
			oldOthers remove: instVar]].
	code := WriteStream on: (String new: 500).
	code cr; cr; tab; nextPutAll: '"From ', SystemVersion current version, ' [', Smalltalk lastUpdateString;
			nextPutAll: '] on ', Date today printString, '"'; cr.
	code tab; nextPutAll: '"These variables are automatically stored into the new instance: '.
	code nextPutAll: copied asArray printString; nextPut: $.; cr.
	code tab; nextPutAll: 'Test for this particular conversion.'; 
		nextPutAll: '  Get values using expressions like (varDict at: ''foo'')."'; cr; cr.
	(newOthers size = 0) & (oldOthers size = 0) & (oldName == nil) ifTrue: [^ self].
		"Instance variables are the same.  Only the order changed.  No conversion needed."
	(newOthers size > 0) ifTrue: [
		code tab; nextPutAll: '"New variables: ', newOthers asArray printString, 
			'.  If a non-nil value is needed, please assign it."'; cr].
	(oldOthers size > 0) ifTrue: [
		code tab; nextPutAll: '"These are going away ', oldOthers asArray printString, 
			'.  Possibly store their info in some other variable?"'; cr].
	oldName ifNotNil: [
		code tab; nextPutAll: '"Test for instances of class ', oldName, '.'; cr.
		code tab; nextPutAll: 'Instance vars with the same name have been moved here."'; cr.
		].
	code tab; nextPutAll: '"Move your code above the ^ super...  Delete extra comments."'; cr. 

	(newClass includesSelector: #convertToCurrentVersion:refStream:) 
		ifTrue: ["append to old methods"
			newCode := (newClass sourceCodeAt: #convertToCurrentVersion:refStream:),
				code contents]
		ifFalse: ["new method"
			newCode := 'convertToCurrentVersion: varDict refStream: smartRefStrm',
				code contents, 
				'	^ super convertToCurrentVersion: varDict refStream: smartRefStrm'].
	newClass compile: newCode classified: 'object fileIn'.


	"If you write a conversion method beware that the class may need a version number change.  This only happens when two conversion methods in the same class have the same selector name.  (A) The inst var lists of the new and old versions intials as some older set of new and old inst var lists.  or (B) Twice in a row, the class needs a conversion method, but the inst vars stay the same the whole time.  (For an internal format change.)
	If either is the case, fileouts already written with the old (wrong) version number, say 2.  Your method must be able to read files that say version 2 but are really 3, until you expunge the erroneous version 2 files from the universe."

 ! !

!SmartRefStream methodsFor: 'class changed shape' stamp: 'tk 6/7/2001 13:02'!
writeConversionMethod: sel class: newClass was: oldName fromInstVars: oldList to: newList
	"The method convertToCurrentVersion:refStream: was not found in newClass.  Write a default conversion method for the author to modify."

	| code newOthers oldOthers copied |

	code := WriteStream on: (String new: 500).
	code nextPutAll: 'convertToCurrentVersion: varDict refStream: smartRefStrm'; cr; tab.
	newOthers := newList asOrderedCollection "copy".
	oldOthers := oldList asOrderedCollection "copy".
	copied := OrderedCollection new.
	newList do: [:instVar |
		(oldList includes: instVar) ifTrue: [
			instVar isInteger ifFalse: [copied add: instVar].
			newOthers remove: instVar.
			oldOthers remove: instVar]].
	code nextPutAll: '"These variables are automatically stored into the new instance '.
	code nextPutAll: copied asArray printString; nextPut: $. .
	code cr; tab; nextPutAll: 'This method is for additional changes.'; 
		nextPutAll: ' Use statements like (foo := varDict at: ''foo'')."'; cr; cr; tab.
	(newOthers size = 0) & (oldOthers size = 0) ifTrue: [^ self].
		"Instance variables are the same.  Only the order changed.  No conversion needed."
	(newOthers size > 0) ifTrue: [code nextPutAll: '"New variables: ', newOthers asArray printString, '  If a non-nil value is needed, please assign it."\' withCRs].
	(oldOthers size > 0) ifTrue: [code nextPutAll: '	"These are going away ', oldOthers asArray printString, '.  Possibly store their info in some other variable?"'].

	code cr; tab.
	code nextPutAll: '^ super convertToCurrentVersion: varDict refStream: smartRefStrm'.
	newClass compile: code contents classified: 'object fileIn'.


	"If you write a conversion method beware that the class may need a version number change.  This only happens when two conversion methods in the same class have the same selector name.  (A) The inst var lists of the new and old versions intials as some older set of new and old inst var lists.  or (B) Twice in a row, the class needs a conversion method, but the inst vars stay the same the whole time.  (For an internal format change.)
	If either is the case, fileouts already written with the old (wrong) version number, say 2.  Your method must be able to read files that say version 2 but are really 3, until you expunge the erroneous version 2 files from the universe."

 ! !

Stream subclass: #DummyStream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Object Storage'!

!DummyStream commentStamp: 'fbs 3/5/2013 09:42' prior: 0!
The purpose of this class is to absorb all stream messages and do nothing.  This is so ReferenceStream can pretend to write on it while traversing all objects it would normally write.  We need to know what those object are.  8/17/96 tk
!

!DummyStream class methodsFor: 'instance creation' stamp: 'jm 12/3/97 20:25'!
on: aFile
	"Return a new DummyStream instance, ignoring the argument."

	^ self basicNew
! !

!DummyStream methodsFor: 'protocol' stamp: ''!
binary
	"do nothing"! !

!DummyStream methodsFor: 'open/close' stamp: 'tk 10/31/97 11:43'!
close
	"do nothing"! !

!DummyStream methodsFor: 'accessing' stamp: ''!
nextInt32Put: arg
	"do nothing"! !

!DummyStream methodsFor: 'accessing' stamp: ''!
nextNumber: cnt put: num
	"do nothing"! !

!DummyStream methodsFor: 'accessing' stamp: 'nice 3/17/2010 20:31'!
nextPutAll: aByteArray
	"do nothing"
	
	^aByteArray! !

!DummyStream methodsFor: 'accessing' stamp: 'nice 3/19/2010 19:03'!
nextPut: aByte
	"do nothing"
	^aByte! !

!DummyStream methodsFor: 'accessing' stamp: 'nice 7/19/2014 00:17'!
nextStringPut: aString
	"do nothing"
	^aString! !

!DummyStream methodsFor: 'accessing' stamp: 'ul 4/25/2016 02:31'!
next: anInteger putAll: aCollection startingAt: anInteger3 

	^aCollection! !

!DummyStream methodsFor: 'accessing' stamp: 'tk 3/6/2000 11:10'!
originalContents

	^ ''! !

!DummyStream methodsFor: 'accessing' stamp: ''!
position
	"Return any random number.  Here is where the real lying begins.  We are a DummyStream afterall.  8/17/96 tk"

	^ 47 ! !

!DummyStream methodsFor: 'accessing' stamp: 'tk 7/12/1998 12:51'!
position: anOffset
	"Pretend to position wherever the caller says!!"
! !

!DummyStream methodsFor: 'positioning' stamp: '6/10/97 17:14 tk'!
skip: aNumber
	"Do nothing."! !

!DummyStream methodsFor: 'error handling' stamp: ''!
subclassResponsibility
	"Do nothing.  Most messages to class Stream are defined as subclassResponsibility.  Just accept them.  8/17/96 tk"

	"No error.  Just go on."! !

Object subclass: #AbstractEvent
	instanceVariableNames: 'item itemKind environment'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!AbstractEvent commentStamp: 'fbs 9/26/2013 17:49' prior: 0!
An AbstractEvent is used by the system change notification framework to indicate that something happened with code - a class was recategorised, a method was added, and so on.

My environment instance variable is a Dictionary containing all the relevant information for an event - selector, class, and so on. It is not an Environment instance!!!

!AbstractEvent class methodsFor: 'accessing' stamp: 'NS 1/16/2004 14:08'!
allChangeKinds
	"AbstractEvent allChangeKinds"

	^AbstractEvent allSubclasses collect: [:cl | cl changeKind]! !

!AbstractEvent class methodsFor: 'accessing' stamp: 'bvs 7/20/2004 12:12'!
allItemKinds
	"self allItemKinds"

	^(AbstractEvent class organization listAtCategoryNamed: #'item kinds') 
		collect: [:sel | self perform: sel]! !

!AbstractEvent class methodsFor: 'item kinds' stamp: 'rw 7/9/2003 11:12'!
categoryKind

	^#category! !

!AbstractEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:08'!
changeKind
	"Return a symbol, with a : as last character, identifying the change kind."

	self subclassResponsibility! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'ab 2/10/2005 16:32'!
classCategory: aName
	^ self item: aName kind: AbstractEvent categoryKind.! !

!AbstractEvent class methodsFor: 'item kinds' stamp: 'rw 7/9/2003 11:12'!
classKind

	^#class! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'NS 1/19/2004 18:42'!
class: aClass
	^ self item: aClass kind: AbstractEvent classKind.! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'NS 1/19/2004 18:42'!
class: aClass category: cat 
	| instance |
	instance := self class: aClass.
	instance itemCategory: cat.
	^instance! !

!AbstractEvent class methodsFor: 'temporary' stamp: 'rw 7/11/2003 10:23'!
comment1

"Smalltalk organization removeElement: #ClassForTestingSystemChanges3
Smalltalk garbageCollect 
Smalltalk organizati

classify:under:


SystemChangeNotifier uniqueInstance releaseAll
SystemChangeNotifier uniqueInstance noMoreNotificationsFor: aDependent.


aDependent := SystemChangeNotifierTest new.
SystemChangeNotifier uniqueInstance
	notifyOfAllSystemChanges: aDependent
	using: #event:

SystemChangeNotifier uniqueInstance classAdded: #Foo inCategory: #FooCat



| eventSource dependentObject |
eventSource := EventManager new.
dependentObject := Object new.

register - dependentObject becomes dependent:
eventSource
	when: #anEvent send: #error to: dependentObject.

unregister dependentObject:
eventSource removeDependent: dependentObject.

[eventSource triggerEvent: #anEvent]
	on: Error
	do: [:exc | self halt: 'Should not be!!']."! !

!AbstractEvent class methodsFor: 'temporary' stamp: 'rw 7/11/2003 10:24'!
comment2

"HTTPSocket useProxyServerNamed: 'proxy.telenet.be' port: 8080
TestRunner open

--------------------
We propose two orthogonal groups to categorize each event:
(1) the 'change type':
	added, removed, modified, renamed
	+ the composite 'changed' (see below for an explanation)
(2) the 'item type':
	class, method, instance variable, pool variable, protocol, category
	+ the composite 'any' (see below for an explanation).
The list of supported events is the cross product of these two lists (see below for an explicit enumeration of the events).

Depending on the change type, certain information related to the change is always present (for adding, the new things that was added, for removals, what was removed, for renaming, the old and the new name, etc.).

Depending on the item type, information regarding the item is present (for a method, which class it belongs to). 

Certain events 'overlap', for example, a method rename triggers a class change. To capture this I impose a hierarchy on the 'item types' (just put some numbers to clearly show the idea. They don't need numbers, really. Items at a certain categories are included by items one category number higher):
level 1 category
level 2 class
level 3 instance variable, pool variable, protocol, method.

Changes propagate according to this tree: any 'added', 'removed' or 'renamed' change type in level X triggers a 'changed' change type in level X - 1. A 'modified' change type does not trigger anything special.
For example, a method additions triggers a class modification. This does not trigger a category modification.

Note that we added 'composite events': wildcards for the 'change type' ('any' - any system additions) and for the 'item type' ('Changed' - all changes related to classes), and one for 'any change systemwide' (systemChanged).

This result is this list of Events:

classAdded
classRemoved
classModified
classRenamed (?)
classChanged (composite)

methodAdded
methodRemoved
methodModified
methodRenamed (?)
methodChanged (composite)

instanceVariableAdded
instanceVariableRemoved
instanceVariableModified 
instanceVariableRenamed (?)
instanceVariableChanged (composite)

protocolAdded
protocolRemoved
protocolModified
protocolRenamed (?)
protocolChanged (composite)

poolVariableAdded
poolVariableRemoved
poolVariableModified
poolVariableRenamed (?)
poolChanged (composite)

categoryAdded
categoryRemoved
categoryModified
categeryRenamed (?)
categoryChanged (composite)

anyAdded (composite)
anyRemoved (composite)
anyModified (composite)
anyRenamed (composite)

anyChanged (composite)



To check: can we pass somehow the 'source' of the change (a browser, a file-in, something else) ? Maybe by checking the context, but should not be too expensive either... I found this useful in some of my tools, but it might be too advanced to have in general. Tools that need this can always write code to check it for them.  But is not always simple...


Utilities (for the recent methods) and ChangeSet are the two main clients at this moment.

Important: make it very explicit that the event is send synchronously (or asynchronously, would we take that route).


					category
						class
							comment
							protocol
								method
OR
				category
				Smalltalk
					class
						comment
						protocol
						method
??



						Smalltalk	category
								\	/
								class
							/	  |	\
						comment  |	protocol
								  |	/
								method

"! !

!AbstractEvent class methodsFor: 'temporary' stamp: 'rw 7/11/2003 15:43'!
comment3

"Things to consider for trapping:
ClassOrganizer>>#changeFromCategorySpecs:
	Problem: I want to trap this to send the appropriate bunch of ReCategorization events, but ClassOrganizer instances do not know where they belong to (what class, or what system); it just uses symbols. So I cannot trigger the change, because not enough information is available. This is a conceptual problem: the organization is stand-alone implementation-wise, while conceptually it belongs to a class. The clean solution could be to reroute this message to a class, but this does not work for all of the senders (that would work from the browserm but not for the file-in).

Browser>>#categorizeAllUncategorizedMethods
	Problem: should be trapped to send a ReCategorization event. However, this is model code that should not be in the Browser. Clean solution is to move it out of there to the model, and then trap it there (or reroute it to one of the trapped places).

Note: Debugger>>#contents:notifying: recompiles methods when needed, so I trapped it to get updates. However, I need to find a way to write a unit test for this. Haven't gotten around yet for doing this though...
"! !

!AbstractEvent class methodsFor: 'item kinds' stamp: 'fbs 9/26/2013 17:46'!
environmentKind

	^#environment! !

!AbstractEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:18'!
eventSelectorBlock

	^[:itemKind :changeKind | itemKind, changeKind, 'Event:']! !

!AbstractEvent class methodsFor: 'item kinds' stamp: 'rw 7/14/2003 11:41'!
expressionKind

	^#expression! !

!AbstractEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:19'!
itemChangeCombinations

	^self supportedKinds collect: [:itemKind | self eventSelectorBlock value: itemKind value: self changeKind]! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'rw 7/9/2003 11:19'!
item: anItem kind: anItemKind

	^self basicNew item: anItem kind: anItemKind! !

!AbstractEvent class methodsFor: 'item kinds' stamp: 'rw 7/9/2003 11:12'!
methodKind

	^#method! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'NS 1/16/2004 14:19'!
method: aMethod class: aClass

	| instance |
	instance := self item: aMethod kind: self methodKind.
	instance itemClass: aClass.
	^instance! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'NS 1/16/2004 14:20'!
method: aMethod protocol: prot class: aClass

	| instance |
	instance := self method: aMethod class: aClass.
	instance itemProtocol: prot.
	^instance! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'NS 1/27/2004 10:48'!
method: aMethod selector: aSymbol class: aClass

	| instance |
	instance := self item: aMethod kind: self methodKind.
	instance itemSelector: aSymbol.
	instance itemClass: aClass.
	^instance! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'NS 1/27/2004 10:49'!
method: aMethod selector: aSymbol class: aClass requestor: requestor

	| instance |
	instance := self method: aMethod selector: aSymbol class: aClass.
	instance itemRequestor: requestor.
	^instance! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'NS 1/27/2004 10:49'!
method: aMethod selector: aSymbol protocol: prot class: aClass

	| instance |
	instance := self method: aMethod selector: aSymbol class: aClass.
	instance itemProtocol: prot.
	^instance! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'NS 1/27/2004 10:50'!
method: aMethod selector: aSymbol protocol: prot class: aClass requestor: requestor

	| instance |
	instance := self method: aMethod selector: aSymbol protocol: prot class: aClass.
	instance itemRequestor: requestor.
	^instance! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'rw 6/30/2003 09:20'!
new
	"Override new to trigger an error, since we want to use specialized methods to create basic and higher-level events."

	^self error: 'Instances can only be created using specialized instance creation methods.'! !

!AbstractEvent class methodsFor: 'item kinds' stamp: 'rw 7/10/2003 12:36'!
protocolKind

	^#protocol! !

!AbstractEvent class methodsFor: 'instance creation' stamp: 'laza 11/2/2011 23:50'!
protocol: aProtocol class: aClass

	| instance |
	instance := self item: aProtocol kind: self protocolKind.
	instance itemClass: aClass.
	^instance! !

!AbstractEvent class methodsFor: 'temporary' stamp: 'ar 9/27/2005 20:05'!
saveChangeNotificationAsSARFileWithNumber: aNumber 
	"Use the SARBuilder package to output the SystemChangeNotification 
	stuff as a SAR file. Put this statement here so that I don't forget it 
	when moving between images :-)"
	"self saveChangeNotificationAsSARFileWithNumber: 6"

	| filename changesText readmeText dumper |
	filename := 'SystemchangeNotification'.
	dumper := self class environment at: #SARChangeSetDumper ifAbsent: [ ^self ].
	changesText := 
'
0.6 Version for Squeak 3.7 (no longer for 3.6!!!!) Changed one hook method to make this version work in Squeak3.7. Download version 5 from http://www.iam.unibe.ch/~wuyts/SystemchangeNotification5.sar if you are working with Squeak 3.6.

0.5 Updated the safeguard mechanism so that clients with halts and errors do not stop all notifications. Added and updated new tests for this. If this interests you have a look at the class WeakActionSequenceTrappingErrors.

0.4 Ported to Squeak 3.6.

0.3 Added the hooks for instance variables (addition, removal and renaming). Refactored the tests.

0.2 Added hooks and tests for method removal and method recategorization.

0.1 First release'.
	readmeText :=
'Implements (part of) the system change notification mechanism. Clients that want to receive notifications about system changes should look at the category #public of the class SystemChangeNotifier, and the unit tests.

VERY IMPORTANT: This version is for Squeak 3.7 only. It will not work in Squeak version 3.6. Download and install the last version that worked in Squeak 3.6 (version 5) from the following URL: http://www.iam.unibe.ch/~wuyts/SystemchangeNotification5.sar'.

	(dumper
		on: Project current changeSet
		including: (ChangeSet allChangeSetNames
				select: [:ea | 'SystemChangeHooks' match: ea])) changesText: changesText;
		 readmeText: readmeText;
		 fileOutAsZipNamed: filename , aNumber printString , '.sar'! !

!AbstractEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:04'!
supportedKinds
	"All the kinds of items that this event can take. By default this is all the kinds in the system. But subclasses can override this to limit the choices. For example, the SuperChangedEvent only works with classes, and not with methods, instance variables, ..."

	^self allItemKinds! !

!AbstractEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 11:39'!
systemEvents
	"Return all the possible events in the system. Make a cross product of 
	the items and the change types."
	"self systemEvents"

	^self allSubclasses
		inject: OrderedCollection new
		into: [:allEvents :eventClass | allEvents addAll: eventClass itemChangeCombinations; yourself]! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'rw 7/10/2003 12:10'!
changeKind

	^self class changeKind! !

!AbstractEvent methodsFor: 'accessing' stamp: 'fbs 9/2/2013 19:03'!
environment
	^ environment.! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'rw 7/10/2003 12:43'!
environmentAt: anItemKind

	(self itemKind = anItemKind) ifTrue: [^self item].
	^environment at: anItemKind ifAbsent: [nil]! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'rw 7/10/2003 12:20'!
eventSelector

	^self class eventSelectorBlock value: itemKind value: self changeKind! !

!AbstractEvent methodsFor: 'testing' stamp: 'rw 6/30/2003 08:34'!
isAdded

	^false! !

!AbstractEvent methodsFor: 'testing' stamp: 'NS 1/19/2004 18:41'!
isCategoryKnown

	^self itemCategory notNil! !

!AbstractEvent methodsFor: 'testing' stamp: 'rw 7/10/2003 15:01'!
isCommented

	^false! !

!AbstractEvent methodsFor: 'testing' stamp: 'rw 7/14/2003 10:15'!
isDoIt

	^false! !

!AbstractEvent methodsFor: 'testing' stamp: 'NS 1/19/2004 15:09'!
isModified

	^false! !

!AbstractEvent methodsFor: 'testing' stamp: 'NS 1/21/2004 09:40'!
isProtocolKnown

	^self itemCategory notNil! !

!AbstractEvent methodsFor: 'testing' stamp: 'rw 7/1/2003 19:53'!
isRecategorized

	^false! !

!AbstractEvent methodsFor: 'testing' stamp: 'rw 6/30/2003 08:34'!
isRemoved

	^false! !

!AbstractEvent methodsFor: 'testing' stamp: 'rw 7/1/2003 11:35'!
isRenamed

	^false! !

!AbstractEvent methodsFor: 'testing' stamp: 'NS 1/27/2004 12:44'!
isReorganized
	^ false! !

!AbstractEvent methodsFor: 'accessing' stamp: 'rw 6/30/2003 08:22'!
item
	"Return the item that triggered the event (typically the name of a class, a category, a protocol, a method)."

	^item! !

!AbstractEvent methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:43'!
itemCategory

	^self environmentAt: self class categoryKind! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'rw 7/10/2003 12:37'!
itemCategory: aCategory

	environment at: self class categoryKind put: aCategory! !

!AbstractEvent methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:43'!
itemClass

	^self environmentAt: self class classKind! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'rw 7/10/2003 12:36'!
itemClass: aClass

	environment at: self class classKind put: aClass! !

!AbstractEvent methodsFor: 'accessing' stamp: 'fbs 9/26/2013 17:58'!
itemEnvironment
	^ environment at: self class environmentKind ifAbsent: [Smalltalk globals].! !

!AbstractEvent methodsFor: 'accessing' stamp: 'fbs 9/26/2013 17:47'!
itemEnvironment: anEnvironment
	^self environmentAt: self class environmentKind put: anEnvironment.! !

!AbstractEvent methodsFor: 'accessing' stamp: 'rw 7/14/2003 12:10'!
itemExpression

	^self environmentAt: self class expressionKind! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'rw 7/14/2003 12:11'!
itemExpression: anExpression

	environment at: self class expressionKind put: anExpression! !

!AbstractEvent methodsFor: 'accessing' stamp: 'rw 6/30/2003 08:22'!
itemKind
	"Return the kind of the item of the event (#category, #class, #protocol, #method, ...)"

	^itemKind! !

!AbstractEvent methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:44'!
itemMethod

	^self environmentAt: self class methodKind! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'rw 7/10/2003 12:38'!
itemMethod: aMethod

	environment at: self class methodKind put: aMethod! !

!AbstractEvent methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:44'!
itemProtocol

	^self environmentAt: self class protocolKind! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'rw 7/10/2003 12:38'!
itemProtocol: aProtocol

	environment at: self class protocolKind put: aProtocol! !

!AbstractEvent methodsFor: 'accessing' stamp: 'NS 1/27/2004 10:38'!
itemRequestor

	^self environmentAt: #requestor! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'NS 1/27/2004 10:38'!
itemRequestor: requestor

	environment at: #requestor put: requestor! !

!AbstractEvent methodsFor: 'accessing' stamp: 'NS 1/27/2004 10:38'!
itemSelector

	^self environmentAt: #selector! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'NS 1/27/2004 10:39'!
itemSelector: aSymbol

	environment at: #selector put: aSymbol! !

!AbstractEvent methodsFor: 'private-accessing' stamp: 'rw 7/10/2003 12:36'!
item: anItem kind: anItemKind

	item := anItem.
	itemKind := anItemKind.
	environment := Dictionary new! !

!AbstractEvent methodsFor: 'printing' stamp: 'NS 1/19/2004 17:52'!
printOn: aStream

	self printEventKindOn: aStream.
	aStream
		nextPutAll: ' Event for item: ';
		print: self item;
		nextPutAll: ' of kind: ';
		print: self itemKind! !

!AbstractEvent methodsFor: 'triggering' stamp: 'rw 7/14/2003 17:06'!
trigger: anEventManager 
	"Trigger the event manager."

	anEventManager triggerEvent: self eventSelector with: self.! !

AbstractEvent subclass: #AddedEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!AddedEvent class methodsFor: 'accessing' stamp: 'rw 7/19/2003 09:52'!
changeKind

	^#Added! !

!AddedEvent class methodsFor: 'accessing' stamp: 'NS 1/20/2004 12:22'!
supportedKinds
	"All the kinds of items that this event can take."
	
	^ Array with: self classKind with: self methodKind with: self categoryKind with: self protocolKind! !

!AddedEvent methodsFor: 'testing' stamp: 'rw 6/30/2003 08:35'!
isAdded

	^true! !

!AddedEvent methodsFor: 'printing' stamp: 'rw 6/30/2003 09:31'!
printEventKindOn: aStream

	aStream nextPutAll: 'Added'! !

AbstractEvent subclass: #CommentedEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!CommentedEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:08'!
changeKind

	^#Commented! !

!CommentedEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 11:20'!
supportedKinds

	^Array with: self classKind! !

!CommentedEvent methodsFor: 'testing' stamp: 'rw 7/1/2003 11:37'!
isCommented

	^true! !

!CommentedEvent methodsFor: 'printing' stamp: 'rw 7/1/2003 11:37'!
printEventKindOn: aStream

	aStream nextPutAll: 'Commented'! !

AbstractEvent subclass: #DoItEvent
	instanceVariableNames: 'context'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!DoItEvent class methodsFor: 'accessing' stamp: 'rw 7/14/2003 10:19'!
changeKind

	^#DoIt! !

!DoItEvent class methodsFor: 'instance creation' stamp: 'NS 1/19/2004 09:47'!
expression: stringOrStream context: aContext
	| instance |
	instance := self item: stringOrStream kind: AbstractEvent expressionKind.
	instance context: aContext.
	^instance! !

!DoItEvent class methodsFor: 'accessing' stamp: 'NS 1/20/2004 12:23'!
supportedKinds
	^ Array with: self expressionKind! !

!DoItEvent methodsFor: 'accessing' stamp: 'rw 7/14/2003 11:29'!
context

	^context! !

!DoItEvent methodsFor: 'private-accessing' stamp: 'rw 7/14/2003 11:29'!
context: aContext

	context := aContext! !

!DoItEvent methodsFor: 'testing' stamp: 'rw 7/14/2003 10:15'!
isDoIt

	^true! !

!DoItEvent methodsFor: 'printing' stamp: 'rw 7/14/2003 10:15'!
printEventKindOn: aStream

	aStream nextPutAll: 'DoIt'! !

AbstractEvent subclass: #ModifiedEvent
	instanceVariableNames: 'oldItem'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

ModifiedEvent subclass: #ModifiedClassDefinitionEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!ModifiedClassDefinitionEvent class methodsFor: 'instance creation' stamp: 'NS 1/20/2004 11:52'!
classDefinitionChangedFrom: oldClass to: newClass
	| instance |
	instance := self item: newClass kind: self classKind.
	instance oldItem: oldClass.
	^instance! !

!ModifiedClassDefinitionEvent class methodsFor: 'accessing' stamp: 'NS 1/20/2004 12:26'!
supportedKinds
	"All the kinds of items that this event can take."
	
	^ Array with: self classKind! !

!ModifiedClassDefinitionEvent methodsFor: 'change logging' stamp: 'al 7/17/2004 21:48'!
anyChanges
	^ self isSuperclassModified or: [self areInstVarsModified or: [self areClassVarsModified or: [self areSharedPoolsModified or: [self isTraitCompositionModified]]]]! !

!ModifiedClassDefinitionEvent methodsFor: 'testing' stamp: 'NS 1/20/2004 19:31'!
areClassVarsModified
	^ self classVarNames ~= self oldClassVarNames! !

!ModifiedClassDefinitionEvent methodsFor: 'testing' stamp: 'NS 1/20/2004 19:30'!
areInstVarsModified
	^ self instVarNames ~= self oldInstVarNames! !

!ModifiedClassDefinitionEvent methodsFor: 'testing' stamp: 'NS 1/20/2004 19:32'!
areSharedPoolsModified
	^ self sharedPools ~= self oldSharedPools! !

!ModifiedClassDefinitionEvent methodsFor: 'accessing' stamp: 'nice 10/20/2009 20:49'!
classVarNames
	^ item classVarNames! !

!ModifiedClassDefinitionEvent methodsFor: 'accessing' stamp: 'NS 1/20/2004 19:29'!
instVarNames
	^ item instVarNames asSet! !

!ModifiedClassDefinitionEvent methodsFor: 'testing' stamp: 'NS 1/20/2004 19:29'!
isSuperclassModified
	^ item superclass ~~ oldItem superclass! !

!ModifiedClassDefinitionEvent methodsFor: 'trait composition' stamp: 'al 7/17/2004 21:59'!
isTraitCompositionModified
	^self traitComposition printString ~= self oldTraitComposition printString! !

!ModifiedClassDefinitionEvent methodsFor: 'accessing' stamp: 'nice 10/20/2009 20:49'!
oldClassVarNames
	^ oldItem classVarNames! !

!ModifiedClassDefinitionEvent methodsFor: 'accessing' stamp: 'NS 1/20/2004 19:29'!
oldInstVarNames
	^ oldItem instVarNames asSet! !

!ModifiedClassDefinitionEvent methodsFor: 'accessing' stamp: 'NS 1/20/2004 19:31'!
oldSharedPools
	^ oldItem sharedPools! !

!ModifiedClassDefinitionEvent methodsFor: 'accessing' stamp: 'NS 1/20/2004 19:28'!
oldSuperclass
	^ oldItem superclass! !

!ModifiedClassDefinitionEvent methodsFor: 'trait composition' stamp: 'al 7/17/2004 21:50'!
oldTraitComposition
	^ oldItem traitComposition! !

!ModifiedClassDefinitionEvent methodsFor: 'printing' stamp: 'al 7/18/2004 10:47'!
printOn: aStream
	super printOn: aStream.
	aStream
		nextPutAll: ' Super: ';
		print: self isSuperclassModified;
		nextPutAll: ' TraitComposition: ';
		print: self isTraitCompositionModified;
		nextPutAll: ' InstVars: ';
		print: self areInstVarsModified;
		nextPutAll: ' ClassVars: ';
		print: self areClassVarsModified;
		nextPutAll: ' SharedPools: ';
		print: self areSharedPoolsModified.! !

!ModifiedClassDefinitionEvent methodsFor: 'accessing' stamp: 'NS 1/20/2004 19:31'!
sharedPools
	^ item sharedPools! !

!ModifiedClassDefinitionEvent methodsFor: 'accessing' stamp: 'NS 1/20/2004 19:28'!
superclass
	^ item superclass! !

!ModifiedClassDefinitionEvent methodsFor: 'trait composition' stamp: 'al 7/17/2004 21:50'!
traitComposition
	^ item traitComposition! !

!ModifiedEvent class methodsFor: 'accessing' stamp: 'NS 1/19/2004 15:10'!
changeKind

	^#Modified! !

!ModifiedEvent class methodsFor: 'instance creation' stamp: 'NS 1/20/2004 19:37'!
classDefinitionChangedFrom: oldClass to: newClass
	^ ModifiedClassDefinitionEvent classDefinitionChangedFrom: oldClass to: newClass! !

!ModifiedEvent class methodsFor: 'instance creation' stamp: 'NS 1/27/2004 11:40'!
methodChangedFrom: oldMethod to: newMethod selector: aSymbol inClass: aClass
	| instance |
	instance := self method: newMethod selector: aSymbol class: aClass.
	instance oldItem: oldMethod.
	^ instance! !

!ModifiedEvent class methodsFor: 'instance creation' stamp: 'NS 1/27/2004 11:40'!
methodChangedFrom: oldMethod to: newMethod selector: aSymbol inClass: aClass requestor: requestor
	| instance |
	instance := self method: newMethod selector: aSymbol class: aClass requestor: requestor.
	instance oldItem: oldMethod.
	^ instance! !

!ModifiedEvent class methodsFor: 'accessing' stamp: 'NS 1/20/2004 12:25'!
supportedKinds
	"All the kinds of items that this event can take."
	
	^ Array with: self classKind with: self methodKind with: self categoryKind with: self protocolKind! !

!ModifiedEvent methodsFor: 'testing' stamp: 'NS 1/19/2004 15:09'!
isModified

	^true! !

!ModifiedEvent methodsFor: 'accessing' stamp: 'NS 1/19/2004 15:08'!
oldItem
	^ oldItem! !

!ModifiedEvent methodsFor: 'private-accessing' stamp: 'NS 1/19/2004 15:08'!
oldItem: anItem
	oldItem := anItem! !

!ModifiedEvent methodsFor: 'printing' stamp: 'NS 1/19/2004 15:10'!
printEventKindOn: aStream

	aStream nextPutAll: 'Modified'! !

!ModifiedEvent methodsFor: 'printing' stamp: 'NS 1/19/2004 17:57'!
printOn: aStream
	super printOn: aStream.
	aStream
		nextPutAll: ' oldItem: ';
		print: oldItem.! !

ModifiedEvent subclass: #ModifiedTraitDefinitionEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!ModifiedTraitDefinitionEvent class methodsFor: 'accessing' stamp: 'al 7/18/2004 10:43'!
supportedKinds
	"All the kinds of items that this event can take."
	
	^ Array with: self classKind! !

!ModifiedTraitDefinitionEvent class methodsFor: 'instance creation' stamp: 'al 7/18/2004 10:50'!
traitDefinitionChangedFrom: oldTrait to: newTrait
	| instance |
	instance := self item: newTrait kind: self classKind.
	instance oldItem: oldTrait.
	^instance! !

!ModifiedTraitDefinitionEvent methodsFor: 'testing' stamp: 'al 7/18/2004 11:08'!
anyChanges
	^ self isTraitCompositionModified! !

!ModifiedTraitDefinitionEvent methodsFor: 'testing' stamp: 'al 7/18/2004 10:43'!
isTraitCompositionModified
	^self traitComposition printString ~= self oldTraitComposition printString! !

!ModifiedTraitDefinitionEvent methodsFor: 'accessing' stamp: 'al 7/18/2004 10:43'!
oldTraitComposition
	^ oldItem traitComposition! !

!ModifiedTraitDefinitionEvent methodsFor: 'printing' stamp: 'al 7/18/2004 10:47'!
printOn: aStream
	super printOn: aStream.
	aStream
		nextPutAll: ' TraitComposition: ';
		print: self isTraitCompositionModified! !

!ModifiedTraitDefinitionEvent methodsFor: 'accessing' stamp: 'al 7/18/2004 10:43'!
traitComposition
	^ item traitComposition! !

AbstractEvent subclass: #RecategorizedEvent
	instanceVariableNames: 'oldCategory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!RecategorizedEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:09'!
changeKind

	^#Recategorized! !

!RecategorizedEvent class methodsFor: 'instance creation' stamp: 'rw 7/9/2003 14:21'!
class: aClass category: cat oldCategory: oldName

	^(self class: aClass category: cat) oldCategory: oldName! !

!RecategorizedEvent class methodsFor: 'instance creation' stamp: 'rw 7/31/2003 16:35'!
method: aMethod protocol: prot class: aClass oldProtocol: oldName

	^(self method: aMethod protocol: prot class: aClass) oldCategory: oldName! !

!RecategorizedEvent class methodsFor: 'instance creation' stamp: 'ar 9/5/2010 21:56'!
selector: aSelector method: aMethod protocol: prot class: aClass oldProtocol: oldName

	^(self method: aMethod protocol: prot class: aClass) 
		itemSelector: aSelector;
		oldCategory: oldName;
		yourself! !

!RecategorizedEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 11:20'!
supportedKinds

	^Array with: self classKind with: self methodKind! !

!RecategorizedEvent methodsFor: 'testing' stamp: 'rw 7/1/2003 19:51'!
isRecategorized

	^true! !

!RecategorizedEvent methodsFor: 'accessing' stamp: 'rw 7/1/2003 20:08'!
oldCategory

	^oldCategory! !

!RecategorizedEvent methodsFor: 'accessing' stamp: 'rw 7/1/2003 20:08'!
oldCategory: aCategoryName

	oldCategory := aCategoryName! !

!RecategorizedEvent methodsFor: 'printing' stamp: 'rw 7/2/2003 09:12'!
printEventKindOn: aStream

	aStream nextPutAll: 'Recategorized'! !

AbstractEvent subclass: #RemovedEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!RemovedEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:09'!
changeKind

	^#Removed! !

!RemovedEvent class methodsFor: 'accessing' stamp: 'NS 1/20/2004 12:28'!
supportedKinds
	"All the kinds of items that this event can take."
	
	^ Array with: self classKind with: self methodKind with: self categoryKind with: self protocolKind! !

!RemovedEvent methodsFor: 'testing' stamp: 'rw 6/30/2003 08:35'!
isRemoved

	^true! !

!RemovedEvent methodsFor: 'printing' stamp: 'rw 6/30/2003 09:31'!
printEventKindOn: aStream

	aStream nextPutAll: 'Removed'! !

AbstractEvent subclass: #RenamedEvent
	instanceVariableNames: 'newName oldName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!RenamedEvent class methodsFor: 'accessing' stamp: 'rw 7/10/2003 12:09'!
changeKind

	^#Renamed! !

!RenamedEvent class methodsFor: 'instance creation' stamp: 'ab 2/10/2005 16:49'!
classCategoryRenamedFrom: anOldClassCategoryName to: aNewClassCategoryName

	^(self classCategory: anOldClassCategoryName) oldName: anOldClassCategoryName; newName: aNewClassCategoryName! !

!RenamedEvent class methodsFor: 'instance creation' stamp: 'NS 1/27/2004 12:19'!
class: aClass category: cat oldName: oldName newName: newName

	^(self class: aClass category: cat) oldName: oldName; newName: newName! !

!RenamedEvent class methodsFor: 'accessing' stamp: 'NS 1/20/2004 12:30'!
supportedKinds
	"All the kinds of items that this event can take."
	
	^ Array with: self classKind with: self categoryKind with: self protocolKind! !

!RenamedEvent methodsFor: 'testing' stamp: 'rw 7/1/2003 11:34'!
isRenamed

	^true! !

!RenamedEvent methodsFor: 'accessing' stamp: 'NS 1/27/2004 12:18'!
newName
	^ newName! !

!RenamedEvent methodsFor: 'accessing' stamp: 'NS 1/27/2004 12:18'!
newName: aName

	newName := aName! !

!RenamedEvent methodsFor: 'accessing' stamp: 'rw 7/1/2003 12:00'!
oldName

	^oldName! !

!RenamedEvent methodsFor: 'accessing' stamp: 'rw 7/1/2003 12:01'!
oldName: aName

	oldName := aName! !

!RenamedEvent methodsFor: 'printing' stamp: 'rw 7/1/2003 11:34'!
printEventKindOn: aStream

	aStream nextPutAll: 'Renamed'! !

AbstractEvent subclass: #ReorganizedEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!ReorganizedEvent class methodsFor: 'accessing' stamp: 'NS 1/27/2004 12:46'!
changeKind

	^#Reorganized! !

!ReorganizedEvent class methodsFor: 'accessing' stamp: 'NS 1/27/2004 12:46'!
supportedKinds

	^Array with: self classKind! !

!ReorganizedEvent methodsFor: 'testing' stamp: 'NS 1/27/2004 12:44'!
isReorganized
	^ true! !

!ReorganizedEvent methodsFor: 'printing' stamp: 'NS 1/27/2004 12:44'!
printEventKindOn: aStream

	aStream nextPutAll: 'Reorganized'! !

Object subclass: #AbstractLauncher
	instanceVariableNames: 'parameters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!AbstractLauncher commentStamp: 'mt 8/14/2016 10:08' prior: 0!
The class AutoStart in combination with the Launcher classes provides a mechanism for starting Squeak from the command line or a web page. Parameters on the command line or in the embed tag in the web page a parsed and stored in the lauchner's parameter dictionary.
Subclasses can access these parameters to determine what to do.

CommandLineLauncherExample provides an example for a command line application. if you start squeak with a command line 'class Integer' it will launch a class browser on class Integer.
To enable this execute
CommandLineLauncherExample activate
before you save the image.
To disable execute
CommandLineLauncherExample deactivate

The PluginLauncher is an example how to use this framework to start Squeak as a browser plugin. It looks for a parameter 'src' which should point to a file containing a squeak script.!

!AbstractLauncher class methodsFor: 'activation' stamp: 'mir 8/6/1999 18:14'!
activate
	"Register this launcher with the auto start class"

	self autoStarter addLauncher: self! !

!AbstractLauncher class methodsFor: 'private' stamp: 'mir 8/4/1999 13:57'!
autoStarter
	^AutoStart! !

!AbstractLauncher class methodsFor: 'activation' stamp: ''!
deactivate
	"Unregister this launcher with the auto start class"
	self autoStarter removeLauncher: self! !

!AbstractLauncher methodsFor: 'private' stamp: 'jm 8/20/1999 15:33'!
commandLine: aString
	"Start up this launcher from within Squeak as if it Squeak been launched the given command line."

	| dict tokens cmd arg |
	dict := Dictionary new.
	tokens := ReadStream on: (aString findTokens: ' ').
	[cmd := tokens next.
	 arg := tokens next.
	 ((cmd ~~ nil) and: [arg ~~ nil])]
		whileTrue: [dict at: cmd put: arg].
	self parameters: dict.
	self startUp.
! !

!AbstractLauncher methodsFor: 'private' stamp: 'mir 9/23/1999 13:18'!
determineParameterNameFrom: alternateParameterNames
	"Determine which of the given alternate parameter names is actually used."

	^alternateParameterNames detect: [:each | self includesParameter: each asUppercase] ifNone: [nil] ! !

!AbstractLauncher methodsFor: 'private' stamp: 'mir 1/11/2000 16:35'!
includesParameter: parName
	"Return if the parameter named parName exists."
	^self parameters
		includesKey: parName asUppercase! !

!AbstractLauncher methodsFor: 'private' stamp: 'mir 8/4/1999 14:19'!
parameterAt: parName
	"Return the parameter named parName or an empty string"
	^self
		parameterAt: parName
		ifAbsent: ['']! !

!AbstractLauncher methodsFor: 'private' stamp: 'ul 8/2/2011 19:51'!
parameterAt: parName ifAbsent: aBlock
	"Return the parameter named parName.
	Evaluate the block if parameter does not exist."
	^self parameters
		at: parName asUppercase
		ifAbsent: aBlock! !

!AbstractLauncher methodsFor: 'private' stamp: 'cmm 10/12/2013 16:22'!
parameters
	parameters == nil
		ifTrue: [parameters := Smalltalk namedArguments].
	^parameters! !

!AbstractLauncher methodsFor: 'private' stamp: 'mir 7/29/1999 10:21'!
parameters: startupParameters
	parameters := startupParameters! !

!AbstractLauncher methodsFor: 'running' stamp: 'tk 10/24/2001 06:40'!
startUp
	"A backstop for subclasses.  Note that this is not a class message (most startUps are class messages)."

! !

AbstractLauncher subclass: #CommandLineLauncherExample
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Download'!

!CommandLineLauncherExample commentStamp: '<historical>' prior: 0!
CommandLineLauncherExample provides an example for a command line application. if you start squeak with a command line 'class Integer' it will launch a class browser on class Integer.
To enable this execute
CommandLineLauncherExample activate
before you save the image.
To disable execute
CommandLineLauncherExample deactivate!

!CommandLineLauncherExample methodsFor: 'running' stamp: 'ar 9/27/2005 20:23'!
startUp
	| className |
	className := self parameterAt: 'class'.
	ToolSet browse: (Smalltalk at: className asSymbol ifAbsent: [Object]) selector: nil! !

AbstractLauncher subclass: #ProjectLauncher
	instanceVariableNames: 'showSplash splashURL whichFlaps eToyAuthentificationServer'
	classVariableNames: 'SplashMorph'
	poolDictionaries: ''
	category: 'System-Download'!

!ProjectLauncher class methodsFor: 'accessing' stamp: 'ar 3/15/2001 23:32'!
splashMorph
	^SplashMorph! !

!ProjectLauncher class methodsFor: 'accessing' stamp: 'ar 3/15/2001 23:33'!
splashMorph: aMorph
	SplashMorph := aMorph.! !

!ProjectLauncher methodsFor: 'eToy login' stamp: 'ar 8/23/2001 22:04'!
cancelLogin
	"This is fine - we just proceed here. Later we may do something utterly different ;-)"
	^self proceedWithLogin! !

!ProjectLauncher methodsFor: 'eToy login' stamp: 'ar 1/3/2010 18:51'!
doEtoyLogin
	"Pop up the eToy login if we have a server that provides us with a known user list"

	"Find us a server who could do eToy authentification for us"
	eToyAuthentificationServer := 
		(ServerDirectory localProjectDirectories, ServerDirectory servers values)
			detect:[:any| any hasEToyUserList]
			ifNone:[nil].
	eToyAuthentificationServer "no server provides user information"
		ifNil:[^self startUpAfterLogin].
	self prepareForLogin.
	(Smalltalk at: #EtoyLoginMorph ifAbsent:[^self cancelLogin])
			loginAndDo:[:userName| self loginAs: userName]
			ifCanceled:[self cancelLogin].
! !

!ProjectLauncher methodsFor: 'running' stamp: 'mir 4/3/2001 15:15'!
hideSplashMorph
	SplashMorph ifNil:[^self].
	self showSplash
		ifFalse: [^self].
	SplashMorph delete.
	World submorphs do:[:m| m visible: true]. "show all"
! !

!ProjectLauncher methodsFor: 'initialization' stamp: 'mir 8/24/2001 20:25'!
initialize
	super initialize.
	showSplash := true.
	HTTPClient isRunningInBrowser
		ifTrue: [whichFlaps := 'etoy']! !

!ProjectLauncher methodsFor: 'running' stamp: 'ar 3/16/2001 12:42'!
installProjectFrom: loader
	self showSplashMorph.
	[[[
		loader installProject
	] on: ProjectViewOpenNotification
	  do:[:ex| ex resume: false] "no project view in plugin launcher"
	] on: ProgressInitiationException "no 'reading aStream' nonsense"
	  do:[:ex| ex sendNotificationsTo: [ :min :max :curr |]]
	] on: ProjectEntryNotification "hide splash morph when entering project"
       do:[:ex| self hideSplashMorph. ex pass].! !

!ProjectLauncher methodsFor: 'eToy login' stamp: 'KR 5/1/2008 01:58'!
loginAs: userName
	"Assuming that we have a valid user url; read its contents and see if the user is really there."
	| actualName userList |
	eToyAuthentificationServer ifNil:[
		self proceedWithLogin.
		^true].
	userList _ eToyAuthentificationServer eToyUserList.
	userList ifNil:[
		self inform:
'Sorry, I cannot find the user list.
(this may be due to a network problem)
Please hit Cancel if you wish to use Squeak.' translated.
		^false].
	"case insensitive search"
	actualName  _ userList detect:[:any| any sameAs: userName] ifNone:[nil].
	actualName isNil ifTrue:[
		self inform: 'Unknown user: ' translated ,userName.
		^false].
	Utilities authorName: actualName.
	eToyAuthentificationServer eToyUserName: actualName.
	self proceedWithLogin.
	^true! !

!ProjectLauncher methodsFor: 'eToy login' stamp: 'ar 8/23/2001 22:06'!
prepareForLogin
	"Prepare for login - e.g., hide everything so only the login morph is visible."
	World submorphsDo:[:m| 
		m isLocked ifFalse:[m hide]]. "hide all those guys"
	World displayWorldSafely.
! !

!ProjectLauncher methodsFor: 'eToy login' stamp: 'ar 8/24/2001 15:17'!
proceedWithLogin
	eToyAuthentificationServer := nil.
	World submorphsDo:[:m| m show].
	WorldState addDeferredUIMessage: [self startUpAfterLogin].! !

!ProjectLauncher methodsFor: 'initialization' stamp: 'mir 8/24/2001 15:51'!
setupFlaps
	"Only called when the image has been launched in a browser.  If I am requested to show etoy flaps, then remove any pre-existing shared flaps and put up the supplies flap only.  if I am requested to show all flaps, then if flaps already exist, use them as is, else set up to show the default set of standard flaps."

	((whichFlaps = 'etoy')
		or: [Preferences eToyFriendly])
		ifTrue:
			[Flaps addAndEnableEToyFlaps].
	whichFlaps = 'all'
		ifTrue: [Flaps sharedFlapsAllowed
				ifFalse: [Flaps enableGlobalFlaps]]! !

!ProjectLauncher methodsFor: 'initialization' stamp: 'mir 8/23/2002 14:52'!
setupFromParameters
	(self includesParameter: 'showSplash')
		ifTrue: [showSplash := (self parameterAt: 'showSplash') asUppercase = 'TRUE'].
	(self includesParameter: 'flaps')
		ifTrue: [whichFlaps := (self parameterAt: 'flaps')].
! !

!ProjectLauncher methodsFor: 'private' stamp: 'mir 4/3/2001 15:15'!
showSplash
	^showSplash! !

!ProjectLauncher methodsFor: 'running' stamp: 'mir 4/3/2001 15:15'!
showSplashMorph
	SplashMorph ifNil:[^self].
	self showSplash
		ifFalse: [^self].
	World submorphs do:[:m| m visible: false]. "hide all"
	World addMorphCentered: SplashMorph.
	World displayWorldSafely.! !

!ProjectLauncher methodsFor: 'running' stamp: 'mt 8/14/2016 10:03'!
startUp

	Preferences eToyLoginEnabled
		ifTrue: [self doEtoyLogin]
		ifFalse:[self startUpAfterLogin].! !

!ProjectLauncher methodsFor: 'running' stamp: 'fbs 1/17/2013 13:36'!
startUpAfterLogin
	| scriptName loader isUrl |
	self setupFlaps.
	Preferences readDocumentAtStartup ifTrue: [
		HTTPClient isRunningInBrowser ifTrue:[
			self setupFromParameters.
			scriptName := self parameterAt: 'src'.
			CodeLoader defaultBaseURL: (self parameterAt: 'Base').
		] ifFalse:[
			scriptName := (Smalltalk documentPath) ifNil:[''].
			scriptName := scriptName convertFromSystemString.
			scriptName isEmpty ifFalse:[
				"figure out if script name is a URL by itself"
				isUrl := (scriptName asLowercase beginsWith:'http://') or:[
						(scriptName asLowercase beginsWith:'file://') or:[
						(scriptName asLowercase beginsWith:'ftp://')]].
				isUrl ifFalse:[ | encodedPath pathTokens |
					"Allow for ../dir/scriptName arguments"
					pathTokens := scriptName splitBy: FileDirectory slash.
					pathTokens := pathTokens collect: [:s | s encodeForHTTP].
					encodedPath := pathTokens reduce: [:acc :each | acc , FileDirectory slash , each].
					scriptName := (FileDirectory default uri 
						resolveRelativeURI: encodedPath) asString]].
		]. ]
	ifFalse: [ scriptName := '' ].

	scriptName isEmptyOrNil
		ifTrue:[^Preferences eToyFriendly ifTrue: [self currentWorld addGlobalFlaps]].
	loader := CodeLoader new.
	loader loadSourceFiles: (Array with: scriptName).
	(scriptName asLowercase endsWith: '.pr') 
		ifTrue:[self installProjectFrom: loader]
		ifFalse:[loader installSourceFiles].
! !

Object subclass: #AbstractSoundSystem
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!AbstractSoundSystem commentStamp: 'gk 2/24/2004 08:34' prior: 0!
This is the abstract base class for a sound system. A sound system offers a small protocol for playing sounds and making beeps and works like a facade towards the rest of Squeak. A sound system is registered in the application registry SoundService and can be accessed by "SoundService default" like for example:

SoundService default playSoundNamed: 'croak'

The idea is that as much sound playing as possible should go through this facade. This way we decouple the sound system from the rest of Squeak and make it pluggable. It also is a perfect spot to check for the Preference class>>soundsEnabled.

Two basic subclasses exist at the time of this writing, the BaseSoundSystem which represents the standard Squeak sound system, and the DummySoundSystem which is a dummy implementation that can be used when there is no sound card available, or when the base sound system isn't in the image, or when you simply don't want to use the available sound card.!

!AbstractSoundSystem methodsFor: 'playing' stamp: 'gk 2/24/2004 23:27'!
beep
	"Make a primitive beep."

	self subclassResponsibility! !

!AbstractSoundSystem methodsFor: 'playing' stamp: 'gk 2/23/2004 19:47'!
playSampledSound: samples rate: rate

	self subclassResponsibility! !

!AbstractSoundSystem methodsFor: 'playing' stamp: 'gk 2/23/2004 19:51'!
playSoundNamedOrBeep: soundName

	self subclassResponsibility! !

!AbstractSoundSystem methodsFor: 'playing' stamp: 'gk 2/23/2004 19:50'!
playSoundNamed: soundName

	self subclassResponsibility! !

!AbstractSoundSystem methodsFor: 'playing' stamp: 'gk 2/23/2004 19:51'!
playSoundNamed: soundName ifAbsentReadFrom: aifFileName

	self subclassResponsibility! !

!AbstractSoundSystem methodsFor: 'playing' stamp: 'mt 4/6/2016 09:37'!
playSound: sound

	self subclassResponsibility.! !

!AbstractSoundSystem methodsFor: 'misc' stamp: 'gk 2/23/2004 19:52'!
randomBitsFromSoundInput: bitCount

	self subclassResponsibility! !

!AbstractSoundSystem methodsFor: 'misc' stamp: 'gk 2/23/2004 19:52'!
sampledSoundChoices

	self subclassResponsibility! !

!AbstractSoundSystem methodsFor: 'misc' stamp: 'gk 2/23/2004 19:53'!
shutDown
	"Default is to do nothing."! !

!AbstractSoundSystem methodsFor: 'misc' stamp: 'gk 2/23/2004 19:56'!
soundNamed: soundName

	self subclassResponsibility! !

AbstractSoundSystem subclass: #DummySoundSystem
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!DummySoundSystem commentStamp: 'gk 2/24/2004 23:14' prior: 0!
This is a dummy sound system registered in SoundService to absorb all sound playing and to use the primitive beep instead of sampled sounds when playing a beep.!

!DummySoundSystem class methodsFor: 'class initialization' stamp: 'topa 4/2/2015 10:37'!
initialize
	SoundService register: self.! !

!DummySoundSystem class methodsFor: 'class initialization' stamp: 'topa 4/2/2015 10:37'!
unload
	SoundService unregister: self.! !

!DummySoundSystem methodsFor: 'playing' stamp: 'gk 2/24/2004 23:53'!
beep
	"Make a primitive beep."

	Beeper beepPrimitive! !

!DummySoundSystem methodsFor: 'playing' stamp: 'gk 2/23/2004 19:53'!
playSampledSound: samples rate: rate
	"Do nothing."
	! !

!DummySoundSystem methodsFor: 'playing' stamp: 'gk 4/8/2005 14:15'!
playSoundNamedOrBeep: soundName 
	"There is no sound support, so we make the beep."

	self beep! !

!DummySoundSystem methodsFor: 'playing' stamp: 'gk 2/23/2004 19:54'!
playSoundNamed: soundName
	"Do nothing."! !

!DummySoundSystem methodsFor: 'playing' stamp: 'gk 2/23/2004 19:54'!
playSoundNamed: soundName ifAbsentReadFrom: aifFileName
	"Do nothing."! !

!DummySoundSystem methodsFor: 'playing' stamp: 'mt 4/6/2016 09:38'!
playSound: sound
	"Do nothing."! !

!DummySoundSystem methodsFor: 'misc' stamp: 'gk 2/23/2004 20:48'!
randomBitsFromSoundInput: bitCount
	"I'm not sure what the right thing to do here is."

	self error: 'Can not provide random data.'! !

!DummySoundSystem methodsFor: 'misc' stamp: 'gk 2/23/2004 19:54'!
sampledSoundChoices
	"No choices other than this."

	^ #('silence')! !

!DummySoundSystem methodsFor: 'misc' stamp: 'gk 2/23/2004 19:55'!
soundNamed: soundName
	"There are no sounds to look up."

	^ nil! !

Object subclass: #AppRegistry
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Applications'!

AppRegistry class
	instanceVariableNames: 'registeredClasses default'!

!AppRegistry commentStamp: 'ads 4/2/2003 15:30' prior: 0!
AppRegistry is a simple little class, not much more than a wrapper around a collection. It's intended to help break dependencies between packages. For example, if you'd like to be able to send e-mail, you could use the bare-bones MailComposition class, or you could use the full-blown Celeste e-mail client. Instead of choosing one or the other, you can call "MailSender default" (where MailSender is a subclass of AppRegistry), and thus avoid creating a hard-coded dependency on either of the two mail senders.

This will only really be useful, of course, for applications that have a very simple, general, well-defined interface. Most of the time, you're probably better off just marking your package as being dependent on a specific other package, and avoiding the hassle of this whole AppRegistry thing. But for simple things like e-mail senders or web browsers, it might be useful.
!

AppRegistry class
	instanceVariableNames: 'registeredClasses default'!

!AppRegistry class methodsFor: 'defaults' stamp: 'ads 3/29/2003 13:36'!
appName
	"Defaults to the class name, which is probably good enough, but you could override this in subclasses if you want to."
	^ self name! !

!AppRegistry class methodsFor: 'defaults' stamp: 'cwp 9/28/2009 09:20'!
askForDefault

	self registeredClasses isEmpty ifTrue:
		[self inform: 'There are no ', self appName, ' applications registered.'.
		^ default := nil].
	self registeredClasses size = 1 ifTrue:
		[^ default := self registeredClasses anyOne].
	default :=  UIManager default 
		chooseFrom: (self registeredClasses collect: [:ea | ea nameForViewer])
		values: self registeredClasses
		title: 'Which ', self appName, ' would you prefer?'.
	default ifNil: [default := self registeredClasses first].
	^default.! !

!AppRegistry class methodsFor: 'class initialization' stamp: 'ar 2/27/2010 11:38'!
cleanUp
	"AppRegistry cleanUp"

	self removeObsolete.! !

!AppRegistry class methodsFor: 'accessing' stamp: 'ads 3/29/2003 13:11'!
default
	^ default ifNil: [self askForDefault]! !

!AppRegistry class methodsFor: 'accessing' stamp: 'nk 3/9/2004 12:35'!
defaultOrNil
	^ default! !

!AppRegistry class methodsFor: 'accessing' stamp: 'nk 3/9/2004 12:33'!
default: aClassOrNil
	"Sets my default to aClassOrNil. 
	Answers the old default."
	| oldDefault |
	oldDefault := default.
	aClassOrNil ifNotNil: [ self register: aClassOrNil ].
	default := aClassOrNil.
	^ oldDefault! !

!AppRegistry class methodsFor: 'registration' stamp: 'ads 3/29/2003 13:01'!
registeredClasses
	^ registeredClasses ifNil: [registeredClasses := OrderedCollection new]! !

!AppRegistry class methodsFor: 'registration' stamp: 'ads 4/2/2003 15:25'!
register: aProviderClass
	(self registeredClasses includes: aProviderClass) ifFalse:
		[default := nil.  "so it'll ask for a new default, since if you're registering a new app you probably want to use it"
		self registeredClasses add: aProviderClass].! !

!AppRegistry class methodsFor: 'registration' stamp: 'ar 9/27/2005 21:48'!
removeObsolete
	"AppRegistry removeObsoleteClasses"
	self registeredClasses copy do:[:cls| 
		(cls class isObsolete or:[cls isBehavior and:[cls isObsolete]]) 
			ifTrue:[self unregister: cls]].
	self subclasses do:[:cls| cls removeObsolete].! !

!AppRegistry class methodsFor: 'registration' stamp: 'ads 3/29/2003 13:03'!
unregister: aProviderClass
	(default = aProviderClass) ifTrue: [default := nil].
	self registeredClasses remove: aProviderClass ifAbsent: [].! !

AppRegistry subclass: #MailSender
	instanceVariableNames: ''
	classVariableNames: 'SmtpServer UserName'
	poolDictionaries: ''
	category: 'System-Applications'!

!MailSender class methodsFor: 'testing' stamp: 'dvf 5/11/2002 01:31'!
isSmtpServerSet
	^ SmtpServer notNil and: [SmtpServer notEmpty]
! !

!MailSender class methodsFor: 'communication' stamp: 'ads 5/11/2003 21:11'!
sendMessage: aMailMessage

	self default ifNotNil: [self default sendMailMessage: aMailMessage]! !

!MailSender class methodsFor: 'settings' stamp: 'rbb 3/1/2005 10:59'!
setSmtpServer
	"Set the SMTP server used to send outgoing messages via"
	SmtpServer ifNil: [SmtpServer := ''].
	SmtpServer := UIManager default
		request: 'What is your mail server for outgoing mail?'
		initialAnswer: SmtpServer.
! !

!MailSender class methodsFor: 'settings' stamp: 'nice 8/27/2010 20:43'!
setUserName
	"Change the user's email name for use in composing messages."

	(UIManager default
			request: 'What is your email address?\(This is the address other people will reply to you)' withCRs
			initialAnswer: (UserName ifNil: [UserName := '']))
		ifNotNil: [:answer | UserName := answer]! !

!MailSender class methodsFor: 'accessing' stamp: 'dvf 5/11/2002 01:29'!
smtpServer
	"Answer the server for sending email"

	self isSmtpServerSet
		ifFalse: [self setSmtpServer].
	SmtpServer isEmpty ifTrue: [
		self error: 'no SMTP server specified' ].

	^SmtpServer! !

!MailSender class methodsFor: 'accessing' stamp: 'dvf 5/11/2002 00:49'!
userName
	"Answer the user name to be used in composing messages."

	(UserName isNil or: [UserName isEmpty])
		ifTrue: [self setUserName].

	UserName isEmpty ifTrue: [ self error: 'no user name specified' ].

	^UserName! !

AppRegistry subclass: #SoundService
	instanceVariableNames: ''
	classVariableNames: 'SoundEnabled'
	poolDictionaries: ''
	category: 'System-Applications'!

!SoundService commentStamp: 'gk 2/24/2004 23:14' prior: 0!
This is the AppRegistry class for the sound system.

A sound system offers a small protocol for playing sounds and making beeps and works like a facade towards the rest of Squeak. A sound system is registered in this registry and can be accessed by "SoundService default". This way we decouple the sound system from the rest of Squeak and make it pluggable. It also is a perfect spot to check for the Preference class>>soundsEnabled.!

!SoundService class methodsFor: 'playing' stamp: 'mt 4/6/2016 10:20'!
beep

	self soundEnabled ifFalse: [^ self].

	self defaultOrNil ifNotNil: [:soundSystem | soundSystem beep].! !

!SoundService class methodsFor: 'accessing' stamp: 'topa 4/2/2015 10:36'!
default

	| cls |
	cls := super default.
	^ cls new! !

!SoundService class methodsFor: 'accessing' stamp: 'topa 4/2/2015 10:51'!
defaultOrNil

	^ super defaultOrNil ifNotNil: [:default |
		default isBehavior ifTrue: [default new] ifFalse: [default]]
! !

!SoundService class methodsFor: 'playing' stamp: 'mt 4/6/2016 12:54'!
playSound: sound

	self soundEnabled ifFalse: [^ self].

	self defaultOrNil ifNotNil: [:soundSystem |
		soundSystem playSound: sound].! !

!SoundService class methodsFor: 'preferences' stamp: 'bf 7/11/2011 11:10'!
soundEnabled
	<preference: 'Enable sound'
		category: 'media'
		description: 'If false, all sound playing is disabled'
		type: #Boolean>
	^SoundEnabled ifNil: [true]! !

!SoundService class methodsFor: 'preferences' stamp: 'HenrikSperreJohansen 1/21/2011 18:33'!
soundEnabled: aBoolean
	SoundEnabled := aBoolean! !

!SoundService class methodsFor: 'preferences' stamp: 'HenrikSperreJohansen 1/21/2011 18:40'!
soundEnablingString
	^ self soundEnabled
		ifFalse:
			['turn sound on' translated]
		ifTrue:
			['turn sound off' translated]! !

!SoundService class methodsFor: 'playing' stamp: 'mt 4/6/2016 10:01'!
stop

	self defaultOrNil ifNotNil: [:soundSystem |
		soundSystem shutDown].! !

!SoundService class methodsFor: 'preferences' stamp: 'HenrikSperreJohansen 1/21/2011 18:36'!
toggleSoundEnabled
	self soundEnabled: self soundEnabled not! !

AppRegistry subclass: #ToolSet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Applications'!

!ToolSet commentStamp: 'ar 7/15/2005 17:58' prior: 0!
ToolSet defines an interface that clients can use to request programmer facilities such as browsers, inspectors, debuggers, message sets etc.!

!ToolSet class methodsFor: 'accessing' stamp: 'ar 7/17/2005 10:47'!
askForDefault
	"Ask for the default implementor"
	self registeredClasses isEmpty 
		ifTrue:[^ default := nil].
	self registeredClasses size = 1 
		ifTrue:[^ default := self registeredClasses anyOne].
	default := UIManager default 
		chooseFrom: (self registeredClasses collect:[:each| each name printString])
		values: self registeredClasses
		title: 'Which ', self appName, ' would you prefer?'.
	^default.! !

!ToolSet class methodsFor: 'inspecting' stamp: 'ar 7/17/2005 10:54'!
basicInspect: anObject
	"Open an inspector on the given object. The tool set must know which inspector type to use for which object - the object cannot possibly know what kind of inspectors the toolset provides."
	self default ifNil:[^self inform: 'Cannot inspect -- no Inspector present'].
	^self default basicInspect: anObject! !

!ToolSet class methodsFor: 'browsing' stamp: 'ar 7/17/2005 10:54'!
browseChangeSetsWithClass: aClass selector: aSelector
	"Browse all the change sets with the given class/selector"
	self default ifNil:[^self inform: 'No ChangeSorter present'].
	^self default browseChangeSetsWithClass: aClass selector: aSelector! !

!ToolSet class methodsFor: 'browsing' stamp: 'fbs 9/13/2013 17:27'!
browseClass: aClass
	"Open a browser"
	self default ifNil:[^self inform: 'Cannot open Browser'].
	^self default browseClass: aClass.! !

!ToolSet class methodsFor: 'browsing' stamp: 'mt 1/18/2016 08:47'!
browseClass: aClass category: aCategory
	"Open a browser"
	self default ifNil:[^self inform: 'Cannot open Browser'].
	^self default browseClass: aClass category: aCategory! !

!ToolSet class methodsFor: 'browsing' stamp: 'ar 7/17/2005 11:14'!
browseHierarchy: aClass selector: aSelector
	"Open a browser"
	self default ifNil:[^self inform: 'No browser present'].
	^self default browseHierarchy: aClass selector: aSelector! !

!ToolSet class methodsFor: 'browsing' stamp: 'mt 1/18/2016 09:01'!
browseMessageCategory: aCategory inClass: aClass
	"Open a message category browser."
	self default ifNil:[^self inform: 'Cannot open Browser'].
	^self default browseMessageCategory: aCategory inClass: aClass! !

!ToolSet class methodsFor: 'browsing' stamp: 'ar 7/17/2005 11:12'!
browseMessageNames: aString
	"Open a MessageNames browser"
	self default ifNil:[^self inform: 'No MessageNames present'].
	^self default browseMessageNames: aString! !

!ToolSet class methodsFor: 'browsing' stamp: 'ar 7/17/2005 11:13'!
browseMessageSet: messageList name: title autoSelect: autoSelectString
	"Open a message set browser"
	self default ifNil:[^self inform: 'Cannot open MessageSet'].
	^self default browseMessageSet: messageList name: title autoSelect: autoSelectString! !

!ToolSet class methodsFor: 'browsing' stamp: 'ar 7/17/2005 10:54'!
browseVersionsOf: aClass selector: aSelector
	"Open a browser"
	self default ifNil:[^self inform: 'Cannot open Browser'].
	^self default browseVersionsOf: aClass selector: aSelector! !

!ToolSet class methodsFor: 'browsing' stamp: 'ar 7/17/2005 10:54'!
browse: aClass selector: aSelector
	"Open a browser"
	self default ifNil:[^self inform: 'Cannot open Browser'].
	^self default browse: aClass selector: aSelector! !

!ToolSet class methodsFor: 'completion' stamp: 'damiencassou 7/29/2009 16:24'!
codeCompletionAround: aBlock textMorph: aTextMorph keyStroke: evt
	self default ifNil: [aBlock value. ^ self].
	self default codeCompletionAround: aBlock textMorph: aTextMorph keyStroke: evt! !

!ToolSet class methodsFor: 'debugging' stamp: 'ar 7/17/2005 10:37'!
debugContext: aContext label: aString contents: contents
	"Open a debugger on the given context."
	self default ifNil:[
		(self confirm: 'Debugger request -- proceed?')
			ifFalse:[Processor terminateActive].
		^self].
	^self default debugContext: aContext label: aString contents: contents! !

!ToolSet class methodsFor: 'debugging' stamp: 'jrd 5/23/2009 00:54'!
debugError: anError
	"Handle an otherwise unhandled error"
	self default ifNil:[ | ctx |
		Smalltalk 
			logSqueakError: anError description 
			inContext: (ctx := anError signalerContext) .
		self inform: (anError description, String cr, ctx shortStack).
		^anError return].
	^self default debugError: anError! !

!ToolSet class methodsFor: 'debugging' stamp: 'mt 4/11/2016 10:29'!
debugMethod: aCompiledMethod forReceiver: anObject inContext: aContext

	self default ifNil:[^ self inform: 'Cannot debug method.'].
	^self default debugMethod: aCompiledMethod forReceiver: anObject inContext: aContext
! !

!ToolSet class methodsFor: 'debugging' stamp: 'ar 9/27/2005 19:12'!
debugSyntaxError: anError
	"Handle a syntax error"
	self default ifNil:[^self debugError: anError]. "handle as usual error"
	^self default debugSyntaxError: anError! !

!ToolSet class methodsFor: 'debugging' stamp: 'ar 7/17/2005 10:37'!
debug: aProcess context: aContext label: aString contents: contents fullView: aBool
	"Open a debugger on the given process and context."
	self default ifNil:[
		(self confirm: 'Debugger request -- proceed?')
			ifFalse:[Processor terminateActive].
		^self].
	^self default debug: aProcess context: aContext label: aString contents: contents fullView: aBool! !

!ToolSet class methodsFor: 'inspecting' stamp: 'al 11/28/2005 20:32'!
explore: anObject
	"Open an explorer on the given object."
	self default ifNil:[^self inform: 'Cannot explore - no ToolSet present'].
	^self default explore: anObject! !

!ToolSet class methodsFor: 'inspecting' stamp: 'ar 7/17/2005 10:43'!
inspectorClassOf: anObject
	"Answer the inspector class for the given object. The tool set must know which inspector type to use for which object - the object cannot possibly know what kind of inspectors the toolset provides."
	self default ifNil:[^nil].
	^self default inspectorClassOf: anObject! !

!ToolSet class methodsFor: 'inspecting' stamp: 'ar 7/17/2005 10:55'!
inspect: anObject
	"Open an inspector on the given object. The tool set must know which inspector type to use for which object - the object cannot possibly know what kind of inspectors the toolset provides."
	self default ifNil:[^self inform: 'Cannot inspect - no ToolSet present'].
	^self default inspect: anObject! !

!ToolSet class methodsFor: 'inspecting' stamp: 'ar 7/17/2005 10:55'!
inspect: anObject label: aString
	"Open an inspector on the given object. The tool set must know which inspector type to use for which object - the object cannot possibly know what kind of inspectors the toolset provides."
	self default ifNil:[^self inform: 'Cannot inspect - no ToolSet present'].
	^self default inspect: anObject label: aString! !

!ToolSet class methodsFor: 'debugging' stamp: 'ar 7/17/2005 10:37'!
interrupt: aProcess label: aString
	"Open a debugger on the given process and context."
	self default ifNil:[
		(self confirm: 'Debugger request -- proceed?')
			ifFalse:[aProcess terminate].
		^self].
	^self default interrupt: aProcess label: aString! !

!ToolSet class methodsFor: 'menu' stamp: 'ar 7/17/2005 13:19'!
menuItems
	"Answer the menu items available for this tool set"
	self default ifNil:[^#()].
	^self default menuItems! !

!ToolSet class methodsFor: 'browsing' stamp: 'ar 7/17/2005 10:44'!
openChangedMessageSet: aChangeSet
	"Open a ChangedMessageSet for aChangeSet"
	self default ifNil:[^self inform: 'Cannot open MessageSet'].
	^self default openChangedMessageSet: aChangeSet! !

!ToolSet class methodsFor: 'browsing' stamp: 'ar 7/17/2005 10:44'!
openClassListBrowser: anArray title: aString
	"Open a class list browser"
	self default ifNil:[^self inform: 'Cannot open ClassListBrowser'].
	^self default openClassListBrowser: anArray title: aString! !

AppRegistry subclass: #WebBrowser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Applications'!

Object subclass: #AutoStart
	instanceVariableNames: 'parameters'
	classVariableNames: 'Active InstalledLaunchers'
	poolDictionaries: ''
	category: 'System-Support'!

!AutoStart class methodsFor: 'private' stamp: 'mir 9/7/2004 13:34'!
active
	^ Active == true! !

!AutoStart class methodsFor: 'private' stamp: 'mir 9/7/2004 13:36'!
active: aBoolean
	Active := aBoolean! !

!AutoStart class methodsFor: 'accessing' stamp: ''!
addLauncherFirst: launcher
	self installedLaunchers addFirst: launcher! !

!AutoStart class methodsFor: 'accessing' stamp: 'mir 7/28/1999 17:47'!
addLauncher: launcher
	self installedLaunchers add: launcher! !

!AutoStart class methodsFor: 'updating' stamp: 'mt 8/14/2016 09:55'!
checkForPluginUpdate
	| pluginVersion updateURL |
	HTTPClient isRunningInBrowser
		ifFalse: [^false].
	pluginVersion := Smalltalk namedArguments
		at: (Smalltalk platformName copyWithout: Character space) asUppercase
		ifAbsent: [^false].
	updateURL := Smalltalk namedArguments
		at: 'UPDATE_URL'
		ifAbsent: [^false].
	^SystemVersion check: pluginVersion andRequestPluginUpdate: updateURL! !

!AutoStart class methodsFor: 'class initialization' stamp: 'mir 7/28/1999 17:44'!
deinstall
	"AutoStart deinstall"

	Smalltalk removeFromStartUpList: AutoStart.
	InstalledLaunchers := nil! !

!AutoStart class methodsFor: 'class initialization' stamp: 'mir 9/30/2004 15:05'!
initialize
	"AutoStart initialize"
	"Order: ExternalSettings, SecurityManager, AutoStart"
	Smalltalk addToStartUpList: AutoStart after: SecurityManager.
	Smalltalk addToShutDownList: AutoStart after: SecurityManager.! !

!AutoStart class methodsFor: 'private' stamp: 'mir 7/28/1999 17:43'!
installedLaunchers
	InstalledLaunchers ifNil: [
		InstalledLaunchers := OrderedCollection new].
	^InstalledLaunchers! !

!AutoStart class methodsFor: 'updating' stamp: 'fbs 12/10/2013 18:09'!
processUpdates
	"Process update files from a well-known update server.  This method is called at system startup time,   Only if the preference #updateFromServerAtStartup is true is the actual update processing undertaken automatically"
	| choice |
	(MCMcmUpdater updateFromServerAtStartup) ifTrue:
		[choice := UIManager default chooseFrom: #('Yes, Update' 'No, Not now' 'Don''t ask again')
			title: 'Shall I look for new code\updates on the server?' withCRs.
		choice = 1 ifTrue: [
			MCMcmUpdater updateFromServer].
		choice = 3 ifTrue: [
			MCMcmUpdater updateFromServerAtStartup: false.
			self inform: 'Remember to save your image to make this setting permant.']].
	^false! !

!AutoStart class methodsFor: 'accessing' stamp: 'mir 7/28/1999 17:47'!
removeLauncher: launcher
	self installedLaunchers remove: launcher ifAbsent: []! !

!AutoStart class methodsFor: 'class initialization' stamp: 'mir 9/30/2004 15:06'!
shutDown: quitting
	self active: false! !

!AutoStart class methodsFor: 'initialization' stamp: 'mt 8/14/2016 09:56'!
startUp: resuming
	"The image is either being newly started (resuming is true), or it's just been snapshotted.
	If this has just been a snapshot, skip all the startup stuff."

	| startupParameters launchers |
	self active ifTrue: [^self].
	self active: true.
	resuming ifFalse: [^self].

	HTTPClient determineIfRunningInBrowser.
	startupParameters := Smalltalk namedArguments.
	(startupParameters includesKey: 'apiSupported' asUppercase )
		ifTrue: [
			HTTPClient browserSupportsAPI: ((startupParameters at: 'apiSupported' asUppercase) asUppercase = 'TRUE').
			HTTPClient isRunningInBrowser
				ifFalse: [HTTPClient isRunningInBrowser: true]].

	World ifNotNil: [:w | w install. w firstHand position: 100 @ 100 ].
			
	"Some images might not have the UpdateStream package."
	((self respondsTo: #checkForUpdates) and: [self checkForUpdates]) ifTrue: [^self].
	self checkForPluginUpdate.
	launchers := self installedLaunchers collect: [:launcher |
		launcher new].
	launchers do: [:launcher |
		launcher parameters: startupParameters].
	launchers do: [:launcher |
		Project current addDeferredUIMessage: [launcher startUp]]! !

Object subclass: #Beeper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

Beeper class
	instanceVariableNames: 'default'!

!Beeper commentStamp: 'gk 2/26/2004 22:44' prior: 0!
Beeper provides simple audio (or in some other way) feedback to the user.

The recommended use is "Beeper beep" to give the user the equivalence of a beep. If you want to force the beep to use the primitive in the VM for beeping, then use "Beeper beepPrimitive". In either case, if sounds are disabled there will be no beep.

The actual beeping, when you use "Beeper beep", is done by sending a #play message to a registered playable object. You can register your own playable object by invoking the class side method #setDefault: passing in an object that responds to the #play message.

The default playable object is an instance of Beeper itself which implements #play on the instance side. That implementation delegates the playing of the beep to the default SoundService.

Note that #play is introduced as a common interface between AbstractSound and Beeper.
This way we can register instances of AbstractSound as playable entities, for example:

	Beeper setDefault: (SampledSound new
						setSamples: self coffeeCupClink
						samplingRate: 12000).

Then "Beeper beep" will play the coffeeCup sound.!

Beeper class
	instanceVariableNames: 'default'!

!Beeper class methodsFor: 'beeping' stamp: 'gk 2/24/2004 08:38'!
beep
	"The preferred way of producing an audible feedback.
	The default playable entity (an instance of Beeper)
	also uses the pluggable SoundService
	mechanism, so it will use the primitive beep only
	if there is no other sound mechanism available."

	self default play
! !

!Beeper class methodsFor: 'beeping' stamp: 'HenrikSperreJohansen 1/21/2011 18:34'!
beepPrimitive
	"Make a primitive beep. Only use this if
	you want to force this to be a primitive beep.
	Otherwise use Beeper class>>beep
	since this method bypasses the current
	registered playable entity."

	SoundService soundEnabled ifTrue: [
		self primitiveBeep]! !

!Beeper class methodsFor: 'customize' stamp: 'gk 2/22/2004 17:51'!
clearDefault
	"Clear the default playable.
	Will be lazily initialized in Beeper class >>default."

	default := nil! !

!Beeper class methodsFor: 'customize' stamp: 'nice 8/27/2010 16:32'!
default
	"When the default is not defined it is
	initialized using #newDefault."

	^ default ifNil: [ default := self newDefault ]! !

!Beeper class methodsFor: 'customize' stamp: 'gk 2/24/2004 22:12'!
newDefault
	"Subclasses may override me to provide a default beep.
	This base implementation returns an instance of Beeper
	which uses the pluggable sound service."

	^ self new! !

!Beeper class methodsFor: 'private' stamp: 'gk 2/24/2004 23:51'!
primitiveBeep
	"Make a primitive beep. Not to be called directly.
	It is much better to use Beeper class>>beep
	or Beeper class>>beepPrimitive
	since this method bypasses the current
	registered playable entity and does not
	check Preferences class>>soundsEnabled."

	<primitive: 140>
	self primitiveFailed! !

!Beeper class methodsFor: 'customize' stamp: 'gk 2/22/2004 17:54'!
setDefault: aPlayableEntity
	"Set the playable entity used when making a beep.
	The playable entity should implement the message #play."

	default := aPlayableEntity! !

!Beeper methodsFor: 'play interface' stamp: 'gk 2/24/2004 23:25'!
play
	"This is how the default Beeper makes a beep,
	by sending beep to the default sound service.
	The sound system will check if sounds are enabled."

	SoundService default beep! !

Object subclass: #BreakpointManager
	instanceVariableNames: ''
	classVariableNames: 'Installed'
	poolDictionaries: ''
	category: 'System-Tools'!

!BreakpointManager commentStamp: 'emm 5/30/2002 14:20' prior: 0!
This class manages methods that include breakpoints.
It has several class methods to install and uninstall breakpoints.

Evaluating "BreakpointManager clear" will remove all installed breakpoints in the system.

Known issues:
- currently, only break-on-entry type of breakpoints are supported
- emphasis change not implemented for MVC browsers
- uninstalling the breakpoint doesn't auto-update other browsers
- uninstalling a breakpoint while debugging should restart-simulate the current method

Ernest Micklei, 2002

Send comments to emicklei@philemonworks.com!

!BreakpointManager class methodsFor: 'intialization-release' stamp: 'emm 5/30/2002 09:08'!
clear
	"BreakpointManager clear"

	self installed copy keysDo:[ :breakMethod |
		self unInstall: breakMethod].
		
		! !

!BreakpointManager class methodsFor: 'private' stamp: 'eem 3/20/2015 12:13'!
compilePrototype: aSymbol in: aClass
	"Compile and answer a new method containing a break statement"

	| oldSource parser methodNode breakOnlyMethodNode sendBreakMessageNode hasTemps newSource mark positionParser markBeforePragmas pragmasFollowTemps |
	oldSource := aClass sourceCodeAt: aSymbol.
	parser := aClass newParser.
	methodNode := parser
						parse: oldSource
						class: aClass
						noPattern: false
						notifying: nil 
						ifFail: [self error: '[breakpoint] unable to install breakpoint'].
	breakOnlyMethodNode := aClass newCompiler
								compile: 'temporaryMethodSelectorForBreakpoint self break. ^self'
								in: aClass 
								notifying: nil 
								ifFail: [self error: '[breakpoint] unable to install breakpoint'].
	sendBreakMessageNode := breakOnlyMethodNode block statements first.
	methodNode block statements addFirst: sendBreakMessageNode.
	"Use anothe rparser to discover positions because if pragmas follow temporaries
	 then we want to insert the break after the last pragma, not after the temporaries."
	(positionParser := aClass newParser)
		initPattern: oldSource
		return:
			[:pattern|
			 positionParser pragmaSequence; temporaries.
			 markBeforePragmas := positionParser startOfNextToken.
			 positionParser pragmaSequence].
	mark := (pragmasFollowTemps := positionParser startOfNextToken > markBeforePragmas)
				ifTrue:
					[positionParser startOfNextToken - 1]
				ifFalse:
					[(hasTemps := positionParser tempsMark <= oldSource size and: [(oldSource at: positionParser tempsMark) = $|])
						ifTrue: [positionParser tempsMark + 2]
						ifFalse: "If no temps, tempsMark points at start of first token in body iff there is a body"
							[positionParser tempsMark >= oldSource size
								ifTrue: "empty body"
									[positionParser startOfNextToken]
								ifFalse:
									[positionParser tempsMark - 1]]].
	newSource := oldSource copyReplaceFrom: mark to: mark - 1 with: '	self break.\' withCRs.
	^methodNode generate copyWithSourceCode: newSource! !

!BreakpointManager class methodsFor: 'install-uninstall' stamp: 'eem 7/4/2015 07:40'!
forget: breakMethod
	"Answer the original unbreakpointed method, removing the original from the BreakpointManager's collection.
	 Use with care!!!!"
	^self installed removeKey: breakMethod ifAbsent: []! !

!BreakpointManager class methodsFor: 'private' stamp: 'nice 8/27/2010 22:14'!
installed
	^Installed ifNil: [Installed := IdentityDictionary new]! !

!BreakpointManager class methodsFor: 'install-uninstall' stamp: 'nice 4/10/2008 22:00'!
installInClass: aClass selector: aSymbol 
	"Install a new method containing a breakpoint.
	The receiver will remember this for unstalling it later"

	| breakMethod |
	breakMethod := self compilePrototype: aSymbol in: aClass.
	breakMethod isNil
		ifTrue: [^ nil].
	self installed at: breakMethod put: aClass >> aSymbol. "old method"
	aClass basicAddSelector: aSymbol withMethod: breakMethod.! !

!BreakpointManager class methodsFor: 'testing' stamp: 'emm 5/30/2002 09:22'!
methodHasBreakpoint: aMethod
	^self installed includesKey: aMethod! !

!BreakpointManager class methodsFor: 'examples' stamp: 'emm 5/30/2002 14:12'!
testBreakpoint
	"In the menu of the methodList, click on -toggle break on entry-
	and evaluate the following:"

	"BreakpointManager testBreakpoint"

	Transcript cr; show: 'Breakpoint test'! !

!BreakpointManager class methodsFor: 'install-uninstall' stamp: 'md 2/15/2006 21:25'!
unInstall: breakMethod 

	| class selector oldMethod |
	oldMethod := self installed at: breakMethod ifAbsent:[^self].
	class := breakMethod methodClass.
	selector := breakMethod selector.
	
	(class>>selector) == breakMethod ifTrue:[	
			class methodDictionary at: selector put: oldMethod].
	self installed removeKey: breakMethod! !

Object subclass: #ChangeRecord
	instanceVariableNames: 'file position type class category meta stamp'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Changes'!

!ChangeRecord commentStamp: '<historical>' prior: 0!
A ChangeRecord represents a change recorded on a file in fileOut format.
It includes a type (more needs to be done here), and additional information
for certain types such as method defs which need class and category.!

!ChangeRecord methodsFor: 'access' stamp: ''!
category
	^category! !

!ChangeRecord methodsFor: 'initialization' stamp: 'tk 6/24/1999 14:51'!
class: clsName category: cat method: method sourceFiles: fileArray
	"This should be enough to find all the information for a method, or method deletion"

	file := fileArray at: method fileIndex.
	position := method filePosition.
	type := #method.
	class := clsName copyUpTo: $ .	"the non-meta part of a class name"
	category := cat.
	meta := clsName endsWith: ' class'.
	self readStamp.! !

!ChangeRecord methodsFor: 'access' stamp: 'eem 4/23/2007 10:37'!
commentClass 
	| commentClass |
	type == #classComment ifFalse: [^ nil].
	(Smalltalk includesKey: class asSymbol) ifFalse: [^ nil].
	commentClass := Smalltalk at: class asSymbol.
	^meta ifTrue: [commentClass class]
		ifFalse: [commentClass]! !

!ChangeRecord methodsFor: 'access' stamp: 'eem 5/29/2015 08:36'!
file
	^file! !

!ChangeRecord methodsFor: 'initialization' stamp: 'fbs 9/18/2013 17:31'!
fileIn
	^ self fileIn: Smalltalk globals.! !

!ChangeRecord methodsFor: 'access' stamp: 'eem 5/18/2010 20:25'!
fileIndex
	SourceFiles withIndexDo: [:sf :i| sf name = file name ifTrue: [^i]].
	^nil! !

!ChangeRecord methodsFor: 'initialization' stamp: 'cwp 3/22/2014 19:38'!
fileIn: anEnvironment
	"File the receiver in.  If I represent a method or a class-comment, file the method in and make a note of it in the recent-submissions list; if I represent a do-it, then, well, do it; if I represent a reorganization then get organized!!"

	Cursor read showWhile:
		[| s cls aSelector |
		type == #doIt
			ifTrue:
				[((s := self string) beginsWith: '----') ifFalse: [Compiler evaluate: s]]
			ifFalse:
				[cls := self methodClass: anEnvironment.
				 type == #method ifTrue:
					[[cls compile: self text classified: category withStamp: stamp notifying: nil]
					on: CurrentEnvironment do: [:e | e resume: anEnvironment].
					(aSelector := self methodSelector) ifNotNil:
						[RecentMessages default recordSelector: aSelector forClass: cls inEnvironment: anEnvironment]].
				 type == #classComment ifTrue:
					[cls comment: self text stamp: stamp.
					RecentMessages default recordSelector: #Comment forClass: cls inEnvironment: anEnvironment].
				 type == #reorganize ifTrue:
					[cls organization changeFromString: self text]]]! !

!ChangeRecord methodsFor: 'access' stamp: 'nk 1/7/2004 10:28'!
fileName
	^(file ifNotNil: [ file name ]) 
			ifNil: [ '<no file>' ]! !

!ChangeRecord methodsFor: 'access' stamp: 'sw 10/20/2002 02:53'!
fileOutOn: aFileStream
	"File the receiver out on the given file stream"

	| aString |
	type == #method
		ifTrue:
			[aFileStream nextPut: $!!.
			aString :=  class asString
							, (meta ifTrue: [' class methodsFor: ']
									ifFalse: [' methodsFor: '])
							, category asString printString.
			stamp ifNotNil:
				[aString := aString, ' stamp: ''', stamp, ''''].
			aFileStream nextChunkPut: aString.
			aFileStream cr].

	type == #preamble ifTrue: [aFileStream nextPut: $!!].

	type == #classComment
		ifTrue:
			[aFileStream nextPut: $!!.
			aFileStream nextChunkPut: class asString, ' commentStamp: ', stamp storeString.
			aFileStream cr].

	aFileStream nextChunkPut: self string.
	type == #method ifTrue: [aFileStream nextChunkPut: ' '].
	aFileStream cr! !

!ChangeRecord methodsFor: 'initialization' stamp: ''!
file: f position: p type: t
	file := f.
	position := p.
	type := t! !

!ChangeRecord methodsFor: 'initialization' stamp: '6/6/97 08:48 dhhi'!
file: f position: p type: t class: c category: cat meta: m stamp: s
	self file: f position: p type: t.
	class := c.
	category := cat.
	meta := m.
	stamp := s! !

!ChangeRecord methodsFor: 'access' stamp: 'tk 6/24/1999 15:27'!
headerFor: selector

	^ '    ' , class , (meta ifTrue: [' class '] ifFalse: [' '])
				, selector
				, (stamp isEmpty ifTrue: [''] ifFalse: ['; ' , stamp])! !

!ChangeRecord methodsFor: 'access' stamp: ''!
isMetaClassChange
	^meta! !

!ChangeRecord methodsFor: 'access' stamp: 'fbs 9/18/2013 08:52'!
methodClass
	^ self methodClass: Smalltalk globals.! !

!ChangeRecord methodsFor: 'access' stamp: 'eem 10/21/2014 09:46'!
methodClassName
	| tokens |
	(class isNil
	and: [type = #doIt
	and: [([(tokens := Scanner new scanTokens: self text) size >= 3]
			on: Error do: [:ex| false])
	and: [(tokens includes: #'.') not "exclude multi-statement doits"
	and: [tokens first isSymbol
	and: [tokens first isKeyword not
	and: [tokens first first canBeGlobalVarInitial
	and: [(tokens includes: #instanceVariableNames:)
		or: [tokens second == #comment:
		or: [tokens third == #comment:]]]]]]]]]) ifTrue:
		["Could be a class definition.
			Class definitions start with Superclass blahSubclass: #ClassName
			Metaclass definitions start with ClassName class instanceVariableNames:"
		 (tokens second isSymbol
		  and: [tokens second isKeyword
		  and: [tokens third isSymbol
		  and: [tokens third isKeyword not
		  and: [(2 to: tokens size by: 2) allSatisfy: [:i| (tokens at: i) isKeyword]]]]]) ifTrue:
			[^tokens third].
		 (tokens size = 4
		  and: [tokens second = #class
		  and: [(tokens third = #instanceVariableNames: or: [tokens third = #comment:])
		  and: [tokens last isString]]]) ifTrue:
			[^tokens first].
		 (tokens size = 3
		  and: [tokens second = #comment:
		  and: [tokens last isString]]) ifTrue:
			[^tokens first]].
	^class! !

!ChangeRecord methodsFor: 'access' stamp: 'ul 8/14/2015 18:05'!
methodClass: anEnvironment
	| methodClassName methodClass |
	(#(method #classComment) includes: type) ifFalse: [ ^ nil ].
	methodClassName := class substrings
		ifEmpty: [ ^ nil ]
		ifNotEmptyDo:
			[ : parts | parts first asSymbol ].
	(anEnvironment includesKey: methodClassName) ifFalse: [ ^ nil ].
	methodClass := anEnvironment at: methodClassName.
	^ meta
		ifTrue: [ methodClass class ]
		ifFalse: [ methodClass ]! !

!ChangeRecord methodsFor: 'access' stamp: 'nice 2/23/2011 21:46'!
methodSelector
	^type == #method ifTrue:
		[(Smalltalk at: class ifAbsent: [Object]) newParser parseSelector: self string]! !

!ChangeRecord methodsFor: 'access' stamp: 'ar 7/15/2005 22:57'!
originalChangeSetForSelector: methodSelector
	"Returns the original changeset which contained this method version.  If it is contained in the .sources file, return #sources.  If it is in neither (e.g. its changeset was deleted), return nil.  (The selector is passed in purely as an optimization.)"

	| likelyChangeSets originalChangeSet |
	(file localName findTokens: '.') last = 'sources'
		ifTrue: [^ #sources].
	likelyChangeSets := ChangeSet allChangeSets select: 
		[:cs | (cs atSelector: methodSelector class: self methodClass) ~~ #none].
	originalChangeSet := likelyChangeSets
		detect: [:cs | cs containsMethodAtPosition: position]
		ifNone: [nil].
	^ originalChangeSet  "(still need to check for sources file)"! !

!ChangeRecord methodsFor: 'access' stamp: 'sumim 9/2/2003 14:07'!
position
	^ position! !

!ChangeRecord methodsFor: 'access' stamp: 'eem 5/19/2010 10:50'!
position: anInteger
	"Set the position.  Dangerous!!  Used in condenseChanges implementation by a very knowledgeable client!!"
	position := anInteger! !

!ChangeRecord methodsFor: 'access' stamp: 'sumim 9/2/2003 13:33'!
prior
	| currFile preamble prevPos tokens prevFileIndex |
	currFile := file readOnlyCopy.
	currFile position: (0 max: position - 150).
	[currFile position < (position - 1)] whileTrue: [preamble := currFile nextChunk].
	currFile close.
	prevPos := nil.
	(preamble findString: 'methodsFor:' startingAt: 1) > 0
		ifTrue: [tokens := Scanner new scanTokens: preamble]
		ifFalse: [tokens := Array new].
	((tokens size between: 7 and: 8)
	and: [(tokens at: tokens size - 5) == #methodsFor:]) ifTrue: [
		(tokens at: tokens size - 3) == #stamp:
		ifTrue: [
			prevPos := tokens last.
			prevFileIndex := SourceFiles fileIndexFromSourcePointer: prevPos.
			prevPos := SourceFiles filePositionFromSourcePointer: prevPos]
		ifFalse: [
			prevPos := tokens at: tokens size - 2.
			prevFileIndex := tokens last].
		(prevPos = 0 or: [prevFileIndex = 0]) ifTrue: [prevPos := nil]].
	prevPos ifNil: [^ nil].
	^ {prevFileIndex. prevPos. 
		SourceFiles sourcePointerFromFileIndex: prevFileIndex andPosition: prevPos}! !

!ChangeRecord methodsFor: 'access' stamp: 'tk 6/21/1999 20:34'!
readStamp
	"Get the time stamp of this method off the file"

	| item tokens anIndex |
	stamp := ''.
	file ifNil: [^ stamp].
	file position: position.
	item := file nextChunk.
	tokens := Scanner new scanTokens: item.
	tokens size < 3 ifTrue: [^ stamp].
	anIndex := tokens indexOf: #stamp: ifAbsent: [^ stamp].
	^ stamp := tokens at: (anIndex + 1).
! !

!ChangeRecord methodsFor: 'access' stamp: '6/6/97 08:56 dhhi'!
stamp
	^ stamp! !

!ChangeRecord methodsFor: 'access' stamp: 'tk 9/7/2000 15:09'!
stamp: threePartString

	stamp := threePartString! !

!ChangeRecord methodsFor: 'access' stamp: 'di 1/13/98 16:57'!
string 
	| string |
	file openReadOnly.
	file position: position.
	string := file nextChunk.
	file close.
	^ string! !

!ChangeRecord methodsFor: 'access' stamp: 'tk 6/23/1999 08:20'!
text
	| text |
	^ file ifNil: ['']
		ifNotNil: [
			file openReadOnly.
			file position: position.
			text := file nextChunkText.
			file close.
			text]! !

!ChangeRecord methodsFor: 'access' stamp: 'nice 12/27/2009 02:51'!
timeStamp
	"Answer a TimeStamp that corresponds to my (text) stamp"
	| tokens |
	tokens := self stamp findTokens: Character separators.
	^ tokens size > 2
		ifTrue: [[| date time |
			date := Date fromString: (tokens at: tokens size - 1).
			time := Time fromString: tokens last.
			TimeStamp date: date time: time]
				on: Error
				do: [:ex | ex
						return: (TimeStamp fromSeconds: 0)]]
		ifFalse: [TimeStamp fromSeconds: 0]! !

!ChangeRecord methodsFor: 'access' stamp: ''!
type
	^ type! !

Object subclass: #ChangeSet
	instanceVariableNames: 'name preamble postscript revertable isolationSet isolatedProject changeRecords structures superclasses'
	classVariableNames: 'AllChangeSets PreviousSet'
	poolDictionaries: ''
	category: 'System-Changes'!

ChangeSet class
	instanceVariableNames: 'current'!

!ChangeSet commentStamp: '<historical>' prior: 0!
ChangeSets keep track of the changes made to a system, so they can be written on a file as source code (a "fileOut").  Every project has an associated changeSet.  For simple projects, a different changeSet may be designated to capture changes at any time.

This implementation of ChangeSet is capable of remembering and manipulating methods for which the classes are not present in the system.  However at the present time, this capability is not used in normal rearranging and fileOuts, but only for invoking and revoking associated with isolation layers.

For isolated projects (see Project class comment), the changeSet binding is semi-permanent.  Every project exists in an isolation layer defined by its closest enclosing parent (or itself) that is isolated.  If a project is not isolated, then changes reported to its designated changeSet must also be reported to the permanent changeSet for that layer, designated in the isolated project.  This ensures that that outer project will be able to revert all changes upon exit.

Note that only certain changes may be reverted.  Classes may not be added, removed, renamed or reshaped except in the layer in which they are defined because these operations on non-local classes are not revertable.

If a Squeak Project is established as being isolated, then its associated changeSet will be declared to be revertable.  In this case all changes stored can be reverted.  The changeSet associated with an isolated project is tied to that project, and cannot be edited in a changeSorter.
------

name - a String used to name the changeSet, and thus any associated project or fileOut.

preamble and postscript:  two strings that serve as prefix (useful for documentation) and suffix (useful for doits) to the fileout of the changeSet.

revertable - a Boolean
If this variable is true, then all of the changes recorded by this changeSet can be reverted.

isolationSet - a ChangeSet or nil
The isolationSet is the designated changeSet for an isolation layer.  If this changeSet is an isolationSet, then this variable will be nil.  If not, then it points to the isolationSet for this layer, and all changes reported here will also be reported to the isolationSet.

isolatedProject - a Project or nil
If this is an isolationSet, then this variable points to the project with which it is associated.

changeRecords -  Dictionary {class name -> a ClassChangeRecord}.
These classChangeRecords (qv) remember all of the system changes.

structures -    Dictionary {#Rectangle -> #(<classVersionInteger> 'origin' 'corner')}.
Of  the names of the instances variables before any changes for all classes in classChanges, and all of their superclasses.  In the same format used in SmartRefStream.  Inst var names are strings.  

superclasses -    Dictionary {#Rectangle -> #Object}.
Of all classes in classChanges, and all of their superclasses.

Structures and superclasses save the instance variable names of this class and all of its superclasses.  Later we can tell how it changed and write a conversion method.  The conversion method is used when old format objects are brought in from the disk from ImageSegment files (.extSeg) or SmartRefStream files (.obj .morph .bo .sp).

NOTE:  It should be fairly simple, by adding a bit more information to the classChangeRecords, to reconstruct the information now stored in 'structures' and 'superclasses'.  This would be a welcome simplification.
!

ChangeSet class
	instanceVariableNames: 'current'!

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:20'!
allChangeSetNames
	^ self allChangeSets collect: [:c | c name]! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:10'!
allChangeSets
	"Return the list of all current ChangeSets"

	^ AllChangeSets! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:22'!
allChangeSetsWithClass: class selector: selector
	class ifNil: [^ #()].
	^ self allChangeSets select: 
		[:cs | (cs atSelector: selector class: class) ~~ #none]! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:17'!
allChangeSets: aCollection
	"Return the list of all current ChangeSets"

	AllChangeSets := aCollection.! !

!ChangeSet class methodsFor: 'services' stamp: 'ar 7/15/2005 21:30'!
assuredChangeSetNamed: aName
	"Answer a change set of the given name.  If one already exists, answer that, else create a new one and answer it."

	| existing |
	^ (existing := self named: aName)
		ifNotNil:
			[existing]
		ifNil:
			[self basicNewChangeSet: aName]! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:14'!
basicNewChangeSet: newName
	| newSet |
	newName ifNil: [^ nil].
	(self named: newName) ifNotNil:
		[self inform: 'Sorry that name is already used'.
		^ nil].
	newSet := self basicNewNamed: newName.
	AllChangeSets add: newSet.
	^ newSet! !

!ChangeSet class methodsFor: 'instance creation' stamp: 'di 4/6/2001 09:43'!
basicNewNamed: aName

	^ (self basicNew name: aName) initialize! !

!ChangeSet class methodsFor: 'current changeset' stamp: 'ar 7/17/2005 10:48'!
browseChangedMessages
	"Create and schedule a message browser on each method that has been 
	changed."

	current isEmpty ifTrue: [^ self inform: 'There are no changed messages
in the current change set.'].
	ToolSet openChangedMessageSet: current.! !

!ChangeSet class methodsFor: 'services' stamp: 'ar 7/15/2005 21:28'!
buildAggregateChangeSet
	"Establish a change-set named Aggregate which bears the union of all the changes in all the existing change-sets in the system (other than any pre-existing Aggregate).  This can be useful when wishing to discover potential conflicts between a disk-resident change-set and an image.  Formerly very useful, now some of its unique contributions have been overtaken by new features"

	| aggregateChangeSet |
	aggregateChangeSet := self existingOrNewChangeSetNamed: 'Aggregate'.
	aggregateChangeSet clear.
	self allChangeSets do:
		[:aChangeSet | aChangeSet == aggregateChangeSet ifFalse:
			[aggregateChangeSet assimilateAllChangesFoundIn: aChangeSet]]

"ChangeSet buildAggregateChangeSet"

	! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:11'!
changeSetsNamedSuchThat: nameBlock
	"(ChangeSet changeSetsNamedSuchThat:
		[:name | name first isDigit and: [name initialInteger >= 373]])
		do: [:cs | AllChangeSets remove: cs wither]"

	^ AllChangeSets select: [:aChangeSet | nameBlock value: aChangeSet name]! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:26'!
changeSet: aChangeSet containsClass: aClass
	| theClass |
	theClass := Smalltalk classNamed: aClass.
	theClass ifNil: [^ false].
	^ aChangeSet containsClass: theClass! !

!ChangeSet class methodsFor: 'current changeset' stamp: 'ar 2/26/2010 23:01'!
cleanUp: aggressive
	"Only delete change sets when being aggressive"

	aggressive ifTrue:[
		ChangeSet  removeChangeSetsNamedSuchThat:
			[:cs| cs name ~= ChangeSet current name].
		ChangeSet current clear.
		ChangeSet current name: 'Unnamed1'.
	].! !

!ChangeSet class methodsFor: 'services' stamp: 'ar 7/15/2005 21:37'!
countOfChangeSetsWithClass: aClass andSelector: aSelector
	"Answer how many change sets record a change for the given class and selector"

	^ (self allChangeSetsWithClass: aClass selector: aSelector) size! !

!ChangeSet class methodsFor: 'current changeset' stamp: 'wiz 9/19/2006 03:21'!
current
	"return the current changeset
	assure first that we have a named changeset. To cure mantis #4535. "
	current isMoribund
		ifTrue: [(ChangeSet newChanges: (ChangeSet assuredChangeSetNamed: 'Unnamed'))] .

	^ current! !

!ChangeSet class methodsFor: 'current changeset' stamp: 'em 3/31/2005 11:48'!
currentChangeSetString
	"ChangeSet current currentChangeSetString"

	^ 'Current Change Set: ' translated, self current name! !

!ChangeSet class methodsFor: 'defaults' stamp: 'nk 7/18/2004 16:13'!
defaultChangeSetDirectory
	"Answer the directory in which to store ChangeSets. 
	Answer the default directory if the preferred directory doesn't exist."
	| dir directoryName |
	directoryName := Preferences
				parameterAt: #defaultChangeSetDirectoryName
				ifAbsentPut: [''].
	dir := directoryName isEmptyOrNil
		ifTrue: [ FileDirectory default ]
		ifFalse: [ FileDirectory default directoryNamed: directoryName ].
	dir exists
		ifTrue: [^ dir].
	^ FileDirectory default! !

!ChangeSet class methodsFor: 'defaults' stamp: 'nk 3/24/2004 15:52'!
defaultChangeSetDirectory: dirOrName 
	"Set the Preference for storing change sets to the given directory or name (possibly relative).
	Rewrite directory names below the default directory as relative names.
	If dirOrName is an empty string, use the default directory."

	"ChangeSet defaultChangeSetDirectory: 'changeSets'"

	| dirName defaultFullName |
	dirName := dirOrName isString
				ifTrue: [FileDirectory default fullNameFor: dirOrName]
				ifFalse: [dirOrName fullName].
	defaultFullName := FileDirectory default fullName.
	dirName = defaultFullName
		ifTrue: [dirName := '']
		ifFalse: [(dirName beginsWith: defaultFullName , FileDirectory slash)
				ifTrue: [dirName := dirName copyFrom: defaultFullName size + 2 to: dirName size]].
	Preferences setParameter: #defaultChangeSetDirectoryName to: dirName! !

!ChangeSet class methodsFor: 'defaults' stamp: 'dgd 9/6/2003 19:56'!
defaultName
	^ self uniqueNameLike: 'Unnamed' translated! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:26'!
deleteChangeSetsNumberedLowerThan: anInteger
	"Delete all changes sets whose names start with integers smaller than anInteger"

	self removeChangeSetsNamedSuchThat:
		[:aName | aName first isDigit and: [aName initialIntegerOrNil < anInteger]].

	"ChangeSet deleteChangeSetsNumberedLowerThan: (ChangeSorterPlus highestNumberedChangeSet name initialIntegerOrNil - 500)"
! !

!ChangeSet class methodsFor: 'scanning' stamp: 'eem 5/29/2015 10:46'!
directAncestryOfVersions: changeRecords
	"Take a sequence of ChangeRecords as answered by scanVersionsOf:class:meta:category:selector:
	 and answer a filtered subsequence consisting only of direct ancestors.  For example, if the input is
		eem 7/7/2009 20:06	7 July 2009 8:06 pm
		eem 6/18/2009 19:21	18 June 2009 7:21 pm
		eem 5/5/2009 12:16	5 May 2009 12:16 pm
		eem 6/18/2009 19:19	18 June 2009 7:19 pm
		eem 6/18/2009 18:57	18 June 2009 6:57 pm
		eem 5/5/2009 12:16	5 May 2009 12:16 pm
		eem 6/18/2009 18:17	18 June 2009 6:17 pm
		eem 6/18/2009 18:14	18 June 2009 6:14 pm
		eem 6/18/2009 18:06	18 June 2009 6:06 pm
		eem 5/5/2009 12:16	5 May 2009 12:16 pm
		eem 5/4/2009 19:19	4 May 2009 7:19 pm
	the output should be
		eem 7/7/2009 20:06	7 July 2009 8:06 pm
		eem 6/18/2009 19:21	18 June 2009 7:21 pm
		eem 5/5/2009 12:16	5 May 2009 12:16 pm
		eem 5/4/2009 19:19	4 May 2009 7:19 pm"

	| filtered i last |
	filtered := OrderedCollection new.
	i := 1.
	[i <= changeRecords size] whileTrue:
		[filtered addLast: (changeRecords at: i).
		 last := changeRecords
					findLast: [:chgRec| filtered last stamp = chgRec stamp and: [filtered last position = chgRec position]]
					startingAt: i.
		 i := last = 0 ifTrue: [i + 1] ifFalse: [last + 1]].
	^filtered! !

!ChangeSet class methodsFor: 'services' stamp: 'ar 7/15/2005 21:38'!
doesAnyChangeSetHaveClass: aClass andSelector: aSelector
	"Answer whether any known change set bears a change for the given class and selector"

	^ (self countOfChangeSetsWithClass: aClass andSelector: aSelector) > 0! !

!ChangeSet class methodsFor: 'fileIn/Out' stamp: 'ar 12/20/2009 13:04'!
doWeFileOut: aClass given: aSet cache: cache
	| aClassAllSuperclasses aClassSoleInstanceAllSuperclasses |
	aClassAllSuperclasses := cache at: aClass ifAbsentPut:[
		aClass allSuperclasses asArray, aClass allTraits.
	].
	(aSet includesAnyOf: aClassAllSuperclasses) ifTrue: [^false].
	aClass isMeta ifFalse: [^true].
	(aSet includes: aClass soleInstance) ifTrue: [^false].
	aClassSoleInstanceAllSuperclasses := cache at: aClass soleInstance
		ifAbsentPut: [aClass soleInstance allSuperclasses asArray, aClass soleInstance allTraits].
	(aSet includesAnyOf: aClassSoleInstanceAllSuperclasses) ifTrue: [^false].
	^true! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:24'!
existingOrNewChangeSetNamed: aName

	| newSet |
	^(self named: aName) ifNil: [
		newSet := self basicNewNamed: aName.
		AllChangeSets add: newSet.
		newSet
	]! !

!ChangeSet class methodsFor: 'services' stamp: 'ar 7/15/2005 21:35'!
fileIntoNewChangeSet: fullName
	"File in all of the contents of the currently selected file, if any, into a new change set." 

	| fn ff |
	fullName ifNil: [^ Beeper beep].
	ff := FileStream readOnlyFileNamed: (fn := GZipReadStream uncompressedFileName: fullName).
	ChangeSet newChangesFromStream: ff named: (FileDirectory localNameFor: fn)! !

!ChangeSet class methodsFor: 'services' stamp: 'nice 12/27/2009 03:11'!
fileOutChangeSetsNamed: nameList
	"File out the list of change sets whose names are provided"
     "ChangeSorter fileOutChangeSetsNamed: #('New Changes' 'miscTidies-sw')"

	| notFound empty infoString |
	notFound := OrderedCollection new.
	empty := OrderedCollection new.
	nameList do:
		[:aName | | aChangeSet | (aChangeSet := self named: aName)
			ifNotNil:
				[aChangeSet isEmpty
					ifTrue:
						[empty add: aName]
					ifFalse:
						[aChangeSet fileOut]]
			ifNil:
				[notFound add: aName]].

	infoString := (nameList size - notFound size) printString, ' change set(s) filed out'.
	notFound size > 0 ifTrue:
		[infoString := infoString, '

', notFound size printString, ' change set(s) not found:'.
		notFound do:
			[:aName | infoString := infoString, '
', aName]].
	empty size > 0 ifTrue:
		[infoString := infoString, '
', empty size printString, ' change set(s) were empty:'.
		empty do:
			[:aName | infoString := infoString, '
', aName]].

	self inform: infoString! !

!ChangeSet class methodsFor: 'file list services' stamp: 'ar 7/15/2005 21:36'!
fileReaderServicesForFile: fullName suffix: suffix

	^ (FileStream isSourceFileSuffix: suffix)
		ifTrue: [ self services]
		ifFalse: [#()]! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'nice 12/27/2009 03:11'!
gatherChangeSets		"ChangeSet gatherChangeSets"
	"Collect any change sets created in other projects"
	| allChangeSets |
	allChangeSets := AllChangeSets asSet.
	ChangeSet allSubInstances do: [:each | | obsolete |
		(allChangeSets includes: each) == (obsolete := each isMoribund) ifTrue:[
			obsolete
				ifTrue: ["Was included and is obsolete."
						AllChangeSets remove: each]
				ifFalse: ["Was not included and is not obsolete."
						AllChangeSets add: each]]].
	^ AllChangeSets! !

!ChangeSet class methodsFor: 'scanning' stamp: 'fbs 7/25/2013 07:07'!
getRecentLocatorWithPrompt: aPrompt
	"Prompt with a menu of how far back to go.  Return nil if user backs out.  Otherwise return the number of characters back from the end of the .changes file the user wishes to include"
	 "ChangeList getRecentPosition"
	| end changesFile banners positions pos chunk i |
	changesFile := (SourceFiles at: 2) readOnlyCopy.
	banners := OrderedCollection new.
	positions := OrderedCollection new.
	end := changesFile size.
	pos := Smalltalk lastQuitLogPosition.
	[pos = 0 or: [banners size > 20]] whileFalse:
		[changesFile position: pos.
		chunk := changesFile nextChunk.
		i := chunk indexOfSubCollection: 'priorSource: ' startingAt: 1.
		i > 0 ifTrue: [positions addLast: pos.
					banners addLast: (chunk copyFrom: 5 to: i-2).
					pos := Number readFrom: (chunk copyFrom: i+13 to: chunk size)]
			ifFalse: [pos := 0]].
	changesFile close.
	pos := UIManager default chooseFrom: banners values: positions title: aPrompt.
	pos == nil ifTrue: [^ nil].
	^ end - pos! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:20'!
highestNumberedChangeSet
	"ChangeSorter highestNumberedChangeSet"
	| aList |
	aList := (self allChangeSetNames select: [:aString | aString startsWithDigit] thenCollect:
		[:aString | aString initialIntegerOrNil]).
	^ (aList size > 0)
		ifTrue:
			[aList max]
		ifFalse:
			[nil]
! !

!ChangeSet class methodsFor: 'class initialization' stamp: 'ar 7/15/2005 21:12'!
initialize
	"ChangeSet initialize"
	AllChangeSets == nil ifTrue:
		[AllChangeSets := OrderedCollection new].
	self gatherChangeSets.
	FileServices registerFileReader: self.
! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:29'!
mostRecentChangeSetWithChangeForClass: class selector: selector
	| hits |
	hits := self allChangeSets select: 
		[:cs | (cs atSelector: selector class: class) ~~ #none].
	hits isEmpty ifTrue: [^ 'not in any change set'].
	^ 'recent cs: ', hits last name! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:11'!
named: aName
	"Return the change set of the given name, or nil if none found.  1/22/96 sw"

	^ AllChangeSets
			detect: [:aChangeSet | aChangeSet name = aName]
			ifNone: [nil]! !

!ChangeSet class methodsFor: 'instance creation' stamp: 'ar 7/16/2005 15:17'!
new
	"All current changeSets must be registered in the AllChangeSets collection.
	Due to a quirk of history, this is maintained as class variable of ChangeSorter."

	^ self basicNewChangeSet: ChangeSet defaultName! !

!ChangeSet class methodsFor: 'services' stamp: 'ar 7/15/2005 21:31'!
newChangeSet
	"Prompt the user for a name, and establish a new change set of
	that name (if ok), making it the current changeset.  Return nil
	of not ok, else return the actual changeset."

	| newName newSet |
	newName := UIManager default
		request: 'Please name the new change set:'
		initialAnswer: ChangeSet defaultName.
	newName isEmptyOrNil ifTrue:
		[^ nil].
	newSet := self basicNewChangeSet: newName.
	newSet ifNotNil:
		[self  newChanges: newSet].
	^ newSet! !

!ChangeSet class methodsFor: 'services' stamp: 'ar 7/15/2005 21:31'!
newChangeSet: aName
	"Makes a new change set called aName, add author initials to try to
	ensure a unique change set name."

	| newName |
	newName := aName , FileDirectory dot , Utilities authorInitials.
	^ self basicNewChangeSet: newName! !

!ChangeSet class methodsFor: 'services' stamp: 'nice 12/27/2009 03:11'!
newChangesFromStream: aStream named: aName
	"File in the code from the stream into a new change set whose
	name is derived from aName. Leave the 'current change set'
	unchanged. Return the new change set or nil on failure."

	| oldChanges newName newSet |
	oldChanges := ChangeSet current.
	PreviousSet := oldChanges name. 		"so a Bumper update can find it"
	newName := aName sansPeriodSuffix.
	newSet := self basicNewChangeSet: newName.
	[ | newStream |newSet ifNotNil:[
		(aStream respondsTo: #converter:) ifFalse: [
			newStream := MultiByteBinaryOrTextStream with: (aStream contentsOfEntireFile).
			newStream reset.
		] ifTrue: [
			newStream := aStream.
		].

		self newChanges: newSet.
		newStream setConverterForCode.
		newStream fileInAnnouncing: 'Loading ', newName, '...'.
		Transcript cr; show: 'File ', aName, ' successfully filed in to change set ', newName].
	aStream close] ensure: [self newChanges: oldChanges].
	PreviousSet := nil.
	^ newSet! !

!ChangeSet class methodsFor: 'current changeset' stamp: 'ar 11/12/2009 01:08'!
newChanges: aChangeSet 
	"Set the system ChangeSet to be the argument, aChangeSet.  Tell the current project that aChangeSet is now its change set.  When called from Project enter:, the setChangeSet: call is redundant but harmless; when called from code that changes the current-change-set from within a project, it's vital"

	SystemChangeNotifier uniqueInstance noMoreNotificationsFor: current.
	current := aChangeSet.
	SystemChangeNotifier uniqueInstance notify: aChangeSet ofAllSystemChangesUsing: #event:.
	Smalltalk currentProjectDo:[:proj |
		proj setChangeSet: aChangeSet]! !

!ChangeSet class methodsFor: 'current changeset' stamp: 'sd 5/22/2003 22:18'!
noChanges 
	"Initialize the system ChangeSet."

	current initialize! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:13'!
promoteToTop: aChangeSet
	"Make aChangeSet the first in the list from now on"

	AllChangeSets remove: aChangeSet ifAbsent: [^ self].
	AllChangeSets add: aChangeSet! !

!ChangeSet class methodsFor: 'defaults' stamp: 'ar 8/6/2009 18:17'!
promptForDefaultChangeSetDirectoryIfNecessary
	"Check the Preference (if any), and prompt the user to change it if necessary.
	The default if the Preference is unset is the current directory.
	Answer the directory."

	"ChangeSet promptForDefaultChangeSetDirectoryIfNecessary"
	| choice directoryName dir |
	directoryName := Preferences
				parameterAt: #defaultChangeSetDirectoryName
				ifAbsentPut: [''].
	[dir := FileDirectory default directoryNamed: directoryName.
	dir exists]
		whileFalse: [choice := UIManager default chooseFrom: {
			'Create directory' translated.
			'Use default directory and forget preference' translated.
			'Choose another directory'  translated
		} title: ('The preferred change set directory (''{1}'') does not exist.
Create it or use the default directory ({2})?' translated format: { directoryName. FileDirectory default pathName }).
			choice = 1
				ifTrue: [dir assureExistence ].
			choice = 3
				ifTrue: [dir := UIManager default chooseDirectory.
					directoryName := dir
					ifNil: [ '' ]
						ifNotNil: [dir pathName ]]].
		self defaultChangeSetDirectory: directoryName.
		^dir! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:26'!
removeChangeSetsNamedSuchThat: nameBlock
	(self changeSetsNamedSuchThat: nameBlock)
		do: [:cs | self removeChangeSet: cs]! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:13'!
removeChangeSet: aChangeSet
	"Remove the given changeSet.  Caller must assure that it's cool to do this"

	AllChangeSets remove: aChangeSet ifAbsent: [].
	aChangeSet wither
! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:26'!
removeEmptyUnnamedChangeSets
	"Remove all change sets that are empty, whose names start with Unnamed,
		and which are not nailed down by belonging to a Project."
	"ChangeSorter removeEmptyUnnamedChangeSets"
	| toGo |
	(toGo := (self changeSetsNamedSuchThat: [:csName | csName beginsWith: 'Unnamed'])
		select: [:cs | cs isEmpty and: [cs okayToRemoveInforming: false]])
		do: [:cs | self removeChangeSet: cs].
	self inform: toGo size printString, ' change set(s) removed.'! !

!ChangeSet class methodsFor: 'current changeset' stamp: 'sd 9/8/2006 21:05'!
resetCurrentToNewUnnamedChangeSet
 
	current := self new.  
	self newChanges: current ! !

!ChangeSet class methodsFor: 'scanning' stamp: 'ul 8/14/2015 18:01'!
scanCategory: file
	"Scan anything that involves more than one chunk; method name is historical only"
	| itemPosition item tokens stamp isComment anIndex |
	itemPosition := file position.
	item := file nextChunk.

	isComment := (item includesSubstring: 'commentStamp:').
	(isComment or: [item includesSubstring: 'methodsFor:']) ifFalse:
		["Maybe a preamble, but not one we recognize; bail out with the preamble trick"
		^{(ChangeRecord new file: file position: itemPosition type: #preamble)}].

	tokens := Scanner new scanTokens: item.
	tokens size >= 3 ifTrue:
		[stamp := ''.
		anIndex := tokens indexOf: #stamp: ifAbsent: [nil].
		anIndex ifNotNil: [stamp := tokens at: (anIndex + 1)].

		tokens second == #methodsFor:
			ifTrue: [^ self scanFile: file category: tokens third class: tokens first
							meta: false stamp: stamp].
		tokens third == #methodsFor:
			ifTrue: [^ self scanFile: file category: tokens fourth class: tokens first
							meta: true stamp: stamp]].

		tokens second == #commentStamp:
			ifTrue:
				[stamp := tokens third.
				item := (ChangeRecord new file: file position: file position type: #classComment
										class: tokens first category: nil meta: false stamp: stamp).
				file nextChunk.
				file skipStyleChunk.
				^Array with: item].
	^#()! !

!ChangeSet class methodsFor: 'scanning' stamp: 'ar 7/16/2005 15:11'!
scanFile: file category: cat class: class meta: meta stamp: stamp
	| itemPosition method items |
	items := OrderedCollection new.
	[itemPosition := file position.
	method := file nextChunk.
	file skipStyleChunk.
	method size > 0] whileTrue:[
		items add: (ChangeRecord new file: file position: itemPosition type: #method
							class: class category: cat meta: meta stamp: stamp)].
	^items! !

!ChangeSet class methodsFor: 'scanning' stamp: 'ul 6/17/2011 12:25'!
scanFile: file from: startPosition to: stopPosition
	| changeList |
	changeList := OrderedCollection new.
	file position: startPosition.
'Scanning ', file localName, '...'
	displayProgressFrom: startPosition to: stopPosition
	during: [:bar | | itemPosition item prevChar |
	[file position < stopPosition] whileTrue:[
		bar value: file position.
		[file atEnd not and: [file peek isSeparator]]
			whileTrue: [prevChar := file next].
		(file peekFor: $!!) ifTrue:[
			(prevChar = Character cr or: [prevChar = Character lf])
				ifTrue: [changeList addAll: (self scanCategory: file)].
		] ifFalse:[
			itemPosition := file position.
			item := file nextChunk.
			file skipStyleChunk.
			item size > 0 ifTrue:[
				changeList add: (ChangeRecord new file: file position: itemPosition type: #doIt).
			].
		].
	]].
	^changeList! !

!ChangeSet class methodsFor: 'scanning' stamp: 'ul 11/15/2010 11:34'!
scanVersionsOf: method class: class meta: meta category: cat selector: selector
	| position prevPos prevFileIndex preamble tokens sourceFilesCopy stamp changeList file |
	changeList := OrderedCollection new.
	position := method filePosition.
	sourceFilesCopy := SourceFiles collect:[:x | x ifNotNil:[x readOnlyCopy]].
	method fileIndex = 0 ifTrue: [^ nil].
	file := sourceFilesCopy at: method fileIndex.
	[position notNil & file notNil] whileTrue:[
		preamble := method getPreambleFrom: file at: (0 max: position - 3).
		"Preamble is likely a linked method preamble, if we're in
			a changes file (not the sources file).  Try to parse it
			for prior source position and file index"
		prevPos := nil.
		stamp := ''.
		(preamble findString: 'methodsFor:' startingAt: 1) > 0
			ifTrue: [tokens := [Scanner new scanTokens: preamble] on: Error do:[#()]]
			ifFalse: [tokens := Array new  "ie cant be back ref"].
		((tokens size between: 7 and: 8)
			and: [(tokens at: tokens size-5) = #methodsFor:]) ifTrue:[
				(tokens at: tokens size-3) = #stamp: ifTrue:[
					"New format gives change stamp and unified prior pointer"
					stamp := tokens at: tokens size-2.
					prevPos := tokens last.
					prevFileIndex := sourceFilesCopy fileIndexFromSourcePointer: prevPos.
					prevPos := sourceFilesCopy filePositionFromSourcePointer: prevPos.
				] ifFalse: ["Old format gives no stamp; prior pointer in two parts"
					prevPos := tokens at: tokens size-2.
					prevFileIndex := tokens last.
				].
				(prevPos = 0 or: [prevFileIndex = 0]) ifTrue: [prevPos := nil]
			].
		((tokens size between: 5 and: 6)
			and: [(tokens at: tokens size-3) = #methodsFor:]) ifTrue:[
				(tokens at: tokens size-1) = #stamp: ifTrue: [
					"New format gives change stamp and unified prior pointer"
					stamp := tokens at: tokens size.
			]
		].
 		changeList add: (ChangeRecord new file: file position: position type: #method
						class: class name category: cat meta: meta stamp: stamp).
		position := prevPos.
		prevPos notNil ifTrue:[file := sourceFilesCopy at: prevFileIndex].
	].
	sourceFilesCopy do: [:x | x ifNotNil:[x close]].
	^changeList! !

!ChangeSet class methodsFor: 'enumerating' stamp: 'ar 7/15/2005 21:13'!
secondaryChangeSet
	"Answer a likely change set to use as the second initial one in a Dual Change Sorter.  "

	AllChangeSets size = 1 ifTrue: [^ AllChangeSets first].
	AllChangeSets last == ChangeSet current
		ifTrue: 	[^ AllChangeSets at: (AllChangeSets size - 1)]
		ifFalse:	[^ AllChangeSets last]! !

!ChangeSet class methodsFor: 'file list services' stamp: 'ar 7/15/2005 21:35'!
serviceFileIntoNewChangeSet
	"Answer a service for installing a file into a new change set"

	^ SimpleServiceEntry 
		provider: self 
		label: 'install into new change set'
		selector: #fileIntoNewChangeSet:
		description: 'install the file as a body of code in the image: create a new change set and file-in the selected file into it'
		buttonLabel: 'install'! !

!ChangeSet class methodsFor: 'file list services' stamp: 'ar 7/15/2005 21:36'!
services
	^ Array with: self serviceFileIntoNewChangeSet! !

!ChangeSet class methodsFor: 'filein/out' stamp: 'nice 12/27/2009 03:11'!
superclassOrder: classes
	"Arrange the classes in the collection, classes, in superclass order so the 
	classes can be properly filed in. Do it in sets instead of ordered collections.
	SqR 4/12/2000 22:04"

	| all list inclusionSet cache |

	list := classes copy. "list is indexable"
	inclusionSet := list asSet. cache := Dictionary new.
	all := OrderedCollection new: list size.
	list size timesRepeat:
		[ | aClass aClassIndex |
			aClassIndex := list findFirst: [:one | one notNil and: 
				[self doWeFileOut: one given: inclusionSet cache: cache]].
			aClass := list at: aClassIndex.
			all addLast: aClass.
			inclusionSet remove: aClass.
			list at: aClassIndex put: nil
		].
	^all! !

!ChangeSet class methodsFor: 'fileIn/Out' stamp: 'al 7/18/2004 18:45'!
traitsOrder: aCollection 
	"Answer an OrderedCollection. The traits 
	are ordered so they can be filed in."

	|  traits |
	traits := aCollection asSortedCollection: [:t1 :t2 |
		(t1 isBaseTrait and: [t1 classTrait == t2]) or: [
			(t2 traitComposition allTraits includes: t1) or: [
				(t1 traitComposition allTraits includes: t2) not]]].
	^traits asArray! !

!ChangeSet class methodsFor: 'defaults' stamp: 'ar 7/15/2005 21:24'!
uniqueNameLike: aString

	| try |
	(self named: aString) ifNil: [^ aString].

	1 to: 999999 do:
		[:i | try := aString , i printString.
		(self named: try) ifNil: [^ try]]! !

!ChangeSet methodsFor: 'moving changes' stamp: 'di 4/4/2000 09:37'!
absorbClass: className from: otherChangeSet
	"Absorb into the receiver all the changes found in the class in the other change set.
	*** Classes renamed in otherChangeSet may have problems"

	| cls |
	(self changeRecorderFor: className)
			assimilateAllChangesIn: (otherChangeSet changeRecorderFor: className).

	(cls := Smalltalk classNamed: className) ifNotNil:
		[self absorbStructureOfClass: cls from: otherChangeSet].
! !

!ChangeSet methodsFor: 'moving changes' stamp: 'di 3/23/2000 11:52'!
absorbMethod: selector class: aClass from: aChangeSet
	"Absorb into the receiver all the changes for the method in the class in the other change set."

	| info |
	info := aChangeSet methodChanges at: aClass name ifAbsent: [Dictionary new].
	self atSelector: selector class: aClass put: (info at: selector).

! !

!ChangeSet methodsFor: 'moving changes' stamp: 'sw 1/30/2001 15:41'!
absorbStructureOfClass: aClass from: otherChangeSet
	"Absorb into the receiver all the structure and superclass info in the other change set.  Used to write conversion methods."

	| sup next |
	otherChangeSet structures ifNil: [^ self].
	(otherChangeSet structures includesKey: aClass name) ifFalse: [^ self].
	structures ifNil:
		[structures := Dictionary new.
		superclasses := Dictionary new].
	sup := aClass name.
	[(structures includesKey: sup) 
		ifTrue: ["use what is here" true]
		ifFalse: [self flag: #noteToDan.  "sw 1/30/2001 13:57 emergency workaround -- a case arose where the otherChangeSet's structures did not have the key, and it gummed up the works."
				(otherChangeSet structures includesKey: sup) ifTrue:
					[structures at: sup put: (otherChangeSet structures at: sup)].
				next := otherChangeSet superclasses at: sup.
				superclasses at: sup put: next.
				(sup := next) = 'nil']
	] whileFalse.


! !

!ChangeSet methodsFor: 'change logging' stamp: 'di 3/29/2000 13:10'!
addClass: class 
	"Include indication that a new class was created."

	class wantsChangeSetLogging ifFalse: [^ self].
	isolationSet ifNotNil:
		["If there is an isolation layer above me, inform it as well."
		isolationSet addClass: class].
	self atClass: class add: #new.
	self atClass: class add: #change.
	self addCoherency: class name! !

!ChangeSet methodsFor: 'private' stamp: 'di 3/23/2000 08:37'!
addCoherency: className
	"SqR!! 19980923: If I recreate the class then don't remove it"

	(self changeRecorderFor: className)
		checkCoherence.
"
	classRemoves remove: className ifAbsent: [].
	(classChanges includesKey: className) ifTrue:
		[(classChanges at: className) remove: #remove ifAbsent: []]
"! !

!ChangeSet methodsFor: 'method changes' stamp: 'sw 12/28/2000 18:08'!
adoptSelector: aSelector forClass: aClass
	"Adopt the given selector/class combination as a change in the receiver"

	self noteNewMethod: (aClass methodDictionary at: aSelector)
			forClass: aClass selector: aSelector priorMethod: nil! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'nice 12/27/2009 03:11'!
askAddedInstVars: classList
	| pairList pairClasses index pls |
	"Ask the author whether these newly added inst vars need to be non-nil"

	pairList := OrderedCollection new.
	pairClasses := OrderedCollection new.
	"Class version numbers:  If it must change, something big happened.  Do need a conversion method then.  Ignore them here."
	classList do: [:cls | | newStruct oldStruct |
		newStruct := (cls allInstVarNames).
		oldStruct := (structures at: cls name ifAbsent: [#(0), newStruct]) allButFirst.
		newStruct do: [:instVarName |
			(oldStruct includes: instVarName) ifFalse: [
				pairList add: cls name, ' ', instVarName.
				pairClasses add: cls]]].

	pairList isEmpty ifTrue: [^ #()].
	[index := UIManager default 
		chooseFrom: pairList, #('all of these need a non-nil value'
						'all of these are OK with a nil value')
		title:
'These instance variables were added.
When an old project comes in, newly added 
instance variables will have the value nil.
Click on items to remove them from the list.
Click on any for which nil is an OK value.'.
	(index <= (pls := pairList size)) & (index > 0) ifTrue: [
		pairList removeAt: index.
		pairClasses removeAt: index].
	index = (pls + 2) ifTrue: ["all are OK" ^ #()].
	pairList isEmpty | (index = (pls + 1)) "all need conversion, exit"] whileFalse.

	^ pairClasses asSet asArray	"non redundant"! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'nice 12/27/2009 03:11'!
askRemovedInstVars: classList
	| pairList pairClasses index pls |
	"Ask the author whether these newly removed inst vars need to have their info saved"

	pairList := OrderedCollection new.
	pairClasses := OrderedCollection new.
	"Class version numbers:  If it must change, something big happened.  Do need a conversion method then.  Ignore them here."
	classList do: [:cls | | oldStruct newStruct |
		newStruct := (cls allInstVarNames).
		oldStruct := (structures at: cls name ifAbsent: [#(0), newStruct]) allButFirst.
		oldStruct do: [:instVarName |
			(newStruct includes: instVarName) ifFalse: [
				pairList add: cls name, ' ', instVarName.
				pairClasses add: cls]]].

	pairList isEmpty ifTrue: [^ #()].
	[index := UIManager default 
		chooseFrom: pairList, #('all of these need a conversion method'
						'all of these have old values that can be erased')
			title:
'These instance variables were removed.
When an old project comes in, instance variables 
that have been removed will lose their contents.
Click on items to remove them from the list.
Click on any whose value is unimportant and need not be saved.'.
	(index <= (pls := pairList size)) & (index > 0) ifTrue: [
		pairList removeAt: index.
		pairClasses removeAt: index].
	index = (pls + 2) ifTrue: ["all are OK" ^ #()].
	pairList isEmpty | (index = (pls + 1))  "all need conversion, exit"] whileFalse.

	^ pairClasses asSet asArray	"non redundant"! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'nice 12/27/2009 03:11'!
askRenames: renamed addTo: msgSet using: smart
	| list |
	"Go through the renamed classes.  Ask the user if it could be in a project.  Add a method in SmartRefStream, and a conversion method in the new class."

	list := OrderedCollection new.
	renamed do: [:cls | | rec ans oldStruct newStruct |
		rec := changeRecords at: cls name.
		rec priorName ifNotNil: [
			ans := UIManager default chooseFrom: 
					#('Yes, write code to convert those instances'
					'No, no instances are in projects')
				title: 'You renamed class ', rec priorName, 
				' to be ', rec thisName,
				'.\Could an instance of ', rec priorName, 
				' be in a project on someone''s disk?'.
			ans = 1 ifTrue: [
					oldStruct := structures at: rec priorName ifAbsent: [nil].
					newStruct := (Array with: cls classVersion), (cls allInstVarNames).
					oldStruct ifNotNil: [
						smart writeConversionMethodIn: cls fromInstVars: oldStruct 
								to: newStruct renamedFrom: rec priorName.
						smart writeClassRename: cls name was: rec priorName.
						list add: cls name, ' convertToCurrentVersion:refStream:']]
				ifFalse: [structures removeKey: rec priorName ifAbsent: []]]].
	list isEmpty ifTrue: [^ msgSet].
	msgSet messageList ifNil: [msgSet initializeMessageList: list]
		ifNotNil: [list do: [:item | msgSet addItem: item]].
	^ msgSet! !

!ChangeSet methodsFor: 'moving changes' stamp: 'di 4/4/2000 11:21'!
assimilateAllChangesFoundIn: otherChangeSet
	"Make all changes in otherChangeSet take effect on self as if they happened just now."

	otherChangeSet changedClassNames do:
		[:className | self absorbClass: className from: otherChangeSet]
! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:05'!
assurePostscriptExists
	"Make sure there is a StringHolder holding the postscript.  "

	"NOTE: FileIn recognizes the postscript by the line with Postscript: on it"
	postscript == nil ifTrue: [postscript := '"Postscript:
Leave the line above, and replace the rest of this comment by a useful one.
Executable statements should follow this comment, and should
be separated by periods, with no exclamation points (!!).
Be sure to put any further comments in double-quotes, like this one."
']! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:02'!
assurePreambleExists
	"Make sure there is a StringHolder holding the preamble; if it's found to have reverted to empty contents, put up the template"

	(preamble isEmptyOrNil)
		ifTrue: [preamble := self preambleTemplate]! !

!ChangeSet methodsFor: 'private' stamp: 'di 3/28/2000 14:40'!
atClass: class add: changeType

	(self changeRecorderFor: class)
		noteChangeType: changeType fromClass: class! !

!ChangeSet methodsFor: 'private' stamp: 'di 4/1/2000 12:00'!
atClass: class includes: changeType

	^(changeRecords at: class name ifAbsent: [^false])
		includesChangeType: changeType! !

!ChangeSet methodsFor: 'private' stamp: 'di 4/1/2000 12:00'!
atSelector: selector class: class

	^ (changeRecords at: class name ifAbsent: [^ #none])
		atSelector: selector ifAbsent: [^ #none]! !

!ChangeSet methodsFor: 'method changes' stamp: 'md 8/27/2005 16:37'!
atSelector: selector class: class put: changeType

	selector isDoIt ifTrue: [^ self].
	(self changeRecorderFor: class) atSelector: selector put: changeType.
! !

!ChangeSet methodsFor: 'accessing' stamp: 'BJP 4/24/2001 00:23'!
author
	| author |
	self assurePreambleExists.
	author := self preambleString lineNumber: 3.
	author := author copyFrom: 8 to: author size. "Strip the 'Author:' prefix. Ugly ugly."	
	^author withBlanksTrimmed.
	! !

!ChangeSet methodsFor: 'initialize-release' stamp: 'di 3/29/2000 20:42'!
beIsolationSetFor: aProject

	self isEmpty ifFalse: [self error: 'Must be empty at the start.'].
	isolatedProject := aProject.
	revertable := true.! !

!ChangeSet methodsFor: 'testing' stamp: 'RAA 10/1/2000 12:28'!
belongsToAProject

	Smalltalk at: #Project ifPresent:
		[:projClass | projClass allProjects do:
			[:proj | proj projectChangeSet == self ifTrue: [^ true]]].
	^ false! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ls 10/21/2001 21:09'!
buildMessageForMailOutWithUser: userName

	| message compressBuffer compressStream data compressedStream compressTarget |

	"prepare the message"
	message := MailMessage empty.
	message setField: 'from' toString: userName.
	message setField: 'to' toString: 'squeak-dev@lists.squeakfoundation.org'.
	message setField: 'subject' toString: (self chooseSubjectPrefixForEmail, name). 

	message body: (MIMEDocument contentType: 'text/plain' content: (String streamContents: [ :str |
		str nextPutAll: 'from preamble:'; cr; cr.
		self fileOutPreambleOn: str ])).

	"Prepare the gzipped data"
	data := WriteStream on: String new.
	data header; timeStamp.
	self fileOutPreambleOn: data.
	self fileOutOn: data.
	self fileOutPostscriptOn: data.
	data trailer.
	data := ReadStream on: data contents.
	compressBuffer := ByteArray new: 1000.
	compressStream := GZipWriteStream on: (compressTarget := WriteStream on: (ByteArray new: 1000)).
	[data atEnd]
		whileFalse: [compressStream nextPutAll: (data nextInto: compressBuffer)].
	compressStream close.
	compressedStream := ReadStream on: compressTarget contents asString.

	message addAttachmentFrom: compressedStream withName: (name, '.cs.gz').

	^ message! !

!ChangeSet methodsFor: 'change logging' stamp: 'NS 1/19/2004 18:30'!
changeClass: class from: oldClass
	"Remember that a class definition has been changed.  Record the original structure, so that a conversion method can be built."

	class wantsChangeSetLogging ifFalse: [^ self]. 
	isolationSet ifNotNil:
		["If there is an isolation layer above me, inform it as well."
		isolationSet changeClass: class from: oldClass].
	class isMeta 
		ifFalse: [self atClass: class add: #change]	"normal"
		ifTrue: [((self classChangeAt: class theNonMetaClass name) includes: #add) 
			ifTrue: [self atClass: class add: #add] 	"When a class is defined, the metaclass
				is not recorded, even though it was added.  A further change is
				really just part of the original add."
			ifFalse: [self atClass: class add: #change]].
	self addCoherency: class name.
	(self changeRecorderFor: class) notePriorDefinition: oldClass.
	self noteClassStructure: oldClass! !

!ChangeSet methodsFor: 'class changes' stamp: 'di 3/23/2000 08:12'!
changedClasses
	"Answer an OrderedCollection of changed or edited classes.
	Does not include removed classes.  Sort alphabetically by name."

	"Much faster to sort names first, then convert back to classes.  Because metaclasses reconstruct their name at every comparison in the sorted collection.
	8/91 sw chgd to filter out non-existent classes (triggered by problems with class-renames"

	^ self changedClassNames
		collect: [:className | Smalltalk classNamed: className]
		thenSelect: [:aClass | aClass notNil]! !

!ChangeSet methodsFor: 'class changes' stamp: 'di 4/1/2000 12:00'!
changedClassNames
	"Answer a OrderedCollection of the names of changed or edited classes.
	DOES include removed classes.  Sort alphabetically."

	^ changeRecords keysSortedSafely ! !

!ChangeSet methodsFor: 'method changes' stamp: 'nice 12/27/2009 03:11'!
changedMessageList
	"Used by a message set browser to access the list view information."

	| messageList |
	messageList := OrderedCollection new.
	changeRecords associationsDo: [:clAssoc | | classNameInParts classNameInFull |
		classNameInFull := clAssoc key asString.
		classNameInParts := classNameInFull findTokens: ' '.

		(clAssoc value allChangeTypes includes: #comment) ifTrue:
			[messageList add:
				(MethodReference new
					setClassSymbol: classNameInParts first asSymbol
					classIsMeta: false 
					methodSymbol: #Comment 
					stringVersion: classNameInFull, ' Comment')].

		clAssoc value methodChangeTypes associationsDo: [:mAssoc |
			(#(remove addedThenRemoved) includes: mAssoc value) ifFalse:
				[messageList add:
					(MethodReference new
						setClassSymbol: classNameInParts first asSymbol
						classIsMeta: classNameInParts size > 1 
						methodSymbol: mAssoc key 
						stringVersion: classNameInFull, ' ' , mAssoc key)]]].
	^ messageList asSortedArray! !

!ChangeSet methodsFor: 'method changes' stamp: 'tk 6/7/1999 18:57'!
changedMessageListAugmented
	"Even added classes have all messages in changedMessageList."
	^ self changedMessageList asArray! !

!ChangeSet methodsFor: 'private' stamp: ''!
changed: anAspectSymbol with: aParameter 
	"Allow objects to depend on the ChangeSet class instead of a particular instance 
	of ChangeSet (which may be switched using projects)."

	ChangeSet changed: anAspectSymbol with: aParameter.
	super changed: anAspectSymbol with: aParameter! !

!ChangeSet methodsFor: 'private' stamp: 'nice 1/2/2013 23:11'!
changeRecorderFor: class

	| cname |
	(class isString)
		ifTrue: [ cname := class ]
		ifFalse: [ cname := class name ].

	"Later this will init the changeRecords so according to whether they should be revertable."
	^ changeRecords at: cname
			ifAbsentPut: [ClassChangeRecord new initFor: cname revertable: revertable]! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'cmm 3/9/2010 10:16'!
checkForAlienAuthorship
	"Check to see if there are any methods in the receiver that have author initials other than that of the current author, and open a browser on all found"
	| aList initials |
	(initials := Utilities authorInitials) isEmptyOrNil ifTrue: [ ^self inform: 'No author initials set in this image' ].
	(aList := self methodsWithInitialsOtherThan: initials) size > 0
		ifTrue:
			[ self systemNavigation
				browseMessageList: aList
				name: 'methods in "' , self name , '" whose authoring stamps do not start with "' , initials , '"' ]
		ifFalse: [ ^self inform: 'All methods in "' , self name , '"
have authoring stamps which start with "' , initials , '"' ]! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'cmm 3/9/2010 10:17'!
checkForAnyAlienAuthorship
	"Check to see if there are any versions of any methods in the receiver that have author initials other than that of the current author, and open a browser on all found"
	| aList initials |
	(initials := Utilities authorInitials) isEmptyOrNil ifTrue: [ ^self inform: 'No author initials set in this image' ].
	(aList := self methodsWithAnyInitialsOtherThan: initials) size > 0
		ifTrue:
			[ self systemNavigation
				browseMessageList: aList
				name: 'methods in "' , self name , '" with any authoring stamps not starting with "' , initials , '"' ]
		ifFalse: [ ^self inform: 'All versions of all methods in "' , self name , '"
have authoring stamps which start with "' , initials , '"' ]! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'cmm 2/17/2010 21:59'!
checkForConversionMethods
	"See if any conversion methods are needed"
	| tell choice smart restore renamed listAdd listDrop msgSet list |

	Preferences conversionMethodsAtFileOut ifFalse: [^ self].	"Check preference"
	structures ifNil: [^ self].

	list := OrderedCollection new.
	renamed := OrderedCollection new.
	self changedClasses do: [:class | | newStruct sel oldStruct need rec |
		need := (self atClass: class includes: #new) not.
		need ifTrue: ["Renamed classes."
			(self atClass: class includes: #rename) ifTrue: [
				rec := changeRecords at: class name.
				rec priorName ifNotNil: [
					(structures includesKey: rec priorName) ifTrue: [
						renamed add: class.  need := false]]]].
		need ifTrue: [need := (self atClass: class includes: #change)].
		need ifTrue: [oldStruct := structures at: class name 
									ifAbsent: [need := false.  #()]].
		need ifTrue: [
			newStruct := (Array with: class classVersion), (class allInstVarNames).
			need := (oldStruct ~= newStruct)].
		need ifTrue: [sel := #convertToCurrentVersion:refStream:.
			(#(add change) includes: (self atSelector: sel class: class)) ifFalse: [
				list add: class]].
		].

	list isEmpty & renamed isEmpty ifTrue: [^ self].
	"Ask user if want to do this"
	tell := 'If there might be instances of ', (list asArray, renamed asArray) printString,
		'\in a project (.pr file) on someone''s disk, \please ask to write a conversion method.\'
			withCRs,
		'After you edit the conversion method, you''ll need to fileOut again.\' withCRs,
		'The preference conversionMethodsAtFileOut in category "fileout" controls this feature.'.
	choice := UIManager default chooseFrom:
'Write a conversion method by editing a prototype
These classes are not used in any object file.  fileOut my changes now.
I''m too busy.  fileOut my changes now.
Don''t ever ask again.  fileOut my changes now.' withCRs title: tell. 
	choice = 4 ifTrue: [Preferences disable: #conversionMethodsAtFileOut].
	choice = 2 ifTrue: ["Don't consider this class again in the changeSet"
			list do: [:cls | structures removeKey: cls name ifAbsent: []].
			renamed do: [:cls | | nn | 
				nn := (changeRecords at: cls name) priorName.
				structures removeKey: nn ifAbsent: []]].
	choice ~= 1 ifTrue: [^ self].	"exit if choice 2,3,4"

	listAdd := self askAddedInstVars: list.	"Go through each inst var that was added"
	listDrop := self askRemovedInstVars: list.	"Go through each inst var that was removed"
	list := (listAdd, listDrop) asSet asArray.

	smart := SmartRefStream on: (RWBinaryOrTextStream on: '12345').
	smart structures: structures.
	smart superclasses: superclasses.
	(restore := self class current) == self ifFalse: [
		self class  newChanges: self].	"if not current one"
	msgSet := smart conversionMethodsFor: list.
		"each new method is added to self (a changeSet).  Then filed out with the rest."
	self askRenames: renamed addTo: msgSet using: smart.	"renamed classes, add 2 methods"
	restore == self ifFalse: [self class newChanges: restore].
	msgSet isEmpty ifTrue: [^ self].
	self inform: 'Remember to fileOut again after modifying these methods.'.
	ToolSet 
		browseMessageSet: msgSet 
		name: 'Conversion methods for ', self name 
		autoSelect: nil! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'nice 12/27/2009 03:11'!
checkForSlips
	"Return a collection of method refs with possible debugging code in them."
	| slips |
	slips := OrderedCollection new.
	self changedClasses do:
		[:aClass |
		(self methodChangesAtClass: aClass name) associationsDo: 
				[:mAssoc | | method | (#(remove addedThenRemoved) includes: mAssoc value) ifFalse:
					[method := aClass compiledMethodAt: mAssoc key ifAbsent: [nil].
					method ifNotNil:
						[method hasReportableSlip
							ifTrue: [slips add: aClass name , ' ' , mAssoc key]]]]].
	^ slips! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'sd 4/16/2003 09:16'!
checkForUnclassifiedMethods
	"Open a message list browser on all methods in the current change set that have not been categorized,"

	| aList |
	(aList := self methodsWithoutClassifications) size > 0
		ifFalse:
			[^ self inform: 'All methods in "', self name, '"
are categorized.']
		ifTrue:
			[self systemNavigation  browseMessageList: aList name: 'methods in "', self name, '" which have not been categorized']! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/17/2005 10:48'!
checkForUncommentedClasses
	"Check to see if any classes involved in this change set do not have class comments.  Open up a browser showing all such classes."

	| aList |
	aList := self changedClasses
		select:
			[:aClass | aClass theNonMetaClass organization classComment isEmptyOrNil]
		thenCollect:
			[:aClass  | aClass theNonMetaClass name].

	aList size > 0
		ifFalse:
			[^ self inform: 'All classes involved in this change set have class comments']
		ifTrue:
			[ToolSet openClassListBrowser: aList asSet asSortedArray title: 'Classes in Change Set ', self name, ': classes that lack class comments']! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'sd 4/16/2003 09:16'!
checkForUncommentedMethods
	| aList |
	"Check to see if there are any methods in the receiver that have no comments, and open a browser on all found"

	(aList := self methodsWithoutComments) size > 0
		ifFalse:
			[^ self inform: 'All methods in "', self name, '" have comments']
		ifTrue:
			[self systemNavigation  browseMessageList: aList name: 'methods in "', self name, '" that lack comments']! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'sd 4/29/2003 20:19'!
checkForUnsentMessages
	"Check the change set for unsent messages, and if any are found, open 
	up a message-list browser on them"
	| nameLine allChangedSelectors augList unsent |
	nameLine := '"' , self name , '"'.
	allChangedSelectors := Set new.
	(augList := self changedMessageListAugmented)
		do: [:each | each isValid
				ifTrue: [allChangedSelectors add: each methodSymbol]].
	unsent := self systemNavigation allUnSentMessagesIn: allChangedSelectors.
	unsent size = 0
		ifTrue: [^ self inform: 'There are no unsent 
messages in change set
' , nameLine].
	self systemNavigation
		browseMessageList: (augList
				select: [:each | unsent includes: each methodSymbol])
		name: 'Unsent messages in ' , nameLine! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'rbb 2/18/2005 14:21'!
chooseSubjectPrefixForEmail

	| subjectIndex |

	subjectIndex :=
		(UIManager default chooseFrom: #('Bug fix [FIX]' 'Enhancement [ENH]' 'Goodie [GOODIE]' 'Test suite [TEST]' 'None of the above (will not be archived)')
			title: 'What type of change set\are you submitting to the list?' withCRs).

	^ #('[CS] ' '[FIX] ' '[ENH] ' '[GOODIE] ' '[TEST] ' '[CS] ') at: subjectIndex + 1! !

!ChangeSet methodsFor: 'class changes' stamp: 'di 4/1/2000 12:00'!
classChangeAt: className
	"Return what we know about class changes to this class."

	^ (changeRecords at: className ifAbsent: [^ Set new])
		allChangeTypes! !

!ChangeSet methodsFor: 'accessing' stamp: 'di 4/1/2000 12:00'!
classRemoves

	^ changeRecords keys select:
		[:className | (changeRecords at: className) isClassRemoval]! !

!ChangeSet methodsFor: 'initialize-release' stamp: 'di 4/1/2000 12:00'!
clear 
	"Reset the receiver to be empty.  "

	changeRecords := Dictionary new.
	preamble := nil.
	postscript := nil! !

!ChangeSet methodsFor: 'class changes' stamp: 'NS 1/26/2004 09:46'!
commentClass: class 
	"Include indication that a class comment has been changed."

	class wantsChangeSetLogging ifFalse: [^ self].
	self atClass: class add: #comment! !

!ChangeSet methodsFor: 'class changes' stamp: 'nk 6/26/2002 12:30'!
containsClass: aClass
	^ self changedClasses includes: aClass! !

!ChangeSet methodsFor: 'testing' stamp: 'nice 10/19/2009 20:52'!
containsMethodAtPosition: aFilePosition
	"Answer whether the receiver contains the method logged at the given file position"

	"class: aClassSymbol" "(need class parameter to speed up?)"  "<- dew 9/6/2001"

	changeRecords do:
		[:classChangeRecord |
		classChangeRecord methodChanges do:
			[:methodChangeRecord | | changeType |
			changeType := methodChangeRecord changeType.
			((changeType == #add or: [changeType == #change]) and:
				[methodChangeRecord currentMethod notNil and: [methodChangeRecord currentMethod filePosition = aFilePosition]])
					ifTrue: [^ true]]].
	^ false! !

!ChangeSet methodsFor: 'converting' stamp: 'nice 12/27/2009 03:11'!
convertApril2000: varDict using: smartRefStrm
	| classChanges methodChanges methodRemoves classRemoves cls info |
	"These variables are automatically stored into the new instance:
		('name' 'preamble' 'postscript' 'structures' 'superclasses' ).
	This method is for additional changes.
	It initializes the isolation variables, and then duplicates the logic fo
		assimilateAllChangesFoundIn:."

	revertable := false.
	isolationSet := nil.
	isolatedProject := nil.
	changeRecords := Dictionary new.

	classChanges := varDict at: 'classChanges'.
	classChanges keysDo:
		[:className |
	  	(cls := Smalltalk classNamed: className) ifNotNil:
			[info := classChanges at: className ifAbsent: [Set new].
			info do: [:each | self atClass: cls add: each]]].

	methodChanges := varDict at: 'methodChanges'.
	methodRemoves := varDict at: 'methodRemoves'.
	methodChanges keysDo:
		[:className |
	  	(cls := Smalltalk classNamed: className) ifNotNil:
			[info := methodChanges at: className ifAbsent: [Dictionary new].
			info associationsDo:
				[:assoc | | selector pair | selector := assoc key.
				(assoc value == #remove or: [assoc value == #addedThenRemoved])
					ifTrue:
						[assoc value == #addedThenRemoved
							ifTrue: [self atSelector: selector class: cls put: #add].
						pair := methodRemoves at: {cls name. selector} ifAbsent: [nil] .
						self removeSelector: selector class: cls priorMethod: nil lastMethodInfo: pair]
					ifFalse: 
						[self atSelector: selector class: cls put: assoc value]]]].

	classRemoves := varDict at: 'classRemoves'.
	classRemoves do:
		[:className | self noteRemovalOf: className].

! !

!ChangeSet methodsFor: 'converting' stamp: 'nice 12/27/2009 03:11'!
convertToCurrentVersion: varDict refStream: smartRefStrm

	"major change - 4/4/2000"
	
	varDict at: 'classChanges' ifPresent: [ :x | | newish |
		newish := self convertApril2000: varDict using: smartRefStrm.
		newish == self ifFalse: [^ newish].
		].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!ChangeSet methodsFor: 'testing' stamp: 'RAA 11/13/2000 17:15'!
correspondingProject
	"If the receiver is the current change set for any project, answer it, else answer nil"

	^Project allProjects 
		detect: [ :proj |
			proj projectChangeSet == self
		]
		ifNone: [nil]

! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'nk 10/15/2003 09:55'!
defaultChangeSetDirectory
	^self class defaultChangeSetDirectory! !

!ChangeSet methodsFor: 'accessing' stamp: 'ar 7/16/2005 18:59'!
editPostscript
	"edit the receiver's postscript, in a separate window.  "
	self assurePostscriptExists.
	UIManager default
		edit: self postscript 
		label: 'Postscript for ChangeSet named ', name
		accept:[:aString| self postscript: aString].! !

!ChangeSet methodsFor: 'moving changes' stamp: 'ar 7/16/2005 18:59'!
editPreamble
	"edit the receiver's preamble, in a separate window.  "
	self assurePreambleExists.
	UIManager default
		edit: self preamble 
		label: 'Preamble for ChangeSet named ', name
		accept:[:aString| self preamble: aString]! !

!ChangeSet methodsFor: 'change logging' stamp: 'rw 10/17/2006 22:26'!
event: anEvent
	"Hook for SystemChangeNotifier"
	anEvent itemKind = SystemChangeNotifier classKind ifTrue: [
		anEvent isRemoved 
			ifTrue: [self noteRemovalOf: anEvent item].
		anEvent isAdded 
			ifTrue: [self addClass: anEvent item].
		anEvent isModified 
			ifTrue: [anEvent anyChanges ifTrue: [self changeClass: anEvent item from: anEvent oldItem]].
		anEvent isCommented 
			ifTrue: [self commentClass: anEvent item].
		anEvent isRenamed 
			ifTrue: [self renameClass: anEvent item from: anEvent oldName to: anEvent newName].
		anEvent isReorganized
			ifTrue: [self reorganizeClass: anEvent item].
		anEvent isRecategorized
			ifTrue: [self changeClass: anEvent item from: anEvent item].
	].

	anEvent itemKind = SystemChangeNotifier methodKind ifTrue: [
		anEvent isAdded
			ifTrue: [self noteNewMethod: anEvent item forClass: anEvent itemClass selector: anEvent itemSelector priorMethod: nil].
		anEvent isModified
			ifTrue: [self noteNewMethod: anEvent item forClass: anEvent itemClass selector: anEvent itemSelector priorMethod: anEvent oldItem].
		anEvent isRemoved
			ifTrue: [self removeSelector: anEvent itemSelector class: anEvent itemClass priorMethod: anEvent item lastMethodInfo: {anEvent item sourcePointer. anEvent itemProtocol}].
		anEvent isRecategorized
			ifTrue: [self reorganizeClass: anEvent itemClass].
	].! !

!ChangeSet methodsFor: 'moving changes' stamp: 'di 4/4/2000 11:49'!
expungeEmptyClassChangeEntries

	changeRecords keysAndValuesRemove:
		[:className :classRecord | classRecord hasNoChanges]! !

!ChangeSet methodsFor: 'moving changes' stamp: 'sw 4/19/2000 16:17'!
expungeUniclasses

	changeRecords keysAndValuesRemove:
		[:className :classRecord | className endsWithDigit]! !

!ChangeSet methodsFor: 'class changes' stamp: 'mt 7/22/2016 10:32'!
fatDefForClass: class

	| newDef oldDef oldStrm newStrm outStrm oldVars newVars addedVars superVars |
	
	class isBehavior ifFalse: [^class definition].
	
	newDef := class definition.
	oldDef := (self changeRecorderFor: class) priorDefinition.
	oldDef ifNil: [^ newDef].
	oldDef = newDef ifTrue: [^ newDef].

	oldStrm := ReadStream on: oldDef.
	newStrm := ReadStream on: newDef.
	outStrm := WriteStream on: (String new: newDef size * 2).

	"Merge inst vars from old and new defs..."
	oldStrm upToAll: 'instanceVariableNames'; upTo: $'.
	outStrm 
		nextPutAll: (newStrm upToAll: 'instanceVariableNames'); 
		nextPutAll: 'instanceVariableNames:'.
	newStrm peek = $: ifTrue: [newStrm next].	"may or may not be there, but already written"
	outStrm
		nextPutAll: (newStrm upTo: $'); nextPut: $'.
	oldVars := (oldStrm upTo: $') findTokens: Character separators.
	newVars := (newStrm upTo: $') findTokens: Character separators.
	addedVars := oldVars asSet addAll: newVars; removeAll: oldVars; asOrderedCollection.
	superVars := class allInstVarNames copyWithoutAll: class instVarNames.
	oldVars := oldVars copyWithoutAll: superVars.
	oldVars , addedVars do: [:var | outStrm nextPutAll: var; space].
	outStrm nextPut: $'.

	class isMeta ifFalse:
		["Merge class vars from old and new defs..."
		oldStrm upToAll: 'classVariableNames:'; upTo: $'.
		outStrm nextPutAll: (newStrm upToAll: 'classVariableNames:'); nextPutAll: 'classVariableNames:';
			nextPutAll: (newStrm upTo: $'); nextPut: $'.
		oldVars := (oldStrm upTo: $') findTokens: Character separators.
		newVars := (newStrm upTo: $') findTokens: Character separators.
		addedVars := oldVars asSet addAll: newVars; removeAll: oldVars; asOrderedCollection.
		superVars := class allClassVarNames copyWithoutAll: class classVarNames.
		oldVars := oldVars copyWithoutAll: superVars.
		oldVars , addedVars do: [:var | outStrm nextPutAll: var; space].
		outStrm nextPut: $'].

	outStrm nextPutAll: newStrm upToEnd.
	^ outStrm contents
! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'mt 7/20/2016 13:37'!
fileOut
	"File out the receiver, to a file whose name is a function of the  
	change-set name and either of the date & time or chosen to have a  
	unique numeric tag, depending on the preference  
	'changeSetVersionNumbers'"
	| slips nameToUse |
	self checkForConversionMethods.
	ChangeSet promptForDefaultChangeSetDirectoryIfNecessary.
	nameToUse := Preferences changeSetVersionNumbers
				ifTrue: [self defaultChangeSetDirectory nextNameFor: self name extension: FileStream cs]
				ifFalse: [self name , FileDirectory dot , Utilities dateTimeSuffix, FileDirectory dot , FileStream cs].
	Cursor write showWhile: [ | internalStream |
			internalStream := WriteStream on: (String new: 10000).
			internalStream header; timeStamp.
			self fileOutPreambleOn: internalStream.
			self fileOutOn: internalStream.
			self fileOutPostscriptOn: internalStream.
			internalStream trailer.

			FileStream writeSourceCodeFrom: internalStream baseName: (nameToUse copyFrom: 1 to: nameToUse size - 3) isSt: false useHtml: false.
	].
	Preferences checkForSlips
		ifFalse: [^ self].
	slips := self checkForSlips.
	(slips size > 0
			and: [UIManager default confirm: 'Changeset was filed out. Still, methods in this fileOut have halts\or references to the Transcript or other ''slips'' in them.\Would you like to browse them?' withCRs
				title: 'Browse Slips?']) 
		ifTrue: [self systemNavigation browseMessageList: slips name: 'Possible slips in ' , name]! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'di 3/28/2000 09:35'!
fileOutChangesFor: class on: stream 
	"Write out all the method changes for this class."

	| changes |
	changes := Set new.
	(self methodChangesAtClass: class name) associationsDo: 
		[:mAssoc | (mAssoc value = #remove or: [mAssoc value = #addedThenRemoved])
			ifFalse: [changes add: mAssoc key]].
	changes isEmpty ifFalse: 
		[class fileOutChangedMessages: changes on: stream.
		stream cr]! !

!ChangeSet methodsFor: 'private' stamp: 'tk 3/7/2001 14:06'!
fileOutClassDefinition: class on: stream 
	"Write out class definition for the given class on the given stream, if the class definition was added or changed."

	(self atClass: class includes: #rename) ifTrue:
		[stream nextChunkPut: 'Smalltalk renameClassNamed: #', (self oldNameFor: class), ' as: #', class name; cr].

	(self atClass: class includes: #change) ifTrue: [ "fat definition only needed for changes"
		stream command: 'H3'; nextChunkPut: (self fatDefForClass: class); cr; command: '/H3'.
		DeepCopier new checkClass: class.	"If veryDeepCopy weakly copies some inst 
			vars in this class, warn author when new ones are added." 
	] ifFalse: [
		(self atClass: class includes: #add) ifTrue: [ "use current definition for add"
			stream command: 'H3'; nextChunkPut: class definition; cr; command: '/H3'.
			DeepCopier new checkClass: class.	"If veryDeepCopy weakly copies some inst 
				vars in this class, warn author when new ones are added." 
		].
	].

	(self atClass: class includes: #comment) ifTrue:
		[class theNonMetaClass organization putCommentOnFile: stream numbered: 0 moveSource: false forClass: class theNonMetaClass.
		stream cr].

! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'dvf 9/27/2005 19:04'!
fileOutOn: stream 
	"Write out all the changes the receiver knows about"

	| classList traits classes traitList list |
	(self isEmpty and: [stream isKindOf: FileStream])
		ifTrue: [self inform: 'Warning: no changes to file out'].
		
	traits := self changedClasses reject: [:each | each isBehavior].
	classes := self changedClasses select: [:each | each isBehavior].
	traitList := self class traitsOrder: traits asOrderedCollection.
	classList := self class superclassOrder: classes asOrderedCollection.
	list := OrderedCollection new
		addAll: traitList;
		addAll: classList;
		yourself.
	
	"First put out rename, max classDef and comment changes."
	list do: [:aClass | self fileOutClassDefinition: aClass on: stream].

	"Then put out all the method changes"
	list do: [:aClass | self fileOutChangesFor: aClass on: stream].

	"Finally put out removals, final class defs and reorganization if any"
	list reverseDo: [:aClass | self fileOutPSFor: aClass on: stream].

	self classRemoves asSortedCollection do:
		[:aClassName | stream nextChunkPut: 'Smalltalk removeClassNamed: #', aClassName; cr].! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'di 3/29/1999 13:35'!
fileOutPostscriptOn: stream 
	"If the receiver has a postscript, put it out onto the stream.  "

	| aString |
	aString := self postscriptString.
	(aString ~~ nil and: [aString size > 0])
		ifTrue:
			[stream nextChunkPut: aString "surroundedBySingleQuotes".
			stream cr; cr]! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'di 3/29/1999 14:58'!
fileOutPreambleOn: stream 
	"If the receiver has a preamble, put it out onto the stream.  "

	| aString |
	aString := self preambleString.
	(aString ~~ nil and: [aString size > 0])
		ifTrue:
			[stream nextChunkPut: aString "surroundedBySingleQuotes".
			stream cr; cr]! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'nice 12/27/2009 03:11'!
fileOutPSFor: class on: stream 
	"Write out removals and initialization for this class."

	| dict classRecord currentDef |
	classRecord := changeRecords at: class name ifAbsent: [^ self].
	dict := classRecord methodChangeTypes.
	dict keysSortedSafely do:
		[:key | | changeType | changeType := dict at: key.
		(#(remove addedThenRemoved) includes: changeType)
			ifTrue: [stream nextChunkPut: class name,
						' removeSelector: ', key storeString; cr]
			ifFalse: [(key = #initialize and: [class isMeta]) ifTrue:
						[stream nextChunkPut: class soleInstance name, ' initialize'; cr]]].
	((classRecord includesChangeType: #change)
		and: [(currentDef := class definition) ~= (self fatDefForClass: class)]) ifTrue:
		[stream command: 'H3'; nextChunkPut: currentDef; cr; command: '/H3'].
	(classRecord includesChangeType: #reorganize) ifTrue:
		[class fileOutOrganizationOn: stream.
		stream cr]! !

!ChangeSet methodsFor: 'moving changes' stamp: 'di 4/4/2000 12:40'!
forgetAllChangesFoundIn: otherChangeSet
	"Remove from the receiver all method changes found in aChangeSet. The intention is facilitate the process of factoring a large set of changes into disjoint change sets.  To use:  in a change sorter, copy over all the changes you want into some new change set, then use the subtract-other-side feature to subtract those changes from the larger change set, and continue in this manner."

	otherChangeSet == self ifTrue: [^ self].
	otherChangeSet changedClassNames do:
		[:className | self forgetChangesForClass: className in: otherChangeSet].
	self expungeEmptyClassChangeEntries.

"  Old code...
	aChangeSet changedClassNames do: 
		[:className |
			(cls := Smalltalk classNamed: className) ~~ nil ifTrue:
				[itsMethodChanges := aChangeSet methodChanges at: className 
						ifAbsent: [Dictionary new].
				itsMethodChanges associationsDo: [:assoc | 
					self forgetChange: assoc value forSelector: assoc key class: cls].
				myClassChange := self classChangeAt: className.
				myClassChange size > 0 ifTrue:
					[(aChangeSet classChangeAt: className) do:
						[:aChange | myClassChange remove: aChange ifAbsent: []]].
				self noteClassForgotten: className]].

	aChangeSet classRemoves do:
		[:className | (recorder := changeRecords at: className ifAbsent: [])
			ifNotNil: [recorder forgetClassRemoval]].
	self expungeEmptyClassChangeEntries
"
! !

!ChangeSet methodsFor: 'moving changes' stamp: 'di 4/4/2000 12:04'!
forgetChangesForClass: className in: otherChangeSet
	"See forgetAllChangesFoundIn:.  Used in culling changeSets."

	(self changeRecorderFor: className)
			forgetChangesIn: (otherChangeSet changeRecorderFor: className).
	self noteClassForgotten: className
! !

!ChangeSet methodsFor: 'method changes' stamp: 'sw 4/19/2001 19:45'!
hasAnyChangeForSelector: aSelector
	"Answer whether the receiver has any change under the given selector, whether it be add, change, or remove, for any class"

	changeRecords do:
		[:aRecord | (aRecord changedSelectors  includes: aSelector)
			ifTrue:	[^ true]].
	^ false! !

!ChangeSet methodsFor: 'accessing' stamp: 'sw 6/29/1999 14:44'!
hasPostscript
	^ postscript notNil! !

!ChangeSet methodsFor: 'moving changes' stamp: 'sw 3/5/1999 19:27'!
hasPreamble
	^ preamble notNil! !

!ChangeSet methodsFor: 'initialize-release' stamp: 'di 4/6/2001 09:40'!
initialize 
	"Initialize the receiver to be empty."

	name ifNil:
		[^ self error: 'All changeSets must be registered, as in ChangeSorter newChangeSet'].
	revertable := false.
	self clear.
! !

!ChangeSet methodsFor: 'isolation layers' stamp: 'di 3/29/2000 14:47'!
invoke

	"Do the first part of the invoke operation -- no particular hurry."
	changeRecords do: [:changeRecord | changeRecord invokePhase1].

	"Complete the invoke process -- this must be very simple."
	"Replace method dicts for any method changes."
	changeRecords do: [:changeRecord | changeRecord invokePhase2].
	Behavior flushCache.

! !

!ChangeSet methodsFor: 'testing' stamp: 'RAA 10/19/2000 13:17'!
isEmpty
	"Answer whether the receiver contains any elements."
	
	changeRecords ifNil: [^true].
	^ changeRecords isEmpty ! !

!ChangeSet methodsFor: 'initialize-release' stamp: ''!
isMoribund
	"Answer whether the receiver is obsolete and about to die; part of an effort to get such guys cleared out from the change sorter.  2/7/96 sw"

	^ name == nil ! !

!ChangeSet methodsFor: 'isolation layers' stamp: 'di 4/13/2000 12:47'!
isolatedProject
	"Return the isolated project for which I am the changeSet."

	^ isolatedProject! !

!ChangeSet methodsFor: 'isolation layers' stamp: 'di 3/29/2000 13:59'!
isolationSet: setOrNil

	setOrNil == self
		ifTrue: [isolationSet := nil]  "Means this IS the isolation set"
		ifFalse: [isolationSet := setOrNil]! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'mt 7/20/2016 13:46'!
lookForSlips
	"Scan the receiver for changes that the user may regard as slips to be remedied"

	| slips nameLine msg |
	nameLine := '
"', self name, '"
'.
	(slips := self checkForSlips) size = 0 ifTrue:
		[^ self inform: ('No slips detected in change set\"{1}".' translated withCRs format: {self name})].

	msg := slips size = 1
		ifTrue:
			[ 'One method in change set "{1}"\has a halt, reference to the Transcript,
and/or some other ''slip'' in it.\Would you like to browse it?' translated withCRs format: {self name}]
		ifFalse:
			[ '{1} methods in change set "{2}"\have halts or references to the
Transcript or other ''slips'' in them.\Would you like to browse them?' translated withCRs format: {slips size. self name}].

	(UIManager default confirm: msg title: 'Browse Slips?')
		ifTrue: [self systemNavigation
			browseMessageList: slips
			name: 'Possible slips in ', self name]! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'wiz 2/12/2010 19:01'!
mailOut
	"Email a compressed version of this changeset to the squeak-dev list, so that it can be shared with everyone.  (You will be able to edit the email before it is sent.)"

	| userName message slips |

	userName := MailSender userName.

	self checkForConversionMethods.
	message := Cursor write showWhile: [self buildMessageForMailOutWithUser: userName].

	MailSender sendMessage: message.

	Preferences checkForSlips ifFalse: [^ self].
	slips := self checkForSlips.
	(slips size > 0 and: [self confirm: 'Methods in this fileOut have halts
or references to the Transcript
or other ''slips'' in them.
Would you like to browse them?'])
		ifTrue: [self systemNavigation browseMessageList: slips name: 'Possible slips in ' , name]
! !

!ChangeSet methodsFor: 'method changes' stamp: 'RAA 5/28/2001 12:05'!
messageListForChangesWhich: aBlock ifNone: ifEmptyBlock

	| answer |

	answer := self changedMessageListAugmented select: [ :each |
		aBlock value: each actualClass value: each methodSymbol
	].
	answer isEmpty ifTrue: [^ifEmptyBlock value].
	^answer
! !

!ChangeSet methodsFor: 'accessing' stamp: 'nice 12/27/2009 03:11'!
methodChanges

	| methodChangeDict |
	methodChangeDict := Dictionary new.
	changeRecords associationsDo:
		[:assn | | changeTypes |
		changeTypes := assn value methodChangeTypes.
		changeTypes isEmpty ifFalse: [methodChangeDict at: assn key put: changeTypes]].
	^ methodChangeDict! !

!ChangeSet methodsFor: 'method changes' stamp: 'di 4/1/2000 12:00'!
methodChangesAtClass: className
	"Return an old-style dictionary of method change types."

	^(changeRecords at: className ifAbsent: [^ Dictionary new])
		methodChangeTypes! !

!ChangeSet methodsFor: 'accessing' stamp: 'di 3/29/2000 16:22'!
methodInfoFromRemoval: classAndSelector

	^ (self changeRecorderFor: classAndSelector first)
		infoFromRemoval: classAndSelector last! !

!ChangeSet methodsFor: 'moving changes' stamp: 'nice 12/27/2009 03:11'!
methodsWithAnyInitialsOtherThan: myInits
	"Return a collection of method refs whose author appears to be different from the given one, even historically"
	| slips |
	slips := Set new.
	self changedClasses do: [:aClass |
		(self methodChangesAtClass: aClass name) associationsDo: [ :mAssoc | | method |
			(#(remove addedThenRemoved) includes: mAssoc value) ifFalse:
				[method := aClass compiledMethodAt: mAssoc key ifAbsent: [nil].
				method ifNotNil: [
					(aClass changeRecordsAt: mAssoc key) do: [ :chg | | aTimeStamp |
						aTimeStamp := chg stamp.
						(aTimeStamp notNil and: [(aTimeStamp beginsWith: myInits) not])
							ifTrue: [slips add: aClass name , ' ' , mAssoc key]]]]]].
	^ slips! !

!ChangeSet methodsFor: 'moving changes' stamp: 'fbs 3/4/2013 20:52'!
methodsWithInitialsOtherThan: myInits
	"Return a collection of method refs whose author appears to be different from the given one"
	| slips |
	slips := OrderedCollection new.
	self changedClasses do:
		[:aClass |
		(self methodChangesAtClass: aClass name) associationsDo: 
				[:mAssoc | | aTimeStamp method | (#(remove addedThenRemoved) includes: mAssoc value) ifFalse:
					[method := aClass compiledMethodAt: mAssoc key ifAbsent: [nil].
					method ifNotNil:
						[((aTimeStamp := method timeStamp) notNil and:
							[(aTimeStamp beginsWith: myInits) not])
								ifTrue: [slips add: aClass name , ' ' , mAssoc key]]]]].
	^ slips

	"Smalltalk browseMessageList: (ChangeSet current methodsWithInitialsOtherThan: 'sw') name: 'authoring problems'"! !

!ChangeSet methodsFor: 'testing' stamp: 'nice 12/27/2009 03:11'!
methodsWithoutClassifications
	"Return a collection representing methods in the receiver which have not been categorized"

	| slips notClassified |

	notClassified := {'as yet unclassified' asSymbol. #all}.
	slips := OrderedCollection new.
	self changedClasses do:
		[:aClass |
		(self methodChangesAtClass: aClass name) associationsDo: 
				[:mAssoc | | aSelector | (aClass includesSelector:  (aSelector := mAssoc key)) ifTrue:
						[(notClassified includes: (aClass organization categoryOfElement: aSelector))
								ifTrue: [slips add: aClass name , ' ' , aSelector]]]].
	^ slips

	"Smalltalk browseMessageList: (ChangeSet current methodsWithoutClassifications) name: 'unclassified methods'"! !

!ChangeSet methodsFor: 'moving changes' stamp: 'nice 10/20/2009 20:55'!
methodsWithoutComments
	"Return a collection representing methods in the receiver which have no precode comments"

	| slips |
	slips := OrderedCollection new.
	self changedClasses do:
		[:aClass |
		(self methodChangesAtClass: aClass name) associationsDo: 
				[:mAssoc | (#(remove addedThenRemoved) includes: mAssoc value) ifFalse:
					[(aClass includesSelector:  mAssoc key) ifTrue:
						[(aClass firstPrecodeCommentFor: mAssoc key) isEmptyOrNil
								ifTrue: [slips add: aClass name , ' ' , mAssoc key]]]]].
	^ slips

	"Smalltalk browseMessageList: (ChangeSet current methodsWithoutComments) name: 'methods lacking comments'"! !

!ChangeSet methodsFor: 'accessing' stamp: ''!
name
	"The name of this changeSet.
	 2/7/96 sw: If name is nil, we've got garbage.  Help to identify."

	^ name == nil
		ifTrue:
			['<no name -- garbage?>']
		ifFalse:
			[name]! !

!ChangeSet methodsFor: 'accessing' stamp: ''!
name: anObject
	name := anObject! !

!ChangeSet methodsFor: 'class changes' stamp: 'tk 6/9/1999 19:54'!
noteClassForgotten: className
	"Remove from structures if class is not a superclass of some other one we are remembering"

	structures ifNil: [^ self].
	Smalltalk at: className ifPresent: [:cls |
		cls subclasses do: [:sub | (structures includesKey: sub) ifTrue: [
			^ self]]].  "No delete"
	structures removeKey: className ifAbsent: [].! !

!ChangeSet methodsFor: 'class changes' stamp: 'dvf 9/27/2005 19:05'!
noteClassStructure: aClass
	"Save the instance variable names of this class and all of its superclasses.  Later we can tell how it changed and write a conversion method.  The conversion method is used when old format objects are brought in from the disk from ImageSegment files (.extSeg) or SmartRefStream files (.obj .morph .bo .sp)."

	| clsName |
	aClass isBehavior ifFalse: [^ self].
	
	structures ifNil: [structures := Dictionary new.
				superclasses := Dictionary new].
	clsName := (aClass name asLowercase beginsWith: 'anobsolete') 
		ifTrue: [(aClass name copyFrom: 11 to: aClass name size) asSymbol]
		ifFalse: [aClass name].
	(structures includesKey: clsName) ifFalse: [
		structures at: clsName put: 
			((Array with: aClass classVersion), (aClass allInstVarNames)).
		superclasses at: clsName put: aClass superclass name].
	"up the superclass chain"
	aClass superclass ifNotNil: [self noteClassStructure: aClass superclass].
! !

!ChangeSet methodsFor: 'change logging' stamp: 'di 3/29/2000 11:08'!
noteNewMethod: newMethod forClass: class selector: selector priorMethod: methodOrNil

	class wantsChangeSetLogging ifFalse: [^ self].
	isolationSet ifNotNil:
		["If there is an isolation layer above me, inform it as well."
		isolationSet noteNewMethod: newMethod forClass: class selector: selector
				priorMethod: methodOrNil].
	(self changeRecorderFor: class)
		noteNewMethod: newMethod selector: selector priorMethod: methodOrNil
! !

!ChangeSet methodsFor: 'class changes' stamp: 'NS 1/19/2004 17:49'!
noteRemovalOf: class
	"The class is about to be removed from the system.
	Adjust the receiver to reflect that fact."

	class wantsChangeSetLogging ifFalse: [^ self].
	(self changeRecorderFor: class)
		noteChangeType: #remove fromClass: class.
	changeRecords removeKey: class class name ifAbsent: [].! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/15/2005 21:27'!
objectForDataStream: refStrm
	"I am about to be written on an object file.  Write a path to me in the other system instead."

	refStrm projectChangeSet == self ifTrue: [^ self].

	"try to write reference for me"
	^ DiskProxy 
		global: #ChangeSet
		selector: #existingOrNewChangeSetNamed: 
		args: (Array with: self name)
"===
	refStrm replace: self with: nil.
	^ nil
==="
! !

!ChangeSet methodsFor: 'testing' stamp: 'sw 8/3/1998 16:25'!
okayToRemove
	^ self okayToRemoveInforming: true! !

!ChangeSet methodsFor: 'testing' stamp: 'sd 5/23/2003 14:24'!
okayToRemoveInforming: aBoolean
	"Answer whether it is okay to remove the receiver.  If aBoolean is true, inform the receiver if it is not okay"

	| aName |
	aName := self name.
	self == self class current ifTrue:
		[aBoolean ifTrue: [self inform: 'Cannot remove "', aName, '"
because it is the 
current change set.'].
		^ false].

	self belongsToAProject ifTrue:
		[aBoolean ifTrue: [self inform: 'Cannot remove "', aName, '" 
because it belongs to a 
project.'].
			^ false].

	^ true
! !

!ChangeSet methodsFor: 'private' stamp: 'di 4/1/2000 12:00'!
oldNameFor: class

	^ (changeRecords at: class name) priorName! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:03'!
postscript
	"Answer the string representing the postscript.  "
	^postscript ifNotNil:[postscript isString ifTrue:[postscript] ifFalse:[postscript contents asString]]! !

!ChangeSet methodsFor: 'accessing' stamp: 'ar 7/16/2005 18:04'!
postscriptHasDependents
	^false! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:03'!
postscriptString
	"Answer the string representing the postscript.  "
	^self postscript! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:03'!
postscriptString: aString
	"Establish aString as the new contents of the postscript.  "
	self postscript: aString! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:03'!
postscript: aString
	"Answer the string representing the postscript.  "
	postscript := aString! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 17:55'!
preamble
	"Answer the string representing the preamble"
	^preamble ifNotNil:[preamble isString ifTrue:[preamble] ifFalse:[preamble contents asString]]! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:00'!
preambleString
	"Answer the string representing the preamble"

	^self preamble! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:00'!
preambleString: aString
	"Establish aString as the new contents of the preamble.  "
	self preamble: aString.! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'nk 7/2/2003 10:47'!
preambleTemplate
	"Answer a string that will form the default contents for a change set's preamble.
	Just a first stab at what the content should be."

	^ String streamContents: [:strm |
		strm nextPutAll: '"Change Set:'.  "NOTE: fileIn recognizes preambles by this string."
		strm tab;tab; nextPutAll: self name.
		strm cr; nextPutAll: 'Date:'; tab; tab; tab; nextPutAll: Date today printString.
		strm cr; nextPutAll: 'Author:'; tab; tab; tab; nextPutAll: Preferences defaultAuthorName.
		strm cr; cr; nextPutAll: '<your descriptive text goes here>"']
"ChangeSet current preambleTemplate"! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:00'!
preamble: aString
	"Establish aString as the new contents of the preamble.  "

	preamble := aString! !

!ChangeSet methodsFor: 'accessing' stamp: ''!
printOn: aStream
	"2/7/96 sw: provide the receiver's name in the printout"
	super printOn: aStream.
	aStream nextPutAll: ' named ', self name! !

!ChangeSet methodsFor: 'testing' stamp: 'RAA 9/27/2000 22:40'!
projectsBelongedTo
	"Answer a list of all the projects for which the receiver is the current change set"

	^ Project allProjects select: [:proj | proj projectChangeSet == self]
! !

!ChangeSet methodsFor: 'moving changes' stamp: 'di 4/1/2000 12:00'!
removeClassAndMetaClassChanges: class
	"Remove all memory of changes associated with this class and its metaclass.  7/18/96 sw"

	changeRecords removeKey: class name ifAbsent: [].
	changeRecords removeKey: class class name ifAbsent: [].
! !

!ChangeSet methodsFor: 'moving changes' stamp: 'yo 8/30/2002 13:59'!
removeClassChanges: class
	"Remove all memory of changes associated with this class"
	| cname |
	(class isString)
		ifTrue: [ cname := class ]
		ifFalse: [ cname := class name ].

	changeRecords removeKey: cname ifAbsent: [].
	self noteClassForgotten: cname.! !

!ChangeSet methodsFor: 'accessing' stamp: 'sw 6/29/1999 14:48'!
removePostscript
	postscript := nil! !

!ChangeSet methodsFor: 'moving changes' stamp: 'sw 3/5/1999 19:32'!
removePreamble
	preamble := nil! !

!ChangeSet methodsFor: 'method changes' stamp: 'di 4/4/2000 11:14'!
removeSelectorChanges: selector class: class 
	"Remove all memory of changes associated with the argument, selector, in 
	this class."

	| chgRecord |
	(chgRecord := changeRecords at: class name ifAbsent: [^ self])
		removeSelector: selector.
	chgRecord hasNoChanges ifTrue: [changeRecords removeKey: class name]! !

!ChangeSet methodsFor: 'change logging' stamp: 'di 3/29/2000 12:29'!
removeSelector: selector class: class priorMethod: priorMethod lastMethodInfo: info
	"Include indication that a method has been forgotten.
	info is a pair of the source code pointer and message category
	for the method that was removed."

	class wantsChangeSetLogging ifFalse: [^ self].
	isolationSet ifNotNil:
		["If there is an isolation layer above me, inform it as well."
		isolationSet removeSelector: selector class: class
				priorMethod: priorMethod lastMethodInfo: info].
	(self changeRecorderFor: class)
		noteRemoveSelector: selector priorMethod: priorMethod lastMethodInfo: info
! !

!ChangeSet methodsFor: 'change logging' stamp: 'rw 10/19/2006 17:52'!
renameClass: class from: oldName to: newName 
	"Include indication that a class has been renamed."

	| recorder oldMetaClassName newMetaClassName |
	isolationSet ifNotNil:
		["If there is an isolation layer above me, inform it as well."
		isolationSet renameClass: class as: newName].
	(recorder := self changeRecorderFor: oldName)
		noteChangeType: #rename;
		noteNewName: newName asSymbol.
		
	"store under new name (metaclass too)"
	changeRecords at: newName put: recorder.
	changeRecords removeKey: oldName.
	self noteClassStructure: class.

	newMetaClassName := newName, ' class'.
	oldMetaClassName := oldName, ' class'.
	recorder := changeRecords at: oldMetaClassName ifAbsent: [^ nil].
	changeRecords at: newMetaClassName put: recorder.
	changeRecords removeKey: oldMetaClassName.
	recorder noteNewName: newMetaClassName! !

!ChangeSet methodsFor: 'class changes' stamp: ''!
reorganizeClass: class 
	"Include indication that a class was reorganized."

	self atClass: class add: #reorganize! !

!ChangeSet methodsFor: 'isolation layers' stamp: 'di 3/29/2000 14:47'!
revoke

	"Do the first part of the revoke operation -- this must be very simple."
	"Replace original method dicts if there are method changes."
	changeRecords do: [:changeRecord | changeRecord revokePhase1].
	Behavior flushCache.

	"Complete the revoke process -- no particular hurry."
	changeRecords do: [:changeRecord | changeRecord revokePhase2].
! !

!ChangeSet methodsFor: 'method changes' stamp: 'SqR 6/13/2000 19:16'!
selectorsInClass: aClassName
	"Used by a ChangeSorter to access the list methods."

	^ (changeRecords at: aClassName ifAbsent: [^#()]) changedSelectors! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'ar 7/16/2005 18:02'!
setPreambleToSay: aString
	"Make aString become the preamble of this change set"
	self preamble: aString! !

!ChangeSet methodsFor: 'accessing' stamp: 'tk 6/8/1999 22:25'!
structures
	^structures! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'di 9/24/1999 12:33'!
summaryString
	"Answer the string summarizing this changeSet"

	^ self summaryStringDelta: 0
"
To summarize all recent changeSets on a file...
(FileStream newFileNamed: 'Summaries.txt') nextPutAll:
	(String streamContents:
		[:s | (ChangeSorter changeSetsNamedSuchThat:
			[:name | name first isDigit and: [name initialIntegerOrNil >= 948]])
			 do: [:cs | s nextPutAll: cs summaryString; cr]]);
		close

To list all changeSets with a certain string in the preamble...
	(FileStream newFileNamed: 'MyUpdates.txt') nextPutAll:
		(String streamContents:
			[:s | ChangeSorter gatherChangeSetRevertables do:
				[:cs | (cs preambleString notNil
					and: [cs preambleString includesSubString: 'Author Name'])
				 	ifTrue: [s nextPutAll: cs summaryString; cr]]]);
		close
"! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'nice 6/11/2010 21:21'!
summaryStringDelta: delta
	"Answer the string summarizing this changeSet"
	
	^ String streamContents:
		[:s | | line author s2 ps intName date |
		intName := self name splitInteger.
		intName first isNumber
			ifTrue: [s nextPutAll: (intName first + delta) printString , intName last]
			ifFalse: [s nextPutAll: intName first  "weird convention of splitInteger"].
		(ps := self preambleString)
			ifNil: [s cr]
			ifNotNil:
			[s2 := ReadStream on: ps.
			s2 match: 'Date:'; skipSeparators.  date := s2 nextLine.
			s2 match: 'Author:'; skipSeparators.  author := s2 nextLine.
			s nextPutAll: ' -- '; nextPutAll: author; nextPutAll: ' -- '; nextPutAll: date; cr.
			[s2 atEnd] whileFalse:
				[line := s2 nextLine.
				(line isEmpty or: [line = '"']) ifFalse: [s nextPutAll: line; cr]]]].
! !

!ChangeSet methodsFor: 'accessing' stamp: 'tk 6/8/1999 22:25'!
superclasses
	^superclasses! !

!ChangeSet methodsFor: 'class changes' stamp: 'di 5/16/2000 09:03'!
trimHistory 
	"Drop non-essential history:  methods added and then removed, as well as rename and reorganization of newly-added classes."

	changeRecords do: [:chgRecord | chgRecord trimHistory]! !

!ChangeSet methodsFor: 'isolation layers' stamp: 'di 3/23/2000 12:00'!
uninstall

	self halt.
! !

!ChangeSet methodsFor: 'fileIn/Out' stamp: 'sd 1/16/2004 21:31'!
verboseFileOut
	"File out the receiver, to a file whose name is a function of the change-set name and either of the date & time or chosen to have a unique numeric tag, depending on the preference 'changeSetVersionNumbers'"

	ChangeSet current fileOut.
	Transcript cr; show: 'Changes filed out ', Date dateAndTimeNow printString! !

!ChangeSet methodsFor: 'initialize-release' stamp: 'sw 3/6/1999 09:31'!
veryDeepCopyWith: deepCopier
	"Return self; this is NOT the way to launch new change sets!! Having this method here allows Change Sorters to be in parts bins"! !

!ChangeSet methodsFor: 'initialize-release' stamp: 'di 3/23/2000 12:14'!
wither
	"The receiver is to be clobbered.  Clear it out.  2/7/96 sw"

	self clear.
	name := nil! !

!ChangeSet methodsFor: 'initialize-release' stamp: 'di 9/21/2000 15:29'!
zapHistory 
	"Much stronger than trimHistory, but it should still leave the changeSet in good shape.
	Must not be done on revertable changeSets
		ChangeSet allInstancesDo: [:cs | cs zapHistory]."

	revertable ifTrue: [^ self].  "No can do"
	changeRecords do: [:chgRecord | chgRecord zapHistory]! !

Object subclass: #ClassChangeRecord
	instanceVariableNames: 'inForce revertable changeTypes thisDefinition priorDefinition thisName priorName thisOrganization priorOrganization thisComment priorComment thisMD priorMD methodChanges'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Changes'!

!ClassChangeRecord commentStamp: 'fbs 9/6/2013 17:32' prior: 0!
A ClassChangeRecorder keeps track of most substantive changes permissible in a project, isolated or not.

Structure:
inForce		a boolean
			Tells whether these changes are in effect.
			true for all changeSets in and above the current project.
			It should be sufficient only to record this for the changeSet
			as a whole, but this redundancy could help in error recovery.
classIsLocal	a boolean
			True if and only if this class is defined in this layer of the
			project structure.
changeTypes an identitySet
			Summarizes which changes have been made in this class.
			Values include #comment, #reorganize, #rename,
			and the four more summarized below.
thisName	a string
			Retains the class name for this layer.
priorName	a string
			Preserves the prior name.
thisComment	a text
			Retains the class comment for this layer.
priorComment	a text
			Preserves the prior comment.
thisOrganization	a classOrganizer
			Retains the class organization for this layer.
priorOrganization	a classOrganizer
			Preserves the prior organization.
thisMD	a methodDictionary
			Used to prepare changes for nearly atomic invocation
			of this layer (see below).
priorMD	a methodDictionary
			Preserves the state of an altered class as it exists in the next
			outer layer of the project structure.
methodChanges		a dictionary of classChangeRecords
			Retains all the method changes for this layer.

Four of the possible changeTypes are maintained in a mutually exclusive set, analogously to MethodChangeRecords.  Here is a simple summary of the relationship between these four changeType symbols and the recording of prior state
			|	prior == nil			|	prior not nil	
	---------	|----------------------------	|--------------------
	add		|	add					|	change
	---------	|----------------------------	|--------------------
	remove	|	addedThenRemoved	|	remove

A classChangeRecorder is notified of changes by the method
		noteMethodChange: <ClassChangeRecord>.
ClassChangeRecorders are designed to invoke a set of changes relative to the definition of a class in an prior layer.  It is important that both invocation and revocation of these changes take place in a nearly atomic fashion so that interdependent changes will be adopted as a whole, and so that only one flush of the method cache should be necessary.  A further reason for revocation to be simple is that it may be requested as an attempt to recover from an error in a project that is failing.!

!ClassChangeRecord methodsFor: 'all changes' stamp: 'di 4/2/2000 21:39'!
allChangeTypes

	| chgs |
	(priorName ~~ nil and: [changeTypes includes: #rename]) ifTrue:
		[(chgs := changeTypes copy) add: 'oldName: ' , priorName.
		^ chgs].
	^ changeTypes! !

!ClassChangeRecord methodsFor: 'all changes' stamp: 'nice 12/27/2009 02:53'!
assimilateAllChangesIn: otherRecord

	otherRecord isClassRemoval ifTrue: [^ self noteChangeType: #remove].

	otherRecord allChangeTypes do:
		[:chg | self noteChangeType: chg fromClass: self realClass].

	otherRecord methodChanges associationsDo:
		[:assn | | selector changeRecord changeType |
		selector := assn key. changeRecord := assn value.
		changeType := changeRecord changeType.
		(changeType == #remove or: [changeType == #addedThenRemoved])
			ifTrue:
				[changeType == #addedThenRemoved
					ifTrue: [self atSelector: selector put: #add].
				self noteRemoveSelector: selector priorMethod: nil
						lastMethodInfo: changeRecord methodInfoFromRemoval]
			ifFalse: 
				[self atSelector: selector put: changeType]].
! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'di 3/28/2000 10:38'!
atSelector: selector ifAbsent: absentBlock

	^ (methodChanges at: selector ifAbsent: absentBlock)
		changeType! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'di 3/28/2000 11:01'!
atSelector: selector put: changeType

	(self findOrMakeMethodChangeAt: selector priorMethod: nil)
		noteChangeType: changeType! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'di 3/28/2000 10:07'!
changedSelectors
	"Return a set of the changed or removed selectors."

	^ methodChanges keys! !

!ClassChangeRecord methodsFor: 'definition' stamp: 'di 3/27/2000 22:06'!
checkCoherence
	"If I recreate the class then don't remove it"

	(changeTypes includes: #remove) ifTrue:
		[changeTypes remove: #remove.
		changeTypes add: #change].
	(changeTypes includes: #addedThenRemoved) ifTrue:
		[changeTypes remove: #addedThenRemoved.
		changeTypes add: #add].
! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'eem 6/11/2008 16:50'!
compileAll: newClass from: oldClass
	"Something about this class has changed.  Locally retained methods must be recompiled.
	NOTE:  You might think that if this changeSet is in force, then we can just note
	the new methods but a lower change set may override and be in force which
	would mean that only the overriding copies go recompiled.  Just do it."

	methodChanges associationsDo:
		[:assn | | sel changeType changeRecord newMethod |
		sel := assn key.
		changeRecord := assn value.
		changeType := changeRecord changeType.
		(changeType == #add or: [changeType == #change]) ifTrue:
			[newMethod := newClass
				recompileNonResidentMethod: changeRecord currentMethod
				atSelector: sel from: oldClass.
			changeRecord noteNewMethod: newMethod]]! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'Lukas Renggli 11/2/2009 00:29'!
findOrMakeMethodChangeAt: selector priorMethod: priorMethod 
	^methodChanges at: selector
		ifAbsentPut: [MethodChangeRecord new priorMethod: priorMethod]! !

!ClassChangeRecord methodsFor: 'removal' stamp: 'eem 6/11/2008 16:50'!
forgetChangesIn: otherRecord
	"See forgetAllChangesFoundIn:.  Used in culling changeSets."

	| cls otherMethodChanges |
	(cls := self realClass) == nil ifTrue: [^ self].  "We can do better now, though..."
	otherMethodChanges := otherRecord methodChangeTypes.
	otherMethodChanges associationsDo:
		[:assoc | | selector actionToSubtract |
		selector := assoc key. actionToSubtract := assoc value.
		(cls includesSelector: selector)
			ifTrue: [(#(add change) includes: actionToSubtract)
					ifTrue: [methodChanges removeKey: selector ifAbsent: []]]
			ifFalse: [(#(remove addedThenRemoved) includes: actionToSubtract)
					ifTrue: [methodChanges removeKey: selector ifAbsent: []]]].
	changeTypes isEmpty ifFalse:
		[changeTypes removeAllFoundIn: otherRecord allChangeTypes.
		(changeTypes includes: #rename) ifFalse:
			[changeTypes removeAllSuchThat: [:x | x beginsWith: 'oldName: ']]]! !

!ClassChangeRecord methodsFor: 'removal' stamp: 'di 3/23/2000 12:27'!
forgetClassRemoval

	self halt.! !

!ClassChangeRecord methodsFor: 'all changes' stamp: 'di 3/28/2000 10:59'!
hasNoChanges

	^ changeTypes isEmpty and: [methodChanges isEmpty]! !

!ClassChangeRecord methodsFor: 'all changes' stamp: 'di 3/24/2000 09:36'!
includesChangeType: changeType

	changeType == #new ifTrue: [^ changeTypes includes: #add].  "Backwd compat"
	^ changeTypes includes: changeType! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'di 3/29/2000 16:26'!
infoFromRemoval: selector

	^ (methodChanges at: selector ifAbsent: [^ nil])
		methodInfoFromRemoval

! !

!ClassChangeRecord methodsFor: 'initialization' stamp: 'di 4/5/2000 08:11'!
initFor: className revertable: isRevertable

	inForce := isRevertable.
	changeTypes := IdentitySet new.
	methodChanges := IdentityDictionary new.
	priorName := thisName := className.
	revertable := isRevertable and: [self realClass notNil].
	revertable ifTrue:
		[priorMD := self realClass methodDict copy.
		priorOrganization := self realClass organization deepCopy].
! !

!ClassChangeRecord methodsFor: 'isolation layers' stamp: 'eem 6/11/2008 16:51'!
invokePhase1

	| elements |
	revertable ifFalse: [^ self].
	inForce ifTrue: [self error: 'Can invoke only when not in force.'].

	"Do the first part of the invoke operation -- no particular hurry."
	"Save the outer method dictionary for quick revert of method changes."
	priorMD := self realClass methodDict.

	"Prepare a methodDictionary for switcheroo."
	thisMD := self realClass methodDict copy.
	methodChanges associationsDo:
		[:assn | | selector changeRecord type |
		selector := assn key.
		changeRecord := assn value.
		type := changeRecord changeType.
		type = #remove ifTrue: [thisMD removeKey: selector].
		type = #add ifTrue: [thisMD at: selector put: changeRecord currentMethod].
		type = #change ifTrue: [thisMD at: selector put: changeRecord currentMethod].
		].

	"Replace the original organization (and comment)."
	priorOrganization := self realClass organization.
	thisOrganization elementArray copy do:
		[:sel | (thisMD includesKey: sel) ifFalse: [thisOrganization removeElement: sel]].
	#(DoIt DoItIn:) do: [:sel | thisMD removeKey: sel ifAbsent: []].
	thisOrganization elementArray size = thisMD size ifFalse:
		[elements := thisOrganization elementArray asSet.
		thisMD keysDo:
			[:sel | (elements includes: sel) ifFalse:
				[thisOrganization classify: sel
					under: (priorOrganization categoryOfElement: sel)]]].
	self realClass organization: thisOrganization.


! !

!ClassChangeRecord methodsFor: 'isolation layers' stamp: 'di 3/29/2000 14:50'!
invokePhase2

	revertable ifFalse: [^ self].

	"Do the second part of the revert operation.  This must be very simple."
	"Replace original method dicts if there are method changes."
	self realClass methodDictionary: thisMD.  "zap.  Must flush Cache in outer loop."
	inForce := true.
! !

!ClassChangeRecord methodsFor: 'removal' stamp: 'di 4/1/2000 23:05'!
isClassRemoval
	"NOTE: there are other removals with changeType #addedThenRemoved,
	but this message is used to write out removals in fileOut, and those
	cases should not be written out."

	^ (changeTypes includes: #remove) or: [changeTypes includes: #removeClass]! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'di 4/1/2000 23:49'!
methodChanges

	^ methodChanges! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'eem 6/11/2008 16:53'!
methodChangeTypes
	"Return an old-style dictionary of method change types."

	| dict |
	dict := IdentityDictionary new.
	methodChanges associationsDo:
		[:assn | | selector record |
		selector := assn key.
		record := assn value.
		dict at: selector put: record changeType].
	^ dict! !

!ClassChangeRecord methodsFor: 'all changes' stamp: 'di 3/28/2000 15:14'!
noteChangeType: changeSymbol

	^ self noteChangeType: changeSymbol fromClass: nil! !

!ClassChangeRecord methodsFor: 'all changes' stamp: 'sw 4/3/2001 14:16'!
noteChangeType: changeSymbol fromClass: class

	(changeSymbol = #new or: [changeSymbol = #add]) ifTrue:
		[changeTypes add: #add.
		changeTypes remove: #change ifAbsent: [].
		revertable := false.
		^ self].
	changeSymbol = #change ifTrue:
		[(changeTypes includes: #add) ifTrue: [^ self].
		^ changeTypes add: changeSymbol].
	changeSymbol == #addedThenRemoved ifTrue:
		[^ self].  "An entire class was added but then removed"
	changeSymbol = #comment ifTrue:
		[^ changeTypes add: changeSymbol].
	changeSymbol = #reorganize ifTrue:
		[^ changeTypes add: changeSymbol].
	changeSymbol = #rename ifTrue:
		[^ changeTypes add: changeSymbol].
	(changeSymbol beginsWith: 'oldName: ') ifTrue:
		["Must only be used when assimilating other changeSets"
		(changeTypes includes: #add) ifTrue: [^ self].
		priorName := changeSymbol copyFrom: 'oldName: ' size + 1 to: changeSymbol size.
		^ changeTypes add: #rename].
	changeSymbol = #remove ifTrue:
		[(changeTypes includes: #add)
			ifTrue: [changeTypes add: #addedThenRemoved]
			ifFalse: [changeTypes add: #remove].
		^ changeTypes removeAllFoundIn: #(add change comment reorganize)].

	self error: 'Unrecognized changeType'! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'di 3/28/2000 23:28'!
noteNewMethod: newMethod selector: selector priorMethod: methodOrNil

	| methodChange |
	methodChange := self findOrMakeMethodChangeAt: selector priorMethod: methodOrNil.
	methodOrNil == nil
		ifTrue: [methodChange noteChangeType: #add]
		ifFalse: [methodChange noteChangeType: #change].
	methodChange noteNewMethod: newMethod.
! !

!ClassChangeRecord methodsFor: 'rename' stamp: 'di 5/8/2000 20:39'!
noteNewName: newName

	thisName := newName! !

!ClassChangeRecord methodsFor: 'definition' stamp: 'di 3/27/2000 22:08'!
notePriorDefinition: oldClass

	oldClass ifNil: [^ self].
	priorDefinition ifNil: [priorDefinition := oldClass definition]! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'di 3/23/2000 23:00'!
noteRemoveSelector: selector priorMethod: priorMethod lastMethodInfo: infoOrNil

	| methodChange |
	methodChange := self findOrMakeMethodChangeAt: selector priorMethod: priorMethod.
	methodChange changeType == #add
		ifTrue: [methodChange noteChangeType: #addedThenRemoved]
		ifFalse: [methodChange noteChangeType: #remove].

	infoOrNil ifNotNil:
		["Save the source code pointer and category so can still browse old versions"
		methodChange noteMethodInfoFromRemoval: infoOrNil]

! !

!ClassChangeRecord methodsFor: 'definition' stamp: 'di 3/28/2000 09:12'!
priorDefinition

	^ priorDefinition! !

!ClassChangeRecord methodsFor: 'rename' stamp: 'di 3/24/2000 09:38'!
priorName

	^ priorName! !

!ClassChangeRecord methodsFor: 'isolation layers' stamp: 'fbs 9/18/2013 08:58'!
realClass
	^ self realClass: Smalltalk globals.! !

!ClassChangeRecord methodsFor: 'isolation layers' stamp: 'fbs 9/18/2013 08:58'!
realClass: anEnvironment
	"Return the actual class (or meta), as determined from my name."
	thisName ifNil: [^ nil].
	(thisName endsWith: ' class')
		ifTrue: [^ (anEnvironment at: (thisName copyFrom: 1 to: thisName size - 6) asSymbol
						ifAbsent: [nil]) class]
		ifFalse: [^ anEnvironment at: thisName ifAbsent: [nil]]! !

!ClassChangeRecord methodsFor: 'method changes' stamp: 'sw 8/14/2002 11:11'!
removeSelector: selector
	"Remove all memory of changes associated with the argument, selector, in this class."

	selector == #Comment
		ifTrue:
			[changeTypes remove: #comment ifAbsent: []]
		ifFalse:
			[methodChanges removeKey: selector ifAbsent: []]! !

!ClassChangeRecord methodsFor: 'isolation layers' stamp: 'di 3/29/2000 14:50'!
revokePhase1

	revertable ifFalse: [^ self].
	inForce ifFalse: [self error: 'Can revoke only when in force.'].

	"Do the first part of the revoke operation.  This must be very simple."
	"Replace original method dict if there are method changes."
	self realClass methodDictionary: priorMD  "zap.  Must flush Cache in outer loop."! !

!ClassChangeRecord methodsFor: 'isolation layers' stamp: 'di 3/29/2000 14:50'!
revokePhase2

	revertable ifFalse: [^ self].

	"Replace the original organization (and comment)."
	thisOrganization := self realClass organization.
	self realClass organization: priorOrganization.
	inForce := false.
! !

!ClassChangeRecord methodsFor: 'rename' stamp: 'tk 6/8/2001 09:11'!
thisName

	^ thisName! !

!ClassChangeRecord methodsFor: 'all changes' stamp: 'eem 1/20/2016 13:39'!
trimHistory
	"Drop non-essential history."
	| realClass |
	realClass := self realClass.

	"Forget methods added and later removed"
	methodChanges keysAndValuesRemove:
		[:sel :chgRecord |
		chgRecord changeType == #addedThenRemoved
		"this removes methods that have been lost, e.g. by being deleted in a different change set, etc"
		or: [(chgRecord changeType == #add or: [chgRecord changeType == #change])
			and: [realClass isNil or: [(realClass includesSelector: sel) not]]]].

	"Forget renaming and reorganization of newly-added classes."
	(changeTypes includes: #add) ifTrue:
		[changeTypes removeAllFoundIn: #(rename reorganize)]! !

!ClassChangeRecord methodsFor: 'initialization' stamp: 'di 9/21/2000 12:34'!
zapHistory
	"Drop all recorded information not needed to simply keep track of what has been changed.
	Saves a lot of space."

	methodChanges do: [:r | r noteNewMethod: nil].  "Drop all refes to old methods"
	thisOrganization := nil.
	priorOrganization := nil.
	thisComment := nil.
	priorComment := nil.
	thisMD := nil.
	priorMD := nil.! !

Object subclass: #ClassReference
	instanceVariableNames: 'classSymbol stringVersion classIsMeta'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Tools'!

!ClassReference commentStamp: '<historical>' prior: 0!
A ClassReference is is a lightweight proxy for a Class's definition.  Allows class definitions to be viewed in MessageLists

Instance Variables
	classSymbol:		Symbol for method's class (without class keyword if meta)
	stringVersion:		the class's definition

!

!ClassReference class methodsFor: 'instance creation' stamp: 'eem 5/7/2008 11:23'!
class: aClass
	^ self new setStandardClass: aClass! !

!ClassReference methodsFor: 'accessing' stamp: 'eem 7/15/2010 17:00'!
actualClass 
	| actualClass |
	actualClass := Smalltalk at: classSymbol ifAbsent: [^nil].
	^classIsMeta ifTrue: [actualClass class] ifFalse: [actualClass]! !

!ClassReference methodsFor: 'accessing' stamp: 'eem 5/7/2008 11:41'!
asStringOrText

	^stringVersion! !

!ClassReference methodsFor: 'accessing' stamp: 'eem 5/7/2008 18:22'!
classIsMeta

	^classIsMeta! !

!ClassReference methodsFor: 'accessing' stamp: 'eem 5/7/2008 16:31'!
classSymbol
	^classSymbol! !

!ClassReference methodsFor: 'accessing' stamp: 'eem 5/8/2008 09:50'!
compiledMethod
	^nil! !

!ClassReference methodsFor: 'comparisons' stamp: 'eem 5/7/2008 11:27'!
isClassReference
	^true! !

!ClassReference methodsFor: 'comparisons' stamp: 'eem 5/7/2008 11:27'!
isMethodReference
	^false! !

!ClassReference methodsFor: 'setting' stamp: 'eem 5/7/2008 11:42'!
setClassAndSelectorIn: csBlock

	^csBlock value: self actualClass value: #Definition! !

!ClassReference methodsFor: 'initialize-release' stamp: 'ul 8/8/2013 12:36'!
setClassSymbol: classSym classIsMeta: isMeta stringVersion: aString 

	classSymbol := classSym.
	classIsMeta := isMeta.
	stringVersion := aString, ' (definition)'! !

!ClassReference methodsFor: 'initialize-release' stamp: 'ul 8/8/2013 12:37'!
setStandardClass: aClass

	self
		setClassSymbol:  aClass theNonMetaClass name
		classIsMeta: aClass isMeta
		stringVersion: aClass name! !

!ClassReference methodsFor: 'queries' stamp: 'eem 7/15/2010 17:34'!
sourceString
	^self actualClass definition! !

!ClassReference methodsFor: 'accessing' stamp: 'eem 5/7/2008 11:41'!
stringVersion

	^stringVersion! !

!ClassReference methodsFor: 'comparisons' stamp: 'eem 5/7/2008 11:33'!
<= anotherMethodOrClassReference

	classSymbol < anotherMethodOrClassReference classSymbol ifTrue: [^true].
	classSymbol > anotherMethodOrClassReference classSymbol ifTrue: [^false].
	classIsMeta = anotherMethodOrClassReference classIsMeta ifFalse: [^classIsMeta not].
	"i.e. if anotherMethodOrClassReference is a MethodReference then we're < it, and so <= to it"
	^true! !

!ClassReference methodsFor: 'comparisons' stamp: 'eem 5/7/2008 11:30'!
= anotherMethodReference 
	"Answer whether the receiver and the argument represent the 
	 same object."
	^ self species == anotherMethodReference species
	   and: [self classSymbol = anotherMethodReference classSymbol
	   and: [self classIsMeta = anotherMethodReference classIsMeta]]! !

Object subclass: #Clipboard
	instanceVariableNames: 'contents recent interpreter'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'System-Support'!

!Clipboard commentStamp: 'fbs 6/15/2013 16:20' prior: 0!
The Clipboard class implements a basic buffering scheme for text. The currently selected text is also exported to the OS so that text can be copied from and to other applications. Commonly only a single instance is used (the default clipboard) but applications are free to use other than the default clipboard if necessary.

If you'd like to be notified of the changes of the clipboard, you can subscribe to its #contentChanged event. For example:

subscriber := [ :newClipboardText :updateSource | 
	Transcript show: ('Clipboard was updated by {1}, the new content is: {2}'
		format: {
			updateSource.
			newClipboardText }); cr ].
Clipboard default
	when: #contentChanged
	send: #value:value:
	to: subscriber.

At this point if you open a Transcript and copy or cut some text, then the message will appear on the Transcript. Make sure that there's a reference kept to subscriber, otherwise it will be garbage collected and no messages will appear.
	
To unsubscribe from the clipboard changes, just evaluate:

Clipboard default removeActionsWithReceiver: subscriber!

!Clipboard class methodsFor: 'accessing' stamp: 'ar 1/15/2001 18:45'!
chooseRecentClipping  "Clipboard chooseRecentClipping"
	"Choose by menu from among the recent clippings"
	^self default chooseRecentClipping! !

!Clipboard class methodsFor: 'class initialization' stamp: 'yo 8/11/2003 22:43'!
clearInterpreters

	self allInstances do: [:each | each clearInterpreter].
! !

!Clipboard class methodsFor: 'accessing' stamp: 'ar 1/15/2001 18:35'!
clipboardText "Clipboard clipboardText"
	^self default clipboardText.! !

!Clipboard class methodsFor: 'accessing' stamp: 'ar 1/15/2001 18:35'!
clipboardText: aText 
	^self default clipboardText: aText! !

!Clipboard class methodsFor: 'accessing' stamp: 'ar 1/15/2001 18:33'!
default
	^Default ifNil:[Default := self new].! !

!Clipboard class methodsFor: 'accessing' stamp: 'ar 1/15/2001 18:48'!
default: aClipboard
	"So that clients can switch between different default clipboards"
	Default := aClipboard.! !

!Clipboard class methodsFor: 'class initialization' stamp: 'fbs 7/15/2013 23:07'!
localeChanged
	self startUp.! !

!Clipboard class methodsFor: 'class initialization' stamp: 'yo 12/29/2003 01:03'!
startUp

	self clearInterpreters.
! !

!Clipboard methodsFor: 'accessing' stamp: 'ar 8/6/2009 18:47'!
chooseRecentClipping  "Clipboard chooseRecentClipping"
	"Choose by menu from among the recent clippings"

	recent ifNil: [^ nil].
	^ UIManager default
		chooseFrom: (recent collect: [:txt | ((txt asString contractTo: 50)
									copyReplaceAll: Character cr asString with: '\')
									copyReplaceAll: Character tab asString with: '|'])
		values: recent! !

!Clipboard methodsFor: 'accessing' stamp: 'yo 8/11/2003 19:07'!
clearInterpreter

	interpreter := nil.
! !

!Clipboard methodsFor: 'accessing' stamp: 'edc 5/8/2007 17:04'!
clipboardText
	"Return the text currently in the clipboard. If the system clipboard is empty, or if it differs from the Smalltalk clipboard text, use the Smalltalk clipboard. This is done since (a) the Mac clipboard gives up on very large chunks of text and (b) since not all platforms support the notion of a clipboard."

	| string decodedString |
	string := self primitiveClipboardText withSqueakLineEndings.
	(string isEmpty
			or: [string = contents asString])
		ifTrue: [^ contents].
	decodedString := self interpreter fromSystemClipboard: string.
	^ decodedString = contents asString 
		ifTrue: [contents]
		ifFalse: [decodedString asText].
! !

!Clipboard methodsFor: 'accessing' stamp: 'ul 8/16/2012 15:24'!
clipboardText: text 

	self clipboardText: text notifyWith: nil! !

!Clipboard methodsFor: 'accessing' stamp: 'ul 8/16/2012 15:23'!
clipboardText: text notifyWith: anObject
	"Set the contents of the clipboard to text. anObject can be used as a marker for the source of the clipboard change, the default value is nil."

	| string |
	string := text asString.
	self noteRecentClipping: text asText.
	contents := text asText.
	string := self interpreter toSystemClipboard: string.
	self primitiveClipboardText: string.
	self triggerEvent: #contentChanged withArguments: { text. anObject }! !

!Clipboard methodsFor: 'initialize' stamp: 'ar 1/15/2001 18:34'!
initialize
	contents := '' asText.
	recent := OrderedCollection new.! !

!Clipboard methodsFor: 'accessing' stamp: 'yo 8/11/2003 18:23'!
interpreter

	interpreter ifNil: [self setInterpreter].
	^ interpreter.
! !

!Clipboard methodsFor: 'private' stamp: 'ar 1/15/2001 18:34'!
noteRecentClipping: text
	"Keep most recent clippings in a queue for pasteRecent (paste... command)"
	text isEmpty ifTrue: [^ self].
	text size > 50000 ifTrue: [^ self].
	(recent includes: text) ifTrue: [^ self].
	recent addFirst: text.
	[recent size > 5] whileTrue: [recent removeLast].
! !

!Clipboard methodsFor: 'primitives' stamp: 'ar 1/15/2001 18:28'!
primitiveClipboardText
	"Get the current clipboard text. Return the empty string if the primitive fails."
	<primitive: 141>
	^ ''! !

!Clipboard methodsFor: 'primitives' stamp: 'ar 1/15/2001 18:30'!
primitiveClipboardText: aString
	"Set the current clipboard text to the given string."

	<primitive: 141>
	"don't fail if the primitive is not implemented"! !

!Clipboard methodsFor: 'accessing' stamp: 'mir 7/20/2004 15:44'!
setInterpreter

	interpreter := LanguageEnvironment defaultClipboardInterpreter.
	interpreter ifNil: [
		"Should never be reached, but just in case."
		interpreter := NoConversionClipboardInterpreter new].
! !

Object subclass: #CodeLoader
	instanceVariableNames: 'baseURL sourceFiles segments publicKey'
	classVariableNames: 'DefaultBaseURL DefaultKey'
	poolDictionaries: ''
	category: 'System-Download'!

!CodeLoader commentStamp: 'hjh 4/22/2010 12:58' prior: 0!
CodeLoader provides a simple facility for loading code from the network.

CodeLoader has a very specific purpose, primarily for loading Etoys projects. It is not meant to be used for anything else. Use Installer for general purposes.

Examples:
	| loader |
	loader := CodeLoader new.
	loader baseURL:'http://isgwww.cs.uni-magdeburg.de/~raab/test/'.
	loader localCache: #('.cache' 'source').
	"Sources and segments can be loaded in parallel"
	loader loadSourceFiles: #('file1.st' 'file2.st.gz').
	loader localCache: #('.cache' 'segments').
	loader loadSegments: #('is1.extseg' 'is2.extseg.gz').
	"Install sources first - will wait until the files are actually loaded"
	loader installSourceFiles.
	"And then the segments"
	loader installSegments.!

CodeLoader subclass: #CachingCodeLoader
	instanceVariableNames: 'cacheDir'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Download'!

!CachingCodeLoader methodsFor: 'accessing' stamp: 'mir 12/22/1999 14:10'!
cacheDir
	^cacheDir! !

!CachingCodeLoader methodsFor: 'accessing' stamp: 'mir 12/22/1999 14:10'!
cacheDir: aString
	cacheDir := aString.! !

!CachingCodeLoader methodsFor: 'private' stamp: 'mir 12/22/1999 14:11'!
createRequestFor: name in: aLoader
	| request |
	request := super createRequestFor: name in: aLoader.
	request cachedName: cacheDir, name.
	^request! !

!CachingCodeLoader methodsFor: 'private' stamp: 'avi 4/30/2004 01:40'!
httpRequestClass
	^CachedHTTPDownloadRequest
! !

!CachingCodeLoader methodsFor: 'accessing' stamp: 'mir 12/22/1999 14:10'!
localCacheDir: aString
	self cacheDir:
		(FileDirectory default pathName,
		FileDirectory slash,
		aString,
		FileDirectory slash)! !

!CachingCodeLoader methodsFor: 'accessing' stamp: 'mir 12/22/1999 14:10'!
localCache: stringArray
	| fd |
	fd := FileDirectory default.
	stringArray do:[:part|
		(fd directoryNames includes: part) 
			ifFalse:[fd createDirectory: part].
		fd := fd directoryNamed: part].
	self cacheDir: (fd pathName copyWith: fd pathNameDelimiter).! !

!CodeLoader class methodsFor: 'utilities' stamp: 'mir 9/6/2000 15:03'!
compressFileNamed: aFileName
	self compressFileNamed: aFileName in: FileDirectory default! !

!CodeLoader class methodsFor: 'utilities' stamp: 'ul 6/17/2011 12:25'!
compressFileNamed: aFileName in: aDirectory
	"Compress the currently selected file"
	| zipped buffer unzipped zipFileName |
	unzipped := aDirectory readOnlyFileNamed: (aDirectory fullNameFor: aFileName).
	unzipped binary.
	zipFileName := aFileName copyUpToLast: $. .
	zipped := aDirectory newFileNamed: (zipFileName, FileDirectory dot, ImageSegment compressedFileExtension).
	zipped binary.
	zipped := GZipWriteStream on: zipped.
	buffer := ByteArray new: 50000.
	'Compressing ', zipFileName
		displayProgressFrom: 0 to: unzipped size
		during:[:bar|
			[unzipped atEnd] whileFalse:[
				bar value: unzipped position.
				zipped nextPutAll: (unzipped nextInto: buffer)].
			zipped close.
			unzipped close].
! !

!CodeLoader class methodsFor: 'accessing' stamp: 'mir 1/11/2000 13:45'!
defaultBaseURL
	^DefaultBaseURL ifNil: ['']! !

!CodeLoader class methodsFor: 'accessing' stamp: 'mir 1/11/2000 13:45'!
defaultBaseURL: aURLString
	DefaultBaseURL := aURLString! !

!CodeLoader class methodsFor: 'accessing' stamp: 'ar 12/22/1999 15:08'!
defaultKey
	"Return the default key used for verifying signatures of loaded code"
	^DefaultKey! !

!CodeLoader class methodsFor: 'accessing' stamp: 'mir 1/10/2000 18:16'!
defaultKey: aPublicKey
	"Store the default key used for verifying signatures of loaded code"
	DefaultKey := aPublicKey
	"CodeLoader defaultKey: DOLPublicKey"
	"CodeLoader defaultKey: (DigitalSignatureAlgorithm testKeySet at: 2)"! !

!CodeLoader class methodsFor: 'utilities' stamp: 'nice 12/26/2009 09:50'!
exportCategories: catList to: aFileName
	"CodeLoader exportCategories: #( 'Game-Animation' 'Game-Framework' ) to: 'Game-Framework'"

	| classList |
	classList := OrderedCollection new.
	catList do: [:catName |
		| list |
		list := SystemOrganization listAtCategoryNamed: catName asSymbol.
		list do: [:nm | classList add: (Smalltalk at: nm); add: (Smalltalk at: nm) class]].
	self exportCodeSegment: aFileName classes: classList keepSource: true! !

!CodeLoader class methodsFor: 'utilities' stamp: 'mir 1/18/2000 20:53'!
exportCategoryNamed: catName
	"CodeLoader exportCategoryNamed: 'OceanicPanic' "

	| list |
	list := SystemOrganization listAtCategoryNamed: catName asSymbol.
	self exportClassesNamed: list to: catName! !

!CodeLoader class methodsFor: 'utilities' stamp: 'mir 1/18/2000 20:53'!
exportClassesNamed: classNameList to: aFileName

	| classList |
	classList := OrderedCollection new.
	classNameList do: [:nm | classList add: (Smalltalk at: nm); add: (Smalltalk at: nm) class].
	self exportCodeSegment: aFileName classes: classList keepSource: true! !

!CodeLoader class methodsFor: 'utilities' stamp: 'nice 2/23/2011 21:40'!
exportCodeSegment: exportName classes: aClassList keepSource: keepSources

	"Code for writing out a specific category of classes as an external image segment.  Perhaps this should be a method."

	| is oldMethods newMethods classList symbolHolder fileName |
	keepSources
		ifTrue: [
			self confirm: 'We are going to abandon sources.
Quit without saving after this has run.' orCancel: [^self]].

	classList := aClassList asArray.

	"Strong pointers to symbols"
	symbolHolder := Symbol allSymbols.

	oldMethods := OrderedCollection new: classList size * 150.
	newMethods := OrderedCollection new: classList size * 150.
	keepSources
		ifTrue: [
			classList do: [:cl |
				cl selectorsAndMethodsDo:
					[:selector :m |
					| oldCodeString methodNode |
					m fileIndex > 0 ifTrue:
						[oldCodeString := cl sourceCodeAt: selector.
						methodNode := cl newCompiler
											parse: oldCodeString in: cl notifying: nil.
						oldMethods addLast: m.
						newMethods addLast: (m copyWithTempsFromMethodNode: methodNode)]]]].
	oldMethods asArray elementsExchangeIdentityWith: newMethods asArray.
	oldMethods := newMethods := nil.

	Smalltalk garbageCollect.
	is := ImageSegment new copyFromRootsForExport: classList.	"Classes and MetaClasses"

	fileName := FileDirectory fileName: exportName extension: ImageSegment fileExtension.
	is writeForExport: fileName.
	self compressFileNamed: fileName

! !

!CodeLoader class methodsFor: 'utilities' stamp: 'mir 2/2/2001 14:56'!
loadCodeSegment: segmentName
	| loader |
	loader := self new.
	loader loadSegments: (Array with: segmentName). 
	loader installSegments.! !

!CodeLoader class methodsFor: 'utilities' stamp: 'mir 10/12/2000 17:39'!
loadCode: codeSegmentName from: baseURL ifClassNotLoaded: testClass

	CodeLoader defaultBaseURL: baseURL.
	(Smalltalk includesKey: testClass)
		ifFalse: [CodeLoader loadCodeSegment: codeSegmentName].
! !

!CodeLoader class methodsFor: 'utilities' stamp: 'ul 6/17/2011 12:25'!
signFilesFrom: sourceNames to: destNames key: privateKey
	"Sign all the given files using the private key.
	This will add an 's' to the extension of the file."
	"| fd oldNames newNames |
	fd := FileDirectory default directoryNamed:'unsigned'.
	oldNames := fd fileNames.
	newNames := oldNames collect:[:name| 'signed', FileDirectory slash, name].
	oldNames := oldNames collect:[:name| 'unsigned', FileDirectory slash, name].
	CodeLoader
		signFilesFrom: oldNames
		to: newNames
		key: DOLPrivateKey."
	| dsa |
	dsa := DigitalSignatureAlgorithm new.
	dsa initRandomNonInteractively.
	'Signing files...' 
		displayProgressFrom: 1 to: sourceNames size during:[:bar|
			1 to: sourceNames size do:[:i|
				bar value: i.
				self signFile: (sourceNames at: i) renameAs: (destNames at: i) key: privateKey dsa: dsa]].
! !

!CodeLoader class methodsFor: 'utilities' stamp: 'mir 2/14/2000 16:47'!
signFiles: fileNames in: dirName key: privateKey
	"Sign the files in the current directory and put them into a folder signed."

	|  newNames oldNames |
	oldNames := fileNames collect:[:fileName | dirName , FileDirectory slash, fileName].
	newNames := fileNames collect:[:fileName | dirName , FileDirectory slash, 'signed', FileDirectory slash, fileName].
	CodeLoader
		signFilesFrom: oldNames
		to: newNames
		key: privateKey! !

!CodeLoader class methodsFor: 'utilities' stamp: 'mir 1/18/2000 18:49'!
signFiles: fileNames key: privateKey
	"Sign the files in the current directory and put them into a folder signed."

	|  newNames |
	newNames := fileNames collect:[:fileName | 'signed', FileDirectory slash, fileName].
	CodeLoader
		signFilesFrom: fileNames
		to: newNames
		key: privateKey! !

!CodeLoader class methodsFor: 'utilities' stamp: 'asm 12/6/2002 08:11'!
signFile: fileName renameAs: destFile key: privateKey dsa: dsa
	"Sign the given file using the private key."
	| in out |
	in := FileStream readOnlyFileNamed: fileName.	in binary.
	out := FileStream newFileNamed: destFile.			out binary.
	[in atEnd] whileFalse:[out nextPutAll: (in next: 4096)].
	in close.	out close.
	FileDirectory activeDirectoryClass splitName: destFile to:[:path :file|
		SecurityManager default signFile: file directory: (FileDirectory on: path).
	].
! !

!CodeLoader class methodsFor: 'utilities' stamp: 'ar 2/6/2001 19:17'!
verifySignedFileNamed: aFileName
	"CodeLoader verifySignedFileNamed: 'signed\dummy1.dsq' "

	| secured signedFileStream |
	signedFileStream := FileStream fileNamed: aFileName.
	secured := SecurityManager default positionToSecureContentsOf: signedFileStream.
	signedFileStream close.
	Transcript show: aFileName , ' verified: '; show: secured printString; cr.

! !

!CodeLoader methodsFor: 'accessing' stamp: 'ar 12/13/1999 18:19'!
baseURL
	^baseURL! !

!CodeLoader methodsFor: 'accessing' stamp: 'ar 12/13/1999 18:19'!
baseURL: aString
	baseURL := aString.! !

!CodeLoader methodsFor: 'private' stamp: 'mir 2/2/2001 14:44'!
createRequestFor: name in: aLoader
	"Create a URL request for the given string, which can be cached locally."
	| request |
	request := HTTPLoader httpRequestClass for: self baseURL , name in: aLoader.
	aLoader addRequest: request. "fetch from URL"
	^request! !

!CodeLoader methodsFor: 'private' stamp: 'avi 4/30/2004 01:40'!
httpRequestClass
	^HTTPDownloadRequest! !

!CodeLoader methodsFor: 'initialize-release' stamp: 'mir 1/11/2000 13:47'!
initialize
	publicKey := DefaultKey.
	baseURL := self class defaultBaseURL! !

!CodeLoader methodsFor: 'installing' stamp: 'tak 9/4/2006 14:50'!
installProject
	"Assume that we're loading a single file and it's a project"
	| aStream |
	aStream _ sourceFiles first contentStream.
	aStream ifNil:[^self error:'Project was not loaded'].
	ProjectLoading openOn: aStream! !

!CodeLoader methodsFor: 'installing' stamp: 'mir 1/20/2000 13:37'!
installSegments
	"Install the previously loaded segments"
	segments == nil ifTrue:[^self].
	segments do:[:req| self installSegment: req].
	segments := nil.! !

!CodeLoader methodsFor: 'installing' stamp: 'nice 7/18/2014 01:23'!
installSegment: reqEntry
	"Install the previously loaded segment"
	| contentStream contents trusted |
	contentStream := reqEntry value contentStream.
	contentStream ifNil:[^self error:'No content to install: ', reqEntry key printString].
	trusted := SecurityManager default positionToSecureContentsOf: contentStream.
	trusted ifFalse:[(SecurityManager default enterRestrictedMode) ifFalse:[
		contentStream close.
		^self error:'Insecure content encountered: ', reqEntry key printString]].
	contents := contentStream upToEnd unzipped asString.
	(contentStream respondsTo: #close) ifTrue:[contentStream close].
	^(RWBinaryOrTextStream with: contents) reset fileInObjectAndCode install.! !

!CodeLoader methodsFor: 'installing' stamp: 'ar 12/22/1999 15:02'!
installSourceFiles
	"Install the previously loaded source files"
	sourceFiles == nil ifTrue:[^self].
	sourceFiles do:[:req| self installSourceFile: req contentStream].
	sourceFiles := nil.! !

!CodeLoader methodsFor: 'installing' stamp: 'nice 7/18/2014 01:22'!
installSourceFile: aStream
	"Install the previously loaded source file"
	| contents trusted |
	aStream ifNil:[^self error:'No content to install'].
	trusted := SecurityManager default positionToSecureContentsOf: aStream.
	trusted ifFalse:[(SecurityManager default enterRestrictedMode) 
					ifFalse:[ aStream close.
							^ self error:'Insecure content encountered']].
	contents := aStream upToEnd unzipped asString.
	(aStream respondsTo: #close) ifTrue:[aStream close].
	^contents readStream fileIn! !

!CodeLoader methodsFor: 'loading' stamp: 'nice 12/27/2009 02:54'!
loadSegments: anArray
	"Load all the source files in the given array."
	| loader |
	loader := HTTPLoader default.
	segments := anArray collect:[:name |
		| reqName request |
		reqName := (FileDirectory extensionFor: name) isEmpty
			ifTrue: [FileDirectory fileName: name extension: ImageSegment compressedFileExtension]
			ifFalse: [name].
		request := self createRequestFor: reqName in: loader.
		name->request].
! !

!CodeLoader methodsFor: 'loading' stamp: 'nice 12/26/2009 09:50'!
loadSourceFiles: anArray
	"Load all the source files in the given array."
	| loader |
	loader := HTTPLoader default.
	sourceFiles := anArray collect: [:name |
		self createRequestFor: name in: loader]! !

!CodeLoader methodsFor: 'accessing' stamp: 'ar 12/22/1999 15:07'!
publicKey
	^publicKey! !

!CodeLoader methodsFor: 'accessing' stamp: 'ar 12/22/1999 15:07'!
publicKey: aPublicKey
	publicKey := aPublicKey! !

Object subclass: #ColorTheme
	instanceVariableNames: ''
	classVariableNames: 'Current'
	poolDictionaries: ''
	category: 'System-Support'!

!ColorTheme class methodsFor: 'applying' stamp: 'dgd 3/31/2006 12:25'!
apply
	^self new apply! !

!ColorTheme class methodsFor: 'applying' stamp: 'dgd 3/12/2006 13:08'!
applyTheme: aThemeClass 
	aThemeClass new apply! !

!ColorTheme class methodsFor: 'accessing' stamp: 'dgd 3/31/2006 12:07'!
current
	^ Current
		ifNil: [self defaultTheme apply]! !

!ColorTheme class methodsFor: 'accessing' stamp: 'dgd 11/2/2004 17:52'!
current: aColorTheme
	Current := aColorTheme! !

!ColorTheme class methodsFor: 'accessing' stamp: 'fbs 7/12/2013 09:21'!
defaultTheme
	^ self new.! !

!ColorTheme methodsFor: 'applying' stamp: 'mt 1/13/2016 08:18'!
apply
	"apply the receiver as the current theme"
	BalloonMorph balloonColor: self balloonColor.

	Preferences setParameter: #defaultWorldColor to: self defaultWorldColor.

	Preferences insertionPointColor: self insertionPointColor.
	Preferences keyboardFocusColor: self keyboardFocusColor.
	Preferences textHighlightColor: self textHighlightColor.

	Preferences setParameter: #menuTitleColor to: self menuTitleColor.
	Preferences setParameter: #menuTitleBorderColor to: self menuTitleBorderColor.
	Preferences setParameter: #menuTitleBorderWidth to: self menuTitleBorderWidth.
	Preferences setParameter: #menuColor to: self menuColor.
	Preferences setParameter: #menuBorderColor to: self menuBorderColor.
	Preferences setParameter: #menuLineColor to: self menuLineColor.
	Preferences setParameter: #menuBorderWidth to: self menuBorderWidth.
	Preferences setParameter: #menuSelectionColor to: self menuSelectionColor.

	SystemProgressMorph reset.

	self class current: self.
! !

!ColorTheme methodsFor: 'theme' stamp: 'fbs 7/6/2013 15:33'!
balloonColor
	^ TranslucentColor
			r: 0.92
			g: 0.92
			b: 0.706
			alpha: 0.75! !

!ColorTheme methodsFor: 'theme' stamp: 'dgd 11/2/2004 21:23'!
cancelColor
	^ Color lightRed! !

!ColorTheme methodsFor: 'theme' stamp: 'fbs 7/6/2013 15:34'!
defaultWorldColor
	^ Color blue muchLighter! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/3/2004 19:27'!
dialog3DTitles
	^ true! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/2/2004 19:56'!
dialogBorderColor
	^ Color fromArray: #(0.355 0.516 1.0 )! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/2/2004 19:54'!
dialogBorderWidth
	^ 4! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/2/2004 19:59'!
dialogButtonBorderWidth
	^ 0! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/2/2004 21:09'!
dialogColor
	^ Color paleYellow! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/3/2004 19:49'!
dialogPaneBorderColor
	^ Color black
! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/3/2004 19:44'!
dialogPaneBorderWidth
	^ 0! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/3/2004 19:30'!
dialogPaneRampOrColor
	^ {0.0 -> (Color r: 0.742 g: 0.871 b: 1.0).
		1.0 -> (Color r: 0.516 g: 0.645 b: 1.0)}! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/3/2004 19:38'!
dialogRampOrColor
	^ {0.0 -> (Color r: 0.516 g: 0.645 b: 1.0).
		1.0 -> (Color r: 0.742 g: 0.871 b: 1.0)}! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/2/2004 21:18'!
dialogTextBoxBorderColor
	^ Color black! !

!ColorTheme methodsFor: 'theme - dialogs' stamp: 'dgd 11/2/2004 21:18'!
dialogTextBoxColor
	^ Color white! !

!ColorTheme methodsFor: 'theme' stamp: 'MAD 3/14/2010 14:58'!
disabledColor
	^ Color lightGray! !

!ColorTheme methodsFor: 'theme - dockingbar' stamp: 'fbs 7/6/2013 15:34'!
dockingBarAutoGradient
	^ true! !

!ColorTheme methodsFor: 'theme - dockingbar' stamp: 'fbs 7/6/2013 15:34'!
dockingBarColor
	^ Color r: 0.6 g: 0.7 b: 1! !

!ColorTheme methodsFor: 'theme - dockingbar' stamp: 'fbs 7/6/2013 15:35'!
dockingBarGradientRamp
	^ { 0.0 -> Color white.
		1.0 -> (Color r: 0.6 g: 0.7 b: 1) }! !

!ColorTheme methodsFor: 'theme' stamp: 'fbs 7/6/2013 15:35'!
helpColor
	^ Color lightGreen! !

!ColorTheme methodsFor: 'theme' stamp: 'fbs 7/6/2013 15:35'!
insertionPointColor
	^ Color red! !

!ColorTheme methodsFor: 'theme' stamp: 'fbs 7/6/2013 15:35'!
keyboardFocusColor
	^ Color r: 0.6 g: 1 b: 1! !

!ColorTheme methodsFor: 'theme - menus' stamp: 'fbs 7/6/2013 15:36'!
menuBorderColor
	^ Color r: 0.2 g: 0.3 b: 0.9! !

!ColorTheme methodsFor: 'theme - menus' stamp: 'fbs 7/6/2013 15:36'!
menuBorderWidth
	^ 2! !

!ColorTheme methodsFor: 'theme - menus' stamp: 'fbs 7/6/2013 15:36'!
menuColor
	^ Color r: 0.85 g: 0.9 b: 1! !

!ColorTheme methodsFor: 'theme - menus' stamp: 'fbs 7/6/2013 15:36'!
menuLineColor
	^ Color r: 0.6 g: 0.7 b: 1! !

!ColorTheme methodsFor: 'theme - menus' stamp: 'fbs 7/6/2013 15:37'!
menuSelectionColor
	^ Color r: 0.2 g: 0.3 b: 0.9! !

!ColorTheme methodsFor: 'theme - menus' stamp: 'fbs 7/6/2013 15:37'!
menuTitleBorderColor
	^ Color r: 0.6 g: 0.7 b: 1! !

!ColorTheme methodsFor: 'theme - menus' stamp: 'fbs 7/6/2013 15:37'!
menuTitleBorderWidth
	^ 6! !

!ColorTheme methodsFor: 'theme - menus' stamp: 'fbs 7/6/2013 15:37'!
menuTitleColor
	^ Color r: 0.6 g: 0.7 b: 1! !

!ColorTheme methodsFor: 'theme' stamp: 'dgd 11/2/2004 21:22'!
okColor
	^ Color lightGreen! !

!ColorTheme methodsFor: 'theme' stamp: 'fbs 7/6/2013 15:37'!
textHighlightColor
	^ Color blue muchLighter alpha: 0.7! !

Object subclass: #DeepCopier
	instanceVariableNames: 'references uniClasses newUniClasses'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Object Storage'!

!DeepCopier commentStamp: 'tk 3/4/2003 19:39' prior: 0!
DeepCopier does a veryDeepCopy.  

It is a complete tree copy using a dictionary.  Any object that is in the tree twice is only copied once.  All references to the object in the copy of the tree will point to the new copy.  See Object|veryDeepCopy which calls (self veryDeepCopyWith: aDeepCopier).

The dictionary of objects that have been seen, holds the correspondance (uniClass -> new uniClass).

When a tree of morphs points at a morph outside of itself, that morph should not be copied.  Use our own kind of weak pointers for the 'potentially outside' morphs.   Default is that any new class will have all of its fields deeply copied.  If a field needs to be weakly copied, define veryDeepInner: and veryDeepFixupWith:.
     veryDeepInner: has the loop that actually copies the fields.  If a class defines its own copy of veryDeepInner: (to leave some fields out), then veryDeepFixupWith: will be called on that object at the end.  veryDeepInner: can compute an alternate object to put in a field.  (Object veryDeepCopyWith: discovers which superclasses did not define veryDeepInner:, and very deeply copies the variables defined in those classes).
	To decide if a class needs veryDeepInner: and veryDeepFixupWith:, ask this about an instance:  If I duplicate this object, does that mean that I also want to make duplicates of the things it holds onto?  If yes, (i.e. a Paragraph does want a new copy of its Text) then do nothing.  If no, (i.e. an undo command does not want to copy the objects it acts upon), then define veryDeepInner: and veryDeepFixupWith:.
	
Here is an analysis for the specific case of a morph being held by another morph.  
Does field X contain a morph (or a Player whose costume is a morph)?  If not, no action needed.
Is the morph in field X already a submorph of the object?  Is it down lower in the submorph tree?
	If so, no action needed.
Could the morph in field X every appear on the screen (be a submorph of some other morph)?
	If not, no action needed.
	If it could, you must write the methods veryDeepFixupWith:   and   veryDeepInner:, and in them, refrain from sending veryDeepCopyWith: to the contents of field X.

newUniClasses =   true in the normal case.  Every duplicated Player gets a new class.  When false, all duplicates will be siblings (sister instances) of existing players.


----- Things Ted is still considering -----
Rule:  If a morph stores a uniClass class (Player 57) as an object in a field, the new uniClass will not be stored there.   Each uniClass instance does have a new class created for it.  (fix this by putting the old class in references and allow lookup?  Wrong if encounter it before seeing an instance?)

Rule: If object A has object C in a field, and A says (^ C) for the copy, but object B has A in a normal field and it gets deepCopied, and A in encountered first, then there will be two copies of C.  (just be aware of it)

Dependents are now fixed up.  Suppose a model has a dependent view.  In the DependentFields dictionary, model -> (view ...).  
	If only the model is copied, no dependents are created (no one knows about the new model).  
	If only the view is copied, it is inserted into DependentFields on the right side.  model -> (view  copiedView ...).  
	If both are copied, the new model has the new view as its dependent.
	If additional things depend on a model that is copied, the caller must add them to its dependents.
!

!DeepCopier methodsFor: 'like fullCopy' stamp: 'mt 7/28/2016 10:54'!
checkBasicClasses
	"Check that no indexes of instance vars have changed in certain classes.  If you get an error in this method, an implementation of veryDeepCopyWith: needs to be updated.  The idea is to catch a change while it is still in the system of the programmer who made it.  
	DeepCopier new checkVariables	"

	| str str2 objCls morphCls playerCls |
	str := '|veryDeepCopyWith: or veryDeepInner: is out of date.'.
	(objCls := self objInMemory: #Object) ifNotNil: [
		objCls instSize = 0 ifFalse: [self error: 
			'Many implementers of veryDeepCopyWith: are out of date']].
	(morphCls := self objInMemory: #Morph) ifNotNil: [
		morphCls superclass == Object ifFalse: [self error: 'Morph', str].
		(morphCls instVarNames copyFrom: 1 to: 6) = #('bounds' 'owner' 'submorphs' 
				'fullBounds' 'color' 'extension') 
			ifFalse: [self error: 'Morph', str]].	"added ones are OK"

	str2 := 'Player|copyUniClassWith: and DeepCopier|mapUniClasses are out of date'.
	(playerCls := self objInMemory: #Player) ifNotNil: [
		playerCls class instVarNames = #('scripts' 'slotInfo' 'variableDocks')
			ifFalse: [self error: str2]].
! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'tk 3/7/2001 15:42'!
checkClass: aClass
	| meth |
	"Check that no indexes of instance vars have changed in certain classes.  If you get an error in this method, an implementation of veryDeepCopyWith: needs to be updated.  The idea is to catch a change while it is still in the system of the programmer who made it."

	self checkBasicClasses.	"Unlikely, but important to catch when it does happen."

	"Every class that implements veryDeepInner: must copy all its inst vars.  Danger is that a user will add a new instance variable and forget to copy it.  So check that the last one is mentioned in the copy method."
	(aClass includesSelector: #veryDeepInner:) ifTrue: [ 
		((aClass compiledMethodAt: #veryDeepInner:) writesField: aClass instSize) ifFalse: [
			aClass instSize > 0 ifTrue: [
				self warnIverNotCopiedIn: aClass sel: #veryDeepInner:]]].
	(aClass includesSelector: #veryDeepCopyWith:) ifTrue: [
		meth := aClass compiledMethodAt: #veryDeepCopyWith:.
		(meth size > 20) & (meth literals includes: #veryDeepCopyWith:) not ifTrue: [
			(meth writesField: aClass instSize) ifFalse: [
				self warnIverNotCopiedIn: aClass sel: #veryDeepCopyWith:]]].
! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'nice 12/26/2009 01:30'!
checkDeep
	"Write exceptions in the Transcript.  Every class that implements veryDeepInner: must copy all its inst vars.  Danger is that a user will add a new instance variable and forget to copy it.  This check is only run by hand once in a while to make sure nothing was forgotten.  
(Please do not remove this method.)
	DeepCopier new checkDeep 	"

	Transcript
		cr;
		show: 'Instance variables shared with the original object when it is copied'.
	(self systemNavigation allClassesImplementing: #veryDeepInner:) do: 
			[:aClass | 
			| mm |
			(mm := aClass instVarNames size) > 0 
				ifTrue: 
					[aClass instSize - mm + 1 to: aClass instSize
						do: 
							[:index | 
							((aClass compiledMethodAt: #veryDeepInner:) writesField: index) 
								ifFalse: 
									[Transcript
										cr;
										show: aClass name;
										space;
										show: (aClass allInstVarNames at: index)]]]]! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'nice 12/26/2009 01:30'!
checkVariables
	"Check that no indexes of instance vars have changed in certain classes.  If you get an error in this method, an implementation of veryDeepCopyWith: needs to be updated.  The idea is to catch a change while it is still in the system of the programmer who made it.  
	DeepCopier new checkVariables	"

	self checkBasicClasses.

	"Every class that implements veryDeepInner: must copy all its inst vars.  Danger is that a user will add a new instance variable and forget to copy it.  So check that the last one is mentioned in the copy method."
	(self systemNavigation allClassesImplementing: #veryDeepInner:) do: 
			[:aClass | 
			((aClass compiledMethodAt: #veryDeepInner:) writesField: aClass instSize) 
				ifFalse: 
					[aClass instSize > 0 
						ifTrue: [self warnIverNotCopiedIn: aClass sel: #veryDeepInner:]]].
	(self systemNavigation allClassesImplementing: #veryDeepCopyWith:) do: 
			[:aClass | 
			| meth |
			meth := aClass compiledMethodAt: #veryDeepCopyWith:.
			meth size > 20 & (meth literals includes: #veryDeepCopyWith:) not 
				ifTrue: 
					[(meth writesField: aClass instSize) 
						ifFalse: [self warnIverNotCopiedIn: aClass sel: #veryDeepCopyWith:]]]! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'nice 12/26/2009 01:31'!
fixDependents
	"They are not used much, but need to be right"

	DependentsFields associationsDo: [:pair |
		pair value do: [:dep | 
			| newDep newModel |
			newDep := references at: dep ifAbsent: [nil].
			newDep ifNotNil: [
				newModel := references at: pair key ifAbsent: [pair key].
				newModel addDependent: newDep]]].
! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'hg 11/23/1999 13:36'!
initialize

	self initialize: 4096.
! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'tk 3/4/2003 19:40'!
initialize: size

	references := IdentityDictionary new: size.
	uniClasses := IdentityDictionary new.	"UniClass -> new UniClass"
	"self isItTimeToCheckVariables ifTrue: [self checkVariables]."
		"no more checking at runtime"
	newUniClasses := true.! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'nice 1/15/2010 22:57'!
mapUniClasses
	"For new Uniclasses, map their class vars to the new objects.  And their additional class instance vars.  (scripts slotInfo) and cross references like (player321)."
	"Players also refer to each other using associations in the References dictionary.  Search the methods of our Players for those.  Make new entries in References and point to them."
| pp newKey |

	newUniClasses ifFalse: [^ self].	"All will be siblings.  uniClasses is empty"
"Uniclasses use class vars to hold onto siblings who are referred to in code"
pp := (Smalltalk at: #Player ifAbsent:[^self]) class superclass instSize.
uniClasses do: [:playersClass | "values = new ones"
	playersClass classPool associationsDo: [:assoc |
		assoc value: (assoc value veryDeepCopyWith: self)].
	playersClass scripts: (playersClass privateScripts veryDeepCopyWith: self).	"pp+1"
	"(pp+2) slotInfo was deepCopied in copyUniClass and that's all it needs"
	pp+3 to: playersClass class instSize do: [:ii | 
		playersClass instVarAt: ii put: 
			((playersClass instVarAt: ii) veryDeepCopyWith: self)].
	].

"Make new entries in References and point to them."
References keys "copy" do: [:playerName | | oldPlayer |
	oldPlayer := References at: playerName.
	(references includesKey: oldPlayer) ifTrue: [
		newKey := (references at: oldPlayer) "new player" uniqueNameForReference.
		"now installed in References"
		(references at: oldPlayer) renameTo: newKey]].
uniClasses "values" do: [:newClass | | newSelList oldSelList |
	oldSelList := OrderedCollection new.   newSelList := OrderedCollection new.
	newClass selectorsAndMethodsDo: [:sel :m | 
		m literals do: [:assoc | | newAssoc |
			assoc isVariableBinding ifTrue: [
				(References associationAt: assoc key ifAbsent: [nil]) == assoc ifTrue: [
					newKey := (references at: assoc value ifAbsent: [assoc value]) 
									externalName asSymbol.
					(assoc key ~= newKey) & (References includesKey: newKey) ifTrue: [
						newAssoc := References associationAt: newKey.
						newClass methodDictionary at: sel put: 
							(newClass compiledMethodAt: sel) clone.	"were sharing it"
						(newClass compiledMethodAt: sel)
							literalAt: ((newClass compiledMethodAt: sel) literals indexOf: assoc)
							put: newAssoc.
						(oldSelList includes: assoc key) ifFalse: [
							oldSelList add: assoc key.  newSelList add: newKey]]]]]].
	oldSelList with: newSelList do: [:old :new |
			newClass replaceSilently: old to: new]].	"This is text replacement and can be wrong"! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'tk 3/11/2003 14:13'!
newUniClasses
	"If false, all new Players are merely siblings of the old players"

	^ newUniClasses! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'tk 3/4/2003 19:44'!
newUniClasses: newVal
	"If false, all new players are merely siblings of the old players"

	newUniClasses := newVal! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'tk 3/7/2001 15:29'!
objInMemory: ClassSymbol
	| cls |
	"Test if this global is in memory and return it if so."

	cls := Smalltalk at: ClassSymbol ifAbsent: [^ nil].
	^ cls isInMemory ifTrue: [cls] ifFalse: [nil].! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'tk 8/20/1998 22:13'!
references
	^ references! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'tk 8/19/1998 15:48'!
uniClasses
	^uniClasses! !

!DeepCopier methodsFor: 'like fullCopy' stamp: 'ar 9/27/2005 20:27'!
warnIverNotCopiedIn: aClass sel: sel
	"Warn the user to update veryDeepCopyWith: or veryDeepInner:"

	self inform: ('An instance variable was added to to class ', aClass name, ',\and it is not copied in the method ', sel, '.\Please rewrite it to handle all instance variables.\See DeepCopier class comment.') withCRs.
	ToolSet browse: aClass selector: sel! !

Object subclass: #DefaultExternalDropHandler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!DefaultExternalDropHandler commentStamp: 'dgd 4/5/2004 19:07' prior: 0!
An alternative default handler that uses the file-list services to process files.
!

!DefaultExternalDropHandler class methodsFor: 'class initialization' stamp: 'dgd 4/5/2004 19:10'!
initialize
	"initialize the receiver"
	ExternalDropHandler defaultHandler: self new! !

!DefaultExternalDropHandler class methodsFor: 'class initialization' stamp: 'dgd 4/5/2004 19:09'!
unload
	"initialize the receiver"
	ExternalDropHandler defaultHandler: nil! !

!DefaultExternalDropHandler methodsFor: 'private' stamp: 'mt 7/22/2016 11:45'!
chooseServiceFrom: aCollection
	"private - choose a service from aCollection asking the user if  needed"
	^aCollection size = 1 
		ifTrue: [aCollection anyOne]
		ifFalse:[UIManager default 
			chooseFrom: (aCollection collect:[:each| each label])
			values: aCollection
			title: 'You dropped a file. Please choose a service:' translated withCRs].! !

!DefaultExternalDropHandler methodsFor: 'event handling' stamp: 'spfa 9/22/2010 09:44'!
handle: dropStream in: pasteUp dropEvent: anEvent 
	"the file was just droped, let's do our job"
	| fileName services chosen |
	fileName := dropStream name.
	services := self servicesForFileNamed: fileName.

	"no service, default behavior"
	services isEmpty 
		ifTrue: [^UIManager default edit: dropStream contentsOfEntireFile].

	chosen := self chooseServiceFrom: services.
	chosen ifNotNil:[chosen performServiceFor: dropStream]! !

!DefaultExternalDropHandler methodsFor: 'private' stamp: 'fbs 8/9/2013 17:30'!
servicesForFileNamed: aString 
	"private - answer a collection of file-services for the file named  
	aString"
	| allServices |
	allServices := FileServices itemsForFile: aString.
	^ allServices
		reject: [:svc | self unwantedSelectors includes: svc selector]! !

!DefaultExternalDropHandler methodsFor: 'private' stamp: 'dgd 4/5/2004 19:23'!
unwantedSelectors
	"private - answer a collection well known unwanted selectors "
	^ #(#removeLineFeeds: #addFileToNewZip: #compressFile: #putUpdate: )! !

Object subclass: #DiffElement
	instanceVariableNames: 'string hash match'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!DiffElement commentStamp: 'ul 9/11/2013 01:01' prior: 0!
My instances are container objects used by TextDiffBuilder for comparison. They hold a string and the precomputed hash of the string to speed up #=. They may reference another DiffElement object which is their pair in the diff.

Instance Variables
	hash:		<Integer>
	match:		<DiffElement>
	string:		<String>

hash
	- the hash of string, stored for fast access

match
	- another DiffElement object which has the same string and turned out to be my pair in the longest common subsequence found by a TextDiffBuilder, or nil if I don't match anything from the other sequence

string
	- a part of a longer text, typically a line
!

!DiffElement class methodsFor: 'as yet unclassified' stamp: 'klub 12/27/2009 01:51'!
string: aString

	^self new
		string: aString;
		yourself! !

!DiffElement methodsFor: 'comparing' stamp: 'ul 12/26/2009 01:10'!
hash

	^hash! !

!DiffElement methodsFor: 'testing' stamp: 'klub 12/27/2009 01:56'!
hasMatch

	^match notNil! !

!DiffElement methodsFor: 'accessing' stamp: 'ul 12/26/2009 04:38'!
match

	^match! !

!DiffElement methodsFor: 'accessing' stamp: 'klub 12/27/2009 01:20'!
matches: aDiffMatch

	match := aDiffMatch.
	aDiffMatch match: self! !

!DiffElement methodsFor: 'accessing' stamp: 'klub 12/27/2009 01:20'!
match: aDiffMatch

	match := aDiffMatch
	! !

!DiffElement methodsFor: 'printing' stamp: 'klub 12/27/2009 01:52'!
printOn: aStream

	super printOn: aStream.
	aStream
		nextPut: $(;
		print: hash;
		nextPutAll: ', ';
		print: string;
		nextPutAll: ', ';
		print: (match class == self class);
		nextPut: $)! !

!DiffElement methodsFor: 'accessing' stamp: 'ul 12/26/2009 01:48'!
string

	^string! !

!DiffElement methodsFor: 'accessing' stamp: 'ul 11/6/2010 02:23'!
string: aString

	string := aString.
	hash := string hash! !

!DiffElement methodsFor: 'comparing' stamp: 'ul 12/26/2009 01:47'!
= anObject

	^anObject class == self class and: [
		anObject hash = hash and: [
			anObject string = string ] ]! !

Object subclass: #DigitalSignatureAlgorithm
	instanceVariableNames: 'randKey randSeed'
	classVariableNames: 'HighBitOfByte SmallPrimes'
	poolDictionaries: ''
	category: 'System-Digital Signatures'!

!DigitalSignatureAlgorithm commentStamp: '<historical>' prior: 0!
This class implements the Digital Signature Algorithm (DSA) of the U.S. government's "Digital Signature Standard" (DSS). The DSA algorithm was proposed in 1991 and became a standard in May 1994. The official description is available as a Federal Information Processing Standards Publication (FIPS PUB 186, May 19, 1994). A companion standard, the Secure Hash Standard, or SHS (FIPS PUB 180-1, April 17, 1995), describes a 160-bit message digest algorithm known as the Secure Hash Algorithm (SHA). This message digest is used to compute the document signature.

Here's how to use it:

  1. The "signer" creates a pair of keys. One of these must be kept private. The other may be freely distributed. For example, it could be built into the signature checking code of an application.

  2. When the signer wishes to sign a packet of data (a "message") , he uses the secure hash algorithm to create a 160-bit message digest (hash) which is used as the input to DSA. The result of this is a pair of large numbers called a "signature" that is attached to the original message.

  3. When someone receives a signed message purported to have come from the signer, they compute the 160-bit hash of the message and pass that, along with the message signature and the signer's public key, to the signature verification algorithm. If the signature checks, then it is virtually guaranteed that the message originated from someone who had the signer's private key. That is, the message is not a forgery and has not been modified since it was signed. For example, if the message contains a program, and the recipient trusts the signer, then the recipient can run the program with the assurance that it won't do anything harmful. (At least, not intentionally. A digital signature is no guarantee against bugs!! :->)

The signer must keep the private key secure, since anyone who has the private key can forge the signer's signature on any message they like. As long as the secret key is not stolen, cryptographers believe it to be virtually impossible either to forge a signature, to find a message that matches an existing sigature, or to discover the signer's private key by analyzing message signatures. Knowing the public key (which, for example, could be recovered from an application that had it built in), does not weaken the security at all.

An excellent reference work on digital signatures and cryptography in general is:

  Schneier, Bruce
  "Applied Cryptography: Protocols, Algorithms, and Source Code in C"
  John Wiley and Sons, 1996.

I used this book as a guide to implementing many of the numerical algorithms required by DSA.

Patents and Export Restrictions:

Many digital signature technologies are patented. DSA is also patented, but the patent is owned by the U.S. government which has made DSA available royalty-free. There is a claim that the government patent infringes on an earlier patent by Schnorr, but the government is requiring the use of DSA, so they apparently believe this claim is not strong enough to be a serious threat to their own patent.

Most cryptography technology, including digital signature technology, requires an export license for it to be distributed outside the U.S. Recent legislation may have relaxed the export license requirements, but it would be prudent to check the current regulations before exporting this code.!

!DigitalSignatureAlgorithm class methodsFor: 'examples' stamp: 'jm 12/22/1999 11:23'!
example
	"Example of signing a message and verifying its signature."
	"Note: Secure random numbers are needed for key generation and message signing, but not for signature verification. There is no need to call initRandomFromUser if you are merely checking a signature."
	"DigitalSignatureAlgorithm example"

	| msg keys sig |
	msg := 'This is a test...'.
	keys := self testKeySet.
	sig := self sign: msg privateKey: keys first.
	self inform: 'Signature created'.
	(self verify: sig isSignatureOf: msg publicKey: keys last)
		ifTrue: [self inform: 'Signature verified.']
		ifFalse: [self error: 'ERROR!! Signature verification failed'].
! !

!DigitalSignatureAlgorithm class methodsFor: 'public' stamp: 'ads 7/31/2003 14:01'!
generateKeySet
	"Generate and answer a key set for code signing. The result is a pair (<private key><public key>). Each key is an array of four large integers. The signer must be sure to record this keys set and must keep the private key secret to prevent someone from forging their signature."
	"Note: Key generation can take some time. Open a transcript so you can see what's happening and take a coffee break!!"
	"Note: Unguessable random numbers are needed for key generation. The user will be prompted to type a really long random string (two or three lines) to initialize the random number generator before generating a key set. A different random string should be typed for every session; it is not a password and we wish to produce different random number streams."
	"DigitalSignatureAlgorithm generateKeySet"

	| dsa |
	dsa := DigitalSignatureAlgorithm new.
	(self confirm: 'Shall I seed the random generator from the current sound input?')
		ifTrue: [dsa initRandomNonInteractively]
		ifFalse: [dsa initRandomFromUser].
	^ dsa generateKeySet
! !

!DigitalSignatureAlgorithm class methodsFor: 'class initialization' stamp: 'jm 12/21/1999 19:15'!
initialize
	"DigitalSignatureAlgorithm initialize"

	"SmallPrimes is a list of small primes greater than two."
	SmallPrimes := Integer primesUpTo: 2000.
	SmallPrimes := SmallPrimes copyFrom: 2 to: SmallPrimes size.

	"HighBitOfByte maps a byte to the index of its top non-zero bit."
	HighBitOfByte := (0 to: 255) collect: [:byte | byte highBit].
! !

!DigitalSignatureAlgorithm class methodsFor: 'public' stamp: 'jm 12/22/1999 11:18'!
sign: aStringOrStream privateKey: privateKey
	"Sign the given message (a stream or string) and answer a signature string."
	"Note: Unguessable random numbers are needed for message signing. The user will be prompted to type a really long random string (two or three lines) to initialize the random number generator before signing a message. A different random string should be typed for every session; it is not a password and we wish to produce different random number streams."

	| dsa hasher h sig |
	dsa := DigitalSignatureAlgorithm new.
	dsa initRandomFromUser.
	hasher := SecureHashAlgorithm new.
	(aStringOrStream class isBytes)
		ifTrue: [h := hasher hashMessage: aStringOrStream]
		ifFalse: [h := hasher hashStream: aStringOrStream].
	sig := dsa computeSignatureForMessageHash: h privateKey: privateKey.
	^ dsa signatureToString: sig
! !

!DigitalSignatureAlgorithm class methodsFor: 'public' stamp: 'RAA 5/31/2000 08:46'!
sign: aStringOrStream privateKey: privateKey dsa: dsa
	"Sign the given message (a stream or string) and answer a signature string."
	"Note: Unguessable random numbers are needed for message signing. The user will be prompted to type a really long random string (two or three lines) to initialize the random number generator before signing a message. A different random string should be typed for every session; it is not a password and we wish to produce different random number streams."

	| hasher h sig |

	hasher := SecureHashAlgorithm new.
	(aStringOrStream class isBytes)
		ifTrue: [h := hasher hashMessage: aStringOrStream]
		ifFalse: [h := hasher hashStream: aStringOrStream].
	sig := dsa computeSignatureForMessageHash: h privateKey: privateKey.
	^ dsa signatureToString: sig
! !

!DigitalSignatureAlgorithm class methodsFor: 'examples' stamp: 'nice 1/12/2010 09:07'!
testExamplesFromDisk
	"verify messages from file on disk"
	"Note: Secure random numbers are needed for key generation and message signing, but not for signature verification. There is no need to call initRandomFromUser if you are merely checking a signature."
	"DigitalSignatureAlgorithm testExamplesFromDisk"

	FileStream readOnlyFileNamed: 'dsa.test.out' do: [:file |
		| msg  sig publicKey |
		[file atEnd] whileFalse: [
			sig := file nextChunk.
			msg := file nextChunk.
			publicKey := Compiler evaluate: file nextChunk.
			(self verify: sig isSignatureOf: msg publicKey: publicKey) ifTrue: [
				Transcript show: 'SUCCESS: ',msg; cr.
			] ifFalse: [
				self error: 'ERROR!! Signature verification failed'
			].
		].
	]! !

!DigitalSignatureAlgorithm class methodsFor: 'examples' stamp: 'jm 12/22/1999 11:28'!
testKeySet
	"Answer a pair of keys for testing. The first key is the private key, the second one is the public key."
	"WARNING: This test key set is public should be used only for testing!! In a real application, the user would create a set of keys using generateKeySet and would keep the private key secret."

	^ #(
		(8343811888543852523216773185009428259187948644369498021763210776677854991854533186365944349987509452133156416880596803846631577352387751880552969116768071 1197175832754339660404549606408619548226315875117 1433467472198821951822151391684734233265646022897503720591270330985699984763922266163182803556189497900262038518780931942996381297743579119123094520048965 957348690772296812)
		(8343811888543852523216773185009428259187948644369498021763210776677854991854533186365944349987509452133156416880596803846631577352387751880552969116768071 1197175832754339660404549606408619548226315875117 1433467472198821951822151391684734233265646022897503720591270330985699984763922266163182803556189497900262038518780931942996381297743579119123094520048965 4645213122572190617807944614677917601101008235397095646475699959851618402406173485853587185431290863173614335452934961425661774118334228449202337038283799))
! !

!DigitalSignatureAlgorithm class methodsFor: 'examples' stamp: 'nice 12/26/2009 01:35'!
timeDecode: count
	"Example of signing a message and verifying its signature."
	"Note: Secure random numbers are needed for key generation and message signing, but not for signature verification. There is no need to call initRandomFromUser if you are merely checking a signature."
	"DigitalSignatureAlgorithm timeDecode: 20"

	| dsa |
	dsa := DigitalSignatureAlgorithm new.
	dsa initRandomFromUser.

	#(1 10 100 1000 10000 100000) do: [ :extraLen |
		| msg keys sig s |
		s := String new: extraLen.
		1 to: s size do: [ :i | s at: i put: (Character value: 200 atRandom)].
		msg := 'This is a test...',s.
		keys := self testKeySet.
		sig := self sign: msg privateKey: keys first dsa: dsa.
		"self inform: 'Signature created'."
		self timeDirect: [
			count timesRepeat: [
				(self verify: sig isSignatureOf: msg publicKey: keys last)
					ifFalse: [self error: 'ERROR!! Signature verification failed'].
			].
		] as: 'verify msgLen = ',msg size printString count: count
	].
! !

!DigitalSignatureAlgorithm class methodsFor: 'testing' stamp: 'RAA 5/31/2000 08:40'!
timeDirect: aBlock as: aString count: anInteger

	Transcript show: anInteger asStringWithCommas,'  ',
		aString ,' took ',
		(Time millisecondsToRun: aBlock) asStringWithCommas,' ms'; cr
! !

!DigitalSignatureAlgorithm class methodsFor: 'testing' stamp: 'RAA 5/31/2000 08:21'!
time: aBlock as: aString count: anInteger

	^{anInteger. aString. (Time millisecondsToRun: aBlock)}! !

!DigitalSignatureAlgorithm class methodsFor: 'public' stamp: 'jm 12/22/1999 11:20'!
verify: signatureString isSignatureOf: aStringOrStream publicKey: publicKey
	"Answer true if the given signature string signs the given message (a stream or string)."
	"Note: Random numbers are not needed for signature verification; thus, there is no need to call initRandomFromUser before verifying a signature."

	| dsa hasher h sig |
	dsa := DigitalSignatureAlgorithm new.
	hasher := SecureHashAlgorithm new.
	(aStringOrStream class isBytes)
		ifTrue: [h := hasher hashMessage: aStringOrStream]
		ifFalse: [h := hasher hashStream: aStringOrStream].
	sig := dsa stringToSignature: signatureString.
	^ dsa verifySignature: sig ofMessageHash: h publicKey: publicKey
! !

!DigitalSignatureAlgorithm class methodsFor: 'examples' stamp: 'nice 1/12/2010 09:07'!
writeExamplesToDisk
	"Example of signing a message and verifying its signature. Used to create samples from one implementation that could later be tested with a different implementation"
	"Note: Secure random numbers are needed for key generation and message signing, but not for signature verification. There is no need to call initRandomFromUser if you are merely checking a signature."
	"DigitalSignatureAlgorithm writeExamplesToDisk"

	| keyList dsa msgList |
	dsa := DigitalSignatureAlgorithm new.
	dsa initRandomFromUser.
	self inform: 'About to generate 5 key sets. Will take a while'.
	keyList := {self testKeySet},((1 to: 5) collect: [ :ignore | self generateKeySet]).
	msgList := {'This is a test...'. 'This is the second test period.'. 'And finally, a third message'}.
	FileStream newFileNamed: 'dsa.test.out' do: [:file |
		msgList do: [ :msg |
			keyList do: [ :keys |
				| sig |
				sig := self sign: msg privateKey: keys first dsa: dsa.
				(self verify: sig isSignatureOf: msg publicKey: keys last) ifTrue: [
					file
						nextChunkPut: sig;
						nextChunkPut: msg;
						nextChunkPut: keys last storeString.
				] ifFalse: [
					self error: 'ERROR!! Signature verification failed'
				].
			].
		].
	]
! !

!DigitalSignatureAlgorithm methodsFor: 'public' stamp: 'nice 9/2/2010 21:48'!
computeSignatureForMessageHash: hash privateKey: privateKey
	"Answer the digital signature of the given message hash using the given private key. A signature is a pair of large integers. The private key is an array of four large integers: (p, q, g, x)."

	| p q g x r s k tmp |
	p := privateKey first.
	q := privateKey second.
	g := privateKey third.
	x := privateKey fourth.

	r := s := 0.
	[r = 0 or: [s = 0]] whileTrue: [
		k := self nextRandom160 \\ q.
		r := (g raisedTo: k modulo: p) \\ q.
		tmp := (hash + (x * r)) \\ q.
		s := ((k reciprocalModulo: q) * tmp) \\ q].

	^ Array with: r with: s
! !

!DigitalSignatureAlgorithm methodsFor: 'public' stamp: 'hh 8/3/2000 18:19'!
generateKeySet
	"Generate and answer a key set for DSA. The result is a pair (<private key><public key>). Each key is an array of four large integers. The private key is (p, q, g, x); the public one is (p, q, g, y). The signer must be sure to record (p, q, g, x), and must keep x secret to prevent someone from forging their signature."
	"Note: Key generation can take some time. Open a transcript so you can see what's happening and take a coffee break!!"

	| qAndPandS q p exp g h x y |
	qAndPandS := self generateQandP.
	Transcript show: 'Computing g...'.
	q := qAndPandS first.
	p := qAndPandS second.
	exp := (p - 1) / q.
	h := 2.
	[g := h raisedTo: exp modulo: p. g = 1] whileTrue: [h := h + 1].
	Transcript show: 'done.'; cr.
	Transcript show: 'Computing x and y...'.
	x := self nextRandom160.
	y := g raisedTo: x modulo: p.
	Transcript show: 'done.'; cr.
	Transcript show: 'Key generation complete!!'; cr.
	^ Array
		with: (Array with: p with: q with: g with: x)
		with: (Array with: p with: q with: g with: y)
! !

!DigitalSignatureAlgorithm methodsFor: 'private' stamp: 'nice 3/31/2011 01:29'!
generateQandP
	"Generate the two industrial-grade primes, q (160-bits) and p (512-bit) needed to build a key set. Answer the array (q, p, s), where s is the seed that from which q and p were created. This seed is normally discarded, but can be used to verify the key generation process if desired."

	| pBits halfTwoToTheP chunkCount sAndq q twoQ n c w x p s |
	pBits := 512.  "desired size of p in bits"
	halfTwoToTheP := 2 raisedTo: (pBits - 1).
	chunkCount := pBits // 160.

	Transcript show: 'Searching for primes q and p...'; cr.
	[
		sAndq := self generateSandQ.
		Transcript show: '  Found a candidate q.'; cr.
		s := sAndq first.
		q := sAndq last.
		twoQ := q bitShift: 1.
		n := 2.
		c := 0.
		[c < 4096] whileTrue: [
			w := self generateRandomLength: pBits s: s n: n.
			x := w + halfTwoToTheP.
			p := (x - ( x \\ twoQ)) + 1.
			p highBit = pBits ifTrue: [
				Transcript show: '    Testing potential p ', (c + 1) printString, '...'; cr.
				(self isProbablyPrime: p) ifTrue: [
					Transcript show: '  Found p!!'; cr.
					^ Array with: q with: p with: s]].
			n := n + chunkCount + 1.
			c := c + 1]] repeat! !

!DigitalSignatureAlgorithm methodsFor: 'private' stamp: 'jm 12/13/1999 16:36'!
generateRandomLength: bitLength s: s n: n
	"Answer a random number of bitLength bits generated using the secure hash algorithm."

	| sha out count extraBits v |
	sha := SecureHashAlgorithm new.
	out := 0.
	count := (bitLength // 160).
	extraBits := bitLength - (count * 160).
	0 to: count do: [:k |
		v := sha hashInteger: (s + n + k).
		k = count ifTrue: [
			v := v - ((v >> extraBits) << extraBits)].
		out := out bitOr: (v bitShift: (160 * k))].
	^ out
! !

!DigitalSignatureAlgorithm methodsFor: 'private' stamp: 'nice 4/11/2011 17:01'!
generateSandQ
	"Generate a 160-bit random seed s and an industrial grade prime q."

	| hasher s sPlusOne u q |
	hasher := SecureHashAlgorithm new.
	[
		s := self nextRandom160.
		sPlusOne := s + 1.
		sPlusOne highBit > 160 ifTrue: [sPlusOne := sPlusOne bitAnd: (1 bitShift: 160) - 1].
		u := (hasher hashInteger: s) bitXor: (hasher hashInteger: sPlusOne).
		q := u bitOr: ((1 bitShift: 159) bitOr: 1).
		(self isProbablyPrime: q) ifTrue: [^ Array with: s with: q]] repeat! !

!DigitalSignatureAlgorithm methodsFor: 'initialization' stamp: 'nice 12/14/2010 15:10'!
initRandomFromString: aString
	"Ask the user to type a long random string and use the result to seed the secure random number generator."

	| s k srcIndex |
	s := aString.
	k := LargePositiveInteger new: (s size min: 64).
	srcIndex := 0.
	k digitLength to: 1 by: -1 do: [:i |
		k digitAt: i put: (s at: (srcIndex := srcIndex + 1)) asciiValue].
	k := k normalize + (Random new next * 16r7FFFFFFF) asInteger.  "a few additional bits randomness"
	k highBit > 512 ifTrue: [k := k bitShift: k highBit - 512].
	self initRandom: k.
! !

!DigitalSignatureAlgorithm methodsFor: 'initialization' stamp: 'fbs 11/19/2013 17:38'!
initRandomFromUser
	"Ask the user to type a long random string and use the result to seed the secure random number generator."

	^ self initRandomFromString: (UIManager default request: 'Enter a long random string to seed the random generator.').! !

!DigitalSignatureAlgorithm methodsFor: 'initialization' stamp: 'fbs 7/25/2013 07:08'!
initRandomNonInteractively
	[self initRandom: (SoundService default randomBitsFromSoundInput: 512)]
		ifError: [self initRandomFromString: 
			Time millisecondClockValue printString, 
			Date today printString, 
			Smalltalk platformName printString].! !

!DigitalSignatureAlgorithm methodsFor: 'initialization' stamp: 'jm 1/11/2000 00:25'!
initRandom: randomInteger
	"Initialize the the secure random number generator with the given value. The argument should be a positive integer of up to 512 bits chosen randomly to avoid someone being able to predict the sequence of random values generated."
	"Note: The random generator must be initialized before generating a key set or signature. Signature verification does not require initialization of the random generator."

	randSeed := 16rEFCDAB8998BADCFE10325476C3D2E1F067452301.  "initial seed"
	randKey := randomInteger.
	Transcript show: 'Random seed: ', randomInteger printString; cr.
! !

!DigitalSignatureAlgorithm methodsFor: 'large integer arithmetic' stamp: 'nice 8/28/2010 21:16'!
inverseOf: x mod: n
	"Answer the inverse of x modulus n. That is, the integer y such that (x * y) \\ n is 1. Both x and n must be positive, and it is assumed that x < n and that x and n are integers."
	"Details: Use the extended Euclidean algorithm, Schneier, p. 247."

	| v u u1 u2 u3 t1 t2 t3 tmp |
	((x <= 0) or: [n <= 0]) ifTrue: [self error: 'x and n must be greater than zero'].
	x >= n ifTrue: [self error: 'x must be < n'].

	v := x.
	u := n.
	(x even and: [n even]) ifTrue: [self error: 'no inverse'].

	u1 := 1. u2 := 0. u3 := u.
	t1 := v. t2 := u - 1. t3 := v.
	[	[u3 even ifTrue: [
			((u1 odd) or: [u2 odd]) ifTrue: [
				u1 := u1 + v.
				u2 := u2 + u].
			u1 := u1 bitShift: -1.
			u2 := u2 bitShift: -1.
			u3 := u3 bitShift: -1].
		((t3 even) or: [u3 < t3]) ifTrue: [
			tmp := u1. u1 := t1. t1 := tmp.
			tmp := u2. u2 := t2. t2 := tmp.
			tmp := u3. u3 := t3. t3 := tmp].
		u3 even and: [u3 > 0]] whileTrue: ["loop while u3 is even"].

		[((u1 < t1) or: [u2 < t2]) and: [u1 > 0]] whileTrue: [
			u1 := u1 + v.
			u2 := u2 + u].
	
		u1 := u1 - t1.
		u2 := u2 - t2.
		u3 := u3 - t3.
		t3 > 0] whileTrue: ["loop while t3 > 0"].

	[u1 >= v and: [u2 >= u]] whileTrue: [
		u1 := u1 - v.
		u2 := u2 - u].

	u3 = 1 ifFalse: [self error: 'no inverse'].
	^ u - u2
! !

!DigitalSignatureAlgorithm methodsFor: 'large integer arithmetic' stamp: 'nice 8/28/2010 21:21'!
isProbablyPrime: p
	"Answer true if p is prime with very high probability. Such a number is sometimes called an 'industrial grade prime'--a large number that is so extremely likely to be prime that it can assumed that it actually is prime for all practical purposes. This implementation uses the Rabin-Miller algorithm (Schneier, p. 159)."

	| iterations factor pMinusOne b m r a j z couldBePrime |
	iterations := 50.  "Note: The DSA spec requires >50 iterations; Schneier says 5 are enough (p. 260)"

	"quick elimination: check for p divisible by a small prime"
	SmallPrimes ifNil: [  "generate list of small primes > 2"
		SmallPrimes := Integer primesUpTo: 2000.
		SmallPrimes := SmallPrimes copyFrom: 2 to: SmallPrimes size].
	factor := SmallPrimes detect: [:f | (p \\ f) = 0] ifNone: [nil].
	factor ifNotNil: [^ p = factor].

	pMinusOne := p - 1.
	b := self logOfLargestPowerOfTwoDividing: pMinusOne.
	m := pMinusOne bitShift: b negated.
	"Assert: pMinusOne = m * (2 raisedTo: b) and m is odd"

	Transcript show: '      Prime test pass '.
	r := Random new.
	1 to: iterations do: [:i |
		Transcript show: i printString; space.
		a := (r next * 16rFFFFFF) truncated.
		j := 0.
		z := (a raisedTo: m modulo: p) normalize.
		couldBePrime := z = 1.
		[couldBePrime] whileFalse: [
			z = 1 ifTrue: [Transcript show: 'failed!!'; cr. ^ false].  "not prime"
			z = pMinusOne
				ifTrue: [couldBePrime := true]
				ifFalse: [
					(j := j + 1) < b
						ifTrue: [z := (z * z) \\ p]
						ifFalse: [Transcript show: 'failed!!'; cr. ^ false]]]].  "not prime"

	Transcript show: 'passed!!'; cr.
	^ true  "passed all tests; probably prime"
! !

!DigitalSignatureAlgorithm methodsFor: 'private' stamp: 'nice 8/28/2010 21:04'!
logOfLargestPowerOfTwoDividing: aPositiveInteger
	"Answer the base-2 log of the largest power of two that divides the given integer. For example, the largest power of two that divides 24 is 8, whose log base-2 is 3. Do this efficiently even when the given number is a large integer. Assume that the given integer is > 0."
	"DigitalSignatureAlgorithm new logOfLargestPowerOfTwoDividing: (32 * 3)"

	^aPositiveInteger lowBit - 1! !

!DigitalSignatureAlgorithm methodsFor: 'private' stamp: 'jm 12/13/1999 14:39'!
nextRandom160
	"Answer a newly generated 160-bit random number in the range [1..(2^160 - 1)]."
	"Details: Try again in the extremely unlikely chance that zero is encountered."

	| result |
	result := 0.
	[result = 0] whileTrue: [
		result := SecureHashAlgorithm new hashInteger: randKey seed: randSeed.
		randKey := randKey + result + 1].
	^ result
! !

!DigitalSignatureAlgorithm methodsFor: 'public' stamp: 'md 7/30/2005 14:12'!
signatureToString: aSignature
	"Answer a string representation of the given signature. This string can be parsed using the stringToSignature: method."

	| s |
	s := WriteStream on: (String new: 2000).
	s nextPutAll: '[DSA digital signature '.
	s nextPutAll: aSignature first printStringHex.
	s space.
	s nextPutAll: aSignature second printStringHex.
	s nextPutAll: ']'.
	^ s contents
! !

!DigitalSignatureAlgorithm methodsFor: 'public' stamp: 'jm 12/14/1999 13:33'!
stringToSignature: aString
	"Answer the signature stored in the given string. A signature string has the format:

		 '[DSA digital signature <r> <s>]'

	where <r> and <s> are large positive integers represented by strings of hexidecimal digits."

	| prefix stream r s |
	prefix := '[DSA digital signature '.
	(aString beginsWith: prefix) ifFalse: [self error: 'bad signature prefix'].
	stream := ReadStream on: aString.
	stream position: prefix size.
	r := Integer readFrom: stream base: 16.
	stream next.
	s := Integer readFrom: stream base: 16.
	^ Array with: r with: s
! !

!DigitalSignatureAlgorithm methodsFor: 'public' stamp: 'nice 9/2/2010 21:48'!
verifySignature: aSignature ofMessageHash: hash publicKey: publicKey
	"Answer true if the given signature is the authentic signature of the given message hash. That is, if the signature must have been computed using the private key set corresponding to the given public key. The public key is an array of four large integers: (p, q, g, y)."

	| p q g y r s w u1 u2 v0 v |
	p := publicKey first.
	q := publicKey second.
	g := publicKey third.
	y := publicKey fourth.
	r := aSignature first.
	s := aSignature last.
	((r > 0) and: [r < q]) ifFalse: [^ false].  "reject"
	((s > 0) and: [s < q]) ifFalse: [^ false].  "reject"

	w := s reciprocalModulo: q.
	u1 := (hash * w) \\ q.
	u2 := (r * w) \\ q.
	v0 := (g raisedTo: u1 modulo: p) * (y raisedTo: u2 modulo: p).
	v := ( v0 \\ p) \\ q.
	^ v = r
! !

Object subclass: #DiskProxy
	instanceVariableNames: 'globalObjectName preSelector constructorSelector constructorArgs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Object Storage'!

!DiskProxy commentStamp: '<historical>' prior: 0!
A DiskProxy is an externalized form of an object to write on a
DataStream. It contains a "constructor" message to regenerate
the object, in context, when sent a comeFullyUpOnReload message
(i.e. "internalize").

We are now using DiskProxy for shared system objects like StrikeFonts.

The idea is to define, for each kind of object that needs special
externalization, a class method that will internalize the object by
reconstructing it from its defining state. We call this a
"constructor" method. Then externalize such an object as a frozen
message that invokes this method--a DiskProxy.

(Here is the old comment:
Constructing a new object is good for any object that (1) can not be
externalized simply by snapshotting and reloading its instance
variables (like a CompiledMethod or a Picture), or (2) wants to be
free to evolve its internal representation without making stored
instances obsolete (and dangerous). Snapshotting and reloading an
object"s instance variables is a dangerous breach of encapsulation.

The internal structure of the class is then free to evolve. All
externalized instances will be useful as long as the
constructor methods are maintained with the same semantics.

There may be several constructor methods for a particular class. This
is useful for (1) instances with characteristically different
defining state, and (2) newer, evolved forms of an object and its
constructors, with the old constructor methods kept around so old
data can still be properly loaded.)

Create one like this example from class Picture

    DiskProxy global: #Picture
            selector: #fromByteArray:
                args: (Array with: self storage asByteArray)

* See also subclass DiskProxyQ that will construct an object in
the above manner and then send it a sequence of messages. This may save
creating a wide variety of constructor methods. It is also useful because
the newly read-in DiskProxyQ can catch messages like #objectContainedIn:
(via #doesNotUnderstand:) and add them to the queue of messages to
send to the new object.

* We may also want a subclass of DiskProxy that evaluates a string
expression to compute the receiver of the constructor message.

My instance variables:
* globalObjectName -- the Symbol name of a global object in the
    System dictionary (usually a class).
* constructorSelector -- the constructor message selector Symbol to
    send to the global object (perform:withArguments:), typically a
    variation on newFrom:.
* constructorArgs -- the Array of arguments to pass in the
    constructor message.

-- 11/9/92 Jerry Morrison
!

!DiskProxy class methodsFor: 'instance creation' stamp: ''!
global: globalNameSymbol selector: selectorSymbol args: argArray
    "Create a new DiskProxy constructor with the given
     globalNameSymbol, selectorSymbol, and argument Array.
     It will internalize itself by looking up the global object name
     in the SystemDictionary (Smalltalk) and sending it this message
     with these arguments."

    ^ self new global: globalNameSymbol
             selector: selectorSymbol
                 args: argArray! !

!DiskProxy methodsFor: 'i/o' stamp: 'cwp 11/27/2015 20:54'!
comeFullyUpOnReload: smartRefStream
	"Internalize myself into a fully alive object after raw loading from a DataStream. (See my class comment.)  DataStream will substitute the object from this eval for the DiskProxy."
	| globalObj symbol pr nn arrayIndex env |

	symbol := globalObjectName.
	"See if class is mapped to another name"
	(smartRefStream respondsTo: #renamed) ifTrue: [
		"If in outPointers in an ImageSegment, remember original class name.  
		 See mapClass:installIn:.  Would be lost otherwise."
		((thisContext sender sender sender sender sender sender 
			sender sender receiver class == ImageSegment) and: [ 
		thisContext sender sender sender sender method == 
			(DataStream compiledMethodAt: #readArray)]) ifTrue: [
				arrayIndex := (thisContext sender sender sender sender) tempAt: 4.
					"index var in readArray.  Later safer to find i on stack of context."
				smartRefStream renamedConv at: arrayIndex put: symbol].	"save original name"
		symbol := smartRefStream renamed at: symbol ifAbsent: [symbol]].	"map"
	env := Environment current.
	globalObj := env valueOf: symbol ifAbsent: [
		preSelector == nil & (constructorSelector = #yourself) ifTrue: [
			Transcript cr; show: symbol, ' is undeclared.'.
			env undeclare: symbol.
			^ nil].
		^ self error: 'Global "', symbol, '" not found'].
	((symbol == #World) and: [Smalltalk isMorphic not]) ifTrue: [
		self inform: 'These objects will work better if opened in a Morphic World.
Dismiss and reopen all menus.'].

	preSelector ifNotNil: [
		Symbol hasInterned: preSelector ifTrue: [:selector |
			[globalObj := globalObj perform: selector] on: Error do: [:ex |
				ex messageText = 'key not found' ifTrue: [^ nil].
				^ ex signal]]
	].
	symbol == #Project ifTrue: [
		(constructorSelector = #fromUrl:) ifTrue: [
			nn := (constructorArgs first findTokens: '/') last.
			nn := (nn findTokens: '.|') first.
			pr := Project named: nn. 
			^ pr ifNil: [self] ifNotNil: [pr]].
		pr := globalObj perform: constructorSelector withArguments: constructorArgs.
		^ pr ifNil: [self] ifNotNil: [pr]].	"keep the Proxy if Project does not exist"

	constructorSelector ifNil: [^ globalObj].
	Symbol hasInterned: constructorSelector ifTrue: [:selector |
		[^ globalObj perform: selector withArguments: constructorArgs] on: Error do: [:ex |
			ex messageText = 'key not found' ifTrue: [^ nil].
			^ ex signal]
	].
				"args not checked against Renamed"
	^ nil 	"was not in proper form"! !

!DiskProxy methodsFor: 'accessing' stamp: 'tk 3/10/2000 23:50'!
constructorArgs
	^ constructorArgs! !

!DiskProxy methodsFor: 'accessing' stamp: 'tk 11/6/2000 22:38'!
constructorSelector
	^ constructorSelector! !

!DiskProxy methodsFor: 'exceptions' stamp: 'tk 3/14/2000 16:27'!
enter
	"Enter the new project"
	self enter: false revert: false saveForRevert: false.! !

!DiskProxy methodsFor: 'exceptions' stamp: 'tfel 8/15/2016 09:31'!
enter: returningFlag revert: revertFlag saveForRevert: saveForRevert
	"Look for our project on the server, then try to enter it!!  DiskProxy is acting as a stub for the real thing.  Called from a ProjectViewMorph in the current project.  If have url, use it.  Else look in current Project's server and folder."

	constructorSelector == #namedExample: ifTrue: ["Project namedUrl: xxx"
		^ ((Smalltalk at: globalObjectName) perform: #fromExampleEtoys:
					withArguments: constructorArgs) ].
	constructorSelector == #namedUrl: ifTrue: ["Project namedUrl: xxx"
		^ ((Smalltalk at: globalObjectName) perform: #fromUrl:
					withArguments: constructorArgs) ].
	constructorSelector == #named: ifTrue: [
		Project current fromMyServerLoad: constructorArgs first].	"name"
! !

!DiskProxy methodsFor: 'accessing' stamp: 'tk 11/6/2000 22:38'!
globalObjectName
	^ globalObjectName! !

!DiskProxy methodsFor: 'initialize-release' stamp: 'tk 4/8/1999 12:58'!
global: globalNameSymbol preSelector: aSelector selector: selectorSymbol args: argArray
	"Initialize self as a DiskProxy constructor with the given
	globalNameSymbol, selectorSymbol, and argument Array.
	I will internalize by looking up the global object name in the
	SystemDictionary (Smalltalk) and sending it this message with
	these arguments."

	globalObjectName := globalNameSymbol asSymbol.
	preSelector := aSelector asSymbol.
	constructorSelector := selectorSymbol asSymbol.
	constructorArgs := argArray.! !

!DiskProxy methodsFor: 'initialize-release' stamp: 'tk 11/4/1999 19:28'!
global: globalNameSymbol selector: selectorSymbol args: argArray
	"Initialize self as a DiskProxy constructor with the given
	globalNameSymbol, selectorSymbol, and argument Array.
	I will internalize by looking up the global object name in the
	SystemDictionary (Smalltalk) and sending it this message with
	these arguments."

	(globalNameSymbol beginsWith: 'AnObsolete') ifTrue: [
		self error: 'Trying to write out, ', globalNameSymbol].
	globalObjectName := globalNameSymbol asSymbol.
	constructorSelector := selectorSymbol asSymbol.
	constructorArgs := argArray.! !

!DiskProxy methodsFor: 'exceptions' stamp: 'RAA 5/17/2000 11:51'!
loadFromServer

	"In support of check for newer version in ProjectViewMorph menu"

	self enter
! !

!DiskProxy methodsFor: 'accessing' stamp: 'tk 11/6/2000 22:35'!
preSelector

	^ preSelector! !

!DiskProxy methodsFor: 'accessing' stamp: 'tk 4/8/1999 12:54'!
preSelector: aSelector

	preSelector := aSelector! !

!DiskProxy methodsFor: 'printing' stamp: 'ar 4/10/2005 18:46'!
printOn: aStream
	"Try to report the name of the project"

	globalObjectName == #Project ifFalse: [^ super printOn: aStream].
	constructorArgs size > 0 ifFalse: [^ super printOn: aStream].
	constructorArgs first isString ifFalse: [^ super printOn: aStream].
	aStream nextPutAll: constructorArgs first, ' (on server)'! !

!DiskProxy methodsFor: 'accessing' stamp: 'tk 10/6/2000 15:18'!
simpleGlobalOrNil
	"Return the object I refer to if it is a simple global in Smalltalk."

	preSelector ifNotNil: [^ nil].
	constructorSelector == #yourself ifFalse: [^ nil].
	^ Smalltalk at: globalObjectName ifAbsent: [nil].
! !

!DiskProxy methodsFor: 'i/o' stamp: 'tk 3/26/98 11:17'!
storeDataOn: aDataStream
	"Besides just storing, get me inserted into references, so structures will know about class DiskProxy."

	super storeDataOn: aDataStream.
	aDataStream references at: self put: #none.
		"just so instVarInfo: will find it and put it into structures"! !

Object subclass: #EventManager
	instanceVariableNames: 'actionMap'
	classVariableNames: 'ActionMaps'
	poolDictionaries: ''
	category: 'System-Object Events'!

!EventManager commentStamp: 'tlk 5/7/2006 20:01' prior: 0!
An EventManager is used to registers a 'observer' object's interest in in changes to an 'observed' object.  Then when the observered object is changed,  EventManager broadcasts the an update message to all objects with a registered interest.  Finally, the Event manager can be used to remove an object from the list of observer object.

An interested object is said to be a dependant on the target object.  Registering an interest in an event is called adding a dependant. Deregistering is called removing  a dependant.  The EventManager's action map is a WeakIdentityDictionary that maps events (selectors) to dependants (objects & selectors) in a way that ensures the mapping is to specific objects (hence identity) and in a way that allows the object to be garbage collected if not other used (hence weak.)  EventManager class has ActionMaps which has one actionMap for each object.

Classic uses of an EventManager are to implement the Observer Pattern, see ChangeNotification or the MorphicModle as examples.!

!EventManager class methodsFor: 'accessing' stamp: 'reThink 2/18/2001 14:42'!
actionMapFor: anObject

    ^self actionMaps
        at: anObject
        ifAbsent: [self createActionMap]! !

!EventManager class methodsFor: 'accessing' stamp: 'rww 10/2/2001 07:20'!
actionMaps

	ActionMaps == nil
		ifTrue: [ActionMaps := WeakIdentityKeyDictionary new].
	^ActionMaps! !

!EventManager class methodsFor: 'class initialization' stamp: 'ar 2/26/2010 23:01'!
cleanUp: aggressive
	"Dump all ActionMaps but only when we're aggressively cleaning"

	aggressive ifTrue:[ActionMaps := nil].! !

!EventManager class methodsFor: 'initialize-release' stamp: 'fbs 6/30/2013 11:05'!
flushEvents
	"EventManager flushEvents"
	
	self actionMaps keysAndValuesDo:[:rcvr :evtDict| rcvr ifNotNil:[
		"make sure we don't modify evtDict while enumerating"
		evtDict keys do:[:evtName| | msgSet |
			msgSet := evtDict at: evtName ifAbsent:[nil].
			(msgSet == nil) ifTrue:[rcvr removeActionsForEvent: evtName]]]].
	EventManager actionMaps finalizeValues. ! !

!EventManager class methodsFor: 'releasing' stamp: 'reThink 2/18/2001 15:34'!
releaseActionMapFor: anObject

	self actionMaps
		removeKey: anObject
		ifAbsent: []! !

!EventManager class methodsFor: 'accessing' stamp: 'reThink 2/25/2001 08:52'!
updateableActionMapFor: anObject

    ^self actionMaps
        at: anObject
        ifAbsentPut: [self createActionMap]! !

!EventManager methodsFor: 'accessing' stamp: 'reThink 2/18/2001 15:37'!
actionMap

    ^actionMap == nil
        ifTrue: [self createActionMap]
        ifFalse: [actionMap]! !

!EventManager methodsFor: 'dependents access' stamp: 'reThink 3/3/2001 10:07'!
addDependent: anObject
	"Make the given object one of the receiver's dependents."

	self
		when: self changedEventSelector
		send: self updateEventSelector
		to: anObject.
	^anObject! !

!EventManager methodsFor: 'dependents access' stamp: 'reThink 3/3/2001 10:07'!
breakDependents
	"Remove all of the receiver's dependents."

	self removeActionsForEvent: self changedEventSelector! !

!EventManager methodsFor: 'accessing' stamp: 'reThink 3/3/2001 10:07'!
changedEventSelector

	^#changed:! !

!EventManager methodsFor: 'updating' stamp: 'reThink 3/3/2001 10:20'!
changed: aParameter 
	"Receiver changed. The change is denoted by the argument aParameter. 
	Usually the argument is a Symbol that is part of the dependent's change 
	protocol. Inform all of the dependents."

	self 
		triggerEvent: self changedEventSelector
		with: aParameter! !

!EventManager methodsFor: 'dependents access' stamp: 'reThink 3/3/2001 10:18'!
dependents

	^(self actionSequenceForEvent: self changedEventSelector) asSet
		collect:
			[:each | each receiver]! !

!EventManager methodsFor: 'copying' stamp: 'nice 10/2/2010 00:09'!
postCopy
	super postCopy.
	self release! !

!EventManager methodsFor: 'accessing' stamp: 'reThink 2/18/2001 15:39'!
releaseActionMap

    actionMap := nil! !

!EventManager methodsFor: 'dependents access' stamp: 'reThink 3/3/2001 10:07'!
removeDependent: anObject
	"Remove the given object as one of the receiver's dependents."

	self 
		removeActionsWithReceiver: anObject
		forEvent: self changedEventSelector.
	^ anObject! !

!EventManager methodsFor: 'accessing' stamp: 'nice 9/21/2010 10:23'!
updateableActionMap
	^actionMap ifNil:
		[actionMap := self createActionMap]! !

!EventManager methodsFor: 'accessing' stamp: 'reThink 3/3/2001 10:07'!
updateEventSelector

	^#update:! !

EventManager subclass: #SystemEventManager
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!SystemEventManager commentStamp: 'tlk 5/7/2006 20:10' prior: 0!
A SystemEventManager is EventManager that overrides Object>>actionSequenceForEvent: anEventSelector to supply WeakActionSequenceTrappingErrors as the default event.

!

!SystemEventManager methodsFor: 'events-accessing' stamp: 'rw 7/20/2003 17:02'!
actionSequenceForEvent: anEventSelector

    ^(self actionMap
        at: anEventSelector asSymbol
        ifAbsent: [^WeakActionSequenceTrappingErrors new])
            asActionSequenceTrappingErrors! !

Object subclass: #ExternalDropHandler
	instanceVariableNames: 'action type extension'
	classVariableNames: 'DefaultHandler RegisteredHandlers'
	poolDictionaries: ''
	category: 'System-Support'!

!ExternalDropHandler class methodsFor: 'private' stamp: 'mt 7/22/2016 11:45'!
chooseServiceFrom: aCollection
	"private - choose a service from aCollection asking the user if  needed"
	^aCollection size = 1 
		ifTrue: [aCollection anyOne]
		ifFalse:[UIManager default 
			chooseFrom: (aCollection collect:[:each| each label])
			values: aCollection
			title: 'You dropped a file. Please choose a service:' translated withCRs].! !

!ExternalDropHandler class methodsFor: 'class initialization' stamp: 'mt 7/13/2016 11:31'!
cleanUp
	"Reset and reinstall the drop handlers"

	self initialize.! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'fbs 6/27/2013 17:34'!
defaultGZipHandler
	^ExternalDropHandler
		type: nil
		extension: 'gz'
		action: [:stream :pasteUp :event | | contents |
			contents := Cursor wait showWhile: [stream viewGZipContents].
				UIManager default
					edit: contents
					label: 'Decompressed contents of: ', stream localName].! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'mir 1/10/2002 17:17'!
defaultHandler
	DefaultHandler ifNil: [DefaultHandler := ExternalDropHandler type: nil extension: nil action: [:dropStream | dropStream edit]].
	^DefaultHandler! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'mir 1/10/2002 16:54'!
defaultHandler: externalDropHandler
	DefaultHandler := externalDropHandler! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'nice 7/22/2014 22:50'!
defaultImageHandler
	^ExternalDropHandler
		type: 'image/'
		extension: nil
		action: [:stream :pasteUp :event |
			| image sketch |
			stream binary.
			image := Form fromBinaryStream: stream contents asByteArray readStream.
			Project current resourceManager 
				addResource: image 
				url: (FileDirectory urlForFileNamed: stream name) asString.
			sketch := World drawingClass withForm: image.
			pasteUp addMorph: sketch centeredNear: event position.
			image := sketch := nil]! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'tak 9/5/2006 11:44'!
defaultProjectHandler
	^ ExternalDropHandler
		type: nil
		extension: 'pr'
		action: [:stream | ProjectLoading openOn: stream]! !

!ExternalDropHandler class methodsFor: 'class initialization' stamp: 'mt 7/13/2016 11:33'!
initialize
	"ExternalDropHandler initialize"

	self resetRegisteredHandlers.
	self
		"registerHandler: self defaultImageHandler;" "we have file services for images"
		registerHandler: self defaultGZipHandler;
		registerHandler: self defaultProjectHandler! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'mir 8/24/2004 15:37'!
lookupExternalDropHandler: stream

	| types extension serviceHandler |
	types := stream mimeTypes.

	types ifNotNil: [
		self registeredHandlers do: [:handler | 
			(handler matchesTypes: types)
				ifTrue: [^handler]]].

	extension := FileDirectory extensionFor: stream name.
	self registeredHandlers do: [:handler | 
		(handler matchesExtension: extension)
				ifTrue: [^handler]].
	serviceHandler := self lookupServiceBasedHandler: stream.
	^serviceHandler
		ifNil: [self defaultHandler]! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'fbs 8/9/2013 17:34'!
lookupServiceBasedHandler: dropStream
	"The file was just dropped; let's do our job"
	| fileName services theOne |
	fileName := dropStream name.

	services := (FileServices itemsForFile: fileName)
		reject: [:svc | self unwantedSelectors includes: svc selector].

	"no service, default behavior"
	services isEmpty
		ifTrue: [^nil].

	theOne := self chooseServiceFrom: services.
	^theOne
		ifNotNil: [ExternalDropHandler type: nil extension: nil action: [:stream | theOne performServiceFor: stream]]! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'mir 1/10/2002 15:57'!
registeredHandlers
	RegisteredHandlers ifNil: [RegisteredHandlers := OrderedCollection new].
	^RegisteredHandlers! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'mir 1/10/2002 17:19'!
registerHandler: aHandler
	self registeredHandlers add: aHandler! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'mir 1/10/2002 15:57'!
resetRegisteredHandlers
	RegisteredHandlers := nil! !

!ExternalDropHandler class methodsFor: 'instance creation' stamp: 'mir 1/10/2002 17:16'!
type: aType extension: anExtension action: anAction 
	^self new type: aType extension: anExtension action: anAction ! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'mir 8/24/2004 15:28'!
unwantedSelectors
	"private - answer a collection well known unwanted selectors "
	^ #(#removeLineFeeds: #addFileToNewZip: #compressFile: #putUpdate: )! !

!ExternalDropHandler methodsFor: 'accessing' stamp: 'mir 1/10/2002 15:54'!
extension
	^extension! !

!ExternalDropHandler methodsFor: 'accessing' stamp: 'ul 11/15/2010 11:39'!
handle: dropStream in: pasteUp dropEvent: anEvent
	| numArgs |
	numArgs := action numArgs.
	numArgs = 1
		ifTrue: [^action value: dropStream].
	numArgs = 2
		ifTrue: [^action value: dropStream value: pasteUp].
	numArgs = 3
		ifTrue: [^action value: dropStream value: pasteUp value: anEvent].
	self error: 'Wrong number of args for dop action.'! !

!ExternalDropHandler methodsFor: 'testing' stamp: 'spfa 5/25/2004 13:38'!
matchesExtension: aExtension
	(self extension isNil or: [aExtension isNil])
		ifTrue: [^false].
	FileDirectory activeDirectoryClass isCaseSensitive
		ifTrue: [^extension = aExtension]
		ifFalse: [^extension sameAs: aExtension]! !

!ExternalDropHandler methodsFor: 'testing' stamp: 'mir 1/10/2002 16:35'!
matchesTypes: types
	(self type isNil or: [types isNil])
		ifTrue: [^false].
	^types anySatisfy: [:mimeType | mimeType beginsWith: self type]! !

!ExternalDropHandler methodsFor: 'accessing' stamp: 'mir 1/10/2002 15:54'!
type
	^type! !

!ExternalDropHandler methodsFor: 'initialize' stamp: 'mir 1/10/2002 17:17'!
type: aType extension: anExtension action: anAction 
	action := anAction.
	type := aType.
	extension := anExtension! !

Object subclass: #ExternalObjectTable
	instanceVariableNames: 'semaphore externalObjectsArray freeSlotIndexes indexesByObjects'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

ExternalObjectTable class
	instanceVariableNames: 'current'!

!ExternalObjectTable commentStamp: 'ul 10/27/2014 16:26' prior: 0!
I'm an enhanced version of JMM's ExternalSemaphoreTable.
I'm responsible for maintaining the externalObjectsArray, which can be found at slot 39 of Smalltalk specialObjectsArray. Manipulating the array directly is highly discouraged.

My services are available via the methods of the "external objects" category of SmalltalkImage.

To register anObject simply use Smalltalk registerExternalObject: anObject. Use #unregisterExternalObject: to unregister it. It your responsibility to always unregister these objects.
If you'd like the create one or more new Semaphores, and register them for external use, then you should use one of #newExternalSemaphore, #newExternalSemaphoreDo: and #newExternalSemaphores:. If you want to unregister more than one external objects at the same time, then #unregisterExternalObjects: is the method you're looking for.

Implementation details:
I maintain a single instance (#current), which can automatically synchronize with externalObjectsArray.
All accesses are protected by the semaphore instance variable.
To ensure fast access, I keep track of the indexes of the external objects in the externalObjectArray in my indexesByObjects IdentityDictionary.
To ensure fast insertion, I keep track of the free slots of the externalObjectsArray in the freeSlotIndexes OrderedCollection. The access pattern is LIFO to avoid unnecessary growth and reordering of the OrderedCollection.
I can grow the externalObjectsArray (see #ensureFreeSlot), but I never shrink it.!

ExternalObjectTable class
	instanceVariableNames: 'current'!

!ExternalObjectTable class methodsFor: 'initialize-release' stamp: 'ul 10/27/2014 15:53'!
cleanUp: aggressive
	"It's safe to release the instance"

	self reset! !

!ExternalObjectTable class methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
current

	^current ifNil: [
		| newInstance |
		newInstance := self new.
		" Check again, because another process might have registered another instance. "
		current ifNil: [ current := newInstance ] ]! !

!ExternalObjectTable class methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
reset

	current := nil! !

!ExternalObjectTable methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
clearExternalObjects
	"Clear the array of objects that have been registered for use in non-Smalltalk code."

	semaphore critical: [
		Smalltalk specialObjectsArray at: 39 put: Array new.
		self 
			initializeCaches;
			synchronizeExternalObjectsArrayAndForceUpdate: true ]! !

!ExternalObjectTable methodsFor: 'private' stamp: 'ul 10/27/2014 16:30'!
ensureFreeSlot
	"Make sure there's at least one free slot."

	| oldSize newSize newExternalObjectsArray specialObjectsArray |
	freeSlotIndexes isEmpty ifFalse: [ ^self ].
	oldSize := externalObjectsArray size.
	newSize := oldSize + 20. "Linear growth is not really a problem, because we never shrink the array."
	newExternalObjectsArray := externalObjectsArray species new: newSize.
	newExternalObjectsArray
		replaceFrom: 1
		to: oldSize
		with: externalObjectsArray
		startingAt: 1.
	specialObjectsArray := Smalltalk specialObjectsArray.
	(specialObjectsArray at: 39) == externalObjectsArray
		ifFalse: [
			"We're not in sync. Try again."
			^self
				synchronizeExternalObjectsArrayAndForceUpdate: true;
				ensureFreeSlot ].
	specialObjectsArray at: 39 put: newExternalObjectsArray.
	externalObjectsArray := newExternalObjectsArray.
	newSize to: oldSize + 1 by: -1 do: [ :each |
		freeSlotIndexes addFirst: each ].
	Smalltalk maxExternalSemaphores ifNotNil: [ :maxExternalSemaphores |
		"Make sure that those new external semaphores can be signaled by the VM. This is not entirely safe, but not incrementing at all is worse."
		newSize > maxExternalSemaphores ifTrue: [
			Smalltalk maxExternalSemaphores: newSize ] ]	

	! !

!ExternalObjectTable methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
externalObjects
	"Return a copy of the externalObjectsArray."

	^semaphore critical: [
		self synchronizeExternalObjectsArrayAndForceUpdate: false.
		externalObjectsArray copy ]! !

!ExternalObjectTable methodsFor: 'private' stamp: 'ul 10/27/2014 16:30'!
externalObjectsArray: anArray
	"Update the index mapping using anArray as the new externalObjectsArray."

	externalObjectsArray := anArray.
	freeSlotIndexes reset.
	indexesByObjects removeAll.
	1 to: externalObjectsArray size do: [ :index |
		(anArray at: index) 
			ifNil: [ freeSlotIndexes addFirst: index ]
			ifNotNil: [ :object | indexesByObjects at: object put: index ] ]! !

!ExternalObjectTable methodsFor: 'initialize-release' stamp: 'ul 10/27/2014 15:53'!
initialize

	semaphore := Semaphore forMutualExclusion.
	self initializeCaches! !

!ExternalObjectTable methodsFor: 'private' stamp: 'ul 10/27/2014 15:53'!
initializeCaches

	freeSlotIndexes := OrderedCollection new.
	indexesByObjects := IdentityDictionary new! !

!ExternalObjectTable methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
newExternalSemaphore
	"Create and register a new Semaphore, and return an array containing itself and its index."

	| newSemaphore |
	^{
		newSemaphore := Semaphore new.
		semaphore critical: [
			self 
				synchronizeExternalObjectsArrayAndForceUpdate: false;
				safelyRegisterNewExternalObject: newSemaphore ] }! !

!ExternalObjectTable methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
newExternalSemaphoreDo: aBlock
	"Create and register a new Semaphore, then evaluate aBlock with it and its index."

	| newSemaphore |
	^aBlock
		value: (newSemaphore := Semaphore new)
		value: (semaphore critical: [
			self 
				synchronizeExternalObjectsArrayAndForceUpdate: false;
				safelyRegisterNewExternalObject: newSemaphore ])! !

!ExternalObjectTable methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
newExternalSemaphores: count
	"Create and register multiple Semaphores, and return an array containing the semaphores and their indexes in separate arrays."

	| semaphores indexes |
	semaphores := Array new: count.
	1 to: count do: [ :index | semaphores at: index put: Semaphore new ].
	indexes := Array new: count.
	semaphore critical: [
		self synchronizeExternalObjectsArrayAndForceUpdate: false.
		1 to: count do: [ :index |
			indexes at: index put: (
				self safelyRegisterNewExternalObject: (semaphores at: index)) ] ].
	^{ semaphores. indexes }! !

!ExternalObjectTable methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
registerExternalObject: anObject
	"Register the given object in the external objects array and return its index. If it is already there, just return its index."

	anObject ifNil: [ ^nil ].
	^semaphore critical: [
		self 
			synchronizeExternalObjectsArrayAndForceUpdate: false;
			safelyRegisterExternalObject: anObject ]! !

!ExternalObjectTable methodsFor: 'private' stamp: 'ul 10/27/2014 15:53'!
safelyRegisterExternalObject: anObject

	| index |
	(index := indexesByObjects at: anObject ifAbsent: nil) ifNotNil: [
		(externalObjectsArray at: index) == anObject ifTrue: [ ^index ].
		"indexesByObjects is out of sync. Try again."
		^self
			synchronizeExternalObjectsArrayAndForceUpdate: true;
			safelyRegisterExternalObject: anObject ].
	^self safelyRegisterNewExternalObject: anObject! !

!ExternalObjectTable methodsFor: 'private' stamp: 'ul 10/27/2014 15:53'!
safelyRegisterNewExternalObject: anObject

	| index |
	self ensureFreeSlot.
	index := freeSlotIndexes removeLast.
	(externalObjectsArray at: index) ifNotNil: [
		"Free slot is not empty. Try again."
		^self 
			synchronizeExternalObjectsArrayAndForceUpdate: true;
			safelyRegisterNewExternalObject: anObject ].
	externalObjectsArray at: index put: anObject.
	indexesByObjects at: anObject put: index.
	^index! !

!ExternalObjectTable methodsFor: 'private' stamp: 'ul 10/27/2014 15:53'!
safelyUnregisterExternalObject: anObject

	(indexesByObjects removeKey: anObject ifAbsent: nil)
		ifNotNil: [ :index |
			(externalObjectsArray at: index) == anObject ifFalse: [
				"We're not in sync. Try again."
				^self
					synchronizeExternalObjectsArrayAndForceUpdate: true;
					safelyUnregisterExternalObject: anObject ].
			externalObjectsArray at: index put: nil.
			freeSlotIndexes add: index ]
		ifNil: [
			"Check if we're just out of sync."
			(externalObjectsArray instVarsInclude: anObject) ifFalse: [ ^self ].
			"We're not in sync. Try again."
			^self
				synchronizeExternalObjectsArrayAndForceUpdate: true;
				safelyUnregisterExternalObject: anObject ]! !

!ExternalObjectTable methodsFor: 'private' stamp: 'ul 10/27/2014 15:53'!
synchronizeExternalObjectsArrayAndForceUpdate: forceUpdate

	| actualExternalObjectsArray |
	actualExternalObjectsArray := Smalltalk specialObjectsArray at: 39.
	(actualExternalObjectsArray == externalObjectsArray and: [ forceUpdate not ])
		ifTrue: [ ^self ].
	self externalObjectsArray: actualExternalObjectsArray! !

!ExternalObjectTable methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
unregisterExternalObjects: aCollection
	"Unregister the given objects from the external objects array. Do nothing if they aren't registered."

	semaphore critical: [
		self synchronizeExternalObjectsArrayAndForceUpdate: false.
		aCollection do: [ :each |
			each ifNotNil: [
				self safelyUnregisterExternalObject: each ] ] ]! !

!ExternalObjectTable methodsFor: 'accessing' stamp: 'ul 10/27/2014 15:53'!
unregisterExternalObject: anObject
	"Unregister the given object from the external objects array. Do nothing if it isn't registered."

	anObject ifNil: [ ^self ].
	semaphore critical: [
		self 
			synchronizeExternalObjectsArrayAndForceUpdate: false;
			safelyUnregisterExternalObject: anObject ]! !

Object subclass: #ExternalSettings
	instanceVariableNames: ''
	classVariableNames: 'RegisteredClients'
	poolDictionaries: ''
	category: 'System-Support'!

!ExternalSettings commentStamp: '<historical>' prior: 0!
ExternalSettings manages settings kept externally, e.g. files.
Objects can register themselves as clients to be notified at startup time to read their settings.

Eventually all the preferences should be managed through this mechanism.
!

!ExternalSettings class methodsFor: 'accessing' stamp: 'bf 8/8/2010 18:08'!
assuredPreferenceDirectory
	"Answer the preference directory, creating it if necessary"

	|  prefDir topDir |
	prefDir := self preferenceDirectory.
	prefDir
		ifNil:
			[topDir := Preferences startInUntrustedDirectory
				ifTrue: [FileDirectory on: SecurityManager default secureUserDirectory]
				ifFalse: [FileDirectory default].
			prefDir := topDir directoryNamed: self preferenceDirectoryName.
			prefDir assureExistence].
	^ prefDir! !

!ExternalSettings class methodsFor: 'class initialization' stamp: 'ar 8/23/2001 22:56'!
initialize
	"ExternalSettings initialize"
	"Order: ExternalSettings, SecurityManager, AutoStart"
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self! !

!ExternalSettings class methodsFor: 'accessing' stamp: 'mir 8/23/2002 14:22'!
parseServerEntryArgsFrom: stream
	"Args are in the form <argName>: <argValueString> delimited by end of line.
	It's not a very robust format and should be replaced by something like XML later.
	But it avoids evaluating the entries for security reasons."

	| entries lineStream entryName entryValue |
	entries := Dictionary new.
	stream skipSeparators.
	[stream atEnd]
		whileFalse: [
			lineStream := ReadStream on: stream nextLine.
			entryName := lineStream upTo: $:.
			lineStream skipSeparators.
			entryValue := lineStream upToEnd.
			(entryName isEmptyOrNil 
				or: [entryValue isEmptyOrNil])
				ifFalse: [entries at: entryName put: entryValue withoutTrailingBlanks].
			stream skipSeparators].
	^entries! !

!ExternalSettings class methodsFor: 'accessing' stamp: 'fbs 7/25/2013 07:08'!
preferenceDirectory
	| prefDirName path |
	prefDirName := self preferenceDirectoryName.
	path := Smalltalk vmPath.
	^(FileDirectory default directoryExists: prefDirName)
		ifTrue: [FileDirectory default directoryNamed: prefDirName]
		ifFalse: [
			((FileDirectory on: path) directoryExists: prefDirName)
				ifTrue: [(FileDirectory on: path) directoryNamed: prefDirName]
				ifFalse: [nil]]
! !

!ExternalSettings class methodsFor: 'accessing' stamp: 'mir 11/16/2001 13:33'!
preferenceDirectoryName
	^'prefs'! !

!ExternalSettings class methodsFor: 'accessing' stamp: 'mir 6/25/2001 18:45'!
registerClient: anObject
	"Register anObject as a settings client to be notified on startup."

	self registeredClients add: anObject! !

!ExternalSettings class methodsFor: 'private' stamp: 'mir 6/25/2001 18:46'!
registeredClients
	RegisteredClients ifNil: [RegisteredClients := Set new].
	^RegisteredClients! !

!ExternalSettings class methodsFor: 'class initialization' stamp: 'mir 8/22/2001 15:17'!
shutDown
	"Look for external defs and load them."
	"ExternalSettings shutDown"

	self registeredClients do: [:client | 
		client releaseExternalSettings]! !

!ExternalSettings class methodsFor: 'class initialization' stamp: 'mir 11/16/2001 13:29'!
startUp
	"Look for external defs and load them."
	"ExternalSettings startUp"

	| prefDir |
	prefDir := self preferenceDirectory.
	prefDir
		ifNil: [^self].
	self registeredClients do: [:client | 
		client fetchExternalSettingsIn: prefDir]! !

Object subclass: #FakeClassPool
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Tools'!

!FakeClassPool commentStamp: '<historical>' prior: 0!
The sole purpose of this class is to allow the Browser code pane to evaluate the class variables of the class whose method it is showing.  It does this by stuffing a pointer to the classpool dictionary of the class being shown into its own classpool.  It does this just around a doIt in the code pane.  An instance of FakeClasspool is then used as the receiver of the doIt.!

!FakeClassPool class methodsFor: 'initialize' stamp: 'dvf 9/27/2005 19:05'!
adopt: classOrNil
	"Temporarily use the classPool and sharedPools of another class"
	classOrNil isBehavior
		ifFalse: [classPool := nil.
				sharedPools := nil]
		ifTrue: [classPool := classOrNil classPool.
				sharedPools := classOrNil sharedPools]
! !

Object subclass: #FilePackage
	instanceVariableNames: 'fullName sourceSystem classes doIts classOrder'
	classVariableNames: 'LogFileStream'
	poolDictionaries: ''
	category: 'System-FilePackage'!

!FilePackage class methodsFor: 'conflict checker logging' stamp: 'dew 2/13/2004 23:33'!
closeLog
	self logFileStream close.
	LogFileStream := nil.! !

!FilePackage class methodsFor: 'instance creation' stamp: ''!
fromFileNamed: aName
	^self new fromFileNamed: aName! !

!FilePackage class methodsFor: 'instance creation' stamp: 'edc 4/1/2007 07:57'!
fromStream: aStream named: aName
	^self new fromStream: aStream named: aName! !

!FilePackage class methodsFor: 'conflict checker logging' stamp: 'dew 2/14/2004 23:03'!
logCr
	Transcript cr.
	self logFileStream nextPut: Character cr.
! !

!FilePackage class methodsFor: 'conflict checker logging' stamp: 'stephaneducasse 11/1/2005 15:39'!
logFileStream

	LogFileStream ifNil:
		[LogFileStream := FileStream fileNamed: 'ConflictChecker.log'.
		LogFileStream setToEnd].
	^ LogFileStream! !

!FilePackage class methodsFor: 'conflict checker logging' stamp: 'dew 2/14/2004 23:03'!
log: aString
	Transcript show: aString.
	self logFileStream nextPutAll: aString.
! !

!FilePackage class methodsFor: 'conflict checker preferences' stamp: 'dew 12/22/2002 00:19'!
showIdenticalConflicts
	"(Not implemented yet.  Need to implement versionFromChangeSet first for this to work for the 'false' case.)"
	"Set this to true if we want to show conflicts with methods which have an identical timestamp to the one being checked.  This type of conflict usually just proves that you've already loaded the changeset (or some part of it) in your image."
	^ true
! !

!FilePackage class methodsFor: 'conflict checker preferences' stamp: 'dew 2/13/2004 23:32'!
verboseConflicts
	"Set this to true if we want to list each replacing (potentially conflicting) method being checked."
	^ true! !

!FilePackage methodsFor: 'fileIn/fileOut' stamp: 'ar 8/6/2009 18:51'!
askForDoits
	| choice choices |
	choices := #('do not process' 'at the beginning' 'at the end' 'cancel').
	choice := nil.
	[choices includes: choice] whileFalse: [
		choice := UIManager default chooseFrom: choices values: choices title:
'The package contains unprocessed doIts.
When would like to process those?'].
	^choices indexOf: choice! !

!FilePackage methodsFor: 'accessing' stamp: ''!
classAt: className
	^self classes at: className! !

!FilePackage methodsFor: 'change record types' stamp: ''!
classComment: chgRec

	(self getClass: chgRec methodClassName) classComment: chgRec! !

!FilePackage methodsFor: 'accessing' stamp: ''!
classes
	^classes! !

!FilePackage methodsFor: 'change record types' stamp: ''!
doIt: chgRec
	| string |
	string := chgRec string.
	('*ubclass:*instanceVariableNames:*classVariableNames:*poolDictionaries:*category:*'
		match: string) ifTrue:[^self classDefinition: string with: chgRec].
	('* class*instanceVariableNames:*'
		match: string) ifTrue:[^self metaClassDefinition: string with: chgRec].
	('* removeSelector: *'
		match: string) ifTrue:[^self removedMethod: string with: chgRec].
	('* comment:*'
		match: string) ifTrue:[^self msgClassComment: string with: chgRec].
	('* initialize'
		match: string) ifTrue:[^self]. "Initialization is done based on class>>initialize"
	('''From *'
		match: string) ifTrue:[^self possibleSystemSource: chgRec].
	doIts add: chgRec.! !

!FilePackage methodsFor: 'fileIn/fileOut' stamp: 'wod 4/15/98 16:00'!
fileIn
	| doitsMark |
	doitsMark := 1.
	doIts isEmpty ifFalse:[doitsMark := self askForDoits].
	doitsMark = 4 ifTrue: [^nil].
	doitsMark = 2 ifTrue:[self fileInDoits].
	classOrder do:[:cls|
		cls fileInDefinition.
	].
	classes do:[:cls|
		Transcript cr; show:'Filing in ', cls name.
		cls fileInMethods.
		cls hasMetaclass ifTrue:[cls metaClass fileInMethods].
	].
	doitsMark = 3 ifTrue:[self fileInDoits].! !

!FilePackage methodsFor: 'fileIn/fileOut' stamp: ''!
fileInDoits
	doIts do:[:chgRec| chgRec fileIn].! !

!FilePackage methodsFor: 'reading' stamp: 'topa 10/7/2015 19:52'!
fileInFrom: aStream
	| changes |
	changes := ChangeSet scanFile: aStream from: 0 to: aStream size.
	aStream close.
	('Processing ', self packageName) 
		displayProgressFrom: 1
		to: changes size
		during:[:bar| | chgRec |
			1 to: changes size do:[:i|
				bar value: i.
				chgRec := changes at: i.
				self perform: chgRec type asMutator with: chgRec.
			].
		].! !

!FilePackage methodsFor: 'fileIn/fileOut' stamp: 'ar 7/17/2005 03:36'!
fileOut
	| fileName stream |
	fileName := UIManager default request: 'Enter the file name' initialAnswer:''.
	stream := FileStream newFileNamed: fileName.
	sourceSystem isEmpty ifFalse:[
		stream nextChunkPut: sourceSystem printString;cr ].
	self fileOutOn: stream.
	stream cr; cr.
	self classes do:[:cls|
		cls needsInitialize ifTrue:[
			stream cr; nextChunkPut: cls name,' initialize']].
	stream cr.
	stream close.

	"DeepCopier new checkVariables."
! !

!FilePackage methodsFor: 'fileIn/fileOut' stamp: ''!
fileOutDoits: aStream
	doIts do:[:chgRec| chgRec fileOutOn: aStream].! !

!FilePackage methodsFor: 'fileIn/fileOut' stamp: 'wod 4/15/98 15:59'!
fileOutOn: aStream
	| doitsMark |
	doitsMark := 1.
	doIts isEmpty ifFalse:[doitsMark := self askForDoits].
	doitsMark = 4 ifTrue: [^nil].
	doitsMark = 2 ifTrue:[self fileOutDoits: aStream].
	classOrder do:[:cls|
		cls fileOutDefinitionOn: aStream.
	].
	classes do:[:cls|
		cls fileOutMethodsOn: aStream.
		cls hasMetaclass ifTrue:[cls metaClass fileOutMethodsOn: aStream].
	].
	doitsMark = 3 ifTrue:[self fileOutDoits: aStream].! !

!FilePackage methodsFor: 'accessing' stamp: 'nk 2/18/2004 18:31'!
fixClassOrder
	"Essentially bubble sort the classOrder so that superclasses appear before subclasses"
	| superClass index subClass superIndex |
	index := 0.
	[index < classOrder size] whileTrue:[
		subClass := classOrder at: (index := index + 1).
		superClass := nil.
		subClass isMeta ifTrue:[
			"Treat non-meta as superclass"
			superClass := self classes at: subClass name ifAbsent:[nil].
		] ifFalse:[
			subClass hasDefinition ifTrue:[
				superClass := self classes 
					at: (Scanner new scanTokens: subClass definition) first ifAbsent:[nil].
				superClass ifNotNil:[superClass hasDefinition ifFalse:[superClass := nil]].
			].
		].
		superClass ifNotNil:[
			superIndex := classOrder indexOf: superClass ifAbsent:[self error:'Where is the class?'].
			superIndex > index ifTrue:[
				"Move superClass before index"
				classOrder remove: superClass.
				classOrder add: superClass before: subClass.
				"Rewind index - we need to check superClass itself"
				index := index - 1.
			].
		].
	].
! !

!FilePackage methodsFor: 'initialize' stamp: 'nice 1/13/2010 08:49'!
fromFileNamed: aName
	fullName := aName.
	FileStream readOnlyFileNamed: aName do: [:stream |
		stream setConverterForCode.
		self fileInFrom: stream]! !

!FilePackage methodsFor: 'initialize' stamp: 'yo 8/17/2004 09:54'!
fromFileNamed: aName encoding: encodingName
	| stream |
	fullName := aName.
	stream := FileStream readOnlyFileNamed: aName.
	stream converter: (TextConverter newForEncoding: encodingName).
	self fileInFrom: stream.! !

!FilePackage methodsFor: 'reading' stamp: 'topa 10/7/2015 20:54'!
fromStream: aStream named: aName
	| changes |
	changes := ChangeSet scanFile: aStream from: 0 to: aStream size.
	aStream close.
	('Processing ', aName) 
		displayProgressFrom: 1
		to: changes size
		during:[:bar| | chgRec |
			1 to: changes size do:[:i|
				bar value: i.
				chgRec := changes at: i.
				self perform: chgRec type asMutator with: chgRec.
			].
		].! !

!FilePackage methodsFor: 'accessing' stamp: 'pnm 8/23/2000 17:10'!
fullName: aString
	fullName := aString! !

!FilePackage methodsFor: 'accessing' stamp: ''!
fullPackageName
	^fullName! !

!FilePackage methodsFor: 'private' stamp: ''!
getClass: className
	| pseudoClass |
	(classes includesKey: className) ifTrue:[
		^classes at: className.
	].
	pseudoClass := PseudoClass new.
	pseudoClass name: className.
	classes at: className put: pseudoClass.
	^pseudoClass.! !

!FilePackage methodsFor: 'initialize' stamp: 'pnm 8/23/2000 14:48'!
initialize
	classes := Dictionary new.
	classOrder := OrderedCollection new.
	sourceSystem := ''.
	doIts := OrderedCollection new.! !

!FilePackage methodsFor: 'private' stamp: ''!
metaClassDefinition: string with: chgRec
	| tokens theClass |
	tokens := Scanner new scanTokens: string.
	theClass := self getClass: (tokens at: 1).
	theClass metaClass definition: string.
	classOrder add: theClass metaClass.! !

!FilePackage methodsFor: 'change record types' stamp: ''!
method: chgRec
	(self getClass: chgRec methodClassName) methodChange: chgRec! !

!FilePackage methodsFor: 'private' stamp: 'ar 4/10/2005 18:46'!
msgClassComment: string with: chgRec
	| tokens theClass |
	tokens := Scanner new scanTokens: string.
	(tokens size = 3 and:[(tokens at: 3) isString]) ifTrue:[
		theClass := self getClass: tokens first.
		^theClass commentString: tokens last].
	(tokens size = 4 and:[(tokens at: 3) asString = 'class' and:[(tokens at: 4) isString]]) ifTrue:[
		theClass := self getClass: tokens first.
		theClass metaClass commentString: tokens last].
! !

!FilePackage methodsFor: 'accessing' stamp: ''!
packageInfo
	^String streamContents:[:s|
		s nextPutAll:'Package: '.
		s nextPutAll: self fullPackageName; cr; cr.
		sourceSystem isEmpty ifFalse:[
			s nextPutAll: sourceSystem; cr; cr].
		doIts isEmpty ifFalse:[
			s nextPutAll:'Unresolvable doIts:'; cr; cr.
			doIts do:[:chgRec|
				s nextPut:$!!; nextPutAll: chgRec string; nextPut: $!!; cr]]].! !

!FilePackage methodsFor: 'accessing' stamp: 'pnm 8/23/2000 17:12'!
packageName
	^FileDirectory localNameFor: self fullPackageName! !

!FilePackage methodsFor: 'private' stamp: 'ar 4/10/2005 18:46'!
possibleSystemSource: chgRec
	| tokens |
	sourceSystem isEmpty ifTrue:[
		tokens := Scanner new scanTokens: chgRec string.
		(tokens size = 1 and:[tokens first isString]) ifTrue:[
			sourceSystem := tokens first.
			^self]].
	doIts add: chgRec.! !

!FilePackage methodsFor: 'change record types' stamp: ''!
preamble: chgRec
	self doIt: chgRec! !

!FilePackage methodsFor: 'accessing' stamp: ''!
removeClass: aPseudoClass
	(self classes removeKey: aPseudoClass name).
	classOrder copy do:[:cls|
		cls name = aPseudoClass name ifTrue:[ classOrder remove: cls].
	].! !

!FilePackage methodsFor: 'accessing' stamp: 'ar 2/5/2004 15:11'!
removeDoIts
	doIts := OrderedCollection new.! !

!FilePackage methodsFor: 'private' stamp: ''!
removedMethod: string with: chgRec
	| class tokens |
	tokens := Scanner new scanTokens: string.
	(tokens size = 3 and:[(tokens at: 2) == #removeSelector: ]) ifTrue:[
		class := self getClass: (tokens at: 1).
		^class removeSelector: (tokens at: 3).
	].
	(tokens size = 4 and:[(tokens at: 2) == #class and:[(tokens at: 3) == #removeSelector:]]) ifTrue:[
		class := self getClass: (tokens at: 1).
		^class metaClass removeSelector: (tokens at: 4).
	].
	doIts add: chgRec! !

!FilePackage methodsFor: 'accessing' stamp: ''!
renameClass: aPseudoClass to: newName
	| oldName |
	oldName := aPseudoClass name.
	self classes removeKey: oldName.
	self classes at: newName put: aPseudoClass.
	aPseudoClass renameTo: newName.! !

!FilePackage methodsFor: 'private' stamp: ''!
sampleMethod
"	In an existing method there are always a number of changes.
	Other stuff
		will be deleted
	Or even better,
		some things may be just modified.
"! !

Object subclass: #FileServices
	instanceVariableNames: ''
	classVariableNames: 'FileReaderRegistry'
	poolDictionaries: ''
	category: 'System-FileRegistry'!

!FileServices class methodsFor: 'accessing' stamp: 'ar 7/16/2005 16:59'!
allRegisteredServices
	"self allRegisteredServices"

	| col |
	col := OrderedCollection new.
	self registeredFileReaderClasses do: [:each | col addAll: (each services)].
	^ col! !

!FileServices class methodsFor: 'class initialization' stamp: 'ar 2/26/2010 23:02'!
cleanUp
	"Remove obsolete services"

	self removeObsolete! !

!FileServices class methodsFor: 'accessing' stamp: 'ar 9/29/2005 12:30'!
initialize
	"FileServices initialize"
	Smalltalk allClassesDo:[:aClass|
		(aClass class includesSelector: #fileReaderServicesForFile:suffix:)
			ifTrue:[self registerFileReader: aClass]].! !

!FileServices class methodsFor: 'accessing' stamp: 'ar 7/16/2005 17:00'!
isReaderNamedRegistered: aSymbol
	"return if a given reader class has been registered. Note that this is on purpose that the argument is
	a symbol and not a class"

	 ^ (self registeredFileReaderClasses collect: [:each | each name]) includes: aSymbol
! !

!FileServices class methodsFor: 'accessing' stamp: 'ar 7/16/2005 17:00'!
itemsForDirectory: aFileDirectory
	"Answer a list of services appropriate when no file is selected."

	| services |
	services := OrderedCollection new.
	self registeredFileReaderClasses do: [:reader |
		reader ifNotNil: [services addAll: (reader fileReaderServicesForDirectory: aFileDirectory) ]].
	^ services! !

!FileServices class methodsFor: 'accessing' stamp: 'ar 7/16/2005 17:00'!
itemsForFile: fullName
	"Answer a list of services appropriate for a file of the given full name"

	| services suffix |
	suffix := self suffixOf: fullName.
	services := OrderedCollection new.
	self registeredFileReaderClasses do: [:reader |
		reader ifNotNil: [services addAll: (reader fileReaderServicesForFile: fullName suffix: suffix)]].
	^ services! !

!FileServices class methodsFor: 'accessing' stamp: 'ar 7/16/2005 16:59'!
registeredFileReaderClasses
	FileReaderRegistry ifNil: [FileReaderRegistry := OrderedCollection new].
	^ FileReaderRegistry! !

!FileServices class methodsFor: 'accessing' stamp: 'ar 7/16/2005 17:00'!
registerFileReader: aProviderClass
	"register the given class as providing services for reading files"

	| registeredReaders |
	registeredReaders := self registeredFileReaderClasses.
	(registeredReaders includes: aProviderClass) 
			ifFalse: [ registeredReaders addLast: aProviderClass ]! !

!FileServices class methodsFor: 'accessing' stamp: 'ar 7/17/2005 02:36'!
removeObsolete
	"FileServices removeObsolete"
	self registeredFileReaderClasses copy 
		do:[:cls| cls isObsolete ifTrue:[self unregisterFileReader: cls]]! !

!FileServices class methodsFor: 'accessing' stamp: 'nk 12/7/2002 12:52'!
suffixOf: aName
	"Answer the file extension of the given file"
	^ aName
		ifNil:
			['']
		ifNotNil:
			[(FileDirectory extensionFor: aName) asLowercase]! !

!FileServices class methodsFor: 'accessing' stamp: 'ar 7/16/2005 17:00'!
unregisterFileReader: aProviderClass
	"unregister the given class as providing services for reading files"

	self registeredFileReaderClasses remove: aProviderClass ifAbsent: [nil]! !

Object subclass: #HTTPClient
	instanceVariableNames: ''
	classVariableNames: 'BrowserSupportsAPI RunningInBrowser'
	poolDictionaries: ''
	category: 'System-Support'!

!HTTPClient class methodsFor: 'class initialization' stamp: 'mir 4/2/2002 15:37'!
browserSupportsAPI
	^BrowserSupportsAPI == true! !

!HTTPClient class methodsFor: 'class initialization' stamp: 'mir 4/2/2002 15:36'!
browserSupportsAPI: aBoolean
	BrowserSupportsAPI := aBoolean! !

!HTTPClient class methodsFor: 'utilities' stamp: 'mir 8/22/2001 12:29'!
composeMailTo: address subject: subject body: body
	"HTTPClient composeMailTo: 'michael.rueger@squeakland.org' subject: 'test subject' body: 'message' "
	| mailTo |
	mailTo := WriteStream on: String new.
	mailTo nextPutAll: 'mailto:'.
	mailTo
		nextPutAll: address;
		nextPut: $?.
	subject isEmptyOrNil
		ifFalse: [mailTo nextPutAll: 'subject='; nextPutAll: subject; nextPut: $&].
	body isEmptyOrNil
		ifFalse: [mailTo nextPutAll: 'body='; nextPutAll: body].

	self httpGet: mailTo contents! !

!HTTPClient class methodsFor: 'class initialization' stamp: 'mir 2/2/2001 17:27'!
determineIfRunningInBrowser
	"HTTPClient determineIfRunningInBrowser"

	RunningInBrowser := StandardFileStream isRunningAsBrowserPlugin
! !

!HTTPClient class methodsFor: 'examples' stamp: 'mir 2/2/2001 17:43'!
exampleMailTo
	"HTTPClient exampleMailTo"

	HTTPClient mailTo: 'm.rueger@acm.org' message: 'A test message from within Squeak'
! !

!HTTPClient class methodsFor: 'examples' stamp: 'ul 1/11/2010 07:17'!
examplePostArgs
	"HTTPClient examplePostArgs"

	| args result |
	args := Dictionary new
		at: 'arg1' put: #('val1');
		at: 'arg2' put: #('val2');
		yourself.
	result := HTTPClient httpPostDocument: 'http://www.squeaklet.com/cgi-bin/thrd.pl [^]' args: args.
	Transcript show: result content; cr; cr.

! !

!HTTPClient class methodsFor: 'examples' stamp: 'ul 1/11/2010 07:17'!
examplePostMultipart
	"HTTPClient examplePostMultipart"

	| args result |
	args := Dictionary new
		at: 'arg1' put: #('val1');
		at: 'arg2' put: #('val2');
		yourself.
	result := HTTPClient httpPostMultipart: 'http://www.squeaklet.com/cgi-bin/thrd.pl'  args: args.
	Transcript show: result content; cr; cr.

! !

!HTTPClient class methodsFor: 'utilities' stamp: 'mir 5/13/2003 10:43'!
getDirectoryListing: dirListURL
	"HTTPClient getDirectoryListing: 'http://www.squeakalpha.org/uploads' "
	| answer ftpEntries |
"	answer := self 
		httpPostDocument: dirListURL
		args: Dictionary new."
	"Workaround for Mac IE problem"
	answer := self httpGetDocument: dirListURL.
	answer isString
		ifTrue: [^self error: 'Listing failed: ' , answer]
		ifFalse: [answer := answer content].
	answer first == $<
		ifTrue: [self error: 'Listing failed: ' , answer].
	ftpEntries := answer findTokens: String crlf.
	^ ftpEntries 
		collect:[:ftpEntry | ServerDirectory parseFTPEntry: ftpEntry]
		thenSelect: [:entry | entry notNil]! !

!HTTPClient class methodsFor: 'post/get' stamp: 'mir 5/11/2001 12:55'!
httpGetDocument: url
	| stream content | 
	^self shouldUsePluginAPI
		ifTrue: [
			stream := FileStream requestURLStream: url ifError: [self error: 'Error in get from ' , url printString].
			stream ifNil: [^''].
			stream position: 0.
			content := stream upToEnd.
			stream close.
			MIMEDocument content: content]
		ifFalse: [HTTPSocket httpGetDocument: url]! !

!HTTPClient class methodsFor: 'post/get' stamp: 'nk 8/30/2004 07:50'!
httpGet: url
	| document |
	document := self httpGetDocument: url.
	^(document isString)
		ifTrue: [
			"strings indicate errors"
			document]
		ifFalse: [(RWBinaryOrTextStream with: document content) reset]! !

!HTTPClient class methodsFor: 'post/get' stamp: 'mir 5/1/2001 15:04'!
httpPostDocument: url args: argsDict
	^self httpPostDocument: url target: nil args: argsDict! !

!HTTPClient class methodsFor: 'post/get' stamp: 'mir 5/1/2001 15:06'!
httpPostDocument: url target: target args: argsDict
	| argString stream content |
	^self shouldUsePluginAPI
		ifTrue: [
			argString := argsDict
				ifNotNil: [argString := HTTPSocket argString: argsDict]
				ifNil: [''].
			stream := FileStream post: argString , ' ' target: target url: url , argString ifError: [self error: 'Error in post to ' , url printString].
			stream position: 0.
			content := stream upToEnd.
			stream close.
			MIMEDocument content: content]
		ifFalse: [HTTPSocket httpPostDocument: url  args: argsDict]! !

!HTTPClient class methodsFor: 'post/get' stamp: 'mir 5/1/2001 12:51'!
httpPostMultipart: url args: argsDict
	" do multipart/form-data encoding rather than x-www-urlencoded "

	^self shouldUsePluginAPI
		ifTrue: [self pluginHttpPostMultipart: url args: argsDict]
		ifFalse: [HTTPSocket httpPostMultipart: url args: argsDict accept: nil request: '']! !

!HTTPClient class methodsFor: 'testing' stamp: 'ccn 3/14/2001 19:56'!
isRunningInBrowser

	RunningInBrowser isNil
		ifTrue: [self determineIfRunningInBrowser].
	^RunningInBrowser! !

!HTTPClient class methodsFor: 'testing' stamp: 'mir 8/4/2003 13:44'!
isRunningInBrowser: aBoolean
	"Override the automatic process.
	This should be used with caution.
	One way to determine it without using the primitive is to check for parameters typically only encountered when running as a plugin."

	RunningInBrowser := aBoolean! !

!HTTPClient class methodsFor: 'utilities' stamp: 'mir 5/1/2001 12:51'!
mailTo: address message: aString
	HTTPClient shouldUsePluginAPI
		ifFalse: [^self error: 'You need to run inside a web browser.'].
	FileStream post: aString url: 'mailto:' , address ifError: [self error: 'Can not send mail']! !

!HTTPClient class methodsFor: 'private' stamp: 'nice 12/27/2009 03:11'!
pluginHttpPostMultipart: url args: argsDict
	| mimeBorder argsStream crLf resultStream result |
	" do multipart/form-data encoding rather than x-www-urlencoded "

	crLf := String crlf.
	mimeBorder := '----squeak-', Time millisecondClockValue printString, '-stuff-----'.
	"encode the arguments dictionary"
	argsStream := WriteStream on: String new.
	argsDict associationsDo: [:assoc |
		assoc value do: [ :value | | fieldValue |
		"print the boundary"
		argsStream nextPutAll: '--', mimeBorder, crLf.
		" check if it's a non-text field "
		argsStream nextPutAll: 'Content-disposition: form-data; name="', assoc key, '"'.
		(value isKindOf: MIMEDocument)
			ifFalse: [fieldValue := value]
			ifTrue: [argsStream nextPutAll: ' filename="', value url pathForFile, '"', crLf, 'Content-Type: ', value contentType.
				fieldValue := (value content
					ifNil: [(FileStream fileNamed: value url pathForFile) contentsOfEntireFile]
					ifNotNil: [value content]) asString].
" Transcript show: 'field=', key, '; value=', fieldValue; cr. "
		argsStream nextPutAll: crLf, crLf, fieldValue, crLf.
	]].
	argsStream nextPutAll: '--', mimeBorder, '--'.
	resultStream := FileStream
		post: 
			('ACCEPT: text/html', crLf,
			'User-Agent: Squeak 3.1', crLf,
			'Content-type: multipart/form-data; boundary=', mimeBorder, crLf,
			'Content-length: ', argsStream contents size printString, crLf, crLf, 
			argsStream contents)
		url: url ifError: [^'Error in post ' url asString].
	"get the header of the reply"
	result := resultStream
		ifNil: ['']
		ifNotNil: [resultStream upToEnd].
	^MIMEDocument content: result! !

!HTTPClient class methodsFor: 'post/get' stamp: 'mir 4/2/2002 15:52'!
requestURL: url target: target
	^self shouldUsePluginAPI
		ifTrue: [FileStream requestURL: url target: target]
		ifFalse: [self error: 'Requesting a new URL target is not supported.']! !

!HTTPClient class methodsFor: 'testing' stamp: 'mt 5/6/2016 15:21'!
shouldUsePluginAPI
	"HTTPClient shouldUsePluginAPI" 

	self isRunningInBrowser
		ifFalse: [^false].
	self browserSupportsAPI
		ifFalse: [^false].
	"The Mac plugin calls do not work in full screen mode"
	^((Smalltalk platformName = 'Mac OS')
		and: [DisplayScreen displayIsFullScreen]) not! !

!HTTPClient class methodsFor: 'utilities' stamp: 'mir 2/2/2001 17:59'!
uploadFileNamed: aFilename to: baseUrl user: user passwd: passwd

	| fileContents remoteFilename |
	remoteFilename := (baseUrl endsWith: '/')
		ifTrue: [baseUrl , '/' , aFilename]
		ifFalse: [baseUrl , aFilename].
	fileContents := (StandardFileStream readOnlyFileNamed: aFilename) contentsOfEntireFile.
	HTTPSocket httpPut: fileContents to: remoteFilename user: user passwd: passwd! !

Object subclass: #HTTPDownloadRequest
	instanceVariableNames: 'semaphore url content loader process'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Download'!

HTTPDownloadRequest subclass: #CachedHTTPDownloadRequest
	instanceVariableNames: 'cachedName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Download'!

!CachedHTTPDownloadRequest methodsFor: 'accessing' stamp: 'ar 12/14/1999 14:53'!
cachedName
	^cachedName! !

!CachedHTTPDownloadRequest methodsFor: 'accessing' stamp: 'ar 12/14/1999 14:53'!
cachedName: aString
	cachedName := aString.! !

!CachedHTTPDownloadRequest methodsFor: 'accessing' stamp: 'ar 12/14/1999 15:00'!
startRetrieval
	| fileStream |
	cachedName == nil ifTrue:[^super startRetrieval].
	(FileDirectory default fileExists: cachedName) ifTrue:[
		fileStream := FileStream concreteStream new open: cachedName forWrite: false.
		fileStream == nil ifFalse:[^self content: 
			(MIMEDocument 
				contentType: 'text/plain' 
				content: fileStream contentsOfEntireFile)].
		FileDirectory default deleteFileNamed: cachedName ifAbsent:[]].
	super startRetrieval. "fetch from URL"
	"and cache in file dir"
	fileStream := FileStream concreteStream new open: cachedName forWrite: true.
	fileStream == nil ifFalse:[
		fileStream nextPutAll: (content content).
		fileStream close].! !

!HTTPDownloadRequest class methodsFor: 'instance creation' stamp: 'mir 10/7/1999 16:59'!
for: aUrl in: aLoader
	^self new for: aUrl in: aLoader! !

!HTTPDownloadRequest methodsFor: 'accessing' stamp: 'ar 5/30/2001 21:03'!
contentStream
	"Return a stream on the content of a previously completed HTTP request"
	semaphore wait.
	^content ifNotNil:[content contentStream]! !

!HTTPDownloadRequest methodsFor: 'accessing' stamp: 'mir 10/4/1999 18:25'!
contents
	semaphore wait.
	^content! !

!HTTPDownloadRequest methodsFor: 'accessing' stamp: 'mir 10/4/1999 18:26'!
content: retrievedContent
	content := retrievedContent.
	semaphore signal! !

!HTTPDownloadRequest methodsFor: 'initialize' stamp: 'KR 9/4/2005 10:33'!
for: aUrl in: aLoader
	url := aUrl.
	loader := aLoader.
	semaphore := Semaphore new.! !

!HTTPDownloadRequest methodsFor: 'private' stamp: 'yo 7/29/2005 16:04'!
httpEncodeSafely: aUrl
	"Encode the url but skip $/ and $:."

	| unescaped |
	unescaped := aUrl unescapePercents.

	^ unescaped encodeForHTTPWithTextEncoding: 'utf-8'
		conditionBlock: [:c | c isSafeForHTTP or: [c = $/ or: [c = $:]]].
! !

!HTTPDownloadRequest methodsFor: 'testing' stamp: 'ar 3/2/2001 16:53'!
isSemaphoreSignaled
	"Return true if the associated semaphore is currently signaled. This information can be used to determine whether the download has finished given that there is no other process waiting on the semaphore."
	^semaphore isSignaled! !

!HTTPDownloadRequest methodsFor: 'accessing' stamp: 'mir 10/7/1999 16:57'!
process: aProcess
	process := aProcess! !

!HTTPDownloadRequest methodsFor: 'accessing' stamp: 'mir 1/15/2000 22:55'!
signalAbort
	loader removeProcess: process.
	self content: 'Retrieval aborted'.
	process ifNotNil: [process terminate]! !

!HTTPDownloadRequest methodsFor: 'accessing' stamp: 'mir 1/10/2000 18:33'!
startRetrieval
	self content: url asUrl retrieveContents! !

HTTPDownloadRequest subclass: #PluginHTTPDownloadRequest
	instanceVariableNames: 'fileStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Download'!

!PluginHTTPDownloadRequest commentStamp: '<historical>' prior: 0!
HTTPBrowserRequest attempts to fetch the contents through a Webbrowser. This works transparently if Squeak is not running in the browser.!

!PluginHTTPDownloadRequest methodsFor: 'accessing' stamp: 'nk 8/30/2004 07:58'!
contentStream
	semaphore wait.
	fileStream
		ifNotNil: [^ fileStream].
	^ content
		ifNotNil: [content isString
				ifTrue: [self error: 'Error loading ' , self url printString]
				ifFalse: [content contentStream]]! !

!PluginHTTPDownloadRequest methodsFor: 'accessing' stamp: 'sd 1/30/2004 15:21'!
contents
	| |
	semaphore wait.
	(content isNil and:[fileStream notNil]) ifTrue:[
"		pos := fileStream position."
		fileStream position: 0.
		content := MIMEDocument content: fileStream upToEnd.
		fileStream close.
	].
	^content! !

!PluginHTTPDownloadRequest methodsFor: 'accessing' stamp: 'ar 12/21/1999 16:36'!
maxAttempts
	"Return the number of attempts to retry before giving up"
	^3! !

!PluginHTTPDownloadRequest methodsFor: 'accessing' stamp: 'sd 1/30/2004 15:21'!
signalAbort
	fileStream ifNotNil: [
		fileStream close].
	fileStream := nil.
	super signalAbort.! !

!PluginHTTPDownloadRequest methodsFor: 'accessing' stamp: 'mir 1/11/2000 11:36'!
startRetrieval
	| attempts |
	attempts := self maxAttempts.
	"Note: Only the first request may fail due to not running in a browser"
	url first = $/
		ifTrue: [url := url copyFrom: 2 to: url size].
	fileStream := FileStream requestURLStream: url ifError:[^super startRetrieval].
	[fileStream == nil] whileTrue:[
		attempts := attempts - 1.
		attempts = 0 ifTrue:[^self content:'Error downloading file'].
		fileStream := FileStream requestURLStream: url].
	semaphore signal.! !

Object subclass: #HTTPLoader
	instanceVariableNames: 'requests downloads'
	classVariableNames: 'DefaultLoader MaxNrOfConnections'
	poolDictionaries: ''
	category: 'System-Download'!

!HTTPLoader class methodsFor: 'accessing' stamp: 'mir 10/4/1999 18:41'!
default
	DefaultLoader ifNil: [
		DefaultLoader := HTTPLoader new].
	^DefaultLoader! !

!HTTPLoader class methodsFor: 'accessing' stamp: 'avi 4/30/2004 01:40'!
httpRequestClass
	^HTTPClient shouldUsePluginAPI
		ifTrue: [PluginHTTPDownloadRequest]
		ifFalse: [HTTPDownloadRequest]! !

!HTTPLoader class methodsFor: 'class initialization' stamp: 'mir 3/8/2001 16:31'!
initialize
	"HTTPLoader initialize"

	MaxNrOfConnections := 4.
	DefaultLoader ifNotNil: [
		DefaultLoader release.
		DefaultLoader := nil]! !

!HTTPLoader methodsFor: 'requests' stamp: 'mir 1/15/2000 22:59'!
abort
	| oldRequests |
	"Abort all requests"
	oldRequests := requests.
	requests := SharedQueue new.
	[oldRequests isEmpty] whileFalse: [
		oldRequests next signalAbort].
	downloads do: [:each | each ifNotNil: [each terminate]].
	downloads := OrderedCollection new
! !

!HTTPLoader methodsFor: 'private' stamp: 'mir 1/16/2000 16:11'!
addRequest: aHTTPRequest
	requests nextPut: aHTTPRequest.
	self startDownload! !

!HTTPLoader methodsFor: 'initialize/release' stamp: 'mir 10/7/1999 18:14'!
initialize
	requests := SharedQueue new.
	downloads := OrderedCollection new! !

!HTTPLoader methodsFor: 'private' stamp: 'mir 10/4/1999 18:31'!
maxNrOfConnections
	^MaxNrOfConnections! !

!HTTPLoader methodsFor: 'private' stamp: 'mir 10/7/1999 18:16'!
nextRequest
	^requests next! !

!HTTPLoader methodsFor: 'initialize/release' stamp: 'mir 10/7/1999 16:59'!
release
	self abort.
	downloads := nil.
	requests := nil! !

!HTTPLoader methodsFor: 'private' stamp: 'md 11/14/2003 16:38'!
removeProcess: downloadProcess
	downloads remove: downloadProcess ifAbsent: []! !

!HTTPLoader methodsFor: 'private' stamp: 'mir 10/7/1999 17:02'!
removeRequest: request
	requests remove: request! !

!HTTPLoader methodsFor: 'requests' stamp: 'mir 4/16/2001 17:48'!
retrieveContentsFor: url
	| request |
	request := self class httpRequestClass for: url in: self.
	self addRequest: request.
	^request contents! !

!HTTPLoader methodsFor: 'requests' stamp: 'nk 8/30/2004 07:50'!
retrieveObjectsFor: aURL
	"Load a remote image segment and extract the root objects.
	Check if the remote file is a zip archive."
	"'http://bradley.online.disney.com/games/subgame/squeak-test/assetInfo.extSeg' 
		asUrl loadRemoteObjects" 
	"'http://bradley.online.disney.com/games/subgame/squeak-test/assetInfo.zip' 
		asUrl loadRemoteObjects" 

	| stream info data |
 	data := self retrieveContentsFor: aURL.
	(data isString)
		ifTrue: [^self error: data]
		ifFalse: [data := data content].
	(data beginsWith: 'error')
		ifTrue: [^self error: data].
	data := data unzipped.
	stream := RWBinaryOrTextStream on: data.
	stream reset.
	info := stream fileInObjectAndCode.
	stream close.
	^info originalRoots! !

!HTTPLoader methodsFor: 'private' stamp: 'mir 5/12/2003 18:10'!
startDownload
	| newDownloadProcess |
	
	downloads size >= self maxNrOfConnections ifTrue: [^self].
	requests size <= 0 ifTrue: [^self].

	newDownloadProcess := [
		[
			self nextRequest startRetrieval
		] on: FTPConnectionException do: [ :ex | 
			Cursor normal show.
			self removeProcess: Processor activeProcess.
			self startDownload
		].
		self removeProcess: Processor activeProcess.
		self startDownload
	] newProcess.
	downloads add: newDownloadProcess.
	newDownloadProcess resume! !

Object subclass: #ImageSegment
	instanceVariableNames: 'arrayOfRoots segment outPointers state segmentName fileName userRootCnt renamedClasses'
	classVariableNames: 'BiggestFileNumber RecentlyRenamedClasses'
	poolDictionaries: ''
	category: 'System-Object Storage'!

!ImageSegment commentStamp: 'eem 8/21/2015 18:55' prior: 0!
I represent a segment of Squeak address space.  I am created from an
array of root objects.  After storing, my segment contains a binary
encoding of every object accessible from my roots but not otherwise
accessible from anywhere else in the system.  My segment contains
outward pointers that are indices into my table of outPointers.
On load my segment is converted back into objects and becommed
into an Array of the loaded objects, so they can be enumerated.

	The main use of ImageSegments is to store Projects.  A dummy
version of SmartRefStream traverses the Project.  Everything it finds
is classified as either an object that is owned by the project (only
pointed to inside the project), or an object outside the project that
is pointed to from inside the project.  The objects that are
completely owned by the project are compressed into pure binary form
in an ImageSegment.  The outside objects are put in the 'outPointers'
array.  The entire ImageSegment (binary part plus outPointers) is
encoded in a SmartRefStream, and saved on the disk.  (aProject
exportSegmentWithChangeSet:fileName:directory:) calls (anImageSegment
writeForExportWithSources:inDirectory:changeSet:).

	Note that every object inside the project is put into the
segment's arrayOfRoots.  This is because a dummy SmartRefStream to
scan the project, in order to make intelligent decisions about what
belongs in the project.
	See Project's class comment for what messages are sent to
objects as they are unpacked in a new image.

---- Older Details ------

	The primary kind of image segment is an Export Segment.  It
can be saved on a server and read into a completely different Squeak
image.
Old way to create one:
(ImageSegment new copyFromRootsForExport: (Array with: Baz with: Baz class))
		writeForExport: 'myFile.extSeg'.
Old way to create one for a project:
	(Project named: 'Play With Me - 3') exportSegment.
To read it into another image:  Select 'myFile.extSeg' in a FileList,
Menu 'load as project'.  It will install its classes automatically.
If you need to see the roots array, it is temporarily stored in
(SmartRefStream scannedObject).

Most of 'states' of an ImageSegment are not used to export a project,
and have been abandoned.

	When a segment is written out onto a file, it goes in a
folder called <image name>_segs.  If your image is called
"Squeak2.6.image", the folder "Squeak2.6_segs" must accompany the
image whenever your move, copy, or rename it.
	Whenever a Class is in arrayOfRoots, its class (aClass class)
must also be in the arrayOfRoots.
	There are two kinds of image segments.  Normal image segments
are a piece of a specific Squeak image, and can only be read back
into that image.  The image holds the array of outPointers that are
necessary to turn the bits in the file into objects.
	To put out a normal segment that holds a Project (not the
current project), execute (Project named: 'xxx') storeSegment.


arrayOfRoots	The objects that head the tree we will trace.
segment			The WordArray of raw bits of all objects in the tree.
outPointers		Oops of all objects outside the segment
pointed to from inside.
state			(see below)
segmentName	Its basic name.  Often the name of a Project.
fileName		The local name of the file.  'Foo-23.seg'
userRootCnt		number of roots submitted by caller.  Extras
are added in preparation for saving.

state that an ImageSegment may exist in...

#activeCopy			(has been copied, with the intent to
become active)
arrayOfRoots, segment, and outPointers have been created by
copyFromRoots:.  The tree of objects has been encoded in the segment,
but those objects are still present in the Squeak system.

#active				(segment is actively holding objects)
The segment is now the only holder of tree of objects.  Each of the
original roots has been transmuted into an ImageSegmentRootStub that
refers back to this image segment.  The original objects in the
segment will all be garbageCollected.

#onFile
The segment has been written out to a file and replaced by a file
pointer.  Only ImageSegmentRootStubs and the array of outPointers
remains in the image.  To get this far:
(ImageSegment new copyFromRoots: (Array with: Baz with: Baz class))
		writeToFile: 'myFile.seg'.

#inactive
The segment has been brought back into memory and turned back into
objects.  rootsArray is set, but the segment is invalid.

#onFileWithSymbols
The segment has been written out to a file, along with the text of
all the symbols in the outPointers array, and replaced by a file
pointer.  This reduces the size of the outPointers array, and also
allows the system to reclaim any symbols that are not referred to
from elsewhere in the image.  The specific format used is that of a
literal array as follows:
	#(symbol1 symbol2 # symbol3 symbol4 'symbolWithSpaces' # symbol5).
In this case, the original outPointers array was 8 long, but the
compacted table of outPointers retains only two entries.  These get
inserted in place of the #'s in the array of symbols after it is read
back in.  Symbols with embedded spaces or other strange characters
are written as strings, and converted back to symbols when read back
in.  The symbol # is never written out.
	NOTE: All IdentitySets or dictionaries must be rehashed when
being read back from this format.  The symbols are effectively
internal.  (No, not if read back into same image.  If a different
image, then use #imported.  -tk)

#imported
The segment is on an external file or just read in from one.  The
segment and outPointers are meant to be read into a foreign image.
In this form, the image segment can be read from a URL, and
installed.  A copy of the original array of root objects is
constructed, with former outPointers bound to existing objects in the
host system.
	(Any Class inside the segment MUST be in the arrayOfRoots.
This is so its association can be inserted into Smalltalk.  The
class's metaclass must be in roots also.  Methods that are in
outPointers because blocks point at them, were found and added to the
roots.
	All IdentitySets and dictionaries are rehashed when being
read back from exported segments.)


To discover why only some of the objects in a project are being
written out, try this (***Destructive Test***).  This breaks lots of
backpointers in the target project, and puts up an array of
suspicious objects, a list of the classes of the outPointers, and a
debugger.
"Close any transcripts in the target project"
World currentHand objectToPaste ifNotNil: [
	self inform: 'Hand is holding a Morph in its paste buffer:\' withCRs,
		World currentHand objectToPaste printString].
PV := Project named: 'xxxx'.
(IS := ImageSegment new) findRogueRootsImSeg:
	(Array with: PV world presenter with: PV world).
IS findOwnersOutPtrs.	"Optionally: write a file with owner chains"
"Quit and DO NOT save"

When an export image segment is brought into an image, it is like an
image starting up.  Certain startUp messages need to be run.  These
are byte and word reversals for nonPointer data that comes from a
machine of the opposite endianness.  #startUpProc passes over all
objects in the segment, and:
	The first time an instance of class X is encountered, (msg _
X startUpFrom: anImageSegment) is sent.  If msg is nil, the usual
case, it means that instances of X do not need special work.  X is
included in the IdentitySet, noStartUpNeeded.  If msg is not nil,
store it in the dictionary, startUps (aClass -> aMessage).
	When a later instance of X is encountered, if X is in
noStartUpNeeded, do nothing.  If X is in startUps, send the message
to the instance.  Typically this is a message like #swapShortObjects.
	Every class that implements #startUp, should see if it needs
a parallel implementation of #startUpFrom:.  !

!ImageSegment class methodsFor: 'testing' stamp: 'di 2/17/2000 21:58'!
activeClasses   "ImageSegment activeClasses"
	"Restore all remaining MD faults and return the active classes"

	| unused active |
	unused := OrderedCollection new.
	active := OrderedCollection new.
	Smalltalk allClasses do:
		[:c | (c instVarNamed: 'methodDict') 
			ifNil: [unused addLast: c]
			ifNotNil: [active addLast: c]].
	unused do: [:c | c recoverFromMDFault].
	^ active
! !

!ImageSegment class methodsFor: 'testing' stamp: 'nice 12/27/2009 03:11'!
activeClassesByCategory   "ImageSegment activeClassesByCategory"
	"Return a dictionary of active classes by system category.
	Useful for finding kernel categories to minimize swapping."

	| active dict |
	active := self activeClasses.
	dict := Dictionary new.
	active do:
		[:c | | cat list | cat := c category.
		list := dict at: cat ifAbsent: [Array new].
		dict at: cat put: (list copyWith: c)].
	^ dict
"
	ImageSegment discoverActiveClasses  <-- do it
		-- do something typical --
	ImageSegment activeClassesByCategory  <-- inspect it
"! !

!ImageSegment class methodsFor: 'accessing' stamp: 'mir 10/11/2000 17:33'!
compressedFileExtension
	^'sqz'! !

!ImageSegment class methodsFor: 'testing' stamp: 'nice 12/27/2009 03:11'!
discoverActiveClasses   "ImageSegment discoverActiveClasses" 
	"Run this method, do a few things, maybe save and resume the image.
	This will leave unused classes with MDFaults.
	You MUST follow this soon by activeClasses, or by swapOutInactiveClasses."

	"NOTE:  discoverActiveClasses uses Squeak's ability to detect and recover from faults due to a nil method dictionary.  It staches the method dict in with the organization during the time when discovery is in progress (Gag me with a spoon).  This is why the faults need to be cleared promptly before resuming normal work with the system.  It is also important that classes *do not* refer directly to their method dictionary, but only via the accessor message."
	
	Smalltalk allClasses do:
		[:c | | ok | ok := true.
		#(Array Object Class Message MethodDictionary) do:
			[:n | ((Smalltalk at: n) == c or:
				[(Smalltalk at: n) inheritsFrom: c]) ifTrue: [ok := false]].
		ok ifTrue: [c induceMDFault]].
"
	ImageSegment discoverActiveClasses.
		-- do something typical --
	PopUpMenu notify: ImageSegment activeClasses size printString , ' classes were active out of ' ,
			Smalltalk allClasses size printString.
"! !

!ImageSegment class methodsFor: 'accessing' stamp: 'mir 10/11/2000 17:32'!
fileExtension
	^'extSeg'! !

!ImageSegment class methodsFor: 'fileIn/Out' stamp: 'tk 9/24/1999 11:41'!
folder
	| im |
	"Full path name of segments folder.  Be sure to duplicate and rename the folder when you duplicate and rename an image.  Is $_ legal in all file systems?"

	im := Smalltalk imageName.
	^ (im copyFrom: 1 to: im size - 6 "'.image' size"), '_segs'! !

!ImageSegment class methodsFor: 'fileIn/Out' stamp: 'nice 12/27/2009 03:11'!
reclaimObsoleteSegmentFiles  "ImageSegment reclaimObsoleteSegmentFiles"
	"Delete segment files that can't be used after this image is saved.
	Note that this is never necessary -- it just saves file space."

	| segDir segFiles folderName byName exists |
	folderName := FileDirectory default class localNameFor: self folder.
	(FileDirectory default includesKey: folderName) ifFalse: [
		^ self "don't create if absent"].
	segDir := self segmentDirectory.
	segFiles := (segDir fileNames select: [:fn | fn endsWith: '.seg']) asSet.
	exists := segFiles copy.
	segFiles isEmpty ifTrue: [^ self].
	byName := Set new.
	"Remove (save) every file owned by a segment in memory"
	ImageSegment allInstancesDo: [:is | | aFileName | 
		(aFileName := is localName) ifNotNil: [
			segFiles remove: aFileName ifAbsent: [].
			(exists includes: aFileName) ifFalse: [
				Transcript cr; show: 'Segment file not found: ', aFileName].
			byName add: is segmentName]].
	"Of the segments we have seen, delete unclaimed the files."
	segFiles do: [:fName | 
		"Delete other file versions with same project name as one known to us"
		(byName includes: (fName sansPeriodSuffix stemAndNumericSuffix first))
			ifTrue: [segDir deleteFileNamed: fName]].! !

!ImageSegment class methodsFor: 'fileIn/Out' stamp: 'di 9/29/1999 15:45'!
segmentDirectory
	"Return a directory object for the folder of segments.
	Create such a folder if none exists."
	| dir folderName |
	dir := FileDirectory default.
	folderName := dir class localNameFor: self folder. "imageName:=segs"
	(dir includesKey: folderName) ifFalse:
		[dir createDirectory: folderName].	"create the folder if necess"
	^ dir directoryNamed: folderName! !

!ImageSegment class methodsFor: 'fileIn/Out' stamp: 'tk 9/23/1999 17:50'!
shutDown
	"Delete segment files that can't be used after this image is saved."

	"This is Optional.  
(1) How tell if saving image now?  Only do if is.
(2) ImageSegmentRootStub allInstancesDo: 
	If more than one file, delete all but one we are using now.
	Leave files with not stubs (could be out in a segment)
	Must forbid two projects from having the same name!!
(3) all Projects do:
	If project is in, delete all files with its name.
"
	! !

!ImageSegment class methodsFor: 'fileIn/Out' stamp: 'fbs 7/25/2013 07:10'!
startUp
	| choice |
	"Minimal thing to assure that a .segs folder is present"

(Preferences valueOfFlag: #projectsSentToDisk) ifTrue: [
	(FileDirectory default includesKey: (FileDirectory localNameFor: self folder)) 
		ifFalse: [
			choice := UIManager default 
				chooseFrom: #('Create folder' 'Quit without saving')
				title: 
					'The folder with segments for this image is missing.\' withCRs,
					self folder, '\If you have moved or renamed the image file,\' withCRs,
					'please Quit and rename the segments folder in the same way'.
			choice = 1 ifTrue: [FileDirectory default createDirectory: self folder].
			choice = 2 ifTrue: [Smalltalk snapshot: false andQuit: true]]]

	! !

!ImageSegment class methodsFor: 'testing' stamp: 'nice 12/27/2009 03:11'!
swapOutInactiveClasses  "ImageSegment swapOutInactiveClasses"  
	"Make up segments by grouping unused classes by system category.
	Read about, and execute discoverActiveClasses, and THEN execute this one."

	| unused groups |
	ImageSegment recoverFromMDFault.
	ImageSegmentRootStub recoverFromMDFault.
	unused := Smalltalk allClasses select: [:c | (c instVarNamed: 'methodDict') == nil].
	unused do: [:c | c recoverFromMDFault].
	groups := Dictionary new.
	SystemOrganization categories do:
		[:cat | | i |
		i := (cat findLast: [:c | c = $-]) - 1.
		i <= 0 ifTrue: [i := cat size].
		groups at: (cat copyFrom: 1 to: i)
			put: (groups at: (cat copyFrom: 1 to: i) ifAbsent: [Array new]) ,
			((SystemOrganization superclassOrder: cat) select: [:c | 
				unused includes: c]) asArray].
	groups keys do:
		[:cat | | roots | roots := groups at: cat.
		Transcript cr; cr; show: cat; cr; print: roots; endEntry.
		roots := roots , (roots collect: [:c | c class]).
		(cat beginsWith: 'Sys' "something here breaks") ifFalse:
			[(ImageSegment new copyFromRoots: roots sizeHint: 0) extract; 
				writeToFile: cat].
		Transcript cr; print: Smalltalk garbageCollect; endEntry]! !

!ImageSegment class methodsFor: 'testing' stamp: 'nice 12/27/2009 03:11'!
swapOutProjects  "ImageSegment swapOutProjects"  
	"Swap out segments for all projects other than the current one."

	| spaceLeft |
	spaceLeft := Smalltalk garbageCollect.
	Project allProjects doWithIndex:
		[:p :i | | newSpaceLeft | p couldBeSwappedOut ifTrue:
			[Transcript cr; cr; nextPutAll: p name.
			(ImageSegment new copyFromRoots: (Array with: p) sizeHint: 0)
				extract; writeToFile: 'project' , i printString.
			newSpaceLeft := Smalltalk garbageCollect.
			Transcript cr; print: newSpaceLeft - spaceLeft; endEntry.
			spaceLeft := newSpaceLeft]].! !

!ImageSegment class methodsFor: 'testing' stamp: 'tk 11/30/1999 22:27'!
testClassFaultOn: someClass  "ImageSegment testClassFaultOn: FileList"  
	"Swap out a class with an existing instance.  Then send a message to the inst.
	This will cause the VM to choke down deep and resend #cannotInterpret:.
	This in turn will send a message to the stubbed class which will choke
	and resend: #doesNotUnderstand:.  Then, if we're lucky, things will start working."

	(ImageSegment new copyFromRoots: (Array with: someClass with: someClass class) 
		sizeHint: 0) extract; writeToFile: 'test'.
! !

!ImageSegment class methodsFor: 'fileIn/Out' stamp: 'tk 12/16/1999 22:33'!
uniqueFileNameFor: segName
	"Choose a unique file name for the segment with this name."
	| segDir fileName listOfFiles |
	segDir := self segmentDirectory.
	listOfFiles := segDir fileNames.
	BiggestFileNumber ifNil: [BiggestFileNumber := 1].
	BiggestFileNumber > 99 ifTrue: [BiggestFileNumber := 1].	"wrap"
	[fileName := segName, BiggestFileNumber printString, '.seg'.
	 (listOfFiles includes: fileName)] whileTrue: [
		BiggestFileNumber := BiggestFileNumber + 1].	"force a unique file name"
	^ fileName! !

!ImageSegment methodsFor: 'compact classes' stamp: 'tk 1/8/2000 17:39'!
aComment
	"Compact classes are a potential problem because a pointer to the class would not ordinarily show up in the outPointers.  We add the classes of all compact classes to outPointers, both for local and export segments.
	Compact classes are never allowed as roots.  No compact class may be in an Environment that is written out to disk.  (In local segments, the compact classes array should never have an ImageSegmentRootStub in it.  For export, fileIn the class first, then load a segment with instances of it.  The fileIn code can be pasted onto the front of the .extSeg file) 
	For local segments, a class may become compact while its instances are out on the disk.  Or it may become un-compact.  A compact class may change shape while some of its instances are on disk.  All three cases go through (ClassDescription updateInstancesFrom:).  If it can't rule out an instance being in the segment, it reads it in to fix the instances.  
	See Behavior.becomeCompact for the rules on Compact classes.  Indexes may not be reused.  This is so that an incoming export segment has its index available.  (Changes may be needed in the way indexes are assigned.)
	For export segments, a compact class may have a different shape.  The normal class reshape mechanism will catch this.  During the installation of the segment, objects will have the wrong version of their class momentarily.  We will change them back before we get caught.
	For export segments, the last two items in outPointers are the number 1717 and an array of the compact classes used in this segment.  (The classes in the array are converted from DiskProxies by SmartRefStream.)  If that class is not compact in the new image, the instances are recopied.
	"! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'nice 2/23/2011 21:51'!
acceptSingleMethodSource: aDictionary

	| oldClassInfo oldClassName ismeta newName actualClass selector |
	oldClassInfo := (aDictionary at: #oldClassName) findTokens: ' '.	"'Class' or 'Class class'"
	oldClassName := oldClassInfo first asSymbol.
	ismeta := oldClassInfo size > 1.

	"must use class var since we may not be the same guy who did the initial work"

	newName := RecentlyRenamedClasses ifNil: [
		oldClassName
	] ifNotNil: [
		RecentlyRenamedClasses at: oldClassName ifAbsent: [oldClassName]
	].
	actualClass := Smalltalk at: newName.
	ismeta ifTrue: [actualClass := actualClass class].
	selector := actualClass newParser parseSelector: (aDictionary at: #methodText).
	(actualClass compiledMethodAt: selector ifAbsent: [^self "hosed input"]) 
		putSource: (aDictionary at: #methodText)
		fromParseNode: nil
		class: actualClass
		category: (aDictionary at: #category)
		withStamp: (aDictionary at: #changeStamp)
		inFile: 2
		priorMethod: nil.

! !

!ImageSegment methodsFor: 'instance change shape' stamp: 'ul 8/2/2011 20:02'!
allInstancesOf: aClass do: aBlock
	| withSymbols oldInstances segSize |
	"Bring me in, locate instances of aClass and submit them to the block.  Write me out again."

	(state = #onFile or: [state = #onFileWithSymbols]) ifFalse: [^ self].
	withSymbols := state = #onFileWithSymbols.
	(outPointers includes: aClass) ifFalse: [^ self].
		"If has instances, they point out at the class"
	state = #onFile ifTrue: [Cursor read showWhile: [self readFromFile]].
	segSize := segment size.
	self install.
	oldInstances := OrderedCollection new.
	self allObjectsDo: [:obj | obj class == aClass ifTrue: [
		oldInstances add: obj]].
	oldInstances do: aBlock.	"do the work"
	self copyFromRoots: arrayOfRoots sizeHint: segSize.
	self extract.
	withSymbols 
		ifTrue: [self writeToFileWithSymbols]
		ifFalse: [self writeToFile].

! !

!ImageSegment methodsFor: 'access' stamp: 'eem 8/21/2015 19:09'!
allObjectsDo: aBlock
	"Enumerate all objects that came from this segment.  NOTE this assumes that the segment was created (and extracted).  After the segment has been installed (install), this method allows you to enumerate its objects."
	self checkAndReportLoadError.
	segment do: aBlock! !

!ImageSegment methodsFor: 'access' stamp: 'tk 4/6/1999 13:15'!
arrayOfRoots
	^ arrayOfRoots! !

!ImageSegment methodsFor: 'access' stamp: 'tk 12/8/1999 21:12'!
arrayOfRoots: array
	arrayOfRoots := array! !

!ImageSegment methodsFor: 'error checking' stamp: 'eem 8/21/2015 19:08'!
checkAndReportLoadError
	"Check that the load has occurred.  A side-efect of the load primitive is to become
	 the segment into an Array of the loaded objects, so they can be enumerated.  If
	 this hasn't happened also check if the segment is a zero-length word array which
	 indicates we're running on an older Spur VM that doesn't do the become."
	segment isArray ifTrue: [^self]. "ok"
	(segment class == WordArrayForSegment
	 and: [segment size = 0]) ifTrue:
		[^self error: 'The load primitive has not becomed segment into an Array of the loaded objects. \Please upgrade your virtual machine to one that does this.' withCRs].
	^self error: 'Segment has not been becommed into the loaded objects'! !

!ImageSegment methodsFor: 'statistics' stamp: 'ar 2/21/2001 18:44'!
classNameAt: index
	| ccIndex |
	ccIndex := self compactIndexAt: index.
	ccIndex = 0 ifFalse:[^(Smalltalk compactClassesArray at: ccIndex) name].
	ccIndex := segment at: index-1.
	(ccIndex bitAnd: 16r80000000) = 0 ifTrue:[
		"within segment; likely a user object"
		^#UserObject].
	ccIndex := (ccIndex bitAnd: 16r7FFFFFFF) bitShift: -2.
	^(outPointers at: ccIndex) name! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'nice 11/13/2015 00:14'!
comeFullyUpOnReload: smartRefStream
	"fix up the objects in the segment that changed size.  An object in the segment is the wrong size for the modern version of the class. Construct a fake class that is the old size.  Replace the modern class with the old one in outPointers.  Load the segment. Traverse the instances, making new instances by copying fields, and running conversion messages.  Keep the new instances.  Bulk forward become the old to the new.  Let go of the fake objects and classes.
	After the install (below), arrayOfRoots is filled in. Globalize new classes.  Caller may want to do some special install on certain objects in arrayOfRoots.
	May want to write the segment out to disk in its new form."

	| mapFakeClassesToReal receiverClasses rootsToUnhiberhate myProject forgetDoItsClasses endianness |

	forgetDoItsClasses := Set new.
	RecentlyRenamedClasses := nil.		"in case old data hanging around"
	mapFakeClassesToReal := smartRefStream reshapedClassesIn: outPointers.
	"Dictionary of just the ones that change shape. Substitute them in outPointers."
	self fixCapitalizationOfSymbols.
	endianness := self endianness.
	arrayOfRoots := self loadSegmentFrom: segment outPointers: outPointers.
	self checkAndReportLoadError.
		"Can't use install.  Not ready for rehashSets"
	mapFakeClassesToReal isEmpty ifFalse: [
		self reshapeClasses: mapFakeClassesToReal refStream: smartRefStream
	].
	"When a Project is stored, arrayOfRoots has all objects in the project, except those in outPointers"
	arrayOfRoots do: [:importedObject | | existing |
		((importedObject isMemberOf: WideString) or: [importedObject isMemberOf: WideSymbol]) ifTrue: [
			importedObject mutateJISX0208StringToUnicode.
			importedObject class = WideSymbol ifTrue: [
				"self halt."
				Symbol hasInterned: importedObject asString ifTrue: [:multiSymbol |
					multiSymbol == importedObject ifFalse: [
						importedObject becomeForward: multiSymbol.
					].
				].
			].
		].
		(importedObject isKindOf: TTCFontSet) ifTrue: [
			existing := TTCFontSet
							familyName: importedObject familyName
							pointSize: importedObject pointSize.	"supplies default"
			existing == importedObject ifFalse: [importedObject becomeForward: existing].
		].
	].
	"Smalltalk garbageCollect.   MultiSymbol rehash.  These take time and are not urgent, so don't to them.  In the normal case, no bad MultiSymbols will be found."

	receiverClasses := self restoreEndianness: (endianness ~~ Smalltalk endianness).		"rehash sets"
	smartRefStream checkFatalReshape: receiverClasses.

	"Classes in this segment."
	arrayOfRoots do: [:importedObject |
		importedObject class class == Metaclass ifTrue: [
			forgetDoItsClasses add: importedObject.
			self declare: importedObject]].
	arrayOfRoots do: [:importedObject |
		importedObject isCompiledMethod ifTrue: [
			importedObject sourcePointer > 0 ifTrue: [importedObject zapSourcePointer]].
		(importedObject isKindOf: Project) ifTrue: [
			myProject := importedObject.
			importedObject ensureChangeSetNameUnique.
			Project addingProject: importedObject.
			importedObject restoreReferences.
			self dependentsRestore: importedObject]].

	rootsToUnhiberhate := arrayOfRoots select: [:importedObject |
		importedObject respondsTo: #unhibernate
	"ScriptEditors and ViewerFlapTabs"
	].
	myProject ifNotNil: [
		myProject world setProperty: #thingsToUnhibernate toValue: rootsToUnhiberhate asArray.
	].

	mapFakeClassesToReal isEmpty ifFalse: [
		mapFakeClassesToReal keysAndValuesDo: [:aFake :aReal |
			aFake removeFromSystemUnlogged.
			"do not assign the fake's hash to the real class"
			aFake becomeForward: aReal copyHash: false].
		SystemOrganization removeEmptyCategories].
	forgetDoItsClasses do: [:c | c forgetDoIts].
	"^ self"
! !

!ImageSegment methodsFor: 'compact classes' stamp: 'ar 2/21/2001 19:26'!
compactClassesArray
	| ccIndexes ind ccArray hdrBits |
	"A copy of the real compactClassesArray, but with only the classes actually used in the segment.  Slow, but OK for export."

	ccIndexes := Set new.
	ind := 2. 	"skip version word, first object"
	"go past extra header words"
	(hdrBits := (segment atPin: ind) bitAnd: 3) = 1 ifTrue: [ind := ind+1].
	hdrBits = 0 ifTrue: [ind := ind+2].

	[ccIndexes add: (self compactIndexAt: ind).	"0 if has class field"
	 ind := self objectAfter: ind.
	 ind > segment size] whileFalse.
	ccArray := Smalltalk compactClassesArray clone.
	1 to: ccArray size do: [:ii | "only the ones we use"
		(ccIndexes includes: ii) ifFalse: [ccArray at: ii put: nil]].
	^ ccArray! !

!ImageSegment methodsFor: 'compact classes' stamp: 'tk 12/21/1999 21:53'!
compactIndexAt: ind
	| word |
	"Look in this header word in the segment and find it's compact class index. *** Warning: When class ObjectMemory change, be sure to change it here. *** "

	((word := segment at: ind) bitAnd: 3) = 2 ifTrue: [^ 0].  "free block"
	^ (word >> 12) bitAnd: 16r1F 	"Compact Class field of header word"

! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'ul 6/17/2011 12:20'!
copyFromRootsForExport: rootArray 
	"When possible, use copySmartRootsExport:.  This way may not copy a complete tree of objects.  Add to roots: all of the methods pointed to from the outside by blocks."
	| newRoots list segSize symbolHolder |
	arrayOfRoots := rootArray.
	"self halt."
	symbolHolder := Symbol allSymbols.	"Hold onto Symbols with strong pointers, 
		so they will be in outPointers"
	(newRoots := self rootsIncludingPlayers) ifNotNil: [
		arrayOfRoots := newRoots].		"world, presenter, and all Player classes"
	"Creation of the segment happens here"
	self copyFromRoots: arrayOfRoots sizeHint: 0.
	segSize := segment size.
	[(newRoots := self rootsIncludingBlockMethods) == nil] whileFalse: [
		arrayOfRoots := newRoots.
		self copyFromRoots: arrayOfRoots sizeHint: segSize].
		"with methods pointed at from outside"
	[(newRoots := self rootsIncludingBlocks) == nil] whileFalse: [
		arrayOfRoots := newRoots.
		self copyFromRoots: arrayOfRoots sizeHint: segSize].
		"with methods, blocks from outPointers"
	"classes of receivers of blocks"
	list := self compactClassesArray.
	outPointers := outPointers, ((list select: [:cls | cls ~~ nil]), (Array with: 1717 with: list)).
	"Zap sender of a homeContext. Can't send live stacks out."
	1 to: outPointers size do: [:ii | 
		(outPointers at: ii) isBlock ifTrue: [outPointers at: ii put: nil].
		(outPointers at: ii) class == MethodContext ifTrue: [outPointers at: ii put: nil]].
	symbolHolder.! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'ul 6/17/2011 12:20'!
copyFromRootsLocalFileFor: rootArray sizeHint: segSize
	"If the roots include a World, add its Player classes to the roots."
	| newRoots |

	arrayOfRoots := rootArray.
	[(newRoots := self rootsIncludingPlayers) == nil] whileFalse: [
		arrayOfRoots := newRoots].		"world, presenter, and all Player classes"
	self copyFromRoots: arrayOfRoots sizeHint: segSize.
! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tk 12/5/2000 11:09'!
copyFromRoots: aRootArray sizeHint: segSizeHint
	"Copy a tree of objects into a WordArray segment.  The copied objects in the segment are not in the normal Squeak space.  If this method yields a very small segment, it is because objects just below the roots are pointed at from the outside.  (See findRogueRootsImSeg: for a *destructive* diagnostic of who is pointing in.)
	Caller must hold onto Symbols.
	To go faster, make sure objects are not repeated in aRootArray and other method directly, with true."

	self copyFromRoots: aRootArray sizeHint: segSizeHint areUnique: false
! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'eem 8/21/2015 18:33'!
copyFromRoots: aRootArray sizeHint: segSizeHint areUnique: areUnique
	"Copy a tree of objects into a WordArray segment.  The copied objects in the segment are not in the normal Squeak space.  
	[1] For exporting a project.  Objects were enumerated by ReferenceStream and aRootArray has them all.
	[2] For exporting some classes.  See copyFromRootsForExport:. (Caller must hold Symbols, or they will not get registered in the target system.)
	[3] For 'local segments'.  outPointers are kept in the image.
	If this method yields a very small segment, it is because objects just below the roots are pointed at from the outside.  (See findRogueRootsImSeg: for a *destructive* diagnostic of who is pointing in.)"
	| segmentWordArray outPointerArray segSize rootSet uniqueRoots |
	aRootArray ifNil: [self errorWrongState].
	uniqueRoots := areUnique 
		ifTrue: [aRootArray]
		ifFalse: [rootSet := IdentitySet new: aRootArray size * 3.
			uniqueRoots := OrderedCollection new.
			1 to: aRootArray size do: [:ii |	"Don't include any roots twice"
				(rootSet includes: (aRootArray at: ii)) 
					ifFalse: [
						uniqueRoots addLast: (aRootArray at: ii).
						rootSet add: (aRootArray at: ii)]
					ifTrue: [userRootCnt ifNotNil: ["adjust the count"
								ii <= userRootCnt ifTrue: [userRootCnt := userRootCnt - 1]]]].
			uniqueRoots].
	arrayOfRoots := uniqueRoots asArray.
	rootSet := uniqueRoots := nil.	"be clean"
	userRootCnt ifNil: [userRootCnt := arrayOfRoots size].
	outPointers := nil.	"may have used this instance before"
	segSize := segSizeHint > 0 ifTrue: [segSizeHint *3 //2] ifFalse: [50000].

	["Guess a reasonable segment size"
	segmentWordArray := WordArrayForSegment new: segSize.
	outPointerArray := [Array new: segSize // 20] ifError: [
		state := #tooBig.  ^ self].
	"Smalltalk garbageCollect."
	(self storeSegmentFor: arrayOfRoots
					into: segmentWordArray
					outPointers: outPointerArray) == nil]
		whileTrue:
			["Double the segment size and try again"
			segmentWordArray := outPointerArray := nil.
			segSize := segSize * 2].
	segment := segmentWordArray.
	outPointers := outPointerArray.
	state := #activeCopy
! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'mt 8/16/2016 10:15'!
copySmartRootsExport: rootArray 
	"Use SmartRefStream to find the object.  Make them all roots.  Create the segment in memory.  Project should be in first five objects in rootArray."
	| newRoots list segSize symbolHolder replacements naughtyBlocks allClasses sizeHint proj dummy |

	"self halt."
	symbolHolder := Symbol allSymbols.	"Hold onto Symbols with strong pointers, 
		so they will be in outPointers"

	dummy := ReferenceStream on: (DummyStream on: nil).
		"Write to a fake Stream, not a file"
	"Collect all objects"
	dummy insideASegment: true.	"So Uniclasses will be traced"
	dummy rootObject: rootArray.	"inform him about the root"
	dummy nextPut: rootArray.
	(proj :=dummy project) ifNotNil: [self dependentsSave: dummy].
	allClasses := SmartRefStream new uniClassInstVarsRefs: dummy.
		"catalog the extra objects in UniClass inst vars.  Put into dummy"
	allClasses do: [:cls | 
		dummy references at: cls class put: false.	"put Player5 class in roots"
		dummy blockers removeKey: cls class ifAbsent: []].
	"refs := dummy references."
	arrayOfRoots := self smartFillRoots: dummy.	"guaranteed none repeat"
	self savePlayerReferences: dummy references.	"for shared References table"
	replacements := dummy blockers.
	dummy project "recompute it" ifNil: [self error: 'lost the project!!'].
	dummy project class == DiskProxy ifTrue: [self error: 'saving the wrong project'].
	dummy := nil.	"force GC?"
	naughtyBlocks := arrayOfRoots select: [ :each |
		(each isKindOf: ContextPart) and: [each hasInstVarRef]
	].

	"since the caller switched ActiveWorld, put the real one back temporarily"
	naughtyBlocks isEmpty ifFalse: [
		World becomeActiveDuring: [World firstHand becomeActiveDuring: [ | goodToGo |
			goodToGo := (UIManager default
				chooseFrom: #('keep going' 'stop and take a look')
				title:
'Some block(s) which reference instance variables 
are included in this segment. These may fail when
the segment is loaded if the class has been reshaped.
What would you like to do?') = 1.
			goodToGo ifFalse: [
				naughtyBlocks inspect.
				self error: 'Here are the bad blocks'].
		]].
	].
	"Creation of the segment happens here"

	"try using one-quarter of memory min: four megs to publish (will get bumped later)"
	sizeHint := (Smalltalk garbageCollect // 4 // 4) min: 1024*1024.
	self copyFromRoots: arrayOfRoots sizeHint: sizeHint areUnique: true.
	segSize := segment size.
	[(newRoots := self rootsIncludingBlockMethods) == nil] whileFalse: [
		arrayOfRoots := newRoots.
		self copyFromRoots: arrayOfRoots sizeHint: segSize areUnique: true].
		"with methods pointed at from outside"
	[(newRoots := self rootsIncludingBlocks) == nil] whileFalse: [
		arrayOfRoots := newRoots.
		self copyFromRoots: arrayOfRoots sizeHint: segSize areUnique: true].
		"with methods, blocks from outPointers"
	list := self compactClassesArray.
	outPointers := outPointers, ((list select: [:cls | cls ~~ nil]), (Array with: 1717 with: list)).
	1 to: outPointers size do: [:ii | 
		(outPointers at: ii) isBlock ifTrue: [outPointers at: ii put: nil].
		(outPointers at: ii) class == MethodContext ifTrue: [outPointers at: ii put: nil].
		"substitute new object in outPointers"
		(replacements includesKey: (outPointers at: ii)) ifTrue: [
			outPointers at: ii put: (replacements at: (outPointers at: ii))]].
	proj ifNotNil: [self dependentsCancel: proj].
	symbolHolder.! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'kfr 5/28/2011 16:39'!
declareAndPossiblyRename: classThatIsARoot
	| existing catInstaller |
	"The class just arrived in this segment.  How fit it into the Smalltalk dictionary?  If it had an association, that was installed with associationDeclareAt:."

	catInstaller _ [
		classThatIsARoot superclass name == #Player 
			ifTrue: [classThatIsARoot category: Object categoryForUniclasses]
			ifFalse: [(classThatIsARoot superclass name beginsWith: 'WonderLandActor')
				ifTrue: [classThatIsARoot category: 'Balloon3D-UserObjects']
				ifFalse: [classThatIsARoot category: Object categoryForUniclasses]].
	].
	classThatIsARoot superclass addSubclass: classThatIsARoot.
	(Smalltalk includesKey: classThatIsARoot name) ifFalse: [
		"Class entry in Smalltalk not referred to in Segment, install anyway."
		catInstaller value.
		^ Smalltalk at: classThatIsARoot name put: classThatIsARoot].
	existing _ Smalltalk at: classThatIsARoot name.
	existing xxxClass == ImageSegmentRootStub ifTrue: [
		"We are that segment!!  Must ask it carefully!!"
		catInstaller value.
		^ Smalltalk at: classThatIsARoot name put: classThatIsARoot].
	existing == false | (existing == nil) ifTrue: [
		"association is in outPointers, just installed"
		catInstaller value.
		^ Smalltalk at: classThatIsARoot name put: classThatIsARoot].
	"Conflict with existing global or copy of the class"
	(existing isKindOf: Class) ifTrue: [
		classThatIsARoot isSystemDefined not ifTrue: [
			"UniClass.  give it a new name"
			classThatIsARoot setName: classThatIsARoot baseUniclass chooseUniqueClassName.
			catInstaller value.	"must be after new name"
			^ Smalltalk at: classThatIsARoot name put: classThatIsARoot].
		"Take the incoming one"
		self inform: 'Using newly arrived version of ', classThatIsARoot name.
		classThatIsARoot superclass removeSubclass: classThatIsARoot.	"just in case"
		(Smalltalk at: classThatIsARoot name) becomeForward: classThatIsARoot.
		catInstaller value.
		^ classThatIsARoot superclass addSubclass: classThatIsARoot].
	self error: 'Name already in use by a non-class: ', classThatIsARoot name.
! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'RAA 1/17/2001 12:06'!
declare: classThatIsARoot

	| nameOnArrival |
	"The class just arrived in this segment.  How fit it into the Smalltalk dictionary?  If it had an association, that was installed with associationDeclareAt:."

	nameOnArrival := classThatIsARoot name.
	self declareAndPossiblyRename: classThatIsARoot.
	nameOnArrival == classThatIsARoot name ifTrue: [^self].
	renamedClasses ifNil: [RecentlyRenamedClasses := renamedClasses := Dictionary new].
	renamedClasses at: nameOnArrival put: classThatIsARoot name.

! !

!ImageSegment methodsFor: 'testing' stamp: 'tk 11/30/1999 22:29'!
deepCopyTest: aRootArray
	"ImageSegment new deepCopyTest: Morph withAllSubclasses asArray"
	"Project allInstances do:
		[:p | p == Project current ifFalse:
			[Transcript cr; cr; nextPutAll: p name.
			ImageSegment new deepCopyTest: (Array with: p)]]."
	| t1 t2 copy |
	t1 := Time millisecondsToRun: [self copyFromRoots: aRootArray sizeHint: 0].
	t2 := Time millisecondsToRun: [copy := self segmentCopy].
	Transcript cr; print: segment size * 4; nextPutAll: ' bytes stored with ';
		print: outPointers size; show: ' outpointers in '; print: t1; show: 'ms.'.
	Transcript cr; nextPutAll: 'Reconstructed in '; print: t2; show: 'ms.'.
	^ copy
"
Smalltalk allClasses do:
	[:m | ImageSegment new deepCopyTest: (Array with: m with: m class)]
"! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'sw 11/19/2002 14:40'!
dependentsCancel: aProject
	"Erase the place we temporarily held the dependents of things in this project.  So we don't carry them around forever."

	aProject projectParameters removeKey: #GlobalDependentsInProject ifAbsent: []! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tk 10/21/2002 16:17'!
dependentsRestore: aProject
	"Retrieve the list of dependents from the exporting system, hook them up, and erase the place we stored them."

	| dict |
	dict := aProject projectParameterAt: #GlobalDependentsInProject.
	dict ifNil: [^ self].
	dict associationsDo: [:assoc |
		assoc value do: [:dd | assoc key addDependent: dd]].

	self dependentsCancel: aProject.! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'nice 12/27/2009 03:11'!
dependentsSave: dummy
	"Object that have dependents are supposed to be instances of subclasses of Model.  But, class Objects still provides 'Global Dependents', and some people still use them.  When both the model and the dependent are in a project that is being saved, remember them, so we can hook them up when this project is loaded in."

	| dict proj |
	proj := dummy project.
	dict := Dictionary new.
	DependentsFields associationsDo: [:assoc | | list |
		(dummy references includesKey: assoc key) ifTrue: [
			list := assoc value select: [:dd | dummy references includesKey: dd].
			list size > 0 ifTrue: [dict at: assoc key put: list]]].

	dict size > 0 ifTrue: [
		proj projectParameterAt: #GlobalDependentsInProject put: dict].
! !

!ImageSegment methodsFor: 'statistics' stamp: 'ar 2/21/2001 19:19'!
doSpaceAnalysis
	"Capture statistics about the IS and print the number of instances per class and space usage"
	| index sz word hdrBits cc instCount instSpace |
	state == #activeCopy ifFalse:[self errorWrongState].
	instCount := IdentityDictionary new.
	instSpace := IdentityDictionary new.
	index := 2. 	"skip version word, first object"
	"go past extra header words"
	hdrBits := (segment at: index) bitAnd: 3.
	hdrBits = 1 ifTrue: [index := index+1].
	hdrBits = 0 ifTrue: [index := index+2].
	[index > segment size] whileFalse:[
		hdrBits := (word := segment at: index) bitAnd: 3.
		hdrBits = 2 ifTrue:[sz := word bitAnd: 16rFFFFFFFC].
		hdrBits = 0 ifTrue:[sz := ((segment at: index-2) bitAnd: 16rFFFFFFFC) + 8].
		hdrBits = 1 ifTrue:[sz := (word bitAnd: "SizeMask" 252) + 4].
		hdrBits = 3 ifTrue:[sz := word bitAnd: "SizeMask" 252].
		hdrBits = 2 
			ifTrue:[cc := #freeChunk]
			ifFalse:[cc := self classNameAt: index].
		instCount at: cc put: (instCount at: cc ifAbsent:[0]) + 1.
		instSpace at: cc put: (instSpace at: cc ifAbsent:[0]) + sz.
		index := self objectAfter: index].
	^{instCount. instSpace}! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'eem 11/12/2015 17:10'!
endianness
	"Return which endian kind the incoming segment came from"

	segment class isBits ifFalse:
		["Hope that primitive 98 did the right thing - anyway, we lost information about endianness, so pretend we share the image's endianness."
		^Smalltalk endianness].
	^(segment first bitShift: -24) asCharacter == $d ifTrue: [#big] ifFalse: [#little]! !

!ImageSegment methodsFor: 'testing' stamp: 'di 3/26/1999 22:51'!
errorWrongState

	^ self error: 'wrong state'! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tk 11/30/1999 22:30'!
extract
	"This operation replaces (using become:) all the original roots of a segment with segmentRootStubs.  Thus the original objects will be reclaimed, and the root stubs will remain to bring the segment back in if it is needed."

	Cursor write showWhile: [
		state = #inactive ifTrue: [self copyFromRoots: arrayOfRoots sizeHint: 0].
		state = #activeCopy ifFalse: [self errorWrongState].
		arrayOfRoots elementsForwardIdentityTo:
			(arrayOfRoots collect: [:r | r rootStubInImageSegment: self]).
		state := #active].
! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'di 3/28/1999 13:47'!
extractThenInstall
	"For testing only"

	| newRoots |
	state = #activeCopy ifFalse: [self errorWrongState].
	arrayOfRoots elementsForwardIdentityTo:
		(arrayOfRoots collect: [:r | r rootStubInImageSegment: self]).
	state := #active.
	newRoots := self loadSegmentFrom: segment outPointers: outPointers.
	state := #inactive.
	arrayOfRoots elementsForwardIdentityTo: newRoots.
! !

!ImageSegment methodsFor: 'testing' stamp: 'tk 9/3/1999 14:11'!
findInOut: anArray
	"Take an array of references to a morph, and try to classify them:  in the segment, in outPointers, or other."

String streamContents: [:strm |
	anArray withIndexDo: [:obj :ind |
		strm cr; nextPutAll: obj printString; space.

		]].! !

!ImageSegment methodsFor: 'testing' stamp: 'ul 8/14/2015 18:02'!
findOwnerMap: morphs
	| st |
	"Construct a string that has a printout of the owner chain for every morph in the list.  Need it as a string so not hold onto them."

st := ''.
morphs do: [:mm |
	(st includesSubstring: mm printString) ifFalse: [
		st := st, '
', mm allOwners printString]].
Smalltalk at: #Owners put: st.
! !

!ImageSegment methodsFor: 'testing' stamp: 'tk 9/7/1999 09:48'!
findOwnersOutPtrs

| ow ff |
ow := Smalltalk at: #Owners ifAbsent: [^ self].
ow ifNil: [^ self].
outPointers do: [:oo | 
	oo isMorph ifTrue: [
		ow := ow copyReplaceAll: oo printString with: oo printString, '[<<<- Pointed at]']].
ff := FileStream fileNamed: 'Owners log'.
ff nextPutAll: ow; close.
Smalltalk at: #Owners put: ow.
ff edit.! !

!ImageSegment methodsFor: 'testing' stamp: 'ar 9/27/2005 18:02'!
findRogueRootsAllMorphs: rootArray 
	"This is a tool to track down unwanted pointers into the segment.  If we don't deal with these pointers, the segment turns out much smaller than it should.  These pointers keep a subtree of objects out of the segment.
1) assemble all objects should be in seg:  morph tree, presenter, scripts, player classes, metaclasses.  Put in a Set.
2) Remove the roots from this list.  Ask for senders of each.  Of the senders, forget the ones that are in the segment already.  Keep others.  The list is now all the 'incorrect' pointers into the segment."

	| inSeg testRoots scriptEditors pointIn wld xRoots |
	Smalltalk garbageCollect.
	inSeg := IdentitySet new: 200.
	arrayOfRoots := rootArray.
	(testRoots := self rootsIncludingPlayers) ifNil: [testRoots := rootArray].
	testRoots do: 
			[:obj | 
			(obj isKindOf: Project) 
				ifTrue: 
					[inSeg add: obj.
					wld := obj world.
					inSeg add: wld presenter].
			(obj isKindOf: Presenter) ifTrue: [inSeg add: obj]].
	xRoots := wld ifNil: [testRoots] ifNotNil: [testRoots , (Array with: wld)].
	xRoots do: 
			[:obj | 
			"root is a project"

			obj isMorph 
				ifTrue: 
					[obj allMorphs do: 
							[:mm | 
							inSeg add: mm.
							mm player ifNotNil: [inSeg add: mm player]].
					obj isWorldMorph ifTrue: [inSeg add: obj presenter]]].
	scriptEditors := IdentitySet new.
	inSeg do: 
			[:obj | 
			obj isPlayerLike 
				ifTrue: 
					[scriptEditors addAll: (obj class tileScriptNames 
								collect: [:nn | obj scriptEditorFor: nn])]].
	scriptEditors do: [:se | inSeg addAll: se allMorphs].
	testRoots do: [:each | inSeg remove: each ifAbsent: []].
	"want them to be pointed at from outside"
	pointIn := IdentitySet new: 400.
	inSeg do: [:ob | pointIn addAll: (Utilities pointersTo: ob except: inSeg)].
	testRoots do: [:each | pointIn remove: each ifAbsent: []].
	pointIn remove: inSeg array ifAbsent: [].
	pointIn remove: pointIn array ifAbsent: [].
	inSeg do: 
			[:obj | 
			obj isMorph 
				ifTrue: 
					[pointIn remove: (obj instVarAt: 3)
						ifAbsent: 
							["submorphs"

							].
					"associations in extension"
					pointIn remove: obj extension ifAbsent: [].
					obj extension ifNotNil: 
							[obj extension otherProperties ifNotNil: 
									[obj extension otherProperties associationsDo: 
											[:ass | 
											pointIn remove: ass ifAbsent: []
											"*** and extension actorState"
											"*** and ActorState instantiatedUserScriptsDictionary ScriptInstantiations"]]]].
			obj isPlayerLike 
				ifTrue: [obj class scripts values do: [:us | pointIn remove: us ifAbsent: []]]].
	"*** presenter playerlist"
	self halt: 'Examine local variables pointIn and inSeg'.
	^pointIn! !

!ImageSegment methodsFor: 'testing' stamp: 'ul 6/17/2011 12:20'!
findRogueRootsImSeg: rootArray
	"This is a tool to track down unwanted pointers into the segment.  If we don't deal with these pointers, the segment turns out much smaller than it should.  These pointers keep a subtree of objects out of the segment.
1) Break all owner pointers in submorphs and all scripts.
2) Create the segment and look at outPointers.
3) Remove those we expect.
4) Remember to quit without saving -- the owner pointers are smashed."

| newRoots suspects bag1 bag2 |
arrayOfRoots := rootArray.
[(newRoots := self rootsIncludingPlayers) == nil] whileFalse: [
	arrayOfRoots := newRoots].		"world, presenter, and all Player classes"
self findRogueRootsPrep.	"and free that context!!"
Smalltalk garbageCollect.
self copyFromRoots: arrayOfRoots sizeHint: 0.

suspects := outPointers select: [:oo | oo isMorph].
suspects size > 0 ifTrue: [suspects inspect].
bag1 := Bag new.  bag2 := Bag new.
outPointers do: [:key | 
	(key isKindOf: Class) 
		ifTrue: [bag2 add: key class name]
		ifFalse: [(#(Symbol Point Rectangle True False String Float Color Form ColorForm StrikeFont Metaclass UndefinedObject TranslucentColor) includes: key class name)
			ifTrue: [bag2 add: key class name]
			ifFalse: [bag1 add: key class name]]].
"(bag sortedCounts) is the SortedCollection"
(StringHolder new contents: bag1 sortedCounts printString, '

', bag2 sortedCounts printString) 
	openLabel: 'Objects pointed at by the outside'.
self halt: 'Examine local variables pointIn and inSeg'.

"Use this in inspectors:
	outPointers select: [:oo | oo class == <a Class>].		"

! !

!ImageSegment methodsFor: 'testing' stamp: 'nice 12/27/2009 03:11'!
findRogueRootsPrep
	"Part of the tool to track down unwanted pointers into the segment.  Break all owner pointers in submorphs, scripts, and viewers in flaps."

| wld players morphs |
wld := arrayOfRoots detect: [:obj | 
	obj isMorph ifTrue: [obj isWorldMorph] ifFalse: [false]] ifNone: [nil].
wld ifNil: [wld := arrayOfRoots detect: [:obj | obj isMorph] 
				ifNone: [^ self error: 'can''t find a root morph']].
morphs := IdentitySet new: 400.
wld allMorphsAndBookPagesInto: morphs.
players := wld presenter allExtantPlayers.	"just the cached list"
players do: [:pp | | scriptEditors |
	scriptEditors := pp class tileScriptNames collect: [:nn | 
			pp scriptEditorFor: nn].
	scriptEditors do: [:se | morphs addAll: se allMorphs]].
wld submorphs do: [:mm | 	"non showing flaps"
	(mm isKindOf: FlapTab) ifTrue: [
		mm referent allMorphsAndBookPagesInto: morphs]].
morphs do: [:mm | 	"break the back pointers"
	mm isInMemory ifTrue: [
	(mm respondsTo: #target) ifTrue: [
		mm nearestOwnerThat: [:ow | ow == mm target 
			ifTrue: [mm target: nil. true]
			ifFalse: [false]]].
	(mm respondsTo: #arguments) ifTrue: [
		mm arguments do: [:arg | arg ifNotNil: [
			mm nearestOwnerThat: [:ow | ow == arg
				ifTrue: [mm arguments at: (mm arguments indexOf: arg) put: nil. true]
				ifFalse: [false]]]]].
	mm eventHandler ifNotNil: ["recipients point back up"
		(morphs includesAllOf: (mm eventHandler allRecipients)) ifTrue: [
			mm eventHandler: nil]].
	"temporary, until using Model for PartsBin"
	(mm isMorphicModel) ifTrue: [
		(mm model isMorphicModel) ifTrue: [
			mm model breakDependents]].
	(mm isTextMorph) ifTrue: [mm setContainer: nil]]].
(Smalltalk includesKey: #Owners) ifTrue: [Smalltalk at: #Owners put: nil].
	"in case findOwnerMap: is commented out"
"self findOwnerMap: morphs."
morphs do: [:mm | 	"break the back pointers"
	mm isInMemory ifTrue: [mm privateOwner: nil]].
"more in extensions?"

! !

!ImageSegment methodsFor: 'testing' stamp: 'nice 12/27/2009 03:11'!
findRogueRootsRefStrm: rootArray 
	"This is a tool to track down unwanted pointers into the segment.  If we don't deal with these pointers, the segment turns out much smaller than it should.  These pointers keep a subtree of objects out of the segment.
1) assemble all objects that should be in the segment by using SmartReference Stream and a dummyReference Stream.  Put in a Set.
2) Remove the roots from this list.  Ask for senders of each.  Of the senders, forget the ones that are in the segment already.  Keep others.  The list is now all the 'incorrect' pointers into the segment."

	| goodInSeg inSeg pointIn dummy |
	dummy := ReferenceStream on: (DummyStream on: nil).
	"Write to a fake Stream, not a file"
	rootArray do: 
			[:root | 
			dummy rootObject: root.	"inform him about the root"
			dummy nextPut: root].
	inSeg := dummy references keys asSet.
	dummy := nil.
	Smalltalk garbageCollect.	"dump refs dictionary"
	rootArray do: [:each | inSeg remove: each ifAbsent: []].
	"want them to be pointed at from outside"
	pointIn := IdentitySet new: 500.
	goodInSeg := IdentitySet new: 2000.
	inSeg do: 
			[:obj | | ok | 
			ok := obj class isPointers.
			obj class == Color ifTrue: [ok := false].
			obj class == TranslucentColor ifTrue: [ok := false].
			obj class == Array ifTrue: [obj size = 0 ifTrue: [ok := false]].
			"shared #() in submorphs of all Morphs"
			ok ifTrue: [goodInSeg add: obj]].
	goodInSeg 
		do: [:ob | pointIn addAll: (Utilities pointersTo: ob except: #())].
	inSeg do: [:each | pointIn remove: each ifAbsent: []].
	rootArray do: [:each | pointIn remove: each ifAbsent: []].
	pointIn remove: inSeg array ifAbsent: [].
	pointIn remove: goodInSeg array ifAbsent: [].
	pointIn remove: pointIn array ifAbsent: [].
	self halt: 'Examine local variables pointIn and inSeg'.
	^pointIn! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'ar 1/3/2010 19:08'!
findStacks
	"Return an array of all the StackMorphs in this project."
	| guys stacks |
	guys := IdentitySet new.
	Smalltalk at: #StackMorph ifPresent:[:aClass|
		guys addAll: aClass withAllSubclasses.
	].
	stacks := OrderedCollection new.
	arrayOfRoots do: [:obj |
		(guys includes: obj class) ifTrue: [stacks add: obj]].
	^ stacks! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'ar 4/12/2005 17:37'!
fixCapitalizationOfSymbols
	"MultiString>>capitalized was not implemented 
correctly. 
	Fix eventual accessors and mutators here."
	| sym ms |
	1 to: outPointers size do:[:i|
		sym := outPointers at: i.
		(sym class == WideSymbol and:[sym size > 3]) ifTrue:[
			((sym beginsWith: 'get')
				and:[(sym at: 4) asInteger < 256
				and:[(sym at: 4) isLowercase]]) ifTrue:[
					ms := sym asString.
					ms at: 4 put: (ms at: 4) asUppercase.
					ms := ms asSymbol.
					sym becomeForward: ms.
			].
			((sym beginsWith: 'set')
				and:[(sym at: 4) asInteger < 256
				and:[(sym at: 4) isLowercase
				and:[sym last = $:
				and:[(sym occurrencesOf: $:) = 1]]]]) ifTrue:[
					ms := sym asString.
					ms at: 4 put: (ms at: 4) asUppercase.
					ms := ms asSymbol.
					sym becomeForward: ms.
				].
			outPointers at: i put: sym.
		].
	].! !

!ImageSegment methodsFor: 'instance change shape' stamp: 'ul 8/2/2011 20:02'!
ifOutPointer: anObject thenAllObjectsDo: aBlock
	| withSymbols segSize |
	"If I point out to anObject, bring me in, Submit all my objects to the block.  Write me out again."

	(state = #onFile or: [state = #onFileWithSymbols]) ifFalse: [^ self].
	withSymbols := state = #onFileWithSymbols.
	(outPointers includes: anObject) ifFalse: [^ self].
	state = #onFile ifTrue: [Cursor read showWhile: [self readFromFile]].
	segSize := segment size.
	self install.
	self allObjectsDo: aBlock.	"do the work"
	self copyFromRoots: arrayOfRoots sizeHint: segSize.
	self extract.
	withSymbols 
		ifTrue: [self writeToFileWithSymbols]
		ifFalse: [self writeToFile].

! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'eem 8/21/2015 19:10'!
install
	"This operation retrieves the segment if necessary from file storage, installs it in memory, and replaces (using become:) all the root stubs with the reconstructed roots of the segment."

	| newRoots |
	state = #onFile ifTrue: [self readFromFile].
	state = #onFileWithSymbols ifTrue:
		[self readFromFileWithSymbols].
	(state = #active) | (state = #imported) ifFalse: [self errorWrongState].
	newRoots := self loadSegmentFrom: segment outPointers: outPointers.
	self checkAndReportLoadError.
	state = #imported 
		ifTrue: ["just came in from exported file"
			arrayOfRoots := newRoots]
		ifFalse: [
			arrayOfRoots elementsForwardIdentityTo: newRoots].
	state := #inactive.
	Beeper beepPrimitive! !

!ImageSegment methodsFor: 'testing' stamp: 'di 9/29/1999 16:50'!
isOnFile
	^ state == #onFile! !

!ImageSegment methodsFor: 'primitives' stamp: 'eem 8/21/2015 18:56'!
loadSegmentFrom: segmentWordArray outPointers: outPointerArray
	"This primitive will install a binary image segment and return as its value the array
	 of roots of the tree of objects represented.  Upon successful completion, the
	 wordArray will have been becomed into anArray of the loaded objects.  If this
	 primitive should fail, it will have destroyed the contents of the segment wordArray."

	<primitive: 99 error: ec>	"successful completion returns the array of roots"
	^nil								"failure returns nil"! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tk 12/17/1999 00:03'!
localName
	| segs ind sep |
	"Return the current file name for this segment, a local name in the segments directory."

	fileName ifNil: [^ nil].
	"^ fileName"	

	"The following is for backward compatibility.  Remove this part after June 2000.
	Check if the fileName is a full path, and make it local.  Regardless of current or previous file system delimiter."

	segs := self class folder copyLast: 4.  ":=segs"
	ind := 1.
	[ind := fileName findString: segs startingAt: ind+1 caseSensitive: false.
		ind = 0 ifTrue: [^ fileName].
		sep := fileName at: ind + (segs size).
		sep isAlphaNumeric ] whileTrue.		"sep is letter or digit, not a separator"

	^ fileName := fileName copyFrom: ind+(segs size)+1 "delimiter" to: fileName size! !

!ImageSegment methodsFor: 'compact classes' stamp: 'tk 3/15/2000 09:51'!
objectAfter: ind
	"Return the object or free chunk immediately following the given object or free chunk in the segment.  *** Warning: When class ObjectMemory change, be sure to change it here. ***"

	| sz word newInd hdrBits |
	sz := ((word := segment at: ind "header") bitAnd: 3) = 2   "free block?"
		ifTrue: [word bitAnd: 16rFFFFFFFC]
		ifFalse: [(word bitAnd: 3) = 0 "HeaderTypeSizeAndClass"
			ifTrue: [(segment at: ind-2) bitAnd: 16rFFFFFFFC]
			ifFalse: [word bitAnd: "SizeMask" 252]].

	newInd := ind + (sz>>2).
	"adjust past extra header words"
	(hdrBits := (segment atPin: newInd) bitAnd: 3) = 3 ifTrue: [^ newInd].
		"If at end, header word will be garbage.  This is OK"
	hdrBits = 1 ifTrue: [^ newInd+1].
	hdrBits = 0 ifTrue: [^ newInd+2].
	^ newInd	"free"! !

!ImageSegment methodsFor: 'access' stamp: 'tk 1/20/2000 20:16'!
originalRoots
	"Return only the roots that the user submitted, not the ones we had to add."

	userRootCnt ifNil: [^ arrayOfRoots].
	^ arrayOfRoots copyFrom: 1 to: userRootCnt! !

!ImageSegment methodsFor: 'access' stamp: 'tk 3/31/1999 21:47'!
outPointers
	^ outPointers! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'cwp 1/18/2014 10:05'!
prepareToBeSaved
	"Prepare objects in outPointers to be written on the disk.  They must be able to match up with existing objects in their new system.  outPointers is already a copy.
	Classes are already converted to a DiskProxy.  
	Associations in outPointers:
1) in Smalltalk.
2) in a classPool.
3) in a shared pool.
4) A pool dict pointed at directly"

| left myClasses outIndexes |
self flag: #environments.
myClasses := Set new.
arrayOfRoots do: [:aRoot | aRoot class class == Metaclass ifTrue: [myClasses add: aRoot]].
outIndexes := IdentityDictionary new.
outPointers withIndexDo: [:anOut :ind | | key | 
	anOut isVariableBinding ifTrue: [
		(myClasses includes: anOut value)
			ifFalse: [outIndexes at: anOut put: ind]
			ifTrue: [(Smalltalk globals associationAt: anOut key ifAbsent: [3]) == anOut 
				ifTrue: [outPointers at: ind put: 
					(DiskProxy global: #Smalltalk selector: #associationDeclareAt: 
						args: (Array with: anOut key))]
				ifFalse: [outIndexes at: anOut put: ind]
				]].
	(anOut isKindOf: Dictionary) ifTrue: ["Pools pointed at directly"
		(key := Smalltalk globals keyAtIdentityValue: anOut ifAbsent: [nil]) ifNotNil: [
			outPointers at: ind put: 
				(DiskProxy global: key selector: #yourself args: #())]].
	anOut isMorph ifTrue: [outPointers at: ind put: 
		(StringMorph contents: anOut printString, ' that was not counted')]
	].
left := outIndexes keys asSet.
left size > 0 ifTrue: ["Globals"
	(left copy) do: [:assoc |	"stay stable while delete items"
		(Smalltalk globals associationAt: assoc key ifAbsent: [3]) == assoc ifTrue: [
			outPointers at: (outIndexes at: assoc) put: 
				(DiskProxy global: #Smalltalk selector: #associationAt: 
					args: (Array with: assoc key)).
			left remove: assoc]]].
left size > 0 ifTrue: ["Class variables"
	Smalltalk allClassesDo: [:cls | cls classPool size > 0 ifTrue: [
		(left copy) do: [:assoc |	"stay stable while delete items"
			(cls classPool associationAt: assoc key ifAbsent: [3]) == assoc ifTrue: [
				outPointers at: (outIndexes at: assoc) put: 
					(DiskProxy new global: cls name
						preSelector: #classPool
						selector: #associationAt: 
						args: (Array with: assoc key)).
				left remove: assoc]]]]].
left size > 0 ifTrue: ["Pool variables"
	Smalltalk globals associationsDo: [:poolAssoc | | pool |
		poolAssoc value class == Dictionary ifTrue: ["a pool"
			pool := poolAssoc value.
			(left copy) do: [:assoc |	"stay stable while delete items"
				(pool associationAt: assoc key ifAbsent: [3]) == assoc ifTrue: [
					outPointers at: (outIndexes at: assoc) put: 
						(DiskProxy global: poolAssoc key selector: #associationAt: 
							args: (Array with: assoc key)).
					left remove: assoc]]]]].
left size > 0 ifTrue: [
	"If points to class in arrayOfRoots, must deal with it separately"
	"OK to have obsolete associations that just get moved to the new system"
	self inform: 'extra associations'.
	left inspect].
! !

!ImageSegment methodsFor: 'statistics' stamp: 'ar 2/21/2001 19:22'!
printSpaceAnalysisOn: aStream
	"Capture statistics about the IS and print the number of instances per class and space usage"
	| instCount instSpace sorted sum1 sum2 |
	instCount := self doSpaceAnalysis.
	instSpace := instCount last.
	instCount := instCount first.
	sorted := SortedCollection sortBlock:[:a1 :a2| a1 value >= a2 value].
	instSpace associationsDo:[:a| sorted add: a].
	sorted do:[:assoc|
		aStream cr; nextPutAll: assoc key; tab.
		aStream print: (instCount at: assoc key); nextPutAll:' instances '.
		aStream print: assoc value; nextPutAll: ' bytes '.
	].
	sum1 := instCount inject: 0 into:[:sum :n| sum + n].
	sum2 := instSpace inject: 0 into:[:sum :n| sum + n].
	aStream cr; cr.
	aStream print: sum1; nextPutAll:' instances '.
	aStream print: sum2; nextPutAll: ' bytes '.
! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'eem 8/21/2015 18:35'!
readFromFile
	"Read in a simple segment.  Use folder of this image, even if remembered as previous location of this image"

	| ff realName |
	realName := self class folder, FileDirectory slash, self localName.
	ff := FileStream readOnlyFileNamed: realName.
	segment := ff nextWordsInto: (WordArrayForSegment new: ff size//4).
	ff close.
	state := #active! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'eem 8/21/2015 18:46'!
rehashSets
	"I have just been brought in and converted to live objects.  Find all Sets and Dictionaries in the newly created objects and rehash them.  Segment is near then end of memory, since is was newly brought in (and a new object created for it).
	Also, collect all classes of receivers of blocks.  Return them.  Caller will check if they have been reshaped."

	| hashedCollections receiverClasses |
	hashedCollections := OrderedCollection new.
		"have to collect them, because Dictionary makes a copy, and that winds up at the end of memory and gets rehashed and makes another one."
	receiverClasses := IdentitySet new.
	self allObjectsDo:
		[:object|
		object isInMemory ifTrue:
			[(object isCollection and: [object isKindOf: HashedCollection]) ifTrue:
				[hashedCollections add: object].
			(object isBlock or: [object isContext]) ifTrue:
				[receiverClasses add: object receiver class]]].
	hashedCollections do: [ :each | each compact ]. "our purpose"
	^receiverClasses	"our secondary job"! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'eem 8/21/2015 11:31'!
reshapeClasses: mapFakeClassesToReal refStream: smartRefStream

	| bads allVarMaps partials in out perfect |

	self flag: #bobconv.	

	partials := OrderedCollection new.
	bads := OrderedCollection new.
	allVarMaps := IdentityDictionary new.
	mapFakeClassesToReal keysAndValuesDo: [ :aFakeClass :theRealClass |
		aFakeClass allInstances do: [ :misShapen |
			perfect := smartRefStream convert1: misShapen to: theRealClass allVarMaps: allVarMaps.
			bads
				detect: [ :x | x == misShapen]
				ifNone: [
					bads add: misShapen.
					partials add: perfect
				].
		].
	].
	bads isEmpty ifFalse: [
		bads asArray elementsForwardIdentityTo: partials asArray
	].

	in := OrderedCollection new.
	out := OrderedCollection new.
	partials do: [ :each |
		perfect := smartRefStream convert2: each allVarMaps: allVarMaps.
		in
			detect: [ :x | x == each]
			ifNone: [
				in add: each.
				out add: perfect
			]
	].
	in isEmpty ifFalse: [
		in asArray elementsForwardIdentityTo: out asArray
	].
! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'nice 11/13/2015 00:35'!
restoreEndianness
	^self restoreEndianness: self endianness ~~ Smalltalk endianness! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'nice 11/13/2015 00:22'!
restoreEndianness: endiannessHasToBeFixed
	"If endiannessHasToBeFixed, then fix endianness (byte order) of any objects not already fixed.  Do this by discovering classes that need a startUp message sent to each instance, and sending it..
	I have just been brought in and converted to live objects.  Find all Sets and Dictionaries in the newly created objects and rehash them.  Segment is near then end of memory, since is was newly brought in (and a new object created for it).
	Also, collect all classes of receivers of blocks which refer to instance variables.  Return them.  Caller will check if they have been reshaped."

	| hashedCollections receiverClasses noStartUpNeeded startUps |

	hashedCollections := OrderedCollection new.
	receiverClasses := IdentitySet new.
	noStartUpNeeded := IdentitySet new.	"classes that don't have a per-instance startUp message"
	startUps := IdentityDictionary new.	"class -> MessageSend of a startUp message"
	self allObjectsDo:
		[:object| | cls msg |
		object isInMemory ifTrue:
			[(object isCollection and: [object isKindOf: HashedCollection]) ifTrue:
				[hashedCollections add: object].
			(object isContext and: [object hasInstVarRef]) ifTrue:
				[receiverClasses add: object receiver class]].
			(noStartUpNeeded includes: object class) ifFalse:
				[cls := object class.
				 (msg := startUps at: cls ifAbsent: nil) ifNil:
					[msg := cls startUpFrom: endiannessHasToBeFixed.	"a Message, if we need to swap bytes this time"
					 msg ifNil: [noStartUpNeeded add: cls]
						ifNotNil: [startUps at: cls put: msg]].
				 msg ifNotNil: [msg sentTo: object]]].
	hashedCollections do: [ :each | each compact ]. "our purpose"
	^ receiverClasses	"our secondary job"! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'eem 8/21/2015 18:47'!
revert
	"Pretend this segment was never brought in.  Check that it has a fileName.  Replace (using become:) all the original roots of a segment with segmentRootStubs.  Thus the original objects will be reclaimed, and the root stubs will remain to bring the segment back in if it is needed.
	How to use revert:  In the project, choose 'save for reverting'.

	ReEnter the project.  Make changes.
	Either exit normally, and change will be kept, or
		Choose 'Revert to saved version'."

	fileName ifNil: [^ self].
	(state = #inactive) | (state = #onFile) ifFalse: [^ self].
	Cursor write showWhile: [
		arrayOfRoots elementsForwardIdentityTo:
			(arrayOfRoots collect: [:r | r rootStubInImageSegment: self]).
		state := #onFile.
		segment := nil]

"Old version:
	How to use revert:  In the project, execute 
(Project current projectParameters at: #frozen put: true)
	Leave the project.  Check that the project went out to disk (it is gray in the Jump to Project list).
	ReEnter the project.  Hear a plink as it comes in from disk.  Make a change.
	Exit the project.  Choose 'Revert to previous version' in the dialog box.
	Check that the project went out to disk (it is gray in the Jump to Project list).
	ReEnter the project and see that it is in the original state."! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'nice 1/15/2010 22:47'!
rootsIncludingBlockMethods
	"Return a new roots array with more objects.  (Caller should store into rootArray.) Any CompiledMethods that create blocks will be in outPointers if the block is held outside of this segment.  Put such methods into the roots list.  Then ask for the segment again."

	| myClasses extras |
	userRootCnt ifNil: [userRootCnt := arrayOfRoots size].
	extras := OrderedCollection new.
	myClasses := OrderedCollection new.
	arrayOfRoots do: [:aRoot | aRoot class class == Metaclass ifTrue: [
					myClasses add: aRoot]].
	myClasses isEmpty ifTrue: [^ nil].	"no change"
	outPointers do: [:anOut | | gotIt | 
		anOut class == CompiledMethod ifTrue: [
			"specialized version of who"
			gotIt := false.
			myClasses detect: [:class |
				class methodsDo: [:m |
					m == anOut 
						ifTrue: [extras add: anOut.  gotIt := true]].
				gotIt] 
				ifNone: []
			].
	].
	extras := extras select: [:ea | (arrayOfRoots includes: ea) not].
	extras isEmpty ifTrue: [^ nil].	"no change"
	^ arrayOfRoots, extras! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'nice 12/28/2009 16:08'!
rootsIncludingBlocks
	"For export segments only.  Return a new roots array with more objects.  (Caller should store into rootArray.)  Collect Blocks and external methods pointed to by them.  Put them into the roots list.  Then ask for the segment again."

	| extras have |
	userRootCnt ifNil: [userRootCnt := arrayOfRoots size].
	extras := OrderedCollection new.
	outPointers do: [:anOut | 
		anOut class == CompiledMethod ifTrue: [extras add: anOut].
		(anOut isBlock) ifTrue: [extras add: anOut].
		(anOut class == MethodContext) ifTrue: [extras add: anOut]].

	[have := extras size.
	 extras copy do: [:anOut |
		anOut isBlock ifTrue: [
			anOut home ifNotNil: [
				(extras includes: anOut home) ifFalse: [extras add: anOut home]]].
		(anOut class == MethodContext) ifTrue: [
			anOut method ifNotNil: [
				(extras includes: anOut method) ifFalse: [extras add: anOut method]]]].
	 have = extras size] whileFalse.
	extras := extras select: [:ea | (arrayOfRoots includes: ea) not].
	extras isEmpty ifTrue: [^ nil].	"no change"

	^ arrayOfRoots, extras! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'md 3/16/2006 22:44'!
rootsIncludingPlayers
	"Return a new roots array with more objects.  (Caller should store into rootArray.) Player (non-systemDefined) gets its class and metaclass put into the Roots array.  Then ask for the segment again."

| extras havePresenter players morphs existing |
userRootCnt ifNil: [userRootCnt := arrayOfRoots size].
extras := OrderedCollection new.
arrayOfRoots do: [:root | 
	(root isKindOf: Presenter) ifTrue: [havePresenter := root].
	(root isKindOf: PasteUpMorph) ifTrue: [
			root isWorldMorph ifTrue: [havePresenter := root presenter]].
	(root isKindOf: Project) ifTrue: [havePresenter := root world presenter]].
havePresenter ifNotNil: [
	havePresenter flushPlayerListCache.		"old and outside guys"
	morphs := IdentitySet new: 400.
	havePresenter associatedMorph allMorphsAndBookPagesInto: morphs.
	players := (morphs select: [:m | m player ~~ nil] 
				thenCollect: [:m | m player]) asArray.
	players := players select: [:ap | (arrayOfRoots includes: ap class) not
		& (ap class isSystemDefined not)].
	extras addAll: (players collect: [:each | each class]).
	extras addAll: (players collect: [:each | each class class]).
	extras addAll: morphs.	"Make then ALL roots!!"
	].
existing := arrayOfRoots asIdentitySet.
extras := extras reject: [ :each | existing includes: each].
extras isEmpty ifTrue: [^ nil].	"no change"
	
		havePresenter := players := morphs := nil.
		^ arrayOfRoots, extras	"will contain multiples of some, but reduced later"
! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tk 12/8/2000 10:00'!
savePlayerReferences: dictOfAllObjects
	| save world |
	"Save our associations we own in the shared References table.  They will be installed when the segment is imported."

	save := OrderedCollection new.
	References associationsDo: [:assoc |
		(dictOfAllObjects includesKey: assoc) ifTrue: [save add: assoc]].
	1 to: 5 do: [:ii | ((arrayOfRoots at: ii) respondsTo: #isCurrentProject) ifTrue: [
					world := (arrayOfRoots at: ii) world]].
	world setProperty: #References toValue: save.
		"assume it is not refed from outside and will be traced"! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'RAA 6/22/2000 17:49'!
scanFrom: aStream
	"Move source code from a fileIn to the changes file for classes in an ImageSegment.  Do not compile the methods.  They already came in via the image segment.  After the ImageSegment in the file, !!ImageSegment new!! captures control, and scanFrom: is called."
	| val chunk |

	[aStream atEnd] whileFalse: 
		[aStream skipSeparators.
		val := (aStream peekFor: $!!)
			ifTrue: ["Move (aStream nextChunk), find the method or class 
						comment, and install the file location bytes"
					(Compiler evaluate: aStream nextChunk logged: false)
						scanFromNoCompile: aStream forSegment: self]
			ifFalse: [chunk := aStream nextChunk.
					aStream checkForPreamble: chunk.
					Compiler evaluate: chunk logged: true].
		aStream skipStyleChunk].
	"regular fileIn will close the file"
	^ val! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'cwp 6/20/2012 17:23'!
scanFrom: aStream environment: anEnvironment
	^ self scanFrom: aStream! !

!ImageSegment methodsFor: 'access' stamp: 'tk 8/18/1999 22:19'!
segment
	^ segment! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'di 3/28/1999 13:48'!
segmentCopy
	"This operation will install a copy of the segment in memory, and return a copy of the array of roots.  The effect is to perform a deep copy of the original structure.  Note that installation destroys the segment, so it must be copied before doing the operation."

	state = #activeCopy ifFalse: [self errorWrongState].
	^ self loadSegmentFrom: segment copy outPointers: outPointers! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tk 10/8/1999 12:39'!
segmentName
	"Return the local file name for this segment."

	^ segmentName! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tk 12/13/1999 09:19'!
segmentName: aString
	"Local file name for this segment."

	segmentName := aString! !

!ImageSegment methodsFor: 'instance change shape' stamp: 'ar 4/10/2005 22:19'!
segUpdateInstancesOf: oldClass toBe: newClass isMeta: isMeta
	| withSymbols oldInstances segSize |
	"Bring me in, locate instances of oldClass and get them converted.  Write me out again."

	(state = #onFile or: [state = #onFileWithSymbols]) ifFalse: [^ self].
	withSymbols := state = #onFileWithSymbols.
	"If has instances, they point out at the class"
	(outPointers includes: oldClass) ifFalse: [
		oldClass == SmallInteger ifTrue: [^ self].	"instance not changable"
		oldClass == Symbol ifTrue: [^ self].	"instance is never in a segment"
		oldClass == ByteSymbol ifTrue: [^ self].	"instance is never in a segment"
		(Smalltalk compactClassesArray includes: oldClass) ifFalse: [^ self]].
		"For a compact class, must search the segment.  Instance does not 
		 point outward to class"
	state = #onFile ifTrue: [Cursor read showWhile: [self readFromFile]].
	segSize := segment size.
	self install.
	oldInstances := OrderedCollection new.
	self allObjectsDo: [:obj | obj class == oldClass ifTrue: [
		oldInstances add: obj]].
	newClass updateInstances: oldInstances asArray from: oldClass isMeta: isMeta.
	self copyFromRoots: arrayOfRoots sizeHint: segSize.
	self extract.
	withSymbols 
		ifTrue: [self writeToFileWithSymbols]
		ifFalse: [self writeToFile].
! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tfel 8/2/2016 10:49'!
smartFillRoots: dummy
	| refs known ours ww blockers |
	"Put all traced objects into my arrayOfRoots.  Remove some
that want to be in outPointers.  Return blockers, an
IdentityDictionary of objects to replace in outPointers."

	blockers _ dummy blockers.
	known _ (refs _ dummy references) size.
	refs keys do: [:obj | "copy keys to be OK with removing items"
		(obj isSymbol) ifTrue: [refs removeKey: obj.  known _ known-1].
		(obj class == PasteUpMorph) ifTrue: [
			obj isWorldMorph & (obj owner == nil) ifTrue: [
				(dummy project ~~ nil and: [obj == dummy project world]) ifFalse: [
					refs removeKey: obj.  known _ known-1.
					blockers at: obj put:
						(StringMorph contents: 'The worldMorph of a different world')]]].
					"Make a ProjectViewMorph here"
		"obj class == Project ifTrue: [Transcript show: obj; cr]."
		(blockers includesKey: obj) ifTrue: [
			refs removeKey: obj ifAbsent: [known _ known+1].  known _ known-1].
		].
	ours _ dummy project ifNotNil: [dummy project world] ifNil: [ActiveWorld].
	refs keysDo: [:obj |
			obj isMorph ifTrue: [
				ww _ obj world.
				(ww == ours) | (ww == nil) ifFalse: [
					refs removeKey: obj.  known _ known-1.
					blockers at: obj put: (StringMorph contents:
								obj printString, ' from another world')]]].
	"keep original roots on the front of the list"
	(dummy rootObject) do: [:rr | refs removeKey: rr ifAbsent: []].
	self classOrganizersBeRoots: dummy.
	^ dummy rootObject, refs fasterKeys asArray.! !

!ImageSegment methodsFor: 'access' stamp: 'tk 9/26/1999 22:54'!
state
	^ state! !

!ImageSegment methodsFor: 'fileIn/Out' stamp: 'eem 7/21/2008 12:09'!
storeDataOn: aDataStream
	"Don't wrote the array of Roots.  Also remember the structures of the classes of objects inside the segment."

	| tempRoots tempOutP list |
	state = #activeCopy ifFalse: [self error: 'wrong state'].
		"real state is activeCopy, but we changed it will be right when coming in"
	tempRoots := arrayOfRoots.
	tempOutP := outPointers.
	outPointers := outPointers clone.
	self prepareToBeSaved.
	arrayOfRoots := nil.
	state := #imported.
	super storeDataOn: aDataStream.		"record my inst vars"
	arrayOfRoots := tempRoots.
	outPointers := tempOutP.
	state := #activeCopy.
	aDataStream references at: #AnImageSegment put: false.	"the false is meaningless"
		"This key in refs is the flag that there is an ImageSegment in this file."

	"Find the receivers of blocks in the segment.  Need to get the structure of their classes into structures.  Put the receivers into references."
	(aDataStream byteStream isKindOf: DummyStream) ifTrue: [
		list := Set new.
		arrayOfRoots do: [:ea | 
			ea isBlock | (ea class == MethodContext) ifTrue: [ 
				list add: ea receiver class ]].
		aDataStream references at: #BlockReceiverClasses put: list].
! !

!ImageSegment methodsFor: 'primitives' stamp: 'di 3/28/1999 13:49'!
storeSegmentFor: rootsArray into: segmentWordArray outPointers: outPointerArray
	"This primitive will store a binary image segment (in the same format as the Squeak image file) of the receiver and every object in its proper tree of subParts (ie, that is not refered to from anywhere else outside the tree).  Note: all elements of the reciever are treated as roots indetermining the extent of the tree.  All pointers from within the tree to objects outside the tree will be copied into the array of outpointers.  In their place in the image segment will be an oop equal to the offset in the outpointer array (the first would be 4). but with the high bit set."

	"The primitive expects the array and wordArray to be more than adequately long.  In this case it returns normally, and truncates the two arrays to exactly the right size.  If either array is too small, the primitive will fail, but in no other case."

	<primitive: 98>	"successful completion returns self"
	^ nil			"failure returns nil"! !

!ImageSegment methodsFor: 'testing' stamp: 'di 3/27/1999 21:36'!
verifyCopy

	| copyOfRoots matchDict |
	copyOfRoots := self segmentCopy.
	matchDict := IdentityDictionary new.
	arrayOfRoots with: copyOfRoots do:
		[:r :c | self verify: r matches: c knowing: matchDict]! !

!ImageSegment methodsFor: 'testing' stamp: 'eem 8/21/2015 11:28'!
verify: ob1 matches: ob2 knowing: matchDict

	| priorMatch first |
	ob1 == ob2 ifTrue:
		["If two pointers are same, they must be immediates or in outPointers"
		(ob1 class isImmediateClass and: [ob1 = ob2]) ifTrue: [^self].
		(outPointers includes: ob1) ifTrue: [^ self].
		self halt].
	priorMatch := matchDict at: ob1 ifAbsent: [nil].
	priorMatch == nil
		ifTrue: [matchDict at: ob1 put: ob2]
		ifFalse: [priorMatch == ob2
					ifTrue: [^ self]
					ifFalse: [self halt]].
	self verify: ob1 class matches: ob2 class knowing: matchDict.
	ob1 class isVariable ifTrue: 
		[ob1 basicSize = ob2 basicSize ifFalse: [self halt].
		first := 1.
		ob1 isCompiledMethod ifTrue: [first := ob1 initialPC].
		first to: ob1 basicSize do:
			[:i | self verify: (ob1 basicAt: i) matches: (ob2 basicAt: i) knowing: matchDict]].
	ob1 class instSize = ob2 class instSize ifFalse: [self halt].
	1 to: ob1 class instSize do:
		[:i | self verify: (ob1 instVarAt: i) matches: (ob2 instVarAt: i) knowing: matchDict].
	ob1 isCompiledMethod ifTrue:
		[ob1 header = ob2 header ifFalse: [self halt].
		ob1 numLiterals = ob2 numLiterals ifFalse: [self halt].
		1 to: ob1 numLiterals do:
			[:i | self verify: (ob1 literalAt: i) matches: (ob2 literalAt: i) knowing: matchDict]]! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'eem 8/21/2015 18:52'!
writeForExportWithSourcesGZ: fName inDirectory: aDirectory
	"Write the segment on the disk with all info needed to reconstruct it in a new image.  For export.  Out pointers are encoded as normal objects on the disk.  Append the source code of any classes in roots.  Target system will quickly transfer the sources to its changes file."

	"this is the gzipped version which I have temporarily suspended until I can get resolve the problem with forward references tring to reposition the stream - RAA 11 june 2000"




	| fileStream allClassesInRoots classesToWriteEntirely methodsWithSource |
	state = #activeCopy ifFalse: [self error: 'wrong state'].
	(fName includes: $.) ifFalse: [
		^ self inform: 'Please use ''.pr'' or ''.extSeg'' at the end of the file name'.].
	fileStream := GZipSurrogateStream newFileNamed: fName inDirectory: aDirectory.
	fileStream fileOutClass: nil andObject: self.
		"remember extra structures.  Note class names."

	"append sources"
	allClassesInRoots := arrayOfRoots select: [:cls | cls isBehavior].
	classesToWriteEntirely := allClassesInRoots select: [ :cls | cls theNonMetaClass isSystemDefined].
	methodsWithSource := OrderedCollection new.
	allClassesInRoots do: [ :cls |
		(classesToWriteEntirely includes: cls) ifFalse: [
			cls selectorsAndMethodsDo: [ :sel :meth |
				meth sourcePointer = 0 ifFalse: [methodsWithSource add: {cls. sel. meth}].
			].
		].
	].
	(classesToWriteEntirely isEmpty and: [methodsWithSource isEmpty]) ifTrue: [
		fileStream reallyClose.	"since #close is ignored"
		^ self
	].
	"fileStream reopen; setToEnd."	"<--not required with gzipped surrogate stream"
	fileStream nextPutAll: '\\!!ImageSegment new!!\\' withCRs.
	methodsWithSource do: [ :each |
		fileStream nextPut: $!!.	"try to pacify ImageSegment>>scanFrom:"
		fileStream nextChunkPut: 'RenamedClassSourceReader formerClassName: ',
				each first name printString,' methodsFor: ',
				(each first organization categoryOfElement: each second) asString printString,
				' stamp: ',(each third timeStamp) printString; cr.
		fileStream nextChunkPut: (each third getSourceFor: each second in: each first) asString.
		fileStream nextChunkPut: ' '; cr.
	].
	classesToWriteEntirely do: [:cls | 
		cls isMeta ifFalse: [fileStream nextPutAll: 
						(cls name, ' category: ''', cls category, '''.!!'); cr; cr].
		cls organization
			putCommentOnFile: fileStream
			numbered: 0
			moveSource: false
			forClass: cls.	"does nothing if metaclass"
		cls organization categories do: 
			[:heading |
			cls fileOutCategory: heading
				on: fileStream
				moveSource: false
				toFile: 0]].
	"no class initialization -- it came in as a real object"
	fileStream reallyClose	"since #close is ignored"
! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'eem 8/21/2015 18:50'!
writeForExportWithSources: fName inDirectory: aDirectory
	"Write the segment on the disk with all info needed to reconstruct it in a new image.  For export.  Out pointers are encoded as normal objects on the disk.  Append the source code of any classes in roots.  Target system will quickly transfer the sources to its changes file."

	"this is the old version which I restored until I solve the gzip problem"

	| fileStream tempFileName zipper allClassesInRoots classesToWriteEntirely methodsWithSource |
	state = #activeCopy ifFalse: [self error: 'wrong state'].
	(fName includes: $.) ifFalse: [
		^ self inform: 'Please use ''.pr'' or ''.extSeg'' at the end of the file name'.].
	tempFileName := aDirectory nextNameFor: 'SqProject' extension: 'temp'.
	zipper := [
		ProgressNotification signal: '3:uncompressedSaveComplete'.
		(aDirectory oldFileNamed: tempFileName) compressFile.	"makes xxx.gz"
		aDirectory 
			rename: (tempFileName, FileDirectory dot, 'gz')
			toBe: fName.
		aDirectory
			deleteFileNamed: tempFileName
			ifAbsent: []
	].
	fileStream := aDirectory newFileNamed: tempFileName.
	fileStream fileOutClass: nil andObject: self.
		"remember extra structures.  Note class names."

	"append sources"
	allClassesInRoots := arrayOfRoots select: [:cls | cls isBehavior].
	classesToWriteEntirely := allClassesInRoots select: [ :cls | cls theNonMetaClass isSystemDefined].
	methodsWithSource := OrderedCollection new.
	allClassesInRoots do: [ :cls |
		(classesToWriteEntirely includes: cls) ifFalse: [
			cls selectorsAndMethodsDo: [ :sel :meth |
				meth sourcePointer = 0 ifFalse: [methodsWithSource add: {cls. sel. meth}].
			].
		].
	].
	(classesToWriteEntirely isEmpty and: [methodsWithSource isEmpty]) ifTrue: [zipper value. ^ self].

	fileStream reopen; setToEnd.
	fileStream nextPutAll: '\\!!ImageSegment new!!\\' withCRs.
	methodsWithSource do: [ :each |
		fileStream nextPut: $!!.	"try to pacify ImageSegment>>scanFrom:"
		fileStream nextChunkPut: 'RenamedClassSourceReader formerClassName: ',
				each first name printString,' methodsFor: ',
				(each first organization categoryOfElement: each second) asString printString,
				' stamp: ',(each third timeStamp) printString; cr.
		fileStream nextChunkPut: (each third getSourceFor: each second in: each first) asString.
		fileStream nextChunkPut: ' '; cr.
	].
	classesToWriteEntirely do: [:cls | 
		cls isMeta ifFalse: [fileStream nextPutAll: 
						(cls name, ' category: ''', cls category, '''.!!'); cr; cr].
		cls organization
			putCommentOnFile: fileStream
			numbered: 0
			moveSource: false
			forClass: cls.	"does nothing if metaclass"
		cls organization categories do: 
			[:heading |
			cls fileOutCategory: heading
				on: fileStream
				moveSource: false
				toFile: 0]].
	"no class initialization -- it came in as a real object"
	fileStream close.
	zipper value! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'eem 8/21/2015 18:52'!
writeForExportWithSources: fName inDirectory: aDirectory changeSet:
aChangeSetOrNil
	"Write the segment on the disk with all info needed to
reconstruct it in a new image.  For export.  Out pointers are encoded
as normal objects on the disk.  Append the source code of any classes
in roots.  Target system will quickly transfer the sources to its
changes file."
	"Files out a changeSet first, so that a project can contain
classes that are unique to the project."

	| fileStream tempFileName zipper allClassesInRoots classesToWriteEntirely methodsWithSource |
	state = #activeCopy ifFalse: [self error: 'wrong state'].
	(fName includes: $.) ifFalse: [
		^ self inform: 'Please use ''.pr'' or ''.extSeg'' at
the end of the file name'].
	tempFileName := aDirectory nextNameFor: 'SqProject' extension: 'temp'.
	zipper := [
		Preferences debugPrintSpaceLog ifTrue:[
			fileStream := aDirectory newFileNamed:
				(fName copyFrom: 1 to: (fName
lastIndexOf: $.)), 'space'.
			self printSpaceAnalysisOn: fileStream.
			fileStream close].
		ProgressNotification signal: '3:uncompressedSaveComplete'.
		(aDirectory oldFileNamed: tempFileName) compressFile.
	"makes xxx.gz"
		aDirectory
			rename: (tempFileName, FileDirectory dot, 'gz')
			toBe: fName.
		aDirectory
			deleteFileNamed: tempFileName
			ifAbsent: []
	].
	fileStream := aDirectory newFileNamed: tempFileName.
	fileStream fileOutChangeSet: aChangeSetOrNil andObject: self.
		"remember extra structures.  Note class names."

	"append sources"
	allClassesInRoots := arrayOfRoots select: [:cls | cls isBehavior].
	classesToWriteEntirely := allClassesInRoots select: [ :cls |
		cls theNonMetaClass isSystemDefined].
	methodsWithSource := OrderedCollection new.
	allClassesInRoots do: [ :cls |
		(classesToWriteEntirely includes: cls) ifFalse: [
			cls selectorsAndMethodsDo: [ :sel :meth |
				meth sourcePointer = 0 ifFalse:
					[methodsWithSource add: {cls. sel. meth}].
			].
		].
	].
	(classesToWriteEntirely isEmpty and: [methodsWithSource isEmpty]) ifTrue:
		[zipper value. ^ self].

	fileStream reopen; setToEnd.
	fileStream nextPutAll: '\\!!ImageSegment new!!\\' withCRs.
	methodsWithSource do: [ :each |
		fileStream nextPut: $!!.	"try to pacify
ImageSegment>>scanFrom:"
		fileStream nextChunkPut: 'RenamedClassSourceReader
formerClassName: ',
				each first name printString,' methodsFor: ',
				(each first organization
categoryOfElement: each second) asString printString,
				' stamp: ',(each third timeStamp) printString; cr.
		fileStream nextChunkPut: (each third getSourceFor:
each second in: each first) asString.
		fileStream nextChunkPut: ' '; cr.
	].
	classesToWriteEntirely do: [:cls |
		cls isMeta ifFalse: [fileStream nextPutAll:
						(cls name, '
category: ''', cls category, '''.!!'); cr; cr].
		cls organization
			putCommentOnFile: fileStream
			numbered: 0
			moveSource: false
			forClass: cls.	"does nothing if metaclass"
		cls organization categories do:
			[:heading |
			cls fileOutCategory: heading
				on: fileStream
				moveSource: false
				toFile: 0]].
	"no class initialization -- it came in as a real object"
	fileStream close.
	zipper value
! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'eem 8/21/2015 18:47'!
writeForExport: shortName
	"Write the segment on the disk with all info needed to reconstruct it in a new image.  For export.  Out pointers are encoded as normal objects on the disk."

	| fileStream |
	state = #activeCopy ifFalse: [self error: 'wrong state'].
	fileStream := FileStream newFileNamed: (FileDirectory fileName: shortName extension: self class fileExtension).
	fileStream fileOutClass: nil andObject: self.
		"remember extra structures.  Note class names."! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'eem 8/21/2015 18:53'!
writeToFile

	state = #active ifFalse: [self error: 'wrong state'. ^ self].
	Cursor write showWhile: [
		segmentName ifNil: [
			segmentName := (FileDirectory localNameFor: fileName) sansPeriodSuffix].
			"OK that still has number on end.  This is an unusual case"
		fileName := self class uniqueFileNameFor: segmentName.	"local name"
		(self class segmentDirectory newFileNamed: fileName) nextPutAll: segment; close.
		segment := nil.
		state := #onFile].! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'ar 4/10/2005 22:19'!
writeToFileWithSymbols
	| symbols nonSymbols pound |

	state = #extracted ifFalse: [self error: 'wrong state'].
	segmentName ifNil: [
		segmentName := (FileDirectory localNameFor: fileName) sansPeriodSuffix].
		"OK that still has number on end.  This is an unusual case"
	fileName := self class uniqueFileNameFor: segmentName.
	symbols := OrderedCollection new.
	nonSymbols := OrderedCollection new.
	pound := '#' asSymbol.
	outPointers do:
		[:s | 
		((s isSymbol) and: [s isLiteral and: [s ~~ pound]])
			ifTrue: [symbols addLast: s]
			ifFalse: [symbols addLast: pound.  nonSymbols addLast: s]].
	(self class segmentDirectory newFileNamed: fileName)
		store: symbols asArray; cr;
		nextPutAll: segment; close.
	outPointers := nonSymbols asArray.
	state := #onFileWithSymbols! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tk 10/8/1999 12:23'!
writeToFileWithSymbols: shortName

	segmentName := (shortName endsWith: '.seg')
		ifTrue: [shortName copyFrom: 1 to: shortName size - 4]
		ifFalse: [shortName].
	segmentName last isDigit ifTrue: [segmentName := segmentName, '-'].
	self writeToFileWithSymbols.! !

!ImageSegment methodsFor: 'read/write segment' stamp: 'tk 10/8/1999 12:23'!
writeToFile: shortName
	"The short name can't have any fileDelimiter characters in it.  It is remembered in case the segment must be brought in and then sent out again (see ClassDescription updateInstancesFrom:)."

	segmentName := (shortName endsWith: '.seg')
		ifTrue: [shortName copyFrom: 1 to: shortName size - 4]
		ifFalse: [shortName].
	segmentName last isDigit ifTrue: [segmentName := segmentName, '-'].
	self writeToFile.! !

Object subclass: #Imports
	instanceVariableNames: 'imports'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

Imports class
	instanceVariableNames: 'default'!

!Imports commentStamp: 'sd 5/11/2003 20:34' prior: 0!
I represent imported resources such as images, sounds, and other kind of files. 
For now I only store images in a simple way. 

To access my default instance use: Imports default. However I'm not a strict singleton
and clients may create several of me using new. !

Imports class
	instanceVariableNames: 'default'!

!Imports class methodsFor: 'instance creation' stamp: 'nk 7/12/2003 10:38'!
default
	"Answer my default instance, creating one if necessary."
	"Imports default"
	^default ifNil: [ default := self new ]! !

!Imports class methodsFor: 'instance creation' stamp: 'nk 7/12/2003 10:36'!
default: anImports
	"Set my default instance. Returns the old value if any."
	| old |
	old := default.
	default := anImports.
	^old! !

!Imports methodsFor: 'images' stamp: 'sd 5/11/2003 20:36'!
images
	"returns all the imported images"

	^ imports values

	! !

!Imports methodsFor: 'icons' stamp: 'nk 6/12/2004 12:44'!
importImageDirectoryWithSubdirectories: directoryOrName 
	| dir forms |
	dir := directoryOrName isString
		ifFalse: [ directoryOrName ]
		ifTrue: [ FileDirectory default directoryNamed: directoryOrName ].
	dir exists
		ifFalse: [self error: dir fullName , ' does not exist'. ^ #()].
	forms := OrderedCollection new.
	dir withAllSubdirectoriesCollect: [ :subdir | forms addAll: (self importImageDirectory: dir) ].
	^ forms! !

!Imports methodsFor: 'icons' stamp: 'ul 12/12/2009 14:10'!
importImageDirectory: directoryOrName 
	| dir extensions forms |
	dir := directoryOrName isString
		ifFalse: [ directoryOrName ]
		ifTrue: [ FileDirectory default directoryNamed: directoryOrName ].
	dir exists
		ifFalse: [self error: dir fullName , ' does not exist'. ^ #()].
	extensions := (ImageReadWriter allTypicalFileExtensions add: 'form';
				 yourself)
				collect: [:ex | '.' , ex].
	forms := OrderedCollection new.
	dir fileNames
		do: [:fileName | | fullName | (fileName endsWithAnyOf: extensions)
				ifTrue: [fullName := dir fullNameFor: fileName.
					(self importImageFromFileNamed: fullName)
						ifNotNil: [:form | forms add: form]]].
	^ forms! !

!Imports methodsFor: 'icons' stamp: 'nk 6/12/2004 12:25'!
importImageFromFileNamed: fullName 
	| localName pathParts form imageName |
	FileDirectory
		splitName: fullName
		to: [:dirPath :lname | 
			localName := lname.
			pathParts := dirPath findTokens: FileDirectory slash].
	form := [Form fromFileNamed: fullName]
				on: Error
				do: [:ex | ex return: nil].
	form
		ifNil: [^ nil].
	imageName := FileDirectory baseNameFor: localName.
	[imports includesKey: imageName]
		whileTrue: [imageName := pathParts isEmpty
						ifTrue: [Utilities
								keyLike: imageName
								satisfying: [:ea | (imports includesKey: ea) not]]
						ifFalse: [pathParts removeLast , '-' , imageName]].
	imports at: imageName put: form.
	^ form! !

!Imports methodsFor: 'images' stamp: 'nk 6/12/2004 12:49'!
importImage: anImage named: aName 
	imports
		at: (Utilities
				keyLike: aName
				satisfying: [:ea | (imports includesKey: ea) not])
		put: anImage! !

!Imports methodsFor: 'images' stamp: 'yo 7/17/2003 00:17'!
imports

	^ imports
! !

!Imports methodsFor: 'initialize' stamp: 'sd 5/11/2003 18:17'!
initialize

	imports := Dictionary new.! !

!Imports methodsFor: 'images' stamp: 'sd 5/11/2003 22:26'!
namesAndImagesDo: aBlock
	"iterate over all the names and image"

	^ imports keysAndValuesDo: aBlock

	! !

!Imports methodsFor: 'images' stamp: 'dtl 2/10/2010 13:22'!
viewImages
	"Open up a special Form inspector on the dictionary of graphical imports."
	"Imports default viewImages"

	imports size isZero ifTrue:
		[^ self inform: 
'The ImageImports repository is currently empty,
so there is nothing to view at this time.  You can
use a file list to import graphics from external files
into Imports, and once you have done that,
you will find this command more interesting.'].
	
	Project current formViewClass openOn:  imports withLabel: 'Graphical Imports'

! !

Object subclass: #ISOLanguageDefinition
	instanceVariableNames: 'iso3 iso2 iso3Alternate language'
	classVariableNames: 'ISO2Countries ISO2Table ISO3Countries ISO3Table'
	poolDictionaries: ''
	category: 'System-Localization'!

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 9/1/2005 14:06'!
buildIso3166CodesTables
	"ISOLanguageDefinition buildIso3166CodesTables"
	| rawdata stream country isoa2 isoa3 unNumeric macName macCode windowsName windowsCode empty table |
	rawdata := self iso3166Codes.
	table := OrderedCollection new: 200. 
	stream := rawdata readStream.
	empty := 160 asCharacter asString.
	[stream atEnd] whileFalse: 
		[country := stream nextLine.
		isoa2 := stream nextLine.
		isoa3 := stream nextLine.
		unNumeric := stream nextLine.
		windowsName := stream nextLine.
		windowsName = empty ifTrue: [windowsName := nil].
		windowsCode := stream nextLine. 
		windowsCode = empty ifTrue: [windowsCode := nil].
		macName := stream nextLine.
		macName = empty ifTrue: [macName := nil].
		macCode := stream nextLine.
		macCode = empty ifTrue: [macCode := nil].
		table add: { country.  isoa2. isoa3.  unNumeric. windowsName.  windowsCode.  macName. macCode. }].
	^table! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 9/1/2005 14:14'!
extraCountryDefinitions
	^{
	{'Kids'. 'KIDS'. 'KIDS'.}.
	}! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 7/15/2004 18:14'!
extraISO3Definitions

	^self readISOLanguagesFrom: 'jpk		Japanese (Kids)
' readStream! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 7/15/2004 18:13'!
initISO3LanguageTable
	"ISOLanguageDefinition initIso3LanguageTable"

	| table |
	table := ISOLanguageDefinition readISOLanguagesFrom: ISOLanguageDefinition isoLanguages readStream.
	table addAll: self extraISO3Definitions.
	^table! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 9/1/2005 14:12'!
initISOCountries
	"ISOLanguageDefinition initISOCountries"
	| iso3166Table |
	iso3166Table := ISOLanguageDefinition buildIso3166CodesTables.
	ISO2Countries := Dictionary new.
	ISO3Countries := Dictionary new.
	iso3166Table do: [:entry | 
		ISO2Countries at: (entry at: 2) put: (entry at: 1).
		ISO3Countries at: (entry at: 3) put: (entry at: 1)].
	self extraCountryDefinitions do: [:entry | 
		ISO2Countries at: (entry at: 2) put: (entry at: 1).
		ISO3Countries at: (entry at: 3) put: (entry at: 1)]! !

!ISOLanguageDefinition class methodsFor: 'class initialization' stamp: 'mir 7/1/2004 18:19'!
initialize
	"ISOLanguageDefinition initialize"

	ISO3Table := nil.
	ISO2Table := nil! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 9/1/2005 14:18'!
iso2Countries
	"ISOLanguageDefinition iso2Countries"
	"ISO2Countries := nil. ISO3Countries := nil"

	ISO2Countries ifNil: [self initISOCountries].
	^ISO2Countries! !

!ISOLanguageDefinition class methodsFor: 'accessing' stamp: 'mir 7/1/2004 18:06'!
iso2LanguageDefinition: aString
	^self iso2LanguageTable at: aString! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 7/1/2004 18:14'!
iso2LanguageTable
	"ISOLanguageDefinition iso2LanguageTable"

	ISO2Table ifNotNil: [^ISO2Table].
	ISO2Table := Dictionary new: self iso3LanguageTable basicSize.
	self iso3LanguageTable do: [:entry |
		ISO2Table at: entry iso2 put: entry].
	^ISO2Table! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 9/1/2005 13:57'!
iso3166Codes
"http://www.unicode.org/onlinedat/countries.html"

^'­LAND ISLANDS
AX






AFGHANISTAN
AF
AFG
004




ALBANIA
AL
ALB
008
CTRY_ALBANIA
355


ALGERIA
DZ
DZA
012
CTRY_ALGERIA
213
verArabic
16
AMERICAN SAMOA
AS
ASM
016




ANDORRA
AD
AND
020




ANGOLA
AO
AGO
024




ANGUILLA
AI
AIA
660




ANTARCTICA
AQ
ATA
010




ANTIGUA AND BARBUDA
AG
ATG
028




ARGENTINA
AR
ARG
032
CTRY_ARGENTINA
54


ARMENIA
AM
ARM
051
CTRY_ARMENIA
374
verArmenian
84
ARUBA
AW
ABW
533




AUSTRALIA
AU
AUS
036
CTRY_AUSTRALIA
61
verAustralia
15
AUSTRIA
AT
AUT
040
CTRY_AUSTRIA
43
verAustria
92
AZERBAIJAN
AZ
AZE
031
CTRY_AZERBAIJAN
994


BAHAMAS
BS
BHS
044




BAHRAIN
BH
BHR
048
CTRY_BAHRAIN
973


BANGLADESH
BD
BGD
050


verBengali
60
BARBADOS
BB
BRB
052




BELARUS
BY
BLR
112
CTRY_BELARUS
375


BELGIUM
BE
BEL
056
CTRY_BELGIUM
32
verFrBelgium, verFlemish
98
BELIZE
BZ
BLZ
084
CTRY_BELIZE
501


BENIN
BJ
BEN
204




BERMUDA
BM
BMU
060




BHUTAN
BT
BTN
064


verBhutan
83
BOLIVIA
BO
BOL
068
CTRY_BOLIVIA
591


BOSNIA AND HERZEGOVINA
BA
BIH
070




BOTSWANA
BW
BWA
072




BOUVET ISLAND
BV
BVT
074




BRAZIL
BR
BRA
076
CTRY_BRAZIL
55
verBrazil
71
BRITISH INDIAN OCEAN TERRITORY
IO
IOT
086




BRUNEI DARUSSALAM
BN
BRN
096
CTRY_BRUNEI_DARUSSALAM
673


BULGARIA
BG
BGR
100
CTRY_BULGARIA
359
verBulgaria
72
BURKINA FASO
BF
BFA
854




BURUNDI
BI
BDI
108




CAMBODIA
KH
KHM
116




CAMEROON
CM
CMR
120




CANADA
CA
CAN
124
CTRY_CANADA
2
verFrCanada, verEndCanada
82
CAPE VERDE
CV
CPV
132




CAYMAN ISLANDS
KY
CYM
136




CENTRAL AFRICAN REPUBLIC
CF
CAF
140




CHAD
TD
TCD
148




CHILE
CL
CHL
152
CTRY_CHILE
56


CHINA
CN
CHN
156
CTRY_PRCHINA
86
verChina
52
CHRISTMAS ISLAND
CX
CXR
162




COCOS (KEELING) ISLANDS
CC
CCK
166




COLOMBIA
CO
COL
170
CTRY_COLOMBIA
57


COMOROS
KM
COM
174




CONGO
CG
COG
178




CONGO, THE DEMOCRATIC REPUBLIC OF THE
CD






COOK ISLANDS
CK
COK
184




COSTA RICA
CR
CRI
188
CTRY_COSTA_RICA
506


COTE D''IVOIRE
CI
CIV
384




CROATIA (local name: Hrvatska)
HR
HRV
191
CTRY_CROATIA
385
verCroatia, verYugoCroatian
68 (c), 25 (y)
CUBA
CU
CUB
192




CYPRUS
CY
CYP
196


verCyprus
23
CZECH REPUBLIC
CZ
CZE
203
CTRY_CZECH
420
verCzech
56
DENMARK
DK
DNK
208
CTRY_DENMARK
45
verDenmark(da), verFaeroeIsl(fo)
9(da), 47(fo)
DJIBOUTI
DJ
DJI
262




DOMINICA
DM
DMA
212




DOMINICAN REPUBLIC
DO
DOM
214
CTRY_DOMINICAN_REPUBLIC
1


EAST TIMOR
TL
TLS
626




ECUADOR
EC
ECU
218
CTRY_ECUADOR
593


EGYPT
EG
EGY
818
CTRY_EGYPT
20
verArabic
16
EL SALVADOR
SV
SLV
222
CTRY_EL_SALVADOR
503


EQUATORIAL GUINEA
GQ
GNQ
226




ERITREA
ER
ERI
232




ESTONIA
EE
EST
233
CTRY_ESTONIA
372
verEstonia
44
ETHIOPIA
ET
ETH
210




FALKLAND ISLANDS (MALVINAS)
FK
FLK
238




FAROE ISLANDS
FO
FRO
234
CTRY_FAEROE_ISLANDS
298


FIJI
FJ
FJI
242




FINLAND
FI
FIN
246
CTRY_FINLAND
358
verFinland
17
FRANCE
FR
FRA
250
CTRY_FRANCE
33
verFrance
1
FRANCE, METROPOLITAN
FX
FXX
249




FRENCH GUIANA
GF
GUF
254




FRENCH POLYNESIA
PF
PYF
258




FRENCH SOUTHERN TERRITORIES
TF
ATF
260




GABON
GA
GAB
266




GAMBIA
GM
GMB
270




GEORGIA
GE
GEO
268
CTRY_GEORGIA
995
verGeorgian
85
GERMANY
DE
DEU
276
CTRY_GERMANY
49
verGermany
3
GHANA
GH
GHA
288




GIBRALTAR
GI
GIB
292




GREECE
GR
GRC
300
CTRY_GREECE
30
verGreece, verGreecePoly
20, 40
GREENLAND
GL
GRL
304


verGreenland
107
GRENADA
GD
GRD
308




GUADELOUPE
GP
GLP
312




GUAM
GU
GUM
316




GUATEMALA
GT
GTM
320
CTRY_GUATEMALA
502


GUINEA
GN
GIN
324




GUINEA-BISSAU
GW
GNB
624




GUYANA
GY
GUY
328




HAITI
HT
HTI
332




HEARD ISLAND & MCDONALD ISLANDS
HM
HMD
334




HONDURAS
HN
HND
340
CTRY_HONDURAS
504


HONG KONG
HK
HKG
344
CTRY_HONG_KONG
852


HUNGARY
HU
HUN
348
CTRY_HUNGARY
36
verHungary
43
ICELAND
IS
ISL
352
CTRY_ICELAND
354
verIceland
21
INDIA
IN
IND
356
CTRY_INDIA
91
verIndiaHindi(hi)
33
INDONESIA
ID
IDN
360
CTRY_INDONESIA
62


IRAN, ISLAMIC REPUBLIC OF
IR
IRN
364
CTRY_IRAN
981
verIran
48
IRAQ
IQ
IRQ
368
CTRY_IRAQ
964
verArabic
16
IRELAND
IE
IRL
372
CTRY_IRELAND
353
verIreland
50
ISRAEL
IL
ISR
376
CTRY_ISRAEL
972
verIsrael
13
ITALY
IT
ITA
380
CTRY_ITALY
39
verItaly
4
JAMAICA
JM
JAM
388
CTRY_JAMAICA
1


JAPAN
JP
JPN
392
CTRY_JAPAN
81
verJapan
14
JORDAN
JO
JOR
400
CTRY_JORDAN
962


KAZAKHSTAN
KZ
KAZ
398
CTRY_KAZAKSTAN
7


KENYA
KE
KEN
404
CTRY_KENYA
254


KIRIBATI
KI
KIR
296




KOREA, DEMOCRATIC PEOPLE''S REPUBLIC OF
KP
PRK
408


verKorea
51
KOREA, REPUBLIC OF
KR
KOR
410
CTRY_SOUTH_KOREA
82
verKorea

KUWAIT
KW
KWT
414
CTRY_KUWAIT
965


KYRGYZSTAN
KG
KGZ
417
CTRY_KYRGYZSTAN
996


LAO PEOPLE''S DEMOCRATIC REPUBLIC
LA
LAO
418




LATVIA
LV
LVA
428
CTRY_LATVIA
371
verLatvia
45
LEBANON
LB
LBN
422
CTRY_LEBANON
961


LESOTHO
LS
LSO
426




LIBERIA
LR
LBR
430




LIBYAN ARAB JAMAHIRIYA
LY
LBY
434
CTRY_LIBYA
218
verArabic
16
LIECHTENSTEIN
LI
LIE
438
CTRY_LIECHTENSTEIN
41


LITHUANIA
LT
LTU
440
CTRY_LITHUANIA
370
verLithuania
41
LUXEMBOURG
LU
LUX
442
CTRY_LUXEMBOURG
352
verFrBelgiumLux
6
MACAU
MO
MAC
446
CTRY_MACAU
853


MACEDONIA, THE FORMER YUGOSLAV REPUBLIC OF
MK
MKD
807
CTRY_MACEDONIA
389
verMacedonian

MADAGASCAR
MG
MDG
450




MALAWI
MW
MWI
454




MALAYSIA
MY
MYS
458
CTRY_MALAYSIA
60


MALDIVES
MV
MDV
462
CTRY_MALDIVES
960


MALI
ML
MLI
466




MALTA
MT
MLT
470


verMalta
22
MARSHALL ISLANDS
MH
MHL
584




MARTINIQUE
MQ
MTQ
474




MAURITANIA
MR
MRT
478




MAURITIUS
MU
MUS
480




MAYOTTE
YT
MYT
175




MEXICO
MX
MEX
484
CTRY_MEXICO
52


MICRONESIA, FEDERATED STATES OF
FM
FSM
583




MOLDOVA, REPUBLIC OF
MD
MDA
498




MONACO
MC
MCO
492
CTRY_MONACO
33


MONGOLIA
MN
MNG
496
CTRY_MONGOLIA
976


MONTSERRAT
MS
MSR
500




MOROCCO
MA
MAR
504
CTRY_MOROCCO
212
verArabic
16
MOZAMBIQUE
MZ
MOZ
508




MYANMAR
MM
MMR
104




NAMIBIA
NA
NAM
516




NAURU
NR
NRU
520




NEPAL
NP
NPL
524


verNepal
106
NETHERLANDS
NL
NLD
528
CTRY_NETHERLANDS
31
verNetherlands
5
NETHERLANDS ANTILLES
AN
ANT
530




NEW CALEDONIA
NC
NCL
540




NEW ZEALAND
NZ
NZL
554
CTRY_NEW_ZEALAND
64


NICARAGUA
NI
NIC
558
CTRY_NICARAGUA
505


NIGER
NE
NER
562




NIGERIA
NG
NGA
566




NIUE
NU
NIU
570




NORFOLK ISLAND
NF
NFK
574




NORTHERN MARIANA ISLANDS
MP
MNP
580




NORWAY
NO
NOR
578
CTRY_NORWAY
47
verNorway
12
OMAN
OM
OMN
512
CTRY_OMAN
968


PAKISTAN
PK
PAK
586
CTRY_PAKISTAN
92
verPakistanUrdu, verPunjabi
34 (U), 95 (P)
PALAU
PW
PLW
585




PANAMA
PA
PAN
591
CTRY_PANAMA
507


PALESTINIAN TERRITORY, OCCUPIED
PS






PAPUA NEW GUINEA
PG
PNG
598




PARAGUAY
PY
PRY
600
CTRY_PARAGUAY
595


PERU
PE
PER
604
CTRY_PERU
51


PHILIPPINES
PH
PHL
608
CTRY_PHILIPPINES
63


PITCAIRN
PN
PCN
612




POLAND
PL
POL
616
CTRY_POLAND
48
verPoland
42
PORTUGAL
PT
PRT
620
CTRY_PORTUGAL
351
verPortugal
10
PUERTO RICO
PR
PRI
630
CTRY_PUERTO_RICO
1


QATAR
QA
QAT
634
CTRY_QATAR
974


REUNION
RE
REU
638




ROMANIA
RO
ROU*
642
CTRY_ROMANIA
40
verRomania
39
RUSSIAN FEDERATION
RU
RUS
643
CTRY_RUSSIA
7
verRussia
49
RWANDA
RW
RWA
646




SAINT KITTS AND NEVIS
KN
KNA
659




SAINT LUCIA
LC
LCA
662




SAINT VINCENT AND THE GRENADINES
VC
VCT
670




SAMOA
WS
WSM
882




SAN MARINO
SM
SMR
674




SAO TOME AND PRINCIPE
ST
STP
678




SAUDI ARABIA
SA
SAU
682
CTRY_SAUDI_ARABIA
966
verArabic
16
SENEGAL
SN
SEN
686




SERBIA AND MONTENEGRO
CS


CTRY_SERBIA
381


SEYCHELLES
SC
SYC
690




SIERRA LEONE
SL
SLE
694




SINGAPORE
SG
SGP
702
CTRY_SINGAPORE
65
verSingapore
100
SLOVAKIA (Slovak Republic)
SK
SVK
703
CTRY_SLOVAK
421
verSlovak
57
SLOVENIA
SI
SVN
705
CTRY_SLOVENIA
386
verSlovenian
66
SOLOMON ISLANDS
SB
SLB
90




SOMALIA
SO
SOM
706




SOUTH AFRICA
ZA
ZAF
710
CTRY_SOUTH_AFRICA
27


SOUTH GEORGIA AND THE SOUTH SANDWICH ISLANDS
GS






SPAIN
ES
ESP
724
CTRY_SPAIN
34
verSpain
8
SRI LANKA
LK
LKA
144




SAINT HELENA
SH
SHN
654




SAINT PIERRE AND MIQUELON
PM
SPM
666




SUDAN
SD
SDN
736




SURINAME
SR
SUR
740




SVALBARD AND JAN MAYEN ISLANDS
SJ
SJM
744




SWAZILAND
SZ
SWZ
748




SWEDEN
SE
SWE
752
CTRY_SWEDEN
46
verSweden
7
SWITZERLAND
CH
CHE
756
CTRY_SWITZERLAND
41
verFrSwiss(fr), verGrSwiss(de)
18(fr), 19(de)
SYRIAN ARAB REPUBLIC
SY
SYR
760
CTRY_SYRIA
963


TAIWAN, PROVINCE OF CHINA
TW
TWN
158
CTRY_TAIWAN
886
verTaiwan
53
TAJIKISTAN
TJ
TJK
762




TANZANIA, UNITED REPUBLIC OF
TZ
TZA
834




TATARSTAN



CTRY_TATARSTAN
7


THAILAND
TH
THA
764
CTRY_THAILAND
66
verThailand
54
TIMOR-LESTE
TL






TOGO
TG
TGO
768




TOKELAU
TK
TKL
772




TONGA
TO
TON
776


verTonga
88
TRINIDAD AND TOBAGO
TT
TTO
780
CTRY_TRINIDAD_Y_TOBAGO
1


TUNISIA
TN
TUN
788
CTRY_TUNISIA
216
verArabic
16
TURKEY
TR
TUR
792
CTRY_TURKEY
90
verTurkey
24
TURKMENISTAN
TM
TKM
795




TURKS AND CAICOS ISLANDS
TC
TCA
796




TUVALU
TV
TUV
798




UGANDA
UG
UGA
800




UKRAINE
UA
UKR
804
CTRY_UKRAINE
380
verUkraine
62
UNITED ARAB EMIRATES
AE
ARE
784
CTRY_UAE
971


UNITED KINGDOM
GB
GBR
826
CTRY_UNITED_KINGDOM
44
verBritain
2
UNITED STATES
US
USA
840
CTRY_UNITED_STATES
1
verUS
0
UNITED STATES MINOR OUTLYING ISLANDS
UM
UMI
581




URUGUAY
UY
URY
858
CTRY_URUGUAY
598


UZBEKISTAN
UZ
UZB
860
CTRY_UZBEKISTAN
7


VANUATU
VU
VUT
548




VATICAN CITY STATE (HOLY SEE)
VA
VAT
336




VENEZUELA
VE
VEN
862
CTRY_VENEZUELA
58


VIET NAM
VN
VNM
704
CTRY_VIET_NAM
84
verVietnam

VIRGIN ISLANDS (BRITISH)
VG
VGB
92




VIRGIN ISLANDS (U.S.)
VI
VIR
850




WALLIS AND FUTUNA ISLANDS
WF
WLF
876




WESTERN SAHARA
EH
ESH
732




YEMEN
YE
YEM
887
CTRY_YEMEN
967


YUGOSLAVIA
YU
YUG
891




ZAIRE
ZR
ZAR
180




ZAMBIA
ZM
ZMB
894




ZIMBABWE
ZW
ZWE
716
CTRY_ZIMBABWE
263


'! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 9/1/2005 14:18'!
iso3Countries
	"ISOLanguageDefinition iso3Countries"
	"ISO2Countries := nil. ISO3Countries := nil"

	ISO3Countries ifNil: [self initISOCountries].
	^ISO3Countries! !

!ISOLanguageDefinition class methodsFor: 'accessing' stamp: 'mir 7/1/2004 18:06'!
iso3LanguageDefinition: aString
	^self iso3LanguageTable at: aString! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 7/21/2004 13:10'!
iso3LanguageTable
	"ISOLanguageDefinition iso3LanguageTable"

	^ISO3Table ifNil: [ISO3Table := self initISO3LanguageTable]! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'yo 9/28/2009 19:55'!
isoLanguages
	"ISO 639: 3-letter codes"
	^'aar	aa	Afar
abk	ab	Abkhazian
ace		Achinese
ach		Acoli
ada		Adangme
ady		Adyghe; Adygei
afa		Afro-Asiatic languages
afh		Afrihili
afr	af	Afrikaans
ain		Ainu
aka	ak	Akan
akk		Akkadian
alb/sqi	sq	Albanian
ale		Aleut
alg		Algonquian languages
alt		Southern Altai
amh	am	Amharic
ang		English, Old (ca.450-1100)
anp		Angika
apa		Apache languages
ara	ar	Arabic
arc		Official Aramaic (700-300 BCE); Imperial Aramaic (700-300 BCE)
arg	an	Aragonese
arm/hye	hy	Armenian
arn		Mapudungun; Mapuche
arp		Arapaho
art		Artificial languages
arw		Arawak
asm	as	Assamese
ast		Asturian; Bable; Leonese; Asturleonese
ath		Athapascan languages
aus		Australian languages
ava	av	Avaric
ave	ae	Avestan
awa		Awadhi
aym	ay	Aymara
aze	az	Azerbaijani
bad		Banda languages
bai		Bamileke languages
bak	ba	Bashkir
bal		Baluchi
bam	bm	Bambara
ban		Balinese
baq/eus	eu	Basque
bas		Basa
bat		Baltic languages
bej		Beja; Bedawiyet
bel	be	Belarusian
bem		Bemba
ben	bn	Bengali
ber		Berber languages)
bho		Bhojpuri
bih	bh	Bihari
bik		Bikol
bin		Bini; Edo
bis	bi	Bislama
bla		Siksika
bnt		Bantu languages
bos	bs	Bosnian
bra		Braj
bre	br	Breton
btk		Batak languages
bua		Buriat
bug		Buginese
bul	bg	Bulgarian
bur/mya	my	Burmese
byn		Blin; Bilin
cad		Caddo
cai		Central American Indian languages
car		Galibi Carib
cat	ca	Catalan; Valencian
cau		Caucasian languages
ceb		Cebuano
cel		Celtic languages
cha	ch	Chamorro
chb		Chibcha
che	ce	Chechen
chg		Chagatai
chi/zho	zh	Chinese
chk		Chuukese
chm		Mari
chn		Chinook jargon
cho		Choctaw
chp		Chipewyan; Dene Suline
chr		Cherokee
chu	cu	Church Slavic; Old Slavonic; Church Slavonic; Old Bulgarian; Old Church Slavonic
chv	cv	Chuvash
chy		Cheyenne
cmc		Chamic languages
cop		Coptic
cor	kw	Cornish
cos	co	Corsican
cpe		Creoles and pidgins, English based
cpf		Creoles and pidgins, French-based
cpp		Creoles and pidgins, Portuguese-based
cre	cr	Cree
crh		Crimean Tatar; Crimean Turkish
crp		Creoles and pidgins
csb		Kashubian
cus		Cushitic languages
cze/ces	cs	Czech
dak		Dakota
dan	da	Danish
dar		Dargwa
day		Land Dayak languages
del		Delaware
den		Slave (Athapascan)
dgr		Dogrib
din		Dinka
div	dv	Divehi; Dhivehi; Maldivian
doi		Dogri
dra		Dravidian languages
dsb		Lower Sorbian
dua		Duala
dum		Dutch, Middle (ca.1050-1350)
dut/nld	nl	Dutch; Flemish
dyu		Dyula
dzo	dz	Dzongkha
efi		Efik
egy		Egyptian (Ancient)
eka		Ekajuk
elx		Elamite
eng	en	English
enm		English, Middle (1100-1500)
epo	eo	Esperanto
est	et	Estonian
ewe	ee	Ewe
ewo		Ewondo
fan		Fang
fao	fo	Faroese
fat		Fanti
fij	fj	Fijian
fil		Filipino; Pilipino
fin	fi	Finnish
fiu		Finno-Ugrian languages)
fon		Fon
fre/fra	fr	French
frm		French, Middle (ca.1400-1600)
fro		French, Old (842-ca.1400)
frr		Northern Frisian
frs		Eastern Frisian
fry	fy	Western Frisian
ful	ff	Fulah
fur		Friulian
gaa		Ga
gay		Gayo
gba		Gbaya
gem		Germanic languages
geo/kat	ka	Georgian
ger/deu	de	German
gez		Geez
gil		Gilbertese
gla	gd	Gaelic; Scottish Gaelic
gle	ga	Irish
glg	gl	Galician
glv	gv	Manx
gmh		German, Middle High (ca.1050-1500)
goh		German, Old High (ca.750-1050)
gon		Gondi
gor		Gorontalo
got		Gothic
grb		Grebo
grc		Greek, Ancient (to 1453)
gre/ell	el	Greek, Modern (1453-)
grn	gn	Guarani
gsw		Swiss German; Alemannic; Alsatian
guj	gu	Gujarati
gwi		Gwich''in
hai		Haida
hat	ht	Haitian; Haitian Creole
hau	ha	Hausa
haw		Hawaiian
heb	he	Hebrew
her	hz	Herero
hil		Hiligaynon
him		Himachali
hin	hi	Hindi
hit		Hittite
hmn		Hmong
hmo	ho	Hiri Motu
hrv	hr	Croatian
hsb		Upper Sorbian
hun	hu	Hungarian
hup		Hupa
iba		Iban
ibo	ig	Igbo
ice/isl	is	Icelandic
ido	io	Ido
iii	ii	Sichuan Yi; Nuosu
ijo		Ijo languages
iku	iu	Inuktitut
ile	ie	Interlingue; Occidental
ilo		Iloko
ina	ia	Interlingua (International Auxiliary Language Association)
inc		Indic languages
ind	id	Indonesian
ine		Indo-European languages
inh		Ingush
ipk	ik	Inupiaq
ira		Iranian languages
iro		Iroquoian languages
ita	it	Italian
jav	jv	Javanese
jbo		Lojban
jpn	ja	Japanese
jpr		Judeo-Persian
jrb		Judeo-Arabic
kaa		Kara-Kalpak
kab		Kabyle
kac		Kachin; Jingpho
kal	kl	Kalaallisut; Greenlandic
kam		Kamba
kan	kn	Kannada
kar		Karen languages
kas	ks	Kashmiri
kau	kr	Kanuri
kaw		Kawi
kaz	kk	Kazakh
kbd		Kabardian
kha		Khasi
khi		Khoisan languages
khm	km	Central Khmer
kho		Khotanese; Sakan
kik	ki	Kikuyu; Gikuyu
kin	rw	Kinyarwanda
kir	ky	Kirghiz; Kyrgyz
kmb		Kimbundu
kok		Konkani
kom	kv	Komi
kon	kg	Kongo
kor	ko	Korean
kos		Kosraean
kpe		Kpelle
krc		Karachay-Balkar
krl		Karelian
kro		Kru languages
kru		Kurukh
kua	kj	Kuanyama; Kwanyama
kum		Kumyk
kur	ku	Kurdish
kut		Kutenai
lad		Ladino
lah		Lahnda
lam		Lamba
lao	lo	Lao
lat	la	Latin
lav	lv	Latvian
lez		Lezghian
lim	li	Limburgan; Limburger; Limburgish
lin	ln	Lingala
lit	lt	Lithuanian
lol		Mongo
loz		Lozi
ltz	lb	Luxembourgish; Letzeburgesch
lua		Luba-Lulua
lub	lu	Luba-Katanga
lug	lg	Ganda
lui		Luiseno
lun		Lunda
luo		Luo (Kenya and Tanzania)
lus		Lushai
mac/mkd	mk	Macedonian
mad		Madurese
mag		Magahi
mah	mh	Marshallese
mai		Maithili
mak		Makasar
mal	ml	Malayalam
man		Mandingo
mao/mri	mi	Maori
map		Austronesian languages
mar	mr	Marathi
mas		Masai
may/msa	ms	Malay
mdf		Moksha
mdr		Mandar
men		Mende
mga		Irish, Middle (900-1200)
mic		Mi''kmaq; Micmac
min		Minangkabau
mis		Uncoded languages
mkh		Mon-Khmer languages
mlg	mg	Malagasy
mlt	mt	Maltese
mnc		Manchu
mni		Manipuri
mno		Manobo languages
moh		Mohawk
mon	mn	Mongolian
mos		Mossi
mul		Multiple languages
mun		Munda languages
mus		Creek
mwl		Mirandese
mwr		Marwari
myn		Mayan languages
myv		Erzya
nah		Nahuatl languages
nai		North American Indian languages
nap		Neapolitan
nau	na	Nauru
nav	nv	Navajo; Navaho
nbl	nr	Ndebele, South; South Ndebele
nde	nd	Ndebele, North; North Ndebele
ndo	ng	Ndonga
nds		Low German; Low Saxon; German, Low; Saxon, Low
nep	ne	Nepali
new		Nepal Bhasa; Newari
nia		Nias
nic		Niger-Kordofanian languages
niu		Niuean
nno	nn	Norwegian Nynorsk; Nynorsk, Norwegian
nob	nb	Bokmal, Norwegian; Norwegian Bokmal
nog		Nogai
non		Norse, Old
nor	no	Norwegian
nqo		N''Ko
nso		Pedi; Sepedi; Northern Sotho
nub		Nubian languages
nwc		Classical Newari; Old Newari; Classical Nepal Bhasa
nya	ny	Chichewa; Chewa; Nyanja
nym		Nyamwezi
nyn		Nyankole
nyo		Nyoro
nzi		Nzima
oci	oc	Occitan (post 1500)
oji	oj	Ojibwa
ori	or	Oriya
orm	om	Oromo
osa		Osage
oss	os	Ossetian; Ossetic
ota		Turkish, Ottoman (1500-1928)
oto		Otomian languages
paa		Papuan languages
pag		Pangasinan
pal		Pahlavi
pam		Pampanga; Kapampangan
pan	pa	Panjabi; Punjabi
pap		Papiamento
pau		Palauan
peo		Persian, Old (ca.600-400 B.C.)
per/fas	fa	Persian
phi		Philippine languages)
phn		Phoenician
pli	pi	Pali
pol	pl	Polish
pon		Pohnpeian
por	pt	Portuguese
pra		Prakrit languages
pro		Provencal, Old (to 1500);Occitan, Old (to 1500)
pus	ps	Pushto; Pashto
qaa-qtz		Reserved for local use
que	qu	Quechua
raj		Rajasthani
rap		Rapanui
rar		Rarotongan; Cook Islands Maori
roa		Romance languages
roh	rm	Romansh
rom		Romany
rum/ron	ro	Romanian; Moldavian; Moldovan
run	rn	Rundi
rup		Aromanian; Arumanian; Macedo-Romanian
rus	ru	Russian
sad		Sandawe
sag	sg	Sango
sah		Yakut
sai		South American Indian languages
sal		Salishan languages
sam		Samaritan Aramaic
san	sa	Sanskrit
sas		Sasak
sat		Santali
scn		Sicilian
sco		Scots
sel		Selkup
sem		Semitic languages
sga		Irish, Old (to 900)
sgn		Sign Languages
shn		Shan
sid		Sidamo
sin	si	Sinhala; Sinhalese
sio		Siouan languages
sit		Sino-Tibetan languages
sla		Slavic languages
slo/slk	sk	Slovak
slv	sl	Slovenian
sma		Southern Sami
sme	se	Northern Sami
smi		Sami languages
smj		Lule Sami
smn		Inari Sami
smo	sm	Samoan
sms		Skolt Sami
sna	sn	Shona
snd	sd	Sindhi
snk		Soninke
sog		Sogdian
som	so	Somali
son		Songhai languages
sot	st	Sotho, Southern
spa	es	Spanish; Castilian
srd	sc	Sardinian
srn		Sranan Tongo
srp	sr	Serbian
srr		Serer
ssa		Nilo-Saharan languages
ssw	ss	Swati
suk		Sukuma
sun	su	Sundanese
sus		Susu
sux		Sumerian
swa	sw	Swahili
swe	sv	Swedish
syc		Classical Syriac
syr		Syriac
tah	ty	Tahitian
tai		Tai languages
tam	ta	Tamil
tat	tt	Tatar
tel	te	Telugu
tem		Timne
ter		Tereno
tet		Tetum
tgk	tg	Tajik
tgl	tl	Tagalog
tha	th	Thai
tib/bod	bo	Tibetan
tig		Tigre
tir	ti	Tigrinya
tiv		Tiv
tkl		Tokelau
tlh		Klingon; tlhIngan-Hol
tli		Tlingit
tmh		Tamashek
tog		Tonga (Nyasa)
ton	to	Tonga (Tonga Islands)
tpi		Tok Pisin
tsi		Tsimshian
tsn	tn	Tswana
tso	ts	Tsonga
tuk	tk	Turkmen
tum		Tumbuka
tup		Tupi languages
tur	tr	Turkish
tut		Altaic languages
tvl		Tuvalu
twi	tw	Twi
tyv		Tuvinian
udm		Udmurt
uga		Ugaritic
uig	ug	Uighur; Uyghur
ukr	uk	Ukrainian
umb		Umbundu
und		Undetermined
urd	ur	Urdu
uzb	uz	Uzbek
vai		Vai
ven	ve	Venda
vie	vi	Vietnamese
vol	vo	Volapuk
vot		Votic
wak		Wakashan languages
wal		Wolaitta; Wolaytta
war		Waray
was		Washo
wel/cym	cy	Welsh
wen		Sorbian languages
wln	wa	Walloon
wol	wo	Wolof
xal		Kalmyk; Oirat
xho	xh	Xhosa
yao		Yao
yap		Yapese
yid	yi	Yiddish
yor	yo	Yoruba
ypk		Yupik languages
zap		Zapotec
zbl		Blissymbols; Blissymbolics; Bliss
zen		Zenaga
zha	za	Zhuang; Chuang
znd		Zande languages
zul	zu	Zulu
zun		Zuni
zxx		No linguistic content; Not applicable
zza		Zaza; Dimili; Dimli; Kirdki; Kirmanjki; Zazaki'! !

!ISOLanguageDefinition class methodsFor: 'private' stamp: 'mir 7/1/2004 18:07'!
readISOLanguagesFrom: stream
	"ISOLanguageDefinition readISOLanguagesFrom: ISOLanguageDefinition isoLanguages readStream "
	| languages language code3 index line |
	languages := Dictionary new.
	[stream atEnd
		or: [(line := stream nextLine readStream) atEnd]]
		whileFalse: [
			language := ISOLanguageDefinition new.
			code3 := line upTo: Character tab.
			(index := code3 indexOf: $/) > 0
				ifTrue: [
					language iso3: (code3 copyFrom: 1 to: index-1).
					language iso3Alternate: (code3 copyFrom: index+1 to: code3 size)]
				ifFalse: [language iso3: code3].
			language
				iso2: (line upTo: Character tab);
				language: line upToEnd.
			languages at: language iso3 put: language].
	^languages! !

!ISOLanguageDefinition methodsFor: 'accessing' stamp: 'mir 7/1/2004 18:20'!
iso2
	^iso2 ifNil: [self iso3]! !

!ISOLanguageDefinition methodsFor: 'initialize' stamp: 'mir 6/30/2004 15:54'!
iso2: aString
	iso2 := aString ifEmpty: [nil] ifNotEmpty: [aString]! !

!ISOLanguageDefinition methodsFor: 'accessing' stamp: 'mir 7/1/2004 18:21'!
iso3
	^iso3 ifNil: ['']! !

!ISOLanguageDefinition methodsFor: 'accessing' stamp: 'mir 6/30/2004 15:47'!
iso3Alternate
	^iso3Alternate ifNil: ['']! !

!ISOLanguageDefinition methodsFor: 'initialize' stamp: 'mir 6/30/2004 15:54'!
iso3Alternate: aString
	iso3Alternate := aString ifEmpty: [nil] ifNotEmpty: [aString]! !

!ISOLanguageDefinition methodsFor: 'initialize' stamp: 'mir 6/30/2004 15:54'!
iso3: aString
	iso3 := aString ifEmpty: [nil] ifNotEmpty: [aString]! !

!ISOLanguageDefinition methodsFor: 'accessing' stamp: 'mir 8/15/2003 13:13'!
language
	^language! !

!ISOLanguageDefinition methodsFor: 'initialize' stamp: 'mir 8/15/2003 13:40'!
language: aString
	language := aString! !

Object subclass: #Locale
	instanceVariableNames: 'id shortDate longDate time decimalSymbol digitGrouping currencySymbol currencyNotation measurement offsetLocalToUTC offsetVMToUTC dstActive'
	classVariableNames: 'Current CurrentPlatform KnownLocales LanguageSymbols LocaleChangeListeners PlatformEncodings Previous'
	poolDictionaries: ''
	category: 'System-Localization'!

!Locale commentStamp: '<historical>' prior: 0!
Main comment stating the purpose of this class and relevant relationship to other classes.



	http://www.w3.org/WAI/ER/IG/ert/iso639.htm
	http://www.oasis-open.org/cover/iso639a.html
	See also
	http://oss.software.ibm.com/cvs/icu/~checkout~/icuhtml/design/language_code_issues.html
	http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.10
	
ISO 3166
http://mitglied.lycos.de/buran/knowhow/codes/locales/
!

!Locale class methodsFor: 'notification' stamp: 'mir 6/30/2004 16:15'!
addLocalChangedListener: anObjectOrClass
	self localeChangedListeners add: anObjectOrClass! !

!Locale class methodsFor: 'accessing' stamp: 'mir 8/16/2003 16:02'!
clipboadInterpreter
	^NoConversionClipboardInterpreter new! !

!Locale class methodsFor: 'accessing' stamp: 'mir 8/31/2005 17:36'!
current
	"Current := nil"
	Current ifNil: [
		Current := self determineCurrentLocale.
		"Transcript show: 'Current locale: ' , Current localeID asString; cr"].
	^Current! !

!Locale class methodsFor: 'accessing' stamp: 'yo 7/28/2004 20:32'!
currentPlatform
	"CurrentPlatform := nil"
	CurrentPlatform ifNil: [CurrentPlatform := self determineCurrentLocale].
	^CurrentPlatform! !

!Locale class methodsFor: 'accessing' stamp: 'yo 7/28/2004 20:39'!
currentPlatform: locale
	CurrentPlatform := locale.
	LanguageEnvironment startUp.
! !

!Locale class methodsFor: 'accessing' stamp: 'tak 10/18/2005 22:33'!
currentPlatform: locale during: aBlock 
	"Alter current language platform during a block"
	| backupPlatform |
	backupPlatform := self currentPlatform.
	[self currentPlatform: locale.
	aBlock value]
		ensure: [self currentPlatform: backupPlatform]! !

!Locale class methodsFor: 'platform specific' stamp: 'fbs 7/25/2013 07:13'!
defaultEncodingName: languageSymbol 
	| encodings platformName osVersion |
	platformName := Smalltalk platformName.
	osVersion := Smalltalk getSystemAttribute: 1002.
	encodings := self platformEncodings at: languageSymbol
				ifAbsent: [self platformEncodings at: #default].
	encodings at: platformName ifPresent: [:encoding | ^encoding].
	encodings at: platformName , ' ' , osVersion
		ifPresent: [:encoding | ^encoding].
	^encodings at: #default! !

!Locale class methodsFor: 'platform specific' stamp: 'fbs 7/25/2013 07:13'!
defaultInputInterpreter
	| platformName osVersion |
	platformName := Smalltalk platformName.
	osVersion := Smalltalk getSystemAttribute: 1002.
	(platformName = 'Win32' and: [osVersion = 'CE']) 
		ifTrue: [^NoInputInterpreter new].
	platformName = 'Win32' ifTrue: [^MacRomanInputInterpreter new].
	^NoInputInterpreter new! !

!Locale class methodsFor: 'private' stamp: 'mir 7/28/2005 00:24'!
determineCurrentLocale
	"For now just return the default locale.
	A smarter way would be to determine the current platforms default locale."
	"Locale determineCurrentLocale"

	^self new determineLocale! !

!Locale class methodsFor: 'class initialization' stamp: 'cmm 2/25/2016 22:05'!
initialize
	"Locale initialize"
	Smalltalk addToStartUpList: Locale.
	Preferences
		addPreference: #useLocale
		categories: #('general' )
		default: false
		balloonHelp: 'Use the system locale to set the system language, etc., at startup.  For time-zone handling, see automaticTimezone.'! !

!Locale class methodsFor: 'class initialization' stamp: 'nk 8/29/2004 13:21'!
initializePlatformEncodings
	"Locale initializePlatformEncodings"

	| platform |
	PlatformEncodings isNil ifTrue: [ PlatformEncodings := Dictionary new ].

	platform := PlatformEncodings at: 'default' ifAbsentPut: Dictionary new.
	platform
		at: 'default' put: 'iso8859-1';
		at: 'Win32 CE' put: 'utf-8';
		yourself.

	platform := PlatformEncodings at: 'ja' ifAbsentPut: Dictionary new.
	platform
		at: 'default' put: 'shift-jis';
		at: 'unix' put: 'euc-jp';
		at: 'Win32 CE' put: 'utf-8';
		yourself.

	platform := PlatformEncodings at: 'ko' ifAbsentPut: Dictionary new.
	platform
		at: 'default' put: 'euc-kr';
		at: 'Win32 CE' put: 'utf-8';
		yourself.

	platform := PlatformEncodings at: 'zh' ifAbsentPut: Dictionary new.
	platform
		at: 'default' put: 'gb2312';
		at: 'unix' put: 'euc-cn';
		at: 'Win32 CE' put: 'utf-8';
		yourself.
! !

!Locale class methodsFor: 'private' stamp: 'tak 10/18/2007 21:14'!
initKnownLocales
	| locales |
	locales := Dictionary new.

	"Init the locales for which we have translations"
	InternalTranslator availableLanguageLocaleIDs do: [:id |
		locales at: id put: (Locale new localeID: id)].
	^locales! !

!Locale class methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:20'!
isoLanguage: isoLanguage
	^self isoLanguage: isoLanguage isoCountry: nil! !

!Locale class methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:31'!
isoLanguage: isoLanguage isoCountry: isoCountry
	^self localeID: (LocaleID  isoLanguage: isoLanguage isoCountry: isoCountry)! !

!Locale class methodsFor: 'accessing' stamp: 'mir 7/15/2004 12:42'!
isoLocale: aString
	! !

!Locale class methodsFor: 'private' stamp: 'mir 7/15/2004 16:44'!
knownLocales
	"KnownLocales := nil"
	^KnownLocales ifNil: [KnownLocales := self initKnownLocales]! !

!Locale class methodsFor: 'accessing' stamp: 'mir 7/13/2004 00:24'!
languageSymbol: languageSymbol
	"Locale languageSymbol: #Deutsch"

	^self isoLanguage: (LanguageSymbols at: languageSymbol)! !

!Locale class methodsFor: 'notification' stamp: 'tfel 7/28/2016 17:35'!
localeChanged
	SystemNavigation default allBehaviorsDo: [:b |
		b == self ifFalse: [b localeChanged]].! !

!Locale class methodsFor: 'notification' stamp: 'tfel 8/8/2016 10:28'!
localeChangedGently
	SystemNavigation default allBehaviorsDo: [:b | b == self ifFalse: [b localeChangedGently]].! !

!Locale class methodsFor: 'notification' stamp: 'mir 6/30/2004 16:15'!
localeChangedListeners
	^LocaleChangeListeners ifNil: [LocaleChangeListeners := OrderedCollection new]! !

!Locale class methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:30'!
localeID: id
	^self knownLocales at: id ifAbsentPut: [Locale new localeID: id]! !

!Locale class methodsFor: 'class initialization' stamp: 'nk 8/29/2004 13:20'!
platformEncodings
	PlatformEncodings isEmptyOrNil ifTrue: [ self initializePlatformEncodings ].
	^PlatformEncodings
! !

!Locale class methodsFor: 'accessing' stamp: 'yo 7/16/2007 13:52'!
previous

	^ Previous
! !

!Locale class methodsFor: 'private' stamp: 'brp 5/25/2005 11:00'!
resetKnownLocales

	KnownLocales := nil
! !

!Locale class methodsFor: 'system startUp' stamp: 'bf 2/23/2016 13:01'!
startUp: resuming
	resuming ifFalse: [^self].
	(Preferences valueOfFlag: #useLocale)
		ifTrue: [
			| newID |
			newID := self current determineLocaleID.
			newID ~= LocaleID current
				ifTrue: [self switchAndInstallFontToID: newID gently: true]]! !

!Locale class methodsFor: 'accessing' stamp: 'dgd 10/7/2004 20:50'!
stringForLanguageNameIs: localeID 
	"Answer a string for a menu determining whether the given  
	symbol is the project's natural language"
	^ (self current localeID = localeID
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, localeID displayName! !

!Locale class methodsFor: 'accessing' stamp: 'tak 8/5/2005 21:11'!
switchAndInstallFontToID: localeID 
	"Locale switchAndInstallFontToID: (LocaleID isoLanguage: 'de')"
	| locale |
	locale := Locale localeID: localeID.
	locale languageEnvironment isFontAvailable
		ifFalse: [(self confirm: 'This language needs additional fonts.
Do you want to install the fonts?' translated)
				ifTrue: [locale languageEnvironment installFont]
				ifFalse: [^ self]].
	self
		switchTo: locale! !

!Locale class methodsFor: 'accessing' stamp: 'nice 3/24/2012 18:22'!
switchAndInstallFontToID: localeID gently: gentlyFlag
	| locale result env envFound ret fontInImage menu |
	"Logic:
		- not in Rainbow, 
			- if the font is in the image, use the font.
			- if the font is not in the image,
				- ask to choose:
					- load font
					- try to enable pango, if pango is available
					- or cancel.
			- if the previous step fails, notify the user that you cannot switch to the language.
	Details:
		- how to check non-pango font is available:
			- if the language environment for the locale doesn't exist, the font is not available.
			- if font loading fails, it is not available.
		- how to check if the language environment doesn't exist:
			- if the locales languageEnvironment is Latin1 but the locale isn't it is not available.
	"	
	locale := Locale localeID: localeID.
	env := locale languageEnvironment.
	result := true.
	envFound := (Latin1Environment supportedLanguages includes: locale isoLanguage) or: [(env isMemberOf: Latin1Environment) not].
	fontInImage := envFound and: [env isFontAvailable].
	fontInImage ifFalse: [
		menu := MenuMorph new.
		menu defaultTarget: menu.
		envFound ifTrue: [menu add: 'load font' translated selector: #modalSelection: argument: #loadFont].
		menu add:  'cancel' translated selector: #modalSelection: argument: #cancel.
		menu addTitle: 'This language needs additional fonts.
Do you want to install the fonts?' translated.
		ret := menu invokeModal.
		ret = #loadFont ifTrue: [result := env installFont. result ifTrue: [StrikeFont setupDefaultFallbackTextStyle]].
		(ret ~~ #loadFont and: [ret ~~ #enablePango]) ifTrue: [result := false]].
	result ifFalse: [self inform: 'Cannot load additional fonts' translated] ifTrue: [self switchTo: locale gently: gentlyFlag].
! !

!Locale class methodsFor: 'accessing' stamp: 'mir 7/15/2004 19:07'!
switchToID: localeID
	"Locale switchToID: (LocaleID isoLanguage: 'de') "

	self switchTo: (Locale localeID: localeID)! !

!Locale class methodsFor: 'accessing' stamp: 'yo 7/16/2007 15:00'!
switchTo: locale
	self switchTo: locale gently: false.
! !

!Locale class methodsFor: 'accessing' stamp: 'KR 10/23/2007 19:39'!
switchTo: locale gently: gentlyFlag
	"Locale switchTo: (Locale isoLanguage: 'de')"
	| availableID |
	availableID := (NaturalLanguageTranslator availableForLocaleID: locale localeID) localeID.
	Current localeID = availableID
		ifFalse: [Previous := Current.
				CurrentPlatform := Current := Locale localeID: availableID.
				NaturalLanguageTranslator localeChanged.
				gentlyFlag ifTrue: [self localeChangedGently] ifFalse: [self localeChanged]]! !

!Locale methodsFor: 'accessing' stamp: 'mir 8/31/2005 17:03'!
determineLocale
	self localeID: self determineLocaleID! !

!Locale methodsFor: 'accessing' stamp: 'tak 9/8/2005 01:45'!
determineLocaleID
	"Locale current determineLocaleID"
	| isoLang isoCountry |
	isoLang := self fetchISO2Language
				ifNil: [^ self localeID].
	isoCountry := self fetchISOCountry
				ifNil: [^ LocaleID isoLanguage: isoLang].
	^ LocaleID isoLanguage: isoLang isoCountry: isoCountry! !

!Locale methodsFor: 'private' stamp: 'mir 8/31/2005 17:36'!
fetchISO2Language
	"Locale current fetchISO2Language"
	| lang isoLang |
	lang := self primLanguage.
	lang ifNil: [^nil].
	lang := lang copyUpTo: 0 asCharacter.
	lang size == 2
		ifTrue: [^lang].
	isoLang := ISOLanguageDefinition iso3LanguageDefinition: lang.
	^isoLang
		ifNil: [nil]
		ifNotNil: [isoLang iso2]! !

!Locale methodsFor: 'private' stamp: 'tak 9/8/2005 01:47'!
fetchISOCountry
	"Locale current fetchISOCountry"
	| countryCode |
	countryCode := self primCountry
				ifNil: [^ nil].
	^ countryCode copyUpTo: 0 asCharacter! !

!Locale methodsFor: 'accessing' stamp: 'yo 9/12/2006 14:00'!
iconForNativeLanguage

	^ (NaturalLanguageFormTranslator localeID: self localeID) translate: 'LanguageNameInNativeLanguage'.
! !

!Locale methodsFor: 'accessing' stamp: 'nk 8/31/2004 09:39'!
isoCountry
	^self localeID isoCountry! !

!Locale methodsFor: 'accessing' stamp: 'nk 8/31/2004 09:39'!
isoLanguage
	^self localeID isoLanguage! !

!Locale methodsFor: 'accessing' stamp: 'mir 7/15/2004 12:41'!
isoLocale
	"<language>-<country>"
	^self isoCountry
		ifNil: [self isoLanguage]
		ifNotNil: [self isoLanguage , '-' , self isoCountry]! !

!Locale methodsFor: 'accessing' stamp: 'mir 7/15/2004 15:52'!
languageEnvironment
	^LanguageEnvironment localeID: self localeID! !

!Locale methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:55'!
localeID
	^id! !

!Locale methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:55'!
localeID: anID
	id := anID! !

!Locale methodsFor: 'accessing' stamp: 'bf 8/26/2008 16:04'!
offsetLocalToUTC
	^self primTimezone! !

!Locale methodsFor: 'system primitives' stamp: 'bf 8/26/2008 14:54'!
primCountry
	"Returns string with country tag according to ISO 639"
	<primitive: 'primitiveCountry' module: 'LocalePlugin'>
	^'US'! !

!Locale methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 18:45'!
primCurrencyNotation
	"Returns boolean if symbol is pre- (true) or post-fix (false)"
	<primitive: 'primitiveCurrencyNotation' module: 'LocalePlugin'>
	^true! !

!Locale methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 18:47'!
primCurrencySymbol
	"Returns string with currency symbol"
	<primitive: 'primitiveCurrencySymbol' module:'LocalePlugin'>
	^'$'! !

!Locale methodsFor: 'system primitives' stamp: 'tpr 6/2/2005 13:42'!
primDecimalSymbol
	"Returns string with e.g. '.' or ','"
	<primitive:'primitiveDecimalSymbol' module: 'LocalePlugin'>
	^'.'! !

!Locale methodsFor: 'system primitives' stamp: 'tpr 6/2/2005 13:42'!
primDigitGrouping
	"Returns string with e.g. '.' or ',' (thousands etc)"
	<primitive:'primitiveDigitGroupingSymbol' module: 'LocalePlugin'>
	^','! !

!Locale methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 18:48'!
primDST
	"Returns boolean if DST  (daylight saving time) is active or not"
	<primitive:'primitiveDaylightSavings' module: 'LocalePlugin'>
	^false! !

!Locale methodsFor: 'system primitives' stamp: 'bf 8/26/2008 14:57'!
primLanguage
	"returns string with language tag according to ISO 639"
	<primitive:'primitiveLanguage' module: 'LocalePlugin'>
	^'en'

! !

!Locale methodsFor: 'system primitives' stamp: 'bf 8/26/2008 15:05'!
primLongDateFormat
	"Returns the long date format
	d day, m month, y year,
	double symbol is null padded, single not padded (m=6, mm=06)
	dddd weekday
	mmmm month name"
	<primitive:'primitiveLongDateFormat' module: 'LocalePlugin'>
	^'dddd, mmmm d, yyyy'! !

!Locale methodsFor: 'system primitives' stamp: 'tpr 6/2/2005 13:43'!
primMeasurement
	"Returns boolean denoting metric(true) or imperial(false)."
	<primitive:'primitiveMeasurementMetric' module: 'LocalePlugin'>
	^true
! !

!Locale methodsFor: 'system primitives' stamp: 'bf 8/26/2008 15:05'!
primShortDateFormat
	"Returns the short date format
	d day, m month, y year,
	double symbol is null padded, single not padded (m=6, mm=06)
	dddd weekday
	mmmm month name"
	<primitive:'primitiveShortDateFormat' module: 'LocalePlugin'>
	^'m/d/yy'! !

!Locale methodsFor: 'system primitives' stamp: 'bf 8/26/2008 15:07'!
primTimeFormat
	"Returns string time format
	Format is made up of 
	h hour (h 12, H 24), m minute, s seconds, x (am/pm String)
	double symbol is null padded, single not padded (h=6, hh=06)"
	<primitive:'primitiveTimeFormat' module: 'LocalePlugin'>
	^'h:mmx'! !

!Locale methodsFor: 'system primitives' stamp: 'bf 8/26/2008 15:45'!
primTimezone
	"The offset from UTC in minutes, with positive offsets being towards the east.
	(San Francisco is in UTC -07*60 and Paris is in UTC +02*60 when daylight savings is in effect)."
	<primitive:'primitiveTimezoneOffset' module: 'LocalePlugin'>
	^0! !

!Locale methodsFor: 'system primitives' stamp: 'tpr 6/2/2005 13:44'!
primVMOffsetToUTC
	"Returns the offset in minutes between the VM and UTC.
	If the VM does not support UTC times, this is 0.
	Also gives us backward compatibility with old VMs as the primitive will fail and we then can return 0."
	<primitive:'primitiveVMOffsetToUTC' module: 'LocalePlugin'>
	^0! !

!Locale methodsFor: 'accessing' stamp: 'tak 8/4/2005 15:18'!
printOn: aStream 
	super printOn: aStream.
	aStream nextPutAll: '(' , id printString , ')'! !

Object subclass: #LocaleID
	instanceVariableNames: 'isoLanguage isoCountry'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Localization'!

!LocaleID class methodsFor: 'accessing' stamp: 'cmm 11/12/2014 14:58'!
countryFor: iso6391Code
	"http://www.loc.gov/standards/iso639-2/php/code_list.php"
	^ iso6391Code
		caseOf:
			{ ['af'] -> ['Afrikaans'].
			['ca'] -> ['Catalan'].
			['cs'] -> [ 'Czech'].
			['da'] -> [ 'Danish'].
			['de'] -> [ 'German'].
			['el'] -> [ 'Greek Modern'].
			['en'] -> [ 'English'].
			['es'] -> [ 'Spanish'].
			['eu'] -> [ 'Basque'].
			['fi'] -> [ 'Finnish'].
			['fo'] -> [ 'Faroese'].
			['fr'] -> [ 'French'].
			['ga'] -> [ 'Irish'].
			['gd'] -> [ 'Gaelic'].
			['hr'] -> [ 'Croatian'].
			['hu'] -> [ 'Hungarian'].
			['is'] -> [ 'Icelandic'].
			['it'] -> [ 'Italian'].
			['ja'] -> [ 'Japanese'].
			['ja-etoys'] -> [ 'Japanese'].
			['ko'] -> [ 'Korean'].
			['nl'] -> [ 'Dutch'].
			['no'] -> [ 'Norwegian'].
			['pt'] -> [ 'Portuguese'].
			['rm'] -> [ 'Romansh'].
			['ro'] -> [ 'Romainian'].
			['sk'] -> [ 'Slovak'].
			['sl'] -> [ 'Slovenian'].
			['sq'] -> [ 'Albanian'].
			['sv'] -> [ 'Swedish'].
			['sw'] -> [ 'Swahili'].
			['zh'] -> [ 'Chinese'] }
		otherwise:
			[ 'other' ]! !

!LocaleID class methodsFor: 'accessing' stamp: 'mir 7/15/2004 15:09'!
current
	^Locale current localeID! !

!LocaleID class methodsFor: 'instance creation' stamp: 'mir 7/15/2004 14:35'!
isoLanguage: langString
	^self isoLanguage: langString isoCountry: nil! !

!LocaleID class methodsFor: 'instance creation' stamp: 'mir 7/15/2004 12:46'!
isoLanguage: langString isoCountry: countryStringOrNil
	^self new isoLanguage: langString isoCountry: countryStringOrNil! !

!LocaleID class methodsFor: 'instance creation' stamp: 'mir 7/21/2004 13:59'!
isoString: isoString
	"Parse the isoString (<language>-<country>) into its components and return the matching LocaleID"
	"LocaleID isoString: 'en' "
	"LocaleID isoString: 'en-us' "

	| parts language country |
	parts := isoString findTokens: #($- ).
	language := parts first.
	parts size > 1
		ifTrue: [country := parts second].
	^self isoLanguage: language isoCountry: country! !

!LocaleID class methodsFor: 'instance creation' stamp: 'tfel 7/28/2016 16:46'!
posixName: aString 
	^ self
		isoString: (aString copyReplaceAll: '_' with: '-')! !

!LocaleID class methodsFor: 'accessing' stamp: 'bf 9/26/2007 16:24'!
previous
	^Locale previous localeID! !

!LocaleID methodsFor: 'accessing' stamp: 'mir 9/1/2005 14:17'!
displayCountry
	^(ISOLanguageDefinition iso2Countries at: self isoCountry asUppercase ifAbsent: [ self isoCountry ]) ! !

!LocaleID methodsFor: 'accessing' stamp: 'tak 3/23/2006 12:26'!
displayLanguage
	| language |
	language := (ISOLanguageDefinition iso2LanguageTable
				at: self isoLanguage
				ifAbsent: [^ self isoLanguage]) language.
	^ self isoCountry
		ifNil: [language]
		ifNotNil: [language , ' (' , self displayCountry , ')']! !

!LocaleID methodsFor: 'accessing' stamp: 'tfel 7/28/2016 16:10'!
displayName
	"Answer a proper name to represent the receiver in GUI. 
	 
	The wording is provided by translations of the magic value 
	'<language display name>'. 
	 
	'English' -> 'English'  
	'German' -> 'Deutsch'  
	"
	| magicPhrase translatedMagicPhrase |
	magicPhrase := '<language display name>'.
	translatedMagicPhrase := NaturalLanguageTranslator translateWithoutLoading: magicPhrase toLocaleID: self.
	^ translatedMagicPhrase = magicPhrase
		ifTrue: [self displayLanguage]
		ifFalse: [translatedMagicPhrase]! !

!LocaleID methodsFor: 'comparing' stamp: 'mir 7/15/2004 14:23'!
hash
	^self isoLanguage hash bitXor: self isoCountry hash! !

!LocaleID methodsFor: 'testing' stamp: 'mir 7/15/2004 14:34'!
hasParent
	^self isoCountry notNil! !

!LocaleID methodsFor: 'accessing' stamp: 'mir 7/15/2004 12:43'!
isoCountry
	^isoCountry! !

!LocaleID methodsFor: 'accessing' stamp: 'tfel 7/28/2016 16:27'!
isoCountry: aValue

	isoCountry := aValue! !

!LocaleID methodsFor: 'accessing' stamp: 'mir 7/15/2004 12:43'!
isoLanguage
	^isoLanguage! !

!LocaleID methodsFor: 'initialize' stamp: 'cmm 11/12/2014 15:09'!
isoLanguage: langString isoCountry: countryStringOrNil 
	isoLanguage := langString.
	isoCountry := countryStringOrNil ifNil: (self class countryFor: langString)! !

!LocaleID methodsFor: 'accessing' stamp: 'mir 7/21/2004 19:17'!
isoString
	^self asString! !

!LocaleID methodsFor: 'accessing' stamp: 'tfel 7/28/2016 16:33'!
parent
	^ (self class
			isoLanguage: self isoLanguage)
			isoCountry: nil;
			yourself! !

!LocaleID methodsFor: 'printing' stamp: 'tfel 7/28/2016 16:18'!
posixName
	"(LocaleID isoString: 'es-MX') posixName"
	"(LocaleID isoString: 'es') posixName"
	"language[_territory]"
	^ self isoCountry
		ifNil: [self isoLanguage]
		ifNotNil: [self isoLanguage , '_' , self isoCountry]! !

!LocaleID methodsFor: 'printing' stamp: 'mir 7/15/2004 12:45'!
printOn: stream
	"<language>-<country>"
	stream nextPutAll: self isoLanguage.
	self isoCountry
		ifNotNil: [stream nextPut: $-; nextPutAll: self isoCountry]! !

!LocaleID methodsFor: 'printing' stamp: 'tak 11/15/2004 12:45'!
storeOn: aStream 
	aStream nextPut: $(.
	aStream nextPutAll: self class name.
	aStream nextPutAll: ' isoString: '.
	aStream nextPutAll: '''' , self printString , ''''.
	aStream nextPut: $).
! !

!LocaleID methodsFor: 'accessing' stamp: 'tak 10/18/2007 21:15'!
translator
	^ InternalTranslator localeID: self ! !

!LocaleID methodsFor: 'comparing' stamp: 'mir 7/15/2004 14:23'!
= anotherObject
	self class == anotherObject class
		ifFalse: [^false].
	^self isoLanguage = anotherObject isoLanguage
		and: [self isoCountry = anotherObject isoCountry]! !

Object subclass: #MczInstaller
	instanceVariableNames: 'stream zip'
	classVariableNames: 'Versions'
	poolDictionaries: ''
	category: 'System-Support'!

!MczInstaller class methodsFor: 'versionInfo' stamp: 'avi 1/19/2004 13:13'!
clearVersionInfo
	Versions := Dictionary new! !

!MczInstaller class methodsFor: 'services' stamp: 'cwp 8/7/2003 18:49'!
extension
	^ 'mcz'! !

!MczInstaller class methodsFor: 'services' stamp: 'nk 6/8/2004 17:29'!
fileReaderServicesForFile: fileName suffix: suffix
	^({ self extension. '*' } includes: suffix)
		ifTrue: [ self services ]
		ifFalse: [#()].
! !

!MczInstaller class methodsFor: 'services' stamp: 'avi 3/7/2004 14:51'!
initialize
	self clearVersionInfo.
	self registerForFileList.! !

!MczInstaller class methodsFor: 'installing' stamp: 'cwp 8/7/2003 18:13'!
installFileNamed: aFileName
	self installStream: (FileStream readOnlyFileNamed: aFileName)! !

!MczInstaller class methodsFor: 'installing' stamp: 'cwp 8/7/2003 17:56'!
installStream: aStream
	(self on: aStream) install! !

!MczInstaller class methodsFor: 'services' stamp: 'cwp 8/7/2003 18:54'!
loadVersionFile: fileName
	self installFileNamed: fileName
! !

!MczInstaller class methodsFor: 'instance creation' stamp: 'cwp 8/7/2003 17:56'!
on: aStream
	^ self new stream: aStream! !

!MczInstaller class methodsFor: 'services' stamp: 'fbs 8/9/2013 17:23'!
registerForFileList
	Smalltalk at: #MCReader ifAbsent: [FileServices registerFileReader: self]! !

!MczInstaller class methodsFor: 'services' stamp: 'KR 3/21/2008 20:57'!
serviceLoadVersion
	^ SimpleServiceEntry
		provider: self
		label: 'load' translatedNoop
		selector: #loadVersionFile:
		description: 'load a package version' translatedNoop! !

!MczInstaller class methodsFor: 'services' stamp: 'ab 8/8/2003 18:01'!
services
	^ Array with: self serviceLoadVersion! !

!MczInstaller class methodsFor: 'versionInfo' stamp: 'cwp 8/11/2003 23:49'!
storeVersionInfo: aVersion
	Versions 
		at: aVersion package name
		put: aVersion info asDictionary! !

!MczInstaller class methodsFor: 'versionInfo' stamp: 'avi 3/7/2004 14:51'!
unloadMonticello
	"self unloadMonticello"
	Utilities breakDependents.
	
	Smalltalk at: #MCWorkingCopy ifPresent:
		[:wc | 
		wc allInstances do:
			[:ea | 
			Versions at: ea package name put: ea currentVersionInfo asDictionary.
			ea breakDependents.
			Smalltalk at: #SystemChangeNotifier ifPresent: [:scn | scn uniqueInstance noMoreNotificationsFor: ea]]
	displayingProgress: 'Saving version info...'].
	
	"keep things simple and don't unload any class extensions"
	(ChangeSet superclassOrder: ((PackageInfo named: 'Monticello') classes)) reverseDo:
		[:ea | 
		ea removeFromSystem].
	
	self registerForFileList.! !

!MczInstaller class methodsFor: 'versionInfo' stamp: 'avi 2/17/2004 02:49'!
versionInfo
	^ Versions! !

!MczInstaller methodsFor: 'utilities' stamp: 'nice 12/7/2009 09:08'!
associate: tokens
	| result |
	result := Dictionary new.
	tokens pairsDo: [:key :value | 
					| overwrittenValue |
					overwrittenValue := value.
					value isString ifFalse: [overwrittenValue := value collect: [:ea | self associate: ea]].
					value == 'nil' ifTrue: [overwrittenValue := ''].
					result at: key put: overwrittenValue].
	^ result! !

!MczInstaller methodsFor: 'utilities' stamp: 'avi 2/17/2004 02:53'!
checkDependencies
	| dependencies unmet |
	dependencies := (zip membersMatching: 'dependencies/*') 
			collect: [:member | self extractInfoFrom: (self parseMember: member)].
	unmet := dependencies reject: [:dep |
		self versions: Versions anySatisfy: (dep at: #id)].
	^ unmet isEmpty or: [
		self confirm: (String streamContents: [:s|
			s nextPutAll: 'The following dependencies seem to be missing:'; cr.
			unmet do: [:each | s nextPutAll: (each at: #name); cr].
			s nextPutAll: 'Do you still want to install this package?'])]! !

!MczInstaller methodsFor: 'private' stamp: 'nice 6/2/2013 18:43'!
contentsForMember: member
	^[(member contentStreamFromEncoding: 'utf8') text contents] on: InvalidUTF8
		do: [:exc | 
			"Case of legacy encoding, presumably it is latin-1.
			But if contents starts with a null character, it might be a case of WideString encoded in UTF-32BE"
			| str |
			str := (member contentStreamFromEncoding: 'latin1') text..
			exc return: ((str peek = Character null and: [ str size \\ 4 = 0 ])
				ifTrue: [WideString fromByteArray: str contents asByteArray]
				ifFalse: [str contents])]! !

!MczInstaller methodsFor: 'utilities' stamp: 'avi 2/17/2004 03:26'!
extractInfoFrom: dict
	dict at: #id put: (UUID fromString: (dict at: #id)).
	dict at: #date ifPresent: [:d | d isEmpty ifFalse: [dict at: #date put: (Date fromString: d)]].
	dict at: #time ifPresent: [:t | t isEmpty ifFalse: [dict at: #time put: (Time readFrom: t readStream)]].
	dict at: #ancestors ifPresent: [:a | dict at: #ancestors put: (a collect: [:ea | self extractInfoFrom: ea])].
	^ dict! !

!MczInstaller methodsFor: 'accessing' stamp: 'cwp 8/13/2003 01:58'!
extractPackageName
	^ (self parseMember: 'package') at: #name.
	! !

!MczInstaller methodsFor: 'accessing' stamp: 'cwp 8/13/2003 02:17'!
extractVersionInfo
	^ self extractInfoFrom: (self parseMember: 'version')! !

!MczInstaller methodsFor: 'installation' stamp: 'avi 2/17/2004 02:56'!
install
	| sources |
	zip := ZipArchive new.
	zip readFrom: stream.
	self checkDependencies ifFalse: [^false].
	self recordVersionInfo.
	sources := (zip membersMatching: 'snapshot/*') 
				asSortedCollection: [:a :b | a fileName < b fileName].
	sources do: [:src | self installMember: src].! !

!MczInstaller methodsFor: 'installation' stamp: 'nice 6/2/2013 16:13'!
installMember: member
	 
	self useNewChangeSetDuring:
		[ | contents |
		contents := self contentsForMember: member.
		contents readStream fileInAnnouncing: 'loading ', member fileName.
		]! !

!MczInstaller methodsFor: 'utilities' stamp: 'nice 6/2/2013 22:43'!
parseMember: memberOrName
	| member contents tokens |
	member := zip member: memberOrName.
	contents := self contentsForMember: member.
	tokens := (self scanner scanTokens: contents) first.
	^ self associate: tokens! !

!MczInstaller methodsFor: 'accessing' stamp: 'cwp 8/7/2003 19:18'!
recordVersionInfo
	Versions 
		at: self extractPackageName 
		put: self extractVersionInfo! !

!MczInstaller methodsFor: 'accessing' stamp: 'cwp 8/13/2003 02:04'!
scanner
	^ Scanner new! !

!MczInstaller methodsFor: 'accessing' stamp: 'avi 2/17/2004 02:55'!
stream: aStream
	stream := aStream! !

!MczInstaller methodsFor: 'utilities' stamp: 'ul 8/2/2011 20:03'!
useNewChangeSetDuring: aBlock
	| changeHolder oldChanges newChanges |
	changeHolder := (ChangeSet respondsTo: #newChanges:)
						ifTrue: [ChangeSet]
						ifFalse: [Smalltalk].
	oldChanges := (ChangeSet respondsTo: #current)
						ifTrue: [ChangeSet current]
						ifFalse: [Smalltalk changes].

	newChanges := ChangeSet new name: (ChangeSet uniqueNameLike: self extractPackageName).
	changeHolder newChanges: newChanges.
	aBlock ensure: [changeHolder newChanges: oldChanges].! !

!MczInstaller methodsFor: 'utilities' stamp: 'bf 2/9/2004 15:00'!
versions: aVersionList anySatisfy: aDependencyID
	^ aVersionList anySatisfy: [:version | 
			aDependencyID = (version at: #id)
				or: [self versions: (version at: #ancestors) anySatisfy: aDependencyID]]! !

Object subclass: #MethodChangeRecord
	instanceVariableNames: 'changeType currentMethod infoFromRemoval'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Changes'!

!MethodChangeRecord commentStamp: '<historical>' prior: 0!
MethodChangeRecords are used to record method changes.  Here is a simple summary of the relationship between the changeType symbol and the recording of prior state

			|	prior == nil			|	prior not nil	
	---------	|----------------------------	|--------------------
	add		|	add					|	change
	---------	|----------------------------	|--------------------
	remove	|	addedThenRemoved	|	remove

Structure:
changeType			symbol -- as summarized above
currentMethod	method
				This is the current version of the method.
				It can be used to assert this change upon entry to a layer. 
infoFromRemoval -- an array of size 2.
				The first element is the source index of the last version of the method.
				The second element is the category in which it was defined, so it
				can be put back there if re-accepted from a version browser.

Note that the above states each have an associated revoke action:
	add --> remove
	change --> change back
	remove --> add back
	addedThenRemoved --> no change
However all of these are accomplished trivially by restoring the original method dictionary.!

!MethodChangeRecord methodsFor: 'as yet unclassified' stamp: 'di 4/1/2000 12:02'!
changeType

	^ changeType! !

!MethodChangeRecord methodsFor: 'as yet unclassified' stamp: 'di 3/28/2000 23:34'!
currentMethod

	^ currentMethod! !

!MethodChangeRecord methodsFor: 'as yet unclassified' stamp: 'di 4/1/2000 12:02'!
methodInfoFromRemoval
	"Return an array with the source index of the last version of the method,
	and the category in which it was defined (so it can be put back there if
	re-accepted from a version browser)."

	(changeType == #remove or: [changeType == #addedThenRemoved])
		ifTrue: [^ infoFromRemoval]
		ifFalse: [^ nil]! !

!MethodChangeRecord methodsFor: 'as yet unclassified' stamp: 'di 4/4/2000 11:05'!
noteChangeType: newChangeType

	(changeType == #addedThenRemoved and: [newChangeType == #change])
		ifTrue: [changeType := #add]
		ifFalse: [changeType := newChangeType]! !

!MethodChangeRecord methodsFor: 'as yet unclassified' stamp: 'di 4/1/2000 11:05'!
noteMethodInfoFromRemoval: info
	"Store an array with the source index of the last version of the method,
	and the category in which it was defined (so it can be put back there if
	re-accepted from a version browser)."

	infoFromRemoval := info! !

!MethodChangeRecord methodsFor: 'as yet unclassified' stamp: 'ar 6/3/2005 17:01'!
noteNewMethod: newMethod
	"NEVER do this. It is evil."
	currentMethod := nil.! !

!MethodChangeRecord methodsFor: 'as yet unclassified' stamp: 'di 4/1/2000 12:02'!
printOn: strm

	super printOn: strm.
	strm nextPutAll: ' ('; print: changeType; nextPutAll: ')'! !

!MethodChangeRecord methodsFor: 'as yet unclassified' stamp: 'di 4/1/2000 10:47'!
priorMethod: ignored

	"We do not save original versions of changed methods because we only
	revoke changes at the level of entire classes, and that is done by
	restoration of the entire methodDictionary."! !

!MethodChangeRecord methodsFor: 'as yet unclassified' stamp: 'ar 5/23/2001 16:16'!
storeDataOn: aDataStream
	| oldMethod |
	oldMethod := currentMethod.
	currentMethod := nil.
	super storeDataOn: aDataStream.
	currentMethod := oldMethod.
! !

Object subclass: #MethodReference
	instanceVariableNames: 'classSymbol classIsMeta methodSymbol stringVersion category environment'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Tools'!

!MethodReference commentStamp: 'tlk 5/9/2006 18:43' prior: 0!
A MethodReference is is a lightweight proxy for a CompiledMethod.  Has methods for pointed to the CompileMethod's source statements, byte codes. Is heavily used my Tools.

Instance Variables
	classIsMeta:		     Boolean class vs. instance
	classSymbol:		Symbol for method's class (without class keyword if meta)
	methodSymbol:		Symbol for method's selector
	stringVersion:		'Class>>selector:' format

!

!MethodReference class methodsFor: 'instance creation' stamp: 'fbs 5/21/2013 22:27'!
class: aClass selector: aSelector
	^ self class: aClass selector: aSelector environment: Smalltalk globals.! !

!MethodReference class methodsFor: 'instance creation' stamp: 'fbs 5/21/2013 22:25'!
class: aClass selector: aSelector environment: anEnvironment
	^ self new setStandardClass: aClass methodSymbol: aSelector environment: anEnvironment.! !

!MethodReference methodsFor: 'accessing' stamp: 'fbs 5/22/2013 07:28'!
actualClass 
	^self environment at: classSymbol ifPresent: [ :actualClass |
		classIsMeta
			ifTrue: [ actualClass classSide ]
			ifFalse: [ actualClass ] ]

! !

!MethodReference methodsFor: 'converting' stamp: 'ul 7/6/2011 02:29'!
asString

	^stringVersion ifNil: [ self stringVersionDefault ]! !

!MethodReference methodsFor: 'converting' stamp: 'cmm 2/9/2014 17:45'!
asValid
	"Sometimes persistent MethodReferences may become stale after a refactoring which moved some methods to a superclass.  This method answers the new valid MethodReference if that happened."
	^ self isValid
		ifTrue: [ self ]
		ifFalse:
			[ | cm |
			cm := self actualClass lookupSelector: self selector.
			cm ifNotNil: [ cm methodReference ] ]! !

!MethodReference methodsFor: 'accessing' stamp: 'ar 12/12/2009 17:43'!
category
	"Answers the class category (cached for reuse via MC and other tools)"
	^ category ifNil: [category := self actualClass organization categoryOfElement: methodSymbol]! !

!MethodReference methodsFor: 'initialize-release' stamp: 'mtf 10/4/2008 00:52'!
category: aSymbol
	category := aSymbol! !

!MethodReference methodsFor: 'testing' stamp: 'RAA 5/28/2001 08:11'!
classIsMeta

	^classIsMeta! !

!MethodReference methodsFor: 'accessing' stamp: 'RAA 5/28/2001 08:10'!
classSymbol

	^classSymbol! !

!MethodReference methodsFor: 'accessing' stamp: 'ab 8/22/2003 17:58'!
compiledMethod
	^ self actualClass compiledMethodAt: methodSymbol! !

!MethodReference methodsFor: 'accessing' stamp: 'fbs 5/22/2013 07:27'!
environment
	^ environment ifNil: [
		"We will probably have MethodReferences already instantiated when this commit lands. We lazily move these over to the new, Environment aware, order of things."
		environment := Smalltalk globals].! !

!MethodReference methodsFor: 'comparing' stamp: 'dgd 3/8/2003 11:54'!
hash
	"Answer a SmallInteger whose value is related to the receiver's  
	identity."
	^ (self species hash bitXor: self classSymbol hash)
		bitXor: self methodSymbol hash! !

!MethodReference methodsFor: 'testing' stamp: 'md 8/27/2005 17:17'!
isValid
	"Answer whether the receiver represents a current selector or Comment"

	| aClass |
	methodSymbol isDoIt ifTrue: [^ false].
	(aClass := self actualClass) ifNil: [^ false].
	^ (aClass includesSelector: methodSymbol) or:
		[methodSymbol == #Comment]! !

!MethodReference methodsFor: 'accessing' stamp: 'RAA 5/28/2001 08:10'!
methodSymbol

	^methodSymbol! !

!MethodReference methodsFor: 'printing' stamp: 'fbs 5/21/2013 21:15'!
printOn: aStream
	| actualClass |
	"Print the receiver on a stream"
	actualClass := classSymbol asString.
	classIsMeta ifTrue: [actualClass := actualClass, ' class'].
	super printOn: aStream.
	aStream nextPutAll: ' ', actualClass, ' >> ', methodSymbol printString.! !

!MethodReference methodsFor: 'accessing' stamp: 'eem 5/25/2010 17:56'!
selector

	^methodSymbol! !

!MethodReference methodsFor: 'initialize-release' stamp: 'RAA 5/28/2001 07:34'!
setClassAndSelectorIn: csBlock

	^csBlock value: self actualClass value: methodSymbol! !

!MethodReference methodsFor: 'initialize-release' stamp: 'RAA 5/28/2001 06:04'!
setClassSymbol: classSym classIsMeta: isMeta methodSymbol: methodSym stringVersion: aString 

	classSymbol := classSym.
	classIsMeta := isMeta.
	methodSymbol := methodSym.
	stringVersion := aString.! !

!MethodReference methodsFor: 'initialize-release' stamp: 'RAA 5/28/2001 08:06'!
setClass: aClass methodSymbol: methodSym stringVersion: aString 

	classSymbol := aClass theNonMetaClass name.
	classIsMeta := aClass isMeta.
	methodSymbol := methodSym.
	stringVersion := aString.! !

!MethodReference methodsFor: 'initialize-release' stamp: 'kph 12/16/2006 10:06'!
setStandardClass: aClass methodSymbol: methodSym

	classSymbol := aClass theNonMetaClass name.
	classIsMeta := aClass isMeta.
	methodSymbol := methodSym.
	stringVersion := nil.! !

!MethodReference methodsFor: 'initialize-release' stamp: 'fbs 5/21/2013 22:26'!
setStandardClass: aClass methodSymbol: aSelector environment: anEnvironment
	classSymbol := aClass theNonMetaClass name.
	classIsMeta := aClass isMeta.
	methodSymbol := aSelector.
	environment := anEnvironment.
	stringVersion := nil.! !

!MethodReference methodsFor: 'accessing' stamp: 'ab 8/22/2003 17:59'!
source
	^ (self actualClass sourceCodeAt: methodSymbol) asString withSqueakLineEndings! !

!MethodReference methodsFor: 'accessing' stamp: 'ab 5/23/2003 22:58'!
sourceCode
	^ self actualClass sourceCodeAt: methodSymbol! !

!MethodReference methodsFor: 'accessing' stamp: 'sr 6/4/2004 01:55'!
sourceString
	^ (self actualClass sourceCodeAt: self methodSymbol) asString! !

!MethodReference methodsFor: 'accessing' stamp: 'RAA 5/29/2001 14:44'!
stringVersion

	^stringVersion! !

!MethodReference methodsFor: 'accessing' stamp: 'kph 12/16/2006 10:11'!
stringVersionDefault

	^classSymbol, (classIsMeta ifTrue: [ ' class ' ] ifFalse: [' ']), methodSymbol  ! !

!MethodReference methodsFor: 'accessing' stamp: 'RAA 5/29/2001 14:44'!
stringVersion: aString

	stringVersion := aString! !

!MethodReference methodsFor: 'accessing' stamp: 'ab 8/22/2003 17:58'!
timeStamp
	^ self compiledMethod timeStamp! !

!MethodReference methodsFor: 'comparing' stamp: 'RAA 5/28/2001 11:56'!
<= anotherMethodReference

	classSymbol < anotherMethodReference classSymbol ifTrue: [^true].
	classSymbol > anotherMethodReference classSymbol ifTrue: [^false].
	classIsMeta = anotherMethodReference classIsMeta ifFalse: [^classIsMeta not].
	^methodSymbol <= anotherMethodReference methodSymbol
! !

!MethodReference methodsFor: 'comparing' stamp: 'nice 1/1/2010 21:57'!
= anotherMethodReference 
	"Answer whether the receiver and the argument represent the 
	same object."
	^ self species == anotherMethodReference species
		and: [self classSymbol = anotherMethodReference classSymbol
		and: [self classIsMeta = anotherMethodReference classIsMeta
		and: [self methodSymbol = anotherMethodReference methodSymbol]]]! !

Object subclass: #MidiPrimTester
	instanceVariableNames: 'port'
	classVariableNames: 'CanSetClock CanUseSemaphore ClockTicksPerSec EchoOn EventsAvailable FlushDriver HasBuffer HasDurs HasInputClock Installed UseControllerCache Version'
	poolDictionaries: ''
	category: 'System-Serial Port'!

!MidiPrimTester commentStamp: '<historical>' prior: 0!
This class simply demonstrates and tests the MIDI primitives. MIDI applications should use Stephen Pope's MIDIPort class, which will replace this one.

The Macintosh, and perhaps some other platforms, can send and receive MIDI data over a serial port by using an external clock signal supplied by an external MIDI adapter to generate the correct MIDI baud rate. Typical clock speeds of such adapters are 1, 2, or 0.5 MHz. This clock speed can be specified when a MIDI port is opened. On other platforms, this clock speed parameter is ignored.
!

!MidiPrimTester class methodsFor: 'class initialization' stamp: 'yo 12/3/2004 17:05'!
initialize
	"Initialize the MIDI parameter constants."
	"MidiPrimTester initialize"

	Installed := 1.
		"Read-only. Return 1 if a MIDI driver is installed, 0 if not.
		 On OMS-based MIDI drivers, this returns 1 only if the OMS
		 system is properly installed and configured."

	Version := 2.
		"Read-only. Return the integer version number of this MIDI driver.
		 The version numbering sequence is relative to a particular driver.
		 That is, version 3 of the Macintosh MIDI driver is not necessarily
		 related to version 3 of the Win95 MIDI driver."

	HasBuffer := 3.
		"Read-only. Return 1 if this MIDI driver has a time-stamped output
		 buffer, 0 otherwise. Such a buffer allows the client to schedule
		 MIDI output packets to be sent later. This can allow more precise
		 timing, since the driver uses timer interrupts to send the data
		 at the right time even if the processor is in the midst of a
		 long-running Squeak primitive or is running some other application
		 or system task."

	HasDurs := 4.
		"Read-only. Return 1 if this MIDI driver supports an extended
		 primitive for note-playing that includes the note duration and
		 schedules both the note-on and the note-off messages in the
		 driver. Otherwise, return 0."

	CanSetClock := 5.
		"Read-only. Return 1 if this MIDI driver's clock can be set
		 via an extended primitive, 0 if not."

	CanUseSemaphore := 6.
		"Read-only. Return 1 if this MIDI driver can signal a semaphore
		 when MIDI input arrives. Otherwise, return 0. If this driver
		 supports controller caching and it is enabled, then incoming
		 controller messages will not signal the semaphore."

	EchoOn := 7.
		"Read-write. If this flag is set to a non-zero value, and if
		 the driver supports echoing, then incoming MIDI events will
		 be echoed immediately. If this driver does not support echoing,
		 then queries of this parameter will always return 0 and
		 attempts to change its value will do nothing."

	UseControllerCache := 8.
		"Read-write. If this flag is set to a non-zero value, and if
		 the driver supports a controller cache, then the driver will
		 maintain a cache of the latest value seen for each MIDI controller,
		 and control update messages will be filtered out of the incoming
		 MIDI stream. An extended MIDI primitive allows the client to
		 poll the driver for the current value of each controller. If
		 this driver does not support a controller cache, then queries
		 of this parameter will always return 0 and attempts to change
		 its value will do nothing."

	EventsAvailable := 9.
		"Read-only. Return the number of MIDI packets in the input queue."

	FlushDriver := 10.
		"Write-only. Setting this parameter to any value forces the driver
		 to flush its I/0 buffer, discarding all unprocessed data. Reading
		 this parameter returns 0. Setting this parameter will do nothing
		 if the driver does not support buffer flushing."

	ClockTicksPerSec := 11.
		"Read-only. Return the MIDI clock rate in ticks per second."

	HasInputClock := 12.
		"Read-only. Return 1 if this MIDI driver timestamps incoming
		 MIDI data with the current value of the MIDI clock, 0 otherwise.
		 If the driver does not support such timestamping, then the
		 client must read input data frequently and provide its own
		 timestamping."
! !

!MidiPrimTester methodsFor: 'private' stamp: 'jm 5/18/1998 12:48'!
bufferTimeStampFrom: aByteArray
	"Return the timestamp from the given MIDI input buffer. Assume the given buffer is at least 4 bytes long."

	^ ((aByteArray at: 1) bitShift: 24) +
	  ((aByteArray at: 2) bitShift: 16) +
	  ((aByteArray at: 3) bitShift: 8) +
	   (aByteArray at: 4)
! !

!MidiPrimTester methodsFor: 'private' stamp: 'jm 5/18/1998 12:48'!
canTurnOnParameter: whichParameter
	"Return true if the given MIDI parameter can be turned on. Leave the parameter in its orginal state."

	| old canSet |
	old := self primMIDIParameterGet: whichParameter.
	self primMIDIParameterSet: whichParameter to: 1.
	canSet := (self primMIDIParameterGet: whichParameter) = 1.
	self primMIDIParameterSet: whichParameter to: old.
	^ canSet
! !

!MidiPrimTester methodsFor: 'tests' stamp: 'jm 5/18/1998 10:30'!
getDriverParameters
	"Return a string that describes this platform's MIDI parameters."
	"MidiPrimTester new getDriverParameters"

	| s parameterNames v |
	parameterNames := #(Installed Version HasBuffer HasDurs CanSetClock CanUseSemaphore EchoOn UseControllerCache EventsAvailable FlushDriver ClockTicksPerSec HasInputClock).

	s := WriteStream on: String new.
	s cr.
	1 to: parameterNames size do: [:i |
		v := self primMIDIParameterGet: i.
		s nextPutAll: (parameterNames at: i).
		s nextPutAll: ' = '.
		s print: v; cr].

	s nextPutAll: 'MIDI Echoing is '.
	(self canTurnOnParameter: EchoOn)
		ifTrue: [s nextPutAll: 'supported.'; cr]
		ifFalse: [s nextPutAll: 'not supported.'; cr].

	s nextPutAll: 'Controller Caching is '.
	(self canTurnOnParameter: UseControllerCache)
		ifTrue: [s nextPutAll: 'supported.'; cr]
		ifFalse: [s nextPutAll: 'not supported.'; cr].

	^ s contents
! !

!MidiPrimTester methodsFor: 'tests' stamp: 'nice 12/27/2009 03:10'!
getInputForSeconds: seconds onPort: portNum
	"Collect MIDI input from the given port for the given number of seconds, and answer a string describing the data read."
	"MidiPrimTester new getInputForSeconds: 5 onPort: 0"

	| s bufList midiStartTime |
	"collect the data"
	self openPort: portNum andDo: [ | endTime buf n |
		buf := ByteArray new: 1000.
		bufList := OrderedCollection new.
		midiStartTime := self primMIDIGetClock.
		endTime := Time millisecondClockValue + (seconds * 1000).
		[Time millisecondClockValue < endTime] whileTrue: [
			n := self primMIDIReadPort: portNum into: buf.
			n > 0 ifTrue: [bufList add: (buf copyFrom: 1 to: n)].
			(Delay forMilliseconds: 5) wait]].

	"format the data into a string"
	s := WriteStream on: String new.
	s cr.
	bufList do: [:b | | t |
		t := (self bufferTimeStampFrom: b) - midiStartTime.
		s print: t.
		s nextPutAll: ': '.
		5 to: b size do: [:i | s print: (b at: i); space].
		s cr].
	^ s contents
! !

!MidiPrimTester methodsFor: 'tests' stamp: 'jm 5/18/1998 10:05'!
getPortList
	"Return a string that describes this platform's MIDI ports."
	"MidiPrimTester new getPortList"

	| s portCount dir directionString |
	s := WriteStream on: String new.
	s cr; nextPutAll: 'MIDI Ports:'; cr.
	portCount := self primMIDIGetPortCount.
	0 to: portCount - 1 do: [:i |
		s tab.
		s print: i; nextPutAll: ': '. 
		s nextPutAll: (self primMIDIGetPortName: i).
		dir := self primMIDIGetPortDirectionality: i.
		directionString := dir printString.  "default"
		dir = 1 ifTrue: [directionString := '(in)'].
		dir = 2 ifTrue: [directionString := '(out)'].
		dir = 3 ifTrue: [directionString := '(in/out)'].
		s space; nextPutAll: directionString; cr].
	^ s contents
! !

!MidiPrimTester methodsFor: 'private' stamp: 'jm 5/18/1998 15:32'!
openPort: portNum andDo: aBlock
	"Open the given MIDI port, evaluate the block, and close the port again. Answer the value of the block."

	| result |
	self primMIDIClosePort: portNum.
	self primMIDIOpenPort: portNum readSemaIndex: 0 interfaceClockRate: 1000000.
	result := aBlock value.
	self primMIDIClosePort: portNum.
	^ result
! !

!MidiPrimTester methodsFor: 'tests' stamp: 'nice 1/18/2010 15:44'!
playDrumRoll: mSecsBetweenNotes count: tapCount onPort: portNum
	"MidiPrimTester new playDrumRoll: 75 count: 64 onPort: 0"
	"Play middle-C tapCount times with the given space between notes. This example works best with a short percussive voice, like a drum."
	"Details: This test can be used to investigate the real-time performance of your system. On a 110 MHz PowerPC Mac, this method can genererate very fast and smooth drum rolls up to about 100 beats/sec (10 mSecs between notes). However, many factors can prevent one from seeing this level of performance including a slow CPU, lack of a level-2 cache, networking or other background processes stealing chunks of processor time from Squeak, or a sluggish MIDI synthesizer."
	"Details: By default, this method does an incremental GC on every note. While not really needed for this example, it illustrates a useful technique for real-time processing in Squeak: do an incremental GC when you know you have a few milliseconds of idle time to avoid triggering one during a time-critical task. In this case, we're also using the GC time to provide a small delay between the note-on and note-off events. If the GC time is too short, as it could be on a fast machine, the note may not sound at all unless you add a few milliseconds of additional delay!!"
	"Note: This example works best if the VM's millisecond clock has 1 millisecond resolution."

	| gcDuringNote noteOn noteOff endTime waitTime |
	gcDuringNote := true.
	"these events use running status, so the command byte is omitted"
	noteOn := #[ 60 100 ].
	noteOff := #[ 60 0 ].
	self primMIDIOpenPort: portNum readSemaIndex: 0 interfaceClockRate: 1000000.

	"send an initial event with command byte to initiate running status"
	self primMIDIWritePort: portNum from: #[144 60 0] at: 0.

	1 to: tapCount do: [:i |
		endTime := Time millisecondClockValue + mSecsBetweenNotes.
		self primMIDIWritePort: portNum from: noteOn at: 0.
		gcDuringNote
			ifTrue: [
				"do quick GC; takes a few milliseconds and provides the note-down time"
				"Note: if GC is too fast on your machine, you need to add a few mSecs delay!!"
				Smalltalk garbageCollectMost]
			ifFalse: [(Delay forMilliseconds: 3) wait].

		self primMIDIWritePort: portNum from: noteOff at: 0.
		waitTime := endTime - Time millisecondClockValue.
		waitTime > 0 ifTrue: [(Delay forMilliseconds: waitTime) wait]].

	self primMIDIClosePort: portNum.
! !

!MidiPrimTester methodsFor: 'tests' stamp: 'nice 1/18/2010 13:40'!
playNoteOnPort: portNum
	"MidiPrimTester new playNoteOnPort: 0"

	| noteOn noteOff bytesWritten |
	noteOn := #[144 60 100].
	noteOff := #[144 60 0].
	self openPort: portNum andDo: [
		bytesWritten := self primMIDIWritePort: portNum from: noteOn at: 0.
		(Delay forMilliseconds: 500) wait.
		bytesWritten := bytesWritten + (self primMIDIWritePort: portNum from: noteOff at: 0)].

	bytesWritten = 6 ifFalse: [self error: 'not all bytes were sent'].
! !

!MidiPrimTester methodsFor: 'tests' stamp: 'nice 1/18/2010 13:40'!
playScale: mSecsPerNote onPort: portNum
	"MidiPrimTester new playScale: 130 onPort: 0"

	| noteOn noteOff |
	noteOn := #[144 0 100].
	noteOff := #[144 0 0].
	self openPort: portNum andDo: [
		#(60 62 64 65 67 69 71 72 74 72 71 69 67 65 64 62 60) do: [:midiKey | 
			noteOn at: 2 put: midiKey.
			noteOff at: 2 put: midiKey.
			self primMIDIWritePort: portNum from: noteOn at: 0.
			(Delay forMilliseconds: mSecsPerNote - 10) wait.
			self primMIDIWritePort: portNum from: noteOff at: 0.
			(Delay forMilliseconds: 10) wait]].
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIClosePort: portNum

	<primitive: 'primitiveMIDIClosePort' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIGetClock

	<primitive: 'primitiveMIDIGetClock' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIGetPortCount

	<primitive: 'primitiveMIDIGetPortCount' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIGetPortDirectionality: portNum

	<primitive: 'primitiveMIDIGetPortDirectionality' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIGetPortName: portNum

	<primitive: 'primitiveMIDIGetPortName' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIOpenPort: portNum readSemaIndex: readSemaIndex interfaceClockRate: interfaceClockRate
	"Open the given MIDI port. If non-zero, readSemaIndex specifies the index in the external objects array of a semaphore to be signalled when incoming MIDI data is available. Not all platforms support signalling the read semaphore. InterfaceClockRate specifies the clock rate of the external MIDI interface adaptor on Macintosh computers; it is ignored on other platforms."

	<primitive: 'primitiveMIDIOpenPort' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIParameterGet: whichParameter

	<primitive: 'primitiveMIDIParameterGetOrSet' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIParameterSet: whichParameter to: newValue

	<primitive: 'primitiveMIDIParameterGetOrSet' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIReadPort: portNum into: byteArray

	<primitive: 'primitiveMIDIRead' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

!MidiPrimTester methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primMIDIWritePort: portNum from: byteArray at: midiClockValue

	<primitive: 'primitiveMIDIWrite' module: 'MIDIPlugin'>
	self primitiveFailed.
! !

Object subclass: #MOFile
	instanceVariableNames: 'localeID fileName isLittleEndian magic revision nStrings originalTableOffset translatedTableOffset hashTableSize hashTableOffset hashTable originalStrings translatedStrings translations'
	classVariableNames: 'Cr Lf'
	poolDictionaries: ''
	category: 'System-Localization'!

!MOFile commentStamp: '<historical>' prior: 0!
Wrapper for MO file of gettext.
Known limitation:  
	currently don't support prural form.
	translation strings have to be encoded in utf-8.

Implementation notes:
	Testing on XO showed emulation of hash search without plugin + on demand loading is slow.
	The test also showed conversion of utf8 string to Squeak's String is really slow (especially for non-latin language).
	so in this version, all of original/translated strings are loaded on initiaization,
	but "translated strings" is left as ByteString on loading time, to reduce loading time.
	After that the translated string is converted on demand. 
!

!MOFile class methodsFor: 'instance creation' stamp: 'KR 10/16/2007 19:05'!
fileName: path localeID: id
	^self new 
			load:path localeID: id! !

!MOFile class methodsFor: 'class initialization' stamp: 'KR 12/12/2007 13:13'!
initialize
	Cr := Character cr.
	Lf := Character lf.
! !

!MOFile methodsFor: 'public' stamp: 'yo 8/27/2008 13:09'!
atRandom

	^ self translatedString:nStrings atRandom.
! !

!MOFile methodsFor: 'public' stamp: 'KR 10/4/2007 01:46'!
fileName
	^fileName! !

!MOFile methodsFor: 'public' stamp: 'KR 10/16/2007 19:03'!
fileName: path
	fileName := path! !

!MOFile methodsFor: 'experimental' stamp: 'KR 10/4/2007 00:18'!
hashPjw: aString
	"So called `hashpjw' function by P.J. Weinberger
   	[see Aho/Sethi/Ullman, COMPILERS: Principles, Techniques and Tools,
   	1986, 1987 Bell Telephone Laboratories, Inc.] "
	| stringSize hash g |
	stringSize := aString size.
	hash := 0.
	1 to: stringSize do: [:pos |
		hash := hash bitShift: 4.
		hash := hash + ((aString at: pos) asInteger).
		g := hash bitAnd: 16rF0000000.
		g = 0 ifFalse: [
			hash := hash  bitXor: (g bitShift: -24).
			hash := hash bitXor: g.
		]
	].
	^hash.
! !

!MOFile methodsFor: 'experimental' stamp: 'KR 10/16/2007 17:40'!
load1: aFileName localeID: id
	"CASE1: 
		all of strings are loaded. 
		translation strings are converted to Squeak format on load time.
		original-string/index pairs are registerd to Dictionary on load time.
		hash search can't be used"
	| strm originalTable translatedTable |
	localeID := id.
	strm := FileStream readOnlyFileNamed: aFileName.
	fileName := aFileName.
	[
		self loadHeader: strm.
		originalTable := self loadStringPointers: strm 
								offset: originalTableOffset.

		originalStrings := self loadStrings: strm 
								pointers: originalTable.

		translatedTable := self loadStringPointers: strm 
								offset: translatedTableOffset.

		translatedStrings := self loadStrings: strm 
								pointers: translatedTable
								encoding: 'utf8'
								languageEnvironment: (Locale localeID: localeID) languageEnvironment .

		translations := Dictionary new.
		1 to: nStrings do: [:index |
			| key |
			key := originalStrings at: index.
			translations at: key put: index.
		].
		originalTable := nil.
	] ensure: [strm close].! !

!MOFile methodsFor: 'experimental' stamp: 'KR 10/16/2007 17:33'!
load4: aFileName localeID: id
	"CASE4: 
		all of strings are loaded. 
		loading and conversion of translation strings to Squeak format is executed on initialization time.
		only hash search can be used"
	| strm originalTable translatedTable |
	localeID := id.
	strm := FileStream readOnlyFileNamed: aFileName.
	fileName := aFileName.
	[
		self loadHeader: strm.
		self loadHashTable: strm.
		originalTable := self loadStringPointers: strm 
								offset: originalTableOffset.

		originalStrings := self loadStrings: strm 
								pointers: originalTable.

		translatedTable := self loadStringPointers: strm 
								offset: translatedTableOffset.

		translatedStrings := self loadStrings: strm 
								pointers: translatedTable
								encoding: 'utf-8'
								languageEnvironment: (Locale localeID: localeID) languageEnvironment .
	] ensure: [strm close].! !

!MOFile methodsFor: 'experimental' stamp: 'KR 10/4/2007 00:27'!
loadHashTable: strm
	| entry |
	hashTable := IntegerArray  ofSize: hashTableSize.
	strm binary.
	strm position: hashTableOffset.
	1 to: hashTableSize do: [:index |
		entry := self nextInt32From: strm.
		hashTable at:  index put: entry.
	]! !

!MOFile methodsFor: 'private' stamp: 'KR 12/22/2007 09:18'!
loadHeader: strm
	strm binary.
	magic :=  strm uint32.
	magic = 16rDE120495 
		ifTrue: [isLittleEndian := true]
		ifFalse: [
			magic = 16r950412DE 
				ifTrue: [isLittleEndian := false]
				ifFalse: [ self error: 'invalid MO']
		].
	revision := self nextInt32From: strm.
	nStrings := self nextInt32From: strm.
	originalTableOffset := self nextInt32From: strm.
	translatedTableOffset := self nextInt32From: strm.
	hashTableSize := self nextInt32From: strm.
	hashTableOffset := self nextInt32From: strm.
! !

!MOFile methodsFor: 'private' stamp: 'KR 10/16/2007 13:31'!
loadStringPointers: strm offset: tableOffset
	"returns tupple {arrayOfOffsetToString  arrayOfLengthOfString}"
	| offsetTable lenTable len offset tupple |
	offsetTable := IntegerArray new: nStrings.
	lenTable := IntegerArray new: nStrings.
	strm binary.
	strm position: tableOffset.
	1 to: nStrings do: [:index |
		len := self nextInt32From: strm.
		offset := self nextInt32From: strm.
		offsetTable at: index put: offset.
		lenTable at: index put: len.
	].
	tupple := Array new: 2.
	tupple at: 1 put: offsetTable.
	tupple at: 2 put:  lenTable.
	^tupple
! !

!MOFile methodsFor: 'private' stamp: 'KR 10/16/2007 17:08'!
loadStrings: strm pointers: table
	^self loadStrings: strm pointers: table encoding: nil languageEnvironment: nil
! !

!MOFile methodsFor: 'private' stamp: 'KR 10/16/2007 16:56'!
loadStrings: strm pointers: tupple encoding: encodingName languageEnvironment: env
	| strings rawStr str offsetTable lenTable |
	offsetTable :=  tupple first.
	lenTable := tupple second.
	strings := Array new: nStrings.
	1 to: nStrings do: [:index |
		rawStr := self loadString: strm 
					pointer:  (offsetTable at: index)
					length: (lenTable at: index).
		str := encodingName isNil ifTrue: [rawStr] 
						ifFalse: [ encodingName = 'utf8' 
									ifTrue: [rawStr utf8ToSqueak applyLanguageInfomation: env]
									ifFalse: [self error: 'this encoding isn''t supported']
						].
		strings at: index put: str.
	].
	^strings.! !

!MOFile methodsFor: 'private' stamp: 'KR 12/12/2007 13:14'!
loadString: strm pointer: top  length: len
	| str |
	str := ByteString new: len.
	strm position:  top.
	strm nextInto: str.
	^str replaceAll: Lf with: Cr.
! !

!MOFile methodsFor: 'public' stamp: 'KR 10/16/2007 17:16'!
load: aFileName localeID: id
	"all of original/translated strings are loaded. 
		but conversion of translation string (in utf-8 bytestring) to Squeak format will be defered.
		original-string/index pairs are registerd to Dictionary on load time.
		hash search can't be used"
	| strm originalTable translatedTable |
	localeID := id.
	strm := FileStream readOnlyFileNamed: aFileName.
	fileName := aFileName.
	[
		self loadHeader: strm.
		originalTable := self loadStringPointers: strm 
								offset: originalTableOffset.

		originalStrings := self loadStrings: strm 
								pointers: originalTable.

		translatedTable := self loadStringPointers: strm 
								offset: translatedTableOffset.

		translatedStrings := self loadStrings: strm 
								pointers: translatedTable.

		translations := Dictionary new: nStrings * 2.  "make too enough room to avoid #grow"
		1 to: nStrings do: [:index |
			| key |
			key := originalStrings at: index.
			translations at: key put: index.
		].
		originalStrings := nil.
	] ensure: [strm close].! !

!MOFile methodsFor: 'private' stamp: 'KR 10/4/2007 00:27'!
nextInt32From: strm
	^isLittleEndian 
			ifTrue: [^strm nextLittleEndianNumber: 4]
			ifFalse: [^strm nextInt32]! !

!MOFile methodsFor: 'private' stamp: 'KR 10/16/2007 13:41'!
originalString: index
	^originalStrings at: index.
! !

!MOFile methodsFor: 'public' stamp: 'tfel 8/4/2016 16:31'!
searchByDictionary: aString
	| index |
	index := translations at: aString ifAbsentPut: [nil].
	index ifNil: [^ nil].
	^self translatedString: index! !

!MOFile methodsFor: 'experimental' stamp: 'KR 10/5/2007 01:43'!
searchByHash: aString
	| hashValue nstr index incr key |
	hashValue :=  self hashPjw: aString.
	incr := 1 + (hashValue \\ (hashTableSize -2)).
	index := (hashValue \\ hashTableSize) .
	[ 	nstr := (hashTable at: index +1 ).
		nstr = 0 ifTrue: [^nil].
		key := self originalString: nstr.
		key = aString ifTrue: [^self translatedString: nstr].
		index >= (hashTableSize - incr) 
				ifTrue: [index := index - (hashTableSize - incr)  ]
				ifFalse:[index := index + incr].	
	] doWhileTrue: true.! !

!MOFile methodsFor: 'experimental' stamp: 'tak 10/18/2007 21:15'!
testSearchByDictionary
	InternalTranslator allKnownPhrases 
		do: [:each |
			self searchByDictionary: each
		].
	! !

!MOFile methodsFor: 'experimental' stamp: 'tak 10/18/2007 21:15'!
testSearchByHash
	InternalTranslator allKnownPhrases 
		do: [:each |
			self searchByHash: each
		].
	! !

!MOFile methodsFor: 'experimental' stamp: 'KR 10/5/2007 02:02'!
translateByHash: aString
	| trans |
	trans := self searchByHash: aString.
	trans isNil ifTrue: [^aString]
			ifFalse: [^trans].
! !

!MOFile methodsFor: 'private' stamp: 'KR 10/17/2007 13:16'!
translatedString: index
	"KNOWN PROBLEM: conversion is executed everytimes this method called"
	| str |
	str := translatedStrings at: index.

	^str utf8ToSqueak applyLanguageInfomation: (Locale localeID: localeID) languageEnvironment.
! !

!MOFile methodsFor: 'public' stamp: 'Richo 5/26/2010 11:17'!
translationFor: aString 
	| |
	aString size = 0 ifTrue: [^ '']. "Gettext header"
	^ (self searchByDictionary: aString) ifNil: [aString]
! !

!MOFile methodsFor: 'private' stamp: 'tfel 8/4/2016 16:40'!
translations

	^ translations! !

Object subclass: #NaturalLanguageFormTranslator
	instanceVariableNames: 'id generics'
	classVariableNames: 'CachedTranslations'
	poolDictionaries: ''
	category: 'System-Localization'!

!NaturalLanguageFormTranslator class methodsFor: 'as yet unclassified' stamp: 'yo 9/12/2006 14:28'!
bitmapForJapanese

	^ (Form
	extent: 54@17
	depth: 16
	fromArray: #( 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 65536 0 1 65537 65537 65537 65537 0 0 0 0 1 0 0 0 0 65537 65537 1 65537 65537 65536 1 0 0 0 1 0 0 1 0 0 0 1 0 0 0 0 1 0 0 0 0 0 0 0 1 0 0 0 65536 0 0 65536 0 0 1 0 0 0 1 0 0 0 0 1 0 0 0 1 65537 65537 65536 1 0 0 0 1 0 0 65536 0 0 1 0 0 0 1 0 1 65537 65537 65537 65537 65537 65537 0 0 0 1 65537 65537 0 0 1 0 1 0 0 0 1 0 0 0 1 0 0 0 0 65537 65536 0 0 0 65537 65537 0 65536 1 0 0 0 65536 1 0 0 0 1 65537 65537 65537 65537 0 0 0 1 1 1 0 0 0 0 0 0 65536 1 0 0 0 65536 1 0 0 0 1 0 0 0 1 0 0 0 1 1 1 0 0 0 65537 65537 65537 65537 65537 65537 0 0 65536 1 0 0 0 1 0 0 0 1 0 0 0 65536 1 0 65536 0 0 0 0 0 0 0 0 0 0 65536 1 0 0 0 1 0 0 0 1 0 0 1 0 1 0 1 0 0 65537 65537 1 65537 65537 65536 0 0 65536 1 0 0 0 1 0 0 0 1 0 0 65536 65537 65537 65537 65536 65536 0 65536 1 1 0 0 65536 0 0 65536 0 65536 0 0 1 0 0 0 1 0 1 0 0 1 0 0 1 0 65536 1 1 0 0 65536 0 1 0 0 65536 0 0 1 65537 65537 65537 65537 0 0 0 0 1 0 0 0 0 65537 65537 1 65537 65537 65536 0 1 0 0 1 0 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 65536 0 0 0 65536 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0)
	offset: 0@0)! !

!NaturalLanguageFormTranslator class methodsFor: 'accessing' stamp: 'yo 1/13/2005 11:13'!
cachedTranslations
	"CachedTranslations := nil" 
	^CachedTranslations ifNil: [CachedTranslations := Dictionary new]! !

!NaturalLanguageFormTranslator class methodsFor: 'class initialization' stamp: 'ar 2/26/2010 23:03'!
cleanUp
	"Flush caches"

	CachedTranslations := nil! !

!NaturalLanguageFormTranslator class methodsFor: 'as yet unclassified' stamp: 'yo 9/12/2006 14:21'!
initializeJapaneseBitmap

	(self localeID: (LocaleID isoString: 'ja')) name: 'LanguageNameInNativeLanguage' form: self bitmapForJapanese.
! !

!NaturalLanguageFormTranslator class methodsFor: 'accessing' stamp: 'yo 1/13/2005 11:13'!
isoLanguage: isoLanguage
	"Return the generic language translator as there is no information about the country code"

	^self isoLanguage: isoLanguage isoCountry: nil! !

!NaturalLanguageFormTranslator class methodsFor: 'accessing' stamp: 'yo 1/13/2005 11:13'!
isoLanguage: isoLanguage isoCountry: isoCountry
	^self localeID: (LocaleID  isoLanguage: isoLanguage isoCountry: isoCountry)! !

!NaturalLanguageFormTranslator class methodsFor: 'i/o' stamp: 'yo 1/13/2005 14:02'!
loadFormsFrom: aStream

	| rr pair inst |
	rr := ReferenceStream on: aStream.
	pair := rr next.
	inst := self localeID: (LocaleID isoString: pair first).
	pair second associationsDo: [:assoc |
		inst name: assoc key form: assoc value.
	].
	^ inst.
! !

!NaturalLanguageFormTranslator class methodsFor: 'accessing' stamp: 'yo 1/13/2005 11:13'!
localeID: localeID 
	^ self cachedTranslations
		at: localeID
		ifAbsentPut: [self new localeID: localeID]! !

!NaturalLanguageFormTranslator methodsFor: 'accessing' stamp: 'yo 1/13/2005 11:15'!
generics
	^generics ifNil: [generics := Dictionary new]! !

!NaturalLanguageFormTranslator methodsFor: 'accessing' stamp: 'yo 1/13/2005 11:27'!
localeID
	^id! !

!NaturalLanguageFormTranslator methodsFor: 'accessing' stamp: 'yo 1/13/2005 11:26'!
localeID: anID
	id := anID! !

!NaturalLanguageFormTranslator methodsFor: 'accessing' stamp: 'yo 1/13/2005 11:17'!
name: formName form: translatedForm 
	self generics at: formName put: translatedForm.
! !

!NaturalLanguageFormTranslator methodsFor: 'i/o' stamp: 'yo 1/13/2005 14:02'!
saveFormsOn: aStream

	| rr |
	rr := ReferenceStream on: aStream.
	rr nextPut: {id isoString. generics}.
	rr close.
! !

!NaturalLanguageFormTranslator methodsFor: 'utilities' stamp: 'yo 1/13/2005 11:35'!
translate: aString

	^ (self generics
		at: aString ifAbsent: [nil]) deepCopy.

	"Do you like to write 'form ifNotNil: [form deepCopy]'?"
! !

Object subclass: #NaturalLanguageTranslator
	instanceVariableNames: 'id'
	classVariableNames: 'Translators'
	poolDictionaries: ''
	category: 'System-Localization'!

!NaturalLanguageTranslator commentStamp: '<historical>' prior: 0!
abstract class of natural language translator.
Class side manages and holds loaded instances of concrete classes.!

NaturalLanguageTranslator subclass: #GetTextTranslator
	instanceVariableNames: 'moFiles'
	classVariableNames: 'LocaleDirsForDomain SystemDefaultLocaleDirs UserDefaultLocaleDirs'
	poolDictionaries: ''
	category: 'System-Localization'!

!GetTextTranslator commentStamp: '<historical>' prior: 0!
emulation of gettext runtime
Known limitation:  
     currently doesn't support plural forms.
!

!GetTextTranslator class methodsFor: 'translation data layout' stamp: 'KR 12/19/2007 11:46'!
addSystemDefaultLocaleDir: dir
	"new dir will be put as first"
 	self systemDefaultLocaleDirs addFirst: dir! !

!GetTextTranslator class methodsFor: 'accessing' stamp: 'tfel 7/28/2016 16:32'!
availableLanguageLocaleIDs
	"GetTextTranslator availableLanguageLocaleIDs"
	| ids dirs localeDirForLang directoryNames |
	ids := Set new.
	dirs := Set new.
	dirs addAll: LocaleDirsForDomain values.
	dirs addAll: self defaultLocaleDirs.
	dirs do: [:dir |
		| localesDir |
		localesDir := FileDirectory on: dir. 
		directoryNames := [localesDir directoryNames] on: InvalidDirectoryError do: [:e | #()].
		directoryNames
				do: [:langDirName | 
					| localeID  |
					localeID := LocaleID posixName: langDirName.
					localeDirForLang := localesDir directoryNamed: (self langDirNameForLocaleID: localeID).
					localeDirForLang ifNotNil: [
						(localeDirForLang fileNamesMatching: '*.mo') ifNotEmpty: [ids add: localeID]].
					localeID hasParent ifTrue: [
						localeDirForLang := localesDir directoryNamed: (self langDirNameForLocaleID: localeID parent).
						localeDirForLang ifNotNil: [
							(localeDirForLang fileNamesMatching: '*.mo') ifNotEmpty: [ids add: localeID parent]]].
				].
	].
	^ids! !

!GetTextTranslator class methodsFor: 'translation data layout' stamp: 'KR 10/16/2007 20:38'!
defaultLocaleDirs
	| dirs |
	dirs := OrderedCollection new.
	UserDefaultLocaleDirs ifNotNil: [dirs addAll: UserDefaultLocaleDirs].
	dirs addAll: self systemDefaultLocaleDirs.
	^dirs
! !

!GetTextTranslator class methodsFor: 'private' stamp: 'tak 10/25/2007 15:56'!
findMOForLocaleID: id domain: aDomainName
	| sepa langSubDir path |
	sepa := FileDirectory slash.
	langSubDir := self langDirNameForLocaleID: id.
	(self localeDirsForDomain: aDomainName)
		do: [:each |
			path := each , sepa , langSubDir, sepa , (self moNameForDomain: aDomainName).
			[(FileDirectory default fileExists: path)
				 ifTrue: [^path]] on: InvalidDirectoryError do: [:e | ^nil]].
	^nil.! !

!GetTextTranslator class methodsFor: 'class initialization' stamp: 'KR 12/19/2007 12:29'!
initialize
	SystemDefaultLocaleDirs := OrderedCollection new.
	UserDefaultLocaleDirs := OrderedCollection new.
	LocaleDirsForDomain := Dictionary new.! !

!GetTextTranslator class methodsFor: 'private' stamp: 'KR 10/16/2007 20:32'!
langDirNameForLocaleID: id
	"returns relative path from locale directory to actual directory containing MOs"
	^(id posixName) , (FileDirectory slash)  , 'LC_MESSAGES'! !

!GetTextTranslator class methodsFor: 'translation data layout' stamp: 'KR 10/16/2007 20:29'!
localeDirForDomain: aDomainName
	"returns registered localeDirectory for the textdomain. returns nil if not registered"
	^LocaleDirsForDomain at: aDomainName ifAbsent: [nil]! !

!GetTextTranslator class methodsFor: 'private' stamp: 'KR 10/16/2007 20:36'!
localeDirsForDomain
	^LocaleDirsForDomain ifNil: [LocaleDirsForDomain := Dictionary new]! !

!GetTextTranslator class methodsFor: 'translation data layout' stamp: 'KR 10/16/2007 20:30'!
localeDirsForDomain: aDomainName
	"returns collection of locale directories for text domain.  
	This includes user defined one for the domain, user defaults and system defaults" 
	| dirs dir |
	dirs := OrderedCollection new.
	dir := self localeDirForDomain: aDomainName.
	dir ifNotNil: [dirs add: dir].
	dirs addAll:  self defaultLocaleDirs.
	^dirs! !

!GetTextTranslator class methodsFor: 'private' stamp: 'KR 10/16/2007 19:40'!
moNameForDomain: domainName
	^domainName , '.mo'! !

!GetTextTranslator class methodsFor: 'instance creation' stamp: 'KR 10/16/2007 20:56'!
newForLocaleID: id
	^self new localeID: id! !

!GetTextTranslator class methodsFor: 'class initialization' stamp: 'KR 12/19/2007 12:32'!
privateStartUp
	self setupLocaleDirs.
	self availableLanguageLocaleIDs do: [ :localeID |
		NaturalLanguageTranslator translators 
				at: localeID 
				put: (self newForLocaleID: localeID).
	]! !

!GetTextTranslator class methodsFor: 'translation data layout' stamp: 'KR 10/16/2007 20:50'!
setLocaleDir: path forDoamin: aDomainName
	self LocaleDirsForDomain
		at: aDomainName
		put: path.! !

!GetTextTranslator class methodsFor: 'translation data layout' stamp: 'fbs 7/25/2013 07:10'!
setupLocaleDirs
	| dirs sepa localesDirName |
	sepa := FileDirectory slash.
	SystemDefaultLocaleDirs := nil.
	dirs := self systemDefaultLocaleDirs.
	localesDirName := 'locale'.
	dirs add:  (Smalltalk imagePath) , sepa , localesDirName.
	dirs add:  (Smalltalk vmPath) , sepa , localesDirName.
	^dirs! !

!GetTextTranslator class methodsFor: 'translation data layout' stamp: 'KR 12/19/2007 12:31'!
systemDefaultLocaleDirs
	^SystemDefaultLocaleDirs ifNil: [SystemDefaultLocaleDirs := OrderedCollection new]
! !

!GetTextTranslator class methodsFor: 'translation data layout' stamp: 'KR 10/16/2007 20:40'!
userDefaultLocaleDirs
	^UserDefaultLocaleDirs ifNil: [UserDefaultLocaleDirs := OrderedCollection new]
! !

!GetTextTranslator methodsFor: 'accessing' stamp: 'yo 8/28/2008 13:16'!
atRandom

	| v |
	moFiles ifEmpty: [^ ''].
	(v := moFiles atRandom value) ifNil: [^ ''].
	^ v atRandom.
! !

!GetTextTranslator methodsFor: 'accessing' stamp: 'KR 10/23/2007 19:16'!
domainRegistered: aDomainName
	"only current translator actually load the MO, to minimize loading time.
	 other translator will load anyway when it goes current"
	(self class current == self) 
		ifTrue: [self moFileForDomain: aDomainName].
	! !

!GetTextTranslator methodsFor: 'accessing' stamp: 'KR 10/23/2007 18:44'!
domainUnregistered: aDomainName
	moFiles removeKey: aDomainName ifAbsent: [^self]
	! !

!GetTextTranslator methodsFor: 'initialize-release' stamp: 'KR 10/3/2007 19:43'!
initialize
	moFiles := Dictionary new.! !

!GetTextTranslator methodsFor: 'accessing' stamp: 'KR 10/30/2007 18:38'!
isDomainLoaded: aDomainName
	| mo |
	mo := moFiles at: aDomainName ifAbsent: [nil].
	^mo isNil not.
! !

!GetTextTranslator methodsFor: 'private' stamp: 'KR 10/17/2007 03:20'!
loadMOFileForDomain: aDomainName
	| moName |
	moName := self class findMOForLocaleID: self localeID 
								domain: aDomainName.
	moName notNil
			 ifTrue: [^MOFile new load: moName
								localeID: self localeID]
			ifFalse: [^nil]
! !

!GetTextTranslator methodsFor: 'accessing' stamp: 'KR 10/16/2007 20:58'!
loadMOFiles
	TextDomainManager allKnownDomains 
		do: [:domainName |
			self moFileForDomain: domainName
		].! !

!GetTextTranslator methodsFor: 'private' stamp: 'KR 10/3/2007 19:46'!
moFileForDomain: domainName
	^moFiles at: domainName ifAbsentPut: [self loadMOFileForDomain: domainName]! !

!GetTextTranslator methodsFor: 'private' stamp: 'tfel 8/4/2016 16:39'!
moFiles

	^ moFiles! !

!GetTextTranslator methodsFor: 'accessing' stamp: 'KR 10/5/2007 00:39'!
reloadMOFiles
	moFiles := Dictionary new.
	self loadMOFiles.! !

!GetTextTranslator methodsFor: 'language switching' stamp: 'KR 10/17/2007 02:00'!
setCurrent
	"ensure actual contents of MOs is loaded on switching language"
	self loadMOFiles! !

!GetTextTranslator methodsFor: 'translation' stamp: 'tak 10/19/2007 13:38'!
translate: aString inDomain: aDomainName
	| mo |
	mo := self moFileForDomain: aDomainName.
	^mo isNil 
		ifTrue: [aString] 
		ifFalse: [mo translationFor: aString]
! !

NaturalLanguageTranslator subclass: #InternalTranslator
	instanceVariableNames: 'generics'
	classVariableNames: 'AllKnownPhrases CachedTranslations'
	poolDictionaries: ''
	category: 'System-Localization'!

!InternalTranslator commentStamp: 'tak 10/19/2007 11:12' prior: 0!
An InternalTranslator is used a translation dictionary in a image.
You can use it without any external translation file.

Structure:
 generics		Dictionary -- msgid -> msgstr
!

!InternalTranslator class methodsFor: 'private' stamp: 'mir 7/15/2004 19:58'!
allKnownPhrases
	^AllKnownPhrases ifNil: [AllKnownPhrases := Dictionary new: 2051]! !

!InternalTranslator class methodsFor: 'accessing' stamp: 'bf 5/2/2010 20:01'!
availableLanguageLocaleIDs
	"Return locale ids for the internal translation dictionary."
	"InternalTranslator availableLanguageLocaleIDs"
	^ self cachedTranslations values collect:[:each | each localeID]! !

!InternalTranslator class methodsFor: 'private' stamp: 'mir 7/13/2004 00:06'!
cachedTranslations
	"CachedTranslations := nil" 
	^CachedTranslations ifNil: [CachedTranslations := Dictionary new]! !

!InternalTranslator class methodsFor: 'private loading' stamp: 'nk 8/21/2004 13:03'!
directoryForLanguage: isoLanguage country: isoCountry create: createDir
	"Try to locate the <prefs>/locale/<language>{/<country>} folder.
	If createDir is set, create the path down to country or language, depending on wether it's specified..
	Return the directory for country or language depending on specification.
	If neither exists, nil"

	"NaturalLanguageTranslator directoryForLanguage: 'es' country: nil create: true"
	"NaturalLanguageTranslator directoryForLanguage: 'de' country: 'DE' create: true"
	"NaturalLanguageTranslator directoryForLanguage: 'en' country: 'US' create: false"
	"NaturalLanguageTranslator directoryForLanguage: 'en' country: nil create: true"

	"If this fails, there is nothing we can do about it here"
	| localeDir  countryDir languageDir |
	localeDir := self localeDirCreate: createDir.
	localeDir ifNil: [^nil].

	isoCountry ifNil: [
		languageDir := localeDir directoryNamed: isoLanguage.
		createDir
			ifTrue: [languageDir assureExistence].
		^languageDir exists
			ifTrue: [languageDir]
			ifFalse: [nil]].

	countryDir := languageDir directoryNamed: isoCountry.
	createDir
		ifTrue: [countryDir assureExistence].

	^countryDir exists
		ifTrue: [countryDir]
		ifFalse: [nil]! !

!InternalTranslator class methodsFor: 'private loading' stamp: 'mir 8/11/2004 10:44'!
directoryForLocaleID: localeID create: createDir
	"Try to locate the <prefs>/locale/<language>{/<country>} folder.
	If createDir is set, create the path down to country or language, depending on locale.
	Return the directory for country or language depending on locale.
	If neither exists, nil"

	"NaturalLanguageTranslator directoryForLanguage: 'de' country: nil readOnly: true"
	"NaturalLanguageTranslator directoryForLanguage: 'de' country: 'DE' readOnly: true"
	"NaturalLanguageTranslator directoryForLanguage: 'en' country: 'US' readOnly: false"
	"NaturalLanguageTranslator directoryForLanguage: 'en' country: nil readOnly: true"

	^self directoryForLanguage: localeID isoLanguage country: localeID isoCountry create: createDir! !

!InternalTranslator class methodsFor: 'class initialization' stamp: 'bf 5/4/2010 16:38'!
discardAllTranslations
	AllKnownPhrases := nil.
	self resetCaches.! !

!InternalTranslator class methodsFor: 'file-services' stamp: 'mir 8/11/2004 10:52'!
fileReaderServicesForFile: fullName suffix: suffix 
	"Answer the file services associated with given file"
	^ (suffix = self translationSuffix) | (suffix = '*')
		ifTrue: [{self serviceMergeLanguageTranslations}]
		ifFalse: [#()]! !

!InternalTranslator class methodsFor: 'private loading' stamp: 'cmm 3/27/2012 18:36'!
loadAvailableExternalLocales
	"private - register locales IDs based on the content of the <prefs>/locale/ directory"
	| localeDir |
	localeDir := self localeDirCreate: false.
	localeDir ifNil: [^ #()].

	localeDir directoryNames
		do: [:langDirName | 
			| langDir | 
			langDir := localeDir directoryNamed: langDirName.

			(langDir fileNamesMatching: '*.' , self translationSuffix)
				ifNotEmpty: [self loadTranslatorForIsoLanguage: langDirName isoCountry: nil].

			langDir directoryNames
				do: [:countryDirName | 
					| countryDir | 
					countryDir := langDir directoryNamed: countryDirName.
					(countryDir fileNamesMatching: '*.' , self translationSuffix)
						ifNotEmpty: [self loadTranslatorForIsoLanguage: langDirName isoCountry: countryDirName]
			]
		].
! !

!InternalTranslator class methodsFor: 'private loading' stamp: 'nk 8/21/2004 13:00'!
loadExternalTranslationsFor: translator
	"Try to load translations from external external files.
	The files are located in the <prefs>/locale/<language>{/<country>} folder.
	There can be more than one file for each location, so applications can install their own partial translation tables. All files in the specific folder are loaded."

	| translationDir |
	translationDir := self directoryForLocaleID: translator localeID create: false.
	translationDir ifNil: [ ^nil ]. 
	(translationDir fileNamesMatching: '*.' , self translationSuffix)
		do: [:fileName | translator loadFromFileNamed: (translationDir fullNameFor: fileName)]! !

!InternalTranslator class methodsFor: 'file-services' stamp: 'nice 7/26/2014 23:04'!
loadForLocaleIsoString: localeString fromGzippedMimeLiteral: mimeString 
	"merge the translation from the mime literal."
	| stream localeID translator gs rbStream s currentPlatform |
	s := Base64MimeConverter mimeDecodeToBytes: mimeString readStream.
	gs := GZipReadStream on: s.
	rbStream := MultiByteBinaryOrTextStream with: gs contents asString.
	rbStream converter: UTF8TextConverter new.
	rbStream reset.
	localeID := LocaleID isoString: localeString.
	currentPlatform := Locale currentPlatform.
	[Locale
		currentPlatform: (Locale localeID: localeID).
	stream := ReadStream on: rbStream contents]
		ensure: [Locale currentPlatform: currentPlatform].
	translator := self localeID: localeID.
	translator loadFromStream: stream.
	LanguageEnvironment resetKnownEnvironments! !

!InternalTranslator class methodsFor: 'private loading' stamp: 'tak 9/7/2005 23:17'!
loadTranslatorForIsoLanguage: isoLanguage isoCountry: isoCountry 
	"private - load the translations from <prefs>/locale/ directory  
	the procedure is to assure the existence of a translator for the  
	given language/country and then load the external translations for this translator"

	| translator |
	translator := self newLocaleID: (LocaleID isoLanguage: isoLanguage isoCountry: isoCountry).

	self loadExternalTranslationsFor: translator! !

!InternalTranslator class methodsFor: 'private loading' stamp: 'mir 8/25/2004 12:03'!
localeDirCreate: createDir
	"Try to locate the <prefs>/locale/ folder.
	If createDir is set, try to create the path.
	If it doesn't exist, return nil"

	"If this fails, there is nothing we can do about it here"
	| prefDir  localeDir |
	(createDir not
			and: [ExternalSettings preferenceDirectory isNil])
		ifTrue: [^ nil].

	prefDir := ExternalSettings assuredPreferenceDirectory.
	prefDir exists
		ifFalse: [^nil].


	localeDir := prefDir directoryNamed: 'locale'.
	createDir
		ifTrue: [localeDir assureExistence].
	^localeDir exists
		ifTrue: [localeDir]
		ifFalse: [nil]! !

!InternalTranslator class methodsFor: 'accessing' stamp: 'tak 9/15/2005 02:08'!
localeID: localeID 
	"For backward compatibility, see NaturalLanguageTranslator >> fileOutHeaderOn:."
	^ self newLocaleID: localeID! !

!InternalTranslator class methodsFor: 'accessing' stamp: 'tak 10/19/2007 13:28'!
mergeLegacyTranslators
	self availableLanguageLocaleIDs
		do: [:localeID | (NaturalLanguageTranslator translators includesKey: localeID)
				ifFalse: [NaturalLanguageTranslator translators
						at: localeID
						put: (self newLocaleID: localeID)]]! !

!InternalTranslator class methodsFor: 'file-services' stamp: 'yo 2/24/2005 21:04'!
mergeTranslationFileNamed: fileFullNameString 
	"merge the translation in the file named fileFullNameString"

	| stream localeID translator |
	stream := FileStream readOnlyFileNamed: fileFullNameString.
	[localeID := LocaleID isoString: stream localName sansPeriodSuffix.
	translator := self localeID: localeID.
	translator loadFromStream: stream]
		ensure: [stream close].
	LanguageEnvironment resetKnownEnvironments.

! !

!InternalTranslator class methodsFor: 'accessing' stamp: 'tak 9/7/2005 23:17'!
newLocaleID: localeID 
	^ self cachedTranslations
		at: localeID
		ifAbsentPut: [self new localeID: localeID]! !

!InternalTranslator class methodsFor: 'class initialization' stamp: 'tak 10/19/2007 15:28'!
privateStartUp
	self loadAvailableExternalLocales.
	self mergeLegacyTranslators.
! !

!InternalTranslator class methodsFor: 'private' stamp: 'mir 7/21/2004 14:18'!
registeredPhraseFor: phrase
	"Using a Dictionary so we can lookup existing string instead of creating needless copies when loading a translation."
	^self allKnownPhrases at: phrase ifAbsentPut: [phrase]! !

!InternalTranslator class methodsFor: 'private' stamp: 'mir 7/15/2004 20:02'!
registerPhrase: phrase
	"Using a Dictionary so we can lookup existing string instead of creating needless copies when loading a translation."
	self allKnownPhrases at: phrase put: phrase! !

!InternalTranslator class methodsFor: 'accessing' stamp: 'tak 10/24/2007 12:46'!
removeLocaleID: localeID 
	"self removeLocaleID: (LocaleID isoString: 'ja-kids')"
	self cachedTranslations
		removeKey: localeID
		ifAbsent: [].
	NaturalLanguageTranslator privateStartUp! !

!InternalTranslator class methodsFor: 'class initialization' stamp: 'tak 10/19/2007 15:34'!
resetCaches
	CachedTranslations := nil.
! !

!InternalTranslator class methodsFor: 'file-services' stamp: 'KR 3/21/2008 20:57'!
serviceMergeLanguageTranslations
	"Answer a service for merging of translation files"
	^ SimpleServiceEntry
		provider: self
		label: 'merge the translation file' translatedNoop
		selector: #mergeTranslationFileNamed:
		description: 'merge the translation file into the language named like the file' translatedNoop
		buttonLabel: 'merge' translatedNoop! !

!InternalTranslator class methodsFor: 'file-services' stamp: 'mir 7/21/2004 13:45'!
services
	"Answer potential file services associated with this class"
	^ {self serviceMergeLanguageTranslations}! !

!InternalTranslator class methodsFor: 'private' stamp: 'mir 8/11/2004 10:52'!
translationSuffix
	^'translation'! !

!InternalTranslator methodsFor: 'accessing' stamp: 'yo 8/27/2008 12:54'!
atRandom

	^ generics atRandom value.
! !

!InternalTranslator methodsFor: 'translation' stamp: 'mir 7/21/2004 18:02'!
checkPhrase: phrase translation: translation! !

!InternalTranslator methodsFor: 'user interface' stamp: 'dgd 8/13/2004 21:54'!
defaultBackgroundColor
	"answer the receiver's defaultBackgroundColor for views"
	^ Color cyan! !

!InternalTranslator methodsFor: 'fileIn/fileOut' stamp: 'tak 11/16/2004 11:04'!
fileOutHeader
	^ '''Translation dictionary'''! !

!InternalTranslator methodsFor: 'user interface' stamp: 'yo 9/9/2005 17:27'!
fileOutHeaderOn: aStream 
	aStream binary.
	UTF8TextConverter writeBOMOn: aStream.
	aStream text.
	aStream nextChunkPut: self fileOutHeader;
		 cr.
	aStream timeStamp; cr.
	aStream nextPut: $!!.
	aStream nextChunkPut: '(' , self class name , ' localeID: ' , id storeString , ')'.
	aStream cr! !

!InternalTranslator methodsFor: 'user interface' stamp: 'yo 6/20/2006 11:47'!
fileOutHeaderOn: aStream withBOM: bomFlag
	bomFlag ifTrue: [
		aStream binary.
		UTF8TextConverter writeBOMOn: aStream.
		aStream text.
	].
	aStream nextChunkPut: self fileOutHeader;
		 cr.
	aStream timeStamp; cr.
	aStream nextPut: $!!.
	aStream nextChunkPut: '(' , self class name , ' localeID: ' , id storeString , ')'.
	aStream cr! !

!InternalTranslator methodsFor: 'fileIn/fileOut' stamp: 'yo 6/20/2006 11:48'!
fileOutOn: aStream 
	"self current fileOutOn: Transcript. Transcript endEntry"
	self fileOutOn: aStream keys: nil withBOM: true.
! !

!InternalTranslator methodsFor: 'fileIn/fileOut' stamp: 'yo 6/20/2006 11:47'!
fileOutOn: aStream keys: keys withBOM: bomFlag
	"self current fileOutOn: Transcript. Transcript endEntry"
	self fileOutHeaderOn: aStream withBOM: bomFlag.
	(keys
		ifNil: [generics keys asSortedCollection])
		do: [:key | self
				nextChunkPut: (generics associationAt: key)
				on: aStream].
	keys
		ifNil: [self untranslated
				do: [:each | self nextChunkPut: each -> '' on: aStream]].
	aStream nextPut: $!!;
		 cr! !

!InternalTranslator methodsFor: 'private' stamp: 'mir 6/30/2004 20:23'!
generics
	^generics ifNil: [generics := Dictionary new]! !

!InternalTranslator methodsFor: 'accessing' stamp: 'KR 10/30/2007 18:42'!
isDomainLoaded: aDomainName
	^true
! !

!InternalTranslator methodsFor: 'private store-retrieve' stamp: 'yo 7/30/2004 13:00'!
loadFromFileNamed: fileNameString 
	"Load translations from an external file"

	| stream |
	[stream := FileStream readOnlyFileNamed: fileNameString.
	self loadFromStream: stream]
		ensure: [stream close].
	self changed: #translations.
	self changed: #untranslated.
! !

!InternalTranslator methodsFor: 'private store-retrieve' stamp: 'tak 11/16/2004 12:37'!
loadFromRefStream: stream 
	"Load translations from an external file"
	| loadedArray refStream |
	refStream := ReferenceStream on: stream.
	[loadedArray := refStream next]
		ensure: [refStream close].
	self processExternalObject: loadedArray ! !

!InternalTranslator methodsFor: 'private store-retrieve' stamp: 'em 3/30/2005 14:32'!
loadFromStream: stream 
	"Load translations from an external file"
	| header isFileIn |
	header := '''Translation dictionary'''.
	isFileIn := (stream next: header size)
				= header.
	stream reset.
	isFileIn
		ifTrue: [stream fileInAnnouncing: 'Loading ' translated, stream localName]
		ifFalse: [self loadFromRefStream: stream]! !

!InternalTranslator methodsFor: 'private store-retrieve' stamp: 'yo 8/2/2004 12:27'!
mergeTranslations: newTranslations
	"Merge a new set of translations into the exiting table.
	Overwrites existing entries."

	newTranslations keysAndValuesDo: [:key :value |
		self rawPhrase: (self class registeredPhraseFor: key) translation: value].
	self changed: #translations.
	self changed: #untranslated.! !

!InternalTranslator methodsFor: 'fileIn/fileOut' stamp: 'tak 11/16/2004 09:26'!
nextChunkPut: anObject on: aStream 
	| i remainder terminator |
	terminator := $!!.
	remainder := anObject storeString.
	[(i := remainder indexOf: terminator) = 0]
		whileFalse: [aStream
				nextPutAll: (remainder copyFrom: 1 to: i).
			aStream nextPut: terminator.
			"double imbedded terminators"
			remainder := remainder copyFrom: i + 1 to: remainder size].
	aStream nextPutAll: remainder.
	aStream nextPut: terminator; cr.! !

!InternalTranslator methodsFor: 'translation' stamp: 'yo 7/30/2004 13:03'!
phrase: phraseString translation: translationString 
	self generics at: phraseString put: translationString asString.
	self changed: #translations.
	self changed: #untranslated.! !

!InternalTranslator methodsFor: 'private store-retrieve' stamp: 'mir 7/15/2004 20:04'!
processExternalObject: anArray 
	"pivate - process the external object"

	"new format -> {translations. untranslated}"

	anArray second do: [:each | self class registerPhrase: each].

	self mergeTranslations: anArray first! !

!InternalTranslator methodsFor: 'translation' stamp: 'yo 8/2/2004 12:27'!
rawPhrase: phraseString translation: translationString 
	self generics at: phraseString put: translationString asString.
! !

!InternalTranslator methodsFor: 'translation' stamp: 'yo 1/14/2005 16:25'!
rawRemoveUntranslated: untranslated

	self class allKnownPhrases removeKey: untranslated ifAbsent: [].
	self changed: #untranslated.! !

!InternalTranslator methodsFor: 'translation' stamp: 'yo 8/1/2004 01:07'!
removeTranslationFor: phraseString
	self generics removeKey: phraseString ifAbsent: [].
	self changed: #translations.
	self changed: #untranslated.! !

!InternalTranslator methodsFor: 'translation' stamp: 'yo 1/14/2005 16:25'!
removeUntranslated: untranslated

	self class allKnownPhrases removeKey: untranslated ifAbsent: [].
! !

!InternalTranslator methodsFor: 'private store-retrieve' stamp: 'yo 2/17/2005 15:45'!
saveToFileNamed: fileNameString 
	"save the receiver's translations to a file named fileNameString"
	| stream |
	"Set true if you need to save as binary"
	false
		ifTrue: [stream := ReferenceStream fileNamed: fileNameString.
			stream nextPut: {self translations. self untranslated}.
			stream close.
			^ self].
	stream := FileStream fileNamed: fileNameString.
	[self fileOutOn: stream]
		ensure: [stream close]! !

!InternalTranslator methodsFor: 'fileIn/fileOut' stamp: 'cwp 3/22/2014 19:45'!
scanFrom: aStream 
	^ self scanFrom: aStream environment: Environment current! !

!InternalTranslator methodsFor: 'fileIn/fileOut' stamp: 'cwp 6/20/2012 17:34'!
scanFrom: aStream environment: anEnvironment
	"Read a definition of dictionary.  
	Make sure current locale corresponds my locale id"
	| aString newTranslations assoc currentPlatform |
	newTranslations := Dictionary new.
	currentPlatform := Locale currentPlatform.
	[Locale
		currentPlatform: (Locale localeID: id).
	[aString := aStream nextChunk withSqueakLineEndings.
	aString size > 0] whileTrue: 
		[assoc := Compiler evaluate: aString environment: anEnvironment.
		assoc value = ''
			ifTrue: [self class registerPhrase: assoc key]
			ifFalse: [newTranslations add: assoc]]]
		ensure: [Locale currentPlatform: currentPlatform].
	self mergeTranslations: newTranslations! !

!InternalTranslator methodsFor: 'translation' stamp: 'tak 10/19/2007 15:11'!
translate: aString inDomain: aDomainName 
	| translator |
	^ self generics
		at: aString
		ifAbsent: [self localeID hasParent
				ifTrue: [translator := self class cachedTranslations
								at: self localeID parent
								ifAbsent: [^ aString].
					translator translate: aString inDomain: aDomainName]
				ifFalse: [aString]]! !

!InternalTranslator methodsFor: 'accessing' stamp: 'mir 7/21/2004 17:00'!
translations
	^self generics! !

!InternalTranslator methodsFor: 'accessing' stamp: 'mir 7/21/2004 17:03'!
untranslated
	| translations |
	translations := self translations.
	^self class allKnownPhrases reject: [:each | translations includesKey: each]! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'topa 10/29/2015 01:03'!
availableForLocaleID: localeID 
	"Answer available locale ID.
	If translator is not found for correct locale ID, then isoLanguage is
	attempted for the key."
	^ self translators
		at: localeID
		ifAbsentPut: [localeID hasParent
				ifTrue: [self translators
						at: localeID parent
						ifAbsent: [self default]]
				ifFalse: [self default]]! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'tak 10/19/2007 13:47'!
availableLanguageLocaleIDs
	"Return the locale ids for the currently available languages.  
	Meaning those which either internally or externally have  
	translations available."
	"NaturalLanguageTranslator availableLanguageLocaleIDs"
	^ self translators values collect:[:each | each localeID]! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'KR 10/15/2007 12:58'!
current
	^ self availableForLocaleID: LocaleID current! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'mt 8/11/2016 08:48'!
default
	"Answer translator for backstop"
	"self default translate: 'test'"
	^ (self == NaturalLanguageTranslator ifTrue: [InternalTranslator] ifFalse: [self]) new
		localeID: (LocaleID isoLanguage: 'en')! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'KR 10/23/2007 18:41'!
domainRegistered: aDomainName
	"notify that new TextDomain is registered"
	self translators do: [:each | each domainRegistered: aDomainName]! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'KR 10/23/2007 18:41'!
domainUnregistered: aDomainName
	"notify that new TextDomain is unregistered"
	self translators do: [:each | each domainUnregistered: aDomainName]! !

!NaturalLanguageTranslator class methodsFor: 'file-services' stamp: 'mir 8/11/2004 10:52'!
fileReaderServicesForFile: fullName suffix: suffix 
	"Answer the file services associated with given file"
	^ (suffix = self translationSuffix) | (suffix = '*')
		ifTrue: [{self serviceMergeLanguageTranslations}]
		ifFalse: [#()]! !

!NaturalLanguageTranslator class methodsFor: 'class initialization' stamp: 'tak 10/19/2007 15:25'!
initialize
	Smalltalk addToStartUpList: NaturalLanguageTranslator after: FileDirectory.
! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:36'!
isoLanguage: isoLanguage
	"Return the generic language translator as there is no information about the country code"

	^self isoLanguage: isoLanguage isoCountry: nil! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:36'!
isoLanguage: isoLanguage isoCountry: isoCountry
	^self localeID: (LocaleID  isoLanguage: isoLanguage isoCountry: isoCountry)! !

!NaturalLanguageTranslator class methodsFor: 'file-services' stamp: 'nice 7/26/2014 23:05'!
loadForLocaleIsoString: localeString fromGzippedMimeLiteral: mimeString 
	"merge the translation from the mime literal."
	| stream localeID translator gs rbStream s currentPlatform |
	s := Base64MimeConverter mimeDecodeToBytes: mimeString readStream.
	gs := GZipReadStream on: s.
	rbStream := MultiByteBinaryOrTextStream with: gs contents asString.
	rbStream converter: UTF8TextConverter new.
	rbStream reset.
	localeID := LocaleID isoString: localeString.
	currentPlatform := Locale currentPlatform.
	[Locale
		currentPlatform: (Locale localeID: localeID).
	stream := ReadStream on: rbStream contents]
		ensure: [Locale currentPlatform: currentPlatform].
	translator := self localeID: localeID.
	translator loadFromStream: stream.
	LanguageEnvironment resetKnownEnvironments! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'KR 10/23/2007 19:42'!
localeChanged
	"notify some project starts to use this locale.
	 this facility may use the event to load translation data dynamically" 
	self current setCurrent
! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'dgd 8/24/2004 19:18'!
localeID: localeID 
	^ self cachedTranslations
		at: localeID
		ifAbsentPut: [self new localeID: localeID]! !

!NaturalLanguageTranslator class methodsFor: 'file-services' stamp: 'laza 3/19/2010 14:01'!
mergeTranslationFileNamed: fileFullNameString 
	"merge the translation in the file named fileFullNameString"

	FileStream readOnlyFileNamed: fileFullNameString do: [:stream |
		| localeID translator |
		localeID := LocaleID isoString: stream localName sansPeriodSuffix.
		translator := self localeID: localeID.
		translator loadFromStream: stream].
	LanguageEnvironment resetKnownEnvironments.

! !

!NaturalLanguageTranslator class methodsFor: 'class initialization' stamp: 'KR 10/23/2007 19:43'!
privateStartUp
	self resetCaches.
	GetTextTranslator privateStartUp.
	InternalTranslator privateStartUp.
	self localeChanged.! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'tak 10/19/2007 11:21'!
removeLocaleID: localeID 
	"self removeLocaleID: (LocaleID isoString: 'ja-kids')"
	^ self translators
		removeKey: localeID
		ifAbsent: []! !

!NaturalLanguageTranslator class methodsFor: 'class initialization' stamp: 'tak 10/19/2007 11:18'!
resetCaches
	Translators := nil.! !

!NaturalLanguageTranslator class methodsFor: 'file-services' stamp: 'mir 7/21/2004 13:45'!
serviceMergeLanguageTranslations
	"Answer a service for merging of translation files"
	^ SimpleServiceEntry
		provider: self
		label: 'merge the translation file'
		selector: #mergeTranslationFileNamed:
		description: 'merge the translation file into the language named like the file'
		buttonLabel: 'merge'! !

!NaturalLanguageTranslator class methodsFor: 'file-services' stamp: 'mir 7/21/2004 13:45'!
services
	"Answer potential file services associated with this class"
	^ {self serviceMergeLanguageTranslations}! !

!NaturalLanguageTranslator class methodsFor: 'class initialization' stamp: 'tak 10/19/2007 15:29'!
startUp: resuming 
	resuming
		ifFalse: [^ self].
	self privateStartUp.! !

!NaturalLanguageTranslator class methodsFor: 'translation' stamp: 'KR 10/30/2007 18:59'!
translateWithoutLoading: aString toLocaleID: localeID
	"translate for default domain"
	^self translateWithoutLoading: aString toLocaleID: localeID inDomain: TextDomainManager defaultDomain.
! !

!NaturalLanguageTranslator class methodsFor: 'translation' stamp: 'tfel 8/6/2016 13:51'!
translateWithoutLoading: aString toLocaleID: localeID inDomain: aDomainName
	"try to translate with small footprint:
		if GetTextTranslator hasn't loaded MO, try to use InternalTranslator.
		if InternalTranslator isn't available, then actually load MO and use it"
	| translator |
	translator := self availableForLocaleID: localeID.
	translator class = NaturalLanguageTranslator ifTrue: [^ aString].
	(translator isDomainLoaded: aDomainName) ifFalse: [
		(InternalTranslator availableLanguageLocaleIDs includes: localeID)
			ifTrue:  [translator := InternalTranslator localeID: localeID].
	].
	^translator translate: aString inDomain: aDomainName! !

!NaturalLanguageTranslator class methodsFor: 'translation' stamp: 'KR 10/30/2007 18:33'!
translate: aString toLocaleID: localeID
	"translate for default domain"
	^ (self availableForLocaleID: localeID)
		translate: aString! !

!NaturalLanguageTranslator class methodsFor: 'translation' stamp: 'KR 10/15/2007 14:00'!
translate: aString toLocaleID: localeID inDomain: aDomainName
	^ (self availableForLocaleID: localeID)
		translate: aString inDomain: aDomainName! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'ul 3/30/2012 12:19'!
translationSuffix

	^'translation'! !

!NaturalLanguageTranslator class methodsFor: 'accessing' stamp: 'tak 10/19/2007 11:15'!
translators
	^ Translators ifNil: [Translators := Dictionary new]	! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'yo 8/27/2008 13:10'!
atRandom

	self subclassResponsibility.
! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'mir 7/21/2004 18:02'!
checkPhrase: phrase translation: translation! !

!NaturalLanguageTranslator methodsFor: 'user interface' stamp: 'dgd 8/13/2004 21:54'!
defaultBackgroundColor
	"answer the receiver's defaultBackgroundColor for views"
	^ Color cyan! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'dgd 8/13/2004 21:12'!
displayLanguage
	^ id displayLanguage! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'dgd 10/7/2004 20:50'!
displayName
	^ id displayName! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'KR 10/23/2007 18:40'!
domainRegistered: aDomainName
	"notify that new TextDomain is registered.  Concrete subclass can responds to this event if needed"! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'KR 10/23/2007 18:40'!
domainUnregistered: aDomainName
	"notify that new TextDomain is unregistered.  Concrete subclass can responds to this event if needed"! !

!NaturalLanguageTranslator methodsFor: 'fileIn/fileOut' stamp: 'tak 11/16/2004 11:04'!
fileOutHeader
	^ '''Translation dictionary'''! !

!NaturalLanguageTranslator methodsFor: 'fileIn/fileOut' stamp: 'tak 11/28/2004 14:50'!
fileOutHeaderOn: aStream 
	aStream nextChunkPut: self fileOutHeader;
		 cr.
	aStream timeStamp; cr.
	aStream nextPut: $!!.
	aStream nextChunkPut: '(' , self class name , ' localeID: ' , id storeString , ')'.
	aStream cr! !

!NaturalLanguageTranslator methodsFor: 'fileIn/fileOut' stamp: 'yo 11/29/2005 11:19'!
fileOutOn: aStream 
	"self current fileOutOn: Transcript. Transcript endEntry"
	self fileOutHeaderOn: aStream.
	self fileOutOn: aStream keys: nil! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'KR 10/30/2007 18:41'!
isDomainLoaded: aDomainName
	"ask if translation dictionary for the domain has been loaded and available now for use"
	self subclassResponsibility.! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:41'!
isoCountry
	^self localeID isoCountry! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:42'!
isoLanguage
	^self localeID isoLanguage! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'mir 7/15/2004 14:42'!
localeID
	^id! !

!NaturalLanguageTranslator methodsFor: 'initialize-release' stamp: 'KR 10/15/2007 12:50'!
localeID: anID
	id := anID! !

!NaturalLanguageTranslator methodsFor: 'fileIn/fileOut' stamp: 'tak 11/16/2004 09:26'!
nextChunkPut: anObject on: aStream 
	| i remainder terminator |
	terminator := $!!.
	remainder := anObject storeString.
	[(i := remainder indexOf: terminator) = 0]
		whileFalse: [aStream
				nextPutAll: (remainder copyFrom: 1 to: i).
			aStream nextPut: terminator.
			"double imbedded terminators"
			remainder := remainder copyFrom: i + 1 to: remainder size].
	aStream nextPutAll: remainder.
	aStream nextPut: terminator; cr.! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'yo 7/30/2004 13:03'!
phrase: phraseString translation: translationString 
	self generics at: phraseString put: translationString asString.
	self changed: #translations.
	self changed: #untranslated.! !

!NaturalLanguageTranslator methodsFor: 'printing' stamp: 'nk 8/29/2004 10:51'!
printOn: aStream
	aStream nextPutAll: self class name; nextPut: $(; print: self localeID; nextPut: $)! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'yo 8/2/2004 12:27'!
rawPhrase: phraseString translation: translationString 
	self generics at: phraseString put: translationString asString.
! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'yo 1/14/2005 16:25'!
rawRemoveUntranslated: untranslated

	self class allKnownPhrases removeKey: untranslated ifAbsent: [].
	self changed: #untranslated.! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'yo 8/1/2004 01:07'!
removeTranslationFor: phraseString
	self generics removeKey: phraseString ifAbsent: [].
	self changed: #translations.
	self changed: #untranslated.! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'yo 1/14/2005 16:25'!
removeUntranslated: untranslated

	self class allKnownPhrases removeKey: untranslated ifAbsent: [].
! !

!NaturalLanguageTranslator methodsFor: 'fileIn/fileOut' stamp: 'cwp 3/22/2014 19:45'!
scanFrom: aStream 
	^ self scanFrom: aStream environment: Environment current! !

!NaturalLanguageTranslator methodsFor: 'fileIn/fileOut' stamp: 'cwp 6/20/2012 17:26'!
scanFrom: aStream environment: anEnvironment
	"Read a definition of dictionary.  
	Make sure current locale corresponds my locale id"
	| newTranslations currentPlatform |
	newTranslations := Dictionary new.
	currentPlatform := Locale currentPlatform.
	[| aString assoc |
	Locale currentPlatform: (Locale localeID: id).
	[aString := aStream nextChunk withSqueakLineEndings.
	aString size > 0] whileTrue: 
		[assoc := Compiler evaluate: aString environment: anEnvironment.
		assoc value = ''
			ifTrue: [self class registerPhrase: assoc key]
			ifFalse: [newTranslations add: assoc]]]
		ensure: [Locale currentPlatform: currentPlatform].
	self mergeTranslations: newTranslations! !

!NaturalLanguageTranslator methodsFor: 'language switching' stamp: 'KR 10/17/2007 01:57'!
setCurrent
	"notify locale of the translator become current"
! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'KR 10/17/2007 01:34'!
translate: aString
	^self translate: aString 
		inDomain: TextDomainManager defaultDomain! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'tak 10/19/2007 13:38'!
translate: aString inDomain: aDomainName
	^ aString! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'mir 6/30/2004 20:22'!
translate: aString in: aContext! !

!NaturalLanguageTranslator methodsFor: 'translation' stamp: 'mir 7/15/2004 14:58'!
translationFor: aString
	^self translate: aString! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'mir 7/21/2004 17:00'!
translations
	^self generics! !

!NaturalLanguageTranslator methodsFor: 'accessing' stamp: 'mir 7/21/2004 17:03'!
untranslated
	| translations |
	translations := self translations.
	^self class allKnownPhrases reject: [:each | translations includesKey: each]! !

!NaturalLanguageTranslator methodsFor: 'fileIn/fileOut' stamp: 'nice 5/1/2015 01:36'!
writeAsMimeString

	| fileName fileStream tmpStream |
	tmpStream := MultiByteBinaryOrTextStream on: ''.
	tmpStream converter: UTF8TextConverter new.
	self fileOutOn: tmpStream.

	fileName := id isoString, '.translation.gz.mime'.
	fileStream := FileStream newFileNamed: fileName.
	fileStream nextPutAll: tmpStream contents asByteArray zipped base64Encoded.
	fileStream close.
! !

Object subclass: #ObjectScanner
	instanceVariableNames: 'pvt3SmartRefStrm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Object Storage'!

!ObjectScanner commentStamp: '<historical>' prior: 0!
An instance of this class is the compiler's context for filing in a SmartRefStream containing instance-specific classes.  When the old name of a new object's class conflicts with an existing class name, install a class var in me.  It has the old name but points at the new class.  The compiler uses it when compiling the code in the fileIn.  Fill the SmartRefStream's renamed class dictionary.

An object fileout:
!!ObjectScanner new initialize!!      "allow me to take control with scanFrom:"

Player subclass: Player23 instanceVariableNames: 'foo' classVariableNames: '' 
	poolDictionaries: nil category: 'Instance Specific'!!
	"I prescan this and (self rename: #Player23 toBe: #Player30)"

!!Player23 methodsFor: 'all' stamp: 'tk 3/9/98 18:58'!!	"actually sent to Player30"
foo
	^ foo!! !!

!!self smartRefStream!!<binary representation of the objects>!!


!

!ObjectScanner methodsFor: 'initialize-release' stamp: 'tk 3/15/98 20:17'!
clear
	"remove all old class vars.  They were UniClasses being remapped to aviod a name conflict."

	self class classPool keys do: [:key |
		self class classPool removeKey: key].	"brute force"! !

!ObjectScanner methodsFor: 'initialize-release' stamp: 'tk 8/15/1998 15:26'!
initialize
	"remove all old class vars that are not instance-specific classes being renamed"

	self clear.
	"Most importantly, return self, so a fileIn will let ObjectScanner seize control.  So UniClasses can be remapped.  See the transfer of control where ReadWriteStream fileIn calls scanFrom:"! !

!ObjectScanner methodsFor: 'scanning' stamp: 'yo 11/11/2002 10:27'!
lookAhead: aChunk
	"See if this chunk is a class Definition, and if the new class name already exists and is instance-specific.  Modify the chunk, and record the rename in the SmartRefStream and in me."

	| pieces sup oldName existing newName newDefn |
	aChunk size < 90 ifTrue: [^ aChunk].		"class defn is big!!"
	(aChunk at: 1) == $!! ifTrue: [^ aChunk].	"method def, fast exit"
	pieces := (aChunk copyFrom: 1 to: (300 min: aChunk size)) findTokens: ' #	\' withCRs.
	pieces size < 3 ifTrue: [^ aChunk].	"really bigger, but just took front"
	(pieces at: 2) = 'subclass:' ifFalse: [^ aChunk].
	sup := Smalltalk at: (pieces at: 1) asSymbol ifAbsent: [^ aChunk].
	sup class class == Metaclass ifFalse: [^ aChunk].
	((oldName := pieces at: 3) at: 1) canBeGlobalVarInitial ifFalse: [^ aChunk].
	oldName := oldName asSymbol.
	(Smalltalk includesKey: oldName) ifFalse: [^ aChunk].	"no conflict"
	existing := Smalltalk at: oldName.
	(existing isKindOf: Class) ifFalse: [^ aChunk].	"Write over non-class global"
	existing isSystemDefined ifTrue: [^ aChunk].	"Go ahead and redefine it!!"
	"Is a UniClass"
	newName := sup chooseUniqueClassName.
	newDefn := aChunk copyReplaceAll: oldName with: newName.
	Compiler evaluate: newDefn for: self logged: true.	"Create the new class"
	self rename: oldName toBe: newName.
	^ newName asString		"to be evaluated"
! !

!ObjectScanner methodsFor: 'utilities' stamp: 'cwp 1/18/2014 10:08'!
rename: existingName toBe: newName
	"See if there is a conflict between what the fileIn wants to call the new UniClass (Player23) and what already exists for another unique instance.  If conflict, make a class variable to intercept the existingName and direct it to class newName."

	self flag: #environments.
	existingName = newName ifFalse: [
		self class ensureClassPool.	"create the dictionary"
		"can't use addClassVarName: because it checks for conflicts with Smalltalk"
		(self class classPool includesKey: existingName) ifFalse: 
			["Pick up any refs in Undeclared"
			self class classPool declare: existingName from: Undeclared].
		self class classPool at: existingName put: (Smalltalk at: newName).
		pvt3SmartRefStrm renamed at: existingName put: newName]! !

!ObjectScanner methodsFor: 'scanning' stamp: 'tk 3/15/98 20:22'!
scanFrom: aByteStream
	"Sieze control of the fileIn.  Put myself in as the context.  If any UniClasses (for just one instance) are defined, they will do it through me, and I will look for conflicting class names.  If so, install the old name as a class var of me, so the compile will work.  Tell my SmartRefStream about renaming the class."

	| valWithOddName47 scannerNamed53 chunkNamed117 |
	pvt3SmartRefStrm := SmartRefStream on: aByteStream.
	aByteStream ascii.
	[aByteStream atEnd] whileFalse:
		[aByteStream skipSeparators.
		valWithOddName47 := (aByteStream peekFor: $!!)
			ifTrue: [chunkNamed117 := aByteStream nextChunk.	"debug"
					scannerNamed53 := Compiler evaluate: chunkNamed117
							for: self logged: false.
					scannerNamed53 class == self class 
						ifTrue: ["I already am the scanner for this file"]
						ifFalse: [scannerNamed53 scanFrom: aByteStream]]
			ifFalse: [chunkNamed117 := aByteStream nextChunk.
					chunkNamed117 := self lookAhead: chunkNamed117.
					Compiler evaluate: chunkNamed117 for: self logged: true].
		aByteStream skipStyleChunk].
	^ valWithOddName47! !

!ObjectScanner methodsFor: 'scanning' stamp: 'cwp 6/20/2012 17:39'!
scanFrom: aByteStream environment: anEnvironment
	"This should probably be reimplemented using an environment
	for compilation. For now, don't change anything"
	^ self scanFrom: aByteStream! !

!ObjectScanner methodsFor: 'accessing' stamp: 'tk 3/15/98 19:33'!
smartRefStream

	^ pvt3SmartRefStrm! !

!Object class methodsFor: '*System-Object Events-class initialization' stamp: 'rw 2/10/2002 13:09'!
flushEvents
	"Object flushEvents"

	EventManager flushEvents. ! !

!Object class methodsFor: '*System-Preferences' stamp: 'mt 7/1/2016 17:44'!
themeProperties
	"Return a list of triples that describe configurable properties in a theme."
	^ {}! !

!Object methodsFor: '*System-Finalization' stamp: 'ar 3/21/98 16:26'!
actAsExecutor
	"Prepare the receiver to act as executor for any resources associated with it"
	self breakDependents! !

!Object methodsFor: '*System-Object Events-accessing' stamp: 'jmv 5/13/2009 10:14'!
actionForEvent: anEventSelector
    "Answer the action to be evaluated when <anEventSelector> has been triggered."

	^self actionMap
		at: anEventSelector asSymbol
		ifAbsent: [nil]! !

!Object methodsFor: '*System-Object Events-accessing' stamp: 'nice 8/22/2010 20:57'!
actionForEvent: anEventSelector
ifAbsent: anExceptionBlock
    "Answer the action to be evaluated when <anEventSelector> has been triggered."

	| actions |
	actions := self actionMap
		at: anEventSelector asSymbol
		ifAbsent: [nil].
	^actions ifNil: [anExceptionBlock value]! !

!Object methodsFor: '*System-Object Events-accessing' stamp: 'reThink 2/18/2001 14:43'!
actionMap

	^EventManager actionMapFor: self! !

!Object methodsFor: '*System-Object Events-accessing' stamp: 'rw 4/27/2002 08:35'!
actionSequenceForEvent: anEventSelector

    ^(self actionMap
        at: anEventSelector asSymbol
        ifAbsent: [^WeakActionSequence new])
            asActionSequence! !

!Object methodsFor: '*System-Object Events-accessing' stamp: 'SqR 6/28/2001 13:19'!
actionsDo: aBlock

	self actionMap do: aBlock! !

!Object methodsFor: '*System-Change Notification-events' stamp: 'nk 8/27/2003 16:23'!
actionsWithReceiver: anObject forEvent: anEventSelector

	^(self actionSequenceForEvent: anEventSelector)
                select: [:anAction | anAction receiver == anObject ]! !

!Object methodsFor: '*System-Preferences' stamp: 'mt 7/1/2016 17:50'!
applyUserInterfaceTheme
	"A new UserInterfaceTheme has been asked to #apply itself onto the current desktop.  The receiver should do what is necessary to change its colors on the screen, update its fonts, form images, etc., if the current UserInterfaceTheme specifies overrides for those values.
	
	Only subclasses which implement new #themeProperties on the class-side should implement this method.
	Implementors of this method should not call 'super applyUserInterfaceTheme'.

	The default implementation does nothing."! !

!Object methodsFor: '*System-Change Notification-converting' stamp: 'rw 4/27/2002 07:48'!
asActionSequence

	^WeakActionSequence with: self! !

!Object methodsFor: '*System-Change Notification-converting' stamp: 'rw 7/20/2003 16:03'!
asActionSequenceTrappingErrors

	^WeakActionSequenceTrappingErrors with: self! !

!Object methodsFor: '*System-Tools-breakpoint' stamp: 'bkv 7/1/2003 12:33'!
break
	"This is a simple message to use for inserting breakpoints during debugging.
	The debugger is opened by sending a signal. This gives a chance to restore
	invariants related to multiple processes."

	BreakPoint signal.

	"nil break."! !

!Object methodsFor: '*System-Preferences' stamp: 'mt 7/1/2016 17:50'!
canApplyUserInterfaceTheme
	"Use this hook to control whether you want to get notified if a new theme gets applied."

	^ true! !

!Object methodsFor: '*System-Object Events-accessing' stamp: 'rw 2/10/2002 13:05'!
createActionMap

	^IdentityDictionary new! !

!Object methodsFor: '*System-Finalization' stamp: 'ar 3/20/98 22:19'!
executor
	"Return an object which can act as executor for finalization of the receiver"
	^self shallowCopy actAsExecutor! !

!Object methodsFor: '*System-Finalization' stamp: 'ar 5/19/2003 20:10'!
finalizationRegistry
	"Answer the finalization registry associated with the receiver."
	^WeakRegistry default! !

!Object methodsFor: '*System-Finalization' stamp: 'ar 3/21/98 16:27'!
finalize
	"Finalize the resource associated with the receiver. This message should only be sent during the finalization process. There is NO garantuee that the resource associated with the receiver hasn't been free'd before so take care that you don't run into trouble - this all may happen with interrupt priority."! !

!Object methodsFor: '*System-Object Events-accessing' stamp: 'SqR 2/19/2001 14:04'!
hasActionForEvent: anEventSelector
    "Answer true if there is an action associated with anEventSelector"

    ^(self actionForEvent: anEventSelector) notNil! !

!Object methodsFor: '*System-Finalization' stamp: 'Igor.Stasenko 5/25/2010 04:59'!
hasMultipleExecutors
	"All objects, except ObjectFinalizerCollection instances should answer false to this message"
	^ false! !

!Object methodsFor: '*System-Support-user interface' stamp: ''!
initialExtent
	"Answer the desired extent for the receiver when a view on it is first opened on the screen. 
	5/22/96 sw: in the absence of any override, obtain from RealEstateAgent"

	^ RealEstateAgent standardWindowExtent! !

!Object methodsFor: '*System-Localization-locales' stamp: 'tfel 7/28/2016 16:52'!
localeChanged
	"Can react to changed locales"! !

!Object methodsFor: '*System-Localization-locales' stamp: 'fbs 5/22/2013 17:13'!
localeChangedGently
	self localeChanged.! !

!Object methodsFor: '*System-Recovery-error handling' stamp: 'mt 5/9/2016 12:59'!
primitiveError: aString 
	Project handlePrimitiveError: aString.! !

!Object methodsFor: '*System-Object Events-accessing-removing' stamp: 'reThink 2/18/2001 15:33'!
releaseActionMap

	EventManager releaseActionMapFor: self! !

!Object methodsFor: '*System-Object Events-accessing-removing' stamp: 'reThink 2/18/2001 15:33'!
removeActionsForEvent: anEventSelector

    | map |
    map := self actionMap.
    map removeKey: anEventSelector asSymbol ifAbsent: [].
    map isEmpty
        ifTrue: [self releaseActionMap]! !

!Object methodsFor: '*System-Object Events-accessing-removing' stamp: 'nk 8/25/2003 21:46'!
removeActionsSatisfying: aBlock

	self actionMap keys do:
		[:eachEventSelector |
			self
   				removeActionsSatisfying: aBlock
				forEvent: eachEventSelector
		]! !

!Object methodsFor: '*System-Object Events-accessing-removing' stamp: 'ul 8/2/2011 20:06'!
removeActionsSatisfying: aOneArgBlock 
forEvent: anEventSelector

    self
        setActionSequence:
            ((self actionSequenceForEvent: anEventSelector)
                reject: aOneArgBlock)
        forEvent: anEventSelector! !

!Object methodsFor: '*System-Object Events-accessing-removing' stamp: 'rw 7/29/2003 17:18'!
removeActionsWithReceiver: anObject

	self actionMap copy keysDo:
		[:eachEventSelector |
			self
   				removeActionsSatisfying: [:anAction | anAction receiver == anObject]
				forEvent: eachEventSelector
		]! !

!Object methodsFor: '*System-Object Events-accessing-removing' stamp: 'reThink 2/18/2001 15:36'!
removeActionsWithReceiver: anObject
forEvent: anEventSelector

    self
        removeActionsSatisfying:
            [:anAction |
            anAction receiver == anObject]
        forEvent: anEventSelector! !

!Object methodsFor: '*System-Object Events-accessing-removing' stamp: 'reThink 2/18/2001 15:31'!
removeAction: anAction
forEvent: anEventSelector

    self
        removeActionsSatisfying: [:action | action = anAction]
        forEvent: anEventSelector! !

!Object methodsFor: '*System-Change Notification-events' stamp: 'nk 8/27/2003 17:45'!
renameActionsWithReceiver: anObject forEvent: anEventSelector toEvent: newEvent

	| oldActions newActions |
	oldActions := Set new.
	newActions := Set new.
	(self actionSequenceForEvent: anEventSelector) do: [ :action |
		action receiver == anObject
			ifTrue: [ oldActions add: anObject ]
			ifFalse: [ newActions add: anObject ]].
	self setActionSequence: (ActionSequence withAll: newActions) forEvent: anEventSelector.
	oldActions do: [ :act | self when: newEvent evaluate: act ].! !

!Object methodsFor: '*System-Finalization' stamp: 'ar 3/21/98 18:38'!
retryWithGC: execBlock until: testBlock
	"Retry execBlock as long as testBlock returns false. Do an incremental GC after the first try, a full GC after the second try."
	| blockValue |
	blockValue := execBlock value.
	(testBlock value: blockValue) ifTrue:[^blockValue].
	Smalltalk garbageCollectMost.
	blockValue := execBlock value.
	(testBlock value: blockValue) ifTrue:[^blockValue].
	Smalltalk garbageCollect.
	^execBlock value.! !

!Object methodsFor: '*System-Object Events-accessing' stamp: 'jmv 5/13/2009 10:16'!
setActionSequence: actionSequence
forEvent: anEventSelector

	| action |
	"This is a good time to compact the action sequence of old, garbage collected stuff."
	action := actionSequence asMinimalRepresentation.
	action
		ifNil: [ self removeActionsForEvent: anEventSelector]
		ifNotNil: [
			self updateableActionMap
				at: anEventSelector asSymbol
				put: action]! !

!Object methodsFor: '*system-support' stamp: 'dvf 8/23/2003 12:27'!
systemNavigation

	^ SystemNavigation default! !

!Object methodsFor: '*System-Finalization' stamp: 'ul 2/26/2010 14:04'!
toFinalizeSend: aSelector to: aFinalizer with: aResourceHandle
	"When I am finalized (e.g., garbage collected) close the associated resource handle by sending aSelector to the appropriate finalizer (the guy who knows how to get rid of the resource).
	WARNING: Neither the finalizer nor the resource handle are allowed to reference me. If they do, then I will NEVER be garbage collected. Since this cannot be validated here, it is up to the client to make sure this invariant is not broken."
	self == aFinalizer ifTrue:[self error: 'I cannot finalize myself'].
	self == aResourceHandle ifTrue:[self error: 'I cannot finalize myself'].
	^self finalizationRegistry add: self executor:
		(ObjectFinalizer
			receiver: aFinalizer
			selector: aSelector
			argument: aResourceHandle)! !

!Object methodsFor: '*System-Localization-locales' stamp: 'tfel 8/10/2016 09:50'!
translatedNoop
	"This is correspondence gettext_noop() in gettext."
	^ self
! !

!Object methodsFor: '*System-Object Events-accessing-triggering' stamp: 'reThink 2/18/2001 15:22'!
triggerEvent: anEventSelector
	"Evaluate all actions registered for <anEventSelector>. Return the value of the last registered action."

    ^(self actionForEvent: anEventSelector) value! !

!Object methodsFor: '*System-Object Events-accessing-triggering' stamp: 'reThink 2/18/2001 17:09'!
triggerEvent: anEventSelector
ifNotHandled: anExceptionBlock
	"Evaluate all actions registered for <anEventSelector>. Return the value of the last registered action."

    ^(self 
		actionForEvent: anEventSelector
		ifAbsent: [^anExceptionBlock value]) value
! !

!Object methodsFor: '*System-Object Events-accessing-triggering' stamp: 'reThink 2/18/2001 15:21'!
triggerEvent: anEventSelector
withArguments: anArgumentList

    ^(self actionForEvent: anEventSelector)
        valueWithArguments: anArgumentList! !

!Object methodsFor: '*System-Object Events-accessing-triggering' stamp: 'reThink 2/18/2001 15:21'!
triggerEvent: anEventSelector
withArguments: anArgumentList
ifNotHandled: anExceptionBlock

    ^(self 
		actionForEvent: anEventSelector
		ifAbsent: [^anExceptionBlock value])
        valueWithArguments: anArgumentList! !

!Object methodsFor: '*System-Object Events-accessing-triggering' stamp: 'reThink 2/18/2001 14:59'!
triggerEvent: anEventSelector
with: anObject

    ^self 
		triggerEvent: anEventSelector
		withArguments: (Array with: anObject)! !

!Object methodsFor: '*System-Object Events-accessing-triggering' stamp: 'reThink 2/18/2001 14:59'!
triggerEvent: anEventSelector
with: anObject
ifNotHandled: anExceptionBlock

    ^self 
		triggerEvent: anEventSelector
		withArguments: (Array with: anObject)
		ifNotHandled: anExceptionBlock! !

!Object methodsFor: '*System-Object Events-accessing' stamp: 'reThink 2/25/2001 08:50'!
updateableActionMap

	^EventManager updateableActionMapFor: self! !

!Object methodsFor: '*System-Preferences' stamp: 'mt 7/1/2016 17:51'!
userInterfaceTheme
	"Call this to conveniently access properties from the current user interface theme."

	^ UserInterfaceTheme current
		pushScope: self;
		yourself! !

!Object methodsFor: '*System-Object Events-accessing-registering' stamp: 'reThink 2/18/2001 15:04'!
when: anEventSelector evaluate: anAction 

	| actions |
	actions := self actionSequenceForEvent: anEventSelector.
	(actions includes: anAction)
		ifTrue: [^ self].
	self 
		setActionSequence: (actions copyWith: anAction)
		forEvent: anEventSelector! !

!Object methodsFor: '*System-Object Events-accessing-registering' stamp: 'rww 12/30/2002 10:37'!
when: anEventSelector
send: aMessageSelector
to: anObject
 
    self
        when: anEventSelector
        evaluate: (WeakMessageSend
            receiver: anObject
            selector: aMessageSelector)! !

!Object methodsFor: '*System-Object Events-accessing-registering' stamp: 'rww 12/30/2002 10:37'!
when: anEventSelector
send: aMessageSelector
to: anObject
withArguments: anArgArray
 
    self
        when: anEventSelector
        evaluate: (WeakMessageSend
            receiver: anObject
            selector: aMessageSelector
		arguments: anArgArray)! !

!Object methodsFor: '*System-Object Events-accessing-registering' stamp: 'rww 12/30/2002 10:37'!
when: anEventSelector
send: aMessageSelector
to: anObject
with: anArg
 
    self
        when: anEventSelector
        evaluate: (WeakMessageSend
            receiver: anObject
            selector: aMessageSelector
		arguments: (Array with: anArg))! !

Object subclass: #PowerManagement
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!PowerManagement commentStamp: '<historical>' prior: 0!
This class is used to restore the VM's view of the power manage option. On the macintosh in a 3.x VM  you can  turn the power manager on or off. This may make your iBook run faster, and will prevents your macintosh from sleeping. This does consume more power so your battery life will be degraded. 

Usually you can turn processor cycling off in the control panel, but this class ensures it's off if you set the preference to turn power management off!

!PowerManagement class methodsFor: 'startup logic' stamp: 'JMM 2/2/2001 13:10'!
deinstall
	"PowerManagement deinstall"

	Smalltalk removeFromStartUpList: self.
! !

!PowerManagement class methodsFor: 'power management' stamp: 'sd 4/29/2003 21:55'!
disablePowerManager
	self disablePowerManager: 1! !

!PowerManagement class methodsFor: 'power management' stamp: 'sd 4/29/2003 21:56'!
disablePowerManager: aInteger 
	"Disable/Enable the architectures power manager by passing in nonzero 
	or zero"
	<primitive: 'primitiveDisablePowerManager'> "primitiveExternalCall" 
	^ self! !

!PowerManagement class methodsFor: 'power management' stamp: 'sd 4/29/2003 21:57'!
enablePowerManager
	self disablePowerManager: 0! !

!PowerManagement class methodsFor: 'class initialization' stamp: 'JMM 2/2/2001 13:09'!
initialize
	"PowerManagement initialize"

	Smalltalk addToStartUpList: self! !

!PowerManagement class methodsFor: 'computing' stamp: 'nice 12/25/2013 22:01'!
itsyVoltage
	"On the Itsy, answer the approximate Vcc voltage. The Itsy will shut 
	itself down when this value reaches 2.0 volts. This method allows one to 
	build a readout of the current battery condition."
	| n |
	n := Smalltalk getSystemAttribute: 1200.
	n
		ifNil: [^ 'no voltage attribute'].
	^ (n asNumber / 150.0 printShowingMaxDecimalPlaces: 2) , ' volts'! !

!PowerManagement class methodsFor: 'startup logic' stamp: 'sd 4/29/2003 22:05'!
startUp
	[Preferences turnOffPowerManager
		ifTrue: [self disablePowerManager]]
		ifError: []! !

Object subclass: #Preference
	instanceVariableNames: 'name value defaultValue helpString localToProject categoryList changeInformee changeSelector type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Preferences'!

!Preference commentStamp: '<historical>' prior: 0!
Represents a true/false flag that is under user control and which can be interrogated by a call to Preferences
	viewRegistry		the registry of the classes responsible for building my view
	name 				a symbol, the formal name of the preference.
	value				a boolean, the current value
	defaultValue		the default value of the preference
	helpString 			string or text, constituting the help message
	localToProject		boolean, whether each project holds its own version
	categoryList			list of categories under which to offer this
	changeInformee 	whom, if anyone, to inform if the value changes:
	changeSelector 		what selector to send to the changeInformee when the value changes!

Preference subclass: #PragmaPreference
	instanceVariableNames: 'provider getter setter'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Preferences'!

!PragmaPreference commentStamp: 'ar 3/9/2009 21:27' prior: 0!
Differs from superclass by redirecting all queries to preference provider.!

!PragmaPreference methodsFor: 'converting' stamp: 'mt 8/26/2015 16:09'!
asPreference
	| preference |
	preference := Preference new.
	preference
		name: name
		defaultValue: defaultValue
		helpString: helpString
		localToProject: self localToProject
		categoryList: categoryList
		changeInformee: changeInformee
		changeSelector:  changeSelector
		type: type.
	preference rawValue: self preferenceValue.
	^preference! !

!PragmaPreference methodsFor: 'accessing' stamp: 'mt 8/26/2015 16:48'!
id

	^ (self provider name, '>>', getter) asSymbol! !

!PragmaPreference methodsFor: 'as yet unclassified' stamp: 'mt 8/26/2015 16:09'!
localToProject
	"Pragma preferences are stored and accessed outside the scope of the preference mechanism. Hence, they cannot be project-local."
	
	^ false! !

!PragmaPreference methodsFor: 'value' stamp: 'ar 3/9/2009 21:31'!
preferenceValue
	"Answer the current value of the preference"
	^provider perform: getter! !

!PragmaPreference methodsFor: 'value' stamp: 'ar 3/9/2009 21:30'!
preferenceValue: aValue
	"set the value as indicated, and invoke the change selector if appropriate"
	self preferenceValue = aValue ifFalse:[
		self rawValue: aValue.
		self notifyInformeeOfChange].! !

!PragmaPreference methodsFor: 'printing' stamp: 'ar 3/9/2009 21:28'!
printOn: aStream
	"Print a string decribing the receiver to the given stream"

	super printOn: aStream.
	aStream nextPutAll: name storeString, ' ', self value storeString! !

!PragmaPreference methodsFor: 'accessing' stamp: 'kb 11/28/2009 12:17'!
provider

	^provider! !

!PragmaPreference methodsFor: 'initialization' stamp: 'ar 3/9/2009 21:35'!
provider: aClass getter: getterSymbol setter: setterSymbol
	provider := aClass.
	getter := getterSymbol.
	setter := setterSymbol.
! !

!PragmaPreference methodsFor: 'value' stamp: 'ar 3/9/2009 21:29'!
rawValue: aValue
	"set the value as indicated, with no side effects"
	provider perform: setter with: aValue! !

!PragmaPreference methodsFor: 'initialization' stamp: 'ar 3/9/2009 21:28'!
restoreDefaultValue
	"Pragma preferences preserve their current value"! !

!PragmaPreference methodsFor: 'value' stamp: 'ar 3/9/2009 21:31'!
togglePreferenceValue
	"Toggle whether the value of the preference. Self must be a boolean preference."
	self preferenceValue: self preferenceValue not.
	self notifyInformeeOfChange! !

!Preference class methodsFor: 'browsing support' stamp: 'bf 3/10/2016 17:39'!
isChangeSelector: aLiteral
	^Preferences isChangeSelector: aLiteral! !

!Preference methodsFor: 'converting' stamp: 'laza 4/7/2010 15:53'!
asPreference
	^self! !

!Preference methodsFor: 'menu' stamp: 'sw 4/12/2001 23:42'!
categoryList
	"Answer the categoryList"

	^ categoryList! !

!Preference methodsFor: 'initialization' stamp: 'sw 4/29/2001 23:51'!
categoryList: aList
	"Set the receiver's categoryList"

	categoryList := aList! !

!Preference methodsFor: 'change notification' stamp: 'sw 4/12/2001 01:39'!
changeInformee: informee changeSelector: aSelector
	"Set the changeInformee and changeSelector as specified"

	changeInformee := informee.
	changeSelector := aSelector! !

!Preference methodsFor: 'accessing' stamp: 'bf 3/10/2016 16:42'!
changeSelector
	^changeSelector! !

!Preference methodsFor: 'menu' stamp: 'sw 4/13/2001 00:01'!
copyName
	"Copy the name of the given preference to the clipboard"

	Clipboard clipboardText: name asString asText! !

!Preference methodsFor: 'value' stamp: 'sw 4/10/2001 15:01'!
defaultValue
	"Answer this preference's defaultValue"

	^ defaultValue! !

!Preference methodsFor: 'value' stamp: 'sw 4/18/2002 12:15'!
defaultValue: aValue
	"Set the receiver's defaultValue"

	defaultValue := aValue.! !

!Preference methodsFor: 'menu' stamp: 'KR 5/11/2008 14:48'!
helpString
	"Answer the help string provided for the receiver"

	^ helpString ifNil: ['no help available' translatedNoop]! !

!Preference methodsFor: 'accessing' stamp: 'mt 8/26/2015 16:48'!
id

	^ self name asSymbol! !

!Preference methodsFor: 'testing' stamp: 'ar 9/27/2005 21:50'!
isObsolete
	^(changeInformee class isObsolete or:[changeInformee isBehavior and:[changeInformee isObsolete]])! !

!Preference methodsFor: 'local to project' stamp: 'mt 8/26/2015 16:09'!
isProjectLocalString
	"Answer a string representing whether sym is a project-local preference or not"

	| aStr |
	aStr :=  'each project has its own setting'.
	^ self localToProject
		ifTrue:
			['<yes>', aStr]
		ifFalse:
			['<no>', aStr]! !

!Preference methodsFor: 'local to project' stamp: 'sw 4/10/2001 01:14'!
localToProject
	"Answer whether this preference is project-local"

	^ localToProject! !

!Preference methodsFor: 'menu' stamp: 'sw 4/10/2001 15:02'!
name
	"Answer this preference's name"

	^ name! !

!Preference methodsFor: 'initialization' stamp: 'eem 6/30/2015 10:54'!
name: aName defaultValue: aValue helpString: aString localToProject: projectBoolean categoryList: aList changeInformee: informee changeSelector:  aChangeSelector type: aType
	"Initialize the preference from the given values.  There is an extra tolerence here for the symbols #true, #false, and #nil, which are interpreted, when appropriate, as meaning true, false, and nil"

	name := aName asSymbol.
	value := defaultValue := aValue
								caseOf: {
									[#true] -> [true].
									[#false] -> [false] }
								otherwise:
									[aValue].
	helpString := aString.
	localToProject := projectBoolean == true or: [projectBoolean = #true].
	type := aType.
	categoryList := aList
						ifNil: [OrderedCollection with: #unclassified]
						ifNotNil: [aList collect: [:elem | elem asSymbol]].

	changeInformee := (informee == nil or: [informee == #nil]) ifFalse:
							[(informee isSymbol)
								ifTrue: [Smalltalk at: informee]
								ifFalse: [informee]].
	changeSelector  := aChangeSelector! !

!Preference methodsFor: 'change notification' stamp: 'bf 3/10/2016 17:38'!
notifyInformeeOfChange
	"If there is a changeInformee, notify her that I have changed value"
	
	<hasLiteralTest: #isChangeSelector:>
	"To find this method as sender of all changeSelectors"

	changeInformee ifNotNil: [changeInformee perform: changeSelector]! !

!Preference methodsFor: 'value' stamp: 'sw 4/10/2001 15:35'!
preferenceValue
	"Answer the current value of the preference"

	^ value! !

!Preference methodsFor: 'value' stamp: 'sw 4/12/2001 23:28'!
preferenceValue: aValue
	"set the value as indicated, and invoke the change selector if appropriate"

	| oldValue |
	oldValue := value.
	value := aValue.
	oldValue ~~ value ifTrue:
		[self notifyInformeeOfChange]! !

!Preference methodsFor: 'debugging' stamp: 'sw 4/13/2001 00:05'!
printOn: aStream
	"Print a string decribing the receiver to the given stream"

	super printOn: aStream.
	aStream nextPutAll: name storeString, ' ', value storeString! !

!Preference methodsFor: 'value' stamp: 'ar 3/9/2009 21:41'!
provider
	"Compatibility with pragma preferences"
	^nil! !

!Preference methodsFor: 'value' stamp: 'sw 4/12/2001 23:28'!
rawValue: aValue
	"set the value as indicated, with no side effects"

	value := aValue! !

!Preference methodsFor: 'value' stamp: 'sw 4/12/2001 00:04'!
restoreDefaultValue
	"restore the default value to the preference"

	value := defaultValue! !

!Preference methodsFor: 'value' stamp: 'hpt 9/26/2004 16:51'!
togglePreferenceValue
	"Toggle whether the value of the preference. Self must be a boolean preference."
	value := value not.
	self notifyInformeeOfChange! !

!Preference methodsFor: 'local to project' stamp: 'ar 8/9/2009 14:45'!
toggleProjectLocalness
	"Toggle whether the preference should be held project-by-project or globally"

	localToProject := localToProject not.! !

!Preference methodsFor: 'accessing' stamp: 'ar 1/2/2010 23:56'!
type
	"The type of this preference. One of #Boolean, #Color, #String, #Font, #Number etc."
	^type! !

!Preference methodsFor: 'accessing' stamp: 'ar 8/9/2009 15:11'!
type: preferenceType
	"The type of this preference. One of #Boolean, #Color, #String, #Font, #Number etc."
	type := preferenceType! !

Object subclass: #Preferences
	instanceVariableNames: ''
	classVariableNames: 'DesktopColor Parameters'
	poolDictionaries: ''
	category: 'System-Preferences'!

Preferences class
	instanceVariableNames: 'preferencesDictionary'!

!Preferences commentStamp: 'eem 6/30/2015 15:10' prior: 0!
A general mechanism to store preference choices.  The default setup treats any symbol as a potential boolean flag; flags unknown to the preference dictionary are always answered as false.  

	To open the control panel:
		PreferenceBrowser open
	To read how to use the panel (and how to make a preference be per-project):
		 Preferences giveHelpWithPreferences

All messages are on the class side.  There are two kinds of preference definition, preference pragmas (which are preferred) and  preferences local to Preferences.

Preference Pragmas
Preferences can be local to a class or system of classes using preference pragmas.  Look at senders of #preference:category:description:type: and #preference:categoryList:description:type: for examples:
	(self systemNavigation browseAllSelect:
		[:m|
		#(preference:category:description:type: preference:categoryList:description:type:) anySatisfy:
			[:s| (m pragmaAt: s) notNil]])
With a preference pragma, the preference is typically kept in a class variable, local to the class whose method(s) contain(s) the pragma.  Good style is to put the preference pragma in the accessor for the variable; see for example BitBlt class>>#subPixelRenderColorFonts. The pragma serves to declare the preference to Preferences.


Preference-local Preferences
To query a a preference:
	Preferences logDebuggerStackToFile
or some people prefer the more verbose
	Preferences valueOfFlag: #logDebuggerStackToFile

You can make up a new preference any time.  Do not define a new message in Preferences class. Accessor methods are compiled automatically when you add a preference, either as as illustrated below, or by using 

To add a non-pragma preference (e.g. in the Postscript of a fileout):
	Preferences
		addPreference: #samplePreference
		categories: #(general browsing)
		default: true
		balloonHelp: 'This is an example of a preference added by a do-it'
		projectLocal: false
		changeInformee: nil
		changeSelector: nil.

To change a preference programatically:
	Preferences disable: #logDebuggerStackToFile.
Or to turn it on,
	Preferences enable: #logDebuggerStackToFile.
!

Preferences class
	instanceVariableNames: 'preferencesDictionary'!

!Preferences class methodsFor: 'prefs - fonts' stamp: 'jmv 8/4/2009 15:09'!
aaFontsColormapDepth
	"Adjust balance between colored AA text quality (especially if subpixel AA is used) and space / performance.
	5 is optimal quality. Each colorMap takes 128kB of RAM, and takes several seconds to build.
	4 is a reasonable balance. Each colorMap takes 16kB of RAM and builds fast on a fast machine.
	3 is good for slow hardware or memory restrictions. Each colorMap takes 2 kb of RAM."
	^self
		valueOfFlag: #aaFontsColormapDepth
		ifAbsent: [4]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
abbreviatedBrowserButtons
	^ self
		valueOfFlag: #abbreviatedBrowserButtons
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'support - misc' stamp: 'sw 2/24/1999 12:26'!
acceptAnnotationsFrom: aSystemWindow
	"This intricate extraction is based on the precise structure of the annotation-request window.  Kindly avert your eyes."
	| aList |
	aList := aSystemWindow paneMorphs first firstSubmorph submorphs collect:
		[:m |  m contents asSymbol].
	self defaultAnnotationRequests: aList
	! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:26'!
addBooleanPreference: prefSymbol categories: categoryList default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system. Default view for this preference is boolean to keep backward compatibility"

	^ self addPreference: prefSymbol  categories: categoryList  default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #Boolean! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:26'!
addBooleanPreference: prefSymbol category: categorySymbol default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system. Default view for this preference is boolean to keep backward compatibility"

	^ self addPreference: prefSymbol  categories: {categorySymbol} default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #Boolean! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:25'!
addColorPreference: prefSymbol categories: categoryList default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system. Default view for this preference is boolean to keep backward compatibility"

	^ self addPreference: prefSymbol  categories: categoryList  default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #Color! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:25'!
addColorPreference: prefSymbol category: categorySymbol default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system. Default view for this preference is boolean to keep backward compatibility"

	^ self addPreference: prefSymbol  categories: {categorySymbol} default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #Color! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:25'!
addFontPreference: prefSymbol categories: categoryList default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system. Default view for this preference is boolean to keep backward compatibility"

	^ self addPreference: prefSymbol  categories: categoryList  default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #Font! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:25'!
addFontPreference: prefSymbol category: categorySymbol default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system. Default view for this preference is boolean to keep backward compatibility"

	^ self addPreference: prefSymbol  categories: {categorySymbol} default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #Font! !

!Preferences class methodsFor: 'support - misc' stamp: 'sw 10/6/1999 15:20'!
addModelItemsToWindowMenu: aMenu
	aMenu addLine.
	aMenu add: 'restore default preference settings' target: self action: #chooseInitialSettings.
	aMenu add: 'restore default text highlighting' target: self action: #initializeTextHighlightingParameters! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:25'!
addNumericPreference: prefSymbol categories: categoryList default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system. "

	^ self addPreference: prefSymbol  categories: categoryList  default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #Number! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:25'!
addNumericPreference: prefSymbol category: categorySymbol default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system."

	^ self addPreference: prefSymbol  categories: {categorySymbol} default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #Number! !

!Preferences class methodsFor: 'add/remove' stamp: 'topa 8/27/2015 23:35'!
addPragmaPreference: pragma
	"Note that there will be no accessor method generated because the pragma's method does already govern that."
	
	| preference |
	
	((pragma keyword beginsWith: #preference:) and: [self respondsTo: pragma keyword])
		ifFalse: [
			"no pragma pref to be defined. do nothing"
			^ self].
	self assert: pragma methodClass isMeta.
	
	preference := self
		perform: pragma keyword
		withArguments: pragma arguments.
			
	preference 
		provider: pragma methodClass theNonMetaClass
		getter: pragma method selector 
		setter: pragma method selector asMutator.
			
	self atomicUpdatePreferences: [ :copyOfDictionaryOfPreferences |
		copyOfDictionaryOfPreferences
			at: preference id
			put: preference].

	^ preference! !

!Preferences class methodsFor: 'add/remove - convenience' stamp: 'mt 8/26/2015 17:25'!
addPreference: prefSymbol categories: categoryList default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system."

	^ self
		addPreference: prefSymbol
		categories: categoryList
		default: defaultValue
		balloonHelp: helpString
		projectLocal: false
		changeInformee: nil
		changeSelector: nil! !

!Preferences class methodsFor: 'add/remove - convenience' stamp: 'mt 8/26/2015 17:25'!
addPreference: prefSymbol categories: categoryList default: aValue balloonHelp: helpString projectLocal: localBoolean changeInformee: informeeSymbol  changeSelector: aChangeSelector
	"Add an item representing the given preference symbol to the system."

	^ self
		addPreference: prefSymbol
		categories: categoryList
		default: aValue
		balloonHelp: helpString
		projectLocal: localBoolean
		changeInformee: informeeSymbol
		changeSelector: aChangeSelector
		type: (self typeForValue: aValue)
! !

!Preferences class methodsFor: 'add/remove' stamp: 'mt 8/27/2015 09:29'!
addPreference: aName categories: categoryList default: aValue balloonHelp: helpString projectLocal: localBoolean changeInformee: informeeSymbol changeSelector: aChangeSelector type: aType
	"Add or replace a preference as indicated.  Reuses the preexisting Preference object for this symbol, if there is one, so that UI artifacts that interact with it will remain valid."

	| newPreference aPreference |
	newPreference := Preference new
		name: aName asSymbol
		defaultValue: aValue
		helpString: helpString
		localToProject: localBoolean
		categoryList: categoryList
		changeInformee: informeeSymbol
		changeSelector: aChangeSelector
		type: aType;
		yourself.
	aPreference := preferencesDictionary
						 at: newPreference name
						 ifAbsent: [newPreference].
	aPreference == newPreference
		ifTrue: "Atomically add the new preference to the dictionary."
			[self atomicUpdatePreferences:
				[:preferenceDictionaryCopy|
				preferenceDictionaryCopy at: newPreference name put: newPreference]]
		ifFalse: "Use the copyFrom: primitive to atomically update the existing preference."
			[aPreference copyFrom: newPreference].
	self compileAccessorForPreference: aPreference.

	^ aPreference! !

!Preferences class methodsFor: 'add/remove - convenience' stamp: 'mt 8/26/2015 17:25'!
addPreference: prefSymbol category: categorySymbol default: defaultValue
	"Add the given preference, putting it in the given category, with the given default value, and with the given balloon help."

	^ self
		addPreference: prefSymbol
		category: categorySymbol
		default: defaultValue
		balloonHelp: nil.! !

!Preferences class methodsFor: 'add/remove - convenience' stamp: 'mt 8/26/2015 17:25'!
addPreference: prefSymbol category: categorySymbol default: defaultValue balloonHelp: helpString 
	"Add the given preference, putting it in the given category, with the given default value, and with the given balloon help."

	^ self
		addPreference: prefSymbol
		categories: {categorySymbol}
		default: defaultValue
		balloonHelp: helpString.! !

!Preferences class methodsFor: 'add/remove - convenience' stamp: 'mt 8/27/2015 09:03'!
addPreference: prefSymbol default: defaultValue

	^ self
		addPreference: prefSymbol
		category: self unclassifiedCategory
		default: defaultValue
		balloonHelp: nil.! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:25'!
addTextPreference: prefSymbol categories: categoryList default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system. Default view for this preference is boolean to keep backward compatibility"

	^ self addPreference: prefSymbol  categories: categoryList  default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #String! !

!Preferences class methodsFor: 'add/remove - specific' stamp: 'mt 8/26/2015 17:25'!
addTextPreference: prefSymbol category: categorySymbol default: defaultValue balloonHelp: helpString 
	"Add an item repreesenting the given preference symbol to the system. Default view for this preference is boolean to keep backward compatibility"

	^ self addPreference: prefSymbol  categories: {categorySymbol} default:  defaultValue balloonHelp: helpString  projectLocal: false  changeInformee: nil changeSelector: nil type: #String! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
allowCelesteTell
	^ self
		valueOfFlag: #allowCelesteTell
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
allowEtoyUserCustomEvents
	^ self
		valueOfFlag: #allowEtoyUserCustomEvents
		ifAbsent: [false]! !

!Preferences class methodsFor: 'accessing' stamp: 'mt 8/27/2015 10:40'!
allPreferences

	^preferencesDictionary values! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alphabeticalProjectMenu
	^ self
		valueOfFlag: #alphabeticalProjectMenu
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alternateHandlesLook
	^ self
		valueOfFlag: #alternateHandlesLook
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alternativeBrowseIt
	^ self
		valueOfFlag: #alternativeBrowseIt
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alternativeButtonsInScrollBars
	^ self
		valueOfFlag: #alternativeButtonsInScrollBars
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alternativeScrollbarLook
	^ self
		valueOfFlag: #alternativeScrollbarLook
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: 'dgd 3/21/2003 19:23'!
alternativeWindowBoxesLook
	^ self
		valueOfFlag: #alternativeWindowBoxesLook
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alternativeWindowLook
	^ self
		valueOfFlag: #alternativeWindowLook
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alwaysHideHScrollbar
	^ self
		valueOfFlag: #alwaysHideHScrollbar
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alwaysShowConnectionVocabulary
	^ self
		valueOfFlag: #alwaysShowConnectionVocabulary
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alwaysShowHScrollbar
	^ self
		valueOfFlag: #alwaysShowHScrollbar
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
alwaysShowVScrollbar
	^ self
		valueOfFlag: #alwaysShowVScrollbar
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'support - misc' stamp: 'tfel 3/27/2010 13:34'!
annotationEditingWindow
	"Answer a window affording editing of annotations"
	| aPanel ins outs current aWindow aButton info standardHeight standardWidth |
	standardHeight := 200.
	standardWidth := (2 sqrt reciprocal * standardHeight) rounded.
	Smalltalk isMorphic
		ifFalse: [self error: 'annotations can be edited only in morphic'].
	aPanel := AlignmentMorph newRow extent: 2 * standardWidth @ standardHeight.
	ins := AlignmentMorph newColumn extent: standardWidth @ standardHeight.
	ins color: Color green muchLighter.
	ins enableDrop: true;
		 beSticky.
	outs := AlignmentMorph newColumn extent: standardWidth @ standardHeight.
	outs color: Color red muchLighter.
	outs enableDrop: true;
		 beSticky.
	aPanel addMorph: outs;
		 addMorphFront: ins.
	outs position: ins position + (standardWidth @ 0).
	current := self defaultAnnotationRequests.
	info := self annotationInfo.
	current
		do: [:sym | | pair aMorph | 
			pair := info
						detect: [:aPair | aPair first == sym].
			aMorph := StringMorph new contents: pair first.
			aMorph setBalloonText: pair last.
			aMorph enableDrag: true.
			aMorph
				on: #startDrag
				send: #startDrag:with:
				to: aMorph.
			ins addMorphBack: aMorph].
	info
		do: [:aPair | (current includes: aPair first)
				ifFalse: [| aMorph |
					aMorph := StringMorph new contents: aPair first.
					aMorph setBalloonText: aPair last.
					aMorph enableDrag: true.
					aMorph
						on: #startDrag
						send: #startDrag:with:
						to: aMorph.
					outs addMorph: aMorph]].
	aPanel layoutChanged.
	aWindow := SystemWindowWithButton new setLabel: 'Annotations'.
	aButton := SimpleButtonMorph new target: Preferences;
				 actionSelector: #acceptAnnotationsFrom:;
				
				arguments: (Array with: aWindow);
				 label: 'apply';
				 borderWidth: 0;
				 borderColor: Color transparent;
				 color: Color transparent.
	aButton submorphs first color: Color blue.
	aButton setBalloonText: 'After moving all the annotations you want to the left (green) side, and all the ones you do NOT want to the right (pink) side, hit this "apply" button to have your choices take effect.'.
	aWindow buttonInTitle: aButton;
		 adjustExtraButton.
	^ aPanel wrappedInWindow: aWindow"Preferences annotationEditingWindow openInHand"! !

!Preferences class methodsFor: 'prefs - annotations' stamp: 'sw 7/12/2001 18:18'!
annotationInfo 
	"Answer a list of pairs characterizing all the available kinds of annotations; in each pair, the first element is a symbol representing the info type, and the second element is a string providing the corresponding balloon help"

	^ #(

		(timeStamp			'The time stamp of the last submission of the method.')
		(firstComment		'The first comment in the method, if any.')
		(masterComment		'The comment at the beginning of the supermost implementor of the method if any.')
		(documentation		'Comment at beginning of the method or, if it has none, comment at the beginning of a superclass''s implementation of the method')
		(messageCategory	'Which method category the method lies in')
		(sendersCount		'A report of how many senders there of the message.')
		(implementorsCount	'A report of how many implementors there are of the message.')
		(recentChangeSet	'The most recent change set bearing the method.')
		(allChangeSets		'A list of all change sets bearing the method.')
		(priorVersionsCount	'A report of how many previous versions there are of the method' )
		(priorTimeStamp		'The time stamp of the penultimate submission of the method, if any'))! !

!Preferences class methodsFor: 'standard queries' stamp: 'sbw 4/24/2010 10:16'!
annotationPanes
	^ self
		valueOfFlag: #annotationPanes
		ifAbsent: [true]! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 6/12/2001 20:17'!
annotationPanesChanged
	"The setting of the annotationPanes preference changed; react.  Formerly, we replaced prototypes in flaps but this is no longer necessary"! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
ansiAssignmentOperatorWhenPrettyPrinting
	^ self
		valueOfFlag: #ansiAssignmentOperatorWhenPrettyPrinting
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
areaFillsAreTolerant
	^ self
		valueOfFlag: #areaFillsAreTolerant
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
areaFillsAreVeryTolerant
	^ self
		valueOfFlag: #areaFillsAreVeryTolerant
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'accessing' stamp: 'ul 7/1/2015 13:01'!
atomicUpdatePreferences: aBlock 
	"Evaluate aBlock with a copy of the preferences dictionary and
	then assign (assignment is atomic) the copy to the dictionary."

	[
		| originalPreferences copyOfPreferences |
		originalPreferences := preferencesDictionary.
		copyOfPreferences := preferencesDictionary
			ifNil: [ IdentityDictionary new ]
			ifNotNil: [ :dictionary | dictionary copy ].
		aBlock value: copyOfPreferences.
		originalPreferences == preferencesDictionary ifTrue: [
	 		preferencesDictionary := copyOfPreferences.
			^self ] ] repeat! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 8/9/2016 17:08'!
attemptToRestoreClassicFonts
	"If certain fonts formerly used in early versions of Squeak happen to be present in the image, restore them to their corresponding roles.  Not called by any other method -- intended to be invoked via do-it, possibly in a postscript"
	"Preferences attemptToRestoreClassicFonts"

	
	self setDefaultFonts: #(	(setButtonFontTo:		NewYork		12)
		(setCodeFontTo:			NewYork		12)
		(setFlapsFontTo:			ComicBold		16)
		(setEToysFontTo:			ComicBold		16)
		(setListFontTo:			NewYork		12)
		(setMenuFontTo:			NewYork		12)
		(setWindowTitleFontTo:	NewYork		15)
		(setSystemFontTo:		NewYork		12)) .! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
autoAccessors
	^ self
		valueOfFlag: #autoAccessors
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
autoIndent
	^ self
		valueOfFlag: #autoIndent
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
automaticFlapLayout
	^ self
		valueOfFlag: #automaticFlapLayout
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'support - misc' stamp: 'dgd 8/31/2003 18:07'!
automaticFlapLayoutString
	"Answer a string for the automaticFlapLayout menu item"
	^ (self automaticFlapLayout
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'automatic flap layout' translated! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
automaticKeyGeneration
	^ self
		valueOfFlag: #automaticKeyGeneration
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
automaticPlatformSettings
	^ self
		valueOfFlag: #automaticPlatformSettings
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
automaticViewerPlacement
	^ self
		valueOfFlag: #automaticViewerPlacement
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'dgd 9/7/2004 18:35'!
balloonHelpDelayTime
	"Answer the number of milliseconds before a balloon help 
	should be put up on morphs."
	^ Parameters
		at: #balloonHelpDelayTime
		ifAbsent: [800]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
balloonHelpEnabled
	^ self
		valueOfFlag: #balloonHelpEnabled
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
balloonHelpInMessageLists
	^ self
		valueOfFlag: #balloonHelpInMessageLists
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
batchPenTrails
	^ self
		valueOfFlag: #batchPenTrails
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
bigDisplay
	^ self
		valueOfFlag: #bigDisplay
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
biggerCursors
	^ self
		valueOfFlag: #biggerCursors
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
biggerHandles
	^ self
		valueOfFlag: #biggerHandles
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
blinkParen
	^ self
		valueOfFlag: #blinkParen
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 2/15/1999 19:44'!
borderColorWhenRunning
	^ Color green! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
browserNagIfNoClassComment
	^ self
		valueOfFlag: #browserNagIfNoClassComment
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
browserShowsPackagePane
	^ self
		valueOfFlag: #browserShowsPackagePane
		ifAbsent: [false]! !

!Preferences class methodsFor: 'themes - tools' stamp: 'fbs 12/6/2013 20:03'!
browseThemes
	"Open up a message-category browser on the theme-defining methods"
	ToolSet browse: Preferences class selector: #outOfTheBox.! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
browseWithDragNDrop
	^ self
		valueOfFlag: #browseWithDragNDrop
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
browseWithPrettyPrint
	^ self
		valueOfFlag: #browseWithPrettyPrint
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
canRecordWhilePlaying
	^ self
		valueOfFlag: #canRecordWhilePlaying
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
capitalizedReferences
	^ self
		valueOfFlag: #capitalizedReferences
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
caseSensitiveFinds
	^ self
		valueOfFlag: #caseSensitiveFinds
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'support' stamp: 'mt 8/27/2015 10:42'!
categoryList
	"Return all available categories. No duplicates."

	| aSet |
	aSet := Set new.
	self allPreferences do: [ :aPreference | 
		aSet addAll: (
			aPreference categoryList collect: [ :aCategory |
				aCategory asSymbol ]) ].
	
	aSet add: self unclassifiedCategory.
	
	^aSet! !

!Preferences class methodsFor: 'support' stamp: 'mt 8/27/2015 09:09'!
categoryListOfPreference: prefSymbol
	"Return a list of all categories in which the preference occurs"

	^ (self preferenceAt: prefSymbol ifAbsent: [^ Error signal: 'Preference not found!!']) categoryList! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
cautionBeforeClosing
	^ self
		valueOfFlag: #cautionBeforeClosing
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
celesteHasStatusPane
	^ self
		valueOfFlag: #celesteHasStatusPane
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
celesteShowsAttachmentsFlag
	^ self
		valueOfFlag: #celesteShowsAttachmentsFlag
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
changeSetVersionNumbers
	^ self
		valueOfFlag: #changeSetVersionNumbers
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
checkForSlips
	^ self
		valueOfFlag: #checkForSlips
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
checkForUnsavedProjects
	^ self
		valueOfFlag: #checkForUnsavedProjects
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'bp 6/13/2004 17:20'!
chooseBalloonHelpFont

	BalloonMorph chooseBalloonFont! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'tween 8/7/2007 01:59'!
chooseCodeFont
	"Not currently sent, but once protocols are sorted out so that we can disriminate on whether a text object being launched is for code or not, will be reincorporated"

	self 
		chooseFontWithPrompt: 'Code font...' translated 
		andSendTo: self 
		withSelector: #setCodeFontTo: 
		highlightSelector: #standardCodeFont.! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'tween 8/7/2007 02:00'!
chooseEToysFont
	"present a menu with the possible fonts for the eToys"
	self
		chooseFontWithPrompt: 'eToys font...' translated
		andSendTo: self
		withSelector: #setEToysFontTo:
		highlightSelector: #standardEToysFont! !

!Preferences class methodsFor: 'fonts' stamp: 'sw 11/9/2007 17:39'!
chooseEToysTitleFont
	"Present a menu with the possible fonts for etoy titles"

	self
		chooseFontWithPrompt: 'Choose the etoy title font' translated
		andSendTo: self
		withSelector: #setEToysTitleFontTo:
		highlight: self standardEToysTitleFont! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'tween 8/7/2007 02:00'!
chooseFlapsFont
	self 
		chooseFontWithPrompt: 'Flaps font...' translated
		andSendTo: self 
		withSelector: #setFlapsFontTo: 
		highlightSelector: #standardFlapFont! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'ar 8/30/2009 16:44'!
chooseFontWithPrompt: aPrompt andSendTo: aReceiver withSelector: aSelector highlightSelector: highlightSelector
	^UIManager default 
			chooseFont: aPrompt 
			for: aReceiver 
			setSelector: aSelector
			getSelector: highlightSelector ! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'tween 8/7/2007 02:00'!
chooseHaloLabelFont
	"present a menu with the possible fonts for label in halo"
	self
		chooseFontWithPrompt: 'Halo Label font...'
		andSendTo: self
		withSelector: #setHaloLabelFontTo:
		highlightSelector: #standardHaloLabelFont! !

!Preferences class methodsFor: 'initialization' stamp: 'mt 8/27/2015 10:42'!
chooseInitialSettings
	"Restore the default choices for all of the standard Preferences."

	self allPreferences do: [:aPreference |
		aPreference restoreDefaultValue].
	Project current installProjectPreferences! !

!Preferences class methodsFor: 'prefs - text' stamp: 'sw 9/6/2000 18:45'!
chooseInsertionPointColor
	"Let the user indicate what color he wishes to have used for insertion points in text"

	ColorPickerMorph new
		choseModalityFromPreference;
		sourceHand: self currentHand;
		target: self;
		selector: #insertionPointColor:;
		originalColor: self insertionPointColor;
		putUpFor: self currentHand near: self currentHand cursorBounds! !

!Preferences class methodsFor: 'prefs - text' stamp: 'sw 12/7/2001 00:44'!
chooseKeyboardFocusColor
	"Let the user indicate what color he wishes to have used for keyboard-focus feedback"

	ColorPickerMorph new
		choseModalityFromPreference;
		sourceHand: self currentHand;
		target: self;
		selector: #keyboardFocusColor:;
		originalColor: self keyboardFocusColor;
		putUpFor: self currentHand near: self currentHand cursorBounds! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'tween 8/7/2007 02:01'!
chooseListFont
	self 
		chooseFontWithPrompt: 'List font...' translated
		andSendTo: self 
		withSelector: #setListFontTo: 
		highlightSelector: #standardListFont! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'tween 8/7/2007 02:01'!
chooseMenuFont
	self 
		chooseFontWithPrompt: 'Menu font...' translated
		andSendTo: self 
		withSelector: #setMenuFontTo: 
		highlightSelector: #standardMenuFont! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'tween 8/7/2007 02:01'!
chooseStandardButtonFont
	self 
		chooseFontWithPrompt: 'Button font...' translated
		andSendTo: self 
		withSelector: #setButtonFontTo: 
		highlightSelector: #standardButtonFont

! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'tween 8/7/2007 02:01'!
chooseSystemFont
	self 
		chooseFontWithPrompt: 'Default font...' translated
		andSendTo: self 
		withSelector: #setSystemFontTo: 
		highlightSelector: #standardSystemFont! !

!Preferences class methodsFor: 'prefs - text' stamp: 'sw 9/6/2000 18:45'!
chooseTextHighlightColor
	"Let the user choose the text-highlight color"

	ColorPickerMorph new
		choseModalityFromPreference;
		sourceHand: self currentHand;
		target: self;
		selector: #textHighlightColor:;
		originalColor: self textHighlightColor;
		putUpFor: self currentHand near: self currentHand cursorBounds! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'tween 8/7/2007 02:01'!
chooseWindowTitleFont
	self 
		chooseFontWithPrompt: 'Window Title font...' translated
		andSendTo: self 
		withSelector: #setWindowTitleFontTo: 
		highlightSelector: #windowTitleFont! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'sw 12/30/2004 01:42'!
classicHaloSpecs
	"Non-iconic halos with traditional placements"

	"Preferences installClassicHaloSpecs"
	"Preferences resetHaloSpecifications"  "  <-  will result in the standard default halos being reinstalled"
	"NB: listed below in clockwise order"

		^ #(
	"  	selector				horiz		vert			color info						icon key
		---------				------		-----------		-------------------------------		---------------"
	(addMenuHandle:		left			top				(red)							none)
	(addDismissHandle:		leftCenter	top				(red		muchLighter)			'Halo-Dismiss')
	(addGrabHandle:			center		top				(black)							none)
	(addDragHandle:			rightCenter	top				(brown)							none)
	(addDupHandle:			right		top				(green)							none)	
	(addMakeSiblingHandle:		right		top				(green muchDarker)				'Halo-Dup')	
	(addDebugHandle:		right		topCenter		(blue	veryMuchLighter)		none)
	(addPoohHandle:			right		center			(white)							none)
	(addPaintBgdHandle:		right		center			(lightGray)						none)
	(addRepaintHandle:		right		center			(lightGray)						none)
	(addGrowHandle:		right		bottom			(yellow)						none)
	(addScaleHandle:		right		bottom			(lightOrange)					none)
	(addFontEmphHandle:	rightCenter	bottom			(lightBrown darker)				none)
	(addFontStyleHandle:		center		bottom			(lightRed)						none)
	(addFontSizeHandle:		leftCenter	bottom			(lightGreen)						none)

	(addRecolorHandle:		right		bottomCenter	(magenta darker)				none)

	(addRotateHandle:		left			bottom			(blue)							none))

! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'hpt 9/24/2004 23:34'!
classicHalosInForce
	^ (self preferenceAt: #haloTheme) preferenceValue == #classicHaloSpecs! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
classicNavigatorEnabled
	^ self
		valueOfFlag: #classicNavigatorEnabled
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
classicNewMorphMenu
	^ self
		valueOfFlag: #classicNewMorphMenu
		ifAbsent: [false]! !

!Preferences class methodsFor: 'updating - system' stamp: 'em 3/24/2005 15:34'!
classicTilesSettingToggled
	"The current value of the largeTiles flag has changed; now react"

	Smalltalk isMorphic ifTrue:
		[Preferences universalTiles
			ifFalse:
				[self inform: 
'note that this will only have a noticeable
effect if the universalTiles preference is
set to true, which it currently is not' translated]
			ifTrue:
				[World recreateScripts]]! !

!Preferences class methodsFor: 'initialization' stamp: 'fbs 4/17/2013 17:08'!
cleanUp
	self removeObsolete.! !

!Preferences class methodsFor: 'initialization' stamp: 'mt 8/2/2016 15:31'!
cleanUp: aggressive

	aggressive ifTrue: [
		self class organization categories
			select: [:ea | ea beginsWith: '*autogenerated']
			thenDo: [:ea | self class removeCategory: ea]].! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
cmdDotEnabled
	^ self
		valueOfFlag: #cmdDotEnabled
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: 'programmatic 7/15/1999 09:55'!
cmdGesturesEnabled
	"compiled programatically -- return hard-coded preference value"
	^ true! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'programmatic 7/15/1999 09:55'!
cmdKeysInText
	"compiled programatically -- return hard-coded preference value"
	^ true! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
collapseWindowsInPlace
	^ self
		valueOfFlag: #collapseWindowsInPlace
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
colorWhenPrettyPrinting
	^ self
		valueOfFlag: #colorWhenPrettyPrinting
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
compactViewerFlaps
	^ self
		valueOfFlag: #compactViewerFlaps
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'private' stamp: 'mt 8/27/2015 09:28'!
compileAccessorForPreferenceNamed: name value: value
	"Compile a method that returns a simple true or false (depending on the value of aBoolean) when Preferences is sent prefName as a message"

	self class 
		compileSilently: (
			'{1} ^{2}'
				format: {
					name asString.
					value storeString })
		classified: 'prefs - misc'.
	
"Preferences compileAccessorForPreferenceNamed: #testing value: false"! !

!Preferences class methodsFor: 'private' stamp: 'mt 8/13/2016 10:18'!
compileAccessorForPreference: aPreference
	"Compile an accessor method for the given preference"

	self class 
		compileSilently: (
			'{1} ^self valueOfFlag: {2} ifAbsent: [ {3} ]'
				format: {
					aPreference name asString.
					aPreference name asSymbol printString.
					aPreference defaultValue storeString })
		classified: 'standard queries'! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
compressFlashImages
	^ self
		valueOfFlag: #compressFlashImages
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
confirmFirstUseOfStyle
	^ self
		valueOfFlag: #confirmFirstUseOfStyle
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
conversionMethodsAtFileOut
	^ self
		valueOfFlag: #conversionMethodsAtFileOut
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
cpuWatcherEnabled
	^ self
		valueOfFlag: #cpuWatcherEnabled
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'laza 3/24/2000 16:05'!
customHaloSpecs
	"Intended for you to modify to suit your personal preference.  What is implemented in the default here is just a skeleton; in comment at the bottom of this method are some useful lines you may wish to paste in to the main body here, possibly modifying positions, colors, etc..
	Note that in this example, we include:
			Dismiss handle, at top-left
			Menu handle, at top-right
			Resize handle, at bottom-right
			Rotate handle, at bottom-left
			Drag handle, at top-center
			Recolor handle, at left-center.  (this one is NOT part of the standard formulary --
											it is included here to illustrate how to
 											add non-standard halos)
			Note that the optional handles for specialized morphs, such as Sketch, Text, PasteUp, are also included"

	^ #(
	(addDismissHandle:		left			top				(red		muchLighter)			'Halo-Dismiss')
	(addMenuHandle:		right		top				(red)							'Halo-Menu')
	(addDragHandle:			center	top					(brown)							'Halo-Drag')
	(addGrowHandle:		right		bottom			(yellow)						'Halo-Scale')
	(addScaleHandle:		right		bottom			(lightOrange)					'Halo-Scale')

	(addRecolorHandle:		left			center			(green muchLighter lighter)		'Halo-Recolor')

	(addPaintBgdHandle:		right		center			(lightGray)						'Halo-Paint')
	(addRepaintHandle:		right		center			(lightGray)						'Halo-Paint')
	(addFontSizeHandle:		leftCenter	bottom			(lightGreen)						'Halo-FontSize')
	(addFontStyleHandle:		center		bottom			(lightRed)						'Halo-FontStyle')
	(addFontEmphHandle:	rightCenter	bottom			(lightBrown darker)				'Halo-FontEmph')
	(addRotateHandle:		left			bottom			(blue)							'Halo-Rot')

	(addDebugHandle:		right		topCenter		(blue	veryMuchLighter)		'Halo-Debug')
	(addPoohHandle:			right		center			(white)							'Halo-Pooh')


			)

	"  Other useful handles...

  		selector				horiz		vert			color info						icon key
		---------				------		-----------		-------------------------------		---------------

	(addTileHandle:			left			bottomCenter	(lightBrown)					'Halo-Tile')
	(addViewHandle:			left			center			(cyan)							'Halo-View')
	(addGrabHandle:			center		top				(black)							'Halo-Grab')
	(addDragHandle:			rightCenter	top				(brown)							'Halo-Drag')
	(addDupHandle:			right		top				(green)							'Halo-Dup')	
	(addHelpHandle:			center		bottom			(lightBlue)						'Halo-Help')
	(addFewerHandlesHandle:	left		topCenter		(paleBuff)						'Halo-FewerHandles')
	(addPaintBgdHandle:		right		center			(lightGray)						'Halo-Paint')
	(addRepaintHandle:		right		center			(lightGray)						'Halo-Paint')
	"
! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'hpt 9/24/2004 23:34'!
customHalosInForce
	^ (self preferenceAt: #haloTheme) preferenceValue == #customHaloSpecs! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
debugHaloHandle
	^ self
		valueOfFlag: #debugHaloHandle
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
debugLogTimestamp
	^ self
		valueOfFlag: #debugLogTimestamp
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 12/1/1999 13:04'!
debugMenuItemsInvokableFromScripts
	"If true, then items occurring in an object's debug menu will be included in the alternatives offered as arguments to a doMenuItem: tile in the scripting system"
	^ false! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
debugPrintSpaceLog
	^ self
		valueOfFlag: #debugPrintSpaceLog
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
debugShowDamage
	^ self
		valueOfFlag: #debugShowDamage
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
decorateBrowserButtons
	^ self
		valueOfFlag: #decorateBrowserButtons
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'prefs - annotations' stamp: 'mt 8/27/2015 10:03'!
defaultAnnotationInfo

	^ #(timeStamp messageCategory implementorsCount allChangeSets)! !

!Preferences class methodsFor: 'prefs - annotations' stamp: 'sw 2/17/1999 00:40'!
defaultAnnotationRequests
	^ Parameters at: #MethodAnnotations ifAbsent:
		[self setDefaultAnnotationInfo]
	"Preferences annotationInfo"! !

!Preferences class methodsFor: 'prefs - annotations' stamp: 'sw 2/8/1999 10:14'!
defaultAnnotationRequests: newList
	^ Parameters at: #MethodAnnotations put: newList! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sma 6/1/2000 12:08'!
defaultAuthorName
	"Answer the author name to be planted, by default, in a changeset-preamble template.  You can hard-code this to hold your name, thus saving you time when writing the preambles of subsequent changesets"

	^ Utilities authorName! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
defaultFileOutFormatMacRoman
	^ self
		valueOfFlag: #defaultFileOutFormatMacRoman
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 2/1/2000 14:05'!
defaultPaintingExtent
	"Answer the preferred size for the onion-skin paint area when launching a new painting within a paste-up morph.  Feel free to change the parameters to suit your configuration."

	^ 800 @ 600! !

!Preferences class methodsFor: 'defaults' stamp: 'mt 7/28/2016 11:52'!
defaultValueTableForCurrentRelease
	"Answer a table defining default values for all the preferences in the release.  Returns a list of (pref-symbol, boolean-symbol) pairs"

	^  #(
		(abbreviatedBrowserButtons false)
		(alternativeBrowseIt false)
		(annotationPanes false)
		(areaFillsAreTolerant false)
		(areaFillsAreVeryTolerant false)
		(automaticFlapLayout true)
		(automaticKeyGeneration false)
		(automaticPlatformSettings true)
		(automaticViewerPlacement true)
		(balloonHelpEnabled true)
		(balloonHelpInMessageLists false)
		(batchPenTrails false)
		(capitalizedReferences true)
		(caseSensitiveFinds false)
		(cautionBeforeClosing false)
		(changeSetVersionNumbers true)
		(checkForSlips true)
		(checkForUnsavedProjects true)
		(classicNavigatorEnabled false)
		(cmdDotEnabled true)
		(collapseWindowsInPlace false)
		(compactViewerFlaps false)
		(compressFlashImages false)
		(confirmFirstUseOfStyle true)
		(conversionMethodsAtFileOut false)
		(debugHaloHandle true)
		(debugPrintSpaceLog false)
		(debugShowDamage false)
		(decorateBrowserButtons true)
		(diffsInChangeList true)
		(diffsWithPrettyPrint false)
		(dismissAllOnOptionClose false)
		(dragNDropWithAnimation false)
		(eToyFriendly false)
		(eToyLoginEnabled false)
		(enableLocalSave true)
		(extractFlashInHighQuality true)
		(extractFlashInHighestQuality false)
		(fastDragWindowForMorphic true)
		(fenceEnabled true)
		(fullScreenLeavesDeskMargins true)
		(haloTransitions false)
		(higherPerformance false)
		(honorDesktopCmdKeys true)
		(includeSoundControlInNavigator false)
		(infiniteUndo false)
		(logDebuggerStackToFile true)
		(magicHalos false)
		(menuButtonInToolPane false)
		(menuColorFromWorld false)
		(menuKeyboardControl false)  
		(modalColorPickers true)
		(mouseOverForKeyboardFocus false)
		(mouseOverHalos false)
		(mvcProjectsAllowed true)
		(navigatorOnLeftEdge true)
		(noviceMode false)
		(okToReinitializeFlaps true)
		(optionalButtons true)
		(passwordsOnPublish false)
		(personalizedWorldMenu true)
		(postscriptStoredAsEPS false)
		(projectViewsInWindows true)
		(projectZoom true)
		(projectsSentToDisk false)
		(propertySheetFromHalo false)
		(readDocumentAtStartup true)
		(restartAlsoProceeds false)
		(reverseWindowStagger true)
		(roundedMenuCorners true)
		(scrollBarsNarrow false)
		(scrollBarsOnRight true)
		(securityChecksEnabled false)
		(selectiveHalos false)
		(showBoundsInHalo false)
		(showDirectionForSketches false)
		(showDirectionHandles false)
		(showFlapsWhenPublishing false)
		(showProjectNavigator false)
		(showSecurityStatus true)
		(showSharedFlaps true)
		(signProjectFiles true)
		(simpleMenus false)
		(smartUpdating true)
		(startInUntrustedDirectory false)
		(systemWindowEmbedOK false)
		(tileTranslucentDrag true)
		(timeStampsInMenuTitles true)
		(turnOffPowerManager false)
		(twentyFourHourFileStamps true)
		(typeCheckingInTileScripting true)
		(uniTilesClassic true)
		(uniqueNamesInHalos false)
		(universalTiles false)
		(unlimitedPaintArea false)
		(useButtonPropertiesToFire false)
		(useUndo true)
		(viewersInFlaps true)
		(warnAboutInsecureContent true)
		(warnIfNoChangesFile true)
		(warnIfNoSourcesFile true))


"
Preferences defaultValueTableForCurrentRelease do:
	[:pair | (Preferences preferenceAt: pair first ifAbsent: [nil]) ifNotNilDo:
			[:pref | pref defaultValue: (pair last == true)]].
Preferences chooseInitialSettings.
"! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'dgd 9/5/2004 16:17'!
defaultWorldColor
	^ Parameters
		at: #defaultWorldColor
		ifAbsent: [ Color r: 0.937 g: 0.937 b: 0.937 ].
! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'jhm 10/15/97 17:31'!
desktopColor
	"Answer the desktop color. Initialize it if necessary."
	
	DesktopColor == nil ifTrue: [DesktopColor := Color gray].
	^ DesktopColor
! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'jhm 10/15/97 17:31'!
desktopColor: aColor
	"Record a new desktop color preference."

	DesktopColor := aColor.
! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 9/6/2000 05:26'!
desktopMenuTitle
	"Answer the title to be used for the 'meta menu'.  For now, you can hard-code this, later someone should make this be a parameter the user can easily change.  sw 9/6/2000"

	^ 'World'    "This is what it has always been"

	"^ 'Desktop'
	^ 'Squeak'
	^ 'Mike''s Control Panel'"! !

!Preferences class methodsFor: 'accessing' stamp: 'eem 6/30/2015 15:21'!
dictionaryOfPreferences
	"N.B.  Does /not/ answer the dictionary itself.  To burrow that deep use e.g. instVarNamed:"
	^preferencesDictionary copy! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
diffsInChangeList
	^ self
		valueOfFlag: #diffsInChangeList
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
diffsWithPrettyPrint
	^ self
		valueOfFlag: #diffsWithPrettyPrint
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'initialization - misc' stamp: 'mt 8/27/2015 09:30'!
disableProgrammerFacilities
	"Warning: do not call this lightly!!  It disables all access to menus, debuggers, halos.  There is no guaranteed return from this, which is to say, you cannot necessarily reenable these things once they are disabled -- you can only use whatever the UI of the current project affords, and you cannot even snapshot -- you can only quit. 

     You can completely reverse the work of this method by calling the dual Preferences method enableProgrammerFacilities, provided you have left yourself leeway to bring about a call to that method.

	To set up a system that will come up in such a state, you have to request the snapshot in the same breath as you disable the programmer facilities.  To do this, put the following line into the 'do' menu and then evaluate it from that 'do' menu:

         Preferences disableProgrammerFacilities.

You will be prompted for a new image name under which to save the resulting image."

	Beeper beep.
	(self 
		confirm: 'CAUTION!!!!
This is a drastic step!!
Do you really want to do this?') 
			ifFalse: 
				[Beeper beep.
				^self inform: 'whew!!'].
	self disable: #cmdDotEnabled.	"No user-interrupt-into-debugger"
	self compileAccessorForPreferenceNamed: #cmdGesturesEnabled value: false.	"No halos, etc."
	self compileAccessorForPreferenceNamed: #cmdKeysInText value: false.	"No user commands invokable via cmd-key combos in text editor"
	self enable: #noviceMode.	"No control-menu"
	self disable: #warnIfNoSourcesFile.
	self disable: #warnIfNoChangesFile.
	Smalltalk saveAs! !

!Preferences class methodsFor: 'get/set - flags' stamp: 'mt 8/26/2015 17:34'!
disable: aSymbol
	"Shorthand access to enabling a preference of the given name.  If there is none in the image, conjure one up"

	^ self setFlag: aSymbol toValue: false! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
dismissAllOnOptionClose
	^ self
		valueOfFlag: #dismissAllOnOptionClose
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
dismissEventTheatreUponPublish
	^ self
		valueOfFlag: #dismissEventTheatreUponPublish
		ifAbsent: [true]! !

!Preferences class methodsFor: 'updating - system' stamp: 'ar 7/21/2010 20:20'!
displaySizeChanged
	self flag: #todo.
	"only change font on small-land image"
	self smallLandFonts.
	self tinyDisplay
		ifTrue: [self enable: #scrollBarsNarrow]
		ifFalse: [self disable: #scrollBarsNarrow].
	self tinyDisplay 
		ifTrue:[self disable: #biggerHandles] 
		ifFalse:[self enable: #biggerHandles]! !

!Preferences class methodsFor: 'get/set' stamp: 'topa 8/27/2015 23:45'!
doesNotUnderstand: aMessage
	"Interpret unary message selectors as preference id."
	
	^ aMessage arguments size > 0
		ifTrue: [super doesNotUnderstand: aMessage]
		ifFalse: [
			self
				valueOfPreference: aMessage selector
				ifAbsent: [super doesNotUnderstand: aMessage]]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
dragNDropWithAnimation
	^ self
		valueOfFlag: #dragNDropWithAnimation
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
dropProducesWatcher
	^ self
		valueOfFlag: #dropProducesWatcher
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
duplicateAllControlAndAltKeys
	^ self
		valueOfFlag: #duplicateAllControlAndAltKeys
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
duplicateControlAndAltKeys
	^ self
		valueOfFlag: #duplicateControlAndAltKeys
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
easySelection
	^ self
		valueOfFlag: #easySelection
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'support - misc' stamp: 'mt 5/17/2016 16:00'!
editAnnotations
	"Put up a window that allows the user to edit annotation specifications"

	| aWindow |
	self currentWorld addMorphCentered: (aWindow := self annotationEditingWindow).
	aWindow beKeyWindow.

	"Preferences editAnnotations"

! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'ar 9/27/2005 20:32'!
editCustomHalos

	ToolSet browse: Preferences class
		selector: #customHaloSpecs! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
enableInternetConfig
	^ self
		valueOfFlag: #enableInternetConfig
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
enableLocalSave
	^ self
		valueOfFlag: #enableLocalSave
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
enablePortraitMode
	^ self
		valueOfFlag: #enablePortraitMode
		ifAbsent: [false]! !

!Preferences class methodsFor: 'initialization - misc' stamp: 'mt 8/27/2015 09:30'!
enableProgrammerFacilities
	"Meant as a one-touch recovery from a #disableProgrammerFacilities call."
	"Preferences enableProgrammerFacilities"

	self enable: #cmdDotEnabled.
	self compileAccessorForPreferenceNamed: #cmdGesturesEnabled value: true. 
	self compileAccessorForPreferenceNamed: #cmdKeysInText value: true.
	self disable: #noviceMode.
	self enable: #warnIfNoSourcesFile.
	self enable: #warnIfNoChangesFile.! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 7/13/2001 21:34'!
enableProjectNavigator
	"Answer whether the project-navigator menu item should be enabled"

	^ true! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
enableVirtualOLPCDisplay
	^ self
		valueOfFlag: #enableVirtualOLPCDisplay
		ifAbsent: [false]! !

!Preferences class methodsFor: 'get/set - flags' stamp: 'mt 8/26/2015 17:34'!
enable: aSymbol
	"Shorthand access to enabling a preference of the given name.  If there is none in the image, conjure one up"

	^ self setFlag: aSymbol toValue: true! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
escapeKeyProducesMenu
	^ self
		valueOfFlag: #escapeKeyProducesMenu
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
eToyFriendly
	^ self
		valueOfFlag: #eToyFriendly
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 4/12/2001 01:31'!
eToyFriendlyChanged
	"The eToyFriendly preference changed; React"
	
	ScriptingSystem customizeForEToyUsers: Preferences eToyFriendly! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
eToyLoginEnabled
	^ self
		valueOfFlag: #eToyLoginEnabled
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
expandedPublishing
	^ self
		valueOfFlag: #expandedPublishing
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'parameters' stamp: 'sw 5/16/2003 00:27'!
expungeParameter: aKey
	"If Parameters holds an entry under the given key, remove the entry.  No senders in the current system, but called from the postscript of the change-set that defines it, and potentially useful otherwise."

	Parameters removeKey: aKey ifAbsent: []! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
extractFlashInHighestQuality
	^ self
		valueOfFlag: #extractFlashInHighestQuality
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
extractFlashInHighQuality
	^ self
		valueOfFlag: #extractFlashInHighQuality
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
extraDebuggerButtons
	^ self
		valueOfFlag: #extraDebuggerButtons
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
fastDragWindowForMorphic
	^ self
		valueOfFlag: #fastDragWindowForMorphic
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
fenceEnabled
	^ self
		valueOfFlag: #fenceEnabled
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
fenceSoundEnabled
	^ self
		valueOfFlag: #fenceSoundEnabled
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: 'mir 6/7/2002 17:10'!
fenceSoundEnabled: aBoolean
	self setPreference: #fenceSoundEnabled toValue: aBoolean! !

!Preferences class methodsFor: 'support - file list services' stamp: 'kfr 4/28/2015 06:59'!
fileReaderServicesForFile: fullName suffix: suffix 
	^(suffix = 'prefs')  | (suffix = '*') 
		ifTrue: [ self services ]
		ifFalse: [ #() ]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'kfr 4/29/2015 12:15'!
fontConfigurationMenu
	| aMenu |
	aMenu := MenuMorph new defaultTarget: Preferences.
	^self fontConfigurationMenu: aMenu.
	
	
	! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'kfr 4/29/2015 16:47'!
fontConfigurationMenu: aMenu

	aMenu removeAllMorphs.
	aMenu addTitle: 'Standard System Fonts' translated.
	aMenu addStayUpIcons.
	
	aMenu add: 'default text font...' translated action: #chooseSystemFont.
	aMenu lastItem font: Preferences standardDefaultTextFont.
	aMenu balloonTextForLastItem: 'Choose the default font to be used for code and  in workspaces, transcripts, etc.' translated.
	
	
	aMenu add: 'list font...' translated action: #chooseListFont.
	aMenu lastItem font: Preferences standardListFont.
	aMenu balloonTextForLastItem: 'Choose the font to be used in list panes' translated.
	
	aMenu add: 'flaps font...' translated action: #chooseFlapsFont.
	aMenu lastItem font: Preferences standardFlapFont.
	aMenu balloonTextForLastItem: 'Choose the font to be used on textual flap tabs' translated.

	aMenu add: 'eToys font...' translated action: #chooseEToysFont.
	aMenu lastItem font: Preferences standardEToysFont.
	aMenu balloonTextForLastItem: 'Choose the font to be used on eToys environment' translated.

	aMenu add: 'eToys title font...' translated action: #chooseEToysTitleFont.
	aMenu lastItem font: Preferences standardEToysTitleFont.
	aMenu balloonTextForLastItem: 'Choose the font to be used in titles on eToys environment' translated.

	aMenu add: 'halo label font...' translated action: #chooseHaloLabelFont.
	aMenu lastItem font: Preferences standardHaloLabelFont.
	aMenu balloonTextForLastItem: 'Choose the font to be used on labels ih halo' translated.

	aMenu add: 'menu font...' translated action: #chooseMenuFont.
	aMenu lastItem font: Preferences standardMenuFont.
	aMenu balloonTextForLastItem: 'Choose the font to be used in menus' translated.
	
	aMenu add: 'window-title font...' translated action: #chooseWindowTitleFont.
	aMenu lastItem font: Preferences windowTitleFont.
	aMenu balloonTextForLastItem: 'Choose the font to be used in window titles.' translated.

	aMenu add: 'balloon-help font...' translated action: #chooseBalloonHelpFont.
	aMenu lastItem font: Preferences standardBalloonHelpFont.
	aMenu balloonTextForLastItem: 'choose the font to be used when presenting balloon help.' translated.
	
	aMenu add: 'code font...' translated action: #chooseCodeFont. 
	aMenu lastItem font: Preferences standardCodeFont. 
	aMenu balloonTextForLastItem: 'Choose the font to be used in code panes.' translated.

	aMenu add: 'button font...' translated action: #chooseStandardButtonFont.
	aMenu lastItem font: Preferences standardButtonFont.
	aMenu balloonTextForLastItem: 'Choose the font to be used in buttons.' translated.

	aMenu addLine.
	aMenu add: 'demo mode' translated action: #setDemoFonts.
	aMenu balloonTextForLastItem: 'Set Fonts usable for giving a presentation' translated.

	aMenu addLine.
	aMenu add: 'restore default font choices' translated action: #restoreDefaultFonts.
	aMenu balloonTextForLastItem: 'Use the standard system font defaults' translated.
	
	aMenu add: 'print default font choices' translated action: #printStandardSystemFonts.
	aMenu balloonTextForLastItem: 'Print the standard system font defaults to the Transcript' translated.
	
	aMenu addLine.
	aMenu add: 'refresh this menu' translated target: self selector: #fontConfigurationMenu:  argument: aMenu.
	aMenu balloonTextForLastItem: 'Update this menu to reflect the current fonts' translated.
	MenuIcons decorateMenu: aMenu.
	^ aMenu! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
fullScreenLeavesDeskMargins
	^ self
		valueOfFlag: #fullScreenLeavesDeskMargins
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
generalizedYellowButtonMenu
	^ self
		valueOfFlag: #generalizedYellowButtonMenu
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'support' stamp: 'mt 8/27/2015 10:42'!
giveHelpWithPreferences
	"Open up a workspace with explanatory info in it about Preferences"

	| aString |
	aString := String streamContents: [:aStream | 
		aStream nextPutAll:

'Many aspects of the system are governed by the settings of various "Preferences".  

Click on any of brown tabs at the top of the panel to see all the preferences in that category.  
Or type in to the box above the Search button, then hit Search, and all Preferences matching whatever you typed in will appear in the "search results" category.  A preference is considered to match your search if either its name matches the characters *or* if anything in the balloon help provided for the preferences matches the search text.

To find out more about any particular Preference, hold the mouse over it for a moment and balloon help will appear.  Also, a complete list of all the Preferences, with documentation for each, is included below.

Preferences whose names are in shown in bold in the Preferences Panel are designated as being allowed to vary from project to project; those whose name are not in bold are "global", which is to say, they apply equally whatever project you are in.

Click on the name of any preference to get a menu which allows you to *change* whether the preference should vary from project to project or should be global, and also allows you to browse all the senders of the preference, and to discover all the categories under which the preference has been classified, and to be handed a button that you can drop wherever you please that will control the preference.

If you like all your current Preferences settings, you may wish to hit the "Save Current Settings as my Personal Preferences" button.  Once you have done that, you can at any point in the future hit "Restore my Personal Preferences" and all your saved settings will get restored immediately.

Also, you can use "themes" to set multiple preferences all at once; click on the "change theme..." button in the Squeak flap or in the Preferences panel, or seek out the themes item in the Appearance menu.' translated.

	aStream cr; cr; nextPutAll: '-----------------------------------------------------------------';
		cr; cr; nextPutAll:  'Alphabetical listing of all Preferences' translated; cr; cr.
   (Preferences allPreferences asSortedCollection: [:a :b | a name < b name]) do:
	[:pref | | aHelpString |
		aStream nextPutAll: pref name; cr.
		aHelpString := pref helpString translated.
		(aHelpString beginsWith: pref name) ifTrue:
			[aHelpString := aHelpString copyFrom: (pref name size ) to: aHelpString size].
		aHelpString := (aHelpString copyReplaceAll: String cr with: ' ')  copyWithout: Character tab.
		aStream nextPutAll: aHelpString capitalized.
		(aHelpString isEmpty or: [aHelpString last == $.]) ifFalse: [aStream nextPut: $.].
        aStream cr; cr]].

	UIManager default edit: aString label: 'About Preferences' translated

"Preferences giveHelpWithPreferences"! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
gradientMenu
	^ self
		valueOfFlag: #gradientMenu
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
gradientScrollBars
	^ self
		valueOfFlag: #gradientScrollBars
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
haloEnclosesFullBounds
	^ self
		valueOfFlag: #haloEnclosesFullBounds
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'sw 10/30/2000 13:32'!
haloSpecifications
	"Answer a list of HaloSpecs that describe which halos are to be used, what they should look like, and where they should be situated"

	^ Parameters at: #HaloSpecs ifAbsent:
			[self installHaloTheme: #iconicHaloSpecifications.
			^ Parameters at: #HaloSpecs]

	"Preferences haloSpecifications"
	"Preferences resetHaloSpecifications"
! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'dgd 10/21/2004 12:17'!
haloSpecificationsForWorld
	| desired |
	"Answer a list of HaloSpecs that describe which halos are to be used on a world halo, what they should look like, and where they should be situated"
	"Preferences resetHaloSpecifications"

	desired := #(addDebugHandle: addMenuHandle: addTileHandle: addViewHandle: addHelpHandle: addScriptHandle: addPaintBgdHandle: addRecolorHandle:).
	^ self haloSpecifications select:
		[:spec | desired includes: spec addHandleSelector]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
haloTheme
	^ self
		valueOfFlag: #haloTheme
		ifAbsent: [#iconicHaloSpecifications]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
haloTransitions
	^ self
		valueOfFlag: #haloTransitions
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
higherPerformance
	^ self
		valueOfFlag: #higherPerformance
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
honorDesktopCmdKeys
	^ self
		valueOfFlag: #honorDesktopCmdKeys
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'halos' stamp: 'tak 3/20/2008 16:30'!
iconicHaloSpecifications
	"Answer an array that characterizes the locations, colors, icons, and selectors of the halo handles that may be used in the iconic halo scheme"

	"Preferences resetHaloSpecifications"

	^ #(
	"  	selector				horiz		vert			color info						icon key
		---------				------		-----------		-------------------------------		---------------"
	(addCollapseHandle:		left			topCenter		(tan)							'Halo-Collapse')
	(addPoohHandle:			right		center			(white)							'Halo-Pooh')
	(addDebugHandle:		right		topCenter		(blue	veryMuchLighter)		'Halo-Debug')
	(addDismissHandle:		left			top				(red		muchLighter)			'Halo-Dismiss')
	(addRotateHandle:		left			bottom			(blue)							'Halo-Rot')
	(addMenuHandle:		leftCenter	top				(white)							'Halo-Menu')
	(addTileHandle:			left			bottomCenter	(lightBrown)					'Halo-Tile')
	(addViewHandle:			left			center			(cyan)							'Halo-View')
	(addGrabHandle:			center		top				(black)							'Halo-Grab')
	(addDragHandle:			rightCenter	top				(brown)							'Halo-Drag')
	(addDupHandle:			right		top				(green)							'Halo-Dup')	
	(addMakeSiblingHandle:	right		top				(green muchDarker)				'Halo-Dup')	
	(addHelpHandle:			center		bottom			(lightBlue)						'Halo-Help')
	(addGrowHandle:		right		bottom			(yellow)						'Halo-Scale')
	(addScaleHandle:		right		bottom			(lightOrange)					'Halo-Scale')
	(addScriptHandle:		rightCenter	bottom			(green muchLighter)			'Halo-Script')
	(addPaintBgdHandle:		right		center			(lightGray)						'Halo-Paint')
	(addViewingHandle:		leftCenter	bottom			(lightGreen lighter)				'Halo-View')
	(addRepaintHandle:		right		center			(lightGray)						'Halo-Paint')
	(addFontSizeHandle:		leftCenter	bottom			(lightGreen)						'Halo-FontSize')
	(addFontStyleHandle:		center		bottom			(lightRed)						'Halo-FontStyle')
	(addFontEmphHandle:	rightCenter	bottom			(lightBrown darker)				'Halo-FontEmph')
	(addRecolorHandle:		right		bottomCenter	(magenta darker)				'Halo-Recolor')
	(addChooseGraphicHandle:	right	bottomCenter	(green muchLighter)			'Halo-ChooseGraphic')
		) ! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'hpt 9/24/2004 23:34'!
iconicHalosInForce
	^ (self preferenceAt: #haloTheme) preferenceValue == #iconicHaloSpecifications! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
ignoreStyleIfOnlyBold
	^ self
		valueOfFlag: #ignoreStyleIfOnlyBold
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
implicitSelfInTiles
	^ self
		valueOfFlag: #implicitSelfInTiles
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
inboardScrollbars
	^ self
		valueOfFlag: #inboardScrollbars
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
includeSoundControlInNavigator
	^ self
		valueOfFlag: #includeSoundControlInNavigator
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
infiniteUndo
	^ self
		valueOfFlag: #infiniteUndo
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 4/12/2001 01:32'!
infiniteUndoChanged
	"The infiniteUndo preference changed; react"
	
	self infiniteUndo ifFalse:
		[CommandHistory resetAllHistory]! !

!Preferences class methodsFor: 'support - misc' stamp: 'md 2/24/2006 21:26'!
initialExtent
	^ Smalltalk isMorphic ifFalse: [219 @ 309] ifTrue: [232 @ 309]! !

!Preferences class methodsFor: 'class initialization' stamp: 'ar 8/9/2009 14:45'!
initialize
	"Preferences initialize"
	self registerForEvents.! !

!Preferences class methodsFor: 'initialization' stamp: 'eem 6/30/2015 15:22'!
initializeDictionaryOfPreferences
	"Initialize the preferencesDictionary to be an empty IdentityDictionary"

	"Preferences initializeDictionaryOfPreferences"

	preferencesDictionary := IdentityDictionary new! !

!Preferences class methodsFor: 'parameters' stamp: 'sw 9/7/1999 12:45'!
initializeParameters
	"Preferences initializeParameters"
	Parameters := IdentityDictionary new.
	self restoreDefaultMenuParameters.
	Parameters at: #maxBalloonHelpLineLength put: 28.
	self initializeTextHighlightingParameters! !

!Preferences class methodsFor: 'prefs - text' stamp: 'tpr 1/20/2016 17:56'!
initializeTextHighlightingParameters
	"Preferences initializeTextHighlightingParameters"

	self
		insertionPointColor: (TranslucentColor r: 0.0 g: 0.0 b: 0.8 alpha: 0.8);
		textHighlightColor: (TranslucentColor r: 0.0 g: 0.0 b: 0.8 alpha: 0.2)! !

!Preferences class methodsFor: 'prefs - text' stamp: 'sw 9/7/1999 12:53'!
insertionPointColor
	^ Parameters at: #insertionPointColor! !

!Preferences class methodsFor: 'prefs - text' stamp: 'sw 9/7/1999 12:54'!
insertionPointColor: aColor
	Parameters at: #insertionPointColor put: aColor! !

!Preferences class methodsFor: 'parameters' stamp: 'sw 1/24/2001 21:44'!
inspectParameters
	"Open up an inspector on the Parameters of Preferences.  This is crude!!"

	Parameters inspectWithLabel: 'Parameters'! !

!Preferences class methodsFor: 'support - misc' stamp: 'eem 6/30/2015 15:26'!
inspectPreferences
	"Open a window on the current preferences dictionary, allowing the user to inspect and change the current preference settings.  This is fallen back upon if Morphic is not present. This is dangerous, the dictionary of preferences should not be accessed concurrently."

	"Preferences inspectPreferences"

	preferencesDictionary inspectWithLabel: 'Preferences'! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'sw 1/28/2000 10:35'!
installClassicHaloSpecs
	"Install an alternative set of halos,  rather more based on the old placements, and without icons, , and lacking the scripting-relating handles.."
	"Preferences installClassicHaloSpecs"
	"Preferences resetHaloSpecifications"  "  <-  will result in the standard default halos being reinstalled"
	self installHaloTheme: #classicHaloSpecs! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'sw 1/28/2000 10:36'!
installCustomHaloSpecs
	"Install an alternative set of halos, as customized by the user"
	"Preferences installCustomHaloSpecs"
	self installHaloTheme: #customHaloSpecs! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'nice 12/27/2009 03:11'!
installHaloSpecsFromArray: anArray

	
	^ Parameters at: #HaloSpecs put: 
		(anArray collect:
			[:quin | | aColor |
				aColor := Color.
				quin fourth do: [:sel | aColor := aColor perform: sel].
				HaloSpec new 
					horizontalPlacement: quin second
					verticalPlacement: quin third 
					color: aColor
					iconSymbol: quin fifth
					addHandleSelector: quin first])! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'hpt 9/24/2004 23:35'!
installHaloTheme: themeSymbol
	self installHaloSpecsFromArray: (self perform: themeSymbol).
	(self preferenceAt: #haloTheme) preferenceValue: themeSymbol.
	! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'sw 1/28/2000 10:36'!
installIconicHaloSpecs
	"Install an alternative set of halos,  rather more based on the old placements, and without icons, , and lacking the scripting-relating handles.."
	"Preferences installIconicHaloSpecs"
	self installHaloTheme: #iconicHaloSpecifications! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'sw 1/28/2000 10:36'!
installSimpleHaloSpecs
	"Preferences installSimpleHaloSpecs"
	self installHaloTheme: #simpleFullHaloSpecifications! !

!Preferences class methodsFor: 'themes - tools' stamp: 'yo 7/2/2004 19:44'!
installTheme: aSymbol
	"Install the theme represented by aSymbol.  The code that makes the theme-specific changes is lodged in a method of the same name as aSymbol, which must reside in category #themes in Preferences class"

	self perform: aSymbol.
	self inform: ('Theme {1} is now installed.
Many of the changes will only be
noticeable in new windows that you
create from now on.' translated format: {aSymbol translated}).! !

!Preferences class methodsFor: 'reacting to change' stamp: 'bf 3/10/2016 17:40'!
isChangeSelector: aLiteral
	^preferencesDictionary anySatisfy: [:pref |
		aLiteral = pref changeSelector]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
keepTickingWhilePainting
	^ self
		valueOfFlag: #keepTickingWhilePainting
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'prefs - text' stamp: 'dew 1/8/2002 01:07'!
keyboardFocusColor
	"Answer the keyboard focus color, initializing it if necessary"

	^ Parameters at: #keyboardFocusColor ifAbsentPut: [Color lightGray]

"
Parameters removeKey: #keyboardFocusColor.
Preferences keyboardFocusColor
"! !

!Preferences class methodsFor: 'prefs - text' stamp: 'sw 12/7/2001 00:44'!
keyboardFocusColor: aColor
	"Set the keyboard focus color"

	Parameters at: #keyboardFocusColor put: aColor! !

!Preferences class methodsFor: 'updating - system' stamp: 'em 3/24/2005 15:34'!
largeTilesSettingToggled
	"The current value of the largeTiles flag has changed; now react"

	Smalltalk isMorphic ifTrue:
		[Preferences universalTiles
			ifFalse:
				[self inform: 
'note that this will only have a noticeable
effect if the universalTiles preference is
set to true, which it currently is not' translated]
			ifTrue:
				[World recreateScripts]]! !

!Preferences class methodsFor: 'support - misc' stamp: 'ar 9/27/2005 20:32'!
letUserPersonalizeMenu
	"Invoked from menu, opens up a single-msg browser on the message that user is invited to customize for rapid morphic access via option-click on morphic desktop"

	ToolSet browse: Preferences class 
		selector: #personalizeUserMenu:! !

!Preferences class methodsFor: 'initialization - save/load' stamp: 'cmm 8/9/2016 20:30'!
loadPreferencesFrom: aFile
	| stream params dict desktopColor |
	stream := ReferenceStream fileNamed: aFile.
	params := stream next.
	self assert: (params isKindOf: IdentityDictionary).
	params removeKey: #PersonalDictionaryOfPreferences.
	dict := stream next.
	self assert: (dict isKindOf: IdentityDictionary).
	desktopColor := stream next.
	stream close.
	dict keysAndValuesDo:
		[:key :value | (self preferenceAt: key ifAbsent: [nil]) ifNotNil:
			[:pref | [pref preferenceValue: value preferenceValue] on: Error do: [ : err | "Ignore preferences which may not be supported anymore."]]].
	params keysAndValuesDo: [ :key :value | self setParameter: key to: value ].
	Smalltalk isMorphic
		ifTrue: [ World fillStyle: desktopColor ]
		ifFalse: [ self desktopColor: desktopColor. ScheduledControllers updateGray ]! !

!Preferences class methodsFor: 'updating' stamp: 'tak 8/3/2005 21:17'!
localeChanged
	LocaleID current isoLanguage = 'ja'
		ifTrue: [Preferences enable: #useFormsInPaintBox]
		ifFalse: [Preferences disable: #useFormsInPaintBox]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
logDebuggerStackToFile
	^ self
		valueOfFlag: #logDebuggerStackToFile
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
magicHalos
	^ self
		valueOfFlag: #magicHalos
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
maintainHalos
	^ self
		valueOfFlag: #maintainHalos
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 11/5/1998 16:49'!
maxBalloonHelpLineLength
	^ Parameters at: #maxBalloonHelpLineLength! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
menuAppearance3d
	^ self
		valueOfFlag: #menuAppearance3d
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'prefs - menus' stamp: 'di 1/14/1999 20:16'!
menuBorderColor
	Display depth <= 2 ifTrue: [^ Color black].
	^ Parameters at: #menuBorderColor! !

!Preferences class methodsFor: 'prefs - menus' stamp: 'sw 11/3/1998 11:16'!
menuBorderWidth
	^ Parameters at: #menuBorderWidth! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
menuButtonInToolPane
	^ self
		valueOfFlag: #menuButtonInToolPane
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'prefs - menus' stamp: 'di 1/14/1999 20:17'!
menuColor
	Display depth <= 2 ifTrue: [^ Color white].
	^ Parameters at: #menuColor! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
menuColorFromWorld
	^ self
		valueOfFlag: #menuColorFromWorld
		ifAbsent: [true
			"success"]! !

!Preferences class methodsFor: 'misc' stamp: 'KR 4/17/2008 21:04'!
menuColorString
	^ ((self valueOfFlag: #menuColorFromWorld)
		ifTrue: ['stop menu-color-from-world' translated]
		ifFalse: ['start menu-color-from-world' translated]) ! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
menuKeyboardControl
	^ self
		valueOfFlag: #menuKeyboardControl
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'prefs - menus' stamp: 'dgd 3/23/2003 11:06'!
menuLineColor
	^ Parameters
		at: #menuLineColor
		ifAbsentPut: [Preferences menuBorderColor lighter]! !

!Preferences class methodsFor: 'prefs - menus' stamp: 'dgd 8/30/2004 20:59'!
menuSelectionColor
	^ Parameters
		at: #menuSelectionColor
		ifAbsent: [nil]! !

!Preferences class methodsFor: 'prefs - menus' stamp: 'di 1/14/1999 20:19'!
menuTitleBorderColor
	Display depth <= 2 ifTrue: [^ Color black].
	^ Parameters at: #menuTitleBorderColor! !

!Preferences class methodsFor: 'prefs - menus' stamp: 'sw 11/3/1998 11:16'!
menuTitleBorderWidth
	^ Parameters at: #menuTitleBorderWidth! !

!Preferences class methodsFor: 'prefs - menus' stamp: 'di 1/14/1999 20:18'!
menuTitleColor
	Display depth = 1 ifTrue: [^ Color white].
	Display depth = 2 ifTrue: [^ Color gray].
	^ Parameters at: #menuTitleColor! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
menuWithIcons
	^ self
		valueOfFlag: #menuWithIcons
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 8/11/2002 02:18'!
messengersInViewers
	"A coming technology..."

	^ false! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 8/18/2000 13:26'!
metaMenuDisabled
	"If true, then click/cmd-click on the desktop will not bring up the World menu.  Can be changed manually right here, and can be programattically changed via a call of the following form:

	Preferences compileHardCodedPref: #metaMenuDisabled enable: true"

	^ false! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
modalColorPickers
	^ self
		valueOfFlag: #modalColorPickers
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
morphicProgressStyle
	^ self
		valueOfFlag: #morphicProgressStyle
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
mouseOverForKeyboardFocus
	^ self
		valueOfFlag: #mouseOverForKeyboardFocus
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
mouseOverHalos
	^ self
		valueOfFlag: #mouseOverHalos
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'updating - system' stamp: 'mir 9/12/2001 15:15'!
mouseOverHalosChanged
	World wantsMouseOverHalos: self mouseOverHalos! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
mvcProjectsAllowed
	^ self
		valueOfFlag: #mvcProjectsAllowed
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
navigatorOnLeftEdge
	^ self
		valueOfFlag: #navigatorOnLeftEdge
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'support - misc' stamp: 'dgd 8/31/2003 18:03'!
navigatorShowingString
	"Answer a string for the show-project-navigator menu item"
	^ (self showProjectNavigator
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'show navigator (N)' translated! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
noviceMode
	^ self
		valueOfFlag: #noviceMode
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'updating - system' stamp: 'bp 1/8/2011 12:59'!
noviceModeSettingChanged
	"The current value of the noviceMode flag has changed;  
	now react"
	TheWorldMainDockingBar updateInstances.
	PasteUpMorph allSubInstances
		select: [:each | each isWorldMorph]
		thenDo: [:each | each initializeDesktopCommandKeySelectors].
	Smalltalk at: #ParagraphEditor ifPresent: [:aClass| aClass initialize]! !

!Preferences class methodsFor: 'themes - tools' stamp: 'mt 8/5/2016 10:14'!
offerThemesMenu
	"Put up a menu offering the user a choice of themes.  Each theme is represented by a method in category #themes in Preferences class.  The comment at the front of each method is used as the balloon help for the theme"

	"Preferences offerThemesMenu"
	| selectors aMenu |
	selectors := self class allMethodsInCategory: #themes.
	selectors := selectors select: [:sel | sel numArgs = 0].
	aMenu := MenuMorph new defaultTarget: self.
	aMenu addTitle: 'Choose a set of preferences to install' translated.
	selectors do:
		[:sel |
			aMenu add: sel target: self selector: #installTheme: argument: sel.
			aMenu balloonTextForLastItem: (self class firstCommentAt: sel)].
	aMenu addLine.
	aMenu add: 'browse details...' translated target: self action: #browseThemes.
	aMenu balloonTextForLastItem: 'Puts up a tool that will allow you to view and edit the code underlying all of the available preference sets' translated.
	aMenu popUpInWorld.
	"(Workspace new contents: 'here is an example of a new window with your new theme installed' translated) openLabel: 'Testing one two three'"! !

!Preferences class methodsFor: 'support' stamp: 'sw 4/24/2001 12:02'!
okayToChangeProjectLocalnessOf: prefSymbol
	"Answer whether it would be okay to allow the user to switch the setting of whether or not the preference symbol is local to a project.  Formerly useful and perhaps again will be, though to be sure this is a non-modular design."

	^ (#() includes: prefSymbol) not! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
okToReinitializeFlaps
	^ self
		valueOfFlag: #okToReinitializeFlaps
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
oliveHandleForScriptedObjects
	^ self
		valueOfFlag: #oliveHandleForScriptedObjects
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
optionalButtons
	^ self
		valueOfFlag: #optionalButtons
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 6/12/2001 20:18'!
optionalButtonsChanged
	"The setting of the optionalButtons preference changed; react.  Formerly, we replaced prototypes in flaps but this is no longer necessary"
! !

!Preferences class methodsFor: 'themes' stamp: 'kfr 5/5/2015 18:21'!
outOfTheBox
	"The default out-of-the-box preference settings for Squeak 3.2.  The 'alternative' window-look and scrollbar-look are used.  Button panes are used but not annotation panes.  Scrollbars are on the right and do not flop out."

	self setPreferencesFrom: self defaultValueTableForCurrentRelease.
	SystemNavigation thoroughSenders: true.
	SystemWindow clickOnLabelToEdit: true.
	Text ignoreStyleIfOnlyBold: true.
	MenuMorph roundedMenuCorners: true.
	ScrollBar scrollBarsWithoutMenuButton: true.
	SoundPlayer soundQuickStart: false.
	SoundPlayer stopSoundWhenDone: false.
	SoundService soundEnabled: true.
	SoundRecorder canRecordWhilePlaying:false.! !

!Preferences class methodsFor: 'parameters' stamp: 'sw 1/27/2000 23:02'!
parameterAt: aKey
	^ Parameters at: aKey ifAbsent: [nil]! !

!Preferences class methodsFor: 'parameters' stamp: 'sw 9/28/2001 08:52'!
parameterAt: aKey default: defaultValueBlock
	"Deprecated interface; no surviving senders in the released image, but clients probably still use"

	^ self parameterAt: aKey ifAbsentPut: defaultValueBlock! !

!Preferences class methodsFor: 'parameters' stamp: 'sw 9/28/2001 08:40'!
parameterAt: aKey ifAbsentPut: defaultValueBlock
	"Return the Parameter setting at the given key.  If there is no entry for this key in the Parameters dictionary, create one with the value of defaultValueBlock as its value"

	^ Parameters at: aKey ifAbsentPut: defaultValueBlock! !

!Preferences class methodsFor: 'parameters' stamp: 'ul 8/2/2011 20:07'!
parameterAt: aKey ifAbsent: aBlock
	"Answer the parameter saved at the given key; if there is no such key in the Parameters dictionary, evaluate aBlock"

	^ Parameters at: aKey ifAbsent: aBlock! !

!Preferences class methodsFor: 'parameters' stamp: 'kfr 1/3/2015 11:58'!
parameters
	
"Preferences parameters explore"
	^Parameters! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
passwordsOnPublish
	^ self
		valueOfFlag: #passwordsOnPublish
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'themes' stamp: 'sw 5/2/2002 10:45'!
personal
	"Settings saved (by sometime earlier having hit the 'Save Current Settings as my Personal Preferences' in a Preferences panel) as my personal preferences"

	self restorePersonalPreferences! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
personalizedWorldMenu
	^ self
		valueOfFlag: #personalizedWorldMenu
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'support - misc' stamp: 'fbs 12/6/2013 21:01'!
personalizeUserMenu: aMenu
	"The user has clicked on the morphic desktop with the yellow mouse button (option+click on the Mac); a menu is being constructed to present to the user in response; its default target is the current world.  In this method, you are invited to add items to the menu as per personal preferences.
	The default implementation, for illustrative purposes, sets the menu title to 'personal', and adds items for go-to-previous-project, show/hide flaps, and load code updates"
	
	aMenu addTitle: 'personal' translated.  "Remove or modify this as per personal choice"

	aMenu addStayUpItem.
	aMenu add: 'previous project' translated action: #goBack.
	aMenu add: 'load latest code updates' translated target: MCMcmUpdater action: #updateFromServer.
	aMenu add: 'about this system...' translated target: Smalltalk action: #aboutThisSystem.
	
	aMenu addLine.
				
	aMenu addUpdating: #suppressFlapsString target: Project current action: #toggleFlapsSuppressed.
	aMenu balloonTextForLastItem: 'Whether prevailing flaps should be shown in the project right now or not.' translated! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
postscriptStoredAsEPS
	^ self
		valueOfFlag: #postscriptStoredAsEPS
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'updating' stamp: 'mt 8/27/2015 10:42'!
prefEvent: anEvent
	"Check if this system event defines or removes a preference."

	| class selector method |
	self flag: #performance. "mt: Maybe defer preference dictionary update?"

	anEvent itemKind = SystemChangeNotifier classKind ifTrue: [
		anEvent isRemoved ifTrue: [
			self removeAllPreferencesSuchThat: [:pref | 
				pref provider == anEvent item]].
		
		anEvent isRenamed ifTrue: [
			self atomicUpdatePreferences: [ :prefs |
				self allPreferences
					select: [:pref | pref provider == anEvent item]
					thenDo: [:pref | prefs at: pref id put: pref].
				prefs keys
					select: [:id | id beginsWith: anEvent oldName]
					thenDo: [:id | prefs removeKey: id] ] ] ].
	
	anEvent itemKind = SystemChangeNotifier methodKind ifTrue: [
		"ignore instance methods"
		anEvent itemClass isMeta
			ifFalse: [^ self]. 

		class := anEvent itemClass theNonMetaClass.
		selector := anEvent itemSelector.
		method := anEvent item.

		anEvent isRemoved ifTrue: [
			self atomicUpdatePreferences: [ :prefs | 
				"See PragmaPreference >> #id."
				prefs removeKey: (class name,'>>', selector) asSymbol ifAbsent: []]].
	
		(anEvent isAdded or: [anEvent isModified]) ifTrue: [
			method pragmas do: [:pragma | self addPragmaPreference: pragma] ] ].! !

!Preferences class methodsFor: 'accessing' stamp: 'ul 1/10/2011 16:22'!
preferenceAt: aSymbol 
	"Answer the Preference object at the given symbol, or nil if not there"

	^self preferenceAt: aSymbol ifAbsent: [ nil ]! !

!Preferences class methodsFor: 'accessing' stamp: 'eem 6/30/2015 15:22'!
preferenceAt: aSymbol ifAbsent: aBlock 
	"Answer the Preference object at the given symbol, or the value of aBlock if not present"

	 ^preferencesDictionary at: aSymbol ifAbsent: aBlock! !

!Preferences class methodsFor: 'support' stamp: 'mt 8/27/2015 10:44'!
preferencesInCategory: aCategorySymbol 
	"Answer a list of Preference objects that reside in the given category."

	^ self allPreferences select: [ :aPreference |
		aPreference categoryList includes: aCategorySymbol ]! !

!Preferences class methodsFor: 'private' stamp: 'mt 8/27/2015 08:54'!
preference: prefName categoryList: categoryList description: helpString type: typeSymbol
	"Create a preference for a preference pragma in a method."
	
	^ PragmaPreference new
		 name: prefName
		 defaultValue: nil	"always nil"
		 helpString: helpString
		 localToProject: false "governed by the method"
		 categoryList: categoryList
		 changeInformee: nil
		 changeSelector: nil
		 type: typeSymbol! !

!Preferences class methodsFor: 'private' stamp: 'mt 8/27/2015 08:55'!
preference: prefName category: categoryName description: helpString type: typeSymbol
	"Create a preference for a preference pragma in a method."
	
	^ self 
		preference: prefName
		categoryList: (categoryName isArray "Alas pragma users are not always careful"
						ifTrue: [categoryName]
						ifFalse: [{categoryName}])
		description: helpString
		type: typeSymbol! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 8/29/2000 15:01'!
preserveCommandExcursions
	"An architecture is in place for storing command excursions to which access is otherwise cut off by having taken a variant branch, but it is not accessible unless you hand-code this preference to true -- which I suggest you do only with fingers crossed."

	^ false! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
preserveTrash
	^ self
		valueOfFlag: #preserveTrash
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
printAlternateSyntax
	^ self
		valueOfFlag: #printAlternateSyntax
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'nk 9/1/2004 11:37'!
printStandardSystemFonts
	"self printStandardSystemFonts"

	| string |
	string := String streamContents: [ :s |

	#(standardDefaultTextFont standardListFont standardFlapFont 
	standardEToysFont standardMenuFont windowTitleFont 
	standardBalloonHelpFont standardCodeFont standardButtonFont) do: [:selector |
		| font |
		font := Preferences perform: selector.
		s
			nextPutAll: selector; space;
			nextPutAll: font familyName; space;
			nextPutAll: (AbstractFont emphasisStringFor: font emphasis);
			nextPutAll: ' points: ';
			print: font pointSize;
			nextPutAll: ' height: ';
			print: font height;
			cr
		]].

	(StringHolder new)
		contents: string;
		openLabel: 'Current system font settings' translated.
! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
projectsSentToDisk
	^ self
		valueOfFlag: #projectsSentToDisk
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
projectViewsInWindows
	^ self
		valueOfFlag: #projectViewsInWindows
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
projectZoom
	^ self
		valueOfFlag: #projectZoom
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
promptForUpdateServer
	^ self
		valueOfFlag: #promptForUpdateServer
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
propertySheetFromHalo
	^ self
		valueOfFlag: #propertySheetFromHalo
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
purgeUndoOnQuit
	^ self
		valueOfFlag: #purgeUndoOnQuit
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
readDocumentAtStartup
	^ self
		valueOfFlag: #readDocumentAtStartup
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: 'cmm 10/12/2013 17:34'!
readDocumentAtStartup: aBoolean
	^ self
		setPreference: #readDocumentAtStartup
		toValue: aBoolean! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
readOnlyMode
	^ self
		valueOfFlag: #readOnlyMode
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 8/9/2016 17:08'!
refreshFontSettings
	"Try to update all the current font settings to make things consistent."

	UserInterfaceTheme current applyAfter: [
		self setFlapsFontTo: (self standardFlapFont);
			setEToysFontTo: (self standardEToysFont);
			setWindowTitleFontTo: (self windowTitleFont);
			setListFontTo: (self standardListFont);
			setMenuFontTo: (self standardMenuFont);
			setSystemFontTo: (TextStyle defaultFont);
			setCodeFontTo: (self standardCodeFont);
			setBalloonHelpFontTo: (BalloonMorph balloonFont)].

	SystemWindow allSubInstancesDo: [ :s | | rawLabel |
		rawLabel := s getRawLabel.
		rawLabel owner vResizing: #spaceFill.
		rawLabel font: rawLabel font.
		s setLabel: s label.
		s replaceBoxes ].! !

!Preferences class methodsFor: 'initialization' stamp: 'mt 8/27/2015 08:38'!
registerForEvents
	"Preferences registerForEvents"
	
	"Do not register pragma preferences with any preferences holder but this one."
	self == Preferences ifFalse: [^ self].
	
	SystemChangeNotifier uniqueInstance
		noMoreNotificationsFor: self;
		notify: self ofAllSystemChangesUsing: #prefEvent:.

	Smalltalk allClassesDo: [:aClass |
		aClass class methodsDo: [:method |
			method pragmas do: [:pragma |
				self addPragmaPreference: pragma] ] ].! !

!Preferences class methodsFor: 'add/remove' stamp: 'mt 8/26/2015 16:58'!
removeAllPreferencesSuchThat: block

	self atomicUpdatePreferences: [ :copyOfDictionaryOfPreferences |
		| map |
		map := copyOfDictionaryOfPreferences select: block.
		map keysDo: [ :prefName |
			copyOfDictionaryOfPreferences removeKey: prefName]]! !

!Preferences class methodsFor: 'initialization' stamp: 'ar 9/27/2005 21:49'!
removeObsolete
	"Remove obsolete preferences"
	Preference allInstancesDo:[:pref|
		pref isObsolete ifTrue:[self removePreference: pref].
	].! !

!Preferences class methodsFor: 'add/remove' stamp: 'mt 8/26/2015 17:26'!
removePreference: aSymbol 
	"Remove all memory of the given preference symbol in my various structures."

	| pref |
	pref := self preferenceAt: aSymbol ifAbsent: [^self].
	
	pref localToProject ifTrue: [
		Project allProjects do: [ :proj | 
			 proj projectPreferenceFlagDictionary ifNotNil: [ :projectpreferences |
				projectpreferences removeKey: aSymbol ifAbsent: [] ] ] ].
	
	self atomicUpdatePreferences: [ :copyOfDictionaryOfPreferences |
		copyOfDictionaryOfPreferences removeKey: aSymbol ifAbsent: nil ].
	
	"Remove auto-generated accessor method."
	self class removeSelector: aSymbol.
	
	^ pref! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'sw 1/25/2000 20:10'!
resetHaloSpecifications
	"Preferences resetHaloSpecifications"

	^ Parameters removeKey: #HaloSpecs ifAbsent: []! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
restartAlsoProceeds
	^ self
		valueOfFlag: #restartAlsoProceeds
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'ul 11/25/2010 23:37'!
restoreDefaultFonts
	"Since this is called from menus, we can take the opportunity to prompt for missing font styles."
	"
	Preferences restoreDefaultFonts
	"

	self setDefaultFonts: #(
		(setSystemFontTo:			'Bitmap DejaVu Sans'		9)
		(setListFontTo:				'Bitmap DejaVu Sans'		9)
		(setFlapsFontTo:			Accushi						12)
		(setEToysFontTo:			BitstreamVeraSansBold		9)
		(setPaintBoxButtonFontTo:	BitstreamVeraSansBold		9)
		(setMenuFontTo:			'Bitmap DejaVu Sans'		9)
		(setWindowTitleFontTo:		'Bitmap DejaVu Sans Bold'	9)
		(setBalloonHelpFontTo:		'Bitmap DejaVu Sans'		7)
		(setCodeFontTo:			'Bitmap DejaVu Sans'		9)
		(setButtonFontTo:			'Bitmap DejaVu Sans'	7)
	)! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'yo 7/28/2004 21:25'!
restoreDefaultFontsForJapanese
	"Preferences restoreDefaultFontsForJapanese"
	#(	"(setButtonFontTo:		ComicBold		15)"
		"(setTextButtonFontTo:		NewYork		12)"
		"(setCodeFontTo:			NewYork		12)"  "Later"
		(setFlapsFontTo:			NewYork		15)
		(setListFontTo:			NewYork		12)
		(setMenuFontTo:			NewYork		12)
		(setWindowTitleFontTo:	NewYork		15)
		(setSystemFontTo:		NewYork		12)) do:
			[:triplet |
				self perform: triplet first with: (StrikeFontSet familyName: triplet second size: triplet third)].

	self setButtonFontTo: (StrikeFont familyName: #ComicBold size: 16).

	Smalltalk at: #BalloonMorph ifPresent:
		[:thatClass | thatClass setBalloonFontTo: (StrikeFontSet familyName: #NewYork size: 12)].

	"Note:  The standardCodeFont is not currently used -- the default font is instead; later hopefully we can split the code font out as  a separate choice, but only after we're able to have the protocols reorganized such that we can know whether it's code or not when we launch the text object.

	Note:  The standard button font is reset by this code but is not otherwise settable by a public UI (too many things can go afoul) "! !

!Preferences class methodsFor: 'initialization - misc' stamp: 'dgd 3/23/2003 11:11'!
restoreDefaultMenuParameters
	"Restore the four color choices of the original implementors of  
	MorphicMenus"
	" 
	Preferences restoreDefaultMenuParameters
	"
	Parameters
		at: #menuColor
		put: (Color
				r: 0.97
				g: 0.97
				b: 0.97).
	Parameters
		at: #menuBorderColor
		put: (Color
				r: 0.167
				g: 0.167
				b: 1.0).
	Parameters at: #menuBorderWidth put: 2.
	Parameters at: #menuTitleColor put: (Color
			r: 0.4
			g: 0.8
			b: 0.9) twiceDarker.
	Parameters
		at: #menuTitleBorderColor
		put: (Color
				r: 0.333
				g: 0.667
				b: 0.751).
	Parameters at: #menuTitleBorderWidth put: 1.
	Parameters
		at: #menuLineColor
		put: (Preferences menuBorderColor lighter)! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 8/9/2016 17:07'!
restoreFontsAfter: aBlock

	"Restore the currently chosen set of standard fonts after 
	evaluating aBlock. Used for tests that modify the default fonts."

	| standardDefaultTextFont standardListFont standardEToysFont standardMenuFont 
	windowTitleFont standardBalloonHelpFont standardCodeFont standardButtonFont |

	standardDefaultTextFont := Preferences standardDefaultTextFont.
	standardListFont := Preferences standardListFont.
	standardEToysFont := Preferences standardEToysFont.
	standardMenuFont := Preferences standardMenuFont.
	windowTitleFont := Preferences windowTitleFont.
	standardBalloonHelpFont := Preferences standardBalloonHelpFont.
	standardCodeFont := Preferences standardCodeFont.
	standardButtonFont := Preferences standardButtonFont.
	^ UserInterfaceTheme current applyAfter: [
		aBlock ensure: [
			Preferences setSystemFontTo: standardDefaultTextFont.
			Preferences setListFontTo: standardListFont.
			Preferences setEToysFontTo: standardEToysFont.
			Preferences setMenuFontTo: standardMenuFont.
			Preferences setWindowTitleFontTo: windowTitleFont.
			Preferences setBalloonHelpFontTo: standardBalloonHelpFont.
			Preferences setCodeFontTo: standardCodeFont.
			Preferences setButtonFontTo: standardButtonFont]].
! !

!Preferences class methodsFor: 'personalization' stamp: 'KR 5/1/2008 01:54'!
restorePersonalPreferences
	"Restore all the user's saved personal preference settings"

	| savedPrefs |
	savedPrefs _ self parameterAt: #PersonalDictionaryOfPreferences ifAbsent: [^ self inform: 'There are no personal preferences saved in this image yet' translated].

	savedPrefs associationsDo:
		[:assoc | (self preferenceAt: assoc key ifAbsent: [nil]) ifNotNilDo:
			[:pref | pref preferenceValue: assoc value preferenceValue]]! !

!Preferences class methodsFor: 'personalization' stamp: 'kfr 10/24/2012 22:56'!
restorePreferencesFromDisk
	| result |
	result := (FileList2 modalFileSelectorForSuffixes: #('prefs')) .
	result ifNil: [^ self].
	self restorePreferencesFromDisk: result fullName
		
	! !

!Preferences class methodsFor: 'initialization - save/load' stamp: 'kfr 1/9/2015 13:48'!
restorePreferencesFromDisk: aFile 
	Cursor wait
		showWhile: [[self loadPreferencesFrom: aFile]
				on: Error
				do: [:ex | self inform: 'there was an error restoring the preferences' translated]]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
reverseWindowStagger
	^ self
		valueOfFlag: #reverseWindowStagger
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
rotationAndScaleHandlesInPaintBox
	^ self
		valueOfFlag: #rotationAndScaleHandlesInPaintBox
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
roundedMenuCorners
	^ self
		valueOfFlag: #roundedMenuCorners
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
roundedScrollBarLook
	^ self
		valueOfFlag: #roundedScrollBarLook
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
roundedWindowCorners
	^ self
		valueOfFlag: #roundedWindowCorners
		ifAbsent: [false]! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 4/12/2001 01:11'!
roundedWindowCornersChanged
	"The user changed the value of the roundedWindowCorners preference.  React"

	ActiveWorld fullRepaintNeeded! !

!Preferences class methodsFor: 'initialization - save/load' stamp: 'eem 6/30/2015 15:23'!
savePersonalPreferences
	"Save the current list of Preference settings as the user's personal choices"

	self
		setParameter: #PersonalDictionaryOfPreferences
		to: preferencesDictionary deepCopy! !

!Preferences class methodsFor: 'prefs - misc' stamp: ''!
scrollBarColor
	"Answer the preferred color for scroll bar elevators."

	^ Color gray! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
scrollBarsNarrow
	^ self
		valueOfFlag: #scrollBarsNarrow
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
scrollBarsOnRight
	^ self
		valueOfFlag: #scrollBarsOnRight
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
scrollBarsWithoutMenuButton
	^ self
		valueOfFlag: #scrollBarsWithoutMenuButton
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - misc' stamp: ''!
scrollBarWidth
	"Answer the preferred width for scroll bars."

	^ 8! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
securityChecksEnabled
	^ self
		valueOfFlag: #securityChecksEnabled
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
selectionsMayShrink
	^ self
		valueOfFlag: #selectionsMayShrink
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
selectiveHalos
	^ self
		valueOfFlag: #selectiveHalos
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: 'al 7/31/2007 16:17'!
serverMode
	^ self
		valueOfFlag: #serverMode
		ifAbsent: [false]! !

!Preferences class methodsFor: 'support - file list services' stamp: 'kfr 4/28/2015 06:54'!
serviceLoadPreferencesFromDisk
	^ SimpleServiceEntry 
		provider: self 
		label: 'load preferences from a saved file'
		selector: #restorePreferencesFromDisk:
		description: 'restore all saved personal preference settings'
		buttonLabel: 'load preferences'! !

!Preferences class methodsFor: 'support - file list services' stamp: 'kfr 4/28/2015 06:49'!
services
	^ Array with: self serviceLoadPreferencesFromDisk! !

!Preferences class methodsFor: 'support - misc' stamp: 'sw 3/2/2004 22:11'!
setArrowheads
	"Let the user edit the size of arrowheads"

	| aParameter result  |
	aParameter := self parameterAt: #arrowSpec ifAbsent: [5 @ 4].
	result := Morph obtainArrowheadFor: 'Default size of arrowheads on pen trails ' translated defaultValue: aParameter asString.
	result ifNotNil:
			[self setParameter: #arrowSpec to: result]
		ifNil:
			[Beeper beep]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:06'!
setBalloonHelpFontTo: aFont

	UserInterfaceTheme current
		set: #balloonHelpFont
		to: aFont;
		apply.! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:06'!
setButtonFontTo: aFont

	UserInterfaceTheme current
		set: #standardButtonFont
		to: aFont;
		apply.! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:06'!
setCodeFontTo: aFont
	"Establish the code font."

	UserInterfaceTheme current
		set: #standardCodeFont
		to: aFont;
		apply.! !

!Preferences class methodsFor: 'initialization - misc' stamp: 'mt 8/27/2015 10:03'!
setDefaultAnnotationInfo
	"Preferences setDefaultAnnotationInfo"
	^ Parameters at: #MethodAnnotations put: self defaultAnnotationInfo! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 8/9/2016 17:03'!
setDefaultFonts: defaultFontsSpec
	"Since this is called from menus, we can take the opportunity to prompt for missing font styles."

	| fontNames map emphases |
	fontNames := defaultFontsSpec collect: [:array | array second].
	map := IdentityDictionary new.
	emphases := IdentityDictionary new.
	fontNames do: [:originalName | | decoded style |
		decoded := TextStyle decodeStyleName: originalName.
		style := map at: originalName put: (TextStyle named: decoded second).
		emphases at: originalName put: decoded first.
		style ifNil: [map at: originalName put: TextStyle default]].

	UserInterfaceTheme current applyAfter: [
		defaultFontsSpec do: [:triplet | self
			perform: triplet first
			with: (((map at: triplet second) fontOfPointSize: triplet third) emphasized: (emphases at: triplet second))]].! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'md 3/7/2006 10:27'!
setDemoFonts
	"Preferences setDemoFonts"

	self setDefaultFonts: #(
		(setSystemFontTo:			BitstreamVeraSans			 		12)
		(setListFontTo:				BitstreamVeraSans					14)
		(setFlapsFontTo:				Accushi								12)
		(setEToysFontTo:				BitstreamVeraSansBold				9)
		(setPaintBoxButtonFontTo:	BitstreamVeraSansBold				9)
		(setMenuFontTo:				BitstreamVeraSans					14)
		(setWindowTitleFontTo:		BitstreamVeraSansBold				12)
		(setBalloonHelpFontTo:		Accujen								18)
		(setCodeFontTo:				BitstreamVeraSans					18)
		(setButtonFontTo:			BitstreamVeraSansMono				14)
	)
! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:06'!
setEToysFontTo: aFont 
	"change the font used in eToys environment"

	UserInterfaceTheme current
		set: #eToysFont
		to: aFont;
		apply.! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:07'!
setEToysTitleFontTo: aFont 
	"change the font used in eToys environment"

	UserInterfaceTheme current
		set: #eToysTitleFont
		to: aFont;
		apply.! !

!Preferences class methodsFor: 'get/set - flags' stamp: 'mt 8/26/2015 17:34'!
setFlag: prefSymbol toValue: aBoolean
	"Convenience method for consistency."
	
	^ self setPreference: prefSymbol toValue: aBoolean! !

!Preferences class methodsFor: 'get/set - flags' stamp: 'mt 8/26/2015 17:32'!
setFlag: prefSymbol toValue: aBoolean during: aBlock
	"Set the flag to the given value for the duration of aBlock"

	(self valueOfFlag: prefSymbol) in: [:previous |
		self setFlag: prefSymbol toValue: aBoolean.
		aBlock ensure: [self setFlag: prefSymbol toValue: previous]].! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:07'!
setFlapsFontTo: aFont

	UserInterfaceTheme current
		set: #standardFlapFont
		to: aFont;
		apply.	
		
	"FlapTab allSubInstancesDo:
		[:aFlapTab | aFlapTab reformatTextualTab]"! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:07'!
setHaloLabelFontTo: aFont 
	"change the font used in eToys environment"

	UserInterfaceTheme current
		set: #haloLabelFont
		to: aFont;
		apply.! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:07'!
setListFontTo: aFont
	"Set the list font as indicated"

	UserInterfaceTheme current
		set: #standardListFont
		to: aFont;
		apply.

	"Smalltalk at: #ListParagraph ifPresent: [:lp | lp initialize].
	Smalltalk at: #Flaps ifPresent: [:flaps | flaps replaceToolsFlap]"! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:07'!
setMenuFontTo: aFont

	UserInterfaceTheme current
		set: #standardMenuFont
		to: aFont;
		apply.
		
	"Smalltalk at: #PopUpMenu ifPresent:[:aClass| aClass setMenuFontTo: aFont].
	TheWorldMainDockingBar updateInstances."! !

!Preferences class methodsFor: 'initialization - misc' stamp: 'mt 7/28/2016 10:40'!
setNotificationParametersForStandardPreferences
	"Set up the notification parameters for the standard preferences that require need them.  When adding new Preferences that require use of the notification mechanism, users declare the notifcation info as part of the call that adds the preference, or afterwards -- the two relevant methods for doing that are:
 	Preferences.addPreference:categories:default:balloonHelp:projectLocal:changeInformee:changeSelector:   and
	Preference changeInformee:changeSelector:"

		"Preferences setNotificationParametersForStandardPreferences"

	
	#(	
		(annotationPanes		annotationPanesChanged)
		(eToyFriendly			eToyFriendlyChanged)
		(infiniteUndo			infiniteUndoChanged)
		(uniTilesClassic			classicTilesSettingToggled)
		(optionalButtons			optionalButtonsChanged)
		(showProjectNavigator	showProjectNavigatorChanged)
		(smartUpdating			smartUpdatingChanged)
		(universalTiles			universalTilesSettingToggled)
		(showSharedFlaps		sharedFlapsSettingChanged)
		(noviceMode		noviceModeSettingChanged)
	)  do:

			[:pair | | aPreference |
				aPreference := self preferenceAt: pair first.
				aPreference changeInformee: self changeSelector: pair second]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:07'!
setPaintBoxButtonFontTo: aFont 
	"change the font used in the buttons in PaintBox."

	UserInterfaceTheme current
		set: #paintBoxButtonFont
		to: aFont;
		apply.! !

!Preferences class methodsFor: 'parameters' stamp: 'stp 01/13/2000 13:29'!
setParameter: paramName to: paramValue
	"Set the given field in the parameters dictionary."

	Parameters at: paramName put: paramValue! !

!Preferences class methodsFor: 'initialization' stamp: 'sw 4/21/2002 05:13'!
setPreferencesFrom: listOfPairs
	"Given a list of <preferenceName, value> pairs, set preference values.  This method is tolerent of the value being supplied either a Boolean or else one of the symbols #true and #false.  Also, a new-value of #noOpinion will result in that 'preference's value not being changed."

	listOfPairs do:
		[:aPair |
			(aPair second == #noOpinion) ifFalse:
				[Preferences setPreference: aPair first toValue: ((aPair second == #true) or: [aPair second == true])]]

"
Preferences setPreferencesFrom: #(( mouseOverForKeyboardFocus false))
Preferences setPreferencesFrom: {{  #mouseOverForKeyboardFocus. true}}
"! !

!Preferences class methodsFor: 'get/set' stamp: 'mt 8/26/2015 17:23'!
setPreference: prefSymbol toValue: anObject
	"Set the given preference to the given value, and answer that value"

	^ (self
		preferenceAt: prefSymbol
		ifAbsent: [^ self addPreference: prefSymbol default: anObject])
			preferenceValue: anObject;
			yourself! !

!Preferences class methodsFor: 'get/set' stamp: 'mt 8/28/2015 10:00'!
setPreference: prefSymbol toValue: anObject during: aBlock

	(self valueOfPreference: prefSymbol) in: [:previous |
		self setPreference: prefSymbol toValue: anObject.
		aBlock ensure: [self setPreference: prefSymbol toValue: previous]].! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'sw 4/17/2001 11:34'!
setSystemFontTo: aFont
	"Establish the default text font and style"

	| aStyle newDefaultStyle |
	aFont ifNil: [^ self].
	aStyle := aFont textStyle ifNil: [^ self].
	newDefaultStyle := aStyle copy.
	newDefaultStyle defaultFontIndex: (aStyle fontIndexOf: aFont).
	TextConstants at: #DefaultTextStyle put: newDefaultStyle.
	Flaps replaceToolsFlap.
	ScriptingSystem resetStandardPartsBin! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:07'!
setWindowTitleFontTo: aFont
	"Set the window-title font to be as indicated"

	UserInterfaceTheme current
		set: #windowTitleFont
		to: aFont;
		apply.

	"(Smalltalk hasClassNamed: #StandardSystemView)
		ifTrue: [(Smalltalk at: #StandardSystemView) setLabelStyle].
	(Smalltalk hasClassNamed: #Flaps)
		ifTrue: [(Smalltalk at: #Flaps) replaceToolsFlap]"
! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 4/30/2001 20:39'!
sharedFlapsSettingChanged
	"The current value of the showSharedFlaps flag has changed; now react"

	self showSharedFlaps  "viz. the new setting"
		ifFalse:		
			[Flaps globalFlapTabsIfAny do:
				[:aFlapTab | Flaps removeFlapTab: aFlapTab keepInList: true]]

		ifTrue:
			[Smalltalk isMorphic ifTrue:
				[self currentWorld addGlobalFlaps]]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
showAdvancedNavigatorButtons
	^ self
		valueOfFlag: #showAdvancedNavigatorButtons
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: 'cmm 1/11/2011 21:10'!
showBoundsInHalo
	^ self
		valueOfFlag: #showBoundsInHalo
		ifAbsent: [true]! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'sw 11/6/2000 10:02'!
showChooseGraphicHaloHandle
	"Hard-coded; reimplement to change behavior.  If this preference is set to true, then a choose-graphic halo handle may appear on the halo of SketchMorphs"

	^ false! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
showDeprecationWarnings
	^ self
		valueOfFlag: #showDeprecationWarnings
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
showDirectionForSketches
	^ self
		valueOfFlag: #showDirectionForSketches
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
showDirectionHandles
	^ self
		valueOfFlag: #showDirectionHandles
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
showFlapsWhenPublishing
	^ self
		valueOfFlag: #showFlapsWhenPublishing
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
showLinesInHierarchyViews
	^ self
		valueOfFlag: #showLinesInHierarchyViews
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
showProjectNavigator
	^ self
		valueOfFlag: #showProjectNavigator
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 4/12/2001 01:33'!
showProjectNavigatorChanged
	"The showProjectNavigatorChanged preference changed; react"
	
	Project current assureNavigatorPresenceMatchesPreference! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
showSecurityStatus
	^ self
		valueOfFlag: #showSecurityStatus
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
showSharedFlaps
	^ self
		valueOfFlag: #showSharedFlaps
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: 'cmm 1/10/2011 14:07'!
showSplitterHandles
	^ self
		valueOfFlag: #showSplitterHandles
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
signProjectFiles
	^ self
		valueOfFlag: #signProjectFiles
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'sw 7/28/2004 16:26'!
simpleFullHaloSpecifications
	"This method gives the specs for the 'full' handles variant when simple halos are in effect"

	"Preferences resetHaloSpecifications"

	^ #(
	"  	selector				horiz		vert			color info						icon key
		---------				------		-----------		-------------------------------		---------------"
	(addDebugHandle:		right		topCenter		(blue	veryMuchLighter)		'Halo-Debug')
	(addPoohHandle:			right		center			(white)							'Halo-Pooh')
	(addDismissHandle:		left			top				(red		muchLighter)			'Halo-Dismiss')
	(addRotateHandle:		left			bottom			(blue)							'Halo-Rot')
	(addMenuHandle:		leftCenter	top				(red)							'Halo-Menu')
	(addTileHandle:			left			bottomCenter	(lightBrown)					'Halo-Tile')
	(addViewHandle:			left			center			(cyan)							'Halo-View')
	(addGrabHandle:			center		top				(black)							'Halo-Grab')
	(addDragHandle:			rightCenter	top				(brown)							'Halo-Drag')
	(addDupHandle:			right		top				(green)							'Halo-Dup')	
	(addMakeSiblingHandle:	right		top				(green muchDarker)				'Halo-Dup')	
	(addHelpHandle:			center		bottom			(lightBlue)						'Halo-Help')
	(addGrowHandle:		right		bottom			(yellow)						'Halo-Scale')
	(addScaleHandle:		right		bottom			(lightOrange)					'Halo-Scale')
	(addFewerHandlesHandle:	left		topCenter		(paleBuff)						'Halo-FewerHandles')
	(addScriptHandle:		right		bottomCenter	(green muchLighter)			'Halo-Script')
	(addPaintBgdHandle:		right		center			(lightGray)						'Halo-Paint')
	(addRepaintHandle:		right		center			(lightGray)						'Halo-Paint')
	(addFontSizeHandle:		leftCenter	bottom			(lightGreen)						'Halo-FontSize')
	(addFontStyleHandle:		center		bottom			(lightRed)						'Halo-FontStyle')
	(addFontEmphHandle:	rightCenter	bottom			(lightBrown darker)		  'Halo-FontEmph')
	(addRecolorHandle:		right		bottomCenter	(magenta darker)				'Halo-Recolor')

		) ! !

!Preferences class methodsFor: 'prefs - halos' stamp: 'hpt 9/24/2004 23:34'!
simpleHalosInForce
	^ (self preferenceAt: #haloTheme) preferenceValue == #simpleFullHaloSpecifications! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
simpleMenus
	^ self
		valueOfFlag: #simpleMenus
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
slideDismissalsToTrash
	^ self
		valueOfFlag: #slideDismissalsToTrash
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
smartUpdating
	^ self
		valueOfFlag: #smartUpdating
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 4/12/2001 01:30'!
smartUpdatingChanged
	"The smartUpdating preference changed. React"

	SystemWindow allSubInstancesDo:
		[:aWindow | aWindow amendSteppingStatus]

	"NOTE: This makes this preference always behave like a global preference, which is problematical"! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
soundQuickStart
	^ self
		valueOfFlag: #soundQuickStart
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
soundReverb
	^ self
		valueOfFlag: #soundReverb
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
soundStopWhenDone
	^ self
		valueOfFlag: #soundStopWhenDone
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
soundsEnabled
	^ self
		valueOfFlag: #soundsEnabled
		ifAbsent: [true]! !

!Preferences class methodsFor: 'support - misc' stamp: 'dgd 9/21/2003 13:46'!
staggerPolicyString
	"Answer the string to be shown in a menu to represent the 
	stagger-policy status"
	^ ((self valueOfFlag: #reverseWindowStagger)
		ifTrue: ['<yes>']
		ifFalse: ['<no>']), 'stagger windows' translated! !

!Preferences class methodsFor: 'standard queries' stamp: 'mir 11/10/2003 14:28'!
standaloneSecurityChecksEnabled
	^ self
		valueOfFlag: #standaloneSecurityChecksEnabled
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:09'!
standardBalloonHelpFont
	
	^ (UserInterfaceTheme current get: #balloonHelpFont)
		ifNil: [TextStyle defaultFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:10'!
standardButtonFont
	
	^ (UserInterfaceTheme current get: #standardButtonFont)
		ifNil: [TextStyle defaultFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:10'!
standardCodeFont

	^ (UserInterfaceTheme current get: #standardCodeFont)
		ifNil: [TextStyle defaultFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:10'!
standardDefaultTextFont

	^ (UserInterfaceTheme current get: #standardDefaultTextFont)
		ifNil: [TextStyle defaultFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:11'!
standardEToysFont

	^ (UserInterfaceTheme current get: #eToysFont)
		ifNil: [self standardButtonFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:11'!
standardEToysTitleFont

	^ (UserInterfaceTheme current get: #eToysTitleFont)
		ifNil: [self standardEToysFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:12'!
standardFlapFont
	
	^ (UserInterfaceTheme current get: #standardFlapFont)
		ifNil: [self standardButtonFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:12'!
standardHaloLabelFont

	^ (UserInterfaceTheme current get: #haloLabelFont)
		ifNil: [TextStyle defaultFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:13'!
standardListFont

	^ (UserInterfaceTheme current get: #standardListFont)
		ifNil: [TextStyle defaultFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:13'!
standardMenuFont

	^ (UserInterfaceTheme current get: #standardMenuFont)
		ifNil: [TextStyle defaultFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:13'!
standardPaintBoxButtonFont

	^ (UserInterfaceTheme current get: #paintBoxButtonFont)
		ifNil: [self standardButtonFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:14'!
standardSystemFont
	"Answer the standard system font "

	^ (UserInterfaceTheme current get: #standardSystemFont)
		ifNil: [TextStyle defaultFont]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
startInUntrustedDirectory
	^ self
		valueOfFlag: #startInUntrustedDirectory
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'initialization - save/load' stamp: 'mt 8/26/2015 14:29'!
storePreferencesIn: aFileName 
	| stream prefsSnapshot |
	#(Prevailing PersonalPreferences) do:
		[:ea |
		 Parameters removeKey: ea ifAbsent: []].
	stream := ReferenceStream fileNamed: aFileName.
	stream nextPut: Parameters.
	prefsSnapshot := preferencesDictionary copy.
	prefsSnapshot keysAndValuesDo: [:key :pref | prefsSnapshot at: key put: pref asPreference].
	stream nextPut: prefsSnapshot.
	stream nextPut: (Smalltalk isMorphic 
						 ifTrue:[World fillStyle]
						 ifFalse:[self desktopColor]).
	stream close! !

!Preferences class methodsFor: 'personalization' stamp: 'kfr 10/24/2012 22:58'!
storePreferencesToDisk
	| newName |
	newName := UIManager default request: 'Please confirm name for save...' initialAnswer: 'myPreferences'.
	newName isEmpty
		ifTrue: [^ self].
	Cursor wait
		showWhile: [[self storePreferencesIn: newName , '.prefs']
				on: Error
				do: [:ex | self inform: 'there was an error storing your preferences to disk. you probably already have stored your preferences' translated]]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'jmv 8/5/2009 23:11'!
subPixelRenderColorFonts
	^ self
		valueOfFlag: #subPixelRenderColorFonts
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'jmv 8/5/2009 23:11'!
subPixelRenderFonts
	^ self
		valueOfFlag: #subPixelRenderFonts
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
sugarAutoSave
	^ self
		valueOfFlag: #sugarAutoSave
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 11/15/2001 08:37'!
suppressWindowTitlesInInstanceBrowsers
	"Hard-coded for the moment: answer whether instance browsers should suppresss their window titles"

	^ false! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
swapControlAndAltKeys
	^ self
		valueOfFlag: #swapControlAndAltKeys
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
swapMouseButtons
	^ self
		valueOfFlag: #swapMouseButtons
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
systemWindowEmbedOK
	^ self
		valueOfFlag: #systemWindowEmbedOK
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
tabAmongFields
	^ self
		valueOfFlag: #tabAmongFields
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
testRunnerShowAbstractClasses
	^ self
		valueOfFlag: #testRunnerShowAbstractClasses
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - text' stamp: 'sw 9/7/1999 12:53'!
textHighlightColor
	^ Parameters at: #textHighlightColor! !

!Preferences class methodsFor: 'prefs - text' stamp: 'sw 9/7/1999 12:54'!
textHighlightColor: aColor
	Parameters at: #textHighlightColor put: aColor! !

!Preferences class methodsFor: 'themes - tools' stamp: 'dgd 9/1/2003 11:43'!
themeChoiceButtonOfColor: aColor font: aFont
	"Answer a button inviting the user to choose a theme"

	| aButton |
	aButton := SimpleButtonMorph new target: self; actionSelector: #offerThemesMenu.
	aButton label: 'change theme...' translated font: aFont.
	aButton color: aColor.
	aButton setBalloonText: 'Numerous "Preferences" govern many things about the way Squeak looks and behaves.  Set individual preferences using a "Preferences" panel.  Set an entire "theme" of many Preferences all at the same time by pressing this "change theme" button and choosing a theme to install.  Look in category "themes" in Preferences class to see what each theme does; add your own methods to the "themes" category and they will show up in the list of theme choices.' translated.
	^ aButton! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
tileTranslucentDrag
	^ self
		valueOfFlag: #tileTranslucentDrag
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
timeStampsInMenuTitles
	^ self
		valueOfFlag: #timeStampsInMenuTitles
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
tinyDisplay
	^ self
		valueOfFlag: #tinyDisplay
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'get/set - flags' stamp: 'mt 8/26/2015 17:40'!
toggle: flagName
	"Toggle the given preference. prefSymbol must be of a boolean preference"
	
	^ self setFlag: flagName toValue: (self valueOfFlag: flagName) not! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
traceMessages
	^ self
		valueOfFlag: #traceMessages
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
turnOffPowerManager
	^ self
		valueOfFlag: #turnOffPowerManager
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
twentyFourHourFileStamps
	^ self
		valueOfFlag: #twentyFourHourFileStamps
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
twoSidedPoohTextures
	^ self
		valueOfFlag: #twoSidedPoohTextures
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
typeCheckingInTileScripting
	^ self
		valueOfFlag: #typeCheckingInTileScripting
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'support' stamp: 'mt 8/26/2015 15:20'!
typeForValue: anObject
	"Returns the preference type for the given value to be distinguished in tools."

	{
		AbstractFont -> #Font.
		Boolean -> #Boolean.
		Color -> #Color.
		String -> #String.
		Number -> #Number
	} do: [:spec | (anObject isKindOf: spec key) ifTrue: [^ spec value]].

	^ #Object
	


! !

!Preferences class methodsFor: 'defaults' stamp: 'mt 8/27/2015 09:02'!
unclassifiedCategory
	"Generic category to be used when no specific alternative is provided."
	
	^ #unclassified! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
unifyNestedProgressBars
	^ self
		valueOfFlag: #unifyNestedProgressBars
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
uniqueNamesInHalos
	^ self
		valueOfFlag: #uniqueNamesInHalos
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
uniTilesClassic
	^ self
		valueOfFlag: #uniTilesClassic
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
universalTiles
	^ self
		valueOfFlag: #universalTiles
		ifAbsent: [false]! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 4/13/2001 11:22'!
universalTilesSettingToggled
	"The current value of the universalTiles flag has changed; now react"

	(self preferenceAt: #universalTiles ifAbsent: [^ self]) localToProject ifFalse:
			[^ self inform: 
'This is troubling -- you may regret having done that, because
the change will apply to *all projects*, including pre-existing ones.  Unfortunately this check is done after the damage is done, so you
may be hosed.  Fortunately, however, you can simply reverse your choice right now and perhaps no deep damage will have been done.'].

	self universalTiles  "User just switched project to classic tiles"
		ifFalse:
			[self inform: 
'CAUTION -- if you had any scripted objects in
this project that already used universal tiles, 
there is no reasonable way to go back to classic
tiles.  Recommended course of action in that case:
just toggle this preference right back to true.']
		ifTrue:
			[Preferences capitalizedReferences ifFalse:
				[Preferences enable: #capitalizedReferences.
				self inform: 
'Note that the "capitalizedReferences" flag
has now been automatically set to true for
you, since this is required for the use of
universal tiles.'].
			World isMorph ifTrue:
				[World recreateScripts]]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
unlimitedPaintArea
	^ self
		valueOfFlag: #unlimitedPaintArea
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 12:38'!
unload
	"Unload the receiver from global registries"

	self environment at: #Flaps ifPresent: [:cl |
	cl unregisterQuadsWithReceiver: self] ! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
updateFromServerAtStartup
	^ self
		valueOfFlag: #updateFromServerAtStartup
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
updateSavesFile
	^ self
		valueOfFlag: #updateSavesFile
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
useArtificialSweetenerBar
	^ self
		valueOfFlag: #useArtificialSweetenerBar
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
useBiggerPaintingBox
	^ self
		valueOfFlag: #useBiggerPaintingBox
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
useButtonPropertiesToFire
	^ self
		valueOfFlag: #useButtonPropertiesToFire
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'sw 2/16/1999 11:24'!
useCategoryListsInViewers
	"Temporarily hard-coded pending viewer work underway"
	^ false! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
useFileList2
	^ self
		valueOfFlag: #useFileList2
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
useFormsInPaintBox
	^ self
		valueOfFlag: #useFormsInPaintBox
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - misc' stamp: 'yo 1/13/2005 11:05'!
useFormsInPaintBox: aBoolean

	self setPreference: #useFormsInPaintBox toValue: aBoolean
! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
usePangoRenderer
	^ self
		valueOfFlag: #usePangoRenderer
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
usePlatformFonts
	^ self
		valueOfFlag: #usePlatformFonts
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
usePopUpArrows
	^ self
		valueOfFlag: #usePopUpArrows
		ifAbsent: [true]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
useSmartLabels
	^ self
		valueOfFlag: #useSmartLabels
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
useUndo
	^ self
		valueOfFlag: #useUndo
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
useVectorVocabulary
	^ self
		valueOfFlag: #useVectorVocabulary
		ifAbsent: [ false ]! !

!Preferences class methodsFor: 'get/set - flags' stamp: 'mt 8/26/2015 17:14'!
valueOfFlag: aFlagName
	"Utility method for all the preferences that are boolean, and for backward compatibility"
	^self valueOfFlag: aFlagName ifAbsent: [false]! !

!Preferences class methodsFor: 'get/set - flags' stamp: 'hpt 9/26/2004 16:48'!
valueOfFlag: aFlagName ifAbsent: booleanValuedBlock
	"the same as in #valueOfFlag:"
	^self valueOfPreference: aFlagName ifAbsent: booleanValuedBlock.! !

!Preferences class methodsFor: 'get/set' stamp: 'hpt 9/26/2004 16:49'!
valueOfPreference: aPreferenceSymbol
	"Answer the value of the given preference"
	^self valueOfPreference: aPreferenceSymbol ifAbsent: []! !

!Preferences class methodsFor: 'get/set' stamp: 'mt 8/26/2015 17:04'!
valueOfPreference: aPreferenceSymbol ifAbsent: block
	"Answer the value of the given preference"
	^ (self preferenceAt: aPreferenceSymbol ifAbsent: [^ block value]) preferenceValue! !

!Preferences class methodsFor: 'updating - system' stamp: 'sw 7/10/2005 03:06'!
vectorVocabularySettingChanged
	"The current value of the useVectorVocabulary flag has changed; now react.  No senders, but invoked by the Preference object associated with the #useVectorVocabulary preference."

	Smalltalk isMorphic ifTrue:
		[ActiveWorld makeVectorUseConformToPreference]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
viewersInFlaps
	^ self
		valueOfFlag: #viewersInFlaps
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
visualExplorer
	^ self
		valueOfFlag: #visualExplorer
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'support - misc' stamp: 'sw 8/29/2000 16:12'!
wantsChangeSetLogging
	"Answer whether method changes in the receiver should be logged to current change set.  This circumlocution avoids such logging for programmatically-compiled methods in Preferences, removing an annoyance"

	^ Utilities authorInitialsPerSe  ~= 'programmatic'! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
warnAboutInsecureContent
	^ self
		valueOfFlag: #warnAboutInsecureContent
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
warnIfChangesFileReadOnly
	^ self
		valueOfFlag: #warnIfChangesFileReadOnly
		ifAbsent: [false]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
warnIfNoChangesFile
	^ self
		valueOfFlag: #warnIfNoChangesFile
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
warnIfNoSourcesFile
	^ self
		valueOfFlag: #warnIfNoSourcesFile
		ifAbsent: [ true ]! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
warningForMacOSFileNameLength
	^ self
		valueOfFlag: #warningForMacOSFileNameLength
		ifAbsent: [false]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'mt 7/25/2016 15:16'!
windowTitleFont
	"Answer the standard font to use for window titles"
	
	^ (UserInterfaceTheme current get: #windowTitleFont)
		ifNil: [TextStyle defaultFont]! !

!Preferences class methodsFor: 'prefs - fonts' stamp: 'sw 12/8/1999 22:18'!
windowTitleStyle
	"Answer the standard style to use for window titles"
	^  self windowTitleFont textStyle! !

!Preferences class methodsFor: 'standard queries' stamp: ''!
wordStyleCursorMovement
	^ self
		valueOfFlag: #wordStyleCursorMovement
		ifAbsent: [ true ]! !

Object subclass: #ProjectLoading
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!ProjectLoading class methodsFor: 'utilities' stamp: 'gk 10/21/2005 11:18'!
bestAccessToFileName: aFileName andDirectory: aDirectoryOrUrlString
	"Answer an array with a stream and a directory. The directory can be nil."
	
	| dir url |
	dir := Project squeakletDirectory.
	(dir fileExists: aFileName) ifTrue: [
		^{dir readOnlyFileNamed: aFileName. dir}].

	aDirectoryOrUrlString isString ifFalse: [
		^{aDirectoryOrUrlString readOnlyFileNamed: aFileName. aDirectoryOrUrlString}].

	url := Url absoluteFromFileNameOrUrlString: aDirectoryOrUrlString.

	(url scheme = 'file') ifTrue: [
		dir := FileDirectory on: url pathForDirectory.
		^{dir readOnlyFileNamed: aFileName. dir}].

	(url path anySatisfy: [:each | each = 'SuperSwikiProj']) ifTrue: [
		dir := SuperSwikiServer new fullPath: url directoryUrl asString.
		^{dir readOnlyFileNamed: url fileName. nil}].

	^{ServerFile new fullPath: aDirectoryOrUrlString. nil}! !

!ProjectLoading class methodsFor: 'utilities' stamp: 'tak 9/3/2006 11:00'!
checkSecurity: aFileName preStream: preStream projStream: projStream
	"Answer true if passed"
	| trusted enterRestricted |
	trusted := SecurityManager default positionToSecureContentsOf:
projStream.
	trusted ifFalse:
		[enterRestricted := (preStream isTypeHTTP or:
[aFileName isNil])
			ifTrue: [Preferences securityChecksEnabled]
			ifFalse: [Preferences standaloneSecurityChecksEnabled].
		enterRestricted
			ifTrue: [SecurityManager default enterRestrictedMode
				ifFalse:
					[preStream close.
					^ false]]].
	^ true! !

!ProjectLoading class methodsFor: 'utilities' stamp: 'tak 9/3/2006 00:17'!
checkStream: aStream 
	(aStream isNil
			or: [aStream size = 0])
		ifFalse: [^ false].
	ProgressNotification signal: '9999 about to enter
project'.
	"the hard part is over"
	self inform: 'It looks like a problem occurred while
getting this project. It may be temporary,
so you may want to try again,' translated.
	^ true! !

!ProjectLoading class methodsFor: 'utilities' stamp: 'tfel 8/10/2016 14:49'!
fileInName: aFileName archive: archive morphOrList: morphOrList  
	| baseChangeSet substituteFont numberOfFontSubstitutes exceptions anObject mgr |
	ResourceCollector current: ResourceCollector new.
	baseChangeSet := ChangeSet current.
	self useTempChangeSet.		"named zzTemp"
	"The actual reading happens here"
	substituteFont := Preferences standardDefaultTextFont copy.
	numberOfFontSubstitutes := 0.
	exceptions := Set new.
	[[anObject := morphOrList fileInObjectAndCodeForProject]
		on: MissingFont do: [ :ex |
				exceptions add: ex.
				numberOfFontSubstitutes := numberOfFontSubstitutes + 1.
				ex resume: substituteFont ]]
			ensure: [ ChangeSet  newChanges: baseChangeSet].
	mgr := ResourceManager new initializeFrom: ResourceCollector current.
	mgr fixJISX0208Resource.
	mgr registerUnloadedResources.
	archive ifNotNil:[mgr preLoadFromArchive: archive cacheName: aFileName].
	ResourceCollector current: nil.
	^ {anObject. numberOfFontSubstitutes. substituteFont. mgr}! !

!ProjectLoading class methodsFor: 'loading' stamp: 'RAA 2/20/2001 20:25'!
installRemoteNamed: remoteFileName from: aServer named: otherProjectName in: currentProject

	| fileAndDir |

	"Find the current ProjectViewMorph, fetch the project, install in ProjectViewMorph without changing size, and jump into new project."

	ProgressNotification signal: '1:foundMostRecent'.
	fileAndDir := self bestAccessToFileName: remoteFileName andDirectory: aServer.
	^self 
		openName: remoteFileName 
		stream: fileAndDir first 
		fromDirectory: fileAndDir second
		withProjectView: (currentProject findProjectView: otherProjectName).
! !

!ProjectLoading class methodsFor: 'loading' stamp: 'tfel 8/10/2016 15:09'!
loadImageSegment: morphOrList  fromDirectory: aDirectoryOrNil withProjectView: existingView numberOfFontSubstitutes: numberOfFontSubstitutes substituteFont: substituteFont mgr: mgr

	| proj projectsToBeDeleted ef |
	Smalltalk at: #Flaps ifPresent: [:flaps |
		(flaps globalFlapTabWithID: 'Navigator' translated)ifNotNil: [:f | f hideFlap]].
	proj := morphOrList arrayOfRoots
			detect: [:mm | mm isKindOf: Project]
			ifNone: [^ nil].
	numberOfFontSubstitutes > 0 ifTrue: [
		proj projectParameterAt: #substitutedFont put: substituteFont].
	ef := proj projectParameterAt: #eToysFont.
	(ef isNil or: [ef ~= substituteFont familySizeFace]) ifTrue: [
		proj projectParameterAt: #substitutedFont put: substituteFont.
	].
	proj projectParameters at: #MultiSymbolInWrongPlace put: false.
		"Yoshiki did not put MultiSymbols into outPointers in older images!!"
	morphOrList arrayOfRoots do: [:obj |
		obj fixUponLoad: proj seg: morphOrList "imageSegment"].
	(proj projectParameters at: #MultiSymbolInWrongPlace) ifTrue: [
		morphOrList arrayOfRoots do: [:obj | (obj isKindOf: Set) ifTrue: [obj rehash]]].

	proj resourceManager: mgr.
	"proj versionFrom: preStream."
	proj lastDirectory: aDirectoryOrNil.
	proj setParent: Project current.
	projectsToBeDeleted := OrderedCollection new.
	existingView == #none ifFalse: [
		self makeExistingView: existingView project: proj projectsToBeDeleted: projectsToBeDeleted].
	ChangeSorter allChangeSets add: proj changeSet.
	Project current projectParameters
		at: #deleteWhenEnteringNewProject
		ifPresent: [ :ignored |
			projectsToBeDeleted add: Project current.
			Project current removeParameter: #deleteWhenEnteringNewProject.
		].
	projectsToBeDeleted isEmpty ifFalse: [
		proj projectParameters
			at: #projectsToBeDeleted
			put: projectsToBeDeleted.
	].
	proj removeParameter: #eToysFont.
	^ proj! !

!ProjectLoading class methodsFor: 'utilities' stamp: 'tfel 8/10/2016 15:12'!
makeExistingView: existingView project: proj projectsToBeDeleted: projectsToBeDeleted
		existingView ifNil: [
			proj createViewIfAppropriate.
			ChangeSorter allChangeSets add: proj changeSet.
			Project current openProject:  proj.
		] ifNotNil: [
			(existingView project isKindOf: DiskProxy) ifFalse: [
				existingView project changeSet name: 
ChangeSet defaultName.
				projectsToBeDeleted add: existingView project.
			].
			(existingView owner isSystemWindow) ifTrue: [
				existingView owner model: proj
			].
			existingView project: proj.
		].! !

!ProjectLoading class methodsFor: 'utilities' stamp: 'cmm 10/23/2013 16:24'!
morphOrList: aFileName stream: preStream fromDirectory: aDirectoryOrNil archive: archive
	"Answer morphOrList or nil if problem happened"
	|  projStream localDir morphOrList |
	projStream := archive
		ifNil: [preStream]
		ifNotNil: [self projectStreamFromArchive: archive].
	(self checkSecurity: aFileName preStream: preStream projStream: projStream)
		ifFalse: [^nil].
	localDir := Project squeakletDirectory.
	aFileName ifNotNil: [
		(aDirectoryOrNil isNil or: [aDirectoryOrNil pathName
~= localDir pathName]) ifTrue: [
			localDir deleteFileNamed: aFileName.
			(localDir fileNamed: aFileName) binary
				nextPutAll: preStream remainingContents;
				close.
		].
	].
	morphOrList := projStream asUnZippedStream.
	preStream sleep.		"if ftp, let the connection close"
	^ morphOrList! !

!ProjectLoading class methodsFor: 'loading' stamp: 'nice 12/26/2009 20:35'!
openFromDirectory: aDirectory andFileName: aFileName

	Project current
		do: [| fileAndDir |
			ProgressNotification signal: '1:foundMostRecent'.
			fileAndDir := self bestAccessToFileName: aFileName andDirectory: aDirectory.
			self 
				openName: aFileName 
				stream: fileAndDir first 
				fromDirectory: fileAndDir second
				withProjectView: nil]
		withProgressInfoOn: nil
		label: 'project loading'
		! !

!ProjectLoading class methodsFor: 'loading' stamp: 'RAA 2/19/2001 08:22'!
openFromFile: preStream fromDirectory: aDirectoryOrNil withProjectView: existingView


	self error: 'use #openFromFile:fromDirectory:withProjectView:'
! !

!ProjectLoading class methodsFor: 'loading' stamp: 'tfel 8/10/2016 14:04'!
openName: aFileName stream: preStream fromDirectory: aDirectoryOrNil
withProjectView: existingView

	^ self openName: aFileName stream: preStream fromDirectory: aDirectoryOrNil
withProjectView: existingView clearOriginFlag: false.! !

!ProjectLoading class methodsFor: 'loading' stamp: 'tfel 8/10/2016 15:16'!
openName: aFileName stream: preStream fromDirectory: aDirectoryOrNil
withProjectView: existingView clearOriginFlag: clearOriginFlag
	"Reconstitute a Morph from the selected file, presumed to
represent a Morph saved via the SmartRefStream mechanism, and open it
in an appropriate Morphic world."

   	| morphOrList archive mgr substituteFont numberOfFontSubstitutes resultArray anObject project manifests dict |
	(self checkStream: preStream) ifTrue: [^ self].
	ProgressNotification signal: '0.2'.
	archive _ preStream isZipArchive
		ifTrue:[ZipArchive new readFrom: preStream]
		ifFalse:[nil].
	archive ifNotNil:[
	manifests _ (archive membersMatching: '*manifest').
	(manifests size = 1 and: [((dict _ self parseManifest: manifests first contents) at: 'Project-Format' ifAbsent: []) = 'S-Expression'])
		ifTrue: [
			^ (self respondsTo: #openSexpProjectDict:stream:fromDirectory:withProjectView:)
				ifTrue: [self openSexpProjectDict: dict stream: preStream fromDirectory: aDirectoryOrNil withProjectView: existingView]
				ifFalse: [self inform: 'Cannot load S-Expression format projects without Etoys' translated]]].

	morphOrList _ self morphOrList: aFileName stream: preStream fromDirectory: aDirectoryOrNil archive: archive.
	morphOrList ifNil: [^ self].
	ProgressNotification  signal: '0.4'.
	resultArray _ self fileInName: aFileName archive: archive morphOrList: morphOrList.
	anObject _ resultArray first.
	numberOfFontSubstitutes _ resultArray second.
	substituteFont _ resultArray third.
	mgr _ resultArray fourth.
	preStream close.
	ProgressNotification  signal: '0.7'.
		"the hard part is over"
	(anObject isKindOf: ImageSegment) ifTrue: [
		project _ self loadImageSegment: anObject
			fromDirectory: aDirectoryOrNil
			withProjectView: existingView
			numberOfFontSubstitutes: numberOfFontSubstitutes
			substituteFont: substituteFont
			mgr: mgr.
		project noteManifestDetailsIn: dict.
		project removeParameter: #sugarProperties.
		Smalltalk at: #SugarPropertiesNotification ifPresent: [:sp |
			sp signal ifNotNilDo: [:props | 
				project keepSugarProperties: props monitor: true]].
		clearOriginFlag ifTrue: [project forgetExistingURL].
		ProgressNotification  signal: '0.8'.
			^ project
				ifNil: [self inform: 'No project found in this file' translated]
				ifNotNil: [ProjectEntryNotification signal: project]].
	Project current openViewAndEnter: anObject! !

!ProjectLoading class methodsFor: 'utilities' stamp: 'yo 2/5/2008 17:15'!
parseManifest: aString

	| dict line index key value aStream |
	aStream := aString readStream.
	dict := Dictionary new.
	[(line := aStream nextLine) notNil] whileTrue: [
		index := line indexOf: $:.
		index > 0 ifTrue: [
			key := line copyFrom: 1 to: index - 1.
			value := (line copyFrom: index + 1 to: line size) withBlanksTrimmed.
			dict at: key put: value.
		].
	].
	^ dict.! !

!ProjectLoading class methodsFor: 'accessing' stamp: 'tfel 8/10/2016 13:54'!
projectStreamFromArchive: archive
	| ext prFiles entry unzipped |
	ext := FileDirectory dot, Project projectExtension.
	prFiles := archive members select:[:any| any fileName endsWith: ext].
	prFiles isEmpty ifTrue:
		[ext := FileDirectory dot, 'sexp'.
		prFiles := archive members select:[:any| any fileName endsWith: ext]].
	prFiles isEmpty ifTrue: [''].
	entry := prFiles first.
	unzipped := MultiByteBinaryOrTextStream on: (ByteArray new: entry uncompressedSize).
	entry extractTo: unzipped.
	^unzipped reset! !

!ProjectLoading class methodsFor: 'accessing' stamp: 'nice 12/26/2009 20:36'!
thumbnailFromUrl: urlString
	"Load the project, and make a thumbnail to it in the current project.
ProjectLoading thumbnailFromUrl: 'http://www.squeak.org/Squeak2.0/2.7segments/SqueakEasy.extSeg'.
"

	Project canWeLoadAProjectNow ifFalse: [^ self].
	Project current
		do: [| fileName fileAndDir |
			ProgressNotification signal: '1:foundMostRecent'.
			fileName := (urlString findTokens: '/') last.
			fileAndDir := self bestAccessToFileName: fileName andDirectory: urlString.
			self
				openName: fileName 
				stream: fileAndDir first 
				fromDirectory: fileAndDir second
				withProjectView: nil]
		withProgressInfoOn: nil
		label: 'project loading'
! !

!ProjectLoading class methodsFor: 'utilities' stamp: 'ar 9/27/2005 20:10'!
useTempChangeSet
	"While reading the project in, use the temporary change set zzTemp"

	| zz |
	zz := ChangeSet named: 'zzTemp'.
	zz ifNil: [zz := ChangeSet basicNewChangeSet: 'zzTemp'].
	ChangeSet  newChanges: zz.! !

Object subclass: #PseudoClass
	instanceVariableNames: 'name definition organization source metaClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!PseudoClass commentStamp: '<historical>' prior: 0!
I provide an inert model of a Class, used by FileContentsBrowser to manipulate filedout code. Instead of a method dictionary or selectors onto CompiledMethods, I have a dictionary ("source") of selectors onto ChangeRecords, which were, in the case of FileContentsBrowser, parsed from a source or change set file.!

!PseudoClass methodsFor: 'methods' stamp: 'nice 2/23/2011 21:50'!
addMethodChange: aChangeRecord
	| selector |
	selector := self newParser parseSelector: aChangeRecord string.
	self organization classify: selector under: aChangeRecord category.
	self sourceCodeAt: selector put: aChangeRecord! !

!PseudoClass methodsFor: 'accessing' stamp: 'nk 4/29/2004 06:59'!
allCallsOn
	^ (self realClass ifNil: [ ^#() ]) allCallsOn! !

!PseudoClass methodsFor: 'accessing' stamp: 'ul 1/9/2016 22:11'!
allInstVarNames

	self realClass ifNotNil: [ :realClass | ^realClass allInstVarNames ].
	^#()! !

!PseudoClass methodsFor: 'private' stamp: 'nk 2/18/2004 18:33'!
allSubclassesWithLevelDo: classAndLevelBlock startingLevel: level
	^ (self realClass ifNil: [ ^self ])  allSubclassesWithLevelDo: classAndLevelBlock startingLevel: level! !

!PseudoClass methodsFor: 'accessing' stamp: 'nk 2/18/2004 18:32'!
allSuperclasses
	^ (self realClass ifNil: [ ^#() ]) allSuperclasses! !

!PseudoClass methodsFor: 'testing method dictionary' stamp: 'ul 12/12/2009 14:07'!
bindingOf: varName
	self exists ifTrue:[
		(self realClass bindingOf: varName) ifNotNil:[:binding| ^binding].
	].
	^Smalltalk bindingOf: varName asSymbol! !

!PseudoClass methodsFor: 'class' stamp: ''!
classComment: aChangeRecord
	self organization classComment: aChangeRecord! !

!PseudoClass methodsFor: 'errors' stamp: ''!
classNotDefined
	^self inform: self name,' is not defined in the system.
You have to define this class first.'.! !

!PseudoClass methodsFor: 'class' stamp: 'di 1/13/1999 12:00'!
classPool 
	self exists ifFalse: [^ nil].
	^ self realClass classPool! !

!PseudoClass methodsFor: 'class' stamp: 'nice 9/21/2010 10:35'!
comment
	^self organization commentRemoteStr
		ifNil: [ self name,' has not been commented in this file']
		ifNotNil: [:rStr | rStr string]! !

!PseudoClass methodsFor: 'class' stamp: 'ar 2/5/2004 15:18'!
commentString
	^self comment asString! !

!PseudoClass methodsFor: 'class' stamp: ''!
commentString: aString
	self classComment: aString asText. "Just wrap it"! !

!PseudoClass methodsFor: 'class' stamp: ''!
comment: aString
	self commentString: aString.! !

!PseudoClass methodsFor: 'accessing' stamp: 'ul 1/9/2016 22:11'!
compilerClass

	^self realClass
		ifNil: [ Compiler ]
		ifNotNil: [ :realClass | realClass compilerClass ]! !

!PseudoClass methodsFor: 'private' stamp: ''!
confirmRemovalOf: aString
	^self confirm:'Remove ',aString,' ?'! !

!PseudoClass methodsFor: 'class' stamp: 'LC 10/8/2001 04:46'!
definition
	| link linkText defText |
	^definition ifNil:
		[defText := Text fromString: 'There is no class definition for '.
		link := TextLink new.
		linkText := link analyze: self name with: 'Definition'.
		linkText := Text string: (linkText ifNil: ['']) attribute: link.
		defText append: linkText; append: ' in this file'].! !

!PseudoClass methodsFor: 'class' stamp: ''!
definition: aString
	definition := aString! !

!PseudoClass methodsFor: 'private' stamp: ''!
evaluate: aString
	^Compiler evaluate: aString for: nil logged: true! !

!PseudoClass methodsFor: 'testing' stamp: ''!
exists
	^(Smalltalk at: self name asSymbol ifAbsent:[^false]) isKindOf: Behavior! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: ''!
fileIn
	"FileIn the receiver"
	self hasDefinition ifTrue:[self fileInDefinition].
	self fileInMethods: self selectors.
	metaClass ifNotNil:[metaClass fileIn].
	self needsInitialize ifTrue:[
		self evaluate: self name,' initialize'.
	].! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: ''!
fileInCategory: aCategory
	^self fileInMethods: (self organization listAtCategoryNamed: aCategory)! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: 'ar 7/16/2005 14:06'!
fileInDefinition
	self hasDefinition ifFalse:[^self].
	(self makeSureSuperClassExists: (definition copyUpTo: Character space)) ifFalse:[^self].
	self hasDefinition ifTrue:[
		Transcript cr; show:'Defining ', self name.
		self evaluate: self definition].
	self exists ifFalse:[^self].
	self hasComment ifTrue:[self realClass classComment: self comment].! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: ''!
fileInMethods
	^self fileInMethods: self selectors! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: 'nice 12/26/2009 20:38'!
fileInMethods: aCollection
	"FileIn all methods with selectors taken from aCollection"
	| theClass |
	self exists ifFalse:[^self classNotDefined].
	theClass := self realClass.
	aCollection do:[:sel|
		| cat |
		cat := self organization categoryOfElement: sel.
		cat = self removedCategoryName ifFalse:[
			theClass 
				compile: (self sourceCodeAt: sel) 
				classified: cat
				withStamp: (self stampAt: sel)
				notifying: nil.
		].
	].! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: ''!
fileInMethod: selector
	^self fileInMethods: (Array with: selector)! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: 'yo 7/5/2004 20:21'!
fileOut
	| internalStream |
	internalStream := WriteStream on: (String new: 1000).
	self fileOutOn: internalStream.
	self needsInitialize ifTrue:[
		internalStream cr; nextChunkPut: self name,' initialize'.
	].

	FileStream writeSourceCodeFrom: internalStream baseName: self name isSt: true useHtml: false.
! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: 'yo 7/5/2004 20:21'!
fileOutCategory: categoryName

	| internalStream |
	internalStream := WriteStream on: (String new: 1000).
	self fileOutMethods: (self organization listAtCategoryNamed: categoryName)
			on: internalStream.
	FileStream writeSourceCodeFrom: internalStream baseName: (self name, '-', categoryName) isSt: true useHtml: false.
! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: 'hg 9/6/2000 12:45'!
fileOutDefinitionOn: aStream
	self hasDefinition ifFalse:[^self].
	aStream nextChunkPut: self definition; cr.
	self hasComment
		ifTrue:
			[aStream cr.
			self organization commentRemoteStr fileOutOn:
aStream]! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: ''!
fileOutMethodsOn: aStream
	^self fileOutMethods: self selectors on: aStream.! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: 'nice 12/26/2009 20:40'!
fileOutMethods: aCollection on: aStream
	"FileOut all methods with selectors taken from aCollection"
	| categories |
	categories := Dictionary new.
	aCollection do:[:sel|
		| cat |
		cat := self organization categoryOfElement: sel.
		cat = self removedCategoryName ifFalse:[
			(categories includesKey: cat) 
				ifFalse:[categories at: cat put: Set new].
			(categories at: cat) add: sel].
	].
	categories associationsDo:[:assoc|
		assoc value do:[:sel|
			aStream cr.
			(self sourceCode at: sel) fileOutOn: aStream.
		].
	].! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: 'yo 7/5/2004 20:21'!
fileOutMethod: selector
	| internalStream |

	internalStream := WriteStream on: (String new: 1000).

	self fileOutMethods: (Array with: selector) on: internalStream.

	FileStream writeSourceCodeFrom: internalStream baseName: (self name , '-' , (selector copyReplaceAll: ':' with: '')) isSt: true useHtml: false.
! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: ''!
fileOutOn: aStream
	"FileOut the receiver"
	self fileOutDefinitionOn: aStream.
	metaClass ifNotNil:[metaClass fileOutDefinitionOn: aStream].
	self fileOutMethods: self selectors on: aStream.
	metaClass ifNotNil:[metaClass fileOutMethods: metaClass selectors on: aStream].! !

!PseudoClass methodsFor: 'accessing' stamp: ''!
fullName
	^self name! !

!PseudoClass methodsFor: 'testing' stamp: 'jmv 3/2/2010 16:24'!
hasChanges

	self sourceCode isEmpty ifFalse:[^true].
	self organization hasNoComment ifFalse:[^true].
	definition ifNotNil: [ ^true].
	metaClass ifNotNil: [ ^metaClass hasChanges].
	^false! !

!PseudoClass methodsFor: 'testing' stamp: ''!
hasComment
	^self organization commentRemoteStr notNil! !

!PseudoClass methodsFor: 'testing' stamp: ''!
hasDefinition
	^definition notNil! !

!PseudoClass methodsFor: 'testing' stamp: ''!
hasMetaclass
	^metaClass notNil! !

!PseudoClass methodsFor: 'testing method dictionary' stamp: 'jcg 1/9/2010 14:23'!
includesSelector: aSymbol 
	self exists
		ifTrue: [^self realClass includesSelector: aSymbol]
		ifFalse: [
			"Not sure what the answer should be in this case (I'm unfamiliar with this class),
			 so I want to force an error rather than answer something nonsensical.  This is
			 no worse than the old behavior, when this method didn't even exist, resulting
			 in a MessageNotUnderstood error."
			self error: 'real class does not exist'
		]! !

!PseudoClass methodsFor: 'accessing' stamp: 'ul 1/9/2016 22:10'!
instVarNames

	self realClass ifNotNil: [ :realClass | ^realClass instVarNames ].
	^#()! !

!PseudoClass methodsFor: 'testing' stamp: 'al 3/1/2006 22:26'!
isTrait
	^false! !

!PseudoClass methodsFor: 'printing' stamp: 'cwp 6/22/2012 15:39'!
literalScannedAs: scannedLiteral environment: anEnvironment notifying: requestor 
	^ scannedLiteral! !

!PseudoClass methodsFor: 'printing' stamp: 'sma 6/17/1999 00:00'!
literalScannedAs: scannedLiteral notifying: requestor 
	^ scannedLiteral! !

!PseudoClass methodsFor: 'private' stamp: ''!
makeSureClassExists: aString
	| theClass |
	theClass := Smalltalk at: (aString asSymbol) ifAbsent:[nil].
	theClass ifNotNil:[^true].
	^self confirm: aString,' does not exist in the system. Use nil instead?'.! !

!PseudoClass methodsFor: 'private' stamp: ''!
makeSureSuperClassExists: aString
	| theClass |
	theClass := Smalltalk at: (aString asSymbol) ifAbsent:[nil].
	theClass ifNotNil:[^true].
	^self confirm: 'The super class ',aString,' does not exist in the system. Use nil instead?'.! !

!PseudoClass methodsFor: 'class' stamp: ''!
metaClass
	^metaClass ifNil:[metaClass := PseudoMetaclass new name: (self name)].! !

!PseudoClass methodsFor: 'methods' stamp: ''!
methodChange: aChangeRecord
	aChangeRecord isMetaClassChange ifTrue:[
		^self metaClass addMethodChange: aChangeRecord.
	] ifFalse:[
		^self addMethodChange: aChangeRecord.
	].
! !

!PseudoClass methodsFor: 'accessing' stamp: ''!
name
	^name! !

!PseudoClass methodsFor: 'testing' stamp: 'ar 3/5/2010 20:40'!
nameExists
	^Smalltalk globals includesKey: self name asSymbol! !

!PseudoClass methodsFor: 'accessing' stamp: ''!
name: anObject
	name := anObject! !

!PseudoClass methodsFor: 'testing' stamp: 'nice 10/20/2009 21:02'!
needsInitialize
	^self hasMetaclass and:[
		self metaClass includesSelector: #initialize]! !

!PseudoClass methodsFor: 'accessing' stamp: 'ul 4/2/2011 14:01'!
newCompiler
	"Answer a Compiler suitable for compiling this Behavior"
	^self compilerClass new parser: self newParser! !

!PseudoClass methodsFor: 'accessing' stamp: 'ul 4/2/2011 00:39'!
newParser
	"Answer a Parser suitable for parsing source code in this Behavior"
	^self parserClass new! !

!PseudoClass methodsFor: 'accessing' stamp: 'NS 4/6/2004 15:46'!
organization
	organization ifNil: [organization := PseudoClassOrganizer defaultList: SortedCollection new].

	"Making sure that subject is set correctly. It should not be necessary."
	organization setSubject: self.
	^ organization! !

!PseudoClass methodsFor: 'private' stamp: 'ul 1/9/2016 22:14'!
parserClass

	^self compilerClass parserClass! !

!PseudoClass methodsFor: 'accessing' stamp: 'nk 2/23/2005 18:00'!
prettyPrinterClass
	^self class prettyPrinterClass! !

!PseudoClass methodsFor: 'printing' stamp: 'ar 2/5/2004 16:04'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut:$(; print: name; nextPut:$)! !

!PseudoClass methodsFor: 'accessing' stamp: 'ul 1/9/2016 22:23'!
realClass

	^Smalltalk globals classNamed: self name! !

!PseudoClass methodsFor: 'removing' stamp: 'jmv 3/2/2010 16:24'!
removeAllUnmodified
	| stClass |
	self exists ifFalse:[^self].
	self removeUnmodifiedMethods: self selectors.
	stClass := self realClass.
	(self hasDefinition and:[stClass definition = self definition]) ifTrue:[definition := nil].
	(self hasComment and:[stClass comment asString = self commentString]) ifTrue:[ self classComment: nil].
	metaClass ifNotNil: [ metaClass removeAllUnmodified]! !

!PseudoClass methodsFor: 'categories' stamp: ''!
removeCategory: selector
	(self organization listAtCategoryNamed: selector) do:[:sel|
		self organization removeElement: sel.
		self sourceCode removeKey: sel.
	].
	self organization removeCategory: selector.! !

!PseudoClass methodsFor: 'categories' stamp: 'nice 9/4/2013 21:29'!
removedCategoryName
	^#'*** removed methods ***'! !

!PseudoClass methodsFor: 'methods' stamp: ''!
removeMethod: selector
	self organization removeElement: selector.
	self sourceCode removeKey: selector.
! !

!PseudoClass methodsFor: 'methods' stamp: ''!
removeSelector: aSelector
	| catName |
	catName := self removedCategoryName.
	self organization addCategory: catName before: self organization categories first.
	self organization classify: aSelector under: catName.
	self sourceCodeAt: aSelector put:'methodWasRemoved' asText.! !

!PseudoClass methodsFor: 'removing' stamp: ''!
removeUnmodifiedMethods: aCollection
	| stClass |
	self exists ifFalse:[^self].
	stClass := self realClass.
	aCollection do:[:sel|
		(self sourceCodeAt: sel) = (stClass sourceCodeAt: sel ifAbsent:['']) asString ifTrue:[
			self removeMethod: sel.
		].
	].
	self organization removeEmptyCategories.! !

!PseudoClass methodsFor: 'class' stamp: 'nk 2/18/2004 18:30'!
renameTo: aString

	self hasDefinition ifTrue:[
		self isMeta ifTrue:[
			self definition: (self definition
				copyReplaceAll: name,' class'
				with: aString, ' class').
		] ifFalse:[
			self definition: (self definition 
					copyReplaceAll:'ubclass: #',name
					with:'ubclass: #', aString)]].
	name := aString.
	metaClass ifNotNil:[metaClass renameTo: aString].! !

!PseudoClass methodsFor: 'methods' stamp: ''!
selectors
	^self sourceCode keys! !

!PseudoClass methodsFor: 'methods' stamp: 'ul 2/5/2010 06:34'!
selectorsDo: aBlock

	^self sourceCode keysDo: aBlock! !

!PseudoClass methodsFor: 'class' stamp: 'di 1/13/1999 12:00'!
sharedPools 
	self exists ifFalse: [^ nil].
	^ self realClass sharedPools! !

!PseudoClass methodsFor: 'methods' stamp: ''!
sourceCode
	^source ifNil:[source := Dictionary new]! !

!PseudoClass methodsFor: 'methods' stamp: ''!
sourceCodeAt: sel
	^(self sourceCode at: sel) string! !

!PseudoClass methodsFor: 'methods' stamp: 'mt 8/12/2016 09:59'!
sourceCodeAt: sel ifAbsent: block
	^ self sourceCode
		at: sel
		ifPresent: [:src | src string]
		ifAbsent: block! !

!PseudoClass methodsFor: 'methods' stamp: ''!
sourceCodeAt: sel put: object
	self sourceCode at: sel put: object! !

!PseudoClass methodsFor: 'methods' stamp: ''!
sourceCodeTemplate
	^''! !

!PseudoClass methodsFor: 'methods' stamp: 'sw 6/10/2003 17:31'!
stampAt: selector
	"Answer the authoring time-stamp of the change"

	| code |
	^ ((code := self sourceCode at: selector) isKindOf: ChangeRecord)
		ifTrue:
			[code stamp]
		ifFalse:
			[code string]! !

!PseudoClass methodsFor: 'accessing' stamp: 'nice 3/30/2010 22:16'!
theMetaClass
	^ self metaClass! !

!PseudoClass methodsFor: 'accessing' stamp: 'wod 5/19/1998 17:42'!
theNonMetaClass
	"Sent to a class or metaclass, always return the class"

	^self! !

!PseudoClass methodsFor: 'categories' stamp: ''!
whichCategoryIncludesSelector: aSelector 
	"Answer the category of the argument, aSelector, in the organization of 
	the receiver, or answer nil if the receiver does not inlcude this selector."

	^ self organization categoryOfElement: aSelector! !

!PseudoClass methodsFor: 'enumerating' stamp: 'ul 1/9/2016 22:06'!
withAllSubAndSuperclassesDo: aBlock

	self realClass ifNotNil: [ :realClass | ^realClass withAllSubAndSuperclassesDo: aBlock ].
	^aBlock value: self! !

!PseudoClass methodsFor: 'enumerating' stamp: 'ul 1/9/2016 22:06'!
withAllSuperclassesDo: aBlock

	self realClass ifNotNil: [ :realClass | ^realClass withAllSuperclassesDo: aBlock ].
	^aBlock value: self! !

PseudoClass subclass: #PseudoMetaclass
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!PseudoMetaclass methodsFor: 'accessing' stamp: ''!
fullName
	^self name,' class'! !

!PseudoMetaclass methodsFor: 'accessing' stamp: ''!
realClass
	^super realClass class! !

!PseudoMetaclass methodsFor: 'accessing' stamp: 'FBS 3/4/2004 14:17'!
theNonMetaClass
	"Sent to a class or metaclass, always return the class"

	^self realClass theNonMetaClass! !

Object subclass: #RealEstateAgent
	instanceVariableNames: ''
	classVariableNames: 'ReverseStaggerOffset StaggerOffset StaggerOrigin StandardSize StandardWindowOrigins'
	poolDictionaries: ''
	category: 'System-Support'!

!RealEstateAgent commentStamp: '<historical>' prior: 0!
Responsible for real-estate management on the screen, which is to say, controlling where new windows appear, with what sizes, etc.  5/20/96 sw!

!RealEstateAgent class methodsFor: 'utilities' stamp: 'kfr 7/21/2014 15:56'!
assignCollapseFrameFor: aSSView 
	"Offer up a location along the left edge of the screen for a collapsed SSView. Make sure it doesn't overlap any other collapsed frames."

	| grid otherFrames topLeft viewBox collapsedFrame extent newFrame verticalBorderDistance top |
	grid := 8.
	verticalBorderDistance := 8.
	aSSView isMorph
		ifTrue: [otherFrames := (SystemWindow windowsIn: aSSView world satisfying: [:w | w ~= aSSView])
						collect: [:w | w collapsedFrame]
						thenSelect: [:rect | rect notNil].
				viewBox := self maximumUsableAreaInWorld: aSSView world]
		ifFalse: [otherFrames := ScheduledControllers scheduledWindowControllers
						collect: [:aController | aController view ~= aSSView ifTrue: [aController view collapsedFrame]]
						thenSelect: [:rect | rect notNil].
				viewBox := Display boundingBox].
	collapsedFrame := aSSView collapsedFrame.
	extent := collapsedFrame notNil
				ifTrue: [collapsedFrame extent]
				ifFalse: [aSSView isMorph
					ifTrue: [aSSView getRawLabel width + aSSView labelWidgetAllowance @ (aSSView labelHeight + 2)]
					ifFalse: [(aSSView labelText extent x + 70) @ aSSView labelHeight
							min: aSSView labelDisplayBox extent]].
	collapsedFrame notNil
		ifTrue: [(otherFrames anySatisfy: [:f | collapsedFrame intersects: f])
				ifFalse: ["non overlapping"
					^ collapsedFrame]].
	top := viewBox top + verticalBorderDistance.
	[topLeft := viewBox left @ top.
	newFrame := topLeft extent: extent.
	newFrame bottom <= (viewBox height - verticalBorderDistance)]
		whileTrue: 
			[(otherFrames anySatisfy: [:w | newFrame intersects: w])
				ifFalse: ["no overlap"
					^ newFrame].
			top := top + grid].
	"If all else fails... (really to many wins here)"
	aSSView isMorph
		ifTrue:[ Project current showWorldMainDockingBar ifTrue:[^ (0 @ Project current world mainDockingBars first height) extent: extent]].
	^ 0 @ 0 extent: extent! !

!RealEstateAgent class methodsFor: 'utilities' stamp: 'nice 12/15/2013 22:35'!
assignCollapsePointFor: aSSView
	"Offer up a location along the left edge of the screen for a collapsed SSView.
	Make sure it doesn't overlap any other collapsed frames."

	| grid otherFrames y free topLeft viewBox |
	grid := 24.  "should be mult of 8, since manual move is gridded by 8"
	aSSView isMorph
		ifTrue: [otherFrames := (SystemWindow windowsIn: aSSView world satisfying: [:w | true])
					collect: [:w | w collapsedFrame]
					thenSelect: [:rect | rect notNil].
				viewBox := self maximumUsableAreaInWorld: aSSView world]
		ifFalse: [otherFrames := ScheduledControllers scheduledWindowControllers
					collect: [:aController | aController view collapsedFrame]
					thenSelect: [:rect | rect notNil].
				viewBox := Display boundingBox].
	y := viewBox top.
	[(y := y + grid) <= (viewBox height - grid)]
		whileTrue:
		[topLeft := viewBox left@y.
		free := true.
		otherFrames do: [:w | free := free & (topLeft ~= w topLeft)].
		free ifTrue: [^ topLeft]].
	"If all else fails..."
	^ 0 @ 0! !

!RealEstateAgent class methodsFor: 'framing' stamp: 'mt 4/19/2015 18:16'!
initialFrameFor: aView initialExtent: initialExtent world: aWorld

	| scaledExtent |
	scaledExtent := Preferences bigDisplay
		ifTrue: [(initialExtent * 1.75) rounded]
		ifFalse: [initialExtent].

	^ Preferences reverseWindowStagger
		ifTrue: [self strictlyStaggeredInitialFrameFor: aView initialExtent: scaledExtent world: aWorld]
		ifFalse: [self normalInitialFrameFor: aView initialExtent: scaledExtent world: aWorld]! !

!RealEstateAgent class methodsFor: 'framing' stamp: 'RAA 5/25/2000 08:13'!
initialFrameFor: aView world: aWorld
	"Find a plausible initial screen area for the supplied view.  See called method."

	^ self initialFrameFor: aView initialExtent: aView initialExtent world: aWorld! !

!RealEstateAgent class methodsFor: 'class initialization' stamp: 'tk 11/26/1998 09:34'!
initialize
	"Initialize the class variables in the receiver.  5/22/96 sw"
	"RealEstateAgent initialize"

	StaggerOffset := 6 @ 20.
	ReverseStaggerOffset := -6 @ 20.
	StaggerOrigin := 200 @ 30.
	StandardSize := 600@400.! !

!RealEstateAgent class methodsFor: 'accessing' stamp: 'nice 12/15/2013 22:32'!
maximumUsableArea
	^self maximumUsableAreaInWorld: ActiveWorld! !

!RealEstateAgent class methodsFor: 'accessing' stamp: 'mt 5/17/2016 14:22'!
maximumUsableAreaInWorld: aWorldOrNil

	| allowedArea |
	allowedArea := Display usableArea.
	aWorldOrNil ifNotNil: [
		allowedArea := allowedArea intersect: aWorldOrNil visibleClearArea.
		Smalltalk isMorphic ifTrue: [
			(((Smalltalk classNamed: 'Flaps') ifNil: [false] ifNotNil: [:cls | cls anyFlapsVisibleIn: aWorldOrNil])
				and: [self respondsTo: #reduceByFlaps:])
					ifTrue: [allowedArea := self reduceByFlaps: allowedArea]]].
	^allowedArea! !

!RealEstateAgent class methodsFor: 'framing - private' stamp: 'mt 4/19/2015 18:15'!
normalInitialFrameFor: aView initialExtent: initialExtent world: aWorld
	"Find a plausible initial screen area for the supplied view, which should be a StandardSystemView, taking into account the 'reverseWindowStagger' Preference, the size needed, and other windows currently on the screen."

	| allOrigins screenRight screenBottom putativeOrigin putativeFrame allowedArea staggerOrigin otherFrames |

	allowedArea := self maximumUsableAreaInWorld: aWorld.
	screenRight := allowedArea right.
	screenBottom := allowedArea bottom.

	otherFrames := Smalltalk isMorphic
		ifTrue: [(SystemWindow windowsIn: aWorld satisfying: [:w | w isCollapsed not])
					collect: [:w | w bounds]]
		ifFalse: [ScheduledControllers scheduledWindowControllers
				select: [:aController | aController view ~~ nil]
				thenCollect: [:aController | aController view isCollapsed
								ifTrue: [aController view expandedFrame]
								ifFalse: [aController view displayBox]]].

	allOrigins := otherFrames collect: [:f | f origin].
	(self standardPositionsInWorld: aWorld) do:  "First see if one of the standard positions is free"
		[:aPosition | (allOrigins includes: aPosition)
			ifFalse:
				[^ (aPosition extent: initialExtent) translatedAndSquishedToBeWithin: allowedArea]].

	staggerOrigin := (self standardPositionsInWorld: aWorld) first.  "Fallback: try offsetting from top left"
	putativeOrigin := staggerOrigin.

	[putativeOrigin := putativeOrigin + StaggerOffset.
	putativeFrame := putativeOrigin extent: initialExtent.
	(putativeFrame bottom < screenBottom) and:
					[putativeFrame right < screenRight]]
				whileTrue:
					[(allOrigins includes: putativeOrigin)
						ifFalse:
							[^ (putativeOrigin extent: initialExtent) translatedAndSquishedToBeWithin: allowedArea]].
	^ (self scrollBarSetback @ self screenTopSetback extent: initialExtent) translatedAndSquishedToBeWithin: allowedArea! !

!RealEstateAgent class methodsFor: 'settings' stamp: 'di 9/22/1998 20:58'!
screenTopSetback
	Smalltalk isMorphic
		ifTrue: [^ 0]
		ifFalse: [^ 18]! !

!RealEstateAgent class methodsFor: 'settings' stamp: 'di 9/22/1998 20:58'!
scrollBarSetback
	Smalltalk isMorphic
		ifTrue: [^ 16-3]  "width = 16; inset from border by 3"
		ifFalse: [^ 24]! !

!RealEstateAgent class methodsFor: 'accessing' stamp: 'RAA 5/25/2000 08:45'!
standardPositions
	
	self error: 'please use #standardPositionsInWorld:'! !

!RealEstateAgent class methodsFor: 'accessing' stamp: 'RAA 5/25/2000 08:43'!
standardPositionsInWorld: aWorldOrNil
	"Return a list of standard window positions -- this may have one, two, or four of them, depending on the size and shape of the display screen.  "

	| anArea aList  midX midY |

	anArea := self maximumUsableAreaInWorld: aWorldOrNil.

	midX := self scrollBarSetback +   ((anArea width - self scrollBarSetback)  // 2).
	midY := self screenTopSetback + ((anArea height - self screenTopSetback) // 2).
	aList := OrderedCollection with: (self scrollBarSetback @ self screenTopSetback).
	self windowColumnsDesired > 1
		ifTrue:
			[aList add: (midX @ self screenTopSetback)].
	self windowRowsDesired > 1
		ifTrue:
			[aList add: (self scrollBarSetback @ (midY+self screenTopSetback)).
			self windowColumnsDesired > 1 ifTrue:
				[aList add: (midX @ (midY+self screenTopSetback))]].
	^ aList! !

!RealEstateAgent class methodsFor: 'accessing' stamp: 'cmm 6/8/2010 18:26'!
standardSize
	^ StandardSize! !

!RealEstateAgent class methodsFor: 'accessing' stamp: 'cmm 6/9/2010 15:12'!
standardSize: extent
	StandardSize := extent! !

!RealEstateAgent class methodsFor: 'settings' stamp: 'cbr 3/30/2010 17:30'!
standardWindowExtent
	"Answer the standard default extent for new windows.  "

	| effectiveExtent width strips height grid allowedArea maxLevel extent |
	effectiveExtent := self maximumUsableArea extent
					- (self scrollBarSetback @ self screenTopSetback).
	Preferences reverseWindowStagger ifTrue:
		["NOTE: following copied from strictlyStaggeredInitialFrameFor:"
		allowedArea := self maximumUsableArea insetBy: (
			self scrollBarSetback @ self screenTopSetback extent: 0@0
		).
		"Number to be staggered at each corner (less on small screens)"
		maxLevel := allowedArea area > 300000 ifTrue: [3] ifFalse: [2].
		"Amount by which to stagger (less on small screens)"
		grid := allowedArea area > 500000 ifTrue: [40] ifFalse: [20].
		extent := (allowedArea extent - (grid*(maxLevel+1*2) + (grid//2))) min: StandardSize "600@400".
		(extent y < 300) ifTrue: [^ (extent x)@300 ].
		^ extent].
	width := (strips := self windowColumnsDesired) > 1
		ifTrue:
			[effectiveExtent x // strips]
		ifFalse:
			[(3 * effectiveExtent x) // 4].
	height := (strips := self windowRowsDesired) > 1
		ifTrue:
			[effectiveExtent y // strips]
		ifFalse:
			[(3 * effectiveExtent y) //4].
	^ width @ height

"RealEstateAgent standardWindowExtent"! !

!RealEstateAgent class methodsFor: 'framing - private' stamp: 'nice 12/15/2013 22:35'!
strictlyStaggeredInitialFrameFor: aStandardSystemView initialExtent: initialExtent world: aWorld
	"This method implements a staggered window placement policy that I (di) like.
	Basically it provides for up to 4 windows, staggered from each of the 4 corners.
	The windows are staggered so that there will always be a corner visible."

	| allowedArea grid initialFrame otherFrames cornerSel corner delta putativeCorner free maxLevel |

	allowedArea :=(self maximumUsableAreaInWorld: aWorld)
		insetBy: (self scrollBarSetback @ self screenTopSetback extent: 0@0).
	"Number to be staggered at each corner (less on small screens)"
	maxLevel := allowedArea area > 300000 ifTrue: [3] ifFalse: [2].
	"Amount by which to stagger (less on small screens)"
	grid := allowedArea area > 500000 ifTrue: [40] ifFalse: [20].
	initialFrame := 0@0 extent: ((initialExtent
							"min: (allowedArea extent - (grid*(maxLevel+1*2) + (grid//2))))
							min: 600@400")).
	otherFrames := Smalltalk isMorphic
		ifTrue: [(SystemWindow windowsIn: aWorld satisfying: [:w | w isCollapsed not])
					collect: [:w | w bounds]]
		ifFalse: [ScheduledControllers scheduledWindowControllers
				select: [:aController | aController view ~~ nil]
				thenCollect: [:aController | aController view isCollapsed
								ifTrue: [aController view expandedFrame]
								ifFalse: [aController view displayBox]]].
	0 to: maxLevel do:
		[:level | 
		1 to: 4 do:
			[:ci | cornerSel := #(topLeft topRight bottomRight bottomLeft) at: ci.
			corner := allowedArea perform: cornerSel.
			"The extra grid//2 in delta helps to keep title tabs distinct"
			delta := (maxLevel-level*grid+(grid//2)) @ (level*grid).
			1 to: ci-1 do: [:i | delta := delta rotateBy: #right centerAt: 0@0]. "slow way"
			putativeCorner := corner + delta.
			free := true.
			otherFrames do:
				[:w |
				free := free & ((w perform: cornerSel) ~= putativeCorner)].
			free ifTrue:
				[^ (initialFrame align: (initialFrame perform: cornerSel)
								with: putativeCorner)
						 translatedAndSquishedToBeWithin: allowedArea]]].
	"If all else fails..."
	^ (self scrollBarSetback @ self screenTopSetback extent: initialFrame extent)
		translatedAndSquishedToBeWithin: allowedArea! !

!RealEstateAgent class methodsFor: 'settings' stamp: 'RAA 11/21/1999 22:54'!
windowColumnsDesired
	"Answer how many separate vertical columns of windows are wanted.  5/22/96 sw"
	^ Preferences reverseWindowStagger
		ifTrue:
			[1]
		ifFalse:
			[(self maximumUsableArea width > 640)
				ifTrue:
					[2]
				ifFalse:
					[1]]! !

!RealEstateAgent class methodsFor: 'settings' stamp: 'RAA 11/21/1999 22:54'!
windowRowsDesired
	"Answer how many separate horizontal rows of windows are wanted.  5/22/96 sw"
	^ Preferences reverseWindowStagger
		ifTrue:
			[1]
		ifFalse:
			[(self maximumUsableArea height > 480)
				ifTrue:
					[2]
				ifFalse:
					[1]]! !

Object subclass: #RecentMessages
	instanceVariableNames: 'methodReferences maximumSubmissionCount isSuspended'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'System-Support'!

!RecentMessages class methodsFor: 'accessing' stamp: 'fbs 9/9/2013 19:56'!
default
	^ Default ifNil: [Default := RecentMessages newRemembering: self numberOfRecentSubmissionsToStore].! !

!RecentMessages class methodsFor: 'instance creation' stamp: 'cmm 10/11/2013 11:08'!
newRemembering: anInteger 
	^ self new
		 maximumSubmissionCount: anInteger ;
		 yourself! !

!RecentMessages class methodsFor: 'preferences' stamp: 'cmm 10/11/2013 11:09'!
numberOfRecentSubmissionsToStore
	<preference: 'Number of recent submissions to store' category: 'Tools' description: 'Answer how many methods back the ''recent method submissions'' history should store' type: #Number>
	^ self default maximumSubmissionCount! !

!RecentMessages class methodsFor: 'preferences' stamp: 'cmm 10/11/2013 11:09'!
numberOfRecentSubmissionsToStore: anInteger
	self default maximumSubmissionCount: anInteger! !

!RecentMessages class methodsFor: 'preferences' stamp: 'cmm 11/30/2013 18:13'!
setRecentHistorySize
	"Let the user specify the recent history size."
	self numberOfRecentSubmissionsToStore:
		(UIManager default
			request: 'How many recent changes
should be remembered?'
			initialAnswer: self numberOfRecentSubmissionsToStore asString)! !

!RecentMessages class methodsFor: 'system startup' stamp: 'fbs 9/14/2013 14:16'!
startUp
	SystemChangeNotifier uniqueInstance
		notify: self default
		ofAllSystemChangesUsing: #event:.! !

!RecentMessages methodsFor: 'private' stamp: 'cmm 10/8/2013 16:47'!
defaultSize
	^ 30! !

!RecentMessages methodsFor: 'change logging' stamp: 'cmm 12/19/2013 10:58'!
event: anEvent 
	"Hook for SystemChangeNotifier"
	(anEvent isCommented and: [ anEvent itemKind = SystemChangeNotifier classKind ]) ifTrue:
		[ self
			recordSelector: #Comment
			forClass: anEvent item
			inEnvironment: anEvent itemEnvironment ].
	((anEvent isAdded or: [ anEvent isModified ]) and: [ anEvent itemKind = SystemChangeNotifier methodKind ]) ifTrue:
		[ anEvent itemRequestor ifNotNil:
			[ self
				recordSelector: anEvent itemSelector
				forClass: anEvent itemClass
				inEnvironment: anEvent itemEnvironment ].
		InMidstOfFileinNotification signal ifFalse: [ Utilities changed: #recentMethodSubmissions ] ]! !

!RecentMessages methodsFor: 'initialize-release' stamp: 'cmm 12/2/2013 10:44'!
initialize
	super initialize.
	methodReferences := OrderedCollection new: (maximumSubmissionCount := 30).
	self resume! !

!RecentMessages methodsFor: 'testing' stamp: 'fbs 9/9/2013 19:53'!
isEmpty
	^ methodReferences isEmpty.! !

!RecentMessages methodsFor: 'accessing' stamp: 'cmm 10/11/2013 13:34'!
leastRecent
	^ methodReferences
		ifEmpty: [ nil ]
		ifNotEmpty: [ methodReferences last ]! !

!RecentMessages methodsFor: 'accessing' stamp: 'fbs 9/2/2013 10:10'!
maximumSubmissionCount
	^ maximumSubmissionCount.! !

!RecentMessages methodsFor: 'accessing' stamp: 'cmm 10/12/2013 15:52'!
maximumSubmissionCount: anInteger
	maximumSubmissionCount := anInteger.
	self trim! !

!RecentMessages methodsFor: 'accessing' stamp: 'cmm 10/11/2013 13:22'!
methodReferences
	^ methodReferences asArray! !

!RecentMessages methodsFor: 'accessing' stamp: 'cmm 10/11/2013 13:35'!
mostRecent
	^ methodReferences first! !

!RecentMessages methodsFor: 'accessing' stamp: 'cmm 11/30/2013 18:23'!
recordSelector: aSelector forClass: aClass inEnvironment: anEnvironment 
	| ref |
	(isSuspended = true or: [ aClass wantsChangeSetLogging not ]) ifTrue: [ ^ self ].
	ref := MethodReference
		class: aClass
		selector: aSelector
		environment: anEnvironment.
	methodReferences
		remove: ref
		ifAbsent: [  ].
	ref methodSymbol isDoIt ifFalse: [ methodReferences addFirst: ref ].
	self trim! !

!RecentMessages methodsFor: 'accessing' stamp: 'cmm 12/2/2013 10:47'!
resume
	isSuspended := false! !

!RecentMessages methodsFor: 'accessing' stamp: 'ul 10/25/2013 18:44'!
revertMostRecent
	"If the most recent method submission was a method change, revert
	that change, and if it was a submission of a brand-new method, 
	remove that method."
	| changeRecords lastSubmission theClass theSelector |
	methodReferences ifEmpty: [ ^ Beeper beep ].
	lastSubmission := self mostRecent.
	theClass := lastSubmission actualClass ifNil: [ ^ Beeper beep ].
	theSelector := lastSubmission methodSymbol.
	changeRecords := theClass changeRecordsAt: theSelector.
	changeRecords isEmptyOrNil ifTrue: [ ^ Beeper beep ].
	changeRecords size = 1
		ifTrue: [ "method has no prior version, so reverting in this case means removing"
			theClass removeSelector: theSelector ]
		ifFalse: [ changeRecords second fileIn: lastSubmission environment ]
! !

!RecentMessages methodsFor: 'accessing' stamp: 'fbs 9/2/2013 10:11'!
size
	^ methodReferences size.! !

!RecentMessages methodsFor: 'accessing' stamp: 'cmm 10/3/2013 14:18'!
suspendWhile: aBlock
	| priorSuspended |
	priorSuspended := isSuspended.
	isSuspended := true.
	aBlock ensure: [ isSuspended := priorSuspended ]! !

!RecentMessages methodsFor: 'private' stamp: 'cmm 10/12/2013 15:56'!
trim
	[ methodReferences size > maximumSubmissionCount ] whileTrue: [ methodReferences removeLast ]! !

Object subclass: #ResourceCollector
	instanceVariableNames: 'stubMap originalMap locatorMap localDirectory baseUrl resourceDirectory internalStubs'
	classVariableNames: 'Current'
	poolDictionaries: ''
	category: 'System-Support'!

!ResourceCollector commentStamp: '<historical>' prior: 0!
The ResourceCollector collects resources that are encountered during project loading or publishing. It merely decouples the places where resources are held from the core object enumeration so that resources can be stored independently from what is enumerated for publishing.!

!ResourceCollector class methodsFor: 'accessing' stamp: 'ar 2/24/2001 21:41'!
current
	^Current! !

!ResourceCollector class methodsFor: 'accessing' stamp: 'ar 2/24/2001 21:41'!
current: aResourceManager
	Current := aResourceManager! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/27/2001 20:32'!
baseUrl
	^baseUrl! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/27/2001 20:39'!
baseUrl: aString
	baseUrl := aString.
	baseUrl isEmpty ifFalse:[
		baseUrl last = $/ ifFalse:[baseUrl := baseUrl copyWith: $/].
	].! !

!ResourceCollector methodsFor: 'initialize' stamp: 'ar 2/27/2001 23:08'!
forgetObsolete
	"Forget obsolete locators, e.g., those that haven't been referenced and not been stored on a file."
	locatorMap keys "copy" do:[:k|
		(locatorMap at: k) localFileName ifNil:[locatorMap removeKey: k]].! !

!ResourceCollector methodsFor: 'initialize' stamp: 'ar 3/3/2001 19:49'!
initialize
	| fd pvt |
	originalMap := IdentityDictionary new.
	stubMap := IdentityDictionary new.
	locatorMap := IdentityDictionary new.
	internalStubs := IdentityDictionary new.
	fd := ScriptingSystem formDictionary.
	pvt := ScriptingSystem privateGraphics asSet.
	fd keysAndValuesDo:[:sel :form|
		(pvt includes: sel) ifFalse:[
			internalStubs at: form put:
				(DiskProxy 
					global: #ScriptingSystem
					selector: #formAtKey:extent:depth:
					args: {sel. form extent. form depth})]].! !

!ResourceCollector methodsFor: 'initialize' stamp: 'ar 2/27/2001 22:36'!
initializeFrom: aResourceManager
	"Initialize the receiver from aResourceManager."
	aResourceManager resourceMap keysAndValuesDo:[:loc :res|
		(res notNil)
			ifTrue:[locatorMap at: res put:  loc.
					loc localFileName: nil].
	].! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/24/2001 22:23'!
localDirectory
	^localDirectory! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/24/2001 22:24'!
localDirectory: aDirectory
	localDirectory := aDirectory! !

!ResourceCollector methodsFor: 'accessing' stamp: 'tk 6/28/2001 15:58'!
locatorMap
	"allow outsiders to store in it.  For files that are not resources that do want to live in the resource directory locally and on the server.  (.t files for example)"

	^locatorMap! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/27/2001 22:54'!
locators
	^locatorMap values! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/27/2001 20:07'!
locatorsDo: aBlock
	^locatorMap valuesDo: aBlock! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/27/2001 17:01'!
noteResource: aResourceStub replacing: anObject
	"Remember the fact that we need to load aResource which will replace anObject."
	stubMap at: aResourceStub put: anObject.! !

!ResourceCollector methodsFor: 'objects from disk' stamp: 'ar 2/24/2001 22:37'!
objectForDataStream: refStream
	"This should never happen; when projects get written they must be decoupled from the resource collector. If you get the error message below something is seriously broken."
	self error:'Cannot write resource manager'! !

!ResourceCollector methodsFor: 'accessing' stamp: 'mir 10/29/2003 13:33'!
objectForDataStream: refStream fromForm: aForm
	"Return a replacement for aForm to be stored instead"
	| stub fName copy loc fullSize nameAndSize |

	"First check if the form is one of the intrinsic Squeak forms"
	stub := internalStubs at: aForm ifAbsent:[nil].
	stub ifNotNil:[
		refStream replace: aForm with: stub. 
		^stub].

	"Now see if we have created the stub already 
	(this may happen if for instance some form is shared)"
	stub := originalMap at: aForm ifAbsent:[nil].
	stub ifNotNil:[^aForm].
	aForm hibernate.
	aForm bits class == FormStub ifTrue:[^nil].	"something is wrong"
	"too small to be of interest"
	"(aForm bits byteSize < 4096) ifTrue:[^aForm]."
	"We'll turn off writing out forms until we figure out how to reliably deal with resources"
	true ifTrue: [^aForm].

	"Create our stub form"
	stub := FormStub 
		extent: (aForm width min: 32) @ (aForm height min: 32) 
		depth: (aForm depth min: 8).
	aForm displayScaledOn: stub.
	aForm hibernate.

	"Create a copy of the original form which we use to store those bits"
	copy := Form extent: aForm extent depth: aForm depth bits: nil.
	copy setResourceBits: aForm bits.

	"Get the locator for the form (if we have any)"
	loc := locatorMap at: aForm ifAbsent:[nil].

	"Store the resource file"
	nameAndSize := self writeResourceForm: copy locator: loc.
	fName := nameAndSize first.
	fullSize := nameAndSize second.

	ProgressNotification signal: '2:resourceFound' extra: stub.
	stub hibernate.
	"See if we need to assign a new locator"
	(loc notNil and:[loc hasRemoteContents not]) ifTrue:[
		"The locator describes some local resource. 
		If we're preparing to upload the entire project to a
		remote server, make it a remote URL instead."
"		(baseUrl isEmpty not and:[baseUrl asUrl hasRemoteContents])
			ifTrue:[loc urlString: baseUrl, fName].
"
		baseUrl isEmpty not
			ifTrue:[loc urlString: self resourceDirectory , fName]].

	loc ifNil:[
		loc := ResourceLocator new urlString: self resourceDirectory , fName.
		locatorMap at: aForm put: loc].
	loc localFileName: (localDirectory fullNameFor: fName).
	loc resourceFileSize: fullSize.
	stub locator: loc.

	"Map old against stub form"
	aForm setResourceBits: stub.
	originalMap at: aForm put: copy.
	stubMap at: stub put: aForm.
	locatorMap at: aForm put: loc.
	"note: *must* force aForm in out pointers if 
	in IS or else won't get #comeFullyUpOnReload:"
	refStream replace: aForm with: aForm.
	^aForm! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/27/2001 22:59'!
removeLocator: loc
	locatorMap keys "copy" do:[:k|
		(locatorMap at: k) = loc ifTrue:[locatorMap removeKey: k]].! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/27/2001 20:21'!
replaceAll
	"Replace all resources by their originals. Done after the resource have been collected to get back to the original state."
	originalMap keysAndValuesDo:[:k :v|
		v ifNotNil:[k replaceByResource: v].
	].! !

!ResourceCollector methodsFor: 'accessing' stamp: 'mir 6/21/2001 14:51'!
resourceDirectory
	resourceDirectory ifNil: [resourceDirectory := self baseUrl copyFrom: 1 to: (self baseUrl lastIndexOf: $/)].
	^resourceDirectory! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/27/2001 20:08'!
resourceFileNames
	"Return a list of all the resource files created"
	^locatorMap values asArray collect:[:loc| loc localFileName].! !

!ResourceCollector methodsFor: 'accessing' stamp: 'ar 2/27/2001 17:01'!
stubMap
	^stubMap! !

!ResourceCollector methodsFor: 'resource writing' stamp: 'yo 11/13/2002 23:30'!
writeResourceForm: aForm fromLocator: aLocator
	"The given form has been externalized before. If it was reasonably compressed, use the bits of the original data - this allows us to recycle GIF, JPEG, PNG etc. data without using the internal compression (which is in most cases inferior). If necessary the data will be retrieved from its URL location. This retrieval is done only if the resouce comes from either
		* the local disk (in which case the file has never been published)
		* the browser cache (in which case we don't cache the resource locally)
	In any other case we will *not* attempt to retrieve it, because doing so can cause the system to connect to the network which is probably not what we want. It should be a rare case anyways; could only happen if one clears the squeak cache selectively."
	| fName fStream url data |
	"Try to be smart about the name of the file"
	fName := (aLocator urlString includes: $:)
		ifTrue: [
			url := aLocator urlString asUrl.
			url path last]
		ifFalse: [aLocator urlString].
	fName isEmptyOrNil ifFalse:[fName := fName asFileName].
	(fName isEmptyOrNil or:[localDirectory isAFileNamed: fName]) ifTrue:[
		"bad luck -- duplicate name"
		fName := localDirectory 
				nextNameFor:'resource' 
				extension: (FileDirectory extensionFor: aLocator urlString)].
	"Let's see if we have cached it locally"
	ResourceManager
		lookupCachedResource: self baseUrl , aLocator urlString
		ifPresentDo:[:stream | data := stream upToEnd].
	"Check if the cache entry is without qualifying baseUrl. Workaround for older versions."
	data ifNil:[
		ResourceManager
			lookupCachedResource: aLocator urlString
			ifPresentDo:[:stream | data := stream upToEnd]].
	data ifNil:[
		"We don't have it cached locally. Retrieve it from its original location."
		((url notNil and: [url hasRemoteContents]) and:[HTTPClient isRunningInBrowser not])
			ifTrue:[^nil]. "see note above"
		(Url schemeNameForString: aLocator urlString)
			ifNil: [^nil].
		data := HTTPLoader default retrieveContentsFor: aLocator urlString.
		data ifNil:[^nil].
		data := data content.
	].
	"data size > aForm bits byteSize ifTrue:[^nil]."
	fStream := localDirectory newFileNamed: fName.
	fStream binary.
	fStream nextPutAll: data.
	fStream close.
	^{fName. data size}! !

!ResourceCollector methodsFor: 'resource writing' stamp: 'ar 9/23/2002 03:34'!
writeResourceForm: aForm locator: aLocator
	"Store the given form on a file. Return an array with the name and the size of the file"
	| fName fStream fullSize result writerClass |
	aLocator ifNotNil:[
		result := self writeResourceForm: aForm fromLocator: aLocator.
		result ifNotNil:[^result]
		"else fall through"
	].
	fName := localDirectory nextNameFor:'resource' extension:'form'.
	fStream := localDirectory newFileNamed: fName.
	fStream binary.
	aForm storeResourceOn: fStream.
false ifTrue:[
	"What follows is a Really, REALLY bad idea. I leave it in as a reminder of what you should NOT do. 
	PART I: Using JPEG or GIF compression on forms where we don't have the original data means loosing both quality and alpha information if present..."
	writerClass := ((Smalltalk includesKey: #JPEGReaderWriter2)
		and: [(Smalltalk at: #JPEGReaderWriter2) new isPluginPresent])
		ifTrue: [(Smalltalk at: #JPEGReaderWriter2)]
		ifFalse: [GIFReadWriter].
	writerClass putForm: aForm onStream: fStream.
	fStream open.
	fullSize := fStream size.
	fStream close.
].

	"Compress contents here"
true ifTrue:[
	"...PART II: Using the builtin compression which combines RLE+ZIP is AT LEAST AS GOOD as PNG and how much more would you want???"
	fStream position: 0.
	fStream compressFile.
	localDirectory deleteFileNamed: fName.
	localDirectory rename: fName, FileDirectory dot, 'gz' toBe: fName.
	fStream := localDirectory readOnlyFileNamed: fName.
	fullSize := fStream size.
	fStream close.
].
	^{fName. fullSize}! !

Object subclass: #ResourceLocator
	instanceVariableNames: 'urlString fileSize localFileName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!ResourceLocator commentStamp: '<historical>' prior: 0!
Describes where a resource can be found.

Instance variables:
	urlString	<String> 	The URL of the resource
	fileSize		<Integer>	The size of the resource
	localFileName	<String>	When non-nil, the place where this resource was/is stored.!

!ResourceLocator class methodsFor: 'utilities' stamp: 'ul 8/14/2015 18:03'!
make: newURLString relativeTo: oldURLString 
	"Local file refs are not handled well, so work around here"
	^((oldURLString includesSubstring: '://') not
		and: [(newURLString includesSubstring: '://') not])
		ifTrue: [oldURLString , (UnixFileDirectory localNameFor: newURLString)]
		ifFalse: [(newURLString asUrlRelativeTo: oldURLString asUrl) asString]! !

!ResourceLocator methodsFor: 'accessing' stamp: ''!
adjustToDownloadUrl: downloadUrl
	"Adjust to the fully qualified URL for this resource."
	self urlString: (ResourceLocator make: self urlString relativeTo: downloadUrl) unescapePercents! !

!ResourceLocator methodsFor: 'accessing' stamp: 'mir 6/19/2001 16:55'!
adjustToRename: newName from: oldName
	"Adjust to the fully qualified URL for this resource."
	self urlString: (self urlString copyReplaceAll: oldName with: newName)! !

!ResourceLocator methodsFor: 'comparing' stamp: 'ar 2/27/2001 20:02'!
hash
	^urlString hash! !

!ResourceLocator methodsFor: 'testing' stamp: 'ar 2/27/2001 22:11'!
hasRemoteContents
	"Return true if we describe a resource which is non-local, e.g., on some remote server."
	(urlString indexOf: $:) = 0 ifTrue:[^false]. "no scheme"
	^urlString asUrl hasRemoteContents! !

!ResourceLocator methodsFor: 'accessing' stamp: 'ar 2/27/2001 20:00'!
localFileName
	^localFileName! !

!ResourceLocator methodsFor: 'accessing' stamp: 'ar 2/27/2001 20:01'!
localFileName: aString
	localFileName := aString! !

!ResourceLocator methodsFor: 'printing' stamp: 'ar 2/27/2001 20:02'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(;
		print: urlString;
		nextPut: $)! !

!ResourceLocator methodsFor: 'accessing' stamp: 'ar 3/2/2001 18:13'!
resourceFileSize
	^fileSize! !

!ResourceLocator methodsFor: 'accessing' stamp: 'ar 3/2/2001 18:13'!
resourceFileSize: aNumber
	fileSize := aNumber! !

!ResourceLocator methodsFor: 'comparing' stamp: 'tk 7/5/2001 22:10'!
species
	^ResourceLocator! !

!ResourceLocator methodsFor: 'accessing' stamp: 'ar 2/27/2001 19:57'!
urlString
	^urlString! !

!ResourceLocator methodsFor: 'accessing' stamp: 'ar 2/27/2001 19:57'!
urlString: aString
	urlString := aString.! !

!ResourceLocator methodsFor: 'comparing' stamp: 'tk 7/5/2001 22:10'!
= aLocator

	^ self species == aLocator species and: [self urlString = aLocator urlString]
! !

Object subclass: #ResourceManager
	instanceVariableNames: 'resourceMap loaded unloaded stopSemaphore stopFlag loaderProcess'
	classVariableNames: 'CachedResources LocalizedExternalResources'
	poolDictionaries: ''
	category: 'System-Support'!

!ResourceManager class methodsFor: 'private-resources' stamp: 'mir 11/29/2001 16:19'!
addCacheLocation: aString for: urlString
	| locations |
	locations := CachedResources at: urlString ifAbsentPut: [#()].
	(locations includes: aString)
		ifFalse: [CachedResources at: urlString put: ({aString} , locations)]! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'ar 5/30/2001 23:21'!
cacheResource: urlString inArchive: archiveName
	"Remember the given url as residing in the given archive"
	| fd file fullName |
	fullName := 'zip://', archiveName.
	((self resourceCache at: urlString ifAbsent:[#()]) 
		anySatisfy:[:cache| cache = fullName]) ifTrue:[^self]. "don't cache twice"
	fd := Project squeakletDirectory.
	"update cache"
	file := [fd oldFileNamed: self resourceCacheName] 
			on: FileDoesNotExistException
			do:[:ex| fd forceNewFileNamed: self resourceCacheName].
	file setToEnd.
	file nextPutAll: urlString; cr.
	file nextPutAll: fullName; cr.
	file close.
	self addCacheLocation: fullName for: urlString.! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'yo 12/20/2003 02:12'!
cacheResource: urlString stream: aStream
	| fd localName file buf |
	HTTPClient shouldUsePluginAPI ifTrue:[^self]. "use browser cache"
	(self resourceCache at: urlString ifAbsent:[#()]) size > 0 
		ifTrue:[^self]. "don't waste space"
	fd := Project squeakletDirectory.
	localName := fd nextNameFor: 'resource' extension:'cache'.
	file := fd forceNewFileNamed: localName.
	buf := ByteArray new: 10000.
	aStream binary.
	file binary.
	[aStream atEnd] whileFalse:[
		buf := aStream next: buf size into: buf.
		file nextPutAll: buf.
	].
	file close.
	"update cache"
	file := [fd oldFileNamed: self resourceCacheName] 
			on: FileDoesNotExistException
			do:[:ex| fd forceNewFileNamed: self resourceCacheName].
	file setToEnd.
	file nextPutAll: urlString; cr.
	file nextPutAll: localName; cr.
	file close.
	self addCacheLocation: localName for: urlString.
	aStream position: 0.
! !

!ResourceManager class methodsFor: 'private-resources' stamp: 'ar 5/30/2001 23:55'!
loadResource: urlString fromCacheFileNamed: fileName in: dir
	| archiveName file archive |
	(fileName beginsWith: 'zip://') ifTrue:[
		archiveName := fileName copyFrom: 7 to: fileName size.
		archive := [dir readOnlyFileNamed: archiveName] 
			on: FileDoesNotExistException
			do:[:ex| ex return: nil].
		archive ifNil:[^nil].
		archive isZipArchive ifTrue:[
			archive := ZipArchive new readFrom: archive.
			file := archive members detect:[:any| any fileName = urlString] ifNone:[nil]].
		file ifNotNil:[file := file contentStream].
		archive close.
	] ifFalse:[
		file := [dir readOnlyFileNamed: fileName] 
				on: FileDoesNotExistException
				do:[:ex| ex return: nil].
	].
	^file! !

!ResourceManager class methodsFor: 'private-resources' stamp: 'mir 8/21/2001 15:50'!
localizedExternalResources
	^LocalizedExternalResources ifNil:[LocalizedExternalResources := Dictionary new]! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'mir 9/15/2002 15:59'!
lookupCachedResource: cachedUrlString ifPresentDo: streamBlock
	"See if we have cached the resource described by the given url and if so, evaluate streamBlock with the cached resource."
	|  urlString candidates url stream |
	CachedResources ifNil:[^self].

	candidates := CachedResources at: cachedUrlString ifAbsent:[nil].
	(self lookupCachedResource: cachedUrlString in: candidates ifPresentDo: streamBlock)
		ifTrue: [^self].

	urlString := self relocatedExternalResource: cachedUrlString.
	urlString ifNil: [^self].
	candidates := CachedResources at: urlString ifAbsent:[nil].
	candidates
		ifNil: [
			(url := urlString asUrl) schemeName = 'file'
				ifTrue: [
					stream := [FileStream readOnlyFileNamed: url pathForFile] 
							on: FileDoesNotExistException do:[:ex| ex return: nil].
					stream
						ifNotNil: [[streamBlock value: stream] ensure: [stream close]]]]
		ifNotNil: [self lookupCachedResource: urlString in: candidates ifPresentDo: streamBlock]! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'nice 12/26/2009 20:55'!
lookupCachedResource: urlString in: candidates ifPresentDo: streamBlock
	"See if we have cached the resource described by the given url and if so, evaluate streamBlock with the cached resource."
	| sortedCandidates dir |
	(candidates isNil or:[candidates size = 0])
		ifTrue:[^false].
	"First, try non-zip members (faster since no decompression is involved)"
	sortedCandidates := (candidates reject:[:each| each beginsWith: 'zip://']),
					(candidates select:[:each| each beginsWith: 'zip://']).
	dir := Project squeakletDirectory.
	sortedCandidates do:[:fileName|
		| file |
		file := self loadResource: urlString fromCacheFileNamed: fileName in: dir.
		file ifNotNil:[
			[streamBlock value: file] ensure:[file close].
			^true]].
	^false! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'nice 10/21/2009 00:12'!
lookupOriginalResourceCacheEntry: resourceFileName for: resourceUrl
	"See if we have cached the resource described by the given url in an earlier version of the same project on the same server. In that case we don't need to upload it again but rather link to it."
	| resourceBase resourceMatch matchingUrls |
	
	CachedResources ifNil:[^nil].

	"Strip the version number from the resource url"
	resourceBase := resourceUrl copyFrom: 1 to: (resourceUrl lastIndexOf: $.) .
	"Now collect all urls that have the same resource base"
	resourceMatch := resourceBase , '*/' , resourceFileName.
	matchingUrls := self resourceCache keys
		select: [:entry | (resourceMatch match: entry) and: [(entry beginsWith: resourceUrl) not]].
	matchingUrls isEmpty
		ifTrue: [^nil].
	matchingUrls asArray sort do: [:entry | | candidates |
			candidates := (self resourceCache at: entry).
			candidates isEmptyOrNil
				ifFalse: [candidates do: [:candidate |
					candidate = resourceFileName
						ifTrue: [^entry]]]].
	^nil! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'nice 6/11/2010 21:27'!
reloadCachedResources	"ResourceManager reloadCachedResources"
	"Reload cached resources from the disk"
	| fd files stream url localName storeBack archiveName |
	CachedResources := Dictionary new.
	LocalizedExternalResources := nil.
	fd := Project squeakletDirectory.
	files := fd fileNames asSet.
	stream := [fd readOnlyFileNamed: self resourceCacheName]
				on: FileDoesNotExistException 
				do:[:ex| fd forceNewFileNamed: self resourceCacheName].
	stream size < 50000 ifTrue:[stream := ReadStream on: stream contentsOfEntireFile].
	storeBack := false.
	[stream atEnd] whileFalse:[
		url := stream nextLine.	
		localName := stream nextLine.
		(localName beginsWith: 'zip://') ifTrue:[
			archiveName := localName copyFrom: 7 to: localName size.
			(files includes: archiveName) 
				ifTrue:[self addCacheLocation: localName for: url]
				ifFalse:[storeBack := true].
		] ifFalse:[
			(files includes: localName) 
				ifTrue:[self addCacheLocation: localName for: url]
				ifFalse:[storeBack := true]
		].
	].
	stream close.
	storeBack ifTrue:[
		stream := fd forceNewFileNamed: self resourceCacheName.
		CachedResources keysAndValuesDo:[:urlString :cacheLocs|
			cacheLocs do:[:cacheLoc|
				stream nextPutAll: urlString; cr.
				stream nextPutAll: cacheLoc; cr].
		].
		stream close.
	].! !

!ResourceManager class methodsFor: 'private-resources' stamp: 'mir 8/21/2001 16:06'!
relocatedExternalResource: urlString
	^self localizedExternalResources at: urlString ifAbsent: [nil]! !

!ResourceManager class methodsFor: 'private-resources' stamp: 'mir 8/21/2001 16:00'!
relocatedExternalResource: urlString to: newUrlString
	| originalURL |
	originalURL := (self localizedExternalResources includesKey: urlString)
		ifTrue: [self localizedExternalResources at: urlString]
		ifFalse: [urlString].
	self localizedExternalResources at: newUrlString put: originalURL! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'mir 8/21/2001 17:24'!
renameCachedResource: urlString to: newUrlString
	"A project was renamed. Reflect this change by duplicating the cache entry to the new url."
	self renameCachedResource: urlString to: newUrlString external: true! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'mir 12/3/2001 13:14'!
renameCachedResource: urlString to: newUrlString external: isExternal
	"A project was renamed. Reflect this change by duplicating the cache entry to the new url."
	| candidates |
	CachedResources
		ifNil:[
			isExternal
				ifTrue: [self resourceCache "force init" ]
				ifFalse: [^self]].
	candidates := CachedResources at: urlString ifAbsent:[nil].
	(candidates isNil or:[candidates size = 0])
		ifFalse: [
		candidates do: [:candidate |
			self addCacheLocation: candidate for: newUrlString]].
	isExternal
		ifTrue: [self relocatedExternalResource: urlString to: newUrlString]! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'ar 8/23/2001 17:52'!
resourceCache
	^CachedResources ifNil:[
		CachedResources := Dictionary new.
		self reloadCachedResources.
		CachedResources].! !

!ResourceManager class methodsFor: 'resource caching' stamp: 'ar 3/3/2001 17:27'!
resourceCacheName
	^'resourceCache.map'! !

!ResourceManager methodsFor: 'private' stamp: 'ar 3/2/2001 19:25'!
abandonResourcesThat: matchBlock
	"Private. Forget resources that match the given argument block"
	resourceMap keys "need copy" do:[:loc|
		(matchBlock value: loc) ifTrue:[
			resourceMap removeKey: loc ifAbsent:[].
			loaded remove: loc ifAbsent:[].
			unloaded remove: loc ifAbsent:[].
		].
	].! !

!ResourceManager methodsFor: 'accessing' stamp: 'ar 2/27/2001 20:50'!
addResource: anObject locator: aLocator
	resourceMap at: aLocator put: anObject.
	loaded add: aLocator.! !

!ResourceManager methodsFor: 'accessing' stamp: 'ar 3/2/2001 20:56'!
addResource: anObject url: urlString
	^self addResource: anObject locator: (ResourceLocator new urlString: urlString)! !

!ResourceManager methodsFor: 'accessing' stamp: 'mir 6/26/2001 17:33'!
adjustToDownloadUrl: downloadUrl
	"Adjust the resource manager to the current download location. A project might have been moved manually to a different location or server."

	downloadUrl isEmptyOrNil ifTrue: [^self].

	self resourceMap
		keysDo:[:locator | locator adjustToDownloadUrl: downloadUrl].
	self resourceMap rehash.
	unloaded rehash! !

!ResourceManager methodsFor: 'accessing' stamp: 'nice 12/26/2009 20:51'!
adjustToNewServer: newResourceUrl from: oldResourceUrl
	"Adjust the resource manager to the current download location. A project might have been moved manually to a different location or server."
	| urlMap |
	newResourceUrl isEmptyOrNil ifTrue: [^self].
	urlMap := Dictionary new.
	self resourceMap
		keysDo: [:locator | 
			"Local file refs are not handled well, so work around here"
			| oldUrl newUrl |
			oldUrl := ResourceLocator make: locator urlString relativeTo: oldResourceUrl.
			newUrl := ResourceLocator make: locator urlString relativeTo: newResourceUrl.
			oldUrl ~= newUrl
				ifTrue: [urlMap at: oldUrl asString unescapePercents put: newUrl asString unescapePercents]].
	self resourceMap rehash.
	unloaded rehash.
	urlMap keysAndValuesDo: [:old :new |
		ResourceManager renameCachedResource: old to: new]! !

!ResourceManager methodsFor: 'accessing' stamp: 'nice 12/26/2009 20:51'!
adjustToRename: newName from: oldName
	"Adjust the resource manager to the current download location. A project might have been moved manually to a different location or server."
	| urlMap |
	newName isEmptyOrNil ifTrue: [^self].
	urlMap := Dictionary new.
	self resourceMap
		keysDo: [:locator | 
			| oldUrl |
			oldUrl := locator urlString.
			locator adjustToRename: newName from: oldName.
			urlMap at: oldUrl put: locator urlString].
	self resourceMap rehash.
	unloaded rehash.
	urlMap keysAndValuesDo: [:old :new |
		ResourceManager renameCachedResource: old to: new]! !

!ResourceManager methodsFor: 'backward-compatibility' stamp: 'fbs 7/25/2013 07:16'!
convertMapNameForBackwardcompatibilityFrom: aString 
	(Smalltalk platformName = 'Mac OS' 
		and: ['10*' match: Smalltalk osVersion]) 
			ifTrue: [^aString convertFromWithConverter: ShiftJISTextConverter new].
	^aString convertFromSystemString! !

!ResourceManager methodsFor: 'private' stamp: 'nice 10/20/2009 23:52'!
fixJISX0208Resource
	resourceMap keys do: [:key |
		| value url |
		value := resourceMap at: key.
		url := key urlString copy.
		url isOctetString not ifTrue: [url mutateJISX0208StringToUnicode].
		resourceMap removeKey: key.
		key urlString: url.
		resourceMap at: key put: value.
	].
! !

!ResourceManager methodsFor: 'private' stamp: 'ar 3/3/2001 15:30'!
formChangedReminder
	^[World newResourceLoaded].! !

!ResourceManager methodsFor: 'private' stamp: 'ar 3/2/2001 17:22'!
hackURL: urlString
	(urlString findString: '/SuperSwikiProj/') > 0 
		ifTrue:[^urlString copyReplaceAll: '/SuperSwikiProj/' with: '/uploads/']
		ifFalse:[^urlString]! !

!ResourceManager methodsFor: 'initialize' stamp: 'ar 2/27/2001 16:54'!
initialize
	"So resources may get garbage collected if possible"
	self reset.! !

!ResourceManager methodsFor: 'initialize' stamp: 'nice 12/26/2009 20:51'!
initializeFrom: aCollector
	"Initialize the receiver from the given resource collector. None of the resources have been loaded yet, so make register all resources as unloaded."
	aCollector stubMap keysAndValuesDo:[:stub :res|
		| newLoc |
		newLoc := stub locator.
		resourceMap at: newLoc put: res.
		"unloaded add: newLoc."
	].! !

!ResourceManager methodsFor: 'loading' stamp: 'ar 5/30/2001 23:11'!
installResource: aResource from: aStream locator: loc
	| repl |
	aResource ifNil:[^false]. "it went away, so somebody might have deleted it"
	(aStream == nil or:[aStream size = 0]) ifTrue:[^false]. "error?!!"
	repl := aResource clone readResourceFrom: aStream asUnZippedStream.
	repl ifNotNil:[
		aResource replaceByResource: repl.
		unloaded remove: loc.
		loaded add: loc.
		^true
	].
	^false! !

!ResourceManager methodsFor: 'loading' stamp: 'nice 12/26/2009 20:52'!
loadCachedResources
	"Load all the resources that we have cached locally"
	self class reloadCachedResources.
	self prioritizedUnloadedResources do:[:loc|
		self class lookupCachedResource: loc urlString ifPresentDo:[:stream|
			| resource |
			resource := resourceMap at: loc ifAbsent:[nil].
			self installResource: resource
				from: stream
				locator: loc.
			(resource isForm) ifTrue:[
				self formChangedReminder value.
				World displayWorldSafely].
		].
	].! !

!ResourceManager methodsFor: 'loading' stamp: 'nice 12/26/2009 20:52'!
loaderProcess
	| loader requests |
	loader := HTTPLoader default.
	requests := Dictionary new.
	self prioritizedUnloadedResources do:[:loc|
		| req |
		req := HTTPLoader httpRequestClass for: (self hackURL: loc urlString) in: loader.
		loader addRequest: req.
		requests at: req put: loc].
	[stopFlag or:[requests isEmpty]] whileFalse:[
		stopSemaphore waitTimeoutMSecs: 500.
		requests keys "need a copy" do:[:r|
			r isSemaphoreSignaled ifTrue:[
				| locator resource stream |
				locator := requests at: r.
				requests removeKey: r.
				stream := r contentStream.
				resource := resourceMap at: locator ifAbsent:[nil].
				self class cacheResource: locator urlString stream: stream.
				self installResource: resource
					from: stream
					locator: locator.
				(resource isForm) ifTrue:[
					WorldState addDeferredUIMessage: self formChangedReminder]
ifFalse: [self halt].
			].
		].
	].
	"Either done downloading or terminating process"
	stopFlag ifTrue:[loader abort].
	loaderProcess := nil.
	stopSemaphore := nil.! !

!ResourceManager methodsFor: 'private' stamp: 'mir 8/20/2001 17:12'!
localizeAllExternalResources
	"Should be a preference later."
	^true! !

!ResourceManager methodsFor: 'accessing' stamp: 'nice 12/26/2009 20:54'!
makeAllProjectResourcesLocalTo: resourceUrl
	"Change the urls in the resource locators so project specific resources are stored and referenced locally. Project specific resources are all those that are kept locally in any of the project's versions."

	| locators projectBase |
 	"Construct the version neutral project base"
	resourceUrl isEmptyOrNil ifTrue: [^self].
	projectBase := resourceUrl copyFrom: 1 to: (resourceUrl lastIndexOf: $.) - 1.
	locators := OrderedCollection new.
	self resourceMap
		keysAndValuesDo:[:loc :res | res ifNotNil: [locators add: loc]].
	locators do: [:locator |
		| locUrl |
		locUrl := locator urlString.
		locUrl ifNotNil: [
			| lastSlash |
			lastSlash := locUrl lastIndexOf: $/.
			lastSlash > 0
				ifTrue: [
					| locBase localResource isExternal |
					locBase := locUrl copyFrom: 1 to: lastSlash - 1.
					locBase := locBase copyFrom: 1 to: (((locBase lastIndexOf: $.) - 1) max: 0).
					isExternal := projectBase ~= locBase.
					(isExternal not
						or: [self localizeAllExternalResources])
						ifTrue: [
							localResource := locUrl copyFrom: lastSlash+1 to: locUrl size.
							"Update the cache entry to point to the new resource location"
							ResourceManager renameCachedResource: locUrl to: (resourceUrl , localResource) external: isExternal.
							locator urlString: localResource]]]].
	self resourceMap rehash
! !

!ResourceManager methodsFor: 'loading' stamp: 'nice 12/26/2009 20:54'!
preLoadFromArchive: aZipArchive cacheName: aFileName
	"Load the resources from the given zip archive"
	| nameMap resMap |
	self class reloadCachedResources.
	resMap := Dictionary new.
	nameMap := Dictionary new.
	unloaded do:[:locator|
		locator localFileName: nil.
		nameMap at: locator urlString put: locator.
		resMap at: locator urlString put: (resourceMap at: locator)].

	aZipArchive members do:[:entry|
		| stream orig loc |
		stream := nil.
		orig := resMap at: (self convertMapNameForBackwardcompatibilityFrom: entry fileName ) ifAbsent:[nil].
		loc := nameMap at: (self convertMapNameForBackwardcompatibilityFrom: entry fileName ) ifAbsent:[nil].
		"note: orig and loc may be nil for non-resource members"
		(orig notNil and:[loc notNil]) ifTrue:[
			stream := entry contentStream.
			self installResource: orig from: stream locator: loc.
			stream reset.
			aFileName 
				ifNil:[self class cacheResource: loc urlString stream: stream]
				ifNotNil:[self class cacheResource: loc urlString inArchive: aFileName]].
	].! !

!ResourceManager methodsFor: 'loading' stamp: 'ar 3/2/2001 18:16'!
prioritizedUnloadedResources
	"Return an array of unloaded resource locators prioritized by some means"
	| list |
	list := unloaded asArray.
	^list sort:[:l1 :l2|
		(l1 resourceFileSize ifNil:[SmallInteger maxVal]) <=
			(l2 resourceFileSize ifNil:[SmallInteger maxVal])]! !

!ResourceManager methodsFor: 'loading' stamp: 'mir 6/18/2001 22:49'!
registerUnloadedResources
	resourceMap keys do: [:newLoc |
		unloaded add: newLoc]
! !

!ResourceManager methodsFor: 'initialize' stamp: 'ar 2/27/2001 16:54'!
reset
	"Clean out everything"
	resourceMap := WeakValueDictionary new.
	loaded := Set new.
	unloaded := Set new.! !

!ResourceManager methodsFor: 'accessing' stamp: 'ar 2/27/2001 20:57'!
resourceMap
	^resourceMap! !

!ResourceManager methodsFor: 'loading' stamp: 'ar 3/3/2001 18:01'!
startDownload
	"Start downloading unloaded resources"
	self stopDownload.
	unloaded isEmpty ifTrue:[^self].
	self loadCachedResources.
	unloaded isEmpty ifTrue:[^self].
	stopFlag := false.
	stopSemaphore := Semaphore new.
	loaderProcess := [self loaderProcess] newProcess.
	loaderProcess priority: Processor lowIOPriority.
	loaderProcess resume.! !

!ResourceManager methodsFor: 'loading' stamp: 'ar 3/2/2001 17:09'!
stopDownload
	"Stop downloading unloaded resources"
	loaderProcess ifNil:[^self].
	stopFlag := true.
	stopSemaphore signal.
	[loaderProcess == nil] whileFalse:[(Delay forMilliseconds: 10) wait].
	stopSemaphore := nil.! !

!ResourceManager methodsFor: 'loading' stamp: 'ar 2/27/2001 21:42'!
updateResourcesFrom: aCollector
	"We just assembled all the resources in a project.
	Include all that were newly found"
	self reset. "start clean"
	aCollector stubMap keysAndValuesDo:[:stub :res|
		"update all entries"
		resourceMap at: stub locator put: res.
		loaded add: stub locator.
	].! !

Object subclass: #SecureHashAlgorithm
	instanceVariableNames: 'totalA totalB totalC totalD totalE totals'
	classVariableNames: 'K1 K2 K3 K4 TA TB TC TD TE TI'
	poolDictionaries: ''
	category: 'System-Digital Signatures'!

!SecureHashAlgorithm commentStamp: '<historical>' prior: 0!
This class implements the Secure Hash Algorithm (SHA) described in the U.S. government's Secure Hash Standard (SHS). This standard is described in FIPS PUB 180-1, "SECURE HASH STANDARD", April 17, 1995.

The Secure Hash Algorithm is also described on p. 442 of 'Applied Cryptography: Protocols, Algorithms, and Source Code in C' by Bruce Scheier, Wiley, 1996.

See the comment in class DigitalSignatureAlgorithm for details on its use.

Implementation notes:
The secure hash standard was created with 32-bit hardware in mind. All arithmetic in the hash computation must be done modulo 2^32. This implementation uses ThirtyTwoBitRegister objects to simulate hardware registers; this implementation is about six times faster than using LargePositiveIntegers (measured on a Macintosh G3 Powerbook). Implementing a primitive to process each 64-byte buffer would probably speed up the computation by a factor of 20 or more.
!

!SecureHashAlgorithm class methodsFor: 'class initialization' stamp: 'ul 10/26/2015 02:11'!
initialize
	"SecureHashAlgorithm initialize"
	"For the curious, here's where these constants come from:
	  #(2 3 5 10) collect: [:x | ((x sqrt / 4.0) * (2.0 raisedTo: 32)) truncated hex]"

	K1 := ThirtyTwoBitRegister fromInteger: 16r5A827999.
	K2 := ThirtyTwoBitRegister fromInteger: 16r6ED9EBA1.
	K3 := ThirtyTwoBitRegister fromInteger: 16r8F1BBCDC.
	K4 := ThirtyTwoBitRegister fromInteger: 16rCA62C1D6.
	
	TA := ThirtyTwoBitRegister fromInteger: 16r67452301.
	TB := ThirtyTwoBitRegister fromInteger: 16rEFCDAB89.
	TC := ThirtyTwoBitRegister fromInteger: 16r98BADCFE.
	TD := ThirtyTwoBitRegister fromInteger: 16r10325476.
	TE := ThirtyTwoBitRegister fromInteger: 16rC3D2E1F0.
	(TI := Bitmap new: 5)
		at: 1 put: 16r67452301;	
		at: 2 put: 16rEFCDAB89;
		at: 3 put: 16r98BADCFE;
		at: 4 put: 16r10325476;
		at: 5 put: 16rC3D2E1F0! !

!SecureHashAlgorithm methodsFor: 'private' stamp: 'jm 12/7/1999 23:25'!
constantForStep: i
	"Answer the constant for the i-th step of the block hash loop. We number our steps 1-80, versus the 0-79 of the standard."

	i <= 20 ifTrue: [^ K1].
	i <= 40 ifTrue: [^ K2].
	i <= 60 ifTrue: [^ K3].
	^ K4
! !

!SecureHashAlgorithm methodsFor: 'private' stamp: 'ul 10/27/2015 01:35'!
expandedBlock: aByteArray
	"Convert the given 64 byte buffer into 80 32-bit registers and answer the result."

	| out src |
	out := Array new: 80.
	src := 1.
	1 to: 16 do: [:i |
		out at: i put: (ThirtyTwoBitRegister fromByteArray: aByteArray at: src).
		src := src + 4].

	17 to: 80 do: [:i |
		out at: i put: (
			(out at: i - 3) copy
				bitXor: (out at: i - 8);
				bitXor: (out at: i - 14);
				bitXor: (out at: i - 16);
				leftRotateBy: 1) ].
	^ out
! !

!SecureHashAlgorithm methodsFor: 'private' stamp: 'ul 5/19/2016 10:39'!
finalHash
	"Concatenate the final totals to build the 160-bit integer result."
	"Details: If the primitives are supported, the results are in the totals array. Otherwise, they are in the instance variables totalA through totalE."

	| result |
	result := ByteArray new: 20.
	totals 
		ifNil: [ "compute final hash when not using primitives"
			result
				unsignedShortAt: 1 put: totalE low bigEndian: false;
				unsignedShortAt: 3 put: totalE hi bigEndian: false;
				unsignedShortAt: 5 put: totalD low bigEndian: false;
				unsignedShortAt: 7 put: totalD hi bigEndian: false;
				unsignedShortAt: 9 put: totalC low bigEndian: false;
				unsignedShortAt: 11 put: totalC hi bigEndian: false;
				unsignedShortAt: 13 put: totalB low bigEndian: false;
				unsignedShortAt: 15 put: totalB hi bigEndian: false;
				unsignedShortAt: 17 put: totalA low bigEndian: false;
				unsignedShortAt: 19 put: totalA hi bigEndian: false ]
		ifNotNil: [ "compute final hash when using primitives"
			result
				unsignedLongAt: 1 put: (totals at: 5) bigEndian: false;
				unsignedLongAt: 5 put: (totals at: 4) bigEndian: false;
				unsignedLongAt: 9 put: (totals at: 3) bigEndian: false;
				unsignedLongAt: 13 put: (totals at: 2) bigEndian: false;
				unsignedLongAt: 17 put: (totals at: 1) bigEndian: false ].
	LargePositiveInteger adoptInstance: result.
	^result normalize! !

!SecureHashAlgorithm methodsFor: 'private' stamp: 'ul 10/27/2015 01:30'!
hashFunction: i of: x with: y with: z using: t1 and: t2
	"Compute the hash function for the i-th step of the block hash loop. We number our steps 1-80, versus the 0-79 of the standard."
	"Details: There are four functions, one for each 20 iterations. The second and fourth are the same."

	t1 loadFrom: x.
	i <= 20 ifTrue: [
		t2
			loadFrom: x;
			bitInvert;
			bitAnd: z.
		^t1
			bitAnd: y;
			bitOr: t2 ].
	i <= 40 ifTrue: [
		^t1
			bitXor: y;
			bitXor: z ].
	i <= 60 ifTrue: [
		t2
			loadFrom: x;
			bitOr: y;
			bitAnd: z.
		^t1
			bitAnd: y;
			bitOr: t2 ].
	^t1
		bitXor: y;
		bitXor: z
! !

!SecureHashAlgorithm methodsFor: 'public' stamp: 'jm 12/14/1999 11:56'!
hashInteger: aPositiveInteger
	"Hash the given positive integer. The integer to be hashed should have 512 or fewer bits. This entry point is used in key generation."

	| buffer dstIndex |
	self initializeTotals.

	"pad integer with zeros"
	aPositiveInteger highBit <= 512
		ifFalse: [self error: 'integer cannot exceed 512 bits'].
	buffer := ByteArray new: 64.
	dstIndex := 0.
	aPositiveInteger digitLength to: 1 by: -1 do: [:i |
		buffer at: (dstIndex := dstIndex + 1) put: (aPositiveInteger digitAt: i)].

	"process that one block"
	self processBuffer: buffer.

	^ self finalHash
! !

!SecureHashAlgorithm methodsFor: 'public' stamp: 'ul 10/26/2015 02:10'!
hashInteger: aPositiveInteger seed: seedInteger
	"Hash the given positive integer. The integer to be hashed should have 512 or fewer bits. This entry point is used in the production of random numbers"

	| buffer dstIndex |
	"Initialize totalA through totalE to their seed values."
	totals
		ifNil: [
			totalA := ThirtyTwoBitRegister
				fromInteger: ((seedInteger bitShift: -128) bitAnd: 16rFFFFFFFF).
			totalB := ThirtyTwoBitRegister
				fromInteger: ((seedInteger bitShift: -96) bitAnd: 16rFFFFFFFF).
			totalC := ThirtyTwoBitRegister
				fromInteger: ((seedInteger bitShift: -64) bitAnd: 16rFFFFFFFF).
			totalD := ThirtyTwoBitRegister
				fromInteger: ((seedInteger bitShift: -32) bitAnd: 16rFFFFFFFF).
			totalE := ThirtyTwoBitRegister
				fromInteger: (seedInteger bitAnd: 16rFFFFFFFF) ]
		ifNotNil: [
			totals
				at: 1 put: ((seedInteger bitShift: -128) bitAnd: 16rFFFFFFFF);
				at: 2 put: ((seedInteger bitShift: -96) bitAnd: 16rFFFFFFFF);
				at: 3 put: ((seedInteger bitShift: -64) bitAnd: 16rFFFFFFFF);
				at: 4 put: ((seedInteger bitShift: -32) bitAnd: 16rFFFFFFFF);
				at: 5 put: (seedInteger bitAnd: 16rFFFFFFFF) ].
	"pad integer with zeros"
	buffer := ByteArray new: 64.
	dstIndex := 0.
	aPositiveInteger digitLength to: 1 by: -1 do: [:i |
		buffer at: (dstIndex := dstIndex + 1) put: (aPositiveInteger digitAt: i)].

	"process that one block"
	self processBuffer: buffer.

	^ self finalHash
! !

!SecureHashAlgorithm methodsFor: 'public' stamp: 'jm 12/14/1999 11:28'!
hashMessage: aStringOrByteArray
	"Hash the given message using the Secure Hash Algorithm."

	^ self hashStream: (ReadStream on: aStringOrByteArray asByteArray)
! !

!SecureHashAlgorithm methodsFor: 'public' stamp: 'ul 10/26/2015 03:54'!
hashStream: aPositionableStream
	"Hash the contents of the given stream from the current position to the end using the Secure Hash Algorithm. The SHA algorithm is defined in FIPS PUB 180-1. It is also described on p. 442 of 'Applied Cryptography: Protocols, Algorithms, and Source Code in C' by Bruce Scheier, Wiley, 1996."
	"SecureHashAlgorithm new hashStream: (ReadStream on: 'foo')"

	| startPosition buf bitLength |
	self initializeTotals.

	"(SecureHashAlgorithm new hashMessage: '') radix: 16 	
	=> 'DA39A3EE5E6B4B0D3255BFEF95601890AFD80709'"
	aPositionableStream atEnd ifTrue: [self processFinalBuffer: #[] bitLength: 0].

	startPosition := aPositionableStream position.
	buf := ByteArray new: 64.
	[aPositionableStream atEnd] whileFalse: [
		buf := aPositionableStream next: 64 into: buf startingAt: 1.
		(aPositionableStream atEnd not and: [buf size = 64])
			ifTrue: [self processBuffer: buf]
			ifFalse: [
				bitLength := (aPositionableStream position - startPosition) * 8.
				self processFinalBuffer: buf bitLength: bitLength]].

	^ self finalHash
! !

!SecureHashAlgorithm methodsFor: 'initialize-release' stamp: 'ul 10/27/2015 01:17'!
initialize

	self primHasSecureHashPrimitive
		ifTrue: [ totals := Bitmap new: 5 ]
		ifFalse: [
			totalA := ThirtyTwoBitRegister new.
			totalB := ThirtyTwoBitRegister new.
			totalC := ThirtyTwoBitRegister new.
			totalD := ThirtyTwoBitRegister new.
			totalE := ThirtyTwoBitRegister new ]! !

!SecureHashAlgorithm methodsFor: 'private' stamp: 'ul 10/26/2015 02:05'!
initializeTotals
	"Initialize totalA through totalE to their seed values."

	totals
		ifNil: [
			"total registers for use when primitives are absent"
			totalA loadFrom: TA.
			totalB loadFrom: TB.
			totalC loadFrom: TC.
			totalD loadFrom: TD.
			totalE loadFrom: TE ]
		ifNotNil: [
			totals
				replaceFrom: 1
				to: totals size
				with: TI
				startingAt: 1 ]! !

!SecureHashAlgorithm methodsFor: 'primitives' stamp: 'jm 12/21/1999 20:11'!
primExpandBlock: aByteArray into: wordBitmap
	"Expand the given 64-byte buffer into the given Bitmap of length 80."

	<primitive: 'primitiveExpandBlock' module: 'DSAPrims'>
	^ self primitiveFailed
! !

!SecureHashAlgorithm methodsFor: 'primitives' stamp: 'jm 12/21/1999 20:13'!
primHashBlock: blockBitmap using: workingTotalsBitmap
	"Hash the given block (a Bitmap) of 80 32-bit words, using the given workingTotals."

	<primitive: 'primitiveHashBlock' module: 'DSAPrims'>
	^ self primitiveFailed
! !

!SecureHashAlgorithm methodsFor: 'primitives' stamp: 'jm 12/21/1999 22:58'!
primHasSecureHashPrimitive
	"Answer true if this platform has primitive support for the Secure Hash Algorithm."

	<primitive: 'primitiveHasSecureHashPrimitive' module: 'DSAPrims'>
	^ false
! !

!SecureHashAlgorithm methodsFor: 'private' stamp: 'jm 12/21/1999 23:32'!
processBufferUsingPrimitives: aByteArray
	"Process given 64-byte buffer using the primitives, accumulating the results in totals."

	| w |
	"expand and process the buffer"
	w := Bitmap new: 80.
	self primExpandBlock: aByteArray into: w.
	self primHashBlock: w using: totals.
! !

!SecureHashAlgorithm methodsFor: 'private' stamp: 'ul 10/27/2015 01:31'!
processBuffer: aByteArray
	"Process given 64-byte buffer, accumulating the results in totalA through totalE."

	| a b c d e t tmp w tmp2 tmp3 |
	totals ifNotNil: [ ^self processBufferUsingPrimitives: aByteArray ].

	"initialize registers a through e from the current totals" 
	a := totalA copy.
	b := totalB copy.
	c := totalC copy.
	d := totalD copy.
	e := totalE copy.

	"expand and process the buffer"
	w := self expandedBlock: aByteArray.
	tmp := ThirtyTwoBitRegister new.
	tmp2 := ThirtyTwoBitRegister new.
	tmp3 := ThirtyTwoBitRegister new.
	1 to: 80 do: [:i |
		tmp
			loadFrom: a;
			leftRotateBy: 5;
			+= (self hashFunction: i of: b with: c with: d using: tmp2 and: tmp3);
			+= e;
			+= (w at: i);
			+= (self constantForStep: i).
		t := e.
		e := d.
		d := c.
		c := b leftRotateBy: 30.
		b := a.
		a := tmp.
		tmp := t ].

	"add a through e into total accumulators"
	totalA += a.
	totalB += b.
	totalC += c.
	totalD += d.
	totalE += e.
! !

!SecureHashAlgorithm methodsFor: 'private' stamp: 'jm 12/14/1999 11:40'!
processFinalBuffer: buffer bitLength: bitLength
	"Process given buffer, whose length may be <= 64 bytes, accumulating the results in totalA through totalE. Also process the final padding bits and length."

	| out |
	out := ByteArray new: 64.
	out replaceFrom: 1 to: buffer size with: buffer startingAt: 1.
	buffer size < 56 ifTrue: [  "padding and length fit in last data block"
		out at: buffer size + 1 put: 128.  "trailing one bit"
		self storeLength: bitLength in: out.  "end with length"
		self processBuffer: out.
		^ self].

	"process the final data block"
	buffer size < 64 ifTrue: [
		out at: buffer size + 1 put: 128].  "trailing one bit"
	self processBuffer: out.

	"process one additional block of padding ending with the length"
	out := ByteArray new: 64.  "filled with zeros"
	buffer size = 64 ifTrue: [
		"add trailing one bit that didn't fit in final data block"
		out at: 1 put: 128].
	self storeLength: bitLength in: out.
	self processBuffer: out.
! !

!SecureHashAlgorithm methodsFor: 'private' stamp: 'jm 12/14/1999 11:10'!
storeLength: bitLength in: aByteArray
	"Fill in the final 8 bytes of the given ByteArray with a 64-bit big-endian representation of the original message length in bits."

	| n i |
	n := bitLength.
	i := aByteArray size.
	[n > 0] whileTrue: [
		aByteArray at: i put: (n bitAnd: 16rFF).
		n := n bitShift: -8.
		i := i - 1].
! !

Object subclass: #SecurityManager
	instanceVariableNames: 'privateKeyPair trustedKeys keysFileName'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'System-Support'!

!SecurityManager class methodsFor: 'accessing' stamp: 'nk 7/30/2004 21:50'!
default
	^Default ifNil: [Default := self new]! !

!SecurityManager class methodsFor: 'class initialization' stamp: 'nk 7/30/2004 21:50'!
initialize
	"SecurityManager initialize"

	"Order: ExternalSettings, SecurityManager, AutoStart"

	Default := self new.
	Smalltalk addToStartUpList: self after: ExternalSettings.
	Smalltalk addToShutDownList: self! !

!SecurityManager class methodsFor: 'class initialization' stamp: 'ar 2/6/2001 16:46'!
shutDown
	self default shutDown.! !

!SecurityManager class methodsFor: 'class initialization' stamp: 'ar 2/6/2001 17:12'!
startUp
	self default startUp.! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:19'!
addTrustedKey: aPublicKey
	"Add a public key to the list of trusted keys"
	trustedKeys := (trustedKeys copyWithout: aPublicKey) copyWith: aPublicKey.! !

!SecurityManager methodsFor: 'testing' stamp: 'ar 2/6/2001 16:14'!
canWriteImage
	"SecurityManager default canWriteImage"
	"Primitive. Return true if the right to write an image hasn't been revoked."
	<primitive: 'primitiveCanWriteImage' module: 'SecurityPlugin'>
	^true "assume so unless otherwise proven"! !

!SecurityManager methodsFor: 'security operations' stamp: 'ar 2/6/2001 16:14'!
disableFileAccess
	"SecurityManager default disableFileAccess"
	"Primitive. Disable unlimited access to files.
	Cannot be revoked from the image."
	<primitive: 'primitiveDisableFileAccess' module: 'FilePlugin'>
	^self primitiveFailed! !

!SecurityManager methodsFor: 'security operations' stamp: 'ar 2/6/2001 16:15'!
disableImageWrite
	"SecurityManager default disableImageWrite"
	"Primitive. Disable writing to an image file.
	Cannot be revoked from the image."
	<primitive: 'primitiveDisableImageWrite' module: 'SecurityPlugin'>
	^self primitiveFailed! !

!SecurityManager methodsFor: 'security operations' stamp: 'ar 2/6/2001 16:15'!
disableSocketAccess
	"SecurityManage default disableSocketAccess"
	"Primitive. Disable access to sockets.
	Cannot be revoked from the image."
	<primitive: 'primitiveDisableSocketAccess' module: 'SocketPlugin'>
	^self primitiveFailed! !

!SecurityManager methodsFor: 'security operations' stamp: 'rbb 2/18/2005 14:27'!
enterRestrictedMode
	"Some insecure contents was encountered. Close all doors and proceed."
	self isInRestrictedMode ifTrue:[^true].
	Preferences securityChecksEnabled ifFalse:[^true]. "it's been your choice..."
	Preferences warnAboutInsecureContent ifTrue:[
		( UIManager default chooseFrom: #('Load it anyways' 'Do not load it')
			title: 
'You are about to load some insecure content.
If you continue, access to files as well as
some other capabilities will be limited.')
			 = 1 ifFalse:[
				"user doesn't really want it"
				^false.
			].
	].
	"here goes the actual restriction"
	self flushSecurityKeys.
	self disableFileAccess.
	self disableImageWrite.
	"self disableSocketAccess."
	FileDirectory setDefaultDirectory: self untrustedUserDirectory.
	^true
! !

!SecurityManager methodsFor: 'security operations' stamp: 'sd 1/30/2004 15:22'!
fileInObjectAndCode: aStream
	| trusted |
	trusted := self positionToSecureContentsOf: aStream.
	trusted ifFalse:[self enterRestrictedMode ifFalse:[
		aStream close.
		^nil]].
	^aStream fileInObjectAndCode! !

!SecurityManager methodsFor: 'initialize-release' stamp: 'ar 2/6/2001 16:23'!
flushSecurityKeys
	"Flush all keys"
	privateKeyPair ifNotNil:[
		self flushSecurityKey: privateKeyPair first.
		self flushSecurityKey: privateKeyPair last.
	].
	privateKeyPair := nil.
	trustedKeys do:[:key| self flushSecurityKey: key].
	trustedKeys := #().! !

!SecurityManager methodsFor: 'initialize-release' stamp: 'ar 2/6/2001 16:24'!
flushSecurityKey: aKey
	"Flush a security key"
	| n |
	n := aKey first.
	1 to: n basicSize do:[:i| n basicAt: i put: 0].
	n := aKey second.
	1 to: n basicSize do:[:i| n basicAt: i put: 0].
! !

!SecurityManager methodsFor: 'private' stamp: 'fbs 6/4/2013 17:25'!
generateKeyPairInBackground
	"SecurityManager default generateKeyPairInBackground"
	"Silently generate a key set on the local machine while running in the background."
	| guesstimate startTime |
	guesstimate := [10 benchmark] timeToRun * 150.
	startTime := Time millisecondClockValue.
	privateKeyPair := nil.
	[self generateLocalKeyPair] fork.
	UIManager default informUserDuring:[:bar|
		[privateKeyPair == nil] whileTrue:[
			bar value:'Initializing Squeak security system (', (Time millisecondClockValue - startTime * 100 // guesstimate) printString,'%)'.
			(Delay forSeconds: 1) wait.
		].
	].
! !

!SecurityManager methodsFor: 'private' stamp: 'fbs 7/25/2013 07:16'!
generateLocalKeyPair
	"SecurityManager default generateLocalKeyPair"
	"Generate a key set on the local machine."
	| dsa |
	dsa := DigitalSignatureAlgorithm new.
	dsa initRandomFromString: 
		Time millisecondClockValue printString, 
		Date today printString, 
		Smalltalk platformName printString.
	privateKeyPair := dsa generateKeySet.
	self storeSecurityKeys.! !

!SecurityManager methodsFor: 'testing' stamp: 'ar 2/6/2001 16:16'!
hasFileAccess
	"SecurityManager default hasFileAccess"
	"Return true if the right to access arbitrary files hasn't been revoked"
	<primitive: 'primitiveHasFileAccess' module: 'FilePlugin'>
	^true "assume so unless otherwise proven"! !

!SecurityManager methodsFor: 'testing' stamp: 'ar 2/6/2001 16:16'!
hasSocketAccess
	"SecurityManager default hasSocketAccess"
	"Return true if the right to access sockets hasn't been revoked"
	<primitive: 'primitiveHasSocketAccess' module: 'SocketPlugin'>
	^true "assume so unless otherwise proven"! !

!SecurityManager methodsFor: 'initialize-release' stamp: 'ar 2/6/2001 16:20'!
initialize
	privateKeyPair := nil.
	trustedKeys := #().
	keysFileName := 'Squeak.keys'.! !

!SecurityManager methodsFor: 'testing' stamp: 'ar 2/6/2001 16:13'!
isInRestrictedMode
	"Return true if we're in restricted mode"
	^(self canWriteImage 
		or:[self hasFileAccess 
		"or:[self hasSocketAccess]"]) not! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:17'!
keysFileName
	^keysFileName! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:17'!
keysFileName: aFileName
	keysFileName := aFileName! !

!SecurityManager methodsFor: 'fileIn/out' stamp: 'ar 3/5/2001 01:55'!
loadSecurityKeys
	"SecurityManager default loadSecurityKeys"
	"Load the keys file for the current user"
	| fd loc file keys |
	self isInRestrictedMode ifTrue:[^self]. "no point in even trying"
	loc := self secureUserDirectory. "where to get it from"
	loc last = FileDirectory pathNameDelimiter ifFalse:[
		loc := loc copyWith: FileDirectory pathNameDelimiter.
	].
	fd := FileDirectory on: loc.
	file := [fd readOnlyFileNamed: keysFileName] 
			on: FileDoesNotExistException do:[:ex| nil].
	file ifNil:[^self]. "no keys file"
	keys := Object readFrom: file.
	privateKeyPair := keys first.
	trustedKeys := keys last.
	file close.! !

!SecurityManager methodsFor: 'security operations' stamp: 'nice 5/8/2011 12:08'!
positionToSecureContentsOf: aStream
	| bytes trusted part1 part2 sig hash dsa okay pos |
	aStream binary.
	pos := aStream position.
	bytes := aStream next: 4.
	bytes = 'SPRJ' asByteArray ifFalse:[
		"was not signed"
		aStream position: pos.
		^false].
	part1 := (aStream nextInto: (LargePositiveInteger basicNew: 20)) normalize.
	part2 := (aStream nextInto: (LargePositiveInteger basicNew: 20)) normalize.
	sig := Array with: part1 with: part2.
	hash := SecureHashAlgorithm new hashStream: aStream.
	dsa := DigitalSignatureAlgorithm new.
	trusted := self trustedKeys.
	okay := trusted anySatisfy: [:key| dsa verifySignature: sig ofMessageHash: hash publicKey: key].
	aStream position: pos+44.
	^okay! !

!SecurityManager methodsFor: 'accessing' stamp: 'tak 3/15/2005 00:45'!
primSecureUserDirectory
	<primitive: 'primitiveGetSecureUserDirectory' module: 'SecurityPlugin'>
	^ nil! !

!SecurityManager methodsFor: 'private' stamp: 'tak 12/17/2004 14:15'!
primUntrustedUserDirectory
	"Primitive. Return the untrusted user directory that is the root directory for files that are visible even in restricted mode."
	<primitive: 'primitiveGetUntrustedUserDirectory' module: 'SecurityPlugin'>
	^ nil! !

!SecurityManager methodsFor: 'private' stamp: 'mir 11/10/2003 16:14'!
printStateOn: stream
	"Print the current state of myself onto stream.
	Used to gather information in the debug log."

	stream
		nextPutAll: 'SecurityManager state:'; cr;
		nextPutAll: 'Restricted: '; nextPutAll: self isInRestrictedMode asString; cr;
		nextPutAll: 'FileAccess: '; nextPutAll: self hasFileAccess asString; cr;
		nextPutAll: 'SocketAccess: '; nextPutAll: self hasSocketAccess asString; cr;
		nextPutAll: 'Working Dir '; nextPutAll: FileDirectory default pathName asString; cr;
		nextPutAll: 'Trusted Dir '; nextPutAll: self secureUserDirectory asString; cr;
		nextPutAll: 'Untrusted Dir '; nextPutAll: self untrustedUserDirectory asString; cr;
		cr! !

!SecurityManager methodsFor: 'accessing' stamp: 'tak 3/15/2005 00:46'!
secureUserDirectory
	"SecurityManager default secureUserDirectory"
	| dir |
	dir := self primSecureUserDirectory.
	^ dir
		ifNil: [FileDirectory default pathName]
		ifNotNil: [(FilePath pathName: dir isEncoded: true) asSqueakPathName]! !

!SecurityManager methodsFor: 'initialize-release' stamp: 'ar 2/6/2001 16:22'!
shutDown
	"Flush existing keys"
	self flushSecurityKeys.! !

!SecurityManager methodsFor: 'security operations' stamp: 'RAA 3/2/2002 14:32'!
signFile: fileName directory: fileDirectory
	"Sign the given project in the directory"
	| bytes file dsa hash sig key |
	Preferences signProjectFiles ifFalse:[^self]. "signing turned off"
	key := self signingKey.
	key ifNil:[^self].
	file := FileStream readOnlyFileNamed: (fileDirectory fullNameFor: fileName).
	bytes := file binary; contentsOfEntireFile.
	fileDirectory deleteFileNamed: fileName ifAbsent:[].
	dsa := DigitalSignatureAlgorithm new.
	dsa initRandom: Time millisecondClockValue + Date today julianDayNumber.
	hash := SecureHashAlgorithm new hashStream: (ReadStream on: bytes).
	sig := dsa computeSignatureForMessageHash: hash privateKey: key.
	file := FileStream newFileNamed: (fileDirectory fullNameFor: fileName).
	file binary.
	"store a header identifying the signed file first"
	file nextPutAll: 'SPRJ' asByteArray.
	"now the signature"
	file 
		nextPutAll: (sig first withAtLeastNDigits: 20); 
		nextPutAll: (sig last withAtLeastNDigits: 20).
	"now the contents"
	file nextPutAll: bytes.
	file close.! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:20'!
signingKey
	"Return the key used for signing projects"
	^privateKeyPair ifNotNil:[privateKeyPair first]! !

!SecurityManager methodsFor: 'initialize-release' stamp: 'ar 2/6/2001 18:28'!
startUp
	"Attempt to load existing keys"
	self loadSecurityKeys.
	(privateKeyPair == nil 
		and:[self isInRestrictedMode not
		and:[Preferences automaticKeyGeneration]]) ifTrue:[
			self generateKeyPairInBackground.
	].! !

!SecurityManager methodsFor: 'fileIn/out' stamp: 'hg 9/29/2001 14:35'!
storeSecurityKeys
	"SecurityManager default storeSecurityKeys"
	"Store the keys file for the current user"
	| fd loc file |
	self isInRestrictedMode ifTrue:[^self]. "no point in even trying"
	loc := self secureUserDirectory. "where to put it"
	loc last = FileDirectory pathNameDelimiter ifFalse:[
		loc := loc copyWith: FileDirectory pathNameDelimiter.
	].
	fd := FileDirectory on: loc.
	fd assureExistence.
	fd deleteFileNamed: self keysFileName ifAbsent:[].
	file := fd newFileNamed: self keysFileName.
	{privateKeyPair. trustedKeys} storeOn: file.
	file close.! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:42'!
trustedKeys
	"Return an array of trusted public keys for verifying some project"
	privateKeyPair ifNil:[^trustedKeys].
	^{privateKeyPair second}, trustedKeys! !

!SecurityManager methodsFor: 'accessing' stamp: 'tak 12/17/2004 14:19'!
untrustedUserDirectory
	"SecurityManager default untrustedUserDirectory"
	| dir |
	dir := self primUntrustedUserDirectory.
	^ dir
		ifNil: [FileDirectory default pathName]
		ifNotNil: [(FilePath pathName: dir isEncoded: true) asSqueakPathName]! !

Object subclass: #SerialPort
	instanceVariableNames: 'port baudRate stopBitsType parityType dataBits outputFlowControlType inputFlowControlType xOnByte xOffByte'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Serial Port'!

!SerialPort commentStamp: '<historical>' prior: 0!
This class supports a simple interface to the serial ports of the underlying platform, if it supports serial ports. The mapping of port numbers to hardware ports is platform specific, but typically follows platform ordering conventions. For example, on the Macintosh, port 0 is the modem port and port 1 is the printer port, since in the programmers documentation these ports are referred to as ports A and B.
!

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:19'!
baudRate

	^ baudRate
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:29'!
baudRate: anInteger
	"Set the baud rate for this serial port."

	baudRate := anInteger.
! !

!SerialPort methodsFor: 'open/close' stamp: 'Sk 1/4/2009 18:29'!
close
	"Close the serial port. Do nothing if the port is not open."
	port
		ifNotNil: [port isString ifTrue:[self primClosePortByName: port]
 ifFalse:[self primClosePort:port]].
	port := nil! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:19'!
dataBits

	^ dataBits
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:29'!
dataBits: anInteger
	"Set the number of data bits for this serial port to 5, 6, 7, or 8."

	dataBits := anInteger.
! !

!SerialPort methodsFor: 'initialization' stamp: 'jm 5/5/1998 15:49'!
initialize
	"Default port settings."

	port := nil.					"set when opened"
	baudRate := 9600.			"9600 baud"
	stopBitsType := 1.				"one stop bit"
	parityType := 0.				"no parity"
	dataBits := 8.					"8 bits"
	outputFlowControlType := 0.	"none"
	inputFlowControlType := 0.	"none"
	xOnByte := 19.				"ctrl-S"
	xOffByte := 24.				"ctrl-X"
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:21'!
inputFlowControlType

	^ inputFlowControlType
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:38'!
inputFlowControlType: anInteger
	"Set the type of input flow control, where:
		0 - none
		1 - XOn/XOff
		2 - hardware handshaking"

	inputFlowControlType := anInteger.
! !

!SerialPort methodsFor: 'input/output' stamp: 'Sk 1/4/2009 18:35'!
nextPutAll: aStringOrByteArray 
	"Send the given bytes out this serial port. The port must be
	open. "
	^ port isString
		ifTrue: [self
		primWritePortByName: port
		from: aStringOrByteArray
		startingAt: 1
		count: aStringOrByteArray size]
		ifFalse: [self
		primWritePort: port
		from: aStringOrByteArray
		startingAt: 1
		count: aStringOrByteArray size]
! !

!SerialPort methodsFor: 'open/close' stamp: 'ml 3/8/2011 19:40'!
openPort: portId 
	"Open the given serial port, using the settings specified by my
	instance variables."
	| p |
	self close.
	p := portId isString
		ifTrue: [self
				primOpenPortByName: portId
				baudRate: baudRate
				stopBitsType: stopBitsType
				parityType: parityType
				dataBits: dataBits
				inFlowControlType: inputFlowControlType
				outFlowControlType: outputFlowControlType
				xOnByte: xOnByte
				xOffByte: xOffByte]
		ifFalse: [self
				primOpenPort: portId
				baudRate: baudRate
				stopBitsType: stopBitsType
				parityType: parityType
				dataBits: dataBits
				inFlowControlType: inputFlowControlType
				outFlowControlType: outputFlowControlType
				xOnByte: xOnByte
				xOffByte: xOffByte].
	^p ifNotNil: [
		port := portId.
		self]! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:19'!
outputFlowControlType

	^ outputFlowControlType
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:38'!
outputFlowControlType: anInteger
	"Set the type of output flow control, where:
		0 - none
		1 - XOn/XOff
		2 - hardware handshaking"

	outputFlowControlType := anInteger.
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:19'!
parityType

	^ parityType
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:29'!
parityType: anInteger
	"Set the parity type for this serial port, where:
		0 - no parity
		1 - odd parity
		2 - even parity"

	parityType := anInteger.
! !

!SerialPort methodsFor: 'primitives' stamp: 'Sk 1/4/2009 18:24'!
primClosePortByName: portName
	<primitive: 'primitiveSerialPortCloseByName' module:'SerialPlugin'>
	^ nil"(DNS)"
	"self primitiveFailed."! !

!SerialPort methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primClosePort: portNumber

	<primitive: 'primitiveSerialPortClose' module: 'SerialPlugin'>
	^ nil  "(DNS)"
	"self primitiveFailed."
! !

!SerialPort methodsFor: 'primitives' stamp: 'Sk 1/4/2009 18:24'!
primOpenPortByName: portName baudRate: baud stopBitsType: stop parityType: parity dataBits: numDataBits inFlowControlType: inFlowCtrl outFlowControlType: outFlowCtrl xOnByte: xOn xOffByte: xOff 
	<primitive: 'primitiveSerialPortOpenByName' module:'SerialPlugin'>
	^ nil"(DNS)"! !

!SerialPort methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primOpenPort: portNumber baudRate: baud stopBitsType: stop
	parityType: parity dataBits: numDataBits
	inFlowControlType: inFlowCtrl outFlowControlType: outFlowCtrl
	xOnByte: xOn xOffByte: xOff

	<primitive: 'primitiveSerialPortOpen' module: 'SerialPlugin'>
	^ nil  "(DNS)"
! !

!SerialPort methodsFor: 'primitives' stamp: 'Sk 1/4/2009 18:24'!
primReadPortByName: portName into: byteArray startingAt: startIndex count: count 
	<primitive: 'primitiveSerialPortReadByName' module:'SerialPlugin'>
	self primitiveFailed! !

!SerialPort methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primReadPort: portNumber into: byteArray startingAt: startIndex count: count

	<primitive: 'primitiveSerialPortRead' module: 'SerialPlugin'>
	self primitiveFailed.
! !

!SerialPort methodsFor: 'primitives' stamp: 'Sk 1/4/2009 18:25'!
primWritePortByName: portName from: byteArray startingAt: startIndex count: count 
	<primitive: 'primitiveSerialPortWriteByName' module:'SerialPlugin'>
	self primitiveFailed! !

!SerialPort methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primWritePort: portNumber from: byteArray startingAt: startIndex count: count

	<primitive: 'primitiveSerialPortWrite' module: 'SerialPlugin'>
	self primitiveFailed.
! !

!SerialPort methodsFor: 'printing' stamp: 'jm 5/1/1998 18:02'!
printOn: aStream

	aStream
		nextPutAll: 'SerialPort(';
		nextPutAll:
			(port ifNil: ['closed'] ifNotNil: ['#', port printString]);
		nextPutAll: ', ';
		print: baudRate; nextPutAll: ' baud, ';
		print: dataBits; nextPutAll: ' bits, ';
		nextPutAll: (#('1.5' '1' '2') at: stopBitsType + 1); nextPutAll: ' stopbits, ';
		nextPutAll: (#('no' 'odd' 'even') at: parityType + 1); nextPutAll: ' parity)'.
! !

!SerialPort methodsFor: 'input/output' stamp: 'Sk 1/4/2009 18:33'!
readByteArray
	"Answer a ByteArray read from this serial port. Answer an
	empty ByteArray if no data is available. The port must be
	open. "
	| buf count |
	buf := ByteArray new: 1000.
	count := port isString
				ifTrue: [self
						primReadPortByName: port
						into: buf
						startingAt: 1
						count: buf size]
				ifFalse: [self
						primReadPort: port
						into: buf
						startingAt: 1
						count: buf size].
	^ buf copyFrom: 1 to: count! !

!SerialPort methodsFor: 'input/output' stamp: 'Sk 1/4/2009 18:34'!
readInto: aStringOrByteArray startingAt: index 
	"Read data into the given String or ByteArray object starting at
	the given index, and answer the number of bytes read. Does
	not go past the end of the given String or ByteArray."
	^ port isString
				ifTrue: [self primReadPortByName: port
		into: aStringOrByteArray
		startingAt: index
		count: aStringOrByteArray size - index + 1]
				ifFalse: [self primReadPort: port
		into: aStringOrByteArray
		startingAt: index
		count: aStringOrByteArray size - index + 1].
		! !

!SerialPort methodsFor: 'input/output' stamp: 'Sk 1/4/2009 18:32'!
readString
	"Answer a String read from this serial port. Answer the empty
	String if no data is available. The port must be open."
	| buf count |
	buf := String new: 1000.
	count := port isString ifTrue:[self
				primReadPortByName: port
				into: buf
				startingAt: 1
				count: buf size.]
ifFalse:[self
				primReadPort: port
				into: buf
				startingAt: 1
				count: buf size.].
	^ buf copyFrom: 1 to: count! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:19'!
stopBitsType

	^ stopBitsType
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 18:02'!
stopBitsType: anInteger
	"Set the stop bits type for this serial port, where:
		0 - 1.5 stop bits
		1 - one stop bit
		2 - two stop bits"

	stopBitsType := anInteger.
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:20'!
xOffByte

	^ xOffByte
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:28'!
xOffByte: anInteger
	"Set the value of the XOff byte to be used if XOn/XOff flow control is enabled."

	xOffByte := anInteger.
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:20'!
xOnByte

	^ xOnByte
! !

!SerialPort methodsFor: 'settings' stamp: 'jm 5/1/1998 17:28'!
xOnByte: anInteger
	"Set the value of the XOn byte to be used if XOn/XOff flow control is enabled."

	xOnByte := anInteger.
! !

Object subclass: #SmalltalkImage
	instanceVariableNames: 'globals'
	classVariableNames: 'EndianCache LastImageName LastQuitLogPosition LastStats LowSpaceProcess LowSpaceSemaphore MemoryHogs ShutDownList SourceFileVersionString SpecialSelectors StartUpList StartupStamp WordSize'
	poolDictionaries: ''
	category: 'System-Support'!

!SmalltalkImage commentStamp: 'dtl 3/6/2010 14:00' prior: 0!
I represent the current image and runtime environment, including system organization, the virtual machine, object memory, plugins and source files. My instance variable #globals is a reference to the system dictionary of global variables and class names.

My singleton instance is called Smalltalk.!

!SmalltalkImage class methodsFor: 'class initialization' stamp: 'cwp 1/18/2014 10:23'!
cleanUp
	"Flush caches"

	Smalltalk removeObsoleteClassesFromCompactClassesArray! !

!SmalltalkImage class methodsFor: 'instance creation' stamp: 'ar 3/5/2010 21:52'!
current
	"Deprecated. Use Smalltalk instead."
	^Smalltalk! !

!SmalltalkImage class methodsFor: 'class initialization' stamp: 'ar 3/5/2010 21:52'!
initialize
	"SmalltalkImage initialize"

	self initializeStartUpList.
	self initializeShutDownList.
! !

!SmalltalkImage class methodsFor: 'class initialization' stamp: 'dtl 1/30/2016 14:33'!
initializeShutDownList
	"SmalltalkImage initialize"

	| oldList |
	oldList := ShutDownList.
	ShutDownList := OrderedCollection new.
	"These get processed from the bottom up..."
	#(
		Delay
		DisplayScreen
		EventSensor
		Form
		ControlManager
		PasteUpMorph
		StrikeFont
		Color
		FileDirectory
		SoundPlayer
		HttpUrl
		Password
		PWS
		MailDB
		ImageSegment
	) do:[:clsName|
		Smalltalk at: clsName ifPresent:[:cls| Smalltalk addToShutDownList: cls].
	].

	oldList ifNotNil: [oldList reverseDo: [:className | Smalltalk at: className
						ifPresent: [:theClass | Smalltalk addToShutDownList: theClass]]].
! !

!SmalltalkImage class methodsFor: 'class initialization' stamp: 'dtl 1/30/2016 14:33'!
initializeStartUpList
	"SmalltalkImage initialize"
	
	| oldList |
	oldList := StartUpList.
	StartUpList := OrderedCollection new.
	"These get processed from the top down..."
	self flag: #'revisit in Squeak 5.3'.
	#(
		SmallInteger
		Delay
		DisplayScreen
		Cursor
		EventSensor
		ProcessorScheduler  "Starts low space watcher and bkground."
		FileDirectory  "Enables file stack dump and opens sources."
		ShortIntegerArray
		ShortRunArray
		CrLfFileStream "Remove this in Squeak 5.3"
	) do:[:clsName|
		Smalltalk at: clsName ifPresent:[:cls| Smalltalk addToStartUpList: cls].
	].
	oldList ifNotNil: [oldList do: [:className | Smalltalk at: className
						ifPresent: [:theClass | Smalltalk addToStartUpList: theClass]]].
	#(
		ImageSegment
		PasteUpMorph
		ControlManager
	) do:[:clsName|
		Smalltalk at: clsName ifPresent:[:cls| Smalltalk addToStartUpList: cls].
	].
! !

!SmalltalkImage class methodsFor: 'instance creation' stamp: 'ar 3/5/2010 21:52'!
new

	self error: 'Use Smalltalk'.! !

!SmalltalkImage class methodsFor: 'class initialization' stamp: 'eem 1/25/2016 19:24'!
startUp: resuming
	resuming ifTrue:
		[LastStats := nil.
		 SystemChangeNotifier uniqueInstance notify: Smalltalk ofAllSystemChangesUsing: #event:]! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'ul 6/17/2011 12:29'!
abandonSources
	"Smalltalk abandonSources"
	"Replaces every method by a copy with the 4-byte source pointer 
	 replaced by a string of all arg and temp names, followed by its
	 length. These names can then be used to inform the decompiler."
	"wod 11/3/1998: zap the organization before rather than after
	 condensing changes."
	"eem 7/1/2009 13:59 update for the closure schematic temp names regime"
	| oldMethods newMethods bTotal bCount |
	(self confirm: 'This method will preserve most temp names
(up to about 15k characters of temporaries)
while allowing the sources file to be discarded.
-- CAUTION --
If you have backed up your system and
are prepared to face the consequences of
abandoning source code files, choose Yes.
If you have any doubts, you may choose No
to back out with no harm done.')
			== true
		ifFalse: [^ self inform: 'Okay - no harm done'].
	oldMethods := OrderedCollection new: CompiledMethod instanceCount.
	newMethods := OrderedCollection new: CompiledMethod instanceCount.
	bTotal := 0.
	bCount := 0.
	self systemNavigation allBehaviorsDo: [:b | bTotal := bTotal + 1].
	'Saving temp names for better decompilation...'
		displayProgressFrom: 0
		to: bTotal
		during:
			[:bar |
			self systemNavigation allBehaviorsDo:
				[:cl |  "for test: (Array with: Arc with: Arc class) do:"
				bar value: (bCount := bCount + 1).
				cl selectorsAndMethodsDo:
					[:selector :m |
					| oldCodeString methodNode |
					m fileIndex > 0 ifTrue:
						[oldCodeString := cl sourceCodeAt: selector.
						methodNode := cl newCompiler
											parse: oldCodeString
											in: cl
											notifying: nil.
						oldMethods addLast: m.
						newMethods addLast: (m copyWithTempsFromMethodNode: methodNode)]]]].
	oldMethods asArray elementsExchangeIdentityWith: newMethods asArray.
	self systemNavigation allBehaviorsDo: [:b | b zapOrganization].
	self condenseChanges.
	Preferences disable: #warnIfNoSourcesFile! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'fbs 7/20/2013 23:29'!
abandonTempNames
	"Replaces every method by a copy with no source pointer or
	encoded temp names."
	"Smalltalk abandonTempNames"
	| continue oldMethods newMethods n |
	continue := self confirm: '-- CAUTION --
If you have backed up your system and
are prepared to face the consequences of
abandoning all source code, hit Yes.
If you have any doubts, hit No,
to back out with no harm done.'.
	continue
		ifFalse: [^ self inform: 'Okay - no harm done'].
	self garbageCollect.
	oldMethods := OrderedCollection new.
	newMethods := OrderedCollection new.
	n := 0.
	'Removing temp names to save space...'
		displayProgressFrom: 0
		to: CompiledMethod instanceCount "This is just a rough guess."
		during: [:bar | self systemNavigation
			allBehaviorsDo: [:cl | cl methodsDo: [:m | 
				bar value: (n := n + 1).
				oldMethods addLast: m.
				newMethods
					addLast: (m copyWithTrailerBytes: CompiledMethodTrailer empty)]]].
	oldMethods asArray elementsExchangeIdentityWith: newMethods asArray.
	self closeSourceFiles.
	self flag: #shouldUseAEnsureBlockToBeSureThatTheFileIsClosed.
	"sd: 17 April 2003"
	Preferences disable: #warnIfNoChangesFile.
	Preferences disable: #warnIfNoSourcesFile! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'nice 5/6/2011 14:34'!
aboutThisSystem 
	"Identify software version"

	^self globals
		at: #SystemReporter
		ifPresent: [:sys | sys open]
		ifAbsent: [self inform: self systemInformationString withCRs]! !

!SmalltalkImage methodsFor: 'startup list' stamp: 'MarcusDenker 4/16/2011 23:07'!
addToShutDownList: aClass
	"This will add a ref to this class at the BEGINNING of the shutDown list."
	"No-op if already in the list."
	
	(ShutDownList includes: aClass name) ifFalse: [ShutDownList addFirst: aClass name]! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/3/1999 22:04'!
addToShutDownList: aClass after: predecessor

	self add: aClass toList: ShutDownList after: predecessor! !

!SmalltalkImage methodsFor: 'startup list' stamp: 'MarcusDenker 4/16/2011 22:36'!
addToShutDownList: aClass before: predecessor

	self add: aClass toList: ShutDownList before: predecessor! !

!SmalltalkImage methodsFor: 'startup list' stamp: 'MarcusDenker 4/16/2011 23:08'!
addToStartUpList: aClass
	"This will add a ref to this class at the END of the startUp list."
	"No-op if already in the list."

	(StartUpList includes: aClass name) ifFalse: [StartUpList addLast: aClass name]! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/3/1999 22:04'!
addToStartUpList: aClass after: predecessor

	self add: aClass toList: StartUpList after: predecessor! !

!SmalltalkImage methodsFor: 'startup list' stamp: 'MarcusDenker 4/16/2011 22:36'!
addToStartUpList: aClass before: predecessor

	self add: aClass toList: StartUpList before: predecessor! !

!SmalltalkImage methodsFor: 'startup list' stamp: 'MarcusDenker 4/16/2011 23:09'!
add: aClass toList: startUpOrShutDownList after: predecessor
	"Add the name of aClass to the startUp or shutDown list.
	Add it after the name of predecessor"

	(Smalltalk globals includes: aClass) 
		ifFalse: [self error: aClass name , ' cannot be found in Smalltalk dictionary.'].

	"Add after predecessor, moving it if already there."
	(Smalltalk globals includes: predecessor)  
		ifFalse: [self error: predecessor name , ' cannot be found in Smalltalk dictionary.'].
	(startUpOrShutDownList includes: predecessor name) 
		ifFalse: [self error: predecessor name , ' cannot be found in the list.'].
	startUpOrShutDownList remove: aClass name ifAbsent:[].
	startUpOrShutDownList add: aClass name after: predecessor name! !

!SmalltalkImage methodsFor: 'startup list' stamp: 'MarcusDenker 4/16/2011 23:09'!
add: aClass toList: startUpOrShutDownList before: successor
	"Add the name of aClass to the startUp or shutDown list.
	Add it before the name of successor"

	(Smalltalk globals includes: aClass) 
		ifFalse: [self error: aClass name , ' cannot be found in Smalltalk dictionary.'].
		
	"Add before successor, moving it if already there."
	(Smalltalk globals includes: successor)  
		ifFalse: [self error: successor name , ' cannot be found in Smalltalk dictionary.'].
	(startUpOrShutDownList includes: successor name) 
		ifFalse: [self error: successor name , ' cannot be found in the list.'].
	startUpOrShutDownList remove: aClass name ifAbsent: [].
	startUpOrShutDownList add: aClass name before: successor name.! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:13'!
allClasses  
	"Return all the class defined in the system"

	^globals allClasses! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:13'!
allClassesAndTraits
	"Return all the classes and traits defined in the system"

	^globals allClassesAndTraits! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:13'!
allClassesAndTraitsDo: aBlock
	"Evaluate the argument, aBlock, for each class and trait in the system."

	globals allClassesAndTraitsDo: aBlock! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:13'!
allClassesDo: aBlock
	"Evaluate the argument, aBlock, for each class in the system."

	globals allClassesDo: aBlock! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:14'!
allTraits
	"Return all traits defined in the system"

	^globals allTraits! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:14'!
allTraitsDo: aBlock
	"Evaluate the argument, aBlock, for each trait in the system."

	globals allTraitsDo: aBlock! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'cmm 3/14/2014 15:15'!
appendChangesTo: sourcesName

	"Condense changes to the end of the given sources file.
	If the file is the same as Smalltalk sourcesName, then just append
	the changes. If the file is different, then copy the sources file and
	append the changes afterwards."

	"Smalltalk appendChangesTo: 'test123.sources'"

	"To verify correctness of the operation run the following code:
	[	| sourceMap |
		sourceMap := Dictionary new.

		(CompiledMethod allInstances select:[:cm| cm isInstalled]) do:[:cm|
			sourceMap at: cm methodReference put: 
				(cm getSourceFor: cm selector in: cm methodClass)].
		Smalltalk allClassesAndTraitsDo:[:aClass|
			sourceMap at: aClass put: aClass comment].

		Smalltalk appendChangesTo: 'verify.sources'.

		(CompiledMethod allInstances select:[:cm| cm isInstalled]) do:[:cm|
			self assert: (sourceMap at: cm methodReference) =
				(cm getSourceFor: cm selector in: cm methodClass)].
		Smalltalk allClassesAndTraitsDo:[:aClass|
			self assert: (sourceMap at: aClass) = aClass comment].
	]"

	| fullName sourcesFile |
	fullName := FileDirectory default fullNameFor: sourcesName.
	(fullName endsWith: '.sources') ifFalse: [ fullName := fullName , '.sources' ].
	fullName = Smalltalk sourcesName ifFalse:[
		"Copy sources file; change file name accordingly"
		FileStream forceNewFileNamed: fullName do: [ :newFile |
			sourcesFile := (SourceFiles at: 1) readOnlyCopy.
			FileDirectory default copyFile: sourcesFile toFile: newFile.
			sourcesFile position: 0 ].
		self setMacFileInfoOn: fullName.
		"Change to the new sources file and reopen"
		self 
			closeSourceFiles ;
			sourceFileVersionString: ((FileDirectory localNameFor: fullName) allButLast: '.sources' size) ;
			openSourceFiles.
	].

	"We've copied the old to the new sources file; reopen the sources file read/write"
	sourcesFile := SourceFiles at: 1.
	sourcesFile close; open: sourcesFile fullName forWrite: true. "should be openReadWrite"
	sourcesFile setToEnd; timeStamp "remember when we did this".

	"Copy method sources from changes to sources"
	CompiledMethod allInstances do:[:method|
		(method isInstalled and:[method fileIndex = 2]) ifTrue:[
			| class selector category preamble changeList index chgRec string source |
			class := method methodClass.
			selector := method selector.
			source := class sourceCodeAt: selector.
			category := class organization categoryOfElement: selector.
			preamble := class name, ' methodsFor: ', category asString printString,
							' stamp: ', method timeStamp printString.

			"Find the last version in the sources file; link up the prior: version"
			changeList := ChangeSet scanVersionsOf: method 
				class: class meta: class isMeta category: category  selector: selector.
			index := changeList findLast:[:any| any fileIndex = 1].
			index > 0 ifTrue:[
				chgRec := changeList at: index.
				preamble := preamble, ' prior: ', (SourceFiles 
					sourcePointerFromFileIndex: chgRec fileIndex 
					andPosition: chgRec position) printString].

			"append to sources file"
			sourcesFile setToEnd; cr; nextPut: $!!; nextChunkPut: preamble; cr.
			string := RemoteString newString: source onFileNumber: 1 toFile: sourcesFile.
			sourcesFile nextChunkPut: ' '.
			method setSourcePosition: string position inFile: 1
		].
	] displayingProgress: 'Moving changes...'.

	"Copy class comments from changes to sources"
	self  allClassesAndTraitsDo: [:classOrTrait | 
		classOrTrait moveClassCommentTo: sourcesFile fileIndex: 1.
	].

	"We've moved everything; reopen the source files"
	self closeSourceFiles; openSourceFiles.

	"Former changes are now in .sources, no need for them in .changes.  Condense them."
	self condenseChanges! !

!SmalltalkImage methodsFor: 'command line' stamp: 'cmm 10/12/2013 17:44'!
argumentAt: argumentIndex 
	"Answer the i-th argument of the command line, or nil if not so many argument."
	^ self getSystemAttribute:
		argumentIndex +
			(Preferences readDocumentAtStartup
				ifTrue: [ 2 ]
				ifFalse: [ 1 ])! !

!SmalltalkImage methodsFor: 'command line' stamp: 'nice 3/7/2010 01:30'!
arguments
	"Answer an array with all the command line arguments.
	This does not include imagePath, documentPath nor any option."
	
	"Smalltalk commandLine arguments"
	
	^Array streamContents: [:str |
		| arg i |
		i := 1.
		[i > 998 or: [(arg := self argumentAt: i) == nil]]
			whileFalse:
				[str nextPut: arg.
				i := i + 1]].! !

!SmalltalkImage methodsFor: 'dictionary access' stamp: 'ar 3/5/2010 21:18'!
associationAt: aKey
	"DO NOT DEPRECATE - used by ImageSegments"

	^globals associationAt: aKey! !

!SmalltalkImage methodsFor: 'dictionary access' stamp: 'ar 3/5/2010 21:17'!
associationAt: key ifAbsent: aBlock
	"Obsoleted."

	self deprecated: 'Use Smalltalk globals'.
	^globals associationAt: key ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'dictionary access' stamp: 'cwp 1/18/2014 10:06'!
associationDeclareAt: aKey
	"DO NOT DEPRECATE - used by ImageSegments"
	self flag: #environments.
	^globals associationDeclareAt: aKey! !

!SmalltalkImage methodsFor: 'dictionary access' stamp: 'ar 3/5/2010 21:19'!
associationOrUndeclaredAt: aKey
	"DO NOT DEPRECATE - used by binary storage"

	^globals associationOrUndeclaredAt: aKey! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'NS 1/16/2004 15:34'!
assureStartupStampLogged
	"If there is a startup stamp not yet actually logged to disk, do it now."
	| changesFile |
	StartupStamp ifNil: [^ self].
	(SourceFiles isNil or: [(changesFile := SourceFiles at: 2) == nil]) ifTrue: [^ self].
	changesFile isReadOnly ifTrue:[^self].
	changesFile setToEnd; cr; cr.
	changesFile nextChunkPut: StartupStamp asString; cr.
	StartupStamp := nil.
	self forceChangesToDisk.! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/5/2010 21:19'!
at: aKey
	"Answer the global associated with the key."

	^globals at: aKey! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/9/2010 16:42'!
at: aKey ifAbsentPut: aBlock 
	"Return the global at the given key. 
	If key is not included in the receiver store the result 
	of evaluating aBlock as new value."

	^globals at: aKey ifAbsentPut: aBlock! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/5/2010 21:20'!
at: key ifAbsent: aBlock
	"Answer the global associated with the key or, if key isn't found,
	answer the result of evaluating aBlock."

	^globals at: key ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ul 4/18/2010 22:22'!
at: key ifPresentAndInMemory: aBlock
	"Lookup the given key in the receiver. If it is present, answer the value of evaluating the given block with the value associated with the key. Otherwise, answer nil."

	^globals at: key ifPresentAndInMemory: aBlock! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/5/2010 21:20'!
at: key ifPresent: aBlock
	"Lookup the given key in the globals. If it is present, answer the value of evaluating the given block with the value associated with the key. Otherwise, answer nil."

	^globals at: key ifPresent: aBlock! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/5/2010 21:21'!
at: aKey put: anObject 
	"Set the global at key to be anObject.  If key is not found, create a
	new entry for key and set is value to anObject. Answer anObject."

	^globals at: aKey put: anObject! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/5/2010 20:34'!
bindingOf: varName
	"Answer the binding of some variable resolved in the scope of the receiver"

	^globals bindingOf: varName! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'nice 3/6/2010 20:57'!
buildDate			
	"Return a String reflecting the build date of the VM"
	"Smalltalk buildDate"

	^self getSystemAttribute: 1006! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
bytesLeft
	"Answer the number of bytes of space available. Does a full garbage collection."

	^ self garbageCollect
! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 2/25/2001 18:00'!
bytesLeftString
	"Return a string describing the amount of memory available"
	| availInternal availPhysical availTotal |
	self garbageCollect.
	availInternal := self primBytesLeft.
	availPhysical := self bytesLeft: false.
	availTotal := self bytesLeft: true.
	(availTotal > (availInternal + 10000)) "compensate for mini allocations inbetween"
		ifFalse:[^availInternal asStringWithCommas, ' bytes available'].
	^String streamContents:[:s|
		s nextPutAll: availInternal asStringWithCommas, 	' bytes (internal) '; cr.
		s nextPutAll: availPhysical asStringWithCommas,	' bytes (physical) '; cr.
		s nextPutAll: availTotal asStringWithCommas, 	' bytes (total)     '].! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 2/25/2001 17:55'!
bytesLeft: aBool
	"Return the amount of available space. If aBool is true, include possibly available swap space. If aBool is false, include possibly available physical memory. For a report on the largest free block currently availabe within Squeak memory but not counting extra memory use #primBytesLeft."
	<primitive: 112>
	^self primBytesLeft! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 2/20/2016 23:38'!
calcEndianness
	"What endian-ness is the current hardware?  The String '1234' will be stored into a machine word.  On BigEndian machines (the Mac), $1 will be the high byte if the word.  On LittleEndian machines (the PC), $4 will be the high byte."
	"Smalltalk endianness"
	| wordThenBytes |
	wordThenBytes := WordArray with: 16r01020304.
	ByteArray adoptInstance: wordThenBytes.
	wordThenBytes first = 1 ifTrue: [^#big].
	wordThenBytes first = 4 ifTrue: [^#little].
	self error: 'The author is confused'! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'yo 7/2/2004 13:24'!
changeImageNameTo: aString
	
	self imageName: aString asSqueakPathName.
	LastImageName := self imageName! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'bf 1/12/2006 19:43'!
changesName
	"Answer the name for the changes file corresponding to the image file name."
	"Smalltalk changesName"

	| imName |
	imName := FileDirectory baseNameFor: self imageName.
	^ imName, FileDirectory dot, 'changes'! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 20:36'!
classNamed: className 
	"Answer the global with the given name."

	^globals classNamed: className! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:05'!
classNames
	"Answer a collection of all class names in the system."

	^globals classNames! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 18:22'!
classOrTraitNamed: aString
	"Answer the global with the given name"

	^globals classOrTraitNamed: aString! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'cwp 1/18/2014 10:21'!
cleanOutUndeclared 
	"This should be deprecated, and senders rewritten to deal with environments directly"
	self flag: #environments.
	
	globals purgeUndeclared.! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'ar 2/27/2010 00:02'!
cleanUp	"Smalltalk cleanUp"
	"Gently clean up"
	^self cleanUp: false! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'ar 2/28/2010 02:28'!
cleanUp: aggressive
	"Clean up. When aggressive is true, this will destroy projects, change sets, etc."
	"Smalltalk cleanUp: false"
	"Smalltalk cleanUp: true"

	^self cleanUp: aggressive except: #()! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'cmm 8/30/2013 16:06'!
cleanUp: aggressive except: exclusions 
	"Clean up. When aggressive is true, this will destroy projects, change sets, etc.
	Leave out any classes specifically listed in exclusions."
	"Smalltalk cleanUp: true except: {Project. ChangeSet}"
	"Find all classes implementing #cleanUp or cleanUp:"
	| classes |
	classes := Smalltalk allClasses select:
		[ : eachClass | (eachClass class includesSelector: #cleanUp) or:
			[ (eachClass class includesSelector: #cleanUp:) and: [ (exclusions includes: eachClass) not ] ] ].
	"Arrange classes in superclass order, superclasses before subclasses, so that specific cleanup (like MethodDictionary compaction) will run after generic superclass cleanup (HashedCollection rehashing). Otherwise generic superclass cleanup might undo specific one (in this case rehashing will undo a good bit of MD compaction)."
	(ChangeSet superclassOrder: classes)
		do: [ : aClass | aClass cleanUp: aggressive ]
		displayingProgress: [ : aClass | 'Cleaning up in ' , aClass name ]! !

!SmalltalkImage methodsFor: 'external objects' stamp: 'ul 10/27/2014 16:37'!
clearExternalObjects
	"Clear the array of objects that have been registered for use in non-Smalltalk code."
	"Smalltalk clearExternalObjects"

	ExternalObjectTable current clearExternalObjects
! !

!SmalltalkImage methodsFor: 'vm profiling' stamp: 'sd 9/24/2003 11:54'!
clearProfile
	"Clear the profile database."

	<primitive: 250>
! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'sd 11/16/2003 13:12'!
closeSourceFiles
	"Shut down the source files if appropriate.  1/29/96 sw: changed so that the closing and nilification only take place if the entry was a FileStream, thus allowing stringified sources to remain in the saved image file"

	1 to: 2 do: [:i |
		((SourceFiles at: i) isKindOf: FileStream)
			ifTrue:
				[(SourceFiles at: i) close.
				SourceFiles at: i put: nil]]! !

!SmalltalkImage methodsFor: 'vm' stamp: 'laza 2/8/2011 11:41'!
cogitClass
	^self getSystemAttribute: 1008! !

!SmalltalkImage methodsFor: 'command line' stamp: 'nice 3/7/2010 01:28'!
commandLine
	"Answer the object to query about command line."
	
	^self! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'eem 3/5/2014 14:53'!
compactClassesArray
	"Smalltalk compactClassesArray"
	"Backward-compatibility support.  Spur does not have compact classes."
	^{}! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'ul 6/17/2011 12:29'!
compressSources	
	"Copy all the source file to a compressed file. Usually preceded by Smalltalk condenseSources."
	"The new file will be created in the default directory, and the code in openSources
	will try to open it if it is there, otherwise it will look for normal sources."
	"Smalltalk compressSources"

	| f cfName cf |
	f := SourceFiles first readOnlyCopy binary.	"binary to preserve utf8 encoding"
	(f localName endsWith: 'sources')
		ifTrue: [cfName := (f localName allButLast: 7) , 'stc']
		ifFalse: [self error: 'Hey, I thought the sources name ended with ''.sources''.'].
	cf := (CompressedSourceStream on: (FileStream newFileNamed: cfName))
				segmentSize: 65536 maxSize: f size.

	"Copy the sources"
'Compressing Sources File...'
	displayProgressFrom: 0 to: f size
	during:
		[:bar | f position: 0.
		[f atEnd] whileFalse:
			[cf nextPutAll: (f next: 65536).
			bar value: f position]].
	cf close.
	self setMacFileInfoOn: cfName.
	self inform: 'You now have a compressed sources file!!
Squeak will use it the next time you start.'! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'fbs 7/20/2013 23:29'!
condenseChanges
	"Move all the changes onto a compacted sources file."
	"Smalltalk condenseChanges"
	| f oldChanges |
	f := FileStream fileNamed: 'ST80.temp'.
	f header; timeStamp.
	'Condensing Changes File...'
		displayProgressFrom: 0
		to: self classNames size + self traitNames size
		during: [:bar | | count | 
			count := 0.
			self
				allClassesAndTraitsDo: [:classOrTrait | 
					bar value: (count := count + 1).
					classOrTrait moveChangesTo: f.
					classOrTrait putClassCommentToCondensedChangesFile: f.
					classOrTrait classSide moveChangesTo: f]].
	self lastQuitLogPosition: f position.
	f trailer; close.
	oldChanges := SourceFiles at: 2.
	oldChanges close.
	FileDirectory default deleteFileNamed: oldChanges name , '.old';
		 rename: oldChanges name toBe: oldChanges name , '.old';
		 rename: f name toBe: oldChanges name.
	self setMacFileInfoOn: oldChanges name.
	SourceFiles
		at: 2
		put: (FileStream oldFileNamed: oldChanges name)! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'eem 5/28/2015 14:33'!
condenseSources
	"Move all the changes onto a compacted sources file."
	"Smalltalk condenseSources"

	| newSourcesFile defaultDirectory newVersion currentVersion |
	Utilities fixUpProblemsWithAllCategory.
	"The above removes any concrete, spurious '-- all --' categories, which mess up the process."
	defaultDirectory := FileDirectory default.
	currentVersion := self sourceFileVersionString.
	newVersion := UIManager default 
		request: 'Please designate the version\for the new source code file...' withCRs
		initialAnswer: currentVersion.
	newVersion ifEmpty: [ ^ self ].
	newVersion = currentVersion ifTrue: [ ^ self error: 'The new source file must not be the same as the old.' ].
	self sourceFileVersionString: newVersion.

	"Write all sources with fileIndex 1"
	newSourcesFile := defaultDirectory newFileNamed: (defaultDirectory localNameFor: self sourcesName).
	newSourcesFile ifNil: [ ^ self error: 'Couldn''t create source code file in\' withCRs,  defaultDirectory name].
	newSourcesFile
		header;
		timeStamp.
	'Condensing Sources File...' 
		displayProgressFrom: 0
		to: self classNames size + self traitNames size
		during: 
			[ :bar | 
			| count |
			count := 0.
			Smalltalk allClassesAndTraitsDo: 
				[ :classOrTrait | 
				bar value: (count := count + 1).
				classOrTrait 
					fileOutOn: newSourcesFile
					moveSource: #historically
					toFile: 1 ] ].
	newSourcesFile
		trailer;
		close.
		
	"Make a new empty changes file"
	self closeSourceFiles.
	defaultDirectory 
		rename: self changesName
		toBe: self changesName , '.old'.
	(FileStream newFileNamed: self changesName)
		header;
		timeStamp;
		close.
	self lastQuitLogPosition: 0.
	self setMacFileInfoOn: self changesName.
	self setMacFileInfoOn: newSourcesFile name.
	self openSourceFiles.
	self inform: 'Source files have been rewritten to\' withCRs, newSourcesFile name, '\Check that all is well,\and then save/quit.' withCRs! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: ''!
copyright
	"The Smalltalk copyright."

	^'Copyright (c) Xerox Corp. 1981, 1982 All rights reserved.
Copyright (c) Apple Computer, Inc. 1985-1996 All rights reserved.'! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
createStackOverflow
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; createStackOverflow"

	self createStackOverflow.  "infinite recursion"! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'fbs 7/25/2013 07:18'!
currentChangeSetString
	"Smalltalk currentChangeSetString"
	^ 'Current Change Set: ' translated, ChangeSet current name! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'sd 9/29/2004 18:27'!
currentProjectDo: aBlock 
	"So that code can work after removal of Projects"
	self
		at: #Project
		ifPresent: [:projClass | aBlock value: projClass current]! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'nk 7/29/2004 10:03'!
datedVersion
	"Answer the version of this release."

	^SystemVersion current datedVersion! !

!SmalltalkImage methodsFor: 'command line' stamp: 'cmm 10/12/2013 18:09'!
documentPath
	"Answer the absolute path of the document passed to the vm or nil if none."
	"Smalltalk commandLine documentPath"
	^ Preferences readDocumentAtStartup ifTrue: [ self getSystemAttribute: 2 ]! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'jrd 6/10/2009 12:54'!
dumpException: ex

	self 
		logSqueakError: ex description 
		inContext: ex signalerContext .
! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'mtf 4/27/2011 15:07'!
dumpException: ex to: aFilename

	self 
		logError: ex description 
		inContext: ex signalerContext
		to: aFilename
! !

!SmalltalkImage methodsFor: 'vm profiling' stamp: 'sd 9/24/2003 11:54'!
dumpProfile
	"Dump the profile database to a file."

	<primitive: 251>
! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'nice 9/25/2010 15:13'!
endianness
	^EndianCache ifNil: [EndianCache := self calcEndianness]! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/5/2010 21:21'!
environment
	"For conversion from Smalltalk to SystemDictionary"

	^globals! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'topa 3/3/2016 16:21'!
event: anEvent
	"Hook for SystemChangeNotifier"

	anEvent isDoIt ifTrue: [^self logChange: anEvent item].
	(anEvent itemKind = SystemChangeNotifier categoryKind) ifTrue: [
		anEvent isAdded
			ifTrue: [^self logChange: 'SystemOrganization addCategory: ', anEvent item storeString].
		anEvent isRemoved
			ifTrue: [^self logChange: 'SystemOrganization removeSystemCategory: ', anEvent item storeString].
		anEvent isRenamed
			ifTrue: [^self logChange: 'SystemOrganization renameCategory: ', anEvent oldName storeString, ' toBe: ', anEvent newName storeString].
	].

	(anEvent itemKind = SystemChangeNotifier classKind) ifTrue: [
		anEvent item acceptsLoggingOfCompilation ifFalse: [^self].
		anEvent isAdded ifTrue: [
			self logChange: anEvent item definition.
		].
		anEvent isModified
			ifTrue: [^self logChange: anEvent item definition].
		anEvent isRemoved
			ifTrue: [^self logChange: 'Smalltalk removeClassNamed: ' , anEvent item name storeString].
		anEvent isRenamed
			ifTrue: [^self logChange: 'Smalltalk renameClassNamed: ', anEvent oldName storeString, ' as: ', anEvent newName storeString].
		anEvent isRecategorized
			ifTrue: [^self logChange: 'SystemOrganization classify: ', anEvent item name storeString, ' under: ', anEvent itemCategory storeString].
	].

	(anEvent itemKind = SystemChangeNotifier protocolKind) ifTrue: [
		anEvent isAdded
			ifTrue: [^self logChange: anEvent itemClass name , ' organization addCategory: ' , anEvent itemProtocol storeString].
		anEvent isRemoved
			ifTrue: [^self logChange: anEvent itemClass name , ' organization removeCategory: ' , anEvent itemProtocol storeString].
		anEvent isRenamed
			ifTrue: [^self logChange: anEvent itemClass name , ' organization renameCategory: ' , anEvent oldName storeString, ' toBe: ', anEvent newName storeString].
	].

	(anEvent itemKind = SystemChangeNotifier methodKind) ifTrue: [
		anEvent itemClass acceptsLoggingOfCompilation ifFalse: [^self].
		anEvent isRemoved
			ifTrue: [^self logChange: anEvent itemClass name , ' removeSelector: ' , anEvent itemSelector storeString].
		anEvent isRecategorized
			ifTrue: [
				^self logChange: anEvent itemClass name , ' organization classify: ', anEvent itemSelector storeString, ' under: ', anEvent itemProtocol storeString.
		].
	].! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'sd 6/28/2003 18:23'!
exitToDebugger
	"Primitive. Enter the machine language debugger, if one exists. Essential.
	See Object documentation whatIsAPrimitive. This primitive is to access the
	debugger when debugging the vm or a plugging in C"

	<primitive: 114>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'external objects' stamp: 'ul 10/27/2014 16:38'!
externalObjects
	"Return an array of objects that have been registered for use in non-Smalltalk code. Smalltalk objects should be referrenced by external code only via indirection through this array, thus allowing the objects to move during compaction. This array can be cleared when the VM re-starts, since variables in external code do not survive snapshots. Note that external code should not attempt to access a Smalltalk object, even via this mechanism, while garbage collection is in progress."
	"Note that this is just a copy of the array used by the VM."
	"Smalltalk externalObjects"

	^ExternalObjectTable current externalObjects
! !

!SmalltalkImage methodsFor: 'command line' stamp: 'cmm 10/12/2013 16:21'!
extractParameters
self deprecated: 'Use #namedArguments'.
	^ self namedArguments! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'sd 6/27/2003 23:27'!
extraVMMemory
	"Answer the current setting of the 'extraVMMemory' VM parameter. See the comment in extraVMMemory: for details."

	^ self vmParameterAt: 23
! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'sd 6/27/2003 23:27'!
extraVMMemory: extraBytesToReserve
	"Request that the given amount of extra memory be reserved for use by the virtual machine to leave extra C heap space available for things like plugins, network and file buffers, and so on. This request is stored when the image is saved and honored when the image is next started up. Answer the previous value of this parameter."

	extraBytesToReserve < 0
		ifTrue: [self error: 'VM memory reservation must be non-negative'].
	^ self vmParameterAt: 23 put: extraBytesToReserve
! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'cwp 1/8/2014 20:07'!
fixObsoleteReferences
	"Smalltalk fixObsoleteReferences"
	

	Smalltalk garbageCollect; garbageCollect.

	Preference allInstances do: [:each || informee realName newInformee | 
		informee := each instVarNamed: #changeInformee.
		((informee isKindOf: Behavior)
			and: [informee isObsolete])
			ifTrue: [
				Transcript show: 'Preference: '; show: each name; cr.
				realName := (informee name copyReplaceAll: 'AnObsolete' with: '') asSymbol.
				newInformee := (informee environment declarationOf: realName) value.
				each instVarNamed: #changeInformee put: newInformee]].
 
	CompiledMethod allInstances do: [:method |
		| obsoleteBindings |
		obsoleteBindings := method literals select: [:literal |
			literal isVariableBinding
				and: [literal value isBehavior
				and: [literal value isObsolete]]].
		obsoleteBindings do: [:binding |
			| obsName realName realClass |
			obsName := binding value name.
			Transcript show: 'Binding: '; show: obsName; cr.
			realName := obsName copyReplaceAll: 'AnObsolete' with: ''.
			realClass := Smalltalk at: realName asSymbol ifAbsent: [UndefinedObject].
			binding isSpecialWriteBinding
				ifTrue: [binding privateSetKey: binding key value: realClass]
				ifFalse: [binding key: binding key value: realClass]]].

	Behavior flushObsoleteSubclasses.
	Smalltalk garbageCollect; garbageCollect.
	
! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 7/22/2010 17:44'!
flagInterpretedMethods: aBoolean
	"The Cog VM can be instructed to set the flag bit of CompiledMethods that
	 it executes but will only interpret.  This can be used e.g. to profile startup.
	 See CompiledMethod>>#flag & CompiledMethod>>#clearFlag.  
	 This flag persists across snapshots, stored in the image header."

	self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 2) + (aBoolean ifTrue: [2] ifFalse: [0])! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 18:23'!
flushClassNameCache
	"Forse recomputation of the cached list of class names."

	^globals flushClassNameCache! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'NS 1/16/2004 15:35'!
forceChangesToDisk
	"Ensure that the changes file has been fully written to disk by closing and re-opening it. This makes the system more robust in the face of a power failure or hard-reboot."

	| changesFile |
	changesFile := SourceFiles at: 2.
	(changesFile isKindOf: FileStream) ifTrue: [
		changesFile flush.
		SecurityManager default hasFileAccess ifTrue:[
			changesFile close.
			changesFile open: changesFile name forWrite: true].
		changesFile setToEnd.
	].
! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'JMM 1/27/2005 13:23'!
forceTenure
	"Primitive. Tell the GC logic to force a tenure on the next increment GC."
	<primitive: 'primitiveForceTenure'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 19:58'!
forgetClass: aClass logged: aBool 
	"Delete the class, aClass, from the system."

	^globals forgetClass: aClass logged: aBool! !

!SmalltalkImage methodsFor: 'modules' stamp: 'dvf 6/30/2005 12:11'!
forgetModule: aString
	"Primitive. If the module named aString is loaded, unloaded. If not, and it is marked an unloadable, unmark it so the VM will try to load it again next time. See comment for #unloadModule:."
	<primitive: 571>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'nice 5/11/2009 23:30'!
fullNameForChangesNamed: aName
	| imgName |
	imgName := self fullNameForImageNamed: aName.
	^FileDirectory fileName: (FileDirectory baseNameFor: imgName) extension: FileDirectory changeSuffix.! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'nice 5/11/2009 23:33'!
fullNameForImageNamed: aName
	| imgDir |
	imgDir := FileDirectory on: self imagePath.
	^FileDirectory fileName: (imgDir fullNameFor: aName) extension: FileDirectory imageSuffix.! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'fbs 6/30/2013 11:04'!
garbageCollect
	"Primitive. Reclaims all garbage and answers the number of bytes of available space."
	Object flushDependents.
	EventManager flushEvents.
	^self primitiveGarbageCollect! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
garbageCollectMost
	"Primitive. Reclaims recently created garbage (which is usually most of it) fairly quickly and answers the number of bytes of available space."

	<primitive: 131>
	^ self primBytesLeft! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ar 3/9/2010 17:00'!
getFileNameFromUser
	"Ask the user for a new image name"
	^self getFileNameFromUserSuggesting: (FileDirectory localNameFor: self imageName)! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ar 3/9/2010 16:59'!
getFileNameFromUserSuggesting: aName
	"Ask the user for a new image name"
	| newName |
	newName := UIManager default
		request: 'New File Name?' translated
		initialAnswer: aName.
	newName isEmpty ifTrue: [^nil].
	((FileDirectory default fileOrDirectoryExists: (self fullNameForImageNamed: newName)) or:
	 [FileDirectory default fileOrDirectoryExists: (self fullNameForChangesNamed: newName)]) ifTrue: [
		(self confirm: ('{1} already exists. Overwrite?' translated format: {newName})) ifFalse: [^nil]].
	^newName
! !

!SmalltalkImage methodsFor: 'private' stamp: 'cmm 10/15/2013 13:27'!
getSystemAttribute: attributeID 
	"Optional. Answer the string for the system attribute with the given 
	integer ID. Answer nil if the given attribute is not defined on this 
	platform. On platforms that support invoking programs from command 
	lines (e.g., Unix), this mechanism can be used to pass command line 
	arguments to programs written in Squeak.

	By convention, the first command line argument that is not a VM
	configuration option is considered a 'document' to be filed in. Such a
	document can add methods and classes, can contain a serialized object,
	can include code to be executed, or any combination of these.

	Currently defined attributes include: 
	-1000	1000th command line argument that specify VM options
	...
	-1		first command line argument that specify VM options
	0		the full path name for currently executing VM
			(or, on some platforms, just the path name of the VM's directory) 
	1		full path name of this image (better use primImageName instead)
	2		first command-line argument for Squeak programs.
			Note: if Preferences readDocumentAtStartup is set, this first argument is treated as a URL to a Squeak document to open.
	3		second command-line argument for Squeak programs
	...
	1000	1000th command line argument for Squeak programs
	1001	this platform's operating system 'Mac OS', 'Win32', 'unix', ...
	1002	operating system version
	1003	this platform's processor type
	1004	vm version
	1005	window system name
	1006	vm build id
	1007	Interpreter class (Cog VM only)
	1008	Cogit class (Cog VM only)
	1009	Platform source version (Cog VM only?)
	1201	max filename length (Mac OS only)
	1202	file last error (Mac OS only)
	10001	hardware details (Win32 only)
	10002	operating system details (Win32 only)
	10003	graphics hardware details (Win32 only)
	"

	<primitive: 149>
	^ nil! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'fbs 7/25/2013 07:18'!
getVMParameters	
	"Answer an Array containing the current values of the VM's internal
	parameter/metric registers.  Each value is stored in the array at the
	index corresponding to its VM register.  (See #vmParameterAt: and
	#vmParameterAt:put:.)"
	"Smalltalk getVMParameters"
	
	<primitive: 254>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/5/2010 21:39'!
globals
	"Answer the global SystemDictionary"

	^globals! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/5/2010 21:40'!
globals: aSystemDictionary
	"Sets the system-wide globals"

	globals ifNotNil:[self error: 'Cannot overwrite existing globals'].
	globals := aSystemDictionary! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'eem 6/23/2016 16:30'!
growMemoryByAtLeast: numBytes
	"Grow memory by at least the requested number of bytes.
	 Primitive.  Fail if no memory is available.  Essential.
	 N.B. In Spur, the heap is composed of segments. Growing memory adds a new
	 segment; it does not extend existing segments. Growth is by at least that
	 established by SmalltalkImage current vmParameterAt: 25, which defaults to 16Mb."
	<primitive: 180>
	^(numBytes isInteger and: [numBytes > 0])
		ifTrue: [OutOfMemory signal]
		ifFalse: [self primitiveFailed]! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'dtl 11/23/2009 15:01'!
handleUserInterrupt
	Preferences cmdDotEnabled ifTrue:
		[[Project current interruptName: 'User Interrupt'] fork]
! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 18:30'!
hasClassNamed: aString
	"Answer whether there is a class of the given name"

	^globals hasClassNamed: aString! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'sd 9/29/2004 18:17'!
hasMorphic
	"Answer whether the Morphic classes are available in the
	system (they may have been stripped, such as by a call to
	Smalltalk removeMorphic"
	^ (self
		at: #Morph
		ifAbsent: [])
		isKindOf: Class! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
hasSpecialSelector: aLiteral ifTrueSetByte: aBlock

	1 to: self specialSelectorSize do:
		[:index | 
		(self specialSelectorAt: index) == aLiteral
			ifTrue: [aBlock value: index + 16rAF. ^true]].
	^false! !

!SmalltalkImage methodsFor: 'image' stamp: 'dtl 4/11/2010 11:45'!
image
	"Answer the object to query about the current object memory and execution environment."
	
	^self! !

!SmalltalkImage methodsFor: 'image' stamp: 'dtl 11/7/2012 23:00'!
imageFormatVersion
	"Answer an integer identifying the type of image in memory. The image version number may
	identify the format of the image (e.g. 32 or 64-bit word size) or specific requirements
	of the image (e.g. block closure support required). This invokes an optional primitive
	that may not be available on all virtual machines."

	"Smalltalk image imageFormatVersion"

	<primitive: 'primitiveImageFormatVersion'>

	"Cog provides a VM parameter"
	^[Smalltalk vm vmParameterAt: 41]
		on: Error
		do: [self notify: 'This virtual machine does not support the optional ',
				'primitive #primitiveImageFormatVersion' translated.
			nil]
! !

!SmalltalkImage methodsFor: 'image' stamp: 'bf 4/18/2011 13:34'!
imageFormatVersionFromFile
	"Answer an integer identifying the type of image on file. The image version number may
	identify the format of the image (e.g. 32 or 64-bit word size) or specific requirements
	of the image (e.g. block closure support required)"

	"Smalltalk image imageFormatVersionFromFile"

	| format |
	format := self imageFormatVersionFromFileAsIs.
	^format <= 16r00FFFFFF
		ifTrue: [  "same endianness as VM"
			format ]
		ifFalse: [ "convert endianness"
			((format bitAnd: 16rFF000000) >> 24)
			+ ((format bitAnd: 16r00FF0000) >> 8)
			+ ((format bitAnd: 16r0000FF00) << 8)
			+ ((format bitAnd: 16r000000FF) << 16)]! !

!SmalltalkImage methodsFor: 'image' stamp: 'bf 7/11/2011 10:54'!
imageFormatVersionFromFileAsIs
	"Answer an integer identifying the type of image on file. The image version number may
	identify the format of the image (e.g. 32 or 64-bit word size) or specific requirements
	of the image (e.g. block closure support required). If the image file has a different
	endianness than the VM, the format version will appear byte-swapped."

	"Smalltalk image imageFormatVersionFromFileAsIs"

	^(FileStream readOnlyFileNamed: Smalltalk imageName do: [ :file | file binary; next: 4 ])
		unsignedLongAt: 1 bigEndian: Smalltalk isBigEndian! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'fbs 7/25/2013 07:18'!
imageName
	"Answer the full path name for the current image."
	"Smalltalk imageName"

	| str |
	str := self primImageName.
	^ (FilePath pathName: str isEncoded: true) asSqueakPathName.
! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'yo 2/24/2005 18:34'!
imageName: newName
	"Set the the full path name for the current image.  All further snapshots will use this."

	| encoded |
	encoded := (FilePath pathName: newName isEncoded: false) asVmPathName.
	self primImageName: encoded.
! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'fbs 7/25/2013 07:18'!
imagePath
	"Answer the path for the directory containing the image file."
	"Smalltalk imagePath"

	^ FileDirectory dirPathFor: self imageName
! !

!SmalltalkImage methodsFor: 'dictionary access' stamp: 'ar 3/5/2010 21:21'!
includesKey: key
	"Answer whether the receiver has a key equal to the argument, key."

	^globals includesKey: key! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
installLowSpaceWatcher
	"Start a process to watch for low-space conditions."
	"Smalltalk installLowSpaceWatcher"

	self primSignalAtBytesLeft: 0.  "disable low-space interrupts"
	LowSpaceProcess == nil ifFalse: [LowSpaceProcess terminate].
	LowSpaceProcess := [self lowSpaceWatcher] newProcess.
	LowSpaceProcess priority: Processor lowIOPriority.
	LowSpaceProcess resume.

! !

!SmalltalkImage methodsFor: 'vm' stamp: 'laza 2/8/2011 11:41'!
interpreterClass
	^self getSystemAttribute: 1007! !

!SmalltalkImage methodsFor: 'vm' stamp: 'dtl 4/11/2010 11:38'!
interpreterSourceVersion
	"Answer a string corresponding to the version of the interpreter source.
	This represents the version level of the Smalltalk source code (interpreter
	and various plugins) that is translated to C by a CCodeGenerator, as distinct
	from the external platform source code, typically written in C and managed
	separately for each platform. An optional primitive is invoked that may not
	be available on all virtual machines."

	"Smalltalk vm interpreterSourceVersion"

	<primitive: 'primitiveInterpreterSourceVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitiveInterpreterSourceVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'sd 6/27/2003 23:25'!
isBigEndian
	^self endianness == #big! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'cmm 10/15/2013 13:32'!
isHeadless
	"Answer whether the command-line specified to launch the VM headless."
	self optionsDo:
		[ : each | (#('display=none' '-headless' '-vm-display-null' ) includes: each) ifTrue: [ ^ true ] ].
	^ false! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'sd 6/27/2003 23:25'!
isLittleEndian
	^self endianness == #little! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'mt 2/18/2016 15:18'!
isMorphic
       
	^ Project current isMorphic! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:46'!
isRoot: oop
	"Primitive. Answer whether the object is currently a root for youngSpace."
	<primitive: 'primitiveIsRoot'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 7/18/2014 10:39'!
isRunningCog
	"Answers if we're running on a Cog VM (JIT or StackInterpreter)"

	^(self vmParameterAt: 42)
		ifNil: [false]
		ifNotNil: [:numStackPages| numStackPages > 0]! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 7/18/2014 10:38'!
isRunningCogit
	"Answers if we're running on the Cog JIT"

	^(self vmParameterAt: 46)
		ifNil: [false]
		ifNotNil: [:machineCodeZoneSize| machineCodeZoneSize > 0]! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 7/18/2014 10:37'!
isRunningSpur
	"Answers if we're running on the Spur memory manager.  Spur has immediate characters.
	 (vmParameterAt: 41 is the image format version for the VM)"

	^Character instSize = 0
	"a.k.a. (self vmParameterAt: 41) anyMask: 16"! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:47'!
isYoung: oop
	"Primitive. Answer whether the object currently resides in youngSpace."
	<primitive: 'primitiveIsYoung'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'sd 11/16/2003 14:02'!
lastQuitLogPosition
	^ LastQuitLogPosition! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'sd 11/16/2003 14:03'!
lastQuitLogPosition: aNumber
	"should be only use to ensure the transition from SystemDictionary to SmalltalkImage, then  	be removed"
		
	LastQuitLogPosition := aNumber! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'fbs 7/25/2013 07:18'!
lastUpdateString
	"Smalltalk lastUpdateString"
	^'latest update: #' translated, SystemVersion current highestUpdate printString! !

!SmalltalkImage methodsFor: 'license' stamp: 'cmm 6/29/2015 11:35'!
license
	"This method contains the text of the license agreement for Squeak."

	^ 

'Copyright (c) The individual, corporate, and institutional contributors who have collectively contributed elements to this software ("The Squeak Community"), 1996-2015 All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

Portions of Squeak are covered by the following license


Copyright (c) Xerox Corp. 1981, 1982 All rights reserved.
Copyright (c) Apple Computer, Inc. 1985-1996 All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
'! !

!SmalltalkImage methodsFor: 'modules' stamp: 'fbs 7/25/2013 07:18'!
listBuiltinModules
	"Smalltalk listBuiltinModules"
	"Return a list of all builtin modules (e.g., plugins). Builtin plugins are those that are 	compiled with the VM directly, as opposed to plugins residing in an external shared library. 	The list will include all builtin plugins regardless of whether they are currently loaded 
	or not. Note that the list returned is not sorted!!"

	| modules index name |
	modules := WriteStream on: (Array new: 20).
	index := 1.
	[
		name := self listBuiltinModule: index.
		name ifNil:[^modules contents].
		modules nextPut: name.
		index := index + 1 ] repeat! !

!SmalltalkImage methodsFor: 'modules' stamp: 'sd 6/27/2003 23:47'!
listBuiltinModule: index
	"Return the name of the n-th builtin module.
	This list is not sorted!!"
	<primitive: 572>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'modules' stamp: 'fbs 7/25/2013 07:18'!
listLoadedModules
	"Smalltalk listLoadedModules"
	"Return a list of all currently loaded modules (e.g., plugins). Loaded modules are those that currently in use (e.g., active). The list returned will contain all currently active modules regardless of whether they're builtin (that is compiled with the VM) or external (e.g., residing in some external shared library). Note that the returned list is not sorted!!"
	| modules index name |
	modules := WriteStream on: (Array new: 20).
	index := 1.
	[
		name := self listLoadedModule: index.
		name ifNil:[^modules contents].
		modules nextPut: name.
		index := index + 1 ] repeat! !

!SmalltalkImage methodsFor: 'modules' stamp: 'sd 6/27/2003 23:48'!
listLoadedModule: index
	"Return the name of the n-th loaded module.
	This list is not sorted!!"
	<primitive: 573>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'cmm 3/14/2014 16:56'!
locateSourcesEntry
	| sourcesFilename |
	sourcesFilename := SourceFileVersionString , FileDirectory dot , 'sources'.
	^ {Smalltalk vmPath. 
	FileDirectory default fullName, FileDirectory slash}
		detect: [ : each | (each , sourcesFilename) asDirectoryEntry notNil ]
		ifFound: [ : foundPath | (foundPath , sourcesFilename) asDirectoryEntry ]
		ifNone: [ nil ]! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'ar 4/10/2005 18:02'!
logChange: aStringOrText 
	"Write the argument, aString, onto the changes file."
	| aString changesFile |
	(SourceFiles isNil or: [(SourceFiles at: 2) == nil]) ifTrue: [^ self].
	self assureStartupStampLogged.

	aString := aStringOrText asString.
	(aString findFirst: [:char | char isSeparator not]) = 0
		ifTrue: [^ self].  "null doits confuse replay"
	(changesFile := SourceFiles at: 2).
	changesFile isReadOnly ifTrue:[^self].
	changesFile setToEnd; cr; cr.
	changesFile nextChunkPut: aString.
		"If want style changes in DoIt, use nextChunkPutWithStyle:, and allow Texts to get here"
	self forceChangesToDisk.! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'tk 10/16/2001 19:24'!
logError: errMsg inContext: aContext to: aFilename
	"Log the error message and a stack trace to the given file."

	| ff |
	FileDirectory default deleteFileNamed: aFilename ifAbsent: [].
	(ff := FileStream fileNamed: aFilename) ifNil: [^ self "avoid recursive errors"].

  	ff nextPutAll: errMsg; cr.
	aContext errorReportOn: ff.
	ff close.! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'mtf 7/1/2009 13:02'!
logSqueakError: errMsg inContext: aContext 
	"Log the error message and a stack trace to the usual file."

	^self logError: errMsg inContext: aContext to: self squeakErrorFileName
! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'eem 2/2/2016 11:45'!
lowSpaceThreshold 
	"Answer the low space threshold. When the amount of free memory (after garbage collection)
	 falls below this limit, the system is in serious danger of completely exhausting memory and
	 crashing. This limit should be made high enough to allow the user open a debugger to diagnose
	 a problem or to save the image.  In a stack-based VM such as Cog contexts for activations in
	 the stack zone will have to be created as the debugger opens, requiring additional headroom."

	| slotsForDebugger slotsForContextsOnStackPages |
	slotsForDebugger := 65536. "Arbitrary guess"
	slotsForContextsOnStackPages :=
		(self vmParameterAt: 42)
			ifNil: [0]
			ifNotNil:
				[:numStackPages| | headerSize numActivationsPerPage maxContextSize |
				numActivationsPerPage := 40. "Design goal of the Cog VM"
				headerSize := 2. "64-bytes for Spur"
				maxContextSize := MethodContext instSize + CompiledMethod fullFrameSize + headerSize.
				numStackPages * numActivationsPerPage * maxContextSize].
	^slotsForDebugger + slotsForContextsOnStackPages * self wordSize! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'dtl 11/23/2009 15:10'!
lowSpaceWatcher
	"Wait until the low space semaphore is signalled, then take appropriate actions."

	| free preemptedProcess |
	self garbageCollectMost <= self lowSpaceThreshold
		ifTrue: [self garbageCollect <= self lowSpaceThreshold
				ifTrue: ["free space must be above threshold before
					starting low space watcher"
					^ Beeper beep]].

	Smalltalk specialObjectsArray at: 23 put: nil.  "process causing low space will be saved here"
	LowSpaceSemaphore := Semaphore new.
	self primLowSpaceSemaphore: LowSpaceSemaphore.
	self primSignalAtBytesLeft: self lowSpaceThreshold.  "enable low space interrupts"

	LowSpaceSemaphore wait.  "wait for a low space condition..."

	self primSignalAtBytesLeft: 0.  "disable low space interrupts"
	self primLowSpaceSemaphore: nil.
	LowSpaceProcess := nil.

	"The process that was active at the time of the low space interrupt."
	preemptedProcess := Smalltalk specialObjectsArray at: 23.
	Smalltalk specialObjectsArray at: 23 put: nil.

	"Note: user now unprotected until the low space watcher is re-installed"

	self memoryHogs isEmpty
		ifFalse: [free := self bytesLeft.
			self memoryHogs
				do: [ :hog | hog freeSomeSpace ].
			self bytesLeft > free
				ifTrue: [ ^ self installLowSpaceWatcher ]].
	Project current
			interruptName: 'Space is low'
			preemptedProcess: preemptedProcess
! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'nk 10/28/2000 20:37'!
lowSpaceWatcherProcess
	^LowSpaceProcess! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'HenrikSperreJohansen 4/20/2011 01:24'!
maxExternalSemaphores
	"The size of table where external semaphores are registered. Only in Cog"
	self isRunningCog ifFalse: [^nil].
	^self vmParameterAt: 49! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'nice 9/22/2013 14:04'!
maxExternalSemaphores: aSize
	"Changes the size of table where external semaphores are registered. 
	The size can only grow, and will always be the next power of two larger than the parameter.
	
	Setting this at any time other than start-up can potentially lose requests.
	 i.e. during the realloc new storage is allocated,
	the old contents are copied and then pointers are switched. 
	 Requests occurring during copying won't be seen if they occur to indices already copied. 
	The intended use is to set the table to some adequate maximum at start-up"
	
	self isRunningCog ifFalse: [^0].
	"The vm-header field is a short, maximum 64k entries. Well, on most platforms anyways "
	(aSize < 0 or: [aSize > 16rFFFF]) ifTrue: [^self error: 'maxExternalSemaphores: is limited to 16rFFFF'].
	^self vmParameterAt: 49 put: aSize! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 9/5/2013 17:10'!
maxIdentityHash
	"Answer the maximum identityHash value supported by the VM."
	<primitive: 176>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'sma 4/22/2000 19:03'!
memoryHogs
	"Answer the list of objects to notify with #freeSomeSpace if memory gets full."

	^ MemoryHogs ifNil: [MemoryHogs := OrderedCollection new]! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'cmm 3/16/2014 22:01'!
moveChanges
	"Move the current version of every method in .changes to my .sources file."
	self appendChangesTo:
		((UIManager default
			request: 'Please specify the name of the\.sources file to move changes to.' withCRs
			initialAnswer: self sourceFileVersionString) ifEmpty: [ ^ self inform: 'Aborted.' ])! !

!SmalltalkImage methodsFor: 'command line' stamp: 'cmm 10/12/2013 18:15'!
namedArguments
	"Assume arguments passed to the image are key->value pairs, answer a Dictionary of the argument names and their values."
	| argName value index paramNameValueDictionary |
	paramNameValueDictionary := Dictionary new.
	index := 1.
	[ argName := self argumentAt: index.
	argName isEmptyOrNil ] whileFalse:
		[ index := index + 1.
		value := (self argumentAt: index) ifNil: [ String empty ].
		paramNameValueDictionary
			at: argName asUppercase
			put: value.
		index := index + 1 ].
	^ paramNameValueDictionary! !

!SmalltalkImage methodsFor: 'external objects' stamp: 'ul 10/27/2014 16:38'!
newExternalSemaphore
	"Create and register a new Semaphore, and return an array containing itself and its index."

	^ExternalObjectTable current newExternalSemaphore! !

!SmalltalkImage methodsFor: 'external objects' stamp: 'ul 10/27/2014 16:38'!
newExternalSemaphoreDo: aBlock 
	"Create and register a new Semaphore, then evaluate aBlock with it and its index."

	^ExternalObjectTable current newExternalSemaphoreDo: aBlock! !

!SmalltalkImage methodsFor: 'external objects' stamp: 'ul 10/27/2014 16:39'!
newExternalSemaphores: count 
	"Create and register multiple Semaphores, and return an array containing the semaphores and their indexes in separate arrays."

	^ExternalObjectTable current newExternalSemaphores: count! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'tk 9/28/2000 15:50'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to Smalltalk instead."

	dp := DiskProxy global: #Smalltalk selector: #yourself
			args: #().
	refStrm replace: self with: dp.
	^ dp! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
okayToProceedEvenIfSpaceIsLow
	"Return true if either there is enough memory to do so safely or if the user gives permission after being given fair warning."

	self garbageCollectMost > self lowSpaceThreshold ifTrue: [^ true].  "quick"
	self garbageCollect > self lowSpaceThreshold ifTrue: [^ true].  "work harder"

	^ self confirm:
'WARNING: There is not enough space to start the low space watcher.
If you proceed, you will not be warned again, and the system may
run out of memory and crash. If you do proceed, you can start the
low space notifier when more space becomes available simply by
opening and then closing a debugger (e.g., by hitting Cmd-period.)
Do you want to proceed?'
! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'dtl 12/24/2009 00:02'!
openSourceFiles

	self imageName = LastImageName ifFalse:
		["Reset the author initials to blank when the image gets moved"
		LastImageName := self imageName.
		Utilities setAuthorInitials: ''].
	FileDirectory
		openSources: self sourcesName
		andChanges: self changesName
		forImage: LastImageName.
	SourceFileArray install! !

!SmalltalkImage methodsFor: 'command line' stamp: 'nice 3/7/2010 01:11'!
optionAt: i
	"Answer the i-th option of the command line, or nil if not so many options."
	
	^self getSystemAttribute: i negated! !

!SmalltalkImage methodsFor: 'command line' stamp: 'cmm 10/12/2013 18:21'!
options
	"Answer an array with all the command line options."
	"Smalltalk commandLine options"
	^ Array streamContents:
		[ : stream | self optionsDo:
			[ : each | stream nextPut: each ] ]! !

!SmalltalkImage methodsFor: 'command line' stamp: 'cmm 10/12/2013 18:21'!
optionsDo: aBlock 
	"Enumerate the command-line arguments passed to the vm only.  This does not include arguments passed to the image."
	1
		to: 1000
		by: 1
		do:
			[ : n | 
			(self optionAt: n)
				ifNil: [ ^ self ]
				ifNotNil: [ : opt | aBlock value: opt ] ]! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'ar 3/5/2010 21:40'!
organization
	"Return the organizer for the receiver"

	^globals organization! !

!SmalltalkImage methodsFor: 'os' stamp: 'nice 3/7/2010 01:28'!
os
	"Answer the object to query about os."
	
	^self! !

!SmalltalkImage methodsFor: 'os' stamp: 'nice 3/7/2010 01:36'!
osVersion
	"Return the version number string of the platform we're running on"
	
	"Smalltalk osVersion"

	^(self getSystemAttribute: 1002) asString! !

!SmalltalkImage methodsFor: 'os' stamp: 'nice 3/7/2010 01:21'!
platformName
	"Return the name of the platform we're running on."
	
	"Smalltalk os platformName"

	^self getSystemAttribute: 1001! !

!SmalltalkImage methodsFor: 'vm' stamp: 'eem 9/23/2011 11:39'!
platformSourceVersion
	"Answer a string corresponding to the version of the external platform source
	code, typically written in C and managed separately for each platform. This
	invokes an optional primitive that may not be available on all virtual machines."

	"Smalltalk vm platformSourceVersion"

	<primitive: 'primitivePlatformSourceVersion'>
	(self getSystemAttribute: 1009) ifNotNil:
		[:platformSourceVersion| ^platformSourceVersion].
	self notify: 'This virtual machine supports neither the optional primitive #primitivePlatformSourceVersion nor the optional system attribute 1009' translated.
	^''! !

!SmalltalkImage methodsFor: 'os' stamp: 'nice 3/7/2010 01:21'!
platformSubtype
	"Return the subType of the platform we're running on."

	"Smalltalk os platformSubtype"
	
	^self getSystemAttribute: 1003! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'fbs 7/4/2013 20:33'!
presumedSentMessages   | sent |
"Smalltalk presumedSentMessages"

	"The following should be preserved for doIts, etc"
	sent := IdentitySet new.
	#(compactSymbolTable rebuildAllProjects
		browseAllSelect:  lastRemoval
		scrollBarValue: vScrollBarValue: scrollBarMenuButtonPressed: 
		withSelectionFrom:  to: removeClassNamed:
		dragon: hilberts: mandala: web test3 factorial tinyBenchmarks benchFib
		newDepth: restoreAfter: zapAllMethods obsoleteClasses
		removeAllUnSentMessages abandonSources removeUnreferencedKeys
		reclaimDependents zapOrganization condenseChanges browseObsoleteReferences
		subclass:instanceVariableNames:classVariableNames:poolDictionaries:category:
		methodsFor:stamp: methodsFor:stamp:prior: instanceVariableNames:
		startTimerInterruptWatcher unusedClasses) do:
		[:sel | sent add: sel].
	"The following may be sent by perform: in dispatchOnChar..."
	Smalltalk at: #ParagraphEditor ifPresent: [:paragraphEditor |
		(paragraphEditor classPool at: #CmdActions) asSet do:
			[:sel | sent add: sel].
		(paragraphEditor classPool at: #ShiftCmdActions) asSet do:
			[:sel | sent add: sel]].
	^ sent! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'eem 3/23/2015 10:00'!
primBytesLeft
	"Primitive. Answer the number of free bytes available in old space.
	 Not accurate unless preceded by
		Smalltalk garbageCollectMost (for reasonable accuracy), or
		Smalltalk garbageCollect (for real accuracy).
	 See Object documentation whatIsAPrimitive."

	<primitive: 112>
	^0! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'fbs 7/25/2013 07:18'!
primImageName
	"Answer the full path name for the current image."
	"Smalltalk imageName"

	<primitive: 121>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'yo 7/2/2004 13:30'!
primImageName: newName
	"Set the the full path name for the current image.  All further snapshots will use this."

	<primitive: 121>
	^ self primitiveFailed! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'eem 10/29/2009 10:52'!
primitiveErrorTable
	"Smalltalk primitiveErrorTable"
	^self specialObjectsArray at: 52! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'eem 3/23/2015 10:00'!
primitiveGarbageCollect
	"Primitive. Reclaims all garbage and answers the size of the largest free chunk in old space.."

	<primitive: 130>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
primLowSpaceSemaphore: aSemaphore
	"Primitive. Register the given Semaphore to be signalled when the
	number of free bytes drops below some threshold. Disable low-space
	interrupts if the argument is nil."

	<primitive: 124>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
primSignalAtBytesLeft: numBytes
	"Tell the interpreter the low-space threshold in bytes. When the free
	space falls below this threshold, the interpreter will signal the low-space
	semaphore, if one has been registered.  Disable low-space interrupts if the
	argument is zero.  Fail if numBytes is not an Integer."

	<primitive: 125>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'mtf 12/16/2010 11:23'!
primVmFileName
	"Answer the absolute file name of the the Smalltalk virtual machine"

	^ self getSystemAttribute: 0! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'fbs 7/25/2013 07:19'!
primVmPath
	"Answer the path for the directory containing the Smalltalk virtual machine. Return the 	empty string if this primitive is not implemented."
	"Smalltalk vmPath"

	<primitive: 142>
	^ ''! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'ar 3/4/2010 11:14'!
printOn: aStream
	self == Smalltalk
		ifTrue: [aStream nextPutAll: #Smalltalk]
		ifFalse: [super printOn: aStream]! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 7/22/2010 17:23'!
processHasThreadIdInstVar: aBoolean
	"The threaded VM needs to know if the 4th inst var of Process
	 is threadId which it uses to control process-to-thread binding.
	 This flag persists across snapshots, stored in the image header."
	aBoolean ifTrue: [self assert: (Process instVarNames at: 4) ='threadId'].
	self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 1) + (aBoolean ifTrue: [1] ifFalse: [0])! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 7/22/2010 17:51'!
processPreemptionYields
	"Answer whether the VM causes a process to yield on process preemption,
	 i.e. to put a preempted process at the back of its run queue.  If the parameter
	 is unavailable (non-Cog VMs) or bit 2 (4) is 0 then preemption yields."

	^(([self vmParameterAt: 48]
			on: Error
			do: [:ex| ^true]) allMask: 4) not! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'dtl 2/5/2016 21:07'!
processPreemptionYields: aBoolean
	"The Cog VM can be instructed not to yield on process preemption,
	 i.e. not to put a preempted process at the back of its run queue.  By
	 default preempting a process causes it to yield (Blue Book semantics)
	 which can have unfortunate effects.
	 This flag persists across snapshots, stored in the image header."

	[self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 4) + (aBoolean ifTrue: [0] ifFalse: [4])]
		on: Error
		do: ["no VM parameter 48, do nothing"]! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'eem 8/31/2010 10:38'!
processShutDownList: quitting
	"Send #shutDown to each class that needs to wrap up before a snapshot.
	 Also void the endianness chace;  this can't safely be done on start-up because
	 Smalltalk is too late in the start-up sequence."

	EndianCache := nil.
	self send: #shutDown: toClassesNamedIn: ShutDownList with: quitting! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ar 11/16/1999 20:12'!
processStartUpList: resuming
	"Send #startUp to each class that needs to run initialization after a snapshot."

	self send: #startUp: toClassesNamedIn: StartUpList with: resuming.
! !

!SmalltalkImage methodsFor: 'vm profiling' stamp: 'sd 9/24/2003 11:54'!
profile: aBlock
	"Make a virtual machine profile of the given block."
	"Note: Profiling support is provided so that VM implementors
	 can better understand and improve the efficiency of the virtual
	 machine. To use it, you must be running a version of the
	 virtual machine compiled with profiling enabled (which
	 makes it much slower than normal even when not profiling).
	 You will also need the CodeWarrior profile reader application."

	self stopProfiling.
	self clearProfile.
	self startProfiling.
	aBlock value.
	self stopProfiling.
	self dumpProfile.! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'sd 6/28/2003 17:32'!
quitPrimitive
	"Primitive. Exit to another operating system on the host machine, if one
	exists. All state changes in the object space since the last snapshot are lost.
	Essential. See Object documentation whatIsAPrimitive."

	<primitive: 113>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ul 5/12/2013 22:31'!
quitPrimitive: exitCode
	"Primitive. Exit with exitCode to another operating system on the host machine, if one exists. All state changes in the object space since the last snapshot are lost. Ignore exitCode if it's not supported by the VM."

	<primitive: 113>
	self quitPrimitive! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'sd 9/29/2004 18:15'!
reclaimDependents
	"No-opped due to weak dictionary in use"
	self garbageCollect! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'bf 1/12/2006 19:20'!
recordStartupStamp
	
	StartupStamp := '----STARTUP----', Time dateAndTimeNow printString, ' as ', self imageName.
! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'eem 11/25/2014 16:41'!
recreateSpecialObjectsArray
	"Smalltalk recreateSpecialObjectsArray"
	
	"To external package developers:
	**** DO NOT OVERRIDE THIS METHOD.  *****
	If you are writing a plugin and need additional special object(s) for your own use, 
	use addGCRoot() function and use own, separate special objects registry "
	
	"The Special Objects Array is an array of objects used by the Squeak virtual machine.
	 Its contents are critical and accesses to it by the VM are unchecked, so don't even
	 think of playing here unless you know what you are doing."
	| newArray |
	newArray := Array new: 60.
	"Nil false and true get used throughout the interpreter"
	newArray at: 1 put: nil.
	newArray at: 2 put: false.
	newArray at: 3 put: true.
	"This association holds the active process (a ProcessScheduler)"
	newArray at: 4 put: (self specialObjectsArray at: 4) "(self bindingOf: #Processor) but it answers an Alias".
	"Numerous classes below used for type checking and instantiation"
	newArray at: 5 put: Bitmap.
	newArray at: 6 put: SmallInteger.
	newArray at: 7 put: ByteString.
	newArray at: 8 put: Array.
	newArray at: 9 put: Smalltalk.
	newArray at: 10 put: BoxedFloat64.
	newArray at: 11 put: (self globals at: #MethodContext ifAbsent: [self globals at: #Context]).
	newArray at: 12 put: nil. "was BlockContext."
	newArray at: 13 put: Point.
	newArray at: 14 put: LargePositiveInteger.
	newArray at: 15 put: Display.
	newArray at: 16 put: Message.
	newArray at: 17 put: CompiledMethod.
	newArray at: 18 put: ((self specialObjectsArray at: 18) ifNil: [Semaphore new]). "low space Semaphore"
	newArray at: 19 put: Semaphore.
	newArray at: 20 put: Character.
	newArray at: 21 put: #doesNotUnderstand:.
	newArray at: 22 put: #cannotReturn:.
	newArray at: 23 put: nil. "This is the process signalling low space."
	"An array of the 32 selectors that are compiled as special bytecodes,
	 paired alternately with the number of arguments each takes."
	newArray at: 24 put: #(	#+ 1 #- 1 #< 1 #> 1 #<= 1 #>= 1 #= 1 #~= 1
							#* 1 #/ 1 #\\ 1 #@ 1 #bitShift: 1 #// 1 #bitAnd: 1 #bitOr: 1
							#at: 1 #at:put: 2 #size 0 #next 0 #nextPut: 1 #atEnd 0 #== 1 #class 0
							#blockCopy: 1 #value 0 #value: 1 #do: 1 #new 0 #new: 1 #x 0 #y 0 ).
	"An array of the 255 Characters in ascii order.
	 Cog inlines table into machine code at: prim so do not regenerate it.
	 This is nil in Spur, which has immediate Characters."
	newArray at: 25 put: (self specialObjectsArray at: 25).
	newArray at: 26 put: #mustBeBoolean.
	newArray at: 27 put: ByteArray.
	newArray at: 28 put: Process.
	"An array of up to 31 classes whose instances will have compact headers; an empty array in Spur"
	newArray at: 29 put: self compactClassesArray.
	newArray at: 30 put: ((self specialObjectsArray at: 30) ifNil: [Semaphore new]). "delay Semaphore"
	newArray at: 31 put: ((self specialObjectsArray at: 31) ifNil: [Semaphore new]). "user interrupt Semaphore"
	"Entries 32 - 34 unreferenced. Previously these contained prototype instances to be copied for fast initialization"
	newArray at: 32 put: nil. "was the prototype Float"
	newArray at: 33 put: nil. "was the prototype 4-byte LargePositiveInteger"
	newArray at: 34 put: nil. "was the prototype Point"
	newArray at: 35 put: #cannotInterpret:.
	newArray at: 36 put: nil. "was the prototype MethodContext"
	newArray at: 37 put: BlockClosure.
	newArray at: 38 put: nil. "was the prototype BlockContext"
	"array of objects referred to by external code"
	newArray at: 39 put: (self specialObjectsArray at: 39).	"external semaphores"
	newArray at: 40 put: nil. "Reserved for Mutex in Cog VMs"
	newArray at: 41 put: ((self specialObjectsArray at: 41) ifNil: [LinkedList new]). "Reserved for a LinkedList instance for overlapped calls in CogMT"
	newArray at: 42 put: ((self specialObjectsArray at: 42) ifNil: [Semaphore new]). "finalization Semaphore"
	newArray at: 43 put: LargeNegativeInteger.
	"External objects for callout.
	 Note: Written so that one can actually completely remove the FFI."
	newArray at: 44 put: (self at: #ExternalAddress ifAbsent: []).
	newArray at: 45 put: (self at: #ExternalStructure ifAbsent: []).
	newArray at: 46 put: (self at: #ExternalData ifAbsent: []).
	newArray at: 47 put: (self at: #ExternalFunction ifAbsent: []).
	newArray at: 48 put: (self at: #ExternalLibrary ifAbsent: []).
	newArray at: 49 put: #aboutToReturn:through:.
	newArray at: 50 put: #run:with:in:.
	"51 reserved for immutability message"
	newArray at: 51 put: #attemptToAssign:withIndex:.
	newArray at: 52 put: #(nil "nil => generic error" #'bad receiver'
							#'bad argument' #'bad index'
							#'bad number of arguments'
							#'inappropriate operation'  #'unsupported operation'
							#'no modification' #'insufficient object memory'
							#'insufficient C memory' #'not found' #'bad method'
							#'internal error in named primitive machinery'
							#'object may move' #'resource limit exceeded'
							#'object is pinned' #'primitive write beyond end of object').
	"53 to 55 are for Alien"
	newArray at: 53 put: (self at: #Alien ifAbsent: []).
	newArray at: 54 put: #invokeCallbackContext:. "use invokeCallback:stack:registers:jmpbuf: for old Alien callbacks."
	newArray at: 55 put: (self at: #UnsafeAlien ifAbsent: []).

	"Used to be WeakFinalizationList for WeakFinalizationList hasNewFinalization, obsoleted by ephemeron support."
	newArray at: 56 put: nil.

	"reserved for foreign callback process"
	newArray at: 57 put: (self specialObjectsArray at: 57 ifAbsent: []).

	newArray at: 58 put: #unusedBytecode.
	"59 reserved for Sista counter tripped message"
	newArray at: 59 put: #conditionalBranchCounterTrippedOn:.
	"60 reserved for Sista class trap message"
	newArray at: 60 put: #classTrapFor:.

	"Now replace the interpreter's reference in one atomic operation"
	self specialObjectsArray becomeForward: newArray! !

!SmalltalkImage methodsFor: 'external objects' stamp: 'ul 10/27/2014 16:36'!
registerExternalObject: anObject
	"Register the given object in the external objects array and return its index. If it is already there, just return its index."

	^ExternalObjectTable current registerExternalObject: anObject! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'ul 9/18/2011 15:27'!
removeAllUnSentMessages
	"Smalltalk removeAllUnSentMessages"
	"[Smalltalk unusedClasses do: [:c | (Smalltalk at: c) removeFromSystem]. 
	Smalltalk removeAllUnSentMessages > 0] whileTrue."
	"Remove all implementations of unsent messages."
	| sels n |
	sels := self systemNavigation allUnSentMessages.
	"The following should be preserved for doIts, etc"
	"needed even after #majorShrink is pulled"
	#(#compactSymbolTable #rebuildAllProjects #browseAllSelect:  #lastRemoval #scrollBarValue: vScrollBarValue: #scrollBarMenuButtonPressed: #withSelectionFrom: #to: #removeClassNamed: #dragon: #hilberts: #mandala: #web #test3 #factorial #tinyBenchmarks #benchFib #newDepth: #restoreAfter: #zapAllMethods #obsoleteClasses #removeAllUnSentMessages #abandonSources #removeUnreferencedKeys #reclaimDependents #zapOrganization #condenseChanges #browseObsoleteReferences #subclass:instanceVariableNames:classVariableNames:poolDictionaries:category: #methodsFor:stamp: #methodsFor:stamp:prior: #instanceVariableNames: #startTimerInterruptWatcher #unusedClasses )
		do: [:sel | sels
				remove: sel
				ifAbsent: []].
	"The following may be sent by perform: in dispatchOnChar..."
	(Smalltalk at: #ParagraphEditor) ifNotNilDo: [:paragraphEditor |
		(paragraphEditor classPool at: #CmdActions) asSet
			do: [:sel | sels
					remove: sel
					ifAbsent: []].
		(paragraphEditor classPool at: #ShiftCmdActions) asSet
			do: [:sel | sels
					remove: sel
					ifAbsent: []]].
	sels size = 0
		ifTrue: [^ 0].
	n := 0.
	self systemNavigation
		allBehaviorsDo: [:x | n := n + 1].
	'Removing ' , sels size printString , ' messages . . .'
		displayProgressFrom: 0
		to: n
		during: [:bar | 
			n := 0.
			self systemNavigation
				allBehaviorsDo: [:class | 
					bar value: (n := n + 1).
					sels
						do: [:sel | class basicRemoveSelector: sel]]].
	^ sels size! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'fbs 12/5/2013 08:58'!
removeChangeSets
	self deprecated: 'Use ChangeSet cleanUp: true'.
	ChangeSet cleanUp: true.! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 18:27'!
removeClassNamed: aName
	"DO NOT DEPRECATE - Invoked from fileouts. If there is currently a class in the system named aName, then remove it.  If anything untoward happens, report it in the Transcript.  "

	^globals removeClassNamed: aName! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'sd 9/29/2004 18:15'!
removeEmptyMessageCategories
	"Smalltalk removeEmptyMessageCategories"
	self garbageCollect.
	(ClassOrganizer allInstances copyWith: SystemOrganization)
		do: [:org | org removeEmptyCategories]! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/3/1999 22:22'!
removeFromShutDownList: aClass

	ShutDownList remove: aClass name ifAbsent: []! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/3/1999 22:22'!
removeFromStartUpList: aClass

	StartUpList remove: aClass name ifAbsent: []! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'ul 3/24/2011 08:29'!
removeObsoleteClassesFromCompactClassesArray

	self compactClassesArray doWithIndex: [ :each :index |
		(each notNil and: [
			each isObsolete and: [
			each instanceCount = 0 ] ]) ifTrue: [
				Smalltalk compactClassesArray at: index put: nil ] ].! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'sd 9/29/2004 18:26'!
removeSelector: descriptor 
	"Safely remove a selector from a class (or metaclass). If the
	class or the method doesn't exist anymore, never mind and
	answer nil.
	This method should be used instead of 'Class removeSelector:
	#method' to omit global class references."
	| class sel |
	class := self
				at: descriptor first
				ifAbsent: [^ nil].
	(descriptor size > 2
			and: [descriptor second == #class])
		ifTrue: [class := class class.
			sel := descriptor third]
		ifFalse: [sel := descriptor second].
	^ class removeSelector: sel! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 18:29'!
renameClassNamed: oldName as: newName
	"DO NOT DEPRECATE - Invoked from fileouts.  If there is currently a class in the system named oldName, then rename it to newName.  If anything untoward happens, report it in the Transcript.  "

	^globals renameClassNamed: oldName as: newName! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 20:02'!
renameClass: aClass as: newName 
	"Rename the class, aClass, to have the title newName."

	^globals renameClass: aClass as: newName! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 20:10'!
renameClass: aClass from: oldName
	"Rename the class, aClass, to have the title newName."

	^globals renameClass: aClass from: oldName! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 20:00'!
renamedClass: aClass from: oldName to: newName
	"Inform Smalltalk that aClass has been renamed"
	(Array with: StartUpList with: ShutDownList) do:[:list | | index |  
		index := list indexOf: oldName ifAbsent: [0].
		index > 0 ifTrue: [list at: index put: newName].
	].
! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:48'!
rootTable
	"Primitive. Answer a snapshot of the VMs root table. 
	Keep in mind that the primitive may itself cause GC."
	<primitive: 'primitiveRootTable'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:49'!
rootTableAt: index
	"Primitive. Answer the nth element of the VMs root table"
	<primitive: 'primitiveRootTableAt'>
	^nil! !

!SmalltalkImage methodsFor: 'command line' stamp: 'cmm 10/27/2013 18:40'!
runAndQuit: aBlock 
	self
		run: aBlock
		snapshotAs: nil
		andQuit: true! !

!SmalltalkImage methodsFor: 'command line' stamp: 'cmm 5/28/2015 21:09'!
run: aBlock
	[ [ (aBlock numArgs = 1 and: [ self arguments size > 1 ])
		ifTrue: [ "Allow a large, variable number of arguments to be passed as an Array to aBlock."
			aBlock value: self arguments ]
		ifFalse: [ aBlock valueWithEnoughArguments: self arguments ] ]
		on: ProgressInitiationException
		do:
			[ : pie | "Don't want to log these notifications."
			pie defaultAction ] ]
		on: Notification , Warning
		do:
			[ : noti | FileStream stdout
				 nextPutAll: DateAndTime now asString ;
				 space ;
				 nextPutAll: noti description ;
				 cr.
			noti resume ]
		on: SyntaxErrorNotification
		do:
			[ : err | FileStream stdout
				 nextPutAll: err errorCode ;
				 cr; flush.
			self isHeadless
				ifTrue: [ self snapshot: false andQuit: true ]
				ifFalse: [ err pass ] ]
		on: Error, MessageNotUnderstood, Halt
		do:
			[ : err | err printVerboseOn: FileStream stderr.
			FileStream stderr flush.
			self isHeadless
				ifTrue: [ self snapshot: false andQuit: true ]
				ifFalse: [ err pass ] ]! !

!SmalltalkImage methodsFor: 'command line' stamp: 'cmm 10/27/2013 18:40'!
run: aBlock snapshotAs: imageName andQuit: shouldQuit 
	self run: aBlock.
	imageName
		ifNil:
			[ shouldQuit ifTrue:
				[ self
					snapshot: false
					andQuit: shouldQuit ] ]
		ifNotNil:
			[ | correctName |
			correctName := (imageName endsWith: '.image')
				ifTrue: [ imageName ]
				ifFalse: [ imageName , '.image' ].
			(FileDirectory default fullNameFor: correctName) = self imageName
				ifTrue:
					[ self
						snapshot: true
						andQuit: shouldQuit ]
				ifFalse:
					[ self saveAs: imageName.
					"Avoid instant-quit when relaunching by resaving."
					shouldQuit ifTrue:
						[ self
							snapshot: true
							andQuit: shouldQuit ] ] ]! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'ar 3/9/2010 16:58'!
saveAs
	"Put up the 'saveAs' prompt, obtain a name, and save the image  under that new name."
	^self saveAs: self getFileNameFromUser.
! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'nice 12/27/2009 03:11'!
saveAsEmbeddedImage
	"Save the current state of the system as an embedded image"

	| dir newName newImageName oldImageSegDir haveSegs |
	dir := FileDirectory default.
	newName := UIManager default request: 'Select existing VM file'
				initialAnswer: (FileDirectory localNameFor: '').
	newName = '' ifTrue: [^Smalltalk].
	newName := FileDirectory baseNameFor: newName asFileName.
	newImageName := newName.
	(dir includesKey: newImageName) 
		ifFalse: 
			[^self 
				inform: 'Unable to find name ' , newName , ' Please choose another name.'].
	haveSegs := false.
	Smalltalk at: #ImageSegment
		ifPresent: 
			[:theClass | 
			(haveSegs := theClass instanceCount ~= 0) 
				ifTrue: [oldImageSegDir := theClass segmentDirectory]].
	self logChange: '----SAVEAS (EMBEDDED) ' , newName , '----' 
				, Date dateAndTimeNow printString.
	self imageName: (dir fullNameFor: newImageName) asSqueakPathName.
	LastImageName := self imageName.
	self closeSourceFiles.
	haveSegs 
		ifTrue: 
			[Smalltalk at: #ImageSegment
				ifPresent: 
					[:theClass | | newImageSegDir | 
					newImageSegDir := theClass segmentDirectory.	"create the folder"
					oldImageSegDir fileNames do: 
							[:theName | 
							"copy all segment files"

							newImageSegDir 
								copyFileNamed: oldImageSegDir pathName , FileDirectory slash , theName
								toFileNamed: theName]]].
	self 
		snapshot: true
		andQuit: true
		embedded: true! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'fbs 7/25/2013 07:19'!
saveAsNewVersion
	"Save the image/changes using the next available version number."
	"Smalltalk saveAsNewVersion"
	
	| newName changesName aName anIndex |
	aName := FileDirectory baseNameFor: (FileDirectory default localNameFor: self imageName).
	anIndex := aName lastIndexOf: FileDirectory dot asCharacter ifAbsent: [nil].
	(anIndex notNil and: [(aName copyFrom: anIndex + 1 to: aName size) isAllDigits])
		ifTrue:
			[aName := aName copyFrom: 1 to: anIndex - 1].

	newName := FileDirectory default nextNameFor: aName extension: FileDirectory imageSuffix.
	changesName := self fullNameForChangesNamed: newName.

	"Check to see if there is a .changes file that would cause a problem if we saved a new .image file with the new version number"
	(FileDirectory default fileOrDirectoryExists: changesName)
		ifTrue:
			[^ self inform:
'There is already .changes file of the desired name,
', newName, '
curiously already present, even though there is
no corresponding .image file.   Please remedy
manually and then repeat your request.'].

	(SourceFiles at: 2) ifNotNil:
		[self closeSourceFiles; "so copying the changes file will always work"
			saveChangesInFileNamed: (self fullNameForChangesNamed: newName)].
	self saveImageInFileNamed: (self fullNameForImageNamed: newName)


! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ar 3/9/2010 17:00'!
saveAsSuggesting: aName
	"Put up the 'saveAs' prompt, obtain a name, and save the image  under that new name."
	^self saveAs: (self getFileNameFromUserSuggesting: aName)! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'ar 3/9/2010 16:58'!
saveAs: newName
	"Save the image  under that new name."
	newName ifNil:[^ self].
	(SourceFiles at: 2) ifNotNil:
		[self closeSourceFiles; "so copying the changes file will always work"
			 saveChangesInFileNamed: (self fullNameForChangesNamed: newName)].
	self saveImageInFileNamed: (self fullNameForImageNamed: newName)! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'fbs 7/20/2013 23:28'!
saveChangesInFileNamed: aString
	| fullChangesName |
	fullChangesName := (FileDirectory default fullNameFor: aString).
	(FileDirectory default directoryNamed:(FileDirectory dirPathFor: fullChangesName )) assureExistence.
	FileDirectory default 
		copyFileWithoutOverwriteConfirmationNamed: self changesName 
		toFileNamed: fullChangesName.
	Smalltalk setMacFileInfoOn: fullChangesName.! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'tpr 12/15/2003 15:58'!
saveImageInFileNamed: aString
	| fullImageName |
	fullImageName := (FileDirectory default fullNameFor: aString).
	(FileDirectory default directoryNamed:(FileDirectory dirPathFor: fullImageName )) assureExistence.
	self
		changeImageNameTo: fullImageName;
		closeSourceFiles;
		openSourceFiles;  "so SNAPSHOT appears in new changes file"
		saveImageSegments;
		snapshot: true andQuit: false! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'nice 12/27/2009 03:11'!
saveImageSegments

	| haveSegs oldImageSegDir |
	haveSegs := false.
	Smalltalk at: #ImageSegment ifPresent: [:theClass | 
		(haveSegs := theClass instanceCount ~= 0) ifTrue: [
			oldImageSegDir := theClass segmentDirectory]].
	haveSegs ifTrue: [
		Smalltalk at: #ImageSegment ifPresent: [:theClass | | newImageSegDir |
			newImageSegDir := theClass segmentDirectory.	"create the folder"
			oldImageSegDir fileNames do: [:theName | "copy all segment files"
				| imageSegmentName |
				imageSegmentName := oldImageSegDir pathName, FileDirectory slash, theName.
				newImageSegDir 
					copyFileWithoutOverwriteConfirmationNamed: imageSegmentName
					toFileNamed: theName]]].
! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'sd 11/16/2003 14:12'!
saveSession
	self snapshot: true andQuit: false! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'nice 12/27/2009 03:10'!
send: startUpOrShutDown toClassesNamedIn: startUpOrShutDownList with: argument
	"Send the message #startUp: or #shutDown: to each class named in the list.
	The argument indicates if the system is about to quit (for #shutDown:) or if
	the image is resuming (for #startUp:).
	If any name cannot be found, then remove it from the list."

	| removals |
	removals := OrderedCollection new.
	startUpOrShutDownList do:
		[:name | | class |
		class := self at: name ifAbsent: [nil].
		class == nil
			ifTrue: [removals add: name]
			ifFalse: [class isInMemory ifTrue:
						[class perform: startUpOrShutDown with: argument]]].

	"Remove any obsolete entries, but after the iteration"
	startUpOrShutDownList removeAll: removals! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'JMM 1/27/2005 12:27'!
setGCBiasToGrowGCLimit: aNumber
	"Primitive. Indicate that the bias to grow logic should do a GC after aNumber Bytes"
	<primitive: 'primitiveSetGCBiasToGrowGCLimit'>
	^self primitiveFailed
"Example:
	Smalltalk setGCBiasToGrowGCLimit: 16*1024*1024.
"! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'JMM 1/27/2005 13:12'!
setGCBiasToGrow: aNumber
	"Primitive. Indicate that the GC logic should be bias to grow"
	<primitive: 'primitiveSetGCBiasToGrow'>
	^self primitiveFailed
"Example:
	Smalltalk setGCBiasToGrowGCLimit: 16*1024*1024.
	Smalltalk setGCBiasToGrow: 1.
"! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'eem 9/21/2013 07:39'!
setGCParameters
	"Adjust the VM's default GC parameters to avoid too much tenuring.
	 Maybe this should be left to the VM?"

	| proportion edenSize survivorSize averageObjectSize numObjects |
	proportion := 0.9. "tenure when 90% of pastSpace is full"
	edenSize := SmalltalkImage current vmParameterAt: 44.
	survivorSize := edenSize / 5.0. "David's paper uses 140Kb eden + 2 x 28kb survivor spaces; Spur uses the same ratios :-)"
	averageObjectSize := 8 * self wordSize. "a good approximation"
	numObjects := (proportion * survivorSize / averageObjectSize) rounded.
	SmalltalkImage current vmParameterAt: 6 put: numObjects  "tenure when more than this many objects survive the GC"! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:54'!
setGCSemaphore: semaIndex
	"Primitive. Indicate the GC semaphore index to be signaled on GC occurance."
	<primitive: 'primitiveSetGCSemaphore'>
	^self primitiveFailed
"Example:

	| index sema process |
	sema := Semaphore new.
	index := Smalltalk registerExternalObject: sema.
	Smalltalk setGCSemaphore: index.
	process := [
		[[true] whileTrue:[
			sema wait.
			Smalltalk beep.
		]] ensure:[
			Smalltalk setGCSemaphore: 0.
			Smalltalk unregisterExternalObject: sema.
		].
	] fork.
	process inspect.
"! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'MPH 10/24/2000 14:27'!
setMacFileInfoOn: aString
	"On Mac, set the file type and creator (noop on other platforms)"
	FileDirectory default
		setMacFileNamed: aString
		type: 'STch'
		creator: 'FAST'.! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'fbs 10/30/2013 22:31'!
setPlatformPreferences
	"Set some platform specific preferences on system startup"
	| platform specs |
	Preferences automaticPlatformSettings ifFalse:[^self].
	platform := self platformName.
	specs := 	#().
	platform = 'Win32' ifTrue:[
		specs := #()].
	platform = 'Mac OS' ifTrue:[
		specs := #()].
	specs do:[:tuple|
		Preferences setPreference: tuple first toValue: (tuple last == true).
	].
! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'topa 1/25/2014 22:47'!
shrink
	"Clean up everything we know how to clean up that can be restored."
	"Prepare unloading"
	self zapMVCprojects.
	Flaps disableGlobalFlaps: false.
	self unloadReloadablePackages.
	"Post-unload aggressive cleanup"
	Smalltalk cleanUp: true.
	SystemOrganization removeSystemCategory: 'UserObjects'.
	Smalltalk at: #ScheduledControllers ifPresent: [Smalltalk at: #ScheduledControllers put: nil].
	Smalltalk garbageCollect! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'cmm 8/31/2013 15:13'!
shrinkAndCleanDesktop
	World removeAllMorphs.
	self shrink.
	MorphicProject defaultFill: (Color gray: 0.9).
	World color: (Color gray: 0.9)! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'sd 11/16/2003 14:20'!
shutDown

	^ self closeSourceFiles! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'gk 2/23/2004 20:51'!
shutDownSound
	"No longer used in the release, but retained for backward compatibility."

	SoundService default shutDown
! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
signalLowSpace
	"Signal the low-space semaphore to alert the user that space is running low."

	LowSpaceSemaphore signal.! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'sd 11/16/2003 13:58'!
snapshotEmbeddedPrimitive
	<primitive: 247>
	^nil "indicates error writing embedded image file"! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'sd 11/16/2003 13:59'!
snapshotPrimitive
	"Primitive. Write the current state of the object memory on a file in the
	same format as the Smalltalk-80 release. The file can later be resumed,
	returning you to this exact state. Return normally after writing the file.
	Essential. See Object documentation whatIsAPrimitive."

	<primitive: 97>
	^nil "indicates error writing image file"! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ul 5/12/2013 22:28'!
snapshot: save andQuitWithExitCode: exitCode

	^self
		snapshot: save
		andQuit: true
		withExitCode: exitCode
		embedded: false! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'sd 11/16/2003 14:12'!
snapshot: save andQuit: quit
	^self snapshot: save andQuit: quit embedded: false! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ul 5/12/2013 22:29'!
snapshot: save andQuit: quit embedded: embeddedFlag

	^self	
		snapshot: save
		andQuit: quit
		withExitCode: nil
		embedded: embeddedFlag! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ul 5/12/2013 22:30'!
snapshot: save andQuit: quit withExitCode: exitCode embedded: embeddedFlag
	"Mark the changes file and close all files as part of #processShutdownList.
	If save is true, save the current state of this Smalltalk in the image file.
	If quit is true, then exit to the outer OS shell.
	If exitCode is not nil, then use it as exit code.
	The latter part of this method runs when resuming a previously saved image. This resume logic checks for a document file to process when starting up."

	| resuming msg |
	Object flushDependents.
	Object flushEvents.

	(SourceFiles at: 2) ifNotNil:[
		msg := String streamContents: [ :s |
			s nextPutAll: '----';
			nextPutAll: (save ifTrue: [ quit ifTrue: [ 'QUIT' ] ifFalse: [ 'SNAPSHOT' ] ]
							ifFalse: [quit ifTrue: [ 'QUIT/NOSAVE' ] ifFalse: [ 'NOP' ]]);
			nextPutAll: '----';
			print: Date dateAndTimeNow; space;
			nextPutAll: (FileDirectory default localNameFor: self imageName);
			nextPutAll: ' priorSource: ';
			print: LastQuitLogPosition ].
		self assureStartupStampLogged.
		save ifTrue: [ LastQuitLogPosition := (SourceFiles at: 2) setToEnd; position ].
		self logChange: msg.
		Transcript cr; show: msg
	].

	Smalltalk processShutDownList: quit.
	Cursor write show.
	save ifTrue: [resuming := embeddedFlag 
					ifTrue: [self snapshotEmbeddedPrimitive] 
					ifFalse: [self snapshotPrimitive].  "<-- PC frozen here on image file"
				resuming == false "guard against failure" ifTrue:
					["Time to reclaim segment files is immediately after a save"
					Smalltalk at: #ImageSegment
						ifPresent: [:theClass | theClass reclaimObsoleteSegmentFiles]]]
		ifFalse: [resuming := false].
	quit & (resuming == false) ifTrue: [
		exitCode
			ifNil: [ self quitPrimitive ]
			ifNotNil: [ self quitPrimitive: exitCode ] ].
	Cursor normal show.
	Smalltalk setGCParameters.
	resuming == true ifTrue: [Smalltalk clearExternalObjects].
	Smalltalk processStartUpList: resuming == true.
	resuming == true ifTrue:[
		self setPlatformPreferences.
		self recordStartupStamp].
	Project current wakeUpTopWindow.
	"Now it's time to raise an error"
	resuming == nil ifTrue: [self error:'Failed to write image file (disk full?)'].
	^ resuming! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'sd 9/24/2003 12:43'!
sourceFileVersionString

	^ SourceFileVersionString! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'sd 9/24/2003 12:42'!
sourceFileVersionString: aString

	SourceFileVersionString := aString! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'cmm 3/14/2014 16:50'!
sourcesName
	"Answer the full path to the version-stable source code if it exists, otherwise the desired location"
	^ self locateSourcesEntry
		ifNil: [ self vmPath , SourceFileVersionString , FileDirectory dot , 'sources' ]
		ifNotNil: [ : entry | entry fullName ]! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialNargsAt: anInteger 
	"Answer the number of arguments for the special selector at: anInteger."

	^ (self specialObjectsArray at: 24) at: anInteger * 2! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialObjectsArray  "Smalltalk specialObjectsArray at: 1"
	<primitive: 129>
	^ self primitiveFailed! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialSelectorAt: anInteger 
	"Answer the special message selector from the interleaved specialSelectors array."

	^ (self specialObjectsArray at: 24) at: anInteger * 2 - 1! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialSelectorSize
	"Answer the number of special selectors in the system."

	^ (self specialObjectsArray at: 24) size // 2! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialSelectors
	"Used by SystemTracer only."

	^SpecialSelectors! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'mtf 12/20/2010 16:23'!
squeakErrorFileName
	"Answer the file name used to dump error reports"

	^ Preferences debugLogTimestamp
		ifTrue: ['SqueakDebug-' , Time totalSeconds printString , '.log']
		ifFalse: ['SqueakDebug.log'].! !

!SmalltalkImage methodsFor: 'vm profiling' stamp: 'sd 9/24/2003 11:55'!
startProfiling
	"Start profiling the virtual machine."

	<primitive: 252>
! !

!SmalltalkImage methodsFor: 'vm profiling' stamp: 'sd 9/24/2003 11:55'!
stopProfiling
	"Stop profiling the virtual machine."

	<primitive: 253>
! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'tk 3/7/2000 18:40'!
storeDataOn: aDataStream
	"I don't get stored.  Use a DiskProxy"

	self error: 'use a DiskProxy to store me'! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 3/3/2016 10:31'!
supportsMultipleBytecodeSets
	"Answer whether the VM supports multiple bytecodeSets."
	"SmalltalkImage current supportsMultipleBytecodeSets"

	^(self vmParameterAt: 65)
		ifNil: [false]
		ifNotNil:
			[:param| "In older VMs this is a boolean answering the vm-internal MULTIPLE_BYTECODE_SETS define"
			 param isInteger "In newer VMs it is a set of integer flags, bit 0 of which is the vm-internal MULTIPLE_BYTECODE_SETS define"
				ifTrue: [param anyMask: 1]
				ifFalse: [param]]! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 9/30/2015 10:58'!
supportsQueueingFinalization
	"Answer whether the VM queues individual weak arrays for finalization, instead
	 of signalling the finalization semaphore once for all arrays and having the
	 WeakRegistry mechanism finalize all weak arrays, whether they need to or not."
	"SmalltalkImage current supportsQueueingFinalization"

	^(self vmParameterAt: 48) anyMask: 16! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 9/30/2015 11:03'!
supportsQueueingFinalization: aBoolean
	"Determine whether the VM queues individual weak arrays for finalization, instead
	 of signalling the finalization semaphore once for all arrays and having the
	 WeakRegistry mechanism finalize all weak arrays, whether they need to or not.
	 This flag persists across snapshots, stored in the image header."
	"SmalltalkImage current supportsQueueingFinalization: true"

	self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 16) + (aBoolean ifTrue: [16] ifFalse: [0])! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 3/3/2016 10:31'!
supportsReadOnlyObjects
	"Answer whether the VM observes the per-object read-only flag and consequently aborts
	 writes to inst vars of, and fails primitives that attempt to modify, read-only objects."
	"SmalltalkImage current supportsReadOnlyObjects"

	^(self vmParameterAt: 65)
		ifNil: [false]
		ifNotNil:
			[:param| "In older VMs this is a boolean answering the vm-internal MULTIPLE_BYTECODE_SETS define"
			 param isInteger "In newer VMs it is a set of integer flags, bit 1 of which is the vm-internal IMMUTABILITY define"
				ifTrue: [param anyMask: 2]
				ifFalse: [false]]! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'nk 7/29/2004 10:09'!
systemInformationString
	"Identify software version"
	^ SystemVersion current version, String cr, self lastUpdateString, String cr, self currentChangeSetString

"
	(eToySystem := self at: #EToySystem ifAbsent: [nil]) ifNotNil:
		[aString := aString, '
Squeak-Central version: ', eToySystem version, ' of ', eToySystem versionDate]."! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: 'sd 7/2/2003 21:45'!
textMarkerForShortReport

	^  'Since last view	'! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'fbs 7/23/2013 09:26'!
timeStamp: aStream 
	"Writes system version and current time on stream aStream."

	| dateTime |
	dateTime := Time dateAndTimeNow.
	aStream nextPutAll: 'From ', self datedVersion, ' [', self lastUpdateString, '] on ', (dateTime at: 1) printString,
						' at ', (dateTime at: 2) printString! !

!SmalltalkImage methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:05'!
traitNames
	"Answer a SortedCollection of all traits (not including class-traits) names."

	^globals traitNames! !

!SmalltalkImage methodsFor: 'modules' stamp: 'sd 6/28/2003 17:38'!
unbindExternalPrimitives
	"Primitive. Force all external primitives to be looked up again afterwards. Since external primitives that have not found are bound for fast failure this method will force the lookup of all primitives again so that after adding some plugin the primitives may be found."
	<primitive: 570>
	"Do nothing if the primitive fails for compatibility with older VMs"
! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'ar 9/6/2010 11:34'!
unloadFonts	"Smalltalk unloadFonts"

	"Unloads all fonts except the style specified initially."

	| fontName fontSize textStyle |
	"Three good choices here:
		* Atlanta 11: Absolutely minimal font.
		* Accuny 10: A few extra sizes over Atlanta.
		* Bitmap DejaVu Sans 9: Nicest fonts but relatively large.
	"
	fontName := #'Bitmap DejaVu Sans'.
	fontSize := 9.

	"Nuke everything other than the desired text style"
	textStyle := TextStyle named: fontName.
	TextStyle defaultFamilyNames do:[:styleName| 
		TextConstants at: styleName put: textStyle
	].
	TextStyle actualTextStyles keys do:[:styleName| 
		TextConstants removeKey: styleName
	].
	TextConstants at: fontName put: textStyle.

	"Use it everywhere"
	Preferences setDefaultFonts: {
		{#setSystemFontTo:.		fontName.	fontSize}.
		{#setListFontTo:.			fontName.	fontSize}.
		{#setFlapsFontTo:.			fontName.	fontSize}.
		{#setEToysFontTo:.			fontName.	fontSize}.
		{#setEToysTitleFontTo:	.	fontName.	fontSize}.
		{#setPaintBoxButtonFontTo:. fontName.	fontSize}.
		{#setMenuFontTo:.			fontName.	fontSize}.
		{#setWindowTitleFontTo:.	fontName.	fontSize}.
		{#setBalloonHelpFontTo:.	fontName.	fontSize}.
		{#setCodeFontTo:.			fontName.	fontSize}.
		{#setButtonFontTo:.		fontName.	fontSize}.
	}.

	"Clean out afterwards"
	Smalltalk at: #TextEditor ifPresent:[:aClass| aClass initialize].
	Smalltalk at: #SmalltalkEditor ifPresent:[:aClass| aClass initialize].
	Smalltalk at: #SystemProgressMorph ifPresent:[:aClass| aClass reset].
	Smalltalk at: #TTCFont ifPresent:[:aClass| aClass registry removeAll].
	Smalltalk at: #Flaps ifPresent:[:aClass | aClass disableGlobalFlaps: false].
	Smalltalk at: #TTFontDescription ifPresent:[:aClass| aClass clearDefault].
	Smalltalk at: #TTFontDescription ifPresent:[:aClass| aClass clearDescriptions].
! !

!SmalltalkImage methodsFor: 'modules' stamp: 'sd 6/27/2003 23:49'!
unloadModule: aString
	"Primitive. Unload the given module.
	This primitive is intended for development only since some
	platform do not implement unloading of DLL's accordingly.
	Also, the mechanism for unloading may not be supported
	on all platforms."
	<primitive: 571>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'topa 1/26/2014 00:54'!
unloadReloadablePackages
	"Unload packages which can be reloaded."

	"One special case to make Services-Base unload"
	Smalltalk
		at: #ServiceRegistry
		ifPresent: [ : aClass | SystemChangeNotifier uniqueInstance noMoreNotificationsFor: aClass ].

	#('CompressionTests'
'CollectionsTests'
'GraphicsTests'
'BalloonTests'
'HelpSystem-Tests'
'KernelTests'
'MorphicExtrasTests'
'MorphicTests'
'MultilingualTests'
'NetworkTests'
'ToolsTests'
'TraitsTests'
'SystemChangeNotification-Tests'
'VersionNumberTests'
'ST80Tests'
"'Tests'" "Currently not unloadable"

'FlexibleVocabularies'
'ReleaseBuilder'
'ScriptLoader'
'UpdateStream'
'CommandLine'
'45Deprecated'
'311Deprecated'
'39Deprecated'
'Universes'
'SMLoader'
'SMBase'
'Services-Base'
'Nebraska'
'ToolBuilder-MVC'
'ST80Tools'
'ST80'
"'EToys'" "Currently not unloadable"
"'PreferenceBrowser'" "Etoys needs them, hence not unloadable"
"'Protocols'" "Etoys needs them, hence not unloadable"
'XML-Parser'
'Help-Squeak-TerseGuide'
'Help-Squeak-Project'
'HelpSystem-Core'
'SystemReporter'
'SUnitGUI'
'VersionNumber' ) do: [ : pkgName | Installer mc unload: pkgName ].
	"Traits use custom unload"
	Smalltalk
		at: #Trait
		ifPresent: [ : aClass | aClass unloadTraits ]! !

!SmalltalkImage methodsFor: 'external objects' stamp: 'ul 10/27/2014 16:36'!
unregisterExternalObjects: aCollection
	"Unregister the given objects from the external objects array. Do nothing if they aren't registered."

	ExternalObjectTable current unregisterExternalObjects: aCollection! !

!SmalltalkImage methodsFor: 'external objects' stamp: 'ul 10/27/2014 16:36'!
unregisterExternalObject: anObject
	"Unregister the given object in the external objects array. Do nothing if it isn't registered."

	ExternalObjectTable current unregisterExternalObject: anObject! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'sd 4/29/2003 19:06'!
unusedClasses
	"Enumerates all classes in the system and returns a list of those that are 
	apparently unused. A class is considered in use if it (a) has subclasses 
	or (b) is referred to by some method or (c) has its name in use as a 
	literal. "
	"Smalltalk unusedClasses asSortedCollection"
	^ self systemNavigation allUnusedClassesWithout: {{}. {}}! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ul 6/17/2011 12:43'!
useUpMemory
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemory"

	| lst |
	lst := nil.
	[ lst := Link nextLink: lst ] repeat! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'di 8/18/2000 21:15'!
useUpMemoryWithArrays 
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemoryWithArrays"

	| b |  "First use up most of memory."
	b := String new: self bytesLeft - self lowSpaceThreshold - 100000.
	b := b.  "Avoid unused value warning"
	(1 to: 10000) collect: [:i | Array new: 10000]! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'di 8/18/2000 16:49'!
useUpMemoryWithContexts 
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemoryWithContexts"

	self useUpMemoryWithContexts! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'di 8/18/2000 16:50'!
useUpMemoryWithTinyObjects 
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemoryWithTinyObjects"

	| b |  "First use up most of memory."
	b := String new: self bytesLeft - self lowSpaceThreshold - 100000.
	b := b.  "Avoid unused value warning"
	(1 to: 10000) collect: [:i | BitBlt new]! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'sd 4/17/2003 21:01'!
verifyChanges		"Smalltalk verifyChanges"
	"Recompile all methods in the changes file."
	self systemNavigation allBehaviorsDo: [:class | class recompileChanges].
! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'nb 6/17/2003 12:25'!
verifyMorphicAvailability
	"If Morphic is available, return true; if not, put up an informer and return false"
	self hasMorphic ifFalse:
		[Beeper beep.
		self inform: 'Sorry, Morphic must
be present to use this feature'.
		^ false].
	^ true! !

!SmalltalkImage methodsFor: 'image' stamp: 'md 5/16/2006 12:34'!
version
	"Answer the version of this release."

	^SystemVersion current version! !

!SmalltalkImage methodsFor: 'vm' stamp: 'dtl 4/11/2010 11:39'!
versionLabel
	"Answer a string corresponding to the version of virtual machine. This
	represents the version level of the Smalltalk source code (interpreter
	and various plugins) that is translated to C by a CCodeGenerator,  in
	addition to the external platform source code, typically written in C and
	managed separately for each platform.
	
	This invokes an optional primitive that may not be available on all virtual
	machines. See also vmVersion, which answers a string identifying the image
	from which virtual machine sources were generated."

	"Smalltalk vm versionLabel"

	<primitive: 'primitiveVMVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitiveVMVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'tk 10/20/2000 11:35'!
veryDeepCopyWith: deepCopier
	"Return self.  I can't be copied.  Do not record me."! !

!SmalltalkImage methodsFor: 'vm' stamp: 'dtl 4/11/2010 11:15'!
vm
	"Answer the object to query about virtual machine."
	
	^self! !

!SmalltalkImage methodsFor: 'command line' stamp: 'mtf 12/16/2010 11:22'!
vmFileName
	"Answer the absolute file name of the the Smalltalk virtual machine"
	"Smalltalk vmFileName"

	^ (FilePath pathName: (self primVmFileName) isEncoded: true) asSqueakPathName.
! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'eem 3/3/2016 11:00'!
vmParameterAt: parameterIndex
	"parameterIndex is a positive integer corresponding to one of the VM's internal parameter/metric registers.
	 Answer with the current value of that register.  Fail if parameterIndex has no corresponding register.
	 VM parameters are numbered as follows:
		1	byte size of old-space (read-only)
		2	byte size of young-space (read-only)
		3	byte size of object memory (read-only)
		4	allocationCount (read-only; nil in Cog VMs)
		5	allocations between GCs (read-write; nil in Cog VMs)
		6	survivor count tenuring threshold (read-write)
		7	full GCs since startup (read-only)
		8	total milliseconds in full GCs since startup (read-only)
		9	incremental GCs since startup (read-only; scavenging GCs on Spur)
		10	total milliseconds in incremental/scavenging GCs since startup (read-only)
		11	tenures of surving objects since startup (read-only)
		12-19 specific to the translating VM (nil in Cog VMs)
		20	utc microseconds at VM start-up (actually at time initialization, which precedes image load) (newer Cog VMs only).
		21	root (remembered) table size (read-only)
		22	root (remembered) table overflows since startup (read-only)
		23	bytes of extra memory to reserve for VM buffers, plugins, etc.
		24	memory threshold above which to shrink object memory (read-write)
		25	ammount to grow by when growing object memory (read-write)
		26	interruptChecksEveryNms - force an ioProcessEvents every N milliseconds (read-write)
		27	number of times mark loop iterated for current IGC/FGC (read-only) includes ALL marking
		28	number of times sweep loop iterated for current IGC/FGC (read-only)
		29	number of times make forward loop iterated for current IGC/FGC (read-only)
		30	number of times compact move loop iterated for current IGC/FGC (read-only)
		31	number of grow memory requests (read-only)
		32	number of shrink memory requests (read-only)
		33	number of root table entries used for current IGC/FGC (read-only)
		34	number of allocations done before current IGC/FGC (read-only)
		35	number of survivor objects after current IGC/FGC (read-only)
		36	millisecond clock when current IGC/FGC completed (read-only)
		37	number of marked objects for Roots of the world, not including Root Table entries for current IGC/FGC (read-only)
		38	milliseconds taken by current IGC (read-only)
		39	Number of finalization signals for Weak Objects pending when current IGC/FGC completed (read-only)
		40	BytesPerWord for this image
		41	imageFormatVersion for the VM
		42	number of stack pages in use (Cog Stack VM only, otherwise nil)
		43	desired number of stack pages (stored in image file header, max 65535; Cog VMs only, otherwise nil)
		44	size of eden, in bytes (Cog VMs only, otherwise nil)
		45	desired size of eden, in bytes (stored in image file header; Cog VMs only, otherwise nil)
		46	size of machine code zone, in bytes (stored in image file header; Cog JIT VM only, otherwise nil)
		47	desired size of machine code zone, in bytes (applies at startup only, stored in image file header; Cog JIT VM only)
		48	various properties of the Cog VM as an integer encoding an array of bit flags.
			Bit 0: tells the VM that the image's Process class has threadId as its 5th inst var (after nextLink, suspendedContext, priority & myList)
			Bit 1: on Cog JIT VMs asks the VM to set the flag bit in interpreted methods
			Bit 2: if set, preempting a process puts it to the head of its run queue, not the back,
					i.e. preempting a process by a higher priority one will not cause the preempted process to yield
						to others at the same priority.
			Bit 3: in a muilt-threaded VM, if set, the Window system will only be accessed from the first VM thread
			Bit 4: in a Spur vm, if set, causes weaklings and ephemerons to be queued individually for finalization
		49	the size of the external semaphore table (read-write; Cog VMs only)
		50-51 reserved for VM parameters that persist in the image (such as eden above)
		52	root (remembered) table maximum size (read-only)
		53	the number of oldSpace segments (Spur only, otherwise nil)
		54	total size of free old space (Spur only, otherwise nil)
		55	ratio of growth and image size at or above which a GC will be performed post scavenge (Spur only, otherwise nil)
		56	number of process switches since startup (read-only)
		57	number of ioProcessEvents calls since startup (read-only)
		58	number of forceInterruptCheck (Cog VMs) or quickCheckInterruptCalls (non-Cog VMs) calls since startup (read-only)
		59	number of check event calls since startup (read-only)
		60	number of stack page overflows since startup (read-only; Cog VMs only)
		61	number of stack page divorces since startup (read-only; Cog VMs only)
		62	number of machine code zone compactions since startup (read-only; Cog VMs only)
		63	milliseconds taken by machine code zone compactions since startup (read-only; Cog VMs only)
		64	current number of machine code methods (read-only; Cog VMs only)
		65	In newer Cog VMs a set of flags describing VM features,
				if non-zero bit 0 implies multiple bytecode set support;
				if non-zero bit 0 implies read-only object support
			(read-only; Cog VMs only; nil in older Cog VMs, a boolean answering multiple bytecode support in not so old Cog VMs)
		66	the byte size of a stack page in the stack zone  (read-only; Cog VMs only)
		67	the maximum allowed size of old space in bytes, 0 implies no internal limit (Spur VMs only).
		68 - 69 reserved for more Cog-related info
		70	the value of VM_PROXY_MAJOR (the interpreterProxy major version number)
		71	the value of VM_PROXY_MINOR (the interpreterProxy minor version number)"

	<primitive: 254>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'eem 3/3/2016 10:59'!
vmParameterAt: parameterIndex put: newValue
	"parameterIndex is a positive integer corresponding to one of the VM's internal
	parameter/metric registers.  Store newValue (a positive integer) into that
	register and answer with the previous value that was stored there.
	Fail if newValue is out of range, if parameterIndex has no corresponding
	register, or if the corresponding register is read-only.

	As of early 2016 the parameters which can be set are
		5	allocations between GCs (read-write; nil in Cog VMs)
		6	survivor count tenuring threshold (read-write)
		23	bytes of extra memory to reserve for VM buffers, plugins, etc.
		24	memory threshold above whichto shrink object memory (read-write)
		25	memory headroom when growing object memory (read-write)
		26	interruptChecksEveryNms - force an ioProcessEvents every N milliseconds (read-write)
		43	desired number of stack pages (stored in image file header, max 65535; Cog VMs only, otherwise nil)
		45	desired size of eden, in bytes (stored in image file header; Cog VMs only, otherwise nil)
		47	desired size of machine code zone, in bytes (applies at startup only, stored in image file header; Cog JIT VM only)
		48	various properties of the Cog VM as an integer encoding an array of bit flags.
			Bit 0: tells the VM that the image's Process class has threadId as its 5th inst var (after nextLink, suspendedContext, priority & myList)
			Bit 1: on Cog JIT VMs asks the VM to set the flag bit in interpreted methods
			Bit 2: if set, preempting a process puts it to the head of its run queue, not the back,
					i.e. preempting a process by a higher priority one will not cause the preempted process to yield
						to others at the same priority.
			Bit 3: in a muilt-threaded VM, if set, the Window system will only be accessed from the first VM thread
			Bit 4: in a Spur vm, if set, causes weaklings and ephemerons to be queued individually for finalization
		49	the size of the external semaphore table (read-write; Cog VMs only)
		55	ratio of growth and image size at or above which a GC will be performed post scavenge (Spur only, otherwise nil)
		67	the maximum allowed size of old space in bytes, 0 implies no internal limit (Spur only)."

	<primitive: 254>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'command line' stamp: 'nice 3/7/2010 02:29'!
vmPath
	"Answer the path for the directory containing the Smalltalk virtual machine. Return the 	empty string if this primitive is not implemented."
	"Smalltalk vmPath"

	^ (FilePath pathName: (self primVmPath) isEncoded: true) asSqueakPathName.
! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: 'dtl 1/29/2016 00:00'!
vmStatisticsReportString
	"(Workspace new contents: Smalltalk vmStatisticsReportString)
		openLabel: 'VM Statistics'"
	"StringHolderView
		open: (StringHolder new contents: Smalltalk vmStatisticsReportString)
		label: 'VM Statistics'"

	| params oldSpaceEnd youngSpaceEnd memorySize fullGCs fullGCTime incrGCs incrGCTime tenureCount upTime upTime2 fullGCs2 fullGCTime2 incrGCs2 incrGCTime2 tenureCount2 str freeSize youngSize used |
	params := self getVMParameters.
	oldSpaceEnd			:= params at: 1.
	youngSpaceEnd		:= params at: 2.
	memorySize			:= params at: 3.
	fullGCs				:= params at: 7.
	fullGCTime			:= params at: 8.
	incrGCs				:= params at: 9.
	incrGCTime			:= params at: 10.
	tenureCount			:= params at: 11.
	upTime := (params at: 20) ~= 0 "utcMicrosecondClock at startupp in later Spur VMs"
				ifTrue: [Time utcMicrosecondClock - (params at: 20) + 500 // 1000]
				ifFalse: [Time eventMillisecondClock]. "Fall back on old microsecond clock; Good for 47.5 days"

	str := WriteStream on: (String new: 700).
	str	nextPutAll: 'uptime			';
		print: (upTime / 1000 / 60 // 60); nextPut: $h;
		print: (upTime / 1000 / 60 \\ 60) asInteger; nextPut: $m;
		print: (upTime / 1000 \\ 60) asInteger; nextPut: $s; cr.

	str	nextPutAll: 'memory		';
		nextPutAll: memorySize asStringWithCommas; nextPutAll: ' bytes'; cr.
	str	nextPutAll:	'	old			';
		nextPutAll: oldSpaceEnd asStringWithCommas; nextPutAll: ' bytes (';
		print: (oldSpaceEnd / memorySize * 100) maxDecimalPlaces: 1; nextPutAll: '%)'; cr.

	params size >= 54 ifTrue: [ "Spur object memory"
		youngSize := params at: 44.
		str	nextPutAll: '	young		';
			nextPutAll: youngSize asStringWithCommas; nextPutAll: ' bytes (';
			print: youngSize / memorySize * 100 maxDecimalPlaces: 1; nextPutAll: '%)'; cr.
		youngSize := youngSpaceEnd. "used eden"
		freeSize := (params at: 54) + (params at: 44) - youngSize.
		used := youngSize + oldSpaceEnd - freeSize.
		str	nextPutAll: '	used		';
			nextPutAll: used asStringWithCommas; nextPutAll: ' bytes (';
			print: used / memorySize * 100 maxDecimalPlaces: 1; nextPutAll: '%)'; cr.
		str	nextPutAll: '	free		';
			nextPutAll: freeSize asStringWithCommas; nextPutAll: ' bytes (';
			print: freeSize / memorySize * 100 maxDecimalPlaces: 1; nextPutAll: '%)'; cr
	] ifFalse: [ "Earlier VM and V3 object memory"
		str	nextPutAll: '	young		';
			nextPutAll: (youngSpaceEnd - oldSpaceEnd) asStringWithCommas; nextPutAll: ' bytes (';
			print: (youngSpaceEnd - oldSpaceEnd / memorySize * 100) maxDecimalPlaces: 1; nextPutAll: '%)'; cr.
		str	nextPutAll: '	used		';
			nextPutAll: youngSpaceEnd asStringWithCommas; nextPutAll: ' bytes (';
			print: (youngSpaceEnd / memorySize * 100) maxDecimalPlaces: 1; nextPutAll: '%)'; cr.
		str	nextPutAll: '	free		';
			nextPutAll: (memorySize - youngSpaceEnd) asStringWithCommas; nextPutAll: ' bytes (';
			print: (memorySize - youngSpaceEnd / memorySize * 100) maxDecimalPlaces: 1; nextPutAll: '%)'; cr.
	].

	str	nextPutAll: 'GCs			';
		nextPutAll: (fullGCs + incrGCs) asStringWithCommas.
	fullGCs + incrGCs > 0 ifTrue: [
		str
			nextPutAll: ' ('; 
			print: (upTime / (fullGCs + incrGCs)) maxDecimalPlaces: 1; 
			nextPutAll: ' ms between GCs)'
	].
	str cr.
	str	nextPutAll: '	full			';
		nextPutAll: fullGCs asStringWithCommas; nextPutAll: ' totalling '; nextPutAll: fullGCTime asStringWithCommas; nextPutAll: ' ms (';
		print: (fullGCTime / upTime * 100) maxDecimalPlaces: 1;
		nextPutAll: '% uptime)'.
	fullGCs = 0 ifFalse:
		[str	nextPutAll: ', avg '; print: (fullGCTime / fullGCs) maxDecimalPlaces: 1; nextPutAll: ' ms'].
	str	cr.
	str	nextPutAll: '	incr			';
		nextPutAll: incrGCs asStringWithCommas; nextPutAll: ' totalling '; nextPutAll: incrGCTime asStringWithCommas; nextPutAll: ' ms (';
		print: (incrGCTime / upTime * 100) maxDecimalPlaces: 1;
		nextPutAll: '% uptime), avg '; print: (incrGCTime / incrGCs) maxDecimalPlaces: 1; nextPutAll: ' ms'; cr.
	str	nextPutAll: '	tenures		';
		nextPutAll: tenureCount asStringWithCommas.
	tenureCount = 0 ifFalse:
		[str nextPutAll: ' (avg '; print: incrGCs // tenureCount; nextPutAll: ' GCs/tenure)'].
	str	cr.

LastStats ifNil: [LastStats := Array new: 6]
ifNotNil: [
	upTime2 := upTime - (LastStats at: 1).
	fullGCs2 := fullGCs - (LastStats at: 2).
	fullGCTime2 := fullGCTime - (LastStats at: 3).
	incrGCs2 := incrGCs - (LastStats at: 4).
	incrGCTime2 := incrGCTime - (LastStats at: 5).
	tenureCount2 := tenureCount - (LastStats at: 6).

	str	nextPutAll: self textMarkerForShortReport ;
		nextPutAll: (fullGCs2 + incrGCs2) asStringWithCommas.
	fullGCs2 + incrGCs2 > 0 ifTrue: [
		str
			nextPutAll: ' ('; 
			print: upTime2 // (fullGCs2 + incrGCs2); 
			nextPutAll: ' ms between GCs)'.
	].
	str cr.
	str	nextPutAll: '	uptime		'; print: (upTime2 / 1000.0) maxDecimalPlaces: 1; nextPutAll: ' s'; cr.
	str	nextPutAll: '	full			';
		nextPutAll: fullGCs2 asStringWithCommas; nextPutAll: ' totalling '; nextPutAll: fullGCTime2 asStringWithCommas; nextPutAll: ' ms (';
		print: (fullGCTime2 / upTime2 * 100) maxDecimalPlaces: 1;
		nextPutAll: '% uptime)'.
	fullGCs2 = 0 ifFalse:
		[str	nextPutAll: ', avg '; print: (fullGCTime2 / fullGCs2) maxDecimalPlaces: 1; nextPutAll: ' ms'].
	str	cr.
	str	nextPutAll: '	incr			';
		nextPutAll: incrGCs2 asStringWithCommas; nextPutAll: ' totalling '; nextPutAll: incrGCTime2 asStringWithCommas; nextPutAll: ' ms (';
		print: (incrGCTime2 / upTime2 * 100) maxDecimalPlaces: 1;
		nextPutAll: '% uptime), avg '.
	incrGCs2 > 0 ifTrue: [
		 str print: (incrGCTime2 / incrGCs2) maxDecimalPlaces: 1; nextPutAll: ' ms'
	].
	str cr.
	str	nextPutAll: '	tenures		';
		nextPutAll: tenureCount2 asStringWithCommas.
	tenureCount2 = 0 ifFalse:
		[str nextPutAll: ' (avg '; print: incrGCs2 // tenureCount2; nextPutAll: ' GCs/tenure)'].
	str	cr.
].
	LastStats at: 1 put: upTime.
	LastStats at: 2 put: fullGCs.
	LastStats at: 3 put: fullGCTime.
	LastStats at: 4 put: incrGCs.
	LastStats at: 5 put: incrGCTime.
	LastStats at: 6 put: tenureCount.

	^ str contents
! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: 'fbs 7/25/2013 07:19'!
vmStatisticsShortString
	"Convenience item for access to recent statistics only"
	"StringHolderView open: (StringHolder new contents: Smalltalk vmStatisticsShortString)
		label: 'VM Recent Statistics'"

	^ (ReadStream on: self vmStatisticsReportString) upToAll: 'Since'; nextLine; upToEnd
! !

!SmalltalkImage methodsFor: 'system attributes' stamp: 'fbs 7/25/2013 07:19'!
vmVersion	
	"Return a string identifying the interpreter version"
	"Smalltalk vmVersion"

	^self getSystemAttribute: 1004! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'eem 8/28/2009 14:52'!
voidCogVMState
	"Void any internal caches the VM maintains other than the method lookup caches.
	 These comprise
		- the stack zone, where method activations are stored, and
		- the machine code zone, where the machine code form of CompiledMethods is held."
	<primitive: 214>
	^self primitiveFailed

	"Time millisecondsToRun: [Smalltalk voidCogVMState]"
	"(1 to: 10) collect: [:ign| Time millisecondsToRun: [Smalltalk voidCogVMState]]"! !

!SmalltalkImage methodsFor: 'os' stamp: 'nice 3/7/2010 01:20'!
windowSystemName			
	"Return the name of the window system currently being used for display."
	
	"Smalltalk os windowSystemName"

	^self getSystemAttribute: 1005! !

!SmalltalkImage methodsFor: 'image' stamp: 'fbs 7/20/2013 23:27'!
wordSize
	"Answer the size in bytes of an object pointer or word in the object memory.
	The value does not change for a given image, but may be modified by a SystemTracer
	when converting the image to another format. The value is cached in WordSize to
	avoid the performance overhead of repeatedly consulting the VM."

	"Smalltalk wordSize"

	^ WordSize ifNil: [WordSize := [self vmParameterAt: 40] on: Error do: [4]]! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'adrian-lienhard 5/27/2009 21:28'!
writeRecentCharacters: nCharacters toFileNamed: aFilename
	"Schedule an editable text view on the last n characters of changes."
	| changes |
	changes := SourceFiles at: 2.
	changes setToEnd; skip: nCharacters negated.
	(FileStream newFileNamed: aFilename) nextPutAll: (changes next: nCharacters); close; open; edit! !

!SmalltalkImage methodsFor: 'sources, changes log' stamp: 'ar 9/27/2005 22:38'!
writeRecentToFile
	"Smalltalk writeRecentToFile"
	| numChars aDirectory aFileName |
	aDirectory := FileDirectory default.
	aFileName := Utilities
				keyLike: 'squeak-recent.01'
				withTrailing: '.log'
				satisfying: [:aKey | (aDirectory includesKey: aKey) not].
	numChars := ChangeSet getRecentLocatorWithPrompt: 'copy logged source as far back as...'.
	numChars
		ifNotNil: [self writeRecentCharacters: numChars toFileNamed: aFileName]! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'dtl 4/30/2016 18:19'!
zapAllOtherProjects 
	"Smalltalk zapAllOtherProjects"

	Project current removeAllOtherProjects
! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'dtl 4/30/2016 18:14'!
zapMVCprojects
	"Smalltalk zapMVCprojects"

	(Smalltalk classNamed: #MVCProject)
		ifNotNilDo: [:mvc | mvc removeProjectsFromSystem]
! !

Object subclass: #SpaceTally
	instanceVariableNames: 'results'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Tools'!

!SpaceTally commentStamp: 'sd 6/20/2003 22:31' prior: 0!
I'm responsible to help getting information about system space usage. The information I compute is represented by a spaceTallyItem

try something like: 

((SpaceTally new spaceTally: (Array with: TextMorph with: Point)) 
	asSortedCollection: [:a :b | a spaceForInstances > b spaceForInstances]) 

SpaceTally new systemWideSpaceTally


This class has been created from a part of SystemDictionary. It still deserves a nice
clean, such as using object instead of array having 4 slots.

sd-20 June 2003!

!SpaceTally methodsFor: 'fileOut' stamp: 'nice 10/20/2009 23:55'!
compareTallyIn: beforeFileName to: afterFileName
	"SpaceTally new compareTallyIn: 'tally' to: 'tally2'"

	| answer s beforeDict a afterDict allKeys |
	beforeDict := Dictionary new.
	s := FileDirectory default fileNamed: beforeFileName.
	[s atEnd] whileFalse: [
		a := Array readFrom: s nextLine.
		beforeDict at: a first put: a allButFirst.
	].
	s close.
	afterDict := Dictionary new.
	s := FileDirectory default fileNamed: afterFileName.
	[s atEnd] whileFalse: [
		a := Array readFrom: s nextLine.
		afterDict at: a first put: a allButFirst.
	].
	s close.
	answer := WriteStream on: String new.
	allKeys := (Set new addAll: beforeDict keys; addAll: afterDict keys; yourself) asSortedCollection.
	allKeys do: [ :each |
		| before after diff |
		before := beforeDict at: each ifAbsent: [#(0 0 0)].
		after := afterDict at: each ifAbsent: [#(0 0 0)].
		diff := before with: after collect: [ :vBefore :vAfter | vAfter - vBefore].
		diff = #(0 0 0) ifFalse: [
			answer nextPutAll: each,'  ',diff printString; cr.
		].
	].
	StringHolder new contents: answer contents; openLabel: 'space diffs'.
	


! !

!SpaceTally methodsFor: 'class analysis' stamp: 'ul 4/1/2013 13:01'!
computeSpaceUsage

	results
		do: [ :entry |
			| class instanceSpaceAndCount |
			class := self class environment at: entry analyzedClassName.
			Smalltalk garbageCollectMost.
			instanceSpaceAndCount := self spaceForInstancesOf: class.
			entry 
				codeSize: class spaceUsed;
				instanceCount: instanceSpaceAndCount second;
				spaceForInstances: instanceSpaceAndCount first ]
		displayingProgress: 'Taking statistics...'! !

!SpaceTally methodsFor: 'class analysis' stamp: 'sd 6/20/2003 22:54'!
preAllocateResultsFor: classes

	results := OrderedCollection new: classes size.
	classes do: [:cl | results add: (SpaceTallyItem analyzedClassName: cl name)].
	results := results asArray.
! !

!SpaceTally methodsFor: 'fileOut' stamp: 'ar 2/26/2010 18:02'!
printSpaceAnalysis	
	"SpaceTally new printSpaceAnalysis"

	^ self printSpaceAnalysis: 1 on: 'STspace.text'! !

!SpaceTally methodsFor: 'fileOut' stamp: 'cmm 1/7/2014 10:48'!
printSpaceAnalysis: threshold onStream: stream
	"If threshold > 0, then only those classes with more than that number
	of instances will be shown, and they will be sorted by total instance space.
	If threshold = 0, then all classes will appear, sorted by name."

	| totalCodeSpace totalInstCount totalInstSpace totalPercent classNameLength printRow |
	self systemWideSpaceTally.
	totalCodeSpace := totalInstCount := totalInstSpace := 0.
	classNameLength := 1.
	results do: [ :each |
		classNameLength := classNameLength max: each analyzedClassName size.
		totalCodeSpace := totalCodeSpace + each codeSize.
		totalInstCount := totalInstCount + each instanceCount.
		totalInstSpace := totalInstSpace + each spaceForInstances ].
	totalPercent := 0.0.

	printRow := [ :class :codeSpace :instanceCount :instanceSpace :percent |
		stream
			nextPutAll: (class padded: #right to: classNameLength + 1 with: $ );
			nextPutAll: (codeSpace padded: #left to: 12 with: $ );
			nextPutAll: (instanceCount padded: #left to: 12 with: $ );
			nextPutAll: (instanceSpace padded: #left to: 14 with: $ );
			nextPutAll: (percent padded: #left to: 8 with: $ );
			cr ].

	stream timeStamp.
	printRow valueWithArguments: { 'Class'. 'code space'. '# instances'. 'inst space'. 'percent' }.

	threshold > 0 ifTrue: [
		"If inst count threshold > 0, then sort by space"
		results := results select: [ :s |
			s instanceCount >= threshold or: [
				s spaceForInstances > (totalInstSpace // 500) ] ].
		results sort: [ :s :s2 | s spaceForInstances > s2 spaceForInstances ] ].

	results do: [ :s |
		| percent | 
		percent := s spaceForInstances * 100.0 / totalInstSpace.
		totalPercent := totalPercent + percent.
		printRow valueWithArguments: {
			s analyzedClassName.
			s codeSize printString.
			s instanceCount printString.
			s spaceForInstances asBytesDescription.
			percent printShowingDecimalPlaces: 1 } ].

	stream cr.
	printRow valueWithArguments: {
		'Total'.
		totalCodeSpace printString.
		totalInstCount printString.
		totalInstSpace printString.
		totalPercent printShowingDecimalPlaces: 1 }! !

!SpaceTally methodsFor: 'fileOut' stamp: 'ul 3/26/2013 09:47'!
printSpaceAnalysis: threshold on: fileName
	"SpaceTally new printSpaceAnalysis: 1000 on: 'STspace.text1'"

	"sd-This method should be rewrote to be more coherent within the rest of the class 
	ie using preAllocate and spaceForInstanceOf:"

	"If threshold > 0, then only those classes with more than that number
	of instances will be shown, and they will be sorted by total instance space.
	If threshold = 0, then all classes will appear, sorted by name."

	FileStream newFileNamed: fileName do: [ :file |
		self printSpaceAnalysis: threshold onStream: file ]! !

!SpaceTally methodsFor: 'fileOut' stamp: 'nice 6/11/2010 21:28'!
printSpaceDifferenceFrom: fileName1 to: fileName2
	"For differential results, run printSpaceAnalysis twice with different fileNames,
	then run this method...
		SpaceTally new printSpaceAnalysis: 0 on: 'STspace.text1'.
			--- do something that uses space here ---
		SpaceTally new printSpaceAnalysis: 0 on: 'STspace.text2'.
		SpaceTally new printSpaceDifferenceFrom: 'STspace.text1' to: 'STspace.text2'
"
	| f coll1 coll2 item |
	f := FileStream readOnlyFileNamed: fileName1.
	coll1 := OrderedCollection new.
	[f atEnd] whileFalse: [coll1 add: f nextLine].
	f close.
	f := FileStream readOnlyFileNamed: fileName2.
	coll2 := OrderedCollection new.
	[f atEnd] whileFalse:
		[item := f nextLine.
		((coll1 includes: item) and: [(item endsWith: 'percent') not])
			ifTrue: [coll1 remove: item]
			ifFalse: [coll2 add: item]].
	f close.
	(StringHolder new contents: (String streamContents: 
			[:s | 
			s nextPutAll: fileName1; cr.
			coll1 do: [:x | s nextPutAll: x; cr].
			s cr; cr.
			s nextPutAll: fileName2; cr.
			coll2 do: [:x | s nextPutAll: x; cr]]))
		openLabel: 'Differential Space Analysis'.
! !

!SpaceTally methodsFor: 'accessing' stamp: 'sd 6/20/2003 22:31'!
results

	^ results! !

!SpaceTally methodsFor: 'fileOut' stamp: 'sd 6/20/2003 22:59'!
saveTo: aFileName
	"| st |
	st := SpaceTally new.
	st spaceTally: (Array with: TextMorph with: Point).
	st saveTo: 'spaceTally2'"
	| s |
	(FileDirectory default fileExists: aFileName) ifTrue: [
		FileDirectory default deleteFileNamed: aFileName].
	s := FileDirectory default fileNamed: aFileName.
	results do: [:each | s nextPutAll: each analyzedClassName asString ; 
						nextPutAll: ' '; nextPutAll: each codeSize printString; 
						nextPutAll: ' '; nextPutAll: each instanceCount printString; 
						nextPutAll: ' '; nextPutAll: each spaceForInstances printString; cr].
	s close! !

!SpaceTally methodsFor: 'instance size' stamp: 'eem 3/12/2014 10:31'!
spaceForInstancesOf: aClass
	"Answer a pair of the number of bytes consumed by all instances of the
	 given class, including their object headers, and the number of instances."

	| instances total |
	instances := aClass allInstances.
	instances isEmpty ifTrue: [^#(0 0)].
	total := 0.
	aClass isVariable
		ifTrue:
			[instances do:
				[:i| total := total + (aClass byteSizeOfInstanceOfSize: i basicSize)]]
		ifFalse:
			[total := instances size * aClass byteSizeOfInstance].
	^{ total. instances size }! !

!SpaceTally methodsFor: 'class analysis' stamp: 'cmm 9/5/2012 10:30'!
spaceTally: classes 
	"Answer a collection of SpaceTallyItems representing the memory space (in bytes) consumed 	by the code and instances of each class in the system. Note that code sizes do not currently 	report memory consumed by class variables. "
	"((SpaceTally new spaceTally: (Array with: TextMorph with: Point)) asSortedCollection: [:a :b | a spaceForInstances > b spaceForInstances]) asArray"
	self preAllocateResultsFor: classes.
	Smalltalk garbageCollect.
	self computeSpaceUsage.
	^ results
		 sort: [ : a : b | a spaceForInstances > b spaceForInstances ] ;
		 yourself! !

!SpaceTally methodsFor: 'class analysis' stamp: 'efc 7/6/2004 00:25'!
systemWideSpaceTally
	"Answer a collection of SpaceTallyItems representing the memory space (in bytes) consumed 	by the code and instances of each class in the system. Note that code sizes do not currently 	report memory consumed by class variables. "

	"(SpaceTally new systemWideSpaceTally asSortedCollection: [:a :b | a spaceForInstances > b spaceForInstances]) asArray"

	^self spaceTally: Smalltalk allClasses.

! !

Object subclass: #SpaceTallyItem
	instanceVariableNames: 'analyzedClassName codeSize instanceCount spaceForInstances'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Tools'!

!SpaceTallyItem commentStamp: 'sd 6/20/2003 22:02' prior: 0!
I'm represent an entry in the spaceTally.!

!SpaceTallyItem class methodsFor: 'instance creation' stamp: 'sd 6/20/2003 22:54'!
analyzedClassName: aClassName

	^ self new
		analyzedClassName: aClassName ; yourself
		! !

!SpaceTallyItem class methodsFor: 'instance creation' stamp: 'sd 6/20/2003 22:54'!
analyzedClassName: aClassName codeSize: codeSize instanceCount: instanceCount spaceForInstances: spaceForInstances

	^ self new
		analyzedClassName: aClassName ;
		codeSize: codeSize ;
		instanceCount: instanceCount ;
		spaceForInstances: spaceForInstances ; yourself! !

!SpaceTallyItem methodsFor: 'accessing' stamp: 'sd 6/20/2003 22:59'!
analyzedClassName

	^ analyzedClassName! !

!SpaceTallyItem methodsFor: 'accessing' stamp: 'sd 6/20/2003 22:59'!
analyzedClassName: aClassName

	analyzedClassName := aClassName! !

!SpaceTallyItem methodsFor: 'accessing' stamp: 'sd 6/20/2003 22:08'!
codeSize

	^ codeSize! !

!SpaceTallyItem methodsFor: 'accessing' stamp: 'sd 6/20/2003 22:09'!
codeSize: aNumber

	codeSize := aNumber! !

!SpaceTallyItem methodsFor: 'accessing' stamp: 'sd 6/20/2003 22:09'!
instanceCount

	^ instanceCount! !

!SpaceTallyItem methodsFor: 'accessing' stamp: 'sd 6/20/2003 22:09'!
instanceCount: aNumber

	instanceCount := aNumber! !

!SpaceTallyItem methodsFor: 'printing' stamp: 'cmm 5/19/2013 12:49'!
printOn: aStream

	analyzedClassName
		ifNotNil: [ aStream nextPutAll: analyzedClassName asString]. 
	aStream nextPutAll: ' ('.
	codeSize
		ifNotNil: [ aStream nextPutAll: 'code size: ' ;  nextPutAll: codeSize asString]. 
	instanceCount
		ifNotNil: [ aStream nextPutAll: ' instance count: ' ;  nextPutAll: instanceCount asString]. 
	spaceForInstances
		ifNotNil: [ aStream nextPutAll: ' space for instances: ' ;  nextPutAll: spaceForInstances asBytesDescription]. 
	aStream nextPut: $).
	! !

!SpaceTallyItem methodsFor: 'accessing' stamp: 'sd 6/20/2003 22:10'!
spaceForInstances

	^ spaceForInstances! !

!SpaceTallyItem methodsFor: 'accessing' stamp: 'sd 6/20/2003 22:10'!
spaceForInstances: aNumber

	spaceForInstances := aNumber! !

Object subclass: #SystemChangeNotifier
	instanceVariableNames: 'eventSource silenceLevel'
	classVariableNames: 'UniqueInstance'
	poolDictionaries: ''
	category: 'System-Change Notification'!

!SystemChangeNotifier class methodsFor: 'item kinds' stamp: 'NS 1/21/2004 09:31'!
categoryKind

	^ AbstractEvent categoryKind! !

!SystemChangeNotifier class methodsFor: 'item kinds' stamp: 'NS 1/21/2004 09:31'!
classKind

	^ AbstractEvent classKind! !

!SystemChangeNotifier class methodsFor: 'private' stamp: 'NS 1/27/2004 16:23'!
createInstance

	^self basicNew initialize! !

!SystemChangeNotifier class methodsFor: 'item kinds' stamp: 'NS 1/21/2004 09:32'!
expressionKind
	^ AbstractEvent expressionKind! !

!SystemChangeNotifier class methodsFor: 'private' stamp: 'rw 6/28/2003 09:41'!
instanceCreationErrorString

	^'This is a singleton implementation, so you are not allowed to create instances yourself. Use #uniqueInstance to access the instance.'! !

!SystemChangeNotifier class methodsFor: 'item kinds' stamp: 'NS 1/21/2004 09:31'!
methodKind

	^ AbstractEvent methodKind! !

!SystemChangeNotifier class methodsFor: 'instance creation' stamp: 'rw 6/28/2003 09:41'!
new

	^self error: self instanceCreationErrorString! !

!SystemChangeNotifier class methodsFor: 'item kinds' stamp: 'NS 1/21/2004 09:32'!
protocolKind
	^ AbstractEvent protocolKind! !

!SystemChangeNotifier class methodsFor: 'private' stamp: 'ul 12/12/2009 14:06'!
resetUniqueInstance
	"self resetUniqueInstance"

	UniqueInstance
		ifNotNil: [:u | UniqueInstance releaseAll.
			UniqueInstance := nil]! !

!SystemChangeNotifier class methodsFor: 'public' stamp: 'NS 1/27/2004 16:23'!
uniqueInstance

	UniqueInstance ifNil: [UniqueInstance := self createInstance].
	^UniqueInstance! !

!SystemChangeNotifier methodsFor: 'private-event lists' stamp: 'rw 7/29/2003 15:14'!
allSystemEvents
	^AbstractEvent systemEvents! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'rw 7/29/2003 15:11'!
classAdded: aClass inCategory: aCategoryName 
	self trigger: (AddedEvent class: aClass category: aCategoryName)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'ab 2/10/2005 16:32'!
classCategoryAdded: aClassCategoryName

	self trigger: (AddedEvent
					classCategory: aClassCategoryName)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'ab 2/10/2005 16:35'!
classCategoryRemoved: aClassCategoryName

	self trigger: (RemovedEvent
					classCategory: aClassCategoryName)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'ab 2/10/2005 16:44'!
classCategoryRenamedFrom: anOldClassCategoryName to: aNewClassCategoryName

	self trigger: (RenamedEvent
					classCategoryRenamedFrom: anOldClassCategoryName 
					to: aNewClassCategoryName)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/26/2004 09:37'!
classCommented: aClass
	"A class with the given name was commented in the system."

	self trigger: (CommentedEvent class: aClass)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'rw 7/29/2003 15:11'!
classCommented: aClass inCategory: aCategoryName 
	"A class with the given name was commented in the system."

	self trigger: (CommentedEvent class: aClass category: aCategoryName)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/20/2004 19:37'!
classDefinitionChangedFrom: oldClass to: newClass
	self trigger: (ModifiedClassDefinitionEvent classDefinitionChangedFrom: oldClass to: newClass)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/16/2004 15:10'!
classRemoved: aClass fromCategory: aCategoryName 
	self trigger: (RemovedEvent class: aClass category: aCategoryName)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/27/2004 12:19'!
classRenamed: aClass from: oldClassName to: newClassName inCategory: aCategoryName 
	self trigger: (RenamedEvent 
				class: aClass
				category: aCategoryName
				oldName: oldClassName
				newName: newClassName)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/27/2004 12:48'!
classReorganized: aClass
	self trigger: (ReorganizedEvent class: aClass)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'rw 7/29/2003 15:12'!
class: aClass recategorizedFrom: oldCategory to: newCategory 
	self trigger: (RecategorizedEvent 
				class: aClass
				category: newCategory
				oldCategory: oldCategory)! !

!SystemChangeNotifier methodsFor: 'public' stamp: 'ul 8/2/2011 20:11'!
doSilently: aBlock
	"Perform the block, and ensure that no system notification are broadcasted while doing so."

	| result |
	silenceLevel := silenceLevel + 1.
	result := aBlock ensure: [silenceLevel > 0 ifTrue: [silenceLevel := silenceLevel - 1]].
	^ result.! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/19/2004 09:48'!
evaluated: textOrStream
	^ self evaluated: textOrStream context: nil.! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/19/2004 09:47'!
evaluated: expression context: aContext
	self trigger: (DoItEvent 
				expression: expression
				context: aContext)! !

!SystemChangeNotifier methodsFor: 'initialize' stamp: 'NS 1/26/2004 20:41'!
initialize

	eventSource := SystemEventManager new.
	silenceLevel := 0.! !

!SystemChangeNotifier methodsFor: 'public' stamp: 'NS 1/26/2004 20:41'!
isBroadcasting

	^ silenceLevel = 0! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/27/2004 11:24'!
methodAdded: aMethod selector: aSymbol inClass: aClass 
	"A method with the given selector was added to aClass, but not put in a protocol."

	self trigger: (AddedEvent
				method: aMethod 
				selector: aSymbol
				class: aClass)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/27/2004 11:24'!
methodAdded: aMethod selector: aSymbol inClass: aClass requestor: requestor
	"A method with the given selector was added to aClass, but not put in a protocol."

	self trigger: (AddedEvent
				method: aMethod 
				selector: aSymbol
				class: aClass
				requestor: requestor)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/27/2004 11:24'!
methodAdded: aMethod selector: aSymbol inProtocol: aCategoryName class: aClass 
	"A method with the given selector was added to aClass in protocol aCategoryName."

	self trigger: (AddedEvent
				method: aMethod
				selector: aSymbol
				protocol: aCategoryName
				class: aClass)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/27/2004 11:24'!
methodAdded: aMethod selector: aSymbol inProtocol: aCategoryName class: aClass requestor: requestor
	"A method with the given selector was added to aClass in protocol aCategoryName."

	self trigger: (AddedEvent
				method: aMethod
				selector: aSymbol
				protocol: aCategoryName
				class: aClass
				requestor: requestor)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/27/2004 11:41'!
methodChangedFrom: oldMethod to: newMethod selector: aSymbol inClass: aClass
	self trigger: (ModifiedEvent
					methodChangedFrom: oldMethod
					to: newMethod
					selector: aSymbol 
					inClass: aClass)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/27/2004 11:41'!
methodChangedFrom: oldMethod to: newMethod selector: aSymbol inClass: aClass requestor: requestor
	self trigger: (ModifiedEvent
					methodChangedFrom: oldMethod
					to: newMethod
					selector: aSymbol 
					inClass: aClass
					requestor: requestor)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/28/2004 11:12'!
methodRemoved: aMethod selector: aSymbol class: aClass 
	"A method with the given selector was removed from the class."

	self trigger: (RemovedEvent
				method: aMethod 
				selector: aSymbol
				class: aClass)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'NS 1/28/2004 11:11'!
methodRemoved: aMethod selector: aSymbol inProtocol: protocol class: aClass 
	"A method with the given selector was removed from the class."

	self trigger: (RemovedEvent
				method: aMethod 
				selector: aSymbol
				protocol: protocol
				class: aClass)! !

!SystemChangeNotifier methodsFor: 'public' stamp: 'rw 7/29/2003 17:01'!
noMoreNotificationsFor: anObject
	"Stop sending system notifications to an object."

	eventSource removeActionsWithReceiver: anObject! !

!SystemChangeNotifier methodsFor: 'public' stamp: 'rw 7/10/2003 12:00'!
notify: anObject ofAllSystemChangesUsing: oneArgumentSelector 
	"Notifies an object of any system changes."

	self 
		notify: anObject
		ofEvents: self allSystemEvents
		using: oneArgumentSelector! !

!SystemChangeNotifier methodsFor: 'private' stamp: 'rw 7/10/2003 15:15'!
notify: anObject ofEvents: eventsCollection using: oneArgumentSelector
	"Notifies an object of any events in the eventsCollection. Send it back a message #oneArgumentSelector, with as argument the particular system event instance."

	eventsCollection do: [:eachEvent |
		eventSource when: eachEvent send: oneArgumentSelector to: anObject]! !

!SystemChangeNotifier methodsFor: 'public' stamp: 'bvs 7/20/2004 12:13'!
notify: anObject ofSystemChangesOfChange: changeKind using: oneArgumentSelector 
	"Notifies an object of system changes of the specified changeKind (#added, #removed, ...). Evaluate 'AbstractEvent allChangeKinds' to get the complete list."

	self 
		notify: anObject
		ofEvents: (self systemEventsForChange: changeKind)
		using: oneArgumentSelector! !

!SystemChangeNotifier methodsFor: 'public' stamp: 'bvs 7/20/2004 12:13'!
notify: anObject ofSystemChangesOfItem: itemKind change: changeKind using: oneArgumentSelector 
	"Notifies an object of system changes of the specified itemKind (#class, #category, ...) and changeKind (#added, #removed, ...). This is the finest granularity possible.
	Evaluate 'AbstractEvent allChangeKinds' to get the complete list of change kinds, and 'AbstractEvent allItemKinds to get all the possible item kinds supported."

	self 
		notify: anObject
		ofEvents: (Bag with: (self systemEventsForItem: itemKind change: changeKind))
		using: oneArgumentSelector! !

!SystemChangeNotifier methodsFor: 'public' stamp: 'bvs 7/20/2004 12:13'!
notify: anObject ofSystemChangesOfItem: itemKind  using: oneArgumentSelector 
	"Notifies an object of system changes of the specified itemKind (#class, #method, #protocol, ...). Evaluate 'AbstractEvent allItemKinds' to get the complete list."

	self 
		notify: anObject
		ofEvents: (self systemEventsForItem: itemKind)
		using: oneArgumentSelector! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'laza 11/3/2011 00:36'!
protocolAdded: aProtocol inClass: aClass 
	self trigger: (
		AddedEvent protocol: aProtocol class: aClass
	)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'laza 11/3/2011 00:35'!
protocolRemoved: aProtocol inClass: aClass 
	self trigger: (
		RemovedEvent protocol: aProtocol class: aClass
	)! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'laza 11/3/2011 00:19'!
protocolRenamedFrom: anOldProtocol to: aNewProtocol inClass: aClass 
	self trigger: (
		(RenamedEvent protocol: anOldProtocol class: aClass)
			oldName: anOldProtocol;
			newName: aNewProtocol
	)! !

!SystemChangeNotifier methodsFor: 'private' stamp: 'rw 7/29/2003 17:05'!
releaseAll
	"Release all the dependents so that nobody receives notifications anymore."

	"Done for cleaning up the system."
	"self uniqueInstance releaseAll"

	eventSource releaseActionMap! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'ar 9/5/2010 21:57'!
selector: selector recategorizedFrom: oldCategory to: newCategory inClass: aClass

	self trigger: (RecategorizedEvent 
				selector: selector
				method: (aClass compiledMethodAt: selector ifAbsent: [nil])
				protocol: newCategory
				class: aClass
				oldProtocol: oldCategory)! !

!SystemChangeNotifier methodsFor: 'private' stamp: 'NS 1/26/2004 20:43'!
setBroadcasting
	silenceLevel := 0.! !

!SystemChangeNotifier methodsFor: 'private-event lists' stamp: 'rw 7/29/2003 15:14'!
systemEventsForChange: changeKind 
	| selectorBlock |
	selectorBlock := AbstractEvent eventSelectorBlock.
	^AbstractEvent allItemKinds 
		collect: [:itemKind | selectorBlock value: itemKind value: changeKind]! !

!SystemChangeNotifier methodsFor: 'private-event lists' stamp: 'rw 7/29/2003 15:14'!
systemEventsForItem: itemKind 
	| selectorBlock |
	selectorBlock := AbstractEvent eventSelectorBlock.
	^AbstractEvent allChangeKinds 
		collect: [:changeKind | selectorBlock value: itemKind value: changeKind]! !

!SystemChangeNotifier methodsFor: 'private-event lists' stamp: 'rw 7/29/2003 15:14'!
systemEventsForItem: itemKind change: changeKind 
	^AbstractEvent eventSelectorBlock value: itemKind value: changeKind! !

!SystemChangeNotifier methodsFor: 'system triggers' stamp: 'al 7/18/2004 10:48'!
traitDefinitionChangedFrom: oldTrait to: newTrait
	self trigger: (ModifiedTraitDefinitionEvent traitDefinitionChangedFrom: oldTrait to: newTrait)! !

!SystemChangeNotifier methodsFor: 'private' stamp: 'NS 1/26/2004 20:41'!
trigger: event

	self isBroadcasting ifTrue: [event trigger: eventSource]

"	| caughtExceptions |
	caughtExceptions := OrderedCollection new.
	self isBroadcasting ifTrue: [
		[(eventSource actionForEvent: event eventSelector) valueWithArguments: (Array with: event)] on: Exception do: [:exc | caughtExceptions add: exc]].
	caughtExceptions do: [:exc | exc resignalAs: exc class new]"! !

Object subclass: #SystemNavigation
	instanceVariableNames: 'browserClass hierarchyBrowserClass environment'
	classVariableNames: 'Authors AuthorsInverted Default ThoroughSenders'
	poolDictionaries: ''
	category: 'System-Support'!

!SystemNavigation commentStamp: 'mha 8/26/2010 09:02' prior: 0!
I support the navigation of the system. I act as a facade but as I could require some state
or different way of navigating the system all my behavior are on the instance side.


For example if you want to look at all methods you have written or changed in the current image do

SystemNavigation new browseAllSelect: [ :method |
       method fileIndex > 1 "only look at changes file"
       and: [ method timeStamp beginsWith: 'your-initials-here' ] ].

!

!SystemNavigation class methodsFor: 'accessing' stamp: 'mt 8/15/2016 11:49'!
authors

	^ Authors ifNil: [self initializeAuthors. Authors]! !

!SystemNavigation class methodsFor: 'accessing' stamp: 'mt 8/15/2016 11:49'!
authorsInverted

	^ AuthorsInverted ifNil: [self initializeAuthors. AuthorsInverted]! !

!SystemNavigation class methodsFor: 'class initialization' stamp: 'mt 8/15/2016 11:48'!
cleanUp: aggressive

	aggressive ifTrue: [Authors := nil. AuthorsInverted := nil].! !

!SystemNavigation class methodsFor: 'accessing' stamp: 'ul 3/14/2011 20:53'!
default
	
	^Default ifNil: [ Default := self new ]! !

!SystemNavigation class methodsFor: 'error handling' stamp: 'eem 4/18/2012 18:54'!
doesNotUnderstand: aMessage
	(self includesSelector: aMessage selector) ifTrue:
		[^self default perform: aMessage selector withArguments: aMessage arguments].
	^super doesNotUnderstand: aMessage! !

!SystemNavigation class methodsFor: 'accessing' stamp: 'fbs 5/9/2013 22:29'!
for: anEnvironment
	^ self basicNew initializeWithEnvironment: anEnvironment.! !

!SystemNavigation class methodsFor: 'class initialization' stamp: 'mt 7/29/2016 11:31'!
initialize

	self initializeAuthors.! !

!SystemNavigation class methodsFor: 'class initialization' stamp: 'mt 7/29/2016 12:59'!
initializeAuthors
	"self initializeAuthors"
	
	| s currentName currentAbbrv |
	Authors := Dictionary new.
	s := self privateAuthorsRaw readStream.
	
	[s atEnd] whileFalse: [
		"Read author name"
		currentName := String streamContents: [:as |
			[s peek ~= $#] whileTrue: [as nextPut: s next]].
		Authors at: currentName put: Set new.
		"Read abbreviations"
		s next. "skip $#"
		currentAbbrv := ''.
		[s atEnd not and: [s peek ~= $!!]] whileTrue: [
			s peek = $# ifTrue: [
				(Authors at: currentName) add: currentAbbrv.
				currentAbbrv := ''.
				s next. "skip $#"].
			currentAbbrv := currentAbbrv, s next asString. "slow..."
			].
		currentAbbrv ifNotEmpty: [(Authors at: currentName) add: currentAbbrv].
		s next. "skip $!!"
	].

	"Fill inverted index for authors."
	AuthorsInverted := Dictionary new.
	Authors keysAndValuesDo: [:fullName :abbreviations |
		abbreviations do: [:abbrv |
			(AuthorsInverted
				at: abbrv asLowercase
				ifAbsentPut: [Set new]) add: fullName]].! !

!SystemNavigation class methodsFor: 'class initialization' stamp: 'mt 7/29/2016 13:33'!
privateAuthorsRaw

	^ 'Aaron Reichow#ajr!!Abigail Sanchez#as!!Adam Engel#ASE!!Adam King#ajk!!Adam Sádovský#Adrin!!Adam Spitz#ads!!Adam Wespiser#wespiser!!Adrian Lienhard#AL#al!!Adrien Sebbane#AS!!Aibek Musaev#AM!!Aik-Siong Koh#askoh!!Alain Fischer#AFi!!Alain Plantec#apl!!Alain Rastoul#alr!!Alan Capewell#arc!!Alan Hodson#TX!!Alan Curtis Kay#ack!!Alan Lovejoy#chronos!!Alan Reider#alnr!!Alberto#AlFe!!Alejandro Aguirre#pato!!Alejandro Fernandez#casco!!Alejandro Gonzalez#ag!!Alejandro Magistrello#asm!!Alejandro Reimondo#afr!!Alessandro Warth#aw!!Alexander Diemand#XLD!!Alexander Lazarevic#laza!!Alexander Martinov#AM!!Alexandre Bergel#ab!!Alexandre Jasmin#alexj!!Alexis Olivier#AO!!Allen Evans#ae!!Alphacode#alpha!!Americo Damasceno#adamascj!!Amir Nanda Gorkhali#Amir!!Andreas Kuckartz#ak!!Andreas Petermann#ap!!Andreas Raab#ar!!Andreas Wiedebruch#sanwb!!Andres Coratella#ac!!Andres Otaduy#aoy!!Andres Valloud#SqR!!Andrew Berg#acb!!Andrew C. Greenberg#acg!!Andrew Capon#ACapon!!Andrew Gaylard#apg!!Andrew Lawson#adl!!Andrew P. Black#apb!!Andrew Tween#amt!!Andy Gelme#andyg!!Andy Milburn#am!!Andy Tween#tween!!Angel Aberbach#aberbacha!!Angela E. Taylor#trxckster!!Anoulak Kictiraz#AK!!Anthony Giaccone#ajg!!Anthony Hannan#ajh!!Antonio Moreno#am!!Antony Blakey#ASB!!Aoki Atsushi#aoJun!!Arian Treffer#ArTr!!Arnaud Jean Baptiste#JBA!!Arndt Jenssen#AJ!!Art Collings#amc!!Arthur Vanderbilt#aav!!Arto Bendiken#arto!!Arturo Frappé#afm!!Ashok Mahajan#ashok!!Attila Magyar#zero!!Avi Bryant#avi!!Aye Aye Khine#AAK!!Baard Ove Kopperud#BOK!!Bakki Kudva#bk!!Balázs Kósi#kb!!Barrie Jehu#jbj!!Barry Bridgens#bjb!!Bas Kok#bkok!!Bastian Steinert#bs!!Bedrich Hovorka#bh3!!Ben Coman#btc!!Ben Goetter#bdg!!Ben Huxham#BHux!!Ben Matasar#ben!!Ben Schroeder#bvs!!Benjamin Pollack#bmp!!Benoît St-Jean#BSTJ!!Bernd Eckardt#be!!Bernd Oldenbuettel#beo!!Bernhard Pieber#bp!!Bert Freudenberg#bf!!Bijan Parsia#BJP!!Bill Dargel#wod!!Bill Holloway#bat#wjh!!Bill Nadvornik#wcn!!Bill Schwab#wks!!Bill Smith#WFS!!Binh Le#binhle2#BL!!Blaine Buxton#btb!!Bob Arning#RAA!!Bob Cowdery#RAC!!Bob Hartwig#bh!!Bob Houston#rkh!!Bob Irving#bi!!Bolot Kerimbaev#bolot!!Boris Gaertner#bg!!Boris Shingarov#bgs!!Bracken Spencer#bra!!Brenda Larcom#BEL!!Brent Pinkney#brp!!Brent Vukmer#bkv!!Brian Brown#rbb!!Brian Byrne#bpb!!Brian Keefer#bmk!!Brian Murphy-Dye#bmd!!Brian T. Rice#btr!!Brian Tabone#hybridmachine!!Bruce Badger#BB!!Bruce Cichowlas#bwc!!Bruce Haugland#BAH!!Bruce Prior#bgp!!Bruno Luca#let!!Bryce Kampjes#wbk!!Burella Juan M.#bjm!!C. David Shaffer#cds!!CalmoSoft#Calmo!!Carl Gruden#cgg!!Carl Gundel#ceg!!Carl Watts#cgw!!Carlos Crosetti#ccrosetti!!Carlos Morales Socorro#cmorsoc!!Casey Ransberger#cbr!!Catherine Pratoff#CP!!Cédrick Béler#cb!!Cees de Groot#CdG#cdg!!Cesare Marilungo#cm!!Chad Nantais#cen!!Chao-Kuo Lin#ckl!!Charles Hixson#cdh!!Chinedu Charles#!!Chris Becker#chb!!Chris Burkert#chbu!!Chris Cunningham#cbc!!Chris Cunnington#CC#chc!!Chris De Vaney#CdV!!Chris Dickerson#erilidon!!Chris Kermiet#ChrisK!!Chris Muller#cmm!!Chris Norton#ccn!!Chris Patrick Schreiner#cps!!Chris Patti#cap!!Chris Reuter#cr!!Chris Wade-Evans#CJWE!!Christoph Budzinski#CHB!!Christoph Lamprecht#chl!!Chuck Valois#cbv#ceevee!!Chucky Ellison#cme!!Claes-Fredrik Mannby#cf!!Clarence Westberg#crw!!Claus Gittinger#cg!!Claus Kick#CK!!Clément Béra#cb!!Clifford Goldberg#cig!!Cocodrilo Astronauta#cocodriloastronauta!!Colin Curtin#cc!!Colin Putney#cwp!!Colm Sean Murdoch O Cinneide#csmk!!Connie Louise#clt!!Connie Tebbe#connielouise#mst303!!Conrad Pöpke#cop!!Cornelis Pols#jayanta!!Cory Redfern#CR!!Costel Vrinceanu#cv#cxv!!Craig Latta#crl!!Cristhiano E. C. De Souza#cecs!!Dafina Zoteva#dafy!!Damien Cassou#dc!!Damien Pollet#dp!!Dan Ingalls#di!!Dan Moniz#dnm!!Dan Norton#dhn!!Dana Gheorghe#DG!!Danie Roux#djr!!Daniel Altman#DEA!!Daniel Poon#danp!!Daniel Sancho Ehlert#dse!!Daniel Vainsencher#dvf!!Daniel Weigel#dewb#dweigel!!Danil Osipchuk#dao!!Danilo Gulamhussene#dg!!Darius Bacon#djb!!Darius Clarke#dgc!!Dary Lalden#DJA!!Dave Faught#daf!!Dave Goehrig#djg!!Dave Mason#dvm!!Dave Mclure#dav0!!Dave Newman#DaN!!Dave Raftery#dgr!!Dave Raymer#dlr#draymer!!David Barroco#DaveBar!!David Belding#autonomad#tzaddi!!David Brown#ddb!!David Corking#dcorking!!David Dykstal#dwd!!David Farber#dvd!!David Harris#dph!!David J. Pennell#djp!!David Mitchell#david-mitchell#dmm!!David N. Smith#dns!!David Pollak#dpp!!David Salamon#dss!!David "Dave" T. Lewis#dtl!!David Tibbe#dt!!David Treffler#canadans!!David Ungar#dmu!!David Weil#adw!!David Wildasin#dtw!!Debora Nascimento#debby!!Delbert Murphy#drm!!Denis Kudriashov#dik!!Dennis Lee Hanchar#smartsphere!!Dennis Schetinin#chaetal!!Derek Brans#djb!!Derek Hadlington#djh!!Derek O''Connell#doc!!Derrell Piper#ddp!!Dick Combeer#DC!!Diego Fernandez#DF!!Diego Gomez Deck#dgd!!Diego Park#drp!!Diogenes Moreira#DAM!!Dipti Ranjan Nayak#nayak!!Dmitry Golubovsky#dmg!!Domianna Caruso#DJC!!Dominic Letz#dl!!Dominique Dutoit#dd!!Don Sawtelle#DS!!Doug Rollwitz#dbr!!Doug Way#dew!!Douglas Brebner#db!!Douglas McPherson#djm!!Duane Maxwell#dsm!!Duane Thomas Williams#dtw!!Dwight Hughes#dwh!!Dwight Mcdowell#dowaito!!E. Glen Paling#egp!!Eddie Cottongim#efc!!Edgar J. De Cleene#edc!!Edwin Cutright#EdnaC!!Eiichiro Ito#e-itoh!!Elena Sebe#gambitside!!Eliot Miranda#eem!!Elod Kironsky#ke!!Emery Gant#eg!!Emilio Oca#efo!!Emmanuel A. Escol#winky!!Enno Schwass#Enno!!Enrico Bertini#eb!!Enrico Spinielli#espin!!Eric Euerlings#EE!!Eric Fournier#emf!!Eric Hochmeister#eh!!Eric Hudson#ejh!!Eric Merritt#EBM!!Eric Rodriguez Guevara#ERG#erg!!Eric Scharff#eds!!Erik Greenwald#erik!!Erik Hinterbichler#erikh!!Ernest Micklei#emm!!Esteban A. Maringolo#eam!!Esteban Lorenzano#EL!!Estrella Demar#!!Etienne Delacroix#eti!!Evan Dembskey#EJD!!Fabio Niephaus#fn!!Fadi Mansour#Redeemer!!Faried Nawaz#fn!!Federico G. Stilman#fgs!!Felix Franz#ff!!Felix Lange#fl!!Felix Munk#FM!!Fernando Soteras#fersoteras#fest!!Filipe Cabecinhas#FilCab#FilipeCabecinhas!!Florian Greese#FlorianGreese#pikapika!!Francisco Pancho Garau#fg!!Francois Stephany#fs#FS!!Frank Barnes#FB#gandermight!!Frank Caggiano#fc!!Frank Kresin#fjk!!Frank Mueller#mue!!Frank Sergeant#fcs!!Frank Shearar#fbs!!Frank Winkler#FWr!!Franz Brummer#fb!!Franziska Krueger#fk!!Frederik Cheeseman#Xantheon!!Frederik De Bleser#fdb!!Frederik Vandendriessche#FVDD!!Fredrik Andersson#el_diego!!G. R. Thushar#grt!!Gabriel Senno#gsenno!!Gale Pedowitz#gep!!Gary Chambers#gvc!!Gary Fisher#gf!!Gen Suzuki#gen!!Geoff Reedy#ger!!Georg Gollmann#go!!George Olive#gao!!Gerald Leeb#LEG#leg!!Gerald Zincke#gzi!!Gerardo Richarte#gera#r++!!German Morales#gm!!Germán S. Arduino#gsa!!Gerry Cornish#gjc!!Gian Holland#gmh!!Gilles Reubi#gr!!Giovanni Corriga#gc!!Giovanni Giorgi#gg!!Giridharan#giridharan!!Giuseppe Luigi Punzi#glp!!Glenn G. Burnside III#ggb!!Gopibrindha#gopi!!Göran Krampe#gh#gk!!Görge Albrecht#gal!!Graham Mcleod#GMc!!Greg Hand#handy!!Greg Harrold#ggh!!Greg Lewin#GL!!Greg Trasuk#gmt!!Gregor Gabrysiak#RogerG!!Gregor Mcnish#igm!!Grit Schuster#gs!!Guillermo Polito#GAP!!Gustavo Rafael Pistoia#grp!!Guy Hillyer#gkh!!Hank Wilde#HFW!!Hannes Hirzel#hjh!!Hans Baveco#jmb!!Hans Kleinwort#HK!!Hans Schippers#HS!!Hans-Martin Mosner#hmm!!Harald Gliebe#hagl!!Harshvardhan#hvg!!Hashim Katalog#Kapashor!!Helge Horch#hh!!Henrik Ekenberg#heg!!Henrik Gedenryd#hg!!Henrique Manela#hm!!Henry Strickland#strick#YAK!!Herbert König#hk!!Herbert Vojèík#hsk!!Hern Tylim#hpt!!Hernan Galante#HJG!!Hernan Parra#HP!!Hernán Tylim#hpt!!Hernan Wilkinson#HAW!!Hiromu-Papa#hp!!Hiroshi Saito#gann!!Holger Schuh#holgersolger!!Houssam Fakih#Geanoubi!!Howard J. Oh#hjo!!Hung Le#hmle!!Ian Piumarta#ikp!!Ian Prince#ijp!!Ian Trudel#it!!Iede Snoek#IS!!Igor Stasenko#sig!!Ikem Nzeribe#ION!!Ingo Hohmann#iho!!Ingo Jaeckel#ij!!Innocent Victor#nkala!!Ioan Gartner#Sharphawk!!Ivo Eduardo Seben De Azevedo#ivoesa!!Ivo Roessling#ir!!Ivor Dt Cook#Ivor Cook!!J. Duijndam#dnmjca!!Jack Keel#jhk!!Jae Yoon Kim#projayjay!!Jahidul Alam#Jahid!!Jaime Silvela#jsm!!James Foster#JGF!!James Roberts#jr!!Jamie Jones#JJ!!Jamie McNeil#!!Jan B. Krejci#jbk!!Jan Barger#jb!!Jan Fietz#jaf!!Jan Van De Sandt#jvds!!Jang Yun Jae#janggoon!!Janko Mivsek#mivsek!!Jaroslav Havlin#jerri!!Jason Burke#jbb!!Jason Dicioccio#jmd!!Jason Dufair#jkd!!Jason L. Smith#jls!!Jason McVay#jlm!!Jason Ostberg#Jason!!Jason Rogers#jlr!!Jason Shoemaker#jts!!Jason Steffler#jrs!!Jason Varner#!!Jason Williams#jdw!!Javier Diaz-Reinoso#jdr!!Javier Musa#jam!!Jay Carlson#nop!!Jay Fearn#JBF!!Jean J. Escoto#jje!!Jeangalineau#JG!!Jecel Assumpcao Jr#jecel!!Jeff Gonis#jmg!!Jeff Pierce#jsp!!Jeff Sparkes#!!Jeffrey Feldman#jbf!!Jeffrey Foreman#jdf!!Jeffrey Haun#JAH!!Jeffrey T. Read#jtr!!Jens Lincke#jl!!Jens Moenig#jens!!Jens Pall Hafsteinsson#jp!!Jens Quade#jq!!Jeremy Johnstone#jstone!!Jeremy Shute#shute!!Jeroen Van Hilst#jvh!!Jerome Peace#wiz!!Jerry Archibald#jla!!Jerry Bryson#JerryB!!Jerry Stokes#js!!Jesse Welton#JW!!Jessy Kuriakose#akhsa!!Jesus Alvarez#ja#jalvarez#jmalvarez!!Jigar H. Doshi#jdoshi2#jhd#jhdoshi!!Jim Bender#jcb!!Jim Benson#jlb!!Jim Buckeye#JR!!Jim Freeman#jmf!!Jim Greer#jbg!!Jimmie Houchin#jlh!!Jivko Jeliazkov#jivko_jeliazkov!!Joan Barrull#joan!!Jochen Riekhof#jfr!!Joel Uzi Eden#jue!!Joern Eyrich#je!!Joey Hagedorn#JAH!!Joey White#jw!!Johannes Dyck#jfd!!John Bailey#jbailey652000!!John D''Ausilio#jld!!John David Miller#jdm!!John Dougan#jrd!!John Ewart#jwe!!John Ferrier#jtf!!John G. Gardner#JGG!!John M. McIntosh#JMM#jmm!!John M. Stephens, Jr.#jms!!John Macchia#JL!!John Magnifico#JM!!John Maloney#jhm#jm!!John Maxwell#jmax!!John McKeon#jmck!!John Pfersich#jgp!!John Pierce#jrp!!John Sarkela#jws#reThink!!John Tobler#jet!!John-Reed Maffeo#jrm!!Jon Hylands#jon!!Jonas Beckman#jobe!!Jonathan Brezin#jpb!!Jonathan D''Andries#jd!!Jonathan Wright#jdw!!Jonty Lochner#jlthedoes!!Jorge Campos#jmc!!Jörn Eyrich#je!!Jose Antonio Romero#HPCPRO!!José L. Redrejo Rodríguez#jlrr!!Jose L. Salmeron#jlsalsil!!Josef Hajas#jhajas#nax!!Joseph M. Paquette#jmp!!Joseph Pelrine#jp!!Joseph Terry#knowself!!Joshua Fein#JTF!!Joshua Gargus#jcg!!Joshua Marker#jlm!!Juan Daugherty#jdaugherty!!Juan Manuel Vuletich#jmv!!Judith Chatelain#jchatelain!!Julian Fitzell#jf!!Julien Bourdon#jb#jbo!!Julio Salcido#!!K. K. Subramaniam#kks!!Kakiya Isao#kakiya!!Kamil Kukura#kamk!!Kang Jin-oh#KJO!!Karel Zelnicek#__Karel__#KZ!!Karl Ramberg#kfr!!Karsten Droste#kd!!Karsten Wolf#kw!!Kazuhiro Abe#ka!!Kazuki Minamitani#minami!!Kees Kremer#kk!!Keith Hodges#kph!!Keith Lazuka#kl!!Keithhall#keithhall513!!Kejing Zhang#novachang!!Kellie Miller#jkm!!Ken Boucher#bons!!Ken Causey#KLC#klc!!Ken Dickey#KenD#KenD1!!Ken G. Brown#kgb!!Ken Happel#kmh!!Ken Treis#kdt!!Kenneth Johnsen#KEJ!!Kevin Fisher#kgf!!Kevin J. Woolley#kjw!!Kevin Kelley#klk!!Kim Yong-Cheon#gal2!!Kirill Jdanov#jdanov!!Klaus D. Witzel#kwl!!Konstantin Käfer#kkx!!Korienek Gene#eneg!!Kosuke Chan#kc!!Krishna Sankar#ks!!Kristin Kaster Lamberty#kk!!Kta Shwele#kta#nantaw!!Kumar Salman#kumar!!Kun Puparussanon#kp!!Kundan Dhumal#kundan_11!!Kurt Thams#ktt!!Lady Sophea#az5!!Lantz Rowland#zL!!Larry Gadallah#ldg!!Larry Pruitt#ldp!!Lars Dornheim#eldeh!!Lars Nilsson#ltn!!Lars Wassermann#lars#lw!!Laura Perez Cerrato#lpc!!Laurence Rozier#!!Laurent Siksous#lss!!Ld Landis#ldlandis!!Leandro Caniglia#LC!!Lee Salzman#lrs!!Lee Yu Hsun#swps567!!Legrand#LGDEKI!!Leo Burd#lb!!Leon Smith#LS!!Leonid Golender#lgol27!!Levente Uzonyi#ul!!Lex Spoon#ls!!Liangbing#Liangbing!!Logan Roguevax#lxr!!Lois Lherbier#LL!!Lorenzo Schiavina#lrs!!Lothar Schenk#los!!Lu Yunhai#luyunhai!!Luca Bruno#let!!Luciano Dell''Olivo#LD-Y1951!!Luciano Esteban Notarfrancesco#len!!Luciano Notarfrancesco#len!!Luigi Panzeri#LPM!!Luis Araujo#LA!!Luis Valente#LV!!Lukas Petrlik#lp!!Lukas Renggli#lr!!Luke Gorrie#lg!!Lushi#YYL!!Lyndon Tremblay#huma!!Magnus Kling#makl!!Mahajan Ashok#ashokmahajan!!Mahamud#encore!!Malcolm Yadack#vmy!!Manuel Trujillo#MTA!!Manuel Wellmann#MaWell#MW!!Marcel Taeumel#mt!!Marcel Weiher#mpw!!Marcelo Pereira Barbosa#mpbarbosa!!Marco Monteiro#masm!!Marco Paga#mpa!!Marco Schmidt#MS!!Marcus Denker#md!!Marcus Kemper#mtk!!Marek Mor#MMOR!!Mariano Abel Coca#nekron!!Mariano Martinez Peck#MMP!!Mark A. Schwenk#mas!!Mark Brady#mby!!Mark Guzdial#mjg!!Mark McCahill#mpm!!Mark Pirogovsky#markp!!Mark Schwenk#mas!!Mark Warren#MAW!!Markus Fritsche#maf#MFritsche!!Markus Gaelli#mga!!Markus Gälli#mga!!Martin Beck#mb!!Martin Holst Swende#mhs!!Martin Hruby#mhaf!!Martin Kuball#maku!!Martin McClure#mrm!!Martin Peschke#MP!!Martin Ruskov#MR!!Martin von Löwis#mvl!!Martin Wirblat#mw!!Masashi Umezawa#mu!!Masato Sumi#sumim!!Matej Kosik#mk!!Mats Nygren#mn!!Matt Kennedy#mck!!Matthew Fulmer#mtf!!Matthew S. Hamrick#msh!!Matti J. Kärki#MJK!!Maurice Rabb#m3r!!Maurizio Boriani#mdb!!Mauro Gardenal#MG!!Max Leske#MaxLeske!!Maximialiano Taborda#mx!!Maximo#mpc!!Mayuresh Kathe#MAK!!McSee#mc!!Merik Voswinkel#mav#mv!!Michael Cole#mc!!Michael Davies#MAD!!Michael Dunn#mrd020!!Michael E. Burke#MEBurke!!Michael Haupt#mha!!Michael Hewner#MPH!!Michael J. Forster#MJF!!Michael Klein#msk!!Michael Maloney#amm!!Michael Naunton#mmn!!Michael Nedzelsky#MichaelNedzelsky!!Michael Parker#kmp!!Michael Perscheid#micper!!Michael Pfurtscheller#mpf!!Michaël Piel#mp!!Michael Pruemm#mip!!Michael Rueger#mir!!Michael Toth#MT181!!Michael Van Der Gulik#mvdg!!Michal Kostal#!!Michal Skrivanek#!!Michal Starke#mist!!Michal Wallace#mjw!!Michal Wendrowski#michal!!Miguel Angel Castro#mac!!Miguel Cobá#mcm!!Mikael Kindborg#miki#mikki!!Mikael Langer#ml!!Mike Janger#mjj!!Mike Kaden#m2k!!Mike Roberts#mjr!!Mike Rutenberg#mdr!!Mike Sersen#Tre!!Mike Thomas#mjt!!Milan Vavra#mva!!Milan Zimmermann#mz!!Min Htut Khaung#minhtutkhaung!!Mirek Hancl#mh!!Mirko Kiefer#MirkoK!!Miron Brezuleanu#MB!!Mohsen Vakilian#mv!!Murphy Stein#mms!!Naala Brewer#nb!!Naheed Misbah#naheed!!NameSpaces#maf!!Nancy Brito#doubles!!Nathaniel Reyes#nreyes!!Ned Konz#nk!!Nevin Pratt#nop!!Nico Rehwaldt#NRe!!Nicolás Cañibano#Cani!!Nicolas Cellier#nice!!Nicolas Petton#np!!Nik Boyd#nik!!Niko Schwarz#nes!!Nikolay Suslov#sn!!Nils-Olof Biderholt#nobi!!Nishihara Satoshi#nsh!!Norbert Hartl#noha!!Norberto Manzanos#nm!!Norman Nunley#ndn!!Noury Bouraqadi#nbo#noury!!Oddbjorn Aasen#obion!!Olaf Buwen#obu!!Oleg Korsak#kamaz!!Pablo Iaría#pi#PI!!Pablo J. E. Rodriguez#pjer!!Paola Zugno#pazu#pz!!Paolo Bernardi#pber!!Paragkamat#kamatp!!Pascal Vollmer#pv!!Patrick C. Castle#pcc!!Patrick Chénais#pch!!Patrick Collison#pc!!Patrick Defrance#pdf!!Patrick Rein#pre!!Patty Gadegast#pg!!Paul Baumann#!!Paul Horth#Phorth!!Paul Jensen#psj!!Paul L. Fries#plf!!Paul McDonough#pnm!!Paul Saitta#PBS!!Paul Steffen#rabit!!Paul W. Mitchell#pwm!!Paul Ward#PW!!Pavel Krivanek#pk!!Pavel Tkachenko#p4s!!Pedro Morte Rolo#pmr!!Peter Armianov#peshe!!Peter Ernicke#per!!Peter H. Meadows#phm!!Peter Hechler#phec!!Peter Olsen#pcolsen!!Peter Osburg#PO!!Peter Schuller#ps!!Peter Willliam Lount#pwl!!Petr Fischer#pf!!Petr Polasek#PP!!Phil B.#pb!!Phil Hargett#PH!!Philip E. Evans#pevans!!Philipp Engelhard#pe!!Philipp Tessenow#phite!!Philippe Back#pba#pbk!!Pierce Ng#pn!!Piero Sartini#PS!!Prashanth Hebbar#hebbarp!!Rabih Nassrallah#laiqui!!Radoslav Hodnicak#rh!!Radoslav Hodnicak#rh!!Radovan Cervenka#RCE!!Rafa Komtur#refek!!Ragnar Hojland Espinosa#RHE!!Rainer Keller#rke!!Rajeev Lochan#rajeev#rajiv!!Ralf Thalau#rt!!Ralph Johnson#rej!!Ralph P. Boland#rpb!!Ram Krishnan#rkris!!Ramiro Diaz Trepat#rdt!!Ramon Leon#rjl!!Randal L. Schwartz#rls!!Randall S. Bohn#rsb!!Ranjan Bagchi#RjB!!Ray Leyva#rlm!!Raymond Asselin#rga!!Reinier van Loon#RvL!!Renaud Lenne#RL!!Rex Morikawa#rxm!!Ricardo J. Ferreira#rjf!!Ricardo Moran#Richo!!Richard A. Harmon#RAH!!Richard Frost#ref!!Richard Karpinski#rhk!!Richard Manning#rsm!!Richard Staehli#rstaehli!!Rick Hedin#rmh!!Rick McGeer#rm!!Ridvan Yildirim#ridvan!!Rik Fischer Smoody#Rik!!Rob Miller#Miller2040!!Rob Vens#RWV!!Robert "Bruce" Carleton#rbc!!Robert B. Livingston#rbl!!Robert Hawley#RJH!!Robert Hirschfeld#rhi!!Robert Krahn#rkrk!!Robert P. Jarvis#rpj!!Robert Rübner#ror!!Robert Withers#rww#rwwj!!Roberto Amorim#rfa!!Robin Joi Green#RJG!!Robin Summers#grs!!Rodney Polkinghorne#resp!!Roel Wuyts#rw!!Roger Verheyden#rv#rver!!Roger Vossler#rav!!Roland Bertuli#rb!!Roland Rutkowski#rolando!!Romain Robbes#rr!!Roman Hubacek#bobr123!!Romolo pranzetti#rp!!Ron Teitelbaum#RON#Ron!!Ron Welch#rpw!!Ronald Cummings#ronald!!Ronaldo M. Ferraz#rmf!!Ruben Bakker#bar!!Rúben Cabaço#rc!!Rudi Angela#AREA!!Russ Shotts#ras!!Russell Allen#rca!!Russell D. Steen#rds!!Russell Penney#rop!!Russell Smithers#!!Ryan Bell#kofno!!Ryan Macnak#ryan!!Sam Adams#ssa!!Sam S. Adams#ssa!!Samir Saidani#sam!!Samuel Henry#sjh!!Samuel Tardieu#st!!Samuel Tesla#sjt!!Santosh Kumar Choure#!!Sascha Baldt#saba!!Savra Osoule#DKS!!Schnulli Pups#sp!!Scott Wallace#sw!!Sean DeNigris#spd!!Sean Glazier#sag!!Sean Hayden#swh!!Sean P. Denigris#spd!!Sean Ross#sdr!!Sebastian Fleissner#SeF!!Sebastian G#gonisebas!!Sebastian Wain#swain!!Serge Stinckwich#zz!!Sergei Zhuravlev#zerg!!Seung Bum Kim#picxenk!!Shahul Hameed#j shahul hameed!!Shane Roberts#srsr!!Shaun Johnson#SJ!!Shawn Macintyre#sdm!!Shawn Murphy#sfm!!Shikatsu Laythe#Shikatsu!!Shuky Chen#SC!!Silvan Graf#sg!!Simon Dobson#sd!!Simon Fielding#sf!!Simon Guest#sjg!!Simon Holland#SimonHolland!!Simon Houlbrooke#nomis!!Simon Michael#sm!!Soehnke Mundt#soehnkemundt!!Srikantha Herath#SH!!Stefan Barthel#sb!!Stefan Berger#stb!!Stefan Matthias Aust#md#sma!!Stefan Udo Hanenberg#sha!!Stefan Welebny#stw!!Steinmann Ruedi#rst!!Stephan B. Wessels#sbw!!Stephan Rudlof#sr!!Stéphane Ducasse#SD#sd!!Stéphane Rollandin#spfa!!Stephane Schitter#stefs!!Stephanie Hamburg#MUTTLYSTEPHANIE!!Stephen Smith#sst!!Stephen Travis Pope#stp!!Stephen Vincent Pair#svp!!Steve Davies#sld!!Steve Elkins#sge!!Steve Fuller#snf!!Steve Gilbert#slg!!Steve Hunter#skh!!Steve Knight#knighty!!Steve Mccusker#smcc!!Steve Messamore#slm!!Steve Sanderson#sms!!Steve Wart#swart!!Steve Wessels#!!Steven Darcy#SMD!!Steven Greenberg#greenbes!!Steven Rodriguez#optionshiftk!!Steven Swerling#sps!!Sudheendra Hangal#hangal!!Sungjin Chun#chunsj!!Suzuki Tetsuya#tetsuya!!Syed Abid#taxman!!Syed Masoodahmad#masden56!!Sylvia Sharma#sharma!!Symon Chalk#symonc!!Takashi Yamamiya#tak!!Tansel Ersavas#mte#MTE!!Tarek Demiati#TD!!Ted Bracht#TB#TB1!!Ted Kaehler#tk!!Terry Jenkins#TCJ!!Thierry Reignier#TREG!!Thijs Janssen#TJ!!Thomas Bernitt#tber!!Thomas Fröb#thf!!Thomas Hemme#Namamazu!!Thomas J Keller#TJK!!Thomas Kowark#tk!!Thomas M. Breuel#tmb!!Thomas Mahler#ThMa!!Thomas Stambaugh#tms!!Thomas Zimmermann#TZ!!Tim Cuthbertson#tec!!Tim Felgentreff#tfel!!Tim Lewis#TimLewis!!Tim Olson#tao!!Tim Rowledge#TPR#tpr!!Timm Knape#tik!!Timothy Falconer#teefal!!Timothy M#tty!!Timothy Retz#tgr!!Tobias Isenberg#ti!!Tobias Pape#topa!!Todd Blanchard#tb!!Tom Counsell#tamc!!Tom Dailey#td!!Tom Koenig#tlk!!Tom Plick#tap!!Tom Rushworth#tbr!!Tommy Thorn#tt!!Tomohiro Oda#TO!!Tony Garnock-Jones#tonyg!!Tony Zampogna#zamp!!Torge Husfeldt#th!!Torsten Bergmann#tbn#TBN!!Torsten Sadowski#ts!!Travis Kay#tkay#tlk!!Trygve Reenskaug#TRee!!Tyler Coumbes#mtc!!Tzaddi Beltaine#tsb!!Udo Schneider#udos!!Vaidotas Didžbalis#vd!!Vassili Bykov#vb!!Vernon Marsden#vmars!!Vijay Mathew Pandyalakal#vmp!!Vladimir Janousek#vj!!Volker Bäcker#volker!!Wally Cash#wac!!Walter Wilhelm#ww!!Ward Cunningham#ward!!Wayne Braun#wb!!Wayne D. Elias#wdelias!!Webb Mcdonald#wxm!!Wilkes Joiner#dwj!!Willem van Asperen#wva!!William Hess#WFH!!William Hidden#whidden!!Wolfgang Eder#edw!!Wolfgang Helbig#whg!!Woon Yeo#!!Wuilmer Olaya Bardales#wob!!Yagendra Dutt Tripathi#yd!!Yang Ha Nguyen#yhm!!Yann Monclair#YM!!Yanni Chiu#yj!!Yasuji Nakayama#yasuji!!Yoshiki Ohshima#yo!!Yuji Ichikawa#ich!!Yunhee Lee#yhl!!Yutaka Kamite#yk!!Zdenek Novy#Zdenye#ZN!!Zeljko Nesic#Poparasan!!Zeynep Besen#zeyno'! !

!SystemNavigation class methodsFor: 'preferences' stamp: 'ul 8/8/2013 13:10'!
thoroughSenders
	"Accessor for the system-wide preference"
	
	<preference: 'Thorough senders.'
		category: #general
		description: 'If true, then ''senders'' browsers will dive inside structured literals in their search.'
		type: #Boolean>
	^ThoroughSenders ifNil: [ true ]! !

!SystemNavigation class methodsFor: 'preferences' stamp: 'ul 8/8/2013 13:00'!
thoroughSenders: aBoolean
	"Accessor for the system-wide preference"
	
	ThoroughSenders := aBoolean! !

!SystemNavigation methodsFor: 'query' stamp: 'eem 11/4/2011 09:20'!
allAccessesTo: instVarName from: aClass
	"Return a sequence of MethodReferences of all methods of aClass or it's sub/superclass that refer to the instance variable instVarName."
	
	| result |
	result := OrderedCollection new.
	aClass withAllSubAndSuperclassesDo:
		[ :class | 
		(class whichSelectorsAccess: instVarName) do:
			[ :selector |
			result add: (MethodReference class: class selector: selector) ] ].
	^result! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 8/8/2013 12:31'!
allBehaviors

	^Generator on: [ :generator |
		self allBehaviorsDo: [ :each |
			generator yield: each ] ]! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 11/2/2013 13:44'!
allBehaviorsDo: aBlock 
	"Evaluate the argument, aBlock, for each kind of Behavior in the system 
	(that is, Object and its subclasses and Traits).
	ar 7/15/1999: The code below will not enumerate any obsolete or anonymous
	behaviors for which the following should be executed:

		Smalltalk allObjectsDo:[:obj| obj isBehavior ifTrue:[aBlock value: obj]].

	but what follows is way faster than enumerating all objects."

	self environment allClassesAndTraitsDo: [ :class |
		aBlock value: class.
		class isTrait ifFalse: [ "class of a Trait is Trait, there are no MetaTraits"
			aBlock value: class class ] ]! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 8/8/2013 13:12'!
allCallsOnClass: aBehavior
	"Answer a sorted collection of all the methods that refer to aBehavior."
	
	| theClass result |
	theClass := aBehavior theNonMetaClass.
	result := self 
		allCallsOn: (
			self class thoroughSenders
				ifTrue: [  theClass name ]
				ifFalse: [ theClass environment bindingOf: theClass name ])
		fromBehaviors: self allBehaviors
		sorted: false.
	theClass environment allClassesDo: [ :class |
		(class sharedPools includes: theClass) ifTrue: [
			result add: (ClassReference class: class) ] ].
	^result sort! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 8/8/2013 12:31'!
allCallsOn: aLiteral 
	"Answer a sorted collection of all the methods that call on aLiteral even deeply embedded in literal array."
	"self default browseAllCallsOn: #open:label:."
	
	^self
		allCallsOn: aLiteral
		fromBehaviors: self allBehaviors
		sorted: true! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 3/13/2011 20:33'!
allCallsOn: firstLiteral and: secondLiteral
	"Answer a sorted collection of all the methods that call on both firstLiteral and secondLiteral."

	| firstList secondList |
	firstList := self allCallsOn: firstLiteral.
	secondList := (self 
		allCallsOn: secondLiteral
		fromBehaviors: (firstList collect: [ :each | each actualClass ] as: IdentitySet)
		sorted: false) asSet.
	firstList removeAllSuchThat: [ :each | (secondList includes: each) not ].
	^firstList
	! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 8/8/2013 13:12'!
allCallsOn: aLiteral fromBehaviors: behaviors sorted: sorted
	"Answer a collection of all the methods implemented by behaviors that call on aLiteral even deeply embedded in literal array."
	
	| result special thorough byte |
	result := OrderedCollection new.
	special := Smalltalk hasSpecialSelector: aLiteral ifTrueSetByte: [ :b | byte := b ].
	"Possibly search for symbols imbedded in literal arrays"
	thorough := aLiteral isSymbol and: [ self class thoroughSenders ].
	behaviors do: [ :behavior |
		| list | 
		list := behavior whichSelectorsReferTo: aLiteral special: special byte: byte thorough: thorough.
		list do: [ :selector |
			result add: (MethodReference class: behavior selector: selector) ] ].
	sorted ifTrue: [ result sort ].
	^result! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 3/30/2012 11:49'!
allCallsOn: aSymbol from: aBehavior
	"Answer a sorted collection of all the methods from aBehavior and it's subclasses that call on aSymbol."

	^self allCallsOn: aSymbol fromBehaviors: aBehavior withAllSubclasses sorted: true! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 3/14/2011 20:50'!
allCallsOn: aLiteral localToPackage: packageNameOrInfo
	"Answer a sorted collection of MethodReferences for all the methods that call on aLiteral in the given package."

	^self
		allCallsOn: aLiteral
		fromBehaviors: (self packageInfoFor: packageNameOrInfo) classesAndMetaClasses
		sorted: true! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 3/14/2011 20:50'!
allCallsOn: aLiteral localTo: aClass
	"Answer a sorted collection of MethodReferences for all the methods that call on aLiteral in, above or below the given class."

	^self
		allCallsOn: aLiteral
		fromBehaviors: (Array streamContents: [ :stream |
			aClass theNonMetaClass withAllSuperAndSubclassesDoGently: [ :each |
				stream nextPut: each ].
			aClass theNonMetaClass class withAllSuperAndSubclassesDoGently: [ :each |
				stream nextPut: each ] ])
		sorted: true! !

!SystemNavigation methodsFor: 'query' stamp: 'fbs 5/10/2013 08:39'!
allClasses
	"currently returns all the classes defined in Smalltalk but could be customized 
	for dealing with environments and in such a case would return on really all the classes"

	^ self environment allClasses! !

!SystemNavigation methodsFor: 'query' stamp: 'fbs 5/10/2013 08:42'!
allClassesAndTraits
	
	^ self environment allClassesAndTraits

	! !

!SystemNavigation methodsFor: 'query' stamp: 'fbs 5/10/2013 08:51'!
allClassesDo: aBlock
	"currently returns all the classes defined in Smalltalk but could be customized 
	for dealing with environments and  in such a case would work on really all the classes"

	^ self environment allClassesDo: aBlock

	! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 3/18/2011 00:15'!
allClassesImplementing: aSelector  
	"Answer an collection of all classes that implement the message aSelector."

	| result |
	result := OrderedCollection new.
	self allBehaviorsDo: [ :behavior | 
		(behavior includesSelector: aSelector) ifTrue: [
			result add: behavior ] ].
	^result! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/18/2003 10:04'!
allGlobalRefs
	"Answer a set of symbols that may be refs to Global names.  In some sense we should only need the associations, but this will also catch, eg, HTML tag types."

	^ self allGlobalRefsWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: 'fbs 5/12/2013 10:55'!
allGlobalRefsWithout: classesAndMessagesPair 
	"Answer a set of symbols that may be refs to Global names. In some  
	sense we should only need the associations, but this will also catch, eg,  
	HTML tag types. This method computes its result in the absence of  
	specified classes and messages."
	"may be a problem if namespaces are introduced as for the moment  
	only Smalltalk is queried. sd 29/4/03"
	| globalRefs absentClasses absentSelectors |
	globalRefs := IdentitySet new: CompiledMethod instanceCount.
	absentClasses := classesAndMessagesPair first.
	absentSelectors := classesAndMessagesPair second.
	"sd 29/04/03"
	Cursor execute
		showWhile: [self environment allClassesDo:
				[:cls | ((absentClasses includes: cls name)
						ifTrue: [{}]
						ifFalse: [{cls. cls class}])
						do: [:cl | (absentSelectors isEmpty
								ifTrue: [cl selectors]
								ifFalse: [cl selectors copyWithoutAll: absentSelectors])
								do: [:sel | "Include all capitalized symbols for good 
									measure"
									(cl compiledMethodAt: sel) literalsDo: [:m |
										((m isSymbol)
												and: [m size > 0
														and: [m first canBeGlobalVarInitial]])
											ifTrue: [globalRefs add: m].
										(m isMemberOf: Array)
											ifTrue: [m
													do: [:x | ((x isSymbol)
																and: [x size > 0
																		and: [x first canBeGlobalVarInitial]])
															ifTrue: [globalRefs add: x]]].
										m isVariableBinding
											ifTrue: [m key
													ifNotNil: [globalRefs add: m key]]]]]]].
	^ globalRefs! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 14:47'!
allImplementedMessages
	"Answer a set of all the messages that are implemented in the system."
	
	^self allImplementedMessagesWithout: #(() ())! !

!SystemNavigation methodsFor: 'query' stamp: 'fbs 5/10/2013 17:26'!
allImplementedMessagesWithout: behaviorsAndSelectorsPair 
	"Answer a set of all the selectors that are implemented in the system, computed in the absence of the supplied behaviors and selectors."
	
	| selectors behaviorsToReject |
	selectors := IdentitySet new.
	behaviorsToReject := behaviorsAndSelectorsPair first asIdentitySet.
	self environment allClassesAndTraitsDo: [ :behavior |
		(behaviorsToReject includes: behavior name) ifFalse: [
			selectors
				addAll: behavior selectors;
				addAll: behavior classSide selectors ] ].
	behaviorsAndSelectorsPair second do: [ :each |
		selectors remove: each ].
	^selectors! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 14:27'!
allImplementorsOf: aSelector 
	"Answer a sorted collection of all the methods which are named aSelector."
	
	| result |
	result := OrderedCollection new.
	self allBehaviorsDo: [ :class |
		(class includesSelector: aSelector) ifTrue: [
			result add: (MethodReference class: class selector: aSelector) ] ].
	^result sort! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 14:30'!
allImplementorsOf: aSelector  localToPackage: packageNameOrInfo
	"Answer a sorted collection of all the methods named aSelector in the given package."

	| result |
	result := OrderedCollection new.
	(self packageInfoFor: packageNameOrInfo) actualMethodsDo: [ :method |
		method selector = aSelector ifTrue: [
			result add: method methodReference ] ].
	^result sort! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 14:29'!
allImplementorsOf: aSelector localTo: aClass 
	"Answer a sorted collection of all the methods that are named aSelector in, above, or below the given class."
	
	| result |
	result := OrderedCollection new.
	{ aClass theNonMetaClass. aClass theMetaClass } do: [ :baseClass |
		baseClass withAllSuperAndSubclassesDoGently: [ :class | 
			(class includesSelector: aSelector) ifTrue: [
				result add: (MethodReference class: class selector: aSelector) ] ] ].
	^result sort! !

!SystemNavigation methodsFor: 'browse' stamp: 'ich. 5/16/2004 01:05'!
allMethodsInCategory: category 
	| aCollection |
	aCollection := SortedCollection new.
	Cursor wait showWhile:
		[self allBehaviorsDo:
			[:x | (x allMethodsInCategory: category) do:
				[:sel | aCollection add: x name , ' ' , sel]]].
	^aCollection.
	! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 13:57'!
allMethodsSelect: aBlock 
	"Answer a sorted collection of each method that, when used as the block  
	argument to aBlock, gives a true result."
	
	| result |
	result := OrderedCollection new.
	self allSelectorsAndMethodsDo: [ :behavior :selector :method |
		(aBlock value: method) ifTrue: [
			result add: (MethodReference class: behavior selector: selector) ] ].
	^result sort! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 14:00'!
allMethodsSelect: aBlock localToPackage: packageNameOrInfo
	"Answer a sorted collection of each method in the given package for which the evaluation of aBlock with the method answers true."

	| result |
	result := OrderedCollection new.
	(self packageInfoFor: packageNameOrInfo) actualMethodsDo: [ :aMethod |
		(aBlock value: aMethod) ifTrue: [
			result add: aMethod methodReference ] ].
	^result sort! !

!SystemNavigation methodsFor: 'query' stamp: 'fbs 5/21/2013 23:04'!
allMethodsSelect: aBlock localTo: aClass
	"Answer a SortedCollection of each methodr in, above, or below the given
	 class that, when used as the argument to aBlock, gives a true result."

	| aSet collectInSet |
	aSet := Set new.
	collectInSet := [:class |
		class selectorsAndMethodsDo:
			[:aSelector :aMethod|
			(aBlock value: aMethod) ifTrue:
				[aSet add: (MethodReference class: class selector: aSelector)]]].
	Cursor wait showWhile:
		[aClass theNonMetaClass
			withAllSuperAndSubclassesDoGently: collectInSet.
		aClass theNonMetaClass class
			withAllSuperAndSubclassesDoGently: collectInSet].
	^aSet! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 15:01'!
allMethodsWithSourceString: aString matchCase: caseSensitive
	"Answer a sorted collection of all the methods that contain, in source code, aString as a substring. Search the class comments also"

	| result |
	result := OrderedCollection new.
	CurrentReadOnlySourceFiles cacheDuring: [
		'Searching all source code...'
			displayProgressFrom: 0 to: Smalltalk classNames size
			during: [ :bar |
				| count previousBehavior |
				count := 0.
				previousBehavior := nil.
				self allSelectorsAndMethodsDo: [ :behavior :selector :method |
					behavior == previousBehavior ifFalse: [
						bar value: (count := count + 1).
						previousBehavior := behavior.
						(behavior organization classComment asString
							includesSubstring: aString
							caseSensitive: caseSensitive) ifTrue: [
								result add: (MethodReference class: behavior selector: #Comment) ] ].
					(method getSource asString
						includesSubstring: aString
						caseSensitive: caseSensitive) ifTrue: [
							result add: (MethodReference class: behavior selector: selector) ] ] ] ].
	^result sort! !

!SystemNavigation methodsFor: 'query' stamp: 'eem 3/10/2014 19:05'!
allObjects
	"Answer an Array of all objects in the system.  Fail if
	 there isn't enough memory to instantiate the result."
	<primitive: 178>
	^self primitiveFailed! !

!SystemNavigation methodsFor: 'query' stamp: 'eem 3/11/2014 13:59'!
allObjectsDo: aBlock 
	"Evaluate the argument, aBlock, for each object in the system, excluding immediates
	 such as SmallInteger and Character."
	self allObjectsOrNil
		ifNotNil: [:allObjects| allObjects do: aBlock]
		ifNil:
			["Fall back on the old single object primitive code.  With closures, this needs
			  to use an end marker (lastObject) since activation of the block will create
			  new contexts and cause an infinite loop.  The lastObject must be created
			  before calling someObject, so that the VM can settle the enumeration (e.g.
			  by flushing new space) as a side effect of  someObject"
			| object lastObject |
			lastObject := Object new.
			object := self someObject.
			[lastObject == object or: [0 == object]] whileFalse:
				[aBlock value: object.
				 object := object nextObject]]! !

!SystemNavigation methodsFor: 'query' stamp: 'eem 3/11/2014 13:56'!
allObjectsOrNil
	"Answer an Array of all objects in the system.  Fail if there isn't
	 enough memory to instantiate the result and answer nil."
	<primitive: 178>
	^nil! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 20:53'!
allObjectsSelect: aBlock 
	"Evaluate the argument, aBlock, for each object in the system excluding 
	SmallIntegers. Return a collection af all objects for whom the value is 
	true. "
	^ Array
		streamContents: [:s | self
				allObjectsDo: [:object | (aBlock value: object)
						ifTrue: [s nextPut: object]]]! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 3/12/2011 15:08'!
allPrimitiveMethods
	"Answer a collection of all the methods that are implemented by primitives."

	| result |
	result := OrderedCollection new.
	self allSelectorsAndMethodsDo: [ :behavior :selector :method |
		method primitive ~= 0 ifTrue: [
			result add: (String streamContents: [ :stream |
				stream
					nextPutAll: behavior name;
					space;
					nextPutAll: selector;
					space;
					print: method primitive ]) ] ].
	^result! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 3/12/2011 15:08'!
allPrimitiveMethodsInCategories: aList 
	"Answer an OrderedCollection of all the methods that are implemented by 
	primitives in the given categories. 1/26/96 sw"
	"SystemNavigation new allPrimitiveMethodsInCategories:  
	#('Collections-Streams' 'Files-Streams' 'Files-Abstract' 'Files-Macintosh')"

	| result categories |
	result := OrderedCollection new.
	categories := aList collect: [ :each | each asSymbol ].
	categories size > 10 ifTrue: [ categories := categories asIdentitySet ].
	self allBehaviorsDo: [ :behavior | 
		(aList includes: behavior category) ifTrue: [
			behavior selectorsAndMethodsDo: [ :selector :method | 
				method primitive ~= 0 ifTrue: [
					result add: (String streamContents: [ :stream |
						stream
							nextPutAll: behavior name;
							space;
							nextPutAll: selector;
							space;
							print: method primitive ]) ] ] ] ].
	^result! !

!SystemNavigation methodsFor: 'query' stamp: 'fbs 5/21/2013 22:40'!
allReferencesToPool: aPool from: aClass
	"Answer all the references to variables from aPool"
	| list |
	list := OrderedCollection new.
	aClass withAllSubclassesDo:[:cls|
		cls selectorsAndMethodsDo:[:sel :meth|
			(meth hasLiteralSuchThat: [:lit| lit isVariableBinding and:[(aPool bindingOf: lit key) notNil]]) ifTrue:
				[list add:(MethodReference class: cls selector: sel)]]].
	^list! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 3/12/2011 14:53'!
allSelectorsAndMethodsDo: aBlock 
	"Evaluate aBlock for all selectors and methods of all behaviors in this image."
	
	self allBehaviorsDo: [ :behavior |
		behavior selectorsAndMethodsDo: [ :selector :method |
			aBlock value: behavior value: selector value: method ] ]! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 5/5/2003 09:18'!
allSelectorsWithAnyImplementorsIn: selectorList 
	"Answer the subset of the given list which represent method selectors 
	which have at least one implementor in the system."
	| good |
	good := OrderedCollection new.
	self allBehaviorsDo: [:class | selectorList
				do: [:aSelector | (class includesSelector: aSelector)
						ifTrue: [good add: aSelector]]].
	^ good asSet asSortedArray" 
	SystemNavigation new selectorsWithAnyImplementorsIn: #( contents 
	contents: nuts)
	"! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 22:35'!
allSelect: aBlock 
	"Answer a sorted collection of each method that, when used as the block argument to aBlock, gives a true result."

	| result |
	result := OrderedCollection new.
	self allSelectorsAndMethodsDo: [ :behavior :selector :method |
		(aBlock value: method) ifTrue: [
			result add: behavior name , ' ' , selector ] ].
	^result sort! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 20:12'!
allSentMessages
	"Answer the set of selectors which are sent somewhere in the system."
	^ self  allSentMessagesWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: 'fbs 5/10/2013 22:03'!
allSentMessagesWithout: classesAndMessagesPair 
	"Answer the set of selectors which are sent somewhere in the system,  
	computed in the absence of the supplied classes and messages."
	| sent absentClasses absentSelectors |
	sent := IdentitySet new: CompiledMethod instanceCount.
	absentClasses := classesAndMessagesPair first.
	absentSelectors := classesAndMessagesPair second.
	"sd 29/04/03"
	Cursor execute showWhile: [
		self environment allClassesAndTraitsDo: [:cls |
			((absentClasses includes: cls name)
				ifTrue: [{}]
				ifFalse: [{cls. cls classSide}])
					do: [:each | (absentSelectors isEmpty
						ifTrue: [each selectors]
						ifFalse: [each selectors copyWithoutAll: absentSelectors])
						do: [:sel | "Include all sels, but not if sent by self"
							(each compiledMethodAt: sel) literalsDo: [:m | 
									(m isSymbol)
										ifTrue: ["might be sent"
											m == sel
												ifFalse: [sent add: m]].
									(m isMemberOf: Array)
										ifTrue: ["might be performed"
											m
												do: [:x | (x isSymbol)
														ifTrue: [x == sel
																ifFalse: [sent add: x]]]]]]]]].
	"The following may be sent without being in any literal frame"
	Smalltalk specialSelectors do: [:sel | sent add: sel].
	Smalltalk presumedSentMessages	do: [:sel | sent add: sel].
	^ sent.! !

!SystemNavigation methodsFor: 'query' stamp: 'cmm 4/27/2015 15:08'!
allStoresInto: varName from: aClass 
	"Answer a sequence of MewthodReferences for all the receiver's methods 
	 or any methods of a subclass/superclass that assign to the instance variable name."
	"self new allStoresInto: 'contents' from: Collection."
	| result roots |
	result := OrderedCollection new.
	(aClass theNonMetaClass inheritsFrom: SharedPool) ifTrue:
		[roots := self allClasses select: [:class| class sharedPools includes: aClass theNonMetaClass].
		  roots do:
			[:root|
			 root withAllSubclassesDo:
				[:class|
				 (class whichMethodsStoreInto: varName), 
				 (class class whichMethodsStoreInto: varName) do:
					[:eachMethod|
					 result add: eachMethod methodReference]]]].
	aClass withAllSubAndSuperclassesDo:
		[ : class |
		(class theNonMetaClass whichMethodsStoreInto: varName),
		(class theMetaClass whichMethodsStoreInto: varName) do:
			[ : eachMethod |
			result add: eachMethod methodReference ] ].
	^result! !

!SystemNavigation methodsFor: 'query' stamp: 'eem 10/7/2015 15:33'!
allUnboundMethods
	"Answer all CompiledMehtods that are not in the class hierarchy"
	"self systemNavigation allUnboundMethods"
	^CompiledMethod allSubInstances select:
		[:m|
		 m methodClass
			ifNil: [true]
			ifNotNil:
				[:mc|
				(mc compiledMethodAt: m selector ifAbsent: []) ~~ m]]! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 22:27'!
allUncommentedMethodsWithInitials: targetInitials
	"Return a sorted collection with all uncommented methods whose initials (in the
	time-stamp, as logged to disk) match the given initials, in chronological order."

	^CurrentReadOnlySourceFiles cacheDuring: [
		| result |
		result := OrderedCollection new.
		self allSelectorsAndMethodsDo: [ :behavior :selector :method |
			method timeStamp ifNotEmptyDo: [ :timeStamp |
				| initials |
				initials := timeStamp substrings first.
				initials first isDigit ifFalse: [
					(initials = targetInitials and: [ 
						(behavior firstPrecodeCommentFor: selector) isNil ]) ifTrue: [
							result add: (MethodReference
								class: behavior
								selector: selector ) ] ] ] ].
		result ]! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 22:30'!
allUnimplementedCalls
	"Answer a collection of each message that is sent by an expression in a method but is not implemented by any object in the system."

	| result implementedMessages |
	implementedMessages := self allImplementedMessages.
	result := OrderedCollection new.
	self allSelectorsAndMethodsDo: [ :behavior :selector :method |
		method messagesDo: [ :each |
			(implementedMessages includes: each) ifFalse: [
					result add: (String streamContents: [ :stream |
						stream
							nextPutAll: behavior name;
							space;
							nextPutAll: selector;
							space;
							nextPutAll: 'calls: ';
							nextPutAll: each ]) ] ] ].
	^result! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 22:31'!
allUnimplementedNonPrimitiveCalls
	"Answer an collection of each message that is sent by an expression in a method but is not implemented by any object in the system. This list won't include primitive methods."

	| result implementedMessages |
	implementedMessages := self allImplementedMessages.
	result := OrderedCollection new.
	self allSelectorsAndMethodsDo: [ :behavior :selector :method |
		method primitive = 0 ifTrue: [
			method messagesDo: [ :each |
				(implementedMessages includes: each) ifFalse: [
					result add: (String streamContents: [ :stream |
						stream
							nextPutAll: behavior name;
							space;
							nextPutAll: selector;
							space;
							nextPutAll: 'calls: ';
							nextPutAll: each ]) ] ] ] ].
	^result! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 13:07'!
allUnreferencedClassVariablesOf: aClass
	"Answer a list of the names of all the receiver's unreferenced class  
	vars, including those defined in superclasses"
	| aList |
	aList := OrderedCollection new.
	aClass withAllSuperclasses
		reverseDo: [:aSuperClass | aSuperClass classVarNames
				do: [:var | (self allCallsOn: (aSuperClass classPool associationAt: var)) isEmpty
						ifTrue: [aList add: var]]].
	^ aList! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 20:12'!
allUnSentMessages
	"SystemNavigation new allUnSentMessages"
	"Answer the set of selectors that are implemented by some object in the  
	system but not sent by any."
	^ self allUnSentMessagesWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 20:13'!
allUnSentMessagesIn: selectorSet 
	"Answer the subset of selectorSet which are not sent anywhere in the 
	system. "
	^ selectorSet copyWithoutAll: self allSentMessages! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 19:19'!
allUnSentMessagesWithout: classesAndMessagesPair 
	"Answer the set of selectors that are implemented but not sent, computed  
	in the absence of the supplied classes and messages."
	^ (self  allImplementedMessagesWithout: classesAndMessagesPair)
		copyWithoutAll: (self  allSentMessagesWithout: classesAndMessagesPair)! !

!SystemNavigation methodsFor: 'query' stamp: 'fbs 5/10/2013 22:26'!
allUnusedClassesWithout: classesAndMessagesPair 
	"Enumerates all classes in the system and returns a list of those that are 
	apparently unused. A class is considered in use if it (a) has subclasses  
	or (b) is referred to by some method or (c) has its name in use as a  
	literal."
	"SystemNavigation new unusedClasses"

	| unused |
	unused := self environment classNames asIdentitySet
				copyWithoutAll: (self allGlobalRefsWithout: classesAndMessagesPair).
	^ unused
		reject: [:cName | | cl | 
			cl := self environment at: cName.
			cl subclasses isEmpty not
				or: [cl inheritsFrom: FileDirectory]]! !

!SystemNavigation methodsFor: 'browse' stamp: 'eem 4/30/2012 16:31'!
browseAllAccessesTo: instVarName from: aClass
	"Create and schedule a Message Set browser for all the receiver's methods or any methods of a subclass/superclass that refer to the instance variable name."
	"self new browseAllAccessesTo: 'contents' from: Collection."
	
	^ self 
		browseMessageList: [ (self allAccessesTo: instVarName from: aClass) sort ]
		name: 'Accesses to ' , instVarName 
		autoSelect: instVarName! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 8/8/2013 12:41'!
browseAllCallsOnClass: aBehavior
	"Create and schedule a message browser on each method that refers to aBehavior."
	"self default browseAllCallsOnClass: Array"
	
	| behaviorName |
	behaviorName := aBehavior theNonMetaClass name.
	self
		browseMessageList: [ self allCallsOnClass: aBehavior ]
		name: 'Users of ', behaviorName
		autoSelect: behaviorName! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 3/14/2011 20:53'!
browseAllCallsOn: aLiteral 
	"Create and schedule a message browser on each method that refers to aLiteral."
	"self default browseAllCallsOn: #open:label:."

	self headingAndAutoselectForLiteral: aLiteral do: [ :label :autoSelect |
		self
			browseMessageList: [ self allCallsOn: aLiteral ]
			name: label
			autoSelect: autoSelect]! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 5/9/2011 02:20'!
browseAllCallsOn: literal1 and: literal2 
	"Create and schedule a message browser on each method that calls on the two Symbols, literal1 and literal2."
	"self default browseAllCallsOn: #at: and: #at:put:."

	^self 
		browseMessageList: [ self allCallsOn: literal1 and: literal2 ]
		name: literal1 printString , ' -and- ' , literal2 printString! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 3/14/2011 21:02'!
browseAllCallsOn: aLiteral from: aBehavior
	"Create and schedule a Message Set browser for all the methods that call on aLiteral within aBehavior."
	"self default browseAllCallsOn: #/ from: Number"

	^self headingAndAutoselectForLiteral: aLiteral do: [ :label :autoSelect |
		self 
			browseMessageList: [ self  allCallsOn: aLiteral from: aBehavior ]
			name: label, ' from ', aBehavior name
			autoSelect: autoSelect ]

	! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 3/14/2011 21:02'!
browseAllCallsOn: aLiteral localToPackage: packageNameOrInfo
	"Create and schedule a message browser on each method in the given package that refers to aLiteral."
	"self default browseAllCallsOn: #open:label: localToPackage: 'Tools'."

	self headingAndAutoselectForLiteral: aLiteral do: [ :label :autoSelect |
		self
			browseMessageList: [ 
				self 
					allCallsOn: aLiteral
					localToPackage: packageNameOrInfo ]
			name: label, ' local to package ', (self packageInfoFor: packageNameOrInfo) name
			autoSelect: autoSelect ]! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 3/14/2011 21:02'!
browseAllCallsOn: aLiteral localTo: aBehavior
	"Create and schedule a message browser on each method in or below the given class that refers to aLiteral."
	"self default browseAllCallsOn: #open:label: localTo: CodeHolder"

	aBehavior ifNil: [ ^self inform: 'No behavior selected.' ].
	self headingAndAutoselectForLiteral: aLiteral do: [ :label :autoSelect |
		self
			browseMessageList:  [ self allCallsOn: aLiteral from: aBehavior ]
			name: label, ' local to ', aBehavior name
			autoSelect: autoSelect ]! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/15/2003 21:43'!
browseAllImplementorsOfList: selectorList
	"Create and schedule a message browser on each method that implements 
	the message whose selector is in the argument selectorList. For example, 
	Smalltalk browseAllImplementorsOf: #(at:put: size).
	1/16/96 sw: defer to the titled version"

	self browseAllImplementorsOfList: selectorList title: 'Implementors of all'! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 7/16/2011 18:48'!
browseAllImplementorsOfList: selectorList title: aTitle 
	"Create and schedule a message browser on each method that implements the message whose selector is in the argument selectorList."
	"For example, self new browseAllImplementorsOf: #(at:put: size). "
	
	| implementorLists |
	implementorLists := selectorList gather: [:each | 
		self allImplementorsOf: each ].
	implementorLists sort.
	^self browseMessageList: implementorLists name: aTitle! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/19/2003 12:15'!
browseAllImplementorsOf: selector 
	"Create and schedule a message browser on each method that implements 
	the message whose selector is the argument, selector. For example,  
	Smalltalk browseAllImplementorsOf: #at:put:."
	^ self
		browseMessageList: (self allImplementorsOf: selector)
		name: 'Implementors of ' , selector! !

!SystemNavigation methodsFor: 'browse' stamp: 'eem 8/30/2010 11:13'!
browseAllImplementorsOf: selector localToPackage: packageNameOrInfo
	"Create and schedule a message browser on each method in the given package
	that implements the message whose selector is the argument, selector. For example, 
	SystemNavigation new browseAllImplementorsOf: #at:put: localToPackage: 'Collections'."

	self browseMessageList: (self
								allImplementorsOf: selector
								localToPackage: packageNameOrInfo) asSortedCollection
		name: 'Implementors of ' , selector,
				' local to package ', (self packageInfoFor: packageNameOrInfo) name! !

!SystemNavigation methodsFor: 'browse' stamp: 'apl 4/4/2005 18:23'!
browseAllImplementorsOf: selector localTo: aClass 
	"Create and schedule a message browser on each method in or below the  
	given class  
	that implements the message whose selector is the argument, selector.  
	For example,  
	SystemNavigation new browseAllImplementorsOf: #at:put: localTo:  
	Dictionary."
	aClass
		ifNil: [^ self inform: 'no class selected'].
	^ self
		browseMessageList: ((self allImplementorsOf: selector localTo: aClass)
				collect: [:methRef | methRef actualClass name , ' ' , methRef methodSymbol])
		name: 'Implementors of ' , selector , ' local to ' , aClass name! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/16/2003 09:17'!
browseAllMethodsInCategory: category 
	^self browseMessageList: (self allMethodsInCategory: category)
		name: category! !

!SystemNavigation methodsFor: 'browse' stamp: 'ar 9/27/2005 22:41'!
browseAllObjectReferencesTo: anObject except: objectsToExclude ifNone: aBlock 
	"Bring up a list inspector on the objects that point to anObject.
	If there are none, then evaluate aBlock on anObject.  "

	| aList shortName |
	aList := Utilities pointersTo: anObject except: objectsToExclude.
	aList size > 0 ifFalse: [^aBlock value: anObject].
	shortName := (anObject name ifNil: [anObject printString]) contractTo: 20.
	aList inspectWithLabel: 'Objects pointing to ' , shortName! !

!SystemNavigation methodsFor: 'browse' stamp: 'ar 9/7/2003 18:05'!
browseAllReferencesToPool: poolOrName from: aClass
	"Open a message list on all messages referencing the given pool"
	| pool list |
	(poolOrName isString)
		ifTrue:[pool := Smalltalk at: poolOrName asSymbol]
		ifFalse:[pool := poolOrName].
	list := self allReferencesToPool: pool from: aClass.
	self
		browseMessageList: list
		name: 'users of ', poolOrName name.
	^list! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/29/2003 20:43'!
browseAllSelect: aBlock 
	"Create and schedule a message browser on each method that, when used 
	as the block argument to aBlock gives a true result. For example,  
	SystemNavigation new browseAllSelect: [:method | method numLiterals >  
	10]."
	^ self
		browseMessageList: (self allMethodsSelect: aBlock)
		name: 'selected messages'! !

!SystemNavigation methodsFor: 'browse' stamp: 'eem 4/3/2012 09:05'!
browseAllSelect: aBlock localToPackage: packageNameOrInfo
	"Create and schedule a message browser on each method in the given package
	 that, when used as the block argument to aBlock answers true. For example,  
		(SystemNavigation new
			browseAllSelect: [:method | method scanFor: 138]
			localToPackage: #System.)"
	^self
		browseMessageList: (self allMethodsSelect: aBlock localToPackage: packageNameOrInfo)
		name: 'selected messages local to package ', (self packageInfoFor: packageNameOrInfo) name! !

!SystemNavigation methodsFor: 'browse' stamp: 'eem 5/19/2008 15:12'!
browseAllSelect: aBlock localTo: aClass
	"Create and schedule a message browser on each method in or below the given class
	 that, when used as the block argument to aBlock gives a true result. For example,  
	 SystemNavigation default browseAllSelect: [:m | m numLiterals > 10] localTo: Morph."
	aClass ifNil: [^self inform: 'no class selected'].
	^self
		browseMessageList: (self allMethodsSelect: aBlock localTo: aClass) asSortedCollection
		name: 'selected messages local to ', aClass name! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 7/16/2011 18:55'!
browseAllSelect: aBlock name: aName autoSelect: autoSelectString 
	"Create and schedule a message browser on each method that, when used 
	as the block argument to aBlock gives a true result. Do not return an  
	#DoIt traces."
	"self new browseAllSelect: [:method | method numLiterals > 10] name:  
	'Methods with more than 10 literals' autoSelect: 'isDigit'"
	^ self
		browseMessageList: [ self allMethodsSelect: aBlock ]
		name: aName
		autoSelect: autoSelectString! !

!SystemNavigation methodsFor: 'browse' stamp: 'eem 11/4/2011 09:22'!
browseAllStoresInto: varName from: aClass
	"Create and schedule a Message Set browser for all the receiver's methods 
	or any methods of a subclass/superclass that refer to the instance variable name."
	"self new browseAllStoresInto: 'contents' from: Collection."
	^ self
		browseMessageList: (self allStoresInto: varName from: aClass)
		name: 'Stores into ' , varName
		autoSelect: varName! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/16/2003 09:16'!
browseAllUnimplementedCalls
	"Create and schedule a message browser on each method that includes a 
	message that is not implemented in any object in the system."

	^self   browseMessageList: self allUnimplementedCalls name: 'Unimplemented calls'! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/15/2003 21:26'!
browseAllUnSentMessages
	"Create and schedule a message browser on each method whose message is  not sent in any method in the system."
	"self new browseAllUnSentMessages"

	self browseAllImplementorsOfList: self allUnSentMessages title: 'Messages implemented but not sent'
! !

!SystemNavigation methodsFor: 'browse' stamp: 'fbs 5/21/2013 22:40'!
browseClassCommentsWithString: aString
	"Smalltalk browseClassCommentsWithString: 'my instances' "
	"Launch a message list browser on all class comments containing aString as a substring."

	| caseSensitive suffix list |

	suffix := (caseSensitive := Sensor shiftPressed)
		ifTrue: [' (case-sensitive)']
		ifFalse: [' (use shift for case-sensitive)'].
	list := Set new.
	Cursor wait showWhile: [
		Smalltalk allClassesDo: [:class | 
			(class organization classComment asString findString: aString 
							startingAt: 1 caseSensitive: caseSensitive) > 0 ifTrue: [
								list add: (
									MethodReference
										class: class
										selector: #Comment
								)
							]
		]
	].
	^ self 
		browseMessageList: list asSortedCollection
		name: 'Class comments containing ' , aString printString , suffix
		autoSelect: aString! !

!SystemNavigation methodsFor: 'browse' stamp: 'ar 9/27/2005 20:33'!
browseClassesWithNamesContaining: aString caseSensitive: caseSensitive 
	"Smalltalk browseClassesWithNamesContaining: 'eMorph' caseSensitive: true "
	"Launch a class-list list browser on all classes whose names containg aString as a substring."

	| suffix aList |
	suffix := caseSensitive
				ifTrue: [' (case-sensitive)']
				ifFalse: [' (use shift for case-sensitive)'].
	aList := OrderedCollection new.
	Cursor wait
		showWhile: [Smalltalk
				allClassesDo: [:class | (class name includesSubstring: aString caseSensitive: caseSensitive)
						ifTrue: [aList add: class name]]].
	aList size > 0
		ifTrue: [ToolSet openClassListBrowser: aList asSet asSortedArray title: 'Classes whose names contain ' , aString , suffix]! !

!SystemNavigation methodsFor: 'browse' stamp: 'ar 9/27/2005 20:34'!
browseClass: aBehavior
	| targetClass |
	targetClass := aBehavior isMeta
				ifTrue: [aBehavior theNonMetaClass]
				ifFalse: [aBehavior ].
	ToolSet browse: targetClass selector: nil! !

!SystemNavigation methodsFor: 'browse' stamp: 'ar 9/27/2005 20:34'!
browseHierarchy: aBehavior
	| targetClass |
	targetClass := aBehavior isMeta
				ifTrue: [aBehavior theNonMetaClass]
				ifFalse: [aBehavior ].
	ToolSet browseHierarchy: targetClass selector: nil.! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/16/2003 09:18'!
browseMessageList: messageList name: label 
	"Create and schedule a MessageSet browser on messageList."
	^ self   
		browseMessageList: messageList 
		name: label 
		autoSelect: nil! !

!SystemNavigation methodsFor: 'browse' stamp: 'cmm 10/16/2015 13:55'!
browseMessageList: messageListOrBlock name: labelString autoSelect: autoSelectString
	"Create and schedule a MessageSet browser on the message list. If messageListOrBlock is a block, then evaluate it to get the message list."

	| messageList title |
	messageList := messageListOrBlock isBlock
		ifTrue: [ Cursor wait showWhile: messageListOrBlock ]
		ifFalse: [ messageListOrBlock ].
	messageList size = 0 ifTrue: [
		^self inform: 'There are no', String cr, labelString ].
	title := messageList size > 1
		ifFalse: [ labelString ]
		ifTrue: [ labelString, ' [', messageList size printString, ']' ].
	^ ToolSet 
		browseMessageSet: messageList 
		name: title 
		autoSelect: autoSelectString! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 7/16/2011 18:49'!
browseMethodsWhoseNamesContain: aString
	"Launch a tool which shows all methods whose names contain the given string; case-insensitive.
·	1/16/1996 sw, at the dawn of Squeak: this was the classic implementation that provided the underpinning for the 'method names containing it' (cmd-shift-W) feature that has always been in Squeak -- the feature that later inspired the MethodFinder (aka SelectorBrowser).
·	sw 7/27/2001:	Switched to showing a MessageNames tool rather than a message-list browser, if in Morphic."

	| aList |
	Smalltalk isMorphic
		ifFalse:
			[aList := Symbol selectorsContaining: aString.
			aList size > 0 ifTrue:
				[self browseAllImplementorsOfList: aList title: 'Methods whose names contain ''', aString, '''']]

		ifTrue:
			[ToolSet browseMessageNames: aString]
	! !

!SystemNavigation methodsFor: 'browse' stamp: 'yo 7/31/2004 18:40'!
browseMethodsWithLiteral: aString
	"Launch a browser on all methods that contain string literals with aString as a substring. Make the search case-sensitive or insensitive as dictated by the caseSensitive boolean parameter"

	self browseAllSelect:
			[:method |
				method hasLiteralSuchThat: [:lit |
					(lit isString and: [lit isSymbol not]) and:
					[lit = aString]]]
		name:  'Methods with string ', aString printString
		autoSelect: aString.
! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/20/2003 14:11'!
browseMethodsWithSourceString: aString 
	"SystemNavigation new browseMethodsWithSourceString: 'SourceString'"
	"Launch a browser on all methods whose source code contains aString as 
	a substring."
	| caseSensitive suffix |
	suffix := (caseSensitive := Sensor shiftPressed)
				ifTrue: [' (case-sensitive)']
				ifFalse: [' (use shift for case-sensitive)'].
	^ self
		browseMessageList: (self allMethodsWithSourceString: aString matchCase: caseSensitive)
		name: 'Methods containing ' , aString printString , suffix
		autoSelect: aString! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/15/2003 22:28'!
browseMethodsWithString: aString
	"Launch a browser on all methods that contain string literals with aString as a substring. The search is case-insensitive, unless the shift key is pressed, in which case the search is case-sensitive."

	'string for testing'.
	^ self browseMethodsWithString: aString matchCase: Sensor shiftPressed

	"SystemNavigation new browseMethodsWithString: 'Testing' matchCase: false"
	"SystemNavigation new browseMethodsWithString: 'Testing' matchCase: true"! !

!SystemNavigation methodsFor: 'browse' stamp: 'eem 4/18/2016 16:35'!
browseMethodsWithString: aString matchCase: caseSensitive
	"Launch a browser on all methods that contain string literals with aString as a substring. Make the search case-sensitive or insensitive as dictated by the caseSensitive boolean parameter"

	self browseAllSelect:
			[:method |
			 method hasLiteralSuchThat:
				[:lit |
				 lit isString
				 and: [(lit includesSubstring: aString caseSensitive: caseSensitive)
				 and: [lit isSymbol
						ifTrue: [(method messages includes: lit) not]
						ifFalse: [true]]]]]
		name:  'Methods with string ', aString printString, (caseSensitive ifTrue: [' (case-sensitive)'] ifFalse: [' (case-insensitive)'])
		autoSelect: aString! !

!SystemNavigation methodsFor: 'browse' stamp: 'dtl 8/1/2010 10:12'!
browseMyChanges
	"Browse only the changes (in the changes file) by the current author."

	"SystemNavigation default browseMyChanges"

	self browseAllSelect: [ :method |
       method fileIndex > 1 "only look at changes file"
               and: [ method timeStamp beginsWith: Utilities authorInitials ]]! !

!SystemNavigation methodsFor: 'browse' stamp: 'ar 9/27/2005 20:34'!
browseObsoleteMethodReferences
	"Open a browser on all referenced behaviors that are obsolete"

	"SystemNavigation default browseObsoleteMethodReferences"

	| list |
	list := self obsoleteMethodReferences.
	self 
		browseMessageList: list
		name: 'Method referencing obsoletes'
		autoSelect: nil! !

!SystemNavigation methodsFor: 'browse' stamp: 'sd 4/15/2003 20:32'!
browseObsoleteReferences  
	"self new browseObsoleteReferences"

	| references |
	references := OrderedCollection new.
	(LookupKey allSubInstances select:
		[:x | ((x value isKindOf: Behavior) and: ['AnOb*' match: x value name]) or:
		['AnOb*' match: x value class name]]) 
		do: [:x | references addAll: (self allCallsOn: x)].
	self  
		browseMessageList: references 
		name: 'References to Obsolete Classes'! !

!SystemNavigation methodsFor: 'browse' stamp: 'ul 7/17/2011 22:33'!
browseUncommentedMethodsWithInitials: targetInitials
	"Browse uncommented methods whose initials (in the time-stamp, as logged to disk) match the given initials.  Present them in chronological order.  CAUTION: It will take several seconds for this to complete."
	"Time millisecondsToRun: [SystemNavigation default browseUncommentedMethodsWithInitials: 'jm']"

	self
		browseMessageList: [ self allUncommentedMethodsWithInitials: targetInitials ] 
		name: 'Uncommented methods with initials ', targetInitials
		autoSelect: nil! !

!SystemNavigation methodsFor: 'browse' stamp: 'cmm 5/2/2011 09:55'!
browseVariableAssignments: aClass 
	aClass chooseVarThenDo:
		[ : aVar | self
			browseAllStoresInto: aVar
			from: aClass ]! !

!SystemNavigation methodsFor: 'browse' stamp: 'cmm 5/2/2011 09:43'!
browseVariableReferences: aClass
	aClass chooseVarThenDo:
		[ : aVar | (aClass allInstVarNames includes: aVar)
			ifTrue:
				[ self
					browseAllAccessesTo: aVar
					from: aClass ]
			ifFalse: [ self browseAllCallsOn: aVar ] ]! !

!SystemNavigation methodsFor: 'ui' stamp: 'mt 7/21/2016 09:46'!
confirmRemovalOf: aSelector on: aClass 
	"Determine if it is okay to remove the given selector. Answer 1 if it  
	should be removed, 2 if it should be removed followed by a senders  
	browse, and 3 if it should not be removed."
	| count answer caption allCalls |
	allCalls := self allCallsOn: aSelector.
	(count := allCalls size) = 0
		ifTrue: [^ 1].
	"no senders -- let the removal happen without warning"
	count = 1
		ifTrue: [(allCalls first actualClass == aClass
					and: [allCalls first methodSymbol == aSelector])
				ifTrue: [^ 1]].
	"only sender is itself"
	caption := 'The message "{1}" has {2} sender{3}.' translated format: {aSelector. count. count > 1 ifTrue: ['s'] ifFalse: ['']}.

	answer := UIManager default 
		chooseFrom: #('Remove it'
				'Remove, then browse senders'
				'Don''t remove, but show me those senders'
				'Forget it -- do nothing -- sorry I asked') title: caption.
	answer = 3
		ifTrue: [self
				browseMessageList: allCalls
				name: 'Senders of ' , aSelector
				autoSelect: aSelector keywords first].
	answer = 0
		ifTrue: [answer := 3].
	"If user didn't answer, treat it as cancel"
	^ answer min: 3! !

!SystemNavigation methodsFor: 'private' stamp: 'fbs 5/9/2013 22:29'!
environment
	^ environment ifNil: [environment := Smalltalk globals].! !

!SystemNavigation methodsFor: 'private' stamp: 'nice 9/11/2013 21:16'!
headingAndAutoselectForLiteral: aLiteral do: binaryBlock 
	"Evaluate binaryBlock with either Users of ... or Senders of ... plus the auto-select string for the given literal.  aLiteral can be a Symbol, a VariableBinding or an arbitrary object."
	| autoSelect |
	^ aLiteral isSymbol
		ifTrue:
			[ binaryBlock
				value: 'Senders of ' , aLiteral
				value: (aLiteral keywords ifEmpty: [aLiteral] ifNotEmpty: [:keys | keys first]) ]
		ifFalse:
			[ autoSelect := aLiteral isVariableBinding
				ifTrue: [ aLiteral key ]
				ifFalse: [ aLiteral printString ].
			binaryBlock
				value: 'Users of ' , autoSelect
				value: autoSelect ]! !

!SystemNavigation methodsFor: 'query' stamp: 'nice 12/27/2009 03:11'!
hierarchyOfClassesSurrounding: aClass
	"Answer a list of classes in the hierarchy both above and below the given class"
	"SystemNavigation default hierarchyOfClassesSurrounding: StringHolder"
	
	| list aClassNonMeta isMeta |
	aClass ifNil: [^ OrderedCollection new].
	aClass ifNil: [^ self].
	aClassNonMeta := aClass theNonMetaClass.
	isMeta := aClassNonMeta ~~ aClass.
	list := OrderedCollection new.
	aClass allSuperclasses reverseDo:
		[:cl | list addLast: cl].
	aClassNonMeta allSubclassesWithLevelDo:
		[:cl :level | | theClassOrMeta |
		theClassOrMeta := isMeta ifTrue: [cl class] ifFalse: [cl].
		list addLast: theClassOrMeta]
	 	startingLevel: 0.
	^ list

! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 1/16/2004 21:03'!
hierarchyOfImplementorsOf: aSelector forClass: aClass
	"Answer a list of classes in the hierarchy both above and below the given class which implement the given selector."
	"SystemNavigation default hierarchyOfImplementorsOf: #contents forClass: StringHolder"

	^ (self hierarchyOfClassesSurrounding: aClass) select:
		[:cl | cl includesSelector: aSelector]
! !

!SystemNavigation methodsFor: 'initialize-release' stamp: 'fbs 5/9/2013 22:30'!
initializeWithEnvironment: anEnvironment
	self initialize.
	environment := anEnvironment.! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/18/2003 10:44'!
isThereAnImplementorOf: aSelector 
	"Answer true if there is at least one implementor of the selector found  
	in the system, false if there are no implementors"
	"self new isThereAnImplementorOf: #contents.  
	self new isThereAnImplementorOf: #nobodyImplementsThis."
	self 
		allBehaviorsDo: [:class | (class includesSelector: aSelector)
				ifTrue: [^ true]].
	^ false! !

!SystemNavigation methodsFor: 'browse' stamp: 'cmm 12/27/2015 14:58'!
methodHierarchyBrowserForClass: aClass selector: selectorSymbol
	"Create and schedule a message set browser on all implementors of the 
	currently selected message selector. Do nothing if no message is selected."
	"SystemNavigation default 
		methodHierarchyBrowserForClass: ParagraphEditor 
		selector: #isControlActive"
	| list aClassNonMeta isMeta tab compiledMethod window |
	aClass ifNil: [^ self].
	aClass isTrait ifTrue: [^ self].
	selectorSymbol ifNil: [^ self].
	aClassNonMeta := aClass theNonMetaClass.
	isMeta := aClassNonMeta ~~ aClass.
	list := OrderedCollection new.
	tab := ''.
	aClass allSuperclasses reverseDo:
		[:cl |
		(cl includesSelector: selectorSymbol) ifTrue:
			[list addLast: tab , cl name, ' ', selectorSymbol].
		tab := tab , '  '].
	aClassNonMeta allSubclassesWithLevelDo:
		[:cl :level | | theClassOrMeta stab |
		theClassOrMeta := isMeta ifTrue: [cl class] ifFalse: [cl].
		(theClassOrMeta includesSelector: selectorSymbol) ifTrue:
			[stab := ''.  1 to: level do: [:i | stab := stab , '  '].
			list addLast: tab , stab , theClassOrMeta name, ' ', selectorSymbol]]
	 	startingLevel: 0.
	window := self browseMessageList: list name: 'Inheritance of ' , selectorSymbol.
	window isSystemWindow ifTrue: 
		[ window model
			deselectAll ; 
			yourself.
		compiledMethod := aClass compiledMethodAt: selectorSymbol ifAbsent:[nil].
		compiledMethod ifNotNil: [ window model selectReference: compiledMethod methodReference ] ]! !

!SystemNavigation methodsFor: 'query' stamp: 'cwp 1/18/2014 10:27'!
methodsWithUnboundGlobals
	"Get all methods that use undeclared global objects that are not listed in Undeclared. For a clean image the result should be empty."

	"SystemNavigation new methodsWithUnboundGlobals"
	self flag: #environments.
	
	^self allSelect:
		[:m|
		m literals anySatisfy:
			[:l|
			l isVariableBinding
			and: [l key isSymbol "avoid class-side methodClass literals"
			and: [(m methodClass bindingOf: l key)
					ifNil: [(Undeclared associationAt: l key ifAbsent: []) ~~ l]
					ifNotNil: [:b| b ~~ l]]]]]! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/20/2003 14:27'!
numberOfImplementorsOf: aSelector 
	"Answer a count of the implementors of the given selector found in the  
	system"
	"self new numberOfImplementorsOf: #contents.  
	self new numberOfImplementorsOf: #nobodyImplementsThis.  
	self new numberOfimplementorsOf: #numberOfImplementorsOf:."
	| aCount |
	aCount := 0.
	self
		allBehaviorsDo: [:class | (class includesSelector: aSelector)
				ifTrue: [aCount := aCount + 1]].
	^ aCount! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 9/23/2004 22:03'!
obsoleteBehaviors
	"SystemNavigation default obsoleteBehaviors inspect"
	"Find all obsolete behaviors including meta classes"

	| obs |
	obs := OrderedCollection new.
	Smalltalk garbageCollect.
	self 
		allObjectsDo: [:cl | (cl isBehavior
					and: [cl isObsolete])
				ifTrue: [obs add: cl]].
	^ obs asArray! !

!SystemNavigation methodsFor: 'query' stamp: 'cmm 12/14/2013 16:08'!
obsoleteClasses   

	"SystemNavigation default obsoleteClasses inspect"
	"NOTE:  Also try inspecting comments below"
	| obs |
	obs := OrderedCollection new.  Smalltalk garbageCollect.
	Metaclass allInstancesDo:
		[:m | | c |
		c := m soleInstance.
		(c ~~ nil and: ['AnOb*' match: c name asString])
			ifTrue: [obs add: c]].
	^ obs asArray

"Likely in a ClassDict or Pool...
(Association allInstances select: [:a | (a value isKindOf: Class) and: ['AnOb*' match: a value name]]) asArray
"
"Obsolete class refs or super pointer in last lit of a method...
| n l found |
SystemNavigation default browseAllSelect:
	[:m | found := false.
	1 to: m numLiterals do:
		[:i | (((l := m literalAt: i) isMemberOf: Association)
				and: [(l value isKindOf: Behavior)
				and: ['AnOb*' match: l value name]])
			ifTrue: [found := true]].
	found]
"! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 6/17/2011 12:31'!
obsoleteMethodReferences
	"SystemNavigation default obsoleteMethodReferences"

	"Open a browser on all referenced behaviors that are obsolete"

	| obsClasses references |
	references := WriteStream on: Array new.
	obsClasses := self obsoleteBehaviors.
	'Scanning for methods referencing obsolete classes' 
		displayProgressFrom: 1
		to: obsClasses size
		during: 
			[:bar | 
			obsClasses keysAndValuesDo: 
					[:index :each | | obsRefs | 
					bar value: index.
					obsRefs := Utilities pointersTo: each except: obsClasses.
					obsRefs do: 
							[:ref | 
							"Figure out if it may be a global"

							(ref isVariableBinding and: [ref key isString	"or Symbol"]) 
								ifTrue: 
									[(Utilities pointersTo: ref) do: 
											[:meth | 
											(meth isKindOf: CompiledMethod) 
												ifTrue: [meth methodReference ifNotNil: [:mref | references nextPut: mref]]]]]]].
	^references contents! !

!SystemNavigation methodsFor: 'private' stamp: 'eem 8/30/2010 10:34'!
packageInfoFor: packageInfoOrString
	^packageInfoOrString isString
		ifTrue: [PackageInfo named: packageInfoOrString]
		ifFalse: [packageInfoOrString]! !

!SystemNavigation methodsFor: 'ui' stamp: 'sd 4/15/2003 15:34'!
showMenuOf: selectorCollection withFirstItem: firstItem ifChosenDo: choiceBlock
	"Show a sorted menu of the given selectors, preceded by firstItem, and all
	abbreviated to 40 characters.  Evaluate choiceBlock if a message is chosen."

	^ self showMenuOf: selectorCollection withFirstItem: firstItem ifChosenDo: choiceBlock withCaption: nil! !

!SystemNavigation methodsFor: 'ui' stamp: 'topa 9/27/2010 11:06'!
showMenuOf: selectorCollection withFirstItem: firstItem ifChosenDo: choiceBlock withCaption: aCaption
	"Show a sorted menu of the given selectors, preceded by firstItem, and all abbreviated to 40 characters.  Use aCaption as the menu title, if it is not nil.  Evaluate choiceBlock if a message is chosen."

	| index menuLabels sortedList |
	sortedList := selectorCollection asSortedCollection.
	menuLabels := Array streamContents: 
		[:strm | strm nextPut: (firstItem contractTo: 40).
		sortedList do: [:sel | strm nextPut: (sel contractTo: 40)]].
	index := UIManager default chooseFrom: menuLabels lines: #(1).
	index = 1 ifTrue: [choiceBlock value: firstItem].
	index > 1 ifTrue: [choiceBlock value: (sortedList at: index - 1)]! !

!SystemNavigation methodsFor: 'browse' stamp: 'fbs 3/16/2011 22:07'!
spawnHierarchyForClass: aClass selector: aSelector
	"Create and schedule a new class hierarchy browser on the requested class/selector."
	"SystemNavigation default spawnHierarchyForClass: SmallInteger selector: #hash"

	(aClass == nil)  ifTrue: [^ self].
	^ ToolSet browseHierarchy: aClass selector: aSelector
! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 15:20'!
unimplemented
	"Answer an collection of each message that is sent by an expression in a method but is not implemented by any object in the system."

	| implemented unimplemented |
	implemented := self allImplementedMessages.
	unimplemented := IdentityDictionary new.
	self allSelectorsAndMethodsDo: [ :behavior :selector :method |
		method messagesDo: [ :each |
			| entry |
			(implemented includes: each) ifFalse: [
				entry := unimplemented 
					at: each 
					ifPresent: [ :oldEntry |
						oldEntry copyWith: behavior name, '>', selector ]
					ifAbsent: [ 
						{ behavior name, '>', selector } ].
				unimplemented at: each put: entry ] ] ].
	^unimplemented! !

!SystemNavigation methodsFor: 'query' stamp: 'ul 7/17/2011 15:13'!
unusedBlocks
	"Answer all methods that contain a block that is not used (not
	 sent a message, returned, passed as an argument, or assigned)."
	"SystemNavigation new unusedBlocks"
	"SystemNavigation new
		browseMessageList: SystemNavigation new unusedBlocks
		name: 'unused blocks'"

	^self allMethodsSelect:
		[:m| | is |
		is := InstructionStream on: m.
		is scanFor: [:b| b = 143 and: [(m at: is thirdByte * 256 + is fourthByte + is pc + 4) = 135]]]! !

Object subclass: #SystemVersion
	instanceVariableNames: 'version date highestUpdate updates'
	classVariableNames: 'Current'
	poolDictionaries: ''
	category: 'System-Support'!

!SystemVersion commentStamp: 'tlk 11/14/2004 10:44' prior: 0!
I am responsible for maintaining what version of Squeak and the VM is running.  I also track all of the update items that have been included in the image.

I'm invoked at auto start to get the latest plugins, etc.!

!SystemVersion class methodsFor: 'updating' stamp: 'fbs 7/25/2013 07:20'!
check: pluginVersion andRequestPluginUpdate: updateURL
	"SystemVersion check: 'zzz' andRequestPluginUpdate: 'http://www.squeakland.org/installers/update.html' "

	"We don't have a decent versioning scheme yet, so we are basically checking for a nil VM version on the mac."
	(self pluginVersion: pluginVersion newerThan: self currentPluginVersion)
		ifFalse: [^true].
	(self confirm: 'There is a newer plugin version available. Do you want to install it now?')
		ifFalse: [^false].
	HTTPClient
		requestURL: updateURL , (Smalltalk platformName copyWithout: Character space) asLowercase , '.html'
		target: '_top'.
	^false! !

!SystemVersion class methodsFor: 'accessing' stamp: ''!
current
	Current ifNil: [Current := SystemVersion new].
	^Current! !

!SystemVersion class methodsFor: 'updating' stamp: 'fbs 7/25/2013 07:20'!
currentPluginVersion
	^Smalltalk vmVersion! !

!SystemVersion class methodsFor: 'instance creation' stamp: 'mir 3/29/2001 18:06'!
newVersion: versionName
	| newVersion |
	newVersion := self new version: versionName.
	newVersion
		highestUpdate: self current highestUpdate.
	Current := newVersion
! !

!SystemVersion class methodsFor: 'accessing' stamp: 'nice 12/27/2009 03:48'!
parseVersionString: versionString
	"Answer the version of this release as version, date, update."
	"SystemVersion parseVersionString: 'Squeak3.1alpha of 28 February 2001 [latest update: #3966]' "

	| version date update |

	[ | stream |
	stream := ReadStream on: versionString.
	version := stream upToAll: ' of '.
	date := Date readFrom: stream.
	stream upToAll: ' #'.
	update := Number readFrom: stream]
		on: Error
		do: [^nil].
	^{version. date. update.}! !

!SystemVersion class methodsFor: 'accessing' stamp: 'mir 8/10/2001 11:53'!
pluginVersion: availableVersionString newerThan: currentVersionString
	| currentVersion availableVersion |
	(currentVersionString isEmptyOrNil
		or: [availableVersionString isEmptyOrNil])
		ifTrue: [^true].
	currentVersion := self parseVersionString: currentVersionString.
	availableVersion := self parseVersionString: availableVersionString.
	(currentVersion isNil
		or: [availableVersion isNil])
		ifTrue: [^false].
	^(currentVersion at: 2) < (availableVersion at: 2)! !

!SystemVersion class methodsFor: 'class initialization' stamp: 'rbb 3/1/2005 11:17'!
setVersion
	"SystemVersion setVersion"

	| newName |
	newName := UIManager default
		request: ('Please name this system version.\The old version is:\',
					self current version, '\set on ', self current date asString) withCRs
 		initialAnswer: self current version.
	newName size > 0 ifTrue:
		[self newVersion: newName]! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
date
	^date! !

!SystemVersion methodsFor: 'accessing' stamp: 'mir 5/1/2001 18:19'!
datedVersion
	"Answer the version of this release."

	^ self version asString , ' of ' , self date printString! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
date: newDate
	date := newDate! !

!SystemVersion methodsFor: 'printing' stamp: 'mt 8/1/2016 15:15'!
description

	self isAlpha ifTrue: [^ 'ALPHA. New features which are not stable yet may come in\with each update. Also, existing features might not work reliably\due to updates and related changes.' translated withCRs].
	
	self isFeatureFreeze ifTrue: [^ 'FEATURE FREEZE. A new release is being prepared.\There will be only bugfixes, but no new features.' translated withCRs].
	
	self isCodeFreeze ifTrue: [^ 'RELEASE CANDIDATE. The new release is almost ready.\There will be only bugfixes, if any.' translated withCRs].
		
	self isRelease ifTrue: [^ ''].! !

!SystemVersion methodsFor: 'accessing' stamp: 'mir 3/29/2001 18:03'!
highestUpdate
	| sortedUpdates |
	highestUpdate ifNil: [
		sortedUpdates := self updates asSortedCollection.
		highestUpdate := (sortedUpdates isEmpty
			ifTrue: [0]
			ifFalse: [sortedUpdates last])].
	^highestUpdate! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
highestUpdate: anInteger
	highestUpdate := anInteger! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
includesUpdate: anUpdate
	^self updates includes: anUpdate! !

!SystemVersion methodsFor: 'initialize' stamp: ''!
initialize
	version := 'No version set'.
	date := Date today.
	updates := Set new.
! !

!SystemVersion methodsFor: 'testing' stamp: 'mt 6/22/2016 15:07'!
isAlpha
	^ (version endsWith: #alpha)! !

!SystemVersion methodsFor: 'testing' stamp: 'mt 6/22/2016 15:04'!
isCodeFreeze
	^ self isFeatureFreeze and: [(version endsWith: #beta) not]! !

!SystemVersion methodsFor: 'testing' stamp: 'mt 6/22/2016 15:07'!
isFeatureFreeze
	^ self isAlpha not! !

!SystemVersion methodsFor: 'testing' stamp: 'ul 8/14/2015 18:05'!
isPharo
	^ version includesSubstring: 'Pharo'! !

!SystemVersion methodsFor: 'testing' stamp: 'mt 6/22/2016 15:08'!
isRelease
	^ self isReleaseCandidate not and: [version last isDigit]! !

!SystemVersion methodsFor: 'testing' stamp: 'mt 6/22/2016 15:16'!
isReleaseCandidate
	^ (version findTokens: 'rc') size > 1! !

!SystemVersion methodsFor: 'testing' stamp: 'ul 8/14/2015 18:05'!
isSqueak
	^ version includesSubstring: 'Squeak'! !

!SystemVersion methodsFor: 'accessing' stamp: 'sw 7/28/2005 12:36'!
majorMinorVersion
	"Return the major/minor version number of the form X.Y, without any 'alpha' or 'beta' or other suffix."
	
	| char stream |
	^ (version includes: $.)
		ifTrue:
			[stream := ReadStream on: version, 'x'.
			stream upTo: $..
			char := stream next.
			[char isDigit]
				whileTrue: [char := stream next].
			version copyFrom: 1 to: stream position - 1]
		ifFalse:
			[version]

"
(SystemVersion new version: 'Squeak3.7alpha') majorMinorVersion
(SystemVersion new version: 'Testing') majorMinorVersion
SystemVersion current majorMinorVersion
"

! !

!SystemVersion methodsFor: 'accessing' stamp: 'mt 6/22/2016 17:56'!
majorVersionNumber

	^ (version copyFrom: 'Squeak' size to: (version indexOf: $.)) asInteger! !

!SystemVersion methodsFor: 'accessing' stamp: 'mt 6/22/2016 18:04'!
minorVersionNumber

	| pointIndex |
	pointIndex := version indexOf: $..
	^ (version
		copyFrom: pointIndex
		to: (self isRelease
			"e.g. Squeak5.1"
			ifTrue: [version size]
			"e.g. Squeak5.1rc5"
			ifFalse: [(pointIndex+1 to: version size) detect: [:ea | (version at: ea) isDigit not]])) asInteger! !

!SystemVersion methodsFor: 'printing' stamp: 'mir 5/1/2001 18:20'!
printOn: stream
	stream
		nextPutAll: self datedVersion;
		nextPutAll: ' update ' , self highestUpdate printString! !

!SystemVersion methodsFor: 'accessing' stamp: 'mir 3/29/2001 18:01'!
registerUpdate: update
	self updates add: update.
	self resetHighestUpdate! !

!SystemVersion methodsFor: 'accessing' stamp: 'mt 6/22/2016 15:18'!
releaseCandidateNumber
	^ self isReleaseCandidate
		ifFalse: [0]
		ifTrue: [(version findTokens: 'rc') last asInteger]! !

!SystemVersion methodsFor: 'accessing' stamp: 'mir 3/29/2001 18:01'!
resetHighestUpdate
	highestUpdate := nil! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
unregisterUpdate: update
	self updates remove: update ifAbsent: []! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
updates
	^updates! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
version
	^version! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
version: newVersion
	version := newVersion! !

Object subclass: #TextDiffBuilder
	instanceVariableNames: 'xLines yLines ignoreLineEndings'
	classVariableNames: 'IgnoreLineEndings InsertTextAttributes NormalTextAttributes RemoveTextAttributes'
	poolDictionaries: ''
	category: 'System-FilePackage'!

!TextDiffBuilder commentStamp: 'fbs 9/23/2013 08:58' prior: 0!
I implement the diff algorithm. I can show the differences between two texts. See my method comments for further information.

Instance Variables
	xLines:				<Array>
	yLines:				<Array>
	ignoreLineEndings:	<Boolean>

xLines
	- an Array of DiffElements which is created from the first input text

yLines
	- an Array of DiffElements which is created from the second input text
	
ignoreLineEndings
	- a Boolean describing whether lines only differing in the line endings should be reported as a difference, or not!

TextDiffBuilder subclass: #ClassDiffBuilder
	instanceVariableNames: ''
	classVariableNames: 'SeparatorSet'
	poolDictionaries: ''
	category: 'System-FilePackage'!

!ClassDiffBuilder commentStamp: 'ul 4/1/2015 23:10' prior: 0!
I'm like TextDiffBuilder, but I split the input text by my own set of separator characters (see #separatorSet), instead of new lines. I'm meant to create diffs of class definitions.!

!ClassDiffBuilder class methodsFor: 'as yet unclassified' stamp: 'ul 4/1/2015 23:08'!
cleanUp: aggressive

	SeparatorSet := nil! !

!ClassDiffBuilder class methodsFor: 'as yet unclassified' stamp: 'ul 4/1/2015 23:08'!
separatorSet

	^SeparatorSet ifNil: [
		SeparatorSet := CharacterSet separators copy
			add: $'; "for variables"
			add: $"; "for comments in mc"
			yourself ]! !

!ClassDiffBuilder methodsFor: 'private' stamp: 'ul 11/4/2010 22:33'!
print: aString withAttributes: attributes on: stream

	stream
		currentAttributes: attributes;
		nextPutAll: aString! !

!ClassDiffBuilder methodsFor: 'private' stamp: 'ul 4/1/2015 23:08'!
split: aString
	"I return an Array or DiffElements containing aString splitted by whitespace ' and ""."

	^Array streamContents: [ :stream |
		| input separators |
		input := aString readStream.
		separators := self class separatorSet.
		[ input atEnd ] whileFalse: [
			| word separator |
			word := input
				upToAnyOf: separators
				do: [ :matchingSeparator | separator := matchingSeparator ].
			stream nextPut: (DiffElement string: word).
			separator ifNotNil: [
				stream nextPut: (DiffElement string: separator asString) ] ] ]! !

TextDiffBuilder subclass: #PrettyTextDiffBuilder
	instanceVariableNames: 'sourceClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!PrettyTextDiffBuilder commentStamp: 'klub 12/28/2009 05:12' prior: 0!
I'm like TextDiffBuilder, but I use the pretty-printed version of the source code if available.

Instance Variables
	sourceClass:		<Class>

sourceClass
	- this class provides the pretty-printer
!

!PrettyTextDiffBuilder class methodsFor: 'instance creation' stamp: 'nk 10/29/2000 12:35'!
from: srcString to: dstString inClass: srcClass 
	^ (self new sourceClass: srcClass) from: srcString to: dstString
		! !

!PrettyTextDiffBuilder methodsFor: 'initialize' stamp: 'nk 10/29/2000 12:16'!
sourceClass: aClass
	sourceClass := aClass.! !

!PrettyTextDiffBuilder methodsFor: 'private' stamp: 'ul 12/27/2009 18:51'!
split: aString

	| formatted |
	aString ifEmpty: [ ^super split: aString ].
	formatted := [
		sourceClass prettyPrinterClass
			format: aString
			in: sourceClass
			notifying: nil
			decorated: false ] 
				on: Error
				do: [ :ex | aString ].
	^super split: formatted! !

!TextDiffBuilder class methodsFor: 'preferences' stamp: 'mt 7/9/2016 10:39'!
applyUserInterfaceTheme

	NormalTextAttributes := nil.
	InsertTextAttributes := nil.
	RemoveTextAttributes := nil.! !

!TextDiffBuilder class methodsFor: 'instance creation' stamp: 'klub 12/27/2009 02:59'!
buildDisplayPatchFrom: sourceText to: destinationText

	^(self from: sourceText to: destinationText) buildDisplayPatch! !

!TextDiffBuilder class methodsFor: 'instance creation' stamp: 'ul 12/27/2009 17:16'!
buildDisplayPatchFrom: sourceText to: destinationText inClass: sourceClass prettyDiffs: prettyDiffs
	
	^((sourceClass notNil and: [ prettyDiffs ])
		ifFalse: [ self from: sourceText to: destinationText ]
		ifTrue: [
			PrettyTextDiffBuilder
				from: sourceText
				to: destinationText
				inClass: sourceClass ]) buildDisplayPatch! !

!TextDiffBuilder class methodsFor: 'instance creation' stamp: 'ul 9/11/2013 01:02'!
from: sourceText to: destinationText

	^self from: sourceText to: destinationText ignoreLineEndings: self ignoreLineEndings! !

!TextDiffBuilder class methodsFor: 'instance creation' stamp: 'ul 9/11/2013 01:02'!
from: sourceText to: destinationText ignoreLineEndings: ignoreLineEndings

	^self new
		from: sourceText to: destinationText ignoreLineEndings: ignoreLineEndings;
		yourself! !

!TextDiffBuilder class methodsFor: 'preferences' stamp: 'nice 2/25/2010 04:07'!
ignoreLineEndings
	"Answer a boolean telling if line endings differences should be ignored or emphasized"
	
	<preference: 'ignoreLineEndings'
		category: 'TextDiff'
		description: 'When enabled, source code differences in line endings will be ignored.'
		type: #Boolean>
	^IgnoreLineEndings ifNil: [ false ]! !

!TextDiffBuilder class methodsFor: 'preferences' stamp: 'nice 2/25/2010 03:57'!
ignoreLineEndings: aBoolean
	"Set the preference telling if line endings differences should be ignored or emphasized"
	
	IgnoreLineEndings := aBoolean! !

!TextDiffBuilder class methodsFor: 'preferences' stamp: 'mt 7/9/2016 10:38'!
themeProperties

	^ {
		{ #insertTextAttributes. 'Styling'. 'Text attributes to use when a patch was inserted.' }.
		{ #normalTextAttributes. 'Styling'. 'Text attributes to use when text did not change.' }.
		{ #removeTextAttributes. 'Styling'. 'Text attributes to use when a patch was removed.' }.
	}! !

!TextDiffBuilder methodsFor: 'updating' stamp: 'mt 7/9/2016 10:46'!
applyUserInterfaceTheme
	"My class-side cache is invalidated there. My dependents have to invoke my services again."! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'mt 7/9/2016 10:44'!
buildDisplayPatch

	| stream |
	stream := AttributedTextStream new.
	
	"Lazy initialize the text attributes cache."
	NormalTextAttributes ifNil: [NormalTextAttributes := self userInterfaceTheme normalTextAttributes
		ifNil: [{TextEmphasis normal}]].
	InsertTextAttributes ifNil: [InsertTextAttributes := self userInterfaceTheme insertTextAttributes
		ifNil: [{TextColor red}]].
	RemoveTextAttributes ifNil: [RemoveTextAttributes := self userInterfaceTheme removeTextAttributes
		ifNil: [{TextEmphasis struckOut. TextColor blue}]].
	
	self 
		patchSequenceDoIfMatch: [ :string |
			self print: string withAttributes: NormalTextAttributes on: stream ]
		ifInsert: [ :string |
			self print: string withAttributes: InsertTextAttributes on: stream ]
		ifRemove: [ :string |
			self print: string withAttributes: RemoveTextAttributes on: stream ].
	^stream contents! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'ul 1/11/2010 01:05'!
buildPatchSequence
	"This method is only implemented for backwards compatibility and testing."

	^Array streamContents: [ :stream |
		self 
			patchSequenceDoIfMatch: [ :string |
				stream nextPut: #match -> (string copyWithout: Character cr) ]
			ifInsert: [ :string | 
				stream nextPut: #insert -> (string copyWithout: Character cr) ]
			ifRemove: [ :string | 
				stream nextPut: #remove -> (string copyWithout: Character cr) ] ]! !

!TextDiffBuilder methodsFor: 'updating' stamp: 'mt 8/16/2016 14:55'!
canApplyUserInterfaceTheme

	^ false! !

!TextDiffBuilder methodsFor: 'private' stamp: 'ul 2/8/2011 03:04'!
findMatches
	"I find the matching pairs of xLines and yLines. First I filter out all lines that can't have a pair, then I find the longest common subsequence of the remaining elements. Finally I mark the matching pairs."

	| temp lcs xFilteredLines yFilteredLines xNumbers yNumbers |
	"Filter out all lines that can't have a pair."
	temp := yLines asSet.
	xFilteredLines := xLines select: [ :each |
		temp includes: each ].
	xFilteredLines size = 0 ifTrue: [ ^self ].
	temp := xLines asSet.
	yFilteredLines := yLines select: [ :each |
		temp includes: each ].
	yFilteredLines size = 0 ifTrue: [ ^self ].
	"Map all lines to SmallIntegers, because they can be compared faster."
	temp := Dictionary new.
	xNumbers := xFilteredLines collect: [ :each |
		temp at: each ifAbsentPut: [ temp size ] ].
	yNumbers := yFilteredLines collect: [ :each |
		temp at: each ifAbsentPut: [ temp size ] ].
	temp := nil.
	"Find the longest common subsequence."
	lcs := self lcsFor: xNumbers and: yNumbers.
	"Mark the matching pairs."
	[ lcs == nil ] whileFalse: [
		(xFilteredLines at: (lcs at: 1)) matches: (yFilteredLines at: (lcs at: 2)).
		lcs := lcs at: 3 ]! !

!TextDiffBuilder methodsFor: 'initialize' stamp: 'ul 9/11/2013 00:56'!
from: xString to: yString

	self from: xString to: yString ignoreLineEndings: self class ignoreLineEndings! !

!TextDiffBuilder methodsFor: 'initialize' stamp: 'ul 9/11/2013 00:56'!
from: xString to: yString ignoreLineEndings: aBoolean

	ignoreLineEndings := aBoolean.
	xLines := self split: xString asString.
	yLines := self split: yString asString.
	self findMatches! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'ul 9/11/2013 00:55'!
ignoreLineEndings: aBoolean

	ignoreLineEndings := aBoolean! !

!TextDiffBuilder methodsFor: 'private' stamp: 'fbs 9/23/2013 08:59'!
lcsFor: xFilteredLines and: yFilteredLines
	"I find one of the longest common subsequences of my arguments. I assume that none of my arguments are empty. I return nil, or an Array which represents a list. The first two elements are the matching ''line'' numbers, the third (and last) is the next node in the list, or nil, if there are no more elements. The list contains the longest common subsequence. I'm a modified version of the Greedy LCS/SES Algorithm from the 6th page of 'An O(ND) Difference Algorithm and Its Variations (1986)' by Eugene W. Myers."

	| n m v lcss max index lcs x y |
	n := xFilteredLines size.
	m := yFilteredLines size.
	max := m + n.
	v := Array new: 2 * max + 1.
	lcss := Array new: 2 * max + 1.
	"Unrolled first iteration (d = 0, k = 0)"
	index := max + 2.
	y := x := v at: index put: 0.	
	lcs := lcss at: index.
	[ x < n and: [ y < m and: [ (xFilteredLines at: x + 1) = (yFilteredLines at: y + 1) ] ] ]
		whileTrue: [ lcs := { x := x + 1. y := y + 1. lcs } ].
	x >= n ifTrue: [ y >= m ifTrue: [ ^lcs ] ].
	v at: max + 1 put: x.
	lcss at: max + 1 put: lcs.
	1 to: max do: [ :d |
		"Unrolled lowest diagonal checks (k = -d)."
		index := max - d + 2.
		x := v at: index.
		y := x + d.
		lcs := lcss at: index.
		[ x < n and: [ y < m and: [ (xFilteredLines at: x + 1) = (yFilteredLines at: y + 1) ] ] ]
			whileTrue: [ lcs := { x := x + 1. y := y + 1. lcs } ].
		x >= n ifTrue: [ y >= m ifTrue: [ ^lcs ] ].
		v at: max - d + 1 put: x.
		lcss at: max - d + 1 put: lcs.
		"Inner diagonals. (k in [2-d..d-2])"
		2 - d to: d - 2 by: 2 do: [ :k |
			index := max + k.
			(v at: index) < (v at: index + 2)
				ifTrue: [ x := v at: (index := index + 2) ]
				ifFalse: [ x := (v at: index) + 1 ].
			y := x - k.
			lcs := lcss at: index.
			[ x < n and: [ y < m and: [ (xFilteredLines at: x + 1) = (yFilteredLines at: y + 1) ] ] ]
				whileTrue: [ lcs := { x := x + 1. y := y + 1. lcs } ].
			x >= n ifTrue: [ y >= m ifTrue: [ ^lcs ] ].
			v at: max + k + 1 put: x.
			lcss at: max + k + 1 put: lcs ].
		"Unrolled highest diagonal checks (k = d)."
		index := max + d.
		x := (v at: index) + 1.
		y := x - d.
		lcs := lcss at: index.
		[ x < n and: [ y < m and: [ (xFilteredLines at: x + 1) = (yFilteredLines at: y + 1) ] ] ]
			whileTrue: [ lcs := { x := x + 1. y := y + 1. lcs } ].
		x >= n ifTrue: [ y >= m ifTrue: [ ^lcs ] ].
		v at: max + d + 1 put: x.
		lcss at: max + d + 1 put: lcs ].
	self error "We should never reach this point."! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'klub 12/29/2009 15:44'!
patchSequenceDoIfMatch: matchBlock ifInsert: insertBlock ifRemove: removeBlock
	"I'm the general purpose method to iterate through the patch sequence. See my senders to learn how to use me."

	| xLine xLineStream |
	xLineStream := xLines readStream.
	yLines do: [ :yLine | 
		yLine hasMatch 
			ifFalse: [ insertBlock value: yLine string  ]
			ifTrue: [
				[ (xLine := xLineStream next) == nil or: [ xLine == yLine match  ] ]
					whileFalse: [ removeBlock value: xLine string ].
				matchBlock value: yLine string ] ].
	[ (xLine := xLineStream next) == nil ] whileFalse: [
		removeBlock value: xLine string ]! !

!TextDiffBuilder methodsFor: 'private' stamp: 'ul 11/4/2010 22:34'!
print: aString withAttributes: attributes on: stream

	stream
		currentAttributes: attributes;
		nextPutAll: aString.
	(aString notEmpty and: [ 
		aString last = Character cr or: [
			aString endsWith: String crlf ] ]) 
				ifFalse: [ stream cr ]! !

!TextDiffBuilder methodsFor: 'private' stamp: 'ul 9/11/2013 00:58'!
split: aString
	"I return an Array of DiffElements containing the strings which are the lines extracted from aString. All lines contain the line separator characters, or not depending on preference."

	^Array streamContents: [ :stream |
		ignoreLineEndings
			ifTrue: [
				aString lineIndicesDo: [ :start :endWithoutSeparators :end |
					stream nextPut: (DiffElement string: (aString copyFrom: start to: endWithoutSeparators)) ] ]
			ifFalse: [
				aString lineIndicesDo: [ :start :endWithoutSeparators :end |
					stream nextPut: (DiffElement string: (aString copyFrom: start to: end)) ] ] ]! !

Object subclass: #TextDomainManager
	instanceVariableNames: ''
	classVariableNames: 'ClassCategories Classes DefaultDomain DomainInfos LoneClasses Packages'
	poolDictionaries: ''
	category: 'System-Localization'!

TextDomainManager class
	instanceVariableNames: 'defaultDomain'!

!TextDomainManager commentStamp: 'fbs 5/12/2013 13:04' prior: 0!
I manage mapping from class category to textdomain.

Class variables:
 ClassCategories	IdentityDictionary -- classCategory -> domainName 
 Classes			IdentityDictionary -- class name (a Symbol) -> domainName   (a cache only!!)
 DefaultDomain	String -- the default domain name
 DomainInfos		Dictionary -- domainName -> a TextDomainInfo
 LoneClasses		IdentityDictionary -- class name (a Symbol) -> domainName.  For classes whose entire category are not all in the same domain (BookMorph and QuickGuideMorph)

TextDomainManager registerCategoryPrefix: 'DrGeoII' domain: 'DrGeoII'.
TextDomainManager unregisterDomain: 'DrGeoII'.

TextDomainManager registerClass: #QuickGuideMorph domain: 'quickguides'.
TextDomainManager registerClass: #QuickGuideHolderMorph  domain: 'quickguides'.
!

TextDomainManager class
	instanceVariableNames: 'defaultDomain'!

!TextDomainManager class methodsFor: 'accessing' stamp: 'fbs 5/12/2013 13:04'!
allKnownDomains
	"Every package has its own text domain now so it's not necessary to keep a registry of all domains, we can simply return all the packages in the image.
	PROBLEM: If a package doesn't contain translations, it won't have a mo file but the GetTextTranslator will try to load it anyway. This happens when we switch languages. So far I tested it briefly and it seems to work..."
	^PackageOrganizer default packageNames , {'Etoys-Tiles'}! !

!TextDomainManager class methodsFor: 'private' stamp: 'fbs 5/7/2013 15:01'!
cleanUp: aggressive
	aggressive ifTrue: [self clearAllDomains].! !

!TextDomainManager class methodsFor: 'private' stamp: 'eem 2/6/2013 12:52'!
clearAllDomains
	"TextDomainManager clearAllDomains"
	self systemNavigation allBehaviorsDo:
		[:b|
		b selectorsAndMethodsDo:
			[:s :m |
			m removeProperty: self textDomainProperty ifAbsent: []]]! !

!TextDomainManager class methodsFor: 'accessing' stamp: 'fbs 5/12/2013 13:05'!
defaultDomain
	"I'm not sure we still need a default domain. AFAIK the default domain will only be used when no domain is found. In that case, wouldn't it be better to just look for a translation in all domains?"
	^defaultDomain! !

!TextDomainManager class methodsFor: 'accessing' stamp: 'Richo 5/13/2010 00:02'!
defaultDomain: aDomainName
	defaultDomain := aDomainName! !

!TextDomainManager class methodsFor: 'accessing' stamp: 'Richo 5/18/2010 11:13'!
domainForClass: aClass
	^'etoys'! !

!TextDomainManager class methodsFor: 'accessing' stamp: 'Richo 5/13/2010 00:10'!
domainForPackage: aPackageInfo
	"Package names and text domains are synonyms now"
	^aPackageInfo name! !

!TextDomainManager class methodsFor: 'accessing' stamp: 'Richo 5/13/2010 00:17'!
domainOfMethod: aCompiledMethod 
	^ aCompiledMethod
		propertyValueAt: self textDomainProperty
		ifAbsent: [self updateDomainOfMethod: aCompiledMethod] ! !

!TextDomainManager class methodsFor: 'class initialization' stamp: 'edc 3/13/2012 10:16'!
initialize
	"	TextDomainManager initialize	"
	self defaultDomain: 'Etoys'! !

!TextDomainManager class methodsFor: 'private' stamp: 'Richo 5/13/2010 00:12'!
textDomainProperty
	^#textDomain! !

!TextDomainManager class methodsFor: 'private' stamp: 'Richo 5/13/2010 00:46'!
updateDomainOfAllMethodsWithTranslations
	self allMethodsWithTranslations do: [:each |
		self updateDomainOfMethod: each]! !

!TextDomainManager class methodsFor: 'private' stamp: 'Richo 5/13/2010 00:27'!
updateDomainOfMethod: aCompiledMethod 
	"First it looks for the package of the method reference (using
	the PackageOrganizer: deadly slow). If the method doesn't
	belong to any package it uses the default domain. Finally it
	stores the text domain of the method using a method
	property, this way we gain performance the next time we
	translate the same method because we avoid the use of
	PackageOrganizer. Have I mentioned it is really slow? :)"
	| package |
	package := PackageOrganizer default
				packageOfMethod: aCompiledMethod methodReference
				ifNone: [].
	^ aCompiledMethod
		propertyValueAt: self textDomainProperty
		put: (package isNil
				ifTrue: [TextDomainManager defaultDomain]
				ifFalse: [package name])! !

Object subclass: #ThirtyTwoBitRegister
	instanceVariableNames: 'hi low'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Digital Signatures'!

!ThirtyTwoBitRegister commentStamp: 'ul 4/13/2015 05:24' prior: 0!
I represent a 32-bit register. An instance of me can hold any non-negative integer in the range [0..(2^32 - 1)]. Operations are performed on my contents in place, like a hardware register, and results are always modulo 2^32. All operations avoid LargeInteger arithmetic as much as possible.

I'm mainly used by the SecureHashAlgorithm class, but I can be used for implementing other algorithms designed for 32-bit arithmetic. For examble George Marsaglia's Xorshift PRNG from http://www.jstatsoft.org/v08/i14/paper :

"Internal state."
x := ThirtyTwoBitRegister fromInteger: 123456789.
y := ThirtyTwoBitRegister fromInteger: 362436069.
z := ThirtyTwoBitRegister fromInteger: 521288629.
w := ThirtyTwoBitRegister fromInteger: 88675123.
"Temporaries."
t := ThirtyTwoBitRegister new.
temp := nil.
"The algorithm: t=(x^(x<<11));x=y;y=z;z=w; return( w=(w^(w>>19))ˆ(t^(t>>8)) );"
xorShift128 := [
	"t=(x^(x<<11));"
	t
		loadFrom: x;
		<< 11;
		bitXor: x.
	"x=y;y=z;z=w;"
	temp := x.
	x := y.
	y := z.
	z := w.
	"w=(w^(w>>19))ˆ(t^(t>>8))"
	w := temp.
	w
		loadFrom: z;
		>> 19;
		bitXor: z;
		bitXor: t.
	t >> 8.
	w bitXor: t.
	"Truncate to 30 bits."
	(w hi bitShift: 14) bitXor: w low ].

The code is about 9.5x faster than an implementation using LargeIntegers:

"Implementation using LargeIntegers."
x := 123456789.
y := 362436069.
z := 521288629.
w := 88675123.
largeIntegerXorShift128 := [
	| t |
	t := ((x bitAnd: 16r1FFFFF) bitShift: 11) bitXor: x.
	x := y.
	y := z.
	z := w.
	w := (((w bitShift: -19) bitXor: w) bitXor: t) bitXor: (t bitShift: -8) ].

xorShift128 bench. '4,990,000 per second. 201 nanoseconds per run.'.
largeIntegerXorShift128 bench. '529,000 per second. 1.89 microseconds per run.'.!

!ThirtyTwoBitRegister class methodsFor: 'instance creation' stamp: 'nice 8/28/2010 22:42'!
fromByteArray: aByteArray at: startIndex
	"Answer a new instance whose initial contents is copied from next four bytes from aByteArray starting at startIndex..
	Convention is Most Significant Byte first (aka big endian)."

	^ self basicNew loadFrom: aByteArray at: startIndex
! !

!ThirtyTwoBitRegister class methodsFor: 'instance creation' stamp: 'nice 8/28/2010 22:38'!
fromInteger: aPositiveInteger
	"Answer a new instance whose initial contents is copied from aPositiveInteger.
	It is required that aPositiveInteger has no more than 32 bits."

	^ self basicNew load: aPositiveInteger
! !

!ThirtyTwoBitRegister class methodsFor: 'instance creation' stamp: 'jm 12/14/1999 16:05'!
new
	"Answer a new instance whose initial contents is zero."

	^ super new load: 0
! !

!ThirtyTwoBitRegister methodsFor: 'converting' stamp: 'jm 12/14/1999 16:03'!
asInteger
	"Answer the integer value of my current contents."

	^ (hi bitShift: 16) + low
! !

!ThirtyTwoBitRegister methodsFor: 'converting' stamp: 'ul 4/10/2015 20:52'!
asSignedInteger
	"Answer the signed integer value of my current contents."

	hi >= 16r8000 ifFalse: [ ^(hi bitShift: 16) + low ].
	^-1 - (low bitXor: 16rFFFF) - ((hi bitXor: 16rFFFF) bitShift: 16)


! !

!ThirtyTwoBitRegister methodsFor: 'accumulator ops' stamp: 'jm 12/7/1999 15:41'!
bitAnd: aThirtTwoBitRegister
	"Replace my contents with the bitwise AND of the given register and my current contents."

	hi := hi bitAnd: aThirtTwoBitRegister hi.
	low := low bitAnd: aThirtTwoBitRegister low.
! !

!ThirtyTwoBitRegister methodsFor: 'accumulator ops' stamp: 'jm 12/7/1999 15:40'!
bitInvert
	"Replace my contents with the bitwise inverse my current contents."

	hi := hi bitXor: 16rFFFF.
	low := low bitXor: 16rFFFF.
! !

!ThirtyTwoBitRegister methodsFor: 'accumulator ops' stamp: 'jm 12/7/1999 15:40'!
bitOr: aThirtTwoBitRegister
	"Replace my contents with the bitwise OR of the given register and my current contents."

	hi := hi bitOr: aThirtTwoBitRegister hi.
	low := low bitOr: aThirtTwoBitRegister low.
! !

!ThirtyTwoBitRegister methodsFor: 'accumulator ops' stamp: 'jm 12/7/1999 15:38'!
bitXor: aThirtTwoBitRegister
	"Replace my contents with the bitwise exclusive OR of the given register and my current contents."

	hi := hi bitXor: aThirtTwoBitRegister hi.
	low := low bitXor: aThirtTwoBitRegister low.
! !

!ThirtyTwoBitRegister methodsFor: 'copying' stamp: 'jm 12/7/1999 15:26'!
copy
	"Use the clone primitive for speed."

	<primitive: 148>
	^ super copy
! !

!ThirtyTwoBitRegister methodsFor: 'comparing' stamp: 'ul 4/6/2015 03:06'!
hash

	^((hi bitShift: 14) bitXor: low) hashMultiply! !

!ThirtyTwoBitRegister methodsFor: 'accessing' stamp: 'jm 12/7/1999 15:26'!
hi

	^ hi
! !

!ThirtyTwoBitRegister methodsFor: 'accumulator ops' stamp: 'ul 3/6/2015 02:50'!
leftRotateBy: bits
	"Rotate my contents left by the given number of bits, retaining exactly 32 bits."
	"Details: Perform this operation with no LargeInteger arithmetic."

	| bitCount newHi |
	bitCount := bits.
	bitCount >= 32 ifTrue: [ bitCount := bitCount \\ 32 ].
	bitCount >= 16 ifTrue: [
		newHi := low.
		low := hi.
		hi := newHi.
		bitCount := bitCount - 16 ].
	bitCount >= 15 ifTrue: [
		newHi := ((hi bitAnd: 16r1) bitShift: 15) bitOr: (low bitShift: -1).
		low := ((low bitAnd: 16r1) bitShift: 15) bitOr: (hi bitShift: -1).
		hi := newHi.
		^self ].
	bitCount >= 1 ifTrue: [
		| shift |
		shift := bitCount - 16.
		newHi := ((hi bitShift: bitCount) bitAnd: 16rFFFF) bitOr: (low bitShift: shift).
		low := ((low bitShift: bitCount) bitAnd: 16rFFFF) bitOr: (hi bitShift: shift).
		hi := newHi ]! !

!ThirtyTwoBitRegister methodsFor: 'accessing' stamp: 'ul 4/11/2015 04:01'!
loadFrom: aThirtyTwoBitRegister
	"Set my contents from the given ThirtyTwoBitRegister."

	hi := aThirtyTwoBitRegister hi.
	low := aThirtyTwoBitRegister low
! !

!ThirtyTwoBitRegister methodsFor: 'accessing' stamp: 'jm 12/14/1999 16:07'!
loadFrom: aByteArray at: index
	"Load my 32-bit value from the four bytes of the given ByteArray starting at the given index. Consider the first byte to contain the most significant bits of the word (i.e., use big-endian byte ordering)."

	hi := ((aByteArray at: index) bitShift: 8) + ( aByteArray at: index + 1).
	low := ((aByteArray at: index + 2) bitShift: 8) + ( aByteArray at: index + 3).
! !

!ThirtyTwoBitRegister methodsFor: 'accessing' stamp: 'nice 8/28/2010 22:34'!
load: anInteger
	"Set my contents to the value of given integer."

	(anInteger positive and: [anInteger digitLength <= 4])
		ifFalse: [self error: 'out of range: ', anInteger printString].
	low := anInteger bitAnd: 16rFFFF.
	hi := (anInteger bitShift: -16) bitAnd: 16rFFFF
! !

!ThirtyTwoBitRegister methodsFor: 'accessing' stamp: 'jm 12/7/1999 15:26'!
low

	^ low! !

!ThirtyTwoBitRegister methodsFor: 'printing' stamp: 'laza 3/29/2004 12:22'!
printOn: aStream
	"Print my contents in hex with a leading 'R' to show that it is a register object being printed."

	aStream nextPutAll: 'R:'.
	self asInteger storeOn: aStream base: 16.
! !

!ThirtyTwoBitRegister methodsFor: 'accumulator ops' stamp: 'ul 4/11/2015 02:24'!
*= aThirtTwoBitRegister
	"Replace my contents with the product of the given register and my current contents."

	| otherLow otherHi mul newLow newHi |
	otherLow := aThirtTwoBitRegister low.
	otherHi := aThirtTwoBitRegister hi.
	"Multiply low with otherLow. Process the two highest bits of low separately if necessary to avoid LargeInteger operations."
	(low bitShift: -8) * (otherLow bitShift: -8) > 16r3FFF
		ifTrue: [ 
			mul := (low bitAnd: 16r3FFF) * otherLow. "Without the two most significant bits of low."
			newLow := (mul bitAnd: 16rFFFF).
			newHi := (mul bitShift: -16).
			mul := (low bitShift: -14) * otherLow. "The two most significant bits of low"
			newLow := newLow + ((mul bitAnd: 16r3) bitShift: 14).
			newHi := newHi + (mul bitShift: -2) + (newLow bitShift: -16) "Carry from newLow" ]
		ifFalse: [
			newLow := low * otherLow. "We'll trim newLow at the end of the method."
			newHi := newLow bitShift: -16 ].
	"Multiply hi with otherLow."
	(hi bitShift: -8) * (otherLow bitShift: -8) > 16r3FFF
		ifTrue: [
			newHi := newHi + 
				((hi bitAnd: 16r3FFF) * otherLow bitAnd: 16rFFFF) +
				(((hi bitShift: -14) * otherLow bitAnd: 16r3) bitShift: 14) ]
		ifFalse: [ newHi := newHi + (hi * otherLow bitAnd: 16rFFFF) ].
	"Multiply low with otherHi."
	(low bitShift: -8) * (otherHi bitShift: -8) > 16r3FFF
		ifTrue: [
			newHi := newHi + 
				((low bitAnd: 16r3FFF) * otherHi bitAnd: 16rFFFF) +
				(((low bitShift: -14) * otherHi bitAnd: 16r3) bitShift: 14) ]
		ifFalse: [ newHi := newHi + (low * otherHi bitAnd: 16rFFFF) ].
	"Truncate and store the results."
	hi := newHi bitAnd: 16rFFFF.
	low := newLow bitAnd: 16rFFFF 
! !

!ThirtyTwoBitRegister methodsFor: 'accumulator ops' stamp: 'jm 12/7/1999 15:36'!
+= aThirtTwoBitRegister
	"Replace my contents with the sum of the given register and my current contents."

	| lowSum |
	lowSum := low + aThirtTwoBitRegister low.
	hi := (hi + aThirtTwoBitRegister hi + (lowSum bitShift: -16)) bitAnd: 16rFFFF.
	low := lowSum bitAnd: 16rFFFF.
! !

!ThirtyTwoBitRegister methodsFor: 'comparing' stamp: 'ul 4/11/2015 04:50'!
< aThirtyTwoBitRegister

	^hi < aThirtyTwoBitRegister hi or: [
		hi = aThirtyTwoBitRegister hi and: [
			low < aThirtyTwoBitRegister low ] ]! !

!ThirtyTwoBitRegister methodsFor: 'accumulator ops' stamp: 'ul 4/11/2015 04:39'!
<< anInteger
	"Unsigned left shift."

	| bitCount |
	bitCount := anInteger.
	bitCount >= 32 ifTrue: [
		hi := low := 0.
		^self ].
	bitCount >= 16 ifTrue: [
		hi := low.
		low := 0.
		bitCount := bitCount - 16 ].
	bitCount >= 15 ifTrue: [
		hi := ((hi bitAnd: 1) bitShift: 15) bitOr: (low bitShift: -1).
		low := (low bitAnd: 1) bitShift: 15.
		^self ].
	bitCount >= 1 ifTrue: [
		hi := ((hi bitShift: bitCount) bitAnd: 16rFFFF) bitOr: (low bitShift: bitCount - 16).
		low := (low bitShift: bitCount) bitAnd: 16rFFFF ]! !

!ThirtyTwoBitRegister methodsFor: 'comparing' stamp: 'ul 4/11/2015 04:51'!
<= aThirtyTwoBitRegister

	^hi < aThirtyTwoBitRegister hi or: [
		hi = aThirtyTwoBitRegister hi and: [
			low <= aThirtyTwoBitRegister low ] ]! !

!ThirtyTwoBitRegister methodsFor: 'comparing' stamp: 'ul 4/6/2015 02:30'!
= anObject

	^self class == anObject class
		and: [ anObject low = low
		and: [ anObject hi = hi ] ]! !

!ThirtyTwoBitRegister methodsFor: 'comparing' stamp: 'ul 4/11/2015 04:51'!
> aThirtyTwoBitRegister

	^hi > aThirtyTwoBitRegister hi or: [
		hi = aThirtyTwoBitRegister hi and: [
			low > aThirtyTwoBitRegister low ] ]! !

!ThirtyTwoBitRegister methodsFor: 'comparing' stamp: 'ul 4/11/2015 04:51'!
>= aThirtyTwoBitRegister

	^hi > aThirtyTwoBitRegister hi or: [
		hi = aThirtyTwoBitRegister hi and: [
			low >= aThirtyTwoBitRegister low ] ]! !

!ThirtyTwoBitRegister methodsFor: 'accumulator ops' stamp: 'ul 4/10/2015 20:54'!
>> anInteger
	"Unsigned right shift."

	| bitCount shift |
	bitCount := anInteger.
	bitCount >= 32 ifTrue: [
		hi := low := 0.
		^self ].
	bitCount >= 16 ifTrue: [
		low := hi.
		hi := 0.
		bitCount := bitCount - 16 ].
	bitCount >= 2 ifTrue: [
		shift := 0 - bitCount.
		low := (low bitShift: shift) bitOr: ((hi bitShift: shift + 16) bitAnd: 16rFFFF).
		hi := hi bitShift: shift.
		^self ].
	bitCount >= 1 ifTrue: [
		low := (low bitShift: -1) bitOr: ((hi bitAnd: 16r1) bitShift: 15).
		hi := hi bitShift: -1 ]! !

Object subclass: #TranslatedReceiverFinder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Localization'!

!TranslatedReceiverFinder class methodsFor: 'utilities' stamp: 'tak 11/8/2006 19:45'!
browseNonLiteralReceivers
	"TranslatedReceiverFinder browseNonLiteralReceivers"
	SystemNavigation default
		browseMessageList: self new nonLiteralReceivers  asSortedCollection
		name: 'Non literal receivers of #translated'
		autoSelect: 'translated'! !

!TranslatedReceiverFinder class methodsFor: 'as yet unclassified' stamp: 'dgd 4/3/2006 13:24'!
makeJapaneseTranslationFile
	| t n |
	NaturalLanguageTranslator initializeKnownPhrases.
	t := TranslatedReceiverFinder new senders.
	n := NaturalLanguageTranslator
				localeID: (LocaleID isoLanguage: 'ja').
	t
		do: [:w | 
			NaturalLanguageTranslator registerPhrase: w.
			self
				at: w
				ifPresent: [:k | n phrase: w translation: k]].
	n saveToFileNamed: 'ja.translation'! !

!TranslatedReceiverFinder methodsFor: 'private' stamp: 'tak 7/30/2007 15:22'!
arraySearch: aSymbol fromArray: anArray addTo: aCollection 
	"Find literals ahead of aSymbol from arrays in the method."
	"BUG: it can handle just one occurrence"
	"self new arraySearch: #hello fromArray: #(ignore (ignore detected
	hello ignore)) addTo: Set new"
	| index |
	(index := anArray identityIndexOf: aSymbol) > 1
		ifTrue: [aCollection add: (anArray at: index - 1) asString].
	(anArray
		select: [:each | each isMemberOf: Array])
		do: [:each | self
				arraySearch: aSymbol
				fromArray: each
				addTo: aCollection].
	^ aCollection! !

!TranslatedReceiverFinder methodsFor: 'private' stamp: 'tak 7/30/2007 15:27'!
arraySearch: aSymbol messageNode: aParseNode addTo: aCollection 
	"Find literals ahead of aSymbol from arrays in the method."
	"self new arraySearch: #hello messageNode: (self
	decompile: #arraySearch:messageNode:addTo:) addTo: Set new"
	self flag: #(#ignore #detected #hello ).
	((aParseNode isMemberOf: LiteralNode)
			and: [aParseNode key isMemberOf: Array])
		ifTrue: [self
				arraySearch: aSymbol
				fromArray: aParseNode key
				addTo: aCollection].
	(aParseNode notNil
			and: [aParseNode isLeaf not])
		ifTrue: [aParseNode getAllChildren
				do: [:child | self
						arraySearch: aSymbol
						messageNode: child
						addTo: aCollection]].
	^ aCollection! !

!TranslatedReceiverFinder methodsFor: 'accessing' stamp: 'tak 9/27/2007 15:59'!
findWordsWith: aSymbol in: aMethodReference 
	"Find words for translation with the symbol in a method. See
	LanguageEditorTest >>testFindTranslatedWords"
	"| message | 
	message := MethodReference new setStandardClass: Morph class
	methodSymbol: #supplementaryPartsDescriptions.
	self new findWordsWIth: #translatedNoop in: message"
	| messages keywords aParseNode |
	aParseNode := aMethodReference decompile.
	"Find from string literal"
	messages := Set new.
	self
		search: aSymbol
		messageNode: aParseNode
		addTo: messages.
	keywords := OrderedCollection new.
	messages
		select: [:aMessageNode | aMessageNode receiver isMemberOf: LiteralNode]
		thenDo: [:aMessageNode | aMessageNode receiver key
				literalStringsDo: [:literal | keywords add: literal]].
	"Find from array literal"
	self
		arraySearch: aSymbol
		messageNode: aParseNode
		addTo: keywords.
	^ keywords! !

!TranslatedReceiverFinder methodsFor: 'accessing' stamp: 'tak 7/24/2007 14:52'!
nonLiteralReceivers
	"self new nonLiteralReceivers"
	| receivers |
	"Answer method references of non literal senders of #translated"
	^ (SystemNavigation default allCallsOn: #translated)
		select: [:message | 
			receivers := OrderedCollection new.
			self search: #translated messageNode: message decompile addTo: receivers.
			receivers
				anySatisfy: [:each | (each receiver isMemberOf: LiteralNode) not]]! !

!TranslatedReceiverFinder methodsFor: 'as yet unclassified' stamp: 'yo 8/2/2004 17:22'!
searchBlockNode: aBlockNode addTo: aCollection

	aBlockNode statements do: [:e |
		(e isMemberOf: MessageNode) ifTrue: [self searchMessageNode: e addTo: aCollection].
		(e isMemberOf: ReturnNode) ifTrue: [self searchReturnNode: e addTo: aCollection].
	].
! !

!TranslatedReceiverFinder methodsFor: 'as yet unclassified' stamp: 'yo 8/2/2004 17:23'!
searchMessageNode: aMessageNode addTo: aCollection

	((aMessageNode receiver isMemberOf: LiteralNode) and: [(aMessageNode selector isMemberOf: SelectorNode) and: [aMessageNode selector key = #translated]]) ifTrue: [
		aCollection add: aMessageNode receiver key.
	].

	(aMessageNode receiver isMemberOf: BlockNode) ifTrue: [self searchBlockNode: aMessageNode receiver addTo: aCollection].
	(aMessageNode receiver isMemberOf: MessageNode) ifTrue: [self searchMessageNode: aMessageNode receiver addTo: aCollection].
	(aMessageNode receiver isMemberOf: ReturnNode) ifTrue: [self searchReturnNode: aMessageNode receiver addTo: aCollection].

	aMessageNode arguments do: [:a |
		(a isMemberOf: BlockNode) ifTrue: [self searchBlockNode: a addTo: aCollection].
		(a isMemberOf: MessageNode) ifTrue: [self searchMessageNode: a addTo: aCollection].
		(a isMemberOf: ReturnNode) ifTrue: [self searchReturnNode: a addTo: aCollection].
	].
! !

!TranslatedReceiverFinder methodsFor: 'as yet unclassified' stamp: 'yo 8/2/2004 17:22'!
searchMethodNode: aMethodNode addTo: aCollection

	(aMethodNode block isMemberOf: BlockNode) ifTrue: [self searchBlockNode: aMethodNode block addTo: aCollection].
	(aMethodNode block isMemberOf: MessageNode) ifTrue: [self searchMessageNode: aMethodNode block addTo: aCollection].
	(aMethodNode block isMemberOf: ReturnNode) ifTrue: [self searchReturnNode: aMethodNode block addTo: aCollection].
! !

!TranslatedReceiverFinder methodsFor: 'as yet unclassified' stamp: 'yo 8/2/2004 17:21'!
searchReturnNode: aReturnNode addTo: aCollection

	(aReturnNode expr isMemberOf: BlockNode) ifTrue: [self searchBlockNode: aReturnNode expr addTo: aCollection].
	(aReturnNode expr isMemberOf: MessageNode) ifTrue: [self searchMessageNode: aReturnNode expr addTo: aCollection].
! !

!TranslatedReceiverFinder methodsFor: 'private' stamp: 'tak 7/24/2007 17:15'!
search: aSymbol messageNode: aParseNode addTo: aCollection 
	"self new search: #translated messageNode: (Project decompile: #updateLocaleDependentsWithPreviousSupplies:gently:) addTo: OrderedCollection new"

	((aParseNode isMemberOf: MessageNode)
			and: [(aParseNode selector isMemberOf: SelectorNode)
					and: [aParseNode selector key = aSymbol]])
		ifTrue: [aCollection add: aParseNode].
	(aParseNode notNil
			and: [aParseNode isLeaf not])
		ifTrue: [aParseNode getAllChildren
				do: [:child | self
						search: aSymbol
						messageNode: child
						addTo: aCollection]].
	^ aCollection! !

!TranslatedReceiverFinder methodsFor: 'as yet unclassified' stamp: 'nice 1/10/2010 15:39'!
senders

	| m o |
	m := SystemNavigation default allCallsOn: #translated.
	m := m collect: [:e |
		e classIsMeta ifTrue: [
			(Smalltalk at: e classSymbol) class decompile: e methodSymbol.
		] ifFalse: [
			(Smalltalk at: e classSymbol) decompile: e methodSymbol.
		]
	].

	o := OrderedCollection new.
	m do: [:e | self searchMethodNode: e addTo: o].
	^ o sort
! !

!TranslatedReceiverFinder methodsFor: 'accessing' stamp: 'tak 7/24/2007 14:53'!
stringReceivers
	"TranslatedReceiverFinder new stringReceivers"
	| stringReceivers messages |
	messages := Set new.
	(SystemNavigation default allCallsOn: #translated)
		do: [:message | self search: #translated messageNode: message decompile addTo: messages].
	stringReceivers := messages
				select: [:each | each receiver isMemberOf: LiteralNode]
				thenCollect: [:each | each receiver key].
	^ stringReceivers asArray sort! !

!TranslatedReceiverFinder methodsFor: 'accessing' stamp: 'KR 7/11/2007 02:06'!
stringReceiversWithContext
	| mrs results rr cls mn t o |
	mrs := SystemNavigation default allCallsOn: #translated.
	results := OrderedCollection new.
	mrs do: [:mr |
		rr := OrderedCollection new.
		cls := Smalltalk at: mr classSymbol.
		rr add: cls category.
		rr add: mr classSymbol.
		rr add: mr methodSymbol.
		mr classIsMeta ifTrue: [
			mn :=  cls class decompile: mr methodSymbol.
		] ifFalse: [
			mn := cls decompile: mr methodSymbol.
		].
		o := OrderedCollection new.
		t := Set new.
		self searchMessageNode: mn addTo: t.
		t do: [ :te |
			(te receiver isMemberOf: LiteralNode) ifTrue: [ 
			    o add: te receiver key.
			].
		].
		o ifNotEmpty: [
			rr add: o.
			results add: rr.
		].
	].
	^ results.

! !

!TranslatedReceiverFinder methodsFor: 'accessing' stamp: 'tak 9/27/2007 15:59'!
stringReceiversWithContext: aSymbol
	"Find string receivers for a symbol.
	Answer a collection of aMethodReference -> {keyword. keyword...}"
	"self new stringReceiversWithContext: #translated"
	| keywords methodReferences |
	methodReferences := SystemNavigation default allCallsOn: aSymbol.
	^ methodReferences inject: OrderedCollection new into: [:list :next |
		keywords := self findWordsWith: aSymbol in: next.
		keywords
			ifNotEmpty: [list add: next -> keywords].
		list]
! !

Object subclass: #UserInterfaceTheme
	instanceVariableNames: 'scope properties name next ignoreApply'
	classVariableNames: 'All Current Default'
	poolDictionaries: ''
	category: 'System-Support'!

!UserInterfaceTheme commentStamp: '<historical>' prior: 0!
A UserInterfaceTheme is a dictionary of preferred visual-properties; colors, borderStyles, borderWidths, fonts, forms, etc. used to color and style the IDE.

Accessing The Theme
To access the proper UserInterfaceTheme instance for an object, send it #userInterfaceTheme.  The default implementation on Object provides the one instance of that is in-use by the IDE at the current time.

Customizing The Theme
We can ask the userInterfaceTheme for the value of any visual-property, by name:

	mySystemWindow userInterfaceTheme closeBoxImage

Initially, the above answers nil, which causes the legacy code to use whatever default it's always used.  To override various visual-properties of any kind of object, the #set: onAny: to: message can be used.  For example, 

	myUserInterfaceTheme
		set: #closeBoxImage 
		for: SystemWindow
		to: MenuIcons smallCancelIcon

Alternatively, values may be derived based on other values in the theme, as in:

	myUserInterfaceTheme
		set: #color 
		for: FillInTheBlankMorph
		to: { MenuMorph->#color.  #twiceDarker }

Now, the accessing expression, above, will answer will answer MenuIcons' smallCancelIcon instead of nil.  SystemWindow's code can be changed to use the expression above to access elements of the theme.

Upgrading Legacy Code
Following the introduction of this class, various client code all around the system must be modified to access it.  This variety of legacy code uses a variety of methods to specify their visual properties:

	1) a hard-coded values.
	2) a values derived from some other value.
	3) providing local storage for a settable value which can be nil.
	4) providing local storage for a settable value which is expected to always have a particular valid value (never nil).

The requirement, for each case, is to let the value be overridden.  

The solution for each of the above should be handled respectively to the above list, as follows:

	1) Check the userInterfaceTheme, if that property returns nil, use the legacy hard-coded value.  (see example: SystemWindow>>#createCloseBox).
	2) Nothing to do -- simply perform the same derivation on the result of (1).
	3) Check the local storage, if present, use it.  If nil, then check the userInterfaceTheme, if it has this property present, use it, else return nil.
	4) Check the userInterfaceTheme, if the property is not nil, use it, otherwise use the local value.

Tool Support
If a new access to #userInterfaceTheme is added to the code, be sure to add the property and its description to the #themeSettings for that class.  See implementors of #themeSettings for examples.!

UserInterfaceTheme subclass: #CommunityTheme
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!CommunityTheme commentStamp: 'mt 7/26/2016 16:22' prior: 0!
A dark theme designed by members the Squeak community such as Karl Ramberg, Chris Muller, and Marcel Taeumel.!

!CommunityTheme class methodsFor: 'instance creation' stamp: 'cmm 8/4/2016 23:36'!
addDarkButtons: aUserInterfaceTheme
	"self createDark apply."
	aUserInterfaceTheme
		set: #borderColor for: #PluggableButtonMorph to: Color darkGray darker ;
		set: #color for: #PluggableButtonMorph to: Color gray darker ;
		set: #textColor for: #PluggableButtonMorph to: Color white;

		set: #selectionModifier for: #PluggableButtonMorph to: [ [:c | self dbSelection twiceDarker] ];
		set: #hoverModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: 0.2] ];
		set: #feedbackModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: 0.3] ].
	"And the plus-version."
	aUserInterfaceTheme set: #disabledTextColor for: #PluggableButtonMorphPlus to: Color lightGray! !

!CommunityTheme class methodsFor: 'instance creation' stamp: 'cmm 8/7/2016 20:08'!
addDarkDialogs: aUserInterfaceTheme
	"self createDark apply."
	aUserInterfaceTheme
		set: #borderColor for: #DialogWindow to: self dbSelection;
		set: #color for: #DialogWindow to: (self dbBlue adjustSaturation: -0.20 brightness: -0.30);

		set: #titleColor for: #DialogWindow to: self dbBlue twiceDarker;
		set: #titleTextColor for: #DialogWindow to: Color veryLightGray;

		set: #textColor for: #DialogWindow to: (Color gray: 0.9);

		set: #okColor for: #DialogWindow to: self dbGreen;
		set: #cancelColor for: #DialogWindow to: self dbOrange;
		set: #buttonColor for: #DialogWindow to: Color darkGray;
		set: #selectionModifier for: #DialogWindow to: [ [:c | self dbSelection twiceDarker ] ].
	"The List Chooser is a dialog, too."
	aUserInterfaceTheme
		set: #addColor for: #ListChooser to: self dbBlue;
		set: #disabledColor for: #ListChooser to: Color transparent.
	"And the system progress bar."
	aUserInterfaceTheme
		set: #color for: #SystemProgressBarMorph to: Color transparent;
		set: #barColor for: #SystemProgressBarMorph to: self dbSelection.
	"And the balloon morphs."
	aUserInterfaceTheme
		set: #borderColor for: #NewBalloonMorph to: Color transparent ;
		set: #color for: #NewBalloonMorph to: (self dbOrange twiceDarker alpha: 0.9) ;
		set: #textColor for: #NewBalloonMorph to: Color white .! !

!CommunityTheme class methodsFor: 'instance creation' stamp: 'cmm 8/7/2016 21:41'!
addDarkFonts: aUserInterfaceTheme
	"Set-up fonts."
	aUserInterfaceTheme
		set: #balloonHelpFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 7 emphasized: TextEmphasis italic emphasisCode);
		set: #standardButtonFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 7);
		set: #standardCodeFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardDefaultTextFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardFlapFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 7 emphasized: TextEmphasis bold emphasisCode);
		set: #haloLabelFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardListFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardMenuFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardSystemFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #windowTitleFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9 emphasized: TextEmphasis bold emphasisCode)! !

!CommunityTheme class methodsFor: 'instance creation' stamp: 'cmm 8/9/2016 14:52'!
addDarkMenusAndDockingBars: aUserInterfaceTheme
	"self createDark apply."
	aUserInterfaceTheme
		set: #borderWidth for: #MenuMorph to: 0;
		set: #color for: #MenuMorph to: Color darkGray twiceDarker;
		set: #titleTextColor for: #MenuMorph to: Color white;
		set: #lineColor for: #MenuMorph to: Color darkGray;
		set: #lineStyle for: #MenuMorph to: BorderStyle default;
		set: #lineWidth for: #MenuMorph to: 1.
	aUserInterfaceTheme
		set: #textColor for: #MenuItemMorph to: self dbForeground;
		set: #selectionColor for: #MenuItemMorph to: self dbSelection;
		set: #selectionTextColor for: #MenuItemMorph to: Color white ;
		set: #disabledTextColor for: #MenuItemMorph to: self dbForeground muchDarker.
		"set: #subMenuMarker for: #MenuItemMorph to: nil." "Use hard-coded default. See MenuItemMorph."
	"The world main docking bar."
	aUserInterfaceTheme
"		set: #color for: #DockingBarMorph to: Color darkGray;"
"		set: #selectionColor for: #DockingBarItemMorph to: self darkContentSecondary;"
		set: #logoColor for: #TheWorldMainDockingBar to: Color white;
		set: #selectionLogoColor for: #TheWorldMainDockingBar to: Color white! !

!CommunityTheme class methodsFor: 'instance creation' stamp: 'cmm 8/11/2016 23:28'!
addDarkScrollables: aUserInterfaceTheme
	"self createDark apply."
	"Scroll bars"
	aUserInterfaceTheme
		set: #thumbColor for: #ScrollBar to: self dbGray;
		set: #thumbBorderColor for: #ScrollBar to: self dbGray twiceDarker.
	"Scroll panes (includes generic stuff for list widgets, tree widgets, and text widgets."
	aUserInterfaceTheme
		set: #borderColor for: #ScrollPane to: (Color transparent) ; "So the search box isn't outlined."
		set: #color for: #ScrollPane to: self dbBackground.
	"List widgets"
	aUserInterfaceTheme
		set: #textColor for: #PluggableListMorph to: (Color gray: 0.9);
		set: #selectionColor for: #PluggableListMorph to: self dbSelection;
		set: #selectionTextColor for: #PluggableListMorph to: Color white ;
		derive: #multiSelectionColor for: #PluggableListMorph from: #PluggableListMorph at: #selectionColor do: [:c | c twiceDarker];
		set: #filterColor for: #PluggableListMorph to: (self dbYellow alpha: 0.4);
		derive: #filterTextColor for: #PluggableListMorph from: #PluggableListMorph at: #textColor ;
		set: #preSelectionModifier for: #PluggableListMorph to: [ [:c | c twiceDarker ] ];
		set: #hoverSelectionModifier for: #PluggableListMorph to: [ [:c | c twiceDarker alpha: 0.5 ] ].
	"Tree widgets"
	aUserInterfaceTheme
		set: #highlightTextColor for: #SimpleHierarchicalListMorph to: self dbYellow lighter lighter;
		set: #lineColor for: #SimpleHierarchicalListMorph to: Color gray.
	"Text widgets"
	aUserInterfaceTheme
		set: #textColor for: #PluggableTextMorph to: (Color gray: 0.9);
		set: #caretColor for: #PluggableTextMorph to: Color orange darker;
		set: #selectionColor for: #PluggableTextMorph to: (self dbSelection duller duller);
		set: #unfocusedSelectionModifier for: #PluggableTextMorph to: [ [:c | c duller] ];
		set: #adornmentReadOnly for: #PluggableTextMorph to: self dbPurple;
		set: #adornmentRefuse for: #PluggableTextMorph to: self dbBlue;
		set: #adornmentConflict for: #PluggableTextMorph to: self dbRed;
		set: #adornmentDiff for: #PluggableTextMorph to: self dbGreen;
		set: #adornmentNormalEdit for: #PluggableTextMorph to: self dbOrange;
		set: #adornmentDiffEdit for: #PluggableTextMorph to: self dbYellow;
		set: #frameAdornmentWidth for: #PluggableTextMorph to: 2.
	aUserInterfaceTheme
		set: #balloonTextColor for: #PluggableTextMorphPlus to: Color lightGray! !

!CommunityTheme class methodsFor: 'instance creation' stamp: 'mt 8/11/2016 16:39'!
addDarkSyntaxHighlighting: aUserInterfaceTheme
	"self createDark apply."
	| normal bold italic underlined darkMap | normal  := TextEmphasis normal.  bold:=TextEmphasis bold.  italic:=TextEmphasis italic.  underlined := TextEmphasis underlined.  darkMap := StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9.
	aUserInterfaceTheme
		set: #color for: #TextAction to: self dbBlue;
	
		set: #default for: #SHTextStylerST80 to: {self dbForeground};
		set: #invalid for: #SHTextStylerST80 to: {self dbInvalid};
		set: #excessCode for: #SHTextStylerST80 to: {self dbInvalid twiceDarker};
		"Descriptive text for humans, italicized."
		set: #comment for: #SHTextStylerST80 to: {self dbComment. italic};
		set: #unfinishedComment for: #SHTextStylerST80 to: {self dbComment darker. italic};
		set: #'$' for: #SHTextStylerST80 to: {self dbConstant};
		set: #character for: #SHTextStylerST80 to: {self dbConstant};
		set: #integer for: #SHTextStylerST80 to: {self dbConstant};
		set: #number for: #SHTextStylerST80 to: {self dbConstant};
		set: #- for: #SHTextStylerST80 to: {self dbForeground. bold};
		set: #= for: #SHTextStylerST80 to: {self dbForeground. bold};
		set: #symbol for: #SHTextStylerST80 to: {self dbBedrock};
		set: #stringSymbol for: #SHTextStylerST80 to: {self dbBedrock};
		set: #literalArray for: #SHTextStylerST80 to: {self dbForeground};
		set: #string for: #SHTextStylerST80 to: {self dbConstant};
		set: #unfinishedString for: #SHTextStylerST80 to: {self dbConstant darker};
		set: #assignment for: #SHTextStylerST80 to: {nil. bold};
		set: #ansiAssignment for: #SHTextStylerST80 to: {nil. bold};
		set: #literal for: #SHTextStylerST80 to: {nil. bold};
		set: #keyword for: #SHTextStylerST80 to: {self dbMessage};
		set: #binary for: #SHTextStylerST80 to: {self dbForeground. bold};
		set: #unary for: #SHTextStylerST80 to: {self dbMessage};
		set: #incompleteKeyword for: #SHTextStylerST80 to: {self dbMessage darker. {underlined. bold}};
		set: #incompleteBinary for: #SHTextStylerST80 to: {self dbMessage darker. underlined};
		set: #incompleteUnary for: #SHTextStylerST80 to: {self dbMessage darker. underlined};
		set: #undefinedKeyword for: #SHTextStylerST80 to: {self dbInvalid};
		set: #undefinedBinary for: #SHTextStylerST80 to: {self dbInvalid};
		set: #undefinedUnary for: #SHTextStylerST80 to: {self dbInvalid};
		"Delineate the selector (good for new users), and make the method look like a mini-document with a title."
		set: #patternKeyword for: #SHTextStylerST80 to: {self dbMessage lighter.  {bold. underlined}};
		set: #patternBinary for: #SHTextStylerST80 to: {nil. bold};
		set: #patternUnary for: #SHTextStylerST80 to: {self dbMessage lighter.  {bold. underlined}};	
		set: #self for: #SHTextStylerST80 to: {self dbBedrock. bold};
		set: #super for: #SHTextStylerST80 to: {self dbBedrock. bold}; 
		set: #true for: #SHTextStylerST80 to: {self dbBedrock. bold};
		set: #false for: #SHTextStylerST80 to: {self dbBedrock. bold};
		set: #nil for: #SHTextStylerST80 to: {self dbBedrock. bold};
		set: #thisContext for: #SHTextStylerST80 to: {self dbBedrock. bold};
		set: #return for: #SHTextStylerST80 to: {self dbForeground. bold};
		set: #patternArg for: #SHTextStylerST80 to: {self dbSelection twiceLighter. TextEmphasis normal. "darkMap"};	
		set: #methodArg for: #SHTextStylerST80 to: {self dbSelection twiceLighter. TextEmphasis normal. "darkMap"};
		set: #blockPatternArg for: #SHTextStylerST80 to: {self dbSelection twiceLighter};
		set: #blockArg for: #SHTextStylerST80 to: {self dbSelection twiceLighter};
		set: #argument for: #SHTextStylerST80 to: {self dbSelection twiceLighter};
		set: #blockArgColon for: #SHTextStylerST80 to: {self dbBedrock}; 
		set: #leftParenthesis for: #SHTextStylerST80 to: {self dbBedrock}; 
		set: #rightParenthesis for: #SHTextStylerST80 to: {self dbBedrock}; 
		set: #leftParenthesis1 for: #SHTextStylerST80 to: {self dbGreen}; 
		set: #rightParenthesis1 for: #SHTextStylerST80 to: {self dbGreen}; 
		set: #leftParenthesis2 for: #SHTextStylerST80 to: {self dbPurple}; 
		set: #rightParenthesis2 for: #SHTextStylerST80 to: {self dbPurple}; 
		set: #leftParenthesis3 for: #SHTextStylerST80 to: {self dbRed}; 
		set: #rightParenthesis3 for: #SHTextStylerST80 to: {self dbRed}; 
		set: #leftParenthesis4 for: #SHTextStylerST80 to: {self dbGreen}; 
		set: #rightParenthesis4 for: #SHTextStylerST80 to: {self dbGreen}; 
		set: #leftParenthesis5 for: #SHTextStylerST80 to: {self dbOrange}; 
		set: #rightParenthesis5 for: #SHTextStylerST80 to: {self dbOrange}; 
		set: #leftParenthesis6 for: #SHTextStylerST80 to: {self dbPurple}; 
		set: #rightParenthesis6 for: #SHTextStylerST80 to: {self dbPurple}; 
		set: #leftParenthesis7 for: #SHTextStylerST80 to: {self dbBlue}; 
		set: #rightParenthesis7 for: #SHTextStylerST80 to: {self dbBlue}; 
		set: #blockStart for: #SHTextStylerST80 to: {self dbBedrock}; 
		set: #blockEnd for: #SHTextStylerST80 to: {self dbBedrock}; 
		set: #blockStart1 for: #SHTextStylerST80 to: {self dbGreen}; 
		set: #blockEnd1 for: #SHTextStylerST80 to: {self dbGreen}; 
		set: #blockStart2 for: #SHTextStylerST80 to: {self dbPurple}; 
		set: #blockEnd2 for: #SHTextStylerST80 to: {self dbPurple}; 
		set: #blockStart3 for: #SHTextStylerST80 to: {self dbRed}; 
		set: #blockEnd3 for: #SHTextStylerST80 to: {self dbRed}; 
		set: #blockStart4 for: #SHTextStylerST80 to: {self dbGreen}; 
		set: #blockEnd4 for: #SHTextStylerST80 to: {self dbGreen}; 
		set: #blockStart5 for: #SHTextStylerST80 to: {self dbOrange}; 
		set: #blockEnd5 for: #SHTextStylerST80 to: {self dbOrange}; 
		set: #blockStart6 for: #SHTextStylerST80 to: {self dbPurple}; 
		set: #blockEnd6 for: #SHTextStylerST80 to: {self dbPurple}; 
		set: #blockStart7 for: #SHTextStylerST80 to: {self dbBlue}; 
		set: #blockEnd7 for: #SHTextStylerST80 to: {self dbBlue}; 																																																		
		set: #arrayStart for: #SHTextStylerST80 to: {self dbBedrock}; 
		set: #arrayEnd for: #SHTextStylerST80 to: {self dbBedrock}; 
		set: #arrayStart1 for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #arrayEnd1 for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #byteArrayStart for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #byteArrayEnd for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #byteArrayStart1 for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #byteArrayEnd1 for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #leftBrace for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #rightBrace for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #cascadeSeparator for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #statementSeparator for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #externalCallType for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #externalCallTypePointerIndicator for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #primitiveOrExternalCallStart for: #SHTextStylerST80 to: {self dbForeground}; 
		set: #primitiveOrExternalCallEnd for: #SHTextStylerST80 to: {self dbForeground};
		set: #methodTempBar for: #SHTextStylerST80 to: {self dbBedrock};
		set: #blockTempBar for: #SHTextStylerST80 to: {self dbBedrock};
		set: #blockArgsBar for: #SHTextStylerST80 to: {self dbBedrock};
		set: #primitive for: #SHTextStylerST80 to: {self dbGreen lighter. bold};
		set: #pragmaKeyword for: #SHTextStylerST80 to: {self dbGreen. bold};
		set: #pragmaUnary for: #SHTextStylerST80 to: {self dbGreen. bold};
		set: #pragmaBinary for: #SHTextStylerST80 to: {self dbGreen. bold};									
		set: #externalFunctionCallingConvention for: #SHTextStylerST80 to: {self dbGreen. bold}; 
		set: #module for: #SHTextStylerST80 to: {self dbGreen. bold};
		set: #blockTempVar for: #SHTextStylerST80 to: {self dbLocal. italic};
		set: #blockPatternTempVar for: #SHTextStylerST80 to: {self dbLocal. italic};
		set: #instVar for: #SHTextStylerST80 to: {self dbYellow. normal };
		set: #workspaceVar for: #SHTextStylerST80 to: {self dbLocal. italic};
		set: #undefinedIdentifier for: #SHTextStylerST80 to: {self dbInvalid};
		set: #incompleteIdentifier for: #SHTextStylerST80 to: {self dbGray. underlined};
		set: #tempVar for: #SHTextStylerST80 to: {self dbLocal. italic};
		set: #patternTempVar for: #SHTextStylerST80 to: {self dbLocal. italic};
		set: #poolConstant for: #SHTextStylerST80 to: {self dbConstant };
		set: #classVar for: #SHTextStylerST80 to: {self dbReference};
		set: #globalVar for: #SHTextStylerST80 to: {self dbClass. normal}.
	"And the text differ"
	aUserInterfaceTheme	
		set: #insertTextAttributes for: #TextDiffBuilder to: { TextColor color: self dbRed };
		set: #removeTextAttributes for: #TextDiffBuilder to: { TextEmphasis struckOut. TextColor color: self dbBlue };
		set: #normalTextAttributes for: #TextDiffBuilder to: { TextEmphasis normal }.! !

!CommunityTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 12:18'!
addDarkToolColors: theme
	"Tool-specific colors."
	
	theme 
		set: #failureColor for: #TestRunner to: self dbYellow;
		set: #errorColor for: #TestRunner to: self dbRed;
		set: #passColor for: #TestRunner to: self dbGreen.! !

!CommunityTheme class methodsFor: 'instance creation' stamp: 'cmm 8/7/2016 21:49'!
addDarkWindowColors: aUserInterfaceTheme
	"self createDark apply."
	aUserInterfaceTheme
		set: #uniformWindowColor for: #Model to: Color darkGray;
		
		set: #unfocusedWindowColorModifier for: #SystemWindow to: [ [:color | color darker] ];
		set: #unfocusedLabelColor for: #SystemWindow to: Color veryLightGray;
		set: #focusedLabelColor for: #SystemWindow to: Color white;

		set: #customWindowColor for: #Browser to: self dbBlue;
		set: #customWindowColor for: #ChangeList to: self dbBlue;
		set: #customWindowColor for: #ChangeSorter to: self dbBlue;
		set: #customWindowColor for: #ChatNotes to: self dbPurple twiceDarker;
		set: #customWindowColor for: #ClassCommentVersionsBrowser to: self dbPurple twiceDarker;
		set: #customWindowColor for: #Debugger to: self dbRed;
		set: #customWindowColor for: #DualChangeSorter to: self dbOrange twiceDarker;
		set: #customWindowColor for: #FileContentsBrowser to: self dbGray;
		set: #customWindowColor for: #FileList to: self dbGray;
		set: #customWindowColor for: #Inspector to: self dbYellow duller;
		set: #customWindowColor for: #InstanceBrowser to: self dbYellow duller;
		set: #customWindowColor for: #Lexicon to: self dbGreen;
		set: #customWindowColor for: #MCTool to: self dbOrange twiceDarker;
		set: #customWindowColor for: #MessageNames to: self dbGreen;
		set: #customWindowColor for: #MessageSet to: self dbGreen;
		set: #customWindowColor for: #ObjectExplorer to: self dbYellow duller;
		set: #customWindowColor for: #PackagePaneBrowser to: self dbBlue;
		set: #customWindowColor for: #PluggableFileList to: self dbGray;
		set: #customWindowColor for: #PreferenceBrowser to: self dbBlue;
		set: #customWindowColor for: #ProcesBrowser to: self dbAqua;
		set: #customWindowColor for: #SMLoader to: self dbOrange twiceDarker;
		set: #customWindowColor for: #SMLoaderPlus to: self dbOrange twiceDarker;
		set: #customWindowColor for: #SMReleaseBrowser to: self dbOrange twiceDarker;
		set: #customWindowColor for: #ScriptingDomain to: self dbYellow duller;
		set: #customWindowColor for: #SelectorBrowser to: self dbBlue;
		set: #customWindowColor for: #StringHolder to: self dbGray;
		set: #customWindowColor for: #TestRunner to: self dbPink darker;
		set: #customWindowColor for: #TranscriptStream to: self dbGray;
		set: #customWindowColor for: #VersionsBrowser to: self dbPurple twiceDarker;
		set: #customWindowColor for: #Workspace to: self dbPink darker.! !

!CommunityTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 12:18'!
createDark
	"self createDark apply."
	| name |
	name := 'Community (dark)'.
	^ (self named: name) in: [:theme |
		theme merge: (self named: 'Squeak') overwrite: true.
		theme name: name.
		"General morph stuff."
		theme
			set: #borderColor for: #ScrollPane to: (Color transparent) ;
			set: #keyboardFocusColor for: #Morph to: (self dbSelection adjustSaturation: -0.3 brightness: 0.10);
			set: #keyboardFocusWidth for: #Morph to: 2;
			set: #softShadowColor for: #Morph to: (self dbSelection muchLighter alpha: 0.025);
			set: #softShadowOffset for: #Morph to: (10@8 corner: 10@12);
			set: #hardShadowColor for: #Morph to: (self dbSelection muchLighter alpha: 0.02);
			set: #hardShadowOffset for: #Morph to: 1@1.
		self
			addDarkFonts: theme;
			addDarkWindowColors: theme;
			addDarkSyntaxHighlighting: theme;
			addDarkScrollables: theme;
			addDarkButtons: theme;
			addDarkDialogs: theme;
			addDarkMenusAndDockingBars: theme;
			addDarkToolColors: theme.
		theme]! !

!CommunityTheme class methodsFor: 'colors' stamp: 'cmm 8/7/2016 22:01'!
dbAqua
	"Suitably saturated to serve as dbSelection."
	^ Color r: 0.1 g: 0.5 b: 0.5! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/7/2016 20:13'!
dbBackground
	"Emptiness."
	^Color black! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/4/2016 22:50'!
dbBedrock
	"Basic syntactical matter which users cannot change."
	^ self dbForeground! !

!CommunityTheme class methodsFor: 'colors' stamp: 'cmm 8/1/2016 21:29'!
dbBlue         
	^Color r: 0.406 g: 0.535 b: 0.645! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/4/2016 21:08'!
dbClass
	^ self dbBlue! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/4/2016 22:31'!
dbComment
	"The code should speak for itself.  Where it cannot, any comments should just be subtle, clarifying whispers."
	^ Color gray darker! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/4/2016 21:16'!
dbConstant
	"Let unparameterized constants stand out in red."
	^ self dbRed! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/4/2016 22:33'!
dbForeground
	"Not too bright."
	^Color r: 0.764 g: 0.776 b: 0.768! !

!CommunityTheme class methodsFor: 'colors' stamp: 'cmm 8/4/2016 13:40'!
dbGray
	^ Color gray slightlyDarker! !

!CommunityTheme class methodsFor: 'colors' stamp: 'cmm 8/1/2016 22:00'!
dbGreen        
	^(Color r: 0.33 g: 0.542 b: 0.287)! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/11/2016 23:26'!
dbInvalid
	"Purposefully bright, so that it conflicts with the rest of the palette, visually indicative of a problem."
	^ Color magenta twiceLighter! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/4/2016 22:52'!
dbLocal
	"Local variables can be descriptive, sometimes saving the need for a comment."
	^ self dbPink! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/4/2016 22:43'!
dbMessage
	"Green is the universal color for okay, positive, go.  In Squeak, sending messages are what make objects 'go'."
	^ self dbGreen
		adjustSaturation: 0.05
		brightness: 0.10! !

!CommunityTheme class methodsFor: 'colors' stamp: 'mt 7/26/2016 15:52'!
dbOrange       
	^Color r: 0.871 g: 0.577 b: 0.372! !

!CommunityTheme class methodsFor: 'colors' stamp: 'cmm 8/3/2016 22:22'!
dbPink       
	^ (Color r: 0.739 g: 0.552 b: 0.652)! !

!CommunityTheme class methodsFor: 'colors' stamp: 'mt 7/26/2016 15:52'!
dbPurple       
	^Color r: 0.698 g: 0.581 b: 0.733! !

!CommunityTheme class methodsFor: 'colors' stamp: 'cmm 8/1/2016 21:30'!
dbRed 
	^Color r: 0.6 g: 0.3 b: 0.3! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/3/2016 22:41'!
dbReference
	"References to other objects.  Yellow like gold."
	^ self dbYellow! !

!CommunityTheme class methodsFor: 'colors by purpose' stamp: 'cmm 8/8/2016 21:21'!
dbSelection
	"Selections are transient, like electricity, so a good neon color.  The arguments passed to methods and blocks could be considered as much, 'selections', as those made from a list or menu."
	^ self dbAqua adjustSaturation: 0.10 brightness: 0.08! !

!CommunityTheme class methodsFor: 'colors' stamp: 'cmm 8/4/2016 14:16'!
dbYellow       
	^ (Color r: 0.675 g: 0.632 b: 0.16)! !

UserInterfaceTheme subclass: #MonokaiTheme
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 8/15/2016 10:59'!
addDarkButtons: theme
	"self createDark apply."
	theme
		set: #borderColor for: #PluggableButtonMorph to: self backgroundColor;
		set: #color for: #PluggableButtonMorph to: self invisibleColor;
		set: #textColor for: #PluggableButtonMorph to: self yellow;

		set: #selectionModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.1] ];
		set: #hoverModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.1] ];
		set: #feedbackModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.3] ].
	
	"And the plus-version."
	theme set: #disabledTextColor for: #PluggableButtonMorphPlus to: self grayDarker.
! !

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 8/3/2016 16:56'!
addDarkDialogs: theme
	"self createDark apply."
	theme
		set: #borderColor for: #DialogWindow to: self backgroundColor muchDarker;
		set: #color for: #DialogWindow to: self invisibleColor;

		set: #titleColor for: #DialogWindow to: self backgroundColor;
		set: #titleTextColor for: #DialogWindow to: self foregroundColor;
		
		set: #textColor for: #DialogWindow to: self yellow;
		
		set: #okColor for: #DialogWindow to: self green muchDarker;
		set: #cancelColor for: #DialogWindow to: self orange muchDarker;
		set: #buttonColor for: #DialogWindow to: self invisibleColor;
		set: #selectionModifier for: #DialogWindow to: [ [:c | c adjustBrightness: -0.1 ] ].
	
	"The List Chooser is a dialog, too."
	theme
		set: #addColor for: #ListChooser to: self blue;
		set: #disabledColor for: #ListChooser to: Color transparent.

		
	"And the system progress bar."
	theme
		set: #color for: #SystemProgressBarMorph to: self invisibleColor;
		set: #barColor for: #SystemProgressBarMorph to: self foregroundColor.
		
	"And the balloon morphs."
	theme
		set: #borderColor for: #NewBalloonMorph to: self backgroundColor muchDarker;
		set: #color for: #NewBalloonMorph to: self invisibleColor.! !

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 7/26/2016 15:11'!
addDarkFonts: theme

	"Set-up fonts."
	theme
		set: #balloonHelpFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 7);
		set: #standardButtonFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 7);
		set: #standardCodeFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardDefaultTextFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardFlapFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 7 emphasized: TextEmphasis bold emphasisCode);
		set: #haloLabelFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardListFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardMenuFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardSystemFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #windowTitleFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9).! !

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 8/5/2016 08:47'!
addDarkMenusAndDockingBars: theme
	"self createDark apply."
	theme
		set: #borderWidth for: #MenuMorph to: 1;
		set: #borderColor for: #MenuMorph to: self invisibleColor;
		set: #color for: #MenuMorph to: self backgroundColor;
		set: #titleTextColor for: #MenuMorph to: self yellow;
		set: #lineColor for: #MenuMorph to: self invisibleColor;
		set: #lineStyle for: #MenuMorph to: BorderStyle default;
		set: #lineWidth for: #MenuMorph to: 1.
		
	theme
		set: #textColor for: #MenuItemMorph to: self foregroundColor;
		set: #disabledTextColor for: #MenuItemMorph to: self grayLight;
		set: #selectionColor for: #MenuItemMorph to: self invisibleColor;
		set: #selectionTextColor for: #MenuItemMorph to: self yellow.
		"set: #subMenuMarker for: #MenuItemMorph to: nil." "Use hard-coded default. See MenuItemMorph."
		
	"The world main docking bar."
	theme
		set: #color for: #DockingBarMorph to: self invisibleColor;
		set: #selectionColor for: #DockingBarItemMorph to: self grayLight;
		set: #logoColor for: #TheWorldMainDockingBar to: self foregroundColor;
		set: #selectionLogoColor for: #TheWorldMainDockingBar to: self yellow.! !

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 7/26/2016 15:22'!
addDarkScrollables: theme
	"self createDark apply."

	"Scroll bars"
	theme
		set: #thumbColor for: #ScrollBar to: self backgroundColor;
		set: #thumbBorderColor for: #ScrollBar to: self backgroundColor;
		set: #thumbBorderWidth for: #ScrollBar to: 1;
		set: #thumbColorModifier for: #ScrollBar to: [ [:c | c adjustBrightness: -0.2] ];
		set: #pagingAreaColorModifier for: #ScrollBar to: [ [:c | Color transparent ] ];
		set: #borderColorModifier for: #ScrollBar to: [ [:c | c adjustBrightness: -0.2] ].
	
	"Scroll panes (includes generic stuff for list widgets, tree widgets, and text widgets."
	theme
		set: #borderWidth for: #ScrollPane to: 0;
	"	set: #borderColor for: #ScrollPane to: Color transparent;"
		set: #color for: #ScrollPane to: self backgroundColor.
		
	"List widgets"
	theme
		set: #textColor for: #PluggableListMorph to: self foregroundColor;
		set: #selectionColor for: #PluggableListMorph to: self invisibleColor;
		set: #selectionTextColor for: #PluggableListMorph to: self yellow;
		set: #filterColor for: #PluggableListMorph to: self yellow;
		set: #filterTextColor for: #PluggableListMorph to: self backgroundColor;
		set: #preSelectionModifier for: #PluggableListMorph to: [ [:c | c lighter alpha: 0.5 ] ];
		set: #hoverSelectionModifier for: #PluggableListMorph to: [ [:c | c lighter alpha: 0.5 ] ].
		
	"Tree widgets"
	theme
		set: #highlightTextColor for: #SimpleHierarchicalListMorph to: self yellow lighter lighter;
		set: #lineColor for: #SimpleHierarchicalListMorph to: self grayLight.
	
	"Text widgets"
	theme
		set: #textColor for: #PluggableTextMorph to: self foregroundColor;
		set: #caretColor for: #PluggableTextMorph to: self yellow;
		set: #selectionColor for: #PluggableTextMorph to: self invisibleColor;
		set: #unfocusedSelectionModifier for: #PluggableTextMorph to: [ [:c | (Color r: 0.285 g: 0.282 b: 0.242) "invisible color"] ];
		set: #adornmentReadOnly for: #PluggableTextMorph to: self magenta;
		set: #adornmentRefuse for: #PluggableTextMorph to: self cyan;
		set: #adornmentConflict for: #PluggableTextMorph to: self red;
		set: #adornmentDiff for: #PluggableTextMorph to: self green;
		set: #adornmentNormalEdit for: #PluggableTextMorph to: self orange;
		set: #adornmentDiffEdit for: #PluggableTextMorph to: self yellow.
	theme
		set: #balloonTextColor for: #PluggableTextMorphPlus to: self grayLight.! !

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 8/11/2016 16:39'!
addDarkSyntaxHighlighting: theme
	"self createDark apply."
	theme
		set: #color for: #TextAction to: self blue;

		set: #default for: #SHTextStylerST80 to: {self foregroundColor};
		set: #invalid for: #SHTextStylerST80 to: {self red};
		set: #excessCode for: #SHTextStylerST80 to: {self red};
		set: #comment for: #SHTextStylerST80 to: {self commentColor};
		set: #unfinishedComment for: #SHTextStylerST80 to: {self red. TextEmphasis italic};
		set: #'$' for: #SHTextStylerST80 to: {self red};
		set: #character for: #SHTextStylerST80 to: {self numberColor};
		set: #integer for: #SHTextStylerST80 to: {self numberColor};
		set: #number for: #SHTextStylerST80 to: {self numberColor};	
		set: #- for: #SHTextStylerST80 to: {self red};
		set: #symbol for: #SHTextStylerST80 to: {self blue};	
		set: #stringSymbol for: #SHTextStylerST80 to: {self blue};	
		set: #literalArray for: #SHTextStylerST80 to: {self blue};
		set: #string for: #SHTextStylerST80 to: {self stringColor. TextEmphasis normal};
		set: #unfinishedString for: #SHTextStylerST80 to: {self red. TextEmphasis normal};
		set: #assignment for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #ansiAssignment for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #literal for: #SHTextStylerST80 to: {nil. TextEmphasis italic};
		set: #keyword for: #SHTextStylerST80 to: {self blue};
		set: #binary for: #SHTextStylerST80 to: {self blue};	
		set: #unary for: #SHTextStylerST80 to: {self blue};
		set: #incompleteKeyword for: #SHTextStylerST80 to: {self foregroundColor. TextEmphasis underlined};
		set: #incompleteBinary for: #SHTextStylerST80 to: {self foregroundColor. TextEmphasis underlined};	
		set: #incompleteUnary for: #SHTextStylerST80 to: {self foregroundColor. TextEmphasis underlined};
		set: #undefinedKeyword for: #SHTextStylerST80 to: {self red};
		set: #undefinedBinary for: #SHTextStylerST80 to: {self red};	
		set: #undefinedUnary for: #SHTextStylerST80 to: {self red};													
		set: #patternKeyword for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #patternBinary for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #patternUnary for: #SHTextStylerST80 to: {nil. TextEmphasis bold};	
		set: #self for: #SHTextStylerST80 to: {self red};
		set: #super for: #SHTextStylerST80 to: {self red}; 
		set: #true for: #SHTextStylerST80 to: {self red};
		set: #false for: #SHTextStylerST80 to: {self red};
		set: #nil for: #SHTextStylerST80 to: {self red};
		set: #thisContext for: #SHTextStylerST80 to: {self red};
		set: #return for: #SHTextStylerST80 to: {self red};
		set: #patternArg for: #SHTextStylerST80 to: {self blue};	
		set: #methodArg for: #SHTextStylerST80 to: {self blue};
		set: #blockPatternArg for: #SHTextStylerST80 to: {self blue};
		set: #blockArg for: #SHTextStylerST80 to: {self blue};
		set: #argument for: #SHTextStylerST80 to: {self blue};
		set: #blockArgColon for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #leftParenthesis for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #rightParenthesis for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #leftParenthesis1 for: #SHTextStylerST80 to: {self green}; 
		set: #rightParenthesis1 for: #SHTextStylerST80 to: {self green}; 
		set: #leftParenthesis2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #rightParenthesis2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #leftParenthesis3 for: #SHTextStylerST80 to: {self red}; 
		set: #rightParenthesis3 for: #SHTextStylerST80 to: {self red}; 
		set: #leftParenthesis4 for: #SHTextStylerST80 to: {self green}; 
		set: #rightParenthesis4 for: #SHTextStylerST80 to: {self green}; 
		set: #leftParenthesis5 for: #SHTextStylerST80 to: {self orange}; 
		set: #rightParenthesis5 for: #SHTextStylerST80 to: {self orange}; 
		set: #leftParenthesis6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #rightParenthesis6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #leftParenthesis7 for: #SHTextStylerST80 to: {self blue}; 
		set: #rightParenthesis7 for: #SHTextStylerST80 to: {self blue}; 
		set: #blockStart for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #blockEnd for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #blockStart1 for: #SHTextStylerST80 to: {self green}; 
		set: #blockEnd1 for: #SHTextStylerST80 to: {self green}; 
		set: #blockStart2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockEnd2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockStart3 for: #SHTextStylerST80 to: {self red}; 
		set: #blockEnd3 for: #SHTextStylerST80 to: {self red}; 
		set: #blockStart4 for: #SHTextStylerST80 to: {self green}; 
		set: #blockEnd4 for: #SHTextStylerST80 to: {self green}; 
		set: #blockStart5 for: #SHTextStylerST80 to: {self orange}; 
		set: #blockEnd5 for: #SHTextStylerST80 to: {self orange}; 
		set: #blockStart6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockEnd6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockStart7 for: #SHTextStylerST80 to: {self blue}; 
		set: #blockEnd7 for: #SHTextStylerST80 to: {self blue}; 																																																		
		set: #arrayStart for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #arrayEnd for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #arrayStart1 for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #arrayEnd1 for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #byteArrayStart for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #byteArrayEnd for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #byteArrayStart1 for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #byteArrayEnd1 for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #leftBrace for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #rightBrace for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #cascadeSeparator for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #statementSeparator for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #externalCallType for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #externalCallTypePointerIndicator for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #primitiveOrExternalCallStart for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #primitiveOrExternalCallEnd for: #SHTextStylerST80 to: {self foregroundColor};
		set: #methodTempBar for: #SHTextStylerST80 to: {self foregroundColor}; 
		set: #blockTempBar for: #SHTextStylerST80 to: {self foregroundColor};
		set: #blockArgsBar for: #SHTextStylerST80 to: {self foregroundColor};
		set: #primitive for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #pragmaKeyword for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #pragmaUnary for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #pragmaBinary for: #SHTextStylerST80 to: {self green. TextEmphasis bold};									
		set: #externalFunctionCallingConvention for: #SHTextStylerST80 to: {self green. TextEmphasis bold}; 
		set: #module for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #blockTempVar for: #SHTextStylerST80 to: {self foregroundColor};
		set: #blockPatternTempVar for: #SHTextStylerST80 to: {self foregroundColor};
		set: #instVar for: #SHTextStylerST80 to: {self foregroundColor};
		set: #workspaceVar for: #SHTextStylerST80 to: {self foregroundColor};
		set: #undefinedIdentifier for: #SHTextStylerST80 to: {self red};
		set: #incompleteIdentifier for: #SHTextStylerST80 to: {self foregroundColor. {TextEmphasis italic. TextEmphasis underlined}};
		set: #tempVar for: #SHTextStylerST80 to: {self foregroundColor};
		set: #patternTempVar for: #SHTextStylerST80 to: {self foregroundColor};
		set: #poolConstant for: #SHTextStylerST80 to: {self foregroundColor};
		set: #classVar for: #SHTextStylerST80 to: {self foregroundColor};
		set: #globalVar for: #SHTextStylerST80 to: {self foregroundColor}.
		
	"And the text differ"
	theme	
		set: #insertTextAttributes for: #TextDiffBuilder to: { TextColor color: self red };
		set: #removeTextAttributes for: #TextDiffBuilder to: { TextEmphasis struckOut. TextColor color: self blue };
		set: #normalTextAttributes for: #TextDiffBuilder to: { TextEmphasis normal }.! !

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 14:10'!
addDarkToolColors: theme
	"Tool-specific colors."
	
	theme 
		set: #failureColor for: #TestRunner to: self yellow;
		set: #errorColor for: #TestRunner to: self red;
		set: #passColor for: #TestRunner to: self green;
		
		set: #failureTextColor for: #TestRunner to: self backgroundColor;
		set: #errorTextColor for: #TestRunner to: self backgroundColor;
		set: #passTextColor for: #TestRunner to: self backgroundColor.
		
	theme
		set: #ignoredOperationAttributes for: #MCOperationsBrowser to: {TextColor color: self grayLight}.! !

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 13:24'!
addDarkWindowColors: theme
	"self createDark apply."
	theme
		set: #uniformWindowColor for: #Model to:( self invisibleColor adjustBrightness: 0.16) "lighter twice";
		
		set: #unfocusedWindowColorModifier for: #SystemWindow to: [ [:color | color adjustBrightness: -0.16 "darker twice"] ];
		set: #unfocusedLabelColor for: #SystemWindow to: [
			Model useColorfulWindows
				ifTrue: [(Color r: 0.285 g: 0.282 b: 0.242) "invisible color"]
				ifFalse: [(Color r: 0.972 g: 0.972 b: 0.948) "foreground color"] ];
		set: #focusedLabelColor for: #SystemWindow to: [
			Model useColorfulWindows
				ifTrue: [(Color r: 0.152 g: 0.156 b: 0.133) "background color"]
				ifFalse: [(Color r: 0.901 g: 0.858 b: 0.455) "yellow"] ];

		set: #customWindowColor for: #Browser to: self green duller;
		set: #customWindowColor for: #ChangeList to: self blue duller;
		set: #customWindowColor for: #ChangeSorter to: self blue duller;
		set: #customWindowColor for: #ChatNotes to: self magenta duller;
		set: #customWindowColor for: #ClassCommentVersionsBrowser to: self violet duller;
		set: #customWindowColor for: #Debugger to: self red duller;
		set: #customWindowColor for: #DualChangeSorter to: self blue duller;
		set: #customWindowColor for: #FileContentsBrowser to: self yellow duller;
		set: #customWindowColor for: #FileList to: self yellow duller;
		set: #customWindowColor for: #InstanceBrowser to: self cyan duller;
		set: #customWindowColor for: #Lexicon to: self cyan duller;
		set: #customWindowColor for: #MCTool to: self violet duller;
		set: #customWindowColor for: #MessageNames to: self green duller;
		set: #customWindowColor for: #MessageSet to: self cyan duller;
		set: #customWindowColor for: #PackagePaneBrowser to: self green duller;
		set: #customWindowColor for: #PluggableFileList to: self yellow duller;
		set: #customWindowColor for: #PreferenceBrowser to: self cyan duller;
		set: #customWindowColor for: #SMLoader to: self orange duller;
		set: #customWindowColor for: #SMLoaderPlus to: self orange duller;
		set: #customWindowColor for: #SMReleaseBrowser to: self orange duller;
		set: #customWindowColor for: #ScriptingDomain to: self yellow duller;
		set: #customWindowColor for: #SelectorBrowser to: self cyan duller;
		set: #customWindowColor for: #StringHolder to: self yellow duller;
		set: #customWindowColor for: #TestRunner to: self orange duller;
		set: #customWindowColor for: #TranscriptStream to: self orange duller;
		set: #customWindowColor for: #VersionsBrowser to: self violet duller.! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:04'!
argumentColor
	
	^ Color fromString: '#FD971F'! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:04'!
backgroundColor

	^ Color fromString: '#272822'! !

!MonokaiTheme class methodsFor: 'colors by accent' stamp: 'mt 7/26/2016 15:09'!
blue

	^ self globalColor! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:05'!
caretColor
	"invalid foreground, invalid deprecated foreground ..."
	
	^ Color fromString: '#F8F8F0'! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:05'!
classColor
	"inherited class, function name, tag attribute,  ..."
	
	^ Color fromString: '#A6E22E'! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:05'!
commentColor
	
	^ Color fromString: '#75715E'! !

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 18:04'!
createDark
	"self createDark apply."
	| themeName |
	themeName := 'Monokai (dark)'.
	^ (self named: themeName) in: [:theme |
		theme merge: (self named: 'Squeak') overwrite: true.
		theme name: themeName.
		
		"General morph stuff."
		theme
			set: #keyboardFocusColor for: #Morph to: self yellow;
			set: #keyboardFocusWidth for: #Morph to: 1.

		theme set: #background for: #MorphicProject to: self darkBackgroundForm.

		self
			addDarkFonts: theme;
			addDarkWindowColors: theme;
			addDarkSyntaxHighlighting: theme;
			addDarkScrollables: theme;
			addDarkButtons: theme;
			addDarkDialogs: theme;
			addDarkMenusAndDockingBars: theme;
			addDarkToolColors: theme.

		theme]! !

!MonokaiTheme class methodsFor: 'colors by accent' stamp: 'mt 7/26/2016 15:10'!
cyan

	^ SolarizedTheme cyan! !

!MonokaiTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 11:39'!
darkBackgroundForm

	| ref |
	ref := self backgroundColor.

	^ (SqueakTheme linenblue  asFormOfDepth: 32) collectColors: [:c | 
		Color
			h:ref hue
			s: ref saturation
			v: c brightness - 0.1
			alpha: c alpha]! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:03'!
foregroundColor

	^ Color fromString: '#F8F8F2'! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:06'!
globalColor
	"library function, library constant, library class type, ..."
	
	^ Color fromString: '#66D9EF'! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:06'!
grayDarker
	
	^ Color fromString: '#383830'! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:06'!
grayLight
	
	^ Color fromString: '#595959'! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:06'!
grayLightest
	
	^ Color fromString: '#E6E6E6'! !

!MonokaiTheme class methodsFor: 'colors by accent' stamp: 'mt 7/26/2016 15:10'!
green

	^ self classColor! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:04'!
invisibleColor
	"line highlight, selection, ..."
	
	^ Color fromString: '#49483E'! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:06'!
keywordColor
	"tag name, invalid background,  ..."
	
	^ Color fromString: '#F92672'! !

!MonokaiTheme class methodsFor: 'colors by accent' stamp: 'mt 7/26/2016 15:10'!
magenta

	^ self keywordColor! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:07'!
numberColor
	"Constant, invalid deprecated background, ..."
	"purple"
	
	^ Color fromString: '#AE81FF'! !

!MonokaiTheme class methodsFor: 'colors by accent' stamp: 'mt 7/26/2016 15:21'!
orange

	^ self argumentColor! !

!MonokaiTheme class methodsFor: 'colors by accent' stamp: 'mt 7/26/2016 15:10'!
red

	^ SolarizedTheme red! !

!MonokaiTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/26/2016 15:07'!
stringColor
	
	^ Color fromString: '#E6DB74'! !

!MonokaiTheme class methodsFor: 'colors by accent' stamp: 'mt 7/26/2016 15:11'!
violet

	^ self numberColor! !

!MonokaiTheme class methodsFor: 'colors by accent' stamp: 'mt 7/26/2016 15:21'!
yellow

	^ self stringColor! !

UserInterfaceTheme subclass: #SolarizedTheme
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SolarizedTheme commentStamp: 'mt 7/26/2016 15:08' prior: 0!
The Solarized color palette.

http://ethanschoonover.com/solarized

SOLARIZED HEX     16/8 TERMCOL  XTERM/HEX   L*A*B      RGB         HSB
--------- ------- ---- -------  ----------- ---------- ----------- -----------
base03    #002b36  8/4 brblack  234 #1c1c1c 15 -12 -12   0  43  54 193 100  21
base02    #073642  0/4 black    235 #262626 20 -12 -12   7  54  66 192  90  26
base01    #586e75 10/7 brgreen  240 #585858 45 -07 -07  88 110 117 194  25  46
base00    #657b83 11/7 bryellow 241 #626262 50 -07 -07 101 123 131 195  23  51
base0     #839496 12/6 brblue   244 #808080 60 -06 -03 131 148 150 186  13  59
base1     #93a1a1 14/4 brcyan   245 #8a8a8a 65 -05 -02 147 161 161 180   9  63
base2     #eee8d5  7/7 white    254 #e4e4e4 92 -00  10 238 232 213  44  11  93
base3     #fdf6e3 15/7 brwhite  230 #ffffd7 97  00  10 253 246 227  44  10  99
yellow    #b58900  3/3 yellow   136 #af8700 60  10  65 181 137   0  45 100  71
orange    #cb4b16  9/3 brred    166 #d75f00 50  50  55 203  75  22  18  89  80
red       #dc322f  1/1 red      160 #d70000 50  65  45 220  50  47   1  79  86
magenta   #d33682  5/5 magenta  125 #af005f 50  65 -05 211  54 130 331  74  83
violet    #6c71c4 13/5 brmagenta 61 #5f5faf 50  15 -45 108 113 196 237  45  77
blue      #268bd2  4/4 blue      33 #0087ff 55 -10 -45  38 139 210 205  82  82
cyan      #2aa198  6/6 cyan      37 #00afaf 60 -35 -05  42 161 152 175  74  63
green     #859900  2/2 green     64 #5f8700 60 -20  65 133 153   0  68 100  60!

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/15/2016 10:58'!
addDarkButtons: theme
	"self createDark apply."
	theme
		set: #borderColor for: #PluggableButtonMorph to: self darkContentSecondary;
		set: #color for: #PluggableButtonMorph to: self darkBackgroundHighlights;
		set: #textColor for: #PluggableButtonMorph to: self darkContentEmphasizedMore;

		set: #selectionModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.1] ];
		set: #hoverModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.1] ];
		set: #feedbackModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.3] ].
	
	"And the plus-version."
	theme set: #disabledTextColor for: #PluggableButtonMorphPlus to: self darkContentSecondary.
! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/3/2016 16:59'!
addDarkDialogs: theme
	"self createDark apply."
	theme
		set: #borderColor for: #DialogWindow to: self darkBackground muchDarker;
		set: #color for: #DialogWindow to: self darkBackgroundHighlights;

		set: #titleColor for: #DialogWindow to: self darkBackground;
		set: #titleTextColor for: #DialogWindow to: self darkContentEmphasized;
		
		set: #textColor for: #DialogWindow to: self darkContentEmphasizedMore;
		
		set: #okColor for: #DialogWindow to: self green;
		set: #cancelColor for: #DialogWindow to: self orange;
		set: #buttonColor for: #DialogWindow to: self darkBackgroundHighlights;
		set: #selectionModifier for: #DialogWindow to: [ [:c | c adjustBrightness: -0.1 ] ].
	
	"The List Chooser is a dialog, too."
	theme
		set: #addColor for: #ListChooser to: self blue;
		set: #disabledColor for: #ListChooser to: Color transparent.

		
	"And the system progress bar."
	theme
		set: #color for: #SystemProgressBarMorph to: self darkBackgroundHighlights;
		set: #barColor for: #SystemProgressBarMorph to: self darkContentSecondary.
		
	"And the balloon morphs."
	theme
		set: #borderColor for: #NewBalloonMorph to: self darkBackground muchDarker;
		set: #color for: #NewBalloonMorph to: self darkBackgroundHighlights.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 7/26/2016 11:08'!
addDarkFonts: theme

	"Set-up fonts."
	theme
		set: #balloonHelpFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 7);
		set: #standardButtonFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 7);
		set: #standardCodeFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardDefaultTextFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardFlapFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 7 emphasized: TextEmphasis bold emphasisCode);
		set: #haloLabelFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardListFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardMenuFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #standardSystemFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9);
		set: #windowTitleFont to: (StrikeFont familyName: 'Darkmap DejaVu Sans' pointSize: 9).! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/5/2016 08:41'!
addDarkMenusAndDockingBars: theme
	"self createDark apply."
	theme
		set: #borderWidth for: #MenuMorph to: 1;
		set: #borderColor for: #MenuMorph to: self darkBackgroundHighlights;
		set: #color for: #MenuMorph to: self darkBackground;
		set: #titleTextColor for: #MenuMorph to: self darkContentEmphasizedMore;
		set: #lineColor for: #MenuMorph to: self darkBackgroundHighlights;
		set: #lineStyle for: #MenuMorph to: BorderStyle default;
		set: #lineWidth for: #MenuMorph to: 1.
		
	theme
		set: #textColor for: #MenuItemMorph to: self darkContentEmphasized;
		set: #disabledTextColor for: #MenuItemMorph to: self darkContentSecondary;
		set: #selectionColor for: #MenuItemMorph to: self darkBackgroundHighlights;
		set: #selectionTextColor for: #MenuItemMorph to: self darkContentEmphasizedMore.
		"set: #subMenuMarker for: #MenuItemMorph to: nil." "Use hard-coded default. See MenuItemMorph."
		
	"The world main docking bar."
	theme
		set: #color for: #DockingBarMorph to: self darkBackgroundHighlights;
		set: #selectionColor for: #DockingBarItemMorph to: self darkContentSecondary;
		set: #logoColor for: #TheWorldMainDockingBar to: self darkContentEmphasized;
		set: #selectionLogoColor for: #TheWorldMainDockingBar to: self darkContentEmphasizedMore.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 7/26/2016 13:15'!
addDarkScrollables: theme
	"self createDark apply."

	"Scroll bars"
	theme
		set: #thumbColor for: #ScrollBar to: self darkBackground;
		set: #thumbBorderColor for: #ScrollBar to: self darkBackground;
		set: #thumbBorderWidth for: #ScrollBar to: 1;
		set: #thumbColorModifier for: #ScrollBar to: [ [:c | c adjustBrightness: -0.1] ];
		set: #pagingAreaColorModifier for: #ScrollBar to: [ [:c | Color transparent ] ];
		set: #borderColorModifier for: #ScrollBar to: [ [:c | c adjustBrightness: -0.1] ].
	
	"Scroll panes (includes generic stuff for list widgets, tree widgets, and text widgets."
	theme
		set: #borderWidth for: #ScrollPane to: 0;
	"	set: #borderColor for: #ScrollPane to: Color transparent;"
		set: #color for: #ScrollPane to: self darkBackground.
		
	"List widgets"
	theme
		set: #textColor for: #PluggableListMorph to: self darkContentEmphasized;
		set: #selectionColor for: #PluggableListMorph to: self darkBackgroundHighlights;
		set: #selectionTextColor for: #PluggableListMorph to: self darkContentEmphasizedMore;
		set: #filterColor for: #PluggableListMorph to: self yellow;
		set: #filterTextColor for: #PluggableListMorph to: self base03;
		set: #preSelectionModifier for: #PluggableListMorph to: [ [:c | c lighter alpha: 0.5 ] ];
		set: #hoverSelectionModifier for: #PluggableListMorph to: [ [:c | c lighter alpha: 0.5 ] ].
		
	"Tree widgets"
	theme
		set: #highlightTextColor for: #SimpleHierarchicalListMorph to: self yellow lighter lighter;
		set: #lineColor for: #SimpleHierarchicalListMorph to: self darkContentSecondary.
	
	"Text widgets"
	theme
		set: #textColor for: #PluggableTextMorph to: self darkContentPrimary;
		set: #caretColor for: #PluggableTextMorph to: self darkContentEmphasizedMore;
		set: #selectionColor for: #PluggableTextMorph to: self darkBackgroundHighlights;
		set: #unfocusedSelectionModifier for: #PluggableTextMorph to: [ [:c | (Color r: 0.027 g: 0.211 b: 0.258) "dark background highlights"] ];
		set: #adornmentReadOnly for: #PluggableTextMorph to: self magenta;
		set: #adornmentRefuse for: #PluggableTextMorph to: self cyan;
		set: #adornmentConflict for: #PluggableTextMorph to: self red;
		set: #adornmentDiff for: #PluggableTextMorph to: self green;
		set: #adornmentNormalEdit for: #PluggableTextMorph to: self orange;
		set: #adornmentDiffEdit for: #PluggableTextMorph to: self yellow.
	theme
		set: #balloonTextColor for: #PluggableTextMorphPlus to: self darkContentSecondary.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/11/2016 16:38'!
addDarkSyntaxHighlighting: theme
	"self createDark apply."
	theme
		set: #color for: #TextAction to: self blue;
		
		set: #default for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #invalid for: #SHTextStylerST80 to: {self red};
		set: #excessCode for: #SHTextStylerST80 to: {self red};
		set: #comment for: #SHTextStylerST80 to: {self cyan};
		set: #unfinishedComment for: #SHTextStylerST80 to: {self red. TextEmphasis italic};
		set: #'$' for: #SHTextStylerST80 to: {self red};
		set: #character for: #SHTextStylerST80 to: {self red};
		set: #integer for: #SHTextStylerST80 to: {self red};
		set: #number for: #SHTextStylerST80 to: {self red};	
		set: #- for: #SHTextStylerST80 to: {self red};
		set: #symbol for: #SHTextStylerST80 to: {self blue};	
		set: #stringSymbol for: #SHTextStylerST80 to: {self blue};	
		set: #literalArray for: #SHTextStylerST80 to: {self blue};
		set: #string for: #SHTextStylerST80 to: {self magenta. TextEmphasis normal};
		set: #unfinishedString for: #SHTextStylerST80 to: {self red. TextEmphasis normal};
		set: #assignment for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #ansiAssignment for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #literal for: #SHTextStylerST80 to: {nil. TextEmphasis italic};
		set: #keyword for: #SHTextStylerST80 to: {self blue};
		set: #binary for: #SHTextStylerST80 to: {self blue};	
		set: #unary for: #SHTextStylerST80 to: {self blue};
		set: #incompleteKeyword for: #SHTextStylerST80 to: {self darkContentPrimary. TextEmphasis underlined};
		set: #incompleteBinary for: #SHTextStylerST80 to: {self darkContentPrimary. TextEmphasis underlined};	
		set: #incompleteUnary for: #SHTextStylerST80 to: {self darkContentPrimary. TextEmphasis underlined};
		set: #undefinedKeyword for: #SHTextStylerST80 to: {self red};
		set: #undefinedBinary for: #SHTextStylerST80 to: {self red};	
		set: #undefinedUnary for: #SHTextStylerST80 to: {self red};													
		set: #patternKeyword for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #patternBinary for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #patternUnary for: #SHTextStylerST80 to: {nil. TextEmphasis bold};	
		set: #self for: #SHTextStylerST80 to: {self red};
		set: #super for: #SHTextStylerST80 to: {self red}; 
		set: #true for: #SHTextStylerST80 to: {self red};
		set: #false for: #SHTextStylerST80 to: {self red};
		set: #nil for: #SHTextStylerST80 to: {self red};
		set: #thisContext for: #SHTextStylerST80 to: {self red};
		set: #return for: #SHTextStylerST80 to: {self red};
		set: #patternArg for: #SHTextStylerST80 to: {self blue};	
		set: #methodArg for: #SHTextStylerST80 to: {self blue};
		set: #blockPatternArg for: #SHTextStylerST80 to: {self blue};
		set: #blockArg for: #SHTextStylerST80 to: {self blue};
		set: #argument for: #SHTextStylerST80 to: {self blue};
		set: #blockArgColon for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #leftParenthesis for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #rightParenthesis for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #leftParenthesis1 for: #SHTextStylerST80 to: {self green}; 
		set: #rightParenthesis1 for: #SHTextStylerST80 to: {self green}; 
		set: #leftParenthesis2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #rightParenthesis2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #leftParenthesis3 for: #SHTextStylerST80 to: {self red}; 
		set: #rightParenthesis3 for: #SHTextStylerST80 to: {self red}; 
		set: #leftParenthesis4 for: #SHTextStylerST80 to: {self green}; 
		set: #rightParenthesis4 for: #SHTextStylerST80 to: {self green}; 
		set: #leftParenthesis5 for: #SHTextStylerST80 to: {self orange}; 
		set: #rightParenthesis5 for: #SHTextStylerST80 to: {self orange}; 
		set: #leftParenthesis6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #rightParenthesis6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #leftParenthesis7 for: #SHTextStylerST80 to: {self blue}; 
		set: #rightParenthesis7 for: #SHTextStylerST80 to: {self blue}; 
		set: #blockStart for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #blockEnd for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #blockStart1 for: #SHTextStylerST80 to: {self green}; 
		set: #blockEnd1 for: #SHTextStylerST80 to: {self green}; 
		set: #blockStart2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockEnd2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockStart3 for: #SHTextStylerST80 to: {self red}; 
		set: #blockEnd3 for: #SHTextStylerST80 to: {self red}; 
		set: #blockStart4 for: #SHTextStylerST80 to: {self green}; 
		set: #blockEnd4 for: #SHTextStylerST80 to: {self green}; 
		set: #blockStart5 for: #SHTextStylerST80 to: {self orange}; 
		set: #blockEnd5 for: #SHTextStylerST80 to: {self orange}; 
		set: #blockStart6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockEnd6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockStart7 for: #SHTextStylerST80 to: {self blue}; 
		set: #blockEnd7 for: #SHTextStylerST80 to: {self blue}; 																																																		
		set: #arrayStart for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #arrayEnd for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #arrayStart1 for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #arrayEnd1 for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #byteArrayStart for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #byteArrayEnd for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #byteArrayStart1 for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #byteArrayEnd1 for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #leftBrace for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #rightBrace for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #cascadeSeparator for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #statementSeparator for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #externalCallType for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #externalCallTypePointerIndicator for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #primitiveOrExternalCallStart for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #primitiveOrExternalCallEnd for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #methodTempBar for: #SHTextStylerST80 to: {self darkContentPrimary}; 
		set: #blockTempBar for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #blockArgsBar for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #primitive for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #pragmaKeyword for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #pragmaUnary for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #pragmaBinary for: #SHTextStylerST80 to: {self green. TextEmphasis bold};									
		set: #externalFunctionCallingConvention for: #SHTextStylerST80 to: {self green. TextEmphasis bold}; 
		set: #module for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #blockTempVar for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #blockPatternTempVar for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #instVar for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #workspaceVar for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #undefinedIdentifier for: #SHTextStylerST80 to: {self red};
		set: #incompleteIdentifier for: #SHTextStylerST80 to: {self darkContentPrimary. {TextEmphasis italic. TextEmphasis underlined}};
		set: #tempVar for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #patternTempVar for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #poolConstant for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #classVar for: #SHTextStylerST80 to: {self darkContentPrimary};
		set: #globalVar for: #SHTextStylerST80 to: {self darkContentPrimary}.
		
	"And the text differ"
	theme	
		set: #insertTextAttributes for: #TextDiffBuilder to: { TextColor color: self red };
		set: #removeTextAttributes for: #TextDiffBuilder to: { TextEmphasis struckOut. TextColor color: self blue };
		set: #normalTextAttributes for: #TextDiffBuilder to: { TextEmphasis normal }.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 14:10'!
addDarkToolColors: theme
	"Tool-specific colors."
	
	theme 
		set: #failureColor for: #TestRunner to: self yellow;
		set: #errorColor for: #TestRunner to: self red;
		set: #passColor for: #TestRunner to: self green;
		
		set: #failureTextColor for: #TestRunner to: self darkBackground;
		set: #errorTextColor for: #TestRunner to: self darkBackground;
		set: #passTextColor for: #TestRunner to: self darkBackground.
		
	theme
		set: #ignoredOperationAttributes for: #MCOperationsBrowser to: {TextColor color: self darkContentSecondary}.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 11:28'!
addDarkWindowColors: theme
	"doIt: [self createDark apply.]"

	theme
		set: #uniformWindowColor for: #Model to: (self darkBackgroundHighlights adjustBrightness: 0.16 "lighter twice");

		set: #unfocusedWindowColorModifier for: #SystemWindow to: [ [:color | color adjustBrightness: -0.16 "darker twice"] ];
		set: #unfocusedLabelColor for: #SystemWindow to: self darkContentEmphasized;
		set: #focusedLabelColor for: #SystemWindow to: self darkContentEmphasizedMore;

		set: #customWindowColor for: #Browser to: self green;
		set: #customWindowColor for: #ChangeList to: self blue;
		set: #customWindowColor for: #ChangeSorter to: self blue;
		set: #customWindowColor for: #ChatNotes to: self magenta;
		set: #customWindowColor for: #ClassCommentVersionsBrowser to: self violet;
		set: #customWindowColor for: #Debugger to: self red;
		set: #customWindowColor for: #DualChangeSorter to: self blue;
		set: #customWindowColor for: #FileContentsBrowser to: self yellow;
		set: #customWindowColor for: #FileList to: self yellow;
		set: #customWindowColor for: #InstanceBrowser to: self cyan;
		set: #customWindowColor for: #Lexicon to: self cyan;
		set: #customWindowColor for: #MCTool to: self violet;
		set: #customWindowColor for: #MessageNames to: self green;
		set: #customWindowColor for: #MessageSet to: self cyan;
		set: #customWindowColor for: #PackagePaneBrowser to: self green;
		set: #customWindowColor for: #PluggableFileList to: self yellow;
		set: #customWindowColor for: #PreferenceBrowser to: self cyan;
		set: #customWindowColor for: #SMLoader to: self orange;
		set: #customWindowColor for: #SMLoaderPlus to: self orange;
		set: #customWindowColor for: #SMReleaseBrowser to: self orange;
		set: #customWindowColor for: #ScriptingDomain to: self yellow;
		set: #customWindowColor for: #SelectorBrowser to: self cyan;
		set: #customWindowColor for: #StringHolder to: self yellow;
		set: #customWindowColor for: #TestRunner to: self orange;
		set: #customWindowColor for: #TranscriptStream to: self orange;
		set: #customWindowColor for: #VersionsBrowser to: self violet.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/15/2016 10:56'!
addLightButtons: theme
	"self createLight apply."
	theme
		set: #borderColor for: #PluggableButtonMorph to: self lightContentSecondary;
		set: #color for: #PluggableButtonMorph to: self lightBackgroundHighlights;
		set: #textColor for: #PluggableButtonMorph to: self lightContentEmphasizedMore;

		set: #selectionModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.1] ];
		set: #hoverModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.1] ];
		set: #feedbackModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.3] ].
	
	"And the plus-version."
	theme set: #disabledTextColor for: #PluggableButtonMorphPlus to: self lightContentSecondary.
! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/3/2016 16:58'!
addLightDialogs: theme
	"self createLight apply."
	theme
		set: #borderColor for: #DialogWindow to: self lightBackground muchDarker;
		set: #color for: #DialogWindow to: self lightBackgroundHighlights;

		set: #titleColor for: #DialogWindow to: self lightBackground;
		set: #titleTextColor for: #DialogWindow to: self lightContentEmphasized;
		
		set: #textColor for: #DialogWindow to: self lightContentEmphasizedMore;
		
		set: #okColor for: #DialogWindow to: self green;
		set: #cancelColor for: #DialogWindow to: self orange;
		set: #buttonColor for: #DialogWindow to: self lightBackgroundHighlights;
		set: #selectionModifier for: #DialogWindow to: [ [:c | c adjustBrightness: -0.1 ] ].
	
	"The List Chooser is a dialog, too."
	theme
		set: #addColor for: #ListChooser to: self blue;
		set: #disabledColor for: #ListChooser to: Color transparent.

		
	"And the system progress bar."
	theme
		set: #color for: #SystemProgressBarMorph to: self lightBackgroundHighlights;
		set: #barColor for: #SystemProgressBarMorph to: self lightContentSecondary.
		
	"And the balloon morphs."
	theme
		set: #borderColor for: #NewBalloonMorph to: self lightBackground muchDarker;
		set: #color for: #NewBalloonMorph to: self lightBackgroundHighlights.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 7/25/2016 17:31'!
addLightFonts: theme

	"Set-up fonts."
	theme
		set: #balloonHelpFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 7);
		set: #standardButtonFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 7);
		set: #standardCodeFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardDefaultTextFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardFlapFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 7 emphasized: TextEmphasis bold emphasisCode);
		set: #haloLabelFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardListFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardMenuFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardSystemFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #windowTitleFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9 emphasized: TextEmphasis bold emphasisCode).! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/5/2016 08:40'!
addLightMenusAndDockingBars: theme
	"self createLight apply."
	theme
		set: #borderWidth for: #MenuMorph to: 1;
		set: #borderColor for: #MenuMorph to: self lightBackgroundHighlights;
		set: #color for: #MenuMorph to: self lightBackground;
		set: #titleTextColor for: #MenuMorph to: self lightContentEmphasizedMore;
		set: #lineColor for: #MenuMorph to: self lightBackgroundHighlights;
		set: #lineStyle for: #MenuMorph to: BorderStyle default;
		set: #lineWidth for: #MenuMorph to: 1.
		
	theme
		set: #textColor for: #MenuItemMorph to: self lightContentEmphasized;
		set: #disabledTextColor for: #MenuItemMorph to: self lightContentSecondary;
		set: #selectionColor for: #MenuItemMorph to: self lightBackgroundHighlights;
		set: #selectionTextColor for: #MenuItemMorph to: self lightContentEmphasizedMore.
		"set: #subMenuMarker for: #MenuItemMorph to: nil." "Use hard-coded default. See MenuItemMorph."
		
	"The world main docking bar."
	theme
		set: #color for: #DockingBarMorph to: self lightBackgroundHighlights;
		set: #selectionColor for: #DockingBarItemMorph to: self lightContentSecondary;
		set: #logoColor for: #TheWorldMainDockingBar to: self lightContentEmphasized;
		set: #selectionLogoColor for: #TheWorldMainDockingBar to: self lightContentEmphasizedMore.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/4/2016 13:28'!
addLightScrollables: theme
	"self createLight apply."

	"Scroll bars"
	theme
		set: #thumbColor for: #ScrollBar to: self lightBackground;
		set: #thumbBorderColor for: #ScrollBar to: self lightBackground;
		set: #thumbBorderWidth for: #ScrollBar to: 1;
		set: #thumbColorModifier for: #ScrollBar to: [ [:c | c adjustBrightness: -0.1] ];
		set: #pagingAreaColorModifier for: #ScrollBar to: [ [:c | Color transparent ] ];
		set: #borderColorModifier for: #ScrollBar to: [ [:c | c adjustBrightness: -0.1] ].
	
	"Scroll panes (includes generic stuff for list widgets, tree widgets, and text widgets."
	theme
		set: #borderWidth for: #ScrollPane to: 0;
	"	set: #borderColor for: #ScrollPane to: Color transparent;"
		set: #color for: #ScrollPane to: self lightBackground.
		
	"List widgets"
	theme
		set: #textColor for: #PluggableListMorph to: self lightContentEmphasized;
		set: #selectionColor for: #PluggableListMorph to: self lightBackgroundHighlights;
		set: #selectionTextColor for: #PluggableListMorph to: self lightContentEmphasizedMore;
		derive: #multiSelectionColor for: #PluggableListMorph from: #PluggableListMorph at: #selectionColor do: [:c | c darker];
		set: #filterColor for: #PluggableListMorph to: self yellow;
		set: #filterTextColor for: #PluggableListMorph to: self lightBackground;
		set: #preSelectionModifier for: #PluggableListMorph to: [ [:c | c darker alpha: 0.5 ] ];
		set: #hoverSelectionModifier for: #PluggableListMorph to: [ [:c | c darker alpha: 0.5 ] ].
		
	"Tree widgets"
	theme
		set: #highlightTextColor for: #SimpleHierarchicalListMorph to: self yellow lighter lighter;
		set: #lineColor for: #SimpleHierarchicalListMorph to: self lightContentSecondary.
	
	"Text widgets"
	theme
		set: #textColor for: #PluggableTextMorph to: self lightContentPrimary;
		set: #caretColor for: #PluggableTextMorph to: self lightContentEmphasizedMore;
		set: #selectionColor for: #PluggableTextMorph to: self lightBackgroundHighlights;
		set: #unfocusedSelectionModifier for: #PluggableTextMorph to: [ [:c | (Color r: 0.933 g: 0.909 b: 0.835) "light background highlights"] ];
		set: #adornmentReadOnly for: #PluggableTextMorph to: self magenta;
		set: #adornmentRefuse for: #PluggableTextMorph to: self cyan;
		set: #adornmentConflict for: #PluggableTextMorph to: self red;
		set: #adornmentDiff for: #PluggableTextMorph to: self green;
		set: #adornmentNormalEdit for: #PluggableTextMorph to: self orange;
		set: #adornmentDiffEdit for: #PluggableTextMorph to: self yellow.
	theme
		set: #balloonTextColor for: #PluggableTextMorphPlus to: self lightContentSecondary.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/11/2016 16:38'!
addLightSyntaxHighlighting: theme
	"self createLight apply."
	theme
		set: #color for: #TextAction to: self blue;

		set: #default for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #invalid for: #SHTextStylerST80 to: {self red};
		set: #excessCode for: #SHTextStylerST80 to: {self red};
		set: #comment for: #SHTextStylerST80 to: {self cyan};
		set: #unfinishedComment for: #SHTextStylerST80 to: {self red. TextEmphasis italic};
		set: #'$' for: #SHTextStylerST80 to: {self red};
		set: #character for: #SHTextStylerST80 to: {self red};
		set: #integer for: #SHTextStylerST80 to: {self red};
		set: #number for: #SHTextStylerST80 to: {self red};	
		set: #- for: #SHTextStylerST80 to: {self red};
		set: #symbol for: #SHTextStylerST80 to: {self blue};	
		set: #stringSymbol for: #SHTextStylerST80 to: {self blue};	
		set: #literalArray for: #SHTextStylerST80 to: {self blue};
		set: #string for: #SHTextStylerST80 to: {self magenta. TextEmphasis normal};
		set: #unfinishedString for: #SHTextStylerST80 to: {self red. TextEmphasis normal};
		set: #assignment for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #ansiAssignment for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #literal for: #SHTextStylerST80 to: {nil. TextEmphasis italic};
		set: #keyword for: #SHTextStylerST80 to: {self blue};
		set: #binary for: #SHTextStylerST80 to: {self blue};	
		set: #unary for: #SHTextStylerST80 to: {self blue};
		set: #incompleteKeyword for: #SHTextStylerST80 to: {self lightContentPrimary. TextEmphasis underlined};
		set: #incompleteBinary for: #SHTextStylerST80 to: {self lightContentPrimary. TextEmphasis underlined};	
		set: #incompleteUnary for: #SHTextStylerST80 to: {self lightContentPrimary. TextEmphasis underlined};
		set: #undefinedKeyword for: #SHTextStylerST80 to: {self red};
		set: #undefinedBinary for: #SHTextStylerST80 to: {self red};	
		set: #undefinedUnary for: #SHTextStylerST80 to: {self red};													
		set: #patternKeyword for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #patternBinary for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #patternUnary for: #SHTextStylerST80 to: {nil. TextEmphasis bold};	
		set: #self for: #SHTextStylerST80 to: {self red};
		set: #super for: #SHTextStylerST80 to: {self red}; 
		set: #true for: #SHTextStylerST80 to: {self red};
		set: #false for: #SHTextStylerST80 to: {self red};
		set: #nil for: #SHTextStylerST80 to: {self red};
		set: #thisContext for: #SHTextStylerST80 to: {self red};
		set: #return for: #SHTextStylerST80 to: {self red};
		set: #patternArg for: #SHTextStylerST80 to: {self blue};	
		set: #methodArg for: #SHTextStylerST80 to: {self blue};
		set: #blockPatternArg for: #SHTextStylerST80 to: {self blue};
		set: #blockArg for: #SHTextStylerST80 to: {self blue};
		set: #argument for: #SHTextStylerST80 to: {self blue};
		set: #blockArgColon for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #leftParenthesis for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #rightParenthesis for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #leftParenthesis1 for: #SHTextStylerST80 to: {self green}; 
		set: #rightParenthesis1 for: #SHTextStylerST80 to: {self green}; 
		set: #leftParenthesis2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #rightParenthesis2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #leftParenthesis3 for: #SHTextStylerST80 to: {self red}; 
		set: #rightParenthesis3 for: #SHTextStylerST80 to: {self red}; 
		set: #leftParenthesis4 for: #SHTextStylerST80 to: {self green}; 
		set: #rightParenthesis4 for: #SHTextStylerST80 to: {self green}; 
		set: #leftParenthesis5 for: #SHTextStylerST80 to: {self orange}; 
		set: #rightParenthesis5 for: #SHTextStylerST80 to: {self orange}; 
		set: #leftParenthesis6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #rightParenthesis6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #leftParenthesis7 for: #SHTextStylerST80 to: {self blue}; 
		set: #rightParenthesis7 for: #SHTextStylerST80 to: {self blue}; 
		set: #blockStart for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #blockEnd for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #blockStart1 for: #SHTextStylerST80 to: {self green}; 
		set: #blockEnd1 for: #SHTextStylerST80 to: {self green}; 
		set: #blockStart2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockEnd2 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockStart3 for: #SHTextStylerST80 to: {self red}; 
		set: #blockEnd3 for: #SHTextStylerST80 to: {self red}; 
		set: #blockStart4 for: #SHTextStylerST80 to: {self green}; 
		set: #blockEnd4 for: #SHTextStylerST80 to: {self green}; 
		set: #blockStart5 for: #SHTextStylerST80 to: {self orange}; 
		set: #blockEnd5 for: #SHTextStylerST80 to: {self orange}; 
		set: #blockStart6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockEnd6 for: #SHTextStylerST80 to: {self magenta}; 
		set: #blockStart7 for: #SHTextStylerST80 to: {self blue}; 
		set: #blockEnd7 for: #SHTextStylerST80 to: {self blue}; 																																																		
		set: #arrayStart for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #arrayEnd for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #arrayStart1 for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #arrayEnd1 for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #byteArrayStart for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #byteArrayEnd for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #byteArrayStart1 for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #byteArrayEnd1 for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #leftBrace for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #rightBrace for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #cascadeSeparator for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #statementSeparator for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #externalCallType for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #externalCallTypePointerIndicator for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #primitiveOrExternalCallStart for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #primitiveOrExternalCallEnd for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #methodTempBar for: #SHTextStylerST80 to: {self lightContentPrimary}; 
		set: #blockTempBar for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #blockArgsBar for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #primitive for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #pragmaKeyword for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #pragmaUnary for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #pragmaBinary for: #SHTextStylerST80 to: {self green. TextEmphasis bold};									
		set: #externalFunctionCallingConvention for: #SHTextStylerST80 to: {self green. TextEmphasis bold}; 
		set: #module for: #SHTextStylerST80 to: {self green. TextEmphasis bold};
		set: #blockTempVar for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #blockPatternTempVar for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #instVar for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #workspaceVar for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #undefinedIdentifier for: #SHTextStylerST80 to: {self red};
		set: #incompleteIdentifier for: #SHTextStylerST80 to: {self lightContentPrimary. {TextEmphasis italic. TextEmphasis underlined}};
		set: #tempVar for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #patternTempVar for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #poolConstant for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #classVar for: #SHTextStylerST80 to: {self lightContentPrimary};
		set: #globalVar for: #SHTextStylerST80 to: {self lightContentPrimary}.
		
	"And the text differ"
	theme	
		set: #insertTextAttributes for: #TextDiffBuilder to: { TextColor color: self red };
		set: #removeTextAttributes for: #TextDiffBuilder to: { TextEmphasis struckOut. TextColor color: self blue };
		set: #normalTextAttributes for: #TextDiffBuilder to: { TextEmphasis normal }.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 14:10'!
addLightToolColors: theme
	"Tool-specific colors."
	
	theme 
		set: #failureColor for: #TestRunner to: self yellow;
		set: #errorColor for: #TestRunner to: self red;
		set: #passColor for: #TestRunner to: self green;

		set: #failureTextColor for: #TestRunner to: self lightBackground;
		set: #errorTextColor for: #TestRunner to: self lightBackground;
		set: #passTextColor for: #TestRunner to: self lightBackground.
		
	theme
		set: #ignoredOperationAttributes for: #MCOperationsBrowser to: {TextColor color: self lightContentSecondary}.! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 11:30'!
addLightWindowColors: theme
	"You have to create dark first.
	doIt: [self createDark. self createLight apply.]"

	theme
		set: #uniformWindowColor for: #Model to: (self lightBackgroundHighlights adjustBrightness: -0.16 "darker twice");

		set: #unfocusedWindowColorModifier for: #SystemWindow to: [ [:color | color adjustBrightness: 0.16 "lighter twice"] ];
		set: #unfocusedLabelColor for: #SystemWindow to: self lightContentEmphasized;
		set: #focusedLabelColor for: #SystemWindow to: self lightContentEmphasizedMore;

		set: #customWindowColor for: #Browser to: self green;
		set: #customWindowColor for: #ChangeList to: self blue;
		set: #customWindowColor for: #ChangeSorter to: self blue;
		set: #customWindowColor for: #ChatNotes to: self magenta;
		set: #customWindowColor for: #ClassCommentVersionsBrowser to: self violet;
		set: #customWindowColor for: #Debugger to: self red;
		set: #customWindowColor for: #DualChangeSorter to: self blue;
		set: #customWindowColor for: #FileContentsBrowser to: self yellow;
		set: #customWindowColor for: #FileList to: self yellow;
		set: #customWindowColor for: #InstanceBrowser to: self cyan;
		set: #customWindowColor for: #Lexicon to: self cyan;
		set: #customWindowColor for: #MCTool to: self violet;
		set: #customWindowColor for: #MessageNames to: self green;
		set: #customWindowColor for: #MessageSet to: self cyan;
		set: #customWindowColor for: #PackagePaneBrowser to: self green;
		set: #customWindowColor for: #PluggableFileList to: self yellow;
		set: #customWindowColor for: #PreferenceBrowser to: self cyan;
		set: #customWindowColor for: #SMLoader to: self orange;
		set: #customWindowColor for: #SMLoaderPlus to: self orange;
		set: #customWindowColor for: #SMReleaseBrowser to: self orange;
		set: #customWindowColor for: #ScriptingDomain to: self yellow;
		set: #customWindowColor for: #SelectorBrowser to: self cyan;
		set: #customWindowColor for: #StringHolder to: self yellow;
		set: #customWindowColor for: #TestRunner to: self orange;
		set: #customWindowColor for: #TranscriptStream to: self orange;
		set: #customWindowColor for: #VersionsBrowser to: self violet.! !

!SolarizedTheme class methodsFor: 'colors by name' stamp: 'mt 7/25/2016 17:00'!
base0

	^ Color fromString: '#839496'! !

!SolarizedTheme class methodsFor: 'colors by name' stamp: 'mt 7/25/2016 17:00'!
base00

	^ Color fromString: '#657b83'! !

!SolarizedTheme class methodsFor: 'colors by name' stamp: 'mt 7/25/2016 17:00'!
base01

	^ Color fromString: '#586e75'! !

!SolarizedTheme class methodsFor: 'colors by name' stamp: 'mt 7/25/2016 17:01'!
base02

	^ Color fromString: '#073642'! !

!SolarizedTheme class methodsFor: 'colors by name' stamp: 'mt 7/25/2016 17:01'!
base03

	^ Color fromString: '#002b36'! !

!SolarizedTheme class methodsFor: 'colors by name' stamp: 'mt 7/25/2016 17:01'!
base1

	^ Color fromString: '#93a1a1'! !

!SolarizedTheme class methodsFor: 'colors by name' stamp: 'mt 7/25/2016 17:01'!
base2

	^ Color fromString: '#eee8d5'! !

!SolarizedTheme class methodsFor: 'colors by name' stamp: 'mt 7/25/2016 17:01'!
base3

	^ Color fromString: '#fdf6e3'! !

!SolarizedTheme class methodsFor: 'colors for accent' stamp: 'mt 7/25/2016 17:01'!
blue

	^ Color fromString: '#268bd2'! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 18:04'!
createDark
	"doIt: [self createDark apply.]"

	| themeName |
	themeName := 'Solarized (dark)'.
	^ (self named: themeName) in: [:theme |
		theme merge: (self named: 'Squeak') overwrite: true.
		theme name: themeName.

		"General morph stuff."
		theme
			set: #keyboardFocusColor for: #Morph to: self darkContentSecondary;
			set: #keyboardFocusWidth for: #Morph to: 1.

		theme set: #background for: #MorphicProject to: self lightBackgroundForm.

		self
			addDarkFonts: theme;
			addDarkWindowColors: theme;
			addDarkSyntaxHighlighting: theme;
			addDarkScrollables: theme;
			addDarkButtons: theme;
			addDarkDialogs: theme;
			addDarkMenusAndDockingBars: theme;
			addDarkToolColors: theme.

		theme]! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 18:04'!
createLight
	"You have to create dark first.
	doIt: [self createDark. self createLight apply.]"

	| themeName |
	themeName := 'Solarized (light)'.
	^ (self named: themeName) in: [:theme |
		theme merge: (self named: 'Solarized (dark)') overwrite: true.
		theme name: themeName.

		"General morph stuff."
		theme
			set: #keyboardFocusColor for: #Morph to: self lightContentSecondary;
			set: #keyboardFocusWidth for: #Morph to: 1.

		theme set: #background for: #MorphicProject to: self darkBackgroundForm.

		self
			addLightFonts: theme;
			addLightWindowColors: theme;
			addLightSyntaxHighlighting: theme;
			addLightScrollables: theme;
			addLightButtons: theme;
			addLightDialogs: theme;
			addLightMenusAndDockingBars: theme;
			addLightToolColors: theme.

		theme]! !

!SolarizedTheme class methodsFor: 'colors for accent' stamp: 'mt 7/25/2016 17:02'!
cyan

	^ Color fromString: '#2aa198'! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:09'!
darkBackground
	^ self base03! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/9/2016 10:00'!
darkBackgroundForm

	| ref |
	ref := self darkBackground.

	^ (SqueakTheme linenblue  asFormOfDepth: 32) collectColors: [:c | 
		Color
			h:ref hue
			s: ref saturation
			v: c brightness
			alpha: c alpha]! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:09'!
darkBackgroundHighlights

	^ self base02! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:09'!
darkContentEmphasized

	^ self base1! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:09'!
darkContentEmphasizedMore

	^ self base2! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:10'!
darkContentPrimary
	"body text, default code, ..."

	^ self base0! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:10'!
darkContentSecondary
	"comments, keyboard focus indication, ..."

	^ self base01! !

!SolarizedTheme class methodsFor: 'colors for accent' stamp: 'mt 7/25/2016 17:02'!
green

	^ Color fromString: '#859900'! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:10'!
lightBackground

	^ self base3! !

!SolarizedTheme class methodsFor: 'instance creation' stamp: 'mt 8/15/2016 10:54'!
lightBackgroundForm

	| ref |
	ref := self lightBackground.

	^ (SqueakTheme linenblue  asFormOfDepth: 32) collectColors: [:c | 
		Color
			h:ref hue
			s: ref saturation
			v: 0.7 - c brightness
			alpha: c alpha]! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:10'!
lightBackgroundHighlights

	^ self base2! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:10'!
lightContentEmphasized

	^ self base01! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:10'!
lightContentEmphasizedMore

	^ self base02! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:10'!
lightContentPrimary
	"body text, default code, ..."

	^ self base00! !

!SolarizedTheme class methodsFor: 'colors by purpose' stamp: 'mt 7/25/2016 17:10'!
lightContentSecondary
	"comments, ..."

	^ self base1! !

!SolarizedTheme class methodsFor: 'colors for accent' stamp: 'mt 7/25/2016 17:02'!
magenta

	^ Color fromString: '#d33682'! !

!SolarizedTheme class methodsFor: 'colors for accent' stamp: 'mt 7/25/2016 17:02'!
orange

	^ Color fromString: '#cb4b16'! !

!SolarizedTheme class methodsFor: 'colors for accent' stamp: 'mt 7/25/2016 17:02'!
red

	^ Color fromString: '#dc322f'! !

!SolarizedTheme class methodsFor: 'colors for accent' stamp: 'mt 7/25/2016 17:02'!
violet

	^ Color fromString: '#6c71c4'! !

!SolarizedTheme class methodsFor: 'colors for accent' stamp: 'mt 7/25/2016 17:02'!
yellow

	^ Color fromString: '#b58900'! !

UserInterfaceTheme subclass: #SqueakTheme
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/14/2016 17:40'!
addButtons: theme

	theme
		set: #borderColor for: #PluggableButtonMorph to: Color gray;
		set: #borderWidth for: #PluggableButtonMorph to: 1;
		set: #borderStyle for: #PluggableButtonMorph to: BorderStyle default;
		set: #color for: #PluggableButtonMorph to: (Color gray: 0.91);
		
		set: #font for: #PluggableButtonMorph to: [Preferences standardButtonFont];
		set: #textColor for: #PluggableButtonMorph to: Color black;

		set: #selectionModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.2] ];
		set: #hoverModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.1] ];
		set: #feedbackModifier for: #PluggableButtonMorph to: [ [:c | c adjustBrightness: -0.3] ].
	
	"And the plus-version."
	theme
		set: #disabledColor for: #PluggableButtonMorphPlus to: Color transparent;
		set: #disabledTextColor for: #PluggableButtonMorphPlus to: (Color gray: 0.6).

	"And the three-phase button."
	theme
		derive: #color for: #ThreePhaseButtonMorph from: #PluggableButtonMorph at: #textColor;
		derive: #font for: #ThreePhaseButtonMorph from: #PluggableButtonMorph;
		derive: #textColor for: #ThreePhaseButtonMorph from: #PluggableButtonMorph.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 7/25/2016 16:52'!
addClassicSyntaxHighlighting: theme
	"This was the former default highlighting. Just overwrite the different values."
		
	theme	
		set: #comment for: #SHTextStylerST80 to: {Color green muchDarker. TextEmphasis italic};
		set: #symbol for: #SHTextStylerST80 to: {Color blue muchDarker. TextEmphasis bold};	
		set: #stringSymbol for: #SHTextStylerST80 to: {Color blue muchDarker. TextEmphasis bold};	
		set: #literalArray for: #SHTextStylerST80 to: {Color blue muchDarker. TextEmphasis bold};

		set: #self for: #SHTextStylerST80 to: {Color red muchDarker. TextEmphasis bold};
		set: #super for: #SHTextStylerST80 to: {Color red muchDarker. TextEmphasis bold}; 
		set: #true for: #SHTextStylerST80 to: {Color red muchDarker. TextEmphasis bold};
		set: #false for: #SHTextStylerST80 to: {Color red muchDarker. TextEmphasis bold};
		set: #nil for: #SHTextStylerST80 to: {Color red muchDarker. TextEmphasis bold};

		set: #thisContext for: #SHTextStylerST80 to: {Color red muchDarker. TextEmphasis bold};
		set: #return for: #SHTextStylerST80 to: {Color red muchDarker. TextEmphasis bold};
		set: #patternArg for: #SHTextStylerST80 to: {Color blue muchDarker. TextEmphasis italic};	
		set: #methodArg for: #SHTextStylerST80 to: {Color blue muchDarker. TextEmphasis italic};
		set: #blockPatternArg for: #SHTextStylerST80 to: {Color blue muchDarker. TextEmphasis italic};
		set: #blockArg for: #SHTextStylerST80 to: {Color blue muchDarker. TextEmphasis italic};
		set: #argument for: #SHTextStylerST80 to: {Color blue muchDarker. TextEmphasis italic};

		set: #primitiveOrExternalCallStart for: #SHTextStylerST80 to: {Color black. TextEmphasis bold}; 
		set: #primitiveOrExternalCallEnd for: #SHTextStylerST80 to: {Color black. TextEmphasis bold};

		set: #blockTempVar for: #SHTextStylerST80 to: {Color gray. TextEmphasis italic};
		set: #blockPatternTempVar for: #SHTextStylerST80 to: {Color gray. TextEmphasis italic};
		set: #instVar for: #SHTextStylerST80 to: {Color black. TextEmphasis bold};
		set: #workspaceVar for: #SHTextStylerST80 to: {Color black. TextEmphasis bold};
		set: #undefinedIdentifier for: #SHTextStylerST80 to: {Color red. TextEmphasis bold};

		set: #tempVar for: #SHTextStylerST80 to: {Color gray darker. TextEmphasis italic};
		set: #patternTempVar for: #SHTextStylerST80 to: {Color gray darker. TextEmphasis italic};
		set: #poolConstant for: #SHTextStylerST80 to: {Color gray muchDarker. TextEmphasis italic};
		set: #classVar for: #SHTextStylerST80 to: {Color gray muchDarker. TextEmphasis bold};
		set: #globalVar for: #SHTextStylerST80 to: {Color black. TextEmphasis bold}.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/5/2016 09:19'!
addDialogs: theme
	"self create apply."
	
	theme
		set: #borderColor for: #DialogWindow to: Color gray;
		set: #borderWidth for: #DialogWindow to: 1;
		set: #borderStyle for: #DialogWindow to: BorderStyle default;
		set: #color for: #DialogWindow to: Color white;

		set: #titleBorderColor for: #DialogWindow to: Color gray;
		set: #titleBorderWidth for: #DialogWindow to: 0;
		set: #titleBorderStyle for: #DialogWindow to: BorderStyle default;
		set: #titleColor for: #DialogWindow to: (Color r: 0.658 g: 0.678 b: 0.78);
		set: #titleFont for: #DialogWindow to: [Preferences windowTitleFont];
		set: #titleTextColor for: #DialogWindow to: Color black;
		
		set: #font for: #DialogWindow to: [Preferences standardSystemFont];
		set: #textColor for: #DialogWindow to: Color black;
		
		set: #okColor for: #DialogWindow to: (Color r: 0.49 g: 0.749 b: 0.49);
		set: #cancelColor for: #DialogWindow to: (Color r: 1 g: 0.6 b: 0.588);
		set: #buttonColor for: #DialogWindow to: (Color r: 0.658 g: 0.678 b: 0.78) twiceLighter;
		set: #selectionModifier for: #DialogWindow to: [ [:c | Color orange muchLighter ] ].
	
	"The List Chooser is a dialog, too."
	theme
		derive: #okColor for: #ListChooser from: #DialogWindow;
		derive: #cancelColor for: #ListChooser from: #DialogWindow;
		set: #addColor for: #ListChooser to: Color blue muchLighter;
		set: #disabledColor for: #ListChooser to: Color gray.

	"And the mulitple list chooser."
	theme
		derive: #okColor for: #ListMultipleChooser from: #DialogWindow;
		derive: #cancelColor for: #ListMultipleChooser from: #DialogWindow.
		
	"And the system progress bar."
	theme
		derive: #borderColor for: #SystemProgressMorph from: #MenuMorph;
		derive: #borderWidth for: #SystemProgressMorph from: #MenuMorph;
		derive: #borderStyle for: #SystemProgressMorph from: #MenuMorph;
		derive: #color for: #SystemProgressMorph from: #MenuMorph;
		derive: #font for: #SystemProgressMorph from: #MenuItemMorph;
		derive: #textColor for: #SystemProgressMorph from: #MenuItemMorph;
		
		set: #borderColor for: #SystemProgressBarMorph to: Color transparent;
		set: #borderWidth for: #SystemProgressBarMorph to: 0;
		set: #borderStyle for: #SystemProgressBarMorph to: BorderStyle default;
		set: #color for: #SystemProgressBarMorph to: (Color r: 0.977 g: 0.977 b: 0.977);
		set: #barColor for: #SystemProgressBarMorph to: (Color r: 0.72 g: 0.72 b: 0.9).
		
	"And the balloon morphs."
	theme
		set: #borderColor for: #NewBalloonMorph to: (Color r: 0.46 g: 0.46 b: 0.353);
		set: #borderWidth for: #NewBalloonMorph to: 1;
		set: #color for: #NewBalloonMorph to:(Color r: 0.92 g: 0.92 b: 0.706);
		set: #font for: #NewBalloonMorph to: [Preferences standardBalloonHelpFont];
		derive: #textColor for: #NewBalloonMorph from: #PluggableButtonMorph.
		
	theme
		derive: #borderColor for: #BalloonMorph from: #NewBalloonMorph;
		set: #borderWidth for: #BalloonMorph to: 0;
		derive: #color for: #BalloonMorph from: #NewBalloonMorph;
		derive: #font for: #BalloonMorph from: #NewBalloonMorph;
		derive: #textColor for: #BalloonMorph from: #NewBalloonMorph.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/14/2016 17:40'!
addDullerButtons: theme

	theme
		set: #borderColor for: #PluggableButtonMorph to: Color gray duller;
		set: #color for: #PluggableButtonMorph to: (Color gray: 0.91) duller.
			
	"And the plus-version."
	theme
		set: #disabledTextColor for: #PluggableButtonMorphPlus to: (Color gray: 0.6) duller.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/14/2016 17:43'!
addDullerDialogs: theme
	"self createDuller apply."
	
	theme
		set: #titleColor for: #DialogWindow to: (Color r: 0.658 g: 0.678 b: 0.78) duller;
		set: #okColor for: #DialogWindow to: (Color r: 0.49 g: 0.749 b: 0.49) duller;
		set: #cancelColor for: #DialogWindow to: (Color r: 1 g: 0.6 b: 0.588) duller;
		set: #buttonColor for: #DialogWindow to: (Color r: 0.658 g: 0.678 b: 0.78) twiceLighter duller;
		set: #selectionModifier for: #DialogWindow to: [ [:c | Color orange duller ] ].
	
	"The List Chooser is a dialog, too."
	theme
		set: #addColor for: #ListChooser to: Color ocean;
		set: #disabledColor for: #ListChooser to: Color gray duller.
		
	"And the system progress bar."
	theme
		set: #color for: #SystemProgressBarMorph to: (Color r: 0.977 g: 0.977 b: 0.977) duller;
		set: #barColor for: #SystemProgressBarMorph to: (Color r: 0.72 g: 0.72 b: 0.9) duller.
		
	"And the balloon morphs."
	theme
		set: #borderColor for: #NewBalloonMorph to: (Color r: 0.46 g: 0.46 b: 0.353) duller;
		set: #color for: #NewBalloonMorph to:(Color r: 0.92 g: 0.92 b: 0.706) duller.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/14/2016 17:39'!
addDullerMenusAndDockingBars: theme

	theme
		set: #borderColor for: #MenuMorph to: Color gray duller;
		set: #color for: #MenuMorph to: (Color gray: 0.9) duller;	
		set: #lineColor for: #MenuMorph to: (Color gray: 0.9) duller.
		
	theme
		set: #disabledTextColor for: #MenuItemMorph to: Color gray duller;
		set: #selectionColor for: #MenuItemMorph to: (Color r: 0.4 g: 0.5 b: 0.7) duller.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/14/2016 17:27'!
addDullerScrollables: theme
	"self createDuller apply"
	
	theme
		set: #filterColor for: #PluggableListMorph to: Color yellow paler duller.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 12:14'!
addDullerToolColors: theme
	"Tool-specific colors."
	
	theme 
		set: #failureColor for: #TestRunner to: Color yellow duller;
		set: #errorColor for: #TestRunner to: Color red duller;
		set: #passColor for: #TestRunner to: Color green duller.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 7/26/2016 15:47'!
addDullerWindowColors: theme

	theme
		set: #uniformWindowColor for: #Model to: Color veryVeryLightGray duller;
		set: #customWindowColor for: #Browser to: (Color r: 0.764 g: 0.9 b: 0.63) duller;
		set: #customWindowColor for: #ChangeList to: (Color r: 0.719 g: 0.9 b: 0.9) duller;
		set: #customWindowColor for: #ChangeSorter to: (Color r: 0.719 g: 0.9 b: 0.9) duller;
		set: #customWindowColor for: #ChatNotes to: (Color r: 1.0 g: 0.7 b: 0.8) duller;
		set: #customWindowColor for: #ClassCommentVersionsBrowser to: (Color r: 0.753 g: 0.677 b: 0.9) duller;
		set: #customWindowColor for: #Debugger to: (Color r: 0.9 g: 0.719 b: 0.719) duller;
		set: #customWindowColor for: #DualChangeSorter to: (Color r: 0.719 g: 0.9 b: 0.9) duller;
		set: #customWindowColor for: #FileContentsBrowser to: (Color r: 0.7 g: 0.7 b: 0.508) duller;
		set: #customWindowColor for: #FileList to: (Color r: 0.65 g: 0.65 b: 0.65) duller;
		set: #customWindowColor for: #InstanceBrowser to: (Color r: 0.726 g: 0.9 b: 0.9) duller;
		set: #customWindowColor for: #Lexicon to: (Color r: 0.79 g: 0.9 b: 0.79) duller;
		set: #customWindowColor for: #MCTool to: (Color r: 0.65 g: 0.691 b: 0.876) duller;
		set: #customWindowColor for: #MessageNames to: (Color r: 0.639 g: 0.9 b: 0.497) duller;
		set: #customWindowColor for: #MessageSet to: (Color r: 0.719 g: 0.9 b: 0.9) duller;
		set: #customWindowColor for: #PackagePaneBrowser to: (Color r: 0.9 g: 0.9 b: 0.63) duller;
		set: #customWindowColor for: #PluggableFileList to: Color lightYellow duller;
		set: #customWindowColor for: #PreferenceBrowser to: (Color r: 0.671 g: 0.9 b: 0.9) duller;
		set: #customWindowColor for: #SMLoader to: (Color r: 0.801 g: 0.801 b: 0.614) duller;
		set: #customWindowColor for: #SMLoaderPlus to: (Color r: 0.801 g: 0.801 b: 0.614) duller;
		set: #customWindowColor for: #SMReleaseBrowser to: (Color r: 0.801 g: 0.801 b: 0.614) duller;
		set: #customWindowColor for: #ScriptingDomain to: (Color r: 0.91 g: 0.91 b: 0.91) duller;
		set: #customWindowColor for: #SelectorBrowser to: (Color r: 0.45 g: 0.9 b: 0.9) duller;
		set: #customWindowColor for: #StringHolder to: (Color r: 0.9 g: 0.9 b: 0.719) duller;
		set: #customWindowColor for: #TestRunner to: (Color r: 0.9 g: 0.576 b: 0.09) duller;
		set: #customWindowColor for: #TranscriptStream to: (Color r: 0.9 g: 0.75 b: 0.45) duller;
		set: #customWindowColor for: #VersionsBrowser to: (Color r: 0.782 g: 0.677 b: 0.9) duller.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 7/30/2016 07:52'!
addFonts: theme

	theme
		set: #balloonHelpFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 7);
		set: #standardButtonFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 7);
		set: #standardCodeFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardDefaultTextFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardFlapFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 7 emphasized: TextEmphasis bold emphasisCode);
		set: #haloLabelFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardListFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardMenuFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #standardSystemFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9);
		set: #windowTitleFont to: (StrikeFont familyName: 'Bitmap DejaVu Sans' pointSize: 9 emphasized: TextEmphasis bold emphasisCode).
! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/11/2016 19:27'!
addMenusAndDockingBars: theme

	theme
		set: #borderColor for: #MenuMorph to: Color gray;
		set: #borderWidth for: #MenuMorph to: 1;
		set: #borderStyle for: #MenuMorph to: BorderStyle default;
		set: #color for: #MenuMorph to: (Color gray: 0.9);

		set: #titleBorderColor for: #MenuMorph to: (Color r: 0.6 g: 0.7 b: 1);
		set: #titleBorderWidth for: #MenuMorph to: 0;
		set: #titleBorderStyle for: #MenuMorph to: BorderStyle default;
		set: #titleColor for: #MenuMorph to: Color transparent;
		set: #titleFont for: #MenuMorph to: [Preferences windowTitleFont];
		set: #titleTextColor for: #MenuMorph to: Color black;
		
		set: #lineColor for: #MenuMorph to: (Color gray: 0.9);
		set: #lineStyle for: #MenuMorph to: BorderStyle inset;
		set: #lineWidth for: #MenuMorph to: 2.
		
	theme
		set: #font for: #MenuItemMorph to: [Preferences standardMenuFont];
		set: #textColor for: #MenuItemMorph to: Color black;
		set: #disabledTextColor for: #MenuItemMorph to: Color gray;
		set: #selectionColor for: #MenuItemMorph to: (Color r: 0.4 g: 0.5 b: 0.7);
		set: #selectionTextColor for: #MenuItemMorph to: Color white.
		
	"Derive some stuff for the docking bar morph, which looks mostly like a menu morph."
	theme
		set: #borderWidth for: #DockingBarMorph to: 0;
		derive: #borderColor for: #DockingBarMorph from: #MenuMorph;
		derive: #borderStyle for: #DockingBarMorph from: #MenuMorph;
		derive: #color for: #DockingBarMorph from: #MenuMorph;
	
		derive: #lineColor for: #DockingBarMorph from: #MenuMorph;
		derive: #lineStyle for: #DockingBarMorph from: #MenuMorph;
		derive: #lineWidth for: #DockingBarMorph from: #MenuMorph.
		
	"The world main docking bar."
	theme
		derive: #font for: #TheWorldMainDockingBar from: #MenuItemMorph;
		derive: #textColor for: #TheWorldMainDockingBar from: #MenuItemMorph;
		set: #logoColor for: #TheWorldMainDockingBar to: Color black;
		set: #selectionLogoColor for: #TheWorldMainDockingBar to: Color white.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/5/2016 08:48'!
addScrollables: theme
	"self create apply"
	
	"Sliders"
	theme
		set: #borderColor for: #Slider to: Color gray;
		set: #borderWidth for: #Slider to: 1;
		set: #color for: #Slider to: Color lightGray;
		set: #thumbBorderColor for: #Slider to: [Color gray: 0.6];
		set: #thumbBorderWidth for: #Slider to: 0;
		set: #thumbColor for: #Slider to: Color veryVeryLightGray;
		set: #thumbShadowModifier for: #Slider to: [ [:c | c alpha: 0.7] ].
				
	"Scroll bars"
	theme
		set: #thumbBorderWidth for: #ScrollBar to: 1;
		set: #thumbColorModifier for: #ScrollBar to: [ [:c | c] ];
		set: #pagingAreaColorModifier for: #ScrollBar to: [ [:c | c darker alpha: 0.35] ];
		set: #borderColorModifier for: #ScrollBar to: [ [:c | c adjustBrightness: -0.3] ].
	
	"Scroll panes (includes generic stuff for list widgets, tree widgets, and text widgets."
	theme
		set: #borderColor for: #ScrollPane to: (Color gray: 0.6);
		set: #borderWidth for: #ScrollPane to: 1;
		set: #borderStyle for: #ScrollPane to: BorderStyle default;
		set: #color for: #ScrollPane to: Color white.
		
	"List widgets"
	theme
		set: #font for: #PluggableListMorph to: [Preferences standardListFont];
		set: #textColor for: #PluggableListMorph to: Color black;
		set: #selectionColor for: #PluggableListMorph to: (Color r: 0.72 g: 0.72 b: 0.9);
		derive: #multiSelectionColor for: #PluggableListMorph from: #PluggableListMorph at: #selectionColor do: [:c | c lighter];
		set: #selectionTextColor for: #PluggableListMorph to: Color black;
		set: #filterColor for: #PluggableListMorph to: Color yellow paler;
		set: #filterTextColor for: #PluggableListMorph to: Color black;
		set: #preSelectionModifier for: #PluggableListMorph to: [ [:c | Color gray: 0.9] ];
		set: #hoverSelectionModifier for: #PluggableListMorph to: [ [:c | c darker alpha: 0.3] ].
		
	"Tree widgets"
	theme
		derive: #font for: #SimpleHierarchicalListMorph from: #PluggableListMorph;
		derive: #textColor for: #SimpleHierarchicalListMorph from: #PluggableListMorph;
		derive: #selectionColor for: #SimpleHierarchicalListMorph from: #PluggableListMorph;
		derive: #selectionTextColor for: #SimpleHierarchicalListMorph from: #PluggableListMorph;
		derive: #filterColor for: #SimpleHierarchicalListMorph from: #PluggableListMorph;
		derive: #filterTextColor for: #SimpleHierarchicalListMorph from: #PluggableListMorph;
		derive: #hoverSelectionModifier for: #SimpleHierarchicalListMorph from: #PluggableListMorph;
		
		set: #higlightTextColor for: #SimpleHierarchicalListMorph to: Color red;
		set: #lineColor for: #SimpleHierarchicalListMorph to: Color veryLightGray.
	
	"Text widgets"
	theme
		set: #font for: #PluggableTextMorph to: [Preferences standardDefaultTextFont];
		set: #textColor for: #PluggableTextMorph to: Color black;
		set: #caretColor for: #PluggableTextMorph to: Color red;
		set: #selectionColor for: #PluggableTextMorph to: (TranslucentColor r: 0.0 g: 0.0 b: 0.8 alpha: 0.2);
		set: #unfocusedSelectionModifier for: #PluggableTextMorph to: [ [:c | Color gray: 0.9] ];
		set: #adornmentReadOnly for: #PluggableTextMorph to: Color black;
		set: #adornmentRefuse for: #PluggableTextMorph to: Color tan;
		set: #adornmentConflict for: #PluggableTextMorph to: Color red;
		set: #adornmentDiff for: #PluggableTextMorph to: Color green;
		set: #adornmentNormalEdit for: #PluggableTextMorph to: Color orange;
		set: #adornmentDiffEdit for: #PluggableTextMorph to: Color yellow;
		set: #frameAdornmentWidth for: #PluggableTextMorph to: 1.
	theme
		set: #balloonTextColor for: #PluggableTextMorphPlus to: (Color gray: 0.7);
		derive: #balloonTextFont for: #PluggableTextMorphPlus from: #PluggableTextMorph at: #font.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/11/2016 16:37'!
addSyntaxHighlighting: theme
	"This was the former sub-dued highlighting."
	
	theme
		set: #color for: #TextAction to: (Color r: 0.4 g: 0.0 b: 1);
		
		set: #default for: #SHTextStylerST80 to: {Color black};
		set: #invalid for: #SHTextStylerST80 to: {Color red};
		set: #excessCode for: #SHTextStylerST80 to: {Color red};
		set: #comment for: #SHTextStylerST80 to: {Color cyan muchDarker};
		set: #unfinishedComment for: #SHTextStylerST80 to: {Color red muchDarker. TextEmphasis italic};
		set: #'$' for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #character for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #integer for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #number for: #SHTextStylerST80 to: {Color red muchDarker};	
		set: #- for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #symbol for: #SHTextStylerST80 to: {Color blue muchDarker};	
		set: #stringSymbol for: #SHTextStylerST80 to: {Color blue muchDarker};	
		set: #literalArray for: #SHTextStylerST80 to: {Color blue muchDarker};
		set: #string for: #SHTextStylerST80 to: {Color magenta muchDarker. TextEmphasis normal};
		set: #unfinishedString for: #SHTextStylerST80 to: {Color red. TextEmphasis normal};
		set: #assignment for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #ansiAssignment for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #literal for: #SHTextStylerST80 to: {nil. TextEmphasis italic};
		set: #keyword for: #SHTextStylerST80 to: {Color blue muchDarker};
		set: #binary for: #SHTextStylerST80 to: {Color blue muchDarker};	
		set: #unary for: #SHTextStylerST80 to: {Color blue muchDarker};
		set: #incompleteKeyword for: #SHTextStylerST80 to: {Color gray muchDarker. TextEmphasis underlined};
		set: #incompleteBinary for: #SHTextStylerST80 to: {Color gray muchDarker. TextEmphasis underlined};	
		set: #incompleteUnary for: #SHTextStylerST80 to: {Color gray muchDarker. TextEmphasis underlined};
		set: #undefinedKeyword for: #SHTextStylerST80 to: {Color red};
		set: #undefinedBinary for: #SHTextStylerST80 to: {Color red};	
		set: #undefinedUnary for: #SHTextStylerST80 to: {Color red};													
		set: #patternKeyword for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #patternBinary for: #SHTextStylerST80 to: {nil. TextEmphasis bold};
		set: #patternUnary for: #SHTextStylerST80 to: {nil. TextEmphasis bold};	
		set: #self for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #super for: #SHTextStylerST80 to: {Color red muchDarker}; 
		set: #true for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #false for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #nil for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #thisContext for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #return for: #SHTextStylerST80 to: {Color red muchDarker};
		set: #patternArg for: #SHTextStylerST80 to: {Color blue muchDarker};	
		set: #methodArg for: #SHTextStylerST80 to: {Color blue muchDarker};
		set: #blockPatternArg for: #SHTextStylerST80 to: {Color blue muchDarker};
		set: #blockArg for: #SHTextStylerST80 to: {Color blue muchDarker};
		set: #argument for: #SHTextStylerST80 to: {Color blue muchDarker};
		set: #blockArgColon for: #SHTextStylerST80 to: {Color black}; 
		set: #leftParenthesis for: #SHTextStylerST80 to: {Color black}; 
		set: #rightParenthesis for: #SHTextStylerST80 to: {Color black}; 
		set: #leftParenthesis1 for: #SHTextStylerST80 to: {Color green muchDarker}; 
		set: #rightParenthesis1 for: #SHTextStylerST80 to: {Color green muchDarker}; 
		set: #leftParenthesis2 for: #SHTextStylerST80 to: {Color magenta muchDarker}; 
		set: #rightParenthesis2 for: #SHTextStylerST80 to: {Color magenta muchDarker}; 
		set: #leftParenthesis3 for: #SHTextStylerST80 to: {Color red muchDarker}; 
		set: #rightParenthesis3 for: #SHTextStylerST80 to: {Color red muchDarker}; 
		set: #leftParenthesis4 for: #SHTextStylerST80 to: {Color green darker}; 
		set: #rightParenthesis4 for: #SHTextStylerST80 to: {Color green darker}; 
		set: #leftParenthesis5 for: #SHTextStylerST80 to: {Color orange darker}; 
		set: #rightParenthesis5 for: #SHTextStylerST80 to: {Color orange darker}; 
		set: #leftParenthesis6 for: #SHTextStylerST80 to: {Color magenta darker}; 
		set: #rightParenthesis6 for: #SHTextStylerST80 to: {Color magenta darker}; 
		set: #leftParenthesis7 for: #SHTextStylerST80 to: {Color blue}; 
		set: #rightParenthesis7 for: #SHTextStylerST80 to: {Color blue}; 
		set: #blockStart for: #SHTextStylerST80 to: {Color black}; 
		set: #blockEnd for: #SHTextStylerST80 to: {Color black}; 
		set: #blockStart1 for: #SHTextStylerST80 to: {Color green muchDarker}; 
		set: #blockEnd1 for: #SHTextStylerST80 to: {Color green muchDarker}; 
		set: #blockStart2 for: #SHTextStylerST80 to: {Color magenta muchDarker}; 
		set: #blockEnd2 for: #SHTextStylerST80 to: {Color magenta muchDarker}; 
		set: #blockStart3 for: #SHTextStylerST80 to: {Color red muchDarker}; 
		set: #blockEnd3 for: #SHTextStylerST80 to: {Color red muchDarker}; 
		set: #blockStart4 for: #SHTextStylerST80 to: {Color green darker}; 
		set: #blockEnd4 for: #SHTextStylerST80 to: {Color green darker}; 
		set: #blockStart5 for: #SHTextStylerST80 to: {Color orange darker}; 
		set: #blockEnd5 for: #SHTextStylerST80 to: {Color orange darker}; 
		set: #blockStart6 for: #SHTextStylerST80 to: {Color magenta darker}; 
		set: #blockEnd6 for: #SHTextStylerST80 to: {Color magenta darker}; 
		set: #blockStart7 for: #SHTextStylerST80 to: {Color blue}; 
		set: #blockEnd7 for: #SHTextStylerST80 to: {Color blue}; 																																																		
		set: #arrayStart for: #SHTextStylerST80 to: {Color black}; 
		set: #arrayEnd for: #SHTextStylerST80 to: {Color black}; 
		set: #arrayStart1 for: #SHTextStylerST80 to: {Color black}; 
		set: #arrayEnd1 for: #SHTextStylerST80 to: {Color black}; 
		set: #byteArrayStart for: #SHTextStylerST80 to: {Color black}; 
		set: #byteArrayEnd for: #SHTextStylerST80 to: {Color black}; 
		set: #byteArrayStart1 for: #SHTextStylerST80 to: {Color black}; 
		set: #byteArrayEnd1 for: #SHTextStylerST80 to: {Color black}; 
		set: #leftBrace for: #SHTextStylerST80 to: {Color black}; 
		set: #rightBrace for: #SHTextStylerST80 to: {Color black}; 
		set: #cascadeSeparator for: #SHTextStylerST80 to: {Color black}; 
		set: #statementSeparator for: #SHTextStylerST80 to: {Color black}; 
		set: #externalCallType for: #SHTextStylerST80 to: {Color black}; 
		set: #externalCallTypePointerIndicator for: #SHTextStylerST80 to: {Color black}; 
		set: #primitiveOrExternalCallStart for: #SHTextStylerST80 to: {Color black}; 
		set: #primitiveOrExternalCallEnd for: #SHTextStylerST80 to: {Color black};
		set: #methodTempBar for: #SHTextStylerST80 to: {Color gray}; 
		set: #blockTempBar for: #SHTextStylerST80 to: {Color gray};
		set: #blockArgsBar for: #SHTextStylerST80 to: {Color gray};
		set: #primitive for: #SHTextStylerST80 to: {Color green muchDarker. TextEmphasis bold};
		set: #pragmaKeyword for: #SHTextStylerST80 to: {Color green muchDarker. TextEmphasis bold};
		set: #pragmaUnary for: #SHTextStylerST80 to: {Color green muchDarker. TextEmphasis bold};
		set: #pragmaBinary for: #SHTextStylerST80 to: {Color green muchDarker. TextEmphasis bold};									
		set: #externalFunctionCallingConvention for: #SHTextStylerST80 to: {Color green muchDarker. TextEmphasis bold}; 
		set: #module for: #SHTextStylerST80 to: {Color green muchDarker. TextEmphasis bold};
		set: #blockTempVar for: #SHTextStylerST80 to: {Color gray};
		set: #blockPatternTempVar for: #SHTextStylerST80 to: {Color gray};
		set: #instVar for: #SHTextStylerST80 to: {Color black};
		set: #workspaceVar for: #SHTextStylerST80 to: {Color black};
		set: #undefinedIdentifier for: #SHTextStylerST80 to: {Color red};
		set: #incompleteIdentifier for: #SHTextStylerST80 to: {Color gray darker. {TextEmphasis italic. TextEmphasis underlined}};
		set: #tempVar for: #SHTextStylerST80 to: {Color gray darker};
		set: #patternTempVar for: #SHTextStylerST80 to: {Color gray darker};
		set: #poolConstant for: #SHTextStylerST80 to: {Color gray muchDarker};
		set: #classVar for: #SHTextStylerST80 to: {Color gray muchDarker};
		set: #globalVar for: #SHTextStylerST80 to: {Color black}.
		
	"And the text differ"
	theme	
		set: #insertTextAttributes for: #TextDiffBuilder to: { TextColor red };
		set: #removeTextAttributes for: #TextDiffBuilder to: { TextEmphasis struckOut. TextColor blue };
		set: #normalTextAttributes for: #TextDiffBuilder to: { TextEmphasis normal }.! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 14:18'!
addToolColors: theme
	"Tool-specific colors."
	
	"SUnit's TestRunner."
	theme 
		set: #failureColor for: #TestRunner to: Color yellow;
		set: #errorColor for: #TestRunner to: Color red;
		set: #passColor for: #TestRunner to: Color green;
		
		derive: #failureTextColor for: #TestRunner from: #PluggableTextMorph at: #textColor;
		derive: #errorTextColor for: #TestRunner from: #PluggableTextMorph at: #textColor;
		derive: #passTextColor for: #TestRunner from: #PluggableTextMorph at: #textColor.
		
	"Monticello Tools."
	theme
		set: #revertedOperationAttributes for: #MCOperationsBrowser to: {TextEmphasis struckOut};
		set: #ignoredOperationAttributes for: #MCOperationsBrowser to: {TextColor color: Color gray}.
		"set: #rejectedOperationAttributes for: #MCOperationsBrowser to: {TextEmphasis struckOut};
		set: #acceptedOperationAttributes for: #MCOperationsBrowser to: {TextEmphasis underlined};
		set: #conflictingOperationAttributes for: #MCOperationsBrowser to: {TextEmphasis bold}."! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 10:51'!
addWindowColors: theme

	theme 
		set: #titleFont for: #SystemWindow to: [Preferences windowTitleFont];
		set: #borderColorModifier for: #SystemWindow to: [ [:c | c adjustBrightness: -0.3] ];
		set: #borderWidth for: #SystemWindow to: 1;
	
		set: #uniformWindowColor for: #Model to: Color veryVeryLightGray;
		derive: #uniformWindowColor for: #TranscriptStream from: #Model;
		derive: #color for: #SystemWindow from: #Model at: #uniformWindowColor; "Fall back for windows without models."
				
		set: #unfocusedWindowColorModifier for: #SystemWindow to: [ [:color | color darker] ];
		set: #unfocusedLabelColor for: #SystemWindow to: Color darkGray;
		set: #focusedLabelColor for: #SystemWindow to: Color black;

		set: #customWindowColor for: #Browser to: (Color r: 0.764 g: 0.9 b: 0.63);
		set: #customWindowColor for: #ChangeList to: (Color r: 0.719 g: 0.9 b: 0.9);
		set: #customWindowColor for: #ChangeSorter to: (Color r: 0.719 g: 0.9 b: 0.9);
		set: #customWindowColor for: #ChatNotes to: (Color r: 1.0 g: 0.7 b: 0.8);
		set: #customWindowColor for: #ClassCommentVersionsBrowser to: (Color r: 0.753 g: 0.677 b: 0.9);
		set: #customWindowColor for: #Debugger to: (Color r: 0.9 g: 0.719 b: 0.719);
		set: #customWindowColor for: #DualChangeSorter to: (Color r: 0.719 g: 0.9 b: 0.9);
		set: #customWindowColor for: #FileContentsBrowser to: (Color r: 0.7 g: 0.7 b: 0.508);
		set: #customWindowColor for: #FileList to: (Color r: 0.65 g: 0.65 b: 0.65);
		set: #customWindowColor for: #InstanceBrowser to: (Color r: 0.726 g: 0.9 b: 0.9);
		set: #customWindowColor for: #Lexicon to: (Color r: 0.79 g: 0.9 b: 0.79);
		set: #customWindowColor for: #MCTool to: (Color r: 0.65 g: 0.691 b: 0.876);
		set: #customWindowColor for: #MessageNames to: (Color r: 0.639 g: 0.9 b: 0.497);
		set: #customWindowColor for: #MessageSet to: (Color r: 0.719 g: 0.9 b: 0.9);
		set: #customWindowColor for: #PackagePaneBrowser to: (Color r: 0.9 g: 0.9 b: 0.63);
		set: #customWindowColor for: #PluggableFileList to: Color lightYellow;
		set: #customWindowColor for: #PreferenceBrowser to: (Color r: 0.671 g: 0.9 b: 0.9);
		set: #customWindowColor for: #SMLoader to: (Color r: 0.801 g: 0.801 b: 0.614);
		set: #customWindowColor for: #SMLoaderPlus to: (Color r: 0.801 g: 0.801 b: 0.614);
		set: #customWindowColor for: #SMReleaseBrowser to: (Color r: 0.801 g: 0.801 b: 0.614);
		set: #customWindowColor for: #ScriptingDomain to: (Color r: 0.91 g: 0.91 b: 0.91);
		set: #customWindowColor for: #SelectorBrowser to: (Color r: 0.45 g: 0.9 b: 0.9);
		set: #customWindowColor for: #StringHolder to: (Color r: 0.9 g: 0.9 b: 0.719);
		set: #customWindowColor for: #TestRunner to: (Color r: 0.9 g: 0.576 b: 0.09);
		set: #customWindowColor for: #TranscriptStream to: (Color r: 0.9 g: 0.75 b: 0.45);
		set: #customWindowColor for: #VersionsBrowser to: (Color r: 0.782 g: 0.677 b: 0.9).! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 18:04'!
create
	"This is the default theme for Squeak.
	
	self create apply. "
	
	^ (self named: 'Squeak') in: [:theme |		
		"General morph stuff."
		theme
			set: #keyboardFocusColor for: #Morph to: (TranslucentColor r: 0.3 g: 0.5 b: 0.5 alpha: 0.5);
			set: #keyboardFocusWidth for: #Morph to: 3;
			set: #softShadowColor for: #Morph to: (Color black alpha: 0.01);
			set: #softShadowOffset for: #Morph to: (10@8 corner: 10@12);
			set: #hardShadowColor for: #Morph to: (Color black alpha: 0.5);
			set: #hardShadowOffset for: #Morph to: 1@1.
			
		theme set: #background for: #MorphicProject to: self linenblue.
			
		self
			addFonts: theme;
			addWindowColors: theme;
			addSyntaxHighlighting: theme;
			addMenusAndDockingBars: theme;
			addDialogs: theme;
			addButtons: theme;
			addScrollables: theme;
			addToolColors: theme.
		
		theme]! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 7/25/2016 16:56'!
createClassic
	"self createClassic"
	
	^ (self named: 'Squeak (classic code style)') in: [:theme |
		theme link: (self named: 'Squeak').
		self addClassicSyntaxHighlighting: theme.
		theme]! !

!SqueakTheme class methodsFor: 'instance creation' stamp: 'mt 8/16/2016 12:14'!
createDuller
	"self createDuller apply"
	
	| name |
	name := 'Squeak (duller)'.
	^ (self named:name) in: [:theme |
		theme merge: (self named: 'Squeak') overwrite: true.
		theme name: name.
		self
			addDullerWindowColors: theme;
			addDullerScrollables: theme;
			addDullerDialogs: theme;
			addDullerMenusAndDockingBars: theme;
			addDullerButtons: theme;
			addDullerToolColors: theme.
			
		theme set: #color for: #TextAction to: Color ocean.
		
		theme]! !

!SqueakTheme class methodsFor: 'images' stamp: 'mt 8/5/2016 11:01'!
linenblue
	<generated>
	^ Form fromBinaryStream: '<~M,6r;%14!!\!!!!!!!!.8Ou6I!!!!!!(&!!!!!!(&#Qk&,!!/\4!!mJm4f2,HtG76Wt''5XIk*6q0X88jl*=5=%Y&
6Ua4+6U*k)7n69E8P_uQ;F<iB4?c%u77Tg:90u!!877Kj56:=%''6U!!e''5!!VJ#6:4%,7Qs+)5!!VM#
5XI_%6U3t,5sdt(5X[e%6pa=45=%V#6:F()6U3t)5X\(-6:F1.77''@15!!hY&6Ua:.6pa:15sn+2
7nZ6>83B+&5X\",6ps:+6psI56UaO?:/jJG7R07+5XRq,6UjF17m]X66UaI26:O@37moa95X7_&
5XRe)77''=05t"+07S6'':7mf[66:FF46Ua:-7R07.6:+(,6q9I/76Nk%5XIq+6UO728NoI-5X7Y#
6q''C17mTO56:4.,5t+.076a.-6psL26qBa983fO06UXI677fm:7R9C45=%_(6UtaZ8;$sC[+S#=
5u`(0]TWZ-i]mo>d;+j^52?1Ndtpm1OM>_6E\%m5Q]=%!!BJg)AD+kARWt!!1''i0jbipIrl#e6#<<
7!!YlHoJYkUO1qeD:8me8i0o9FiZV8ipm%(&4s"T\e)Ybd4jub;+$`?V@@rBfR\`4k!!I36aP%2%1
>4=cEK9u%&]&@P/qt%;%<cAW;-\\lE-Ci@mH^Kd^"KT!!:d!!]_<`,XU-5l;P8.4k6c@I_OXdVH_J
[gMHi#D<i]42TDR_U9C?o#HF6#%.O)GYqR$f[lWN^A3SW&XK%(QGhkDG2^fn;1(c6_#s.tI@(kp
!!q7>6i7**Wcfa-H6^MB+^$LD@N:m@IT+Mj0C%HlHnPMOnL''T;KL@`OGVl/"^(aCYXRPk7D45%a!!
0qTXZ"pY4r;$B^@cIoH;DXjk.Rb-]NU0_:l%s9&tGu7]Bd+#iGXFVu50*s-nhY`;CLPu-Fd4Gu$
5CmS`QY"jHpMmdDl-<-:/B8trW5.?%lT]R;KY^M+L;B!!]f_b[J^NC''+.YXa9g?H5ApZN<G%#r$C
MA]n''#7*BfGhA9IpCb''84fTj%W3dZq\/O1)?kEX_b!!eUH[W!!KCR>s`]/Lj=nX?<kVbKN9-W:>ui
0MJ`(ai_GOX>PU:M6b+fX*RsSE^&\:"@//joHbSCI.5[(b1S;r(b?2QS*Tb3;oKAOlae_1!!Z#^+
Q)@@D",`4MnO9:Yn0!!0C=>i(Z9-rj>UlOQED9R1h;6gPeM.#;q@)Y>HX;/ni/7^jG<kJPdi$i''m
*jBSc/-%J6"\gFoP5Rj>3UKFI^<V>lJDLSJ>MK>h=P"K3EXW_3m.VU^;euq''G5G,%gU%GJp8BjS
ii+1uJVCg[+lqh3_)dn!!59)6M$@+0s_JZa?_Gb;X#%]`g@5'']DDAUMUZtPe+>cW_ZbD^_UM[k2f
]Rmd,kQNq<m=Gg3>5Y,U`4UU!!8:HEiN!!@lePFDq7"P+?k=QLUCo#XM.r*I;>C%K,h8.2C%WsH-I
Y[m\''bA(@!!2ta"CdW&?ue"r=ZOkj!!%,[1o/-d5QOGDZRZ:osQ1!!X+G=b8;da-Z&>%(>`rET\Y\G
B<ijWTrDo3C8)CLXN$''aGBPI9ghKBh*qt<KMKN_oC_4>e/DuQh$R]?@V-m^IXMn!!8%i/fn8fm7l
Y\JU@[2\BP>C5=1O:A3^CiLP!!.[+e`>irk9n#Ng"NO65oGRB`Me+^+&5/(CoEj2Fc4%H-o^#>C"
d[\C!!A>sF<WC@D!!E\d2fCq[MZ]:o+iI4g+Q?od>;RbGd)KKK,r==#LdYjV2%LJ`eF<XR)]\ROMk
Yo89UT6,CM9chGk%h1;FT>=nOUAib2''E!!-ghe9_OE;=ruOO=qC-;]G\*kLfjOtm*^4Jg@sn$.BC
r=BkB`Y=N=VWf(&j\fN[8jaT712.D(I45k4j>C&6ruA*q!!)#QuGD3VZ*X8b/G+PN(>AbljFt(QN
/?HY&]MEY+7ps:A/5(F3+X''.$lnTuChI;g<qu"#fN>:cl''Q2<(jA-%p\I_l2<&-o-DZU^S)R/^!!
YV+hIp0bl<@$q-mJlPR]MT<L^O+Yu$kmL8%3pb-McQupDeS\_),l0Jqp%NqZ=Sb&PK&ulR+Qm0]
cS8Pc9-In>U!!!!=o$`CeWPJAYZ#-=J^&Y$NkYTdf@eZP!!h!!1I&l+YYTdO:!!ONGgG^mR*jdB0`NT5
V`^84-s/<7a6dmgqheqbf!!6jgXPEl\+(''o5Fl%\W%;/H^W$/=gHWqIX&!!k8*l\uBqhY*;=-`MmA
Dh%b7\%VR1mC=lcVDm-C#1,;"1UlOTOm8#Pa;lfbKgWg''gY`QBDE)rePXTFl-P["!!Ke&J45CVL8
ViUA0@Vo@qS:bDI6"E(bQ?W-&e]g2CGguo^(n#Sb!!0krUI^goHrF"Rif$E7_=F$&-O<,suJlTt2
@$R(2KitY2-\/TUs!!"[>)1tr3.9Llb''J/,g\#=H-Z1<63?kJ06!!*&o/`+*1P+*Q_M`kOfc%bAd&
AI''''HB8+$56)6^I^@bmAXa#k()[tT/0M"a[J41Nqi\SAJ<SGW2>+ue_T)#eO_eSAl,T''rt\W<]p
o?J5md_jS>`r\lV[tEf6-rG9.]]X$NLuWY#MZf=fm(E^l)`b4"$aTi[9m''cJe[4W0qn=`P]pX0G
s-LtY)0XHV"9_";.:lL:cCqbF1DoZ(h;E5Pl-7T%TDWWI?tG<6''<Lh.0ubd51@>nF/IEn''on$i:
of!!)r">%1iMdM"nB%fpiD9@U689*DKU%Qd,C/@a5\p2iCi''Ji6p]M?7,.NDSmIgQ&5kbD\NR$Dr
/01*B[qgUoRX7DQ`;h?I!!Sb\*fr.S[p,Dks''SoZ%e_`oR*@sn"+oqJ4ot2>=Do:IV+D#F(77s^$
=>dnJ^HeJK2:`HmHqU^=*PcDsK>m/lH>X5uM_hS0.Gl$pY(FV]YTad1EL9+Q[Fo%3>&j:A!!TqbV
!!a*NcK7I,0$P$5549/#%Dd52Q)-FmJ"+VKd`L"A,J7UK+ol"J+o0G6?.dU&]pm5%fMCrJ4DZboa
4dP3)874(W,)V"AEWrOBb+ZfV!!!!B8_C_!!*!!^dr/jI3oq&p#cCtq[Z_\W-aN;HJO[hC#2+3L$%C.
\#,''m?8ap3Id^t$O]0#nZ+O0%c&TX?(Qq7ZMg(e3;#RZ\,/o9;ZdCfU;eJG.k/QO"GR$m^o.8:j
RmfDq<=jqTeO`gIXEpiNlid.a%QYS.O#7/X+V_)W)Z]V7Iitu@rrB/-OuZWe$*54I][ObBXYVs>
bNtP+hfK<L/2#fX^$H@.Yag".;(.+6%_B<XYtMe@a)oY0;#%30Ed`+TYb\VZ1JnU`oiCJa<dPN_
j]d;4(N''G8;<RPA_4j-ZoiM_*$g;i4$G?Q*%8T!!mahdosAnDs^T+-9V82kp?Yi,=1PTopE#I*ps
X0F##K]ek[P*o0KXlPnCe8+ocSq"ii2:>S>N!!uolb!!Y]:#uZMtE]rH.]+22[:510."BIg?V#qUu
Yo9&uJFlrM+nG:K1dK]I%^:%B<1RX)cl]lDV0V*4O&\c$:I+f1%KLk\kBT`Vdg_KI%=0P@WiKN1
5/p6c]kh;PDX`b[<f^8IpEWj%l-WG:Hbsho?QTMWCV;qo/;#''lKk>K^Uk<nV4I1^2T''m=D>R/i$
]."HF,EG9V#$,"0$f;%;h;/-4$*_pj4/UTc%TY$T4j193&hH_liX@@T.,P"UHj?mQ6TVW$\W"(@
LM\''5$b\l7''eDR8C8?pWG`"kji>s''+$WsnV/8kl`!!>YfDC9/8DZcA;IMr[&d<%3aM8o^jP=%O.k
,ou:P=K\o(0,faK&FD\/rWg\2(12\_aJ>s.Y`D^\)U$''P)B9N2+l87mUuquKp9B:Cg)n1sEW^EW
1QReB\P<d5HX?cUZ#UV#PRgdo%GoNC-!!97uYk2+Bg4njHn5q`$T#]QLa.d@NEOru!!Ta\_7YOG5h
WS><VBR=_8X?''EAK-eoOFT6BE$TJQK,qiTjH`sajI;TLqZ3&V0A,.k,]^AWF`#($+GYKRhYm;*r
7MblK"VsG`cN#N!!K2iO8!!(:Vi7^/O!!M?eZ3<l"j7@$LtI`ReMV,$;>a/?e(OBe6s:*7BDiL1_RN
SkMb?D[/qD][s[oQAZFWE3Rh0d%)2h?O&H:`/Eodd?Ck?%MSZ"Nc''BM`sf$f,sg((F,"5Z%G-3n
E0re!!YL%>U.Gb`>al4:$(tl!!:TBWUYF3MJK4\]EdBroN9l)*T''VZsrU1^;r&-88-0@Y/0Q(uGX#
Sq$&$Q=(#>.iBFE/Eon*jUm0mr>7()X4.OFE!!2J$KW.$M94.e1*"%[d3mP^''_7DB9l=4n>TQ<.6
KP$&#hj/[aVtjC6KB:eWf3i`PeaVUJ!!<Og?]tn2%1[btL,E5^q)$hdl*IB?.2qa*Dg''#S^f)#G_
[rDNdXoOAeK>f!!9ra[*uf''Vj4WV)"lF0OG`4''7l$"hYq]J46\5<k1V6RE*a<07oI_&6#h*LKT[L
#AgjTKi>5@#V=[]kMrXoALF''VY9D<IS\WZdF*YdT=>&GH_)_jU,^kKj8n&JAj_R&GVe][_Ft/u$
Y]?;8_-lDW`o8l]9A$Sf!!)O\9T%F0]]iUR>M^4&rX4q(mH`L!!qade$b4H<4qXl3r>Vlsh"f8:'',
_"%MT#K(nBmNl>DOH@_C5+ZYrb`&`#1[Yu=r"H;4@f`P!!=>+&V6D9$sK=V/h/pT#3@,,5uLuhHH
NIIV"?qj!!se,@L#PsMuiZKb.3J#L]\L[F+lErfU''#A_qCKaClC!!FV2Y^qr#2C]!!@IKdErq3<BR/
FH?+$"q/U^NC''>hG8(-RA^A]IlCg(OF5$*EnfJ.'':R8UIh(T$@A$BnT((;aAb"u2IQE5CX=`a03
pnm%Vi5*kOJhr.''i,&oSV8Z\UlWC^N2.dhu/L.t9N&m>";2k-fa^`^$[F\i.=<eVl:a+[-iZ7>''
bDQfY@(?8<6C)pkc,%7>o:B@_%eZD)k"%dG(P=b4CZa`gq=V!!@+L"<Vb97hs$ATZQ=1@m0]C''*G
`d+^"hg?)>A.CqQFb?/NAq5"G*m<8`-sQo@0GL6=B+9%BrjYY<C+XE''UJ!!s8K!!\P`J&6UC,;u>B
3&""m`;*FVb2Jss_=H(XWH>s+=;ZS-#D+hG&spG_9<2J!!q@6mf^tL<197g13$5P;sMg#:3[tBrW
1MSC4<^"`d8GMU\AMPttQB+EcJ>MuD(-N?aJ>KI5"M43On2\<7R"3/Z_;ZFc*CFSCJTtN,90''0o
+d>/SRcCFTQAP@bFr''N[Ntp0bF8D6UajNaOYS!!,#%5eo!!@nf*e>^5:G.24CL/:3A(S]?''1-Y%6O
8F?<;/bpAX<<p\J[WW&p8dV_4la"O1o3UC#E[uo$j7>LEoa9eLmKSuJKN58!!I6um]YkU(XIgg+p
24`N<6=DN./l0*J)Rb=&2iOWgQV9=XJlPVP4[#0aJVhYP$e+R8LI*]X+h6-ZLViK_<rL<Oaj+,,
C2759iF%rAcG1@E/#9eqe0>m^YU+7g%u6Z)*tMDL)q6&tC2iuiSX7o7K@P,f*Qhlr;Y5;DYj*-/
raU`hJ`WSE&M/ZI$*Z2@5YAZs''"W>^NL=kjT\$SjRb-Li]O)+Kq0GMi?Y!!\V;E6PHBD-LEquW-"
''D`bEh6`L66j]Lm2>_[_:ON#;o0]Y77ZuTC1''e(2H*[5GXPQKo4k:job5\*%GoqZm/-He88G2tF
a3lh?-XaVSVg]Yf\#1/b/@T-8U[YLqf3iSDU^)KO4:XU(raZ:8k''F@833U@@/CplZ5=mf4/VQOD
h]Veh&IQd+%]_#m<dl^+GSdUNF":W!!#dAbh?kGY>CLHgr]-]=t=[REq=L.a$!!t/:]o'':RT#o6!!<
#qFich!!\%L%R]B`9H+`$Ag8s>SE0l(@Ca*R#lH-ceju(d^\,U-n@''AgC\(A)?dG<aYf8Sd_XZ"@
H)qlcIM!!PN)1W5M56e4*+S<>#5p*H&)IS\FN\I\<4]XHGaBr-empp_JoEL\^B"=F?YAdQSI*0r@
K%7]b`!!<td!!%X[8c>Bj*lpJ!!4Ntm1TSfg2N&a258A$GG@U=BhDK:1))!!6]i\2Xji;$S:>6Pi2ne
Ke;M=GUl]FDBm#!!7XAO[-7rbf6=H:;_K#Z.Pd!!f5Ij^V16ntpI;WZ/2Oo/NniB^kZT+>&c.^2OU
(K_Pe"dZ?UZjrC[pEd%S,DH5lHCrQ"\T_I>ru>7?;3.''[ZM%o2>%pf>CE4sE#/P0bTAKcP=TOG%
Yf[\gq0Aah#V>))@QHa=&?KsBeCk?l#[Odf>L.G5BRHQ=KgBD%(o`kB!!^I#.CL/j\J2fO[6"km2
nj(N/2*bQW\aUtsN_Tq)D`TH&FrRT>"GU4;$=:"u#>9?D>l!!(mrBcH*QJ_#;g+Co!!p"p#;@aBbr
lb$/G\5C2%cBq`I*fQ-[bM4hbfgYIbm,e:!!K,%^dWT/=F$8$6T[%jN#Fe3JAYSM''/1$N\&XK0/e
,HtX=@1H=MhA(GqCM+M9o9gB9L1&R+$PnL6n''5=+f#C''<A$GGc^uE8Ji^L(nK*C/$)TsGQQK\U#
4Bo,S0TQrIb8i,B^iF;SiFO6Lg?lRH^e_d-T?t1"No#o[2a[l5U!!S:L^fn)_?km66FiQe2gQ''1,
/8Op#6h(I.(IOmE7t:a[=Fg*C*eQ<cmfH,>&$$XM4k@,E3*qh-C7"5;K>c-YE9bn^l/N4<9/19"
ftX.-%A(h9D,&G^Ji31&ZF+5lY,1TJ,7h)*#7LemOP]`a@f_q4ZK''7S*J$WfH\DOcCTs=T!!iQK7
Pm]r?ANARS-^QWn/D4i9c+ke:@-RPK=rVRXYrE8J$.bP>]cS:*eC"hp4brB1$.^BnY1"?A.1L7r
br2\^FM2].=T7I5?lA_M"KPi4<[lXQ-!!NF%7tLmqK2pb8<[kM.71M\YV&Q-n/<.%<i)IA,YNp0V
`7JBc*m5I)<MrXE;#os$lM2RQL1#Gtke5_eqUV*d''mn4_\8E1;/H]cWA"DA[blk)e<[G)/)"sHl
fkPYG=+Q&3=OMF0$>gmKUNQ1Y?;IguO%#:RPK1O/O#>WS1rO_ik((:LDAZ**r?RH#(*:Y^QY[+1
Mn6''+Ro$@o?HU=A[UK#Lc0!!jc%s\;D,*:J8)k4Y&-n7[JIC,TF"6\\B<j9aD!!3:?G\(YKhAZG,^
:kD+k_;a''SoaDoDCUP3\/k+[F/,A?foN4]k0A8g?NgOu9>Jm1f$Q_t0L#cX+BjV"d(N(aI"q+La
KaV9s"r?Y4(@DQr`ZO`q_;]eF$ZiPb<l<NZf_dN5?pQ0$##h,FBtD,aMYA(+qYZ#<94i=^_"(0S
,sEtJ:s,um_33(;"''1.CI:b>N!!la<2V0<dlHVM;uos$cHm''X_?dOn]q%*7?]=K\nka0t:QX5VsF
m/(`_I5odBgE-):TL\6HLM^Wgn''GGL(''23`<O_ZP@L/:&SUn1VF4@;B^L@Du]N;$b=n>_reo7]e
Mh7$GFo[gcI>Od>[_Zn<M*1tL2#CFkItPd.''Ws1]EmU,Z&amDaK`oT<]:I<+FqH:[UBe^k6$b[r
jYBrN!!Nb:1K5n)V3B[ce66Z"`B40[:YY+`j@$Paa=;=(K3W\"T/3X%bh?tlp_(,0_bGK<VkK-O[
#7I]N>^4fh@W&8OeB1i7?b@C1rLR#nrfB9#(-[TZ2''uG#n]=<JY_-''"A!!O?;''0;FQ2k7^aQ=I5(
bLoIp.c''$V0_H9)N=Dtm2&7/_gR!!\eDQ(*Brj<e_*@u0:Hc;\''\(Aj;<r]8h7pp<N$;:1N$ppY(
V8r-74Noh5lOWsgc99`Y2PX:P/:`^?d#aGJ!!\@]ur&+#jo8XiU\iZOt+LB("o$Z0r$\:)BKoZ3A
!!\;GWk+T#D-*YlL''F:!!I[0d8ofYREB!!6[J]EbAO@$Xq4S''q5+[[(qs#E-+eW4T2VFYu\[2:rior
dUL,Sp(f`Y0VDPNX-B4q=FUYLm`0JVerds%jYj>nFdmHVE)3^\,J^T=Tjr9h/<@c0j=+>U[@2<&
F0H-4EoIF.S%qWG>[tmcK;GI>dBbp\WQ%j-FNHRGi&DFJbo=IHZoc#<e9Ybb?,#Icku916NVb+m
jBt.#A*GoB=J_6V-Z.]j<&7)#rkaO?hCT@"V]75:Q`YV9$K0O<!!6r@.,5OWAm]i:?/I-h''?bNOb
Z+Z`mQRa%/g(`m&hRt2i5;+E3]u!!n9[ef[i$C5T=(8kQ*;B>n*2\?du&.o9Bq?SXaNm5hX<p6bK
=9oFPXiAd[oXILX3U)''t[Ai:r.239?%FG91K4PGBiQ=uQ+]tSR40\KC^r":M#Mo_%/q[_RY(;:F
$s''n`=NEE5km!!)^E\%pe=J;oO*RgAIe/qX_W9HJ!!bb$r=hl''W"*A!!_siX8l3U!!UW5?D2uTf2ui@
+TMMt)m$SED]&mTf`6%r2IiXN)Ku7KN>#?OB>5*.@\e#MedE-1XoKuP)e=i1$U+)`XqrkV10Qok
hOnR6T#%j"G4EiXWt\Nr"0d/2FDD>.:cruOl-kh''%Cs1OYB-`/$m9Rbqj[:nD''lJe[5FHNbHA*(
MA4Gg'']AlOe17o%X[''d;nIU%a5<&[6maj_e8d1USiI^](%''VK$^(IGOZ''S;!!#?\Rj_$=(!!Yl*.B
L4fgX=Adm.o.DVU)k0"9l3dS"2@E!!BGIemgM_P]q&6PBK>`-edbDVVV&+)UKq*"=Ok1LdCA[''=(
!!fO\o<a>^kY=g<g"fmoC%KPn1@$/),=S5sO[?,:)Ynh]ZKY<"+(OcZk.7[cS$5S.p*fK=W]=Ci=
#''u:><@(=+i;s-Vn.!!1M)o7sQl0PX6^ku!!*PpNc"2+$7\!!)&M;q-:CNOri9$@msa2r32_N*;GNa
De]F1A+<2^D,6%=\UAa2*bu*`WLaO*WtA`f\^LRm?s@,EVaq6Ie$a+Yo`)HB]q`!!`G(kWImRLV7
=<-<tcDdSVQ$Jqe8ti\&dl<eJ[94-sKst@_NT@O1Fhb)%^UWGS]$Rn[AgO,1cm4mgGUrh''O#=$^
h=i*q#"s-7AfG)@aDp>_3lVWo@GCPPn*u,-1\4o[pY4TZcFr)s@''(_2QbY)o$.@I:4m!!+*pO#!!C
)VX2DqcQq[7t=$]PR@dq!!7uQK*B.ph=IaqrpC!!9n;^hQ*fj=hs1%YB+Y^fro;FH]A;m`eQN9BBH
48F03''s)Ea(+t1B,ZlJK`?*`7.>Em]@kE;o^sop7qO)lqCPD''KUeAeRlddROb@"td:p<rd5J(2^
=SmNLB;n!!cXq>[[(9HmAX6''NoYK*XlhFX7@X!!`d?A[^*1h9g=pa5WZ''GMq``>m%=jag"RPW!!QLh
ALAH&@&[40b,%4d]EWQdFu%=Uhk0h''$4n%^.>qs_CB5soI7\Z\Hb7=%"\\#+4Vk*6&o4?A4b./!!
6L/kd@.iZcjmYq@^,G^f&M;UOk2r=Nn53Rj0>#9qo6c#?X!!N3l@+>9[rE?G0$<C@9$bm[cI/b(''
Yh%6%]R,Y4[0]agF0Bm^C9TV]BcdM0eTOgaj%rje!!E^e+k$YUR8IAL)JtW(2<J?K]B`R^\gUmP<
J?LUqY''6c+8g.jiapJ-<J:JofIda,TQ''E%IX''imT=6N0>IcRb+)1&A]K%)o;"#q8N,XC%]nM^IX
5<)b''3i[rr.EXCRXSqs?XH*2`hZm*#@Bd=^mXpq?9W+]Jr2ao]4I0\_!!,FJc)(hX"`UOLAd(,2?
Trm(Be;MjeKa0GqQ$FoF(B4gA"?KYPJH[aC_J1L;/Fs@e1:nS7H1+RC>PHT0(''-BsnlUNu7WNtl
N[k\u=P1:A2q&6*`:''Ep#:gJ-m]k?Z;CGcS9.''''-(@\X8bF=&O"&%^7+IIG@[t)Hl"r;(X92j-<
$`U]Ig-''k4FoO#tcN(''CcG7;L^tK(@b8].i=L,VTe+KQ?!!H&$u(((X/,/sF+)*o2ooB/##ln$?`
OaF1''Gc1KMdo_DlAfl]W8[.4qC91''<KhttTWrTD\7i&:f_!!:E4Up7)m=M=G5#j`#W>[sm_/C$AD
7@3L)2?&e,*7[s0MT>:fCASP)iglB\-XdrV"Bau`nD&7"$M?c@*_W]m!!7pc%NF_D&IMj[WYch6C
$<Are,tF(i[GTHc40T8''93JZ:JD:,?^):&2=C^"frn2q#PI:N.YR]6h`kjAA/4"B-Ye5amI*)B7
[L+pPa-/\Q*ecq"XKc0Mo13B=(OT<K''YVF#i"B.,>#jYTIDF!!%B_Q6dYd/*-=CWL3(3l$fVh`-U
ELtGG405]2SI7q^))C,i6fpNpQ7d8SVXR&NOG_[^K%+P?cd/FKopbZ%Fs*=a(j/d5IX_huB.3+R
+T9?IADq70,nNYNgFR>!!kMi&^kA+\d?''>oZd4?;hjjuo`/2Y+uGH:$Z#*9%%;oqgI,cA[d#:jZf
B_g?$%''3s"FXcpb1\1*@g&IaibeYbF]ZV.nVl-FFTEffO?#;AQn&(t7/O#G@W=j24=>EC6$(]f_
=FtG>aZfc`2hHaU&m0\"6G!!n5V0<RP1^=E''>d2:aGnXEe$O&esF(J8YD4fjT*u/0h&IQQdnI8n!!
R<Y(SVILW"\04G@+&72mN"li=>.83R+i1I:f/@a!!5$P/>f57as_Mq7Z*fI1hGTVZ,cXZ?:hcG``
<_3NhFm^`c0p*SZ$MD%!!k-;9=O$`tq@Uk*^idD0"mCu3Mci4mNYo7fgrQ.qj;YI];VVb>sI]@/D
;l-[m$>]o.>i`''Be90,Z<j)*&PtO1Smr7gB&F]Mfj''H0FiE(6t7]UHIiNEug\E_rF!!%^gZiXDnj
Xui,N)r)!!DTi]<5Z<,]YfdcVdBmlUPdCFH9E@rk<O$-6MHC"JY]H4$on+<i%F!!X+5$ADVD,QJ_H
#+@^<<2OYP]&B31L$o/q^DL%S''F:erNU[EX3sI3Zk@+[HgVIp"Ns:NoK8$ddjX[EQ%[MBIF.Gf-
;JE+K$5Qe/U%&!!L;t''&Q55%aT]YV.Mk)!!F@YRZu5n/*]=$ER,F/<ie8Pqi@JjC!!>#b[XqJ+LQOu
bNl4Fc"l%@W0]?oId^/^_1M<?LM&,tX0[YO+1?5ob)=G2`QVh#&.80(dp&=)=F9KtS+''p@37Z2Q
&_MGfGMh1Jmk*9FR#?0<c+ro(37h7=QGL,iZ#g9c/AF(]p$e0[%+[?2=EW;,H1F`i")B=ADp>p6
=>luHYC@\!!MOOl0RsPtO>^/gnKkW36OE6JjDDT''X,HqqGQCJ!!W9Yp;d"e9''c=8P#mb0CW%V3Had
<Q8G:%:F/f/-,H(Gq8sXoNn^6hQI0bThN;^]tj3>r?/=T&%j8lr''k)AQ^b\mX>q#e,2''V&H2j(h
llAN!!o"M=G@''sbm^^_+_(>]Ic*m_[nYj*0s<^2ZB3S6GIFure+/qGn*A*s(*<WVc)O6+Pk?PMr/
AM:$m"BIDIepHkf2:sd/#r/L7Uu,gQ&UOa\OMq,T][is@/`n.6-_0rO8tSpb:VJH]nmn%2''82ZB
AG*-`hdbS[G1X6*piqk-$s$L9CJKQonlC(%,m^T>-j''`A0=c9UXT>&2IBETQf@l0)H*5/ggSE3B
j8`#?@Xo.O^]IUkXN\J3]AFRC_G;/Q:B7uJR+S3AkHeG%<WR[IH%H2F==bt1\]1XPRVWMCYLc0D
M7h7G24$ZUBu:Ua-37@Ul/&fK]]E8SAL[a4n(Qf#n"8h,G@CU[JSoOieW.IL\%)n_/qV%O(kZC@
33,gee9/+1P-1SFh`"+7!!X=tTOb(NH%6AeDf4lE\YaHO?Z8DccIZXK4!!l#m\*.p-&YP8Y[!!79SV
qVilGB9S5B.2>`okJFj!!Hg3X5JV+O&2qZV3HA_MZb%N4n945eV#i5=NGn$(39@B1a!!+3`&F\N7f
9`U]1X@''4>nU5uM%aKd<]"4JYLMi<H+`_J3nm#:k_?RrR,-m\p;Tdc_Rm>mp;gY8?J''FfriOVa&
hCJ6m`aJU&.''IuOAQUlnI.\X`4cJ3O%?"C1fF)*G/mst8:p=L1?=<"r[tE3LrjdD0Yf[D_iGi8m
cNdiXi''f/qI,fjNqO0=31fg`#HWUFZ!!+%u8B4Wk1=Gp/&_1Gu(rEiJ&GC[;]o]7m<X]E-4dUssB
$EX"T9,fG:S:d(Fb<5Q%]GFNQXO(R_FfgLE<nGh%@HA''%W4k).=fkb=QK[>aYEFf-4nlGAXiZ71
1acsiJ1W''1(;OK2cF4cD"Nc/.$7n9dG2"<q^!!XGMqt(EMR,CQ=g.X52l\mP6NOSXI>QDs_;p,2W
\m?D071X]E^Ln"WfZ@m"DP2\MR2@:M@_1A,XE";A]cJ^A=DE>q$s)OP"?X[X5CP?nbq!!XcJ0''\@
7.1_sJYh[BP`9=_lS5+YKjQ$+hYmg%I2krfGk.GE!!=O"RlF-g&\0"hgQ!!jF(Habdui=01Wj>[j*
^p==U8i,a/]Z5Gcp/n--Q8DKl?!!CRX=R`:j9)Is18/a:=6n165LJRFqYddj">-ka>Q=ZfLmj2)H
W#=P''N15Mf#7h$)V&_\U2=jqH''H(troUahI>a7D+H9S>OSTl^>)?qFj$50F?.a0-[F31K]BbQgL
dL?X&,r<f[06m]7hI:=EhY4Hndr!!&6BnX8p0C8ba_=@]PYL>?NKB4.)@+PE#J;%nE<eeKE;Mp-m
e1OT1@-R2!!XS,]!!G5L74gDI!!t)(RT]^k-G_KYfHG\,_h%hneS9^B:Nr_#cCLFU!!caiKo4j2>jrg
f"+TUrATAJ]7Qlf$C6AY<aQrlFP''/\m7j4bVV!!^-$bus.Dg''WK##"(o`&l:a9lmDm/,WLV`lMhT
BZEX]6jGb7\''<''VXG+YtZ`-''QQS>8Ui`MZs==(cRQAh+B@*R+/?\>2!!m#0>&*JjGp"jh<lY+(tH
/ZcN\/4J?[`Z!!sDr-r!!8cFbN,HHk2/[aob?BZElO!!I[1S2$n-!!<n_cjNo!!WKCb,bV2XJ@5gDkjB
h_`hb4q,*WGcW[Q&B,lTFT@`8b"T/+T1K!!oEra7E98B2pQG8iC6;''7F(^d)af"`c96TXWe9-.GU
rfbYT\t1ja,nq@;f(,^4@_c).kfLE\3EY<<(k]5L9iMF[[/ec]..e%_,I#>X?mQQR@/\1I/:]5#
U=Q_X#>;0=K#K7V_?*h!!!!3:QEYfZuK\._uS#?[:0!!!!"8ZH^fc,;@rO7/DWJkkaKr2#Q[#o@0&XK
"Y:QS2=joN(.OdK<Z5(`m^V03%4j@Qms2"bWYmc6/+d#8I4rL%%r:ulqomfW`[[[e!!d`\Om?CM7
_%MhdV4b;$R_ViQ>M@e=@]sJ8)pj+i+6eK*ejq;-?+)cX(%1fV5]*)23X:FJEGp(5=;ZZjdVDCd
]2sOn50HrEnlQ^Do&SX9/6itZAgU7%S&*fC83=8Ba*K^Ird<QQeGV`sf''gChc''.lt94.c[!!:05N
?ibMe@$S#-QEEUaMB7L=(&GnHgo</RYl&0[&$W^:6gd''lFD_)Z[I^;fc84`U@)8%p\?)ra3s=jl
6&K657>AgfFG$29lR4B;7GVY>j''9pU@>/><dUQ#B=D(eH/?d+:Fe/RA_#d3_SD[?Oq\smb<Nga%
aD9co6#Rgq9S*2DI<r7Z^"%Y(M6=Ta#'',\2hmM>hcLRscb($PcA%+k#,rF4jP-!!h:-_6dV[[k-G
g+W383UL\\de(QD6`ls#W`61+At!!AHpb(V9r''T<@>RlIG]u!!o#"Na#l7=[#IZ/Dn:EgQZca5X3>
WrZB^EXJ,9"HsCUjMW1RR*g&GZZLhT69WPDrVYLB=SDnjNF`B]SrCK8,Hq=$$nORDXAU)@<bkNY
4''8Vl&QnsbJ48''lrWA1rdTIZI_Y\q=0j$C<Ml^h%90q4M8UD6VUkspo,.$`GeHo@WfbL!!cK*@4M
P**R-Qb''Al=0m5Ed#6a1bU1eKfnMb7,FLl<!!NaOQ(CUt]AqJB)E&7s;R<d,d[^MNNN&U!!`&#\%H
5CWY,:YYRb5Q#.QQ?Q!!N?K&/5l)g3Pr0u=&4l?&DPO''$eH`Xc&2bdUt"BIdu,2Te=nXa+h$._qc
%sebm;!!8Qn&&BaW&?JEkpu=+3/(3Qinc$+de[q@`2kLg;%TXX;hZQtR7Y(GT^qa-Z).8F2mAfjU
D?7pUY7lda2`M2''@$+[o$hj_''g_VNu#hD5QJCX&qej%@G:D9fj"RFif"^(8GTQqrUJn9CqS%QGr
`Y8HHkt(kMQ:`:;N4sN;/7q6C!!%_!!frRU^ZBLhN[LdPO@^f#P%cc:,1_&`pm!!#s[o]%Ns1FiC4a
GHfSO@(i.l8&W7-!!enY*1KY,c/ARSE7$^G+`Iud.eNgS7S_!!jU6K[lPc?Ya0%tF34?L,]$&0GX<
HVK)i:t>b"J^8)ZrV"ih8`JQr*XSI_gj0bRb,:4UVKO6[bJM=oMpMH>aX/Vc8E&lA8:52n0k4\4
rnSgh?>1?^k49DDJo&)f6j:YiU!!SQ*AF%$@i?/gY6fEt%9Qf,1$HMCc;C5fM=!![OTD!!n<iZ:]?Y
$?OB&Ph,Sn!!s^-d)%*WZdsAlt4k@1"kcqYcZ0#F+!!.Nm7rSG6''.LoJuJlSrWX/pqR''!!1Z[_4kO=
enT0Z#$*279jjBg(`urcCIr$PD#T".B%Fe=83#fM4EUZ%F6(b7XSs%<CWRo?qYE)gZ<@9e[QJeW
D0#`ZIGIeT67k"(7Jib"Of''l4]#VZm]UQp0k-:`(m)KqgAuFqg3<Ubec7fI5iJM.n$fktTVoD&T
e]ogj5K>60A$tn"qXbJ;?I9CNra3+C;OT_6T-K+b^gPEaMt])iA_t%d^"d.=1''PpdceQA-9f^D@
A*jdhO/9lOU_=eB+0P`2m4D_VI@m<4k$VI#L$N,(%-DC4K:3,9Fg->\gVcPCh-@je^aeG`@+dmL
b;p\!!mdk%paoj9p%)FFTDIAb''-f54]4-fJcJS"]eqS]fD?kk4r''3G<jZ1MP&kHQcV"uEEe-=HX''
`Y^6/H]lE9a4F]JgC[hA3i<oJBAe+uqH2]ZNu`kLhmm>49X"Tlc-mDSdurIkV\@#L.o()]EqJ.$
<jP3]IL+)O''XQpG1;V*VSXSOY1lo$Kf)C+F8A3/dgaa@pV<>;6.fH1.ZSfR6r<^$[Hbft=$7qW"
c5bND''AUFpC.aLeO#aZQS(;_IF,CF]:Wd0MC"o44ACg7;IIgS#g@J@HeXjtPVU&7C*O%1'';.UgU
etU4UpB6)j>&me*m9UO''$it+&H>Qu"lGmP+E$5o.&@goVWi]jD!!Fej5(W)Jj"UFAe:cJL!!+LspN
NPVndE-X_C=Q@K-1DO<^=Dn3J''<oUr/:JD>m"3[b`k_F&*s*2jXR\NiRV==#$a@;+*MjklUpEn=
;Xbg5E&C`53]R5ANktQLLMQ?7G''^KZ7#]#Pc=tm44sgd]JAjsHqeOV^MdVYeQa7/8UiZoOqQ+G$
Pej/I3X!!VB?B_''t>$M,H0`S4P(X<q/R2u#M%]7#NYl*m._7?-U0Oq=DT(k''AH*YC$:)^l0f18''t
==!!"T@u9IfY&TM&$?0Sm4RuUt2U,`C["r$O@BXUHg$<b1O#Zl2-=bAWflm+cSm5h]>f1(1fOMT8
XR''J&%#''"E!!A''?VEP%W`Y5iEhSEAEV66W(tFT_`b?nI,M\]m3eV/p(cEY0iu58@i^\co;hh=,Zn
pe(^iI:odG@$PaG=SPp)q>(L>$s-E$kXSmYYqCS2J2YEAYdQcmf''QG9Lb29EaoemA*aF0qCDmOY
@r]/HR_\b7!!Usbn#6:*,3`pk!!pQg(UAff.8q<V9:KYg`n9]dX''F\ph!!"!!265]g;d=(@SQln"4G_
NJsB?k-;R6Dt?#dQOD0e=tJlHidClo5_h_"(RcgNJ*#:&$J(A7lX4E<g''9M@?uZ5`(;LQIb$bbN
=?;^n0u9jlO+fr^<JlE/Xp#3BJYb:B@F7<cl5''tG.KZu0!!@Z@Naqu](=FTko==0^Q+ONBJAC!!8I
_n&Q<N)f)<;6S@G<mG//''Y7dZ$AAKX)HO"kKmSG@-<OLSX-@N''f$bJ;H4*jOfImjg3`nR)h[TW(
$<(Hro_9?d/<uM^2ulU5W$PPlhggH>G6$r:q)T7'';r,7k0Zdh4.0O2^<bq(s@$Ng*2:P_L9$`g/
_Y"@G!!>CM_T1d0i]m6`g_Ghk5_IIHVkI6&>''5/Ep^rsUu9./jZ=N/,)!!Hn+]+%T$(letn1BqtM3
:[Gf5=&^#iG],cHBO],QIeT:*^qioQ"3ZGWK%bo@QC<H&lT5>e4VT)-+mho)XKi+t,s:AT#&C^[
$;l6.U4K08A5hj"PQlM%b7Ts?-kuORD%8jX(PeY$=9].''WL3pZMdTNus#Hm-,HsY_Lpqo2C[aEi
b&:M]eB0gS#RJQYJhI=p_-42`U$sD5dWg@t0j)''tXBIPe>)54#f-QI_b9;uS#HJu^9$LqrAn<t\
J@au.?k<\"P7J;M_9-kP[B-lT<_.3I$3@<`FKhc/3sQ/&Ju,@9a*h7i''@ja3VOfWH9U.]o?G!!T[
.kD>niB@^PiSYqC2f^j]CqPVP:U''bS-V-sAbrcaE<n6AZqrm@<0^>pC"5QjglTW/-l_c,ek2O'':
$EQ8mbEHTJY]S4Y^gds0ol]#<_RC_/ILB8S*."^bIbS;"KB4J,,i6e"j:;K%LI*LY;*+mgW%Gg5
_u,oND[/pYmLC$*o1QV4X>1@U*XXn\9s/(Y6K.eD$f;f&E`]Jqqh#(-*NZOJQ5nQb=?qDE7i!!bI
WF)":O@$b3%6KW0m5?/\F^hFV:p!!O.5IC]UpFI;:Q)2UpmijV-+cW-$EA,f;q[`O%R)r2Rm(,H;
H3dnsc&W7DdNe:rC=A.;#+H/p,hKkQMDP2+A"AG[nI>gUTK_7GnU?AioUN3MQm@4[iWG6qn5s7&
kbmlP"Uh80#ius4!!7GA^^OYE+?:d*15I48n2XA1FDsbZoOa)pU:mPt`Kb362O!!Qm<EeXW''#4PI%
LL%a>oN2/&<-+$V(%9,#NPIkC3<j8;nN$?Rf)2=IiS-KNOEuS"^mq8@!!;,ul&dkVu_u^(dYFe+s
E@<A_4BTKtLiop/-_\mMWNT9&B417i@$q52-Wmr#JV?a)i>2P2D?l($$TE]LFX^;s;ElOXJKkkl
Fka"9_RAP&nGJ$+Y)G]@CrtJ;`[]o[^r#JHWQiA.Ju*6jAsR:k</oXX$2L023U*eDnO?JN#2]iT
GlW+C\AkOqoHS''X!!Lm&Bs&WWTXK6<Dk&]OHOgO;LqGMr66l$C\6)Is$lkf`i#GiE_b7:&rBQ>`^
]bp7r4.]d,%9[uf+(B"GgZ?4\2;lgI*8!!:"e,@0uo2^nU>uS7XJ+8+#(2-E]W"=NJI?"d#GEDI%
c+p?dWF$;.LeX"n@"euA&UmTie&khqXQ+C)J-65dekE`s67c!!i@%e,iV"5q;o_cd?e]>rcfN\R&
6_T.?(^a[m$gBA;3/Tm/b)0)s07h54YQ,qNGb8C&4b8;I=PI&q.MK&/G6Y)X_''3du97K//bNcTb
fkkVU6*i+5jo_^7=:??r.p,i\_m.2_r9JZAm8&V5kO%kgGEJ6T5Q"@XT`KB=b<k/ZjndCqXE0m1
r3WZ:$%R."(''\/l]&?t''W@8O!!6a4#jPe0JJekH7qMVVdEj-7BPg[S?21\+XFes./s@sqD7\TfbP
nNO-oE3Qq3$g-\"h+t]pG3BdH5hH)s==$>.R9?phR.l^jo%?PqXFA<rq.B0b/p4?Z5O!!''0%$p=(
Jp8Iu%79A''[R+l,"R1VH+moQc38G_''Q=)^0pAc^,ehH\J=I-d<V)8)U^qb3u=<ZN-O?T(Am''1(K
2j+DsEN`25TClNaGqAe#mO;8R:BN0+)hHPB"RG[*"g"/#''aLe($!!=TQ";38G^_/mlel\;SXFhZ:
PfR:X6_P#;HJQ2.Ge.eE=O`FmH%#jR(Vq$A:''Cd&NKt8#M>(SN!!>Yt91.KUdbiIKI^s_0K/09=q
,I!!D/6EHkuo>dk;duOM^H`Xn0O6*ntmP)hmCFfmWFTo[R(W^@HD7cI*;*qgZ;PFefkIX+O*aiCW
!!e!!Sf#8ete3ETkKk-.pcA)-eVJj+!!SP@=d`_k+s(cT.<-d_14nT$YFS=hKqu%`C_CKpTh@=I.]o
]EDKPh[UFG99rG%**>b9k37WCE@>''_%H''_gU<f:Kb#%lAFMiP[2\1UeSSfi+l%T;EI#.mMeCT+&
R$''RrXUAOeCVdl''YkW8CX''>e.%)Lp(=QL$#iJh78#V9?m/>X:P]E_R!!m:Ke/g=A3_GC"GSg=kWc
q"1%M3LPl6]jq[>7J:i!!YpajuSkEAFD83R6(P@#u/s<-7erSlA=L];P)%+u4!!n++(DinoK=QTn)
.9OXT"qW_F7<0oVP@ASZ@>ja#Mj_Y.b,q(.B!!nn\rH>1LO!!PG$#uVR%XpYBk@BYb=4dL41GWs,q
AH''>8Z?L1@kB72uFV[:S7D2h[Y^ch[eno^WfUmr>JjAJ:A*$I''Jq(mPq/JN(gKK)JmPm=MU`RHG
inYa*QWs\pFn4o)+/Koi@bkjXV`0,0YM$(%;KoVR(-6u[P''PZiSjEuGiqU.desajQ_8>;1Lk3R''
cnY/phCC-!!e13''RhCC]O(bE=0Yl8&>(>fVQh\Ej!!9u''Z"R"86GXE"QME%P2=)`oT''G=Fe&VKZn^
$h+5*H3dQA9c*7U`;ACL+3p7)-$'',;UGdH0#R2dfq%1XI!!s1XhU9R$?^fn1GV]pq,Y]B!!(WlNP"
.A`''?_TcJN)n]63pK4*&>fOs),f/Ks"J-&hH`akp//L5D5LX=;(II%`YU&/-k.u25(O8Z0AIcLZ
nh)49g6P%9_0"a42KP6NSJ!!$jVGF]I?[@P2NE+1Dr=*i&2@:(B/-i5?We93F.LTWe-o>!!DTCkmj
!!==\i6)gdHZ*J0mKVjF%?PJ(QZWI??DSNIn@06H^cn*''&pSjRs91(c6ehU@=^Xf:=J:6"==F),R
^''6.?-57-(ZmYQ_FkS''j%XJC]Gs>]a^qq6sqr:nr!!e</4J//^K]dbsZ#A`.-)1i#S''Wc<Y=QO/!!
FF<C^Q$\.;oN4YgQ9ufgJX_k">ZRmIJd$\DH\HAB.r;6Q3bA@-7pqGc(/0XSJoG"9&$Te3_)3!!"
+"IIZT7LEXDAp[K0O4oVpK)#Z,tM=S=0K$0/QAEe3nFf)Qqb6^VH\Zo*e=JV:Cd<T]DN!!9";Ei\
''e\3Fn5q>)b6nWAXOcEq/V4l^.sbFK"^jmZC[nP5i;Mm\XE)f@PTg"!!2d)bu/DLsk<8,DT,8;n"
+[$O"dg\QeXLN@/cZ0*jNs0=8A&7j3(@FE''=q$*bmtJe+L1$/AemR$D^]GK,nGZf%Un9DE''mj-b
MiS$WgHKQgX9r>@QI>h=!!/oCNT7*f%,kWmOrC,?R$s/(hH!!5fa`qo_V+]u;Y#$1%QdMqoYe1m@$
h!!Ze\6p?B;HVRmcZV31k^''4rGf%tM_;e,V::H)MYlU.c]Koj]ae9[j'',bO00qZ8"Y@Y2#*)#Fm0
@$Mb.''bb2=cRACVia"f<Z>T`/rCmH''r5iII''hPjbU0#YrkS8:4`e)e#c,F.sij!!C*\I;MrG>"%#
Ju.LXJ\M6JIiPOm.)`o29u2^0!!BBUOqV4CcS:TeVW6@nup"[:-@15^&J%%?+_STNm!!o1C)a>\dh
<LX;\RTuG:2'':8o>LO`92a%Y)Qbajf%@40R:EQ\3o)oG"Yf=(=p4Q8G0KHWKepa*a)s/&3G7Kql
[nQUTGc,A(pMCqP%#"^_CFqB/K!!jpZb!!+0>A;2Z@Mnjqa>.5rbnpdGN"l43g("L&,*]sub,Zn!!)
"jk#Sj7U4rpY0AiZp1M_fK'',[JL*hr/EbFFE:L*kY_C1]71jmueFVg)r!!7f^XgN>Ba3pCb*1*6N
FraChJgGTa@kHGoh+p>od+Q.R9]^ad^nK-R4HD?)9:%iY.co;m]`fs:L`''=##P^r9!!@p=!!:pE65
#-3pe''0n(MPa->X.1L\2JYr_1*_QJF7qFKW2@#oe7:8CtrCpZ''9P\%,?aHQnn%a6q"q/U^[0]sF
''eISF_G:TAnf"T557m_''S0WM%Z''2WNfCD/eE4]CZk+2h_CqA&<!!7jhRj`6H/i45k-\mYT)5%7pm
hY54KRXp/Tq0Lh;h>=J(DJ^+>J9A]j$<W8/0*F*_#NiUlFp_s/`''^5F:1&2fFmK@=`X*jtYa&M9
)O]^s_IMuB"=pIU+#al%W2fB?g2mmp^2cjd$YG*,0=`o6ATR[''1KZnojjk*HL,L*=EhQ47/&>Do
WA`F6dOr+F]*+$Ma`X@=%I/e"T06K\bIoNI#Z`Aj_9,Sf0FU+1;<K"D)b4[1l*^SXDh[MtmWQ>:
$I8^\A$E77dMgoQ3!!=3hEZJ7t2HN#,]SlrZ$Ng2nQAKj!!HE7C+n;DJDJ''TTHHl,t^]F1qD#7CS@
#@OsjJ7]Abp=ss!!cr\%5jkaBR!!#Da>X-EL5F]#euU=8/+JTrDAYL<r,lTnoe"I<5A"jAh>HKV3_
5le$pf3l!!u)GAK7f''H;rMI*dfOS36%_.&!!R%Y_iP`i?Soj;]`Gr=N/_5^=m\pi<PG@NN_.nhZdB
d^dAdK,#HN=o]^E)ShY(cj<U&0=W\>lUAI]Q):V2g''UqY-0Z9q[+;T7f^?.<TG%N1Gc]!!VJWgto
$(7BGf&Y_Q+9am?Tb8c+*$^*r94.e9525f6?ghREs&E3Cngth4Hjs,EKgGc^%BKV^<Gr+Xa^^Sd
4Ve(2=D2:(YqQ4eHaYbP2qWhQmcAjob"Z9''b%>+3Jf4d_W4I0YQCU@/1m<$6f>Y:"*dFd,JAhk2
mhi-(b3`TE.9OAI#-3EtMU*''^^s_U"YY8QSPE2tUTlcXW[d%GJVAH>uZX,[t[6NY7*IH7N<H%>D
?/[E[XFh6u7+odRi1HYO8kNmDLNKV/(S9Gh3sM`\`&j"M=?-YXHM3A,Pp<N;F!!$+K)^9e,+iLu=
8:fDUWG^G.-jN9P]]^(?CT-Hb:O>dia:RB$bCMBSF,aaW_`#BmMeseCnlX5a2,0h2)B6R=nIG3J
>4[ELcB%Bt_f]>r<s$K4m;OXZ3p.!!1cfqfX8hUqcNTFPojrH)2"$Y;8a3#r[DQUYRQ4@:D$?9W*
2"g7OG`7RASKG<]=+\spF)Hc,YR<"d[,t[G1T@JsHE8*,@C[7,]Z"Ahm]l!!&Y(2elhsfT4=g*qU
%kOE[$C9Xm$l4#jp>^H\=MR]BXR-9E4jBA0''^u#''7EZlW=9GVZ%+Zg@!!a>gVYhJCXW8Z!!A=B<36
oRjSRd99q)Ga?`,ODO=[''+$<pFtFZgg61N:0@`t"#&2jeApda\>JYorI]0OBKI''_%@+Hies&F-$
AIV+_HOa9/eG#GP;-.NK]r;#Pe_b_r3?*CJ&?#S(Q2F"5(.q)B?X]S$\GaJA6[`G<`*@f?LFVfi
_KUO&$SQe,!!/lfkPeZB&$b9"_^l@&CM"H_Sl"(<%^!!?TLc,L30NIM@Pap_D@''uE(6<&XeWklqat
E3QE-B".rq7?:G&R-pU&GKA8b''V^hhEgX$dn"XFg,$4j?cW>`VmM0?F+i(B,*RBVTa,5\Z],e_/
^*dB%PCjU@Ylm^lM3=n-%9B"KH[pfmYWhM`Ts]=P<kfQS5Bu@!!eSbWB+HPV&CF%amNK=HE]=NG*
oWc:%$C:[e2=[I*Rp`KG_><er=uJq+_gH/"X2faDKmMGEH1B*t"ea]PQ>8&5%PNH''eju2G=GM/Q
,/Aejf/8l5qXY61.,H6Q>nk.=>_3-M=<Hufm=^>9b$mOoPl@Fg6i+k-YNZ@qk5=*lZ`h4=&HKFk
jkO@6>I04r%IJP''U5sj''^gdg2<e@$cSn+CKpJDI3.Ku!!uE1FsPoN3I24VZ?&cinSU>c-b?!!)$++
n6"7lr,Ip2Wuls;_264\9,5"F>m";HFH4".^rknT3"Jjc8-(RP/(+k:Q?h(lO/8/90G:/`cgJK`
"KO)+<?)$X^j?*?DZC/(JlPE7*VFCWX$^jj61:@D=uh=Z''!!1f''Q(J4^GW]*(:CXWi>SV>*CSsfT
f_EgBVn$_G@tb]]f"iOJ@FQH#S4g*anZIj8]L9o&f%LM(c]=^gJEDQJSKk?E.-?4+rG5jZ(c1^]
=Kmlp=X^)P/MDQ=A*),E"0g)&pUnAX/S;!!(''f>)-m1if6NA8.(33im!!lES!!`hbGA6n5s-NFbtIH
56/LIJC^T-gJ-SmVm5omF)Fs=''+S&5s0E&>pV?B7ejA+\VeMg/`aJ$6eJME6YsR9-^k.+pc+H,?
YhUAZol7'';&Zt<&ADCGK,$<EJ17rGJ/?sUc=EUk7\]To8!!NS%!!K_CnTLs@%l4jGkuqSh&@XF%#K
9:s>OeD%g^Q*@0[E#C4bB!!A@F"r<99_ILDcW-D''u2$Zc#>7O]E!!j&[1QC^;P/<#A;*X#N(''e25R
Vu&"/kL\_%rHWo>:p:3@opoilaRVT^DQ''.9ZL(aX<,m]Mqbb=R7t:MkIj([soa[YJ5kh_lb:>dB
*%[/rQ691r9.,h4e\.r"IgN2tT\KkI1M\G;R<WrE;<qVKDNK6=!!M6@LHf9HlNbNt^*\.InK?DjF
Fi1c$Y63f]BMFmIQ`@2om.P6lR[s^!!4SMcd)U";)<8uI.A,%1R,uma:NLmlh$WE,Zq@G-#OHfN_
YR%L8h>WSXWcO4iG<%U8_K<*bjOdte_#c,Qmn$t/XEBL;@"!!>1<bi#-",+EaYl+D*b!!WC`\]5U\
"0uqfXC8<:W,lgarU3C6_Oi>\AHEVUXE"WKk>6C@XT.X67-s:K<"6VW/><^-JQ71D@,\eL-(s2L
#KCdH''sqRjhR3r\HWJh?B%tbVc3:HEA$nAjs#+#%<:Tn98T*nUVb`q`P\Fk`bq8S(gt3''"dLA-e
-$2HlW"__a9qdfm<!!Fg4e6!!!!\U2_2)jcsYTBojI>%Rt!!cFT;HnmsSnrW)o)?SH''GDH[7#)PcpEo
S]CU,JTDIr6N(;1IZ3Y6$3>7Z\\3m-l^I5XZW7*<@QD9a#cr:m$P"PN^`r`SSde895mI/XJ,])o
f%Y<?C'')K.@$NJPCY<4"K,*X*Cl\cnV%*=nFg*&c6/@&DM"Y%!!J/MJ_=g-O1U\%.oqW6poG4=+)
!!rrH%=m.aFTK&6h5g(=/cS''eP]I0pU!!e`XdeOisF$T#"j>6_,\J;''UD=I=ZY\q`U&;:S6XYhU,u
/BFCZnb\0*GmB7+Y],sPrBJf&#?$3u%##i**-Acid/2_HOr\d#5Hqh5ebsjn-=kO=NWKgJcHh5"
UJFc;:po[_3?mKDQd!!QHGi1RuOP%>mhiFdZB3J#(30l&"#4"P.beeVf(ORQ3-k^MWP9)7m^uF4:
h2g)1QtZ%F1Z$4R7-/u_o;j8dWgKlf>6!![>-<0ZUAkZ*6D^N2b`k/[VE8+W;Hlo,g9_-$t_/*am
Yi,)sR5>\2^ESMOXLQ\\&1-KlmTCZpqHLp5EaWbRdXP6*LYaBICJoA7gSQj).,>-:dTE):R;IPm
>!!/]]_*W;4b!!A''XaU]<Y/E^b!!S%^(K[L"dVJ!!qs/*g<*Fa"@rGh=9c>Wj0EW..h>fkN@S+J7u//
crf;qq$LNK(C*V$$C2Y@<f!!C,IUHRmQN=0)*su$_f@\"6Ak4Q!!QcCEUc)qPB9s;@<lP1<6O7K[Q
lCUL9-[aV@#;_tCWo^Jf/!!n[_[djoml[X\F)#(KNe1j_7&csR''!!,F!!FCHdP$I1hTdB\#M56B("X
K^W9bY!!Y3M*!!iQeFKYe#"0B2TR_$OdFl)V.c8k(^s''Kh),lu_Af+&/3^dr306h\^$b0jKCK9;>H
&aSBbYN<+R04#^GUd@EOC1VmG=*#CS]M:+H2jaM<m^6<b7Ug#5]t1lVp2,.\RWZLCl/!!IPK''s]6
^&u@V:qdt/rr!!0rM"OZU$/kcVX@Hp@mP@18_IO_Rpu-K?jdtm3+"W=#pST$KZLO_C/\Cr,4.aKs
bYDVm?qK4I"[t9hYt**7/t`"3fgiK78PcGXf^usN[HAO_O(+^\Fldn?c0dc%`]h_McTcj''^O4.9
095Y-Di1%%Gpu/*f0!!(^''?cZgY?:DbM''%_WB%j\u@;t84WqQ-[mj?e?lW*cK$VGHg@o[UjdAj3H
O2Zr-?&Prjb=MA>c:q-!!Jq6+pK`$#lk5=n=n''+!!c7p6Wn&uY?Ml.L/>-(Uf=<=hEU.VIKZ@9"r0
5D?lP@q")_M8;FNfJKG*m[Ab+OVCNtX]eK(Ndm!!EjW710]F4qr&Z`%Di[j9ljtQj-h/6PD#n''G!!
\<anN`O\!!V!!UaZ@FMX3q9uaePd_9bEK\2g#-!!&RUo\.XV?-0FGLB''W[l\Zrb+@0-ke!!<>j1?P%9
Y0YEI+frPQ0AaDHTT''+VYDr%@L/8Y)^T>S9[_a($[(;d70)T5(C97pQAAC72@OhbRq4qfcO;XFW
[s;2dPQUeu<h66uMY%&rMh%[=J=pWB<.FnXY3Tt''NQ*c^Lh+_(qS$!!&/C\gke>D[5FS1*e)!!_a+
SE*90ALi]kMAEI\;X[G7%X_.Y)<ahg>7</VAfEB<!!TK9)5lX>*EYkuS4u'',`(-,XS%NUl\!!QgR.
4XBi59=BTq_V.V`*nE?N+K^HXAX"iQO3Xa6L2Fk\V8F^F9SaF5BFEpM*a`EAFpQ2L583DY_"CNL
!!4pEg4euT3V1%04#&oAWnpsnHLR2OK[h7qNGl'']k-Jcf91`PJdIH$!!5&GnUnO_&s(\eMpFHR0!!''
_E:G()mPRX\e\]p!!ne$0&\%$14iMn75k7r#"/''-s$$c-+@<iun+B)!!jj[19Ao+si]AOTXif[P4X
Rm;Kp8g^"V$(F^)&;^UB?7)#7+[Y"*[q5RP2977''.:;t:J2Y.h&1h&]RK-EGV<AKfMOk<8Bh!!i_
TOcZeJ8+J$q`i!!5<N8Q(&&dacSNnaL`6C_\UJu3Zh"df9%:[:*g];lYPL6Y&8Ne/Gge0,]N)9#Y
?neeAc%3^-:mMNCPp3JfWk^eOKg7Xo%1%nIIo8h''pDI6A_RY0,47acMEFua+dEIP0%ZX"!!F(W)?
II"Li\"CHc?giIDGO[ToALIk9&D>n&gCVauM''CImn5H#W-rcI:q]cJA7<Y@X`NAgeWU:Z7_[g@B
n6=oge]4>F4H`E(IA@s.Tm(PBB!!;Vj.nJ$6Ai,GFhj"2"0bd*X>qD6Jn?ocD-n[WMJmK^W6X_nd
JYU$%C7Wu]G*YJm=DC$Y$i7*k,\sO1#-:4:jgG6LDBY+!!7oflG!!1If>5.P80<\[]SfaNSq_4^81
_ue]<KHFM,<^\H:J40I.7fRZe#17_''*&7mcRNLIkHprR>P@/rZlgnLQG^)''L$NJ?fX>1"G(9CZ%
!!\=d2O<U3Jg^lul!!m#"^7192Y!!<IEJX4@8DV;RnP\RL(h!!S&W$BG*u75N"*h5=nr?dHHp)@/uVm
YJ$pl%i.m[&?PA`-RTM8/OU!!&OQ>QAJjJ[.$'':fp4\W6gX,E5S,$Yo>,jH/H\cWD&JHb,9E!!qts
pUW8fi$sQDTkjej`2iO`=W7/PS&/n3J$QCJ*X7P:QQj=cnpCf@!!tspF?t?<sq9+/K`Lb>^oid[^
CKc:?#Z&]W\XC(].n@^0_;6"`:&mB[X0=a^8n#M=Gp]"GT7^B.!!nPeN23%T2EXs_,4eGm@L3aVS
YRC$n_(+OP==f/b:?!!7pPZ[poo*Y+t@0Dn>()hZ?=_;''l6XVbd?Skl=i7VoJS](*^-c)6@s(;b9
6]&2Sb9tC.KS$rZYiZP,ai=0+YjDC4q$:(>JOKXJ2`EG,C8J\B`7ob8EUN_%^r"jAE@RHn(bDfJ
Dj.n,en&Un4=ktHEuP"LcjlA;T4pD%!!`aUJMaT"l&irZVP>)kY+a''l#?t99>r=QsV/*("YJ/-@_
dgEJ1Mf>sk;MPB*rjsU*=Z/$EE)$0oLJP;SeNdf5!!"e5%C]]b4X?<j1s3K0HgJn+sqS/8pIjYQB
@(CgfM?''UPQ/q0DoMtPq=Ja6Xnlf/kcKb[J_oZPB2%\^d7nq!!*J*NrLUMCM#Xi*Ou/@8_.c=R_,
k9)->\BMc]\E;3=f*ti)(T]@lMoCNu@o,2m()^E];7_jYHkLF''%$RsYC#rS@_EDs<JOd23pOH7-
%T_aHGr"U\3*#TrYRX,%T7I_dTi^*mc$ImK"Ughhgm5,B(r%/#OZbN(+#^FT%tK@0#0Ud&8!!NaO
_''@A-oBs0-MNO4c*?c+q%-!!M3KgT5[S4;9l9=P>-FUV=NK<uT,=5a:Q!!3:3L79klMTTG%%;R-<R
YF4Gh^)cgXc=[gK+9sn7DWGt%eg-5<O$s"L#F;f+Pn@,?;aUc4>!!4fP23rDad$)j.=TH7bN"6(]
7e,ts/()c!!NP7ZE]%7B"BNB[ZK[X;rJ1@$m61V]sKuCi-@\&M''9sRQ)Qsn;:*hRrd*Jj\J>ngZq
_/eek64qerZ74]oTlV*%A_q`aE=+7;lVVsZ-Zl4J+X\"l81]hL?*rnC!!t1_)ehEjPb<*N!!Z0;4g
!!(a]68lG-L^(4713<rS<Z$67"<,''lFijNtnHjhRo<+%8ob8dV)pJO<2K.gf9?Z5=6)6=ssnu5_;
qF4kPFf_rF:rf=1/!!a5Xj''_D+[D;lh.L#c]..lb:&\#m+,B,i)I>$k=Y<Z1J,@HXu;c(<5rLLmD
;i0&sd^3MpI]''LYOA)n<`t[dQi3GGJWrPp`rs''HE^b`3FNPpDf-^<Pe@%$r+&:f%pSB%fZ2q7NF
44]1Bai:$g!!g<iY7pRIgr/=^UJ]7,]m_LqE''S[:ra_V"h3gP.H;IX2TVm.W7H8tc9!!6FY''FY''2C
em66hI%4A+*L9fT"O$EW#E(%VhBT:''nG''T@!!NW<K(!!m_tqDhpunRU#@cC`k_[.@[rMZ/*j<;Y=F
IbB2k89bj1*`rJm\39U(/@$5s$f;":%)A#9Ju''&]K7mA8hC_]++;7d8!!9l:D,4<$,n+,Ic''J)JD
i6l0Jkl%d+@%E/R+&tE''(6\oB)m3[fVA)Z!!Th:>@Nc*Y\Pq40m4o7,k,72_\Ltb_AW]aJEThmLL
=b0n4>I8T^\UCqp@C\=8I0*I1WdBP$df0:TW"[heIe\t>]uiTCC:uE_4\-WSr_L3D//HYHb?u>_
6*I0b"2*=`UfCA*,ENV=-P<hc?@XnTX9?l>Yp+H+8YX5LHQrA==<s)hh''F''tEWdrCj4OQ;`DERV
q%-Y^4<"0J#aWY%H$:;LAo92cEWdrY5/edm,Dn#tnE<2''#A`>#Kp,[+HV$HsV=>s9D''''/%1''.h*
!!K0(3:#N2CG\$b,=Jqnn''A^CrM8UV-Vj?@CkeRmWRj^lbYY;%;>ONC[qNg"s=Q1#GB%;rpK`b`=
_?$IebjssUdp@S^AV\''a<iNNpYe%KUY?fb<Q+AW/7hfs8,=obr92nC/dHLTm#aXAS&&:ha^aB:t
2BtRo_L9#g@$jJ?OG*t''!!>`YfLte7rR!!Sk.Sd(*`GlWa*i*8Qj>kL:j/R$)91+5\K_JWDA3gNR>
Ll"mbM/35+$V@T;p"i^@I?*rSEN(\\<_:W3s1d-7VcCSQc[gd?Y[T>UHXMs]Q2>fS9F>`kf;M_E
j4DSH[EaA8])bLOf0?i%c)^glYt-tZ$`32<migd!!>Hh698U<O_9n;l]p\FkjbR0CAKh%p<HZ+49
Zpp^sQPP4T*W>_*H$ruT#iH1n^"''h+@NAS(i%,$oe2p>]O.QXmAmkH4AtPm?O+L@q!!`k:<ADY*7
>:8]$5Z5?9K#Q2HOU%W\bCN*4X/mb0,uuW>a>9"_-bpY3R^f"1##!!:XiX@Br6X[G9*s4"tc--XK
"o98MKa"qN&R#3#no?0[8K^N20u<\u%EP4V%0g3K$C9X''gl;<:bpkkB!!%5Qk.E#!!^$IVd;\&OSH
/kJ!!X\MqLGmHBA4/II*N=N2UCLq#&3bh*<Ma%1L8!!rKM:;tp"?i6G:bGNQZI4??AGHoA8@F[62R
IIBKsN(''b7nH%^rH''3CTb9:Cu<!!kXJ=<88&REgs%FLFc7V:I:M5XHG[rs+@;qc62n&pP%jYE*''_
(72.bU*D8]eB]0=Mul7`&>W,X=;*@''3_D;j?bj6#:s+(1Q[=mL]!!b)E4!!bX\1"''?:\ar;540YB=
5>=ITV3_]$<eM*[@p+X6\_=%)_(k([L"Jse^W1I"P''?A&MX7);"0%[!!O>L(7`V^4I$9l&l,J^U>
R:$J5-Q^X>(4YpN`9eep=rg`*%''+f^EDPl?PF/+Q!!@nIM^aeFp=MRhs@)Ws_rhI@@iB@]%m(nPl
?q''==-S.Z9"gIO`:e./kX=+d&">''.Kb/deq]#sBH]RhS7AIkl/!!/h_7qtC;rZ3-EQc]5iImqRGS
lP;#aOtOr&''DK\U=NB.V?g1go?LY165X%6V\1lJ@pVA''Bk+U.mad]l\ISDMRV''_4k)t6cS*k\Ti
\8n?9jq%D)Zb?D0dK3YUW(ZL"Zu^<#^"0clEg!!-(dgPXr=K,dfV(9A[?!!Z!!u<DTip.]-J:o5,aT
^W-''*(e>Ylj@^arW]..e?94YJ?ef.Vqa9>%]b?B\F:u<lcY.Sg?8#=%!!hf2hi(mg1BSeCuPhL[g
Sq7]`65#B%h^l;J-CGi''Cc-M[2IEq%H(<W@&.+]6(SEXQSc&kAau`j+@NDOrI%OHQ!!F)d;\YD(k
8?Z#F=R@H@?4MZ,+FbrI^A+:U5=P(Xf>2D",ut_gi(:n>ehYgc#?^''gXR]f(C0`%=mt\\.AHZ$q
d[u!!3/G:]/_+JhnVY)(fM01Ie\?n62?fUEq!!LAeC0=AM5XMW>h7VYUc8co9"hVaQ)nQUQErp[])
8-W1-#hIb,QGVnmO.QMT>J!!]md<oq[_:E8&Sf@j2cbT*\:)f>E''fVE,)@Kd;O9o%R*!!:R7+9A;N
-M:c+k5rtU#8L@p834o5j6n8--!!5.9E3AJ7Z!!(I`;qA"X%af/<eS]SaT;%Fp[C8a0ma2ikGOSf:
"0cMm#*\BdK>DI4%4%6dQ%^+CWde\@1a^BB=g7[&L!!3Z=keP%9Y_H(,?TcWRA$G/=`tL,A3Wc?T
2K$#5>hYWtF&dF/QHNGX`G"H`Z8TL;PF&&fSj!!?o4tCI$^\IWO0Dm/!!;.q0ldnE.F&e8UH).7.u
=_LAgXA+uaTTYb?cA''3RF7seE//qCY_qYV2a2$tMn+qLIfA$pGlgJLQ^GE)R",lII==PRu$;Q+O
%t_2F";JWg$W$j6J5tG4<]pF`3=;)0cbNE`<r!!`>9GWEa/!!cK(;?R4aT"=??[<0G*CpSdID?*kO
j43E+m)Qpm!!*?.ZJsREE-3/D8],U;Y=<&ind;G2UhY:NdEX*@TJE=W>ocBQHOu28C=O,4m!!W]bA
<hHSb"B`AoS`a3dN,@!!#!!HPe3G^@BM9<_kk.MT^enfGZi^D);8SP:uKYLa>[%Yba,)3b52\^"Ui
>A=r>K;rC2j^W18-fqPhJ45L6g8Jg5WDX"br_r*jCP4#UJed078ELJ&Nm@<??.W80q7Br:&JX_X
D)"O\f2\hDS)*;b:''J_-dpB-*I!!4;u`O[G=O=6bmJu"/AOE@;53EYqW=EOj"+38J8g)kk]U_1m,
c#!!\3<p(j<NC$3s_#g8V])D;`bq`VI2HZokb2V=C`^>q!!/:o,R3&er!!/-N&;S.(#]5erCF%V\K8
W"-BJMiK<,X`[o!!n-07X7T6d.(.n''o-(sC.<@/KLM9i"?hEMs\RUC7gk]Hl+EVVM#[=5lO*jjrT
MH)`/Fb-?(?#G,.^O#n\=SKZ&:+qa__?<+B1(okJ"WHm13EVBInQl:?.)P5;bN4a"][hC3"g*:T
b&1UY._-Ki%`+<O_-1X-JqX5-(k<@E2-ZJMHQ?lH&)D_T>iL,gYMK3Y!!#YTO$Pl>PGc\/P:m-NN
W;[rQZP;DFPuk,aKRagC7:be3S[>^M$YlrQ''dUIbdKhBf`_,mhbh-./=hCdJ3*F;->_Ms"]>l?$
>TJIYl3kC*-/JE0XTJDr?4pPHiJpcRJu"G$Hd!!d:J\RnF=<XBLj!!h3I2$"!!q:sUZ\;m84]X&m-Z
gOKT"Z`k"h*@u$CaqcN\L;a2*dLIQpHL^//@/4u<]YF.tfpk]WH^noV]u4GA_HR7I`?TZ-)puXl
XO&.Ur.BFa(C;f%Fk[@(XZQ[nmhlK0d,>+(%Y<__]B4hgYW)m20.T[M=EH$t/G6i>jjtK&Q"/Fs
^fg,oUf]r.TFQM6O-%U#^c:iVP/Gcd/6BM558]e-B[qAJd2)UHi<c_B"q/0JYf[Q^.!!%-"q%JcX
IRXAAW_L_;c$+I0PO;eg[+aUY:U&7IUdBZ-k`LS66a/TYr,%k''M&udLg=kT:IGF7TpOcikb.''<6
5n"Fk5U<$W^(W&V_#d9_""Tf25DCcZ#2c83R3X$nf6(d[9_QqM0:QHr3hbH:F,B-J6,F]>[7"!!8
E,BMH>LOj"28H3u(<l>HYZM`KL%L>_)RJ6&MeBjujkX<e#1Defo>:''Hb@rrOKlTC\JjC[h!!5D)\
iL7TbeW\`qF.H.Vs!!Z@87G[9F("b2%PQ=mWhaN&a!!EB1-Kcuod(-?=SG=gA9>n@_QBu:YM/fQ)*
0@`"@_&DT@IDE.''kdjnL3l>9DT]?Ah(0%1?BWh";L$i&0qg/BF[TY[DZ)2;`B>?iajn^UsS9[^+
CpK#8_@a$1.6%=i#_`Bh-c&,X(Rb=HI@3`7d=^PqZ,EsGSI^GR3ZIdJ@&5:*JAnJJ>[mZ:JgICE
)Y18h1Up)ao()+o"sH2ZXatpB=;u(,HDsto)HcYg_."QP#$,=gfs,$=$''rcD!!mBMt+Lfaq#7F),
16rUWVQ/<l>C=pl//\uRk&3&[*YhkfSqL.G4F)5:5mJbu3:l[?s6f3SY^9>+`>#E%K4XA&J!!d;#
UuXoXNE5)]4g5:>K!!i8D+"J)-e?SpK`L`-:m6!!S+a&Q5GB80g)e[qGqof7+7/cP5=rV[4.3!!,BT
:*_m)3!!%3CF"":aEWB.(Xdf6dG`Z#m8?<#b2+IW6?Z%Uj^-lm_''8!!K=U]@It,kIg:E6\FjP4kCu
S.`jWQ0OefhU^IipOQaP7td/oZn(ZK@UXa7/5O)3>Hkh+$Wbr4AJhuuQ!!,iadO#DI''.fIkXIN-N
Jjl<R:Zl,N)Xm,BJJo+a.n3Dj-''6GAefJsJl[.aL#8(8@d6\WBZbepjYR!!11\iq7nB?S;Ad1:!!(
''mke`F#oU[/B$-?$R@qOX>0gWmXYP`mP%`,)-DP?K11sfqrhRF2a/lb^`''<AKb-6*UCM_UXE\2a
''/$u0aqRi`?k@aj]]UcMqj2I^EH;W+$5gbq$SKAI0;Pe%!!O#EBb7:B+(+`"a^>@T8ol@So\Js)T
<fp6_0+g!!Gp[rQ)5_W#7BFs)PBB%o:Y`?=.ZZi[n!!O1o82H8Nha+44<7nBk''%P6L?p2qp,$l4DD
C0d"Fdskq\jf[I;.h-VU8$>nKSWdd#<nP3nSj!!;C::,N3ZB&DGYWK)+m7ia%6#uk(liL<@N<r0Q
#8Wf!!UeUdJkCOE]lPs$?oe2SSHiP_=j/(O>m+>2/b$8E;%oj5r"D]]0c(<&eHpb,!!M<-,`52,^s
f8,!!SMZ].10EY3n.!!0f0^iJnmr^%Ldec$?7Q9opuIN5F-n&mMkr&r*:U9rL#..j3+F.*2*#F9m,
XoQ7t(H9ApY^erlZs&FELI1R@";^KEY]()RiebF\+Uo2Yd8`^Za,jII/gOYW>+Y7]-X?MA0VRj^
i"#-V5\F<]?0J1PZlRlXh/7h/=R:c/)rE_`.%3^?6a/PsZ(UR,]QuCKmb"/YgTb(Z.@+j"4:;qN
J]72n1VT@S$B''T76<a`s@]i6/J2X+5F*tXds!!u2mQ9pbmE?Fk3Rjd7[lrXR[bDTN6Jb%oh^W&m`
B<?+Poga,l0n&Ll!!''<PV@=m%(I0%HVOC4?*OkU:k9MgO0A\Y&$]8''dJ"qY/_QWQDD"0bCJPrb4&
)1Y0Tmf7.=TtdVAFN-+.dMi2QK#PS\2nVVH*0OF[;!!HD$X?M$[BIF;D9Q]N3J0e3P2*H4Zgc;TD
W;cth"B^,$f%_q]WslcnjcImn?[iV</4$rW!!F456@K\;iG,nM0#?K4O:9a4^_!!8t837[7qb\eOs
J6''0N-HlfmkZC?j.ep0*-''k_u%MhNAiO8qj/e4EW/$G#%WrZ''Sn4SF-^i()A4G)ZNH0#aAac6XZ
"k#%,i.1A^#&ARnlcJ\WoFYSa!!%YX^p0bGaUJ?QMpKg00E<**-rg2oAfSl=K2+GkG=Ch[PRJ=cS
kQ;H5*J&o.l/TPa\2RqgbAZcM$!!)LH^5_mY*>,7>8^EjI%9P^n(+3McUUVLH79WitWOs1Oit?,Y
J\F@ENB8O#J/tku!!@sZBmRtdi\nRlGi"T:''+*E_jTP/leJ`UtA#A_$uJjbl7Z(\0I!!)cM2^jWH6
Jb<(hYi\<HY?Z+M*N4Ge/\i9bD=."G''lY(moX08NBP27s\?5"RPG$3,M<X9FqnXcD,Zn>lfAPkQ
Lq63''.+d+Wj,u`n9LMcfOgP\7.\@&Kq*nbS''M*k1,@E)*Yf9:1/Th##"P,U:MhQjKo`dG),Bgm1
%XK.rC_kL.6Qo7j=Haa;Ohe:rM''C>1g=mgaensEQ#ip4Y9UYR(prM5ej#q/R4betS!!36UF3i7n#
,.YQ/G(:=@(`[^,\]hk;">#R]e-44.fF/X/&6T/:SD><XWP,h/o1S9@:6Hn!!RFc[3SF23C<"F8g
4MnQ3=R;IKqk#fp4%E(.OaeVc8!!K*#$''o_8$t:#Io]"oucfN,K?h3DE5H''QG9I3dnafWC2_2>9Q
!!K.g^C''k^X/@%Yt''$e<3%>?P.bk''Yld$k)>Hia1)#&CLh63*Z%=AfeO,?k;dh4ZMD0)pBRb7q#t
$E]i.?4lJG+88WbEUu6aFnV=JT.<WOK%8$?I]t(,!!b:3cRQm&i@^7V6"A@AEF\)YQDZLUoRY`:6
*gd,a#V>=m,sae>cf4NWp!!nDpH<\<%XQ#jOs7[#-JAYc?60!!B$OTCaa3A7F6$W"Z[lYX[[h?Rg9
L\j;.FX")po"dh%JFXJ=<H&h17Ek_nI`sfS5*@1F7W1ubXo_R"!!0hL(/4ncCbXSMcO)5J-YXjMP
R(N9g[eg>sHsq)K?b34GoK3-)iXCpjRdZe^j1K*lgl?k!!EYS&RI!!O)D?95>?!!TPji+qB0h53c[.
bC>b`8NcWS9d!!X2ep3;h.pLJX]CJ^CmO?MrDkHG#C[/EI7uo*$415Cp_EXj-h3[LrBFL2L-A*&;
1d\WY[8/i]O5>*:7pd`AdeC.h!!iUJHl`G''@)@qO^gQZ$--c&,Z@-$kRGfsmch=d41e;*irfPS90
2e=8P-i+NZ@[Vcn@F#^o<_7]^ZPckUDE@1oUieFC1b^FjIu\#H!!<JK>)/A1],Tu:_/6!!1NJQ3_a
TaQL_$WajA4G;_.O<%HAj"::>$Pr4DYY''F78tIDEdFK[k?.V`\d7eaP5iBFj30Ta@D%TbZSq7\E
iQ&_HE5`Q:03FZq70rGt%DK.f$8De/WFYW<LKbR_b?Ghe[XA9oI*2Uo^*nc+4RXc3oO(hSRRD>%
WsgJU3gT&=rON5LcPFnHmOt&5gQ->S+F<._beZn1J=A()(MAlEe9Y8)f@Bm-3>i?T,HRt(O1NnL
4l"YI;DA]!!]^pLmk\Zl+HmnGD''`+M@DNn%&hK(*hrgADH1eJj.R@"r9&[7!!&bN19W%kGc/6e2r&
m=83?VWe,_FWH"O:.PJ[CH_3*!!$\NGDaR14BFI3$^0OA392GCA)!!5*$m[V3&YY;G`m/e,W"mFuS
]s7kN"0d0%1_cMdX09KQ7g''+u)M82mW(q1AJtFKLR:V!!30]_[)b?WY*A''CUmc$f$R89ZphcbO#9
NBR=X!!["E0JAkY`\_]iI*+VTC";^]=@1H<u"BNgVf_OipP\''c?7tcEYZ7_&VY.aabN*j61rl8"P
>D0cu=Fa9LO9Q(<82f:\j9NL.-Zd^.\7;cGMY42L!!pa$8!!3FK$&7sBKV4=p_;3qJikl0J9qO\1<
XA4OH[<:0q>`r+oA5L;OT4e8lS^Y+Q.ta/V%>Elu(,5uN$''pK?$37WE,AF6l>L4WtgjnL:e;VM''
k`%BNGh`C7\p*fb=LVWVDuidW)1a)!!HXMDp-YQESW(ZD$Q2jp89)$;^a+W/Q''FRG*clP9Q/1;.4
JVPDZq4(gnH$:/1ametM+'';!!T.t-5qJr:18WTGeK;N#oj>l8Ft\1[WcA>hB)/FnC>\.Q"`XS.g-
3j@I?l_RP(V.3m>;M)m]''/*Sk]P?dDd">nn1cYSp!!aI?''nHd]MJFt=qCm`L.[/WbeELa6$JfYb]
F\gLsh5^DcgU%Cm-Xb29&I%5$/7Vs!!XLPs(^`k3?.f^90+P#a[q\ZD,dN2tDc7>+W?=IUVN=JQ3
#u[^(A?/b)81pHY''^5J"jR&6Y0EnLmD\</Vn1Lq!!"H-fB"VN8he;%*`=F:*!!SG"?rWLmfD27#a(
;ThS78/>Q)\&:O@`KRg5!!(;EReQT/aY-&UJ7UWL*o=0t<Rp2>J!!E`7_#:.&&T=''f6A33d6&dq7G
VE3SRn$eN`V"5`2e!!+kUJ-M`iA)u0b#?VQ*701^;XI-9s==+\,]:lBhL$X@lO<<JXo+>L)''Th\6
(3(CsqfbXq=B;BEFuI\.et>hi<rgTnBtD-Ig5=J)Y6S''hIf9@MDTDMq8$NtshH?ukLM&1Zg:oIH
P-\neFEF=)D:ND`lZ<AiK\2/Ojp#D$?!!(1:<Yl3:.NXfs/[=*1&$s,(N.n"V`q6:N(>9k#M8B;k
ah.1G%i.rlV(l(_$lo[:!!_KW#d$JX.[^^r84Z.G4P:UY>/e<ZFHPuPsl8r+pFu(;$Y?I,hj"B8<
?m3?1+@qW-Lk''$@:\QV17e;reYV61frY?rW"Yi#opTT,%ks&$9QBF:o-!!SM\KB3;i_&DNLIN9O3
OCBVS0;$oAIe^lRctb10Qsc''C^iImpFfc,u1bZ_\s3p\#=3)cF,aMIC1I7FFd9r)5KdTj8cTbWV
iWLfU@!!OLu)Q`sb1Q$G?r)]1^rXf\N(<]d.r<qP]Y6:OqNV/>sIJs!!WGQhcblosj:Q,++5MLY2X
s7du<QPF56F86<NY/0[Ir]aIc_'':NBKQ!!=F5SCrg47n/J.9fgsI9kKHGB`ra5N7=`;DpHG*3)-M
;&3&r7enY(K"*4de0^-sPTk0tOLf;IV?"(,&7*AS9BdF.&l9AJn1,YirE:E0=&:E#88\DsYaEXh
lJ@L0M?m&#,4IhIE1:[pYk)!!=c#u"@)H]0Scuf\u)juqbT"TrSLG9''j8(F+316Wi@0Wfu)MF#2q
&P9\_j-*Y`%+H"V!!8#1NJ.GJbHf;mkHE1_jB!![Vl70#tV,0+DDNX#hN)AWPYfi9s65M?*VJWmW@
Bn-06]''J5adYEr_+^+)H5-+B!!qgR_mrjF\T_6Ik3?O.:#jn$[!!6\S##6-<Z;H+h:o(tX&Qo)6"l
V;<BBVHfIg@#-_;D6N;E1a>MIdm#E7'';uET3m1-/9.f@#H0!!$JA`]4ddde!!rUHdbt5gmH7nQH6O
Ti47N6)@$pcZPRF(j8n`H''JU]]iL[21<(4_`QW^CXTXL*3G)CVod[j\9(i`.cK1pT2YtOdZL#hV
4C)F$dr+=RmA7dl-:T<V$M/=r]nmnL"2HPi(^q6#C"&C+FcX9TDk-i"EK#4#T9.50Q4m]q[^BAg
FG.0J1:7%p:Z!!`EMW3C9Ee*V.)-coE"89I0;p5AZ>cjH,;$^52[QI(]ljU%-gii=C#+%Yt?#jfg
!!jk*A"_")<3R#XOKP`)=jk#p_)s_]s3Z6saMeaqo\$_VN,g4gCiW&W.K1g7Gf)ESk,P[M*eniZ>
1OnkQgj8G&p\(!!II_YnQfN4?NVs_)+_''%3+=18<,MG.W)P)T/^J3\X@^DRWc]s_:F(Yd;G_PrAU
get5jks`C^,SR2#9$M.2kV6a3j2OO)7&sbC9S&?@hGh3kW_[QT9q''c-j_?MO?ih*ia,^l5(9i/i
^fAAig2:i^T]*c,+](a?+)os[OcoD*oa,IeYeoc8O(g46^0P5A2Q,8W%;bRRg\*28D%$I''o#f4Z
?jTXcYfHhAK+N<U;7"H7Qi!!WCn[.`uO\3.8L;.7WD="''!!"jd>l1;"HTQ)d*#_jU["lg&`U1o7<^
"S"q,K&r\G[VIWla5_4/$C3:UO*e^[#6HF5%kJl>%b?->^4N3-?UP>=CDt/=[Dq4[($PTY=E+FF
9u-lb!!"6DHAMuXQ/C''][5"E!!1g2<?0b71qK.GJ2fS`S0r@-&G3Gn4#-Js?Pk5MG+r''<pP+90Z!!_
a>Q''bekZX=N\iE(r,fKZ]H$e0f''X_[!!&RUs!!F6*;/$fNA[kUuk*bphObSaMC!!:+4fb"Yk.dt&51
j]S.N`WC*f`s=L5&.X:<Kgm=`bKfPa)E4^ClOC?T"]u*/7O.@XYV`7n"r7C]lg7lLU6?AWa$jdD
`l^n`=Wo>q1Tt107iKS0o5uDm:p$qo$$2#h?e86Q8ml)%Z''HDiKmeqlkn10.epkg->s"/%kjmFQ
$!!&]Q+W#7Gbj?RrJLMjHdPe[32?0598868r@''si%""6u\Yh5U+=T0>p?H`$N6j)ptYi=C=9D9GX
LsMK^?f;OF2st/BO=0bB@-&'';aq"VU,?XG]"*k,RTNLWo)@[3949Vab.9uZ''\Ye?`he]Qn[p9^"
F]YPd220iQ=pb=q/D=VG5%t--#J`qSQ"r0cFP\OPcC;\p)-CV\gR]jIkF`1s."/n^ke&SB?sI1r
!!KPOkO>_S"DPrkk0Da`n(.-!!C>DmZnQ@b3E''J)WYep7V''M8g0:]3R%j.'']&)\@hakg?!!I$/&ne6
QQj/qh<RU+_:!!Z=4kE-C%#YPLQoNhG(.c!!eIPn@R''ZLh<J%LPOX?<Jeg_>(ZXVc[J*YmC_[RQRi
ZQBJ($EO8CZr:@dBl1B"8h95+V@^-aX>Lb''K1.I#etsfMe_g#7m:A)-q+b>:[f@0WM4.Ldf(=2N
Mn71)p@ZZGXn%iiNt2@/J,rd;N]e#V9_Yf>"c9+s`aIb#Yh8qt0t.gnYUJAS"!!b$N(5blZ%Q=0+
)$bGgMtbmhs7kX?6X3u!!7,VCUg_;p`/;n!!M*.<QKXldU<&jK@3C+TlGPm]f4+);.JFY):8psSAT
o!!p`>Zd$<)(NXBY,R"Po.aCJH""F%hD3os9.+d''+=%G?R.9O%>9*`;MR_mF_Y-#,3,dYeJDf<@$
j^Y.0X1Q7jkB:R=*=]Gu1U(Zqf)24Y)&Q]_Usa-T^l`C[==nS&(,],l!!G)Ea#Gt=g_9BP;2Q=*R
Je`@2VA`)Jn>XFU+YsGOUsl)HThBP>ehu1i7piPJX.0(_@m^KtS^:KUK5Oo"-p$G0V-Tj74@!!W`
*%t@GC]IIq$8qTc9b2+"L;REXh7s*K*NY(L"cHT)$"WkJh]UPM,#F%3&ap2F`aD*J_E3=6B>N#)
ESAMfAq)rh+lS/b?%R&N[qo&,"V9$`"QC\/I"jY$jg#7r4XM)J0$Gg]>3$S<RY''YmWEFa_c`8H]
A$ElS3`rp#<lA^3KnqKLRV]LE`mOrJBe0dVL''_ZqLlo]g&M3?^Irj7FW)?''3''lKJ%_]KbGB7/)9
F,f:DpBiJCYp]h-Y/[Ji/GBf(,N8<<8YQLK&Ra]TYhIo@EH;Li&OWe6/]ac''=ga0$87,+Qq+u4:
gHYLX7:WkGW@MdAeR6fB":?_/k;?;Y*%]Y42Q<;cBe9)!!/GKQ[6,8*I=M%rANs#gmkC&&9-/aj"
Z7OQ0T7q]7(HQLk=?$Re,3!!<EiGPl5apVja;$BT[>D$lf+,aO3dmY^:eWYko,Q#I8_ut%CSsg''$
Zi;U''DPJA*D(6m0Beoop[T\iEeQ3lqYqn8>(M]BZ/;f"tm2<.ha*S`+\/DAa/SHKg"Hl''Nq,"5J
@]6OO/G[,iZHDpKR%''\n7YR6I#pI\+SG?:&&1ee%MohiLW-AC$9?1.\R7[K\V6Po77\giR^IX2<
h6mGuUI$"DJ9=T]>@\;n/<3[78;3pRJgK;7#>>&EW77[UHmqbq#7F''`WhSJu@a0q]o^.sd-^iWo
*C5M2ICKZdfJA5dVm3Up2Lr]?6IT$X,4T601KDWC^qr#=<ZD7cI4U#6!!hX*]NO=@Gp"o]c_1I<)
(A[)''G*1M),#*:FMe2%KD+i^N_p_8mlJkq@=;d<3DIMXM1I<#\B>?:ch8(SY%0H/A/0LZ+=?o3$
/9frf>p(1a#raF#!!s+1n8cZ>N"U-F94BU.4n-VjpP,c[!!Grmciq9,*VS`@]*C=b/[;Qi`tjKP#@
IOL$Xo'']FHqYo-fH3@>Y?pT/k?_BkseA*l;5*gaG+D=u`Xe]BNLcPMb8HQr&GOA+cg^$oaEZZjc
MBls''?bnHqLk."6Jb,J#Ns\[6V>$24VD/!!=CuL+*;jc-%OS<c&ID9s^F73?VLZSt@O<+ZC$2;u,
RqP=B=P!!!!HQM?ql]_*+D&Z<FeZXDlr"KMMA"^(eY4T5b@ZV^i/9msXi3osY)j)XU(K*;@8It&8:
!!"OOeR5:g2jZZ]X/@N(ioe''X/p5li)X3f''=@SJZph\N*YT,/SB+0Z$O7Y>bi!!><F^''Bm=SC<#<P
;=h''$fBlt2V@_u&_5\rCJ;a8O(4HfC;KK6LL+FVs?Y%Ho`BI7:]K,Es[Yd+h6@FdEHn8QiJ8b1>
\afQM8Y*8ChQT+Q+>(/qp&1bq)h<]pKpCh+''4l\n6&Ip5[BZ]9kMV_gQ(%a`MYgLh#,Qm<[Yf;:
=DA$S8j@ObNoc#''$soJ<lD@B/MmTS"QL0;B=M(?F=h$c6ileW''.oY(gV5Jk?2mKNXq.gk96Aq#E
K<u->qKci9]^qS-YR6R(RF3N@Q=r(:dpPVC`cC6jD@/0WS(5=VEXYuD^hVaL=MO@UFO_DYAU!!oQ
UkF83XDXp)6\?+tJ0gLl#*u3''>GKEYWhsl9$57>^<V^st"W2jACqHp''\O$Z(-c*rn`p2BWF?F84
\.^8?;],H]=P9\X"na1L=6Eu=''["0`?&u..f:ihYV7p=''gaY+adS4.[X*`ou2lm>]IZi3IK4SU2
fsuIT]-sRhjX5*^Gs;F4DS.oiYl''!!$ra;s4-i5"#Jn7589CU<*1q9tC-''k6@D>CPQ:\/^7A*o9`
;=p[i84t%_qtaX+LM<h`"\W#Q0`rp-''92*]jU&0^=^\e[MNPp4,f^;Nlb_\H%[(!!kl0o[&;co"d
[uO^@K^ZHgf/g@D0Dt*H-UPZM)Q>r>$7"]1E!!B?l+L!!Uf_8;aN_8.&8''iHAdXaJ2[_!!95N/,$c)
@>YkC''WcO9=R*O,&3<M2jesW`%\(^u<c3"$90P`6*3sF%S/TGo+h`::XFl_`o%''!!6fA9GWgVAtD
BV9l%iU=a+9u.HXpHYs-l]26#e\H5SZq)2&\m_ts2-.YU`X;FI#>!!rGI<uqbJ+?%E_B)g1^O.\:
r^=(,1H"@O5cQZ,J]2!!O@a>e[8''cThOPS(R#Zq5m="f/ls-*<uJsAH"2@=4[["\A`&0pJ!!*K>bH
TSY]OVZ3O;&(N*\;m:5!!.>64[<CtlBNs5TmRU>2S;cnS#$aB8q4''8Rh[KdSf)Co6-#u"hl*\:^>
Gu#"<Vh#RAJ;"mm]H^PY;npAn!!kQ7eD^`E<!!O-MSXF(R]nIVpT!!EbjAOQd7@VO;u]2jKXWZsd_P
b]uDY()p2?<.Bq$.E-A4kRd:AQ7qTn_Z''B&AdoN`''%]S*p8<"Gf%$9`Wc4i''_RmXIIldop,igul
Pdj3u=P:>&.4lflc%b$5*"#e)#<"?tI%<@f%[I8h;_+<@`e$BYa4r;<*2mUG#\^@iFAMu%/H5U_
Nk<*lAW8G[:Y3"e-?n;WCFmtC8[;s''?ZsiMK7)]B-$)&ZlCuV!!PkJG;aFI(A%G''\"naEhN-=CO@
`,UcJ@,55uC!!1At$P*9]/29Y")@+nAXdf%S6t=%jQLa%AA:F[HnR/&71pod-GE%M''%=`J-`=]<u
$nP[h6-BC;<TrPE:76ri&Qk4nY$m5LNn"gpb1?^<]N0s8m=HY]ik01E[m70.>/Z^Qc9sQpJ1Ll*
<3U&fE>gD)^3>,C''j1<qms7F$N4b3/(''QnD`VfmH[=trYl<pTP`K(Y6!!P#F2&>B).U[pYQlREYm
NG!!bFD''.;1$Fe8;NkhZBXR[Ol#,(NI%BNNj&>AdMFH068LIb5KYe2fD(;W.J$B<mO;''2bp9C(+W
*RBS?W"Z-KMe5rgWIf7h@o5)<J0"hd9VH"A/5<R97[''5lfA75nMLe4Dc>^#_.Q&!!Lg;F]rle&+J
_4p&cCf''l4VA^s/.^6(]#8R`Lp;+S&4g;NiL;)sA476TA&?NMH//2KgqTB(Dh>ik])!!/&A#0Ta"
H*FCMeGXrKAa/q`i]i?S3b3F30NeAQ:N7IY(16`_0-[,;2XMALRN=2''6a.sC(/cnd;VS-Ab0;-0
g:9c]Q6([7^sD$[X9+^;6BVRJc$=54K;UT7]Ad''&J9jNe$[E5`coJl4cMJ;,i#]+spDU+Cr''?6a
,*U6_`e''fY$<mM0:Q[86mF7''biWFek%h5XC6`8!!pLHR^Ak''c7%`s.hnr!!GLR2HR^&o!!FfYHY-''h
$@H&c_64UI"E95T+S?<#@]$F.l/6uo,3?Ln%+%gXD@Ou1/2`,YB6iQK<&cUdaEI_:,+\s=h@!!@7
^iIP+DXX.4MB/qliEu\22mBfqAs[RA(<8bb7*`SW-^nD?6,]JgcG)RR(,eauWPL&R3pqc\.7A"[
V`kmVJR8S^HMP,j/ENV/n&kC([8*W-*>oU7"kRgL[(V15\d9p<:?J:"L<9njP!!cZkc*sQ%fm>\\
E<<3I<''fcOm1XZHg@&''qXgR+m+i2S6Rg,13%V<M(Bc9XWG!!pX8lLn8_&.2G@''J/_642CtL.fg;^
#W7;@23qN0F:rm]hY=@F_R0Kl#>97=Habmc_34MWo.lZt<F5ABUM-.W/&Mg2MB+8Dr2_=U$TEJ5
@"j.8"`%Cb@C^*U7j2f?*EO<jprZtFGCtSU=L];$\KFI&ET`mUoWaKLi?CJp9n4S3IB7[#Yh1!!(
44#2j9^4I);pP7O/F,Q^@kA>6C1[hA!!\5J;!!9qtJb%s^J&9`tg/F''Zg9a$kpd_<gkcq(e?%_2@5
Lk''Rhl]GTFe>X/;$!!+eA/@e7OlaXrT5SYgT_2g$I"NPF''JU20''X8?3M!!:o+#i6B*;7''fn(G"nYO
P"(jk''o0TaZ!!k0$rP7t=B_rVt9Jk<_gi+773!!#=IKu9dL?kKP^BfY*''M=<[E_W/FhY6gM+&/\N;
]h:#L!!t,iGpa3i*E^BaO^="c4JL.mje,QO-?*9M<-5E8+gWa,7$5O&OmI\0BE!!%D`?j22#]`Z"H
\6!!.=hIZgu$3`!!V0rKXMmO9t`>m<4O@gQ6s&!!l_C-0Q9q-6923h!!/B\[UP\SZ+@A:AQS-/$dFi9
6F=9M*d<_l;>a7=hC]?\H.R.\A@"7]qZ.).))YNjYjMV:9d0<202!!sfQS3E#FOjB]GCjs8A&6pG
1m\)j3Oh\`2f@r-HTeX1kup:0>1j"Rp,$b]nMP)djGg$54''.H7!!=JrE.;h7;<Msu?ase>M(&%W!!
]S\m9UVA:N/!!S%V<tY_1+#/JT,aYC4,i:fi/DZmSL"5@9!!mACm/b2j/=T2c0kj,PD,+k0T&,pI)
#eY^WVU3qmJed/[=K,\^JJAdc$^N@rdKkpt\Rk9.R_fqE+^jP%Zj"g4M=]$#1o[O9Jjo3PFlY\`
ImE.+q%V&D@)W/N?o)r-W.QF21L''f(g/>JL-++u*"2tLP1@sVbk9g,WZcX([%YuV[5*k''Tp@fk8
r-(&.8ffdE6c!!db9>''W($`nJVfq`Dd\\9!!$lWa9^M%NV6-=A.XCYrfSgqrIO7H&:''b=/mD3/mp:
q+h)#Wl=GS=D3]?/Mq2L-Q57&BWSS`b''K1%Zd`+9)@MpH/<!!;39YlT,YYr:M6I(,cBN!!sT=,K[8
^Yi]Scq_AS;b65jHdI^H?t2n6$rDjoLB912MfP:91ssDA8[-qC!!*7aN0T3uB1*hEV3a;c_T8S@Q
jGF[rJ-?nsM]2S!!!!Gg!!q\sg;MXLJWh:on8A''lJ8]V+G&J/3oYESUTEYYYr9r5&r.gq01<>''P^?d
(iJ7aI4Y%Ik?p.R)[^A2RmK9("o''O9KSeS50F+Y9@oZn;+BXk3A(a;o01EW+iKK\M^nM%8SG2+/
"O^6SehpK7\XJ)7,I#`"=SYC!!Wkhk0L\b!!B=Tr&iCs0Ju!!Nm=k1]=XJ?*kUSYR(%-XL&s!!F5A9P
Yp5$3!!OrB9P8[)s`u:CC9YfZN$<FLkWdJdf(<9n"G6QLNBp?f%5IM:6.E,AoY=F9J<[1_^0sCW]
C[A_"goWU*$HaU0d`,KHE=5lo!!j(#X9.5$#=MHM0hAP(CP[O0?5"Js]d2rQ.Sc`$K!!>]5VQ4-VJ
''rLQ,(6;9jSL_f#Q,_C1r3pFL.T7bBq5e3L`ra\,*n!!>K)`Mo]?1rf$j\n52?6ItKWPQGKYf%6a
XnoJlK3WN''=''1!!?-;A7=C\^N''kt]qBJ<ak*<=$%`aF\jl^VRoZ9DA-=cs1''4#sd6I*jMW=47<Ll
+;`!!)ZA*D+Cl1T43.1Gi[19c:c<lX/W,3Q3Wg7uo=E*p=#7KSY#+o-qdC?.VhGd3%^da!!0eP[Eu
)-E2cjZXLg"]13d+^c/rH6ZV!!q(,XJ`m^H%MJFbMijU[sM*b^_=WN-;Q-]C_=CU[ibhOLer@q]Z
2]se&i8/rm"]0ml0JeD8hR7`)Eh5''pi;m?*6*Thj_91`8E"ep1_*[dMrc,`jf)og3MisQN;j_;,
K8kQOH)damjDBMBq-X62O+gVd<Vol_oUnH2("V@?U[*^q%i2!!bhR/L&''<F/''<WY<"4p28Pc?lOq
''J+6gOpb11>TjdWGT[$0l:S9G/Z2u*EI''d;8-aD[9+,"#3U[I\CTYG$7siC4N][cJOZtLBX#7*G
;.8j%nadu5=Z]^055@<kJb/<J*s7F1''iQ@rXqYm:P-2&6mK&6/a1s0!!iEp3N70Q^O=DoSbdQU<A
br1^1`1$D;QW%9X_#c:nbC96N7A=RWQ/$RSmK]fD=?Q<EJ/.TfYe#f$1F7h_os%b#$r%mV%TZ9=
>+Pg3h?tmZ4\lLg>280]=K/0Z_G[7*Q^lkhKjoAPa3l#)S1oX+iF)V;hYRR0`M=4Vk)@,[hc^SA
Ca_tSU488dFgr<I]''(''bC1`^["(4P$QBWlGp)JssUQi+oTE\&;/:t6''?Ejc<l+@6h3(%VZ_D8Go
Bc"$WGCh+d^5+o7aD5<r=$\r?$.B2@qH2[h4<CRF4^;ab?-QZSZ05`!!"jpm-?D"H_jIoNp9:lTR
L1tRfUo>J5-?6,g$B.O2N0/;n=GdHXLDufiQKr''5e>L?7<%+98(/nG>VU.3OMj,6"g3hG]6Ek%q
_6V,Z=Z6F>=#MNr"5G.r^oBnYr6pm*V.O_Ok+fd#/;95%!!pp\(&`47(_0dd7CS7r>''70D@p?V,a
YcnKNh?,=eG-E`_JCPVj8G)ot;j\#Y2m<os6josW^S/QC0f8r0Ydi4MXe+:Ch="a:VGNpM!!>^4b
pOE;rlQ@;FQ);pE)UqE).TdHQnIS&r$8\"bi?W(`r=OiG(?\sa?GV5EAH>[iSdU(7XY$uscb`7;
8V5^4EfEMVWZ32f9*]Ee/0aQHdFT>Ql<$tZ@C&hUC-J&-,W4cgX?bM@9:Y`%4DLU@cioi+^]>D[
/+Y`&>.\t2C2RjhW5c[k/CR]mYY]sPBe0e?V;F6k''2rT/Z<"ZbL;bH@kkMHd3a1S:MqGElI:6q!!
48-Jl?tHZ[;`G!!i(k@<-?kpkO?mjafahq+)''^C8\!!#BMa=F?13XH/C9VI9AP"/<%OT?4kUPqcD,
*a$3P@)Y&4(''"k(BK0b]^GNr0,]eaUo"2`d()J&c883N/''[p/U;,c0/T5ge''QWNWn$JE4A@.bPE
eeK7Hilb$>`5e$jkC=s_GS!!t04E_bs_"4=+A!!uq+UI+4(_dMkm$KD\"!!bQ-R!!ufWCkFM(P0R1VI
?pPg4Pd7an!!nM9QWuN@D`Ac]aV?Th;@nl[O5$.q6`HbhU[=<p0B/!!^5X:ueZ+;dV&a2,n3@)9Y<
pF^bo"=V(FnM@%=bXDi("5D)#lN<s2Tt]a?[8oXU2<]u>eB[po/:#Rq9*''9fPR"]uYZ)j.^rgj)
$R``I-\6u"Yi6''nE=OP2mnE+KBh?nV:p\OG8g7MM7E6JX=Os0r`j+Z6nM<oclf$uid%2HaJFd%f
`rb&:&lEVW<uVmE+ojC;,8GIAa5XX--ZMF*/$lUgg:.iKeN9p:UE)K!!"KRoCk$/m:gCu3o%TZ2/
fHXf4J;]-h]NXK,Yh0c5[u_1JZUoFrp3HdICJOJK>[t#T`M^Iu\m/RI7[QU7o(MWgfq^Jh/D>uh
LdgWHA-`pOAY=nJ\OgkZ/b6%noqnul"#9"&!!N`uT&1h_J_4pVg]l1*i=C[nBhUTZhJWJ@:a0KsH
!!K@8m"dfHaWH>cn,Krq8^qpR`UfX+5%lXKfcFZ!!7JQ7a<c/2HZ-20S(90:Bm+WH*H5d#C''jDJ+P
-0HC!!]s"u\r3uBKS\jWWcJi4XJAYdJW,SK9*3=*XLgsh4=''r?ae5>2:JL*K\4:j&+*iu!!u]L-:F
#+!!)XVB+J+Q"9om''!!.M@63UD=0X07CYioS2+g[ErePb7o"\YZ%.>3So1BC-7Ej0p_>E<9faY>,r
$!!<no5n9:!!kck6@TD<Usk:X/[-YKcOKf+ChACOEB7HdA1JsBn/g^TSIb-Z(CeJ''JW8UlH/,[\dM
07oh(\5?btEK)gq''ro@4<uT7''c6u2:R1]Cqa+:''9RiT6S==%^t2.<>doN+RJ[d^Qs7[8NhW2e.T
4\]46Pn$j"^dD.1_1"+J=jDOtLE#T:)!!ck)WJ??E3Htr`4_A*T;a5+(XC(_[=.KPrY<LH6=OnWt
i>jCbNi%e)9h,b5e6!!YlOG@/gJjmMY.?"N;$\+s[rUU(k+8]/3D4p$.Pt#nU''u5I;MV%sq<PB5N
R"=qAUN%hj,0](LW8^ni!!\<Sl#,e>[:5\1f1K;1qH`aQO^&\"+@pu7>p;&61*NBpCHtTZQENEb[
q7QgMZXE#h*A!!HZ(@ERQV#6!!=b/-_dG''X5S_+f731QB?cLk8`.A4_B=F=b,\G/VN_(4)$j''!!/eL
<oEEI^1>1cXCQ66$*:L&"r+/+BWPj!!e_jg9,I#TdL$H=/=n$-Vj''WUjc),Ti>''"6^s)PlocWBN3
,@L0\Y_8)^<5/fA4;b04%BO^e/2cVZ"''4P]YjqmJom:;rCppH/''<GeR_>_OWYV6%b:BE%a#E''fp
#7Dr./<=.s^u>ZB\MQ*r$<DJk>3FDJ%[3?i!!\?<5!!)T%LqM)Zs(*G=89?^f$jp=h7-">8b2@bh#
NF#?DA=YL"O-e)ge9HKWm<mJ`l=jO%M\B\q''eQDHfeS?AL3DZM3ffc_d9&]&N!!]T]`N`6Q/BVsl
2:$/jRMPEb[3O<H/%/kp^Ol-("r=II)-,=44<4M!!:M_V!!RC3T+Xg.(qHeWc_:!!X7<liu>H-2Bj"
#mZC>^0&$IWBqiM(;b+qXGs"_/.FP.!!=Le8,ZqPk/F8!!(qd=]j)8HuL2$)WXDuDE"[j\UV_B1s1
B3n"&Wg)`FGGuMH4i''Gk<]F2<LX[5BaY3<#8_!!jGJ>6L\#*se%*uunI.b8P=efVYSro$A8HBAE/
$$t-B$:eGRYkl[''=:p>)_K3Q1Zb^C>);1gWHY!!QUW5b"U.=Jf)1<=e#`Qc5f47AOsn''!!CQi6TcR
cMP15\cR;sR*EkaG0BOoYV`2>NNM.ML"-i5;g^.@=Dc;(\s=ccQLO&")Q-frid/+jcWc"?%gHE5
5J=2nXX>A]Hk-J*PP;0k;Mjtuic)-Y37.IO!!SkJRYBF=KrZ8;49`B(lr=uEjf*#DSiOZ1afA<^D
''O]#s+[WJ5<s,M<SP*b.2k@$UCe^FX9gWcjF2''_;lCGCnXtTmH"fq&50*7n4(<+)sUcN1thFrRD
;6V0Nd1ebUXRPJ;RnVd!!=4iGMbBj9@.dmPB8??/^YF<AAY3S:^eQBC,=^Vk[O9W:JX8\ZZ66P#j
<1*0r]8%c\F>fpML#L''75cUR;L''^r`^ip)ce6mF''N;Jn\gA;7[f1c8]!!7P&#KCpUrpqd`H3p0en
.])5j==kOK$K)_SJLS58XIjgW)?uKhc]f11b%QB$8L=4''":H&sQ&!!/4\-":eiL=PbI4U)-4#.+*
dTD=@A5l+7<Z?f<O''Lr/8>gm;$EQY6cS5U(o*M"!!@Q_k[kL(dh<`>V\IB:S:h\rR*_&>f"0`c%j
grBI\;]dOGa)\Dsd%6hjIQX6NBL1hPUa^,@!!K3+Yeu#=k1minA+D%OWACT4hg@h\S''W;YGhFYn5
L$68m0cCcbJi!!SQ_[2UJ1@Vfm3K9j9i&\Fof93/=`o2XjXEZL&nI2W%g^$q#"tQDN9+PHP`n8Y9
^]JHfXAH-?&?PSIYVI##7usE7''PX`h0"EfR^?*4[.:_sY?S>aIa*fL^gloN?ZZ(Q9Lj6,h?N^`_
=FCTY!!e"A#JX)/=c#FW^XWdM7$]r.(/:dc?jO+\:>d-&FFkps''<(m14HH5NnQ&ha[M[cJ?!!bjr_
P^mOtiFB)C5WT.nP5[b=#aE8_kUW`=6[LuFe;&$MSZObP''VuBHSk!!>i]RH1</_jVG/2i5Emr7Yb
YUJc:>DX?ih]rK`grGRAA''Hdt7oq*s`S2.;OD5gWG9L?SX^VV@3hK.I;BHQr;eY)O"fu[I>[qX1
''nXthg!!%><Q39Ub*V%AtF*dDD&KQ1gn2-_-Ji!!]_r$gu`oBrJ0F9[;5:hqU&("NC`)t?R;2W-bO
MBuB^$`s+o@/]bRJL,g@*3aB,=L]G(T&L/\5g$q''cF\\)(%_QE&BCO:mU3kL_W"[Z=-e&*&4WPD
kG\Tc5$/0%r5ju8ar+<CBr%Up-?P[_!!0a8/]P%E''";ZN8!!a+P01aQ6p:$aC[HqQuAO-Qdg.`"ed
$5T;[]6l5<4U"Ilrd5.nLiWu1BM''772LOCb''d&mt%]kc''"dBI\"t9/r^JuA6%Z9C)dJ;ck_,@d@
Jl2W?/''JfTom((SEG:@ZaZ/3Yf[^Y&D\''0dmC_338diA[a3oMD%oB!!L^pqlrD@-Dpi41;d%Ma&a
?idYa9XiQ;.HK&2B5%_GM%cl%aN45i/5%P\&.s^9X)]?t13bM9h)&0t1:1d`YPXDsq:/O;njOb1
A)<M<:2)b+*+06lIsCru^(Nn[+0f,@4b8m]_>8Lk_!!3k/eLIdY%tTH/"cMPj%Toj@7P-3l[%.:,
S/6QsYgi:u]GDMO0.B1<@-muu0u8n<]pE+Qs%^i<q>d`^^WEBg#hDW`h^MZ:`]+^ni<B8k#l:O[
6b4_RdcX-:Z7-dMpp.Srel+YclS8uuX@;`Y(Jj[%hbO-W6q\<rg+iG150fH%jha>i0:UPqG4#)0
<pdOUIHKp<1+!!mH''KY\4^F6o*F]]K?j]:Z-.Tj,3&/*@;%iW9KXk[O;+4f1_a]N&prLMMA3g/Zd
Ns_#;9f19P5hZ^9_H/anZ!!p45^&CoPkQcqM].&t2,YGdES,=<W3XOV(WB(BP8sc/r\uAGB"V+Nc
n.(Y59,5fk3rS(LA/D+>/BGGU"bN#-1U-G<@)^_#bAsm!!K%1qEQ3!!5<LVSc\?tB@>cG>BJ@**MA
RJ(HrY0$Qn''bPY@[E+SmMO;7\_?fcMZWbK!!S4n1F$W>UVahinH@**''HUY''<^''<p-A"^$(8!!825V
btL(0h$2q''@''TZ._/c[=)GA=R'',>pG=D7e6EtkN[@8O)''b&2aBjEc:VXSM[/,4+mkC7i9=YhDIN
%8]uZM8fT&QK;TWn,R//(2C(r6>6nga>(u;4r2,mQVEr-[9@E_Iu=UkY99ASc]_m,^)qp&jfC8q
/<[L''0sJ)r$C4C43EYfH@/]I6:n*fQs5-6iW)I[9/Elc&7c&]dQpLn5U>"88,/DbT2Ph''NV.3o4
_LR/*C!!&o&R3f*5;8C''GJV>!!n5"r5]2oQ-HM!!keS@c`]q<3Yp''YjsJM:p@0\^sK`=L"C4@j&i3R
U$a1fq?K2ckZO\fcl_A/\b"Zb_''3#i>cS"./)]p-1l=CuDCIH"2mC`&_b,n5ZEYoYGX#:eKtK><
Z#0]f;O''TVW&#8?q*WcOC&KG6R5:?CQGCMBPb.TPRtBbT%1ZF(oUh_eeYM5mR"/s8U\3^`"0bhN
N5uuQ/=(BC@usaPg''@bGh2;;U87OM3&,%Q0=qGpBf$[`c;00qHRnS+24r4#b_#b#r1)"]K4t8DP
^!!cjU&n9k5HX9LIZ02imgJ-./$F$<^CjA(1+9eeZXB-jeZKip.6U&CpW\bJ]L%g0Vf3$b(g0SM$
gRBWc22-@+C.1mJQ2),u!!''CE9?jINc#-/gNX6Z;8/m''WQs''`a$,@EsYSRAFsk!!?_dWL$,*[;QD[
aGPUmF[1fuj(<!!S<c1FUKWOr\d]:6RKYX*j-s3PG;72ImJb;sWVf^AM$@+m+BiVa+bi,F5D.YP8
o6M2_Q;!!,MVQd=XRn)''dfp&l1NAG+)q)FP?e$rD;`J^/(Z.P58r&0NlG2<BRpTaXoUq>!!79OheO
CJV-L<a]L=l)cgD6>KF)E1-OtbKk`+@iYXZ+qosT_ALN5?1<.+e3LN*mrI7hM@E=hO$/N"Ml/ne
CB7btQD;=45\''U[!!#eUaK`b_.\3Y:X">''?lOGY+,X!!`HtZlk11VWV"]rbAYoim\ZG27/.,6>75p
L$-8R&u@%1[_6^%*\?(S<lB?:''mj>2TLXJ5=E+m''SaXc"-GXAcV0WJNZuZsa=pW`s!!AbfjpTn&9
M*]03D8sh\^+\MK&b5`-XH^I5BAY-&Hj2=''@/Ut`kl\.r@5:bPWD$a_''J-GinPHHC"gtUFY_iT5
*N[qm:jah==^TTNh[thOT[Hq!!hP@*Of2+llTa/_;iDr7amU''0uC&dOVIZW_)7@#%pq&mU;?ib!!)
95,U_*LOW6G-DO/'';NED+QV-0S:jp69!!Hqa"^#aKX=QOJpnsuc[+l3Dd<PYT2!!tDq_SFk_KrZ`/
QP@^s9@uqsh-^[:mk_3-2R%34/=Zemigmp&2\&k0"Y1W''(2t$?TJ2)0''eDVdas19p%bC/h"7>Xj
/@n''IQQWOfVCSS+V(]Q%HqTDl+ZHD"O_TgWV@^Gu%otYlJSanUPV,,3aL4\*[a\BfRUliW0TH<Z
cSB:?MDu@h93m7p<eP8RK.\)V^94:.`Eq_599Z7Y8G-=BlL6NIB5oJ)1*@`D>''it5QXV/*9FBM!!
W0#U[OZ7W0Y&0?#1_tu8L7`eDe`"j\k7F_t4r0tESq:bed[FO7FW=%$F=HQYJlSe6CCkZo:6rFp
IqE(&a(k%T(QinQABer@3FF_t3,T''L.&T''f!!%P4_@jAV>''ZtWdO`uGQ>5TNE$r%lFf2u2ubJ9Q\
X9;cO.`?''/2`VCL+SNXGUds;''**%O=9d''_<Duu>3?]$<1$=H;M8?h)8!!"10jb-)3h%Y06Z;f/V%
1Wk(L=P@626a_sY]A_8*MuH/3TPHA3ndg70YdW>hgV<k''icffsS);_P2S*3=TNdQcD\G!!&1K.sc
3]""mbR0Z&!!Q])K%L$k%''D5V(>Rr:>*u.gr/9Th&BHA:]PL&D!!7F$L]SXmRm,.Q$Gf/Zj1kFU<(
c?!!R^"/hkfQAL[!!dPB''mVq>[erdj?q3a?1"=U"oTe<hZ_CIl''p7ME!!gpI5;Bd:^?+4C2eS#E*V?
&mF/7qk5H<3pmG985/ab(3I:Z-[8.\=UT9od[hOkbKWc?*%W]$k$Z<Epo"bZ(8fI),KtUbYjarp
b/]&*1N4:8pccKF2Ym]\])JU6?rb<FcUP-<V,jJ?->bQ"R_h/&dW4,QY^44`[M6!!K0E(a=qms3X
>lY2QQ!!e;7KC*Go,54/LY`KI)O!!S3\hLm7uNke9[YqcudrT-huTCoHW*%^bAAH",u!!fN%[LQ=m''
SG-XSiQfI#JFF\H">t[6]8d4!!JMq''BS-1ulXJ*nD>^%aiDW3jGs4nQPV:h\dNA^F71/l]qckM4=
(E4)Q&hG+>;ES`Y>99hOR$g8^W5)<Eg^gJ[dBIK-Beth]+COa;d(^0PPpKNj23Vjn4%u?"72d4Y
$B*=%''WeqdXG@Ah<]F4/f67HM6GUGZ=S_oGje`UI,aS9?f9[3#@[hUFDU6?ARmN"_^q(+/T@k!!P
6RRWKY6/,J/0"mf^Ve=/^ju&KB`2g)Kjs=NlIsC)K2jo][G<KQI%dfC23rhOeQDCm1(RMITB>''L
_]8hP)Li%D?O`OT=M8L$2G.agLC2,HCK!!(e6"rkuJV,8VZ*)M''(I#1p<b(dp%JO("XeT:k,ILIH
/h2L[*/n]+!!_%+:$7u"sOu)oP7pnW%2+LYR4;$XBiTDjl=[U$2<Mr;mTt[C3O^2)L2"P47E5P@Z
Z13^.!!,=''.W.8HUY80j4e.:>t<1,+dZB`f*o$eo+3j9%I<_Pj?(g,?`PQIq6,$:NTL`.gl"gI!!]
J/)RuE!!37i&m,b9<3U=#=LBR\V$9RdTrqeSA1M@N=uo2I9-+=Rf+dpG@/Ui:T\EMi(U-#SK8#-E
JHb8=F9u!!sg.0nE24a)dXm@BagubjYPd87<=S:1Q2#t$\g=%%''.Y6`XBV!!AE&#Wu!!1))Ihd7YC8
O4Db#;MP''n/KM#jZ/jHWC?[UM''J.]IS%+NE(W%hSPsOH*AJFMBXnCDL*E0E5I;K3IThe!!&e=UT9
euO9LlH8]X=>3cjYfnA_"Ud1i=VgN=&V09MCK,_9f1SEuj!!@K;6JFR''W&9kOr=orfhG[L[0=!!JG
K-g'')l`1k8feKVcB4)HT9M/VDZq9:TiFB[?*3AoK=H\B5UI#29iZ)''bqnQt#<t>DV^EdGUQ7ni*
NN./+D85m-%[KoTppk7m(FddK)&EbI:;Fa8JXig<P;e,*cF\8*K>kU+_0Ro0f_;XKgY(#/`CI>H
6Fm0p)dW;a%T]jS-k^U*fOJZ:[/rank#lCYFkMP!!JS?0LTV:+@JRtS-CX3/2O%N[EAV]h\]P=O=
[<)[!!_<U5,J0O!!rI/X/^ML&7#o4o.^N$ZU7V+<be<#=t-s$*!!]J`!!grn##2Ob@okj7Z\<2md$<=
S1E(mX3,ueW;(+63uTX&E0_VTV@Z(S"tdl''EPDW,U7/CWT8Q4uG;RmSA%L`XJYa-\el_7l?c;qu
T8TbD!!E.-8U&=:$Yfa<s3?Zd=@LCG6<G3q/Ii^0-J;#A&''orMKrq`a)l-$o#La-<8B7O"ZS@[`g
l&<LkXNVZ9"eYi''E7(A#H?gUTNF:.pn.iteU(=3^H!!i)NoiaDOE5M[W<A,$#B2](<''S=<a:B4&6
U[E]jChT3t"Ban=SG>icd(lQ,@-Z4^(TEu"/=^oigFn]"''?<`$f7tUDAeb5Va485YdZpOU9M4L(
lW=-d#1r,s*Emmr]gqN3e,nIBb4Kg$(Ehs!!;fo1hMPA2[P;H<;M)mc*Q=+"E`e^fO=1GVhPOi0N
*!!$HDGd_B=s4++uXL+BfZSnD*V,.6W&8>a^U&sWqTa`O[\q,Qch-6&F)-D_ip,#fH/CWQAYZ-\c
(=1r<3b)+`-a4Ws()3jB^,<-6''Wa-SVI6B?S]7qYr-351Y6C4-2;<lVQYj83l*;m"[r#L5s8$/0
F04''BGlr]]Bq$W+7k;-R?n!!3;35ikRngQ1GKoJ0JO0o@]#@k<^6EX#$aio-H&m:QaP0F''(8&<'';
1mU5F!!iSiX(g)u/eH7X/l+BB.7`X]SC[#N;<fJ]EJ>s><XEXg[lO)]Rh,_kj3IHUPG<Ecgf7LQ!!
m@,/@XYlCKnMHW57>)U$#F9,<j''^t?Y[RqT7t(ZTE,6cA%TgF[2hh2(d`WD7VXai''=g4jh,''Q^h
DU*Mf.NF''ES-:FW=g.mr"tO-c((YWIk?Gc[mfBMG''TlYoCA;R1:4IN<9d+8U;Er2G&Ft4SO!!R5*
c:_^t@Vh$c3EQneP87#8YR''\//1W3d`Hg7ZEp"H;?rj1R$C8M=l,S`X9%o?WLD]egS*8e2%H:L7
69OTm#obH6^_eUfiA%T,0J?GkJO;`^Z.I=2K(\gf"JfJY)"-jkHs/M#[VU6EaQB87Ypo<rMoW*P
G?B3A**E$6Z.-N+B788AEaRO[_!!_*''1+Gtj''m#:CNMZ#S$UUgAKSUG@(kP0n9!!E,5hP_mIN.%q1
R#0r[`t@LJ!!%Vfg.(KF\Y%qWB93#iY#qtj0frnup]h+?5VD."g`e+&8GlXp*#8SXo_>Z:/YO.AZ
XFL$k=DN\>?lBD&^Rp,nF0P/!!^rd6kE>U6D:bc0d]16lW/BG).CE;7<(''(9T-J$elK?L]7XqJe7
-/B(ImKEtB[4,(]l<nmAL*#4SMnU51.##Vi#[4K;kd.)k!!9ju@%@05^G7lGbRVOHcL196FD=/l@
!!7ci+/>#"TE`=HXqY<1*>;]3H(/AKRkVIZJ!!7n`e*UC\J!!!!l[`)m3=Ya_F8O)-DsN=>5`RJR,MI
1E5s..K_.1QL(@`''cmK.&f,AHXkob''V&F+R(,,rNgOIZ6,-"LKPODg$P_f0tTkpmt$Eq%:5)Q\!!
$!!?t9@)\-TPjc:+XSDX"jdff@8[-R,BQ#rBiHGJV^L+3Ya!!I7[mH5VIYV`;>APTo5=:T''J!!O=\N
''om.!!l=84Ie''tPu-YoqsX)#t%KRRk?/>*3*]+a@5k59JHE$%^S:c@KMjZ/*9XO]L!!RK/fR=:0''^
^f&=[6c&=4[Dp<93:?m@jTZ4@Y@(q06-=[JWPN'',B\)?S;5JZbp+2`mXMO1K@''MLho^5n^MIW>D
+(t%@^nG_QGQUKc^aD"*eAhoQ''C0^k=6_OY?<JQ4C:Pb6^<TLX<j<_8.-+3>/14YOH"mZmeoDNW
%<9I9X&r%/3X#oB!!,e6A&YmVR_o2!!/Sq@33'',le\aH%6(hD9k7:FErp<+A?^2:frOA*K$U/''`aC
5PU2Q<sIu7keg02[\SbuNd?''fY^lHJ9*&dOK/N2(VI1hYD@3:L!!I6+d<SdhD`,)G5(3Jg@9`-t$
R=5<AjCfM$*<n%R(ZGkL_28JLKnPm]:e$YqbZi8[_#fK_S.l#4OD7X;]ic)+k''Bg_Q/b^b+H&-^
#,shMrZA=iD.Bf#!!"rR=T0fU$d_ln<q_oF/B"bqB9C@fIl[6$$&0TL.<%n]AhQ=%,9!!C-;lG"YJ
b=!!:[%$nb3B9nABYY:Ya:GX@\VV30O,M$/)oK</PD``$EJi''4tU65,^_kt2/Jebqj$!!=44*%ZR/
X^IuZ@XKj`eQ@=f@/VD^Jea=K$\)u#-"(k6aom3P@gP+B_)c6:CTt[lK(d&c+a2F''j`!!3\ot,[o
b2ueXeSY>Ap)3fg=A]#E''A1WGQIB&O6Dp.cIIBA[g_>W.*N[,+:Ecj#AQaYroBs@I`]n"GFa1O6
*uT1)i5''c9(&2S,o\3VZ@QZ.<Is^W7Hc@58aErB%hG*aL^l.49nLq^[f\"A^G3;$DH''a)8WHe#a
"<\dgaaW?a92''.n`j#M,cMh]Tcs44YKXNam@#79#(n!!^ehZKZgUI*F3LVV?Pi8<mj_Pk98?+stR
()";JWno<->]l!!"+cj;u=G=cCWL&,>8@O".mg<Qj)XG:mAIEOm1K_EU&cF[kiaWESenM?4Ne%Al
,ZslM_#j*oYQ+pobnOL#''%Z.4%2tihFAmEpY)nN0)#=24F;A2%YgONRbdHnk,qqgW5MZ>Un&J_j
fB/5F*PULDe,#")X''ZA6ZN,pHH(@"-8?GaMDe@J<8cIq]4pQVF%NUT;!!Hg$&Zam1cTnaNoQg=MU
/2jec2:,H&,Ql1?&snu8ftRI[.kub5!!>s981mX+."o<n9^Qie(_IJ/''XQJ*V7B])p>2fl?(=VA"
/JPB=PC!!/1D>dFIXA#uJ7>*,5V#bu-QM1B0@(/6,QMnCr!![Q9@pGKs1&hLo`@(@`I/)t6G\;_:=
3o:cV4ATFhZ#XpAd/gG"F!!1:C+(?U7-FKjpg10uVZ&f`''f"OeFV3(]#W(it$XG"mG.mo^6[t:IT
?@rD$"tT;Y_2:t6hd!!5trALm58#5K9*@q1P?nE%riWrX\QoGIS0%VS,YB"o6rOFK7ZHg:`:FiX`
.#kLq%@`DYK>hD?[/blV9[g`%Q<6ePMPu,i^jc_W#CX8fd%]k>a(gOPZ!!On8"3Uof(k\+N?u9A7
4WUVG\&,htUsR4Ei4N-NU.&6''%[q,`>KBATaM=h<0U)KdeNMN''c@9D$[U7.N-8=VrjXP7$nZf/8
@#k:M&ck\4RPbrqJ3op8''fmhgQ9ub[@QmiA72E="5Ai.Pn^.:t&!!f<G3mAi#c:3N?#7D$-Ro9OA
(QZ+/C,!!jGauO*BkqICY62h;sCB/.)RmKN%SG+,VFk*CKo\;sCRaWH/(4&LT.tr!!?iAJ&*_[A)V
UI)cZqAV$/+OrO-eh.ElE5Ql+<ZD5-Ik/>"I10.XYgPWV!!%Y2sbru0/4V0/l5d2pGX.Ib%(fC.@
)HGL6,*.B,C9V5>bA4DIn=Z1kNF05D0#?gLYb2nM#$*BlN_[F=bLOR"EA^rUe:7.:`<T@,0O&=%
Ia<P0%99o(VT-V`!!NA<[)-Ba"1nj]u>bUq+5!!>`(l=3,-/@kPA?3jp(`iEc(a>N9kR6#[1oO"tQ
6#.,9.jFg%Y^2tB(k"52OIoU2I_*7R#)dVu=LD8])7C)'')TYFBYp.VgP*1rQ''K+uJK10csjpqPV
#V:Bt>YSXtEZtad@*+pLgArfSYQ3TO!!*IkGF76fG^!!l)P;N>sQ7<)7N*3A*[Vc)T"]<;DtKj/`<
G6(_S_29C<guaT(r@K[,^B`=?594SCC[HQ#n,:C:`N<OA"q.pO/6e%cS745%;TDCW/B2qu$bQbS
>?HhGgZu/1+o:<Ds8>TW9jU"n<?dhpOD!!O<Q3C@l])O2L!!*I!!%<UKli.j(g!!c!!B?CceG(o$C7)H
M''C9PROR8V<dP0U^gc_c9KAfCio"2F''EbCl][pQ%q=1A0`+hk/YitjUM(^L4:nV[WA#?O+&h+3e
BbHq`Vb6q.l>fMJq@6Va9u\*p_3uaDcGB''KTnA%jJ2COI:/bb2Zj%D??_m]/@&1''#Ze&#<])jMO
!!DAI+kBnH1VFk;p&D7RX7H5n(*&N?P;J@u0?Ro7LX6dNd5*h4jIhRt5"VqNJHScL,9OYL]N#biQ
6=Zi\=2$^tp,MkpKCr%>FU*h&JG$GTo5`OA)-FZgXG@^F"cj%J+6CK/F6:.(Jd%t*31?>RpAkYk
1p[lJQ`L&Z_hoMk2-*43;:ScX8dD\k.;b;-A+;1_!!0&?@c.[+@X9]''S.j4"==LBSGWNe2ba\64/
`Qro_-s0ZNJ;r^mc[?m.ILL6M*[''"YOO!!^3!!PfLLI?WN)<7he\_.(.-nZ6?O=<buiFH,t)Bfqa;
k[#7P-]n#t)GadZD[,hqZK3]@ZD43[''!!-m;9]9-6Kh(r@A$A#1i[`S1=E,$+&s=WKRM@BD#A^KI
''J.X6''2ut(DmU,9DZ_QK,Cp"!!X092Y/XLl$XIOg+:Bh1;"UeR9:0jpSGqVg%6\,j`UM1pD8t%k>
6onGFE.\9ao>+g#G%''idmfY.IpWo)I?Lb@Q2:$$?IB4><EOr<IAdRs`jbYE#VuJ/I5a)3#2;T`[
CcV*lfV(+*Zk+e\!!JXi30Ch\t#X;U$jX!!0<h-ZK!!1<2QoJU8A=fOWe=/"FS"jg''@sX0GPQp4NDn
Ff=N:C\!!^#gIdRUr3WJ''PJo0-hCm^WM8AV^lP@LD>N5tj$\2J=Fp?K!!TUZ2/0%q:\q$_*0b''B?B
Sn]f=:h(*=^(8fkAeq/6/!!oWACRWdi+GcD#TadLF#PWg]c9$D6''+$7X.7A6@qgF)C.^]KHJ!!8Wr
i&ip,WaGSe]q6a?NP;U$n:AB\#.PbVLiFD-D)S4?(nr#,%@6*%q=B#\(b-+!!\sujJkF`=#ErYjE
?1Gd35@,tSGF>8gd!!q5/49mC2#"s=h\aK_1VeKdas7''K:8\0#C_&C.&Z+<ERY<G84:@\R2?o,VP
fok#J#q_!!d9R''31Wn`?Q''l$+]_+K/SJAjD\<b>?k''s(aj=:I"0C1`Z]JP_J_>NWDqO/;U3)Ttis
!!EP(,''^$LY!!iRr<2M,TR7SI<''mQLrJ#p0o!!nMRY`P;_GgS0f[2D@4i_VHRA$7juf\.[ZSiMLpP1
Y^eN;U&R4ZbB!!gJfWVI(VW#+Kc`IZ+p5Zq?K;G,QVNWs>52[Ja"BVIBCn8b1%otk-XP^*dr#lj+
a\*s&1=()+;(`g$58:3o?spqf/7<?&J]oiN3<dS_:n*''ElXNZ*_VE2kXo''-.CbFo!!9k,1sQi4hb
0A6CbJ/&5il(oT;L)KkRjsMNaA\#OapfJ85M_Ns"$@%l^]6gc]?s-"sW6P9^?MKQP^<el>T0_e1
90Np823p*t?-krB]CrL1-s2$<=NATrG:6ROh=J_.U,7AGXZls$g]NkJ?#)!!6hODeZ7ch:@FieSs
1]g>LmV]7R@98R^]elIcOH*P]ii<:Ls-nY[pF!!!![@g85''cEul"V&fpC3[u[Q35$\AWF-$?9f1E8
/-o@h3p=$W<Yf(kp9Y]a-sPTl7fk1^"''41Z(4[Q1F&OJ1!!Jmjp5S7l^I&mo^J1\+$@#]T9[''J"q
&njI?T<Hi<H8i&Z"5+pfWVme7Ge%YhE65No@*,N-e3.2b8ZBq.SCAsIAZMb,(B1:dN\2=5>cpqa
$_>H-=JkekPO,]VoHVmU3uTVXctlqPM)eZu!!DBjbj(cm:ckQKKY?k?l#hIt7`qfkd1=^Z2On4U6
&;heG1iD"G:p;-5UI#<8a7hTV]ZGj[n*:tPQ)KDmZ<T\$;DctEK4q>*_*[K`l"")W>lK<SNJAI5
R6ic,!!6QVApo;,Es6@AqE3NBEG[6_h2ND1mr9o3Em9[W7+4j)g]T4`JbME>hMtd@W)TtU(#t0Du
"aigJ7GJ''t^[^(/U*9Z,XesPDTlSgnejMq+Nkrk#"BJ43M@(o.EX#Q,#qS&<eS#2>9aTA]D_i\C
%5B"E9>R&s2=APnV7''qfD7OJRF.qK#CC;auP)IX`2N#,.i[\<k_D@/=2Td4Ha]#EccF<qV153$0
?_a7\c?lE_9Po!!CEi"(l\aR0ai_?<H9JZ6D4NU=`^Gi,ma$8:9]VEiFk5:p?MTMJPI#70_.0Sd@
Z(_N<n:He1`[XK4>+%ks?:r_EQ^t5'']DfFI`gT<Pc/4/7$K6&[_E]l:Pq]JjL`B?Q<BBU''X3!!Z.
a&?]RX17c(@%##$m?GmjTn2l<i^$uBUY`J\r#b](aJ5Ud*!!C+G"j9W6AN(S59/g@b1<XXS&1h2o
/Gq5Q20f7/qMJ3:l`=@b5IR_2Tm$\Kpg@K@4pNV=B>J>N3i/%K/1ciTbtW%SPT%O>7\f/XVncJc
R$GkoEF$&>q!!6\B#Zu<_=CGU6R_lo3-D%5iA#<)OUDhm0&YFu1h3''4)?+)Yi-d(]>U%9<_=BE"s
GQ>W@Q7os_cFZ<76D4*4I''4:Co(\H04R`bh*;cUPhuhFBZKp0ebHa\"4\ut%T,Z8b'';u@TYRID5
"aac=EGSSk8&,2QUb.[7g\9*''9h_5q#<$Y#g^.(H.[9O(UI#sK=@1>t^rc_[k.sl:hGIA>UrKtb
XI+RXf?=q@Z4@qVp+W`AR:$%c^l!!KSlJfIK!!37eXh^M\*T$q*;+0/L,#"q[.mU4*oEA(F/$H6/5
#j/J/U8VE0os**@5e66H-i<FSfcHd^je''^sqYF-;''f76"&6PM`g"qULW_]Yn2)9jf2^OqVXsg.a
R@l"i3eo/$OT*/>:dtG2a\-.?Ut5sgJu)dd)''Br)r)lr8d^''GfM_[#@EWDJ.2--qd@f\?q.uP[/
O!!T49"r9;%MIYm]4!!YNNGg39*8HCJCY?_."C97''/%L5X"ib6Ag`i;O_1U-ZM4f2&ldc6Q`\,h0e
dSiT+"dU&WNXB!!+1bl^&V40g$)Co9-$4TQ)\+7NrQ>:EbjK68P<fLgfp!!kp>+);3aScfGY#?^IP
Y9W/7J,J9KHL:Pnqs\"TA&4$W=:/$h@n#HB=<P.j82Z)R%Ie>S5p&`omM%5:8bn=nUN[go<t/YI
M''W@,*%RAhJb<7D@tc@Z(g+=rA[''ZN*R\''<Wg/e4@+dDFlF^o6"PD=.2-B$i/8`2_?mQ>V+/ucZ
Y.8ucJBU_l]f>LB$*?Bldrd`L06YVOj=MglJu/3@''qUL8#nl["F#%!!m4h,sJk@85/SZIdTYfRV2
j>jHK"''-1F_+Z#tRR3!!pg]:I2#ZetXY^dIh`qa9T?_uGjdD!!X-NC<liKXZqn<JMOO,+<Z:]H_QQ
AT?SuS\U,d#Sq#[iRe#eq,s_IbB)FPT&eW&O`K,*efObq=F`GV?ONt/*3`E>eim-#E@gXp7!!6t@
?2WApW`bCJnHt"Y)8O-lO$<XG>gKD$U&"/A05F)PJV7u6NJ.i^/XP;Rl!!SM''WD*\p(S2IrV12Hs
`E`e,plnO^_A&0&/5d=GJOQL3Yf9$.&;%_Qb''D;I+)@J(&@KLZdUTo:UIPcZ*]d^:PA<j%''$#=S
6GYG7eYnIR<&b$(*_8j>dE-1tRuo6*3e)7n5aV^+akCF:&r2/!!$&4Oj^_D+)_)kn9E-3a>/pp7N
aYa[3R"9D4''Lk1cXDMMLd9&-T=>R00d/Ct5f.W>7oA[/6Yen4sNCEGSVD''35(Kaj%n:Q=HF''s-$
AlIf-79JE^l:9$/YNqgr1?\VX@&^$E)o5E$IB918AS+f;."QO.h.BB/26`=?a3/$Y)_>g28H*ut
l.:?&f4RMeR$>$pXnc3FgrC%+k[6\"%fsQ!!An`j]+hkCI/\]<l(N\qCAuno7/3$b">9l;L_g\Ys
[4Eo["`&AjX?pEjCI(47>YX\ZI\L*@Rb=_-#''D:><lFlN,#VXhi`L6L6Q.Ir=(?pjJAt%ia#W)s
lP*IF;56?e)G;-$B;9^Y)]&8#nG5=N&Kma+3s)dlA3\XIJ_JuoQ)YI:''J/kijg*LQM,^]Y9C!!3I
3l,]qk1GBTKH&b]XG@+6"cU7frP4''$icYj6S:cN"_)bLIHX0EuKa36q?QUh?=<obH"r:5.''Wb8@
aT]+D?>J/($fu*;f%P;''J5NW::/8SsW6fp6X`Tq.lRb8bl!!V&<3q$^t_Mqq\0/[>0o2U1RU#DQZ
a''sBf\=n2XJJ6tCXf>%IAYlY&+HKZK"2oI\beUNN2i\r@?UhWK;TO!!`0I)%C!!"&!!!!"5$-glL;5+
g!!@M"rCg5OT<<%/c9''m3fa=#S%Lam+QVIUX=up=(QDhmCY/ZY(hBnqAbVWAaMIZEP?tj1[Gq[cM
>k?.0g>Ak]''2!!"<paiX9^rgKUS)p=A]ddGM=K^bS=:SG?>+^esF(a8<(/L?r_9+-VKoZ:9mg`A"
a^^f_].C(4rhn@s\6O2IhZjN""=l\b=MBfgeg&M\AM&$''N@+tbcBJfCJIPP/Kh$_`LXIQD6]NOU
,(EFnC5jObqg,JOp:@EAq28YB!!O+i2@R5cr^reaT64HsbmCXg`*PTdLAg:S:Q^9W^kHHY[PR\IZ
>%8^Q(M;5&<Z_]/$s%+?/d2X;f(Mf>k_1JH''071PkQf:BpPsB%.$m_Z/h_''fI>NH?RG)hLo6uV;
!!hkiU@?]]dbcbpuk?[3WP=8)''(jUho/IdC+Zn@<>jo5iG)MZXo&sPW]!!!!!!!!j78?7R6=>B~>' ascii85Decoded asByteArray readStream! !

!UserInterfaceTheme class methodsFor: 'tools' stamp: 'mt 7/1/2016 14:25'!
allThemeProperties
	"Answer an Array of 3-element Array's.  Each inner Array are the information needed to present a Theme editor tool; the property name, category, and description.
	
	self allThemeProperties"
	
	^ Array streamContents:
		[ : stream | self allThemePropertiesDo: [ : cls : prop | stream nextPut: {cls}, prop ]]! !

!UserInterfaceTheme class methodsFor: 'tools' stamp: 'mt 7/1/2016 14:23'!
allThemePropertiesDo: block
	"Iterate over all classes that provide theme properties."
	
	Smalltalk allClassesDo: [:eachClass |
		(eachClass respondsTo: #themeProperties) ifTrue: [
			eachClass themeProperties do: [:eachProperty |
				block cull: eachClass cull: eachProperty]]].! !

!UserInterfaceTheme class methodsFor: 'accessing' stamp: 'mt 7/1/2016 14:49'!
allThemes
	^ All ifNil: [All := IdentitySet new]! !

!UserInterfaceTheme class methodsFor: 'tools' stamp: 'mt 7/1/2016 14:25'!
categories
	"self categories"
	
	^ Set new in: [:result |
		self allThemePropertiesDo: [:cls :prop | result add: prop second].
		result]! !

!UserInterfaceTheme class methodsFor: 'initialize-release' stamp: 'mt 8/14/2016 17:24'!
cleanUp: aggressive

	aggressive ifTrue: [
		All := nil.
		SqueakTheme create; createDuller.
		SolarizedTheme createDark; createLight.
		MonokaiTheme createDark.
		CommunityTheme createDark].! !

!UserInterfaceTheme class methodsFor: 'private' stamp: 'mt 8/16/2016 14:39'!
clientClasses

	^ IdentitySet new in: [:result |
		self allThemePropertiesDo: [:cls :prop | result add: cls].
		result]! !

!UserInterfaceTheme class methodsFor: 'private' stamp: 'mt 8/16/2016 14:39'!
clientClassesToReapply
	"All client classes plus their unique subclasses."
	
	^ IdentitySet new in: [:result | self clientClasses do: [:cc | cc withAllSubclassesDo: [:sc |
		result add: sc]]. result]
! !

!UserInterfaceTheme class methodsFor: 'accessing' stamp: 'mt 7/1/2016 14:08'!
current

	^ Current ifNil: [Current := Default]! !

!UserInterfaceTheme class methodsFor: 'accessing' stamp: 'cmm 5/14/2016 22:00'!
current: aUserInterfaceTheme
	"Replace the current system theme with aUserInterfaceTheme."
	Current := aUserInterfaceTheme
"Notify?"! !

!UserInterfaceTheme class methodsFor: 'accessing' stamp: 'mt 7/2/2016 13:56'!
default
	^ Default ifNil: [Default := self new name: 'Autogenerated Default'; yourself]! !

!UserInterfaceTheme class methodsFor: 'accessing' stamp: 'mt 7/1/2016 14:08'!
default: aUserInterfaceTheme
	Default := aUserInterfaceTheme.! !

!UserInterfaceTheme class methodsFor: 'initialize-release' stamp: 'mt 7/1/2016 14:11'!
initialize
	
	self reset.! !

!UserInterfaceTheme class methodsFor: 'initialize-release' stamp: 'mt 7/1/2016 14:55'!
named: aString
	^ self allThemes
		detect: [:ea | ea name = aString]
		ifNone: [self new name: aString; register]! !

!UserInterfaceTheme class methodsFor: 'tools' stamp: 'cmm 5/30/2016 19:00'!
propertiesForCategory: categoryName
	^ self allThemeProperties select: [ : each | each third = categoryName ]! !

!UserInterfaceTheme class methodsFor: 'tools' stamp: 'mt 7/1/2016 14:26'!
propertiesForClass: aClass
	^ self allThemeProperties select: [ : each | each first = aClass ]! !

!UserInterfaceTheme class methodsFor: 'initialize-release' stamp: 'mt 7/1/2016 14:09'!
reset
	self default apply.! !

!UserInterfaceTheme methodsFor: 'actions' stamp: 'mt 8/16/2016 14:59'!
apply
	"Apply this theme to all affected objects. Let classes decide on how to iterate and call their instances."

	ignoreApply == true ifTrue: [^ self].
	
	UserInterfaceTheme current: self.

	self class clientClassesToReapply in: [:cc |
		cc do: [:eachClass | eachClass applyUserInterfaceTheme].
		Cursor wait showWhile: [
			SystemNavigation default allObjectsDo: [:o |
				((cc includes: o class)
					and: [o canApplyUserInterfaceTheme])
						ifTrue: [o applyUserInterfaceTheme]]]].

	Project current restoreDisplay.! !

!UserInterfaceTheme methodsFor: 'actions' stamp: 'mt 8/9/2016 17:07'!
applyAfter: block

	ignoreApply := true.
	^ block ensure: [ignoreApply := false. self apply]! !

!UserInterfaceTheme methodsFor: 'actions' stamp: 'mt 7/29/2016 14:45'!
applyTo: someObjects
	"Apply this theme to the given objects. Useful if you have to re-theme a specific set of objects. Restore the current theme after that."
	
	UserInterfaceTheme current in: [:priorTheme |
		[
			UserInterfaceTheme current: self.

			someObjects
				select: [:ea | ea canApplyUserInterfaceTheme]
				thenDo: [:ea | ea applyUserInterfaceTheme].

			Project current restoreDisplay.
		] ensure: [UserInterfaceTheme current: priorTheme]].! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'cmm 5/13/2016 19:26'!
atomicUpdate: aBlock
	"Like our Preferences >> #atomicUpdatePreference"
	[
		| original copy returnValue |
		original := properties.
		copy := properties copy.
		returnValue := aBlock value: copy.
		original == properties ifTrue: [
			properties := copy.
			^ returnValue]
	] repeat! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'mt 7/25/2016 14:59'!
clear: propertySymbol
	
	^ self
		set: propertySymbol
		to: nil! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'cmm 5/23/2016 14:11'!
clear: propertySymbol for: aClass
	"No longer override the default visual property specified by propertySymbol for any kinds of aClass.  Use the default."
	^ self
		set: propertySymbol
		for: aClass
		to: nil! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'mt 7/3/2016 13:08'!
derive: propertySymbol for: aClassOrSymbol from: anotherClassOrSymbol

	^ self
		derive: propertySymbol
		for: aClassOrSymbol
		from: anotherClassOrSymbol
		at: propertySymbol
! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'mt 7/3/2016 13:09'!
derive: propertySymbol for: aClassOrSymbol from: anotherClassOrSymbol at: anotherPropertySymbol
	"Derive this property from another property."
	
	self
		set: propertySymbol
		for: aClassOrSymbol
		to: (MessageSend
			receiver: self
			selector: #get:for:
			arguments: {anotherPropertySymbol. anotherClassOrSymbol}).! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'mt 7/3/2016 13:08'!
derive: propertySymbol for: aClassOrSymbol from: anotherClassOrSymbol at: anotherPropertySymbol do: block
	
	self
		set: propertySymbol
		for: aClassOrSymbol
		to: (MessageSend
			receiver: self
			selector: #get:for:do:
			arguments: {anotherPropertySymbol. anotherClassOrSymbol. block}).! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'mt 7/3/2016 13:09'!
derive: propertySymbol for: aClassOrSymbol from: anotherClassOrSymbol do: block

	^ self
		derive: propertySymbol
		for: aClassOrSymbol
		from: anotherClassOrSymbol
		at: propertySymbol
		do: block! !

!UserInterfaceTheme methodsFor: 'lookup' stamp: 'mt 7/3/2016 13:48'!
doesNotUnderstand: aMessage 
	"Answer whether I have, or inherit, a value for the visual-attribute specified by aMessage's #selector."

	aMessage numArgs > 0 ifTrue: [^ super doesNotUnderstand: aMessage].
	scope isEmpty ifTrue: [^ super doesNotUnderstand: aMessage].
	
	^ [self get: scope top class -> aMessage selector]
		ensure: [scope pop]! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'mt 7/3/2016 13:49'!
getViaLink: keyObject 
	"keyObject is intended to be an Association"
	
	^ next ifNotNil: [next get: keyObject]! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'mt 7/25/2016 17:56'!
getViaSuperclasses: keyObject 
	"keyObject is intended to be an Association"
		
	"We know we're the only referencer of keyObject.  Update it rather than create new ones, for performance reasons."
	keyObject key: keyObject key superclass.

	keyObject key ifNil: [^ nil].
	
	properties
		at: keyObject
		ifPresent: [:prop | ^ prop value].
	
	^ self getViaSuperclasses: keyObject! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'mt 7/25/2016 18:02'!
get: keyObject 
	"keyObject is intended to be an Association. We have two lookup strategies: 1) along the superclass chain, 2) via a linke theme. Evaluate the result because there can be message sends stored or blocks."
	
	| k |
	properties
		at: keyObject
		ifPresent: [:prop | ^ prop value].
	
	keyObject isVariableBinding "simple key objects"
		ifFalse: [^ self getViaLink: keyObject].
	
	k := keyObject key.
	(self getViaSuperclasses: keyObject)
		ifNotNil: [:prop | ^ prop].
		
	keyObject key: k. "restore"
	^ self getViaLink: keyObject! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'mt 7/9/2016 11:29'!
get: propertySymbol for: scope
	"For convenience. Does support access to non-class keys."
	
	| aClass |
	aClass := (scope isNil or: [scope isBehavior])
		ifTrue: [scope]
		ifFalse: [Smalltalk classNamed: scope].

	aClass ifNotNil: [^ self get: aClass -> propertySymbol].
		
	properties
		at: scope -> propertySymbol
		ifPresent: [:prop | ^ prop value].
		
	^ self getViaLink: scope -> propertySymbol! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'mt 7/3/2016 13:04'!
get: propertySymbol for: scope do: block

	^ block cull: (self get: propertySymbol for: scope)! !

!UserInterfaceTheme methodsFor: 'initialization' stamp: 'cmm 5/23/2016 14:16'!
initialize
	super initialize.
	name := 'unnamed'.
	properties := Dictionary new.
	scope := Stack new! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'cmm 5/25/2016 18:15'!
link: aUserInterfaceTheme
	"When accessing properties by name, if a property is not specified in the receiver, give aUserInterfaceTheme a chance to provide it."
	next := aUserInterfaceTheme! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'cmm 5/25/2016 18:16'!
merge: aUserInterfaceTheme 
	"Merge aUserInterfaceTheme into my properties.  Same as #link:, except merges all the properties of aUserInterfaceTheme into the  receiver"
	self
		merge: aUserInterfaceTheme
		overwrite: false! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'mt 7/3/2016 13:22'!
merge: aUserInterfaceTheme overwrite: aBoolean 
	"Merge aUserInterfaceTheme into my properties. Only overwrite properties if it sais so."
	
	self
		 validateCanMerge: aUserInterfaceTheme ;
		 preMerge: aUserInterfaceTheme.

	self atomicUpdate: [:props |
		aUserInterfaceTheme properties keysAndValuesDo: [:key :value |
			(aBoolean or: [(props includesKey: key) not])
				ifTrue: [ | mergedValue |
					mergedValue := value copy.
					"Update receiver when deriving properties."
					mergedValue isMessageSend ifTrue: [mergedValue receiver: self].
					props at: key put: mergedValue.
					]]].! !

!UserInterfaceTheme methodsFor: 'accessing' stamp: 'cmm 5/23/2016 14:15'!
name
	^ name! !

!UserInterfaceTheme methodsFor: 'accessing' stamp: 'cmm 5/23/2016 14:15'!
name: aString 
	name := aString! !

!UserInterfaceTheme methodsFor: 'accessing' stamp: 'cmm 5/25/2016 16:24'!
next
	"The next theme to look for properties when I don't contain a particular property."
	^ next! !

!UserInterfaceTheme methodsFor: 'copying' stamp: 'cmm 5/26/2016 15:14'!
postCopy
	"Keep same name and linked next."
	super postCopy.
	properties := properties copy.
	scope := nil! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'cmm 5/25/2016 17:57'!
preMerge: aUserInterfaceTheme 
	name := name , ' + ' , aUserInterfaceTheme name.
	next ifNil: [ next := aUserInterfaceTheme next ]! !

!UserInterfaceTheme methodsFor: 'printing' stamp: 'cmm 5/23/2016 14:17'!
printOn: aStream
	super printOn: aStream.
	aStream space; print: name! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'cmm 5/25/2016 15:01'!
properties
	^ properties! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'cmm 5/13/2016 22:07'!
pushScope: anObject
	scope push: anObject! !

!UserInterfaceTheme methodsFor: 'initialization' stamp: 'mt 7/1/2016 14:49'!
register
	self class allThemes add: self.! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'mt 7/2/2016 12:25'!
set: propertySymbol for: aClassOrSymbol to: valueObject
	"Where aClass asks its userInterfaceTheme for propertySymbol, provide valueObject."
	
	| aClass |
	aClass := aClassOrSymbol isBehavior ifTrue: [aClassOrSymbol] ifFalse: [Smalltalk classNamed: aClassOrSymbol].
	aClass ifNil: [^ self].
	^ self atomicUpdate:
		[ : props | | key |
		key := aClass -> propertySymbol.
		valueObject
			ifNil:
				[ props
					removeKey: key
					ifAbsent: [ "already cleared, don't error" ] ]
			ifNotNil:
				[ props
					at: key
					put: valueObject ] ]! !

!UserInterfaceTheme methodsFor: 'building' stamp: 'mt 7/25/2016 14:54'!
set: propertySymbol to: valueObject
	"For convenience."
	
	^ self atomicUpdate:
		[ : props |
		valueObject
			ifNil:
				[ props
					removeKey: propertySymbol
					ifAbsent: [ "already cleared, don't error" ] ]
			ifNotNil:
				[ props
					at: propertySymbol
					put: valueObject ] ]! !

!UserInterfaceTheme methodsFor: 'initialization' stamp: 'mt 7/1/2016 14:50'!
unregister
	self class allThemes remove: self ifAbsent: [].! !

!UserInterfaceTheme methodsFor: 'accessing' stamp: 'cmm 6/1/2016 21:31'!
unsetProperties
	"Answer the property keys which are unset."! !

!UserInterfaceTheme methodsFor: 'private' stamp: 'cmm 5/25/2016 17:59'!
validateCanMerge: aUserInterfaceTheme 
	(next notNil and:
		[ aUserInterfaceTheme next notNil and: [ next ~= aUserInterfaceTheme next ] ]) ifTrue: [ self error: 'Links to different themes.  Cannot merge.' ]! !

Object subclass: #Utilities
	instanceVariableNames: ''
	classVariableNames: 'AuthorInitials AuthorName CommonRequestStrings LastStats ScrapsBook'
	poolDictionaries: ''
	category: 'System-Support'!

!Utilities commentStamp: '<historical>' prior: 0!
A repository for general and miscellaneous utilities; much of what is here are in effect global methods that don't naturally attach to anything else.  1/96 sw!

!Utilities class methodsFor: 'common requests' stamp: 'sw 12/1/1999 16:09'!
appendToCommonRequests: aString
	self commonRequestStrings: (CommonRequestStrings contents, String cr, aString)

"Utilities appendToCommonRequests: 'self beep'"
! !

!Utilities class methodsFor: 'identification' stamp: 'dhhi 11/6/97 16:48'!
authorInitials
	"Answer the initials to be used to identify the current code author.  "

	[AuthorInitials isEmpty] whileTrue: [self setAuthorInitials].
	^ AuthorInitials! !

!Utilities class methodsFor: 'identification' stamp: 'sw 7/6/1998 11:49'!
authorInitialsPerSe
	"Answer the currently-prevailing author initials, such as they, empty or not"

	^ AuthorInitials! !

!Utilities class methodsFor: 'identification' stamp: 'fc 3/11/2004 12:22'!
authorName
	AuthorName ifEmpty: [self setAuthorName].
	^ AuthorName! !

!Utilities class methodsFor: 'identification' stamp: 'RAA 12/28/2000 16:32'!
authorNamePerSe

	^ AuthorName! !

!Utilities class methodsFor: 'identification' stamp: 'ar 8/23/2001 22:44'!
authorName: aString
	AuthorName := aString! !

!Utilities class methodsFor: 'miscellaneous' stamp: ''!
awaitMouseUpIn: box repeating: doBlock ifSucceed: succBlock
	"The mouse has gone down in box; track the mouse, inverting the box while it's within, and if, on mouse up, the cursor was still within the box, execute succBlock.  While waiting for the mouse to come up, repeatedly execute doBlock. 5/11/96 sw
	6/10/96 sw: call new method that adds extra feature"

	^ self awaitMouseUpIn: box whileMouseDownDo: doBlock whileMouseDownInsideDo: [] ifSucceed: succBlock! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'JMM 3/31/2000 20:41'!
awaitMouseUpIn: box whileMouseDownDo: doBlock1 whileMouseDownInsideDo: doBlock2 ifSucceed: succBlock
	"The mouse has gone down in box; track the mouse, inverting the box while it's within, and if, on mouse up, the cursor was still within the box, execute succBlock.  While waiting for the mouse to come up, repeatedly execute doBlock1, and also, if the cursor is within the box, execute doBlock2.  6/10/96 sw
3/31/00 JMM added logic to stop multiple redraws"

	| p inside lightForm darkForm isLight |

	p := Sensor cursorPoint.
	inside := box insetBy: 1.
	isLight := true.
	lightForm := Form fromDisplay: inside.
	darkForm := lightForm deepCopy reverse.
	[Sensor anyButtonPressed] whileTrue:
		[doBlock1 value.
		(box containsPoint: (p := Sensor cursorPoint))
			ifTrue: [doBlock2 value.
					isLight ifTrue: 
						[isLight := false.
						darkForm displayAt: inside origin]]
			ifFalse: [isLight ifFalse:
						[isLight := true.
						lightForm displayAt: inside origin]]].
	(box containsPoint: p)
		ifTrue: [lightForm displayAt: inside origin.
				^ succBlock value]
! !

!Utilities class methodsFor: 'identification' stamp: 'tk 4/10/98 06:25'!
changeStamp 
	"Answer a string to be pasted into source code to mark who changed it and when."
	^ self authorInitials , ' ' , Date today mmddyyyy, ' ',
		((String streamContents: [:s | Time now print24: true on: s]) copyFrom: 1 to: 5)! !

!Utilities class methodsFor: 'identification' stamp: 'cmm 3/9/2010 15:19'!
changeStampPerSe
	"Answer a string to be pasted into source code to mark who changed it and when."
	^String streamContents:
		[ : stream | 
		self authorInitialsPerSe isEmptyOrNil
			ifTrue: [ '·' ]
			ifFalse: [ self authorInitialsPerSe ].
		stream
			 space ;
			 nextPutAll: Date today mmddyyyy ;
			 space.
		(String streamContents:
			[ : s | 
			Time now
				print24: true
				on: s ])
			copyFrom: 1
			to: 5 ]! !

!Utilities class methodsFor: 'summer97 additions' stamp: 'ar 8/6/2009 19:20'!
chooseFileWithSuffixFromList: aSuffixList withCaption: aCaption
	"Pop up a list of all files in the default directory which have a suffix in the list.  Return #none if there are none; return nil if the user backs out of the menu without making a choice."
	"Utilities chooseFileWithSuffixFromList: #('.gif' '.jpg')"
	| aList aName |
	aList := OrderedCollection new.
	aSuffixList do:
		[:aSuffix | aList addAll: (FileDirectory default fileNamesMatching: '*', aSuffix)].
	^ aList size > 0
		ifTrue:
			[aName := UIManager default chooseFrom: aList values: aList title: aCaption.
			aName]
		ifFalse:
			[#none]! !

!Utilities class methodsFor: 'summer97 additions' stamp: 'ar 8/6/2009 19:19'!
chooseFileWithSuffix: aSuffix
	"Utilities chooseFileWithSuffix: '.gif'"
	| aList aName |
	aList := FileDirectory default fileNamesMatching: '*', aSuffix.
	aList size > 0
		ifTrue:
			[aName := UIManager default chooseFrom: aList values: aList title: 'Choose a file'.
			^ aName]
		ifFalse:
			[self inform: 'Sorry, there are no files
whose names end with "', aSuffix, '".'.
			^ nil]! !

!Utilities class methodsFor: 'summer97 additions' stamp: 'sw 9/13/97 20:44'!
classCategoriesStartingWith: aPrefix
	"Answer a list of system class categories beginning with the given prefix"

	"Utilities classCategoriesStartingWith: 'Files'"

	^ SystemOrganization categories select:
		[:aCat | (aCat asString findString:  aPrefix startingAt: 1) = 1]! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'sw 9/6/2000 09:06'!
cleanseOtherworldlySteppers
	"If the current project is a morphic one, then remove from its steplist those morphs that are not really in the world"

	| old delta |
	Smalltalk isMorphic ifTrue:
		[old := self currentWorld stepListSize.
		self currentWorld steppingMorphsNotInWorld do: [:m | m delete].
		self currentWorld cleanseStepList.
		(delta := (old - self currentWorld stepListSize)) > 0 ifTrue:
			[Transcript cr; show: (delta asString, ' morphs removed from steplist')]]

	"Utilities cleanseOtherworldlySteppers"! !

!Utilities class methodsFor: 'class initialization' stamp: 'mt 8/9/2016 14:36'!
cleanUp: aggressive
	aggressive ifTrue: [self setAuthorInitials: String empty].! !

!Utilities class methodsFor: 'common requests' stamp: 'mt 5/14/2016 09:40'!
commonRequestStrings

	(CommonRequestStrings == nil or: [CommonRequestStrings isKindOf: Array])
		ifTrue: [self initializeCommonRequestStrings].
	^ CommonRequestStrings! !

!Utilities class methodsFor: 'common requests' stamp: 'sw 12/12/96'!
commonRequestStrings: aString
	"Initialize the common request strings from aString.  "

	CommonRequestStrings := StringHolder new contents: aString! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'mdr 8/31/2000 18:54'!
convertCRtoLF: fileName
	"Convert the given file to LF line endings. Put the result in a file with the extention '.lf'"

	| in out c justPutCR |
	in := (FileStream readOnlyFileNamed: fileName) binary.
	out :=  (FileStream newFileNamed: fileName, '.lf') binary.
	justPutCR := false.
	[in atEnd] whileFalse: [
		c := in next.
		c = 10
			ifTrue: [
				out nextPut: 13.
				justPutCR := true]
			ifFalse: [
				(justPutCR and: [c = 10]) ifFalse: [out nextPut: c].
				justPutCR := false]].
	in close.
	out close.
! !

!Utilities class methodsFor: 'identification' stamp: ''!
copyrightNotice
	^ 'Copyright 1985-96, Apple Computer, Inc.'! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'nice 2/24/2013 17:45'!
createPageTestWorkspace
	"Used to generate a workspace window for testing page up and page down stuff."
	"Utilities createPageTestWorkspace"

	| numberOfLines maxStringLength minLineCounterSize lineCounterSize offsetSize stream headerConstant prevStart prevStrLen prevLineNumber |
	numberOfLines := 400.
	maxStringLength := 22.
	minLineCounterSize := 3.
	lineCounterSize := (numberOfLines log asInteger + 1) max: minLineCounterSize.
	offsetSize := 5.
	stream := String new writeStream.
	headerConstant := lineCounterSize + 1 + offsetSize + 1.
	prevStart := headerConstant negated.
	prevStrLen := 0.
	prevLineNumber := 0.
	numberOfLines timesRepeat: [ | log stringLen pad lineNumber charIndex start |
		stringLen := maxStringLength atRandom max: 1.
		lineNumber := prevLineNumber + 1.
		start := prevStart + prevStrLen + headerConstant + 1.
		prevStart := start.
		prevStrLen := stringLen.
		prevLineNumber := lineNumber.
		log := lineNumber log asInteger.
		pad := lineCounterSize - log - 1.
		pad timesRepeat: [stream nextPutAll: '0'].
		stream nextPutAll: lineNumber printString.
		stream space.
		log := start log asInteger.
		pad := offsetSize - log - 1.
		pad timesRepeat: [stream nextPutAll: '0'].
		stream nextPutAll: start printString.
		stream space.
		charIndex := 'a' first asInteger.
		stringLen timesRepeat: [ | char |
			char := Character value: charIndex.
			charIndex := charIndex + 1.
			stream nextPut: char].
		lineNumber = numberOfLines ifFalse: [stream cr]
		].
	UIManager default edit: stream contents label: 'Test Data'.
! !

!Utilities class methodsFor: 'identification' stamp: 'tk 4/10/1998 07:16'!
dateStamp
	"Utilities dateStamp"
	^ Date today mmddyyyy, ' ',
		((String streamContents: [:s | Time now print24: true on: s]) copyFrom: 1 to: 5)! !

!Utilities class methodsFor: 'identification' stamp: 'eem 1/5/2016 12:21'!
dateTimeSuffix
	"Answer a string which indicates the date and time, intended for use in building fileout filenames, etc."

	"Utilities dateTimeSuffix"
	^Preferences twentyFourHourFileStamps
		ifFalse:
			[self monthDayTimeStringFrom: Time totalSeconds]
		ifTrue:
			[self monthDayTime24StringFrom: Time totalSeconds]! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'nk 5/18/2003 13:03'!
decimalPlacesForFloatPrecision: precision
	"Answer the number of decimal places that correspond to the given floatPrecision"

	^ (#(1 0.1 0.01 0.001 0.0001 0.00001 0.000001 0.0000001 0.00000001 0.000000001) indexOf: precision ifAbsent: [ ^precision log negated floor ]) - 1

"
#(1 0.1 0.01 0.001 0.0001 0.00001 0.000001 0.0000001 0.00000001 0.000000001) collect: [:prec | Utilities decimalPlacesForFloatPrecision: prec]
"! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'nice 12/27/2009 03:11'!
decommissionTheAllCategory
	"Utilities decommissionTheAllCategory"
	"Moves all methods that are in a category named 'all' into the default 'as yet unclassified' category"

	| methodCount classCount |
	self flag: #ShouldBeMovedIntoClassOrganization. "sd"
	methodCount := 0.
	classCount := 0.
	self systemNavigation allBehaviorsDo:
		[:aClass | | org any | org := aClass organization.
			any := false.
			aClass selectorsDo:
				[:aSelector | | aCategory |
					aCategory := org categoryOfElement: aSelector.
					aCategory = #all ifTrue:
						[org classify: aSelector under: ClassOrganizer default suppressIfDefault: false.
						methodCount := methodCount + 1.
						any := true]].
			any ifTrue: [classCount := classCount + 1].
			org removeEmptyCategories].
	Transcript cr; show: methodCount printString, ' methods in ', classCount printString, ' classes moved
from "all" to "as yet unclassified"'
! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'fbs 3/4/2013 20:51'!
doesMethod: aSelector forClass: aClass bearInitials: initials
	"Answer whether a method bears the given initials at the head of its time stamp"

	| aTimeStamp implementingClass aMethod |
	implementingClass := aClass whichClassIncludesSelector: aSelector.
	implementingClass ifNil: [^ false].
	(aMethod := implementingClass compiledMethodAt: aSelector)
		ifNil: [^ false].
	^ (aTimeStamp := aMethod timeStamp) notNil and:
		[aTimeStamp beginsWith: initials]! !

!Utilities class methodsFor: 'debugging' stamp: 'sw 4/29/2001 23:42'!
doesNotUnderstand: aMessage
	"A temporary expedient for revectoring various messages formerly sent to Utilities that now are instead implemented by Flaps; this is only for the benefit of pre-existing buttons and menu items that were set up to call the old interface"

	| aSelector |
	aSelector := aMessage selector.
	(#(addLocalFlap explainFlaps addMenuFlap addPaintingFlap addStackToolsFlap addGlobalFlap offerGlobalFlapsMenu toggleWhetherToUseGlobalFlaps ) includes: aSelector)
		ifTrue:
			[^ self inform: 
'Sorry, this is an obsolete menu.  Please
dismiss it and get a fresh one.  Thank you'].

	^ super doesNotUnderstand: aMessage! !

!Utilities class methodsFor: 'common requests' stamp: 'tk 5/4/1998 17:35'!
editCommonRequestStrings
	"Let the user edit the common request strings.  2/1/96 sw"

	CommonRequestStrings openLabel: 'Common Request Strings'! !

!Utilities class methodsFor: 'common requests' stamp: 'nk 8/30/2004 08:02'!
eval: aString
	"Evaluate the string in a neutral context, and under certain circumstances print the 
	result in the transcript"

	| result |
	result := Compiler new evaluate: aString in: nil to: nil.
	(result isNumber) | (result isString)
		ifTrue:
			[Transcript cr; nextPutAll: result printString]! !

!Utilities class methodsFor: 'identification' stamp: 'di 6/13/97 13:52'!
fixStamp: changeStamp 
	| parts |
	parts := changeStamp findTokens: ' '.
	(parts size > 0 and: [parts last first isLetter]) ifTrue:
		["Put initials first in all time stamps..."
		^ String streamContents:
				[:s | s nextPutAll: parts last.
				parts allButLast do: [:p | s space; nextPutAll: p]]].
	^ changeStamp! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'nice 12/27/2009 03:11'!
fixUpProblemsWithAllCategory
	"Moves all methods that are in formally classified a category named '-- all --' into the default 'as yet unclassified' category"

	"Utilities fixUpProblemsWithAllCategory"

	| methodCount classCount |
	self flag: #ShouldBeMovedInClassOrganization.
	methodCount := 0.
	classCount := 0.
	self systemNavigation allBehaviorsDo:
		[:aClass | | org any | org := aClass organization.
			(org categories includes: #'-- all --') ifTrue:
				[any := false.
				aClass selectorsDo:
					[:aSelector | | aCategory |
						aCategory := org categoryOfElement: aSelector.
						aCategory = #'-- all --' ifTrue:
							[org classify: aSelector under: ClassOrganizer default suppressIfDefault: false.
							Transcript cr; show: aClass name, ' >> ', aSelector.
							methodCount := methodCount + 1.
							any := true]].
			any ifTrue: [classCount := classCount + 1].
			org removeEmptyCategories]].
	Transcript cr; show: methodCount printString, ' methods in ', classCount printString, ' classes moved from "-- all --" to "as yet unclassified"'
! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'sw 1/15/2007 20:48'!
floatPrecisionForDecimalPlaces: places
	"Answer the floatPrecision that corresponds to the given number of decimal places"

	^ places caseOf:
			{[0]->[1] .
			[1]-> [0.1] . 
			[2]-> [0.01] .
			[3]-> [0.001] .
			[4]-> [0.0001] .
			[5]-> [0.00001] .
			[6]-> [0.000001] .
			[7]-> [0.0000001] .
			[8]-> [0.00000001] .
			[9]-> [0.000000001].
			[10]->[0.0000000001]}
		otherwise:
			[(10.0 raisedTo: places negated) asFloat]

"
(0 to: 6) collect: [:i | Utilities floatPrecisionForDecimalPlaces: i]
(-10 to: 20) collect: [:i | Utilities floatPrecisionForDecimalPlaces: i]
"! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'nice 12/27/2009 03:11'!
garbageCollectAndReport
	"Do a garbage collection, and report results to the user."

	| reportString |
	reportString := String streamContents:
		[:aStream | 
			aStream nextPutAll: Smalltalk bytesLeftString.
			Smalltalk at: #Command ifPresent:
				[:cmdClass | | cc |
				(cc := cmdClass instanceCount) > 0 ifTrue:
					[aStream cr; nextPutAll:
		('(note: there are ', cc printString,
		                         ' undo record(s) present in your
system; purging them may free up more space.)')]]].
			
	self inform: reportString
! !

!Utilities class methodsFor: 'summer97 additions' stamp: 'sw 10/6/1998 14:09'!
graphicsFileSuffixes
	"Answer a list of filename suffixes which signal file content which we are able to internalize"

	^#('.gif' '.bmp' '.jpg' '.jpeg' '.jpe', '.form')! !

!Utilities class methodsFor: 'user interface' stamp: 'ar 8/6/2009 19:22'!
informUserDuring: aBlock
	"Display a message above (or below if insufficient room) the cursor during execution of the given block. This is now handled by UIManager."
	"Utilities informUserDuring:[:bar|
		#(one two three) do:[:info|
			bar value: info.
			(Delay forSeconds: 1) wait]]"
	^UIManager default informUserDuring: aBlock! !

!Utilities class methodsFor: 'user interface' stamp: 'ar 8/6/2009 19:21'!
informUser: aString during: aBlock
	"Display a message above (or below if insufficient room) the cursor during execution of the given block. This is now handled by UIManager."
	"Utilities informUser: 'Just a sec!!' during: [(Delay forSeconds: 1) wait]"
	^UIManager default informUser: aString during: aBlock.! !

!Utilities class methodsFor: 'class initialization' stamp: 'fbs 5/14/2013 17:23'!
initialize
	(self classVarNames includes: 'ScrapsBook') ifTrue: [self removeClassVarName: 'ScrapsBook'].! !

!Utilities class methodsFor: 'common requests' stamp: 'sw 2/5/2002 13:17'!
initializeCommonRequestStrings
	"Initialize the common request strings, a directly-editable list of expressions that can be evaluated from the 'do...' menu."

	CommonRequestStrings := StringHolder new contents: 
'Utilities emergencyCollapse.
Utilities closeAllDebuggers.
-
Sensor keyboard.
ParagraphEditor abandonChangeText.
Cursor normal show.
-
CommandHistory resetAllHistory.
Project allInstancesDo: [:p | p displayDepth: 16].
ScriptingSystem inspectFormDictionary.
Form fromUser bitEdit.
Display border: (0@0 extent: 640@480) width: 2.
-
Undeclared inspect.
Undeclared removeUnreferencedKeys; inspect.
Transcript clear.
Utilities grabScreenAndSaveOnDisk.
FrameRateMorph new openInHand.
-
Utilities reconstructTextWindowsFromFileNamed: ''TW''.
Utilities storeTextWindowContentsToFileNamed: ''TW''.
ChangeSorter removeEmptyUnnamedChangeSets.
ChangeSorter reorderChangeSets.
-
ActiveWorld installVectorVocabulary.
ActiveWorld abandonVocabularyPreference.
Smalltalk saveAsNewVersion'

"Utilities initializeCommonRequestStrings"! !

!Utilities class methodsFor: 'debugging' stamp: ''!
inspectCollection: aCollection notifying: aView
	aCollection size = 0 
		ifTrue: [aView notNil 
			ifTrue: [^ aView flash]
			ifFalse: [^ self]].
	aCollection size = 1
		ifTrue: [aCollection first inspect]
		ifFalse: [aCollection asArray inspect]! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'nice 2/24/2013 17:44'!
instanceComparisonsBetween: fileName1 and: fileName2
	"For differential results, run printSpaceAnalysis twice with different fileNames,
	then run this method...
		SpaceTally new printSpaceAnalysis: 0 on: 'STspace.text1'.
			--- do something that uses space here ---
		SpaceTally new printSpaceAnalysis: 0 on: 'STspace.text2'.
		Utilities instanceComparisonsBetween: 'STspace.text1' and: 'STspace.text2'"

	| instCountDict report f aString items className newInstCount oldInstCount newSpace oldPair oldSpace |
	instCountDict := Dictionary new.
	report := String new writeStream.
	f := FileStream readOnlyFileNamed: fileName1.
	[f atEnd] whileFalse:
		[aString := f nextLine.
		items := aString findTokens: ' '.
		(items size = 4 or: [items size = 5]) ifTrue:
			[instCountDict at: items first put: (Array with: items third asNumber with: items fourth asNumber)]].
	f close.

	f := FileStream readOnlyFileNamed: fileName2.
	[f atEnd] whileFalse:
		[aString := f nextLine.
		items := aString findTokens: ' '.
		(items size = 4 or: [items size = 5]) ifTrue:
			[className := items first.
			newInstCount := items third asNumber.
			newSpace := items fourth asNumber.
			oldPair := instCountDict at: className ifAbsent: [nil].
			oldInstCount := oldPair ifNil: [0] ifNotNil: [oldPair first].
			oldSpace := oldPair ifNil: [0] ifNotNil: [oldPair second].
			oldInstCount ~= newInstCount ifTrue:
				[report nextPutAll: (newInstCount - oldInstCount) printString; tab; nextPutAll: (newSpace - oldSpace) printString; tab; nextPutAll: className asString; cr]]].
	f close.

	(StringHolder new contents: report contents)
		openLabel: 'Instance count differentials between ', fileName1, ' and ', fileName2! !

!Utilities class methodsFor: 'miscellaneous' stamp: ''!
keyLike: aString satisfying: aBlock
	"Return a key like aString that satisfies aBlock.  The block should provide a test for acceptability -- typically the test is about whether the key is already in use.  aBlock should return a boolean.  8/11/96 sw"

	| stemAndSuffix suffix stem newKey |
	(aBlock value: aString) ifTrue: [^ aString].
	stemAndSuffix := aString stemAndNumericSuffix.
	suffix := stemAndSuffix last + 1.
	stem := stemAndSuffix first.
	[aBlock value: (newKey := stem, suffix printString)]
		whileFalse:
			[suffix := suffix + 1].
	^ newKey
! !

!Utilities class methodsFor: 'miscellaneous' stamp: ''!
keyLike: aString withTrailing: trailerString satisfying: aBlock
	"Return a key like (aString, trailerString) that satisfies aBlock.  The block should provide a test for acceptability -- typically the test is about whether the key is already in use.  aBlock should return a boolean.  8/11/96 sw"

	| stemAndSuffix suffix stem composite |
	composite := aString, trailerString.
	(aBlock value: composite) ifTrue: [^ composite].
	stemAndSuffix := aString stemAndNumericSuffix.
	suffix := stemAndSuffix last + 1.
	stem := stemAndSuffix first.
	[aBlock value: (composite := stem, suffix printString, trailerString)]
		whileFalse:
			[suffix := suffix + 1].
	^ composite
! !

!Utilities class methodsFor: 'identification' stamp: 'fbs 3/4/2013 20:51'!
methodsWithInitials: targetInitials
	"Based on a do-it contributed to the Squeak mailing list by Göran Hultgen:
 Browse methods whose initials (in the time-stamp, as logged to disk) match the given initials.
 Print out the complete time-stamp table to the Transcript.
 Answer a list of (initials -> count) associations.

CAUTION: It may take several minutes for this to complete."

	"Time millisecondsToRun: [Utilities methodsWithInitials: 'bf']"

	| initials timeStamp  allSubmitters |
	initials := ''.
	timeStamp := ''.
	allSubmitters := Bag new.
	self systemNavigation
		browseAllSelect:
			[:cm |
				timeStamp := cm timeStamp.
				initials := timeStamp isEmpty
					ifTrue:
						['']
					ifFalse:
						[timeStamp substrings first].
				initials := initials isEmpty
					ifTrue:
						['<no time stamp>']
					ifFalse:
						[initials first isDigit
							ifTrue:
								['<date>']
							ifFalse:
								[initials]].
				allSubmitters add: initials.
				(initials = targetInitials)]
		name: ('Methods with initials ', targetInitials)
		autoSelect: nil.

	allSubmitters sortedCounts do: [:elem | Transcript cr; show: elem asString].
	^ allSubmitters
! !

!Utilities class methodsFor: 'identification' stamp: 'sw 11/13/1999 23:03'!
monthDayTime24StringFrom: aSecondCount
	| aDate aTime |
	"From the date/time represented by aSecondCount, produce a string which indicates the date and time in the compact form
             ddMMMhhmm		where dd is a two-digit day-of-month, MMM is the alpha month abbreviation and hhmm is the time on a 24-hr clock.

          Utilities monthDayTime24StringFrom: Time primSecondsClock
"

	aDate := Date fromSeconds: aSecondCount.
	aTime := Time fromSeconds: aSecondCount \\ 86400.

	^ (aDate dayOfMonth asTwoCharacterString), 
		(aDate monthName copyFrom: 1 to: 3), 
		(aTime hhmm24)! !

!Utilities class methodsFor: 'identification' stamp: 'sw 11/13/1999 23:03'!
monthDayTimeStringFrom: aSecondCount
	| aDate aTime |
	"From the date/time represented by aSecondCount, produce a string which indicates the date and time in the form:
		ddMMMhhmmPP	  where:
							dd is a two-digit day-of-month,
							MMM is the alpha month abbreviation,
							hhmm is the time,
							PP is either am or pm

          Utilities monthDayTimeStringFrom: Time primSecondsClock
"

	aDate := Date fromSeconds: aSecondCount.
	aTime := Time fromSeconds: aSecondCount \\ 86400.

	^ (aDate dayOfMonth asTwoCharacterString), 
		(aDate monthName copyFrom: 1 to: 3), 
		((aTime hours \\ 12) asTwoCharacterString), 
		(aTime minutes asTwoCharacterString),
		(aTime hours > 12 ifTrue: ['pm'] ifFalse: ['am'])! !

!Utilities class methodsFor: 'miscellaneous' stamp: ''!
nextClockwiseSideAfter: aSide
 	aSide == #left ifTrue:
		[^ #top].
	aSide == #right ifTrue:
		[^ #bottom].
	aSide == #top ifTrue:
		[^ #right].
	^ #left! !

!Utilities class methodsFor: 'recent method submissions' stamp: 'fbs 9/3/2013 09:12'!
numberOfRecentSubmissionsToStore
	self deprecated: 'Use RecentMessages >> #numberOfRecentSubmissionsToStore'.
	^ RecentMessages numberOfRecentSubmissionsToStore.! !

!Utilities class methodsFor: 'recent method submissions' stamp: 'fbs 9/3/2013 09:13'!
numberOfRecentSubmissionsToStore: aNumber
	self deprecated: 'Use RecentMessages >> #numberOfRecentSubmissionsToStore'.
	RecentMessages numberOfRecentSubmissionsToStore: aNumber.! !

!Utilities class methodsFor: 'common requests' stamp: 'mt 4/12/2015 19:25'!
offerCommonRequests
	"Offer up the common-requests menu.  If the user chooses one, then evaluate it, and -- provided the value is a number or string -- show it in the Transcript."

	"Utilities offerCommonRequests"

	| reply result index normalItemCount strings lines |

	Smalltalk isMorphic ifTrue: [^ self offerCommonRequestsInMorphic].

	(CommonRequestStrings == nil or: [CommonRequestStrings isKindOf: Array])
		ifTrue:
			[self initializeCommonRequestStrings].
	
	strings := CommonRequestStrings contents.
	lines := strings asString lines copyWith: 'edit this menu'.
	normalItemCount := lines size - 1.
	
	index := UIManager default 
		chooseFrom: lines
		lines: (Array with: normalItemCount).

	index = 0 ifTrue: [^ self].
	index > normalItemCount ifTrue:
		[^ self editCommonRequestStrings].

	reply := lines at: index.
	reply size = 0 ifTrue: [^ self].

	result := Compiler evaluate: reply.
	(result isNumber) | (result isString)
		ifTrue:
			[Transcript cr; nextPutAll: result printString]! !

!Utilities class methodsFor: 'common requests' stamp: 'nice 12/28/2009 16:04'!
offerCommonRequestsInMorphic
	"Offer up the common-requests menu.  If the user chooses one, then evaluate it, and -- provided the value is a number or string -- show it in the Transcript."

	"Utilities offerCommonRequests"

	| aMenu  strings |

	(CommonRequestStrings == nil or: [CommonRequestStrings isKindOf: Array])
		ifTrue:
			[self initializeCommonRequestStrings].
	strings := CommonRequestStrings contents.
	aMenu := MenuMorph new.
	aMenu title: 'Common Requests' translated.
	aMenu addStayUpItem.
	strings asString linesDo:
		[:aString |
			aString = '-'
				ifTrue:
					[aMenu addLine]
				ifFalse:
					[aMenu add: (aString ifEmpty: [' ']) target: self selector: #eval: argument: aString]].

	aMenu addLine.
	aMenu add: 'edit this list' translated target: self action: #editCommonRequestStrings.
	aMenu popUpInWorld: self currentWorld! !

!Utilities class methodsFor: 'support windows' stamp: 'di 9/23/1998 02:02'!
openStandardWorkspace 
	"Open up a throwaway workspace with useful expressions in it.  1/22/96 sw"
	"Utilities openStandardWorkspace"

	(StringHolder new contents: self standardWorkspaceContents)
		openLabel: 'Useful Expressions ', Date today printString! !

!Utilities class methodsFor: 'user interface' stamp: 'fbs 12/6/2013 21:02'!
openUpdateButton
	PluggableButtonMorph new
		label: 'Load Code Updates';
		model: MCMcmUpdater;
		action: #updateFromServer;
		openInHand! !

!Utilities class methodsFor: 'miscellaneous' stamp: ''!
oppositeCornerFrom: aCorner
	"Answer the corner diagonally opposite to aCorner.  6/27/96 sw"

	aCorner == #topLeft
		ifTrue:
			[^ #bottomRight].
	aCorner == #topRight
		ifTrue:
			[^ #bottomLeft].
	aCorner == #bottomLeft
		ifTrue:
			[^ #topRight].
	^ #topLeft! !

!Utilities class methodsFor: 'miscellaneous' stamp: ''!
oppositeModeTo: aMode
 	aMode == #readOnly ifTrue: [^ #writeOnly].
	aMode == #writeOnly ifTrue: [^ #readOnly].
	^ aMode! !

!Utilities class methodsFor: 'miscellaneous' stamp: ''!
oppositeSideTo: aSide
 	aSide == #left ifTrue:
		[^ #right].
	aSide == #right ifTrue:
		[^ #left].
	aSide == #top ifTrue:
		[^ #bottom].
	^ #top! !

!Utilities class methodsFor: 'finding pointers' stamp: 'ar 7/16/2005 19:56'!
pointersToItem: index of: anArray
	"Find all occurrences in the system of pointers to the given element of the given array. This is useful for tracing up a pointer chain from an inspector on the results of a previous call of pointersTo:. To find out who points to the second element of the results, one would evaluate:

	Utilities pointersToItem: 2 of: self

in the inspector."
	^ self pointersTo: (anArray at: index) except: (Array with: anArray)! !

!Utilities class methodsFor: 'finding pointers' stamp: 'ar 7/16/2005 19:56'!
pointersTo: anObject
	"Find all occurrences in the system of pointers to the argument anObject."
	"(Utilities pointersTo: Browser) inspect."
	^ self pointersTo: anObject except: #()
! !

!Utilities class methodsFor: 'finding pointers' stamp: 'mtf 8/15/2008 23:54'!
pointersTo: anObject except: objectsToExclude 
	"Find all occurrences in the system of pointers to the argument
	anObject. Remove objects in the exclusion list from the
	results. "
	^ anObject inboundPointersExcluding: objectsToExclude! !

!Utilities class methodsFor: 'recent method submissions' stamp: 'fbs 9/9/2013 20:24'!
recentMethodSubmissions
	"Answer the list of recent method submissions, in order.  5/16/96 sw"
	self deprecated: 'Use RecentMessages default methodReferences'.
	^ RecentMessages default methodReferences.! !

!Utilities class methodsFor: 'class initialization' stamp: 'KR 8/12/2007 11:02'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: {#Utilities. #recentSubmissionsWindow. 	'Recent' translatedNoop.		'A message browser that tracks the most recently-submitted methods' translatedNoop}
						forFlapNamed: 'Tools'.]! !

!Utilities class methodsFor: 'investigations' stamp: 'nice 12/27/2009 03:11'!
reportSenderCountsFor: selectorList 
	"Produce a report on the number of senders of each of the selectors in 
	the list. 1/27/96 sw"
	| total report |
	total := 0.
	report := '
'.
	selectorList
		do: [:selector | | thisSize | 
			thisSize := (self systemNavigation allCallsOn: selector) size.
			report := report , thisSize printString , String tab , selector printString , String cr.
			total := total + thisSize].
	report := report , '--- ------------------
'.
	report := report , total printString , String tab , 'TOTAL
'.
	^ report! !

!Utilities class methodsFor: 'recent method submissions' stamp: 'fbs 9/9/2013 20:23'!
revertLastMethodSubmission
	self deprecated: 'Use RecentMessages default revertLastMethodSubmission'.
	RecentMessages default revertLastMethodSubmission.! !

!Utilities class methodsFor: 'identification' stamp: 'CdG 10/17/2005 20:56'!
setAuthorInitials
	"Put up a dialog allowing the user to specify the author's initials.  "

	self setAuthorInitials:
		(UIManager default request: 'Please type your initials: ' translated
					initialAnswer: AuthorInitials)! !

!Utilities class methodsFor: 'identification' stamp: 'fc 3/11/2004 12:13'!
setAuthorInitials: aString

	AuthorInitials := aString.

	"Case of being reset due to, eg, copy of image."
	aString isEmpty ifTrue: [AuthorName := '']! !

!Utilities class methodsFor: 'identification' stamp: 'CdG 10/17/2005 20:57'!
setAuthorName
	AuthorName := UIManager default 
			request: 'Please type your name:' translated
			initialAnswer: 'Your Name' translated! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'RAA 5/28/2001 10:02'!
setClassAndSelectorFrom: messageIDString in: csBlock
	"Decode strings of the form <className> [class] <selectorName>.   If <className> does not exist as a class, use nil for the class in the block"

	| aStream aClass maybeClass sel |

	(messageIDString isKindOf: MethodReference) ifTrue: [
		^messageIDString setClassAndSelectorIn: csBlock
	].

	aStream := ReadStream on: messageIDString.
	aClass := Smalltalk at: (aStream upTo: $ ) asSymbol ifAbsent: [nil].
	maybeClass := aStream upTo: $ .
	sel := aStream upTo: $ .
	((maybeClass = 'class') & (sel size ~= 0))
		ifTrue:
			[aClass
				ifNil:
					[csBlock value: nil value: sel asSymbol]
				ifNotNil:
					[csBlock value: aClass class value: sel asSymbol]]
		ifFalse:
			[csBlock value: aClass value: maybeClass asSymbol]


"
Utilities setClassAndSelectorFrom: 'Utilities class oppositeModeTo:' in: [:aClass :aSelector | Transcript cr; show: 'Class = ', aClass name printString, ' selector = ', aSelector printString].

Utilities setClassAndSelectorFrom: 'MessageSet setClassAndSelectorIn:' in: [:aClass :aSelector | Transcript cr; show: 'Class = ', aClass name printString, ' selector = ', aSelector printString].
"
! !

!Utilities class methodsFor: 'fetching updates' stamp: 'fbs 11/26/2013 14:45'!
setSystemVersionFromConfig: aConfig
	self deprecated: 'Use MCConfiguration >> #setSystemVersion'.
	aConfig setSystemVersion.! !

!Utilities class methodsFor: 'fetching updates' stamp: 'fbs 11/25/2013 15:09'!
setSystemVersion: aConfig
	self deprecated: 'Use MCConfiguration >> #setSystemVersion'.
	aConfig setSystemVersion.! !

!Utilities class methodsFor: 'graphical support' stamp: ''!
showFormsAcrossTopOfScreen: aFormList
	"Display the given array of forms across the top of the screen, wrapping to subsequent lines if needed.    Useful for example for looking at sets of rotations and animations.  6/10/96 sw"

	| position maxHeight screenBox ceiling |

	position := 20.
	maxHeight := 0.
	ceiling := 0.
	screenBox := Display boundingBox.
	aFormList do:
		[:elem | elem displayAt: (position @ ceiling).
			maxHeight := maxHeight max: elem boundingBox height.
			position := position + elem boundingBox width + 5.
			position > (screenBox right - 100) ifTrue:
				[position := 20.
				ceiling := ceiling + maxHeight + 10.
				maxHeight := 0]]! !

!Utilities class methodsFor: 'graphical support' stamp: 'nice 12/27/2009 03:48'!
showFormsDictAcrossTopOfScreen: formDict
	"Display the given Dictionary of forms across the top of the screen, wrapping to subsequent lines if needed.  Beneath each, put the name of the associated key."

	"Utilities showFormsDictAcrossTopOfScreen: HaloIcons"

	| screenBox position maxHeight ceiling |

	position := 20.
	maxHeight := 0.
	ceiling := 0.
	screenBox := Display boundingBox.
	formDict associationsDo:
		[:assoc | | box elem labelWidth keyString h |
			(elem := assoc value) displayAt: (position @ ceiling).
			box := elem boundingBox.
			h := box height.
			keyString := (assoc key isString) ifTrue: [assoc key] ifFalse: [assoc key printString].
			keyString displayAt: (position @ (ceiling + h)).
			labelWidth := TextStyle default defaultFont widthOfString: keyString.
			maxHeight := maxHeight max: h.
			position := position + (box width max: labelWidth) + 5.
			position > (screenBox right - 100) ifTrue:
				[position := 20.
				ceiling := ceiling + maxHeight + 15.
				maxHeight := 0]]! !

!Utilities class methodsFor: 'support windows' stamp: 'laza 7/14/2012 11:26'!
standardWorkspaceContents
	^ self class firstCommentAt: #standardWorkspaceContents

	"ChangeList browseRecentLog.
ChangeList browseRecent: 2000.

Preferences editAnnotations.
Flaps reinstateDefaultFlaps. 

(FileStream oldFileNamed: 'Lives of the Wolves') edit.
(FileStream oldFileNamed: 'tuesdayFixes.cs') fileIn.
ChangeList browseFile: 'myChanges.st'.

TextStyle default fontAt: 7 put: (StrikeFont new readMacFontHex: 'Cairo 18').

StandardSystemView doCacheBits.  ""restore fast windows mode in mvc""

Symbol selectorsContaining: 'rsCon'.
SystemNavigation default browseMethodsWhoseNamesContain: 'screen'.

Browser newOnClass: Utilities.
Browser fullOnClass: SystemDictionary.

FormView allInstances inspect.
StandardSystemView someInstance inspect.

Utilities storeTextWindowContentsToFileNamed: 'TextWindows'.

ScriptingSystem resetStandardPartsBin.

SystemOrganization categoryOfElement: #Controller. 
ParagraphEditor organization categoryOfElement: #changeEmphasis.

Cursor wait showWhile: [Sensor waitButton].

Smalltalk bytesLeft asStringWithCommas.
ByteSymbol instanceCount.
Time millisecondsToRun:
	[SystemNavigation default allCallsOn: #asOop].
MessageTally spyOn: [SystemNavigation default allCallsOn: #asOop].

"

"Utilities openStandardWorkspace"! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'ar 8/7/2009 22:31'!
steplistToolsWorkspace
	^ ((StringHolder new contents:  'self currentWorld listOfSteppingMorphs asArray inspectWithLabel: ''stepping morphs''.
Utilities cleanseOtherworldlySteppers.
self currentWorld steppingMorphsNotInWorld do: [:m | m delete].
self currentWorld stepListSummary.
self currentWorld stepListSize.
self currentHand attachMorph: FrameRateMorph new') openLabel: 'Steplist workspace')

setWindowColor: (Color r: 0.9 g: 0.7 b: 0.5);
			openInWorld: self currentWorld extent: (550 @ 140)

"Utilities steplistToolsWorkspace"! !

!Utilities class methodsFor: 'miscellaneous' stamp: 'dtl 2/14/2010 20:59'!
storeTextWindowContentsToFileNamed: aName
	"Utilities storeTextWindowContentsToFileNamed: 'TextWindows'"
	| aDict aRefStream |

	aDict := Project current textWindows..
	aDict size = 0 ifTrue: [^ self inform: 'no windows found to export.'].
	aRefStream := ReferenceStream fileNamed: aName.
	aRefStream nextPut: aDict.
	aRefStream close.
	self inform: 'Done!!  ', aDict size printString, ' window(s) exported.'! !

!Utilities class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 12:42'!
unload
	"Unload the receiver from global registries"

	self environment at: #Flaps ifPresent: [:cl |
	cl unregisterQuadsWithReceiver: self] ! !

!Utilities class methodsFor: 'fetching updates' stamp: 'fbs 11/23/2013 21:09'!
updateFromServer
	"Update the image by loading all pending updates from the server."
	self deprecated: 'Use MCMcmUpdater updateFromServer'.
	MCMcmUpdater updateFromServer.! !

Object subclass: #WeakFinalizationList
	instanceVariableNames: 'first'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Finalization'!

!WeakFinalizationList commentStamp: 'Igor.Stasenko 9/22/2010 21:09' prior: 0!
IMPORTANT!!!!!!

This class is a special object, recognized by VM.
Its only purpose is to 
a) identify a special kind of objects who usually having a weak references but
  also having an instance of me held by first non-weak fixed slot (instance variable).

b) a 'first' instance variable points to the head of a list of items, reported by VM which has weak references which became garbage during last garbage collection

At my class side, there are some public behavior, which is used by finalization process to detect if VM supports new finalization scheme or should use the old one.
Weak registry using #hasNewFinalization for switching to correct finalization logic,
depending on VM it currently runs on.
!

!WeakFinalizationList class methodsFor: 'vm capability test' stamp: 'eem 8/25/2015 10:33'!
hasNewFinalization
	^false! !

!WeakFinalizationList methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
first
	^ first! !

!WeakFinalizationList methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
swapWithNil

	| head |
	head := first.
	first := nil.
	^ head! !

Object weakSubclass: #WeakFinalizerItem
	instanceVariableNames: 'list next executor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Finalization'!

!WeakFinalizerItem commentStamp: 'Igor.Stasenko 9/22/2010 20:59' prior: 0!
My instances is used by weak registry to hold a single weak reference
and executor(s).

Once object, referenced weakly by my instance become garbage, a weak registry triggers its execution
by sending #finalizeValues to my instance.

Note, that a new VM finalization scheme does not implies to use this particular class
in order to implement finalization scheme. VM refers only to WeakFinalizationList class. 

In this way, my class and its implementation can serve as an example for implementing various finalization actions, which may differ from this one, provided by default for use by weak registry.

Once initialized, my instance should:
 - point to particular list (an instance of WeakFinalizationList),
 - next should be nil
 - executor or multiple executors initialized
 - weak reference slot should point to some object of interest

At the moment, when object, referenced weakly, become garbage, VM checks if its fist instance variable is an instance of WeakFinalizationList.
If it so, then it adds a given object to this list, and also links the tail of list through 'next' instance variable. 

So, as a result of garbage collection, a list will contain all objects, which had weak references to garbage collected objects. 
It is a responsibility of application to manage the instances of WeakFinalizationList's , as well as clear this list before the next garbage collection.
As a consequence of that you can:
 - use multiple different lists and manage them differently in order to react differently when some objects became garbage
 - you are not obliged to handle/clear the list(s) immediately after GC. You can clean up them periodically.
 - you can implement own kind of weak referencing object(s), which could use same finalization, provided by newer VMs.

VM requires only that an object with weak reference having at least two instance variables,
and its first instance variable points to instance of WeakFinalizationList. Everything else is optional.
!

!WeakFinalizerItem class methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 3/8/2010 21:00'!
new
	^ self basicNew: 1! !

!WeakFinalizerItem methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
add: newExecutor

	executor 
		ifNil: [ executor := newExecutor ]
		ifNotNil: [
			executor hasMultipleExecutors
				ifTrue: [ executor add: newExecutor]
				ifFalse: [ executor := ObjectFinalizerCollection with: executor with: newExecutor ]
		]! !

!WeakFinalizerItem methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
clear
	list := next := nil.! !

!WeakFinalizerItem methodsFor: 'copying' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
copyWithList: aList

	^ self copy list: aList! !

!WeakFinalizerItem methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
executor
	^ executor! !

!WeakFinalizerItem methodsFor: 'finalizing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
finalizeValues
	" cleanup the receiver, so it could be reused "
	| ex |
	ex := executor.
	executor := nil.
	next := nil.
	ex finalize.! !

!WeakFinalizerItem methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
list
	^ list! !

!WeakFinalizerItem methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
list: aList
	list := aList! !

!WeakFinalizerItem methodsFor: 'initialize-release' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
list: weakFinalizationList object: anObject
	self assert: (weakFinalizationList class == WeakFinalizationList).
	list := weakFinalizationList.
	self at: 1 put: anObject.
! !

!WeakFinalizerItem methodsFor: 'initialize-release' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
list: weakFinalizationList object: anObject executor: anExecutor
	self assert: (weakFinalizationList class == WeakFinalizationList).
	list := weakFinalizationList.
	self at: 1 put: anObject.
	executor := anExecutor! !

!WeakFinalizerItem methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
next
	^ next! !

!WeakFinalizerItem methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:06'!
object
	^ self at: 1! !

!WeakFinalizerItem methodsFor: 'copying' stamp: 'Igor.Stasenko 10/11/2010 15:06'!
postCopy
	executor hasMultipleExecutors ifTrue: [ executor := executor copy ].! !

Magnitude subclass: #MessageTally
	instanceVariableNames: 'class method blockNesting process tally receivers senders time gcStats maxClassNameSize maxClassPlusSelectorSize maxTabs reportOtherProcesses time0 startTime endTime preHibernationCopy'
	classVariableNames: 'DefaultPollPeriod ShowProcesses Timer'
	poolDictionaries: ''
	category: 'System-Tools'!

!MessageTally commentStamp: 'eem 7/16/2012 17:10' prior: 0!
My instances observe and report the amount of time spent in methods.

NOTE: a higher-level user interface (combining the MessageTally result tree with a method browser) is available from TimeProfileBrowser. Note that TimeProfileBrowser was not fancy with the different setting possibilities.

	TimeProfileBrowser spyOn:  [20 timesRepeat: 
			[Transcript show: 100 factorial printString]]
	

Strategies
-----------
MessageTally provides two different strategies available for profiling:

* spyOn: and friends use a high-priority Process to interrupt the block or process being spied on at periodic intervals. The interrupted call stack is then examined for caller information. See below for an example showing different settings

* tallySends: and friends use the interpreter simulator to run the block, recording every method call.

The two give you different results:

	* spyOn: gives you a view of where the time is being spent in your program, at least on a rough statistical level (assuming you've run the 	block for long enough and have a high enough poll rate). If you're trying to optimize your code, start here and optimize the methods where 	most of the time is being spent first.

	* tallySends: gives you accurate counts of how many times methods get called, and by exactly which route. If you're debugging, or trying to 	figure out if a given method is getting called too many times, this is your tool.

Q: How do you interpret MessageTally>>tallySends
A: The methods #tallySends and #spyOn: measure two very different quantities, but broken down in the same who-called-who format.  #spyOn: is approximate, but more indicative of real time spent, whereas #tallySends is exact and a precise record of how many times each method got executed.

Examples
----------

Here you can see all the processes computation time
	
		[1000 timesRepeat: [3.14159 printString. Processor yield]] fork.
		[1000 timesRepeat: [30 factorial. Processor yield]] fork.
		[1000 timesRepeat: [30 factorial. Processor yield]] fork.
		MessageTally spyAllOn: [ (Delay forMilliseconds: 100) wait] 

This profiles across a snapshot, writing the output to {imageName}.N.profile
	| fn |
	fn := (FileDirectory baseNameFor: Smalltalk image imageName), '.profile'.
	[MessageTally
			spyOn: [Smalltalk image snapshot: true andQuit: true]
			toFileNamed: fn
			reportOtherProcesses: true]
		on: FileExistsException
		do: [:ex|
			ex resume:
				(FileStream forceNewFileNamed: 
					(FileDirectory default
						nextNameFor:
							(FileDirectory localNameFor: 
								(FileDirectory baseNameFor: fn))
						extension: 'profile'))]

Settings
---------
You can change the printing format (that is, the whitespace and string compression) by using these instance methods: 
	maxClassNameSize:
	maxClassPlusSelectorSize:
	maxTabs:

You can change the default polling period (initially set to 1) by calling
	MessageTally defaultPollPeriod: numberOfMilliseconds


To understand the difference
----------------------------------
Here we see all the processes
	[1000 timesRepeat: [
		100 timesRepeat: [120 factorial].
		(Delay forMilliseconds: 10) wait
		]] forkAt: 45 named: '45'.
	MessageTally spyAllOn: [10000 timesRepeat: [1.23 printString]]
	
	
Here we only see the execution of the expression [10000 timesRepeat: [1.23 printString]
	[1000 timesRepeat: [
		100 timesRepeat: [120 factorial].
		(Delay forMilliseconds: 10) wait
		]] forkAt: 45 named: '45'.
	MessageTally spyOn: [10000 timesRepeat: [1.23 printString]]
	
Here we only check the exact message sends: this is not a pc-sampling approach
	[1000 timesRepeat: [
		100 timesRepeat: [120 factorial].
		(Delay forMilliseconds: 10) wait
		]] forkAt: 45 named: '45'.
	MessageTally tallySends: [10000 timesRepeat: [1.23 printString]]
	


!

!MessageTally class methodsFor: 'defaults' stamp: 'jmv 4/30/2010 14:10'!
defaultMaxClassNameSize
	"Return the default maximum width of the class name alone"
	^50! !

!MessageTally class methodsFor: 'defaults' stamp: 'jmv 4/30/2010 14:10'!
defaultMaxClassPlusSelectorSize
	"Return the default maximum width of the class plus selector together (not counting the '>>')"
	^100! !

!MessageTally class methodsFor: 'defaults' stamp: 'jmv 3/2/2009 12:32'!
defaultMaxTabs
	"Return the default number of tabs after which leading white space is compressed"
	^120! !

!MessageTally class methodsFor: 'defaults' stamp: 'nk 3/8/2004 12:41'!
defaultPollPeriod
	"Answer the number of milliseconds between interrupts for spyOn: and friends.
	This should be faster for faster machines."
	^DefaultPollPeriod ifNil: [ DefaultPollPeriod := 1 ]! !

!MessageTally class methodsFor: 'defaults' stamp: 'nk 3/8/2004 12:41'!
defaultPollPeriod: numberOfMilliseconds
	"Set the default number of milliseconds between interrupts for spyOn: and friends.
	This should be faster for faster machines."
	DefaultPollPeriod := numberOfMilliseconds! !

!MessageTally class methodsFor: 'class initialization' stamp: 'eem 7/13/2012 16:15'!
initialize
	"MessageTally initialize"
	"By default, show each process separately"
	ShowProcesses ifNil: [ShowProcesses := true].
	Smalltalk
		addToShutDownList: self after: Delay; "i.e. convert to relative times & stats after Delay sleeps."
		addToStartUpList: self before: Delay "i.e. convert back to absolute stats before Delay wakes."! !

!MessageTally class methodsFor: 'system startup' stamp: 'eem 7/16/2012 17:05'!
runningInstance
	"Answer the running instance if the Timer is running."
	^Timer ifNotNil:
		[((Processor activeProcess == Timer
			ifTrue: [thisContext]
			ifFalse: [Timer suspendedContext]) findContextSuchThat:
				[:ctxt| ctxt receiver isKindOf: self]) ifNotNil:
					[:ctxt| ctxt receiver]]! !

!MessageTally class methodsFor: 'defaults' stamp: 'ar 9/4/2009 00:09'!
showProcesses
	"Indicates whether to show each process separately or cumulatively.
	For example, compare the spy results of the following with both values:
	
		[1000 timesRepeat: [3.14159 printString. Processor yield]] fork.
		[1000 timesRepeat: [30 factorial. Processor yield]] fork.
		[1000 timesRepeat: [30 factorial. Processor yield]] fork.
		MessageTally spyAllOn: [ (Delay forMilliseconds: 100) wait] 

	"
	<preference: 'Show Tally Processes'
		category: 'debug'
		description: 'When true, show individual processes in MessageTally'
		type: #Boolean>
	^ShowProcesses! !

!MessageTally class methodsFor: 'defaults' stamp: 'jmv 3/4/2009 10:29'!
showProcesses: aBool
	"Indicates whether to show each process separately or cumulatively.
	For example, compare the spy results of the following with both values:
	
		[1000 timesRepeat: [3.14159 printString. Processor yield]] fork.
		[1000 timesRepeat: [30 factorial. Processor yield]] fork.
		[1000 timesRepeat: [30 factorial. Processor yield]] fork.
		MessageTally spyAllOn: [ (Delay forMilliseconds: 100) wait]

	"
	ShowProcesses := aBool.! !

!MessageTally class methodsFor: 'spying' stamp: 'ul 2/22/2010 16:47'!
showReport: node
	"Open a string holder with the reports from the given node"
	(StringHolder new contents: 
		(String streamContents: [:s | node report: s ]))
			openLabel: 'Spy Results'! !

!MessageTally class methodsFor: 'system startup' stamp: 'eem 7/13/2012 15:14'!
shutDown: quitting
	"This message is sent on system shutdown to registered classes"
	self runningInstance ifNotNil: [:runningInstance| runningInstance hibernate]! !

!MessageTally class methodsFor: 'spying' stamp: 'ul 2/22/2010 16:47'!
spyAllOn: aBlock
	"Spy on all the processes in the system
	
	[1000 timesRepeat: [3.14159 printString. Processor yield]] fork.
	[1000 timesRepeat: [20 factorial. Processor yield]] fork.
	[1000 timesRepeat: [20 factorial. Processor yield]] fork.
	MessageTally spyAllOn: [ (Delay forMilliseconds: 100) wait]
	
	"
	| node result |
	node := self new.
	node reportOtherProcesses: true. "Irrelevant in this case. All processes will be reported on their own."
	result := node spyAllEvery: self defaultPollPeriod on: aBlock.
	self showReport: node.
	^ result! !

!MessageTally class methodsFor: 'spying' stamp: 'ul 2/22/2010 16:50'!
spyOnProcess: aProcess forMilliseconds: msecDuration 
	"
	Spy on aProcess for a certain amount of time
	| p1 p2 |  
	p1 := [100000 timesRepeat: [3.14159 printString. Processor yield]] newProcess.  
	p2 := [100000 timesRepeat: [3.14159 printString. Processor yield]] newProcess.
	p1 resume.
	p2 resume.  
	(Delay forMilliseconds: 100) wait.  
	MessageTally spyOnProcess: p1 forMilliseconds: 1000
	"
	
	^self 
		spyOnProcess: aProcess
		forMilliseconds: msecDuration
		reportOtherProcesses: ShowProcesses
! !

!MessageTally class methodsFor: 'spying' stamp: 'ul 2/22/2010 16:50'!
spyOnProcess: aProcess forMilliseconds: msecDuration reportOtherProcesses: aBoolean
	"
	Spy on aProcess for a certain amount of time
	| p1 p2 |  
	p1 := [100000 timesRepeat: [3.14159 printString. Processor yield]] newProcess.  
	p2 := [100000 timesRepeat: [3.14159 printString. Processor yield]] newProcess.
	p1 resume.
	p2 resume.  
	(Delay forMilliseconds: 100) wait.  
	MessageTally spyOnProcess: p1 forMilliseconds: 1000 reportOtherProcesses: true
	"
	| node |
	node := self new.
	node reportOtherProcesses: aBoolean.
	node
		spyEvery: self defaultPollPeriod
		onProcess: aProcess
		forMilliseconds: msecDuration.
	self showReport: node.! !

!MessageTally class methodsFor: 'spying' stamp: 'ul 2/22/2010 16:50'!
spyOnProcess: aProcess forMilliseconds: msecDuration toFileNamed: fileName 
	"Spy on the evaluation of aProcess. Write the data collected on a file  
	named fileName. Will overwrite fileName"
	| node |
	node := self new.
	node
		spyEvery: self defaultPollPeriod
		onProcess: aProcess
		forMilliseconds: msecDuration.
	FileStream fileNamed: fileName do: [ :file |
		node report: file ]! !

!MessageTally class methodsFor: 'spying' stamp: 'ul 2/24/2010 00:06'!
spyOnProcess: aProcess forMilliseconds: msecDuration toFileNamed: fileName reportOtherProcesses: aBoolean
	"Spy on the evaluation of aProcess. Write the data collected on a file  
	named fileName. Will overwrite fileName"

	| node |
	node := self new.
	node reportOtherProcesses: aBoolean.
	node
		spyEvery: self defaultPollPeriod
		onProcess: aProcess
		forMilliseconds: msecDuration.
	FileStream fileNamed: fileName do: [ :file |
		node report: file ]! !

!MessageTally class methodsFor: 'spying' stamp: 'ar 2/3/2010 21:13'!
spyOn: aBlock
	"Spy on aBlock, in the current process. Can include or not statistics on other processes in the report.
	[1000 timesRepeat: [
		100 timesRepeat: [120 factorial].
		(Delay forMilliseconds: 10) wait
		]] forkAt: 45 named: '45'.
	MessageTally spyOn: [10000 timesRepeat: [1.23 printString]]
	"
	^self spyOn: aBlock reportOtherProcesses: ShowProcesses! !

!MessageTally class methodsFor: 'spying' stamp: 'ul 2/22/2010 16:48'!
spyOn: aBlock reportOtherProcesses: aBoolean
	"
	Spy on aBlock, in the current process. Can include or not statistics on other processes in the report.
	[1000 timesRepeat: [
		100 timesRepeat: [120 factorial].
		(Delay forMilliseconds: 10) wait
		]] forkAt: 45 named: '45'.
	MessageTally spyOn: [10000 timesRepeat: [1.23 printString]] reportOtherProcesses: true
	"
	| node result |
	node := self new.
	node reportOtherProcesses: aBoolean.
	result := node spyEvery: self defaultPollPeriod on: aBlock.
	self showReport: node.
	^ result! !

!MessageTally class methodsFor: 'spying' stamp: 'ul 2/22/2010 16:48'!
spyOn: aBlock toFileNamed: fileName 
	"Spy on the evaluation of aBlock. Write the data collected on a file
	named fileName."

	| value node |
	node := self new.
	value := node spyEvery: self defaultPollPeriod on: aBlock.
	FileStream newFileNamed: fileName do: [ :file |
		node report: file ].
	^value! !

!MessageTally class methodsFor: 'spying' stamp: 'ul 2/24/2010 00:06'!
spyOn: aBlock toFileNamed: fileName reportOtherProcesses: aBoolean
	"Spy on the evaluation of aBlock. Write the data collected on a file
	named fileName."

	| value node |
	node := self new.
	node reportOtherProcesses: aBoolean.
	value := node spyEvery: self defaultPollPeriod on: aBlock.
	FileStream newFileNamed: fileName do: [ :file |
		node report: file ].
	^value! !

!MessageTally class methodsFor: 'system startup' stamp: 'eem 7/13/2012 16:58'!
startUp: resuming
	"This message is sent to registered classes when the system is coming up."
	self runningInstance ifNotNil: [:runningInstance| runningInstance unhibernate: resuming]! !

!MessageTally class methodsFor: 'spying' stamp: 'nice 12/25/2013 22:04'!
tallySendsTo: receiver inBlock: aBlock showTree: treeOption
	"
	MessageTally tallySends: [3.14159 printString]
	"
	"This method uses the simulator to count the number of calls on each method
	invoked in evaluating aBlock. If receiver is not nil, then only sends
	to that receiver are tallied.
	Results are presented as leaves, sorted by frequency,
	preceded, optionally, by the whole tree."
	| prev tallies startTime totalTime |
	startTime := Time millisecondClockValue.
	tallies := MessageTally new class: aBlock receiver class method: aBlock method.
	tallies reportOtherProcesses: true.	"Do NOT filter nodes with nil process"
	prev := aBlock.
	thisContext sender
		runSimulated: aBlock
		contextAtEachStep: [ :current |
			current == prev ifFalse: [ "call or return"
				prev sender == nil ifFalse: [ "call only"
					(receiver == nil or: [ current receiver == receiver ])
						ifTrue: [ tallies tally: current by: 1 ] ].
				prev := current ] ].

	totalTime := Time millisecondClockValue - startTime // 1000.0.
	(StringHolder new contents:
		(String streamContents:
			[:s |
			s nextPutAll: 'This simulation took ' , (totalTime printShowingMaxDecimalPlaces: 2)
							, ' seconds.'; cr.
			treeOption
				ifTrue: [ tallies fullPrintExactOn: s ]
				ifFalse: [ tallies leavesPrintExactOn: s ] ]))
		openLabel: 'Spy Results'! !

!MessageTally class methodsFor: 'spying' stamp: 'jmv 9/24/2009 16:17'!
tallySends: aBlock
	"
	MessageTally tallySends: [3.14159 printString]
	"

	^ self tallySendsTo: nil inBlock: aBlock showTree: true! !

!MessageTally class methodsFor: 'as yet unclassified' stamp: 'ul 2/22/2010 16:47'!
terminateTimerProcess

	Timer ifNotNil: [
		Timer terminate.
		Timer := nil ].
! !

!MessageTally class methodsFor: 'spying' stamp: ''!
time: aBlock

	^ Time millisecondsToRun: aBlock! !

!MessageTally methodsFor: 'accessing' stamp: 'eem 1/28/2015 16:37'!
blockNesting
	^blockNesting! !

!MessageTally methodsFor: 'tallying' stamp: 'eem 1/28/2015 16:28'!
blockNestingCountOf: context
	| count nest |
	count := 0.
	nest := context.
	[nest closure notNil] whileTrue:
		[count := count + 1.
		 nest := nest closure outerContext].
	^count! !

!MessageTally methodsFor: 'tallying' stamp: ''!
bumpBy: count

	tally := tally + count! !

!MessageTally methodsFor: 'collecting leaves' stamp: ''!
bump: hitCount
	tally := tally + hitCount! !

!MessageTally methodsFor: 'collecting leaves' stamp: ''!
bump: hitCount fromSender: senderTally
	"Add this hitCount to the total, and include a reference to the
	sender responsible for the increment"
	self bump: hitCount.
	senders == nil ifTrue: [senders := OrderedCollection new].
	senderTally == nil
		ifFalse: [senders add: (senderTally copyWithTally: hitCount)]! !

!MessageTally methodsFor: 'private' stamp: 'eem 1/28/2015 16:30'!
class: aClass method: aMethod

	class := aClass.
	method := aMethod.
	tally := blockNesting := 0.
	receivers := {}! !

!MessageTally methodsFor: 'private' stamp: 'eem 1/28/2015 16:30'!
class: aClass method: aMethod nesting: blockNestingCount

	class := aClass.
	method := aMethod.
	blockNesting := blockNestingCount.
	tally := 0.
	receivers := {}! !

!MessageTally methodsFor: 'private' stamp: 'fbs 7/25/2013 07:14'!
computeGCStats
	"Compute the deltas in the GC stats.  Serves for reporting, hibernating and unhibernating."
	Smalltalk getVMParameters keysAndValuesDo:
		[ :idx :gcVal |
		gcVal isNumber ifTrue: [gcStats at: idx put: (gcVal - (gcStats at: idx))]]! !

!MessageTally methodsFor: 'private' stamp: 'eem 7/13/2012 16:55'!
convertToOrFromDeltaForHibernation
	"Convert between absolute and relative times and stats (and back!!)."
	| now |
	now := Time millisecondClockValue.
	startTime ifNotNil: [startTime := now - startTime].
	time0 ifNotNil: [time0 := now - time0].
	endTime ifNotNil: [endTime := now - endTime].
	self computeGCStats! !

!MessageTally methodsFor: 'private' stamp: 'eem 1/28/2015 16:41'!
copyWithTally: hitCount
	^ (MessageTally new class: class method: method nesting: blockNesting)
		reportOtherProcesses: reportOtherProcesses;
		process: process;
		bump: hitCount! !

!MessageTally methodsFor: 'printing' stamp: 'jmv 9/24/2009 15:49'!
fullPrintExactOn: aStream
	aStream nextPutAll: '**Tree**'; cr.
	self
		treePrintOn: aStream
		tabs: OrderedCollection new
		thisTab: ''
		total: tally
		totalTime: time
		tallyExact: true
		orThreshold: nil.
	aStream nextPut: Character newPage; cr.
	aStream nextPutAll: '**Leaves**'; cr.
	self leavesPrintExactOn: aStream! !

!MessageTally methodsFor: 'printing' stamp: 'jmv 9/24/2009 15:49'!
fullPrintOn: aStream threshold: perCent
	| threshold |  
	threshold := (perCent asFloat / 100 * tally) rounded.
	aStream nextPutAll: '**Tree**'; cr.
	self
		rootPrintOn: aStream
		total: tally
		totalTime: time
		threshold: threshold.
	aStream nextPut: Character newPage; cr.
	aStream nextPutAll: '**Leaves**'; cr.
	self
		leavesPrintOn: aStream
		threshold: threshold! !

!MessageTally methodsFor: 'comparing' stamp: 'nice 12/24/2009 03:02'!
hash
	"Hash is reimplemented because = is implemented."

	^method scaledIdentityHash! !

!MessageTally methodsFor: 'private' stamp: 'eem 7/13/2012 16:57'!
hibernate
	"Sent before snapshot.  Convert absolute times and stats into relative times and stats
	 Also copy the receiver for the non-snapshot path."
	preHibernationCopy := self shallowCopy.
	gcStats := gcStats copy. "So as to leave preHibernationCopy's alone."
	self convertToOrFromDeltaForHibernation! !

!MessageTally methodsFor: 'initialization' stamp: 'ar 2/3/2010 21:22'!
initialize
	"Initialize the receiver"
	super initialize.
	maxClassNameSize := self class defaultMaxClassNameSize.
	maxClassPlusSelectorSize := self class defaultMaxClassPlusSelectorSize.
	maxTabs := self class defaultMaxTabs.
	reportOtherProcesses := false.
! !

!MessageTally methodsFor: 'collecting leaves' stamp: 'eem 1/28/2015 16:42'!
into: leafDict fromSender: senderTally
	| leafNode |
	leafNode := leafDict
					at: method
					ifAbsentPut:
						[(self class new class: class method: method nesting: blockNesting)
							process: process;
							reportOtherProcesses: reportOtherProcesses].
	leafNode bump: tally fromSender: senderTally! !

!MessageTally methodsFor: 'comparing' stamp: ''!
isPrimitives
	"Detect pseudo node used to carry tally of local hits"
	^ receivers == nil! !

!MessageTally methodsFor: 'collecting leaves' stamp: 'jmv 9/24/2009 16:07'!
leavesInto: leafDict fromSender: senderTally

	| rcvrs |
	rcvrs := self sonsOver: 0.
	rcvrs size = 0
		ifTrue: [ self into: leafDict fromSender: senderTally ]
		ifFalse: [
			
			(reportOtherProcesses not and: [ rcvrs anyOne process isNil ]) ifTrue: [
				^self].
			
			rcvrs do: [ :node |
				node isPrimitives
					ifTrue: [ node leavesInto: leafDict fromSender: senderTally ]
					ifFalse: [ node leavesInto: leafDict fromSender: self ]]]! !

!MessageTally methodsFor: 'printing' stamp: 'ul 2/4/2010 15:34'!
leavesPrintExactOn: aStream
	| dict |
	dict := IdentityDictionary new: 100.
	self leavesInto: dict fromSender: nil.
	dict values sort
		do: [ :node |
			node printOn: aStream total: tally totalTime: nil tallyExact: true.
			node printSenderCountsOn: aStream ]! !

!MessageTally methodsFor: 'printing' stamp: 'ul 2/4/2010 15:34'!
leavesPrintOn: aStream threshold: threshold
	| dict |
	dict := IdentityDictionary new: 100.
	self leavesInto: dict fromSender: nil.
	(dict values select: [:node | node tally > threshold])
		sort do: [:node |
			node printOn: aStream total: tally totalTime: time tallyExact: false ]! !

!MessageTally methodsFor: 'printing format' stamp: 'nk 3/8/2004 12:29'!
maxClassNameSize
	^maxClassNameSize! !

!MessageTally methodsFor: 'printing format' stamp: 'nk 3/8/2004 12:30'!
maxClassNameSize: aNumber
	maxClassNameSize := aNumber! !

!MessageTally methodsFor: 'printing format' stamp: 'nk 3/8/2004 12:29'!
maxClassPlusSelectorSize
	^maxClassPlusSelectorSize! !

!MessageTally methodsFor: 'printing format' stamp: 'nk 3/8/2004 12:30'!
maxClassPlusSelectorSize: aNumber
	maxClassPlusSelectorSize := aNumber! !

!MessageTally methodsFor: 'printing format' stamp: 'nk 3/8/2004 12:29'!
maxTabs
	^maxTabs! !

!MessageTally methodsFor: 'printing format' stamp: 'nk 3/8/2004 12:30'!
maxTabs: aNumber
	maxTabs := aNumber! !

!MessageTally methodsFor: 'private' stamp: ''!
method

	^method! !

!MessageTally methodsFor: 'private' stamp: ''!
primitives: anInteger

	tally := anInteger.
	receivers := nil! !

!MessageTally methodsFor: 'printing' stamp: 'eem 1/28/2015 16:46'!
printOn: aStream 
	| className |
	(class isNil or: [method isNil]) ifTrue: [^super printOn: aStream].
	className := method methodClass name contractTo: self maxClassNameSize.
	blockNesting ifNotNil:
		[blockNesting timesRepeat:
			[aStream nextPutAll: '[] in ']].
	aStream
		nextPutAll: className;
		nextPutAll: ' >> ';
		nextPutAll: (method selector contractTo: self maxClassPlusSelectorSize - className size)! !

!MessageTally methodsFor: 'printing' stamp: 'eem 1/28/2015 16:46'!
printOn: aStream total: total totalTime: totalTime tallyExact: isExact 
	| aSelector className myTally aClass percentage |
	isExact 
		ifTrue: 
			[myTally := tally.
			receivers == nil 
				ifFalse: [receivers do: [:r | myTally := myTally - r tally]].
			aStream
				print: myTally;
				space]
		ifFalse: 
			[percentage := tally asFloat / total * 100.0.
			aStream
				nextPutAll: (percentage printShowingDecimalPlaces: 1);
				nextPutAll: '% {';
				print: (percentage * totalTime / 100) rounded;
				nextPutAll: 'ms} '].
	receivers == nil 
		ifTrue: 
			[aStream
				nextPutAll: 'primitives';
				cr]
		ifFalse: 
			[aSelector := method selector.
			aClass := method methodClass.
			className := aClass name contractTo: self maxClassNameSize.
			blockNesting > 0 ifTrue:
				[aStream
					next: blockNesting put: $[;
					next: blockNesting put: $];
					space].
			aStream
				nextPutAll: class name;
				nextPutAll: (aClass = class 
							ifTrue: ['>>']
							ifFalse: ['(' , aClass name , ')>>']);
				nextPutAll: (aSelector 
							contractTo: self maxClassPlusSelectorSize - className size);
				cr]! !

!MessageTally methodsFor: 'printing' stamp: 'ul 2/5/2010 21:45'!
printSenderCountsOn: aStream
	| mergedSenders |
	mergedSenders := IdentityDictionary new.
	senders do:
		[:node |
		| mergedNode |
		mergedNode := mergedSenders at: node method ifAbsent: [nil].
		mergedNode == nil
			ifTrue: [mergedSenders at: node method put: node]
			ifFalse: [mergedNode bump: node tally]].
	mergedSenders values sort do:
		[:node | 
		10 to: node tally printString size by: -1 do: [:i | aStream space].
		node printOn: aStream total: tally totalTime: nil tallyExact: true]! !

!MessageTally methodsFor: 'private' stamp: 'ar 3/3/2009 19:29'!
process
	^process! !

!MessageTally methodsFor: 'private' stamp: 'ar 3/3/2009 19:29'!
process: aProcess
	process := aProcess! !

!MessageTally methodsFor: 'reporting' stamp: 'cmm 7/28/2012 11:22'!
reportGCStatsOn: str
	| oldSpaceEnd youngSpaceEnd memoryEnd fullGCs fullGCTime incrGCs incrGCTime tenureCount upTime rootOverflows |
	upTime := time.
	oldSpaceEnd			:= gcStats at: 1.
	youngSpaceEnd		:= gcStats at: 2.
	memoryEnd			:= gcStats at: 3.
	fullGCs				:= gcStats at: 7.
	fullGCTime			:= gcStats at: 8.
	incrGCs				:= gcStats at: 9.
	incrGCTime			:= gcStats at: 10.
	tenureCount			:= gcStats at: 11.
	rootOverflows		:= gcStats at: 22.

	str cr.
	str	nextPutAll: '**Memory**'; cr.
	str	nextPutAll:	'	old			';
		nextPutAll: (oldSpaceEnd asStringWithCommasSigned: true); nextPutAll: ' bytes'; cr.
	str	nextPutAll: '	young		';
		nextPutAll: ((youngSpaceEnd - oldSpaceEnd) asStringWithCommasSigned: true); nextPutAll: ' bytes'; cr.
	str	nextPutAll: '	used		';
		nextPutAll: (youngSpaceEnd asStringWithCommasSigned: true); nextPutAll: ' bytes'; cr.
	str	nextPutAll: '	free		';
		nextPutAll: ((memoryEnd - youngSpaceEnd) asStringWithCommasSigned: true); nextPutAll: ' bytes'; cr.

	str cr.
	str	nextPutAll: '**GCs**'; cr.
	str	nextPutAll: '	full			';
		print: fullGCs; nextPutAll: ' totalling '; nextPutAll: fullGCTime asStringWithCommas; nextPutAll: ' ms (';
		print: (fullGCTime / upTime * 100) maxDecimalPlaces: 2;
		nextPutAll: '% uptime)'.
	fullGCs = 0 ifFalse:
		[str	nextPutAll: ', avg '; print: (fullGCTime / fullGCs) maxDecimalPlaces: 1; nextPutAll: ' ms'].
	str	cr.
	str	nextPutAll: '	incr			';
		print: incrGCs; nextPutAll: ' totalling '; nextPutAll: incrGCTime asStringWithCommas; nextPutAll: ' ms (';
		print: (incrGCTime / upTime * 100) maxDecimalPlaces: 1;
		nextPutAll: '% uptime)'.
	incrGCs = 0 ifFalse:
		[str nextPutAll: ', avg '; print: (incrGCTime / incrGCs) maxDecimalPlaces: 1; nextPutAll: ' ms'].
	str cr.
	str	nextPutAll: '	tenures		';
		nextPutAll: tenureCount asStringWithCommas.
	tenureCount = 0 ifFalse:
		[str nextPutAll: ' (avg '; print: incrGCs // tenureCount; nextPutAll: ' GCs/tenure)'].
	str	cr.
	str	nextPutAll: '	root table	';
		nextPutAll: rootOverflows asStringWithCommas; nextPutAll: ' overflows'.
	str cr.
! !

!MessageTally methodsFor: 'accessing' stamp: 'ar 2/3/2010 21:21'!
reportOtherProcesses
	"If true, reports all processes in the system"
	^reportOtherProcesses! !

!MessageTally methodsFor: 'accessing' stamp: 'ar 2/3/2010 21:21'!
reportOtherProcesses: aBoolean
	"If true, reports all processes in the system"
	reportOtherProcesses := aBoolean! !

!MessageTally methodsFor: 'reporting' stamp: 'jmv 3/4/2009 09:27'!
report: strm 
	"Print a report, with cutoff percentage of each element of the tree 
	(leaves, roots, tree), on the stream, strm."

	self report: strm cutoff: 1! !

!MessageTally methodsFor: 'reporting' stamp: 'jmv 9/24/2009 15:35'!
report: strm cutoff: threshold 
	tally = 0
		ifTrue: [strm nextPutAll: ' - no tallies obtained']
		ifFalse: 
			[strm nextPutAll: ' - '; print: tally; nextPutAll: ' tallies, ', time printString, ' msec.'; cr; cr.
			self fullPrintOn: strm threshold: threshold].
		
	time isZero ifFalse:	
		[self reportGCStatsOn: strm].! !

!MessageTally methodsFor: 'printing' stamp: 'ul 5/6/2015 13:03'!
rootPrintOn: aStream total: total totalTime: totalTime threshold: threshold

	| groups sons |
	sons := self sonsOver: threshold.
	groups := sons groupBy: [ :aTally | aTally process ].
	groups keysAndValuesDo: [ :p :g |
		(reportOtherProcesses or: [ p notNil ]) ifTrue: [
			aStream nextPutAll: '--------------------------------'; cr.
			aStream nextPutAll: 'Process: ',  (p ifNil: [ 'other processes'] ifNotNil: [ p browserPrintString]); cr.
			aStream nextPutAll: '--------------------------------'; cr.
			g sort do:[:aSon | 
				aSon 
					treePrintOn: aStream
					tabs: OrderedCollection new
					thisTab: ''
					total: total
					totalTime: totalTime
					tallyExact: false
					orThreshold: threshold]].
	]! !

!MessageTally methodsFor: 'comparing' stamp: 'eem 1/28/2015 16:43'!
sonsOver: threshold

	| hereTally last sons |
	receivers basicSize = 0 ifTrue: [^#()].
	hereTally := tally.
	sons := receivers select:  "subtract subNode tallies for primitive hits here"
		[:son |
		hereTally := hereTally - son tally.
		son tally > threshold].
	hereTally > threshold ifTrue:
		[last := MessageTally new class: class method: method nesting: blockNesting.
		 last process: process.
		 last reportOtherProcesses: reportOtherProcesses.
		 ^sons copyWith: (last primitives: hereTally)].
	^sons! !

!MessageTally methodsFor: 'comparing' stamp: 'tk 7/5/2001 22:04'!
species
	^MessageTally! !

!MessageTally methodsFor: 'initialize-release' stamp: 'fbs 7/25/2013 07:14'!
spyAllEvery: millisecs on: aBlock
	"Create a spy and spy on the given block at the specified rate."
	"Spy all the system processes"

	| myDelay |
	aBlock isBlock
		ifFalse: [ self error: 'spy needs a block here' ].
	self class: aBlock receiver class method: aBlock method.
		"set up the probe"
	myDelay := Delay forMilliseconds: millisecs.
	time0 := Time millisecondClockValue.
	gcStats := Smalltalk getVMParameters.
	Timer ifNotNil: [ self error: 'it seems a tally is already running' ].
	Timer := [
		[true] whileTrue: [
			| observedProcess |
			startTime := Time millisecondClockValue.
			myDelay wait.
			observedProcess := Processor preemptedProcess.
			self
				tally: observedProcess suspendedContext
				in: observedProcess
				"tally can be > 1 if ran a long primitive"
				by: (Time millisecondClockValue - startTime) // millisecs].
		nil] newProcess.
	Timer priority: Processor timingPriority-1.
		"activate the probe and evaluate the block"
	Timer resume.
	^ aBlock ensure: [
		"cancel the probe and return the value"
		"Could have already been terminated. See #terminateTimerProcess"
		self class terminateTimerProcess.
		self computeGCStats.
		time := Time millisecondClockValue - time0]! !

!MessageTally methodsFor: 'initialize-release' stamp: 'fbs 7/25/2013 07:15'!
spyEvery: millisecs onProcess: aProcess forMilliseconds: msecDuration 
	"Create a spy and spy on the given process at the specified rate."
	| myDelay observedProcess sem |
	(aProcess isKindOf: Process)
		ifFalse: [self error: 'spy needs a Process here'].
	self class: aProcess suspendedContext receiver class method: aProcess suspendedContext method.
	"set up the probe"
	observedProcess := aProcess.
	myDelay := Delay forMilliseconds: millisecs.
	time0 := Time millisecondClockValue.
	endTime := time0 + msecDuration.
	sem := Semaphore new.
	gcStats := Smalltalk getVMParameters.
	Timer ifNotNil: [ self error: 'it seems a tally is already running' ].
	Timer := [
			[
				startTime := Time millisecondClockValue.
				myDelay wait.
				self
					tally: Processor preemptedProcess suspendedContext
					in: (observedProcess == Processor preemptedProcess 
						ifTrue: [ observedProcess ]
						ifFalse: [ nil ])
					"tally can be > 1 if ran a long primitive"
					by: (Time millisecondClockValue - startTime) // millisecs.
				startTime < endTime
			] whileTrue.
			sem signal.
		] newProcess.
	Timer priority: Processor timingPriority-1.
		"activate the probe and evaluate the block"
	Timer resume.
	"activate the probe and wait for it to finish"
	sem wait.
	self computeGCStats.
	time := Time millisecondClockValue - time0! !

!MessageTally methodsFor: 'initialize-release' stamp: 'fbs 7/25/2013 07:14'!
spyEvery: millisecs on: aBlock
	"Create a spy and spy on the given block at the specified rate."
	"Spy only on the active process (in which aBlock is run)"

	| myDelay observedProcess |
	aBlock isBlock
		ifFalse: [ self error: 'spy needs a block here' ].
	self class: aBlock receiver class method: aBlock method.
		"set up the probe"
	observedProcess := Processor activeProcess.
	myDelay := Delay forMilliseconds: millisecs.
	gcStats := Smalltalk getVMParameters.
	Timer ifNotNil: [ self error: 'it seems a tally is already running' ].
	time0 := Time millisecondClockValue.
	Timer := [
		[ true ] whileTrue: [
			startTime := Time millisecondClockValue.
			myDelay wait.
			self
				tally: Processor preemptedProcess suspendedContext
				in: (observedProcess == Processor preemptedProcess ifTrue: [observedProcess] ifFalse: [nil])
				"tally can be > 1 if ran a long primitive"
				by: (Time millisecondClockValue - startTime) // millisecs].
		nil] newProcess.
	Timer priority: Processor timingPriority-1.
		"activate the probe and evaluate the block"
	Timer resume.
	^ aBlock ensure: [
		"cancel the probe and return the value"
		"Could have already been terminated. See #terminateTimerProcess"
		self class terminateTimerProcess.
		self computeGCStats.
		time := Time millisecondClockValue - time0]! !

!MessageTally methodsFor: 'reporting' stamp: 'stp 05/08/1999 12:06'!
tally
	"Answer the receiver's number of tally."

	^tally! !

!MessageTally methodsFor: 'tallying' stamp: 'eem 1/28/2015 16:40'!
tallyPath: context by: count
	| aMethod nesting path |
	aMethod := context method.
	nesting := self blockNestingCountOf: context.

	"Find the correct child (if there)"
	path := receivers
				detect: [:oldTally | 
					oldTally method == aMethod
					and: [oldTally blockNesting = nesting]]
				ifNone: [].
	
	"Add new child if needed"
	path ifNil:
		[path := MessageTally new class: context receiver class method: aMethod nesting: nesting.
		path reportOtherProcesses: reportOtherProcesses.
		receivers := receivers copyWith: path].
	
	^ path bumpBy: count! !

!MessageTally methodsFor: 'tallying' stamp: 'eem 1/28/2015 16:39'!
tallyPath: context in: aProcess by: count
	| aMethod nesting path |
	aMethod := context method.
	nesting := self blockNestingCountOf: context.

	"Find the correct child (if there)"
	path := receivers
				detect: [:oldTally | 
					oldTally method == aMethod
					and: [oldTally process == aProcess
					and: [oldTally blockNesting = nesting]]]
				ifNone: [].

	"Add new child if needed"
	path ifNil:
		[path := MessageTally new
					class: context receiver class method: aMethod nesting: nesting;
					process: aProcess;
					reportOtherProcesses: reportOtherProcesses;
					maxClassNameSize: maxClassNameSize;
					maxClassPlusSelectorSize: maxClassPlusSelectorSize;
					maxTabs: maxTabs.
		receivers := receivers copyWith: path].

	^path bumpBy: count! !

!MessageTally methodsFor: 'tallying' stamp: 'eem 2/1/2015 17:49'!
tally: context by: count
	"Explicitly tally the specified context and its stack."
	"Add to this node if appropriate"
	(context method == method
	 and: [blockNesting = (self blockNestingCountOf: context)]) ifTrue:
		[^self bumpBy: count].

	^context sender
		ifNil: "No sender? Add new branch to the tree."
			[(self bumpBy: count) tallyPath: context by: count]
		ifNotNil: "Find the node for the sending context (or add it if necessary)"
			[:sender|
			 (self tally: sender by: count) tallyPath: context by: count]! !

!MessageTally methodsFor: 'tallying' stamp: 'eem 1/28/2015 16:38'!
tally: context in: aProcess by: count
	"Explicitly tally the specified context and its stack."
	| sender nesting |
	nesting := self blockNestingCountOf: context.

	"Add to this node if appropriate"
	(context method == method
	 and: [blockNesting = nesting]) ifTrue:
		[^self bumpBy: count].
	
	"No sender? Add new branch to the tree."
	(sender := context sender) ifNil:
		[^(self bumpBy: count) tallyPath: context in: aProcess by: count].
	
	"Find the node for the sending context (or add it if necessary)"
	^ (self tally: sender in: aProcess by: count) tallyPath: context in: aProcess by: count! !

!MessageTally methodsFor: 'reporting' stamp: 'stp 05/08/1999 11:47'!
time
	"Answer the receiver's run time."

	^time! !

!MessageTally methodsFor: 'printing' stamp: 'ul 2/5/2010 21:48'!
treePrintOn: aStream tabs: tabs thisTab: myTab total: total totalTime: totalTime tallyExact: isExact orThreshold: threshold 
	| sons |
	tabs do: [:tab | aStream nextPutAll: tab].
	tabs size > 0 
		ifTrue: 
			[self 
				printOn: aStream
				total: total
				totalTime: totalTime
				tallyExact: isExact].
	sons := isExact ifTrue: [receivers] ifFalse: [self sonsOver: threshold].
	sons isEmpty 
		ifFalse: 
			[tabs addLast: myTab.
			sons sort.
			1 to: sons size do: [ :i | 
				| sonTab |
				sonTab := i < sons size ifTrue: ['  |'] ifFalse: ['  '].
				(sons at: i) 
					treePrintOn: aStream
					tabs: (tabs size < self maxTabs 
							ifTrue: [tabs]
							ifFalse: [(tabs select: [:x | x = '[']) copyWith: '['])
					thisTab: sonTab
					total: total
					totalTime: totalTime
					tallyExact: isExact
					orThreshold: threshold].
			tabs removeLast]! !

!MessageTally methodsFor: 'private' stamp: 'eem 7/13/2012 16:59'!
unhibernate: resuming
	"Sent after snapshot.  If resuming (starting a shapshot), then convert relative
	 times back into absolute times.  If not, revert to pre-hibernation state."
	resuming
		ifTrue: [self convertToOrFromDeltaForHibernation]
		ifFalse: [self copyFrom: preHibernationCopy]! !

!MessageTally methodsFor: 'comparing' stamp: ''!
< aMessageTally 
	"Refer to the comment in Magnitude|<."

	^tally > aMessageTally tally! !

!MessageTally methodsFor: 'comparing' stamp: 'ar 3/3/2009 19:36'!
= aMessageTally

	self species == aMessageTally species ifFalse: [^ false].
	^ aMessageTally method == method and:[aMessageTally process == process]! !

!MessageTally methodsFor: 'comparing' stamp: ''!
> aMessageTally 
	"Refer to the comment in Magnitude|>."

	^tally < aMessageTally tally! !

!String methodsFor: '*System-Localization' stamp: 'bf 9/27/2007 12:20'!
literalStringsDo: aBlock 
	"Assuming the receiver receiver is a literal, evaluate aBlock with all Strings (but not Symbols) within it."
	aBlock value: self! !

!String methodsFor: '*system' stamp: 'ar 2/13/2010 15:50'!
suggestedTypeNames
	^ Array streamContents:
		[ : stream |
		self withIndexDo:
		[ : eachChar : index |
		eachChar isUppercase ifTrue:
			[ stream nextPut: 
				(self
					copyFrom: index
					to: self size) withBlanksTrimmed.
			"Often, argument names that refer to Collections end in the letter s, which can cause the suggested type-name to not be found.  Account for this."
			self last = $s
				ifTrue:
					[ stream nextPut:
						(self
							copyFrom: index
							to: self size-1) withBlanksTrimmed ] ] ] ]! !

!String methodsFor: '*System-Localization' stamp: 'Richo 5/28/2010 10:39'!
translated
	"answer the receiver translated to the default language"
	| translation |
	translation := self
		translatedTo: LocaleID current
		inDomain: (TextDomainManager domainOfMethod: thisContext sender method).
	self == translation ifTrue: [^self translatedInAllDomains].
	^translation! !

!String methodsFor: '*System-Localization' stamp: 'dgd 8/27/2004 18:43'!
translatedIfCorresponds
	"answer the receiver translated to the default language only if 
	the receiver begins and ends with an underscore (_)"
	^ ('_*_' match: self)
		ifTrue: [(self copyFrom: 2 to: self size - 1) translated]
		ifFalse: [self]! !

!String methodsFor: '*System-Localization' stamp: 'Richo 8/16/2010 22:25'!
translatedInAllDomains
	| translation |
	"Transcript show: self printString, ' translatedInAllDomains'; cr."
	TextDomainManager allKnownDomains do: [:domain |
		translation := self translatedTo: LocaleID current inDomain: domain.
		self = translation ifFalse: [^translation]
	].
	^self! !

!String methodsFor: '*System-Localization' stamp: 'Richo 5/17/2010 23:06'!
translatedInAnyDomain
	| translation |
	Transcript show: self printString, ' translatedInAnyDomain'; cr.
	TextDomainManager allKnownDomains do: [:domain |
		translation := self translatedInDomain: domain.
		self = translation ifFalse: [^translation]].
	^self! !

!String methodsFor: '*System-Localization' stamp: 'Richo 8/16/2010 22:26'!
translatedInDomain: aDomainName
	| translation |
	translation := self translatedTo: LocaleID current inDomain: aDomainName.
	self == translation ifTrue: [^self translatedInAllDomains].
	^translation
! !

!String methodsFor: '*System-Localization' stamp: 'Richo 8/16/2010 22:27'!
translatedInDomain: aDomainName or: anotherDomainName
	| translation |
	translation := self translatedTo: LocaleID current inDomain: aDomainName.
	self == translation ifTrue: [^self translatedInDomain: anotherDomainName].
	^translation
! !

!String methodsFor: '*System-Localization' stamp: 'tak 7/25/2007 00:01'!
translatedNoop
	"This is correspondence gettext_noop() in gettext."
	^ self! !

!String methodsFor: '*System-Localization' stamp: 'Richo 5/12/2010 16:43'!
translatedTo: localeID 
	"answer the receiver translated to the given locale id"
	^ self translatedTo: localeID inDomain: (TextDomainManager domainOfMethod: thisContext sender method).! !

!String methodsFor: '*System-Localization' stamp: 'tk 1/6/2008 12:21'!
translatedTo: localeID inDomain: aDomainName
	"answer the receiver translated to the given locale id in the textdomain"

	^ NaturalLanguageTranslator translate: self 
								toLocaleID: localeID 
								inDomain:  aDomainName! !

!FileDirectory class methodsFor: '*System-Files' stamp: 'tpr 10/9/2003 16:27'!
openChanges: changesName forImage: imageName
"find the changes file by looking in
a) the directory derived from the image name
b) the DefaultDirectory (which will normally be the directory derived from the image name or the SecurityManager's choice)
If an old file is not found in either place, check for a read-only file in the same places. If that fails, return nil"
	| changes fd |
	"look for the changes file or an alias to it in the image directory"
	fd := FileDirectory on: (FileDirectory dirPathFor: imageName).
	(fd fileExists: changesName)
		ifTrue: [changes := fd oldFileNamed: changesName].
	changes ifNotNil:[^changes].

	"look for the changes in the default directory"
	fd := DefaultDirectory.
	(fd fileExists: changesName)
		ifTrue: [changes := fd oldFileNamed: changesName].
	changes ifNotNil:[^changes].

	"look for read-only changes in the image directory"
	fd := FileDirectory on: (FileDirectory dirPathFor: imageName).
	(fd fileExists: changesName)
		ifTrue: [changes := fd readOnlyFileNamed: changesName].
	changes ifNotNil:[^changes].

	"look for read-only changes in the default directory"
	fd := DefaultDirectory.
	(fd fileExists: changesName)
		ifTrue: [changes := fd readOnlyFileNamed: changesName].
	"this may be nil if the last try above failed to open a file"
	^changes
! !

!FileDirectory class methodsFor: '*System-Files' stamp: 'ul 8/14/2015 18:02'!
openSources: sourcesName andChanges: changesName forImage: imageName 
	"Open the changes and sources files and install them in SourceFiles. Inform the user of problems regarding write permissions or CR/CRLF mixups."
	"Note: SourcesName and imageName are full paths; changesName is a  
	local name."
	| sources changes msg wmsg |
	msg := 'Squeak cannot locate &fileRef.

Please check that the file is named properly and is in the
same directory as this image.'.
	wmsg := 'Squeak cannot write to &fileRef.

Please check that you have write permission for this file.

You won''t be able to save this image correctly until you fix this.'.

	sources := self openSources: sourcesName forImage: imageName.
	changes := self openChanges: changesName forImage: imageName.

	((sources == nil or: [sources atEnd])
			and: [Preferences valueOfFlag: #warnIfNoSourcesFile])
		ifTrue: [Smalltalk platformName = 'Mac OS'
				ifTrue: [msg := msg , '
Make sure the sources file is not an Alias.'].
self inform: (msg copyReplaceAll: '&fileRef' with: 'the sources file named ' , sourcesName)].

	(changes == nil
			and: [Preferences valueOfFlag: #warnIfNoChangesFile])
		ifTrue: [self inform: (msg copyReplaceAll: '&fileRef' with: 'the changes file named ' , changesName)].

	((Preferences valueOfFlag: #warnIfNoChangesFile) and: [changes notNil])
		ifTrue: [changes isReadOnly
				ifTrue: [self inform: (wmsg copyReplaceAll: '&fileRef' with: 'the changes file named ' , changesName)].

			((changes next: 200)
					includesSubstring: String crlf)
				ifTrue: [self inform: 'The changes file named ' , changesName , '
has been injured by an unpacking utility.  Crs were changed to CrLfs.
Please set the preferences in your decompressing program to 
"do not convert text files" and unpack the system again.']].

	SourceFiles := Array with: sources with: changes! !

!FileDirectory class methodsFor: '*System-Files' stamp: 'fbs 7/25/2013 07:09'!
openSources: fullSourcesName forImage: imageName 
"We first do a check to see if a compressed version ofthe sources file is present.
Open the .sources file read-only after searching in:
a) the directory where the VM lives
b) the directory where the image came from
c) the DefaultDirectory (which is likely the same as b unless the SecurityManager has changed it).
"

	| sources fd sourcesName |
	(fullSourcesName endsWith: 'sources') ifTrue:
		["Look first for a sources file in compressed format."
		sources := self openSources: (fullSourcesName allButLast: 7) , 'stc'
						forImage: imageName.
		sources ifNotNil: [^ CompressedSourceStream on: sources]].

	sourcesName := FileDirectory localNameFor: fullSourcesName.
	"look for the sources file or an alias to it in the VM's directory"
	fd := FileDirectory on: Smalltalk vmPath.
	(fd fileExists: sourcesName)
		ifTrue: [sources := fd readOnlyFileNamed: sourcesName].
	sources ifNotNil: [^ sources].
	"look for the sources file or an alias to it in the image directory"
	fd := FileDirectory on: (FileDirectory dirPathFor: imageName).
	(fd fileExists: sourcesName)
		ifTrue: [sources := fd readOnlyFileNamed: sourcesName].
	sources ifNotNil: [^ sources].
	"look for the sources in the current directory"
	fd := DefaultDirectory.
	(fd fileExists: sourcesName)
		ifTrue: [sources := fd readOnlyFileNamed: sourcesName].
	"sources may still be nil here"
	^sources
! !

!FileDirectory class methodsFor: '*System-Files' stamp: 'tak 12/17/2004 14:03'!
setDefaultDirectory: directoryName
	"Initialize the default directory to the directory supplied. This method is called when the image starts up."
	| dirName |
	DirectoryClass := self activeDirectoryClass.
	dirName := (FilePath pathName: directoryName) asSqueakPathName.
	[dirName endsWith: self slash] whileTrue:[
		dirName := dirName copyFrom: 1 to: dirName size - self slash size.
	].
	DefaultDirectory := self on: dirName.! !

BasicClassOrganizer subclass: #PseudoClassOrganizer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!PseudoClassOrganizer methodsFor: 'comment accessing' stamp: 'NS 4/6/2004 16:44'!
classComment
	"Answer the comment associated with the object that refers to the receiver."
	classComment == nil ifTrue: [^''].
	^classComment! !

!PseudoClassOrganizer methodsFor: 'comment accessing' stamp: 'NS 4/6/2004 16:44'!
classComment: aChangeRecord
	classComment := aChangeRecord! !

!PseudoClassOrganizer methodsFor: 'accessing' stamp: 'NS 4/6/2004 12:27'!
setDefaultList: aCollection
	super setDefaultList: aCollection.
	self classComment: nil.! !

!TextStyle class methodsFor: '*System-Fonts' stamp: 'rbb 3/1/2005 11:18'!
chooseTTCFontSize: args
	"Prompt for a point size and, if one is given, add a new font size to the font named by the first member of args. If args' length is three, send a message with the selector equal to the third of args, and the receiver equal to the second of args, passing the selected style as an argument."

	| f n style |
	f := UIManager default request: 'New Point Size' initialAnswer: '0'.
	n := f asNumber.
	style := (TextConstants at: args first) addNewFontSize: n.
	style ifNotNil: [
		args second ifNotNil: [args second perform: args third with: style].
	].
! !

!TextStyle class methodsFor: '*System-Changes-mime file in/out' stamp: 'tpr 6/10/2005 18:29'!
writeSF2FamilyNamed: familyName inDirectory: directoryName toChangeSet: csName 
	"
	TextStyle writeSF2FamilyNamed: 'Accuny' inDirectory: 'AccunyCorrectedFeb252004Beta Folder' toChangeSet: 'AccunyInstall'.
	"

	|  family |
	family := OrderedCollection new.
	family addAll: (StrikeFont readStrikeFont2Family: familyName fromDirectory: (FileDirectory default fullNameFor: directoryName)) .
	family do: [:f | f reset].
	self 
		writeStyle: (TextStyle fontArray: family asArray)
		named: familyName
		toChangeSet: csName! !

!TextStyle class methodsFor: '*System-Changes-mime file in/out' stamp: 'nice 7/17/2014 23:41:33.7'!
writeStyle: aTextStyle named: familyName toChangeSet: csName
	"Write the text style to a change set, with a postscript that will re-load it.
	NOTE: to do TTCFonts, you have to have a working ShortPointArray endianness conversion."
	"
	TTCFont recreateCache.
	TextStyle writeStyle: (TextStyle named: #Arial) named: 'Arial' toChangeSet: 'ArialInstall'.
	
	TextStyle writeStyle: (TextStyle named: #Accuny) named: 'Accuny' toChangeSet: 'AccunyInstall2'.
	"

	| cs mime |

	cs := ChangeSet basicNewNamed: csName.
	cs adoptSelector: #collectionFromCompressedMIMEString: forClass: self class.
	cs adoptSelector: #replaceStyle:with: forClass: self class.
	cs adoptSelector: #replaceFontsIn:with: forClass: self class.
	cs adoptSelector: #looseFontsFromFamily: forClass: self class.
	((aTextStyle fontArray copyWithout: nil) collect: [ :f | f class ]) asSet do: [ :cls  | 
		cs adoptSelector: #derivativeFonts forClass: cls.
		cs adoptSelector: #releaseCachedState forClass: cls ].

	cs preambleString: (String streamContents: [ :s |
		s nextPutAll: '"Change Set:		'; nextPutAll: csName; cr;
		nextPutAll: 'Date:		'; print: Date today; cr;
		nextPutAll: 'Author:		'; nextPutAll: Utilities authorName; cr; cr;
		nextPutAll: 'Installs the text style '''; nextPutAll: familyName; nextPutAll: ''''; cr;
		nextPutAll: 'from a compressed MIME encoding in the postscript."'; cr. ]).

	mime := aTextStyle compressedMIMEEncoded.

	cs postscriptString: (String streamContents: [ :s | s
		nextPutAll: '"Postscript:'; cr;
		nextPutAll: 'Install the text style from the compressed MIME encoding, and replace the old one.';
		nextPut: $"; cr;
		nextPutAll: 'TextConstants at: #';
		nextPutAll: familyName;
		nextPutAll: ' ifPresent: [ :oldStyle | TextConstants at: #';
		nextPutAll: familyName;
		nextPutAll: 'OLD put: oldStyle. TextConstants removeKey: #';
		nextPutAll: familyName;
		nextPutAll: ' ].';
		cr;
		nextPutAll: 'TextStyle collectionFromCompressedMIMEString: ';
		cr;
		print: mime;
		nextPut: $.; cr; cr;
		nextPutAll: 'TextConstants at: #';
		nextPutAll: familyName;
		nextPutAll: 'OLD ifPresent: [ :oldStyle | TextStyle replaceStyle: oldStyle with: (TextStyle named: ''';
		nextPutAll: familyName;
		nextPutAll: ''') ].';
		cr;
		nextPutAll: 'TextStyle replaceFontsIn: (TextStyle looseFontsFromFamily: ''';
		nextPutAll: familyName;
		nextPutAll: ''') with: (TextStyle named: ''';
		nextPutAll: familyName;
		nextPutAll: ''').';		
		cr ]).

	cs fileOut.

! !

!TextStyle methodsFor: '*System-Fonts' stamp: 'rbb 3/1/2005 11:18'!
addNewFontSizeDialog: args
	"This is called from a modal menu and call back the menu with entered argument."
	| f n r |
	f := UIManager default request: 'Enter the point size' initialAnswer: '12'.
	n := f asNumber.
	r := self addNewFontSize: n.
	r ifNotNil: [
		args second ifNotNil: [args second modalSelection: {args first. n}].
	].
! !

!TextStyle methodsFor: '*System-Fonts' stamp: 'tfel 8/5/2016 16:04'!
addNewFontSize: pointSize
	"Add a font in specified size to the array of fonts."
	| f d newArray t isSet |
	fontArray first emphasis ~= 0 ifTrue: [
		t := TextConstants at: self fontArray first familyName asSymbol.
		t fonts first emphasis = 0 ifTrue: [
			^ t addNewFontSize: pointSize.
		].
	].

	pointSize <= 0 ifTrue: [^ nil].
	fontArray do: [:s |
		s pointSize = pointSize ifTrue: [^ s].
	].

	(isSet := fontArray first isKindOf: TTCFontSet) 
	ifTrue:[
		| fonts |
		fonts := fontArray first fontArray collect: [ :font |
			| newFont |
			(font isNil)
			ifTrue: [newFont := nil]
			ifFalse: [
				newFont := (font ttcDescription size > 256)
					ifTrue: [MultiTTCFont new initialize]
					ifFalse: [TTCFont new initialize].
				newFont ttcDescription: font ttcDescription.
				newFont pixelSize: pointSize * 96 // 72.
				font derivativeFonts notEmpty ifTrue: [font derivativeFonts do: [ :proto |
					proto ifNotNil: [
						d := proto class new initialize.
						d ttcDescription: proto ttcDescription.
						d pixelSize: newFont pixelSize.
						newFont derivativeFont: d]]].
				].
			newFont].
		f := TTCFontSet newFontArray: fonts]
	ifFalse: [
		f := fontArray first class new initialize: fontArray first.
		f pointSize: pointSize.
		fontArray first derivativeFonts do: [:proto |
			proto ifNotNil: [
				d := TTCFont new initialize: proto.
				d pointSize: f pointSize.
				f derivativeFont: d.
			].
		].
	].
	newArray := (fontArray copyWith: f) asArray sort: [:a :b | a pointSize <= b pointSize].
	self newFontArray: newArray.
	isSet ifTrue: [
		TTCFontSet register: newArray at: newArray first familyName asSymbol.
	].
	^ self fontOfPointSize: pointSize
! !

!TextStyle methodsFor: '*System-Fonts' stamp: 'yo 6/23/2003 19:58'!
discardOtherSizes
	"This method trys to discard the fonts in non-standard size.  If the size is still in use, there will be a problem."
	| newArray |
	self isTTCStyle ifFalse: [^ self].
	newArray := fontArray select: [:s | TTCFont pointSizes includes: s pointSize].
	self newFontArray: newArray.

"(TextConstants at: #ComicSansMS) discardOtherSizes"! !

!FileDoesNotExistException methodsFor: '*System-Files-error handling' stamp: 'mtf 1/20/2011 14:16'!
defaultAction
	"The default action taken if the exception is signaled."

	^self readOnly
		ifTrue: [StandardFileStream readOnlyFileDoesNotExistUserHandling: self fileName
			ifDebug: [super defaultAction]]
		ifFalse: [StandardFileStream fileDoesNotExistUserHandling: self fileName
			ifDebug: [super defaultAction]]
! !

Collection subclass: #WeakRegistry
	instanceVariableNames: 'list valueDictionary sema executors'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'System-Finalization'!

!WeakRegistry commentStamp: 'ul 9/26/2010 02:51' prior: 0!
I am a registry for objects needing finalization. When an object is added the object as well as its executor is stored. When the object is garbage collected, the executor can take the appropriate action for any resources associated with the object.

This kind of WeakRegistry is using a new VM feature, which allows a more robust finalization support. In contrast to the old implementation, it doesn't spend linear time checking which elements became garbage.

See also:
	Object executor
	Object actAsExecutor
	Object finalize!

!WeakRegistry class methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 21:22'!
default
	^Default ifNil:[Default := self new]! !

!WeakRegistry class methodsFor: 'instance creation' stamp: 'Igor.Stasenko 3/8/2010 21:47'!
new
	| registry |
	registry := super new.
	WeakArray addWeakDependent: registry.
	^registry
! !

!WeakRegistry class methodsFor: 'instance creation' stamp: 'Igor.Stasenko 5/25/2010 05:31'!
new: n
	^ self new! !

!WeakRegistry methodsFor: 'adding' stamp: 'Igor.Stasenko 3/8/2010 21:03'!
add: anObject
	"Add anObject to the receiver. Store the object as well as the associated executor."
	
	^self add: anObject executor: anObject executor! !

!WeakRegistry methodsFor: 'adding' stamp: 'Igor.Stasenko 5/25/2010 05:30'!
add: anObject executor: anExecutor

	self protected: [ | finItem |
		finItem := valueDictionary at: anObject ifAbsentPut: [
			WeakFinalizerItem new list: list object: anObject ].
		finItem add: anExecutor ].
	^ anObject
! !

!WeakRegistry methodsFor: 'enumerating' stamp: 'Igor.Stasenko 5/25/2010 05:11'!
do: aBlock
	^self protected: [
		valueDictionary keysDo: aBlock.
	].
! !

!WeakRegistry methodsFor: '*System-Finalization' stamp: 'eem 8/25/2015 10:36'!
finalizeValues
	"Finalize any values, which happen to stocked in our list, due to some weak references become garbage"

	| finalizer |
	self protected:
		[valueDictionary finalizeValues.
		 finalizer := executors.
		 executors := nil ].
	finalizer ifNotNil: [finalizer do: [ :each | each finalizeValues]]! !

!WeakRegistry methodsFor: 'initialize-release' stamp: 'Igor.Stasenko 5/28/2010 05:12'!
initialize
	valueDictionary := WeakIdentityKeyDictionary new.
	list := WeakFinalizationList new.
	sema := Semaphore forMutualExclusion.
	self installFinalizer.! !

!WeakRegistry methodsFor: 'initialize-release' stamp: 'eem 8/25/2015 10:34'!
installFinalizer

	valueDictionary finalizer:
		[:executor|
		 (executors ifNil: [executors := OrderedCollection new]) add: executor]! !

!WeakRegistry methodsFor: 'accessing' stamp: 'Igor.Stasenko 10/11/2010 15:11'!
keys

	^self protected: [ valueDictionary keys ]
! !

!WeakRegistry methodsFor: 'copying' stamp: 'Igor.Stasenko 5/28/2010 05:16'!
postCopy
	"should we prohibit any attempts to copy receiver?"
	self protected: [ | oldDict |
		sema := Semaphore forMutualExclusion.
		oldDict := valueDictionary.
		list := WeakFinalizationList new.
		valueDictionary := WeakIdentityKeyDictionary new.
		self installFinalizer.
	
		oldDict keysAndValuesDo: [:key :value |
			valueDictionary at: key put: (value copyWithList: list)
		].
	]! !

!WeakRegistry methodsFor: 'printing' stamp: 'Igor.Stasenko 5/25/2010 05:13'!
printElementsOn: aStream
	sema ifNil: [^super printElementsOn: aStream].
	aStream nextPutAll: '(<this WeakRegistry is locked>)'! !

!WeakRegistry methodsFor: 'private' stamp: 'Igor.Stasenko 5/25/2010 05:34'!
protected: aBlock
	"Execute aBlock protected by the accessLock"

	^ sema
		critical: aBlock
		ifError: [ :msg :rcvr |
		rcvr error: msg ] ! !

!WeakRegistry methodsFor: 'printing' stamp: 'Igor.Stasenko 5/25/2010 05:16'!
removeAll
	"See super"
	
	self protected:[
		valueDictionary removeAll.
	].! !

!WeakRegistry methodsFor: 'printing' stamp: 'Igor.Stasenko 5/25/2010 05:21'!
remove: oldObject ifAbsent: exceptionBlock
	"Remove oldObject as one of the receiver's elements."
	
	oldObject ifNil: [ ^nil ].
	^(self protected: [ valueDictionary removeKey: oldObject ifAbsent: nil ])
		ifNil: [ exceptionBlock value ]! !

!WeakRegistry methodsFor: 'accessing' stamp: 'Igor.Stasenko 5/25/2010 05:48'!
size
	^ self protected: [valueDictionary slowSize]! !

!WeakRegistry methodsFor: 'private' stamp: 'ar 3/20/98 19:31'!
species
	^Set! !

!ReadWriteStream methodsFor: '*System-Object Storage-fileIn/Out' stamp: 'di 7/17/97 16:12'!
fileInObjectAndCode
	"This file may contain:
1) a fileIn of code  
2) just an object in SmartReferenceStream format 
3) both code and an object.
	File it in and return the object.  Note that self must be a FileStream or RWBinaryOrTextStream.  Maybe ReadWriteStream incorporate RWBinaryOrTextStream?"
	| refStream object |
	self text.
	self peek asciiValue = 4
		ifTrue: [  "pure object file"
			refStream := SmartRefStream on: self.
			object := refStream nextAndClose]
		ifFalse: [  "objects mixed with a fileIn"
			self fileIn.  "reads code and objects, then closes the file"
			object := SmartRefStream scannedObject].	"set by side effect of one of the chunks"
	SmartRefStream scannedObject: nil.  "clear scannedObject"
	^ object! !

!ReadWriteStream methodsFor: '*System-Object Storage-fileIn/Out' stamp: ''!
fileNameEndsWith: aString
	"See comment in FileStream fileNameEndsWith:"

	^false! !

ReadWriteStream subclass: #Transcripter
	instanceVariableNames: 'frame para'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Recovery'!

!Transcripter commentStamp: '<historical>' prior: 0!
Transcripter is a dog-simple scrolling stream with display.  It is intended to operate with no support from MVC or color in a minimal, or headless version of Squeak.  No attention has been paid to appearance or performance.!

!Transcripter class methodsFor: 'utilities' stamp: 'di 11/3/2000 18:47'!
emergencyEvaluator
	(Transcripter newInFrame: (0@0 corner: 320@200))
		show: 'Type ''revert'' to revert your last method change.
Type ''exit'' to exit the emergency evaluator.';
		readEvalPrint! !

!Transcripter class methodsFor: 'instance creation' stamp: 'di 8/14/97 12:09'!
newInFrame: frame
"
(Transcripter newInFrame: (0@0 extent: 100@200))
	nextPutAll: 'Hello there'; endEntry;
	cr; print: 355.0/113; endEntry;
	readEvalPrint.
"
	| transcript |
	transcript := self on: (String new: 100).
	transcript initInFrame: frame.
	^ transcript clear! !

!Transcripter class methodsFor: 'instance creation' stamp: 'ar 11/16/1999 20:16'!
startTranscriptProcess   "Transcripter startTranscriptProcess"
	| activeProcess |
	Transcript := self newInFrame: Display boundingBox.
	activeProcess := [Transcript readEvalPrint.
					Smalltalk processShutDownList: true; quitPrimitive]
						newProcess
					priority: Processor userSchedulingPriority.
	activeProcess resume.
	Processor terminateActive
! !

!Transcripter methodsFor: 'private' stamp: 'di 8/14/97 12:12'!
black
	Display depth = 1 ifTrue: [^ Bitmap with: 16rFFFFFFFF "Works without color support"].
	^ Color black! !

!Transcripter methodsFor: 'accessing' stamp: 'di 8/14/97 12:41'!
clear
	Display fill: (frame insetBy: -2) fillColor: self black;
			fill: frame fillColor: self white.
	self on: (String new: 100); endEntry! !

!Transcripter methodsFor: 'command line' stamp: 'ul 6/17/2011 12:48'!
confirm: queryString 

	| choice |
	[choice := self request: queryString , '
Please type yes or no followed by return'.
	choice first asUppercase = $Y ifTrue: [^ true].
		choice first asUppercase = $N ifTrue: [^ false]] repeat! !

!Transcripter methodsFor: 'accessing' stamp: 'dtl 12/14/2013 16:17'!
endEntry
	| c d cb |
	c := self contents.
	Display extent ~= DisplayScreen actualScreenSize ifTrue:
		["Handle case of user resizing physical window"
		DisplayScreen startUp.
		frame := frame intersect: Display boundingBox.
		^ self clear; show: c].
	Project current displayTranscripter: self.
	d := para compositionRectangle bottom - frame bottom.
	d > 0 ifTrue:
		["Scroll up to keep all contents visible"
		cb := para characterBlockAtPoint: para compositionRectangle topLeft
											+ (0@(d+para lineGrid)).
		self on: (c copyFrom: cb stringIndex to: c size).
		readLimit:= position:= collection size.
		^ self endEntry].
! !

!Transcripter methodsFor: 'initialization' stamp: 'dtl 12/14/2013 13:32'!
initInFrame: rect
	frame := rect insetBy: 2.  "Leave room for border"
	Project current initializeParagraphForTranscripter: self
! !

!Transcripter methodsFor: 'command line' stamp: 'cmm 10/12/2013 16:42'!
readEvalPrint
	| line okToRevert |
	okToRevert := true.
	[#('quit' 'exit' 'done' ) includes: (line := self request: '>')]
		whileFalse:
		[line = 'revert'
		ifTrue: [okToRevert
			ifTrue: [RecentMessages default revertMostRecent.
					self cr; show: 'reverted: ' , RecentMessages default mostRecent.
					okToRevert := false]
			ifFalse: [self cr; show: 'Only one level of revert currently supported']]
		ifFalse: [self cr; show: ([Compiler evaluate: line] ifError: [:err :ex | err])]]! !

!Transcripter methodsFor: 'command line' stamp: 'sma 2/26/2000 19:39'!
request: prompt
	| startPos char contents | 
	self cr; show: prompt.
	startPos := position.
	[[Sensor keyboardPressed] whileFalse.
	(char := Sensor keyboard) = Character cr]
		whileFalse:
		[char = Character backspace
			ifTrue: [readLimit := position := (position - 1 max: startPos)]
			ifFalse: [self nextPut: char].
		self endEntry].
	contents := self contents.
	^ contents copyFrom: startPos + 1 to: contents size! !

!Transcripter methodsFor: 'accessing' stamp: 'sma 2/26/2000 19:35'!
show: anObject
	self nextPutAll: anObject asString; endEntry! !

!Transcripter methodsFor: 'private' stamp: 'di 8/14/97 12:12'!
white
	Display depth = 1 ifTrue: [^ Bitmap with: 0 "Works without color support"].
	^ Color white! !

!FileStream class methodsFor: '*System-Files' stamp: 'yo 8/18/2004 20:24'!
fileIn: fullName
	"File in the entire contents of the file specified by the name provided"

	| ff |
	fullName ifNil: [^ Beeper beep].
	ff := self readOnlyFileNamed: (GZipReadStream uncompressedFileName: fullName).
	ff fileIn.
! !

!ClassDescription methodsFor: '*system-support' stamp: 'dvf 8/23/2003 12:45'!
allUnreferencedClassVariables
	"Answer a list of the names of all the receiver's unreferenced class vars, including those defined in superclasses"

	^ self systemNavigation allUnreferencedClassVariablesOf: self! !

!ClassDescription methodsFor: '*System' stamp: 'dvf 8/19/2005 11:31'!
methodReferencesInCategory: aCategoryName
	^(self organization listAtCategoryNamed: aCategoryName)
		collect: [:ea | MethodReference new
						setClassSymbol: self theNonMetaClass name
						classIsMeta: self isMeta
						methodSymbol: ea
						stringVersion: '']
! !

!ReplaceExistingFileException methodsFor: '*System-Files-error handling' stamp: 'fbs 7/12/2013 17:37'!
defaultAction
	| selection |
	selection := UIManager default
		chooseFrom: #('delete version in target directory' 'cancel' )
		title: fileName , ' already exists'.
	^ selection = 1.! !

!StrikeFont class methodsFor: '*System-Fonts' stamp: 'ar 1/5/2002 21:41'!
fromUser
	"StrikeFont fromUser"
	^self fromUser: TextStyle defaultFont! !

!StrikeFont class methodsFor: '*System-Fonts' stamp: 'nk 9/1/2004 14:29'!
fromUser: priorFont
	^self fromUser: priorFont allowKeyboard: true! !

!StrikeFont class methodsFor: '*System-Fonts' stamp: 'nice 12/27/2009 03:11'!
fromUser: priorFont allowKeyboard: aBoolean	"StrikeFont fromUser"
	"Present a menu of available fonts, and if one is chosen, return it.
	Otherwise return nil. Using ToolBuilder for added abstraction."
	| fontList fontMenu font builder resultBlock widget style result item |
	builder := ToolBuilder default.
	fontList := StrikeFont actualFamilyNames.
	fontMenu := builder pluggableMenuSpec new.
	resultBlock := [:value| result := value].
	fontList do: [:fontName | | active ptMenu |
		style := TextStyle named: fontName.
		active := priorFont familyName sameAs: fontName.
		ptMenu := builder pluggableMenuSpec new.
		style pointSizes do: [:pt | | label |
			label := pt printString, ' pt'.
			item := ptMenu add: label 
				target: resultBlock
				selector: #value:
				argumentList: {{fontName. pt}}.
			item checked: (active and:[pt = priorFont pointSize]).
		].
		style isTTCStyle ifTrue: [
			ptMenu add: 'new size'
				target: style 
				selector: #addNewFontSizeDialog: 
				argumentList: {{fontName. fontMenu}}.
		].
		item := fontMenu add: fontName action: nil.
		item subMenu: ptMenu.
		item checked: active.
	].
	TTFileDescription offerNonPortableFonts ifTrue:[
		fontMenu addSeparator.
		item := fontMenu add: 'More (non-portable) Fonts...' 
			target: resultBlock 
			selector: #value:
			argumentList: #(TTFileDescription).
	].
	widget := builder open: fontMenu.
	builder runModal: widget.
	result ifNil:[^nil].
	"Pick a non-portable font if requested"
	result == #TTFileDescription ifTrue:[
		^Smalltalk at: #TTFileDescription ifPresent:[:ttf| 
			ttf fontFromUser: priorFont allowKeyboard: aBoolean]].
	style := TextStyle named: result first.
	style ifNil: [^ self].
	font := style fonts detect: [:any | any pointSize = result last] ifNone: [nil].
	^ font! !

!StrikeFont class methodsFor: '*System-Fonts' stamp: 'tak 11/11/2004 21:14'!
setupDefaultFallbackFont
"
	StrikeFont setupDefaultFallbackFont
"

	(#(#Accuat #Accujen #Accula #Accumon #Accusf #Accushi #Accuve #Atlanta) collect: [:e | TextStyle named: e]) do: [:style |
		style fontArray do: [:e |
			e reset.
			e setupDefaultFallbackFont.
		].
	].
	TTCFont allSubInstances
		do: [:font | font reset.
			font setupDefaultFallbackFont]

! !

!Model methodsFor: '*System-preferences' stamp: 'mt 7/9/2016 08:51'!
defaultWindowColor

	^ self uniformWindowColor! !

!Model methodsFor: '*System-preferences' stamp: 'mt 7/2/2016 11:21'!
uniformWindowColor

	^ self userInterfaceTheme uniformWindowColor ifNil: [Color veryVeryLightGray]! !

!Model methodsFor: '*System-preferences' stamp: 'mt 7/10/2016 12:14'!
windowColorToUse

	^ Color colorFrom: (self class useColorfulWindows
		ifTrue: [self userInterfaceTheme customWindowColor ifNil: [self defaultWindowColor]]
		ifFalse: [self uniformWindowColor])! !

Model subclass: #Project
	instanceVariableNames: 'world uiManager changeSet transcript parentProject previousProject displayDepth viewSize thumbnail nextProject projectParameters version urlList environment lastDirectory lastSavedAtSeconds projectPreferenceFlagDictionary resourceManager'
	classVariableNames: 'AllProjects CurrentProject GoalFreePercent GoalNotMoreThan'
	poolDictionaries: ''
	category: 'System-Support'!

!Project commentStamp: 'cbr 7/27/2010 21:36' prior: 0!
A Project stores the state of a complete Squeak desktop, including
the windows, and the currently active changeSet.  A project knows who
its parent project is.  When you change projects, whether by entering
or exiting, the screen state of the project being exited is saved in
that project.

A project is retained by its view in the parent world.  It is
effectively named by the name of its changeSet, which can be changed
either by renaming in a changeSorter, or by editing the label of its
view from the parent project.

As the site of major context switch, Projects are the locus of
swapping between the old MVC and the new Morphic worlds.  The
distinction is based on whether the variable 'world' contains a
WorldMorph or a ControlManager.

Saving and Loading
Projects may be stored on the disk in external format.  (Project
named: 'xxx') exportSegment, or choose 'store project on file...'.
Projects may be loaded from a server and stored back.  Storing on a
server never overwrites;  it always makes a new version.  A project
remembers the url of where it lives in urlList.  The list is length
one, for now.  The url may point to a local disk instead of a server.
All projects that the user looks at are cached in the Squeaklet
folder.  Sorted by server.  The cache holds the most recent version
only.

When a project is loaded into Squeak, its objects are converted to
the current version.  There are three levels of conversion.  First,
each object is converted from raw bits to an object in its old
format.  Then it is sent some or all of these messages:
	comeFullyUpOnReload: smartRefStream  		Used to
re-discover an object that already exists in this image, such as a
resource, global variable, Character, or Symbol.  (sent to objects in
outPointers)
	convertToCurrentVersion: varDict refStream: smartRefStrm
		fill in fields that have been added to a class since
the object was stored.  Used to set the extra inst var to a default
value.  Or, return a new object of a different class.  (sent to
objects that changed instance variables)
	fixUponLoad: aProject refStream: smartRefStrm
	change the object due to conventions that have changed on the
project level.  (sent to all objects in the incoming project)

Here is the calling sequence for storing out a Project:
Project saveAs
Project storeOnServer
Project storeOnServerWithProgressInfo
Project storeOnServerInnards
Project exportSegmentFileName:directory:
Project exportSegmentWithChangeSet:fileName:directory:
ImageSegment writeForExportWithSources:inDirectory:changeSet:

!

!Project class methodsFor: 'utilities' stamp: 'RAA 11/16/2000 12:04'!
addingProject: newProject

	(self allProjects includes: newProject) ifTrue: [^self].
	AllProjects := self allProjects copyWith: newProject.! !

!Project class methodsFor: 'utilities' stamp: 'tk 10/26/1999 14:25'!
advanceToNextProject
	| nextProj |
	(nextProj := CurrentProject nextProject) ifNotNil:
		 [nextProj enter: true revert: false saveForRevert: false]
! !

!Project class methodsFor: 'utilities' stamp: 'di 6/10/1999 11:30'!
allNames
	^ (self allProjects collect: [:p | p name]) asSortedCollection: [:n1 :n2 | n1 asLowercase < n2 asLowercase]! !

!Project class methodsFor: 'utilities' stamp: 'di 6/10/1999 11:30'!
allNamesAndProjects
	^ (self allProjects asSortedCollection: [:p1 :p2 | p1 name asLowercase < p2 name asLowercase]) collect:
		[:aProject | Array with: aProject name with: aProject]! !

!Project class methodsFor: 'utilities' stamp: 'RAA 11/13/2000 17:14'!
allProjects

	^AllProjects ifNil: [
		Smalltalk garbageCollect.
		AllProjects := self allSubInstances select: [:p | p name notNil].
	].! !

!Project class methodsFor: 'utilities' stamp: 'jla 5/28/2001 21:48'!
allProjectsOrdered
	"Answer a list of all projects in hierarchical order, depth first"
	
	| allProjects  |
	allProjects := OrderedCollection new.
	Project topProject withChildrenDo:
		[:p | allProjects add: p].
	^ allProjects

"
Project allProjectsOrdered
"! !

!Project class methodsFor: 'utilities' stamp: 'dtl 11/27/2009 13:25'!
canWeLoadAProjectNow

	^ Smalltalk verifyMorphicAvailability
! !

!Project class methodsFor: 'utilities' stamp: 'sw 9/12/2001 23:05'!
chooseNaturalLanguage
	"Have the current project choose a new natural language"

	self current chooseNaturalLanguage! !

!Project class methodsFor: 'class initialization' stamp: 'mt 8/15/2016 19:24'!
cleanUp: agressive
	"Remove all projects but only when cleaning aggressively"

	agressive ifTrue:[
		self removeAllButCurrent.
		self current resourceManager reset].! !

!Project class methodsFor: 'constants' stamp: ''!
current
	"Answer the project that is currently being used."

	^CurrentProject! !

!Project class methodsFor: 'utilities' stamp: 'mt 5/1/2016 17:01'!
deletingProject: outgoingProject
	"Delete the outgoing project and all its sub-projects. Start with the sub-projects."
	
	"Delete all sub-projects."
	outgoingProject subProjects do: [:p | self deletingProject: p].

	"Forget that it ever existed. We have to do that now to avoid other code iterating over the partially removed project."
	self forget: outgoingProject.	

	"Give all registered projects the chance to release their references."
	self allProjects do: [:p | p deletingProject: outgoingProject].
	
	"Really clean-up everything related to the outgoing project."
	self releaseProjectReferences: outgoingProject.! !

!Project class methodsFor: 'squeaklet on server' stamp: 'dtl 11/27/2009 13:43'!
enterIfThereOrFind: aProjectName

	^ self current enterIfThereOrFind: aProjectName
! !

!Project class methodsFor: 'utilities' stamp: 'RAA 6/3/2000 09:52'!
enter: aString
	"Enter the project with the given name"
	^ ((self named: aString) ifNil: [^ CurrentProject]) enter! !

!Project class methodsFor: 'utilities' stamp: 'mt 5/1/2016 14:11'!
flattenProjectHierarchy
	"Clean-up project hierarchy. Make all projects be sub-projects of the top project."
	
	self allSubInstancesDo: [:ea | ea setParent: self topProject].! !

!Project class methodsFor: 'utilities' stamp: 'mt 5/1/2016 11:52'!
forget: aProject

	AllProjects := self allProjects copyWithout: aProject.! !

!Project class methodsFor: 'squeaklet on server' stamp: 'dtl 11/27/2009 14:11'!
fromUrl: urlString
	"Load the project, and make a thumbnail to it in the current project.  Replace the old one if necessary.
To test it, make sure you don't have AAEmptyTest.001.pr in your Squeaklets directory. Then do:
Project fromUrl: 'http://209.143.91.36/super/SuperSwikiProj/AAEmptyTest.001.pr'.
The project should open and you should enter it. Return to the previous project and delete the project. Then do:
Project fromUrl: 'AAEmptyTest.001.pr'.
The project should open again - this time faster - and you should enter it. 
"
	| absoluteUrl projectFilename projName proj serverDir pair |
	Project canWeLoadAProjectNow ifFalse: [^ self].
	absoluteUrl := Url absoluteFromFileNameOrUrlString: urlString.
	projectFilename := absoluteUrl fileName.
	projName := (Project parseProjectFileName: projectFilename unescapePercents) first.
	(proj := Project named: projName) ifNotNil: [
		"it appeared"
		^ProjectEntryNotification signal: proj].

	serverDir := ServerDirectory serverForURL: absoluteUrl directoryUrl asString.
	serverDir ifNotNil: [
		pair := self mostRecent: projectFilename onServer: serverDir.
		"Pair first is name exactly as it is on the server"
		pair first ifNil: [^self current openBlankProjectNamed: projName].
		projectFilename := pair first].

	ProjectLoading
		installRemoteNamed: projectFilename
		from: absoluteUrl asString unescapePercents
		named: projName
		in: CurrentProject! !

!Project class methodsFor: 'error recovery' stamp: 'mt 5/9/2016 12:59'!
handlePrimitiveError: errorMessage
	"This method is called when the error handling results in a recursion in 
	calling on error: or halt or halt:.."

	self tryOtherProjectForRecovery: errorMessage.
	self tryEmergencyEvaluatorForRecovery: errorMessage.
	
	Project current restoreDisplay.! !

!Project class methodsFor: 'utilities' stamp: 'dgd 11/5/2004 15:03'!
hierarchyOfNamesAndProjects
	"Answer a list of all project names, with each entry preceded  
	by white space commensurate with its depth beneath the top  
	project"
	^ self allProjectsOrdered
		select: [:project | (#('<no name -- garbage?>' 'Building with Squeak' 'Fun With Music' ) includes: project name) not]
		thenCollect: [:project | Array with: project nameAdjustedForDepth with: project]! !

!Project class methodsFor: 'class initialization' stamp: 'mt 5/6/2016 14:37'!
initialize

	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self.! !

!Project class methodsFor: 'squeaklet on server' stamp: 'RAA 1/28/2001 08:39'!
isBadNameForStoring: aString

	| badChars |

	"will the name of this project cause problems when stored on an arbitrary file system?"
	badChars := #( $: $< $> $| $/ $\ $? $* $" $.) asSet.
	^aString size > 24 or: [
		aString anySatisfy: [ :each | 
			each asciiValue < 32 or: [badChars includes: each]
		]
	]
! !

!Project class methodsFor: 'utilities' stamp: 'dtl 11/23/2009 20:43'!
jumpToProject
	"Project jumpToProject"
	"Present a list of potential projects and enter the one selected."

	self current jumpToProject! !

!Project class methodsFor: 'utilities' stamp: 'dtl 11/23/2009 22:27'!
jumpToSelection: selection
	"Enter the project corresponding to this menu selection."

	self flag: #toRemove. "Project current jumpToProject"
	^ self current jumpToSelection: selection
! !

!Project class methodsFor: 'squeaklet on server' stamp: 'cmm 10/15/2013 13:19'!
loaderUrl
	"Return a url that will launch a project in a browser by composing a url like <loaderURL>?<projectURL>"
	^ Smalltalk namedArguments
		at: 'LOADER_URL'
		ifAbsent: [  ]! !

!Project class methodsFor: 'class initialization' stamp: 'tfel 7/28/2016 17:34'!
localeChanged
	self current class == self
		ifTrue: [self current localeChanged].! !

!Project class methodsFor: 'utilities' stamp: 'dtl 11/23/2009 15:05'!
maybeForkInterrupt

	self flag: #toRemove. "unreferenced in image, check eToys"
	Preferences cmdDotEnabled ifFalse: [^self].
	[self current interruptName: 'User Interrupt'] fork
! !

!Project class methodsFor: 'squeaklet on server' stamp: 'KR 5/1/2008 01:58'!
mostRecent: projName onServer: aServerDirectory
	| stem list max goodName triple num stem1 stem2 rawList nothingFound unEscName |
	"Find the exact fileName of the most recent version of project with the stem name of projName.  Names are of the form 'projName|mm.pr' where mm is a mime-encoded integer version number.
	File names may or may not be HTTP escaped, %20 on the server."

	self flag: #bob.		"do we want to handle unversioned projects as well?"
						"I think we do now - Yoshiki."

	nothingFound _ {nil. -1}.
	aServerDirectory ifNil: [^nothingFound].
	"23 sept 2000 - some old projects have periods in name so be more careful"
	unEscName _ projName unescapePercents.
	triple _ Project parseProjectFileName: unEscName.
	stem _ triple first.
	rawList _ aServerDirectory fileNames.

	rawList isString ifTrue: [self inform: 'server is unavailable' translated. ^nothingFound].
	list _ rawList collect: [:nnn | nnn unescapePercents].
	max _ -1.  goodName _ nil.
	list withIndexDo: [:aName :ind |
		((aName beginsWith: stem)) ifTrue: [
			((aName endsWith: triple last) or: [triple last = '' and: [aName endsWith: '.pr']]) ifTrue: [
			num _ (Project parseProjectFileName: aName) second.
			num > max ifTrue: [max _ num.  goodName _ (rawList at: ind)]]]].

	max = -1 ifFalse: [^ Array with: goodName with: max].

	"try with underbar for spaces on server"
	(stem includes: $ ) ifTrue: [
		stem1 _ stem copyReplaceAll: ' ' with: '_'.
		list withIndexDo: [:aName :ind |
			(aName beginsWith: stem1) ifTrue: [
				num _ (Project parseProjectFileName: aName) second.
				num > max ifTrue: [max _ num.  goodName _ (rawList at: ind)]]]].
	max = -1 ifFalse: [^ Array with: goodName with: max].
	
	"try without the marker | "
	stem1 _ stem allButLast, '.pr'.
	stem2 _ stem1 copyReplaceAll: ' ' with: '_'.	"and with spaces replaced"
	list withIndexDo: [:aName :ind |
		(aName beginsWith: stem1) | (aName beginsWith: stem2) ifTrue: [
			(triple _ aName findTokens: '.') size >= 2 ifTrue: [
				max _ 0.  goodName _ (rawList at: ind)]]].	"no other versions"
	max = -1 ifFalse: [^ Array with: goodName with: max].

	^nothingFound		"no matches"
! !

!Project class methodsFor: 'squeaklet on server' stamp: 'RAA 9/27/2000 14:11'!
namedUrl: urlString
	| projName |
	"Return project if in, else nil"

	"Ted's fix for unreachable projects"

	projName := (urlString findTokens: '/') last.
	projName := (Project parseProjectFileName: projName unescapePercents) first.
	^ Project named: projName
! !

!Project class methodsFor: 'utilities' stamp: 'jla 4/2/2001 15:57'!
namedWithDepth: projName
	"Answer the project with the given name, or nil if there is no project of that given name."
	"(Project named: 'New Changes') enter"

	^ self allProjects
		detect: [:proj | 
			  proj name = projName or:
				[proj nameAdjustedForDepth = projName]]
		ifNone: [nil]! !

!Project class methodsFor: 'utilities' stamp: 'tk 3/10/2000 21:10'!
named: projName
	"Answer the project with the given name, or nil if there is no project of that given name."
	"(Project named: 'New Changes') enter"

	^ self allProjects
		detect: [:proj | proj name = projName]
		ifNone: [nil]
! !

!Project class methodsFor: 'utilities' stamp: 'RAA 11/11/2000 23:05'!
named: projName in: aListOfProjects
	"Answer the project with the given name, or nil if there is no project of that given name."
	"Use given collection for speed until we get faster #allProjects"

	^ aListOfProjects
		detect: [:proj | proj name = projName]
		ifNone: [nil]
! !

!Project class methodsFor: 'utilities' stamp: 'mt 2/17/2016 10:01'!
ofWorld: anObject
	"Find the project of a world."

	"Usually it is the current project"
	CurrentProject world == anObject ifTrue: [^ CurrentProject].

	"Inefficient enumeration if it is not..."
	^ self allProjects detect: [:pr |
		pr world isInMemory 
			ifTrue: [pr world == anObject]
			ifFalse: [false]]
		ifNone: [nil]! !

!Project class methodsFor: 'squeaklet on server' stamp: 'ar 10/11/2000 15:42'!
parseProjectFileName: aString
	"It was formerly possible to have periods in projct names and this messed up some parsing methods. Try to handle that more gracefully and allow for a change in scheme at a later time.
	ar 10/11/2000: Switch to a different version encoding scheme. The new scheme is
		baseName.NNN.ext
	where NNN is at least three digits wide and encodes the version in a human readable form.
	Examples:
		Project parseProjectFileName: 'My Project.007.pr'.
		Project parseProjectFileName: 'My.First.Project.042.prj'.
		Project parseProjectFileName: 'My Project.123456.p r o j e c t'.
	The 'dot' is determined on FileDirectory>>dot to compensate for platforms wishing to use something different from a period. Also allows parsing the former encoding of file using Base64 encoded versions of the form
	Project parseProjectFileName: 'aa.bb.cc|AQ.ss'
	"
	| suffix baseName version versionAndSuffix index tokens |
	"answer an array with: 
		1 = basic project name
		2 = version string
		3 = suffix (pr)"

	"First check for the old style versions"
	index := aString findLast:[:ch| ch = $|].
	index = 0 ifFalse:["Old style version"
		baseName := aString copyFrom: 1 to: index-1.
		versionAndSuffix := aString copyFrom: index+1 to: aString size.
		(versionAndSuffix occurrencesOf: $.) = 0 ifTrue: [^ #('no suffix')].
		version := versionAndSuffix copyUpTo: $..
		suffix := versionAndSuffix copyFrom: version size+1 to: versionAndSuffix size.
		"Decode Base64 encoded version"
		version isEmpty
			ifTrue:[version := 0]
			ifFalse:[version := Base64MimeConverter decodeInteger: version unescapePercents].
		^{baseName. version. suffix}].
	"New style versions"
	tokens := aString findTokens: FileDirectory dot.
	tokens size < 2 "Not even a single dot"
		ifTrue:[^{aString. 0. ''}].
	tokens size < 3 ifTrue:["Only one dot"
		self flag: #arNote. "We could allow project file names of the form 'project.001' (e.g., no project extension) or '.001.pr' (without a base name) but I don't think its a good idea."
		^{tokens first. 0. tokens last}].
	suffix := tokens last.
	version := tokens at: tokens size - 1.
	(version anySatisfy:[:ch| ch isDigit not]) ifTrue:[
		"Non-digit version??? I don't think so..."
		baseName := aString copyFrom: 1 to: aString size - suffix size - 1.
		^{baseName. 0. suffix}].
	baseName := aString copyFrom: 1 to: aString size - suffix size - version size - 2.
	version := version asInteger.
	^{baseName. version. suffix}! !

!Project class methodsFor: 'squeaklet on server' stamp: 'ar 2/27/2001 13:43'!
projectExtension
	^'pr'! !

!Project class methodsFor: 'utilities' stamp: 'sbw 4/13/2003 12:53'!
projectHierarchy
	"Answer a string representing all the projects in the system in  
	hierarchical order."
	"Project projectHierarchy"
	^ String
		streamContents: [:aStream | self hierarchyOfNamesAndProjects
				do: [:aPair | aStream nextPutAll: aPair first;
						 cr]]! !

!Project class methodsFor: 'preferences' stamp: 'tfel 8/10/2016 14:23'!
publishInSexp

	Smalltalk at: #SISSDictionaryForScanning ifPresent: [:siss | ^ siss publishInSexp].
	^ false! !

!Project class methodsFor: 'class initialization' stamp: 'RAA 12/17/2000 12:37'!
rebuildAllProjects
	"Project rebuildAllProjects"

	AllProjects := nil.
	self allProjects.! !

!Project class methodsFor: 'utilities' stamp: 'mt 5/1/2016 17:02'!
releaseProjectReferences: outgoingProject
	"Give all kinds of projects the chance to do generic clean-up to their best knowledge. Consider image segments first."
	
	Smalltalk garbageCollect.
		
	ImageSegment allSubInstancesDo: [:seg |
		seg ifOutPointer: outgoingProject thenAllObjectsDo: [:obj |
			(obj isKindOf: Project) ifTrue: [obj deletingProject: outgoingProject]]].
	
	self allSubclassesDo: [:projectClass |
		(projectClass selectors includes: #releaseProjectReferences:)
			ifTrue: [projectClass releaseProjectReferences: outgoingProject]].! !

!Project class methodsFor: 'utilities' stamp: 'mt 5/13/2016 13:53'!
removeAllButCurrent
	"Project removeAllButCurrent"

	self current removeAllOtherProjects.! !

!Project class methodsFor: 'utilities' stamp: 'mir 11/26/2004 16:15'!
removeAll: projects
	"Project removeAll: (Project allSubInstances copyWithout: Project current)"

	AllProjects := nil.
	Smalltalk garbageCollect.

	ProjectHistory currentHistory initialize.
	projects do: [:project |
		Project deletingProject: project.
		StandardScriptingSystem removePlayersIn: project].

	Smalltalk garbageCollect.
	Smalltalk garbageCollect.
! !

!Project class methodsFor: 'shrinking' stamp: 'mt 5/1/2016 17:21'!
removeProjectsFromSystem
	"Remove all projects of this kind from the system, reorganizing the project hierarchy as needed.
	This method was originally implemented as SmalltalkImage>>zapMVCProjects in earlier
	versions of Squeak."

	"MVCProject removeProjectsFromSystem"
	"MorphicProject removeProjectsFromSystem"

	| projectsToRemove projectsToReorganize |
	
	(Project current isKindOf: self) ifTrue: [
		"We have to enter another project to unload this project kind."
		(Project subclasses detect: [:ea | (ea includesBehavior: self) not] ifNone: [])
			ifNil: [^ Error signal: 'You have to enter another kind of project.']
			ifNotNil: [:projectClass |
				^ projectClass new
					addDeferredUIMessage: [self removeProjectsFromSystem];
					enter]].

	(Project topProject isKindOf: self) ifTrue: [
		"We have to become the top project now."
		Project current beTopProject].
		
	Smalltalk garbageCollect. "So allInstances is precise"
	projectsToRemove := OrderedCollection new.
	projectsToReorganize := OrderedCollection new.
	
	Project allSubInstancesDo: [:proj | 
		(proj isKindOf: self)
			ifTrue: [projectsToRemove add: proj]
			ifFalse: [(proj parent isKindOf: self)
				ifTrue: [projectsToReorganize add: proj]]].
			
	projectsToReorganize do: [:proj |
		"Remove views and refs."
		proj parent deletingProject: proj.
		"Add as sub-project to top project to avoid cycles and confusion. Ignore the current project."
		Project topProject addProject: proj].

	projectsToRemove do: [:proj | proj delete].! !

!Project class methodsFor: 'utilities' stamp: 'dtl 7/11/2013 19:38'!
resumeProcess: aProcess
	"Adopt aProcess as the project process -- probably because of proceeding from a debugger"

	self current uiProcess: aProcess.
	aProcess resume! !

!Project class methodsFor: 'utilities' stamp: 'mt 5/7/2016 17:23'!
returnToParentProject

	Project current isTopProject
		ifFalse: [Project current parent enter].! !

!Project class methodsFor: 'utilities' stamp: 'tk 10/26/1999 14:25'!
returnToPreviousProject
	"Return to the project from which this project was entered. Do nothing if the current project has no link to its previous project."

	| prevProj |
	prevProj := CurrentProject previousProject.
	prevProj ifNotNil: [prevProj enter: true revert: false saveForRevert: false].
! !

!Project class methodsFor: 'squeaklet on server' stamp: 'RAA 10/17/2000 19:11'!
serverDirectoryFromURL: aURLString

	| dir |
	self flag: #bob.		"need to include swikis in this - hacked for now"

	(aURLString findString: '/SuperSwikiProj/') > 0 ifTrue: [
		dir := SuperSwikiServer new fullPath: (aURLString copyUpToLast: $/).
		^dir
	].

	^ServerDirectory new fullPath: aURLString! !

!Project class methodsFor: 'squeaklet on server' stamp: 'RAA 10/17/2000 19:11'!
serverFileFromURL: aURLString

	| dir |
	self flag: #bob.		"need to include swikis in this - hacked for now"

	(aURLString findString: '/SuperSwikiProj/') > 0 ifTrue: [
		dir := SuperSwikiServer new fullPath: (aURLString copyUpToLast: $/).
		^dir readOnlyFileNamed: (aURLString findTokens: '/') last
	].
	^ServerFile new fullPath: aURLString! !

!Project class methodsFor: 'snapshots' stamp: 'mt 5/6/2016 14:34'!
shutDown: quitting

	quitting ifTrue: [
		Project current world triggerEvent: #aboutToLeaveWorld].! !

!Project class methodsFor: 'squeaklet on server' stamp: 'tfel 8/10/2016 10:28'!
squeakletDirectory

	| squeakletDirectoryName |
	squeakletDirectoryName := (Smalltalk at: #SugarLauncher ifPresent: [:c |
		c current parameterAt: 'SQUEAKLETS' ifAbsent: []]) ifNil: ['Squeaklets'].
	(FileDirectory default directoryExists: squeakletDirectoryName) ifFalse: [
		FileDirectory default createDirectory: squeakletDirectoryName
	].
	^FileDirectory default directoryNamed: squeakletDirectoryName! !

!Project class methodsFor: 'snapshots' stamp: 'mt 5/6/2016 15:09'!
startUp: startAfresh

	Project current restoreDisplay.

	startAfresh ifTrue: [
		Project current world triggerEvent: #aboutToEnterWorld].! !

!Project class methodsFor: 'utilities' stamp: 'RAA 9/27/2000 18:52'!
storeAllInSegments
	"Write out all Projects in this Image.
	Project storeAllInSegments.		"

	| all ff ll |
all := Project allProjects.
Transcript show: 'Initial Space Left: ', (ff := Smalltalk garbageCollect) printString; cr.
all do: [:proj |
	Transcript show: proj name; cr.
	proj storeSegment  "storeSegmentNoFile"].
Transcript show: 'After writing all: ', (ll := Smalltalk garbageCollect) printString; cr.
Transcript show: 'Space gained: ', (ll - ff) printString; cr.
"some will come back in"! !

!Project class methodsFor: 'squeaklet on server' stamp: 'KR 5/1/2008 01:58'!
sweep: aServerDirectory
	| repository list parts ind entry projectName versions |
	"On the server, move all but the three most recent versions of each Squeaklet to a folder called 'older'"
	"Project sweep: ((ServerDirectory serverNamed: 'DaniOnJumbo') clone 
				directory: '/vol0/people/dani/Squeaklets/2.7')"

	"Ensure the 'older' directory"
	(aServerDirectory includesKey: 'older') 
		ifFalse: [aServerDirectory createDirectory: 'older'].
	repository _ aServerDirectory clone directory: aServerDirectory directory, '/older'.

	"Collect each name, and decide on versions"
	list _ aServerDirectory fileNames.
	list isString ifTrue: [^ self inform: 'server is unavailable' translated].
	list _ list asSortedCollection asOrderedCollection.
	parts _ list collect: [:en | Project parseProjectFileName: en].
	parts _ parts select: [:en | en third = 'pr'].
	ind _ 1.
	[entry _ list at: ind.
		projectName _ entry first asLowercase.
		versions _ OrderedCollection new.  versions add: entry.
		[(ind _ ind + 1) > list size 
			ifFalse: [(parts at: ind) first asLowercase = projectName 
				ifTrue: [versions add: (parts at: ind).  true]
				ifFalse: [false]]
			ifTrue: [false]] whileTrue.
		aServerDirectory moveYoungest: 3 in: versions to: repository.
		ind > list size] whileFalse.
! !

!Project class methodsFor: 'utilities' stamp: 'RAA 9/27/2000 19:00'!
topProject
	"Answer the top project.  There is only one"

	^ self allProjects detect: [:p | p isTopProject]! !

!Project class methodsFor: 'error recovery' stamp: 'mt 5/9/2016 12:58'!
tryEmergencyEvaluatorForRecovery: errorMessage

	| hasTranscripter transcripter |
	hasTranscripter := (Smalltalk classNamed: #Transcripter)
		ifNotNil: [ :t | transcripter := t. true]
		ifNil: [false].
	(String
		streamContents: 
			[:s |
			| context |
			s nextPutAll: '***System error handling failed***'.
			s cr; nextPutAll: errorMessage.
			context := thisContext sender sender.
			20 timesRepeat: [context == nil ifFalse: [s cr; print: (context := context sender)]].
			s cr; nextPutAll: '-------------------------------'.
			hasTranscripter
				ifTrue: [
					s cr; nextPutAll: 'Type CR to enter an emergency evaluator.'.
					s cr; nextPutAll: 'Type any other character to restart.']
				ifFalse: [
					s cr; nextPutAll: 'Type any character to restart.']])
		displayAt: 0 @ 0.

	[Sensor keyboardPressed] whileFalse.

	Sensor keyboard = Character cr ifTrue: [
		hasTranscripter ifTrue: [transcripter emergencyEvaluator]].! !

!Project class methodsFor: 'error recovery' stamp: 'mt 5/9/2016 13:01'!
tryOtherProjectForRecovery: errorMessage
	"Try entering the parent project if it uses a different user interface. We determine this by comparing the project's class."
	
	| safeProject nextProject |
	nextProject := Project current.
	safeProject := nil.
	[safeProject notNil or: [nextProject isTopProject]] whileFalse: [
		nextProject := nextProject parent.
		nextProject class == Project current class
			ifFalse: [safeProject := nextProject]].
	safeProject ifNotNil: [:p | 
		p enterForEmergencyRecovery.
		"Active process will usually suspend after this."].! !

!Project class methodsFor: 'constants' stamp: 'mt 8/16/2016 17:42'!
uiManager

	^ self current uiManager! !

!Project class methodsFor: 'constants' stamp: 'dtl 7/11/2013 19:40'!
uiProcess

	^ self current uiProcess! !

!Project class methodsFor: 'utilities' stamp: 'nice 12/27/2009 03:11'!
versionForFileName: version
	"Project versionForFileName: 7"
	
	^String streamContents:[:s| | v |
		v := version printString.
		v size < 3 ifTrue:[v := '0', v].
		v size < 3 ifTrue:[v := '0', v].
		s nextPutAll: v.
	]
! !

!Project methodsFor: 'resources' stamp: 'ar 3/2/2001 17:25'!
abortResourceLoading
	"Abort loading resources"
	resourceManager ifNil:[^self].
	resourceManager stopDownload.! !

!Project methodsFor: 'file in/out' stamp: 'sw 9/28/2009 15:48'!
acceptProjectDetails: details
	"Store project details back into a property of the world, and if a name is provided, make sure the name is properly installed in the project."

	world setProperty: #ProjectDetails toValue: details.
	details at: 'projectname' ifPresent: [ :newName | 
		self renameTo: newName]! !

!Project methodsFor: 'scheduling' stamp: 'dtl 9/20/2009 18:53'!
addDeferredUIMessage: valuableObject 
	"Arrange for valuableObject to be evaluated at a time when the user interface
	is in a coherent state."

	self subclassResponsibility! !

!Project methodsFor: 'release' stamp: 'di 9/28/1999 23:46'!
addDependent: aMorph

	"Projects do not keep track of their dependents, lest they point into other projects and thus foul up the tree structure for image segmentation."

	^ self  "Ignore this request"! !

!Project methodsFor: 'utilities' stamp: 'dtl 11/25/2009 13:53'!
addItem: item toMenu: menu selection: action color: aColor thumbnail: aForm
	"Add menu item representing the sender to a menu"

	self subclassResponsibility! !

!Project methodsFor: 'sub-projects & hierarchy' stamp: 'mt 5/1/2016 13:54'!
addProject: aProject
	"Add the given project as sub-project."
	
	aProject setParent: self.! !

!Project methodsFor: 'printing' stamp: 'sw 9/7/2000 06:56'!
addSubProjectNamesTo: aStream indentation: anIndentation
	"Add the names of the receiver and all its subprojects, and all *their* subprojects recursively, to aStream, indenting by the specified number of tab stops "

	self isTopProject ifFalse:  "circumvent an annoying cr at the top "
		[aStream cr].  
	aStream tab: anIndentation; nextPutAll: self name.
	self subProjects do:
		[:p |
			p addSubProjectNamesTo: aStream indentation: anIndentation + 1]! !

!Project methodsFor: 'file in/out' stamp: 'dtl 11/2/2009 23:40'!
armsLengthCommand: aCommand withDescription: aString
	"Set things up so that this aCommand is sent to self as a message
after jumping to the parentProject.  For things that can't be executed
while in this project, such as saveAs, loadFromServer, storeOnServer.  See
ProjectViewMorph step."

	self subclassResponsibility! !

!Project methodsFor: 'file in/out' stamp: 'ar 10/11/2000 15:25'!
assureIntegerVersion
	"For converting the project versions"
	self currentVersionNumber. "Does it for us"! !

!Project methodsFor: 'menu messages' stamp: 'dtl 11/6/2009 21:35'!
assureNavigatorPresenceMatchesPreference
	"Make sure that the current project conforms to the presence/absence of the navigator"
! !

!Project methodsFor: 'sub-projects & hierarchy' stamp: 'mt 5/1/2016 17:01'!
beTopProject

	self isTopProject ifTrue: [^ self].
	
	self parent deletingProject: self. "Just remove views and refs to me."
	self addProject: self class topProject.
	self setParent: self.! !

!Project methodsFor: 'editors' stamp: 'dtl 1/23/2010 18:24'!
bitEdit: aForm
	"Create and schedule a view located in an area designated by the user 
	that contains a view of aForm magnified by 8@8 that can be modified using
	a bit editor. It also contains a view of the original form."

	self subclassResponsibility
! !

!Project methodsFor: 'editors' stamp: 'dtl 1/23/2010 18:34'!
bitEdit: aForm at: magnifiedFormLocation scale: scaleFactor 
	"Create and schedule a view whose top left corner is magnifiedLocation 
	and that contains a view of aForm magnified by scaleFactor that  can be
	modified using the Bit Editor. It also contains a view of the original form."

	self subclassResponsibility! !

!Project methodsFor: 'utilities' stamp: 'mt 4/12/2016 19:07'!
buildJumpToMenu: menu
	"Make the supplied menu offer a list of potential projects, consisting of:
		*	The previous-project chain
		*	The next project, if any
		*	The parent project, if any
		*	All projects, alphabetically or hierarchically"

	| prev listed i next |
	listed := OrderedCollection with: self.
	i := 0.

	"The previous Project chain"
	prev := self previousProject.
	[(prev ~~ nil and: [(listed includes: prev) not])] whileTrue:
	  [i := i + 1.
		listed add: prev.
		self 	addItem: prev name , ' (', ('back {1}' translated format:{i}  ), ')'
				toMenu: menu 
				selection: ('%back' , i printString) 
				color: prev color
				thumbnail: prev thumbnail.
		prev := prev previousProject].
	i > 0 ifTrue: [menu addLine].


	"Then the next Project"
	(((next := self nextProject) ~~ nil) and: [(listed includes: next) not]) ifTrue:
		[self	addItem: (next name, ' (', ('forward {1}' translated format:{1}), ')') 
				toMenu: menu 
				selection: next name 
				color: next color
				thumbnail: next thumbnail]. 
	next ~~ nil ifTrue: [menu addLine].

	"Then the parent"
	self isTopProject ifFalse: 
		[self	addItem: self parent name , ' (', 'parent' translated, ')' 
				toMenu: menu 
				selection: #parent 
				color: self parent color
				thumbnail: self parent thumbnail.
		  menu addLine].

	"Finally all the projects, in hierarchical or alphabetical order:"
	(Preferences alphabeticalProjectMenu
			ifTrue:
				[Project allNamesAndProjects]
			ifFalse:
				[Project hierarchyOfNamesAndProjects]) do:

		[:aPair | | toAdd | 
			toAdd := aPair last isCurrentProject
				ifTrue:
				  [aPair first, ' (', 'current' translated, ')']
				ifFalse:
				  [aPair first].
			self	addItem: toAdd 
				toMenu: menu 
				selection: aPair first 
				color: aPair last color
				thumbnail: aPair last thumbnail].
	^ menu! !

!Project methodsFor: 'file in/out' stamp: 'ar 10/11/2000 15:06'!
bumpVersion: versionNumber
	"Make a new version after the previous version number"
	versionNumber ifNil:[^0].
	^versionNumber + 1! !

!Project methodsFor: 'release' stamp: 'sw 7/6/1998 11:16'!
canDiscardEdits
	"Don't regard a window of mine as one to be discarded as part of a 'closeUnchangedWindows' sweep"

	^ false! !

!Project methodsFor: 'accessing' stamp: 'tk 4/5/2000 16:29'!
changeSet
	^ changeSet! !

!Project methodsFor: 'sub-projects & hierarchy' stamp: 'jla 5/28/2001 21:50'!
children
	"Answer a list of all the subprojects of the receiver"
	
	| children |
	children := OrderedCollection new.
	Project allProjects do: [ :p | 
		(self == p parent and: [self ~~ p]) ifTrue:
			[ children add: p ]].
	^ children

"
Project topProject children
"! !

!Project methodsFor: 'language' stamp: 'dtl 11/29/2009 13:05'!
chooseNaturalLanguage
	"Put up a menu allowing the user to choose the natural language for the project"

	"Project current chooseNaturalLanguage"

	self subclassResponsibility
! !

!Project methodsFor: 'accessing' stamp: 'mt 4/12/2016 19:09'!
color
	"Color representation for this project."
	
	^ Color black! !

!Project methodsFor: 'utilities' stamp: 'mt 5/4/2016 15:29'!
composeDisplayTextIntoForm: displayText
	"Compose the given display text into a form. This is at project level so that UI-specific mechanisms can be used. We may want to make this independent from Morphic or MVC in the future. Good-looking text rendering is good for any user interface."
	
	| port form font width height lines |
	font := displayText textStyle defaultFont.
	lines := displayText string lines.
	
	width := lines
		ifEmpty: [1]
		ifNotEmpty: [(lines collect: [:line | font widthOfString: line]) max].
	height := font height * lines size.
	
	form := Form extent: width@height depth: 32.
	port := BitBlt toForm: form.
	
	"1) Fill background."
	port
		fill: form boundingBox
		fillColor: displayText backgroundColor
		rule: Form over.
	
	"2) Draw text lines."
	port combinationRule: Form paint.
	font
		installOn: port
		foregroundColor: displayText foregroundColor
		backgroundColor: Color transparent.
		
	lines withIndexDo: [:line :index |
		font
			displayString: line
			on: port
			from: 1
			to: line size
			at: 0@((index-1) * font height)
			kern: 0].
	
	^ form! !

!Project methodsFor: 'file in/out' stamp: 'bf 9/17/2010 21:38'!
compressFilesIn: tempDir to: localName in: localDirectory
	"Compress all the files in tempDir making up a zip file in localDirectory named localName"

	| archive archiveName entry fileNames |
	archive := ZipArchive new.
	fileNames := tempDir fileNames.
	(fileNames includes: 'manifest')
		ifTrue: [fileNames := #('manifest'), (fileNames copyWithout: 'manifest')].
	fileNames do:[:fn|
		archiveName := fn.
		entry := archive addFile: (tempDir fullNameFor: fn) as: archiveName.
		entry desiredCompressionMethod: (
			fn = 'manifest'
				ifTrue: [ZipArchive compressionLevelNone] 
				ifFalse: [ZipArchive compressionDeflated]).
	].
	archive writeToFileNamed: (localDirectory fullNameFor: localName).
	archive close.
	tempDir fileNames do:[:fn|
		tempDir deleteFileNamed: fn ifAbsent:[]].
	localDirectory deleteDirectory: tempDir localName.! !

!Project methodsFor: 'file in/out' stamp: 'nice 12/27/2009 03:10'!
compressFilesIn: tempDir to: localName in: localDirectory resources: collector
	"Compress all the files in tempDir making up a zip file in localDirectory named localName"
	| archive urlMap |
	urlMap := Dictionary new.
	collector locatorsDo:[:loc|
		"map local file names to urls"
		urlMap at: (tempDir localNameFor: loc localFileName) put: loc urlString.
		ResourceManager cacheResource: loc urlString inArchive: localName].
	archive := ZipArchive new.
	tempDir fileNames do:[:fn| | archiveName entry |
		archiveName := urlMap at: fn ifAbsent:[fn].
		entry := archive addFile: (tempDir fullNameFor: fn) as: archiveName.
		entry desiredCompressionMethod: ZipArchive compressionStored.
	].
	archive writeToFileNamed: (localDirectory fullNameFor: localName).
	archive close.
	tempDir fileNames do:[:fn|
		tempDir deleteFileNamed: fn ifAbsent:[]].
	localDirectory deleteDirectory: tempDir localName.! !

!Project methodsFor: 'file in/out' stamp: 'RAA 6/3/2000 10:25'!
couldBeSwappedOut

	self flag: #bob.		"need a better test in multi-project world"
	^self isCurrentProject not! !

!Project methodsFor: 'utilities' stamp: 'tfel 8/10/2016 15:11'!
createViewIfAppropriate
	"overridden in subclasses"
	^ self! !

!Project methodsFor: 'file in/out' stamp: 'ar 10/11/2000 15:25'!
currentVersionNumber

	version ifNil: [^0].
	version isInteger ifTrue:[^version].
	version := Base64MimeConverter decodeInteger: version unescapePercents.
	^version! !

!Project methodsFor: 'protocols' stamp: 'dtl 2/27/2010 10:18'!
currentVocabulary
	"Answer the currently-prevailing default vocabulary."

	self subclassResponsibility! !

!Project methodsFor: 'debugging' stamp: 'mt 4/18/2016 16:33'!
debugMethod: aCompiledMethod forReceiver: anObject inContext: aContextOrNil
	"Just execute the method and return the result. We cannot know how interactive debugging works for arbitrary projects."
	
	^ aCompiledMethod
		valueWithReceiver: anObject
		 arguments: (aContextOrNil
							ifNil: [#()]
							ifNotNil: [{aContextOrNil}])! !

!Project methodsFor: 'file in/out' stamp: 'dtl 11/27/2009 14:11'!
decideAboutCreatingBlank: otherProjectName

	| resp |

	"20 Oct - just do it"
	true "version isNil" ifFalse: [	"if saved, then maybe don't create"
		resp := (UIManager default chooseFrom: #('Yes, make it up' 'No, skip it') 
			title: ('I cannot locate the project\',
				otherProjectName,
				'\Would you like me to create a new project\with that name?'
			) withCRs).
		resp = 1 ifFalse: [^ nil]
	].
	^Project current openBlankProjectNamed: otherProjectName! !

!Project methodsFor: 'release' stamp: 'mt 5/1/2016 13:50'!
delete

	self isCurrentProject
		ifTrue: [^ Error signal: 'Cannot delete the current project.'].

	self removeChangeSetIfPossible.

	ProjectHistory forget: self.
	Project deletingProject: self.! !

!Project methodsFor: 'release' stamp: 'mt 5/1/2016 13:16'!
deletingProject: outgoingProject
	"Clear my previousProject link if it points at the given Project, which is being deleted."

	parentProject == outgoingProject
		ifTrue: [parentProject := parentProject parent].
	
	previousProject == outgoingProject
		ifTrue: [previousProject := previousProject previousProject].
	previousProject == outgoingProject
		ifTrue: [previousProject := nil].
	
	nextProject == outgoingProject
		ifTrue: [nextProject := nextProject nextProject].
	nextProject == outgoingProject
		ifTrue: [nextProject := nil].! !

!Project methodsFor: 'active process' stamp: 'tk 11/14/2007 15:29'!
depth
	"Return the depth of this project from the top.
	 topProject = 0, next = 1, etc."
	"Project current depth."

	| depth project |
	depth _ 0.
	project _ self.
	
	[project class == DiskProxy ifTrue: [^ depth].
	 project isTopProject]
		whileFalse:
			[project _ project parent.
			depth _ depth + 1].
	^ depth! !

!Project methodsFor: 'dispatching' stamp: 'dtl 7/7/2013 18:42'!
dispatchTo: requestor addPrefixAndSend: baseSelector withArguments: arguments
	"Sender wants to perform a method with dependencies on the type of project.
	Dispatch to an appropriate method for the current project.
	
	This method is a workaround for lack of proper ToolBuilder support in the sender,
	and should be eliminated when possible (dtl Feb 2010)"

	| selector |
	"Flag target selectors to help identify senders"
	true ifFalse: [
		self flag: #mvcOpenLabel:in: .
		self flag: #morphicOpenLabel:in: .
		self flag: #mvcStartUpLeftFlush .
		self flag: #morphicStartUpLeftFlush .
		self flag: #mvcStartUpWithCaption:icon:at:allowKeyboard: .
		self flag: #morphicStartUpWithCaption:icon:at:allowKeyboard: .
		self flag: #mvcOpen: .
		self flag: #morphicOpen: .
		self flag: #mvcOpen .
		self flag: #morphicOpen .
		self flag: #mvcOpenOn:context:label:contents:fullView: .
		self flag: #morphicOpenOn:context:label:contents:fullView: .
		self flag: #mvcResumeProcess: .
		self flag: #morphicResumeProcess:
	].
	selector := (self selectorPrefixForDispatch, baseSelector) asSymbol.
	^ requestor perform: selector withArguments: arguments! !

!Project methodsFor: 'displaying' stamp: 'mt 5/9/2016 11:12'!
displayDepthChanged
	"The depth of the display has changed."
	! !

!Project methodsFor: 'accessing' stamp: ''!
displayDepth: newDepth
	displayDepth := newDepth! !

!Project methodsFor: 'displaying' stamp: 'mt 5/9/2016 10:39'!
displaySizeChanged
	"Inform the current project that its display size has changed. Usually, projects invalidate all their drawing caches and restore their graphical contents."

	self invalidate; restore.! !

!Project methodsFor: 'transcripter' stamp: 'dtl 12/14/2013 14:52'!
displayTranscripter: transcripter
	"A transcripter is a minimal user interface to support an emergency evaluator.
	Different kinds of project may use different paragraph implementations."
	self subclassResponsibility! !

!Project methodsFor: 'displaying' stamp: 'ar 5/28/2000 19:43'!
displayZoom: entering
	"Show the project transition when entering a new project"
	| newDisplay vanishingPoint |
	"Play the flash transition if any."
	self projectParameters at: #flashTransition ifPresent:[:dict|
		dict at: CurrentProject ifPresent:[:player| ^player playProjectTransitionFrom: CurrentProject to: self entering: entering]].
	"Show animated zoom to new display"
	newDisplay := self imageForm.
	entering
		ifTrue: [vanishingPoint := Sensor cursorPoint]
		ifFalse: [vanishingPoint := self viewLocFor: CurrentProject].
	Display zoomIn: entering orOutTo: newDisplay at: 0@0
			vanishingPoint: vanishingPoint.! !

!Project methodsFor: 'file in/out' stamp: 'tak 6/12/2005 14:56'!
doArmsLengthCommand: aCommand

	"We are no longer the active project, so do it"

	[self perform: aCommand]
		ensure: [self enter: #specialReturn.	"re-enter me and forget the temp project"]

! !

!Project methodsFor: 'menu messages' stamp: 'yo 3/6/2005 00:09'!
doWeWantToRename

	| want |

	self hasBadNameForStoring ifTrue: [^true].
	(self name beginsWith: 'Unnamed' translated) ifTrue: [^true].
	want _ world valueOfProperty: #SuperSwikiRename ifAbsent: [false].
	world removeProperty: #SuperSwikiRename.
	^want

! !

!Project methodsFor: 'file in/out' stamp: 'mir 6/25/2001 10:50'!
downloadUrl
	"^(self primaryServerIfNil: [^'']) downloadUrl"
	^lastDirectory
		ifNil: [(self primaryServerIfNil: [^'']) downloadUrl]
		ifNotNil: [lastDirectory downloadUrl]! !

!Project methodsFor: 'utilities' stamp: 'dtl 11/30/2009 22:35'!
do: aBlock withProgressInfoOn: aMorphOrNil label: aString
	"Evaluate aBlock with a labeled progress bar"

	self subclassResponsibility

! !

!Project methodsFor: 'editors' stamp: 'dtl 1/24/2010 11:26'!
editCharacter: character ofFont: strikeFont
	"Open a bit editor on a character in the given strike font."

	self subclassResponsibility ! !

!Project methodsFor: 'file in/out' stamp: 'RAA 10/26/2000 14:12'!
ensureChangeSetNameUnique

	| myName |

	myName := self name.
	Project allProjects do: [:pp | 
		pp == self ifFalse: [
			(pp name = myName and: [pp projectChangeSet ~~ changeSet]) ifTrue: [
				(pp parameterAt: #loadingNewerVersion ifAbsent: [false]) ifTrue: [
					pp projectParameters at: #loadingNewerVersion put: false.
				] ifFalse: [
					changeSet ifNil: [^ changeSet := ChangeSet new].
					^changeSet name: (ChangeSet uniqueNameLike: myName)
				].
			]
		]
	]
! !

!Project methodsFor: 'enter' stamp: 'tk 10/26/1999 14:25'!
enter
	"Enter the new project"
	self enter: (CurrentProject parent == self) revert: false saveForRevert: false.! !

!Project methodsFor: 'enter - recovery' stamp: 'mt 5/6/2016 15:29'!
enterForEmergencyRecovery
	"Stripped down verion of #enter:revert:saveForRevert:. More error handling. Less features."
	
	| leavingProject process |
	self isCurrentProject ifTrue: [^ self].
	
	ProjectHistory remember: CurrentProject.
	nextProject := CurrentProject.
	
	[ CurrentProject world triggerEvent: #aboutToLeaveWorld.
	CurrentProject abortResourceLoading.
	CurrentProject finalExitActions: self.
	CurrentProject saveState ] on: Error do: [:ex | "Ignore." ].
	
	"********** SWITCHING CURRENT PROJECT **********"
	leavingProject := CurrentProject.
	CurrentProject := self.
	ProjectHistory remember: self.
	"********** SWITCHING CURRENT PROJECT **********"

	self loadState.
	self finalEnterActions: leavingProject.
	self addDeferredUIMessage: [self startResourceLoading].
	self world triggerEvent: #aboutToEnterWorld.
	
	"Now that everything is set up, we can show zoom animation.
	Do we really need this in case of an emergency?"
	self showZoom
		ifTrue: [self displayZoom: leavingProject parent ~~ self "Entering?"]
		ifFalse: [self restore].
	
	"Update processes at last."
	self scheduleProcessForEnter.
	
	"Do not terminate but suspend the projects ui process to support debugging."
	process := leavingProject uiProcess.
	self addDeferredUIMessage: [process debugWithTitle: 'FATAL PROJECT ERROR!!'].
	leavingProject suspendProcessForDebug.! !

!Project methodsFor: 'squeaklet on server' stamp: 'dtl 11/27/2009 13:41'!
enterIfThereOrFind: aProjectName

	self subclassResponsibility
! !

!Project methodsFor: 'enter' stamp: 'sw 11/10/1999 10:29'!
enter: returningFlag	
	self enter: returningFlag revert: false saveForRevert: false! !

!Project methodsFor: 'enter' stamp: 'mt 8/11/2016 11:24'!
enter: returningFlag revert: revertFlag saveForRevert: saveForRevert
	"Install my ChangeSet, Transcript, and scheduled views as current globals. If returningFlag is true, we will return to the project from whence the current project was entered; don't change its previousProject link in this case.
	If saveForRevert is true, save the ImageSegment of the project being left.
	If revertFlag is true, make stubs for the world of the project being left.
	If revertWithoutAsking is true in the project being left, then always revert."

	| leavingProject forceRevert response seg |

	self isIncompletelyLoaded
		ifTrue: [^ self loadFromServer: true].
	self isCurrentProject
		ifTrue: [^ self].
	
	forceRevert := false.
	CurrentProject rawParameters 
		ifNil: [revertFlag ifTrue: [^ self inform: 'nothing to revert to' translated]]
		ifNotNil: [saveForRevert ifFalse: [
				forceRevert := CurrentProject projectParameters 
								at: #revertWithoutAsking ifAbsent: [false]]].
	forceRevert not & revertFlag ifTrue: [
		response := (UIManager default chooseFrom: {
			'Revert to saved version' translated.
			'Cancel' translated.
		} title: 'Are you sure you want to destroy this Project\ and revert to an older version?\\(From the parent project, click on this project''s thumbnail.)' translated withCRs) = 1.
		response ifFalse: [^ self]].

	revertFlag | forceRevert 
		ifTrue: [seg := CurrentProject projectParameters at: #revertToMe ifAbsent: [
					^ self inform: 'nothing to revert to' translated]]
		ifFalse: [
			CurrentProject makeThumbnail.
			returningFlag == #specialReturn
				ifTrue:
					[ProjectHistory forget: CurrentProject.		"this guy is irrelevant"
					Project forget: CurrentProject]
				ifFalse:
					[ProjectHistory remember: CurrentProject]].

	(revertFlag | saveForRevert | forceRevert) ifFalse: [
		(Preferences valueOfFlag: #projectsSentToDisk)
			ifTrue: [
				self inform: 'Project serialization via image segments\does not work at the moment. Disabling the\preference #projectsSentToDisk now...' withCRs.
				Preferences disable: #projectsSentToDisk.
				"self storeToMakeRoom"]].

	"Update display depth for leaving and entring project."
	CurrentProject displayDepth: Display depth.
	displayDepth == nil ifTrue: [displayDepth := Display depth].
	self installNewDisplay: Display extent depth: displayDepth.

	returningFlag == #specialReturn ifTrue: [
		CurrentProject removeChangeSetIfPossible.	"keep this stuff from accumulating"
		nextProject := nil
	] ifFalse: [
		returningFlag
			ifTrue: [nextProject := CurrentProject]
			ifFalse: [previousProject := CurrentProject].
	].

	CurrentProject world triggerEvent: #aboutToLeaveWorld.
	CurrentProject abortResourceLoading.
	CurrentProject finalExitActions: self.
	CurrentProject saveState.
	
	"********** SWITCHING CURRENT PROJECT **********"
	leavingProject := CurrentProject.
	CurrentProject := self.
	ProjectHistory remember: self.
	"********** SWITCHING CURRENT PROJECT **********"

	self loadState.
	self finalEnterActions: leavingProject.
	self addDeferredUIMessage: [self startResourceLoading].
	self world triggerEvent: #aboutToEnterWorld.

	"Save project for revert."
	saveForRevert ifTrue: [
		Smalltalk garbageCollect.	"let go of pointers"
		leavingProject storeSegment.
		"result :=" leavingProject world isInMemory 
			ifTrue: ['Can''t seem to write the project.']
			ifFalse: [leavingProject projectParameters at: #revertToMe put: 
					leavingProject world xxxSegment clone].
				'Project written.'].
			"original is for coming back in and continuing."
	revertFlag | forceRevert ifTrue: [
		seg clone revert].	"non-cloned one is for reverting again later"
	self removeParameter: #exportState.
	
	"Now that everything is set up, we can show zoom animation."
	self showZoom
		ifTrue: [self displayZoom: leavingProject parent ~~ self "Entering?"]
		ifFalse: [self restore].
	
	"Update processes at last."
	self scheduleProcessForEnter.
	leavingProject terminateProcessForLeave.
! !

!Project methodsFor: 'menu messages' stamp: 'yo 7/2/2004 19:46'!
exit
	"Leave the current project and return to the project in which this one was created."

	self isTopProject ifTrue: [^ self inform: 'Can''t exit the top project' translated].
	parentProject enter: false revert: false saveForRevert: false.
! !

!Project methodsFor: 'file in/out' stamp: 'tfel 8/10/2016 13:24'!
exportSegmentFileName: aFileName directory: aDirectory

	^ self exportSegmentFileName: aFileName directory: aDirectory withoutInteraction: false! !

!Project methodsFor: 'file in/out' stamp: 'tfel 8/10/2016 14:22'!
exportSegmentFileName: aFileName directory: aDirectory withoutInteraction: noInteraction

	| exportChangeSet |

	"An experimental version to fileout a changeSet first so that a project can contain its own classes"

	"Store my project out on the disk as an *exported* ImageSegment.  Put all outPointers in a form that can be resolved in the target image.  Name it <project name>.extSeg.
	Player classes are included automatically."
	exportChangeSet := nil.
	(changeSet notNil and: [changeSet isEmpty not]) ifTrue: [
		(noInteraction or: [self confirm: 
	'Would you like to include all the changes in the change set
	as part of this publishing operation?' translated]) ifTrue: [
				exportChangeSet := changeSet
		].
	].

	Project publishInSexp ifTrue: [
		^ self exportSegmentInSexpWithChangeSet: exportChangeSet fileName: aFileName directory: aDirectory withoutInteraction: noInteraction
	].
	^ self 
		exportSegmentWithChangeSet: exportChangeSet
		fileName: aFileName 
		directory: aDirectory
		withoutInteraction: noInteraction! !

!Project methodsFor: 'file in/out' stamp: 'tfel 8/10/2016 14:25'!
exportSegmentInSexpWithChangeSet: aChangeSetOrNil fileName: aFileName directory: aDirectory withoutInteraction: noInteraction

	self subclassResponsibility! !

!Project methodsFor: 'file in/out' stamp: 'dtl 11/29/2009 13:27'!
exportSegmentWithCatagories: catList classes: classList fileName: aFileName directory: aDirectory
	"Store my project out on the disk as an *exported* ImageSegment.  All outPointers will be in a form that can be resolved in the target image.  Name it <project name>.extSeg.  What do we do about subProjects, especially if they are out as local image segments?  Force them to come in?
	Player classes are included automatically."

	self flag: #toRemove.
	self subclassResponsibility
! !

!Project methodsFor: 'file in/out' stamp: 'dtl 11/29/2009 13:24'!
exportSegmentWithChangeSet: aChangeSetOrNil fileName: aFileName
directory: aDirectory
	"Store my project out on the disk as an *exported*
ImageSegment.  All outPointers will be in a form that can be resolved
in the target image.  Name it <project name>.extSeg.  Whatdo we do
about subProjects, especially if they are out as local image
segments?  Force them to come in?
	Player classes are included automatically."

	self subclassResponsibility! !

!Project methodsFor: 'menu messages' stamp: ''!
fileOut
	changeSet fileOut! !

!Project methodsFor: 'enter' stamp: 'mt 4/17/2016 18:22'!
finalEnterActions: leavingProject

! !

!Project methodsFor: 'enter' stamp: 'mt 4/17/2016 18:23'!
finalExitActions: enteringProject

	SoundService stop.! !

!Project methodsFor: 'utilities' stamp: 'dtl 12/2/2009 20:29'!
findAFolderForProject: aProject label: dialogLabel
	"Find a folder for saving or loading a project"

	self subclassResponsibility
! !

!Project methodsFor: 'file in/out' stamp: 'dtl 12/2/2009 20:30'!
findAFolderToLoadProjectFrom

	^ CurrentProject findAFolderForProject: self label: 'Select a folder on a server:'
! !

!Project methodsFor: 'file in/out' stamp: 'dtl 12/2/2009 20:30'!
findAFolderToStoreProjectIn

	^ CurrentProject findAFolderForProject: self label: 'Select a folder on a server:'
! !

!Project methodsFor: 'accessing' stamp: 'dtl 11/22/2009 20:15'!
findProjectView: projectDescription
	"In this world, find the morph that holds onto the project described by projectDescription.
	projectDescription can be a project, or the name of a project.  The project may be
	represented by a DiskProxy.  The holder morph may be at any depth in the world.
	Need to fix this if Projects have subclasses, or if a class other than ProjectViewMorph
	can officially hold onto a project.  (Buttons, links, etc)
	If parent is an MVC world, return the ProjectController."

	self flag: #bob.		"read the comment"
	self subclassResponsibility! !

!Project methodsFor: 'flaps support' stamp: 'sw 4/24/2001 11:03'!
flapsSuppressed
	"Answer whether flaps are suppressed in this project"

	^ self showSharedFlaps not! !

!Project methodsFor: 'flaps support' stamp: 'dtl 11/13/2009 18:21'!
flapsSuppressed: aBoolean
	"Make the setting of the flag that governs whether global flaps are suppressed in the project be as indicated and add or remove the actual flaps"

	self projectPreferenceFlagDictionary at: #showSharedFlaps put: aBoolean not.
	self == Project current 
		ifTrue: "Typical case"
			[Preferences setPreference: #showSharedFlaps toValue: aBoolean not]
! !

!Project methodsFor: 'release' stamp: 'mt 5/1/2016 13:50'!
forget

	self isCurrentProject
		ifTrue: [^ Error signal: 'Cannot forget the current project.'].

	self class forget: self.! !

!Project methodsFor: 'accessing' stamp: 'mir 6/22/2001 20:06'!
forgetExistingURL
	self resourceManager makeAllProjectResourcesLocalTo: self resourceUrl.
	urlList := nil! !

!Project methodsFor: 'editors' stamp: 'dtl 1/24/2010 16:07'!
formEdit: aForm
	"Start up an instance of the form editor on a form."
 
	self subclassResponsibility! !

!Project methodsFor: 'editors' stamp: 'dtl 2/10/2010 13:19'!
formViewClass
	"Answer a class suitable for a view on a form or collection of forms"

	self subclassResponsibility! !

!Project methodsFor: 'file in/out' stamp: 'mir 6/7/2001 14:41'!
fromMyServerLoad: otherProjectName
	| pair pr dirToUse |
	"If a newer version of me is on the server, load it."

	(pr := Project named: otherProjectName) ifNotNil: ["it appeared"
		^ pr enter
	].
	dirToUse := self primaryServerIfNil: [
		lastDirectory ifNil: [
			self inform: 'Current project does not know a server either.'.
			^nil].
		lastDirectory].

	pair := self class mostRecent: otherProjectName onServer: dirToUse.
	pair first ifNil: [^self decideAboutCreatingBlank: otherProjectName].	"nothing to load"
	^ProjectLoading
		installRemoteNamed: pair first
		from: dirToUse
		named: otherProjectName
		in: self

! !

!Project methodsFor: 'futures' stamp: 'jcg 12/17/2009 01:27'!
future: receiver do: aSelector at: deltaMSecs args: args
	"Send a message deltaSeconds into the future.  No response is expected."
	| msg |
	msg := MessageSend receiver: receiver selector: aSelector arguments: args.
	deltaMSecs = 0 
		ifTrue: [self addDeferredUIMessage: msg]
		ifFalse: [
			[	(Delay forMilliseconds: deltaMSecs) wait.
				self addDeferredUIMessage: msg.
			] forkAt: Processor userSchedulingPriority + 1.
		]..
	^nil! !

!Project methodsFor: 'futures' stamp: 'fbs 5/15/2013 17:20'!
future: receiver send: aSelector at: deltaMSecs args: args
	"Send a message deltaSeconds into the future.  Answers a Promise that will be resolved at some time in the future."
	| pr closure |
	pr := Promise new.
	closure := [pr resolveWith: (receiver perform: aSelector withArguments: args)].
	deltaMSecs = 0
		ifTrue: [self addDeferredUIMessage: closure]
		ifFalse: [
			[	(Delay forMilliseconds: deltaMSecs) wait.
				self addDeferredUIMessage: 
					closure
			] forkAt: Processor userSchedulingPriority + 1.
		].
	^pr
		! !

!Project methodsFor: 'displaying' stamp: 'mt 5/6/2016 14:51'!
handleFatalDrawingError: errMsg
	"Handle a fatal drawing error."

	Display deferUpdates: false. "Just in case"
	self primitiveError: errMsg.

	"Hm... we should jump into a 'safe' worldState here, but how do we find it?!!"! !

!Project methodsFor: 'file in/out' stamp: 'RAA 9/21/2000 15:30'!
hasBadNameForStoring

	^Project isBadNameForStoring: self name
! !

!Project methodsFor: 'file in/out' stamp: 'sw 7/17/2009 21:58'!
htmlPagePrototype
	"Return the HTML page prototype"
^'<html>
<head>
<title>Squeak Project</title>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8">
</head>

<body bgcolor="#FFFFFF">
<EMBED 
	type="application/x-squeak-source"
	ALIGN="CENTER"
	WIDTH="$$WIDTH$$"
	HEIGHT="$$HEIGHT$$"
	src="$$PROJECT$$"
	pluginspage="http://www.squeakland.org/download/">

</EMBED>

</body>
</html>
'! !

!Project methodsFor: 'displaying' stamp: 'ar 6/2/1999 01:58'!
imageForm
	^self imageFormOfSize: Display extent
		depth: (displayDepth ifNil:[Display depth])! !

!Project methodsFor: 'displaying' stamp: 'mt 4/13/2016 11:30'!
imageFormOfSize: extentPoint depth: d
	| newDisplay |
	newDisplay := DisplayScreen extent: extentPoint depth: d.
	Display replacedBy: newDisplay do:[self invalidate; restore].
	^newDisplay! !

!Project methodsFor: 'initialization' stamp: ''!
initialExtent
	^ (Display extent // 6) + (0@17)! !

!Project methodsFor: 'initialization' stamp: 'ar 11/12/2009 01:11'!
initialize
	"Initialize the project, seting the CurrentProject as my parentProject and initializing my project preferences from those of the CurrentProject"
	Project addingProject: self.
	changeSet := ChangeSet new.
	transcript := TranscriptStream new.
	displayDepth := Display depth.
	parentProject := CurrentProject.
	self initializeProjectPreferences
! !

!Project methodsFor: 'transcripter' stamp: 'dtl 12/14/2013 13:28'!
initializeParagraphForTranscripter: transcripter
	"A transcripter is a minimal user interface to support an emergency evaluator.
	Different kinds of project may use different paragraph implementations."
	self subclassResponsibility! !

!Project methodsFor: 'project parameters' stamp: 'sw 4/24/2001 11:58'!
initializeProjectParameters
	"Initialize the project parameters."

	projectParameters := IdentityDictionary new.
	^ projectParameters! !

!Project methodsFor: 'project parameters' stamp: 'mt 8/27/2015 10:43'!
initializeProjectPreferences
	"Initialize the project's preferences from currently-prevailing preferences that are currently being held in projects in this system"
	
	projectPreferenceFlagDictionary := Project current projectPreferenceFlagDictionary deepCopy.    "Project overrides in the new project start out being the same set of overrides in the calling project"

	Preferences allPreferences do:  "in case we missed some"
		[:aPreference |
			aPreference localToProject ifTrue:
				[(projectPreferenceFlagDictionary includesKey: aPreference name) ifFalse:
			[projectPreferenceFlagDictionary at: aPreference name put: aPreference preferenceValue]]].

	(Project current projectParameterAt: #disabledGlobalFlapIDs  ifAbsent: [nil]) ifNotNil:
		[:idList | self projectParameterAt: #disabledGlobalFlapIDs put: idList copy]
! !

!Project methodsFor: 'initialization' stamp: 'di 4/14/1999 13:55'!
initialProject
	self saveState.
	parentProject := self.
	previousProject := nextProject := nil! !

!Project methodsFor: 'initialization' stamp: 'ar 5/16/2001 17:08'!
installNewDisplay: extent depth: depth
	"When entering a new project, install a new Display if necessary."
	^Display setExtent: extent depth: depth! !

!Project methodsFor: 'menu messages' stamp: 'mt 8/27/2015 10:43'!
installProjectPreferences
	"Install the settings of all preferences presently held individually by projects in the receiver's projectPreferenceFlagDictionary"

	
	Preferences allPreferences do:
		[:aPreference | | localValue | 
			aPreference localToProject ifTrue:
				[localValue := self projectPreferenceFlagDictionary at: aPreference name ifAbsent: [nil].
				localValue ifNotNil:
					[aPreference rawValue: localValue]]]! !

!Project methodsFor: 'debugging' stamp: 'fm 9/11/2010 00:56'!
interruptName: labelString
	"Create a Notifier on the active scheduling process with the given label."

	^ self subclassResponsibility
! !

!Project methodsFor: 'debugging' stamp: 'dtl 11/23/2009 14:54'!
interruptName: labelString preemptedProcess: theInterruptedProcess
	"Create a Notifier on the active scheduling process with the given label."

	^ self subclassResponsibility
! !

!Project methodsFor: 'displaying' stamp: 'ar 9/18/2009 20:24'!
invalidate
	"Invalidate the entire project so that a redraw will be forced later."
	^self subclassResponsibility! !

!Project methodsFor: 'testing' stamp: 'RAA 6/3/2000 10:23'!
isCurrentProject

	^self == CurrentProject! !

!Project methodsFor: 'testing' stamp: 'dtl 11/7/2009 21:09'!
isIncompletelyLoaded
	"Answer true if project is incomplete and should be loaded from server "

	^ false! !

!Project methodsFor: 'testing' stamp: 'mt 5/9/2016 10:03'!
isMorphic

	self flag: #moveEventually. "mt: Once we manage to remove all calls to this from outside the Morphic package, we can add this as an extension of Morphic to Project."
	^ false! !

!Project methodsFor: 'sub-projects & hierarchy' stamp: 'di 3/29/2000 15:37'!
isTopProject
	"Return true only if this is the top project (its own parent).
	Also include the test here for malformed project hierarchy."

	parentProject == self ifTrue: [^ true].
	parentProject == nil ifTrue: [self error: 'No project should have a nil parent'].
	^ false! !

!Project methodsFor: 'utilities' stamp: 'dtl 11/23/2009 21:44'!
jumpToProject
	"Present a list of potential projects and enter the one selected."

	"Project current jumpToProject"

	self subclassResponsibility! !

!Project methodsFor: 'utilities' stamp: 'dtl 11/23/2009 20:42'!
jumpToSelection: selection
	"Enter the project corresponding to this menu selection."
	
	"Project jumpToProject"
	| nBack prev pr |
	selection ifNil: [^ self].
	(selection beginsWith: '%back') ifTrue:
		[nBack := (selection copyFrom: 6 to: selection size) asNumber.
		prev := CurrentProject previousProject.
		1 to: nBack-1 do:
			[:i | prev ifNotNil: [prev := prev previousProject]].
		prev ifNotNil: [prev enter: true revert: false saveForRevert: false]].
	selection = #parent ifTrue:
		[CurrentProject parent enter: false revert: false saveForRevert: false.
		^ self].
	(pr := Project namedWithDepth: selection) ifNil: [^ self].
	pr enter: false revert: false saveForRevert: false! !

!Project methodsFor: 'accessing' stamp: 'tk 6/26/1998 22:17'!
labelString
	^ changeSet name! !

!Project methodsFor: 'accessing' stamp: 'mir 6/7/2001 16:18'!
lastDirectory: aDirectoryOrNil
	lastDirectory := aDirectoryOrNil! !

!Project methodsFor: 'accessing' stamp: 'RAA 9/28/2000 18:33'!
lastSavedAtSeconds

	^lastSavedAtSeconds! !

!Project methodsFor: 'sub-projects & hierarchy' stamp: 'mt 5/1/2016 13:59'!
liftSubProjects
	"Lift my sub-projects to my parent project."
	
	self parent ifNil: [^ Error signal: 'Cannot lift sub-project because I have to parent.'].
	
	self subProjects do: [:ea | self parent addProject: ea].! !

!Project methodsFor: 'file in/out' stamp: 'RAA 5/17/2000 18:52'!
loadFromServer
	
	^self loadFromServer: false! !

!Project methodsFor: 'file in/out' stamp: 'dtl 11/30/2009 22:43'!
loadFromServer: newerAutomatically
	"If a newer version of me is on the server, load it."
	| pair resp server |
	self assureIntegerVersion.

	self isCurrentProject ifTrue: ["exit, then do the command"
		^ self armsLengthCommand: #loadFromServer withDescription: 'Loading' translated
	].
	server := self tryToFindAServerWithMe ifNil: [^ nil].
	pair := self class mostRecent: self name onServer: server.
	pair first ifNil: [^ self inform: ('can''t find file on server for {1}' translated format: {self name})].
	self currentVersionNumber > pair second ifTrue: [
		^ self inform: ('That server has an older version of the project.' translated)].
	version = (Project parseProjectFileName: pair first) second ifTrue: [
		resp := (UIManager default chooseFrom: 
				(Array with: 'Reload anyway' translated 
						with: 'Cancel' translated withCRs) 
				title:  'The only changes are the ones you made here.' translated).
		resp ~= 1 ifTrue: [^ nil]
	] ifFalse: [
		newerAutomatically ifFalse: [
			resp := (UIManager default 
						chooseFrom: #('Load it' 'Cancel') 
						title:  'A newer version exists on the server.').
			resp ~= 1 ifTrue: [^ nil]
		].
	].

	"let's avoid renaming the loaded change set since it will be replacing ours"
	self projectParameters at: #loadingNewerVersion put: true.

	CurrentProject
		do: [ProjectLoading
				installRemoteNamed: pair first
				from: server
				named: self name
				in: parentProject]
		withProgressInfoOn: nil
		label: 'project loading'
! !

!Project methodsFor: 'enter' stamp: 'mt 4/17/2016 12:32'!
loadState	

	self installProjectPreferences.
	ChangeSet  newChanges: changeSet.
	thumbnail ifNotNil: [:form | form unhibernate].
	TranscriptStream newTranscript: transcript.
! !

!Project methodsFor: 'language' stamp: 'tak 8/4/2005 16:30'!
localeChanged
	"Set the project's natural language as indicated"
	self projectParameterAt: #localeID put: LocaleID current.
	self updateLocaleDependents! !

!Project methodsFor: 'language' stamp: 'nice 12/27/2009 03:10'!
localeID
	"Answer the natural language for the project"

	
	^ self projectParameterAt: #localeID
		ifAbsentPut: [ | prev |
			(prev := self previousProject)
				ifNotNil: [prev projectParameterAt: #localeID ifAbsent: [LocaleID current]]
				ifNil: [LocaleID current]]! !

!Project methodsFor: 'menu messages' stamp: 'tpr 3/15/2013 12:42'!
makeThumbnail
	"Make a thumbnail image of this project from the Display."
	viewSize ifNil: [viewSize := Display extent // 8].
	thumbnail := Form extent: viewSize depth: Display depth.
	(WarpBlt toForm: thumbnail)
			sourceForm: Display;
			cellSize: 2;  "installs a colormap"
			combinationRule: Form over;
			copyQuad: (Display boundingBox) innerCorners
			toRect: (0@0 extent: viewSize).
	(Smalltalk at: #InternalThreadNavigationMorph) ifNotNil: [:tnMorph |
			tnMorph  cacheThumbnailFor: self].
	^thumbnail
! !

!Project methodsFor: 'accessing' stamp: 'di 7/19/1999 15:06'!
name
	changeSet == nil ifTrue: [^ 'no name'].
	^ changeSet name! !

!Project methodsFor: 'accessing' stamp: 'jla 5/28/2001 20:01'!
nameAdjustedForDepth
	"Answer the name of the project, prepended with spaces reflecting the receiver's depth from the top project"
	"	Project current nameAdjustedForDepth	"

	| stream |
	stream := WriteStream on: String new.
	self depth timesRepeat: 
	  [2 timesRepeat: [stream nextPut: $ ]].
	stream nextPutAll: self name.
	^ stream contents! !

!Project methodsFor: 'language' stamp: 'dgd 10/7/2004 20:51'!
naturalLanguage
	"Answer the natural language for the project"
	^ self localeID displayName! !

!Project methodsFor: 'accessing' stamp: 'sw 2/15/1999 12:12'!
nextProject
	^ nextProject
! !

!Project methodsFor: 'displaying' stamp: 'mt 7/28/2016 15:05'!
noDisplayDuring: block
	"Projects can provide means to bypass screen updates so that graphics code can be updated safely."! !

!Project methodsFor: 'project parameters' stamp: 'sw 2/16/2001 22:35'!
noteThatParameter: prefSymbol justChangedTo: aBoolean
	"Provides a hook so that a user's toggling of a project parameter might precipitate some immediate action"

! !

!Project methodsFor: 'file in/out' stamp: 'RAA 10/26/2000 16:58'!
objectForDataStream: refStrm
	| uu dp |
	"I am about to be written on an object file.  Write a path to me in the other system instead."

	"Use a copy with no parent, previous or next to reduce extra stuff copied"
	refStrm project == self ifTrue: [^ self copy setParent: nil].

	dp := (uu := self url) size > 0 ifTrue: [
		DiskProxy global: #Project selector: #namedUrl: args: {uu}.
	] ifFalse: [
		DiskProxy global: #Project selector: #named: args: {self name}
	].
	refStrm replace: self with: dp.
	^ dp
! !

!Project methodsFor: 'utilities' stamp: 'ar 1/10/2010 10:31'!
offerMenu: menuSelector from: aModel shifted: aBoolean
	"Pop up a menu whose target is aModel and whose contents are provided
	by sending the menuSelector to the model. The menuSelector takes two
	arguments: a menu, and a boolean representing the shift state."

	self subclassResponsibility! !

!Project methodsFor: 'release' stamp: 'mt 5/1/2016 14:03'!
okToChange
	"Answer whether the window in which the project is housed can be dismissed -- which is destructive. We never clobber a project without confirmation"

	| answer |
	(self confirm: ('Do you really want to delete the project\{1}\and all its content?' withCRs translated format:{self name}))
		ifFalse: [^ false].

	self subProjects ifNotEmpty: [:sp |
		answer := Project current uiManager
			chooseFrom: #(
				"1" 'Lift all sub-projects'
				"2" 'Discard all sub-projects (NO UNDO!!)'
				"3 or 0" 'Cancel')
			lines: #(2)
			title: ('The project {1}\contains {2} sub-project(s).' withCRs translated format:{self name. sp size}).
		
		(answer = 0 or: [answer = 3]) ifTrue: [^ false].
		answer = 1 ifTrue: [self liftSubProjects. ^ true].
		answer = 2 ifTrue: [^ sp allSatisfy: [:ea | ea okToChange]]].
	
	^ true! !

!Project methodsFor: 'squeaklet on server' stamp: 'dtl 11/27/2009 14:12'!
openBlankProjectNamed: projName

	"Project current openBlankProjectNamed: 'My new project'"

	self subclassResponsibility
! !

!Project methodsFor: 'editors' stamp: 'dtl 2/10/2010 15:26'!
openImage: aForm name: fullName saveResource: aBoolean
	"Open a view on an image. If aBoolean is true, save the image as a project resource."

	self subclassResponsibility! !

!Project methodsFor: 'initialization' stamp: 'ar 9/18/2009 21:43'!
openProject: aProject
	"Create a new for a new project in the context of the receiver"
	^self subclassResponsibility! !

!Project methodsFor: 'project parameters' stamp: 'ar 5/25/2000 23:23'!
parameterAt: aSymbol
	^self parameterAt: aSymbol ifAbsent:[nil]! !

!Project methodsFor: 'project parameters' stamp: 'ar 5/25/2000 23:23'!
parameterAt: aSymbol ifAbsent: aBlock
	projectParameters ifNil:[^aBlock value].
	^projectParameters at: aSymbol ifAbsent: aBlock! !

!Project methodsFor: 'sub-projects & hierarchy' stamp: 'di 3/6/98 10:22'!
parent
	^ parentProject! !

!Project methodsFor: 'utilities' stamp: 'dtl 12/28/2009 13:08'!
pointerMoved
	"Pointer has moved during a drag operation. Perform any necessary updates."

	^ self! !

!Project methodsFor: 'release' stamp: 'dtl 11/22/2009 17:48'!
prepareForDelete
	"The window in which the project is housed is about to deleted. Perform
	any necessary actions to prepare for deletion. For an MVC project, no
	action is required."
! !

!Project methodsFor: 'accessing' stamp: 'jm 5/20/1998 23:31'!
previousProject
	^ previousProject
! !

!Project methodsFor: 'file in/out' stamp: 'mir 6/7/2001 14:39'!
primaryServer
	"Return my primary server, that is the one I was downloaded from or are about to be stored on."
	^self primaryServerIfNil: [nil]! !

!Project methodsFor: 'file in/out' stamp: 'mir 6/7/2001 14:39'!
primaryServerIfNil: aBlock
	"Return my primary server, that is the one I was downloaded from or are about to be stored on. If none is set execute the exception block"
	| serverList | 
	serverList := self serverList.
	^serverList isEmptyOrNil
		ifTrue: [aBlock value]
		ifFalse: [serverList first]! !

!Project methodsFor: 'printing' stamp: 'dtl 11/22/2009 15:54'!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' (', self name, ') in '.
	world printOn: aStream
! !

!Project methodsFor: 'accessing' stamp: ''!
projectChangeSet
	^ changeSet! !

!Project methodsFor: 'file in/out' stamp: 'ar 2/27/2001 13:44'!
projectExtension
	^self class projectExtension! !

!Project methodsFor: 'project parameters' stamp: 'sw 10/30/2000 11:14'!
projectParameterAt: aSymbol
	"Answer the project parameter stored at the given symbol, or nil if none"

	^ self projectParameters at: aSymbol ifAbsent: [nil]! !

!Project methodsFor: 'project parameters' stamp: 'sw 9/28/2001 08:49'!
projectParameterAt: aKey ifAbsentPut: defaultValueBlock
	"Return the project parameter setting at the given key.  If there is no entry for this key in the Parameters dictionary, create one with the value of defaultValueBlock as its value"

	^ self projectParameters at: aKey ifAbsentPut: defaultValueBlock! !

!Project methodsFor: 'project parameters' stamp: 'ul 8/2/2011 20:07'!
projectParameterAt: aSymbol ifAbsent: aBlock
	"Answer the project parameter stored at the given symbol, or the result of evaluating the block"

	^ self projectParameters at: aSymbol ifAbsent: aBlock! !

!Project methodsFor: 'project parameters' stamp: 'sw 2/17/2001 21:36'!
projectParameterAt: aSymbol put: aValue
	"Set the given project parameter to the given value"

	self projectParameters at: aSymbol put: aValue.
	self noteThatParameter: aSymbol justChangedTo: aValue.
	^ aValue! !

!Project methodsFor: 'project parameters' stamp: 'sw 4/22/1999 15:14'!
projectParameters
	^ projectParameters ifNil: [self initializeProjectParameters]! !

!Project methodsFor: 'project parameters' stamp: 'nice 12/27/2009 03:45'!
projectPreferenceAt: aSymbol
	"Answer the project preference stored at the given symbol.  If there is none in the local preference dictionary, it must be because it was only latterly declared to be a project-local preference, so obtain its initial value instead from the last-known global or default setting"

	
	^ self projectPreferenceAt: aSymbol ifAbsent: 
		[ | aValue |
		aValue := Preferences valueOfFlag: aSymbol.
		self projectPreferenceFlagDictionary at: aSymbol put: aValue.
		^ aValue]! !

!Project methodsFor: 'project parameters' stamp: 'ul 8/2/2011 20:07'!
projectPreferenceAt: aSymbol ifAbsent: aBlock
	"Answer the project preference stored at the given symbol, or the result of evaluating the block"

	^ self projectPreferenceFlagDictionary at: aSymbol ifAbsent: aBlock! !

!Project methodsFor: 'project parameters' stamp: 'sw 2/16/2001 22:23'!
projectPreferenceFlagDictionary
	"Answer the dictionary that holds the project preferences, creating it if necessary"

	^ projectPreferenceFlagDictionary ifNil: [projectPreferenceFlagDictionary := IdentityDictionary new]! !

!Project methodsFor: 'project parameters' stamp: 'tk 10/26/1999 13:55'!
rawParameters
	^ projectParameters! !

!Project methodsFor: 'release' stamp: 'mt 5/1/2016 14:52'!
release

	self delete.

	world == nil ifFalse: [
		world release.
		world := nil].

	^ super release! !

!Project methodsFor: 'shrinking' stamp: 'mt 5/1/2016 18:35'!
removeAllOtherProjects
	"Remove all other projects from the system, and set the current project as the root
	project. This method was originally implemented as SmalltalkImage>>zapAllOtherProjects
	in earlier versions of Squeak."

	"Project current removeAllOtherProjects"

	Project current == self ifFalse: [
		^ self
			addDeferredUIMessage: [self removeAllOtherProjects];
			enter].
	
	self beTopProject.

	Project rebuildAllProjects.  "Does a GC"
	Project allProjects do: [:p | p == Project current ifFalse: [p delete]].
		
	TheWorldMenu allInstancesDo: [:m | 1 to: m class instSize do: [:i | m instVarAt: i put: nil]].
	ChangeSet classPool at: #AllChangeSets put: nil.
	ProjectHistory currentHistory initialize.
	CommandHistory resetAllHistory.
	ChangeSet initialize.

	self assert: Project current == self.! !

!Project methodsFor: 'release' stamp: 'ar 9/27/2005 20:09'!
removeChangeSetIfPossible

	| itsName |

	changeSet ifNil: [^self].
	changeSet isEmpty ifFalse: [^self].
	(changeSet projectsBelongedTo copyWithout: self) isEmpty ifFalse: [^self].
	itsName := changeSet name.
	ChangeSet removeChangeSet: changeSet.
	"Transcript cr; show: 'project change set ', itsName, ' deleted.'"
! !

!Project methodsFor: 'project parameters' stamp: 'ar 6/2/1999 05:29'!
removeParameter: aKey
	projectParameters ifNil:[^self].
	projectParameters removeKey: aKey ifAbsent:[].! !

!Project methodsFor: 'accessing' stamp: ''!
renameTo: newName
	| oldBase |
	newName = self name
		ifFalse: [
			oldBase := self resourceDirectoryName.
			version := nil.
			self resourceManager adjustToRename: self resourceDirectoryName from: oldBase.
			self changeSet name: newName.
			].! !

!Project methodsFor: 'resources' stamp: 'mir 6/21/2001 15:43'!
resourceDirectoryName
	"Project current resourceDirectoryName"
	^String streamContents:[:s|
		s nextPutAll: self name.
		s nextPutAll: FileDirectory dot.
		s nextPutAll: self versionForFileName.
	]
! !

!Project methodsFor: 'resources' stamp: 'ar 2/27/2001 17:02'!
resourceManager
	^resourceManager ifNil:[resourceManager := ResourceManager new]! !

!Project methodsFor: 'resources' stamp: 'ar 2/27/2001 15:49'!
resourceManager: aResourceManager
	resourceManager := aResourceManager! !

!Project methodsFor: 'resources' stamp: 'mir 6/26/2001 17:34'!
resourceUrl
	"compose my base url for resources on the server"
	| firstURL | 
"
	primaryServer := self primaryServerIfNil: [^''].
	firstURL := primaryServer altUrl ifNil: [primaryServer url]."
	firstURL := self downloadUrl.
	firstURL isEmpty
		ifFalse: [firstURL last == $/ ifFalse: [firstURL := firstURL, '/']].
	^ firstURL, self resourceDirectoryName , '/'
! !

!Project methodsFor: 'displaying' stamp: 'mt 4/13/2016 10:32'!
restore
	"Redraw the entire project."
	
	self subclassResponsibility.! !

!Project methodsFor: 'displaying' stamp: 'mt 5/9/2016 10:46'!
restoreDisplay 
	"Force re-initialization of the display. Uses display-changed notification to re-draw everything."

	Display restore.! !

!Project methodsFor: 'file in/out' stamp: 'bf 4/9/2013 15:45'!
restoreReferences

	| refs newPool |
	refs := world valueOfProperty: #References ifAbsent: [nil].
	(refs isMemberOf: OrderedCollection) ifTrue: [
		world removeProperty: #References.
		newPool := world referencePool.
		refs do: [:assoc | newPool add: assoc].
	].
! !

!Project methodsFor: 'file in/out' stamp: 'KR 5/1/2008 01:57'!
revert
	| |
	"Exit this project and do not save it.  Warn user unless in dangerous projectRevertNoAsk mode.  Exit to the parent project.  Do a revert on a clone of the segment, to allow later reverts."

	projectParameters ifNil: [^ self inform: 'nothing to revert to' translated].
	parentProject enter: false revert: true saveForRevert: false.
	"does not return!!"
! !

!Project methodsFor: 'file in/out' stamp: 'RAA 10/13/2000 18:03'!
saveAs
	"Forget where stored before, and store.  Will ask user where."

	self forgetExistingURL.
	self storeOnServer.! !

!Project methodsFor: 'file in/out' stamp: 'yo 7/2/2004 19:50'!
saveForRevert
	"Exit to the parent project.  Do a GC.  Save the project in a segment.  Record the ImageSegment object as the revertToMe in Project parameters"

	self isTopProject ifTrue: [^ self inform: 'Can''t exit the top project' translated].
	parentProject enter: false revert: false saveForRevert: true.
	"does not return!!"

! !

!Project methodsFor: 'menu messages' stamp: 'mt 8/27/2015 10:43'!
saveProjectPreferences
	"Preserve the settings of all preferences presently held individually by projects in the receiver's projectPreferenceFlagDictionary"

	Preferences allPreferences do:
		[:aPreference | 
			aPreference localToProject ifTrue:
				[projectPreferenceFlagDictionary at: aPreference name put: aPreference preferenceValue]]! !

!Project methodsFor: 'enter' stamp: 'mt 4/17/2016 12:32'!
saveState
	"Save the current state in me prior to leaving this project"

	self saveProjectPreferences.
	changeSet := ChangeSet current.
	thumbnail ifNotNil: [:form | form hibernate].
	transcript := Transcript.! !

!Project methodsFor: 'enter' stamp: 'mt 4/17/2016 13:27'!
scheduleProcessForEnter
	"Complete the enter: by launching a new process"

	self subclassResponsibility! !

!Project methodsFor: 'dispatching' stamp: 'dtl 1/30/2010 15:48'!
selectorPrefixForDispatch
	"A string to be preppended to selectors for project specific methods"

	^ self subclassResponsibility! !

!Project methodsFor: 'file in/out' stamp: 'nice 12/27/2009 03:10'!
serverList
	| servers |
	"Take my list of server URLs and return a list of ServerDirectories to write on."

	urlList isEmptyOrNil ifTrue: [^ nil].
	servers := OrderedCollection new.
	urlList do: [:url | | server |
		server := ServerDirectory serverForURL: url.
		server ifNotNil: [servers add: server].
		server := ServerDirectory serverForURL: url asUrl downloadUrl.
		server ifNotNil: [servers add: server]].
	^servers isEmpty
		ifTrue: [nil]
		ifFalse: [servers]! !

!Project methodsFor: 'utilities' stamp: 'dtl 2/27/2010 09:52'!
setAsBackground: aForm
	"Set  aForm as a background image."

	self subclassResponsibility
! !

!Project methodsFor: 'initialization' stamp: 'ar 11/12/2009 01:11'!
setChangeSet: aChangeSet

	changeSet := aChangeSet
! !

!Project methodsFor: 'sub-projects & hierarchy' stamp: 'mt 5/9/2016 14:30'!
setParent: newParent

	"Notify the current parent to discard views and other references. Avoids deleting that project later if the former parent is deleted. You can only have one parent anyway.."
	self parent ~= newParent ifTrue: [self parent deletingProject: self].

	parentProject := newParent.
	nextProject := previousProject := nil.! !

!Project methodsFor: 'initialization' stamp: 'RAA 10/13/2000 18:21'!
setServer
	"Mark me as a new project.  See if a server is known, remember it."

	self projectParameters at: #exportState put: #nacent.
	urlList isEmptyOrNil ifTrue: [urlList := parentProject urlList].! !

!Project methodsFor: 'accessing' stamp: 'RAA 5/10/2001 14:57'!
setThumbnail: aForm

	self flag: #bob.		"no longer used??"
	thumbnail := aForm! !

!Project methodsFor: 'accessing' stamp: 'di 1/21/98 11:06'!
setViewSize: aPoint
	viewSize := aPoint! !

!Project methodsFor: 'utilities' stamp: 'dtl 2/12/2010 20:41'!
showImage: aForm named: imageName
	"Show an image, possibly attached to the pointer for positioning"

	self subclassResponsibility
! !

!Project methodsFor: 'flaps support' stamp: 'sw 4/24/2001 11:02'!
showSharedFlaps
	"Answer whether shared flaps are shown or suppressed in this project"

	| result |
	result := Preferences showSharedFlaps.
	^ self == Project current
		ifTrue:
			[result]
		ifFalse:
			[self projectPreferenceAt: #showSharedFlaps ifAbsent: [result]]! !

!Project methodsFor: 'displaying' stamp: 'sw 1/12/2000 18:44'!
showZoom
	"Decide if user wants a zoom transition, and if there is enough memory to do it."

	^ Preferences projectZoom and:
		"Only show zoom if there is room for both displays plus a megabyte"
		[Smalltalk garbageCollectMost > 
						(Display boundingBox area*displayDepth //8+1000000)]! !

!Project methodsFor: 'active process' stamp: 'dtl 7/6/2013 13:19'!
spawnNewProcessIfThisIsUI: suspendedProcess
	"Initialize a UI process if needed. Answer true if suspendedProcess was interrupted
	from a UI process."
	self subclassResponsibility! !

!Project methodsFor: 'file in/out' stamp: 'RAA 2/19/2001 07:37'!
squeakletDirectory

	^self class squeakletDirectory! !

!Project methodsFor: 'resources' stamp: 'mir 6/18/2001 16:19'!
startResourceLoading
	"Abort loading resources"
	resourceManager ifNil:[^self].
	resourceManager adjustToDownloadUrl: self resourceUrl.
	resourceManager startDownload! !

!Project methodsFor: 'file in/out' stamp: 'yo 12/14/2004 18:18'!
storeAttributeKey: key value: value on: aStream

	(key includes: $:) ifTrue: [self error: 'key should not contain :'].
	aStream nextPutAll: key.
	aStream nextPutAll: ': '.
	aStream nextPutAll: value.
	aStream cr.
! !

!Project methodsFor: 'file in/out' stamp: 'KR 12/16/2005 11:07'!
storeAttributesOn: aStream

	| details |
	self storeAttributeKey: 'Squeak-Version' value: SystemVersion current version on: aStream.
	self storeAttributeKey: 'Squeak-LatestUpdate' value: SystemVersion current highestUpdate printString on: aStream.
	self storeAttributeKey: 'File-Name-Encoding' value: LanguageEnvironment defaultFileNameConverter class encodingNames first on: aStream.

	details := self world valueOfProperty: #ProjectDetails ifAbsent: [Dictionary new].
	details associationsDo: [:assoc |
		self storeAttributeKey: assoc key asString value: assoc value asString on: aStream.].
! !

!Project methodsFor: 'file in/out' stamp: 'tk 8/21/1999 07:31'!
storeDataOn: aDataStream
	"Write me out.  All references to other projects are weak references.  They only go out if they are written for another reason."
	| cntInstVars cntIndexedVars localInstVars offset |

	cntInstVars := self class instSize.
	cntIndexedVars := self basicSize.
	localInstVars := Project instVarNames.
	offset := Project superclass instSize.
	aDataStream
		beginInstance: self class
		size: cntInstVars + cntIndexedVars.
	1 to: cntInstVars do:
		[:ii | 
		(ii between: offset+1 and: offset + localInstVars size) 
			ifTrue: [(#('parentProject' 'previousProject' 'nextProject') includes: 
				(localInstVars at: ii-offset)) 
					ifTrue: [aDataStream nextPutWeak: (self instVarAt: ii)]
								"owner only written if in our tree"
					ifFalse: [aDataStream nextPut: (self instVarAt: ii)]]
			ifFalse: [aDataStream nextPut: (self instVarAt: ii)]].

	1 to: cntIndexedVars do:
		[:i | aDataStream nextPut: (self basicAt: i)]! !

!Project methodsFor: 'file in/out' stamp: 'yo 12/14/2004 19:01'!
storeHtmlPageIn: aFileDirectory
	"Prepare the HTML wrapper for the current project"
	| file page |
	file := aFileDirectory forceNewFileNamed: (self name, FileDirectory dot,'html').
	file ifNil: [^self].
	file converter: UTF8TextConverter new.
	page := self htmlPagePrototype.
	page := page copyReplaceAll: '$$PROJECT$$' with: self versionedFileName.
	page := page copyReplaceAll: '$$WIDTH$$' with: world bounds width printString.
	page := page copyReplaceAll: '$$HEIGHT$$' with: world bounds height printString.
	page := page copyReplaceAll: String cr with: String lf. "not sure if necessary..."
	file nextPutAll: page.
	file close.! !

!Project methodsFor: 'file in/out' stamp: 'KR 12/16/2005 10:07'!
storeManifestFileIn: aFileDirectory

	| file |
	file := aFileDirectory forceNewFileNamed: (self name, FileDirectory dot,'manifest').
	file ifNil: [^ self].
	file converter: UTF8TextConverter new.
	self storeAttributesOn: file.
	file close.
! !

!Project methodsFor: 'accessing' stamp: 'mir 6/26/2001 17:09'!
storeNewPrimaryURL: aURLString
	| oldResourceUrl |
	oldResourceUrl := self resourceUrl.
	urlList isEmptyOrNil ifTrue: [urlList := Array new: 1].
	urlList at: 1 put: aURLString.
	self lastDirectory: nil.
	self resourceManager adjustToNewServer: self resourceUrl from: oldResourceUrl
! !

!Project methodsFor: 'file in/out' stamp: 'tfel 8/4/2016 14:53'!
storeOnServer

	"Save to disk as an Export Segment.  Then put that file on the server I came from, as a new version.  Version is literal piece of file name.  Mime encoded and http encoded."

	world setProperty: #optimumExtentFromAuthor toValue: world extent.
	self validateProjectNameIfOK: [:details |
		self acceptProjectDetails: details.
		self isCurrentProject ifTrue: ["exit, then do the command"
			^ self 
				armsLengthCommand: #storeOnServerAssumingNameValid
				withDescription: 'Publishing' translated
		].
		self storeOnServerWithProgressInfo.
	].! !

!Project methodsFor: 'file in/out' stamp: 'yo 9/4/2007 14:17'!
storeOnServerAssumingNameValid

	"Save to disk as an Export Segment.  Then put that file on the server I came from, as a new version.  Version is literal piece of file name.  Mime encoded and http encoded."
	world setProperty: #optimumExtentFromAuthor toValue: world extent.
	self isCurrentProject ifTrue: ["exit, then do the command"
		Flaps globalFlapTabsIfAny do: [:each | Flaps removeFlapTab: each keepInList: true].
		^ self 
			armsLengthCommand: #storeOnServerAssumingNameValid
			withDescription: 'Publishing' translated
	].
	self storeOnServerWithProgressInfo.
! !

!Project methodsFor: 'file in/out' stamp: 'rbb 3/1/2005 11:08'!
storeOnServerInnards
	"Save to disk as an Export Segment.  Then put that file on the server I came from, as a new version.  Version is literal piece of file name.  Mime encoded and http encoded."

	| resp newName primaryServerDirectory serverVersionPair localDirectory localVersionPair myVersionNumber warning maxNumber suppliedPassword oldResourceUrl |
	self assureIntegerVersion.

	"Find out what version"
	primaryServerDirectory := self primaryServerIfNil: [
		(primaryServerDirectory := self findAFolderToStoreProjectIn) ifNil: [^self].
		oldResourceUrl := self resourceUrl.
		primaryServerDirectory == #localOnly ifTrue: [
			self storeNewPrimaryURL: FileDirectory default url.
			nil
		] ifFalse: [
			self storeNewPrimaryURL: primaryServerDirectory downloadUrl.
			primaryServerDirectory
		].
	].

	localDirectory := self squeakletDirectory.
	serverVersionPair := self class mostRecent: self name onServer: primaryServerDirectory.
	localVersionPair := self class mostRecent: self name onServer: localDirectory.
	maxNumber := myVersionNumber := self currentVersionNumber.

	ProgressNotification signal: '2:versionsDetected'.

	warning := ''.
	myVersionNumber < serverVersionPair second ifTrue: [
		warning := warning,'\There are newer version(s) on the server' translated.
		maxNumber := maxNumber max: serverVersionPair second.
	].
	myVersionNumber < localVersionPair second ifTrue: [
		warning := warning,'\There are newer version(s) in the local directory' translated.
		maxNumber := maxNumber max: localVersionPair second.
	].
	"8 Nov 2000 - only check on the first attempt to publish"
	myVersionNumber = 0 ifTrue: [
		warning isEmpty ifFalse: [
			myVersionNumber = 0 ifTrue: [
				warning := warning,'\THIS PROJECT HAS NEVER BEEN SAVED' translated.
			].
			warning := 'WARNING' translated, '\Project: ' translated, self name,warning.
			resp := (UIManager default 
					chooseFrom: (Array with: 'Store anyway' translated 
										with: 'Cancel' translated)
  					title: (warning, '\Please cancel, rename this project, and see what is there.' translated) withCRs).
				resp ~= 1 ifTrue: [^ nil]
		].
	].
	version := self bumpVersion: maxNumber.

	oldResourceUrl
		ifNotNil: [self resourceManager adjustToNewServer: self resourceUrl from: oldResourceUrl].

	"write locally - now zipped automatically"
	newName := self versionedFileName.
	lastSavedAtSeconds := Time totalSeconds.
	self exportSegmentFileName: newName directory: localDirectory.
	(localDirectory readOnlyFileNamed: newName) setFileTypeToObject; close.

	ProgressNotification signal: '4:localSaveComplete'.	"3 is deep in export logic"

	primaryServerDirectory ifNotNil: [
		suppliedPassword := ''.
		Preferences passwordsOnPublish ifTrue: [
			suppliedPassword := UIManager default requestPassword: 'Project password' translated
		].
		[
		primaryServerDirectory
			writeProject: self
			inFileNamed: newName asFileName
			fromDirectory: localDirectory.
		] on: ProjectPasswordNotification do: [ :ex |
			ex resume: (suppliedPassword ifNil: [''])
		].
	].
	ProgressNotification signal: '9999 save complete'.

	"Later, store with same name on secondary servers.  Still can be race conditions.  All machines will go through the server list in the same order."
	"2 to: servers size do: [:aServer | aServer putFile: local named: newName]."
! !

!Project methodsFor: 'file in/out' stamp: 'tfel 8/4/2016 14:55'!
storeOnServerShowProgressOn: aMorphOrNil forgetURL: forget

	"Save to disk as an Export Segment.  Then put that file on the server I came from, as a new version.  Version is literal piece of file name.  Mime encoded and http encoded."

	world setProperty: #optimumExtentFromAuthor toValue: world extent.
	self validateProjectNameIfOK: [:details |
		self acceptProjectDetails: details.
		self isCurrentProject ifTrue: ["exit, then do the command"
			forget
				ifTrue: [self forgetExistingURL]
				ifFalse: [urlList isEmptyOrNil ifTrue: [urlList := parentProject urlList copy]].
			^self
				armsLengthCommand: #storeOnServerAssumingNameValid
				withDescription: 'Publishing' translated
		].
		self storeOnServerWithProgressInfoOn: aMorphOrNil.
	].
! !

!Project methodsFor: 'file in/out' stamp: 'dtl 11/30/2009 22:40'!
storeOnServerWithProgressInfo

	"Save to disk as an Export Segment.  Then put that file on the server I came from, as a new version.  Version is literal piece of file name.  Mime encoded and http encoded."

	CurrentProject
		do: [self storeOnServerInnards]
		withProgressInfoOn: nil
		label: 'project storing'
! !

!Project methodsFor: 'file in/out' stamp: 'dtl 11/30/2009 22:40'!
storeOnServerWithProgressInfoOn: aMorphOrNil

	"Save to disk as an Export Segment.  Then put that file on the server I came from, as a new version.  Version is literal piece of file name.  Mime encoded and http encoded."

	CurrentProject
		do: [self storeOnServerInnards]
		withProgressInfoOn: aMorphOrNil
		label: 'project storing'

	! !

!Project methodsFor: 'resources' stamp: 'ar 3/2/2001 15:16'!
storeResourceList: collector in: fd
	"Store a list of all used resources in the given directory. Used for maintenance."
	| file rcName |
	rcName := self resourceDirectoryName,'.rc'.
	file := fd forceNewFileNamed: rcName.
	collector locatorsDo:[:loc| file nextPutAll: loc urlString; cr].
	file close.
	file := fd readOnlyFileNamed: rcName.
	file compressFile.
	fd deleteFileNamed: rcName ifAbsent:[].! !

!Project methodsFor: 'file in/out' stamp: 'dtl 11/29/2009 13:15'!
storeSegment
	"Store my project out on the disk as an ImageSegment.  Keep the outPointers in memory.  Name it <project name>.seg.  *** Caller must be holding (Project alInstances) to keep subprojects from going out. ***"

	self subclassResponsibility
! !

!Project methodsFor: 'file in/out' stamp: 'dtl 11/29/2009 13:08'!
storeSegmentNoFile
	"For testing.  Make an ImageSegment.  Keep the outPointers in memory.  Also useful if you want to enumerate the objects in the segment afterwards (allObjectsDo:)"

	self subclassResponsibility
! !

!Project methodsFor: 'file in/out' stamp: 'nk 7/30/2004 17:52'!
storeSomeSegment
	"Try all projects to see if any is ready to go out.  Send at most three of them.
	Previous one has to wait for a garbage collection before it can go out."

	| cnt pList start proj gain |
	cnt := 0.
	gain := 0.
	pList := Project allProjects.
	start := pList size atRandom.	"start in a random place"
	start to: pList size + start
		do: 
			[:ii | 
			proj := pList atWrap: ii.
			proj storeSegment 
				ifTrue: 
					["Yes, did send its morphs to the disk"

					gain := gain + (proj projectParameters at: #segmentSize ifAbsent: [0]).	"a guess"
					Beeper beep.
					(cnt := cnt + 1) >= 2 ifTrue: [^gain]]].
	Beeper  beep.
	^gain! !

!Project methodsFor: 'file in/out' stamp: 'fbs 7/25/2013 07:16'!
storeToMakeRoom
	"Write out enough projects to fulfill the space goals.
	Include the size of the project about to come in."

	| params memoryEnd goalFree cnt gain proj skip tried |
	GoalFreePercent ifNil: [GoalFreePercent := 33].
	GoalNotMoreThan ifNil: [GoalNotMoreThan := 20000000].
	params := Smalltalk  getVMParameters.
	memoryEnd	:= params at: 3.
"	youngSpaceEnd	:= params at: 2.
	free := memoryEnd - youngSpaceEnd.
"
	goalFree := GoalFreePercent asFloat / 100.0 * memoryEnd.
	goalFree := goalFree min: GoalNotMoreThan.
	world isInMemory ifFalse: ["enough room to bring it in"
		goalFree := goalFree + (self projectParameters at: #segmentSize ifAbsent: [0])].
	cnt := 30.
	gain := Smalltalk garbageCollectMost.
	"skip a random number of projects that are in memory"
	proj := self.  skip := 6 atRandom.
	[proj := proj nextInstance ifNil: [Project someInstance].
		proj world isInMemory ifTrue: [skip := skip - 1].
		skip > 0] whileTrue.
	cnt := 0.  tried := 0.

	[gain > goalFree] whileFalse: [
		proj := proj nextInstance ifNil: [Project someInstance].
		proj storeSegment ifTrue: ["Yes, did send its morphs to the disk"
			gain := gain + (proj projectParameters at: #segmentSize 
						ifAbsent: [20000]).	"a guess"
			Beeper beep.
			(cnt := cnt + 1) > 5 ifTrue: [^ self]].	"put out 5 at most"
		(tried := tried + 1) > 23 ifTrue: [^ self]].	"don't get stuck in a loop"! !

!Project methodsFor: 'sub-projects & hierarchy' stamp: 'mt 4/14/2016 21:12'!
subProjects
	"Answer a list of all the subprojects of the receiver. By default, use the list of all projects and traverse the parent."
	
	^ Project allProjects select: [:p | p parent == self and: [p isTopProject not]]! !

!Project methodsFor: 'enter - recovery' stamp: 'mt 4/17/2016 21:55'!
suspendProcessForDebug
	"If this project fails, let another project debug this project's main loop/process. Default behavior is terminating the process. Hence, no debugging."
	
	self terminateProcessForLeave.! !

!Project methodsFor: 'SuperSwiki' stamp: 'RAA 7/13/2000 11:11'!
tellAFriend

	self tellAFriend: nil
! !

!Project methodsFor: 'SuperSwiki' stamp: 'yo 2/12/2005 19:25'!
tellAFriend: emailAddressOrNil
	| urlForLoading |
"
Project current tellAFriend
"

	(urlForLoading := self urlForLoading) ifNil: [
		urlForLoading := self url		"fallback for dtp servers"
	].
	urlForLoading isEmptyOrNil ifTrue: [
		^self inform: 'Since this project has not been saved yet,
I cannot tell someone where it is.' translated
	].
	HTTPClient tellAFriend: emailAddressOrNil url: urlForLoading name: self name! !

!Project methodsFor: 'enter' stamp: 'mt 4/12/2016 18:11'!
terminateProcessForLeave
	"Clean-up processes that this project uses."

	self subclassResponsibility.! !

!Project methodsFor: 'utilities' stamp: 'dtl 2/14/2010 20:44'!
textWindows
	"Answer a dictionary of all system windows for text display keyed by window title.
	Generate new window titles as required to ensure unique keys in the dictionary."

	self subclassResponsibility! !

!Project methodsFor: 'accessing' stamp: 'di 1/21/98 11:07'!
thumbnail
	^ thumbnail! !

!Project methodsFor: 'menu messages' stamp: 'cbr 5/1/2010 01:53'!
toggleFullScreen
	"Toggle between full screen and windowed mode."
	self lastScreenModeSelected
		ifTrue: [ self fullScreenOff ]
		ifFalse: [ self fullScreenOn ]! !

!Project methodsFor: 'file in/out' stamp: 'rbb 2/16/2005 17:15'!
tryToFindAServerWithMe

	| resp primaryServerDirectory |

	urlList isEmptyOrNil ifTrue: [urlList := parentProject urlList copy].
	[self primaryServer isNil] whileTrue: [
		resp := (UIManager default 
					chooseFrom: #('Try to find a server' 'Cancel')
					title: 'This project thinks it has never been on a server').
		resp ~= 1 ifTrue: [^ nil].
		(primaryServerDirectory := self findAFolderToLoadProjectFrom) ifNil: [^nil].
		self storeNewPrimaryURL: primaryServerDirectory downloadUrl.
	].
	^self primaryServer
! !

!Project methodsFor: 'accessing' stamp: 'dtl 5/2/2010 19:09'!
uiManager
	"Answer the manager that provides user interface services for this project "
	^ uiManager
		ifNil: [uiManager := Smalltalk
						at: #UIManager
						ifPresent: [:mgr | uiManager := mgr getDefault]]! !

!Project methodsFor: 'active process' stamp: 'mt 4/12/2016 19:19'!
uiProcess
	"Answer the most important process of this project."

	self subclassResponsibility.! !

!Project methodsFor: 'language' stamp: 'mt 2/18/2016 14:54'!
updateLocaleDependents
	"Set the project's natural language as indicated"

	LanguageEnvironment localeChanged.! !

!Project methodsFor: 'file in/out' stamp: 'mir 6/25/2001 10:55'!
url
	| firstURL |
	"compose my url on the server"

	urlList isEmptyOrNil ifTrue: [^''].
	firstURL := urlList first.
	firstURL isEmpty
		ifFalse: [
			firstURL last == $/
				ifFalse: [firstURL := firstURL, '/']].
	^ firstURL, self versionedFileName
! !

!Project methodsFor: 'file in/out' stamp: 'mir 8/8/2001 17:58'!
urlForLoading
	"compose a url that will load me in someone's browser"
	| myServer serverList myUrl downloadUrl |
	serverList := self serverList.
	serverList isEmptyOrNil
		ifTrue: [
			urlList isEmptyOrNil ifTrue: [^nil].
			downloadUrl := urlList first asUrl downloadUrl]
		ifFalse: [
			myServer := serverList first.
			myUrl := myServer altUrl.
			myUrl last == $/
				ifFalse: [myUrl := myUrl , '/'].
			downloadUrl := myUrl].
	^downloadUrl , (self name, FileDirectory dot,'html') encodeForHTTP
! !

!Project methodsFor: 'accessing' stamp: 'tk 4/5/2000 13:57'!
urlList
	^ urlList! !

!Project methodsFor: 'menu messages' stamp: 'tfel 8/4/2016 14:54'!
validateProjectNameIfOK: aBlock

	| details |

	details := world valueOfProperty: #ProjectDetails.
	details ifNotNil: ["ensure project info matches real project name"
		details at: 'projectname' put: self name.
	].
	self doWeWantToRename ifFalse: [^ aBlock value: details].
	(Smalltalk at: #EToyProjectDetailsMorph) ifNotNil: [:etpdm |
		etpdm
			getFullInfoFor: self 
			ifValid: [:d |
				World displayWorldSafely.
				aBlock value: d
			]
			expandedFormat: false]
! !

!Project methodsFor: 'file in/out' stamp: 'mir 6/21/2001 15:43'!
versionedFileName
	"Project current versionedFileName"
	^String streamContents:[:s|
		s nextPutAll: self name.
		s nextPutAll: FileDirectory dot.
		s nextPutAll: self versionForFileName.
		s nextPutAll: FileDirectory dot.
		s nextPutAll: self projectExtension.
	]
! !

!Project methodsFor: 'file in/out' stamp: 'mir 6/21/2001 15:45'!
versionForFileName
	"Project current versionForFileName"
	^self class versionForFileName: self currentVersionNumber! !

!Project methodsFor: 'file in/out' stamp: 'RAA 10/15/2000 19:10'!
versionFrom: aServerFile
	"Store the version of the file I actually came from.  My stored version was recorded before I knew the latest version number on the server!!"
	| theName serverUrl |

	self flag: #bob.		"this may become unnecessary once we get the version before writing"
	self flag: #bob.		"need to recognize swiki servers"

	serverUrl := aServerFile directoryUrl.
	theName := aServerFile localName.
	version := (Project parseProjectFileName: theName) second.
	(serverUrl beginsWith: 'ftp:') ifTrue: ["update our server location"
		self storeNewPrimaryURL: serverUrl
	].
! !

!Project methodsFor: 'displaying' stamp: 'mt 2/18/2016 15:47'!
viewLocFor: exitedProject 
	"Look for a view of the exitedProject, and return its center"

	^ Display center! !

!Project methodsFor: 'accessing' stamp: 'di 1/21/98 11:06'!
viewSize
	^ viewSize! !

!Project methodsFor: 'enter' stamp: 'dtl 12/28/2009 22:55'!
wakeUpTopWindow
	"Image has been restarted, and the startUp list has been processed. Perform
	any additional actions needed to restart the user interface."

	^ self! !

!Project methodsFor: 'initialization' stamp: 'di 6/10/1998 13:54'!
windowActiveOnFirstClick

	^ true! !

!Project methodsFor: 'release' stamp: 'mt 5/1/2016 14:51'!
windowIsClosing

	self prepareForDelete.! !

!Project methodsFor: 'initialization' stamp: 'KLC 12/19/2005 14:32'!
windowReqNewLabel: newLabel
	newLabel isEmpty ifTrue: [^ false].
	newLabel = changeSet name ifTrue: [^ true].
	(ChangeSet named: newLabel) == nil
		ifFalse: [self inform: 'Sorry that name is already used'.
				^ false].
	changeSet name: newLabel.
	^ true! !

!Project methodsFor: 'sub-projects & hierarchy' stamp: 'ul 8/2/2011 20:08'!
withChildrenDo: aBlock
	"Evaluate the block first with the receiver as argument, then, recursively and depth first, with each of the receiver's children as argument"
	
	aBlock value: self.
	self children do: [:p | 
		p withChildrenDo: aBlock ]! !

!Project methodsFor: 'accessing' stamp: ''!
world
	^ world! !

!Project methodsFor: 'file in/out' stamp: 'nice 1/18/2010 19:07'!
writeFileNamed: localFileName fromDirectory: localDirectory toServer: primaryServerDirectory

	| local resp gifFileName f |

	local := localDirectory oldFileNamed: localFileName.
	resp := primaryServerDirectory upLoadProject: local named: localFileName resourceUrl: self resourceUrl retry: false.
	local close.
	resp == true ifFalse: [
		"abandon resources that would've been stored with the project"
		self resourceManager abandonResourcesThat:
			[:loc| loc urlString beginsWith: self resourceUrl].
		self error: 'the primary server of this project seems to be down (',
							resp printString,')'. 
		^ self
	].

	gifFileName := self name,'.gif'.
	localDirectory deleteFileNamed: gifFileName ifAbsent: [].
	local := localDirectory fileNamed: gifFileName.
	thumbnail ifNil: [
		(thumbnail := Form extent: 100@80) fillColor: Color orange
	] ifNotNil: [
		thumbnail unhibernate.
	].
	f := thumbnail colorReduced.  "minimize depth"
	f depth > 8 ifTrue: [
		f := thumbnail asFormOfDepth: 8
	].
	GIFReadWriter putForm: f onStream: local.
	local close.

	[local := StandardFileStream readOnlyFileNamed: (localDirectory fullNameFor: gifFileName).
	(primaryServerDirectory isKindOf: FileDirectory)
		ifTrue: [primaryServerDirectory deleteFileNamed: gifFileName ifAbsent: []].
	primaryServerDirectory putFile: local named: gifFileName retry: false.
	] on: Error do: [:ex |].
	local close.

	primaryServerDirectory updateProjectInfoFor: self.
	primaryServerDirectory sleep.	"if ftp, close the connection"
! !

!Project methodsFor: 'file in/out' stamp: 'nice 12/27/2009 03:10'!
writeStackText: stacks in: resourceDirectory registerIn: aCollector
	"The user's text is very valuable.  Write an extra file with just the text.  It can be read in case the Project can't be opened." 
	"Find allText for each stack, storeOn a local file in the resources folder, with a name like myProj.005.myStack.t.  Make the names be unique."

	"get project name and version"
	| resourceURL ind |
	resourceURL := self resourceUrl.
	stacks do: [:stackObj | | sn localName char trial textLoc fs |	"Construct a good file name"
		localName := self versionedFileName allButLast: 2.	"projectName.005."
		stacks size = 1 ifFalse: ["must distinguish between stacks in the project"
			(sn := stackObj knownName) ifNil: [
				sn := stackObj hash printString].	"easy name, or use hash"
			localName := localName , sn, FileDirectory dot]. 	"projectName.005.myStack."
		localName := localName , 't'.
		"See if in use because truncates same as another, fix last char, try again"
		[trial := resourceDirectory checkName: localName fixErrors: true.
		 trial endsWith: '.t'] whileFalse: [
				localName := (localName allButLast: 3) , FileDirectory dot, 't'].
		[resourceDirectory fileExists: trial] whileTrue: [
			char := trial at: (ind := trial size - 3).
			trial at: ind put: (char asciiValue + 1) asCharacter].	"twiddle it a little"
		
		"write allText in file"
		fs := resourceDirectory newFileNamed: trial.
		fs timeStamp; cr; nextPutAll: '''This is the text for a stack in this project.  Use only in an emergency, if the project file is ever unreadable.''.'; cr; cr.
		stackObj getAllText storeOn: fs.    fs close.
		textLoc := (ResourceLocator new) localFileName: trial; 
			urlString: resourceURL, '/', trial.
		aCollector locatorMap at: trial "any distinct object" put: textLoc.
		].! !

Model subclass: #ProjectHistory
	instanceVariableNames: 'mostRecent'
	classVariableNames: 'ChangeCounter CurrentHistory'
	poolDictionaries: ''
	category: 'System-Support'!

!ProjectHistory class methodsFor: 'accessing' stamp: 'RAA 7/10/2000 15:56'!
changeCounter

	^ChangeCounter ifNil: [ChangeCounter := 0].! !

!ProjectHistory class methodsFor: 'class initialization' stamp: 'ar 2/26/2010 23:05'!
cleanUp
	"Re-initialize current history"

	self currentHistory initialize! !

!ProjectHistory class methodsFor: 'accessing' stamp: 'nk 7/30/2004 21:51'!
currentHistory
	^CurrentHistory ifNil: [CurrentHistory := self new]! !

!ProjectHistory class methodsFor: 'project lifetime' stamp: 'RAA 7/12/2000 10:16'!
forget: aProject

	self currentHistory forget: aProject
! !

!ProjectHistory class methodsFor: 'project lifetime' stamp: 'RAA 7/10/2000 15:31'!
remember: aProject

	self currentHistory remember: aProject
! !

!ProjectHistory methodsFor: 'updating' stamp: 'RAA 7/10/2000 15:57'!
changed

	ChangeCounter := (ChangeCounter ifNil: [0]) + 1.
	super changed.! !

!ProjectHistory methodsFor: 'project lifetime' stamp: 'nice 12/26/2009 20:35'!
cleanUp
	mostRecent := mostRecent reject: [ :each |
		| proj |
		proj := each fourth first.
		proj isNil or: [proj world isNil]
	].
	self changed.
! !

!ProjectHistory methodsFor: 'project lifetime' stamp: 'RAA 9/27/2000 20:27'!
forget: aProject

	| newTuple |
	newTuple := {
		aProject name.
		aProject thumbnail.
		aProject url.
		WeakArray with: aProject.
	}.
	mostRecent := mostRecent reject: [ :each |
		each fourth first == aProject or: [
			each fourth first isNil & (each first = newTuple first)
		].
	].
	self changed.
	^newTuple! !

!ProjectHistory methodsFor: 'initialize-release' stamp: 'RAA 7/10/2000 15:27'!
initialize

	mostRecent := OrderedCollection new.
! !

!ProjectHistory methodsFor: 'accessing' stamp: 'RAA 9/27/2000 20:29'!
mostRecentCopy

	self cleanUp.
	^mostRecent copy! !

!ProjectHistory methodsFor: 'accessing' stamp: 'RAA 11/10/2000 10:11'!
mostRecentNames

	self cleanUp.
	^mostRecent collect: [ :each |
		each first
	].
! !

!ProjectHistory methodsFor: 'accessing' stamp: 'rbb 3/1/2005 11:10'!
mostRecentThread

	| projectNames threadName |
	self cleanUp.
	projectNames := (mostRecent collect: [ :each | {each first} ]) reversed.

	threadName := UIManager default 
		request: 'Please name this thread.' 
		initialAnswer: 'Recent projects @ ',Time now printString.
	threadName isEmptyOrNil ifTrue: [^nil].
	"rbb 3/1/2005 - Should be refactored to eliminate Morphic dependency"
	InternalThreadNavigationMorph know: projectNames as: threadName.
	^threadName
! !

!ProjectHistory methodsFor: 'project lifetime' stamp: 'RAA 7/12/2000 10:17'!
remember: aProject

	| newTuple |

	newTuple := self forget: aProject.
	mostRecent addFirst: newTuple.
	mostRecent size > 10 ifTrue: [mostRecent := mostRecent copyFrom: 1 to: 10].
	self changed! !

Model subclass: #SARInstaller
	instanceVariableNames: 'zip directory fileName installed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SARInstaller commentStamp: 'nk 7/5/2003 21:12' prior: 0!
I am an object that handles the loading of SAR (Squeak ARchive) files.

A SAR file is a Zip file that follows certain simple conventions:

* it may have a member named "install/preamble".

This member, if present, will be filed in as Smalltalk source code at the beginning of installation.
Typically, the code in the preamble will make whatever installation preparations are necessary,
and will then call methods in the "client services" method category to extract or install other zip members.

* It may have a member named "install/postscript".

This member, if present, will be filed in as Smalltalk source code at the end of installation.
Typically, the code in the postscript will set up the operating environment,
and will perhaps put objects in flaps, open projects or README files, or launch samples.

Within the code in the preamble and postscript, "self" is set to the instance of the SARInstaller.

If neither an "install/preamble" nor an "install/postscript" file is present,
all the members will be installed after prompting the user,
based on a best guess of the member file types that is based on member filename extensions.

This is new behavior.!

!SARInstaller class methodsFor: 'change set utilities' stamp: 'ar 3/5/2010 20:33'!
basicNewChangeSet: newName
	Smalltalk at: #ChangesOrganizer ifPresent: [ :cs | ^cs basicNewChangeSet: newName ].
	(self changeSetNamed: newName) ifNotNil: [ self inform: 'Sorry that name is already used'. ^nil ].
	^ChangeSet basicNewNamed: newName.! !

!SARInstaller class methodsFor: 'SqueakMap' stamp: 'ul 12/12/2009 14:07'!
cardForSqueakMap: aSqueakMap
	"Answer the current card or a new card."

	(aSqueakMap cardWithId: self squeakMapPackageID)
		ifNotNil: [ :card |
			(card installedVersion = self squeakMapPackageVersion) ifTrue: [ ^card ]
		].

	^self newCardForSqueakMap: aSqueakMap
! !

!SARInstaller class methodsFor: 'change set utilities' stamp: 'ar 3/5/2010 20:33'!
changeSetNamed: newName
	Smalltalk at: #ChangesOrganizer ifPresent: [ :cs | ^cs changeSetNamed: newName ].
	^ChangeSet allInstances detect: [ :cs | cs name = newName ] ifNone: [ nil ].! !

!SARInstaller class methodsFor: 'change set utilities' stamp: 'nk 7/5/2003 22:49'!
currentChangeSet
	"Answer the current change set, in a way that should work in 3.5 as well"

	"SARInstaller currentChangeSet"

	^[ ChangeSet current ]
		on: MessageNotUnderstood
		do: [ :ex | ex return: Smalltalk changes ]! !

!SARInstaller class methodsFor: 'instance creation' stamp: 'nk 10/27/2002 10:29'!
directory: dir fileName: fn
	^(self new) directory: dir; fileName: fn; yourself.! !

!SARInstaller class methodsFor: 'package format support' stamp: 'nice 12/27/2009 03:18'!
ensurePackageWithId: anIdString

	self squeakMapDo: [ :sm | | newCS |
		self withCurrentChangeSetNamed: 'updates' do: [ :cs |
			| card |
			newCS := cs.
			card := sm cardWithId: anIdString.
			(card isNil or: [ card isInstalled not or: [ card isOld ]])
				ifTrue: [ sm installPackageWithId: anIdString ]
		].
		newCS isEmpty ifTrue: [ ChangeSet removeChangeSet: newCS ]
	].! !

!SARInstaller class methodsFor: 'change set utilities' stamp: 'yo 8/17/2004 10:04'!
fileIntoChangeSetNamed: aString fromStream: stream 
	"We let the user confirm filing into an existing ChangeSet
	or specify another ChangeSet name if
	the name derived from the filename already exists.
	Duplicated from SMSimpleInstaller.
	Should be a class-side method."

	^self withCurrentChangeSetNamed: aString
		do: [ :cs | | newName |
			newName := cs name.
			stream setConverterForCode.
			stream 
				fileInAnnouncing: 'Loading ' , newName , ' into change set ''' , newName, ''''.
			stream close]! !

!SARInstaller class methodsFor: 'class initialization' stamp: 'nk 11/13/2002 07:33'!
fileReaderServicesForFile: fullName suffix: suffix 

	^(suffix = 'sar') | (suffix = '*') 
		ifTrue: [Array with: self serviceFileInSAR]
		ifFalse: [#()]
! !

!SARInstaller class methodsFor: 'class initialization' stamp: 'fbs 8/9/2013 17:32'!
initialize
	"SARInstaller initialize"
	FileServices registerFileReader: self.! !

!SARInstaller class methodsFor: 'class initialization' stamp: 'nk 7/5/2003 21:05'!
installSAR: relativeOrFullName
	FileDirectory splitName: (FileDirectory default fullNameFor: relativeOrFullName)
		to: [ :dir :fileName | (self directory: (FileDirectory on: dir) fileName: fileName) fileIn ]! !

!SARInstaller class methodsFor: 'package format support' stamp: 'nk 7/25/2003 14:05'!
loadDVS
	"Load the DVS support from SqueakMap"

	self ensurePackageWithId: '100d59d0-bf81-4e74-a4fe-5a2fd0c6b4ec'! !

!SARInstaller class methodsFor: 'package format support' stamp: 'nk 9/9/2003 12:08'!
loadMonticello
	"Load Monticello support (MCInstaller and Monticello) from SqueakMap"

	self ensurePackageWithId: 'af9d090d-2896-4a4e-82d0-c61cf2fdf40e'.
	self ensurePackageWithId: '66236497-7026-45f5-bcf6-ad00ba7a8a4e'.! !

!SARInstaller class methodsFor: 'package format support' stamp: 'nk 7/25/2003 14:39'!
loadMonticelloCVS
	"Load MonticelloCVS support from SqueakMap"

	self ensurePackageWithId: '2be9f7e2-1de2-4eb6-89bd-ec9b60593a93'.
! !

!SARInstaller class methodsFor: 'SqueakMap' stamp: 'nk 7/21/2003 17:17'!
newCardForSqueakMap: aSqueakMap
	"Answer a new card."

	^(aSqueakMap newCardWithId: self squeakMapPackageID)
	created: 3236292323
	updated:3236292323
	name: 'SARInstaller for 3.6'
	currentVersion:'16'
	summary: 'Lets you load SAR (Squeak ARchive) files from SqueakMap and the File List. For 3.6 and later images.'
	description:'Support for installing SAR (Squeak ARchive) packages from SqueakMap and the File List.
For 3.6 and later images.

SMSARInstaller will use this if it''s present to load SAR packages.

Use SARBuilder for making these packages easily.'
	url: 'http://bike-nomad.com/squeak/'
	downloadUrl:'http://bike-nomad.com/squeak/SARInstallerFor36-nk.16.cs.gz'
	author: 'Ned Konz <ned@bike-nomad.com>'
	maintainer:'Ned Konz <ned@bike-nomad.com>'
	registrator:'Ned Konz <ned@bike-nomad.com>'
	password:240495131608326995113451940367316491071470713347
	categories: #('6ba57b6e-946a-4009-beaa-0ac93c08c5d1' '94277ca9-4d8f-4f0e-a0cb-57f4b48f1c8a' 'a71a6233-c7a5-4146-b5e3-30f28e4d3f6b' '8209da9b-8d6e-40dd-b23a-eb7e05d4677b' );
	modulePath: ''
	moduleVersion:''
	moduleTag:''
	versionComment:'v16: same as v16 of SARInstaller for 3.4 but doesn''t include any classes other than SARInstaller.

To be loaded into 3.6 images only. Will de-register the 3.4 version if it''s registered.

Added a default (DWIM) mode in which SAR files that are missing both a preamble and postscript have all their members loaded in a default manner.

Changed the behavior of #extractMemberWithoutPath: to use the same directory as the SAR itself.

Added #extractMemberWithoutPath:inDirectory:

Moved several change set methods to the class side.

Made change set methods work with 3.5 or 3.6a/b

Now supports the following file types:

Projects (with or without construction of a ViewMorph)
Genie gesture dictionaries
Change sets
DVS packages
Monticello packages
Graphics files (loaded as SketchMorphs)
Text files (loaded as text editor windows)
Morph(s) in files

Now keeps track of installed members.'! !

!SARInstaller class methodsFor: 'change set utilities' stamp: 'nk 7/5/2003 22:51'!
newChanges: aChangeSet
	"Change the current change set, in a way that should work in 3.5 as well"
	"SARInstaller newChanges: SARInstaller currentChangeSet"

	^[ ChangeSet newChanges: aChangeSet ]
		on: MessageNotUnderstood
		do: [ :ex | ex return: (Smalltalk newChanges: aChangeSet) ]! !

!SARInstaller class methodsFor: 'class initialization' stamp: 'KR 3/21/2008 21:01'!
serviceFileInSAR
	"Answer a service for opening a changelist browser on a file"

	^ SimpleServiceEntry 
		provider: self 
		label: 'install SAR' translatedNoop
		selector: #installSAR:
		description: 'install this Squeak ARchive into the image.' translatedNoop
		buttonLabel: 'install' translatedNoop! !

!SARInstaller class methodsFor: 'class initialization' stamp: 'nk 11/21/2002 09:46'!
services
	^Array with: self serviceFileInSAR
! !

!SARInstaller class methodsFor: 'package format support' stamp: 'nk 7/25/2003 08:27'!
squeakMapDo: aBlock
	"If SqueakMap is installed, evaluate aBlock with the default map.
	Otherwise, offer to install SqueakMap and continue."

	Smalltalk at: #SMSqueakMap ifPresent: [ :smClass | ^aBlock value: smClass default ].

	(self confirm: 'SqueakMap is not installed in this image.
Would you like to load it from the network?')
		ifTrue: [ TheWorldMenu loadSqueakMap.
			^self squeakMapDo: aBlock ].

	^nil! !

!SARInstaller class methodsFor: 'SqueakMap' stamp: 'nk 7/21/2003 17:16'!
squeakMapPackageID
	^'75c970ab-dca7-48ee-af42-5a013912c880'! !

!SARInstaller class methodsFor: 'SqueakMap' stamp: 'nk 7/21/2003 17:18'!
squeakMapPackageVersion
	^'16'! !

!SARInstaller class methodsFor: 'class initialization' stamp: 'fbs 8/9/2013 17:32'!
unload
	FileServices unregisterFileReader: self.! !

!SARInstaller class methodsFor: 'change set utilities' stamp: 'rbb 3/1/2005 11:11'!
withCurrentChangeSetNamed: aString do: aOneArgumentBlock 
	"Evaluate the one-argument block aOneArgumentBlock while the named change set is active.
	We let the user confirm operating on an existing ChangeSet 
	or specify another ChangeSet name if 
	the name derived from the filename already exists. 
	Duplicated from SMSimpleInstaller. 
	Returns change set."

	| changeSet newName oldChanges |
	newName := aString.
	changeSet := self changeSetNamed: newName.
	changeSet ifNotNil: 
			[newName := UIManager default 
						request: 'ChangeSet already present, just confirm to overwrite or enter a new name:'
						initialAnswer: newName.
			newName isEmpty ifTrue: [self error: 'Cancelled by user'].
			changeSet := self changeSetNamed: newName].
	changeSet ifNil: [changeSet := self basicNewChangeSet: newName].
	changeSet 
		ifNil: [self error: 'User did not specify a valid ChangeSet name'].
	oldChanges := self currentChangeSet.
	
	[ self newChanges: changeSet.
	aOneArgumentBlock value: changeSet] 
			ensure: [ self newChanges: oldChanges].
	^changeSet! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 10/25/2002 12:16'!
directory
	^directory! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 10/25/2002 12:16'!
directory: anObject
	directory := anObject! !

!SARInstaller methodsFor: 'private' stamp: 'nk 10/13/2003 12:56'!
errorNoSuchMember: aMemberName
	(self confirm: 'No member named ', aMemberName, '. Do you want to stop loading?')
		== true ifTrue: [ self error: 'aborted' ].! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 7/5/2003 22:40'!
extractMemberWithoutPath: aMemberOrName
	"Extract aMemberOrName to its own filename, but ignore any directory paths, using my directory instead."
	self extractMemberWithoutPath: aMemberOrName inDirectory: self directory.
! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 7/5/2003 22:40'!
extractMemberWithoutPath: aMemberOrName inDirectory: aDirectory
	"Extract aMemberOrName to its own filename, but ignore any directory paths, using aDirectory instead"
	| member |
	member := self memberNamed: aMemberOrName.
	member ifNil: [ ^self errorNoSuchMember: aMemberOrName ].
	self zip extractMemberWithoutPath: member inDirectory: aDirectory.
	self installed: member.! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 7/5/2003 22:25'!
extractMember: aMemberOrName
	"Extract aMemberOrName to a file using its filename"
	(self zip extractMember: aMemberOrName)
		ifNil: [ self errorNoSuchMember: aMemberOrName ]
		ifNotNil: [ self installed: aMemberOrName ].! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 7/5/2003 22:25'!
extractMember: aMemberOrName toFileNamed: aFileName
	"Extract aMemberOrName to a specified filename"
	(self zip extractMember: aMemberOrName toFileNamed: aFileName)
		ifNil: [ self errorNoSuchMember: aMemberOrName ]
		ifNotNil: [ self installed: aMemberOrName ].! !

!SARInstaller methodsFor: 'fileIn' stamp: 'nice 1/13/2010 08:46'!
fileIn
	"File in to a change set named like my file"
	directory readOnlyFileNamed: fileName do: [:stream |
		| newCS |
		newCS := self class withCurrentChangeSetNamed: fileName
			do: [:cs | self fileInFrom: stream].
		newCS isEmpty ifTrue: [ ChangeSet removeChangeSet: newCS ]]! !

!SARInstaller methodsFor: 'fileIn' stamp: 'nice 12/27/2009 03:10'!
fileInFrom: stream
	"The zip has been saved already by the download.
	Read the zip into my instvar, then file in the correct members"

	

	[ | postscript preamble |
		stream position: 0.
		zip := ZipArchive new readFrom: stream.

		preamble := zip memberNamed: 'install/preamble'.
		preamble ifNotNil: [
			preamble contentStream text setConverterForCode fileInFor: self announcing: 'Preamble'.
			self class currentChangeSet preambleString: preamble contents.
		].

		postscript := zip memberNamed: 'install/postscript'.
		postscript ifNotNil: [
			postscript contentStream text setConverterForCode fileInFor: self announcing: 'Postscript'.
			self class currentChangeSet postscriptString: postscript contents.
		].

		preamble isNil & postscript isNil ifTrue: [
			(self confirm: 'No install/preamble or install/postscript member were found.
	Install all the members automatically?') ifTrue: [ self installAllMembers ]
		].

	] ensure: [ stream close ].

! !

!SARInstaller methodsFor: 'client services' stamp: 'nice 1/18/2010 19:08'!
fileInGenieDictionaryNamed: memberName 
	"This is to be used from preamble/postscript code to file in zip 
	members as Genie gesture dictionaries.
	Answers a dictionary."

	| member object crDictionary stream |

	crDictionary := Smalltalk at: #CRDictionary ifAbsent: [ ^self error: 'Genie not installed' ].
	"don't know how to recursively load"

	member := self memberNamed: memberName.
	member ifNil: [ ^self errorNoSuchMember: memberName ].

	stream := ReferenceStream on: member contentStream.

	object := [ stream next ]
		on: Error do: 
		[:ex |  stream close.
		self inform: 'Error on loading: ' , ex description. ^ nil ].
	stream close.

	(object notNil and: [object name isEmptyOrNil])
		ifTrue: [object := crDictionary name: object storedName].

	self installed: member.

	^ object
! !

!SARInstaller methodsFor: 'private' stamp: 'ar 9/27/2005 20:10'!
fileInMCVersion: member withBootstrap: mcBootstrap
	"This will use the MCBootstrapLoader to load a (non-compressed) Monticello file (.mc or .mcv)"
	| newCS |
	self class withCurrentChangeSetNamed: member localFileName
		do: [ :cs | 
			newCS := cs.
			mcBootstrap loadStream: member contentStream ascii ].

	newCS isEmpty ifTrue: [ ChangeSet removeChangeSet: newCS ].

	World doOneCycle.

	self installed: member.! !

!SARInstaller methodsFor: 'client services' stamp: 'yo 8/17/2004 10:01'!
fileInMemberNamed: csName
	"This is to be used from preamble/postscript code to file in zip members as ChangeSets."
	| cs |
	cs := self memberNamed: csName.
	cs ifNil: [ ^self errorNoSuchMember: csName ].
	self class fileIntoChangeSetNamed: csName fromStream: cs contentStream text setConverterForCode.
	self installed: cs.
! !

!SARInstaller methodsFor: 'client services' stamp: 'nice 12/27/2009 03:10'!
fileInMonticelloPackageNamed: memberName 
	"This is to be used from preamble/postscript code to file in zip 
	members as Monticello packages (.mc)."

	| member file mcPackagePanel mcRevisionInfo mcSnapshot mcFilePackageManager mcPackage mcBootstrap newCS |

	mcPackagePanel := Smalltalk at: #MCPackagePanel ifAbsent: [ ].
	mcRevisionInfo := Smalltalk at: #MCRevisionInfo ifAbsent: [ ].
	mcSnapshot := Smalltalk at: #MCSnapshot ifAbsent: [ ].
	mcFilePackageManager := Smalltalk at: #MCFilePackageManager ifAbsent: [ ].
	mcPackage := Smalltalk at: #MCPackage ifAbsent: [ ].
	member := self memberNamed: memberName.
	member ifNil: [ ^self errorNoSuchMember: memberName ].

	"We are missing MCInstaller, Monticello and/or MonticelloCVS.
	If the bootstrap is present, use it. Otherwise interact with the user."
	({ mcPackagePanel. mcRevisionInfo. mcSnapshot. mcFilePackageManager. mcPackage } includes: nil)
		ifTrue: [
			mcBootstrap := self getMCBootstrapLoaderClass.
			mcBootstrap ifNotNil: [ ^self fileInMCVersion: member withBootstrap: mcBootstrap ].

			(self confirm: ('Monticello support is not installed, but must be to load member named ', memberName, '.
Load it from SqueakMap?'))
				ifTrue: [ self class loadMonticello; loadMonticelloCVS.
					^self fileInMonticelloPackageNamed: memberName ]
				ifFalse: [ ^false ] ].

	member extractToFileNamed: member localFileName inDirectory: self directory.
	file := (Smalltalk at: #MCFile)
				name: member localFileName
				directory: self directory.

	self class withCurrentChangeSetNamed: file name do: [ :cs | | snapshot info |
		newCS := cs.
		file readStreamDo: [ :stream |
			info := mcRevisionInfo readFrom: stream nextChunk.
			snapshot := mcSnapshot fromStream: stream ].
			snapshot install.
			(mcFilePackageManager forPackage:
				(mcPackage named: info packageName))
					file: file
		].

	newCS isEmpty ifTrue: [ ChangeSet removeChangeSet: newCS ].

	mcPackagePanel allSubInstancesDo: [ :ea | ea refresh ].
	World doOneCycle.

	self installed: member.
! !

!SARInstaller methodsFor: 'client services' stamp: 'ar 9/27/2005 20:10'!
fileInMonticelloVersionNamed: memberName 
	"This is to be used from preamble/postscript code to file in zip 
	members as Monticello version (.mcv) files."

	| member newCS mcMcvReader |
	mcMcvReader := Smalltalk at: #MCMcvReader ifAbsent: [].
	member := self memberNamed: memberName.
	member ifNil: [^self errorNoSuchMember: memberName].

	"If we don't have Monticello, offer to get it."
	mcMcvReader ifNil:  [
		(self confirm: 'Monticello is not installed, but must be to load member named ', memberName , '.
Load it from SqueakMap?') 
			ifTrue:  [ self class loadMonticello.
						^self fileInMonticelloVersionNamed: memberName]
					ifFalse: [^false]].

	self class withCurrentChangeSetNamed: member localFileName
		do: 
			[:cs | 
			newCS := cs.
			(mcMcvReader versionFromStream: member contentStream ascii) load ].
	newCS isEmpty ifTrue: [ChangeSet removeChangeSet: newCS].
	World doOneCycle.
	self installed: member! !

!SARInstaller methodsFor: 'client services' stamp: 'ar 9/27/2005 20:10'!
fileInMonticelloZipVersionNamed: memberName 
	"This is to be used from preamble/postscript code to file in zip 
	members as Monticello version (.mcz) files."

	| member mczInstaller newCS mcMczReader |
	mcMczReader := Smalltalk at: #MCMczReader ifAbsent: [].
	mczInstaller := Smalltalk at: #MczInstaller ifAbsent: [].
	member := self memberNamed: memberName.
	member ifNil: [^self errorNoSuchMember: memberName].

	"If we don't have Monticello, but have the bootstrap, use it silently."
	mcMczReader ifNil:  [
		mczInstaller ifNotNil: [ ^mczInstaller installStream: member contentStream ].
		(self confirm: 'Monticello is not installed, but must be to load member named ', memberName , '.
Load it from SqueakMap?') 
			ifTrue:  [ self class loadMonticello.
						^self fileInMonticelloZipVersionNamed: memberName]
					ifFalse: [^false]].

	self class withCurrentChangeSetNamed: member localFileName
		do: 
			[:cs | 
			newCS := cs.
			(mcMczReader versionFromStream: member contentStream) load ].
	newCS isEmpty ifTrue: [ChangeSet removeChangeSet: newCS].
	World doOneCycle.
	self installed: member! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 7/5/2003 22:27'!
fileInMorphsNamed: memberName addToWorld: aBoolean
	"This will load the Morph (or Morphs) from the given member.
	Answers a Morph, or a list of Morphs, or nil if no such member or error.
	If aBoolean is true, also adds them and their models to the World."

	| member morphOrList |
	member := self memberNamed: memberName.
	member ifNil: [ ^self errorNoSuchMember: memberName ].
	self installed: member.

	morphOrList := member contentStream fileInObjectAndCode.
	morphOrList ifNil: [ ^nil ].
	aBoolean ifTrue: [ ActiveWorld addMorphsAndModel: morphOrList ].

	^morphOrList
! !

!SARInstaller methodsFor: 'client services' stamp: 'ul 9/20/2011 12:07'!
fileInPackageNamed: memberName 
	"This is to be used from preamble/postscript code to file in zip 
	members as DVS packages."
	| member current new baseName imagePackageLoader packageInfo streamPackageLoader packageManager |
	member := self zip memberNamed: memberName.
	member ifNil: [ ^self errorNoSuchMember: memberName ].

	imagePackageLoader := Smalltalk at: #ImagePackageLoader ifAbsent: [].
	streamPackageLoader := Smalltalk at: #StreamPackageLoader ifAbsent: [].
	packageInfo := Smalltalk at: #PackageInfo ifAbsent: [].
	packageManager := Smalltalk at: #FilePackageManager ifAbsent: [].

	"If DVS isn't present, do a simple file-in"
	(packageInfo isNil or: [imagePackageLoader isNil or: [streamPackageLoader isNil]])
		ifTrue: [ ^ self fileInMemberNamed: memberName ].

	baseName := memberName copyReplaceAll: '.st' with: ''.
	(packageManager allManagers anySatisfy: [ :pm | pm packageName = baseName ])
		ifTrue: [
			current := imagePackageLoader new package: (packageInfo named: baseName).
			new := streamPackageLoader new stream: member contentStream ascii.
			(new changesFromBase: current) fileIn ]
		ifFalse: [ self class fileIntoChangeSetNamed: baseName fromStream: member contentStream ascii setConverterForCode. ].

	packageManager registerPackage: baseName.

	self installed: member.! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 7/27/2003 09:49'!
fileInProjectNamed: projectOrMemberName createView: aBoolean 
	"This is to be used from preamble/postscript code to file in SAR members 
	as Projects. 
	Answers the loaded project, or nil. 
	Does not enter the loaded project. 
	If aBoolean is true, also creates a ProjectViewMorph 
	(possibly in a window, depending on your Preferences)."
	| member project triple memberName |
	member := self memberNamed: projectOrMemberName.
	member
		ifNotNil: [ memberName := member fileName ]
		ifNil: [ 	member := self memberNamed: (memberName := self memberNameForProjectNamed: projectOrMemberName) ].
	member ifNil: [ ^self errorNoSuchMember: projectOrMemberName ].
	triple := Project parseProjectFileName: memberName unescapePercents.
	project := nil.
	[[ProjectLoading
		openName: triple first
		stream: member contentStream
		fromDirectory: nil
		withProjectView: nil]
		on: ProjectViewOpenNotification
		do: [:ex | ex resume: aBoolean]]
		on: ProjectEntryNotification
		do: [:ex | 
			project := ex projectToEnter.
			ex resume].
	self installed: member.
	^ project! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 9/26/2003 17:17'!
fileInTrueTypeFontNamed: memberOrName

	| member description |
	member := self memberNamed: memberOrName.
	member ifNil: [^self errorNoSuchMember: memberOrName].

	description := TTFontDescription addFromTTStream: member contentStream.
	TTCFont newTextStyleFromTT: description.

	World doOneCycle.
	self installed: member! !

!SARInstaller methodsFor: 'fileIn' stamp: 'nk 7/27/2003 14:02'!
fileIntoChangeSetNamed: aString fromStream: stream
	"Not recommended for new code"
	^self class fileIntoChangeSetNamed: aString fromStream: stream! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 10/25/2002 12:16'!
fileName
	^fileName! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 10/25/2002 12:16'!
fileName: anObject
	fileName := anObject! !

!SARInstaller methodsFor: 'client services' stamp: 'ul 12/12/2009 14:07'!
getMCBootstrapLoaderClass
	^Smalltalk at: #MCBootstrapLoader
		ifAbsent: 
			[(self memberNamed: 'MCBootstrapLoader.st') 
				ifNotNil: [:m | self fileInMemberNamed: m.
					Smalltalk at: #MCBootstrapLoader ifAbsent: []]]! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 10/14/2003 15:40'!
importImage: memberOrName
	| member form |
	member := self memberNamed: memberOrName.
	member ifNil: [ ^self errorNoSuchMember: memberOrName ].
	form := ImageReadWriter formFromStream: member contentStream binary.
	form ifNil: [ ^self ].
	Imports default importImage: form named: (FileDirectory localNameFor: member fileName) sansPeriodSuffix.
	self installed: member.! !

!SARInstaller methodsFor: 'initialization' stamp: 'nk 7/5/2003 22:24'!
initialize
	installed := OrderedCollection new.! !

!SARInstaller methodsFor: 'fileIn' stamp: 'nk 10/12/2003 20:41'!
installAllMembers
	"Try to install all the members, in order, based on their filenames and/or contents."
	| uninstalled |
	uninstalled := OrderedCollection new.
	zip members do: [ :member | self installMember: member ].
	uninstalled := self uninstalledMembers.
	uninstalled isEmpty ifTrue: [ ^self ].
	uninstalled inspect.! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 7/5/2003 23:01'!
installedMemberNames
	"Answer the names of the zip members that have been installed already."
	^self installedMembers collect: [ :ea | ea fileName ]! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 7/10/2003 16:53'!
installedMembers
	"Answer the zip members that have been installed already."
	^installed ifNil: [ installed := OrderedCollection new ]! !

!SARInstaller methodsFor: 'private' stamp: 'nk 7/10/2003 16:55'!
installed: aMemberOrName
	self installedMembers add: (self zip member: aMemberOrName)! !

!SARInstaller methodsFor: 'client services' stamp: 'tak 1/24/2005 19:12'!
installMember: memberOrName
	| memberName extension isGraphic stream member |
	member := self memberNamed: memberOrName.
	member ifNil: [ ^false ].
	memberName := member fileName.
	extension := (FileDirectory extensionFor: memberName) asLowercase.
	Smalltalk at: #CRDictionary ifPresent: [ :crDictionary |
		(extension = crDictionary fileNameSuffix) ifTrue: [  self fileInGenieDictionaryNamed: memberName. ^true ] ].
	extension caseOf: {
		[ Project projectExtension ] -> [ self fileInProjectNamed: memberName createView: true ].
		[ FileStream st ] -> [ self fileInPackageNamed: memberName ].
		[ FileStream cs ] -> [  self fileInMemberNamed: memberName  ].
"		[ FileStream multiSt ] -> [  self fileInMemberNamedAsUTF8: memberName  ].
		[ FileStream multiCs ] -> [  self fileInMemberNamedAsUTF8: memberName  ].
"
		[ 'mc' ] -> [ self fileInMonticelloPackageNamed: memberName ].
		[ 'mcv' ] -> [ self fileInMonticelloVersionNamed: memberName ].
		[ 'mcz' ] -> [ self fileInMonticelloZipVersionNamed: memberName ].
		[ 'morph' ] -> [ self fileInMorphsNamed: member addToWorld: true ].
		[ 'ttf' ] -> [ self fileInTrueTypeFontNamed: memberName ].
		[ 'translation' ] -> [  self fileInMemberNamed: memberName  ].
	} otherwise: [
		('t*xt' match: extension) ifTrue: [ self openTextFile: memberName ]
			ifFalse: [ stream := member contentStream.
		isGraphic := ImageReadWriter understandsImageFormat: stream.
		stream reset.
		isGraphic
			ifTrue: [ self openGraphicsFile: member ]
			ifFalse: [ "now what?" ^false ]]
	].
	^true
! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 10/14/2003 18:58'!
memberNamed: aString
	^(zip member: aString)
		ifNil: [ | matching |
			matching := zip membersMatching: aString.
			matching isEmpty ifFalse: [ matching last ]].! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 7/5/2003 10:02'!
memberNameForProjectNamed: projectName
	"Answer my member name for the given project, or nil.
	Ignores version numbers and suffixes, and also unescapes percents in filenames."

	^self zip memberNames detect: [ :memberName | | triple |
		triple := Project parseProjectFileName: memberName unescapePercents.
		triple first asLowercase = projectName asLowercase
	] ifNone: [ nil ].! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 7/5/2003 21:57'!
memberNames
	^self zip memberNames! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 10/27/2002 10:34'!
membersMatching: aString
	^self zip membersMatching: aString! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 6/12/2004 10:03'!
openGraphicsFile: memberOrName
	| member morph |
	member := self memberNamed: memberOrName.
	member ifNil: [ ^self errorNoSuchMember: memberOrName ].
	morph := (World drawingClass fromStream: member contentStream binary).
	morph ifNotNil: [ morph openInWorld ].
	self installed: member.! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 7/5/2003 22:28'!
openTextFile: memberOrName
	"Open a text window on the given member"
	| member |
	member := self memberNamed: memberOrName.
	member ifNil: [ ^self errorNoSuchMember: memberOrName ].
	StringHolder new
		acceptContents: member contents;
		openLabel: member fileName.
	self installed: member.! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 10/27/2002 10:36'!
prependedDataSize
	^self zip prependedDataSize! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 7/5/2003 23:00'!
uninstalledMemberNames
	"Answer the names of the zip members that have not yet been installed."
	^self uninstalledMembers collect: [ :ea | ea fileName ]! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 7/10/2003 16:55'!
uninstalledMembers
	"Answer the zip members that haven't been installed or extracted yet."
	^zip members copyWithoutAll: self installedMembers! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 10/25/2002 12:16'!
zip
	^zip! !

!SARInstaller methodsFor: 'client services' stamp: 'nk 10/27/2002 10:35'!
zipFileComment
	^self zip zipFileComment! !

!SARInstaller methodsFor: 'accessing' stamp: 'nk 10/25/2002 12:16'!
zip: anObject
	^zip := anObject! !

Model subclass: #SimpleServiceEntry
	instanceVariableNames: 'provider label selector useLineAfter stateSelector description argumentGetter buttonLabel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FileRegistry'!

!SimpleServiceEntry commentStamp: '<historical>' prior: 0!
I represent a service

provider : the service provider
label : to be display in a menu
selector : to do the service
useLineAfter
stateSelector : a secondary selector (to be able to query state of the provider for example)
description : a description for balloon for example
argumentGetter : a selector to get additional arguments with (if selector requres them)
buttonLabel : a short label

The entire client interface (provided by FileList and other users of the registry)
is this (browse #getArgumentsFrom: and the 
senders of #argumentGetter:):

fullName (returns a String with the full filename)
dirAndFileName (returns {directory. fileName})
readOnlyStream (returns an open read-only stream)
!

SimpleServiceEntry subclass: #FileModifyingSimpleServiceEntry
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FileRegistry'!

!FileModifyingSimpleServiceEntry commentStamp: 'nk 11/26/2002 12:03' prior: 0!
I represent a service that may change the contents of a directory.
Such changes include:
* file creation
* file deletion
* file modification!

!FileModifyingSimpleServiceEntry methodsFor: 'as yet unclassified' stamp: 'nk 11/26/2002 12:08'!
performServiceFor: anObject
	| retval |
	retval := super performServiceFor: anObject.
	self changed: #fileListChanged.
	^retval	"is this used anywhere?"! !

!SimpleServiceEntry class methodsFor: 'instance creation' stamp: 'hg 8/1/2000 18:57'!
provider: anObject label: aString selector: aSymbol

	^self new provider: anObject label: aString selector: aSymbol! !

!SimpleServiceEntry class methodsFor: 'instance creation' stamp: 'sd 1/31/2002 22:05'!
provider: anObject label: aString selector: aSymbol description: anotherString

	^self new provider: anObject label: aString selector: aSymbol description: anotherString! !

!SimpleServiceEntry class methodsFor: 'instance creation' stamp: 'sw 2/17/2002 02:48'!
provider: anObject label: aString selector: aSymbol description: anotherString buttonLabel: aLabel
	"Answer a new service object with the given initializations.  This variant allows a custom button label to be provided, in order to preserve precious horizontal real-estate in the button pane, while still allowing more descriptive wordings in the popup menu"

	^ self new provider: anObject label: aString selector: aSymbol description: anotherString; buttonLabel: aLabel; yourself! !

!SimpleServiceEntry methodsFor: 'services menu' stamp: 'nk 2/15/2004 16:16'!
addServiceFor: served toMenu: aMenu
	aMenu add: self label 
		target: self 
		selector: self requestSelector 
		argument: served.
	self useLineAfter ifTrue: [ aMenu addLine ].! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'RAA 2/2/2002 08:14'!
argumentGetter: aBlock

	argumentGetter := aBlock! !

!SimpleServiceEntry methodsFor: 'toolbuilder' stamp: 'ar 2/12/2005 14:34'!
buildWith: builder in: aModel
	"Answer a button spec that will trigger the receiver service in aModel"
	| buttonSpec |
	buttonSpec := builder pluggableActionButtonSpec new.
	buttonSpec 
		model: self; 
		label: self buttonLabel;
		help: self description;
		action: (MessageSend receiver: aModel selector: #executeService: argument: self).
	^builder build: buttonSpec! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'sw 2/15/2002 17:53'!
buttonLabel
	"Answer the label to be emblazoned on a button representing the service in a file list, for example"

	^ buttonLabel ifNil: [label]! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'dgd 9/1/2003 12:16'!
buttonLabel: aString 
	"Set the receiver's buttonLabel, to be used on a button in a 
	tool-pane; this is split out so that a more abbreviated wording 
	can be deployed if desired"
	buttonLabel := aString translated! !

!SimpleServiceEntry methodsFor: 'performing service' stamp: 'jrp 7/30/2005 22:35'!
buttonToTriggerIn: aFileList
	"Answer a button that will trigger the receiver service in a file list"

	| aButton |
	aButton :=  PluggableButtonMorph
					on: self
					getState: nil
					action: #performServiceFor:.
	aButton 
		arguments: { aFileList }.

	aButton
		color: Color transparent;
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		label: self buttonLabel;
		askBeforeChanging: true;
		onColor: Color white offColor: Color white.
		aButton setBalloonText: self description.

	^ aButton! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'sd 1/31/2002 21:03'!
description
	"may be used for balloon or other"

	^ description
! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'sd 1/31/2002 21:03'!
description: aString
	"may be used for balloon or other"

	description := aString
! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'sd 1/31/2002 21:10'!
extraSelector
	"normally should not be used directly"

	^stateSelector! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'sd 1/31/2002 21:11'!
extraSelector: aSymbol

	stateSelector := aSymbol! !

!SimpleServiceEntry methodsFor: 'performing service' stamp: 'RAA 2/2/2002 08:31'!
getArgumentsFrom: aProvider

	argumentGetter ifNil: [^aProvider fullName].
	^argumentGetter value: aProvider! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'hg 8/3/2000 13:06'!
label

	^label! !

!SimpleServiceEntry methodsFor: 'extra' stamp: 'nk 6/8/2004 16:42'!
performExtraFor: anObject
	"carry out the extra service I provide"
	"the stateSelector can be used to ask state of the provider to be reflected in button or other"

	^stateSelector numArgs = 0
		ifTrue: [provider perform: stateSelector]
		ifFalse: [provider perform: stateSelector with: (self getArgumentsFrom: anObject) ]
! !

!SimpleServiceEntry methodsFor: 'performing service' stamp: 'nk 8/31/2004 19:30'!
performServiceFor: anObject
	"carry out the service I provide"

	^selector numArgs = 0
		ifTrue: [provider perform: selector]
		ifFalse: [
			selector numArgs = 1
				ifTrue: [ provider perform: selector with: (self getArgumentsFrom: anObject) ]
				ifFalse: [ provider perform: selector withArguments: (self getArgumentsFrom: anObject) ]]! !

!SimpleServiceEntry methodsFor: 'printing' stamp: 'jmv 3/2/2010 17:15'!
printOn: aStream

	aStream nextPutAll: self class name; nextPutAll: ': ('.
	self provider
		ifNotNil: [ aStream nextPutAll: provider printString].
	aStream nextPutAll: ' --- '. 
	self selector
		ifNotNil: [ aStream nextPutAll: selector asString].
	aStream nextPut: $)

! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'sd 1/31/2002 21:38'!
provider

	^ provider! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'dgd 9/1/2003 12:12'!
provider: anObject label: aString selector: aSymbol 
	"basic initialization message"
	provider := anObject.
	label := aString translated.
	selector := aSymbol.
	stateSelector := #none.
	description := ''! !

!SimpleServiceEntry methodsFor: 'performing service' stamp: 'dgd 9/1/2003 12:13'!
provider: anObject label: aString selector: aSymbol description: anotherString 
	"basic initialization message"
	self
		provider: anObject
		label: aString
		selector: aSymbol.
	stateSelector := #none.
	description := anotherString translated! !

!SimpleServiceEntry methodsFor: 'extra' stamp: 'sd 1/31/2002 21:08'!
requestExtraSelector
	"send me this message to ask me to perform secondary service"

	^#performExtraFor:
! !

!SimpleServiceEntry methodsFor: 'performing service' stamp: 'hg 8/1/2000 19:49'!
requestSelector
	"send me this message to ask me to perform my service"

	^#performServiceFor:
! !

!SimpleServiceEntry methodsFor: 'accessing' stamp: 'sd 1/31/2002 21:09'!
selector
	"normally should not be used directly"

	^selector! !

!SimpleServiceEntry methodsFor: 'services menu' stamp: 'hg 8/1/2000 19:53'!
useLineAfter

	^useLineAfter == true! !

!SimpleServiceEntry methodsFor: 'services menu' stamp: 'hg 8/1/2000 19:54'!
useLineAfter: aBoolean

	useLineAfter := aBoolean
! !

IdentityDictionary subclass: #SystemDictionary
	instanceVariableNames: 'cachedClassNames cachedNonClassNames'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SystemDictionary commentStamp: 'nice 3/6/2010 21:56' prior: 0!
I represent a special dictionary used as global namespace for class names :

	Smalltalk globals classNames.

and for traits too:

	Smalltalk globals traitNames.
	
and a few other globals:

	(Smalltalk globals keys reject: [:k | (Smalltalk globals at: k) isBehavior])
			collect: [:k | k -> (Smalltalk globals at: k) class].

As the above example let you guess, the global namespace of Smalltalk system is accessed through:

	Smalltalk globals.!

!SystemDictionary methodsFor: 'classes and traits' stamp: 'sd 4/17/2003 21:15'!
allClasses  
	"Return all the class defines in the Smalltalk SystemDictionary"
	"Smalltalk allClasses"

	^ self classNames collect: [:name | self at: name]! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:07'!
allClassesAndTraits
	"Return all the classes and traits defined in the Smalltalk SystemDictionary"

	^Array streamContents: [ :stream |
		self allClassesAndTraitsDo: [ :each | stream nextPut: each ] ]! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:13'!
allClassesAndTraitsDo: aBlock
	
	self
		allClassesDo: aBlock;
		allTraitsDo: aBlock! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:02'!
allClassesDo: aBlock
	"Evaluate the argument, aBlock, for each class in the system."

	self classNames do: [ :name |
		aBlock value: (self at: name) ]! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'al 2/23/2006 21:39'!
allTraits
	"Return all traits defined in the Smalltalk SystemDictionary"

	^ self traitNames collect: [:each | self at: each]! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:03'!
allTraitsDo: aBlock
	"Evaluate the argument, aBlock, for each trait in the system."

	self traitNames do: [ :name |
		aBlock value: (self at: name) ]! !

!SystemDictionary methodsFor: 'dictionary access' stamp: 'cwp 1/18/2014 10:25'!
associationOrUndeclaredAt: key 
	"return an association or install in undeclared.  Used for mating up ImageSegments."
	self flag: #environments.

	^ self associationAt: key ifAbsent: [
		Undeclared at: key put: nil.
		Undeclared associationAt: key]! !

!SystemDictionary methodsFor: 'dictionary access' stamp: 'cwp 6/22/2012 15:58'!
at: aKey put: anObject 
	"Override from Dictionary to check Undeclared and fix up
	references to undeclared variables."
	| index element |
	(self includesKey: aKey) ifFalse: 
		[self declare: aKey from: (self at: #Undeclared).
		self flushClassNameCache].
	super at: aKey put: anObject.
	^ anObject! !

!SystemDictionary methodsFor: 'accessing' stamp: 'ar 3/9/2010 17:07'!
bindingOf: varName
	"SystemDictionary includes Symbols only"
	^super bindingOf: varName asSymbol! !

!SystemDictionary methodsFor: 'accessing' stamp: 'cwp 6/22/2012 09:16'!
bindingOf: varName ifAbsent: aBlock
	"SystemDictionary includes Symbols only"
	^super bindingOf: varName asSymbol ifAbsent: aBlock! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 2/20/2011 07:08'!
classAndTraitNames
	"Answer a sorted collection of all class and trait (not including class-traits) names. The performance of this algorithm is O(n) if the classNames are already cached, otherwise O(n*log(n)) where n is self size."

	| classNames traitNames result temporary |
	classNames := self classNames.
	traitNames := self traitNames ifEmpty: [ ^classNames ].
	temporary := Array new: classNames size + traitNames size.
	result := temporary shallowCopy.
	temporary
		replaceFrom: 1
		to: classNames size
		with: classNames
		startingAt: 1;
		replaceFrom: classNames size + 1
		to: temporary size
		with: traitNames
		startingAt: 1;
		mergeFirst: 1
		middle: classNames size
		last: temporary size
		into: result
		by: nil.
	^result! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'al 1/12/2006 23:57'!
classNamed: className 
	^self classOrTraitNamed: className.! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 10/8/2010 21:22'!
classNames
	"Answer a sorted collection of all class names. Use the return value of #fillCaches to avoid concurrency issues."
	
	^cachedClassNames ifNil: [ self fillCaches at: 1 ]! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'fbs 1/9/2011 16:56'!
classOrTraitNamed: aString 
	"aString is either a class or trait name or a class or trait name followed by ' class' or 'classTrait' respectively.
	Answer the class or metaclass it names."

	| meta baseName |
	(aString endsWith: ' class')
		ifTrue: [meta := true.
				baseName := aString copyFrom: 1 to: aString size - 6]
		ifFalse: [
			(aString endsWith: ' classTrait')
				ifTrue: [
					meta := true.
					baseName := aString copyFrom: 1 to: aString size - 11]
				ifFalse: [
					meta := false.
					baseName := aString]].

	^self at: baseName asSymbol ifPresent:
		[ :global |
 		global isBehavior ifTrue:
			[ meta
				ifFalse: [ global ]
				ifTrue: [ global classSide ]]]! !

!SystemDictionary methodsFor: 'accessing' stamp: 'ar 3/5/2010 14:38'!
environment
	"For conversion from Smalltalk to SystemDictionary"
	^self! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 10/8/2010 21:33'!
fillCaches
	"Fill cachedClassNames and cachedNonClassNames. Return an array with the calculated values."

	| classNames nonClassNames |
	classNames := OrderedCollection new: self size.
	nonClassNames := OrderedCollection new.
	self keysAndValuesDo: [ :key :value |
		value isInMemory ifTrue: [
			"The key == value name test below addresses two separate issues:
				1) Obsolete classes, where key = #Foo and value name = 'AnObsoleteFoo'
				2) Aliases, i.e., Smalltalk at: #OtherName put: aClass"
			((value isKindOf: Class) and: [ key == value name ])
				ifTrue: [ classNames add: key ]
				ifFalse: [ nonClassNames add: key ] ] ].
	classNames sort.
	cachedNonClassNames := nonClassNames sort.
	cachedClassNames := classNames.
	^{ classNames. nonClassNames }! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 10/8/2010 21:23'!
flushClassNameCache
	"Smalltalk flushClassNameCache"
	"Force recomputation of the cached list of class names and non-class names."

	cachedClassNames := cachedNonClassNames := nil! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'cwp 11/7/2012 21:56'!
forgetClass: aClass logged: aBool 
	"Delete the class, aClass, from the system.
	Note that this doesn't do everything required to dispose of a class - to do that use Class>>removeFromSystem."

	aBool ifTrue: [SystemChangeNotifier uniqueInstance classRemoved: aClass fromCategory: aClass category].
	self organization removeElement: aClass name.
	Smalltalk removeFromStartUpList: aClass.
	Smalltalk removeFromShutDownList: aClass.
	aClass isObsolete ifFalse: [
		(self associationAt: aClass name ifAbsent: [ nil ]) ifNotNil: [ :oldBinding |
			| newBinding |
			newBinding := oldBinding shallowCopy.
			self replaceBinding: oldBinding in: aClass with: newBinding.
			(self isThisBindingReferred: oldBinding) ifTrue: [
				[oldBinding value: nil]
					on: AttemptToWriteReadOnlyGlobal
					do: [:ex | true].
				Undeclared add: oldBinding ] ] ].
	self removeKey: aClass name ifAbsent: [].
	self flushClassNameCache! !

!SystemDictionary methodsFor: 'memory space' stamp: 'eem 6/23/2016 16:16'!
growMemoryByAtLeast: numBytes
	"Grow memory by at least the requested number of bytes.
	 Primitive.  Fail if no memory is available.  Essential.
	 N.B. In Spur, the heap is composed of segments. Growing memory adds a new
	 segment; it does not extend existing segments. Growth is by at least that
	 established by SmalltalkImage current vmParameterAt: 25, which defaults to 16Mb."
	<primitive: 180>
	^(numBytes isInteger and: [numBytes > 0])
		ifTrue: [OutOfMemory signal]
		ifFalse: [self primitiveFailed]! !

!SystemDictionary methodsFor: 'classes and traits' stamp: ''!
hasClassNamed: aString
	"Answer whether there is a class of the given name, but don't intern aString if it's not alrady interned.  4/29/96 sw"

	Symbol hasInterned: aString ifTrue: 
		[:aSymbol | ^ (self at: aSymbol ifAbsent: [nil]) isKindOf: Class].
	^ false! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 9/28/2012 11:02'!
isThisBindingReferred: binding

	self systemNavigation allSelectorsAndMethodsDo: [ :behavior :selector :method |
		(method instVarsInclude: binding) ifTrue: [ 
			method literalsDo: [ :literal | "Avoid possible false positives from the primitive."
				literal == binding ifTrue: [ ^true ] ] ] ].
	^false! !

!SystemDictionary methodsFor: 'system attributes' stamp: 'eem 9/5/2013 17:10'!
maxIdentityHash
	"Answer the maximum identityHash value supported by the VM."
	<primitive: 176>
	^self primitiveFailed! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 10/8/2010 21:22'!
nonClassNames
	"Answer a sorted collection of all non-class names. Use the return value of #fillCaches to avoid concurrency issues."
	
	^cachedNonClassNames ifNil: [ self fillCaches at: 2 ]! !

!SystemDictionary methodsFor: 'objects from disk' stamp: 'ar 3/5/2010 21:25'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to Smalltalk instead."

	dp := DiskProxy global: #Smalltalk selector: #globals
			args: #().
	refStrm replace: self with: dp.
	^ dp! !

!SystemDictionary methodsFor: 'accessing' stamp: 'ar 7/11/1999 21:56'!
organization
	"Return the organizer for the receiver"
	^SystemOrganization! !

!SystemDictionary methodsFor: 'accessing' stamp: 'nice 10/20/2009 23:58'!
poolUsers
	"Answer a dictionary of pool name -> classes that refer to it.
	Also includes any globally know dictionaries (such as
	Smalltalk, Undeclared etc) which although not strictly
	accurate is potentially useful information"
	"Smalltalk poolUsers"
	| poolUsers |
	poolUsers := Dictionary new.
	self keys
		do: [:k | "yes, using isKindOf: is tacky but for reflective code like
			this it is very useful. If you really object you can:-
			a) go boil your head.
			b) provide a better answer.
			your choice."
			| pool refs |
			(((pool := self at: k) isKindOf: Dictionary)
					or: [pool isKindOf: SharedPool class])
				ifTrue: [refs := self systemNavigation allClasses
								select: [:c | c sharedPools identityIncludes: pool]
								thenCollect: [:c | c name].
					refs
						add: (self systemNavigation
								allCallsOn: (self associationAt: k)).
					poolUsers at: k put: refs]].
	^ poolUsers! !

!SystemDictionary methodsFor: 'printing' stamp: 'sma 6/1/2000 09:53'!
printElementsOn: aStream
	aStream nextPutAll:'(lots of globals)'! !

!SystemDictionary methodsFor: 'printing' stamp: 'ar 3/5/2010 21:25'!
printOn: aStream
	self == Smalltalk globals
		ifTrue: [aStream nextPutAll: 'Smalltalk globals'.
			aStream nextPutAll:' "a SystemDictionary with lots of globals"']
		ifFalse: [super printOn: aStream]! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'sw 9/5/97 18:30'!
removeClassNamed: aName
	"Invoked from fileouts:  if there is currently a class in the system named aName, then remove it.  If anything untoward happens, report it in the Transcript.  "

	| oldClass |
	(oldClass := self at: aName asSymbol ifAbsent: [nil]) == nil
		ifTrue:
			[Transcript cr; show: 'Removal of class named ', aName, ' ignored because ', aName, ' does not exist.'.
			^ self].

	oldClass removeFromSystem! !

!SystemDictionary methodsFor: 'removing' stamp: 'eem 5/19/2010 15:35'!
removeKey: key ifAbsent: aBlock
	"Remove key (and its associated value) from the receiver. If key is not in
	the receiver, answer the result of evaluating aBlock. Otherwise, answer
	the value externally named by key."

	self flushClassNameCache.
	^super removeKey: key ifAbsent: aBlock! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'sw 10/28/96'!
renameClassNamed: oldName as: newName
	"Invoked from fileouts:  if there is currently a class in the system named oldName, then rename it to newName.  If anything untoward happens, report it in the Transcript.  "

	| oldClass |
	(oldClass := self at: oldName asSymbol ifAbsent: [nil]) == nil
		ifTrue:
			[Transcript cr; show: 'Class-rename for ', oldName, ' ignored because ', oldName, ' does not exist.'.
			^ self].

	oldClass rename: newName! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 20:17'!
renameClass: aClass as: newName 
	"Rename the class, aClass, to have the title newName."

	^self renameClass: aClass from: aClass name to: newName! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ar 3/5/2010 20:16'!
renameClass: aClass from: oldName 
	"Rename the class, aClass, to have the title newName."

	^self renameClass: aClass from: oldName to: aClass name! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'laza 12/14/2011 15:46'!
renameClass: aClass from: oldName to: newName
	"Rename the class, aClass, to have the title newName."

	| oldref category |
	category := SystemOrganization categoryOfElement: oldName.
	self organization classify: newName under: category suppressIfDefault: true.
	self organization removeElement: oldName.
	oldref := self associationAt: oldName.
	self removeKey: oldName.
	oldref key: newName.
	self add: oldref.  "Old association preserves old refs"
	Smalltalk renamedClass: aClass from: oldName to: newName.
	self flushClassNameCache.
	SystemChangeNotifier uniqueInstance classRenamed: aClass from: oldName to: newName inCategory: category! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 9/28/2012 10:44'!
replaceBinding: oldBinding inAllMethodsWith: newBinding
	"Replace oldBinding to newBinding in all methods in the system except for those which are the methods of the class of oldBinding. Return true if such replacement were made, false otherwise."

	| foundReferringMethod |
	foundReferringMethod := false.
	self systemNavigation allSelectorsAndMethodsDo: [ :behavior :selector :method |
		(behavior ~~ oldBinding value and: [
			method instVarsInclude: oldBinding ]) ifTrue: [ "Use the fast primitive first."
				"The penultimate literal holds the selector of the method, the last literal is the class."
				1 to: method numLiterals - 2 do: [ :index |
					(method literalAt: index) == oldBinding ifTrue: [
						foundReferringMethod := true.
						method 
							literalAt: index put: newBinding;
							flushCache ] ] ] ].
	^foundReferringMethod! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 9/28/2012 10:56'!
replaceBinding: oldBinding in: aClass with: newBinding
	"Replace oldBinding with newBinding in aClass. This method is expected to be used during the removal of aClass."

	aClass methodsDo: [ :method |
		1 to: method numLiterals do: [ :index |
			(method literalAt: index) == oldBinding ifTrue: [
				method literalAt: index put: newBinding ] ] ]! !

!SystemDictionary methodsFor: 'private' stamp: 'ul 11/30/2009 05:14'!
scanForEmptySlotFor: anObject
	"Scan the key array for the first slot containing an empty slot (indicated by a nil). Answer the index of that slot. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := anObject hash \\ array size + 1.
	[ 
		(array at: index) ifNil: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!SystemDictionary methodsFor: 'private' stamp: 'ul 11/30/2009 05:16'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	| index start |
	index := start := anObject hash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [ element key == anObject ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!SystemDictionary methodsFor: 'dictionary access' stamp: 'di 12/19/1999 21:17'!
scopeFor: varName from: lower envtAndPathIfFound: envtAndPathBlock
	"Null compatibility with partitioning into environments."

	(self includesKey: varName)
		ifTrue: [^ envtAndPathBlock value: self value: String new]
		ifFalse: [^ nil]! !

!SystemDictionary methodsFor: 'snapshot and quit' stamp: 'eem 9/21/2013 07:39'!
setGCParameters
	"Adjust the VM's default GC parameters to avoid too much tenuring.
	 Maybe this should be left to the VM?"

	| proportion edenSize survivorSize averageObjectSize numObjects |
	proportion := 0.9. "tenure when 90% of pastSpace is full"
	edenSize := SmalltalkImage current vmParameterAt: 44.
	survivorSize := edenSize / 5.0. "David's paper uses 140Kb eden + 2 x 28kb survivor spaces; Spur uses the same ratios :-)"
	averageObjectSize := 8 * self wordSize. "a good approximation"
	numObjects := (proportion * survivorSize / averageObjectSize) rounded.
	SmalltalkImage current vmParameterAt: 6 put: numObjects  "tenure when more than this many objects survive the GC"! !

!SystemDictionary methodsFor: 'objects from disk' stamp: 'tk 3/7/2000 18:40'!
storeDataOn: aDataStream
	"I don't get stored.  Use a DiskProxy"

	self error: 'use a DiskProxy to store me'! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 12/8/2010 21:06'!
traitNames
	"Answer a sorted collection of all traits (not including class-traits) names."

	^self nonClassNames select: [ :name |
		(name beginsWith: 'AnObsolete') not and: [
			self 
				at: name 
				ifPresent: [ :global | global isInMemory and: [ global isTrait ] ]
				ifAbsent: [ false ] ] ]! !

!SystemDictionary methodsFor: 'accessing' stamp: 'cwp 6/22/2012 15:48'!
undeclared
	^ self at: #Undeclared! !

!SystemDictionary methodsFor: 'copying' stamp: 'tk 10/20/2000 11:35'!
veryDeepCopyWith: deepCopier
	"Return self.  I can't be copied.  Do not record me."! !

!Text class methodsFor: '*System-Support-instance creation' stamp: 'rbb 3/1/2005 11:18'!
fromUser
	"Answer an instance of me obtained by requesting the user to type a string."
	"Text fromUser"

	^ self fromString:
		(UIManager default request: 'Enter text followed by carriage return')
! !

Array variableSubclass: #ActionSequence
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Object Events'!

!ActionSequence commentStamp: 'tlk 5/7/2006 20:02' prior: 0!
An ActionSequence is an array that lists the object's dependant objects.!

!ActionSequence methodsFor: 'converting' stamp: 'reThink 2/18/2001 15:12'!
asActionSequence

	^self! !

!ActionSequence methodsFor: 'converting' stamp: 'rw 7/20/2003 16:03'!
asActionSequenceTrappingErrors

	^WeakActionSequenceTrappingErrors withAll: self! !

!ActionSequence methodsFor: 'converting' stamp: 'reThink 2/18/2001 15:28'!
asMinimalRepresentation

	self size = 0
		ifTrue: [^nil].
	self size = 1
		ifTrue: [^self first].
	^self! !

!ActionSequence methodsFor: 'printing' stamp: 'SqR 07/28/2001 18:25'!
printOn: aStream

	self size < 2 ifTrue: [^super printOn: aStream].
	aStream nextPutAll: '#('.
	self
		do: [:each | each printOn: aStream]
		separatedBy: [aStream cr].
	aStream nextPut: $)! !

!ActionSequence methodsFor: 'evaluating' stamp: 'nice 9/21/2010 09:32'!
value
	"Evaluate the elements of the receiver and answer the result of last evaluation."

	^self inject: nil into:
		[:previousAnswer :each | each value]! !

!ActionSequence methodsFor: 'evaluating' stamp: 'nice 9/21/2010 09:33'!
valueWithArguments: anArray
	"Evaluate the elements of the receiver with provided arguments and answer the result of last evaluation."

	^self inject: nil into:
		[:previousAnswer :each | each valueWithArguments: anArray]! !

!Array methodsFor: '*System-Object Storage-file in/out' stamp: 'tk 9/28/2000 15:35'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  If I am one of two shared global arrays, write a proxy instead."

self == (TextConstants at: #DefaultTabsArray) ifTrue: [
	dp := DiskProxy global: #TextConstants selector: #at: args: #(DefaultTabsArray).
	refStrm replace: self with: dp.
	^ dp].
self == (TextConstants at: #DefaultMarginTabsArray) ifTrue: [
	dp := DiskProxy global: #TextConstants selector: #at: args: #(DefaultMarginTabsArray).
	refStrm replace: self with: dp.
	^ dp].
^ super objectForDataStream: refStrm! !

Array variableSubclass: #WeakActionSequence
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

WeakActionSequence variableSubclass: #WeakActionSequenceTrappingErrors
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Change Notification'!

!WeakActionSequenceTrappingErrors methodsFor: 'converting' stamp: 'rw 7/20/2003 16:03'!
asActionSequenceTrappingErrors

	^self! !

!WeakActionSequenceTrappingErrors methodsFor: 'evaluating' stamp: 'rw 8/6/2003 12:09'!
value
	"Do the same as my parent, but make sure that all actions that do not  
	give errors are evaluated before resignaling the ones that gave errors  
	(giving the chance to clients to handle them)."

	^self valueStartingFrom: 1! !

!WeakActionSequenceTrappingErrors methodsFor: 'evaluating' stamp: 'nice 8/22/2010 17:47'!
valueStartingFrom: startIndex
	"Do the same as my parent, but make sure that all actions that do not 
	give errors are evaluated before resignaling the ones that gave errors 
	(giving the chance to clients to handle them)."

	"Note: I currently trap Halt,Error so that I am sure to get a Halt event in case of a Halt. This is being fixed in the exception system - when the fix is done it will be enough to capture only Error."

	| each answer |
	answer := nil.
	startIndex to: self size do: [:index |
		each := self at: index.
		[
			answer := each valueOtherwise: [ answer ].
		]
			on: Halt , Error
			do: [:exc | 
				self valueStartingFrom: index + 1.
				exc pass]].
	^ answer! !

!WeakActionSequenceTrappingErrors methodsFor: 'evaluating' stamp: 'rw 8/6/2003 12:07'!
valueWithArguments: anArray 
	"Do the same as my parent, but make sure that all actions that do not 
	give errors are evaluated before resignaling the ones that gave errors 
	(giving the chance to clients to handle them)."

	^self valueWithArguments: anArray startingFrom: 1! !

!WeakActionSequenceTrappingErrors methodsFor: 'evaluating' stamp: 'nice 8/22/2010 17:48'!
valueWithArguments: anArray startingFrom: startIndex
	"Do the same as my parent, but make sure that all actions that do not 
	give errors are evaluated before resignaling the ones that gave errors 
	(giving the chance to clients to handle them)."

	"Note: I currently trap Halt,Error so that I am sure to get a Halt event in case of a Halt. This is being fixed in the exception system - when the fix is done it will be enough to capture only Error."

	| each answer |
	answer := nil.
	startIndex to: self size do: [:index |
		each := self at: index.
		[
			answer := each valueWithArguments: anArray otherwise: [ answer ].
		]
			on: Halt , Error
			do: [:exc | 
				self valueWithArguments: anArray startingFrom: index + 1.
				exc pass]].
	^ answer! !

!WeakActionSequence methodsFor: 'converting' stamp: 'rw 4/27/2002 07:44'!
asActionSequence

	^self! !

!WeakActionSequence methodsFor: 'converting' stamp: 'rw 7/20/2003 16:03'!
asActionSequenceTrappingErrors

	^WeakActionSequenceTrappingErrors withAll: self! !

!WeakActionSequence methodsFor: 'converting' stamp: 'jmv 5/13/2009 10:20'!
asMinimalRepresentation

	| valid |
	valid := self reject: [:e | e isReceiverOrAnyArgumentGarbage ].
	valid size = 0
		ifTrue: [ ^nil ].
	valid size = 1
		ifTrue: [ ^valid first ].
	^valid! !

!WeakActionSequence methodsFor: 'printing' stamp: 'rw 4/27/2002 07:46'!
printOn: aStream

	self size < 2 ifTrue: [^super printOn: aStream].
	aStream nextPutAll: '#('.
	self
		do: [:each | each printOn: aStream]
		separatedBy: [aStream cr].
	aStream nextPut: $)! !

!WeakActionSequence methodsFor: 'evaluating' stamp: 'nice 8/22/2010 21:05'!
value
	"Answer the result of evaluating the elements of the receiver.
	Actually, return just the last result."

	^self inject: nil into: [ :previousAnswer :each | 
		each valueOtherwise: [ previousAnswer ]].! !

!WeakActionSequence methodsFor: 'evaluating' stamp: 'nice 8/22/2010 21:02'!
valueWithArguments: anArray
	"Return the last result"

	^self inject: nil into: [ :previousAnswer :each |
		each valueWithArguments: anArray otherwise: [ previousAnswer ]]! !

!ParserNotification methodsFor: '*System-default handlers' stamp: 'cwp 8/25/2009 20:04'!
defaultAction
	
	self openMenuIn: 
		[:labels :lines :caption | 
		UIManager default chooseFrom: labels lines: lines title: caption]! !

!FileExistsException methodsFor: '*System-Files-error handling' stamp: 'mtf 1/20/2011 14:17'!
defaultAction
	"The default action taken if the exception is signaled."

	^ self fileClass fileExistsUserHandling: self fileName ifDebug: [super defaultAction]
! !

!CompiledMethod methodsFor: '*System-Tools-debugger support' stamp: 'eem 7/4/2015 07:39'!
getAndForgetUnbreakpointedOriginal
	"Answer the original unbreakpointed method, removing the original from the BreakpointManager's collection.
	 Use with care!!!!"
	self assert: self hasBreakpoint.
	^BreakpointManager forget: self! !

!CompiledMethod methodsFor: '*System-Support' stamp: 'md 2/16/2006 14:00'!
methodReference
	| class selector |
	class := self methodClass ifNil: [^nil].
	selector := self selector ifNil: [^nil].
	^MethodReference class: class selector: selector.
	! !
DataStream initialize!
DummySoundSystem initialize!
AutoStart initialize!
ChangeSet initialize!
DefaultExternalDropHandler initialize!
DigitalSignatureAlgorithm initialize!
ExternalDropHandler initialize!
ExternalSettings initialize!
FileServices initialize!
HTTPLoader initialize!
ISOLanguageDefinition initialize!
Locale initialize!
MczInstaller initialize!
MidiPrimTester initialize!
MOFile initialize!
GetTextTranslator initialize!
NaturalLanguageTranslator initialize!
PowerManagement initialize!
Preferences initialize!
RealEstateAgent initialize!
SecureHashAlgorithm initialize!
SecurityManager initialize!
SmalltalkImage initialize!
SystemNavigation initialize!
TextDomainManager initialize!
UserInterfaceTheme initialize!
Utilities initialize!
MessageTally initialize!
Project initialize!
SARInstaller initialize!
